#ifndef BPFIT_DEDXTABLES_H
#define BPFIT_DEDXTABLES_H

namespace bpfit {
  ///
  /// \brief Tables of mean energy loss
  ///
  /// Implements:
  ///...................................................................
  ///
  ///      MUON STOPPING POWER AND RANGE TABLES 10 MeV - 100 TeV
  ///                         DONALD GROOM
  ///                        NIKOLAI MOKHOV
  ///                       SERGEI STRIGANOV
  ///      Atomic data and Nuclear Data Tables 78, 183-356 (2001)
  ///...................................................................
  ///
  /// Below 10 MeV the tables are extrapolated linearly in log-log and
  /// are pretty accurate down to a beta*gamma of ~0.1 where the dE/dx
  /// ~= 1/beta^2 behavior begins to turn over. I've only input the
  /// tables up to 800 GeV.
  ///
  class dEdxTable {
  public:
    ///
    /// \brief Build dEdx table for named material
    /// \param mat - Material name
    ///
    /// \warning If the material is unknown will quietly return 0's always.
    ///
    dEdxTable(const char* mat);
    
    ///
    /// Return dE/dx in units of GeV/cm
    ///
    /// \param T - kinetic energy of particle (GeV)
    /// \param m - mass of particle (GeV)
    /// \return dEdx in units of GeV/cm
    ///
    double operator()(double T, double m) const;
    
  private:
    void InitPVC();
    void InitSteel();
    void InitParaffin();
    void InitMineralOil();
    // Added because I encounter vacuum and code aborts because
    // we do not have a dEdx table for vacuum.
    void InitVac();

  private:
    double TtoBetaGamma(double t,  double m) const;
    double BetaGammaToT(double bg, double m) const;
    
    double Interp(double t,
		  double m,
		  const double T[],
		  const double dedx[],
		  unsigned int imax) const;
  private:
    unsigned int  fSz;   ///< Table size
    const double* fT;    ///< Muon kinetic energy
    const double* fdEdx; ///< Muon energy loss
    double        fRho;  ///< Material density
  };
}
#endif
////////////////////////////////////////////////////////////////////////
