#pragma once

// This file defines the basic Cut object. For specific cuts, and examples of
// how to implement your own, see Cuts.h

#include <functional>
#include <set>
#include <string>

#include "CAFAna/Var.h"

namespace caf{class StandardRecord;}

namespace ana
{
  /// \brief Representation of a cut (selection) to be applied to a \ref
  /// caf::StandardRecord object
  ///
  /// A Cut consists of a function, taking a StandardRecord and returning a
  /// boolean indicating if that event passes the cut, plus a list of the
  /// branch names that must be active in the tree. This explicit naming of
  /// required branches leads to a significant speedup in CAF loading. However,
  /// if you neglect a necessary branch, that field will be uninitialized in
  /// the StandardRecord you are passed.
  ///
  /// Cut objects may be combined with the standard boolean operations && ||
  /// and !
  class Cut
  {
  public:
    /// The type of the function part of a cut
    typedef bool (CutFunc_t)(const caf::StandardRecord* sr);

    /// std::function can wrap a real function, function object, or lambda
    Cut(const std::set<std::string>& reqs,
        const std::function<CutFunc_t>& fun) : fReqs(reqs), fFunc(fun) {}
    Cut(const Cut& copy) : fReqs(copy.fReqs), fFunc(copy.fFunc) {}

    /// The list of branches required to be active
    const std::set<std::string>& Requires() const {return fReqs;}

    /// Allows a cut to be called with bool result = myCut(sr) syntax
    bool operator()(const caf::StandardRecord* sr) const
    {
      return fFunc(sr);
    }
  protected:
    std::set<std::string> fReqs;
    std::function<CutFunc_t> fFunc;
  };

  Cut operator&&(const Cut& a, const Cut& b);
  Cut operator||(const Cut& a, const Cut& b);
  Cut operator!(const Cut& a);

  Cut operator>(const Var& v, double c);
  Cut operator<(const Var& v, double c);
  Cut operator>=(const Var& v, double c);
  Cut operator<=(const Var& v, double c);
  Cut operator==(const Var& v, double c);

  Cut operator>(const Var& a, const Var& b);
  Cut operator<(const Var& a, const Var& b);
  Cut operator>=(const Var& a, const Var& b);
  Cut operator<=(const Var& a, const Var& b);
  Cut operator==(const Var& a, const Var& b);

  Cut operator>(double c, const Var& v);
  Cut operator<(double c, const Var& v);
  Cut operator>=(double c, const Var& v);
  Cut operator<=(double c, const Var& v);

  /// The simplest possible cut: pass everything, used as a default
  const Cut kNoCut({}, [](const caf::StandardRecord*){return true;});
} // namespace
