#pragma once

// Specific cut implementations. When this file gets too crowded it should be
// split up.

// Definition of the generic Cut object
#include "Cut.h"

#include "StandardRecord/StandardRecord.h"

#include <cassert>

namespace ana
{
  /// \brief Is this a Neutral %Current event?
  ///
  /// We use uniform-initializer syntax to concisely pass the list of necessary
  /// branches. In this case the selection function is simple enough that we
  /// can include it inline as a lambda function.
  const Cut kIsNC({"mc.nu", "mc.nnu", "mc.nu.iscc"},
                  [](const caf::StandardRecord* sr)
                  {
                    if(sr->mc.nnu == 0) return false;
                    assert(sr->mc.nnu == 1);
                    return !sr->mc.nu[0].iscc;
                  });

  /// Helper for \ref kPassesPresel
  inline bool PassesPreselFunc(const caf::StandardRecord* sr)
  {
    // TODO HACK this should be cutting those out anyway
    if(sr->sel.lem.pid < -4) return false;
    return (sr->slc.nhit >= 43 && sr->slc.nhit <= 120 &&
            (sr->trk.ndiscrete == 0 || sr->trk.discrete[0].len < 267) &&
            2*sr->slc.calE > 1 && 2*sr->slc.calE < 3);
  }

  /// \brief Does this event pass the nue preselection?
  ///
  /// In this case the selection is complicated enough to be split out into a
  /// separate usual function.
  const Cut kPassesPresel({"sel.lem.pid", // HACK
	                   "slc.nhit",
                           "trk.discrete",
                           "trk.ndiscrete",
                           "trk.discrete.len",
                           "slc.calE"},
                          PassesPreselFunc);

  //----------------------------------------------------------------------
  /// Helper for defining true CC event cuts
  class CCFlavSel
  {
  public:
    CCFlavSel(int pdg, int pdgorig) : fPdg(pdg), fPdgOrig(pdgorig)
    {
    }

    bool operator()(const caf::StandardRecord* sr) const
    {
      if(sr->mc.nnu == 0) return false;
      assert(sr->mc.nnu == 1);
      return (sr->mc.nu[0].iscc &&
              abs(sr->mc.nu[0].pdg) == fPdg &&
              abs(sr->mc.nu[0].pdgorig) == fPdgOrig);
    }
  protected:
    int fPdg, fPdgOrig;
  };

  /// List of CAF variables that are required to select on CC flavour
  const std::set<std::string> kCCFlavorReqs = {"mc.nu",
                                               "mc.nnu",
                                               "mc.nu.iscc",
                                               "mc.nu.pdg",
                                               "mc.nu.pdgorig"};

  // Finally, the function argument to the Cut constructor can be a "functor"
  // object (one with operator()). This allows similar logic but with different
  // constants to be easily duplicated.

  /// Select CC \f$ \nu_\mu\to\nu_e \f$
  const Cut kIsSig    (kCCFlavorReqs, CCFlavSel(12, 14));
  /// Select CC \f$ \nu_\mu\to\nu_\mu \f$
  const Cut kIsNumuCC (kCCFlavorReqs, CCFlavSel(14, 14));
  /// Select CC \f$ \nu_e\to\nu_e \f$
  const Cut kIsBeamNue(kCCFlavorReqs, CCFlavSel(12, 12));
  /// Select CC \f$ \nu_e\to\nu_\mu \f$
  const Cut kIsNumuApp(kCCFlavorReqs, CCFlavSel(14, 12));
  /// Select CC \f$ \nu_\mu\to\nu_\tau \f$
  const Cut kIsTauFromMu(kCCFlavorReqs, CCFlavSel(16, 14));
  /// Select CC \f$ \nu_e\to\nu_\tau \f$
  const Cut kIsTauFromE(kCCFlavorReqs, CCFlavSel(16, 12));

  /// Is this truly an antineutrino?
  const Cut kIsAntiNu({"mc.nu", "mc.nnu", "mc.nu.pdg"},
                      [](const caf::StandardRecord* sr)
                      {
			if(sr->mc.nnu == 0) return false;  
			assert(sr->mc.nnu == 1);
                        return sr->mc.nu[0].pdg < 0;
                      });

  /// Extremely crude CC selection, only for use in some old demos
  const Cut kCrudeCCSel({"trk.nkalman", "trk.kalman.len"},
                        [](const caf::StandardRecord* sr)
                        {
                          if(sr->trk.nkalman == 0) return false;
                          return sr->trk.kalman[0].len > 200;
                        });

  // Gareth's stuff

  /// Select truly quasielastic events
  const Cut kIsQE({"mc.nnu", "mc.nu.mode"},
                  [](const caf::StandardRecord* sr)
                  {
                    if(sr->mc.nnu == 0) return false;
                    assert(sr->mc.nnu == 1);
                    return (sr->mc.nu[0].mode == 0);
                  });

  // Kirk's stuff

  /// numu group quasielastic selection
  const Cut kQE({"sel.qepid.ntrk", "sel.qepid.pid"},
		[](const caf::StandardRecord* sr)
		{
		  return ( (sr->sel.qepid.ntrk == 1 && sr->sel.qepid.pid > 0.45) ||
			   (sr->sel.qepid.ntrk == 2 && sr->sel.qepid.pid > 0.4));
		});

  /// numu group selection
  const Cut kCC({"slc.calE", "sel.remid.pid", "trk.kalman.nhit","slc.nhit","slc.ncontplanes", "sel.remid.bestidx3d"},
                [](const caf::StandardRecord* sr)
                {
                  if(sr->sel.remid.bestidx3d == 999) return false;

                  return (sr->slc.calE > 0 &&
                          sr->sel.remid.pid > 0.75 &&
                          sr->trk.kalman[sr->sel.remid.bestidx3d].nhit > 10 &&
                          sr->slc.nhit > 20 &&
                          sr->slc.ncontplanes > 4);
                });

  /// Is it contained?
  const Cut kContainNick({"sand.numu.vtxdst", "sand.numu.mudst"},  // Nick's old containment
			 [](const caf::StandardRecord* sr)
			 {
			   return (sr->sand.numu.vtxdst > 50 &&
				   sr->sand.numu.mudst > 40);
			 });

  const Cut kContain({"sel.contain.numucontain"},
		     [](const caf::StandardRecord* sr)
		     {
		       return (sr->sel.contain.numucontain);
		     });

  const Cut kUCCosmicRej({"sel.cosrej.numuuncontpid", "sel.cosrej.anglekal"},  // NOT UPDATED
                         [](const caf::StandardRecord* sr)
                         {
                           return (sr->sel.cosrej.anglekal > 0.5 &&
                                   sr->sel.cosrej.numuuncontpid > 0.545);
                         });

  const Cut kCosmicRej({"sel.cosrej.anglekal", "sel.cosrej.numucontpid"},
                       [](const caf::StandardRecord* sr)
                       {
                         return (sr->sel.cosrej.anglekal  > 0.5 &&
                                 sr->sel.cosrej.numucontpid > 0.545);
                       });

  /// Cut set to select numu CC original events for MRCC.
  /// Checks that the highest track PID is more than 0.725,
  /// the higest PID track is 3D and the vertex dist > 50cm
  const Cut kIsGoodNumuForMrcc({"sel.remid.pid", "sel.remid.bestidx", "sel.remid.bestidx3d", "sand.numu.vtxdst"},
			       [](const caf::StandardRecord* sr)
			       {
				 if( sr->sel.remid.bestidx != sr->sel.remid.bestidx3d)
				   return false;
				 return ( sr->sel.remid.pid > 0.725  /*&&
								       sr->sand.numu.vtxdst > 50*/);
			       });
  
  /// Cut set to select contained nue events in ND
  const Cut kNDContainment({"vtx.nelastic", "vtx.elastic.vtx.fX",
	"vtx.elastic.vtx.fZ","vtx.elastic.vtx.fY",
	"shw.nshwlid", "shw.shwlid.stop.fX",
	"shw.shwlid.stop.fY", "shw.shwlid.stop.fZ"},
    [](const caf::StandardRecord* sr)
    {
      if( sr->vtx.nelastic <= 0 || sr->shw.nshwlid <=0 )
	return false;
      return ( abs(sr->vtx.elastic[0].vtx.X()) < 140  &&
	       abs(sr->vtx.elastic[0].vtx.Y()) < 140  &&
	       sr->vtx.elastic[0].vtx.Z()      < 1200 &&
	       sr->vtx.elastic[0].vtx.Z()      > 50   &&
	       abs(sr->shw.shwlid[0].stop.X()) < 185  &&
	       abs(sr->shw.shwlid[0].stop.Y()) < 185  &&
	       sr->shw.shwlid[0].stop.Z()      > 50   &&
	       sr->shw.shwlid[0].stop.Z()      < 1270);
    });
  
  
  //--------------------------------------------------------------------------
  // Data neutrino search useful cuts - likely temporary, but here for 
  // convenience for now as multiple folks use same cuts
  
  /// Cut events that are not in time 208us - 236us
  const Cut kInTime({"slc.meantime"},
		    [](const caf::StandardRecord* sr)
		    {
		      return (sr->slc.meantime >= 208000 && 
			      sr->slc.meantime <= 236000);
		    });
  
  
  /// An "angle-variable" type cut based on looking at events with a track-hit to slice-hit ratio > 0.88.
  /// Cuts vary based on number of hits in the slice to carve out a curved 2D space. \hideinitializer
  const Cut kAngVar({"slc.nhit","trk.cosmic.nhit","sand.nue.anglecosmic","trk.cosmic.dir.fY"},
                    [](const caf::StandardRecord* sr)
                    {
                      if(((sr->trk.cosmic[0].nhit*1.0)/(1.0*sr->slc.nhit))>0.88){
                        if(sr->slc.nhit>50 && sr->slc.nhit<=135 && 
			   (((sr->sand.nue.anglecosmic*sr->sand.nue.anglecosmic)
			     - (sr->trk.cosmic[0].dir.Y()*sr->trk.cosmic[0].dir.Y()))
			    <((1.1/85)*sr->slc.nhit-1.65))) 
			  return false;
			
                        if(sr->slc.nhit>135 && sr->slc.nhit<=250 && 
			   (((sr->sand.nue.anglecosmic*sr->sand.nue.anglecosmic)
			     - (sr->trk.cosmic[0].dir.Y()*sr->trk.cosmic[0].dir.Y()))
			    <((.6/115)*sr->slc.nhit-0.6))) 
			  return false;
			
                        if(sr->slc.nhit>250 && sr->slc.nhit<=500 && 
			   (((sr->sand.nue.anglecosmic*sr->sand.nue.anglecosmic)
			     - (sr->trk.cosmic[0].dir.Y()*sr->trk.cosmic[0].dir.Y()))
			    < ((0.1/250)*sr->slc.nhit+0.6))) 
			  return false;
			
                        if(sr->slc.nhit>500 && 
			   (((sr->sand.nue.anglecosmic*sr->sand.nue.anglecosmic)
			     - (sr->trk.cosmic[0].dir.Y()*sr->trk.cosmic[0].dir.Y())) > 0.8)) 
			  return false;
                      }
                      return true;
                    });
  
  /// XY containment in 2 di-block era based on slice box coordinates
  const Cut kXYContainment({"slc.boxmin.fX","slc.boxmin.fY","slc.boxmax.fX","slc.boxmax.fY"},
                           [](const caf::StandardRecord* sr)
                           {
                             return (sr->slc.boxmin.X() > -700 &&
                                     sr->slc.boxmin.Y() > -700 &&
                                     sr->slc.boxmax.X() < 700 &&
                                     sr->slc.boxmax.Y() < 660);
                           });
  
  /// 2 di-block era contained-type event selector
  const Cut kContained({"slc.boxmin.fZ","slc.boxmax.fZ"},
                       [](const caf::StandardRecord* sr)
                       {
                         return (sr->slc.boxmin.Z() > 100 &&
                                 sr->slc.boxmax.Z() < 800 );
                       });
  
  /// 2 di-block era through-going-type event selector
  const Cut kThroughGoing({"slc.boxmin.fZ","slc.boxmax.fZ"},
                          [](const caf::StandardRecord* sr)
                          {
                            return (sr->slc.boxmin.Z() < 100 &&
                                    sr->slc.boxmax.Z() > 800 );
			  });
  
  /// 2 di-block era Entering-type event selector
  const Cut kEnter({"slc.boxmin.fZ","slc.boxmax.fZ"},
                   [](const caf::StandardRecord* sr)
                   {
                     return (sr->slc.boxmin.Z() < 100 &&
                             sr->slc.boxmax.Z() < 800 );
                   });
  
  /// 2 di-block era Exiting-type event selector
  const Cut kExit({"slc.boxmin.fZ","slc.boxmax.fZ"},
                  [](const caf::StandardRecord* sr)
                  {
                    return (sr->slc.boxmin.Z() > 100 &&
                            sr->slc.boxmax.Z() > 800 );
                  });

  
  /// 2 di-block era cleanup before applying other cuts
  const Cut kCleanup({"slc.nhit","trk.cosmic.view"},
		     [](const caf::StandardRecord* sr)
		     {
		       return (sr->slc.nhit > 20 &&
			       sr->trk.cosmic[0].view == 2);
		     });
  
  
  
  
  /// functions for sliding containment, useful for modeling smaller detector for early reach
  class ContainFxn
  {
  public:
    ContainFxn(double zmax) : fZmax(zmax)
      {
      }
      
      bool operator()(const caf::StandardRecord* sr) const
      {
	return (sr->slc.boxmin.X() > -745 &&
		sr->slc.boxmin.Y() > -745 &&
		sr->slc.boxmax.X() < 745 &&
		sr->slc.boxmax.Y() < 720 &&
		sr->slc.boxmin.Z() > fZmax+12 &&
		sr->slc.boxmax.Z() < 5950 &&
		sr->sel.cosrej.kalfwdcell > 10 &&
		sr->sel.cosrej.kalbakcell > 10 &&
		sr->sel.cosrej.cosfwdcell > 10 &&
		sr->sel.cosrej.cosbakcell > 10);
	
      }
  protected:
      double fZmax;
  };

  /// Variables required by \ref ContainFxn
  const std::set<std::string> kContainFxnReqs = {"slc.boxmax.fX", "slc.boxmax.fY", "slc.boxmax.fZ", "slc.boxmin.fX", "slc.boxmin.fY", "slc.boxmin.fZ", "mc.nu", "mc.nu.vtx.fZ", "sel.contain.kalfwdcell", "sel.contain.kalbakcell", "sel.contain.cosfwdcell", "sel.contain.cosbakcell"};

  class NotContainFxn
  {
  public:
    NotContainFxn(double zmax) : fZmax(zmax)
      {
      }
      
      bool operator()(const caf::StandardRecord* sr) const
      {
	return ((sr->slc.boxmin.X() < -745 ||
		 sr->slc.boxmin.Y() < -745 ||
		 sr->slc.boxmax.X() > 745 ||
		 sr->slc.boxmax.Y() > 720 ||
		 sr->slc.boxmin.Z() > 5950 ||
		 sr->slc.boxmin.Z() < fZmax+12 ||
		 sr->sel.contain.kalfwdcell < 11 ||
		 sr->sel.contain.kalbakcell < 11 ||
		 sr->sel.contain.cosfwdcell < 11 ||
		 sr->sel.contain.cosbakcell < 11) && sr->mc.nu[0].vtx.Z() > fZmax);
	
      }
  protected:
      double fZmax;
  };

} // namespace
