#pragma once

#include <string>

namespace osc{class IOscCalculatorAdjustable;}

namespace ana
{
  /// Interface definition for fittable variables
  class FitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const = 0;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const = 0;
    virtual std::string ShortName() const = 0;
    virtual std::string LatexName() const = 0;
    virtual double Penalty(double /*val*/) const {return 0;}
  };

  //----------------------------------------------------------------------

  /// Base class for variables with constraints. Apply penalty outside range
  class ConstrainedFitVar: public FitVar
  {
  public:
    virtual double Penalty(double val) const;
    virtual double LowLimit() const = 0;
    virtual double HighLimit() const = 0;
  protected:
    double Clamp(double val) const;
  };

  //----------------------------------------------------------------------

  /// \f$ \theta_{13} \f$
  class FitTheta13: public FitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "th13";}
    virtual std::string LatexName() const {return "#theta_{13}";}
  };

  /// \f$ \theta_{13} \f$
  const FitTheta13 kFitTheta13 = FitTheta13();

  //----------------------------------------------------------------------

  /// \f$ \sin^22\theta_{13} \f$
  class FitSinSq2Theta13: public ConstrainedFitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "ss2th13";}
    virtual std::string LatexName() const {return "sin^{2}2#theta_{13}";}

    virtual double LowLimit() const {return 0;}
    virtual double HighLimit() const {return 1;}
  };

  /// \f$ \sin^22\theta_{13} \f$
  const FitSinSq2Theta13 kFitSinSq2Theta13 = FitSinSq2Theta13();

  //----------------------------------------------------------------------

  /// \f$ \delta_{CP}/\pi \f$
  class FitDeltaInPiUnits: public FitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "delta(pi)";}
    virtual std::string LatexName() const {return "#delta / #pi";}
  };

  /// \f$ \delta_{CP}/\pi \f$
  const FitDeltaInPiUnits kFitDeltaInPiUnits = FitDeltaInPiUnits();

  //----------------------------------------------------------------------

  /// \f$ \sin^2\theta_{23} \f$
  class FitSinSqTheta23: public ConstrainedFitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "ssth23";}
    virtual std::string LatexName() const {return "sin^{2}#theta_{23}";}

    virtual double LowLimit() const {return 0;}
    virtual double HighLimit() const {return 1;}
  };

  /// \f$ \sin^2\theta_{23} \f$
  const FitSinSqTheta23 kFitSinSqTheta23 = FitSinSqTheta23();

  //----------------------------------------------------------------------

  /// \f$ \sin^22\theta_{23} \f$
  class FitSinSq2Theta23: public ConstrainedFitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "ss2th23";}
    virtual std::string LatexName() const {return "sin^{2}2#theta_{23}";}


    virtual double LowLimit() const {return 0;}
    virtual double HighLimit() const {return 1;}
  };

  /// \f$ \sin^22\theta_{23} \f$
  const FitSinSq2Theta23 kFitSinSq2Theta23 = FitSinSq2Theta23();

  //----------------------------------------------------------------------

  /// \f$ \Delta m^2_{32} \f$
  class FitDmSq32: public ConstrainedFitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "dmsq32";}
    virtual std::string LatexName() const {return "#Deltam^{2}_{32}";}

    // "1eV^2 splitting should be enough for anyone"
    // OscCalculatorPMNS freaks out at large splittings
    virtual double LowLimit() const {return -1;}
    virtual double HighLimit() const {return +1;}
  };

  /// \f$ \Delta m^2_{32} \f$
  const FitDmSq32 kFitDmSq32 = FitDmSq32();

  //-------------------------------------------------------------------------

  /// \f$ \Delta m^2_{32}\times10^3{\rm eV}^2 \f$
  class FitDmSq32Scaled: public ConstrainedFitVar
  {
  public:
    virtual double GetValue(const osc::IOscCalculatorAdjustable* osc) const;
    virtual void SetValue(osc::IOscCalculatorAdjustable* osc, double val) const;
    virtual std::string ShortName() const {return "dmsq32scaled";}
    virtual std::string LatexName() const {return "#Deltam^{2}_{32} (10^{-3} eV^{2})";}

    // "1eV^2 splitting should be enough for anyone"
    // OscCalculatorPMNS freaks out at large splittings
    virtual double LowLimit() const {return -1000;}
    virtual double HighLimit() const {return +1000;}
  };

  /// \f$ \Delta m^2_{32}\times10^3{\rm eV}^2 \f$
  const FitDmSq32Scaled kFitDmSq32Scaled = FitDmSq32Scaled();
} // namespace
