//////////////////////////////////////////////////////////////////////////////
/// \file   MRCCDecomp.h
/// \brief  Defines the framework for decomposing ND spectra based on the MRCC 
///         technique
/// \author Kanika Sachdev (ksachdev@physics.umn.edu)
//////////////////////////////////////////////////////////////////////////////

#pragma once

#include "CAFAna/Binning.h"
#include "CAFAna/Cut.h"
#include "CAFAna/Cuts.h"
#include "CAFAna/Decomp.h"
#include "CAFAna/Prediction.h"
#include "CAFAna/Spectrum.h"
#include "CAFAna/Ratio.h"
#include "CAFAna/MRCCLoader.h"
#include "CAFAna/Var.h"
#include "CAFAna/Vars.h"

namespace ana
{
  /// \brief Decomposes ND spectra based on the MRCC technique
  ///
  /// The NC component is estimated using the NC/MRCC ratio from
  /// MC and MRCC spectrum from data. The intinsic \f$ \nu_e \f$ CC
  /// spectrum is just taken from the MC and scaled down to the data
  /// POT. The \f$ \nu_\mu \f$ CC component is the total data spectrum
  /// less NC component, less \f$ \nu_e \f$ CC component.
  class MRCCDecomp: public IDecomp
  {
  public:

    /// \brief 
    /// MRCCDecomp Constructor. This one takes in the MRCC MC and MRCC-Parent MC	
    /// SpectrumLoaders. It is only useful for construction of NC/MRCC ratio.  
    MRCCDecomp( MRCCLoader&           mrcc,
    		SpectrumLoaderBase&   nearMC,
    		std::string   label = "Reconstructed Neutrino Energy [GeV]",
    		const Binning& bins = kNueEnergyBinning,
    		const Var& eBins    = kCaloE,
    		const Cut& allCuts  = kPassesPresel && kNDContainment,
    		const Var& weight   = kUnweighted);
    

    /// \brief 
    /// MRCCDecomp Constructor. This one takes in the MRCC MC and MRCC-Parent MC	
    /// spectrum loaders, as well as MRCC data and MRCC-Parent data spectrum loaders. It 
    /// is useful for construction of NC/MRCC ratio and also decompose the spectrum in the 
    /// MRCC-Parent data files using that ratio.
    MRCCDecomp( MRCCLoader&         mrccMC,
                SpectrumLoaderBase& nearMC,
		MRCCLoader&         mrccDATA,
		SpectrumLoaderBase& NDDATA,
                std::string         label = "Reconstructed Neutrino Energy [GeV]",
		const Binning& bins = kNueEnergyBinning,
		const Var& eBins    = kCaloE,
		const Cut& allCuts  = kPassesPresel && kNDContainment,
		const Var& weight   = kUnweighted);

    ///\brief 
    /// Return the decomposed NC spectrum 
    /// In MRCC Decomposition, NC Spectrum is 
    /// MRCC data * (NC MC/ MRCC MC)
    virtual Spectrum NCComponent()      const override;


    ///\brief
    /// Return the decomposed intrinsic \f$ \nu_e \f$ CC spectrum    
    /// Since MRCC decomposition only really tells you the NC component,
    /// \f$ \nu_e \f$ CC spectrum is just taken directly from MC, scaled
    /// to the correct data POT.
    virtual Spectrum NueComponent()     const override;
    virtual Spectrum AntiNueComponent() const override;

    ///\brief 
    /// Return the decomposed \f$ \nu_\mu\f$ CC spectrum   
    /// Since MRCC decomposition only really tells you the NC component,
    /// \f$ \nu_\mu \f$ CC spectrum is just Total - NC - \f$\nu_e \f$ CC
    virtual Spectrum NumuComponent() const override; 
    virtual Spectrum AntiNumuComponent() const override;

    ///\brief 
    /// Save decomposed spectra to specified directory
    virtual void SaveTo(TDirectory* dir) const override;

    ///\brief 
    /// Returns the true NC spectrum from MC that was used to make
    /// the MRCC ratio.
    Spectrum NCSpectrumMC() const { return fNCSpec;}

    ///\brief
    /// Returns the MRCC spectrum from MC that was used to make
    /// the MRCC ratio.
    Spectrum MRCCSpectrumMC() const { return fMrccSpec;}

    ///\brief 
    ///Returns the MRCC spectrum from data that was used to make
    /// NC Component estimate in that data set.
    Spectrum MRCCSpectrumData() const { return fDataMrccSpec;}

    ///\brief 
    ///Returns the histogram of NC/MRCC ratio used for estimating
    /// the NC Component.
    Ratio NCToMRCCRatio() const;

  private:

    // All the spectra below are post selection cuts

    Spectrum fMrccSpec;       ///< MRCC Spectrum from MC
    Spectrum fNCSpec;         ///< NC Spectrum from MC
    Spectrum fNueCCSpec;      ///< Beam Nue CC spectrum from MC
    Spectrum fAntiNueCCSpec;  ///< Beam AntiNue CC spectrum from MC

    /// NuMu CC spectrum from MC, for splitting numu component by charge
    Spectrum fNumuCCSpec;
    /// AntiNuMu CC spectrum from MC, for splitting numu component by charge
    Spectrum fAntiNumuCCSpec;

    Spectrum fDataMrccSpec;   ///< MRCC Spectrum from Data
    Spectrum fDataAll;        ///< Spectrum of all events in ND, undecomposed
    
  };// end of MRCCDecomp class definition 

}// end of namespace ana
