#pragma once

#include <vector>

#include "CAFAna/Decomp.h"
#include "CAFAna/Var.h"
#include "CAFAna/Vars.h"
#include "CAFAna/Cuts.h"
#include "CAFAna/Binning.h"

class TDirectory;

namespace ana
{

  const float fracInt = 1;

  // This is the Binning Scheme I'll use for the # Michel Templates
  const Binning SBinning = Binning::Simple(25,0,25);
  const Binning MBinning = Binning::Simple(4,0,4);
  // This is just a way to access # Michel info from the CAFs
  const Var kNMichels({"me.nslc"},
		      [](const caf::StandardRecord* sr)
  		      {return sr->me.nslc;});
  // This is the Number of Slices in each spill in the ND
  const Var kNSlcs({"slc.nnonnoise"},
		      [](const caf::StandardRecord* sr)
		      {return sr->slc.nnonnoise;});
  // This is a reweight: used for mocking different beam intensities in MC
  const Var kIntReWei({"slc.nnonnoise"},
		      [](const caf::StandardRecord* sr)
		      {return pow(fracInt,sr->slc.nnonnoise)*
			  exp(10.06*(1-fracInt));});
  // The selection I'm using right now has a containment cut && a PID cut
  const Cut kMDecompSel({"slc.boxmin.fX","slc.boxmin.fY","slc.boxmin.fZ",
			    "slc.boxmax.fX","slc.boxmax.fY","slc.boxmax.fZ",
			    "sel.lem.pid"},
			[](const caf::StandardRecord* sr){
			  return (sr->slc.boxmax.X() <  177 && 
				  sr->slc.boxmax.Y() <  177 &&
				  sr->slc.boxmax.Z() < 1300 &&
				  sr->slc.boxmin.X() > -177 && 
				  sr->slc.boxmin.Y() > -177 &&
				  sr->slc.boxmin.Z() >   645 &&
				  sr->sel.lem.pid > 0.8);
			});
  // This Cut helps to differentiate the background into three proccesses
  const Cut kIsNumu({"mc.nu","mc.nnu","mc.nu.pdg"},
		    [](const caf::StandardRecord* sr){
		      if (sr->mc.nnu == 0)
			return false;
		      return (abs(sr->mc.nu[0].pdg) == 14);
		    });
 
  /// Just return the ND truth spectra as the decomposition
  class MichelDecomp: public IDecomp
  {
  public:
    MichelDecomp(SpectrumLoaderBase& NDMCloader,
		 SpectrumLoaderBase& NDdataloader,
		 const HistAxis& axis,
		 const Cut& cut = kMDecompSel,
		 const Var& wei = kUnweighted);

    MichelDecomp(SpectrumLoaderBase& NDMCloader,
		 SpectrumLoaderBase& NDdataloader,
		 std::string       label,
		 const Binning&    bins,
		 const Var&        var,
		 const Cut& cut = kMDecompSel,
		 const Var& wei = kUnweighted);

    virtual Spectrum NCComponent()       const override;
    virtual Spectrum NumuComponent()     const override;
    virtual Spectrum AntiNumuComponent() const override;
    virtual Spectrum NueComponent()      const override;
    virtual Spectrum AntiNueComponent()  const override;

    virtual void Fitter(TH2* CC, TH2* NC, TH2* NE, TH2* Data,
    			std::vector<double>& cc_scale_factors,
			std::vector<double>& nc_scale_factors) const;
    void DataFluctuate();

    std::vector<float> SlcWei() const;
    TH2* IntensityReweight(Spectrum in, std::vector<float> wei) const;

    virtual void SaveTo(TDirectory* dir) const override;
    static std::unique_ptr<MichelDecomp> LoadFrom(TDirectory* dir);

  protected:
    MichelDecomp() :
      fNC( 0, 0),
      fNumu( 0, 0),
      fAntiNumu( 0, 0),
      fNue( 0, 0),
      fAntiNue( 0, 0),
      fData( 0, 0),
      fMCNSlcs( 0, 0),
      fDataNSlcs( 0, 0),
      fMichelNC( 0, 0),
      fMichelNumu( 0, 0),
      fMichelNue( 0, 0),
      fMichelData( 0, 0),
      fMockData( 0, 0)
	{};
    /// Helper for \ref NumuComponent and \ref AntiNumuComponent
    Spectrum ScaleNumu(const Spectrum& spect) const;

    Spectrum fNC;
    Spectrum fNumu;
    Spectrum fAntiNumu;
    Spectrum fNue;
    Spectrum fAntiNue;
    Spectrum fData;

    Spectrum fMCNSlcs;
    Spectrum fDataNSlcs;

    Spectrum fMichelNC;
    Spectrum fMichelNumu;
    Spectrum fMichelNue;
    Spectrum fMichelData;
    Spectrum fMockData;

  };
}
