#pragma once

#include "CAFAna/Extrap.h"
#include "CAFAna/ModularExtrapComponent.h"

class TDirectory;

namespace ana
{

  class Loaders;
  class SpectrumLoaderBase;
  class IDecomp;
  class HistAxis;
  class Cut;
  class Var;
  class OscillatableSpectrum;

  /// Extrapolates each component using a separate SingleComponentExtrap.
  /// Only extrapolates one sample.
  class ModularExtrap: public IExtrap
  {
    public:

      /// Creates a nue-like extraploation.
      /** numuDecomp and numuNDCut are for the signal component (mu->e).
          nueDecomp and nueNDCut are for the backgound components. */
      static ModularExtrap Nue(
        Loaders& loaders,
        const IDecomp& nueDecomp,
        const IDecomp& numuDecomp,
        const HistAxis& axis,
        const Cut& fdcut,
        const Cut& nueNDcut,
        const Cut& numuNDcut,
        const Var& weight = kUnweighted
      );

      /// Creates a numu-like extrapolation.
      static ModularExtrap Numu(
        Loaders& loaders,
        const IDecomp& numuDecomp,
        const HistAxis& axis,
        const Cut& fdcut,
        const Cut& ndcut,
        const Var& weight = kUnweighted
      );

      /// Creates a nue-like extraploation with individual spectrum loaders.
      /** numuDecomp and numuNDCut are for the signal component (mu->e).
          nueDecomp and nueNDCut are for the backgound components. */
      static ModularExtrap Nue(
        SpectrumLoaderBase& nearMCLoader,
        SpectrumLoaderBase& farMCswapLoader,
        SpectrumLoaderBase& farMCnonswapLoader,
        SpectrumLoaderBase& farMCtauswapLoader,
        const IDecomp& nueDecomp,
        const IDecomp& numuDecomp,
        const HistAxis& axis,
        const Cut& fdcut,
        const Cut& nueNDcut,
        const Cut& numuNDcut,
        const Var& weight = kUnweighted
      );

      /// Creates a numu-like extrapolation with individual spectrum loaders.
      static ModularExtrap Numu(
        SpectrumLoaderBase& nearMCLoader,
        SpectrumLoaderBase& farMCswapLoader,
        SpectrumLoaderBase& farMCnonswapLoader,
        SpectrumLoaderBase& farMCtauswapLoader,
        const IDecomp& numuDecomp,
        const HistAxis& axis,
        const Cut& fdcut,
        const Cut& ndcut,
        const Var& weight = kUnweighted
      );

      // Prevent copying because we own objects on the free store.
      ModularExtrap(const ModularExtrap&) = delete;
      ModularExtrap& operator=(const ModularExtrap&) = delete;
      ModularExtrap(ModularExtrap&&) = default;
      ModularExtrap& operator=(ModularExtrap&&) = default;

      void SaveTo(TDirectory* dir) const override;
      static std::unique_ptr<ModularExtrap> LoadFrom(TDirectory* dir);

      // override abstract methods
      OscillatableSpectrum NueSurvComponent()      override;
      OscillatableSpectrum AntiNueSurvComponent()  override;
      OscillatableSpectrum NumuSurvComponent()     override;
      OscillatableSpectrum AntiNumuSurvComponent() override;
      OscillatableSpectrum NueAppComponent()       override;
      OscillatableSpectrum AntiNueAppComponent()   override;
      OscillatableSpectrum NumuAppComponent()      override;
      OscillatableSpectrum AntiNumuAppComponent()  override;
      Spectrum             NCComponent()           override;
      // Not yet in base class.  Names not set in stone.
      OscillatableSpectrum TauFromMuComponent();
      OscillatableSpectrum AntiTauFromMuComponent();
      OscillatableSpectrum TauFromEComponent();
      OscillatableSpectrum AntiTauFromEComponent();

      ~ModularExtrap();

    private:

      /// Sets up components common to both styles.
      ModularExtrap(
        SpectrumLoaderBase& farMCswapLoader,
        SpectrumLoaderBase& farMCnonswapLoader,
        SpectrumLoaderBase& farMCtauswapLoader,
        const HistAxis& axis,
        const Cut& fdcut,
        const Var& weight
      );

      ModularExtrap(){};

      SingleComponentExtrap* fEEextrap;
      SingleComponentExtrap* fMMextrap;
      SingleComponentExtrap* fMEextrap;
      SingleComponentExtrap* fEMextrap;
      SingleComponentExtrap* fNCextrap;
      SingleComponentExtrap* fEEAntiextrap;
      SingleComponentExtrap* fMMAntiextrap;
      SingleComponentExtrap* fMEAntiextrap;
      SingleComponentExtrap* fEMAntiextrap;
      SingleComponentExtrap* fMTextrap;
      SingleComponentExtrap* fETextrap;
      SingleComponentExtrap* fMTAntiextrap;
      SingleComponentExtrap* fETAntiextrap;

  };

}
