#pragma once

#include "CAFAna/Binning.h"
#include "CAFAna/Cut.h"
#include "CAFAna/Cuts.h"
#include "CAFAna/Decomp.h"
#include "CAFAna/Extrap.h"
#include "CAFAna/Loaders.h"
#include "CAFAna/Ratio.h"
#include "CAFAna/Spectrum.h"
#include "CAFAna/SpectrumLoaderBase.h"
#include "CAFAna/Var.h"
#include "CAFAna/Vars.h"

namespace ana
{
  /// Perform a direct extrapolation using far over near ratios and an appearance technique.
  class NDDataMCRatioExtrap: public IExtrap
  {
    public:
      /// Class constructor: note the differences between the 4 input Var and Cut parameters.
      /// The first set is for the survival method and final spectrum.
      /// The second set is for the appearance method oscillating nue to numu.
      /// The last two sets are for the appearance method oscillating numu to nue.
      /// Before oscillation of numu to nue, the numu spectrum is broken by QE and non-QE,
      /// hence two sets of Var and Cut inputs.
      /// \param isNuE Determines whether cutSurvNuE or cutSurvNuMu is applied to oscillated spectra
      /// \param signSplit Option to split events by sign from MC
      NDDataMCRatioExtrap(Loaders& loaders,
                          IDecomp* decomp,
                          std::string label   = "Reconstructed Energy (GeV)",
                          const Binning& bins = kNumuEnergyBinning,
                          const bool isNuE     = true,
                          const bool signSplit = true,
                          const Var& varSurv          = kCaloE,
                          const Cut& cutSurvPre       = kPassesPresel,
                          const Cut& cutSurvNC        = kIsNC,
                          const Cut& cutSurvNuE       = kIsBeamNue,
                          const Cut& cutSurvNuMu      = kIsNumuCC,
                          const Var& varApprCCNuE     = kCaloE,
                          const Cut& cutApprCCNuEPre  = kPassesPresel,
                          const Cut& cutApprCCNuESel  = kIsBeamNue,
                          const Var& varApprCCNuMu    = kCCE,
                          const Cut& cutApprCCNuMuPre = kCrudeCCSel,
                          const Cut& cutApprCCNuMuSel = kIsNumuCC && !kIsQE,
                          const Var& varApprQENuMu    = kQEE,
                          const Cut& cutApprQENuMuPre = kCrudeCCSel,
                          const Cut& cutApprQENuMuSel = kIsNumuCC &&  kIsQE,
                          const Cut& cutContainND = kNoCut,
                          const Cut& cutContainFD = kNoCut,
                          const Var& wei = kUnweighted);

      virtual OscillatableSpectrum NueSurvComponent();
      virtual OscillatableSpectrum AntiNueSurvComponent();

      virtual OscillatableSpectrum NumuSurvComponent();
      virtual OscillatableSpectrum AntiNumuSurvComponent();

      virtual OscillatableSpectrum NueAppComponent();
      virtual OscillatableSpectrum AntiNueAppComponent();

      virtual OscillatableSpectrum NumuAppComponent();
      virtual OscillatableSpectrum AntiNumuAppComponent();

      virtual Spectrum NCComponent();

      Spectrum NearMC()   const; ///< Outputs the summed ND components from MC (components used for the survival analysis)
      Spectrum FarMC()    const; ///< Outputs the summed FD components from MC without oscillations (components used for survival method)
      Spectrum NearData() const; ///< Outputs the summed ND components from data (components used for the survival analysis)

      Ratio SignFraction(int pdg) const; ///< Shows the ratio of events of a given flavor that are neutrinos or anti-neutrinos

      Ratio NDRatio(int pdg = 0, bool appr = false) const; ///< Shows the ND Data/MC ratio for a given current and flavor

      Ratio Purity    (int pdg) const; ///< Shows the purity of a given neutrino flavor for the appearance analysis
      Ratio Efficiency(int pdg) const; ///< Shows the efficiency of a given neutrino flavor for the appearance analysis

      virtual void SaveTo(TDirectory* dir) const;
      static std::unique_ptr<NDDataMCRatioExtrap> LoadFrom(TDirectory* dir);

    protected:
      NDDataMCRatioExtrap()
        : fSignSplit(true), fDecomp(nullptr),
          fApprDataCCNuE(0, 0), fApprDataCCNuMu(0, 0), fApprDataQENuMu(0, 0),

          fSurvNDNC(0, 0),          fSurvFDNC(0, 0),
          fSurvNDCCNuE(0, 0),       fSurvNDCCANuE(0, 0),
          fSurvNDCCNuMu(0, 0),      fSurvNDCCANuMu(0, 0),
          fOSpecSurvFDCCNuE(0, 0),  fOSpecSurvFDCCANuE(0, 0),
          fOSpecSurvFDCCNuMu(0, 0), fOSpecSurvFDCCANuMu(0, 0),

          fApprNDCCNuE(0, 0),       fApprNDCCANuE(0, 0),
          fApprNDCCNuMu(0, 0),      fApprNDCCANuMu(0, 0),
          fApprNDQENuMu(0, 0),      fApprNDQEANuMu(0, 0),
          fOSpecApprFDCCNuE(0, 0),  fOSpecApprFDCCANuE(0, 0),
          fOSpecApprFDCCNuMu(0, 0), fOSpecApprFDCCANuMu(0, 0),
          fOSpecApprFDQENuMu(0, 0), fOSpecApprFDQEANuMu(0, 0),

          fOSpecOscdFDCCNuE(0, 0),   fOSpecOscdFDCCANuE(0, 0),
          fOSpecOscdFDCCNuMu(0, 0),  fOSpecOscdFDCCANuMu(0, 0),
          fOSpecOscdFDCCNuTau(0, 0), fOSpecOscdFDCCANuTau(0, 0),

          fApprFDPurEffNumCCNuE(0, 0),  fApprFDPurEffNumCCANuE(0, 0),
          fApprFDPurDenCCNuE(0, 0),     fApprFDPurDenCCANuE(0, 0),
          fApprFDEffDenCCNuE(0, 0),     fApprFDEffDenCCANuE(0, 0),
          fApprFDPurEffNumCCNuMu(0, 0), fApprFDPurEffNumCCANuMu(0, 0),
          fApprFDPurDenCCNuMu(0, 0),    fApprFDPurDenCCANuMu(0, 0),
          fApprFDEffDenCCNuMu(0, 0),    fApprFDEffDenCCANuMu(0, 0),
          fApprFDPurEffNumQENuMu(0, 0), fApprFDPurEffNumQEANuMu(0, 0),
          fApprFDPurDenQENuMu(0, 0),    fApprFDPurDenQEANuMu(0, 0),
          fApprFDEffDenQENuMu(0, 0),    fApprFDEffDenQEANuMu(0, 0)
      {}

      /// SpectrumLoader based constructor. This constructor always splits by sign.
      NDDataMCRatioExtrap(SpectrumLoaderBase& nearMC,
                          SpectrumLoaderBase& far_MC,
                          SpectrumLoaderBase& swapMC,
                          SpectrumLoaderBase& tau_MC,
                          SpectrumLoaderBase& NDDATA,
                          IDecomp*            decomp,
                          std::string label   = "Reconstructed Energy (GeV)",
                          const Binning& bins = kNumuEnergyBinning,
                          const bool isNuE     = true,
                          const bool signSplit = true,
                          const Var& varSurv          = kCaloE,
                          const Cut& cutSurvPre       = kPassesPresel,
                          const Cut& cutSurvNC        = kIsNC,
                          const Cut& cutSurvNuE       = kIsBeamNue,
                          const Cut& cutSurvNuMu      = kIsNumuCC,
                          const Var& varApprCCNuE     = kCaloE,
                          const Cut& cutApprCCNuEPre  = kPassesPresel,
                          const Cut& cutApprCCNuESel  = kIsBeamNue,
                          const Var& varApprCCNuMu    = kCCE,
                          const Cut& cutApprCCNuMuPre = kCrudeCCSel,
                          const Cut& cutApprCCNuMuSel = kIsNumuCC && !kIsQE,
                          const Var& varApprQENuMu    = kQEE,
                          const Cut& cutApprQENuMuPre = kCrudeCCSel,
                          const Cut& cutApprQENuMuSel = kIsNumuCC &&  kIsQE,
                          const Cut& cutContainND = kNoCut,
                          const Cut& cutContainFD = kNoCut,
                          const Var& wei = kUnweighted);


      bool fSignSplit; ///< Helper variable to determine whether to split neutrinos by sign
      void NoSignSplitCorrection(); ///< Combines neutrino and anti-neutrino plots if sign-splitting is turned off

      // Spectra to be filled from CAFs

      // Contains Spectra that have data information
      IDecomp* fDecomp;

      Spectrum fApprDataCCNuE;
      Spectrum fApprDataCCNuMu;
      Spectrum fApprDataQENuMu;

      Spectrum fSurvNDNC;
      Spectrum fSurvFDNC;

      Spectrum fSurvNDCCNuE;
      Spectrum fSurvNDCCANuE;
      Spectrum fSurvNDCCNuMu;
      Spectrum fSurvNDCCANuMu;
      OscillatableSpectrum fOSpecSurvFDCCNuE;
      OscillatableSpectrum fOSpecSurvFDCCANuE;
      OscillatableSpectrum fOSpecSurvFDCCNuMu;
      OscillatableSpectrum fOSpecSurvFDCCANuMu;

      Spectrum fApprNDCCNuE;
      Spectrum fApprNDCCANuE;
      Spectrum fApprNDCCNuMu;
      Spectrum fApprNDCCANuMu;
      Spectrum fApprNDQENuMu;
      Spectrum fApprNDQEANuMu;
      OscillatableSpectrum fOSpecApprFDCCNuE;
      OscillatableSpectrum fOSpecApprFDCCANuE;
      OscillatableSpectrum fOSpecApprFDCCNuMu;
      OscillatableSpectrum fOSpecApprFDCCANuMu;
      OscillatableSpectrum fOSpecApprFDQENuMu;
      OscillatableSpectrum fOSpecApprFDQEANuMu;

      OscillatableSpectrum fOSpecOscdFDCCNuE;
      OscillatableSpectrum fOSpecOscdFDCCANuE;
      OscillatableSpectrum fOSpecOscdFDCCNuMu;
      OscillatableSpectrum fOSpecOscdFDCCANuMu;
      OscillatableSpectrum fOSpecOscdFDCCNuTau;
      OscillatableSpectrum fOSpecOscdFDCCANuTau;

      Spectrum fApprFDPurEffNumCCNuE;
      Spectrum fApprFDPurEffNumCCANuE;
      Spectrum fApprFDPurDenCCNuE;
      Spectrum fApprFDPurDenCCANuE;
      Spectrum fApprFDEffDenCCNuE;
      Spectrum fApprFDEffDenCCANuE;
      Spectrum fApprFDPurEffNumCCNuMu;
      Spectrum fApprFDPurEffNumCCANuMu;
      Spectrum fApprFDPurDenCCNuMu;
      Spectrum fApprFDPurDenCCANuMu;
      Spectrum fApprFDEffDenCCNuMu;
      Spectrum fApprFDEffDenCCANuMu;
      Spectrum fApprFDPurEffNumQENuMu;
      Spectrum fApprFDPurEffNumQEANuMu;
      Spectrum fApprFDPurDenQENuMu;
      Spectrum fApprFDPurDenQEANuMu;
      Spectrum fApprFDEffDenQENuMu;
      Spectrum fApprFDEffDenQEANuMu;
  };
}
