#pragma once

#include "CAFAna/Binning.h"
#include "CAFAna/Var.h"
#include "CAFAna/Cut.h"
#include "CAFAna/HistAxis.h"
#include "CAFAna/SpectrumLoaderBase.h"

#include <memory>
#include <string>
#include <vector>

class TDirectory;
class TH1;
class TH2;
class TH3;
class TH1D;

/// Oscillation analysis framework, runs over CAF files outside of ART
namespace ana
{
  class Ratio;

  /// Representation of a spectrum in any variable, with associated POT
  class Spectrum
  {
  public:
    friend class SpectrumLoader;
    friend class MRCCLoader;

    Spectrum(const std::string& label, const Binning& bins,
             SpectrumLoaderBase& loader,
             const Var& var,
             const Cut& cut = kNoCut,
             const Var& wei = kUnweighted);

    Spectrum(SpectrumLoaderBase& loader,
             const HistAxis& axis,
             const Cut& cut = kNoCut,
             const Var& wei = kUnweighted);

    Spectrum(const std::string& label, const Binning& bins);
    Spectrum(const std::string& label, double pot, const Binning& bins);

    /// Takes posession of \a h. ie, don't try to delete \a h or use it for anything else.
    Spectrum(TH1* h, double pot, std::string varName = "");

    /// 2D Spectrum of two Vars
    Spectrum(const std::string& label, SpectrumLoaderBase& loader,
             const Binning& binsx, const Var& varx,
             const Binning& binsy, const Var& vary,
             const Cut& cut = kNoCut,
             const Var& wei = kUnweighted);

    /// 3D Spectrum of three Vars
    Spectrum(const std::string& label, SpectrumLoaderBase& loader,
             const Binning& binsx, const Var& varx,
             const Binning& binsy, const Var& vary,
             const Binning& binsz, const Var& varz,
             const Cut& cut = kNoCut,
             const Var& wei = kUnweighted);

    virtual ~Spectrum();

    Spectrum(const Spectrum& rhs);
    Spectrum& operator=(const Spectrum& rhs);

    void Fill(double x, double w = 1);

    TH1D* ToTH1(double pot) const;

    /// Spectrum must be 2D to obtain TH2
    TH2*  ToTH2     (double pot) const;
    /// Spectrum must be 2D to obtain TH2. Normalized to X axis.
    TH2*  ToTH2NormX(double pot) const;

    /// Spectrum must be 3D to obtain TH3
    TH3*  ToTH3     (double pot) const;

    /// \brief Return total number of events scaled to \a pot
    ///
    /// \param pot POT to scale to
    /// \param err The statistical error on this total (optional)
    double Integral(double pot, double* err = 0) const;

    /// \brief Mock data is \ref FakeData with Poisson fluctuations applied
    ///
    /// Use for low-budget MDCs, or just getting a sense of the expected scale
    /// of statistical variation
    Spectrum MockData(double pot) const;
    /// \brief Fake data is a MC spectrum scaled to the POT expected in the data
    ///
    /// Use for sensitivity plots and testing fit convergence
    Spectrum FakeData(double pot) const;

    double POT() const {return fPOT;}

    /// DO NOT USE UNLESS DEALING WITH BROKEN FILES!
    void OverridePOT(double newpot) {fPOT = newpot;}

    void Clear();

    // Arithmetic operators are as if these are unlike samples, each a
    // contribution to one total, not seperate sources of stats for the same
    // sample.
    Spectrum& operator+=(const Spectrum& rhs);
    Spectrum operator+(const Spectrum& rhs) const;

    Spectrum& operator-=(const Spectrum& rhs);
    Spectrum operator-(const Spectrum& rhs) const;

    Spectrum& operator*=(const Ratio& rhs);
    Spectrum operator*(const Ratio& rhs) const;

    Spectrum& operator/=(const Ratio& rhs);
    Spectrum operator/(const Ratio& rhs) const;

    void SaveTo(TDirectory* dir) const;
    static std::unique_ptr<Spectrum> LoadFrom(TDirectory* dir);

  protected:
    void ConstructHistogram(const std::string& label, const Binning& bins);
  protected:
    TH1D* fHist;
    double fPOT;

    // The following internal variables are used in case Spectrum is
    // Multi Dimensional. Currently only 2D are supported.
    std::vector<Binning> fBinning;      ///< Binning for Multi-Dimensional Spectrum
    std::string          fLabel;        ///< Labels for 2D histo
  };

  // Commutative
  inline Spectrum operator*(const Ratio& lhs, const Spectrum& rhs){return rhs*lhs;}
  inline Spectrum operator/(const Ratio& lhs, const Spectrum& rhs){return rhs/lhs;}
}
