#pragma once

#include "Rtypes.h"

#include "CAFAna/SystShifts.h"

class TH2;
class TH2F;

namespace osc{class IOscCalculatorAdjustable;}

namespace ana
{
  class Experiment;
  class FitVar;

  /// Log-likelihood scan across two parameters
  class Surface
  {
  public:
    friend class NumuSurface;

    /// \param calc Values for oscillation parameters to be held fixed
    /// \param margVars Oscillation parameters to marginalize over
    /// \param margSysts Systematic parameters to marginalize over
    /// \param parallel Use all the cores on this machine? Be careful...
    Surface(const Experiment* expt,
            osc::IOscCalculatorAdjustable* calc,
            const FitVar* xvar, int nbinsx, double xmin, double xmax,
            const FitVar* yvar, int nbinsy, double ymin, double ymax,
            const std::vector<const FitVar*>& margVars = {},
            const std::vector<Syst>& margSysts = {},
            bool parallel = false);

    /// Draw the surface itself
    void Draw() const;
    /// Draw the best fit point
    void DrawBestFit(Color_t color) const;
    double MinChi() const {return fMinChi;}
    /// \param fc Surface to compare against for this significance level
    /// \param style Line style for TAttLine, solid, dotted, dashed etc
    /// \param color Line color for TAttLine
    /// \param minchi \f$\chi^2\f$ of best fit to compare against.
    ///               Default: best fit from this surface.
    void DrawContour(TH2* fc, Style_t style, Color_t color,
                     double minchi = -1);
    TH2* ToTH2(double minchi = -1) const;
    void SetTitle(const char* str);

    /// Maps of the values taken on by the marginalized parameters
    std::vector<TH2*> GetMarginalizedHists() {return fMargHists;}
  protected:
    void EnsureAxes() const;

    void FillSurface(const std::string& progTitle,
                     const Experiment* expt,
                     osc::IOscCalculatorAdjustable* calc,
                     const FitVar* xvar, const FitVar* yvar,
                     const std::vector<const FitVar*>& margVars,
                     const std::vector<Syst>& margSysts);

    void FillSurfacePoint(const Experiment* expt,
                          osc::IOscCalculatorAdjustable* calc,
                          const FitVar* xvar, double x,
                          const FitVar* yvar, double y,
                          const std::vector<const FitVar*>& margVars,
                          const std::vector<Syst>& margSysts);

    bool fParallel;

    double fMinChi;
    double fMinX, fMinY; // Best fit point
    TH2F* fHist;
    std::vector<TH2*> fMargHists;
  };

  /// Up-value surface for 68% confidence in 2D in gaussian approximation
  TH2* Gaussian68Percent2D(const Surface& s);
  /// Up-value surface for 90% confidence in 2D in gaussian approximation
  TH2* Gaussian90Percent2D(const Surface& s);
  /// Up-value surface for 95% confidence in 2D in gaussian approximation
  TH2* Gaussian95Percent2D(const Surface& s);
  /// Up-value surface for 2 sigma confidence in 2D in gaussian approximation
  TH2* Gaussian2Sigma2D   (const Surface& s);
  /// Up-value surface for 99% confidence in 2D in gaussian approximation
  TH2* Gaussian99Percent2D(const Surface& s);
  /// Up-value surface for 3 sigma confidence in 2D in gaussian approximation
  TH2* Gaussian3Sigma2D   (const Surface& s);

  // First approximation of the correct up-values to use for ss2th13 vs delta

  /// Up-value surface for 68% confidence in 2D in gaussian approximation
  TH2* Gaussian68Percent1D(const Surface& s);
  /// Up-value surface for 90% confidence in 2D in gaussian approximation
  TH2* Gaussian90Percent1D(const Surface& s);
  /// Up-value surface for 95% confidence in 2D in gaussian approximation
  TH2* Gaussian95Percent1D(const Surface& s);
  /// Up-value surface for 2 sigma confidence in 2D in gaussian approximation
  TH2* Gaussian2Sigma1D   (const Surface& s);
  /// Up-value surface for 99% confidence in 2D in gaussian approximation
  TH2* Gaussian99Percent1D(const Surface& s);
  /// Up-value surface for 3 sigma confidence in 2D in gaussian approximation
  TH2* Gaussian3Sigma1D   (const Surface& s);
}
