///////////////////////////////////////////////////////////////////////////
/// \brief   Histograms used by attenuation calibration
/// \author  Christopher Backhouse - bckhouse@caltech.edu
/// \date    Apr 2013
///////////////////////////////////////////////////////////////////////////
#ifndef CALDP_ATTENPROFILES_H
#define CALDP_ATTENPROFILES_H

#include "TH2.h"

#include "GeometryObjects/OfflineChan.h"

#include <cassert>
#include <cmath>
#include <map>

namespace caldp
{
  /// Lightweight TProfile equivalent
  class LiteProfile
  {
  public:

    // No-one in Calibration wants default constructors, they just lead to
    // things accidentally getting the wrong binning, much better to be
    // explicit. But generating reflex dictionaries relies on them, so control
    // them with this #define
#ifdef ATTENPROFILES_DEFAULT_CONSTRUCTORS
    LiteProfile() : fMinW(-1), fMaxW(-1), fCount(), fSum(), fSqSum() {}
#endif

#ifndef __GCCXML__
    LiteProfile(float w0, float w1)
      : fMinW(w0), fMaxW(w1),
        fCount(), fSum(), fSqSum() // zero-initialize arrays
    {
    }

    void Fill(float x, float y)
    {
      const int bin = ((x-fMinW)*kNumBins)/(fMaxW-fMinW);
      if(bin < 0 || bin >= kNumBins) return; // Drop overflows

      assert(y >= 0);

      ++fCount[bin];
      fSum[bin] += y;
      fSqSum[bin] += y*y;
    }

    LiteProfile& operator+=(const LiteProfile& rhs)
    {
      assert(rhs.fMinW == fMinW);
      assert(rhs.fMaxW == fMaxW);

      for(int i = 0; i < kNumBins; ++i){
        fCount[i] += rhs.fCount[i];
        fSum[i] += rhs.fSum[i];
        fSqSum[i] += rhs.fSqSum[i];
      }

      return *this;
    }

    TH1F* ToTH1F() const
    {
      TH1F* h = new TH1F("", "", kNumBins, fMinW, fMaxW);

      for(int i = 0; i < kNumBins; ++i){
        const int N = fCount[i];
        if(N == 0) continue;
	h->SetEntries(h->GetEntries()+N);

        h->SetBinContent(i+1, fSum[i]/N);

        const float arg = fSqSum[i]-fSum[i]*fSum[i]/N;
        if(arg >= 0){
          h->SetBinError(i+1, sqrt(arg)/N);
        }
        else{
          h->SetBinError(i+1, fSum[i]/N);
        }
      }

      return h;
    }

    int TotalHits() const
    {
      int ret = 0;
      for(int n: fSum) ret += n;
      return ret;
    }
#endif
  protected:
    float fMinW, fMaxW;

    static const int kNumBins = 100;

    int fCount[kNumBins];
    float fSum[kNumBins];
    float fSqSum[kNumBins];
  };

  /// Profiles used by attenuation calibration
  class AttenProfiles
  {
  public:
#ifdef ATTENPROFILES_DEFAULT_CONSTRUCTORS
    AttenProfiles() {}
#endif

#ifndef __GCCXML__
    AttenProfiles(float minW, float maxW)
      : WPE(minW, maxW),
	WPE_corr(minW, maxW),
        WPE_corr_xy(minW, maxW),
        WPE_corr_z(minW, maxW),
	WPE_corr_xy_truth(minW, maxW)
    {
    }

    AttenProfiles& operator+=(const AttenProfiles& rhs)
    {
      WPE += rhs.WPE;
      WPE_corr += rhs.WPE_corr;
      WPE_corr_xy += rhs.WPE_corr_xy;
      WPE_corr_z += rhs.WPE_corr_z;
      WPE_corr_xy_truth += rhs.WPE_corr_xy_truth;

      return *this;
    }
#endif

    LiteProfile WPE, WPE_corr, WPE_corr_xy, WPE_corr_z, WPE_corr_xy_truth;
  };

  /// AttenProfiles for many channels
  class AttenProfilesMap
  {
  public:

#ifdef ATTENPROFILES_DEFAULT_CONSTRUCTORS
    AttenProfilesMap() : fMinW(-1), fMaxW(-1) {}
#endif

#ifndef __GCCXML__
    AttenProfilesMap(float w0, float w1) : fMinW(w0), fMaxW(w1) {}

    AttenProfilesMap(const AttenProfiles&) = delete; // uncopyable

    bool HasProfiles(const geo::OfflineChan& c)
    {
      return fMap.find(c) != fMap.end();
    }

    bool HasProfiles(int plane, int cell)
    {
      return HasProfiles(geo::OfflineChan(plane, cell));
    }

    AttenProfiles& GetProfiles(const geo::OfflineChan& c)
    {
      auto it = fMap.find(c);
      if(it != fMap.end()) return it->second;

      fMap.emplace(std::make_pair(c, AttenProfiles(fMinW, fMaxW)));
      return fMap.find(c)->second;
    }

    AttenProfiles& GetProfiles(int plane, int cell)
    {
      return GetProfiles(geo::OfflineChan(plane, cell));
    }

    /// List of channels with profiles on
    std::vector<geo::OfflineChan> Channels() const
    {
      std::vector<geo::OfflineChan> ret;
      ret.reserve(fMap.size());
      for(auto& it: fMap) ret.push_back(it.first);
      return ret;
    }

    /// Intended for writing out/drawing all the histograms
    std::map<int, std::map<int, std::vector<TH1F*>>> GetAllProfilesByPlaneAndCell()
    {
      std::map<int, std::map<int, std::vector<TH1F*>>> ret;
      for(auto& it: fMap){
        const int plane = it.first.Plane();
        const int cell = it.first.Cell();
        const AttenProfiles& hs = it.second;
        std::vector<TH1F*>& v = ret[plane][cell];
	/*        if(hs.WPE.TotalHits()){
          v.push_back(hs.WPE.ToTH1F());
          v.back()->SetName(TString::Format("h_WPE_%03d_%03d", plane, cell));
          v.back()->SetTitle(TString::Format("Plane %d, Cell %d;W;PE", plane, cell));
        }

        if(hs.WPE_corr.TotalHits()){
          v.push_back(hs.WPE_corr.ToTH1F());
          v.back()->SetName(TString::Format("h_WPE_corr_avg_%03d_%03d", plane, cell));
          v.back()->SetTitle(TString::Format("Avg Corr, Plane %d, Cell %d;W;PE/cm", plane, cell));
	  }*/

        if(hs.WPE_corr_xy.TotalHits()){
          v.push_back(hs.WPE_corr_xy.ToTH1F());
          v.back()->SetName(TString::Format("h_WPE_corr_xy_%03d_%03d", plane, cell));
          v.back()->SetTitle(TString::Format("XY, Plane %d, Cell %d;W;PE/cm", plane, cell));
        }

        if(hs.WPE_corr_z.TotalHits()){
          v.push_back(hs.WPE_corr_z.ToTH1F());
          v.back()->SetName(TString::Format("h_WPE_corr_z_%03d_%03d", plane, cell));
          v.back()->SetTitle(TString::Format("Z, Plane %d, Cell %d;W;PE/cm", plane, cell));
        }
	/*
        if(hs.WPE_corr_xy_truth.TotalHits()){
          v.push_back(hs.WPE_corr_xy_truth.ToTH1F());
          v.back()->SetName(TString::Format("h_WPE_corr_xy_truth_%03d_%03d", plane, cell));
          v.back()->SetTitle(TString::Format("XY, Plane %d, Cell %d;W;MeV/cm", plane, cell));
	  }*/
      }
      return ret;
    }

    AttenProfilesMap& operator+=(const AttenProfilesMap& rhs)
    {
      for(auto& it: rhs.fMap){
	GetProfiles(it.first) += it.second;
      }
      return *this;
    }

    void Clear()
    {
      fMap.clear();
    }

    double MinW() const {return fMinW;}
    double MaxW() const {return fMaxW;}
#endif
  protected:
    float fMinW, fMaxW;

    std::map<geo::OfflineChan, AttenProfiles> fMap;
  };

}

#endif
