#!/bin/env python

import os, sys
import shutil 
import samweb_client, ifdh
import pprint
import argparse
import subprocess
import ROOT
import re
import md5, json
import resource

from samweb_client.utility import fileEnstoreChecksum

#  Stop ROOT from complaining about not having dictionaries, 3000 is below kError but above kWarning. 
ROOT.gErrorIgnoreLevel=3000

detectors = dict()
detectors["fd"]   = "fardet"
detectors["nd"]   = "neardet"
detectors["ndos"] = "ndos"

#Necessary for createMetadata function
_lineregex = re.compile(r'\d+: ([^ \t]+)[ \t]+(.+)')

#For extracting the data stream from the file name
_stream_pattern = re.compile(r"^.*?_r[0-9]+_s[0-9]+_(.*?)[\.,_].*")


#files 
_skip_pattern = re.compile(r"^.*(RootOutput).*\.root", re.IGNORECASE)

def setVMemLimit():
    #memLimit = 3.9*1024**3
    memLimit = 3.9*1000**3
    print "Old virtual memory limit:", resource.getrlimit(resource.RLIMIT_AS)
    print "Limiting the virtual memory of the nova job to 3.9 GB"
    resource.setrlimit(resource.RLIMIT_AS, (memLimit, memLimit))
    print "New virtual memory limit:", resource.getrlimit(resource.RLIMIT_AS)

def unCamelCase(s):
    """ Convert CamelCase to camel_case """
    # \B matches an empty string which is NOT at the beginning of a word
    # so requiring this means no _ will be inserted at the start
    return re.sub( r'\B([A-Z])', r'_\1', s).lower()
                

def createMetadata(inFile):
    filename = os.path.basename(inFile)
    filesize = os.path.getsize(inFile)

    md = {'file_name': filename, 'group': 'nova', 'file_size': filesize, 'file_format': 'artroot'}
    
    try:
        data = subprocess.check_output(["sam_metadata_dumper", str(inFile)])
    except subprocess.CalledProcessError:
        print "sam_metadata_dumper failed!"
        return None

    dumperDict = json.loads(data)

    for dumperKey in dumperDict:
        if str(inFile) in str(dumperKey):
            tmpMD = dumperDict[dumperKey]

            for k,v in tmpMD.iteritems():
                if '.' not in k:
                    k = unCamelCase(k)
                else:
                    k.lower()
                    v = str(v)
                    
                if k == 'parents':
                    vTmp = []
                    for entry in v:
                        if isinstance(entry, int):
                            pass
                        elif isinstance(entry, basestring):
                            entry = os.path.basename(entry)
                        elif isinstance(entry, dict):
                            if "file_name" in entry:
                                entry["file_name"] = os.path.basename( entry["file_name"] )

                        vTmp.append(entry)
                    v = vTmp

                if k == 'process_name' and md.get('application',[]).get('name') is None:
                    k = 'application_name'
                elif k == 'stream_name':
                    k = 'data_stream'
                    
                if k in ('start_date', 'end_date'):
                    v = long(v)
                    
                if k in ('first_event', 'last_event'):
                    try:
                        v = int(v)
                    except TypeError:
                        if len(v) == 3:
                            v = v[2]
                        else:
                            continue
                        
                if k.startswith('application_'):
                    if 'application' not in md:
                        md['application'] = {}
                    md['application'][k[12:]] = v
                elif k in ('file_format_era', 'file_format_version'):
                    pass
                else:
                    md[k] = v
                        
    md['crc'] = fileEnstoreChecksum(inFile)
    
    return md

def getOutDir(pathname, hashDirs=False):
    if "DEST" in os.environ:
        dest = os.environ["DEST"]
    else:
        print "No destination set!"
        exit(1)
    dirs = [dest]
    if hashDirs:
      head, tail = os.path.split(pathname)
      hash = md5.new(tail)
      dirs += list(hash.hexdigest()[:3])
    return os.path.join(*dirs)

def checkAndMoveFiles(inFile, declareFiles, outputs):
    """Checks all root files in the current directories for zombie or recovered status. Bad files are deleted while good files are moved to the results subdirectory for copy out ease"""
    inFileBase = os.path.basename(inFile)
    baseDir = "."
    for root, dirs, filenames in os.walk(baseDir):
        if root == baseDir:
            for file in filenames:
                if file.endswith (".root") and file != inFileBase:
                    fileWPath = os.path.join(root, file)
                    rootFile = ROOT.TFile(fileWPath)
                    if rootFile.IsZombie() or rootFile.TestBit(ROOT.TFile.kRecovered):
                        #do not copy over, just delete 
                        print "File", fileWPath, "is Zombie - remove it"
                        os.remove(fileWPath)
                    elif not file in outputs:
                        #do not copy over, just delete 
                        print "File", fileWPath, " is not among requested outputs, removing"
                        os.remove(fileWPath)                        
                    else:
                        if declareFiles == True:
                            declareFile(fileWPath)
                        newFilePath = os.path.join(root, "results", file)
                        os.renames(fileWPath, newFilePath)
                    rootFile.Close()
    return

def declareFile(fileWPath):
    """Checks file for a TKey of RootFileDB.  If it exists, run sam_metadata_dumper and construct appropriate metadata for the file. Use that metadata to declare the file to SAM"""
    samweb = samweb_client.SAMWebClient(experiment='nova')
    rootFile = ROOT.TFile(fileWPath)
    if rootFile.FindKey("RootFileDB"):
        print "Found key (RootFileDB): constructing metadata for", fileWPath
        md = createMetadata(fileWPath)
        if (md == None):
            print "No metadata found!"
        else:
            pprint.pprint(md)
            print ''
            print "Declaring", fileWPath, "to SAM"
            try:
                samweb.declareFile(md)
                #samweb.validateFileMetadata(md=md)
            except:
                print fileWPath, "already exists in SAM"
    else:
        print fileWPath, "does not contain RootFileDB, do not try to declare"
    rootFile.Close()
    return

def copyOutFiles(hashDirs=False):
    """Builtin facility to copy out art files. This adds in a subdirectories with a single hex digit each corresponding to the first three digits in the hash of the output file name. This splits up files into 4096 separate subdirectories, preventing overfull directories. Copy out does not happen if the file already exists in the output"""
    print "Copying out files"
    dh = ifdh.ifdh("http://samweb.fnal.gov:8480/sam/nova/api")
    baseDir = "./results"
    for root, dirs, filenames in os.walk(baseDir):
        if root == baseDir:
            for file in filenames:
                if file.endswith (".root") and file != inFileBase:
                    fileWPath = os.path.join(root, file)
                    dest = getOutDir(file, hashDirs)
                    skip_match = _skip_pattern.match(file)
                    if skip_match == None:
                        try:
                            print "Checking if", dest + "/" + file, "exists"
                            dh.ls(dest + "/" + file, 1, "")
                        except RuntimeError:
                            print "It doesn't - copy", fileWPath, "to", dest
                            dh.cp(["-D", fileWPath, dest])
                    print "Removing", fileWPath
                    os.remove(fileWPath)
    dh.cleanup()
    return

def getOutputFileName(detector, runNum, subNum, streamEntry, release, subversion, dataFlag, tier, nevt, specialName=None):
  if dataFlag == "sim":
    pattern = re.compile(r"^.*?_(.*?)_\d+.*_v\d+_(.*?)(\.|\_)sim\..+")
    match = pattern.match(inFileBase)
    if match != None:
      simLabel = match.groups()[0]
      uniq     = match.groups()[1]
      outName = '{0}_{1}_{2}_r{3:08d}_s{4:02d}_{5}_v{6}_{7}_{8}.{9}.root'.format(detector, simLabel, nevt, runNum, subNum, release, subversion, uniq, dataFlag, tier)
    else:
      print "regex couldn't find the sim label and the uniqueifier"
      exit(1)
  else:
    outName = '{0}_r{1:08d}_s{2:02d}_{3}_{4}_v{5}_{6}.{7}.root'.format(detector, runNum, subNum, streamEntry, release, subversion, dataFlag, tier)

  if specialName != None:
    oldExtension = "." + tier + ".root"
    newExtension = "_" + specialName + oldExtension
    outName = outName.replace(oldExtension, newExtension)
  
  return outName
  
def resolveFclPath(fcl):
   # Check if we have an absolute path name, return it if so
  if fcl[0] == "/":
    return fcl 

  # Otherwise, we need to do some searching 
  fclPaths = os.environ["FHICL_FILE_PATH"].split(":")
  for path in fclPaths:
    if os.path.isfile(path + fcl):
      return path + fcl
      
  # If we haven't found it, we have a problem.
  raise IOError(sys.argv[0] + ": config  file "+ fcl+" not found in FHICL_FILE_PATH")
    

if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Run the nova command using SAM metadata')
    parser.add_argument('inFile', help='The input file to run over', type=str)
    parser.add_argument('--config', '-c', help='FHiCL file to use as configuration for nova executable', type=str)
    parser.add_argument('--outTier', help='Data tier of the output file, multiple allowed, formatted as <name_in_fcl_outputs>:<data_tier>', type=str, action='append')
    parser.add_argument('--cafTier', help='Module label for CAF output, multiple allowed.  Format as <cafmaker_module_label>:<data_tier>', type=str, action='append')
    parser.add_argument('--histTier', help='File identifier string for TFileService output, only one allowed.  Supply as --histTier <id> for output_name.<id>.root, where output_name is assembled based on the input file.', type=str)
    parser.add_argument('--specialName', help='Additional name to add before data tier in output.', type=str)
    parser.add_argument('-n', help='Number of events to run over', type=int)
    parser.add_argument('--copyOut', help='Use the built in copy out mechanism', action='store_true')
    parser.add_argument('--hashDirs', help='Use hash directory structure in destination directory.', action='store_true')
    parser.add_argument('--declareFiles', help='Declare files with metadata on worker node', action='store_true')
    parser.add_argument('--noCleanup', help='Skip working directory cleanup step, good for interactive debugging', action='store_true')
    parser.add_argument('--gdb', help='Run nova executable under gdb, print full stack trace, then quit gdb.', action='store_true')
    
    args = parser.parse_args()

    setVMemLimit()

    inFile = args.inFile
    
    samweb = samweb_client.SAMWebClient(experiment='nova')

    inFileBase = os.path.basename(inFile)

    metadata = samweb.getMetadata(inFileBase)

    #make all lower case keys
    metadata = {k.lower():v for k,v in metadata.items()}

    print "Input File = ", inFile, "with metadata:"
    pprint.pprint(metadata)
    print ''

    #if new format "runs" metadata, use that, otherwise, get from file name
    if "runs" in metadata and len(metadata["runs"][0]) == 3:
        runNum = int(metadata["runs"][0][0])
        subNum = int(metadata["runs"][0][1])
    else:
        run_pattern = re.compile(r"^.*?_r([0-9]+)_s([0-9]+).+")
        run_match = run_pattern.match(inFileBase)
        if run_match != None:
            runNum = int(run_match.groups()[0])
            subNum = int(run_match.groups()[1])
        else:
            print "No run number/subrun number found!"
            exit(1)
            
    print "Run Number:    ", runNum
    print "Subrun Number: ", subNum
    
    #get data_stream from metadata.  If it is "all", try extracting from the file name.
    if "data_stream" in metadata:
        tmpStream = metadata["data_stream"]
    elif "online.stream" in metadata:
        tmpStream = metadata["online.stream"]


    if tmpStream == 'all':
        stream_match = _stream_pattern.match(inFileBase)
        if stream_match != None:
            tmpStream = stream_match.groups()[0]
        if tmpStream[0] == 't':
            tmpStream = tmpStream[1:]
            
    try:
        stream = int(tmpStream)
        streamEntry = 't{0:02d}'.format(stream)
    except ValueError:
        stream = tmpStream
        streamEntry = stream
    except:
        print "No data stream found!"
        exit(1)

    if "online.detector" in metadata:
        detector = metadata["online.detector"].lower()
    elif "nova.detectorid" in metadata:
        detId = metadata["nova.detectorid"].lower()
        if detId in detectors.keys():
            detector = detectors[detId] 
        else:
            print "No detector name found for detectorid:", detId
            exit(1)
    else:
        print "No detector name found!"
        exit(1)
    
    if "SRT_BASE_RELEASE" in os.environ:
        release = os.environ["SRT_BASE_RELEASE"]
    else:
        print "No release set!"
        exit(1)

    if "file_type" in metadata:
        file_type = metadata["file_type"]
    else:
        print "No file_type set!"
        exit(1)

    if file_type == "importedDetector":
        dataFlag = "data"
    elif file_type == "importedSimulated":
        dataFlag = "sim"
    else:
        print "Unrecognized file_type:", file_type
        exit(1)

    if "NPASS" in os.environ:
        subversion = os.environ["NPASS"]
    elif "nova.subversion" in metadata:
        subversion = metadata["nova.subversion"]
    else:
        subversion = 1

    nevt = None
    if "simulated.number_of_spills" in metadata:
        nevt = metadata["simulated.number_of_spills"]

    fclFile = args.config

    # Get a copy of the fcl file and bring it here
    # First step, find the fcl
    fclPath = resolveFclPath(fclFile)
    print "Found fcl file here: ", fclPath
    # Next step, make a temporary name for the fcl in the local directory and copy it here
    tmpFclName = os.path.basename(fclPath).replace(".fcl", "_" + os.path.splitext(inFileBase)[0] + ".fcl") 
    print "Creating local copy : ", tmpFclName
    shutil.copy(fclPath, tmpFclName)

    # Open the fcl file so that we can append output filenames to it 
    fclFileObj = open(tmpFclName, 'a')

    doMeta = True
    if not (args.outTier or args.cafTier):
        doMeta = False
    
    # Start setting up the nova command, add SAM parameters 
    cmdList = []
    cmdList.append('nova')
    cmdList.append('-c')
    cmdList.append(tmpFclName)

    if doMeta:
        cmdList.append('--sam-application-family=nova')
        cmdList.append('--sam-application-version=' + release)
        cmdList.append('--sam-file-type=' + file_type)

    if not args.outTier:
      args.outTier = []
    if not args.cafTier:
      args.cafTier = []

    
    outList = [] # list of files to be supplied with -o 
    outputs = [] # list of files that will be moved to results directory, includes CAFs
    # Loop over output tiers 
    for outTier in args.outTier:
        try:
            output = outTier.split(":")[0]
            tier   = outTier.split(":")[1]
        except:
            raise ValueError("Output data tier: " + outTier + "not formatted corectly, should be <output_name>:<data_tier>")
        cmdList.append('--sam-data-tier=' + outTier)
        cmdList.append('--sam-stream-name=' +output + ':' + str(stream))
        outName = getOutputFileName(detector, runNum, subNum, streamEntry, release, subversion, dataFlag, tier, nevt, specialName=args.specialName)
        outList.append(outName)
        outputs.append(outName)
        print "Output file name: ", outName, " for tier ", tier, " and output ", output  
        fclFileObj.write("\noutputs." + output + '.fileName: "'+ outName + '"\n')

    for cafTier in args.cafTier:
      try:
        cafLabel = cafTier.split(":")[0]
        tier   = cafTier.split(":")[1]
      except:
        raise ValueError("Output data tier: " + outTier + "not formatted corectly, should be <output_name>:<data_tier>")
      cafName = getOutputFileName(detector, runNum, subNum, streamEntry, release, subversion, dataFlag, tier, nevt, specialName=args.specialName)
      print "Adding CAF: ", cafLabel, tier, cafName
      fclFileObj.write("\nphysics.producers." + cafLabel + '.CAFFilename: "' + cafName + '" \n')
      fclFileObj.write("physics.producers." + cafLabel + '.DataTier: "' + tier + '" \n')
      outputs.append(cafName)
    
    if args.histTier:
      try:
        tier = str(args.histTier)
      except:
        raise Exception("Histogram identifier supplied by --histTier could not be converted to a string.")
      histName = getOutputFileName(detector, runNum, subNum, streamEntry, release, subversion, dataFlag, tier, nevt, specialName=args.specialName)
      outputs.append(histName)
      fclFileObj.write("\nservices.TFileService.fileName: " + '"' + histName + '"\n')
      
    fclFileObj.close()

    print "Config: "
    for line in open(tmpFclName):
        print line.strip()

    if args.n != None:
        cmdList.append('-n')
        cmdList.append(str(args.n))
    cmdList.append(inFile)


    if args.gdb:
      gdbArgs = ["gdb", "-return-child-result", "--ex", "run", "--ex", "bt", "full", "--ex", "q", "--args"]
      cmdList = gdbArgs + cmdList

    cmd =  ' '.join(cmdList)

    print 'Running:', cmd
    sys.stdout.flush() # flush the stdout buffer before running the nova executable, cleans up output. 
    retCode = subprocess.call(cmdList)

    if retCode == 0:
        if args.declareFiles == True:
            checkAndMoveFiles(inFile, True, outputs)
        else:
            checkAndMoveFiles(inFile, False, outputs)

        if args.copyOut == True:
            copyOutFiles(args.hashDirs)
    else:
        #job didn't succeed, remove output files if they exist
        for file in outList:
            try:
                os.remove("./" + file)
            except OSError:
                pass

    #clean up section
    dirList = os.listdir(".")
    for file in dirList:
        skip_match = _skip_pattern.match(file)
        if skip_match != None:
            print file, "contains RootOutput*.root: clean up"
            os.remove("./" + file)
    if not args.noCleanup:
      os.remove(tmpFclName)
      os.remove(inFile)
    dh = ifdh.ifdh("http://samweb.fnal.gov:8480/sam/nova/api")
    dh.cleanup()
    
    exit(retCode)
