////////////////////////////////////////////////////////////////////////
// Class:       AirShower
// Module Type: analyzer
// File:        AirKalmanAna_module.cc
// author:      Mehreen Sultana ( sultana@fnal.gov )
// Description: Analyzing Cosmic Ray Muon Showers to get directionality
//              and multiplicity 
// 
////////////////////////////////////////////////////////////////////////

#include <boost/format.hpp>
#include <math.h>

#include <string>
#include <vector>

// ROOT includes
#include "TH1D.h"
#include "TH2D.h"
#include "TVector3.h"
#include "TTree.h"
#include <TH1.h>
#include "TH1F.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/FindManyP.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/RecoHit.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/FilterList.h"
#include "CMap/CMap.h"


#include <memory>
#include <boost/format.hpp>
#include <fstream>
#include <vector>


////////////////////////////////////// AIR NAMESPACE ///////////////////////////////////////////

namespace air {
  class AirKalmanAna; 
}

class air::AirKalmanAna : public art::EDAnalyzer {
public:
  explicit AirKalmanAna(fhicl::ParameterSet const & p);
  virtual ~AirKalmanAna();

  void beginJob();
  void analyze(art::Event const & e) override;

private:
  std::string hit_label_, slice_label_, track_label_;
  
  std::map<std::string, TH1D*> h_;


  /////////////////////////////////// DEFINING HISTOS ////////////////////////////////////////

  //Individual Track Level Histograms

  TH1D* fCosY;    // Cosine theta of 3D tracks with respect ot the Y axis
  TH1D* fthetaY;  // Theta Y of each track in degrees

  
  TH1D* fCosX;    // Cosine theta of 3D tracks with respect ot the X axis
  TH1D* fthetaX;  // Theta X of each track in degrees


  TH1D* fCosZ;    // Cosine theta of 3D tracks with respect ot the Z axis
  TH1D* fthetaZ;  // Theta Z of each track in degrees
  
  TH1F* TotalADC; // Total ADC Of Track
  
  TH1F* AvgADC;   // Average ADC of Track Track.TotalADC/Track.NCell

  TH1F* TotalLength; //Total Track Length

  TH1F* NCells; //Number of Cells Per Track

  // Slice Level Histograms

  TH1F* STotalADC;  // Total ADC of Slice
  
  TH1F* SAvgADC; // Average ADC of Slice


  // Event Level Histograms
  TH1D* f3DTracksPerEvent;   // number of 3d tracks found per event - 
                             // NEED TO RECONSTRUCT TO ONLY USE MUONS. HOW TO DO THAT? 


  TH1D* f3DTracksPerSlice;   // 3D tracks per slice - For Debugging/Improvement 
                             // How different is it for most events from 3D tracks per event? 


  TH1D* fTracksPerSlice;     // Total reconstructed tracks within a slice
  TH1D* fTracksPerEvent;     // Total reconstructed tracks within an event
  
  

  ////////////////////////////////  DEFINING VARIABLES /////////////////////////////////////////

  unsigned HighEnergyThresh; // High Energy ADC Threshold defined in the trigger - unsure of
  // what it is right now.

  unsigned ntracksinslice; // all tracks reconstructed in slice
  unsigned ntracksinevent;
  
  unsigned n3Dtracksinslice; // only 3D tracks reconstructed in slice

  unsigned n3Dtracksinevent; // Total 3D tracks in an event 


  // Computational Variables; 

  double cosY; // cosY = (end - start)/total_length // end and start of only Y components 
  double thetaY; // theta relative to the Y axis only
  double cosX; // (end - start)/total_length // end and start of only X components 
  double thetaX; // theta relative to the X axis only
  double cosZ; // (end - start)/total_length // end and start of only Z components 
  double thetaZ; // theta relative to the X axis only

  double savgadc; //Average ADC per slice
  double avgadc; //Average ADC per track
  double totallength; //totallength of track
  double totaladc; // total adc of track
  unsigned ncells;

    

};

air::AirKalmanAna::AirKalmanAna(fhicl::ParameterSet const & p) : 
  EDAnalyzer(p),
  hit_label_(p.get<std::string>("hit_label")),
  slice_label_(p.get<std::string>("slice_label")),
  track_label_(p.get<std::string>("track_label"))
{
}



air::AirKalmanAna::~AirKalmanAna()
{
}



void air::AirKalmanAna::beginJob()
{
  art::ServiceHandle<art::TFileService> tfs;

  // Individual Track Level Histograms

  fCosY                        = tfs->make<TH1D>("cosY","Cos #Theta_{Y} Multiplicity;Cos #theta_{Y};N 3DTracks / 0.025 CosY", 100,-1.5,1.0); // 
  fthetaY                      = tfs->make<TH1D>("thetaY","#Theta_{Y} Multiplicity;#theta_{Y};N 3DTracks / 0.50 #theat_{Y}", 361,-0.5, 180);

  fCosX                        = tfs->make<TH1D>("cosX","Cos #Theta_{X} Multiplicity;Cos #theta_{X};N 3DTracks / 0.025 CosX", 100,-1.5,1.0); // 
  fthetaX                      = tfs->make<TH1D>("thetaX","#Theta_{X} Multiplicity;#theta_{X};N 3DTracks / 0.50 #theat_{X}", 361,-0.5, 180);

  fCosZ                        = tfs->make<TH1D>("cosZ","Cos #Theta_{Z} Multiplicity;Cos #theta_{Z};N 3DTracks / 0.025 CosZ", 100,-1.5,1.0); // 
  fthetaZ                      = tfs->make<TH1D>("thetaZ","#Theta_{Z} Multiplicity;#theta_{Z};N 3DTracks / 0.50 #theat_{Z}", 361,-0.5, 180);

  TotalADC                     = tfs->make<TH1F>("TotalADC","TotalADC Per 3D Track;TotalADC;N 3DTracks / 100 TOTALADC", 1000,0,100000); //total ADC

  AvgADC                       = tfs->make<TH1F>("AverageADC","AverageADC Per 3D Track;Average ADC;N 3DTracks / 20 Average ADC ", 250,0,5000);
  // Average ADC [TotalADC/NCells/Slice]

  NCells                       = tfs->make<TH1F>("NCells","N Cells Per 3D Track; N Cells; N 3DTracks / 20 N Cells", 200, 0, 2000); 

  TotalLength                  = tfs->make<TH1F>("TotalLength","Total Length (cm) of 3D Tracks; Length (cm); N 3D Tracks / 10 cm", 300, 0, 3000);

  // Event Level Histograms

  // Need to fix/Check it to really give tracks per event and not tracks per slice. 

  fTracksPerEvent              = tfs->make<TH1D>("TracksPerEvent","Tracks Per Event;3D Tracks Per Event; NEvents",101,-0.5,200.5); 
  fTracksPerSlice              = tfs->make<TH1D>("TracksPerSlice","Tracks Per Slice;3D Tracks Per Slice; NSlices",101,-0.5,200.5);

  f3DTracksPerEvent            = tfs->make<TH1D>("3DTracksPerEvent","3D Tracks Per Event;3D Tracks Per Event; NEvents",101,-0.5,200.5);
  f3DTracksPerSlice            = tfs->make<TH1D>("3DTracksPerSlice","3D Tracks Per Slice;3D Tracks Per Slice; NSlices",101,-0.5,100.5);

  //Slice Level Histograms

  STotalADC                     = tfs->make<TH1F>("STotalADC","Total ADC Per Slice; Total ADC Per Slice; N Slices / 1000 Total ADC", 1000,0,100000); //total ADC
  SAvgADC                       = tfs->make<TH1F>("SAverageADC","Average ADC per Slice; Average ADC Per Slice; N Slice / 10 Average ADC ", 100,0,1000); // Average ADC [TotalADC/NCells/Slice]


}


void air::AirKalmanAna::analyze(art::Event const & e)
{
  // Get Cell Hits - although they aren't explicitly used yet for analysis
  // I want to use the hits and clusters associated with HoughResults/HoughLines as input for Kalman Reconstruction
  // Need to figure out associations 

  art::Handle<std::vector<rb::CellHit> > hits;
  e.getByLabel(hit_label_, hits);

  // Get Slices 

  art::Handle< std::vector<rb::Cluster> > slicecol;	
  e.getByLabel(slice_label_,slicecol);
    
    
  // Make Slice List
  art::PtrVector<rb::Cluster> slicelist;	
  for(unsigned int i = 0; i<slicecol->size();++i)
  {	
    art::Ptr<rb::Cluster>slice(slicecol,i);
    slicelist.push_back(slice);
  }

  // setup the associations between tracks and slices	
  art::FindManyP<rb::Track> fmtrack(slicecol,e,track_label_);
	

  // initializing track counts
  
  n3Dtracksinevent = 0;
  ntracksinevent = 0;


  // Looping over slices;


  for (unsigned int islice = 0; islice < slicelist.size(); ++islice)
  {
    
    std::vector<art::Ptr<rb::Track> > tracks;	
    if(fmtrack.isValid())
    {	
      tracks = fmtrack.at(islice);	
    }
    
    if(slicelist[islice]->IsNoise()){ continue; }
    
    n3Dtracksinslice = 0;

    if (slicelist[islice]->TotalADC() > HighEnergyThresh)
    {
      savgadc = (slicelist[islice]->TotalADC())/(slicelist[islice]->NCell());
      STotalADC->Fill(slicelist[islice]->TotalADC());
      SAvgADC->Fill(savgadc);
      
      for (unsigned int i=0;i<tracks.size();i++)
      {
	art::Ptr<rb::Track> track = tracks[i];
	
	// Interating over ALL tracks in slice - to double check counts/debug/reco
	
	++ntracksinslice;
	++ntracksinevent;

	if (track->Is3D())
	{
	  
	  // iterate 3D tracks
	  ++n3Dtracksinslice;
	  ++n3Dtracksinevent;
	  
	  // DEFINE TRACK INFO - need to see if I can choose where to start and stop on a track - what is allowed to go in the ()?? 
	  totaladc = (track->TotalADC());
	  ncells = track->NCell();
	  avgadc = totaladc/ncells;
	  
	  TVector3 start(track->TrajectoryPoint(1));
	  TVector3 end(track->Stop());
	  TVector3 length = end-start;
	  totallength = length.Mag();

	  // Y - DIRECTIONALITY
	  
	  cosY = (end.Y() - start.Y())/totallength;
	  thetaY = (M_PI - acos(cosY))*180/M_PI;
	  
	  // X - DIRECTIONALITY
	  
	  cosX = (end.X() - start.X())/totallength;
	  thetaX = (M_PI - acos(cosX))*180/M_PI;

	  // Z - DIRECTIONALITY
	  
	  cosZ = (end.Z() - start.Z())/totallength;
	  thetaZ = (M_PI - acos(cosZ))*180/M_PI;
	  
	  // OVERALL INDIVIDUAL TRACK HISTOFILL
	  
	  TotalADC->Fill(totaladc);
	  AvgADC->Fill(avgadc);
	  TotalLength->Fill(totallength);
	  NCells->Fill(ncells);
	  
	  // FILL HISTOGRAMS FOR Y 

	  fthetaY->Fill(thetaY);
	  fCosY->Fill(cosY);
      
	  // FILL HISTOGRAMS FOR X
      
	  fthetaX->Fill(thetaX);
	  fCosX->Fill(cosX);
      
	  // FILL HISTOGRAMS FOR Z

	  fthetaZ->Fill(thetaZ);
	  fCosZ->Fill(cosZ);
	  std::cout << thetaZ << std::endl;
			 

	} // end Is3D loop    
      } // end For (Tracks) loop

      f3DTracksPerSlice->Fill(n3Dtracksinslice);
      fTracksPerSlice->Fill(ntracksinslice);
    } // end Slice Threshold Loop


  } // End Slice loop
 
  f3DTracksPerEvent->Fill(n3Dtracksinevent);
  fTracksPerEvent->Fill(ntracksinevent);

} // end analyze event



DEFINE_ART_MODULE(air::AirKalmanAna)
