////////////////////////////////////////////////////////////////////////
/// \brief   This module reads a text file and then stores BNB spill info
/// \author  Chad Johnson
/// \date
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <fstream>
#include <map>
#include <set>
#include <dirent.h>
#include <iostream>
#include <sys/types.h>
#include <vector>

// ROOT includes
#include "TH1.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "SummaryData/SpillData.h"
#include "NovaDatabase/Table.h"


/// Information about the Booster beam
namespace bnb {

  class BNBInfo : public art::EDProducer {
  public:
    explicit BNBInfo(fhicl::ParameterSet const& pset); // Required! explicit tag tells the compiler this is not a copy constructor      
    virtual ~BNBInfo();                        
      
    // Optional, read/write access to event
    void produce(art::Event& evt);

    // Optional if you want to be able to configure from event display, for example
    /* void reconfigure(const fhicl::ParameterSet& pset); */
      
    void beginSubRun(art::SubRun &subrun);
    void readFromDB(unsigned long int start_sec,unsigned long int end_sec);
    void readBNBAllFiles();
    int  readBNBFileRange(unsigned long int sec, unsigned long int nsec);
    int  readBNBDir();
    int  readFromDB();
    bool fGotGoodTable;

  private:
    unsigned int            readfileerror;
    std::string             fMBFileDirectory;
    std::set<std::string>   fMBFiles;
    std::string             fMBSpillsTable;
    std::string             fTextFileName; ///> file name for input text file containing BNB info
    unsigned int            fSubRunLength;
    std::ifstream           fTextFile;
    unsigned int            fTimeWindow; //number of nsec allowed between bnb time and event time
    nova::database::Table*  fTable;

    std::map<unsigned long long int,std::pair<unsigned long long int,double> > fBNBSpillTimesMap;
    std::multimap<unsigned long int,std::pair<unsigned long int,std::pair<unsigned long long int,double> > > fBNBSpillTimesMap2;
  };
}

////////////////////////////////////////////////////////////////////////
namespace bnb
{
  //................................................................
  BNBInfo::BNBInfo(fhicl::ParameterSet const& pset) :
    fTable(0)
  {
    // tell the module what it is making as this is a EDProducer
    produces< sumdata::SpillData >();
 
    fTextFileName     = pset.get< std::string >("TextFileName");
    fTimeWindow       = pset.get< int>("TimeWindow");
    fMBFileDirectory  = pset.get< std::string>("MBFileDirectory");
    fSubRunLength     = pset.get< int>("SubRunLength");

    readfileerror = 0;
  }

  //......................................................................
  BNBInfo::~BNBInfo()
  {
  }

  //......................................................................
  // void BNBInfo::reconfigure(const fhicl::ParameterSet& pset)
  // {
  // }

  //----------------------------------------------------------
   void BNBInfo::beginSubRun(art::SubRun &subrun)
   {
    // insert check hear to see if we need to load more MB spills
    mf::LogInfo("BNBInfo") << " we are in beginSubRun " << std::endl;
     
    // get the time stamp.
    art::Timestamp ts_begin = subrun.beginTime();
    art::Timestamp ts_end = subrun.endTime();
    mf::LogInfo("BNBInfo") << " begin time: " << ts_begin.value() << std::endl;
    mf::LogInfo("BNBInfo") << " end time: " << ts_end.value() << std::endl;
    this->readFromDB(ts_begin.timeHigh()-1,ts_begin.timeHigh()+fSubRunLength);
     return;
   }

  //----------------------------------------------------------
  void BNBInfo::readFromDB(unsigned long int start_sec, unsigned long int end_sec) 
  {
    std::pair<std::map<unsigned long long int,std::pair<unsigned long long int,double> >::iterator,bool> ret;
    std::multimap<unsigned long int,std::pair<unsigned long int,std::pair<unsigned long long int,double> > >::iterator ret2;
    unsigned long int ubnb_sec,ubnb_nsec;
    unsigned long long int ubnb;
    //signed long int sgps_sec,sgps_nsec;
    unsigned long int ugps_sec,ugps_nsec;
    unsigned long long int ugps;
    double pot;
    //double curr;

    char start_subrun_sec[100],end_subrun_sec[100];

    fGotGoodTable= true; 

    try{
      delete fTable;
      fTable = new nova::database::Table(std::string("/BNBInfo/MBSpills.xml"));
      fTable->SetDetector("NDOS");
//       fTable->Clear();
      sprintf(start_subrun_sec,"%lu",start_sec);
      sprintf(end_subrun_sec,"%lu",end_sec);
      mf::LogError("BNBInfo") << "DB start: " << start_sec << " DB end: " << end_sec;
      fTable->SetValidityRange("det_sec",start_sec,end_sec);
      fTable->LoadFromDB();

      if(fTable->NRow()==0){
	fGotGoodTable=false;
	mf::LogError("BNBInfo") << " No MB booster spills available for this run/subrun " << std::endl;
      }
      else{
	mf::LogError("BNBInfo") << " loading MB booster spills for this run " << std::endl;
	fBNBSpillTimesMap.clear();
	for(int i=0;i<fTable->NRow();i++){
	  bool ok = fTable->GetRow(i)->Col("det_sec")->Get(ubnb_sec);
	  ok = ok && fTable->GetRow(i)->Col("det_nsec")->Get(ubnb_nsec);
	  ok = ok && fTable->GetRow(i)->Col("pot")->Get(pot);
	  ok = ok && fTable->GetRow(i)->Col("gps_sec")->Get(ugps_sec);
	  ok = ok && fTable->GetRow(i)->Col("gps_nsec")->Get(ugps_nsec);
          if(!ok) throw;

	  ubnb = (unsigned long long int)(static_cast<unsigned long long int>(ubnb_sec) << 32 | ubnb_nsec);
	  ugps = (unsigned long long int)(static_cast<unsigned long long int>(ugps_sec) << 32 | ugps_nsec);
	  ret = fBNBSpillTimesMap.insert(std::pair<unsigned long long int,std::pair<unsigned long long int,double> >(ubnb,std::pair<unsigned long long int,double>(ugps,pot)));
	  ret2 = fBNBSpillTimesMap2.insert(std::pair<unsigned long int,std::pair<unsigned long int,std::pair<unsigned long long int,double> > >(ubnb_sec,std::pair<unsigned long int,std::pair<unsigned long long int,double> >(ubnb_nsec,std::pair<unsigned long long int,double>(ugps,pot))));
	}
	mf::LogError("BNBInfo") << " n rows: " << fTable->NRow() << std::endl;
      }
    }
    catch (...) {
      fGotGoodTable=false;
    }
  }

  //----------------------------------------------------------
  int BNBInfo::readBNBDir()
  {
    DIR *dp;
    struct dirent *dirp;
    std::pair<std::set<std::string>::iterator,bool> ret;
    std::string fname,sdirp;
    if((dp  = opendir(fMBFileDirectory.c_str())) == NULL) {
      mf::LogError("BNBInfo") << "Error opening " << fMBFileDirectory << std::endl;
      return 0;
    }
    mf::LogInfo("BNBInfo") << "opening " << fMBFileDirectory << std::endl;
    
    while ((dirp = readdir(dp)) != NULL) {
      sdirp = std::string(dirp->d_name);
      if(sdirp.size()>4) {
	if(sdirp.compare(sdirp.size()-4,4,".dat") == 0) {
	  fname = fMBFileDirectory + std::string("/");
	  fname += std::string(dirp->d_name);
	  ret = fMBFiles.insert(fname);
	  if(!ret.second) return 0;
	}
      }
    }
    closedir(dp);
    return 1;
  }
    
  int BNBInfo::readBNBFileRange(unsigned long int sec, unsigned long int nsec)
  {
    std::set<std::string>::iterator it;
    if(!readBNBDir())
      return 0;

    for(it = fMBFiles.begin();it!=fMBFiles.end();it++) {
      
    }
    return 1;
  }

  void BNBInfo::readBNBAllFiles()
  {
    std::pair<std::map<unsigned long long int,std::pair<unsigned long long int,double> >::iterator,bool> ret;
    std::ifstream tempfile;
    std::string tempfilename,tempbuf;

    unsigned long int ubnb_sec,ubnb_nsec;
    unsigned long long int ubnb;

    //signed long int sgps_sec,sgps_nsec;
    unsigned long int ugps_sec,ugps_nsec;
    unsigned long long int ugps;

    double pot,curr;
    for(std::set<std::string>::iterator it = fMBFiles.begin();it!=fMBFiles.end();it++) {
    // fTextFile.open(fTextFileName.c_str());
    // mf::LogInfo("BNBInfo") << "opened text file, " << fTextFileName << "\n";
    // while(!fTextFile.eof()) {
    //   getline(fTextFile,tempfilename);

      tempfile.open((*it).c_str());
      mf::LogInfo("BNBInfo") << "opened text file, " << (*it).c_str() << "\n";
      if(tempfile.bad()) {mf::LogError("BNBInfo") << tempfilename << "file is bad!" << "\n";}
      while(!tempfile.eof()) {
	getline(tempfile,tempbuf);
	sscanf(tempbuf.c_str(),"%lu %lu %lf %lf %lu %lu",&ubnb_sec,&ubnb_nsec,&pot,&curr,&ugps_sec,&ugps_nsec);
	ubnb = (unsigned long long int)(static_cast<unsigned long long int>(ubnb_sec) << 32 | ubnb_nsec);
	// 	 ugps_sec = (unsigned long int) sgps_sec;
	// 	 ugps_nsec = (unsigned long int) sgps_nsec;
	ugps = (unsigned long long int)(static_cast<unsigned long long int>(ugps_sec) << 32 | ugps_nsec);

	ret = fBNBSpillTimesMap.insert(std::pair<unsigned long long int,std::pair<unsigned long long int,double> >(ubnb,std::pair<unsigned long long int,double>(ugps,pot)));
	//for some reason it tries to store the last value twice. if there is no final newline in the text file then it doesn't complain
	//  	 if(!ret.second){
	//  	   mf::LogError("BNBInfo") << "----we found another spill at the same spill time!----"<< "\n";
	// 	      	   mf::LogError("BNBInfo") << ret.first->first << " " << ret.first->second.second  << " " << ubnb_sec << " " << ubnb_nsec << " " << ubnb << " " << pot << "\n";
	// 	    	   mf::LogError("BNBInfo") << ugps << " " << ugps_sec << " " << ugps_nsec << "\n";
	//  	 }
      }
      tempfile.close();
    }
    fTextFile.close();
  }
  
  //......................................................................
  void BNBInfo::produce(art::Event& evt)
  {

    if(!fGotGoodTable) return;
    
    std::multimap<unsigned long int,std::pair<unsigned long int,std::pair<unsigned long long int,double> > >::iterator ret_lo,ret_up;

    // get the time stamp.
    art::Timestamp ts = evt.time();
    unsigned long int uevt_sec = ts.timeHigh(); //utval >> 32 & 0xFFFFFFFFF;
    unsigned long int uevt_nsec = ts.timeLow(); //utval & 0xFFFFFFFFF;
    unsigned long long int uabs_dt=0,uabs_dt_last=0;
    signed long long int sdt=0;

    unsigned long int ubnb_sec,ubnb_nsec;
    //    unsigned long int ubnb_sec_lo_temp,ubnb_nsec_lo_temp,ubnb_sec_up_temp,ubnb_nsec_up_temp;
    //    unsigned long long int uabs_dt_lo_temp=0,uabs_dt_up_temp=0;
    double bnb_pot;

    //signed long int sgps_sec,sgps_nsec;
    unsigned long int ugps_sec,ugps_nsec;

    int ncorr=0;
    int nspills=0;
    
    mf::LogInfo("BNBTiming") << "evt sec: " << uevt_sec << " evt nsec: " << uevt_nsec << "\n";

    std::unique_ptr<sumdata::SpillData> spill(new sumdata::SpillData);

    ret_lo = fBNBSpillTimesMap2.lower_bound(uevt_sec-1);
    ret_up = fBNBSpillTimesMap2.upper_bound(uevt_sec);
    mf::LogInfo("BNBTiming") << "sec lo: " << ret_lo->first << " sec up: " << ret_up->first << " nsec lo: " << ret_lo->second.first << " nsec up: " << ret_up->second.first << "\n";

    if(ret_lo == ret_up) {
      ubnb_sec = ret_lo->first;
      ubnb_nsec = ret_lo->second.first;
      ugps_sec = ret_lo->second.second.first >> 32 & 0xFFFFFFFF;
      ugps_nsec = ret_lo->second.second.first & 0xFFFFFFFF;
      sdt = (signed long long int)(((ubnb_sec - uevt_sec)*1000000000) + (ubnb_nsec - uevt_nsec));
      bnb_pot = ret_lo->second.second.second;
      uabs_dt = (unsigned long long int)abs(sdt);

      if(uabs_dt<fTimeWindow) {
	spill->spilltimesec = ubnb_sec;
	spill->spilltimensec = ubnb_nsec;
	spill->gpsspilltimesec = ugps_sec;
	spill->gpsspilltimensec = ugps_nsec;
	spill->spillpot = bnb_pot;
	spill->deltaspilltimensec = sdt;
      }
    }
    else {
      for(std::multimap<unsigned long int,std::pair<unsigned long int,std::pair<unsigned long long int,double> > >::iterator it = ret_lo; it != ret_up; it++) {
	ubnb_sec = it->first;
	ubnb_nsec = it->second.first;
	ugps_sec = it->second.second.first >> 32 & 0xFFFFFFFF;
	ugps_nsec = it->second.second.first & 0xFFFFFFFF;
	sdt = (signed long long int)(((ubnb_sec - uevt_sec)*1000000000) + (ubnb_nsec - uevt_nsec));
	bnb_pot = it->second.second.second;
	uabs_dt = (unsigned long long int)abs(sdt);
	// mf::LogInfo("BNBTiming") << "sdt: " << sdt << " ubnb_sec: " << ubnb_sec << " uevt_sec: " << uevt_sec << " ubnb_nsec: " << ubnb_nsec << " uevt_nsec: " << uevt_nsec << "\n";

	if((uabs_dt < uabs_dt_last)||(ncorr==0)) {
	  if(uabs_dt<fTimeWindow) {
	    spill->spilltimesec = ubnb_sec;
	    spill->spilltimensec = ubnb_nsec;
	    spill->gpsspilltimesec = ugps_sec;
	    spill->gpsspilltimensec = ugps_nsec;
	    spill->spillpot = it->second.second.second;
	    spill->deltaspilltimensec = sdt;
	    nspills++;
	  }
	  uabs_dt_last = uabs_dt;
	  ncorr++;
	}
      }
    }

    if(spill->deltaspilltimensec != -9999999) {
      mf::LogInfo("BNBTiming") << "evt sec: " << uevt_sec << " evt nsec: " << uevt_nsec << " final dt was: " << spill->deltaspilltimensec << " after ncorr: " << ncorr << "\n"; }
    else
      mf::LogInfo("BNBTiming") << "nothing found" << "\n";

    if(nspills > 1)  { 
      mf::LogError("BNBInfo") << "-----found " << nspills << " spills within the time window corresponding to event run: " << evt.run() << " and subRun: " << evt.subRun() << "------\n";
    } //error, found more than one corresponding BNB spill!
    evt.put(std::move(spill));
    return;
  }
} //namespace bnbinfo


//   //......................................................................
//   void BNBInfo::produce(art::Event& evt)
//   {

//     if(!fGotGoodTable) return;

//     std::map<unsigned long long int,std::pair<unsigned long long int,double> >::iterator ret_up,ret_lo;

//     // get the time stamp.
//     art::Timestamp ts = evt.time();
//     unsigned long int uevt_sec = ts.timeHigh(); //utval >> 32 & 0xFFFFFFFFF;
//     unsigned long int uevt_nsec = ts.timeLow(); //utval & 0xFFFFFFFFF;
//     unsigned long long int uevt_bound_lo;
//     unsigned long long int uevt_bound_hi;
//     unsigned long long int uabs_dt=0,uabs_dt_last=0;
//     signed long long int sdt=0;

//     unsigned long int ubnb_sec,ubnb_nsec;
//     unsigned long int ubnb_sec_lo_temp,ubnb_nsec_lo_temp,ubnb_sec_up_temp,ubnb_nsec_up_temp;
//     unsigned long long int uabs_dt_lo_temp=0,uabs_dt_up_temp=0;

//     signed long int sgps_sec,sgps_nsec;
//     unsigned long int ugps_sec,ugps_nsec;

//     int ncorr=0;
    
//     mf::LogInfo("BNBTiming") << "evt sec: " << uevt_sec << " evt nsec: " << uevt_nsec << "\n";

//     std::unique_ptr<sumdata::SpillData> spill(new sumdata::SpillData);


//     if(fTimeWindow<500000000) {
//       if(uevt_nsec < fTimeWindow)
//     	uevt_bound_lo = (unsigned long long int)(((uevt_sec-1) << 32) | ((uevt_nsec+1000000000)- fTimeWindow));
//       else
//     	uevt_bound_lo = (unsigned long long int)((uevt_sec << 32) | (uevt_nsec- fTimeWindow));
//       uevt_bound_hi = (unsigned long long int)((uevt_sec << 32) | (uevt_nsec));
//     }
//     else
//       mf::LogError("BNBInfo") << "TimeWindow must be less than 0.5 sec" << "\n";

//     ret_lo = fBNBSpillTimesMap.lower_bound(uevt_bound_lo);
//     ret_up = fBNBSpillTimesMap.lower_bound(uevt_bound_hi);

//     ubnb_sec_lo_temp = ret_lo->first >> 32 & 0xFFFFFFFF;
//     ubnb_nsec_lo_temp = ret_lo->first & 0xFFFFFFFF;
//     uabs_dt_lo_temp = (unsigned long long int)abs((signed long long int)((((ubnb_sec_lo_temp - uevt_sec)*1000000000) + (ubnb_nsec_lo_temp - uevt_nsec))));
//     ubnb_sec_up_temp = ret_up->first >> 32 & 0xFFFFFFFF;
//     ubnb_nsec_up_temp = ret_up->first & 0xFFFFFFFF;
//     uabs_dt_up_temp = (unsigned long long int)abs((signed long long int)((((ubnb_sec_up_temp - uevt_sec)*1000000000) + (ubnb_nsec_up_temp - uevt_nsec))));

//     mf::LogInfo("BNBTiming") << "ubnb_sec_lo_temp " << ubnb_sec_lo_temp << " ubnb_nsec_lo_temp " << ubnb_nsec_lo_temp << " uabs_dt_lo_temp " << uabs_dt_lo_temp << " ubnb_sec_up_temp " << ubnb_sec_up_temp << " ubnb_nsec_up_temp " << ubnb_nsec_up_temp << " uabs_dt_up_temp " << uabs_dt_up_temp <<  "\n";

//     if((uabs_dt_lo_temp < fTimeWindow) || (uabs_dt_up_temp < fTimeWindow)) {
//       if(ret_lo == ret_up) {
// 	ubnb_sec = ret_lo->first >> 32 & 0xFFFFFFFF;
// 	ubnb_nsec = ret_lo->first & 0xFFFFFFFF;
// 	ugps_sec = ret_lo->second.first >> 32 & 0xFFFFFFFF;
// 	ugps_nsec = ret_lo->second.first & 0xFFFFFFFF;
// 	sdt = (signed long long int)(((ubnb_sec - uevt_sec)*1000000000) + (ubnb_nsec - uevt_nsec));

// 	spill->spilltimesec = ubnb_sec;
// 	spill->spilltimensec = ubnb_nsec;
// 	spill->gpsspilltimesec = ugps_sec;
// 	spill->gpsspilltimensec = ugps_nsec;
// 	spill->spillpot = ret_lo->second.second;
// 	spill->deltaspilltimensec = sdt;
// 	uabs_dt_last = uabs_dt;
//       }
//       else {
// 	for(std::map<unsigned long long int,std::pair<unsigned long long int,double> >::iterator it = ret_lo; it != ret_up; it++) {
// 	  ubnb_sec = it->first >> 32 & 0xFFFFFFFF;
// 	  ubnb_nsec = it->first & 0xFFFFFFFF;
// 	  ugps_sec = it->second.first >> 32 & 0xFFFFFFFF;
// 	  ugps_nsec = it->second.first & 0xFFFFFFFF;

// 	  sdt = (signed long long int)(((ubnb_sec - uevt_sec)*1000000000) + (ubnb_nsec - uevt_nsec));
// 	  uabs_dt = (unsigned long long int)abs(sdt);
// 	  if((uabs_dt < uabs_dt_last)||(ncorr==0)) {
// 	    mf::LogInfo("BNBTiming") << "sdt: " << sdt << " ubnb_sec: " << ubnb_sec << " uevt_sec: " << uevt_sec << " ubnb_nsec: " << ubnb_nsec << " uevt_nsec: " << uevt_nsec << " it->first: " << it->first << " ret_lo: " << ret_lo->first << " ret_up: " << ret_up->first << " uevt_bound_lo: " << uevt_bound_lo << " uevt_bound_hi: " << uevt_bound_hi  << "\n";
// 	    spill->spilltimesec = ubnb_sec;
// 	    spill->spilltimensec = ubnb_nsec;
// 	    spill->gpsspilltimesec = ugps_sec;
// 	    spill->gpsspilltimensec = ugps_nsec;
// 	    spill->spillpot = it->second.second;
// 	    spill->deltaspilltimensec = sdt;
// 	    uabs_dt_last = uabs_dt;
// 	    //there should be only one BNB spill corresponding to this event
// 	    ncorr++;
// 	    //break;
// 	  }
// 	}
//       }
//      }
//     if(spill->deltaspilltimensec != -9999999) {
//       mf::LogInfo("BNBTiming") << "evt sec: " << uevt_sec << " evt nsec: " << uevt_nsec << " final dt was: " << spill->deltaspilltimensec << " after ncorr: " << ncorr << "\n"; }

//     if(ncorr > 1)  { 
//       mf::LogError("BNBInfo") << "-----found " << ncorr << " spills corresponding to event run: " << evt.run() << " and subRun: " << evt.subRun() << "------\n";
//     } //error, found more than one corresponding BNB spill!
//     evt.put(std::move(spill));
//     return;
//   }
// } //namespace bnbinfo
////////////////////////////////////////////////////////////////////////

namespace bnb
{
  DEFINE_ART_MODULE(BNBInfo);
}


