////////////////////////////////////////////////////////////////////////
// Class:       BPFitter
// Module Type: producer
// File:        BPFitter_module.cc
//
// Brief:  Fits tracks to clusters using the "Break Point Fitter" model.
//
// Generated at Thu Apr 25 11:07:44 2013 by Michael Baird using artmod
// from art v1_02_06.
// authors: mbaird42@fnal.gov
////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>

// ROOT includes
#include "TH1F.h"
#include "TVector3.h"

// ART includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"

// NOvAsoft includes
#include "Geometry/Geometry.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/HitList.h"
#include "RecoBase/CellHit.h"

#include "Utilities/AssociationUtil.h"

#include "BreakPointFitter/func/Lutz.h"
#include "BreakPointFitter/func/Path.h"
#include "BreakPointFitter/func/ScatteringSurfaces.h"

namespace bpfit {
  class BPFitter;
}

class bpfit::BPFitter : public art::EDProducer
{
public:
  explicit BPFitter(fhicl::ParameterSet const & pset);
  virtual ~BPFitter();

  void produce(art::Event & evt);
  void reconfigure(fhicl::ParameterSet const & pset);

  void SortByZ(double* XY, double *Z,
	       double* XYrot, double *Zrot,
	       unsigned int N);
  // void Merge2DTracks( something... );

private:
  void TestPath();
  void TestScatteringSurfaces();

private:
  std::string fSliceLabel;
  std::string fVertexLabel;
  std::string fClusterLabel;
  bool        fWeightByQ;

  TH1F* fChiSqX;   ///< chi squared distribution for X-view fit
  TH1F* fChiSqY;   ///< chi squared distribution for Y-view fit
};

//--------------------------------------------------------------------
bpfit::BPFitter::BPFitter(fhicl::ParameterSet const & pset)
{

  ///\TODO: initalize all variable properly here...

  produces< std::vector<rb::Track> >();

  // Tracks are associated with the cluster they came from.

  ///\TODO: should I make more assns than this???
  //        e.g. - tracks to slices or tracks to tracks (from which they came)
  produces< art::Assns<rb::Track, rb::Vertex> >();
  
  this->reconfigure(pset);

  // book the histos
  art::ServiceHandle<art::TFileService> tfs;
  fChiSqX = tfs->make<TH1F>("ChiSqX",
			    "#chi^{2} per d.o.f. for x-view fit",
			    100,0.0,20.0);
  fChiSqY = tfs->make<TH1F>("ChiSqY",
			    "#chi^{2} per d.o.f. for y-view fit",
			    100,0.0,20.0);
}

//--------------------------------------------------------------------
void bpfit::BPFitter::reconfigure(fhicl::ParameterSet const & pset)
{
  fSliceLabel   = pset.get< std::string >("SliceLabel");
  fVertexLabel  = pset.get< std::string >("VertexLabel");
  fClusterLabel = pset.get< std::string >("ClusterLabel");
  fWeightByQ    = pset.get< bool        >("WeightByQ");
}

//......................................................................

bpfit::BPFitter::~BPFitter()
{ }

//......................................................................

void bpfit::BPFitter::TestPath() 
{
  art::ServiceHandle<geo::Geometry> geo;
  
  double p1[3] = {-100,-100,100};
  double p2[3] = { 100, 100,1100};
  
  bpfit::Path path(*geo, p1, p2);
  double slen = path.Length();

  unsigned int i;
  double s;
  double x0;
  double tmu, pmu, betamu, gammamu;
  double tpi, ppi, betapi, gammapi;
  double tp,  pp,  betap,  gammap;
  for (i=0; i<=1000; ++i) {
    s = slen*(float)i/1000.0;
    x0 = path.X0(s);
    path.MuonParams  (s, &tmu, &pmu, &betamu, &gammamu);
    path.PionParams  (s, &tpi, &ppi, &betapi, &gammapi);
    path.ProtonParams(s, &tp,  &pp,  &betap,  &gammap);
    
    std::cout << "11121970 " 
	      << s << " " 
	      << x0 << " "
	      << tmu << " "
	      << pmu << " "
	      << betamu << " "
	      << gammamu << " "
	      << tpi << " "
	      << ppi << " "
	      << betapi << " "
	      << gammapi << " "
	      << tp << " "
	      << pp << " "
	      << betap << " "
	      << gammap << std::endl;
  }
}

//......................................................................

void bpfit::BPFitter::TestScatteringSurfaces() 
{
  unsigned int i;
  //
  // Some bogus measurement planes every 12 cm
  //
  std::vector<double> m(101);
  for (i=0; i<101; ++i) m[i] = 12.0*(float)i;
  
  // Some bogus track parameters to go with that. Start and end points
  // needed.
  double p0[3]  = {-100, -100, 100};
  double p1[3];
  double dir[3] = {0.3, 0.3, sqrt(1.0-0.3*0.3-0.3*0.3) };
  p1[0] = p0[0] + dir[0]*m[100];
  p1[1] = p0[1] + dir[1]*m[100];
  p1[2] = p0[2] + dir[2]*m[100];
  
  art::ServiceHandle<geo::Geometry> geo;
  int pdgmuon = 13;
  ScatteringSurfaces splanes(pdgmuon, 100.0, 2.0, 0.25);
  splanes.MakeSurfaces(*geo, p0, p1, m);
  
  for (i=0; i<splanes.N(); ++i) {
    std::cout << "12031970 " 
	      << i << " "
	      << splanes.S(i) << " "
	      << splanes.SigAlpha(i) << std::endl;
  }
}

//......................................................................

void bpfit::BPFitter::produce(art::Event & evt)
{
  // this->TestPath();
  this->TestScatteringSurfaces();
  return;

  art::Handle<std::vector<rb::Cluster> > sHandle;  
  evt.getByLabel(fSliceLabel, sHandle);
    
  art::PtrVector<rb::Cluster> slices;
  for(unsigned int i=0; i<sHandle->size(); ++i){
    slices.push_back(art::Ptr<rb::Cluster>(sHandle, i));
  }



  // get associations between slices and vertices
  art::FindManyP<rb::Vertex> fmVert(sHandle, evt, fVertexLabel);
  // try to get associations between tracks and slices
  // OR between vertices and tracks
  art::FindManyP<rb::Track> fmSlTrack(sHandle, evt, fClusterLabel);


  // define the vector of tracks to be created
  std::unique_ptr<std::vector<rb::Track> >
    trackcol(new std::vector<rb::Track>);
  std::unique_ptr< art::Assns<rb::Track, rb::Vertex> >
    trkAssns(new art::Assns<rb::Track, rb::Vertex>);



  //loop over slices
  for(unsigned int sli = 0; sli < slices.size(); ++sli){

    //skip noise slice    
    if (slices[sli]->IsNoise()) continue;

    // get the vertex and the "tracks" (which I will use as clusters)
    // associated with this slice
    std::vector< art::Ptr< rb::Vertex > > v;
    std::vector< art::Ptr< rb::Track > > tracks;

    // get the vertex, then try getting the tracks from the slice first
    // followed by trying to get them from the vertex
    if(fmVert.isValid()) {
      v = fmVert.at(sli);

      ///\ TODO: It is possible that in the future I may need to generalize
      //         this to handle multiple vertices

      // There can be only one vertex!
      if(v.size() != 1) continue;

      if(fmSlTrack.isValid())
	tracks = fmSlTrack.at(sli);
      else {
	art::FindManyP<rb::Track> fmVertTrack(v,evt,fClusterLabel);
	if(fmVertTrack.isValid())
	  tracks = fmVertTrack.at(0);  // assuming only 1 vertex
	else continue;
      }
      
    } // end if fmVert isValid
    else continue;



    // loop over all tracks
    for(unsigned int trki = 0; trki < tracks.size(); ++trki) {
      // Determine if track is to be fit:
      // 2D tracks are skipped, could put a remid or some other kind of
      // track quality cut here as well...
      unsigned int nX = tracks[trki]->NXCell();
      unsigned int nY = tracks[trki]->NYCell();

      // skip 2D tracks
      if(tracks[trki]->Is2D()) continue;

      // for now, require 4 or more hits in each view to make tracks
      ///\TODO: make this able to handle events with small number of hits
      //        (e.g. - 1 in each view or 2 in one and 1 in the other or
      //         onw view all in the same plane)
      if(nX < 4 || nY < 4) continue;



      // Use the vertex origin as the new origin of the coordinate system.
      double vX = v[0]->GetX();      
      double vY = v[0]->GetY();
      double vZ = v[0]->GetZ();



      //
      // split into X-hit list and Y-hit list
      //
      const art::PtrVector<rb::CellHit>& XCells = tracks[trki]->XCells();
      const art::PtrVector<rb::CellHit>& YCells = tracks[trki]->YCells();

      rb::Cluster Xclust(geo::kX, XCells);
      rb::Cluster Yclust(geo::kY, YCells);      
      
      rb::HitList hlX(Xclust, fWeightByQ);
      rb::HitList hlY(Yclust, fWeightByQ);

      // Make a list of XYZ hit coordinates in a frame where the vertex
      // is the origin.
      double X[nX], XZ[nX];
      double Y[nY], YZ[nY];
      for(unsigned int i = 0; i < nX; ++i) {
	X [i] = hlX[i].fXY - vX;
	XZ[i] = hlX[i].fZ  - vZ;
      }
      for(unsigned int j = 0; j < nY; ++j) {
	Y [j] = hlY[j].fXY - vY;
	YZ[j] = hlY[j].fZ  - vZ;
      }



      //
      // rotate hits
      //

      // make new lists for the rotated frames
      double Xrot[nX], XZrot[nX];
      double Yrot[nY], YZrot[nY];

      double dx = tracks[trki]->Dir().X();
      double dy = tracks[trki]->Dir().Y();
      double dz = tracks[trki]->Dir().Z();

      double thetaX = atan2(dx,dz);
      double thetaY = atan2(dy,dz);

      double CX = cos(thetaX);
      double SX = sin(thetaX);
      double CY = cos(thetaY);
      double SY = sin(thetaY);

      for(unsigned int i = 0; i < nX; ++i) {
	XZrot[i] = CX*XZ[i] - SX*X[i];
	Xrot [i] = SX*XZ[i] + CX*X[i];
      }
      for(unsigned int i = 0; i < nY; ++i) {
	YZrot[i] = CY*YZ[i] - SY*Y[i];
	Yrot [i] = SY*YZ[i] + CY*Y[i];
      }

      // sort hits by Zrot
      this->SortByZ(X,XZ,Xrot,XZrot,nX);
      this->SortByZ(Y,YZ,Yrot,YZrot,nY);



      //
      // fit X-view 2D track:
      //

      // decide on number and location of scattering planes
      unsigned int NX = 1;
      NX = nX-1;

      double planeXZ[NX];
      // temporarily have planes and measurements coincide
      for(unsigned int i = 0; i < NX; ++i) planeXZ[i] = XZrot[i];

      // initialize and call Lutz
      Lutz lutzX(nX, NX);
      for(unsigned int hiti = 0; hiti < nX; ++hiti) {
	///\TODO: make the decisions about sigma x smarter!!!
	lutzX.SetMeasurement(hiti, XZrot[hiti], Xrot[hiti], 2.0/sqrt(12.0));     
      }
      // for now just use the same number of scattering planes as hits...
      ///\TODO: make the placement of planes smarter!!!
      ///\TODO: make picking sigmaSJ smarter!!!
      for(unsigned int planei = 0; planei < NX; ++planei) {
	lutzX.SetScatteringPlane(planei, planeXZ[planei], 0.005);
      }
      double aX, bX, alphaX[NX], chi2X;
      lutzX.Fit(&aX, &bX, alphaX, &chi2X);
      
      // fill with chi^2 per degrees of freedom
      fChiSqX->Fill(chi2X/(nX-2.0));


     
      // calculate trajectory, rotate back, and translate back to the original
      // detector origin
      double Xfinal[nX], XZfinal[nX];

      // add the trajectory points
      for(unsigned int traji = 0; traji < nX; ++traji) {
	double xi = aX + bX*XZrot[traji];
	for(unsigned int planej = 0; planej < NX; ++planej) {
	  if(planeXZ[planej] <= XZrot[traji])
	    xi += alphaX[planej]*(XZrot[traji]-planeXZ[planej]);
	}
	XZfinal[traji] =  CX*XZrot[traji] + SX*xi + vZ;
	Xfinal [traji] = -SX*XZrot[traji] + CX*xi + vX;
      }



      //
      // fit Y-view 2D track:
      //

      // decide on number and location of scattering planes
      unsigned int NY = 1;
      NY = nY-1;

      double planeYZ[NY];
      // temporarily have planes and measurements coincide
      for(unsigned int i = 0; i < NY; ++i) planeYZ[i] = YZrot[i];

      // initialize and call Lutz
      Lutz lutzY(nY, NY);
      for(unsigned int hiti = 0; hiti < nY; ++hiti) {
	///\TODO: make the decisions about sigma x smarter!!!
	lutzY.SetMeasurement(hiti, YZrot[hiti], Yrot[hiti], 2.0/sqrt(12.0));     
      }
      // for now just use the same number of scattering planes as hits...
      ///\TODO: make the placement of planes smarter!!!
      ///\TODO: make picking sigmaSJ smarter!!!
      for(unsigned int planei = 0; planei < NY; ++planei) {
	lutzY.SetScatteringPlane(planei, planeYZ[planei], 0.005);
      }
      double aY, bY, alphaY[NY], chi2Y;
      lutzY.Fit(&aY, &bY, alphaY, &chi2Y);

      // fill with chi^2 per degrees of freedom
      fChiSqY->Fill(chi2Y/(nY-2.0));


      
      // calculate trajectory, rotate back, and translate back to the original
      // detector origin
      double Yfinal[nY], YZfinal[nY];

      // add the trajectory points
      for(unsigned int traji = 0; traji < nY; ++traji) {
	double xi = aY + bY*YZrot[traji];
	for(unsigned int planej = 0; planej < NY; ++planej) {
	  if(planeYZ[planej] <= YZrot[traji])
	    xi += alphaY[planej]*(YZrot[traji]-planeYZ[planej]);
	}
	YZfinal[traji] =  CY*YZrot[traji] + SY*xi + vZ;
	Yfinal [traji] = -SY*YZrot[traji] + CY*xi + vY;
      }



      //
      // merge both 2D views into one 3D track 
      //
      ///\ TODO: I have to think about the logic of this more, but this might
      //         be hard coding the assumption of a forward going particle.

      unsigned int Xi = 0, Yi = 0;
      
      // Determine which view has the starting point and calculate the
      // track start and direction from that point.

      double startX = 0.0, startY = 0.0, startZ = 0.0;
      // double dirX = 0.0, dirY = 0.0, dirZ = 0.0;

      if(XZfinal[0] < YZfinal[0]) {
	// use the first X point and extrapolate from the other view
	startX =  Xfinal[0];
	startZ = XZfinal[0];
	double y1 =  Yfinal[0];
	double y2 =  Yfinal[1];
	double z1 = YZfinal[0];
	double z2 = YZfinal[1];
	double bottom = z2-z1;
	startY = (y1*z2-y2*z1)/bottom + (y2-y1)/bottom*XZfinal[0];
	Xi++;
      }
      else {
	// use the first Y point and extrapolate from the other view
	startY =  Yfinal[0];
	startZ = YZfinal[0];
	double x1 =  Xfinal[0];
	double x2 =  Xfinal[1];
	double z1 = XZfinal[0];
	double z2 = XZfinal[1];
	double bottom = z2-z1;
	startX = (x1*z2-x2*z1)/bottom + (x2-x1)/bottom*YZfinal[0];
	Yi++;
      }

      TVector3 start(startX, startY, startZ);
      TVector3 dir(dx, dy, dz);  ///\ TODO: properly calculate this!!!
      rb::Track track(tracks[trki]->AllCells(), start, dir);

      bool done = false;

      //
      ///\ TODO: I shouldn't be doing this in y = mx + b format (that will mess up
      //         tracks angled at 90 degrees where the slope is infinity.)  I
      //         should try to do this in polar or Hough-type coordinates.
      while(!done) {
	// add X point
	if(XZfinal[Xi] < YZfinal[Yi]) {
	  // Adding X point to trajectory...
	  // Interpolate/extrapolate from the other view to find the
	  // associated Y coordinate.

	  double Ycalc = 0.0;

	  unsigned int i;
	  for(i = 0; i < nY; ++i) {
	    if(YZfinal[i] > XZfinal[Xi]) break;
	  }

	  // If i = 0, then we are at the beginning of the list for this view,
	  // so we have to extrapolate back from the first 2 points.
	  if(i == 0) i = 1;

	  double y1 =  Yfinal[i-1];
	  double y2 =  Yfinal[i];
	  double z1 = YZfinal[i-1];
	  double z2 = YZfinal[i];
	  double bottom = z2-z1;
	  if(bottom == 0.0) bottom = 1.0e-9;
	  Ycalc = (y1*z2-y2*z1)/bottom + (y2-y1)/bottom*XZfinal[Xi];

	  TVector3 point(Xfinal[Xi], Ycalc, XZfinal[Xi]);
	  track.AppendTrajectoryPoint(point);
	  Xi++;
	}

	// else add Y point
	else {
	  // Adding Y point to trajectory...
	  // Interpolate/extrapolate from the other view to find the
	  // associated X coordinate.

	  double Xcalc = 0.0;

	  unsigned int i;
	  for(i = 0; i < nX; ++i) {
	    if(XZfinal[i] > YZfinal[Yi]) break;
	  }

	  // If i = 0, then we are at the beginning of the list for this view,
	  // so we have to extrapolate back from the first 2 points.
	  if(i == 0) i = 1;

	  double x1 =  Xfinal[i-1];
	  double x2 =  Xfinal[i];
	  double z1 = XZfinal[i-1];
	  double z2 = XZfinal[i];
	  double bottom = z2-z1;
	  if(bottom == 0.0) bottom = 1.0e-9;
	  Xcalc = (x1*z2-x2*z1)/bottom + (x2-x1)/bottom*YZfinal[Yi];

	  TVector3 point(Xcalc, Yfinal[Yi], YZfinal[Yi]);
	  track.AppendTrajectoryPoint(point);
	  Yi++;
	}
	// If we have reached the end of either list, then quit and handle
	// the remainder of the other list seperately.
	if(Xi == nX || Yi == nY) done = true;

	///\TODO: Put in a saftey counter to make sure that this never
	//        becomes an infinite loop.
      }

      // Finish adding the rest of the unfinished view.
      if(Xi == nX) {
	double x1 =  Xfinal[nX-2];
	double x2 =  Xfinal[nX-1];
	double z1 = XZfinal[nX-2];
	double z2 = XZfinal[nX-1];
	double bottom = z2-z1;
	if(bottom == 0.0) bottom = 1.0e-9;
	double Xcalc = 0.0;
	for(unsigned int i = Yi; i < nY; ++i) {
	  Xcalc = (x1*z2-x2*z1)/bottom + (x2-x1)/bottom*YZfinal[i];
	  TVector3 point(Xcalc, Yfinal[i], YZfinal[i]);
	  track.AppendTrajectoryPoint(point);
	}
      }	  
      if(Yi == nY) {
	double y1 =  Yfinal[nY-2];
	double y2 =  Yfinal[nY-1];
	double z1 = YZfinal[nY-2];
	double z2 = YZfinal[nY-1];
	double bottom = z2-z1;
	if(bottom == 0.0) bottom = 1.0e-9;
	double Ycalc = 0.0;
	for(unsigned int i = Xi; i < nX; ++i) {
	  Ycalc = (y1*z2-y2*z1)/bottom + (y2-y1)/bottom*XZfinal[i];
	  TVector3 point(Xfinal[i], Ycalc, XZfinal[i]);
	  track.AppendTrajectoryPoint(point);
	}
      }	 



      // add 3D track to the collection and make the association to
      // the vertex
      trackcol->push_back(track);
      util::CreateAssn(*this, evt, *trackcol, v[0], *trkAssns);

      // make the track/slice association as well???

    } // end loop over tracks
  } // end loop over slices



  // put the tracks and assns into the event
  evt.put(std::move(trackcol));
  evt.put(std::move(trkAssns));
}

//--------------------------------------------------------------------
void bpfit::BPFitter::SortByZ(double* XY, double* Z,
			      double* XYrot, double* Zrot,
			      unsigned int N)
{
  // For simplicity, this is a bubble sort.
  ///\TODO: optimize the sorting method
  bool done = false;
  while(!done) {
    done = true;
    for(unsigned int i = 1; i < N; ++i) {
      if(Zrot[i-1] > Zrot[i]) {
	double Ztemp  = Z[i];
	double XYtemp = XY[i];
	Z[i]    = Z[i-1];
	Z[i-1]  = Ztemp;
	XY[i]   = XY[i-1];
	XY[i-1] = XYtemp;

	Ztemp  = Zrot[i];
	XYtemp = XYrot[i];
	Zrot[i]    = Zrot[i-1];
	Zrot[i-1]  = Ztemp;
	XYrot[i]   = XYrot[i-1];
	XYrot[i-1] = XYtemp;

	done = false;
      } // end if
    } // end for i
  } // end while !done

}

//////////////////////////////////////////////////////////////////////


DEFINE_ART_MODULE(bpfit::BPFitter)
