///
/// \file    BreakPoint_module.cc
/// \brief   Run the break-point fitter
/// \author  messier@indiana.edu
/// \version $Id:$
///
#include "TH2F.h"
#include "TNtuple.h"
//
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/FindManyP.h"
//
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
//
#include "Utilities/AssociationUtil.h"
//
#include "BreakPointFitter/func/HitList3D.h"
#include "BreakPointFitter/func/TrackBasis.h"
#include "BreakPointFitter/func/Path.h"
#include "BreakPointFitter/func/ScatteringSurfaces.h"
#include "BreakPointFitter/func/Lutz.h"
#include "BreakPointFitter/func/BPException.h"

namespace bpfit { class BreakPoint; };

class bpfit::BreakPoint : public art::EDProducer
{
public:
  // FHICL Configuration data
  bool         fDebugHitList0;   ///< Debug the detector-frame hit list?
  bool         fDebugHitList1;   ///< Debug the track-frame hit list?
  bool         fDebugScatSurf;   ///< Debug the scattering surfaces
  bool         fDebugFitList0;   ///< Debug the fitted track locations (det. frame)
  bool         fDebugFitList1;   ///< Debug the fitted track locations (track frame)
  bool         fLogResiduals;    ///< Make ntuple of residuals?
  bool         fFillTrackNt;     ///< Fill ntuple of tracking results?
  bool         fFillExceptNt;    ///< Fill ntuple of exception cases?
  std::string  fSliceLabel;      ///< Where to find slices in event
  std::string  fVertexLabel;     ///< Where to find vertices in event
  std::string  fClusterLabel;    ///< Where to find clusters in event
  std::string  fClusterInstance; ///< Which instance of clusters to find
  unsigned int fNhitX;           ///< Require prongs to have this many x hits
  unsigned int fNhitY;           ///< Require prongs to have this many y hits
  unsigned int fNhit;            ///< Require this many hits total
  double       fChi2Cutoff;      ///< Drop hits with this chi2 or greater
  double       fdStol;           ///< Scattering surface ds tolerance
  double       fdXtol;           ///< Scattering surface dx tolerance
  double       fdTtol;           ///< Scattering surface dt tolerance
public:
  // Member data
  int          fRun;
  int          fEvent;
  unsigned int fSlice;
  unsigned int fProng;
  TNtuple*     fHitList0Nt;
  TNtuple*     fHitList1Nt;
  TNtuple*     fScatSurfNt;
  TNtuple*     fFitList0Nt;
  TNtuple*     fFitList1Nt;
  TNtuple*     fResidNt;
  TNtuple*     fTrackNt;
  TNtuple*     fExceptNt;
public:
  //......................................................................
  
  void reconfigure(fhicl::ParameterSet const& p)
  {
    fDebugHitList0 = p.get<bool>("DebugHitList0");
    fDebugHitList1 = p.get<bool>("DebugHitList1");
    fDebugScatSurf = p.get<bool>("DebugScatSurf");
    fDebugFitList0 = p.get<bool>("DebugFitList0");
    fDebugFitList1 = p.get<bool>("DebugFitList1");
    fLogResiduals  = p.get<bool>("LogResiduals");
    fFillTrackNt   = p.get<bool>("FillTrackNt");
    fFillExceptNt  = p.get<bool>("FillExceptNt");

    fSliceLabel      = p.get<std::string>("SliceLabel");
    fVertexLabel     = p.get<std::string>("VertexLabel");
    fClusterLabel    = p.get<std::string>("ClusterLabel");
    fClusterInstance = p.get<std::string>("ClusterInstance");

    fNhitX      = p.get<unsigned int>("NhitX");
    fNhitY      = p.get<unsigned int>("NhitY");
    fNhit       = p.get<unsigned int>("Nhit");
    fChi2Cutoff = p.get<double>("Chi2Cutoff");
    fdStol      = p.get<double>("dStol");
    fdXtol      = p.get<double>("dXtol");
    fdTtol      = p.get<double>("dTtol");
  }
  
  //......................................................................

  explicit BreakPoint(fhicl::ParameterSet const& p) :
    fRun(0),
    fEvent(0),
    fProng(0),
    fHitList0Nt(0),
    fHitList1Nt(0),
    fScatSurfNt(0),
    fFitList0Nt(0),
    fFitList1Nt(0),
    fResidNt(0),
    fTrackNt(0),
    fExceptNt(0)
  {
    produces< std::vector<rb::Track> >();
    produces< art::Assns<rb::Track, rb::Vertex> >();
    produces< art::Assns<rb::Track, rb::Cluster> >();
    this->reconfigure(p);
  }

  //......................................................................

  virtual ~BreakPoint() { }

  //......................................................................

  void FindSlices(const art::Event& evt, 
		  art::Handle< std::vector<rb::Cluster> >& h,
		  art::PtrVector<rb::Cluster>& s) 
  {
    unsigned int i;
    evt.getByLabel(fSliceLabel, h);
    for (i=0; i<h->size(); ++i) {
      s.push_back(art::Ptr<rb::Cluster>(h, i));
    }
  }

  //......................................................................
  
  void FindProngs(art::Event&                          evt,
		  unsigned int                         islice,
		  art::FindManyP<rb::Prong>&           prongh,
		  std::vector< art::Ptr<rb::Vertex> >& vertexp,
		  std::vector< art::Ptr<rb::Prong> >&  prongp)
  {
    if (prongh.isValid()) {
      prongp = prongh.at(islice);
    }
    else {
      //
      // Hmm... No prongs associated with slice - try the vertex.
      //
      art::FindManyP<rb::Prong> prongh2(vertexp, evt, fClusterLabel);
      if (prongh2.isValid()) {
	if (prongh2.size()==1) {
	  prongp = prongh2.at(0);
	}
	else {
	  std::cerr << __FILE__ << ":" << __LINE__ 
		    << " Multiple vertices breaks our assumptions."
		    << std::endl;
	}
      }
    }
  }
  
  //......................................................................
  
  bool GoodProng(art::Ptr<rb::Prong>& t)
  {
    if (t->Is2D()) return false;
    
    unsigned int nx = t->NXCell();
    unsigned int ny = t->NYCell();
    if (nx<fNhitX || ny<fNhitY) return false;
    if (nx+ny<fNhit)            return false;
    
    return true;
  }
  
  //......................................................................
  
  void FillHitList(art::Ptr<rb::Prong>& prongp,
		   bpfit::HitList3D&    hits)
  {
    art::ServiceHandle<geo::Geometry> geom;
    unsigned int itr, i;
    unsigned int n = prongp->NCell();
    //
    // Build the 3D hits in two passes. In the first we don't have
    // information about where transversly we are along a cell. In the
    // second iteration, we do.
    //
    for (itr=0; itr<2; ++itr) {
      for (i=0; i<n; ++i) {
	const rb::CellHit* h = prongp->Cell(i).get();
	
	unsigned int ip = h->Plane();
	unsigned int ic = h->Cell();
	
	const geo::PlaneGeo* p = geom->Plane(ip);
	const geo::CellGeo*  c = p->   Cell(ic);
	
	double xyz[3] = {0,0,0};;
	double sigx=0, sigy=0, sigz=0, sigt=50;
	double w = 1.0;
	if (itr==0) {
	  //
	  // On first iteration we don't have information from other
	  // view. Use cell centers for position estimates.
	  //
	  static const double sqrt3 = sqrt(3.0);
	  c->GetCenter(xyz);
	  if (p->View()==geo::kX) sigx = c->HalfW()/sqrt3;
	  if (p->View()==geo::kY) sigy = c->HalfW()/sqrt3;
	  sigz  = c->HalfD()/sqrt3;
	}
	else {
	  //
	  // On second iteration, we have more information about where
	  // in the cell we are. Use the estimated location from the
	  // other view. Otherwise, most information is a copy of what
	  // is already there.
	  //
	  if (p->View()==geo::kX) c->GetCenter(xyz, hits[i].fY);
	  if (p->View()==geo::kY) c->GetCenter(xyz, hits[i].fX);
	  sigx = hits[i].fSigX;
	  sigy = hits[i].fSigY;
	  sigz = hits[i].fSigZ;
	  sigt = hits[i].fSigT;
	  w    = hits[i].fW;
	}
	
	if (p->View()==geo::kX) {
	  hits.SetHit(i, geo::kX, xyz[0], xyz[1], xyz[2], h->TNS(),
		      sigx, sigy, sigz, sigt, w);
	}
	else if (p->View()==geo::kY) {
	  hits.SetHit(i,
		      geo::kY, xyz[0], xyz[1], xyz[2], h->TNS(),
		      sigx, sigy, sigz, sigt, w);
	}
	else abort();
      } // loop on hits
      hits.CompleteOrtho();
    } // iterations
    //
    // Keep all hits inside the detector
    //
    double xyz[3];
    for (i=0; i<hits.size(); ++i) {
      xyz[0] = hits[i].fX;
      xyz[1] = hits[i].fY;
      xyz[2] = hits[i].fZ;
      this->ClampToDetector(xyz);
      hits[i].fX = xyz[0];
      hits[i].fY = xyz[1];
      hits[i].fZ = xyz[2];
    }
    
    hits.SortByZ();
  }

  //......................................................................
  ///
  /// Dump the hit list to an ntuple we can study offline for
  /// debugging
  ///
  void DebugHitList(bpfit::HitList3D& h, int which) 
  {
    TNtuple* nt = 0;
    if (which==0) {
      if (fHitList0Nt==0) {
	art::ServiceHandle<art::TFileService> tfs;
	fHitList0Nt = tfs->make<TNtuple>("hits0",
					 "Det Frame Hit List Debugging",
					 "run:evt:trk:v:x:y:z:sx:sy:sz");
      }
      nt = fHitList0Nt;
    }
    if (which==1) {
      if (fHitList1Nt==0) {
	art::ServiceHandle<art::TFileService> tfs;
	fHitList1Nt = tfs->make<TNtuple>("hits1",
					 "Trk Frame Hit List Debugging",
					 "run:evt:trk:v:x:y:z:sx:sy:sz");
      }
      nt = fHitList1Nt;
    }
    float x[10];
    x[0] = fRun;
    x[1] = fEvent;
    x[2] = fProng;
    for (unsigned int i=0; i<h.size(); ++i) {
      x[3] = h[i].fView;
      x[4] = h[i].fX;
      x[5] = h[i].fY;
      x[6] = h[i].fZ;
      x[7] = h[i].fSigX;
      x[8] = h[i].fSigY;
      x[9] = h[i].fSigZ;
      nt->Fill(x);
    }
  }

  //......................................................................
  
  void FillTrackNt(const HitList3D& h,
		   const Lutz& lx,
		   const Lutz& ly,
		   const Path& path)
  {
    if (fTrackNt==0) {
      art::ServiceHandle<art::TFileService> tfs;
      fTrackNt = tfs->make<TNtuple>("trks",
				    "Track summary",
				    "run:evt:slice:trk:"
				    "nhitx:nhity:x0:y0:z0:x1:y1:z1:l:p:"
				    "theta:phi:"
				    "chi2xh:chi2xa:chi2yh:chi2ya");
    }
    unsigned int i;
    unsigned int nhitx = 0, nhity = 0;
    for (i=0; i<h.size(); ++i) {
      if (h[i].fView==geo::kX) ++nhitx;
      if (h[i].fView==geo::kY) ++nhity;
    }
    
    double p, len;
    len = path.Length();
    path.MuonParams(0, 0, &p, 0, 0);
    i = h.size()-1;
    double dx = path.fXYZ[1].X() - path.fXYZ[0].X();
    double dy = path.fXYZ[1].Y() - path.fXYZ[0].Y();
    double dz = path.fXYZ[1].Z() - path.fXYZ[0].Z();
    if (dz==0.0) dz = 1.0E-9;
    
    float x[20];
    x[0]  = fRun;
    x[1]  = fEvent;
    x[2]  = fSlice;
    x[3]  = fProng;
    x[4]  = nhitx;
    x[5]  = nhity;
    x[6]  = h[0].fX;
    x[7]  = h[0].fY;
    x[8]  = h[0].fZ;
    x[9]  = h[i].fX;
    x[10] = h[i].fY;
    x[11] = h[i].fZ;
    x[12] = len;
    x[13] = p;
    x[14] = 180.0*acos(dz/sqrt(dx*dx+dy*dy+dz*dz))/M_PI;
    x[15] = 180.0*atan2(dy,dx)/M_PI;
    x[16] = lx.Chi2XI();
    x[17] = lx.Chi2Beta();
    x[18] = ly.Chi2XI();
    x[19] = ly.Chi2Beta();
    fTrackNt->Fill(x);
  }

  //......................................................................
  
  void FillExceptNt(int run, int subrun, int event, BPException Ex)
  {

   if (fExceptNt == 0) {
      art::ServiceHandle<art::TFileService> tfs;
      fExceptNt = tfs->make<TNtuple>("except",
				    "Exception summary",
				    "run:subrun:evt:slice:prong:flag:value");
   }

   float x[7];
   x[0]  = run;
   x[1]  = subrun;
   x[2]  = event;
   x[3]  = fSlice;
   x[4]  = fProng;
   x[5]  = Ex.fFlag;
   x[6]  = Ex.fValue;
   fExceptNt->Fill(x);

  }

  //......................................................................
  
  void MakeBasis(art::Ptr<rb::Vertex>& vtx,
		 const HitList3D&      hits,
		 bpfit::TrackBasis&    sys) 
  {
    //
    // Extract hits into format expected by TrackBasis
    //
    unsigned int i;
    std::vector<double> x(hits.size());
    std::vector<double> y(hits.size());
    std::vector<double> z(hits.size());
    std::vector<double> dx(hits.size());
    std::vector<double> dy(hits.size());
    std::vector<double> dz(hits.size());

    for (i=0; i<hits.size(); ++i) {
      x[i]  = hits[i].fX;
      y[i]  = hits[i].fY;
      z[i]  = hits[i].fZ;
      dx[i] = hits[i].fSigX;
      dy[i] = hits[i].fSigY;
      dz[i] = hits[i].fSigZ;
    }

    sys.MakeBasis(x,y,z,dx,dy,dz,vtx->GetX(),vtx->GetY(),vtx->GetZ());
  }

  //......................................................................

  void ClampToDetector(double* p)
  {
    art::ServiceHandle<geo::Geometry> geom;
    if (p[0]>geom->DetHalfWidth())  p[0] = geom->DetHalfWidth();
    if (p[1]>geom->DetHalfHeight()) p[1] = geom->DetHalfHeight();
    if (p[2]>geom->DetLength())     p[2] = geom->DetLength();

    if (p[0]<-geom->DetHalfWidth())  p[0] = -geom->DetHalfWidth();
    if (p[1]<-geom->DetHalfHeight()) p[1] = -geom->DetHalfHeight();
    if (p[2]<0)                      p[2] = 0;
  }

  //......................................................................

  void FindEndPoints(HitList3D&         h,
		     bpfit::TrackBasis& sys,
		     double*            p0,
		     double*            p1) 
  {
    unsigned int i;
    double zmin =  9e9;
    double zmax = -9e9;
    for (i=0; i<h.size(); ++i) {
      if (h[i].fZ<zmin) zmin = h[i].fZ;
      if (h[i].fZ>zmax) zmax = h[i].fZ;
    }
    p0[0] = 0.0; p0[1] = 0.0; p0[2] = zmin;
    p1[0] = 0.0; p1[1] = 0.0; p1[2] = zmax;
    double dx;
    double dy;
    double dz;
    sys.TrkToDet(0,      0,      zmin,   0,   0,   0,
		 &p0[0], &p0[1], &p0[2], &dx, &dy, &dz);
    sys.TrkToDet(0,      0,      zmax,   0,   0,   0,
		 &p1[0], &p1[1], &p1[2], &dx, &dy, &dz);
    this->ClampToDetector(p0);
    this->ClampToDetector(p1);
  }
  
  //......................................................................

  void DebugScatSurf(bpfit::ScatteringSurfaces& s)
  {
    if (fScatSurfNt==0) {
      art::ServiceHandle<art::TFileService> tfs;
      fScatSurfNt = tfs->make<TNtuple>("scat",
				       "Scattering Surfaces",
				       "run:evt:trk:s:sigalpha");
    }
    float x[5];
    x[0] = fRun;
    x[1] = fEvent;
    x[2] = fProng;
    for (unsigned int i=0; i<s.N(); ++i) {
      x[3] = s.S(i);
      x[4] = s.SigAlpha(i);
      fScatSurfNt->Fill(x);
    }
  }

  //......................................................................

  void DebugFitList(unsigned int which,
		    const std::vector<double>& x,
		    const std::vector<double>& y,
		    const std::vector<double>& z) 
  {
    TNtuple* nt = 0;
    if (which==0) {
      if (fFitList0Nt==0) {
	art::ServiceHandle<art::TFileService> tfs;
	fFitList0Nt = tfs->make<TNtuple>("fit0",
					 "fitted track - det. frame",
					 "run:evt:trk:x:y:z");
      }
      nt = fFitList0Nt;
    }
    if (which==1) {
      if (fFitList1Nt==0) {
	art::ServiceHandle<art::TFileService> tfs;
	fFitList1Nt = tfs->make<TNtuple>("fit1",
					 "fitted track - track frame",
					 "run:evt:trk:x:y:z");
      }
      nt = fFitList1Nt;
    }

    unsigned int i;
    float v[6];
    v[0] = fRun;
    v[1] = fEvent;
    v[2] = fProng;
    for (i=0; i<z.size(); ++i) {
      v[3] = x[i];
      v[4] = y[i];
      v[5] = z[i];
      nt->Fill(v);
    }
  }
  
  //......................................................................
  //
  // Create an ntuple so we can examine the residuals. Variables are:
  // m   - measured value. x/y/alpha in the two views
  // sig - estimated 1-sigma error on measured value
  // fit - fitted value of parameter
  // w   - which kind of measurement is it. Code follows:
  //       0 = X measurement from x view
  //       1 = X measurement interpolated
  //       2 = Y measurement from the y view
  //       3 = Y measurement from the y view
  //       4 = angle "measurement" from the x view
  //       5 = angle "measurement" from the y view
  //
  void LogResiduals(bpfit::HitList3D&          hit,
		    bpfit::ScatteringSurfaces& scat,
		    bpfit::Lutz&               xfit,
		    bpfit::Lutz&               yfit,
		    double*                    alphax,
		    double*                    alphay)
  {
    if (fResidNt==0) {
      art::ServiceHandle<art::TFileService> tfs;
      fResidNt = tfs->make<TNtuple>("resid", 
				    "resid", 
				    "run:evt:trk:m:sig:fit:w");
    }
    unsigned int i;
    float v[7];
    v[0] = fRun;
    v[1] = fEvent;
    v[2] = fProng;
    for (i=0; i<hit.size(); ++i) {
      v[3] = hit[i].fX;
      v[4] = hit[i].fSigX;
      v[5] = xfit.X(hit[i].fZ);
      if (hit[i].fView==geo::kX) v[6] = 0;
      if (hit[i].fView==geo::kY) v[6] = 1;
      fResidNt->Fill(v);

      v[3] = hit[i].fY;
      v[4] = hit[i].fSigY;
      v[5] = yfit.X(hit[i].fZ);
      if (hit[i].fView==geo::kY) v[6] = 2;
      if (hit[i].fView==geo::kX) v[6] = 3;
      fResidNt->Fill(v);
    }
    for (i=0; i<scat.N(); ++i) {
      v[3] = 0;
      v[4] = scat.SigAlpha(i);
      v[5] = alphax[i];
      v[6] = 4;
      fResidNt->Fill(v);

      v[3] = 0;
      v[4] = scat.SigAlpha(i);
      v[5] = alphay[i];
      v[6] = 5;
      fResidNt->Fill(v);
    }
  }

  //......................................................................

  void FitTrack(art::Ptr<rb::Vertex>& vertexp, 
		art::Ptr<rb::Prong>&  prongp,
		rb::Track& track)
  {
    unsigned int i, j;
    //
    // Track fitting starts with a hit list
    //
    bpfit::HitList3D hits0(prongp->NCell());
    this->FillHitList(prongp, hits0);
    if (fDebugHitList0) this->DebugHitList(hits0, 0);
    //
    // Then we want to swing around to a track-based coordinate system
    // so that we can properly locate the scattering planes
    //
    bpfit::TrackBasis sys;
    this->MakeBasis(vertexp, hits0, sys);
    bpfit::HitList3D hits1(hits0);
    for (i=0; i<hits0.size(); ++i) {
      sys.DetToTrk(hits0[i].fX,     hits0[i].fY,     hits0[i].fZ,
		   hits0[i].fSigX,  hits0[i].fSigY,  hits0[i].fSigZ,
		   &hits1[i].fX,    &hits1[i].fY,    &hits1[i].fZ,
		   &hits1[i].fSigX, &hits1[i].fSigY, &hits1[i].fSigZ);
    }
    if (fDebugHitList1) this->DebugHitList(hits1, 1);
    hits1.SortByZ();

    //
    // Get the extent of the track and z1 locations of measurement
    // planes
    //
    double p0[3], p1[3];
    this->FindEndPoints(hits1, sys, p0, p1);

    int    PDG_MUON  = 13;

    std::vector<double> s(hits1.size());
    for (i=0; i<hits1.size(); ++i) {
      s[i] = hits1[i].fZ;
    }

    art::ServiceHandle<geo::Geometry> geom;    
    bpfit::ScatteringSurfaces scat_surf(PDG_MUON, fdStol, fdXtol, fdTtol);
    scat_surf.MakeSurfaces(*geom, p0, p1, s);
    if (fDebugScatSurf) this->DebugScatSurf(scat_surf);
    
    bpfit::Lutz lutzx(hits1.size(), scat_surf.N());
    bpfit::Lutz lutzy(hits1.size(), scat_surf.N());
    for (i=0; i<hits1.size(); ++i) {
      lutzx.SetMeasurement(i, hits1[i].fZ, hits1[i].fX, hits1[i].fSigX);
      lutzy.SetMeasurement(i, hits1[i].fZ, hits1[i].fY, hits1[i].fSigY);
    }
    for (j=0; j<scat_surf.N(); ++j) {
      lutzx.SetScatteringPlane(j, scat_surf.S(j), scat_surf.SigAlpha(j));
      lutzy.SetScatteringPlane(j, scat_surf.S(j), scat_surf.SigAlpha(j));
    }
    double alphax[scat_surf.N()];
    double alphay[scat_surf.N()];
    lutzx.Fit(0, 0, alphax, 0);
    lutzy.Fit(0, 0, alphay, 0);
    //
    // Check the fit results and remove far-away outliers
    //
    double chix, chiy, chim, fac;
    for (i=0; i<hits1.size(); ++i) {
      chix = lutzx.Chi2XIi(i);
      chiy = lutzy.Chi2XIi(i);
      chim = (chix>chiy) ? chix : chiy;
      fac  = sqrt(chim);
      if (chim>fChi2Cutoff) {
	hits1[i].fW = 1.0E-9;
	hits1[i].fSigX *= fac;
	hits1[i].fSigY *= fac;
	hits1[i].fSigZ *= fac;
      }
    }
    //
    // Need to recompute the projection from orthogonal projection as
    // outliers would have biased the estimates
    //
    hits1.CompleteOrtho();
    for (i=0; i<hits1.size(); ++i) {
      lutzx.SetMeasurement(i, hits1[i].fZ, hits1[i].fX, hits1[i].fSigX);
      lutzy.SetMeasurement(i, hits1[i].fZ, hits1[i].fY, hits1[i].fSigY);
    }
    lutzx.Fit(0, 0, alphax, 0);
    lutzy.Fit(0, 0, alphay, 0);
    
    //
    // Get the fitted track positions in the track frame. Take the
    // start and end points from the hits and insert the track
    // locations from the scattering surfaces.
    //
    std::vector<double> xfit(scat_surf.N()+2);
    std::vector<double> yfit(scat_surf.N()+2);
    std::vector<double> zfit(scat_surf.N()+2);
    zfit[0] = hits1[0].fZ;
    for (j=0; j<scat_surf.N(); ++j) {
      zfit[1+j] = scat_surf.S(j);
    }
    zfit[1+j] = hits1[hits1.size()-1].fZ;
    for (j=0; j<zfit.size(); ++j) {
      xfit[j] = lutzx.X(zfit[j]);
      yfit[j] = lutzy.X(zfit[j]);
    }

    if (fDebugFitList1) this->DebugFitList(1, xfit, yfit, zfit);
    if (fLogResiduals) {
      this->LogResiduals(hits1, scat_surf, lutzx, lutzy, alphax, alphay);
    }

    //
    // Translate back to the detector frame and build a path
    //
    double xyz[3], dx, dy, dz;
    bpfit::Path path(*geom, zfit.size());
    for (j=0; j<zfit.size(); ++j) {
      sys.TrkToDet(xfit[j], yfit[j], zfit[j], 0,   0,   0,
		   &xyz[0], &xyz[1], &xyz[2], &dx, &dy, &dz);
      path.SetPoint(j, xyz);
    }
    //
    // Log the track parameters
    //
    if (fFillTrackNt) this->FillTrackNt(hits0, lutzx, lutzy, path);

    //
    // Build the track before returning
    // \todo Where to store chi^2's, GeV, ... you know, the track parameters...
    //
    TVector3 dir(path.fXYZ[1].X() - path.fXYZ[0].X(),
		 path.fXYZ[1].Y() - path.fXYZ[0].Y(),
		 path.fXYZ[1].Z() - path.fXYZ[0].Z());
    dir.Unit();
    track.SetDir(dir);
    track.SetStart(path.fXYZ[0]);

    for (i=0; i<path.fXYZ.size(); ++i) {
      track.AppendTrajectoryPoint(path.fXYZ[i]);
    }

  }

  //......................................................................
  
  void produce(art::Event& evt) 
  {
    //
    // Allocations for the track products and their associations
    //
    std::unique_ptr<std::vector<rb::Track> >
      trackcol(new std::vector<rb::Track>);
    std::unique_ptr< art::Assns<rb::Track, rb::Vertex> >
      trkAssns(new art::Assns<rb::Track, rb::Vertex>);
    std::unique_ptr< art::Assns<rb::Track, rb::Cluster> >
      trkslcAssns(new art::Assns<rb::Track, rb::Cluster>);

    fRun   = evt.run();
    fEvent = evt.id().event();
    fSlice = 0;
    //
    // Extract the slices, vertices, and tracks we need as inputs
    //
    art::Handle< std::vector<rb::Cluster> > sliceh;
    art::PtrVector<rb::Cluster>             slicep;
    this->FindSlices(evt, sliceh, slicep);
    art::FindManyP<rb::Vertex> vertexh(sliceh, evt, fVertexLabel);
    art::FindManyP<rb::Prong>
      prongh(sliceh, evt, art::InputTag(fClusterLabel,fClusterInstance));
    if (!vertexh.isValid()) {
      return;
    }

    unsigned int islice;
    for (islice=0; islice<slicep.size(); ++islice) {
      fSlice = islice;
      //
      // Skip the noise slice - no physics there
      //
      if (slicep[islice]->IsNoise()) continue;

      //
      // Find vertex associated with this slice
      //
      std::vector< art::Ptr<rb::Vertex> > vertexp;
      vertexp = vertexh.at(islice);
      if (vertexp.size()!=1) continue;
      
      //
      // Find the tracks associated with this slice
      //
      std::vector< art::Ptr<rb::Prong> > prongp;
      this->FindProngs(evt, islice, prongh, vertexp, prongp);
      if (prongp.size()==0) continue;

      //
      // At this point we should have a list of tracks to fit. Go
      // ahead and process them
      //
      unsigned int nprong_good = 0; // How many tracks pass selection cuts
      for (fProng=0; fProng<prongp.size(); ++fProng) {
	//
	// Figure out if the track has enough information to be
	// fit. If not, skip to the next track
	//
	if (!this->GoodProng(prongp[fProng])) continue;
	++nprong_good;

	//
	// Attempt to fit the track.
	//
	rb::Track track(*prongp[fProng]);

	try {
	  ////////////////////////////////////////////////////////////////////////////////
	  // Put for loop here to fit 3 times (muon, pion, proton).
	  // Add input parameter to FitTrack that denotes what type of particle.
	  // Eg. FitTrack(vertexp[0], prongp[fProng], track, evt, islice, particle)
	  // Then inside of FitTrack if particle=0 fit for muon, if particle=1
	  // fit for pion, if particle=2 fit for proton.  Having trackcol->push_back(track)
	  // (and the associations) in the for loop will push the three variations
	  // of tracks into trackcol and allow us to store them for downstream comparison.
	  ////////////////////////////////////////////////////////////////////////////////
	  this->FitTrack(vertexp[0], prongp[fProng], track);
	  trackcol->push_back(track);
	  util::CreateAssn(*this, evt, *trackcol, vertexp[0], *trkAssns);
	  util::CreateAssn(*this, evt, *trackcol, slicep[islice] , *trkslcAssns);		    
	}
	catch (BPException& Ex) {
	  Ex.Print();
	  if(fFillExceptNt) this->FillExceptNt(evt.run(),evt.subRun(),
					       evt.id().event(),Ex);
	}
		
      } // Loop on prongs
    } // Loop on slices

    evt.put(std::move(trackcol));
    evt.put(std::move(trkAssns));
    evt.put(std::move(trkslcAssns));

  } // produce() method
};

//......................................................................

DEFINE_ART_MODULE(bpfit::BreakPoint);

////////////////////////////////////////////////////////////////////////
