#include "BreakPointFitter/func/dEdxTable.h"
#include <string>
#include <cmath>
#include <iostream>
using namespace bpfit;
//
// Muon mass used by Groom, Mokhov, Striganov to compute their tables
//
static const double gsMmu = 105.6583568E-3;

//......................................................................

dEdxTable::dEdxTable(const char* mat) : 
  fSz(0),
  fT(0),
  fdEdx(0),
  fRho(0)
{
  static const std::string pvc   ("PVC");
  static const std::string scint ("Scintillator");
  static const std::string paraf ("Paraffin");
  static const std::string glue  ("Glue");
  static const std::string steel ("Steel");
  static const std::string vac   ("Vacuum");

  if      (scint==mat) this->InitMineralOil();
  else if (pvc  ==mat) this->InitPVC();
  else if (steel==mat) this->InitSteel();
  else if (glue ==mat) this->InitPVC();
  else if (paraf==mat) this->InitParaffin();
  else if (vac  ==mat) this->InitVac();
  else {
    static unsigned int nwarn = 50;
    if (nwarn>0) {
      --nwarn;
      std::cerr << __FILE__ << ":" << __LINE__ 
		<< " Have no table for material " << mat
		<< " Defaulting to Vacuum.";
      if (nwarn==0) {
	std::cerr << " Last warning.";
      }
      std::cerr << std::endl;
      this->InitVac();
    }
  }
}

//......................................................................

double dEdxTable::TtoBetaGamma(double t, double m) const
{
  double x = t/m;
  return sqrt(x*(x+2));
}

//......................................................................

double dEdxTable::BetaGammaToT(double bg, double m) const
{
  return m*(sqrt(bg*bg+1)-1);
}

//......................................................................

double dEdxTable::Interp(double t,
			 double m,
			 const double T[],
			 const double dedx[],
			 unsigned int imax) const
{
  //
  // Compute equivalent muon kinetic energy. Ie. a muon with
  // the same beta*gamma
  //
  double tmu = this->BetaGammaToT( this->TtoBetaGamma(t, m), gsMmu );
  
  unsigned int i0=0, i1=1;
  if      (tmu<=T[0])      { i0=0;      i1=1;}
  else if (tmu>=T[imax-1]) { i0=imax-2; i1=imax-1;}
  else {
    for (; i1<imax; ++i0, ++i1) {
      if (tmu>=T[i0] && tmu<T[i1]) break;
    }
  }

  double x  = log(tmu);
  double y;
  double x0 = log(T[i0]);
  double x1 = log(T[i1]);
  double y0 = log(dedx[i0]);
  double y1 = log(dedx[i1]);
  
  y = y0 + (x-x0)*(y1-y0)/(x1-x0);

  return exp(y)*1E-3;
}

//......................................................................
//
// TABLE III-8
//
void dEdxTable::InitPVC()
{
  fSz = 32;
  static const double T[] = {
    10.0E-3,  14.0E-3,  20.0E-3,  30.0E-3,  40.0E-3, 80.0E-3,
    100.0E-3, 140.0E-3, 200.0E-3, 298.0E-3, 300.0E-3, 400.0E-3, 800.0E-3,
    1.0,      1.4,      2.0,      3.0,      4.0,      8.0,
    10.0,     14.0,     20.0,     30.0,     40.0,     80.0,
    100.0,    140.0,    200.0,    300.,     400.0,    696.0,    800.0
  };
  static const double dEdx[] = {
    6.998, 5.469, 4.279, 3.326, 2.843, 2.140,
    2.009, 1.877, 1.803, 1.780, 1.780, 1.790, 1.870,
    1.905, 1.961, 2.023, 2.094, 2.144, 2.262,
    2.299, 2.357, 2.420, 2.500, 2.566, 2.780,
    2.877, 3.062, 3.334, 3.776, 4.218, 5.531, 5.996
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 1.49;
}

//......................................................................
///
/// TABLE I-35
///
/// SPECIAL NOTE: I am currently using the values for Iron as a temporary
/// approximation for steel.
///
/// ///\TODO: Replace the values below with the corect values for steel.
void dEdxTable::InitSteel()
{
  fSz = 32;
  static const double T[] = {
    10.0E-3,  14.0E-3,  20.0E-3,  30.0E-3,  40.0E-3, 80.0E-3,
    100.0E-3, 140.0E-3, 200.0E-3, 274.0E-3, 300.0E-3, 400.0E-3, 800.0E-3,
    1.0,      1.4,      2.0,      3.0,      4.0,      8.0,
    10.0,     14.0,     20.0,     30.0,     40.0,     80.0,
    100.0,    140.0,    200.0,    300.0,    345.0,    400.0,    800.0
  };
  static const double dEdx[] = {
    5.494, 4.321, 3.399, 2.654, 2.274, 1.717,
    1.616, 1.516, 1.463, 1.451, 1.453, 1.467, 1.548,
    1.582, 1.637, 1.697, 1.767, 1.816, 1.936,
    1.975, 2.039, 2.113, 2.214, 2.303, 2.623,
    2.777, 3.082, 3.543, 4.304, 4.651, 5.079, 8.221
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 7.874;
}

//......................................................................
//
// TABLE II-21
// 
void dEdxTable::InitParaffin()
{
  fSz = 31;
  static const double T[] = {
    10.0E-3,  14.0E-3,  20.0E-3,  30.0E-3,  40.0E-3,  80.0E-3,
    100.0E-3, 140.0E-3, 200.0E-3, 300.0E-3, 328.E-3,  400.0E-3, 800.0E-3,
    1.0,      1.4,      2.0,      3.0,      4.0,      8.0,
    10.0,     14.0,     20.0,     30.0,     40.0,     80.0,
    100.0,    140.0,    200.0,    300.0,    345.0,    400.0,    800.0
  };
  static const double dEdx[] = {
    8.531, 6.645, 5.183, 4.015, 3.426, 2.561,
    2.397, 2.228, 2.130, 2.089, 2.088, 2.093, 2.164,
    2.198, 2.254, 2.315, 2.386, 2.436, 2.554,
    2.591, 2.647, 2.707, 2.779, 2.836, 3.003,
    3.072, 3.201, 3.381, 3.664, 3.942, 5.038
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 0.930;
}

//......................................................................
//
// Vaccuum is a valid GEANT material. Set things very close to
// zero. Keep the tables small for sake of efficiency
//
void dEdxTable::InitVac()
{
  fSz = 2;
  static const double T[]    = { 10.0E-3,  800.0 };
  static const double dEdx[] = { 1.0E-6,   1.0E-6};
  fT    = T;
  fdEdx = dEdx;
  fRho  = 1.0E-6;
}

//......................................................................
//
// Model mineral oil as less-dense paraffin
//
void dEdxTable::InitMineralOil()
{
  this->InitParaffin();
  fRho = 0.859;
}

//......................................................................

double dEdxTable::operator()(double T, double m) const
{
  if (fT==0) return 0.0;
  return fRho*this->Interp(T, m, fT, fdEdx, fSz);
}

////////////////////////////////////////////////////////////////////////
