#include "CAFAna/Binning.h"

#include "TDirectory.h"
#include "TH1.h"
#include "TObjString.h"

namespace ana
{
  //----------------------------------------------------------------------
  Binning Binning::Simple(int n, double lo, double hi)
  {
    Binning bins;
    bins.fNBins = n;
    bins.fMin = lo;
    bins.fMax = hi;
    return bins;
  }

  //----------------------------------------------------------------------
  Binning Binning::Custom(const std::vector<double>& edges)
  {
    assert(edges.size() > 1);

    Binning bins;
    bins.fEdges = edges;
    bins.fNBins = edges.size()-1;
    bins.fMin = edges.front();
    bins.fMax = edges.back();
    return bins;
  }

  //----------------------------------------------------------------------
  Binning TrueEnergyBins()
  {
    // Osc P is ~sin^2(1/E). Difference in prob across a bin is ~dP/dE which
    // goes as 1/E^2 times a trigonometric term depending on the parameters but
    // bounded between -1 and +1. So bins should have width ~1/E^2. E_i~1/i
    // achieves that.

    const int kNumTrueEnergyBins = 100;

    // N+1 bin low edges
    std::vector<double> edges(kNumTrueEnergyBins+1);

    const double Emin = .5; // 500 MeV: there's really no events below there

    // How many edges to generate. Allow room for 0-Emin bin
    const double N = kNumTrueEnergyBins-1;
    const double A = N*Emin;

    edges[0] = 0;

    for(int i = 1; i <= N; ++i){
      edges[kNumTrueEnergyBins-i] = A/i;
    }

    edges[kNumTrueEnergyBins] = 120; // Replace the infinity that would be here

    return Binning::Custom(edges);
  }

  //----------------------------------------------------------------------
  void Binning::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();

    TObjString("Binning").Write("type");

    TH1D hMinMax("", "N, min and max", 2, 0, 2);
    hMinMax.SetBinContent(1, fNBins);
    hMinMax.SetBinContent(2, fMin);
    hMinMax.SetBinContent(3, fMax);
    hMinMax.Write("nminmax");

    if(!fEdges.empty()){
      TH1D edges("", "Edges", fEdges.size(), 0, fEdges.size());
      for(unsigned int i = 0; i < fEdges.size(); ++i)
        edges.SetBinContent(i+1, fEdges[i]);
      edges.Write("edges");
    }

    tmp->cd();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<Binning> Binning::LoadFrom(TDirectory* dir)
  {
    TObjString* tag = (TObjString*)dir->Get("type");
    assert(tag);
    assert(tag->GetString() == "Binning");

    TH1D* hMinMax = (TH1D*)dir->Get("nminmax");
    assert(hMinMax);

    std::unique_ptr<Binning> ret(new Binning(Binning::Simple(hMinMax->GetBinContent(1),
                                                             hMinMax->GetBinContent(2),
                                                             hMinMax->GetBinContent(3))));

    TH1* hEdges = (TH1*)dir->Get("edges");
    if(hEdges){
      for(int i = 0; i < hEdges->GetNbinsX(); ++i)
        ret->fEdges.push_back(hEdges->GetBinContent(i+1));
    }

    return ret;
  }
}
