#include "CAFAna/ExposureAssumptions.h"

#include <algorithm>
#include <cassert>

#include "TAxis.h"
#include "TCanvas.h"
#include "TGaxis.h"
#include "TGraph.h"
#include "TH2.h"
#include "TPad.h"

namespace ana
{
  //----------------------------------------------------------------------
  ExposureAssumptions::ExposureAssumptions() : fDirtyPM(true), fDirtyExp(true)
  {
  }

  //----------------------------------------------------------------------
  void ExposureAssumptions::AddMassPt(TTimeStamp t, double m)
  {
    assert(fMassPts.empty() || t >= fMassPts.back().first);
    fMassPts.push_back(std::make_pair(t, m));
    fDirtyPM = fDirtyExp = true;
  }

  //----------------------------------------------------------------------
  void ExposureAssumptions::AddPowerPt(TTimeStamp t, double m)
  {
    assert(fPowerPts.empty() || t >= fPowerPts.back().first);
    fPowerPts.push_back(std::make_pair(t, m));
    fDirtyPM = fDirtyExp = true;
  }

  //----------------------------------------------------------------------
  double ExposureAssumptions::MassAt(TTimeStamp t) const
  {
    assert(!fMassPts.empty());
    if(t <= fMassPts.front().first) return fMassPts.front().second;
    if(t >= fMassPts.back().first) return fMassPts.back().second;

    for(unsigned int i = 0; i+1 < fMassPts.size(); ++i){
      const TTimeStamp t0 = fMassPts[i].first;
      const TTimeStamp t1 = fMassPts[i+1].first;
      if(t >= t0 && t <= t1){
        const double m0 = fMassPts[i].second;
        const double m1 = fMassPts[i+1].second;

        return m0+(t-t0)*(m1-m0)/(t1-t0);
      }
    }
    assert(0 && "not reached");
  }

  //----------------------------------------------------------------------
  double ExposureAssumptions::PowerAt(TTimeStamp t) const
  {
    assert(!fPowerPts.empty());
    if(t <= fPowerPts.front().first) return fPowerPts.front().second;
    if(t >= fPowerPts.back().first) return fPowerPts.back().second;

    for(unsigned int i = 0; i+1 < fPowerPts.size(); ++i){
      const TTimeStamp t0 = fPowerPts[i].first;
      const TTimeStamp t1 = fPowerPts[i+1].first;
      if(t >= t0 && t <= t1){
        const double p0 = fPowerPts[i].second;
        const double p1 = fPowerPts[i+1].second;

        return p0+(t-t0)*(p1-p0)/(t1-t0);
      }
    }
    assert(0 && "not reached");
  }

  //----------------------------------------------------------------------
  std::vector<std::pair<TTimeStamp, double>> ExposureAssumptions::
  PowerMassPts() const
  {
    assert(!fMassPts.empty() && !fPowerPts.empty());

    if(!fDirtyPM) return fPowerMassPts;
    fDirtyPM = false;

    std::vector<TTimeStamp> ts;
    for(unsigned int i = 0; i < fMassPts.size(); ++i)
      ts.push_back(fMassPts[i].first);
    for(unsigned int i = 0; i < fPowerPts.size(); ++i)
      ts.push_back(fPowerPts[i].first);

    std::sort(ts.begin(), ts.end());
    assert(!ts.empty());

    fPowerMassPts.clear();

    for(unsigned int i = 1; i < ts.size(); ++i){
      const int kNumSteps = 10;
      const double dt = (ts[i]-ts[i-1])/kNumSteps;
      if(dt == 0) continue;
      for(int j = 0; j <= kNumSteps; ++j){
        const double t = ts[i-1]+j*dt;

        const double pm = PowerAt(t)*MassAt(t)/1000; // Convert to MW

        fPowerMassPts.push_back(std::make_pair(t, pm));
      } // end for j
    } // end for i

    return fPowerMassPts;
  }

  //----------------------------------------------------------------------
  double ExposureAssumptions::PowerMassAt(TTimeStamp t) const
  {
    const std::vector<std::pair<TTimeStamp, double>> pms = PowerMassPts();

    if(t <= pms.front().first) return pms.front().second;
    if(t >= pms.back().first) return pms.back().second;

    for(unsigned int i = 0; i+1 < pms.size(); ++i){
      const TTimeStamp t0 = pms[i].first;
      const TTimeStamp t1 = pms[i+1].first;
      if(t >= t0 && t <= t1){
        const double pm0 = pms[i].second;
        const double pm1 = pms[i+1].second;

        return pm0+(t-t0)*(pm1-pm0)/(t1-t0);
      }
    }
    assert(0 && "not reached");
  }

  //----------------------------------------------------------------------
  std::vector<std::pair<TTimeStamp, double>> ExposureAssumptions::
  ExposurePts() const
  {
    if(!fDirtyExp) return fExposurePts;
    fDirtyExp = false;

    const std::vector<std::pair<TTimeStamp, double>> pms = PowerMassPts();

    fExposurePts.clear();

    fExposurePts.push_back(std::make_pair(pms[0].first, 0));

    const double kSecPerYear = 365.25*24*60*60;

    double integ = 0;
    for(unsigned int i = 1; i < pms.size(); ++i){
      const int kNumSteps = 10;
      const double dt = (pms[i].first-pms[i-1].first)/kNumSteps;
      for(int j = 0; j < kNumSteps; ++j){
        const double t0 = pms[i-1].first+j*dt;
        const double t1 = pms[i-1].first+(j+1)*dt;

        const double pm0 = PowerMassAt(t0);
        const double pm1 = PowerMassAt(t1);

        // Parallelogram rule
        integ += dt*(pm0+pm1)/2;

        fExposurePts.push_back(std::make_pair(t1, integ/kSecPerYear));
      } // end for j
    } // end for i

    return fExposurePts;
  }

  //----------------------------------------------------------------------
  double ExposureAssumptions::ExposureAt(TTimeStamp t) const
  {
    const std::vector<std::pair<TTimeStamp, double>> exps = ExposurePts();

    if(t <= exps.front().first) return exps.front().second;
    if(t >= exps.back().first) return exps.back().second;

    for(unsigned int i = 0; i+1 < exps.size(); ++i){
      const TTimeStamp t0 = exps[i].first;
      const TTimeStamp t1 = exps[i+1].first;
      if(t >= t0 && t <= t1){
        const double exp0 = exps[i].second;
        const double exp1 = exps[i+1].second;

        return exp0+(t-t0)*(exp1-exp0)/(t1-t0);
      }
    }
    assert(0 && "not reached");
  }

  //----------------------------------------------------------------------
  double ExposureAssumptions::ScaleFactor(TTimeStamp t,
                                          double power,
                                          double mass,
                                          double years) const
  {
    const double exposure = power*mass*years/1000; // convert to MW
    double ret = ExposureAt(t)/exposure;
    return ret;
  }

  //----------------------------------------------------------------------
  TGraph* ExposureAssumptions::MassGraph() const
  {
    TGraph* ret = new TGraph;
    for(unsigned int i = 0; i < fMassPts.size(); ++i){
      ret->SetPoint(i, fMassPts[i].first, fMassPts[i].second);
    }
    return ret;
  }

  //----------------------------------------------------------------------
  TGraph* ExposureAssumptions::PowerGraph() const
  {
    TGraph* ret = new TGraph;
    for(unsigned int i = 0; i < fPowerPts.size(); ++i){
      ret->SetPoint(i, fPowerPts[i].first, fPowerPts[i].second);
    }
    return ret;
  }

  //----------------------------------------------------------------------
  TGraph* ExposureAssumptions::PowerMassGraph() const
  {
    TGraph* ret = new TGraph;
    const std::vector<std::pair<TTimeStamp, double>> lums = PowerMassPts();
    for(unsigned int i = 0; i < lums.size(); ++i){
      ret->SetPoint(i, lums[i].first, lums[i].second);
    }
    return ret;
  }

  //----------------------------------------------------------------------
  TGraph* ExposureAssumptions::ExposureGraph() const
  {
    TGraph* ret = new TGraph;
    const std::vector<std::pair<TTimeStamp, double>> exps = ExposurePts();
    for(unsigned int i = 0; i < exps.size(); ++i){
      ret->SetPoint(i, exps[i].first, exps[i].second);
    }
    return ret;
  }

  //----------------------------------------------------------------------
  void ExposureAssumptions::DrawMassPower() const
  {
    const TTimeStamp kMinT(2013, 1, 1, 0, 0, 0);
    const TTimeStamp kMaxT(2016, 1, 1, 0, 0, 0);

    const double kMaxMass = 20;
    const double kMaxPow = 800;

    const int kLineWidth = 4;

    const int kPowerColor = 46; // dull red
    const int kMassColor  =  9; // dull blue

    TH2* mtaxes = new TH2F("", "NO#nuA early reach. Exposure assumptions;;mass (kton)", 100, kMinT, kMaxT, 100, 0, kMaxMass);
    mtaxes->Draw();

    TAxis* tax = mtaxes->GetXaxis();
    tax->SetTimeOffset(0);
    tax->SetTimeDisplay(1);
    tax->SetTimeFormat("%b %Y");

    TAxis* max = mtaxes->GetYaxis();
    max->SetTitleColor(kMassColor);
    max->SetLabelColor(kMassColor);
    max->CenterTitle();

    TGaxis* pax = new TGaxis(kMaxT, 0, kMaxT, kMaxMass, 0, kMaxPow, 510, "+L");
    pax->SetTitle("Beam power (kW)");
    pax->SetTitleColor(kPowerColor);
    pax->SetLabelColor(kPowerColor);
    pax->CenterTitle();
    pax->Draw();

    TGraph* gm = MassGraph();
    gm->SetLineWidth(kLineWidth);
    gm->SetLineColor(kMassColor);
    gm->Draw("l same");

    TGraph* gp = PowerGraph();
    for(int i = 0; i < gp->GetN(); ++i){
      double x, y;
      gp->GetPoint(i, x, y);
      gp->SetPoint(i, x, kMaxMass/kMaxPow*y);
    }
    gp->SetLineWidth(kLineWidth);
    gp->SetLineColor(kPowerColor);
    gp->Draw("l same");

    gPad->SetGrid();
  }

  //----------------------------------------------------------------------
  void ExposureAssumptions::DrawExposure() const
  {
    const TTimeStamp kMinT(2013, 1, 1, 0, 0, 0);
    const TTimeStamp kMaxT(2016, 1, 1, 0, 0, 0);

    const double kMaxMP = 12;
    const double kMaxExposure = 24;

    const int kLineWidth = 4;

    const int kMPColor       = 46; // brown
    const int kExposureColor = 37; // gray

    TH2* mptaxes = new TH2F("", "NO#nuA early reach. Exposure assumptions;;mass #times beam power (kton MW)", 100, kMinT, kMaxT, 100, 0, kMaxMP);
    mptaxes->Draw();
    TAxis* tax = mptaxes->GetXaxis();
    tax->SetTimeOffset(0);
    tax->SetTimeDisplay(1);
    tax->SetTimeFormat("%b %Y");
    TAxis* mpax = mptaxes->GetYaxis();
    mpax->SetTitleColor(kMPColor);
    mpax->SetLabelColor(kMPColor);
    mpax->CenterTitle();

    TGaxis* eax = new TGaxis(kMaxT, 0, kMaxT, kMaxMP, 0, kMaxExposure, 510, "+L");
    eax->SetTitle("Integrated exposure (kton MW yr)");
    eax->SetTitleColor(kExposureColor);
    eax->SetLabelColor(kExposureColor);
    eax->CenterTitle();
    eax->Draw();

    TGraph* lg = PowerMassGraph();
    lg->SetLineWidth(kLineWidth);
    lg->SetLineColor(kMPColor);
    lg->Draw("l same");

    TGraph* eg = ExposureGraph();
    for(int i = 0; i < eg->GetN(); ++i){
      double x, y;
      eg->GetPoint(i, x, y);
      eg->SetPoint(i, x, kMaxMP/kMaxExposure*y);
    }
    eg->SetLineWidth(kLineWidth);
    eg->SetLineColor(kExposureColor);
    eg->Draw("l same");

    gPad->SetGrid();
  }

  //----------------------------------------------------------------------
  void ExposureAssumptions::DrawAll() const
  {
    TVirtualPad* pad = gPad ? gPad : new TCanvas;
    pad->Divide(1, 2);
    pad->cd(1);
    DrawMassPower();
    pad->cd(2);
    DrawExposure();
    pad->cd();
  }

  //----------------------------------------------------------------------
  NominalExposureAssumptions::NominalExposureAssumptions()
  {
    //   Mass:
    // - 2.5 kton instrumented on Aug 1
    // - rate of increase after Jul 1 = 0.83 kton / month
    // - rate of increase before Jul 1 is such that the x-intercept is mid-April

    //   Beam:
    // - turn-on at 30 kW on Aug 1
    // - ramp-up to 300 kW from Aug 1 to Sep 1
    // - ramp-up to 500 kW from Sep 1 to Feb 1
    // - continue ramping at same slope until 700 kW

    AddMassPt(TTimeStamp(2013, 4, 15, 0, 0, 0), 0);
    AddMassPt(TTimeStamp(2013, 8,  1, 0, 0, 0), 2.5);
    AddMassPt(TTimeStamp(2014, 9, 26, 0, 0, 0), 14);
    AddMassPt(TTimeStamp(2037, 3,  1, 0, 0, 0), 14);

    AddPowerPt(TTimeStamp(2013,  8, 1, 0, 0, 0), 0);
    AddPowerPt(TTimeStamp(2013,  8, 1, 0, 0, 0), 30);
    AddPowerPt(TTimeStamp(2013,  9, 1, 0, 0, 0), 300);
    //    AddPowerPt(TTimeStamp(2014, 2, 1, 0, 0, 0), 500);
    AddPowerPt(TTimeStamp(2014,  7, 1, 0, 0, 0), 700);
    AddPowerPt(TTimeStamp(2037,  1, 1, 0, 0, 0), 700);
  }
}
