#include "ExtrapUtil.h"

#include <cassert>

#include "CAFAna/Binning.h"

#include "TAxis.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TH2.h"
#include "TStyle.h"

namespace ana
{
  //----------------------------------------------------------------------
  TH1* MultiplyWithStats(TH1* h1, TH1* h2, double c1, double c2, std::string opt)
  {
    TH1* ret = (TH1*)h1->Clone(); // Create return histogram with correct features
    if(ret->GetSumw2N() == 0) ret->Sumw2(); // Turn on errors

    // Multiply two histograms with scale factors, break if error occurs
    if(!ret->Multiply(h1, h2, c1, c2)) {
      bool compatible_hists = false;
      assert(compatible_hists);
    }

    // Variables for readability
    const int nbins = h1->GetNbinsX();
    double bin1, bin2, err1, err2, result;

    for(int i=0; i <= nbins+1; i++) {
      bin1 = h1->GetBinContent(i);
      bin2 = h2->GetBinContent(i);
      result = ret->GetBinContent(i);

      if(opt == "1off") // Set err1 to 0 if option "1off" is specified
        err1 = 0;
      else // If stats not already turned on, err1 is the square root of the bin
        err1 = ( (h1->GetSumw2N() == 0) ? sqrt(bin1) : h1->GetBinError(i) );

      if(opt == "2off")
        err2 = 0;
      else
        err2 = ( (h2->GetSumw2N() == 0) ? sqrt(bin2) : h2->GetBinError(i) );

      // For C = A*B, err_c = C*[(err_A/A)^2 + (err_B/B)^2]^0.5, ignoring correlation
      if(result != 0)
        ret->SetBinError(i, result*sqrt(err1*err1/(bin1*bin1) + err2*err2/(bin2*bin2)) );
      else
        ret->SetBinError(i, 0);
    }

    return ret;
  }

  //----------------------------------------------------------------------
  TH1* DivideWithStats(TH1* hn, TH1* hd, double cn, double cd, std::string opt)
  {
    TH1* ret = (TH1*)hn->Clone();
    if(ret->GetSumw2N() == 0) ret->Sumw2();

    if(!ret->Divide(hn, hd, cn, cd)) {
      bool compatible_hists = false;
      assert(compatible_hists);
    }

    const int nbins = hn->GetNbinsX();
    double binN, binD, errN, errD, result;

    for(int i=0; i <= nbins+1; i++) {
      binN = hn->GetBinContent(i);
      binD = hd->GetBinContent(i);
      result = ret->GetBinContent(i);

      if(opt == "1off")
        errN = 0;
      else
        errN = ( (hn->GetSumw2N() == 0) ? sqrt(binN) : hn->GetBinError(i) );

      if(opt == "2off")
        errD = 0;
      else
        errD = ( (hd->GetSumw2N() == 0) ? sqrt(binD) : hd->GetBinError(i) );

      if(result != 0)
        ret->SetBinError(i, result*sqrt(errN*errN/(binN*binN) + errD*errD/(binD*binD)) );
      else
        ret->SetBinError(i, 0);
    }

    return ret;
  }

  //----------------------------------------------------------------------
  TH2* RecoToTrueMatrix(OscillatableSpectrum ospecrt, bool stat_errors)
  {
    TH2* rt = ospecrt.ToTH2(ospecrt.Unoscillated().POT());
    if(stat_errors) 
      if(rt->GetSumw2N() == 0) rt->Sumw2();

    TH1D* RecoScale = rt->ProjectionX(); // Total number of events within given reco energy column
    if(stat_errors)
      if(RecoScale->GetSumw2N() == 0) RecoScale->Sumw2();

    const int X = rt->GetNbinsX();
    const int Y = rt->GetNbinsY();

    // Normalize each column (for given reco energy, all true energy bins add to 1)
    for(int x = 0; x < X+2; x++) {
      for(int y = 0; y < Y+2; y++) {
        if(rt->GetBinContent(x,y) != 0 && RecoScale->GetBinContent(x) != 0)
          rt->SetBinContent(x,y,rt->GetBinContent(x,y)/RecoScale->GetBinContent(x));
        if(!stat_errors) rt->SetBinError(x,y,0);
      }
    }

    return rt;
  }

  //----------------------------------------------------------------------
  TH2* RecoToTrueMatrix(TH2* rt, bool stat_errors)
  {
    if(stat_errors)
      if(rt->GetSumw2N() == 0) rt->Sumw2();

    TH1D* RecoScale = rt->ProjectionX(); // Total number of events within given reco energy column
    if(stat_errors)
      if(RecoScale->GetSumw2N() == 0) RecoScale->Sumw2();

    const int X = rt->GetNbinsX();
    const int Y = rt->GetNbinsY();

    // Normalize each column (for given reco energy, all true energy bins add to 1)
    for(int x = 0; x < X+2; x++) {
      for(int y = 0; y < Y+2; y++) {
        if(rt->GetBinContent(x,y) != 0 && RecoScale->GetBinContent(x) != 0)
          rt->SetBinContent(x,y,rt->GetBinContent(x,y)/RecoScale->GetBinContent(x));
        if(!stat_errors) rt->SetBinError(x,y,0);
      }
    }

    return rt;
  }

  //----------------------------------------------------------------------
  TH2* TrueToRecoMatrix(OscillatableSpectrum ospectr, bool stat_errors)
  {
    TH2* tr = ospectr.ToTH2(ospectr.Unoscillated().POT());
    if(stat_errors)
      if(tr->GetSumw2N() == 0) tr->Sumw2();

    TH1D* TrueScale = tr->ProjectionY(); // Total number of events within given true energy row
    if(stat_errors)
      if(TrueScale->GetSumw2N() == 0) TrueScale->Sumw2();

    const int X = tr->GetNbinsX();
    const int Y = tr->GetNbinsY();

    // Normalize each row (for given true energy, all reco energy bins add to 1)
    for(int x = 0; x < X+2; x++) {
      for(int y = 0; y < Y+2; y++) {
        if(tr->GetBinContent(x,y) != 0 && TrueScale->GetBinContent(y) != 0)
          tr->SetBinContent(x,y,tr->GetBinContent(x,y)/TrueScale->GetBinContent(y));
        if(!stat_errors) tr->SetBinError(x,y,0);
      }
    }

    return tr;
  }

  //----------------------------------------------------------------------
  TH2* TrueToRecoMatrix(TH2* tr, bool stat_errors)
  {
    if(stat_errors)
      if(tr->GetSumw2N() == 0) tr->Sumw2();

    TH1D* TrueScale = tr->ProjectionY(); // Total number of events within given true energy row
    if(stat_errors)
      if(TrueScale->GetSumw2N() == 0) TrueScale->Sumw2();

    const int X = tr->GetNbinsX();
    const int Y = tr->GetNbinsY();

    // Normalize each row (for given true energy, all reco energy bins add to 1)
    for(int x = 0; x < X+2; x++) {
      for(int y = 0; y < Y+2; y++) {
        if(tr->GetBinContent(x,y) != 0 && TrueScale->GetBinContent(y) != 0)
          tr->SetBinContent(x,y,tr->GetBinContent(x,y)/TrueScale->GetBinContent(y));
        if(!stat_errors) tr->SetBinError(x,y,0);
      }
    }

    return tr;
  }

  //----------------------------------------------------------------------
  void MigrateToTrue(TH1* hreco, TH2* hrt, TH1* htrue, bool stat_errors)
  {
    assert(hreco->GetNbinsX() == hrt->GetNbinsX());
    assert(htrue->GetNbinsX() == hrt->GetNbinsY());

    for(int itrue = 0; itrue <= htrue->GetNbinsX(); itrue++) {
      htrue->SetBinContent(itrue,0);
      if(stat_errors)
        htrue->SetBinError(itrue,0);
      
      for(int ireco = 0; ireco <= hreco->GetNbinsX(); ireco++) {
        htrue->SetBinContent(itrue, htrue->GetBinContent(itrue) 
                                   +hreco->GetBinContent(ireco)*((hrt->GetBinContent(ireco,itrue)>0) ? hrt->GetBinContent(ireco,itrue) : 0));
      if(stat_errors)
        htrue->SetBinError(itrue, htrue->GetBinError(itrue)
                                 +hreco->GetBinError(ireco)*hreco->GetBinError(ireco)*hrt->GetBinContent(ireco,itrue)*hrt->GetBinContent(ireco,itrue)
                                 +hreco->GetBinContent(ireco)*hreco->GetBinContent(ireco)*hrt->GetBinError(ireco,itrue)*hrt->GetBinError(ireco,itrue));
      }

      if(stat_errors) htrue->SetBinError(itrue, sqrt(htrue->GetBinError(itrue)));
    }

    return;
  }

  //----------------------------------------------------------------------
  void MigrateToReco(TH1* htrue, TH2* htr, TH1* hreco, bool stat_errors)
  {
    assert(htrue->GetNbinsX() == htr->GetNbinsY());
    assert(hreco->GetNbinsX() == htr->GetNbinsX());

    for(int ireco = 0; ireco <= hreco->GetNbinsX(); ireco++) {
      hreco->SetBinContent(ireco,0);
      if(stat_errors)
        hreco->SetBinError(ireco,0);

      for(int itrue = 0; itrue <= htrue->GetNbinsX(); itrue++) {
        hreco->SetBinContent(ireco, hreco->GetBinContent(ireco) 
                                   +htrue->GetBinContent(itrue)*((htr->GetBinContent(ireco,itrue)>0) ? htr->GetBinContent(ireco,itrue) : 0));
        if(stat_errors)
          hreco->SetBinError(ireco, hreco->GetBinError(ireco)
                                   +htrue->GetBinError(itrue)*htrue->GetBinError(itrue)*htr->GetBinContent(ireco,itrue)*htr->GetBinContent(ireco,itrue)
                                   +htrue->GetBinContent(itrue)*htrue->GetBinContent(itrue)*htr->GetBinError(ireco,itrue)*htr->GetBinError(ireco,itrue));
      }

      if(stat_errors) hreco->SetBinError(ireco, sqrt(hreco->GetBinError(ireco)));
    }

    return;
  }

  //----------------------------------------------------------------------
  TH1* PurityOrEfficiency(Spectrum numerator, Spectrum denominator, bool stat_errors)
  {
    TH1* num =   numerator.ToTH1(  numerator.POT());
    TH1* den = denominator.ToTH1(denominator.POT());
    assert( num->GetNbinsX() == den->GetNbinsX() );

    TH1* ret = (TH1D*)num->Clone(""); // Create histogram with same attributes as inputs

    if(stat_errors) {
      if(ret->GetSumw2N() == 0) ret->Sumw2();
      if(num->GetSumw2N() == 0) num->Sumw2();
      if(den->GetSumw2N() == 0) den->Sumw2();
      den->Scale(numerator.POT()/denominator.POT());
      ret->Divide(num, den, 1, 1, "B"); // Use binary errors
    }
    else ret->Divide(num, den, denominator.POT(), numerator.POT());

    return ret;
  }

  //----------------------------------------------------------------------
  TH1* PurityOrEfficiency(OscillatableSpectrum numerator, OscillatableSpectrum denominator, bool stat_errors)
  {
    TH1* num =   numerator.TrueEnergy().ToTH1(   numerator.Unoscillated().POT() );
    TH1* den = denominator.TrueEnergy().ToTH1( denominator.Unoscillated().POT() );
    assert( num->GetNbinsX() == den->GetNbinsX() );

    TH1* ret = new TH1D("",";Energy (GeV);Events",
                        kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);

    if(stat_errors) {
      if(ret->GetSumw2N() == 0) ret->Sumw2();
      if(num->GetSumw2N() == 0) num->Sumw2();
      if(den->GetSumw2N() == 0) den->Sumw2();
      den->Scale(numerator.Unoscillated().POT()/denominator.Unoscillated().POT());
      ret->Divide(num, den, 1, 1, "B"); // Use binary errors
    }
    else ret->Divide(num, den, denominator.Unoscillated().POT(), numerator.Unoscillated().POT());

    return ret;
  }

  //----------------------------------------------------------------------
  TH1* PurityOrEfficiency(TH1* num, TH1* den, bool stat_errors)
  {
    assert( num->GetNbinsX() == den->GetNbinsX() );

    TAxis* ax = num->GetXaxis();
    int nbins = num->GetNbinsX();
    double xmin = ax->GetBinLowEdge(1);
    double xmax = ax->GetBinUpEdge(nbins);
    TH1* ret = new TH1D("",";Reconstructed Energy (GeV);Events",nbins,xmin,xmax);

    if(stat_errors) {
      if(ret->GetSumw2N() == 0) ret->Sumw2();
      if(num->GetSumw2N() == 0) num->Sumw2();
      if(den->GetSumw2N() == 0) den->Sumw2();
      ret->Divide(num, den, 1, 1, "B"); // Use binary errors
    }
    else ret->Divide(num, den);

    return ret;
  }

  //----------------------------------------------------------------------
  TH1* FarOverNearRatio(Spectrum sFar, Spectrum sNear, bool stat_errors)
  {
    TH1* hf =  sFar.ToTH1( sFar.POT());
    TH1* hn = sNear.ToTH1(sNear.POT());
    assert( hf->GetNbinsX() == hn->GetNbinsX() );

    TH1* ret = (TH1D*)hf->Clone(""); // Create histogram with same attributes as inputs
    ret->SetYTitle("F/N Ratio");

    if(stat_errors) {
      if(ret->GetSumw2N() == 0) ret->Sumw2();
      if(hf->GetSumw2N() == 0)   hf->Sumw2();
      if(hn->GetSumw2N() == 0)   hn->Sumw2();

      ret = DivideWithStats(hf, hn, sNear.POT(), sFar.POT());
    }
    else
      ret->Divide(hf, hn, sNear.POT(), sFar.POT());

    return ret;
  }

}
