#include "CAFAna/FCBin.h"

#include "CAFAna/FCPoint.h"

#include <algorithm>
#include <cassert>

#include "TH1.h"
#include "TTree.h"

// Anonymous namespace, no-one else needs to see these
namespace{
  int gNVals;
  std::vector<float> gVals;
}

namespace ana
{
  //----------------------------------------------------------------------
  void FCBin::Add(const FCPoint& pt)
  {
    fVals.push_back(pt.Up());
  }

  //----------------------------------------------------------------------
  double FCBin::UpValueForSignificance(double sig)
  {
    assert(sig > 0 && sig < 1);

    std::sort(fVals.begin(), fVals.end());

    const int N = fVals.size();

    const int idx = sig*(N+1)-1;
    // Bin must have enough points for this significance
    assert(idx >= 0);
    assert(idx < N-1);

    // Up values for the points either side of this significance
    const double up0 = fVals[idx];
    const double up1 = fVals[idx+1];
    // What significance taking the points themselves would be worth. Consider
    // some small-N cases. If there is only one point it represents the 50%
    // mark. If there are two, 50% is halfway between them, and they are 33%
    // and 67%.
    const double sig0 = double(idx+1)/(N+1);
    const double sig1 = double(idx+2)/(N+1);
    const double dsig = 1./(N+1); //sig1-sig0;

    // Make sure we picked the right points
    assert(sig >= sig0 && sig <= sig1);

    // Linear interpolation
    return (up0*(sig1-sig)+up1*(sig-sig0))/dsig;
  }

  //----------------------------------------------------------------------
  void FCBin::Draw() const
  {
    TH1* h = new TH1F("", ";#Delta#chi^{2};Experiments", 200, 0, 10);
    for(float v: fVals) h->Fill(v);
    h->Draw();

    /*
    double* integ = h->GetIntegral();
    for(int n = 0; n < 102; ++n) h->SetBinContent(n, integ[n]);
    */
  }

  //----------------------------------------------------------------------
  void FCBin::InitToTree(TTree* tr)
  {
    tr->Branch("nvals", &gNVals);
    tr->Branch("vals", &gVals.front(), "vals[nvals]/F");
  }

  //----------------------------------------------------------------------
  void FCBin::ToTree(TTree* tr) const
  {
    gNVals = fVals.size();

    gVals = fVals;
    tr->SetBranchAddress("vals", &gVals.front());

    tr->Fill();
  }

  //----------------------------------------------------------------------
  void FCBin::InitFromTree(TTree* tr)
  {
    tr->SetBranchAddress("nvals", &gNVals);

    // It's a shape we have to run through the tree upfront to determine this,
    // but TTree doesn't seem to like having the branch address changed
    // mid-entry as you'd have to do otherwise.
    int maxN = 0;
    for(int i = 0; i < tr->GetEntries(); ++i){
      tr->GetEntry(i);
      maxN = std::max(maxN, gNVals);
    }

    gVals.resize(maxN);
    tr->SetBranchAddress("vals", &gVals.front());
  }

  //----------------------------------------------------------------------
  FCBin FCBin::FromTree(TTree* tr, int idx)
  {
    tr->GetEntry(idx);

    assert(gNVals <= int(gVals.size()));

    FCBin ret;
    ret.fVals = std::vector<float>(gVals.begin(), gVals.begin()+gNVals);
    return ret;
  }
} // namespace
