#include "CAFAna/FCSurface.h"

#include "CAFAna/FCPoint.h"
#include "CAFAna/Progress.h"
#include "CAFAna/Utilities.h"

#include "TCanvas.h"
#include "TFile.h"
#include "TH2.h"
#include "TTree.h"
#include "TVirtualPad.h"

#include <cassert>
#include <cmath>

namespace ana
{
  //----------------------------------------------------------------------
  FCSurface::FCSurface(int nbinsx, double xmin, double xmax,
                       int nbinsy, double ymin, double ymax)
  {
    DontAddDirectory guard;

    fBinning = ExpandedHistogram("", nbinsx, xmin, xmax, nbinsy, ymin, ymax);

    fBins.resize((nbinsx+2)*(nbinsy+2));
  }

  //----------------------------------------------------------------------
  FCSurface::~FCSurface()
  {
    delete fBinning;
  }

  //----------------------------------------------------------------------
  void FCSurface::Add(const FCPoint& pt)
  {
    const int bin = fBinning->FindBin(pt.TrueX(), pt.TrueY());
    fBins[bin].Add(pt);
  }

  //----------------------------------------------------------------------
  void FCSurface::Add(const FCCollection& coln)
  {
    Progress prog("Filling FCSurface from collection");
    const unsigned int N = coln.NPoints();
    for(unsigned int n = 0; n < N; ++n){
      if(n%100000 == 0) prog.SetProgress(double(n)/N);
      Add(coln.Point(n));
    }
    prog.Done();
  }

  //----------------------------------------------------------------------
  TH2* FCSurface::SurfaceForSignificance(double sig)
  {
    DontAddDirectory guard;

    TH2* hRet = new TH2F(*fBinning);
    hRet->SetName(UniqueName().c_str());

    // Don't do overflow and underflow
    for(int x = 1; x < hRet->GetNbinsX()+1; ++x){
      for(int y = 1; y < hRet->GetNbinsY()+1; ++y){
        const int bin = hRet->GetBin(x, y);
        if(fBins[bin].Empty()) continue;
        const double up = fBins[bin].UpValueForSignificance(sig);
        assert(!std::isinf(up) && !std::isnan(up));
        hRet->SetBinContent(bin, up);
      }
    }

    hRet->SetMinimum(0);

    return hRet;
  }

  //----------------------------------------------------------------------
  void FCSurface::DrawBins()
  {
    TVirtualPad* c = gPad;
    if(!c) c = new TCanvas;

    const int X = fBinning->GetNbinsX();
    const int Y = fBinning->GetNbinsY();
    c->Divide(X, Y);

    for(int x = 1; x < X+1; ++x){
      for(int y = 1; y < Y+1; ++y){
        c->cd(x+(y-1)*X);
        const int bin = fBinning->GetBin(x, y);
        fBins[bin].Draw();
      }
    }
  }

  //----------------------------------------------------------------------
  const FCBin* FCSurface::GetBin(int x, int y) const
  {
    return &fBins[fBinning->GetBin(x, y)];
  }

  //----------------------------------------------------------------------
  void FCSurface::SaveToFile(const std::string& fname) const
  {
    TFile fout(fname.c_str(), "RECREATE");
    fBinning->Write("bins");

    TTree tr("fcsurf", "fcsurf");
    FCBin::InitToTree(&tr);
    for(const FCBin& bin : fBins) bin.ToTree(&tr);
    tr.Write();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<FCSurface> FCSurface::LoadFromFile(const std::string& fname)
  {
    DontAddDirectory guard;

    std::unique_ptr<FCSurface> ret(new FCSurface);

    TFile fin(fname.c_str());
    assert(!fin.IsZombie());

    ret->fBinning = (TH2F*)(fin.Get("bins")->Clone());

    TTree* tr = (TTree*)fin.Get("fcsurf");
    assert(tr);

    FCBin::InitFromTree(tr);

    for(int i = 0; i < tr->GetEntries(); ++i)
      ret->fBins.push_back(FCBin::FromTree(tr, i));

    return ret;
  }

} // namespace
