#include "CAFAna/FileReducer.h"

#include "CAFAna/Progress.h"

#include "StandardRecord/StandardRecord.h"

#include <cassert>
#include <iostream>

#include <fenv.h>

#include "TFile.h"
#include "TH2.h"
#include "TTree.h"


namespace ana
{
  //----------------------------------------------------------------------
  FileReducer::FileReducer(const std::string& wildcard,
                           const std::string& outfile)
    : SpectrumLoaderBase(wildcard, kBeam), fOutfile(outfile), fEventCut(nullptr), fOmitBadFiles(true)
  {
  }

  //----------------------------------------------------------------------
  FileReducer::FileReducer(const std::vector<std::string>& fnames,
                           const std::string& outfile)
    : SpectrumLoaderBase(fnames, kBeam), fOutfile(outfile), fEventCut(nullptr), fOmitBadFiles(true)
  {
  }

  //----------------------------------------------------------------------
  FileReducer::~FileReducer()
  {
    if(fEventCut) delete fEventCut;
  }

  //----------------------------------------------------------------------
  void FileReducer::AddRequiredVar(const Var& var)
  {
    RequireVariables(var.Requires());
  }

  //----------------------------------------------------------------------
  void FileReducer::AddRequiredCut(const Cut& cut)
  {
    RequireVariables(cut.Requires());
  }

  //----------------------------------------------------------------------
  void FileReducer::AddEventCut(const Cut& cut)
  {
    if(fEventCut){
      *fEventCut = *fEventCut && cut;
    }
    else{
      fEventCut = new Cut(cut);
    }
  }

  //----------------------------------------------------------------------
  void FileReducer::SetEventList(const std::string& fname)
  {
    FILE* f = fopen(fname.c_str(), "r");
    assert(f);

    while(!feof(f)){
      int run, subrun, event;
      fscanf(f, "%d %d %d", &run, &subrun, &event);
      fEventList.insert(std::make_tuple(run, subrun, event));
    }

    fclose(f);
  }

  //----------------------------------------------------------------------
  void FileReducer::Go()
  {
    feenableexcept(FE_INVALID);

    if(fGone){
      std::cerr << "Error: can only call Go() once on a FileReducer" << std::endl;
      abort();
    }
    fGone = true;

    if(!fBranchNames.empty()){
      // If the user is requesting something we need to give them true energy
      // as well so that they can oscillate things.
      //
      // Don't do this if the list of requirements was blank because we need to
      // leave it blank so that the later stages detect that and propagate
      // everything instead.
      RequireVariable("mc.nu");
      RequireVariable("mc.nnu");
      RequireVariable("mc.nu.E");

      if(!fEventList.empty()){
        // We're going to need the header information so we can actually select
        // the requested events.
        RequireVariable("hdr.run");
        RequireVariable("hdr.subrun");
        RequireVariable("hdr.evt");
      }
    }

    if(fBranchNames.empty()){
      std::cout << "No specific branches requested. Copying all." << std::endl;
    }
    else{
      std::cout << "Active branches: ";
      for(const std::string& s: fBranchNames) std::cout << s << " ";
      std::cout << std::endl;
    }

    const unsigned int Nfiles = fFileNames.size();

    Progress prog(TString::Format("Filling %lu variables from %d files matching '%s'", fBranchNames.size(), Nfiles, fWildcard.c_str()).Data());

    TFile* fout = new TFile(fOutfile.c_str(), "RECREATE");
    TTree* trOut = new TTree("recTree", "recTree");
    caf::StandardRecord dummy;
    trOut->Branch("rec", &dummy);
    TH1* hPOTOut = new TH1F("TotalPOT", "", 1, 0, 1);
    TH1* hEventsOut = new TH1F("TotalEvents", "", 1, 0, 1);
    TH1* hLivetimeOut = new TH1F("TotalLivetime", "", 1, 0, 1);

    caf::StandardRecord* oldsr = 0;

    for(unsigned int fileIdx = 0; fileIdx < Nfiles; ++fileIdx){
      TFile f(fFileNames[fileIdx].c_str());


      if(fOmitBadFiles){
        if(f.IsZombie()) continue;
      }
      else{
        assert(!f.IsZombie());
      }

      TH1* hPOT = (TH1*)f.Get("TotalPOT");
      if(fOmitBadFiles && !hPOT) continue;

      assert(hPOT);
      hPOTOut->Add(hPOT);

      TH1* hEvents = (TH1*)f.Get("TotalEvents");
      
      assert(hEvents);
      hEventsOut->Add(hEvents);
      
      TH1* hLivetime = (TH1*)f.Get("TotalLivetime");
      assert(hLivetime);
      hLivetimeOut->Add(hLivetime);

      TTree* tr = (TTree*)f.Get("recTree");
      assert(tr);

      BranchList branches(tr, fBranchNames);

      const int Nentries = tr->GetEntries();
      for(int n = 0; n < Nentries; ++n){
        caf::StandardRecord* sr = branches.GetEntry(n);

        // Apply EventList cut if it's been enabled
        if(!fEventList.empty() &&
           !fEventList.count(std::make_tuple(sr->hdr.run,
                                             sr->hdr.subrun,
                                             sr->hdr.evt))) continue;

        /// see if we want to omit the event
        if(!fEventCut || (*fEventCut)(sr)){
          if(sr != oldsr){
            trOut->SetBranchAddress("rec", &sr);
            oldsr = sr;
          }
          trOut->Fill();
        }

        if(n%1000 == 0 && Nfiles == 1)
          prog.SetProgress(double(n)/Nentries);
      } // end for n
      prog.SetProgress((fileIdx+1.)/Nfiles);
    } // end for fileIdx

    fout->cd();
    trOut->Write();
    hPOTOut->Write();
    hEventsOut->Write();
    hLivetimeOut->Write();

    prog.Done();
  }

} // namespace
