#include "CAFAna/Fit.h"

#include "CAFAna/Experiment.h"
#include "CAFAna/FitVars.h"
#include "CAFAna/Utilities.h"

#include "OscLib/func/IOscCalculator.h"
#include "Utilities/func/MathUtil.h"

#include "TError.h"
#include "TH1.h"

#include <cassert>
#include <memory>
#include <iostream>

namespace ana
{
  //----------------------------------------------------------------------
  double SimpleFOM(const Spectrum& obs, const Spectrum& unosc, double pot)
  {
    if(pot == 0) pot = obs.POT();

    TH1* oh = obs.ToTH1(pot);
    TH1* bh = unosc.ToTH1(pot);
    assert(oh->GetNbinsX() == bh->GetNbinsX());

    double fomSq = 0;

    // Combine s/sqrt(s+b) in quadrature between bins
    for(int i = 0; i < oh->GetNbinsX(); ++i){
      const double o = oh->GetBinContent(i);
      const double b = bh->GetBinContent(i);
      const double s = o-b;

      if(s <= 0) continue;

      fomSq += util::sqr(s)/(s+b);
    }

    return sqrt(fomSq);
  }

  //----------------------------------------------------------------------
  Fitter::Fitter(const Experiment* expt,
                 std::vector<const FitVar*> vars,
                 std::vector<Syst> systs)
    : fExpt(expt), fVars(vars), fSysts(systs)
  {
  }

  //----------------------------------------------------------------------
  double Fitter::Fit(osc::IOscCalculatorAdjustable* seed,
                     SystShifts& systSeed,
                     Verbosity verb) const
  {
    if(verb == kVerbose){
      std::cout << "Finding best fit for";
      for(const FitVar* v: fVars) std::cout << " " << v->ShortName();
      for(Syst s: fSysts) std::cout << " " << ShortName(s);
      std::cout << "..." << std::endl;
    }

    fCalc = seed;
    fShifts = systSeed;

    ROOT::Minuit2::MnUserParameters mnPars;

    for(const FitVar* v: fVars){
      const double val = v->GetValue(seed);
      // name, value, error
      mnPars.Add(v->ShortName(), val, val ? val/2 : .1);
    }
    // One way this can go wrong is if two variables have the same ShortName
    assert(mnPars.Params().size() == fVars.size());
    for(Syst s: fSysts){
      const double val = systSeed.GetShift(s);
      // name, value, error
      mnPars.Add(ShortName(s), val, 1);
    }
    // One way this can go wrong is if two variables have the same ShortName
    assert(mnPars.Params().size() == fVars.size()+fSysts.size());

    ROOT::Minuit2::MnMigrad mnMigrad(*this, mnPars);

    // Minuit2 doesn't give a good way to control verbosity...
    const int olderr = gErrorIgnoreLevel;
    if(verb == kQuiet) gErrorIgnoreLevel = 1001; // Ignore warnings
    ROOT::Minuit2::FunctionMinimum minpt = mnMigrad();
    gErrorIgnoreLevel = olderr;

    if(!minpt.IsValid())
      std::cout << "*** ERROR: minimum is not valid ***" << std::endl;

    const std::vector<double> minvec = minpt.UserParameters().Params();

    // Store results back to the "seed" variable
    for(unsigned int i = 0; i < fVars.size(); ++i){
      fVars[i]->SetValue(seed, minvec[i]);
    }
    // Store systematic results back into "systSeed"
    for(unsigned int j = 0; j < fSysts.size(); ++j){
      systSeed.SetShift(fSysts[j], minvec[fVars.size()+j]);
    }

    if(verb == kVerbose){
      std::cout << "Best fit";
      for(const FitVar* v: fVars){
        std::cout << ", " << v->ShortName() << " = " << v->GetValue(seed);
      }
      for(Syst s: fSysts){
        std::cout << ", " << ShortName(s) << " = " << systSeed.GetShift(s);
      }
      std::cout << ", LL = " << minpt.Fval() << std::endl;
    }

    return minpt.Fval();
  }

  //----------------------------------------------------------------------
  double Fitter::Fit(osc::IOscCalculatorAdjustable* seed,
                     Verbosity verb) const
  {
    SystShifts systs = SystShifts::Nominal();
    return Fit(seed, systs, verb);
  }

  //----------------------------------------------------------------------
  double Fitter::operator()(const std::vector<double>& pars) const
  {
    assert(pars.size() == fVars.size()+fSysts.size());

    double penalty = 0;
    for(unsigned int i = 0; i < fVars.size(); ++i){
      const double val = pars[i];
      penalty += fVars[i]->Penalty(val);
      fVars[i]->SetValue(fCalc, val);
    }
    for(unsigned int j = 0; j < fSysts.size(); ++j){
      const double val = pars[fVars.size()+j];
      fShifts.SetShift(fSysts[j], val);
    }

    return fExpt->ChiSq(fCalc, fShifts)+penalty+fShifts.Penalty();
  }

  //----------------------------------------------------------------------
  TH1* Slice(const Experiment* expt,
             osc::IOscCalculatorAdjustable* calc, const FitVar* v,
             int nbinsx, double minx, double maxx,
             double input_minchi)
  {
    TH1* ret = new TH1F(UniqueName().c_str(),
                        (";"+v->LatexName()+";#Delta#chi^{2}").c_str(),
                        nbinsx, minx, maxx);

    double minpos = 0;
    double minchi = 1e10;
    for(int n = 0; n < nbinsx; ++n){
      const double x = ret->GetXaxis()->GetBinCenter(n+1);
      v->SetValue(calc, x);

      const double chi = expt->ChiSq(calc);
      ret->Fill(x, chi);
      if(chi < minchi){
        minchi = chi;
        minpos = x;
      }
    }


    if(input_minchi == -1){
      Fitter fit(expt, {v});
      // Seed from best grid point
      v->SetValue(calc, minpos);
      minchi = fit.Fit(calc); // get a better value
    }
    else{
      minchi = input_minchi;
    }

    for(int n = 0; n < nbinsx; ++n) ret->AddBinContent(n+1, -minchi);

    return ret;
  }

  //----------------------------------------------------------------------
  TH1* SqrtSlice(const Experiment* expt,
                 osc::IOscCalculatorAdjustable* calc, const FitVar* v,
                 int nbinsx, double minx, double maxx, double minchi)
  {
    TH1* ret = Slice(expt, calc, v, nbinsx, minx, maxx, minchi);
    for(int n = 0; n < ret->GetNbinsX()+2; ++n){
      const double v = ret->GetBinContent(n);
      ret->SetBinContent(n, v > 0 ? sqrt(v) : 0);
    }
    ret->GetYaxis()->SetTitle("#sigma");
    return ret;
  }
}
