#pragma once

#include "CAFAna/Spectrum.h"
#include "CAFAna/Prediction.h"
#include "CAFAna/SystShifts.h"

#include "Minuit2/MnMigrad.h"
#include "Minuit2/FCNBase.h"
#include "Minuit2/FunctionMinimum.h"
#include "Minuit2/MnUserParameters.h"


namespace osc{class IOscCalculatorAdjustable;}

namespace ana
{
  class Experiment;
  class FitVar;

  /// \brief Figure-of-merit with no systematics, for binned data
  ///
  /// \param obs The observed data
  /// \param unosc A spectrum representing the null hypothesis
  /// \param pot POT to scale to. Leave at default to adopt POT from \a obs
  ///
  /// \returns Sum in quadrature over the bins
  /// \f[ \sqrt{\sum_i^{\rm bins}\left({s_i\over\sqrt{s_i+b_i}}\right)^2} \f]
  double SimpleFOM(const Spectrum& obs, const Spectrum& unosc, double pot = 0);

  /// Perform MINUIT fits in one or two dimensions
  class Fitter: public ROOT::Minuit2::FCNBase
  {
  public:
    Fitter(const Experiment* expt,
           std::vector<const FitVar*> vars,
           std::vector<Syst> systs = {});

    enum Verbosity{kQuiet, kVerbose};
    /// \param[out] seed Seed parameter and output best-fit point
    /// \param      verb If quiet, no printout
    /// \return     The log-likelihood of the best-fit point
    double Fit(osc::IOscCalculatorAdjustable* seed,
               SystShifts& systSeed,
               Verbosity verb = kVerbose) const;

    /// Variant with no systematics result returned
    double Fit(osc::IOscCalculatorAdjustable* seed,
               Verbosity verb = kVerbose) const;

    /// Evaluate the log-likelihood, as required by MINUT interface
    virtual double operator()(const std::vector<double>& pars) const;

    /// Definition of one-sigma, required by MINUIT
    virtual double Up() const {return 1;}
  protected:
    const Experiment* fExpt;
    std::vector<const FitVar*> fVars;
    std::vector<Syst> fSysts;
    mutable osc::IOscCalculatorAdjustable* fCalc;
    mutable SystShifts fShifts;
  };

  /// \f$\chi^2\f$ scan in one variable, holding all others constant
  TH1* Slice(const Experiment* expt,
             osc::IOscCalculatorAdjustable* calc, const FitVar* v,
             int nbinsx, double minx, double maxx,
             double minchi = -1);

  /// Super cheesy...
  TH1* SqrtSlice(const Experiment* expt,
                 osc::IOscCalculatorAdjustable* calc, const FitVar* v,
                 int nbinsx, double minx, double maxx, double minchi);
}
