#include "CAFAna/Loaders.h"

#include "CAFAna/SpectrumLoader.h"
#include "CAFAna/Utilities.h"

#include <cassert>
#include <iostream>

namespace ana
{
  //----------------------------------------------------------------------
  Loaders::Loaders()
    : fNumChildCallbacks(0)
  {
  }

  //----------------------------------------------------------------------
  Loaders::Loaders(const std::string& dir, FluxType flux)
    : fNumChildCallbacks(0)
  {
    // Loop through all configurations. Try and construct the wildcard string
    // that will find the right files. May well need adjusting as our naming
    // conventions become clearer.

    const std::string fluxStr = (flux == kFHC) ? "fhc" : "rhc";

    for(caf::Det_t det: {caf::kNEARDET, caf::kFARDET}){
      const std::string detStr = (det == caf::kNEARDET) ? "neardet" : "fardet";

      for(DataMC mc: {kData, kMC}){
        const std::string mcStr = (mc == kMC) ? "sim" : "data";

        for(DataSource src: {kBeam, kCosmic}){
          const std::string cosmicFluxStr = (src == kBeam) ? fluxStr : "";

          for(SwappingConfig swap: {kNonSwap, kFluxSwap, kTauSwap}){
            // Skip the physically meaningless ones
            if((det == caf::kNEARDET || mc == kData) && swap != kNonSwap) continue;
            if(src == kCosmic && swap != kNonSwap) continue;

            std::string swapStr;
            if(swap == kNonSwap && mc == kMC && det == caf::kFARDET) swapStr = "_nonswap_";
            if(swap == kFluxSwap) swapStr = "_swap_";
            if(swap == kTauSwap) swapStr = "_fluxtau_";

            // Can't have both a swapStr and a srcStr
            if(src == kCosmic) swapStr = "cosmic";

            const std::string pattern = dir+"/"+detStr+"*"+cosmicFluxStr+"*"+swapStr+"*"+mcStr+".caf.root";

            const std::vector<std::string> fnames = Wildcard(pattern);

            std::cout << "Found " << fnames.size() << " files matching \"" << pattern << "\"." << std::endl;

            if(!fnames.empty()) SetLoaderPath(pattern, det, mc, src, swap);
          }
        }
      }
    }
  }

  //----------------------------------------------------------------------
  Loaders::~Loaders()
  {
    for(auto it: fLoaders) delete it.second;
  }

  //----------------------------------------------------------------------
  void Loaders::SetLoaderPath(const std::string& path,
                              caf::Det_t det,
                              DataMC datamc,
                              DataSource src,
                              SwappingConfig swap)
  {
    assert(datamc == kMC || swap == kNonSwap);
    assert(det == caf::kFARDET || swap == kNonSwap);
    assert(src == kBeam || swap == kNonSwap);

    Key_t key(det, datamc, src, swap);

    // Clear out the old one if necessary
    auto it = fLoaders.find(key);
    if(it != fLoaders.end()) delete it->second;

    fLoaders[key] = new SpectrumLoader(path, src);
    fLoaders[key]->RegisterCompletionCallback(this, &Loaders::LoadedCallback);
  }

  //----------------------------------------------------------------------
  void Loaders::DisableLoader(caf::Det_t det,
                              DataMC datamc,
                              DataSource src,
                              SwappingConfig swap)
  {
    assert(datamc == kMC || swap == kNonSwap);
    assert(det == caf::kFARDET || swap == kNonSwap);
    assert(src == kBeam || swap == kNonSwap);

    Key_t key(det, datamc, src, swap);

    // Clear out the current one if possible
    auto it = fLoaders.find(key);
    if(it != fLoaders.end()){
      delete it->second;
      fLoaders.erase(it);
    }
  }

  //----------------------------------------------------------------------
  SpectrumLoaderBase& Loaders::GetLoader(caf::Det_t det,
                                         DataMC datamc,
                                         DataSource src,
                                         SwappingConfig swap)
  {
    assert(datamc == kMC || swap == kNonSwap);
    assert(det == caf::kFARDET || swap == kNonSwap);
    assert(src == kBeam || swap == kNonSwap);

    Key_t key(det, datamc, src, swap);

    // Look up and return. Use fNull if no loader is set for this config
    auto it = fLoaders.find(key);
    if(it == fLoaders.end()) return fNull;
    return *it->second;
  }

  //----------------------------------------------------------------------
  void Loaders::RegisterCompletionCallback(std::function<CallbackFunc_t> func)
  {
    fCallbacks.push_back(func);
  }

  //----------------------------------------------------------------------
  void Loaders::Go()
  {
    for(auto it: fLoaders) it.second->Go();

    // Ensure anyone who signed up to one of these gets properly notified
    fNull.Go();

    // No need to notify our callbacks here. Our LoadedCallback() should have
    // noticed we're done and taken care of it.
  }

  //----------------------------------------------------------------------
  void Loaders::SetShifter(SystematicShifter* s)
  {
    for(auto it: fLoaders) it.second->SetShifter(s);
  }

  //----------------------------------------------------------------------
  void Loaders::LoadedCallback()
  {
    ++fNumChildCallbacks;
    // Someone got loaded more than once??
    assert(fNumChildCallbacks <= fLoaders.size());

    // One all of our children have notified, it's time to notify people
    // waiting on us
    if(fNumChildCallbacks == fLoaders.size()){
      for(auto c: fCallbacks) c();
    }
  }
}
