#include "CAFAna/MRCCDecomp.h"

#include <cassert>
#include <vector>

#include "CAFAna/Binning.h"
#include "CAFAna/ExtrapUtil.h"
#include "CAFAna/Ratio.h"

#include "TAxis.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TGraph.h"
#include "TH1.h"
#include "TH2.h"

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////                  

namespace ana{

  MRCCDecomp::MRCCDecomp(MRCCLoader&         mrccMCLoader,
  			 SpectrumLoaderBase& nearMCLoader,
  			 std::string         label,
  			 const Binning&      bins,
  			 const Var&          energyBin,
  			 const Cut&          allCuts,
  			 const Var&          weight):
    fMrccSpec      (label, bins, mrccMCLoader, energyBin, allCuts , weight),
    fNCSpec        (label, bins, nearMCLoader, energyBin, allCuts && kIsNC, weight),
    fNueCCSpec     (label, bins, nearMCLoader, energyBin, allCuts && kIsBeamNue && !kIsAntiNu, weight),
    fAntiNueCCSpec (label, bins, nearMCLoader, energyBin, allCuts && kIsBeamNue &&  kIsAntiNu, weight),
    fNumuCCSpec    (label, bins, nearMCLoader, energyBin, allCuts && kIsNumuCC  && !kIsAntiNu, weight),
    fAntiNumuCCSpec(label, bins, nearMCLoader, energyBin, allCuts && kIsNumuCC  &&  kIsAntiNu, weight),
    fDataMrccSpec  (label, bins),
    fDataAll       (label, bins)
  {
  }// end of simple constructor
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  


  MRCCDecomp::MRCCDecomp(MRCCLoader&         mrccMCLoader,
			 SpectrumLoaderBase& nearMCLoader,
			 MRCCLoader&         mrccDATALoader,
			 SpectrumLoaderBase& NDDATALoader,
			 std::string         label,
			 const Binning&      bins,
			 const Var&          energyBin,
			 const Cut&          allCuts,
			 const Var&          weight):
    MRCCDecomp   ( mrccMCLoader, nearMCLoader, label, bins, energyBin, allCuts, weight)
  {
    mrccDATALoader.AddSpectrum( fDataMrccSpec, energyBin, allCuts , weight);
    NDDATALoader.AddSpectrum( fDataAll, energyBin, allCuts , weight);
    
  }// end of constructor


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  


  Spectrum MRCCDecomp::NCComponent() const
  {
    return NCToMRCCRatio() * fDataMrccSpec;
  }// end of NCComponent

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  
  
  // MRCC is best suited to predict the NC component.
  // Its estimation of other components of the spectrum
  // is based on the NC estimate. Nue CC is not really 
  // estimated. Its taken directly from MC
  Spectrum MRCCDecomp::NueComponent() const
  {
    return fNueCCSpec;
  }// end of NueComponent

  Spectrum MRCCDecomp::AntiNueComponent() const
  {
    return fAntiNueCCSpec;
  }// end of NueComponent

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  

  Spectrum MRCCDecomp::NumuComponent() const
  {
    const Ratio nuFrac = fNumuCCSpec/(fNumuCCSpec+fAntiNumuCCSpec);
    return nuFrac*(fDataAll-NCComponent()-NueComponent()-AntiNueComponent());
  }// end of NumuComponent


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  

  Spectrum MRCCDecomp::AntiNumuComponent() const
  {
    const Ratio antiFrac = fAntiNumuCCSpec/(fNumuCCSpec+fAntiNumuCCSpec);
    return antiFrac*(fDataAll-NCComponent()-NueComponent()-AntiNueComponent());
  }// end of NumuComponent


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  

  Ratio MRCCDecomp::NCToMRCCRatio() const
  {
    return Ratio(fNCSpec, fMrccSpec);
  }// end of NCToMRCCRatio

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                  

  void MRCCDecomp::SaveTo( TDirectory* dir ) const
  {
    assert(0 && "Not implemented");
  }

}// end of namespace ana
