#include "CAFAna/MRCCLoader.h"


#include "CAFAna/OscillatableSpectrum.h"
#include "CAFAna/Progress.h"
#include "CAFAna/Spectrum.h"
#include "CAFAna/SystematicShifter.h"
#include "CAFAna/Utilities.h"

#include "StandardRecord/StandardRecord.h"

#include <cassert>
#include <iostream>

#include <fenv.h>

#include "TFile.h"
#include "TH2.h"
#include "TTree.h"


namespace ana
{
  //----------------------------------------------------------------------
  MRCCLoader::MRCCLoader(const std::string& childWildcard,
                         const std::string& parentWildcard,
                         const Cut& childCut,
                         const Cut& parentCut)
    : SpectrumLoaderBase(childWildcard, kBeam), fShifter(0),
      fParentFileNames(WildcardOrSAMQuery(parentWildcard)),
      fChildCut(childCut), fParentCut(parentCut)
  {
    assert(fParentFileNames.size() == fFileNames.size());

    // Files need to be in the same order. This isn't foolproof if they're in
    // differing subdirectory structures, but should be good enough in
    // practice?
    std::sort(fFileNames.begin(), fFileNames.end());
    std::sort(fParentFileNames.begin(), fParentFileNames.end());

    fParentBranchNames.insert(parentCut.Requires().begin(),
			      parentCut.Requires().end());
    fParentBranchNames.insert("hdr.run");
    fParentBranchNames.insert("hdr.subrun");
    fParentBranchNames.insert("hdr.evt");
    fParentBranchNames.insert("hdr.subevt");

    RequireVariable("hdr.run");
    RequireVariable("hdr.subrun");
    RequireVariable("hdr.evt");
    RequireVariable("hdr.subevt");
    RequireVariable("parent.mrccpar.slcidx");
    RequireVariables(childCut.Requires());
  }

  //----------------------------------------------------------------------
  MRCCLoader::MRCCLoader(const std::vector<std::string>& childFileNames,
                         const std::vector<std::string>& parentFileNames,
                         const Cut& childCut,
                         const Cut& parentCut)
    : SpectrumLoaderBase(childFileNames, kBeam), fShifter(0),
      fParentFileNames(parentFileNames), fChildCut(childCut), fParentCut(parentCut)
  {
    assert(fParentFileNames.size() == fFileNames.size());

    // Files need to be in the same order. This isn't foolproof if they're in
    // differing subdirectory structures, but should be good enough in
    // practice?
    std::sort(fFileNames.begin(), fFileNames.end());
    std::sort(fParentFileNames.begin(), fParentFileNames.end());

    fParentBranchNames.insert(parentCut.Requires().begin(),
			      parentCut.Requires().end());
    fParentBranchNames.insert("hdr.run");
    fParentBranchNames.insert("hdr.subrun");
    fParentBranchNames.insert("hdr.evt");
    fParentBranchNames.insert("hdr.subevt");

    RequireVariable("hdr.run");
    RequireVariable("hdr.subrun");
    RequireVariable("hdr.evt");
    RequireVariable("hdr.subevt");
    RequireVariable("parent.mrccpar.slcidx");
    RequireVariables(childCut.Requires());
  }

  //----------------------------------------------------------------------
  MRCCLoader::~MRCCLoader()
  {
  }

  //----------------------------------------------------------------------
  void MRCCLoader::AddSpectrum(Spectrum& spect,
                               const Var& var,
                               const Cut& cut,
                               const Var& wei)
  {
    SpectrumLoaderBase::AddSpectrum(spect, var, cut, wei);

    fHistDefs.back().shifter = fShifter;
  }

  //----------------------------------------------------------------------
  void MRCCLoader::AddOscillatableSpectrum(OscillatableSpectrum& spect,
                                           const Var& var,
                                           const Cut& cut,
                                           const Var& wei)
  {
    SpectrumLoaderBase::AddOscillatableSpectrum(spect, var, cut, wei);

    fHistDefs.back().shifter = fShifter;
  }
  
  //----------------------------------------------------------------------
  std::tuple<int, int, int, int> MRCCLoader::
  CurrentRecordTuple(const caf::StandardRecord* sr) const
  {
    return std::make_tuple(sr->hdr.run,
			   sr->hdr.subrun,
			   sr->hdr.evt,
			   sr->hdr.subevt);
  }

  //----------------------------------------------------------------------
  std::tuple<int, int, int, int> MRCCLoader::
  ParentRecordTuple(const caf::StandardRecord* sr) const
  {
    return std::make_tuple(sr->hdr.run,
			   sr->hdr.subrun,
			   sr->hdr.evt,
			   sr->parent.mrccpar.slcidx);
  }
    
  //----------------------------------------------------------------------
  void MRCCLoader::Go()
  {
    feenableexcept(FE_INVALID);

    if(fGone){
      std::cerr << "Error: can only call Go() once on a MRCCLoader" << std::endl;
      abort();
    }
    fGone = true;

    for(const HistDef& hd: fHistDefs){
      if(hd.shifter) RequireVariables(hd.shifter->Requires());
    }

    std::cout << "Active branches: ";
    for(const std::string& s: fBranchNames) std::cout << s << " ";
    std::cout << std::endl;

    const unsigned int Nfiles = fFileNames.size();

    double pot = 0;

    Progress prog(TString::Format("Filling %lu spectra from %d files matching '%s'", fHistDefs.size(), Nfiles, fWildcard.c_str()).Data());

    for(unsigned int fileIdx = 0; fileIdx < Nfiles; ++fileIdx){
      TFile fChild(fFileNames[fileIdx].c_str());
      assert(!fChild.IsZombie());

      TFile fParent(fParentFileNames[fileIdx].c_str());
      assert(!fParent.IsZombie());

      TH1* hPOTChild = (TH1*)fChild.Get("TotalPOT");
      assert(hPOTChild);
      pot += hPOTChild->Integral();

      TTree* trChild = (TTree*)fChild.Get("recTree");
      assert(trChild);

      TTree* trParent = (TTree*)fParent.Get("recTree");
      assert(trParent);

      BranchList childBranches(trChild, fBranchNames);
      BranchList parentBranches(trParent, fParentBranchNames);

      int nParent = 0;
      caf::StandardRecord* parentSR = parentBranches.GetEntry(nParent);

      const int Nentries = trChild->GetEntries();
      for(int nChild = 0; nChild < Nentries; ++nChild){
        caf::StandardRecord* childSR = childBranches.GetEntry(nChild);

	// This is a bug in the upstream MuonRemove. Work around it for now,
	// but one day this should be a fatal error. Message commented out
	// because this happens every few thousand slices.
	if(childSR->parent.mrccpar.slcidx == 0){
	  //	  std::cerr << "\nWarning! MRCC parent slice (parent.mrccpar.slcidx) is set to noise slice!\n"
	  //		    << "Skipping record " << nChild
	  //		    << " in " << fFileNames[fileIdx]
	  //		    << std::endl;
	  continue;
	}

        // Navigate to the right event and then the slice index that matches
        // the MRCC parent. There are assumptions here that everything is in
        // order.

        // Tuples compare by earlier parameters first
        while(CurrentRecordTuple(parentSR) < ParentRecordTuple(childSR)){
          parentSR = parentBranches.GetEntry(++nParent);
        }

	while(CurrentRecordTuple(parentSR) > ParentRecordTuple(childSR)){
          parentSR = parentBranches.GetEntry(--nParent);
        }

	// Make sure we found the right one
	assert(CurrentRecordTuple(parentSR) == ParentRecordTuple(childSR));

        HandleRecord(childSR, parentSR);

        if(nChild%10000 == 0 && Nfiles == 1)
	  prog.SetProgress(double(nChild)/Nentries);
      } // end for n
      prog.SetProgress((fileIdx+1.)/Nfiles);
    } // end for fileIdx

    prog.Done();

    std::cout << "Total " << pot << " POT" << std::endl;
    assert(pot > 0);

    for(HistDef& hd: fHistDefs){
      // OK because we're a friend class
      if(hd.spect)
        hd.spect->fPOT += pot;
      else
        hd.oscSpect->fPOT += pot;
    }

    // Notify everyone this call to Go() has completed
    for(auto c: fCallbacks) c();
  }

  //----------------------------------------------------------------------
  void MRCCLoader::HandleRecord(caf::StandardRecord* childSR,
                                caf::StandardRecord* parentSR)
  {
    // TODO: what to really do about no true nu?
    const double trueE = (fAnyOsc && childSR->mc.nu.size() == 1) ? childSR->mc.nu[0].E : 0;

    SystematicShifter* lastShift = 0;

    double systWeight = 1;

    for(const HistDef& hd: fHistDefs){

      // When the systematic shifter changes, need to provide a clean slate for
      // the next one to work from. Unfortunately, copying the whole
      // StandardRecord is pretty expensive. So we need to rely on this
      // slightly dangerous "Unshift" mechanism.
      if(hd.shifter != lastShift){
        if(lastShift) lastShift->Unshift(childSR);
        systWeight = 1;
        lastShift = hd.shifter;
        if(hd.shifter) hd.shifter->Shift(childSR, systWeight);
      }

      // No cut, or cut passes
      if(fParentCut(parentSR) && fChildCut(childSR) && hd.cut(childSR)){
        const double wei = hd.wei(childSR)*systWeight;
        if(wei){
          const double val = hd.var(childSR);
          if(hd.spect) hd.spect->fHist->Fill(val, wei);
          // TODO: ignoring events with no true neutrino
          if(hd.oscSpect && trueE > 0){
            hd.oscSpect->fHist->Fill(val, trueE, wei);
          }
        }
      }
    } // end for hd
  }

  //----------------------------------------------------------------------
  void MRCCLoader::SetShifter(SystematicShifter* s)
  {
    fShifter = s;
  }
} // namespace
