#include "CAFAna/MichelDecomp.h"
#include "CAFAna/Ratio.h"
#include "CAFAna/Utilities.h"
#include "Utilities/func/MathUtil.h"

#include <cassert>

#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TAxis.h"
#include "TMatrixD.h"
#include "TVectorD.h"
#include "TDirectory.h"
#include "TObjString.h"
#include "TRandom.h"

namespace ana
{
  //----------------------------------------------------------------------
  //----------------------------------------------------------------------
  MichelDecomp::MichelDecomp(SpectrumLoaderBase& NDMCloader,
			     SpectrumLoaderBase& NDdataloader,
			     const HistAxis& EAxis,
			     const Cut& cut,
			     const Var& wei)
    : fNC         (NDMCloader, EAxis,
		   cut && kIsNC,
		   kUnweighted),
      fNumu       (NDMCloader, EAxis,
		   cut && kIsNumuCC && !kIsAntiNu,
		   kUnweighted),
      fAntiNumu   (NDMCloader, EAxis,
		   cut && kIsNumuCC && kIsAntiNu,
		   kUnweighted),
      fNue        (NDMCloader, EAxis,
		   cut && kIsBeamNue && !kIsAntiNu,
		   kUnweighted),
      fAntiNue    (NDMCloader, EAxis,
		   cut && kIsBeamNue && kIsAntiNu,
		   kUnweighted),
      fData       (NDdataloader, EAxis,
		   cut,
		   kIntReWei),
      fMCNSlcs    ("N Slcs", SBinning, NDMCloader, kNSlcs,
		   cut, 
		   kUnweighted),
      fDataNSlcs  ("N Slcs", SBinning, NDdataloader, kNSlcs,
		   cut,
		   kIntReWei),
      fMichelNC   (EAxis.label, NDMCloader,
		   EAxis.bins, EAxis.var, MBinning, kNMichels, 
		   SBinning, kNSlcs, cut && kIsNC,
		   kUnweighted),
      fMichelNumu (EAxis.label, NDMCloader,
		   EAxis.bins, EAxis.var, MBinning, kNMichels,
		   SBinning, kNSlcs, cut && !kIsNC && kIsNumu,  
		   kUnweighted),
      fMichelNue  (EAxis.label, NDMCloader,
		   EAxis.bins, EAxis.var, MBinning, kNMichels, 
		   SBinning, kNSlcs, cut && !kIsNC && !kIsNumu, 
		   kUnweighted),
      fMichelData (EAxis.label, NDdataloader,
		   EAxis.bins, EAxis.var, MBinning, kNMichels,
		   cut, kIntReWei),
      fMockData   (EAxis.label, NDdataloader,
		   EAxis.bins, EAxis.var, MBinning, kNMichels,
		   cut, kIntReWei)
  {}

  MichelDecomp::MichelDecomp(SpectrumLoaderBase& NDMCloader,
			     SpectrumLoaderBase& NDdataloader,
			     std::string       label,
			     const Binning&    bins,
			     const Var&        var,
			     const Cut& cut,
			     const Var& wei)
    : MichelDecomp(NDMCloader, NDdataloader, 
		   HistAxis(label, bins, var),
		   cut, wei)
  {}


  //----------------------------------------------------------------------
  void MichelDecomp::DataFluctuate(){
    fMockData = fMichelData.MockData(fMockData.POT());
  }

  //----------------------------------------------------------------------
  std::vector<float> MichelDecomp::SlcWei() const {
    TH1* dataNSlcs = fDataNSlcs.ToTH1(fDataNSlcs.POT());
    TH1* mcNSlcs = fMCNSlcs.ToTH1(fMCNSlcs.POT());
    // We don't want any POT differences between the two spectra 
    dataNSlcs->Scale(pow(dataNSlcs->Integral(),-1));
    mcNSlcs->Scale(pow(mcNSlcs->Integral(),-1));

    std::vector<float> wei;

    if (dataNSlcs->GetNbinsX() != mcNSlcs->GetNbinsX())
      return wei;

    for (int bIdx = 1; bIdx < dataNSlcs->GetNbinsX() + 1; bIdx++){
      float mcslcs = mcNSlcs->GetBinContent(bIdx);
      float dataslcs = dataNSlcs->GetBinContent(bIdx);
      float curWei = 1;
      if (mcslcs > 0)
	curWei = dataslcs / mcslcs;
      wei.push_back(curWei);
    }
    return wei;
  }

  //----------------------------------------------------------------------
  TH2* MichelDecomp::IntensityReweight(Spectrum in, std::vector<float> wei) const {
    TH3* inHist = in.ToTH3(in.POT());
    TAxis* inX = inHist->GetXaxis();
    TAxis* inY = inHist->GetYaxis();
    TAxis* inZ = inHist->GetZaxis();
    TH2* out = new TH2F(UniqueName().c_str(), "foo", 
			inX->GetNbins(), inX->GetXmin(), inX->GetXmax(),
			inY->GetNbins(), inY->GetXmin(), inY->GetXmax());
    for (int xIdx = 1; xIdx < inX->GetNbins() + 1; xIdx++){
      for (int yIdx = 1; yIdx < inY->GetNbins() + 1; yIdx++){
	float outXY = 0;
	for (int zIdx = 1; zIdx < inZ->GetNbins() + 1; zIdx++)
	  outXY += wei[zIdx - 1]*inHist->GetBinContent(xIdx, yIdx, zIdx);
	out->SetBinContent(xIdx, yIdx, outXY);
      }
    }
    return out;
  }

  //----------------------------------------------------------------------
  void MichelDecomp::Fitter(TH2* tcc, TH2* tnc, TH2* tnue, TH2* tdata, 
    			    std::vector<double>& cc_scale_factors,
			    std::vector<double>& nc_scale_factors) const {
    // We have to fit out scale factors for the decomposition.  
    // But it turns out that there 
    // is an exact analytic solution for my chi square.  
    // So I got to fill a huge matrix with 
    // a bunch of big sums.

    int E_Bins = tcc->GetNbinsX();
    int M_Bins = tcc->GetNbinsY();
    for (int E_idx = 0; E_idx < E_Bins; E_idx++){
      // Do the Minimization for each Energy Bin
      double form_matrix_cc_cc = 0, form_matrix_cc_nc = 0, form_matrix_nc_nc = 0;
      double column_matrix_cc = 0, column_matrix_nc = 0;
      for (int M_idx = 0; M_idx < M_Bins; M_idx++){
	double cur_data = tdata->GetBinContent(E_idx+1,M_idx+1);
	if (cur_data != 0){
	  form_matrix_cc_cc += util::sqr(tcc->GetBinContent(E_idx+1,M_idx+1))*
	    pow(cur_data,-1);
	  form_matrix_cc_nc += tcc->GetBinContent(E_idx+1,M_idx+1)*
	    tnc->GetBinContent(E_idx+1,M_idx+1)*pow(cur_data,-1);
	  form_matrix_nc_nc += util::sqr(tnc->GetBinContent(E_idx+1,M_idx+1))*
	    pow(cur_data,-1);
	  
	  column_matrix_cc += tcc->GetBinContent(E_idx+1,M_idx+1) - 
	    tcc->GetBinContent(E_idx+1,M_idx+1)*
	    tnue->GetBinContent(E_idx+1,M_idx+1)*pow(cur_data,-1);
	  column_matrix_nc += tnc->GetBinContent(E_idx+1,M_idx+1) - 
	    tnc->GetBinContent(E_idx+1,M_idx+1)*
	    tnue->GetBinContent(E_idx+1,M_idx+1)*pow(cur_data,-1);
	}
      }

      // And now we fill our form matrix and our column matrix
      TMatrixD Form_Matrix(2,2);
      Form_Matrix(0,0) = form_matrix_cc_cc;
      Form_Matrix(1,0) = form_matrix_cc_nc;
      Form_Matrix(0,1) = form_matrix_cc_nc;
      Form_Matrix(1,1) = form_matrix_nc_nc;
      
      TVectorD Column_Matrix(2);
      Column_Matrix(0) = column_matrix_cc;
      Column_Matrix(1) = column_matrix_nc;
      
      // And now we can invert and multiply the two for the 
      // reconstructed scale factors
      TMatrixD Form_Inverse = Form_Matrix;
      Form_Inverse.Invert();
      TVectorD Result(2);
      Result = Form_Inverse * Column_Matrix;
      
      cc_scale_factors.push_back(Result(0));
      nc_scale_factors.push_back(Result(1));
    } // End loop over Energy Bins
    delete tcc; delete tnc; delete tnue; delete tdata;
  }
  
  //----------------------------------------------------------------------
  Spectrum MichelDecomp::NueComponent() const {

    // We're using the MC for the NuE component
    return fNue;
  }

  //----------------------------------------------------------------------
  Spectrum MichelDecomp::AntiNueComponent() const {

    // We're using the MC for the NuE component
    return fAntiNue;
  }
  
  //----------------------------------------------------------------------
  Spectrum MichelDecomp::ScaleNumu(const Spectrum& spect) const {

    std::vector<float> wei = SlcWei();
    TH2* MCC   = IntensityReweight(fMichelNumu, wei);
    TH2* MNC   = IntensityReweight(fMichelNC,   wei);
    TH2* MNE   = IntensityReweight(fMichelNue,  wei);
    TH2* MData = fMockData.ToTH2(fMockData.POT());

    std::vector<double> cc_scale_factors;
    std::vector<double> nc_scale_factors;
    Fitter(MCC, MNC, MNE, MData,
	   cc_scale_factors, nc_scale_factors);

    TH1* cc_scale = spect.ToTH1(spect.POT()); // for the binning

    int E_Bins = cc_scale->GetNbinsX();
    for (int E_idx = 0; E_idx < E_Bins; E_idx++){
      std::cout << "Michel Decomp scale factor:   " <<
	cc_scale_factors[E_idx] << std::endl;
      cc_scale->SetBinContent(E_idx+1,cc_scale_factors[E_idx]);
    }
    Ratio cc_scale_spectrum = Ratio(cc_scale,"CC Scale Factor");

    delete cc_scale;

    return cc_scale_spectrum*spect;
  }

  //----------------------------------------------------------------------
  Spectrum MichelDecomp::NumuComponent() const {
    return ScaleNumu(fNumu);
  }

  //----------------------------------------------------------------------
  Spectrum MichelDecomp::AntiNumuComponent() const {
    return ScaleNumu(fAntiNumu);
  }

  
  //----------------------------------------------------------------------
  Spectrum MichelDecomp::NCComponent() const {

    std::vector<float> wei = SlcWei();
    TH2* MCC   = IntensityReweight(fMichelNumu, wei);
    TH2* MNC   = IntensityReweight(fMichelNC,   wei);
    TH2* MNE   = IntensityReweight(fMichelNue,  wei);
    TH2* MData = fMockData.ToTH2(fMockData.POT());
    
    std::vector<double> cc_scale_factors;
    std::vector<double> nc_scale_factors;
    Fitter(MCC, MNC, MNE, MData, 
	   cc_scale_factors, nc_scale_factors);
    
    TH1* nc_scale = fNC.ToTH1(fNC.POT());
    
    int E_Bins = nc_scale->GetNbinsX();
    for (int E_idx = 0; E_idx < E_Bins; E_idx++){
      nc_scale->SetBinContent(E_idx+1,nc_scale_factors[E_idx]);
    }
    Ratio nc_scale_spectrum = Ratio(nc_scale,"NC Scale Factor");
    

    return nc_scale_spectrum*fNC;
    //return fData - fNue - nc_scale_spectrum*fNumu;
  }

  //----------------------------------------------------------------------
  void MichelDecomp::SaveTo(TDirectory* dir) const{
    TDirectory* tmp = gDirectory;
    dir->cd();

    TObjString("MichelDecomp").Write("type");

    fNC.SaveTo(          dir->mkdir("NCComp"));
    fNue.SaveTo(         dir->mkdir("NueComp"));
    fAntiNue.SaveTo(     dir->mkdir("AntiNueComp"));
    fNumu.SaveTo(        dir->mkdir("NumuComp"));
    fAntiNumu.SaveTo(    dir->mkdir("AntiNumuComp"));
    fData.SaveTo(        dir->mkdir("Data"));

    fMCNSlcs.SaveTo(     dir->mkdir("NSlcsInMC"));
    fDataNSlcs.SaveTo(   dir->mkdir("NSlcsInData"));
    
    fMichelNC.SaveTo(    dir->mkdir("NCTemplate"));
    fMichelNumu.SaveTo(  dir->mkdir("NumuTemplate"));
    fMichelNue.SaveTo(   dir->mkdir("NueTemplate"));

    fMichelData.SaveTo(  dir->mkdir("MichelData"));
    fMockData.SaveTo(    dir->mkdir("MockData"));

    tmp->cd();
  }

  std::unique_ptr<MichelDecomp> MichelDecomp::LoadFrom(TDirectory* dir){
    std::unique_ptr<MichelDecomp> ret(new MichelDecomp);

    // This is a lot of repetitive typing.  Define some macros
#define LOAD_SPECT(FIELD, LABEL) assert(dir->GetDirectory(LABEL)); ret->FIELD = *Spectrum::LoadFrom(dir->GetDirectory(LABEL));

      LOAD_SPECT(fNC,         "NCComp");
      LOAD_SPECT(fNue,        "NueComp");
      LOAD_SPECT(fAntiNue,    "AntiNueComp");
      LOAD_SPECT(fNumu,       "NumuComp");
      LOAD_SPECT(fAntiNumu,   "AntiNumuComp");
      LOAD_SPECT(fData,       "Data");

      LOAD_SPECT(fMCNSlcs,    "NSlcsInMC");
      LOAD_SPECT(fDataNSlcs,  "NSlcsInData");
      
      LOAD_SPECT(fMichelNC,   "NCTemplate");
      LOAD_SPECT(fMichelNumu, "NumuTemplate");
      LOAD_SPECT(fMichelNue,  "NueTemplate");

      LOAD_SPECT(fMichelData, "MichelData");
      LOAD_SPECT(fMockData,   "MockData");

    return ret;
  }


}
