#include "CAFAna/LoadFromFile.h"
#include "CAFAna/ModularExtrapComponent.h"
#include "CAFAna/SpectrumLoaderBase.h"
#include "CAFAna/Decomp.h"
#include "CAFAna/Ratio.h"

#include "TH1.h"
#include "TDirectory.h"
#include "TObjString.h"

namespace ana
{

  OscillatableSpectrum SingleComponentExtrap::Return() const
  {
    if (!fEvaluated)
    {
      fCache = Eval();
      fEvaluated = true;
    }
    return fCache;
  }

  //----------------------------------------------------------------------------

  Spectrum SingleComponentExtrap::GetDecompResult(
    const IDecomp& decomp,
    const DecompResult dr
  ){
    switch (dr){
      case DecompResult::nue     : return decomp.NueComponent();
      case DecompResult::numu    : return decomp.NumuComponent();
      case DecompResult::nuebar  : return decomp.AntiNueComponent();
      case DecompResult::numubar : return decomp.AntiNumuComponent();
      case DecompResult::NC      : return decomp.NCComponent();
    }
    assert( 0 && "Bad DecompResult" );
  }

  //----------------------------------------------------------------------------

  std::string SingleComponentExtrap::DRToString(
    const DecompResult dr
  ){
    switch (dr){
      case DecompResult::nue     : return "Nue";
      case DecompResult::numu    : return "Numu";
      case DecompResult::nuebar  : return "NueBar";
      case DecompResult::numubar : return "NumuBar";
      case DecompResult::NC      : return "NC";
    }
    assert( 0 && "Bad DecompResult" );
  }

  //----------------------------------------------------------------------------

  DecompResult SingleComponentExtrap::StringToDR(
    const std::string str
  ){
    if      (str=="Nue")     return DecompResult::nue;
    else if (str=="Numu")    return DecompResult::numu;
    else if (str=="NueBar")  return DecompResult::nuebar;
    else if (str=="NumuBar") return DecompResult::numubar;
    else if (str=="NC")      return DecompResult::NC;
    else assert( 0 && "Bad DecompResult String" );
  }

  //----------------------------------------------------------------------------

  void SingleComponentExtrap::ZeroDivisionWarn(
    const Spectrum& num,
    const Spectrum& denom,
    const std::string component,
    const std::string location
  ){

    TH1* nh = num.ToTH1(1e20);
    TH1* dh = denom.ToTH1(1e20);

    int nbins = nh->GetNbinsX();
    assert( (nbins == dh->GetNbinsX()) && "Bin Mismatch" );

    for (int bin=0; bin<=nbins+1; bin++)
    {
       if (    ( nh->GetBinContent(bin) > 0 )
            && ( dh->GetBinContent(bin) == 0) )
         std::cout << "\nWARNING: can't rescale zero to non-zero. "
                      "Events lost. Check MC stats."
                   << "\nComponent: " << component
                   << "\nLocation: " << location
                   << "\nBin Center: " << dh->GetBinCenter(bin)
                   << "\n" ;
    }

  }

  //----------------------------------------------------------------------------

  ComponentFDMonteCarlo::ComponentFDMonteCarlo(
    SpectrumLoaderBase& loader,
    const HistAxis& axis,
    const Cut& fdcut,
    const Var& weight,
    const Cut& flavors,
    SpectrumLoaderBase& extraloader
  ) : fRecoFD(loader, axis, fdcut && flavors, weight)
  {
    extraloader.AddOscillatableSpectrum(
      fRecoFD, axis.var, fdcut && flavors, weight );
  }

  OscillatableSpectrum ComponentFDMonteCarlo::Eval() const
    {return fRecoFD;}

  void ComponentFDMonteCarlo::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();
    TObjString("ComponentFDMonteCarlo").Write("type");
    fRecoFD.SaveTo(dir->mkdir("RecoFD"));
    tmp->cd();
  }

  std::unique_ptr<ComponentFDMonteCarlo>
    ComponentFDMonteCarlo::LoadFrom(TDirectory* dir)
  {
    assert(dir->GetDirectory("RecoFD"));
    return std::unique_ptr<ComponentFDMonteCarlo>(new ComponentFDMonteCarlo(
      *(OscillatableSpectrum::LoadFrom(dir->GetDirectory("RecoFD")).release())
    ));
  }

  //----------------------------------------------------------------------------

  ComponentFarOverNear::ComponentFarOverNear(
    SpectrumLoaderBase& ndloader,
    const HistAxis& axis,
    const Cut& fdcut,
    const Var& weight,
    std::string label,
    const Cut& ndcut,
    const IDecomp& decomposition,
    const DecompResult dr,
    const Cut& ndflavor,
    SpectrumLoaderBase& fdloader,
    const Cut& fdflavors
  )
    : fRecoToTrueND(ndloader, axis, ndcut && ndflavor, weight),
      fTrueToRecoFD(fdloader, axis, fdcut && fdflavors, weight),
      fDecomp(decomposition),
      fDecompRes(dr),
      fLabel(label)
  {}

  OscillatableSpectrum ComponentFarOverNear::Eval() const
  {

    //Compute F/N Ratio
    Ratio ndTofd( fTrueToRecoFD.TrueEnergy(), fRecoToTrueND.TrueEnergy() );

    //Copy to local variables because reweighting is in-place
    OscillatableSpectrum recoToTrueND(fRecoToTrueND);
    OscillatableSpectrum trueToRecoFD(fTrueToRecoFD);

    //Get ND data from Decomp
    Spectrum decompresult(GetDecompResult(fDecomp,fDecompRes));

    //ND Reco->True
    ZeroDivisionWarn(
      decompresult, fRecoToTrueND.Unoscillated(),
      fLabel, "MC ND Reco" );
    recoToTrueND.ReweightToRecoSpectrum( decompresult );

    // Multiply by F/N and FD True->Reco
    ZeroDivisionWarn(
      fTrueToRecoFD.TrueEnergy(), fRecoToTrueND.TrueEnergy(),
      fLabel, "MC ND Truth" );
    trueToRecoFD.ReweightToTrueSpectrum( recoToTrueND.TrueEnergy() * ndTofd );

    return trueToRecoFD;

  }

  void ComponentFarOverNear::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();
    TObjString("ComponentFarOverNear").Write("type");
    fRecoToTrueND.SaveTo(dir->mkdir("RecoToTrueND"));
    fTrueToRecoFD.SaveTo(dir->mkdir("TrueToRecoFD"));
    fDecomp.SaveTo(dir->mkdir("Decomp"));
    TObjString(DRToString(fDecompRes).c_str()).Write("DecompRes");
    TObjString(fLabel.c_str()).Write("Label");
    tmp->cd();
  }

  std::unique_ptr<ComponentFarOverNear>
    ComponentFarOverNear::LoadFrom(TDirectory* dir)
  {

    assert(dir->GetDirectory("RecoToTrueND"));
    assert(dir->GetDirectory("TrueToRecoFD"));
    assert(dir->GetDirectory("Decomp"));
    TObjString* dr = (TObjString*)dir->Get("DecompRes");
    assert(dr);
    TObjString* label = (TObjString*)dir->Get("Label");
    assert(label);

    return std::unique_ptr<ComponentFarOverNear>(new ComponentFarOverNear(
      *(OscillatableSpectrum::LoadFrom(dir->GetDirectory("RecoToTrueND"))),
      *(OscillatableSpectrum::LoadFrom(dir->GetDirectory("TrueToRecoFD"))),
      *(ana::LoadFrom<IDecomp>(
        dir->GetDirectory("Decomp")).release()), //leaks!
      StringToDR(dr->GetString().Data()),
      label->GetString().Data()
    ));

  }

  //----------------------------------------------------------------------------

  ComponentRecoOverReco::ComponentRecoOverReco(
    SpectrumLoaderBase& ndloader,
    const HistAxis& axis,
    const Cut& fdcut,
    const Var& weight,
    std::string label,
    const Cut& ndcut,
    const IDecomp& decomposition,
    const DecompResult dr,
    const Cut& ndflavor,
    SpectrumLoaderBase& fdloader,
    const Cut& fdflavors,
    SpectrumLoaderBase& extrafdloader
  )
    : fRecoND(ndloader, axis, ndcut && ndflavor, weight),
      fTrueToRecoFD(fdloader, axis, fdcut && fdflavors, weight),
      fDecomp(decomposition),
      fDecompRes(dr),
      fLabel(label)
  {
    extrafdloader.AddOscillatableSpectrum(
      fTrueToRecoFD, axis.var, fdcut && fdflavors, weight );
  }

  OscillatableSpectrum ComponentRecoOverReco::Eval() const
  {

    //Compute F/N Ratio
    Ratio ndTofd( fTrueToRecoFD.Unoscillated(), fRecoND );

    //Copy to local variable because reweighting is in-place
    OscillatableSpectrum result(fTrueToRecoFD);

    //Get ND data from Decomp
    Spectrum decompresult(GetDecompResult(fDecomp,fDecompRes));

    // Multiply by F/N and add in FD truth information
    ZeroDivisionWarn(
      decompresult, fRecoND,
      fLabel, "MC ND Reco" );
    ZeroDivisionWarn(
      fTrueToRecoFD.Unoscillated(), fRecoND,
      fLabel, "MC ND Reco" );
    result.ReweightToRecoSpectrum( decompresult * ndTofd );

    return result;

  }

  void ComponentRecoOverReco::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();
    TObjString("ComponentRecoOverReco").Write("type");
    fRecoND.SaveTo(dir->mkdir("RecoND"));
    fTrueToRecoFD.SaveTo(dir->mkdir("TrueToRecoFD"));
    fDecomp.SaveTo(dir->mkdir("Decomp"));
    TObjString(DRToString(fDecompRes).c_str()).Write("DecompRes");
    TObjString(fLabel.c_str()).Write("Label");
    tmp->cd();
  }

  std::unique_ptr<ComponentRecoOverReco>
    ComponentRecoOverReco::LoadFrom(TDirectory* dir)
  {

    assert(dir->GetDirectory("RecoND"));
    assert(dir->GetDirectory("TrueToRecoFD"));
    assert(dir->GetDirectory("Decomp"));
    TObjString* dr = (TObjString*)dir->Get("DecompRes");
    assert(dr);
    TObjString* label = (TObjString*)dir->Get("Label");
    assert(label);

    return std::unique_ptr<ComponentRecoOverReco>(new ComponentRecoOverReco(
      *(Spectrum::LoadFrom(dir->GetDirectory("RecoND"))),
      *(OscillatableSpectrum::LoadFrom(dir->GetDirectory("TrueToRecoFD"))),
      *(ana::LoadFrom<IDecomp>(
        dir->GetDirectory("Decomp")).release()), //leaks!
      StringToDR(dr->GetString().Data()),
      label->GetString().Data()
    ));

  }

}
