#pragma once

#include "CAFAna/OscillatableSpectrum.h"

class TDirectory;

namespace ana
{

  class SpectrumLoaderBase;
  class IDecomp;
  class HistAxis;
  class Cut;
  class Var;

  /// Simple way to remember what to ask the decomposition for.
  enum class DecompResult {nue, numu, NC, nuebar, numubar};

  /// Base class for component extrapolation.
  class SingleComponentExtrap
  {
    public:

      SingleComponentExtrap() : fEvaluated(false), fCache(nullptr, 0) {}
      virtual ~SingleComponentExtrap() {}
      virtual void SaveTo(TDirectory* dir) const = 0;

      /// Interface so that result of Eval() is called only once and cached.
      OscillatableSpectrum Return() const;

    protected:

      /// Core extrapolation math.
      virtual OscillatableSpectrum Eval() const = 0;

      /// Helper function to pick out single Spectrum from a decomposition.
      static Spectrum GetDecompResult(const IDecomp&, DecompResult);

      /// Helper function to turn a DecompResult into a string (for storage).
      static std::string DRToString(DecompResult);

      /// Helper function to turn a string into a DecompResult (for loading).
      static DecompResult StringToDR(std::string);

      /// Helper function to warn about empty bins before Spectrum division.
      /** Necessary because root thinks n/0==0, so we lose events when
          low stats cause empty bins. */
      static void ZeroDivisionWarn(
        const Spectrum& num,
        const Spectrum& denom,
        std::string component,
        std::string location
      );

    private:
      mutable bool fEvaluated;
      mutable OscillatableSpectrum fCache;

  };

  /// "Extrapolates" component by returning FD Monte Carlo.
  class ComponentFDMonteCarlo: public SingleComponentExtrap
  {

    public:
      ComponentFDMonteCarlo(
        SpectrumLoaderBase& loader,
        const HistAxis& axis,
        const Cut& fdcut,
        const Var& weight,
        const Cut& flavors,
        SpectrumLoaderBase& extraloader = kNullLoader // for NC
      );
      OscillatableSpectrum Eval() const override;
      void SaveTo(TDirectory* dir) const override;
      static std::unique_ptr<ComponentFDMonteCarlo> LoadFrom(TDirectory* dir);
    private:
      ComponentFDMonteCarlo(OscillatableSpectrum RecoFD) : fRecoFD(RecoFD) {}

      OscillatableSpectrum fRecoFD;

  };

  /// Extrapolates component using truth-over-truth method.
  class ComponentFarOverNear: public SingleComponentExtrap
  {

    public:
      ComponentFarOverNear(
        SpectrumLoaderBase& ndloader,
        const HistAxis& axis,
        const Cut& fdcut,
        const Var& weight,
        std::string label,
        const Cut& ndcut,
        const IDecomp& decomposition,
        const DecompResult dr,
        const Cut& ndflavor,
        SpectrumLoaderBase& fdloader,
        const Cut& fdflavors
      );
      OscillatableSpectrum Eval() const override;
      void SaveTo(TDirectory* dir) const override;
      static std::unique_ptr<ComponentFarOverNear> LoadFrom(TDirectory* dir);
    private:
      ComponentFarOverNear(
        OscillatableSpectrum RecoToTrueND,
        OscillatableSpectrum TrueToRecoFD,
        const IDecomp& Decomp,
        const DecompResult DecompRes,
        std::string Label
      ) : fRecoToTrueND(RecoToTrueND), fTrueToRecoFD(TrueToRecoFD),
          fDecomp(Decomp), fDecompRes(DecompRes), fLabel(Label) {}

      OscillatableSpectrum fRecoToTrueND;
      OscillatableSpectrum fTrueToRecoFD;
      const IDecomp& fDecomp;
      const DecompResult fDecompRes;
      std::string fLabel;

  };

  /// Extrapolates using reco-over-reco method.
  class ComponentRecoOverReco: public SingleComponentExtrap
  {

    public:
      ComponentRecoOverReco(
        SpectrumLoaderBase& ndloader,
        const HistAxis& axis,
        const Cut& fdcut,
        const Var& weight,
        std::string label,
        const Cut& ndcut,
        const IDecomp& decomposition,
        const DecompResult dr,
        const Cut& ndflavor,
        SpectrumLoaderBase& fdloader,
        const Cut& fdflavors,
        SpectrumLoaderBase& extrafdloader = kNullLoader // for NC
      );
      OscillatableSpectrum Eval() const override;
      void SaveTo(TDirectory* dir) const override;
      static std::unique_ptr<ComponentRecoOverReco> LoadFrom(TDirectory* dir);
    private:
      ComponentRecoOverReco(
        Spectrum RecoND,
        OscillatableSpectrum TrueToRecoFD,
        const IDecomp& Decomp,
        const DecompResult DecompRes,
        std::string Label
      ) : fRecoND(RecoND), fTrueToRecoFD(TrueToRecoFD),
          fDecomp(Decomp), fDecompRes(DecompRes), fLabel(Label) {}

      Spectrum fRecoND;
      OscillatableSpectrum fTrueToRecoFD;
      const IDecomp& fDecomp;
      const DecompResult fDecompRes;
      std::string fLabel;

  };

}
