#include "CAFAna/NDDataMCRatioExtrap.h"

#include "OscLib/func/IOscCalculator.h"

#include <cassert>
#include <vector>

#include "CAFAna/LoadFromFile.h"
#include "CAFAna/OscCurve.h"
#include "CAFAna/XSec.h"

#include "TAxis.h"
#include "TCanvas.h"
#include "TDirectory.h"
#include "TFile.h"
#include "TGraph.h"
#include "TH1.h"
#include "TH2.h"
#include "TObjString.h"

namespace ana
{
  //----------------------------------------------------------------------
  NDDataMCRatioExtrap::NDDataMCRatioExtrap(Loaders& loaders,
                                           IDecomp* decomp,
                                           std::string label,
                                           const Binning& bins,
                                           const bool isNuE,
                                           const bool signSplit,
                                           const Var& varSurv,
                                           const Cut& cutSurvPre,
                                           const Cut& cutSurvNC,
                                           const Cut& cutSurvNuE,
                                           const Cut& cutSurvNuMu,
                                           const Var& varApprCCNuE,
                                           const Cut& cutApprCCNuEPre,
                                           const Cut& cutApprCCNuESel,
                                           const Var& varApprCCNuMu,
                                           const Cut& cutApprCCNuMuPre,
                                           const Cut& cutApprCCNuMuSel,
                                           const Var& varApprQENuMu,
                                           const Cut& cutApprQENuMuPre,
                                           const Cut& cutApprQENuMuSel,
                                           const Cut& cutContainND,
                                           const Cut& cutContainFD,
                                           const Var& wei)
    : NDDataMCRatioExtrap(loaders.GetLoader(caf::kNEARDET, Loaders::kMC,   ana::kBeam, Loaders::kNonSwap),
                          loaders.GetLoader(caf::kFARDET,  Loaders::kMC,   ana::kBeam, Loaders::kNonSwap),
                          loaders.GetLoader(caf::kFARDET,  Loaders::kMC,   ana::kBeam, Loaders::kFluxSwap),
                          loaders.GetLoader(caf::kFARDET,  Loaders::kMC,   ana::kBeam, Loaders::kTauSwap),
                          loaders.GetLoader(caf::kNEARDET, Loaders::kData, ana::kBeam, Loaders::kNonSwap),
                          decomp, label, bins, isNuE, signSplit,
                          varSurv,       cutSurvPre,       cutSurvNC, cutSurvNuE, cutSurvNuMu,
                          varApprCCNuE,  cutApprCCNuEPre,  cutApprCCNuESel,
                          varApprCCNuMu, cutApprCCNuMuPre, cutApprCCNuMuSel,
                          varApprQENuMu, cutApprQENuMuPre, cutApprQENuMuSel,
                          cutContainND, cutContainFD, wei)
  {
    if(!signSplit) {
      loaders.RegisterCompletionCallback(this, &NDDataMCRatioExtrap::NoSignSplitCorrection);
    }
  }

  //----------------------------------------------------------------------
  NDDataMCRatioExtrap::NDDataMCRatioExtrap(SpectrumLoaderBase& nearMC,
                                           SpectrumLoaderBase& far_MC,
                                           SpectrumLoaderBase& swapMC,
                                           SpectrumLoaderBase& tau_MC,
                                           SpectrumLoaderBase& NDDATA,
                                           IDecomp*            decomp,
                                           std::string label,
                                           const Binning& bins,
                                           const bool isNuE,
                                           const bool signSplit,
                                           const Var& varSurv,
                                           const Cut& cutSurvPre,
                                           const Cut& cutSurvNC,
                                           const Cut& cutSurvNuE,
                                           const Cut& cutSurvNuMu,
                                           const Var& varApprCCNuE,
                                           const Cut& cutApprCCNuEPre,
                                           const Cut& cutApprCCNuESel,
                                           const Var& varApprCCNuMu,
                                           const Cut& cutApprCCNuMuPre,
                                           const Cut& cutApprCCNuMuSel,
                                           const Var& varApprQENuMu,
                                           const Cut& cutApprQENuMuPre,
                                           const Cut& cutApprQENuMuSel,
                                           const Cut& cutContainND,
                                           const Cut& cutContainFD,
                                           const Var& wei)
    : // Fill spectra
      fApprDataCCNuE (label, bins, NDDATA, varApprCCNuE,  cutContainND && cutApprCCNuEPre  && cutApprCCNuESel,  wei),
      fApprDataCCNuMu(label, bins, NDDATA, varApprCCNuMu, cutContainND && cutApprCCNuMuPre && cutApprCCNuMuSel, wei),
      fApprDataQENuMu(label, bins, NDDATA, varApprQENuMu, cutContainND && cutApprQENuMuPre && cutApprQENuMuSel, wei),

      fSurvNDNC(label, bins, nearMC, varSurv, cutContainND && cutSurvPre && cutSurvNC, wei),
      fSurvFDNC(label, bins, far_MC, varSurv, cutContainFD && cutSurvPre && cutSurvNC, wei),

      fSurvNDCCNuE       (label, bins, nearMC, varSurv, cutContainND && cutSurvPre && cutSurvNuE  && !kIsAntiNu, wei),
      fSurvNDCCANuE      (label, bins, nearMC, varSurv, cutContainND && cutSurvPre && cutSurvNuE  &&  kIsAntiNu, wei),
      fSurvNDCCNuMu      (label, bins, nearMC, varSurv, cutContainND && cutSurvPre && cutSurvNuMu && !kIsAntiNu, wei),
      fSurvNDCCANuMu     (label, bins, nearMC, varSurv, cutContainND && cutSurvPre && cutSurvNuMu &&  kIsAntiNu, wei),
      fOSpecSurvFDCCNuE  (label, bins, far_MC, varSurv, cutContainFD && cutSurvPre && cutSurvNuE  && !kIsAntiNu, wei),
      fOSpecSurvFDCCANuE (label, bins, far_MC, varSurv, cutContainFD && cutSurvPre && cutSurvNuE  &&  kIsAntiNu, wei),
      fOSpecSurvFDCCNuMu (label, bins, far_MC, varSurv, cutContainFD && cutSurvPre && cutSurvNuMu && !kIsAntiNu, wei),
      fOSpecSurvFDCCANuMu(label, bins, far_MC, varSurv, cutContainFD && cutSurvPre && cutSurvNuMu &&  kIsAntiNu, wei),

      fApprNDCCNuE       (label, bins, nearMC, varApprCCNuE,  cutContainND && cutApprCCNuEPre  && cutApprCCNuESel  && !kIsAntiNu, wei),
      fApprNDCCANuE      (label, bins, nearMC, varApprCCNuE,  cutContainND && cutApprCCNuEPre  && cutApprCCNuESel  &&  kIsAntiNu, wei),
      fApprNDCCNuMu      (label, bins, nearMC, varApprCCNuMu, cutContainND && cutApprCCNuMuPre && cutApprCCNuMuSel && !kIsAntiNu, wei),
      fApprNDCCANuMu     (label, bins, nearMC, varApprCCNuMu, cutContainND && cutApprCCNuMuPre && cutApprCCNuMuSel &&  kIsAntiNu, wei),
      fApprNDQENuMu      (label, bins, nearMC, varApprQENuMu, cutContainND && cutApprQENuMuPre && cutApprQENuMuSel && !kIsAntiNu, wei),
      fApprNDQEANuMu     (label, bins, nearMC, varApprQENuMu, cutContainND && cutApprQENuMuPre && cutApprQENuMuSel &&  kIsAntiNu, wei),
      fOSpecApprFDCCNuE  (label, bins, far_MC, varApprCCNuE,  cutContainFD && cutApprCCNuEPre  && cutApprCCNuESel  && !kIsAntiNu, wei),
      fOSpecApprFDCCANuE (label, bins, far_MC, varApprCCNuE,  cutContainFD && cutApprCCNuEPre  && cutApprCCNuESel  &&  kIsAntiNu, wei),
      fOSpecApprFDCCNuMu (label, bins, far_MC, varApprCCNuMu, cutContainFD && cutApprCCNuMuPre && cutApprCCNuMuSel && !kIsAntiNu, wei),
      fOSpecApprFDCCANuMu(label, bins, far_MC, varApprCCNuMu, cutContainFD && cutApprCCNuMuPre && cutApprCCNuMuSel &&  kIsAntiNu, wei),
      fOSpecApprFDQENuMu (label, bins, far_MC, varApprQENuMu, cutContainFD && cutApprQENuMuPre && cutApprQENuMuSel && !kIsAntiNu, wei),
      fOSpecApprFDQEANuMu(label, bins, far_MC, varApprQENuMu, cutContainFD && cutApprQENuMuPre && cutApprQENuMuSel &&  kIsAntiNu, wei),

      fOSpecOscdFDCCNuE   (label, bins),
      fOSpecOscdFDCCANuE  (label, bins),
      fOSpecOscdFDCCNuMu  (label, bins),
      fOSpecOscdFDCCANuMu (label, bins),
      fOSpecOscdFDCCNuTau (label, bins),
      fOSpecOscdFDCCANuTau(label, bins),

      // Fill purities, efficiencies generators
      fApprFDPurEffNumCCNuE  (label, kTrueEnergyBins, far_MC, varApprCCNuE,
                              cutContainFD && cutApprCCNuEPre && cutApprCCNuESel && kIsBeamNue && !kIsAntiNu, wei),
      fApprFDPurEffNumCCANuE (label, kTrueEnergyBins, far_MC, varApprCCNuE,
                              cutContainFD && cutApprCCNuEPre && cutApprCCNuESel && kIsBeamNue &&  kIsAntiNu, wei),
      fApprFDPurDenCCNuE     (label, kTrueEnergyBins, far_MC, varApprCCNuE,
                              cutContainFD && cutApprCCNuEPre && cutApprCCNuESel &&               !kIsAntiNu, wei),
      fApprFDPurDenCCANuE    (label, kTrueEnergyBins, far_MC, varApprCCNuE,
                              cutContainFD && cutApprCCNuEPre && cutApprCCNuESel &&                kIsAntiNu, wei),
      fApprFDEffDenCCNuE     (label, kTrueEnergyBins, far_MC, varApprCCNuE,
                              cutContainFD &&                                       kIsBeamNue && !kIsAntiNu, wei),
      fApprFDEffDenCCANuE    (label, kTrueEnergyBins, far_MC, varApprCCNuE,
                              cutContainFD &&                                       kIsBeamNue &&  kIsAntiNu, wei),
      fApprFDPurEffNumCCNuMu (label, kTrueEnergyBins, far_MC, varApprCCNuMu,
                              cutContainFD && cutApprCCNuMuPre && cutApprCCNuMuSel && kIsNumuCC && !kIsQE && !kIsAntiNu, wei),
      fApprFDPurEffNumCCANuMu(label, kTrueEnergyBins, far_MC, varApprCCNuMu,
                              cutContainFD && cutApprCCNuMuPre && cutApprCCNuMuSel && kIsNumuCC && !kIsQE &&  kIsAntiNu, wei),
      fApprFDPurDenCCNuMu    (label, kTrueEnergyBins, far_MC, varApprCCNuMu,
                              cutContainFD && cutApprCCNuMuPre && cutApprCCNuMuSel &&                        !kIsAntiNu, wei),
      fApprFDPurDenCCANuMu   (label, kTrueEnergyBins, far_MC, varApprCCNuMu,
                              cutContainFD && cutApprCCNuMuPre && cutApprCCNuMuSel &&                         kIsAntiNu, wei),
      fApprFDEffDenCCNuMu    (label, kTrueEnergyBins, far_MC, varApprCCNuMu,
                              cutContainFD &&                                         kIsNumuCC && !kIsQE && !kIsAntiNu, wei),
      fApprFDEffDenCCANuMu   (label, kTrueEnergyBins, far_MC, varApprCCNuMu,
                              cutContainFD &&                                         kIsNumuCC && !kIsQE &&  kIsAntiNu, wei),
      fApprFDPurEffNumQENuMu (label, kTrueEnergyBins, far_MC, varApprQENuMu,
                              cutContainFD && cutApprQENuMuPre && cutApprQENuMuSel && kIsNumuCC &&  kIsQE && !kIsAntiNu, wei),
      fApprFDPurEffNumQEANuMu(label, kTrueEnergyBins, far_MC, varApprQENuMu,
                              cutContainFD && cutApprQENuMuPre && cutApprQENuMuSel && kIsNumuCC &&  kIsQE &&  kIsAntiNu, wei),
      fApprFDPurDenQENuMu    (label, kTrueEnergyBins, far_MC, varApprQENuMu,
                              cutContainFD && cutApprQENuMuPre && cutApprQENuMuSel &&                        !kIsAntiNu, wei),
      fApprFDPurDenQEANuMu   (label, kTrueEnergyBins, far_MC, varApprQENuMu,
                              cutContainFD && cutApprQENuMuPre && cutApprQENuMuSel &&                         kIsAntiNu, wei),
      fApprFDEffDenQENuMu    (label, kTrueEnergyBins, far_MC, varApprQENuMu,
                              cutContainFD &&                                         kIsNumuCC &&  kIsQE && !kIsAntiNu, wei),
      fApprFDEffDenQEANuMu   (label, kTrueEnergyBins, far_MC, varApprQENuMu,
                              cutContainFD &&                                         kIsNumuCC &&  kIsQE &&  kIsAntiNu, wei)
  {
    fSignSplit = signSplit;
    fDecomp = decomp;

    // Swapped files are just as valid a source of NCs as unswapped. This
    // approximately doubles our statistics. SpectrumLoader just adds events
    // and POT for both cases, which is the right thing to do.
    swapMC.AddSpectrum(fSurvFDNC, varSurv, cutContainFD && cutSurvPre && kIsNC, wei);

    // Fill oscillated spectra with the appropriate cut
    const Cut& cutFinal = (isNuE ? cutSurvNuE : cutSurvNuMu);
    swapMC.AddOscillatableSpectrum(fOSpecOscdFDCCNuE,    varSurv, cutContainFD && cutSurvPre && cutFinal && kIsSig       && !kIsAntiNu, wei);
    swapMC.AddOscillatableSpectrum(fOSpecOscdFDCCANuE,   varSurv, cutContainFD && cutSurvPre && cutFinal && kIsSig       &&  kIsAntiNu, wei);
    swapMC.AddOscillatableSpectrum(fOSpecOscdFDCCNuMu,   varSurv, cutContainFD && cutSurvPre && cutFinal && kIsNumuApp   && !kIsAntiNu, wei);
    swapMC.AddOscillatableSpectrum(fOSpecOscdFDCCANuMu,  varSurv, cutContainFD && cutSurvPre && cutFinal && kIsNumuApp   &&  kIsAntiNu, wei);
    tau_MC.AddOscillatableSpectrum(fOSpecOscdFDCCNuTau,  varSurv, cutContainFD && cutSurvPre && cutFinal && kIsTauFromMu && !kIsAntiNu, wei);
    tau_MC.AddOscillatableSpectrum(fOSpecOscdFDCCANuTau, varSurv, cutContainFD && cutSurvPre && cutFinal && kIsTauFromMu &&  kIsAntiNu, wei);
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::NueSurvComponent()
  {
    OscillatableSpectrum oSurvFDCCNuE = fOSpecSurvFDCCNuE; // Copy FD MC OscSpec into local variable
    // Reweight the reco bins by the ND Data/MC ratio
    oSurvFDCCNuE.ReweightToRecoSpectrum( oSurvFDCCNuE.Unoscillated()*NDRatio(+12) );

    return oSurvFDCCNuE;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::AntiNueSurvComponent()
  {
    OscillatableSpectrum oSurvFDCCANuE = fOSpecSurvFDCCANuE;
    oSurvFDCCANuE.ReweightToRecoSpectrum( oSurvFDCCANuE.Unoscillated()*NDRatio(-12) );

    return oSurvFDCCANuE;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::NumuSurvComponent()
  {
    OscillatableSpectrum oSurvFDCCNuMu = fOSpecSurvFDCCNuMu;
    oSurvFDCCNuMu.ReweightToRecoSpectrum( oSurvFDCCNuMu.Unoscillated()*NDRatio(+14) );

    return oSurvFDCCNuMu;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::AntiNumuSurvComponent()
  {
    OscillatableSpectrum oSurvFDCCANuMu = fOSpecSurvFDCCANuMu;
    oSurvFDCCANuMu.ReweightToRecoSpectrum( oSurvFDCCANuMu.Unoscillated()*NDRatio(-14) );

    return oSurvFDCCANuMu;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::NueAppComponent()
  {
    XSec* xsec = new XSec();
    Ratio xsecratio = xsec->GetRatio(12,+1,"CH2","tot_cc", 14,+1,"CH2","tot_cc", kTrueEnergyBins); // Get cross section ratio

    // Copy and combine FD MC OscSpecs into local variabe
    OscillatableSpectrum oApprFDCCNuMu = fOSpecApprFDCCNuMu + fOSpecApprFDQENuMu;
    // Reweight FD MC OscSpec by ND Data/MC ratio
    oApprFDCCNuMu.ReweightToRecoSpectrum( oApprFDCCNuMu.Unoscillated()*NDRatio(+14, true) );

    OscillatableSpectrum oOscdFDCCNuE = fOSpecOscdFDCCNuE; // Copy FD MC OscSpec into local variable
    // Reweight oscillated spectrum by appearance spectrum with cross section applied (multiplied in),
    // corrected for purity (multiplied in) and efficiency (divided out)
    oOscdFDCCNuE.ReweightToTrueSpectrum( oApprFDCCNuMu.TrueEnergy()*
                                         xsecratio*Purity(+14)/Efficiency(+14) );

    delete xsec;
    return oOscdFDCCNuE;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::AntiNueAppComponent()
  {
    XSec* xsec = new XSec();
    Ratio xsecratio = xsec->GetRatio(12,-1,"CH2","tot_cc", 14,-1,"CH2","tot_cc", kTrueEnergyBins);

    OscillatableSpectrum oApprFDCCANuMu = fOSpecApprFDCCANuMu + fOSpecApprFDQEANuMu;
    oApprFDCCANuMu.ReweightToRecoSpectrum( oApprFDCCANuMu.Unoscillated()*NDRatio(-14, true) );

    OscillatableSpectrum oOscdFDCCANuE = fOSpecOscdFDCCANuE;
    oOscdFDCCANuE.ReweightToTrueSpectrum( oApprFDCCANuMu.TrueEnergy()*
                                          xsecratio*Purity(-14)/Efficiency(-14) );

    delete xsec;
    return oOscdFDCCANuE;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::NumuAppComponent()
  {
    XSec* xsec = new XSec();
    Ratio xsecratio = xsec->GetRatio(14,+1,"CH2","tot_cc", 12,+1,"CH2","tot_cc", kTrueEnergyBins);

    OscillatableSpectrum oApprFDCCNuE = fOSpecApprFDCCNuE;
    oApprFDCCNuE.ReweightToRecoSpectrum( oApprFDCCNuE.Unoscillated()*NDRatio(+12, true) );

    OscillatableSpectrum oOscdFDCCNuMu = fOSpecOscdFDCCNuMu;
    oOscdFDCCNuMu.ReweightToTrueSpectrum( oApprFDCCNuE.TrueEnergy()*
                                          xsecratio*Purity(+12)/Efficiency(+12) );

    delete xsec;
    return oOscdFDCCNuMu;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum NDDataMCRatioExtrap::AntiNumuAppComponent()
  {
    XSec* xsec = new XSec();
    Ratio xsecratio = xsec->GetRatio(14,-1,"CH2","tot_cc", 12,-1,"CH2","tot_cc", kTrueEnergyBins);

    OscillatableSpectrum oApprFDCCANuE = fOSpecApprFDCCANuE;
    oApprFDCCANuE.ReweightToRecoSpectrum( oApprFDCCANuE.Unoscillated()*NDRatio(-12, true) );

    OscillatableSpectrum oOscdFDCCANuMu = fOSpecOscdFDCCANuMu;
    oOscdFDCCANuMu.ReweightToTrueSpectrum( oApprFDCCANuE.TrueEnergy()*
                                          xsecratio*Purity(-12)/Efficiency(-12) );

    delete xsec;
    return oOscdFDCCANuMu;
  }

  Spectrum NDDataMCRatioExtrap::NCComponent()
  {
    return fSurvFDNC*NDRatio();
  }

  //----------------------------------------------------------------------
  Spectrum NDDataMCRatioExtrap::NearMC() const
  {
    Spectrum ret = fSurvNDNC;
    ret += fSurvNDCCNuE;
    ret += fSurvNDCCANuE;
    ret += fSurvNDCCNuMu;
    ret += fSurvNDCCANuMu;
    // Sums the three components
    return ret;
  }

  //----------------------------------------------------------------------
  Spectrum NDDataMCRatioExtrap::FarMC() const
  {
    Spectrum ret = fSurvFDNC;
    ret += fOSpecSurvFDCCNuE.Unoscillated();
    ret += fOSpecSurvFDCCANuE.Unoscillated();
    ret += fOSpecSurvFDCCNuMu.Unoscillated();
    ret += fOSpecSurvFDCCANuMu.Unoscillated();
    
    return ret;
  }

  //----------------------------------------------------------------------
  Spectrum NDDataMCRatioExtrap::NearData() const
  {
    Spectrum ret = fDecomp->NCComponent();
    ret += fDecomp->NueComponent();
    ret += fDecomp->AntiNueComponent();
    ret += fDecomp->NumuComponent();
    ret += fDecomp->AntiNumuComponent();
    
    return ret;
  }

  //----------------------------------------------------------------------
  Ratio NDDataMCRatioExtrap::SignFraction(int pdg) const
  {
    assert(abs(pdg) == 12 || abs(pdg) == 14);

    if(pdg == +12)
      return Ratio(fApprNDCCNuE,
                   fApprNDCCNuE + fApprNDCCANuE);
    if(pdg == -12)
      return Ratio(fApprNDCCANuE,
                   fApprNDCCANuE + fApprNDCCNuE);
    if(pdg == +14)
      return Ratio(fApprNDCCNuMu + fApprNDQENuMu,
                   fApprNDCCNuMu + fApprNDCCANuMu +
                   fApprNDQENuMu + fApprNDQEANuMu);

    return Ratio(fApprNDCCANuMu + fApprNDQEANuMu,
                 fApprNDCCANuMu + fApprNDCCNuMu +
                 fApprNDQEANuMu + fApprNDQENuMu);
  }

  //----------------------------------------------------------------------
  Ratio NDDataMCRatioExtrap::NDRatio(int pdg, bool appr) const
  {
    assert(pdg == 0 || abs(pdg) == 12 || abs(pdg) == 14);

    if(appr) {
      if(pdg == +12)
        return Ratio(fApprDataCCNuE*SignFraction(+12),
                     fApprNDCCNuE);
      if(pdg == -12)
        return Ratio(fApprDataCCNuE*SignFraction(-12),
                     fApprNDCCANuE);
      if(pdg == +14)
        return Ratio((fApprDataCCNuMu + fApprDataQENuMu)*SignFraction(+14),
                     fApprNDCCNuMu + fApprNDQENuMu);
      if(pdg == -14)
        return Ratio((fApprDataCCNuMu + fApprDataQENuMu)*SignFraction(-14),
                     fApprNDCCANuMu + fApprNDQEANuMu);
    }

    if(pdg == +12) {
      if(fSignSplit) {
        return Ratio(fDecomp->NueComponent(), fSurvNDCCNuE);
      }
      return Ratio(fDecomp->NueComponent() + fDecomp->AntiNueComponent(),
                   fSurvNDCCNuE);
    }
    if(pdg == -12) {
      if(fSignSplit) {
        return Ratio(fDecomp->AntiNueComponent(), fSurvNDCCANuE);
      }
      return Ratio(fDecomp->NueComponent() - fDecomp->NueComponent(),
                   fSurvNDCCNuE);
    }
    if(pdg == +14) {
      if(fSignSplit) {
        return Ratio(fDecomp->NumuComponent(), fSurvNDCCNuMu);
      }
      return Ratio(fDecomp->NumuComponent() + fDecomp->AntiNumuComponent(),
                   fSurvNDCCNuMu);
    }
    if(pdg == -14) {
      if(fSignSplit) {
        return Ratio(fDecomp->AntiNumuComponent(), fSurvNDCCANuMu);
      }
      return Ratio(fDecomp->NumuComponent() - fDecomp->NumuComponent(),
                   fSurvNDCCNuMu);
    }

    return Ratio(fDecomp->NCComponent(), fSurvNDNC);
  }

  //----------------------------------------------------------------------
  Ratio NDDataMCRatioExtrap::Purity(int pdg) const
  {
    assert(abs(pdg) == 12 || abs(pdg) == 14);

    if(pdg == +12)
      return Ratio(fApprFDPurEffNumCCNuE, fApprFDPurDenCCNuE, true);
    if(pdg == -12)
      return Ratio(fApprFDPurEffNumCCANuE, fApprFDPurDenCCANuE, true);
    if(pdg == +14)
      return Ratio(fApprFDPurEffNumCCNuMu + fApprFDPurEffNumQENuMu,
                   fApprFDPurDenCCNuMu    + fApprFDPurDenQENuMu);

    return Ratio(fApprFDPurEffNumCCANuMu + fApprFDPurEffNumQEANuMu,
                 fApprFDPurDenCCANuMu    + fApprFDPurDenQEANuMu);
  }

  //----------------------------------------------------------------------
  Ratio NDDataMCRatioExtrap::Efficiency(int pdg) const
  {
    assert(abs(pdg) == 12 || abs(pdg) == 14);

    if(pdg == +12)
      return Ratio(fApprFDPurEffNumCCNuE, fApprFDEffDenCCNuE, true);
    if(pdg == -12)
      return Ratio(fApprFDPurEffNumCCANuE, fApprFDEffDenCCANuE, true);
    if(pdg == +14)
      return Ratio(fApprFDPurEffNumCCNuMu + fApprFDPurEffNumQENuMu,
                   fApprFDEffDenCCNuMu    + fApprFDEffDenQENuMu);

    return Ratio(fApprFDPurEffNumCCANuMu + fApprFDPurEffNumQEANuMu,
                 fApprFDEffDenCCANuMu    + fApprFDEffDenQEANuMu);
  }

  //----------------------------------------------------------------------
  void NDDataMCRatioExtrap::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;

    dir->cd();

    TObjString("NDDataMCRatioExtrap").Write("type");

    TObjString(fSignSplit ? "true" : "false").Write("sign_split");

    fDecomp->SaveTo(dir->mkdir("surv_decomp"));

    fApprDataCCNuE.SaveTo(dir->mkdir("app_data_nue"));
    fApprDataCCNuMu.SaveTo(dir->mkdir("app_data_numu_cc"));
    fApprDataQENuMu.SaveTo(dir->mkdir("app_data_numu_qe"));

    fSurvNDNC.SaveTo(dir->mkdir("surv_nd_nc"));
    fSurvFDNC.SaveTo(dir->mkdir("surv_fd_nc"));

    fSurvNDCCNuE.SaveTo(dir->mkdir("surv_nd_nue"));
    fSurvNDCCANuE.SaveTo(dir->mkdir("surv_nd_anue"));
    fSurvNDCCNuMu.SaveTo(dir->mkdir("surv_nd_numu"));
    fSurvNDCCANuMu.SaveTo(dir->mkdir("surv_nd_anumu"));
    fOSpecSurvFDCCNuE.SaveTo(dir->mkdir("surv_fd_nue"));
    fOSpecSurvFDCCANuE.SaveTo(dir->mkdir("surv_fd_anue"));
    fOSpecSurvFDCCNuMu.SaveTo(dir->mkdir("surv_fd_numu"));
    fOSpecSurvFDCCANuMu.SaveTo(dir->mkdir("surv_fd_anumu"));

    fApprNDCCNuE.SaveTo(dir->mkdir("app_nd_nue"));
    fApprNDCCANuE.SaveTo(dir->mkdir("app_nd_anue"));
    fApprNDCCNuMu.SaveTo(dir->mkdir("app_nd_numu_cc"));
    fApprNDCCANuMu.SaveTo(dir->mkdir("app_nd_anumu_cc"));
    fApprNDQENuMu.SaveTo(dir->mkdir("app_nd_numu_qe"));
    fApprNDQEANuMu.SaveTo(dir->mkdir("app_nd_anumu_qe"));
    fOSpecApprFDCCNuE.SaveTo(dir->mkdir("app_fd_nue"));
    fOSpecApprFDCCANuE.SaveTo(dir->mkdir("app_fd_anue"));
    fOSpecApprFDCCNuMu.SaveTo(dir->mkdir("app_fd_numu_cc"));
    fOSpecApprFDCCANuMu.SaveTo(dir->mkdir("app_fd_anumu_cc"));
    fOSpecApprFDQENuMu.SaveTo(dir->mkdir("app_fd_numu_qe"));
    fOSpecApprFDQEANuMu.SaveTo(dir->mkdir("app_fd_anumu_qe"));

    fOSpecOscdFDCCNuE.SaveTo(dir->mkdir("oscd_fd_nue"));
    fOSpecOscdFDCCANuE.SaveTo(dir->mkdir("oscd_fd_anue"));
    fOSpecOscdFDCCNuMu.SaveTo(dir->mkdir("oscd_fd_numu"));
    fOSpecOscdFDCCANuMu.SaveTo(dir->mkdir("oscd_fd_anumu"));
    fOSpecOscdFDCCNuTau.SaveTo(dir->mkdir("oscd_fd_nutau"));
    fOSpecOscdFDCCANuTau.SaveTo(dir->mkdir("oscd_fd_anutau"));

    fApprFDPurEffNumCCNuE.SaveTo(dir->mkdir("nue_pur_eff_num"));
    fApprFDPurEffNumCCANuE.SaveTo(dir->mkdir("anue_pur_eff_num"));
    fApprFDPurDenCCNuE.SaveTo(dir->mkdir("nue_pur_den"));
    fApprFDPurDenCCANuE.SaveTo(dir->mkdir("anue_pur_den"));
    fApprFDEffDenCCNuE.SaveTo(dir->mkdir("nue_eff_den"));
    fApprFDEffDenCCANuE.SaveTo(dir->mkdir("anue_eff_den"));
    fApprFDPurEffNumCCNuMu.SaveTo(dir->mkdir("numu_cc_pur_eff_num"));
    fApprFDPurEffNumCCANuMu.SaveTo(dir->mkdir("anumu_cc_pur_eff_num"));
    fApprFDPurDenCCNuMu.SaveTo(dir->mkdir("numu_cc_pur_den"));
    fApprFDPurDenCCANuMu.SaveTo(dir->mkdir("anumu_cc_pur_den"));
    fApprFDEffDenCCNuMu.SaveTo(dir->mkdir("numu_cc_eff_den"));
    fApprFDEffDenCCANuMu.SaveTo(dir->mkdir("anumu_cc_eff_den"));
    fApprFDPurEffNumQENuMu.SaveTo(dir->mkdir("numu_qe_pur_eff_num"));
    fApprFDPurEffNumQEANuMu.SaveTo(dir->mkdir("anumu_qe_pur_eff_num"));
    fApprFDPurDenQENuMu.SaveTo(dir->mkdir("numu_qe_pur_den"));
    fApprFDPurDenQEANuMu.SaveTo(dir->mkdir("anumu_qe_pur_den"));
    fApprFDEffDenQENuMu.SaveTo(dir->mkdir("numu_qe_eff_den"));
    fApprFDEffDenQEANuMu.SaveTo(dir->mkdir("anumu_qe_eff_den"));

    tmp->cd();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<NDDataMCRatioExtrap> NDDataMCRatioExtrap::LoadFrom(TDirectory* dir)
  {
    std::unique_ptr<NDDataMCRatioExtrap> ret(new NDDataMCRatioExtrap);

    // This is a lot of repetitive typing. Define some macros
#define LOAD_OSC(FIELD, LABEL) assert(dir->GetDirectory(LABEL)); ret->FIELD = *OscillatableSpectrum::LoadFrom(dir->GetDirectory(LABEL));
#define LOAD_SPECT(FIELD, LABEL) assert(dir->GetDirectory(LABEL)); ret->FIELD = *Spectrum::LoadFrom(dir->GetDirectory(LABEL));

    TObjString* signtag = (TObjString*)dir->Get("sign_split");
    ret->fSignSplit = !(signtag->GetString().CompareTo("true"));

    ret->fDecomp = ana::LoadFrom<IDecomp>(dir->GetDirectory("surv_decomp")).release();

    LOAD_SPECT(fApprDataCCNuE,  "app_data_nue");
    LOAD_SPECT(fApprDataCCNuMu, "app_data_numu_cc");
    LOAD_SPECT(fApprDataQENuMu, "app_data_numu_qe");

    LOAD_SPECT(fSurvNDNC, "surv_nd_nc");
    LOAD_SPECT(fSurvFDNC, "surv_fd_nc");

    LOAD_SPECT(fSurvNDCCNuE,   "surv_nd_nue");
    LOAD_SPECT(fSurvNDCCANuE,  "surv_nd_anue");
    LOAD_SPECT(fSurvNDCCNuMu,  "surv_nd_numu");
    LOAD_SPECT(fSurvNDCCANuMu, "surv_nd_anumu");
    LOAD_OSC(fOSpecSurvFDCCNuE,   "surv_fd_nue");
    LOAD_OSC(fOSpecSurvFDCCANuE,  "surv_fd_anue");
    LOAD_OSC(fOSpecSurvFDCCNuMu,  "surv_fd_numu");
    LOAD_OSC(fOSpecSurvFDCCANuMu, "surv_fd_anumu");

    LOAD_SPECT(fApprNDCCNuE,      "app_nd_nue");
    LOAD_SPECT(fApprNDCCANuE,     "app_nd_anue");
    LOAD_SPECT(fApprNDCCNuMu,     "app_nd_numu_cc");
    LOAD_SPECT(fApprNDCCANuMu,    "app_nd_anumu_cc");
    LOAD_SPECT(fApprNDQENuMu,     "app_nd_numu_qe");
    LOAD_SPECT(fApprNDQEANuMu,    "app_nd_anumu_qe");
    LOAD_OSC(fOSpecApprFDCCNuE,   "app_fd_nue");
    LOAD_OSC(fOSpecApprFDCCANuE,  "app_fd_anue");
    LOAD_OSC(fOSpecApprFDCCNuMu,  "app_fd_numu_cc");
    LOAD_OSC(fOSpecApprFDCCANuMu, "app_fd_anumu_cc");
    LOAD_OSC(fOSpecApprFDQENuMu,  "app_fd_numu_qe");
    LOAD_OSC(fOSpecApprFDQEANuMu, "app_fd_anumu_qe");

    LOAD_OSC(fOSpecOscdFDCCNuE,    "oscd_fd_nue");
    LOAD_OSC(fOSpecOscdFDCCANuE,   "oscd_fd_anue");
    LOAD_OSC(fOSpecOscdFDCCNuMu,   "oscd_fd_numu");
    LOAD_OSC(fOSpecOscdFDCCANuMu,  "oscd_fd_anumu");
    LOAD_OSC(fOSpecOscdFDCCNuTau,  "oscd_fd_nutau");
    LOAD_OSC(fOSpecOscdFDCCANuTau, "oscd_fd_anutau");

    LOAD_SPECT(fApprFDPurEffNumCCNuE,   "nue_pur_eff_num");
    LOAD_SPECT(fApprFDPurEffNumCCANuE,  "anue_pur_eff_num");
    LOAD_SPECT(fApprFDPurDenCCNuE,      "nue_pur_den");
    LOAD_SPECT(fApprFDPurDenCCANuE,     "anue_pur_den");
    LOAD_SPECT(fApprFDEffDenCCNuE,      "nue_eff_den");
    LOAD_SPECT(fApprFDEffDenCCANuE,     "anue_eff_den");
    LOAD_SPECT(fApprFDPurEffNumCCNuMu,  "numu_cc_pur_eff_num");
    LOAD_SPECT(fApprFDPurEffNumCCANuMu, "anumu_cc_pur_eff_num");
    LOAD_SPECT(fApprFDPurDenCCNuMu,     "numu_cc_pur_den");
    LOAD_SPECT(fApprFDPurDenCCANuMu,    "anumu_cc_pur_den");
    LOAD_SPECT(fApprFDEffDenCCNuMu,     "numu_cc_eff_den");
    LOAD_SPECT(fApprFDEffDenCCANuMu,    "anumu_cc_eff_den");
    LOAD_SPECT(fApprFDPurEffNumQENuMu,  "numu_qe_pur_eff_num");
    LOAD_SPECT(fApprFDPurEffNumQEANuMu, "anumu_qe_pur_eff_num");
    LOAD_SPECT(fApprFDPurDenQENuMu,     "numu_qe_pur_den");
    LOAD_SPECT(fApprFDPurDenQEANuMu,    "anumu_qe_pur_den");
    LOAD_SPECT(fApprFDEffDenQENuMu,     "numu_qe_eff_den");
    LOAD_SPECT(fApprFDEffDenQEANuMu,    "anumu_qe_eff_den");

    return ret;
  }

  //----------------------------------------------------------------------
  void NDDataMCRatioExtrap::NoSignSplitCorrection()
  {
    // Add the anti-neutrino events to the neutrino spectra
    fSurvNDCCNuE  += fSurvNDCCANuE;
    fSurvNDCCNuMu += fSurvNDCCANuMu;
    fOSpecSurvFDCCNuE  += fOSpecSurvFDCCANuE;
    fOSpecSurvFDCCNuMu += fOSpecSurvFDCCANuMu;

    fApprNDCCNuE  += fApprNDCCANuE;
    fApprNDCCNuMu += fApprNDCCANuMu;
    fApprNDQENuMu += fApprNDQEANuMu;
    fOSpecApprFDCCNuE  += fOSpecApprFDCCANuE;
    fOSpecApprFDCCNuMu += fOSpecApprFDCCANuMu;
    fOSpecApprFDQENuMu += fOSpecApprFDQEANuMu;

    fOSpecOscdFDCCNuE   += fOSpecOscdFDCCANuE;
    fOSpecOscdFDCCNuMu  += fOSpecOscdFDCCANuMu;
    fOSpecOscdFDCCNuTau += fOSpecOscdFDCCANuTau;

    fApprFDPurEffNumCCNuE  += fApprFDPurEffNumCCANuE;
    fApprFDPurDenCCNuE     += fApprFDPurDenCCANuE;
    fApprFDEffDenCCNuE     += fApprFDEffDenCCANuE;
    fApprFDPurEffNumCCNuMu += fApprFDPurEffNumCCANuMu;
    fApprFDPurDenCCNuMu    += fApprFDPurDenCCANuMu;
    fApprFDEffDenCCNuMu    += fApprFDEffDenCCANuMu;
    fApprFDPurEffNumQENuMu += fApprFDPurEffNumQEANuMu;
    fApprFDPurDenQENuMu    += fApprFDPurDenQEANuMu;
    fApprFDEffDenQENuMu    += fApprFDEffDenQEANuMu;

    // Clear all of the anti-neutrino plots
    // This causes the anti-neutrino component functions to return 0 events
    fSurvNDCCANuE.Clear();
    fSurvNDCCANuMu.Clear();
    fOSpecSurvFDCCANuE.Clear();
    fOSpecSurvFDCCANuMu.Clear();

    fApprNDCCANuE.Clear();
    fApprNDCCANuMu.Clear();
    fApprNDQEANuMu.Clear();
    fOSpecApprFDCCANuE.Clear();
    fOSpecApprFDCCANuMu.Clear();
    fOSpecApprFDQEANuMu.Clear();

    fOSpecOscdFDCCANuE.Clear();
    fOSpecOscdFDCCANuMu.Clear();
    fOSpecOscdFDCCANuTau.Clear();

    fApprFDPurEffNumCCANuE.Clear();
    fApprFDPurDenCCANuE.Clear();
    fApprFDEffDenCCANuE.Clear();
    fApprFDPurEffNumCCANuMu.Clear();
    fApprFDPurDenCCANuMu.Clear();
    fApprFDEffDenCCANuMu.Clear();
    fApprFDPurEffNumQEANuMu.Clear();
    fApprFDPurDenQEANuMu.Clear();
    fApprFDEffDenQEANuMu.Clear();

    return;
  }
}
