#include "CAFAna/NumuSurface.h"

#include "CAFAna/FitVars.h"

#include "OscLib/func/IOscCalculator.h"

#include "TCanvas.h"
#include "TH2.h"
#include "TLatex.h"
#include "TPad.h"

#include <cmath>

namespace ana
{
  //----------------------------------------------------------------------
  NumuSurface::NumuSurface(const Experiment* expt,
			   osc::IOscCalculatorAdjustable* calc,
			   int nbinsx, double xmin, double xmax,
			   int nbinsy, double ymin, double ymax,
			   const std::vector<const FitVar*>& margVars,
			   const std::vector<Syst>& margSysts)
    : fPadNorm(0), fPadInv(0)
  {
    calc->SetDmsq32(+fabs(calc->GetDmsq32()));
    fSurfNorm = new Surface(expt, calc,
			    &kFitSinSqTheta23, nbinsx,  xmin,  xmax,
			    &kFitDmSq32Scaled, nbinsy, +ymin, +ymax,
			    margVars, margSysts);
    calc->SetDmsq32(-fabs(calc->GetDmsq32()));
    fSurfInv  = new Surface(expt, calc,
			    &kFitSinSqTheta23, nbinsx,  xmin,  xmax,
			    &kFitDmSq32Scaled, nbinsy, -ymax, -ymin,
			    margVars, margSysts);
  }

  //---------------------------------------------------------------------
  NumuSurface::~NumuSurface()
  {
    delete fSurfNorm;
    delete fSurfInv;
    // Can't delete the fPad members unless we record who actually created them
  }

  //---------------------------------------------------------------------
  void NumuSurface::EnsurePad() const
  {
    if(!gPad) new TCanvas;

    // The pads already exist, remember them
    if(gPad->FindObject("norm_pad")){
      fPadNorm = (TPad*)gPad->FindObject("norm_pad");
      fPadInv = (TPad*)gPad->FindObject("inv_pad");
      return;
    }

    // Otherwise, create the two pads
    fPadNorm = new TPad("norm_pad", "", 0, 0, 1, 1);
    fPadInv = new TPad("inv_pad", "", 0, 0, 1, 1);

    fPadNorm->SetFillStyle(0);
    fPadInv->SetFillStyle(0);

    fPadNorm->SetBottomMargin(.5);
    fPadInv->SetTopMargin(.5);

    fPadNorm->Draw();
    fPadInv->Draw();

    // This is the first time we're drawing, so need to turn off the normal
    // hierachy's x-axis
    fSurfNorm->fHist->GetXaxis()->SetLabelSize(0);
    fSurfNorm->fHist->GetXaxis()->SetTitle("");

    // Draw our own y-axis title, shared between the pads
    // TODO, for some reason this disappears when you resize the canvas, until
    // you click in the canvas
    TLatex* dmsq = new TLatex(0, .5, fSurfNorm->fHist->GetYaxis()->GetTitle());
    dmsq->SetTextAlign(23);
    dmsq->SetTextSize(fSurfNorm->fHist->GetYaxis()->GetTitleSize());
    dmsq->SetTextAngle(90);
    dmsq->Draw();

    // So don't need the usual ones
    fSurfNorm->fHist->GetYaxis()->SetTitle("");
    fSurfInv ->fHist->GetYaxis()->SetTitle("");
  }

  //---------------------------------------------------------------------
  void NumuSurface::Draw() const
  {
    EnsurePad();

    TVirtualPad* top = gPad;

    fPadNorm->cd();
    fSurfNorm->Draw();
    fPadInv->cd();
    fSurfInv->Draw();
    top->cd(0);

    gPad->Update();
  }

  //---------------------------------------------------------------------
  void NumuSurface::DrawBestFit(Color_t color) const
  {
    EnsurePad();

    TVirtualPad* top = gPad;

    if(fSurfNorm->MinChi() < fSurfInv->MinChi()){
      fPadNorm->cd();
      fSurfNorm->DrawBestFit(color);
    }
    else{
      fPadInv->cd();
      fSurfInv->DrawBestFit(color);
    }
    top->cd(0);

    gPad->Update();
  }

  //----------------------------------------------------------------------
  double NumuSurface::MinChi() const
  {
    return std::min(fSurfNorm->MinChi(), fSurfInv->MinChi());
  }

  //----------------------------------------------------------------------
  void NumuSurface::DrawContour68(Style_t style, Color_t color, double minchi)
  {
    DrawContour(Gaussian68Percent2D(*fSurfNorm),
		Gaussian68Percent2D(*fSurfInv),
		style, color, minchi);
  }

  //----------------------------------------------------------------------
  void NumuSurface::DrawContour90(Style_t style, Color_t color, double minchi)
  {
    DrawContour(Gaussian90Percent2D(*fSurfNorm),
		Gaussian90Percent2D(*fSurfInv),
		style, color, minchi);
  }

  //----------------------------------------------------------------------
  void NumuSurface::DrawContour(TH2* fcNorm, TH2* fcInv,
				Style_t style, Color_t color,
				double minchi)
  {
    EnsurePad();

    // If user didn't explicitly pass a minimum chisq, each contour has to use
    // the same value, the minimum from the two of them.
    if(minchi < 0) minchi = MinChi();

    TVirtualPad* top = gPad;
    fPadNorm->cd();
    fSurfNorm->DrawContour(fcNorm, style, color, minchi);
    fPadInv->cd();
    fSurfInv->DrawContour(fcInv, style, color, minchi);
    top->cd(0);

    gPad->Update();
  }

  //----------------------------------------------------------------------
  void NumuSurface::SetTitle(const char* str)
  {
    fSurfNorm->SetTitle(str);
  }
} // namespace
