#include "CAFAna/OscillatableSpectrum.h"

#include "CAFAna/Binning.h"
#include "CAFAna/Ratio.h"
#include "CAFAna/Utilities.h"

#include "TDirectory.h"
#include "TH2.h"
#include "TObjString.h"

#include <cassert>
#include <memory>

namespace ana
{
  //----------------------------------------------------------------------
  OscillatableSpectrum::
  OscillatableSpectrum(std::string label, const Binning& bins,
                       SpectrumLoaderBase& loader,
                       const Var& var,
                       const Cut& cut,
                       const Var& wei)
    : ReweightableSpectrum()
  {
    DontAddDirectory guard;

    fPOT = 0;

    if(bins.IsSimple()){
      fHist = new TH2D("", (";"+label+";True Energy (GeV);Events").c_str(),
                       bins.NBins(), bins.Min(), bins.Max(),
                       kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);
    }
    else{
      fHist = new TH2D("", (";"+label+";True Energy (GeV);Events").c_str(),
                       bins.NBins(), &bins.Edges()[0],
                       kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);
    }

    loader.AddOscillatableSpectrum(*this, var, cut, wei);
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum::OscillatableSpectrum(SpectrumLoaderBase& loader,
                                             const HistAxis& axis,
                                             const Cut& cut,
                                             const Var& wei)
    : OscillatableSpectrum(axis.label, axis.bins, loader, axis.var, cut, wei)
  {
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum::OscillatableSpectrum(std::string label,
                                             const Binning& bins)
    : ReweightableSpectrum()
  {
    DontAddDirectory guard;

    fPOT = 0;

    if(bins.IsSimple()){
      fHist = new TH2D("", (";"+label+";True Energy (GeV);Events").c_str(),
                       bins.NBins(), bins.Min(), bins.Max(),
                       kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);
    }
    else{
      fHist = new TH2D("", (";"+label+";True Energy (GeV);Events").c_str(),
                       bins.NBins(), &bins.Edges()[0],
                       kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);
    }
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum::OscillatableSpectrum(std::string label, double pot,
                                             const Binning& bins)
    : ReweightableSpectrum()
  {
    DontAddDirectory guard;

    fPOT = pot;

    if(bins.IsSimple()){
      fHist = new TH2D("", (";"+label+";True Energy (GeV);Events").c_str(),
                       bins.NBins(), bins.Min(), bins.Max(),
                       kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);
    }
    else{
      fHist = new TH2D("", (";"+label+";True Energy (GeV);Events").c_str(),
                       bins.NBins(), &bins.Edges()[0],
                       kTrueEnergyBins.NBins(), &kTrueEnergyBins.Edges()[0]);
    }
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum::OscillatableSpectrum(TH2* h, double pot, std::string label)
    : ReweightableSpectrum(h, pot, label)
  {
  }

  //----------------------------------------------------------------------
  Spectrum OscillatableSpectrum::Oscillated(const OscCurve& curve) const
  {
    std::unique_ptr<TH1> Ps(curve.ToTH1()); // Don't leak the histogram

    return WeightedBy(Ps.get());
  }

  //----------------------------------------------------------------------
  Spectrum OscillatableSpectrum::Oscillated(osc::IOscCalculator* calc,
                                            int from, int to) const
  {
    const OscCurve curve(calc, from, to);
    return Oscillated(curve);
  }

  //----------------------------------------------------------------------
  void OscillatableSpectrum::ReweightToTrueSpectrum(const Spectrum& target)
  {
    // This is a big component of what extrapolations do, so it has been
    // optimized for speed

    Ratio corr(target, TrueEnergy());
    TH1D* hcorr = corr.ToTH1();

    assert(hcorr->GetNbinsX() == fHist->GetNbinsY());

    const int X = fHist->GetNbinsX();
    const int Y = fHist->GetNbinsY();

    // Direct access to the bins is faster
    double* histArr = fHist->GetArray();
    double* corrArr = hcorr->GetArray();

    int bin = 0;
    for(int y = 0; y < Y+2; ++y){
      const double w = corrArr[y];
      for(int x = 0; x < X+2; ++x){
        // Our loops go over the bins in the order they are internally in
        // fHist, and we do overflows, so we keep up exactly. If you get
        // paranoid, reenable this briefly.

        // assert(bin == fHist->GetBin(x, y));

        histArr[bin] *= w;
        ++bin;
      }
    }

    delete hcorr;
  }

  //----------------------------------------------------------------------
  void OscillatableSpectrum::ReweightToRecoSpectrum(const Spectrum& target)
  {
    // This is a big component of what extrapolations do, so it has been
    // optimized for speed

    Ratio corr(target, Unoscillated());
    TH1D* hcorr = corr.ToTH1();

    assert(hcorr->GetNbinsX() == fHist->GetNbinsX());

    const int X = fHist->GetNbinsX();
    const int Y = fHist->GetNbinsY();

    // Direct access to the bins is faster
    double* histArr = fHist->GetArray();
    double* corrArr = hcorr->GetArray();

    int bin = 0;
    for(int y = 0; y < Y+2; ++y){
      for(int x = 0; x < X+2; ++x){
        // Our loops go over the bins in the order they are internally in
        // fHist, and we do overflows, so we keep up exactly. If you get
        // paranoid, reenable this briefly.

        // assert(bin == fHist->GetBin(x, y));

        histArr[bin] *= corrArr[x];
        ++bin;
      }
    }

    delete hcorr;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum& OscillatableSpectrum::operator+=(const OscillatableSpectrum& rhs)
  {
    if(rhs.fPOT){
      fHist->Add(rhs.fHist, fPOT/rhs.fPOT);
    }
    else{
      // How can it have events but no POT?
      assert(rhs.fHist->Integral() == 0);
    }
    return *this;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum OscillatableSpectrum::operator+(const OscillatableSpectrum& rhs) const
  {
    OscillatableSpectrum ret = *this;
    ret += rhs;
    return ret;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum& OscillatableSpectrum::operator-=(const OscillatableSpectrum& rhs)
  {
    if(rhs.fPOT){
      fHist->Add(rhs.fHist, -fPOT/rhs.fPOT);
    }
    else{
      // How can it have events but no POT?
      assert(rhs.fHist->Integral() == 0);
    }
    return *this;
  }

  //----------------------------------------------------------------------
  OscillatableSpectrum OscillatableSpectrum::operator-(const OscillatableSpectrum& rhs) const
  {
    OscillatableSpectrum ret = *this;
    ret -= rhs;
    return ret;
  }

  //----------------------------------------------------------------------
  void OscillatableSpectrum::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();

    TObjString("OscillatableSpectrum").Write("type");

    fHist->Write("hist");
    TH1D hPot("", "", 1, 0, 1);
    hPot.Fill(.5, fPOT);
    hPot.Write("pot");
    tmp->cd();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<OscillatableSpectrum> OscillatableSpectrum::LoadFrom(TDirectory* dir)
  {
    TObjString* tag = (TObjString*)dir->Get("type");
    assert(tag);
    assert(tag->GetString() == "OscillatableSpectrum");

    TH2D* spect = (TH2D*)dir->Get("hist");
    assert(spect);
    TH1* hPot = (TH1*)dir->Get("pot");
    assert(hPot);
    return std::unique_ptr<OscillatableSpectrum>(new OscillatableSpectrum(spect, hPot->GetBinContent(1)));
  }
}
