#include "CAFAna/Plots.h"

#include "CAFAna/Prediction.h"
#include "CAFAna/Spectrum.h"
#include "CAFAna/SystShifts.h"

#include "Utilities/func/MathUtil.h"

#include "TCanvas.h"
#include "TGraphAsymmErrors.h"
#include "TH2.h"
#include "TLegend.h"
#include "TLine.h"
#include "TList.h"

#include <cmath>

namespace ana
{
  //----------------------------------------------------------------------
  void DataMCComparison(const Spectrum& data, const Spectrum& mc)
  {
    TH1* hMC = mc.ToTH1(data.POT());
    hMC->SetLineColor(kRed);

    TH1* hData = data.ToTH1(data.POT());
    hData->Sumw2();
    hData->SetMarkerStyle(kFullCircle);

    // Need to draw the biggest thing first to get correct axis limits
    if(hMC->GetMaximum() > hData->GetMaximum()+sqrt(hData->GetMaximum())){
      hMC->Draw("hist");
      hData->Draw("ep same");
    }
    else{
      hData->Draw("ep");
      hMC->Draw("hist same");
      hData->Draw("ep same"); // Data always goes on top
    }

    gPad->Update();
  }

  //----------------------------------------------------------------------
  void DataMCComparisonAreaNormalized(const Spectrum& data, const Spectrum& mc)
  {
    TH1* hMC = mc.ToTH1(mc.POT());
    hMC->SetLineColor(kRed);

    TH1* hData = data.ToTH1(data.POT());
    hData->Sumw2();
    hData->SetMarkerStyle(kFullCircle);

    hMC->Scale(hData->Integral()/hMC->Integral());

    // Need to draw the biggest thing first to get correct axis limits
    if(hMC->GetMaximum() > hData->GetMaximum()+sqrt(hData->GetMaximum())){
      hMC->Draw("hist");
      hData->Draw("ep same");
    }
    else{
      hData->Draw("ep");
      hMC->Draw("hist same");
      hData->Draw("ep same"); // Data always goes on top
    }

    gPad->Update();
  }

  //----------------------------------------------------------------------
  void DataMCComparisonComponents(const Spectrum& data,
                                  const IPrediction* mc,
                                  osc::IOscCalculator* calc)
  {
    TH1* hMC = mc->Predict(calc).ToTH1(data.POT());
    hMC->SetLineColor(kBlack);

    TH1* hNC = mc->PredictComponent(calc,
                                    Flavors::kAll,
                                    Current::kNC,
                                    Sign::kBoth).ToTH1(data.POT());
    hNC->SetLineColor(kBlue);

    TH1* hCC = mc->PredictComponent(calc,
                                    Flavors::kAllNuMu,
                                    Current::kCC,
                                    Sign::kBoth).ToTH1(data.POT());
    hCC->SetLineColor(kGray+2);

    TH1* hBeam = mc->PredictComponent(calc,
                                      Flavors::kNuEToNuE,
                                      Current::kCC,
                                      Sign::kBoth).ToTH1(data.POT());
    hBeam->SetLineColor(kMagenta);

    TH1* hCosmic = mc->PredictCosmicComponent().ToTH1(data.POT());
    hCosmic->SetLineColor(kViolet);

    TH1* hData = data.ToTH1(data.POT());
    hData->SetMarkerStyle(kFullCircle);

    // Need to draw the biggest thing first to get correct axis limits
    if(hMC->GetMaximum() > hData->GetMaximum()+sqrt(hData->GetMaximum())){
      hMC->Draw("hist");
      hData->Draw("ep same");
    }
    else{
      hData->Draw("ep");
      hMC->Draw("hist same");
      hData->Draw("ep same"); // Data always goes on top
    }

    hNC->Draw("hist same");
    hCC->Draw("hist same");
    hBeam->Draw("hist same");
    hCosmic->Draw("hist same");

    gPad->Update();
  }

  //----------------------------------------------------------------------
  void RatioPlot(const Spectrum& data,
                 const Spectrum& expected,
                 const Spectrum& fit)
  {
    TH1* hExp = expected.ToTH1(data.POT());

    TH1* hData = data.ToTH1(data.POT());
    hData->Sumw2();
    hData->SetMarkerStyle(kFullCircle);
    hData->Divide(hExp);

    TH1* hFit = fit.ToTH1(data.POT());
    hFit->SetLineColor(kRed);
    hFit->Divide(hExp);

    hFit->Draw("][");
    hData->Draw("same");
    hFit->GetYaxis()->SetRangeUser(0, 1.2);
    hFit->GetYaxis()->SetTitle("Ratio to expectation");

    TLine* one = new TLine(hData->GetXaxis()->GetXmin(), 1,
                           hData->GetXaxis()->GetXmax(), 1);
    one->SetLineWidth(2);
    one->SetLineStyle(7);
    one->Draw();

    gPad->Update();
  }

 //----------------------------------------------------------------------
  void PlotWithSystErrorBand(IPrediction* pred,
                             const std::vector<Syst>& systs,
                             osc::IOscCalculator* calc,
                             double pot,
                             int col, int errCol)
  {
    if(col == -1) col = kRed;
    if(errCol == -1) errCol = col-7; // hopefully a lighter version

    TH1* nom = pred->Predict(calc).ToTH1(pot);

    std::vector<TH1*> ups, dns;

    for(Syst syst: systs){
      SystShifts shifts;
      shifts.SetShift(syst, +1);
      ups.push_back(pred->PredictSyst(calc, shifts).ToTH1(pot));
      shifts.SetShift(syst, -1);
      dns.push_back(pred->PredictSyst(calc, shifts).ToTH1(pot));
    }

    nom->SetLineColor(col);
    nom->Draw("hist ]["); // Set the axes up

    TGraphAsymmErrors* g = new TGraphAsymmErrors;

    for(int binIdx = 0; binIdx < nom->GetNbinsX()+2; ++binIdx){
      const double y = nom->GetBinContent(binIdx);
      g->SetPoint(binIdx, nom->GetXaxis()->GetBinCenter(binIdx), y);

      const double w = nom->GetXaxis()->GetBinWidth(binIdx);

      double errUp = 0;
      double errDn = 0;

      for(unsigned int systIdx = 0; systIdx < ups.size(); ++systIdx){
        double hi = ups[systIdx]->GetBinContent(binIdx)-y;
        double lo = y-dns[systIdx]->GetBinContent(binIdx);

        if(lo <= 0 && hi <= 0) std::swap(lo, hi);

        errUp += hi*hi;
        errDn += lo*lo;

        // TODO: what happens if they're both high or both low?
      } // end for systIdx

      g->SetPointError(binIdx, w/2, w/2, sqrt(errUp), sqrt(errDn));
    } // end for i

    g->SetFillColor(errCol);
    g->Draw("e2 same");

    nom->Draw("hist ][ same");

    for(TH1* up: ups) delete up;
    for(TH1* dn: dns) delete dn;
  }

  //----------------------------------------------------------------------
  /// Helper for \ref AutoPlaceLegend
  double PointDistanceToBox(double x, double y,
                            double x0, double y0, double x1, double y1)
  {
    // Inside
    if(x > x0 && x < x1 && y > y0 && y < y1) return 0;

    // Corners
    double d = util::sqr(x-x0)+util::sqr(y-y0);
    d = std::min(d, util::sqr(x-x1)+util::sqr(y-y0));
    d = std::min(d, util::sqr(x-x1)+util::sqr(y-y1));
    d = std::min(d, util::sqr(x-x0)+util::sqr(y-y1));

    // Top and bottom edges
    if(x > x0 && x < x1){
      d = std::min(d, util::sqr(y-y0));
      d = std::min(d, util::sqr(y-y1));
    }
    // Left and right
    if(y > y0 && y < y1){
      d = std::min(d, util::sqr(x-x0));
      d = std::min(d, util::sqr(x-x1));
    }

    return d;
  }

  //----------------------------------------------------------------------
  TLegend* AutoPlaceLegend(double dx, double dy)
  {
    gPad->Update();

    // Convert requested width and height into physics coordinates
    dx *= (gPad->GetX2()-gPad->GetX1());
    dy *= (gPad->GetY2()-gPad->GetY1());

    // Range of axes in physics units
    const double x0 = gPad->GetUxmin();
    const double x1 = gPad->GetUxmax();
    const double y0 = gPad->GetUymin();
    const double y1 = gPad->GetUymax();

    const double X = x1-x0;
    const double Y = y1-y0;

    double bestd = 0;
    double bestx = 0;
    double besty = 0;

    for(int fallback = false; fallback <= true; ++fallback){
      for(double x = x0+dx/2; x < x1-dx/2; x += X/50){
        for(double y = y0+dy/2; y < y1-dy/2; y += Y/50){
          double d = 999999;

          // Repel from edges
          d = std::min(d, util::sqr((x-dx/2-x0)/X));
          d = std::min(d, util::sqr((x+dx/2-x1)/X));
          d = std::min(d, util::sqr((y-dy/2-y0)/Y));
          d = std::min(d, util::sqr((y+dy/2-y1)/Y));
          if(d < bestd) continue;

          TIter next(gPad->GetListOfPrimitives());
          while(TObject* obj = next()){

            if(!obj->InheritsFrom(TH1::Class())) continue;
            if( obj->InheritsFrom(TH2::Class())) continue;

            TH1* h = (TH1*)obj;

            for(int n = 1; n <= h->GetNbinsX(); ++n){
              const double px = h->GetBinCenter(n);
              const double py = h->GetBinContent(n);

              if(fallback){
                d = std::min(d, util::sqr((px-x)/X)+util::sqr((py-y)/Y));
              }
              else{
                d = std::min(d, PointDistanceToBox(px/X, py/Y,
                                                   (x-dx/2)/X, (y-dy/2)/Y,
                                                   (x+dx/2)/X, (y+dy/2)/Y));
              }
              if(d < bestd) break;
            }
            if(d < bestd) break;
          } // end while

          if(d > bestd){
            bestd = d;
            bestx = x;
            besty = y;
          }
        } // end for y
      } // end for x

      if(bestd != 0) break; // If we always collide, have to do fallback
    } // end for fallback

    // Convert to pad coordinates
    const double nx = (bestx-gPad->GetX1())/(gPad->GetX2()-gPad->GetX1());
    const double ny = (besty-gPad->GetY1())/(gPad->GetY2()-gPad->GetY1());

    const double ndx = dx/(gPad->GetX2()-gPad->GetX1());
    const double ndy = dy/(gPad->GetY2()-gPad->GetY1());

    return new TLegend(nx-ndx/2, ny-ndy/2, nx+ndx/2, ny+ndy/2);
  }

} // namespace
