#include "CAFAna/PredictionInterp.h"

#include "CAFAna/LoadFromFile.h"
#include "CAFAna/Ratio.h"
#include "CAFAna/SystematicShifter.h"

#include "TDirectory.h"
#include "TH2.h"
#include "CAFAna/Utilities.h"
#include "TMatrixD.h"
#include "TMD5.h"
#include "TObjString.h"
#include "TVectorD.h"

#include "OscLib/func/IOscCalculator.h"
#include "Utilities/func/MathUtil.h"

#include "CAFAna/Loaders.h"

namespace ana
{
  //----------------------------------------------------------------------
  PredictionInterp::PredictionInterp(std::vector<Syst> systs,
                                     osc::IOscCalculator* osc,
                                     PredConsFunc predCons,
                                     Loaders& loaders,
                                     const HistAxis& axis,
                                     const Cut& cut,
                                     const Var& wei)
    : fSyst(systs.back()), fOscOrigin(osc), fPrevNom("", Binning::Simple(1, 0, 1)), fPrevHash(0)
  {
    // Get a snapshot of osc's state now so we can check it doesn't change
    // under us.
    fOscOriginHash = fOscOrigin->GetParamsHash();

    assert(!systs.empty());

    // -3sigma to +3sigma in 1-sigma steps
    for(int i = 0; i <= 6; ++i){
      if(i == 3) continue; // Don't need nominal again
      fShifts.push_back(i-3);
    }

    // We deal with multiple systematics recursively. This object takes care
    // of the last one in the list, and the rest are handed off to be our
    // "nominal". This works because the correction is multiplicative, so we
    // can apply it for each systematic one at a time.

    loaders.SetShifter(0);

    std::vector<IPrediction*> preds;

    for(unsigned int i = 0; i < fShifts.size(); ++i){
      SystematicShifter* s = new SystematicShifter(fSyst, fShifts[i]);
      loaders.SetShifter(s);

      fPreds.push_back(predCons(loaders, axis, cut, wei));
    }

    // We've taken care of this one
    systs.pop_back();

    loaders.SetShifter(0);

    if(systs.empty()){
      fPredNom = predCons(loaders, axis, cut, wei);
      fBase = true;
    }
    else{
      fPredNom = new PredictionInterp(systs, fOscOrigin, predCons, loaders, axis, cut, wei);
      fBase = false;
    }

    // We need to know when all our predictions are filled so that we can
    // compute the fits
    loaders.RegisterCompletionCallback(this, &PredictionInterp::LoadedCallback);
  }

  //----------------------------------------------------------------------
  PredictionInterp::~PredictionInterp()
  {
    delete fPredNom;
    delete fPrevHash;
  }

  //----------------------------------------------------------------------
  void PredictionInterp::LoadedCallback()
  {
    // Make sure the oscillations we were given haven't changed since then, so
    // as not to surprise anyone.
    assert(*fOscOrigin->GetParamsHash() == *fOscOriginHash);
    const Spectrum nom = fPredNom->Predict(fOscOrigin);

    TH1* ratios[fPreds.size()];
    for(unsigned int i = 0; i < fPreds.size(); ++i){
      ratios[i] = Ratio(fPreds[i]->Predict(fOscOrigin), nom).ToTH1();
      delete fPreds[i];
    }

    const int binMax = ratios[0]->GetNbinsX();
    for(int binIdx = 0; binIdx < binMax+2; ++binIdx){
      // Analytic minimization for cubic fit, with value at origin fixed to 1
      TMatrixD m(3, 3);
      TVectorD v(3);

      for(unsigned int shiftIdx = 0; shiftIdx < fPreds.size(); ++shiftIdx){
        const double x = fShifts[shiftIdx];
        const double y = ratios[shiftIdx]->GetBinContent(binIdx);

        double xpow[7];
        xpow[0] = 1;
        for(int i = 1; i <= 6; ++i) xpow[i] = x*xpow[i-1];

        for(int i = 0; i <= 2; ++i){
          v(i) += xpow[i+1]*(y-1);
          for(int j = 0; j <= 2; ++j){
            m(i, j) += xpow[i+j+2];
          }
        }
      } // end for shiftIdx

      m = m.Invert();
      v = m*v;

      fFits.push_back(Coeffs(v(0), v(1), v(2)));
    } // end for binIdx

    for(TH1* r: ratios) delete r;
  }

  //----------------------------------------------------------------------
  Spectrum PredictionInterp::Predict(osc::IOscCalculator* calc) const
  {
    return fPredNom->Predict(calc);
  }

  //----------------------------------------------------------------------
  Spectrum PredictionInterp::PredictComponent(osc::IOscCalculator* calc,
                                              Flavors::Flavors_t flav,
                                              Current::Current_t curr,
                                              Sign::Sign_t sign) const
  {
    return fPredNom->PredictComponent(calc, flav, curr, sign);
  }

  //----------------------------------------------------------------------
  Spectrum PredictionInterp::PredictSyst(osc::IOscCalculator* calc,
                                         const SystShifts& syst) const
  {
    // TODO: shouldn't need to redo the nominal prediction when the osc pars
    // haven't changed. How to detect?
    // This is the biggest component in the slowness of systs

    return PredictComponentSyst(calc, syst,
                                Flavors::kAll,
                                Current::kBoth,
                                Sign::kBoth) + PredictCosmicComponent();
  }

  //----------------------------------------------------------------------
  Spectrum PredictionInterp::PredictComponentSyst(osc::IOscCalculator* calc,
                                                  const SystShifts& syst,
                                                  Flavors::Flavors_t flav,
                                                  Current::Current_t curr,
                                                  Sign::Sign_t sign) const
  {
    Spectrum ret("", Binning::Simple(1, 0, 1));

    // For base case, shouldn't pass systs through
    if(fBase){
      // If it's the common case, predicting the whole spectrum, then we might
      // have cached it.
      if(flav == Flavors::kAll &&
         curr == Current::kBoth &&
         sign == Sign::kBoth){

        // If the oscillation parameters haven't changed, can keep old nominal
        // spectrum and just have to apply new systematic scaling. Making the
        // nominal spectrum is the slowest part, and when making a surface it's
        // often true that we're just altering systematic parameters.
        const TMD5* hash = calc->GetParamsHash();
        if(!hash || !fPrevHash || *hash != *fPrevHash){
          fPrevNom = fPredNom->PredictComponent(calc, flav, curr, sign);
          delete fPrevHash;
          fPrevHash = hash;
        }
        else{
          delete hash;
        }
        ret = fPrevNom;
      }
      else{
        // Not the common case. Doesn't need to be as fast.
        ret = fPredNom->PredictComponent(calc, flav, curr, sign);
      }
    }
    else{
      // Not base. Pass systs through.
      ret = fPredNom->PredictComponentSyst(calc, syst, flav, curr, sign);
    }

    const double shift = syst.GetShift(fSyst);

    // TODO histogram operations could be too slow
    TH1* h = ret.ToTH1(ret.POT());
    for(int n = 0; n < h->GetNbinsX()+2; ++n){
      const Coeffs f = fFits[n];
      const double corr = 1 + f.a*shift + f.b*util::sqr(shift) + f.c*util::cube(shift);
      h->SetBinContent(n, std::max(h->GetBinContent(n)*corr, 0.));
    }

    Spectrum ret2(h, ret.POT());
    delete h;
    return ret2;
  }

  //----------------------------------------------------------------------
  Spectrum PredictionInterp::PredictCosmicComponent() const
  {
    // No interpolation on cosmic component, for now
    return fPredNom->PredictCosmicComponent();
  }

  //----------------------------------------------------------------------
  void PredictionInterp::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;

    dir->cd();
    TObjString("PredictionInterp").Write("type");

    // No need to save the other members, they're only used to derive the
    // fFits coefficients.

    TH1D hSyst("", "", 1, 0, 1);
    hSyst.SetBinContent(1, fSyst);
    hSyst.Write("syst");

    TH1D hBase("", "", 1, 0, 1);
    hBase.SetBinContent(1, fBase);
    hBase.Write("base");

    fPredNom->SaveTo(dir->mkdir("nom"));

    TH2D hFits("", "Coefficients;Bin;Variable",
               fFits.size(), 0, fFits.size(),
               3, 0, 3);
    hFits.GetYaxis()->SetBinLabel(1, "a");
    hFits.GetYaxis()->SetBinLabel(2, "b");
    hFits.GetYaxis()->SetBinLabel(3, "c");

    for(unsigned int i = 0 ; i < fFits.size(); ++i){
      hFits.SetBinContent(i+1, 1, fFits[i].a);
      hFits.SetBinContent(i+1, 2, fFits[i].b);
      hFits.SetBinContent(i+1, 3, fFits[i].c);
    }
    hFits.Write("coeffs");

    tmp->cd();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<PredictionInterp> PredictionInterp::LoadFrom(TDirectory* dir)
  {
    TObjString* tag = (TObjString*)dir->Get("type");
    assert(tag);
    assert(tag->GetString() == "PredictionInterp");

    std::unique_ptr<PredictionInterp> ret(new PredictionInterp);

    TH1* hSyst = (TH1*)dir->Get("syst");
    assert(hSyst);
    ret->fSyst = Syst(int(hSyst->GetBinContent(1)+.5));

    TH1* hBase = (TH1*)dir->Get("base");
    assert(hBase);
    ret->fBase = bool(hBase->GetBinContent(1));

    ret->fPredNom = ana::LoadFrom<IPrediction>(dir->GetDirectory("nom")).release();

    TH2* hFits = (TH2*)dir->Get("coeffs");
    const int nbins = hFits->GetNbinsX();
    for(int i = 0; i < nbins; ++i){
      const double a = hFits->GetBinContent(i+1, 1);
      const double b = hFits->GetBinContent(i+1, 2);
      const double c = hFits->GetBinContent(i+1, 3);

      ret->fFits.push_back(Coeffs(a, b, c));
    }

    return ret;
  }

} // namespace
