#pragma once

#include "CAFAna/Prediction.h"

#include "CAFAna/SpectrumLoader.h"
#include "CAFAna/SystShifts.h"

#include <memory>

class TMD5;

namespace ana
{
  class Loaders;

  /// Implements systematic errors by interpolation between shifted templates
  class PredictionInterp: public IPrediction
  {
  public:
    typedef std::function<IPrediction* (Loaders&, const HistAxis&, const Cut&, const Var&)> PredConsFunc;

    /// \param systs What systematics we should be capable of interpolating
    /// \param osc The oscillation point to expand around. Please don't
    ///            alter the parameters after passing it here.
    /// \param predCons Construct an IPrediction from the following
    ///                 information. See \ref SimplePredictionCreator for an
    ///                 example
    PredictionInterp(std::vector<Syst> systs,
                     osc::IOscCalculator* osc,
                     PredConsFunc predCons,
                     Loaders& loaders,
                     const HistAxis& axis,
                     const Cut& cut = kNoCut,
                     const Var& wei = kUnweighted);

    virtual ~PredictionInterp();



    virtual Spectrum Predict(osc::IOscCalculator* calc) const override;
    virtual Spectrum PredictSyst(osc::IOscCalculator* calc,
                                 const SystShifts& syst) const override;

    virtual Spectrum PredictComponent(osc::IOscCalculator* calc,
                                      Flavors::Flavors_t flav,
                                      Current::Current_t curr,
                                      Sign::Sign_t sign) const override;
    virtual Spectrum PredictComponentSyst(osc::IOscCalculator* calc,
                                          const SystShifts& syst,
                                          Flavors::Flavors_t flav,
                                          Current::Current_t curr,
                                          Sign::Sign_t sign) const override;

    virtual Spectrum PredictCosmicComponent() const override;

    virtual void SaveTo(TDirectory* dir) const;
    static std::unique_ptr<PredictionInterp> LoadFrom(TDirectory* dir);
  protected:
    PredictionInterp() : fPredNom(0), fPrevNom(0, 0), fPrevHash(0) {};

    void LoadedCallback();

    Syst fSyst; ///< What systematic we're interpolating

    std::vector<double> fShifts; ///< Shift values sampled

    bool fBase; ///< Is this the base case of the recursion?
    IPrediction* fPredNom; ///< The nominal prediction
    std::vector<IPrediction*> fPreds; ///< The predictions at shifted points

    /// The oscillation values we assume when evaluating the coefficients
    osc::IOscCalculator* fOscOrigin;
    const TMD5* fOscOriginHash;

    struct Coeffs{
      Coeffs(double _a, double _b, double _c) : a(_a), b(_b), c(_c) {}
      double a, b, c;
    };
    std::vector<Coeffs> fFits; ///< Interpolation functions for each bin

    mutable Spectrum fPrevNom;
    mutable const TMD5* fPrevHash;
  };

  /// \brief Can use this for the predCons argument if your Prediction requires
  /// no additional information
  template<class T> IPrediction*
  SimplePredictionCreator(Loaders& l, const HistAxis& a, const Cut& c, const Var& w)
  {
    return new T(l, a, c, w);
  }
}
