#include "CAFAna/ReweightableSpectrum.h"

#include "CAFAna/Binning.h"
#include "CAFAna/Utilities.h"

#include "TDirectory.h"
#include "TH2.h"
#include "TObjString.h"

#include <cassert>
#include <iostream>

namespace ana
{
  //----------------------------------------------------------------------
  ReweightableSpectrum::ReweightableSpectrum(std::string xlabel, std::string ylabel, double pot,
                                             int nbinsx, double xmin, double xmax,
                                             int nbinsy, double ymin, double ymax)
    : fPOT(pot)
  {
    DontAddDirectory guard;

    fHist = new TH2D("", (";"+xlabel+";"+ylabel).c_str(),
                     nbinsx, xmin, xmax, nbinsy, ymin, ymax);

    // Ensure errors get accumulated properly
    fHist->Sumw2();
  }

  //----------------------------------------------------------------------
  ReweightableSpectrum::ReweightableSpectrum(TH2* h, double pot, std::string label)
    : fPOT(pot)
  {
    if(!h){
      fHist = 0;
      return;
    }

    DontAddDirectory guard;

    const TString className = h->ClassName();

    if(className == "TH2D"){
      // Shortcut if types match
      fHist = new TH2D(*((TH2D*)h));
      fHist->GetZaxis()->SetTitle("Events");
    }
    else{
      const TAxis* ax = h->GetXaxis();
      const TAxis* ay = h->GetYaxis();
      fHist = new TH2D("", (";"+std::string(ax->GetTitle())+";"+std::string(ay->GetTitle())).c_str(),
                       ax->GetNbins(), ax->GetXbins()->GetArray(),
                       ay->GetNbins(), ay->GetXbins()->GetArray());
      fHist->Add(h);
    }

    if(!label.empty()) fHist->GetXaxis()->SetTitle(label.c_str());
  }

  //----------------------------------------------------------------------
  ReweightableSpectrum::~ReweightableSpectrum()
  {
    if(fHist && fHist->GetDirectory()){
      static bool once = true;
      if(once){
        once = false;
        std::cerr << "ReweightableSpectrum's fHist is associated with a directory. How did that happen?" << std::endl;
      }
    }

    delete fHist;
  }

  //----------------------------------------------------------------------
  ReweightableSpectrum::ReweightableSpectrum(const ReweightableSpectrum& rhs)
  {
    DontAddDirectory guard;

    fHist = new TH2D(*rhs.fHist);

    fPOT = rhs.fPOT;
  }

  //----------------------------------------------------------------------
  ReweightableSpectrum& ReweightableSpectrum::operator=(const ReweightableSpectrum& rhs)
  {
    if(this == &rhs) return *this;

    DontAddDirectory guard;

    delete fHist;
    fHist = new TH2D(*rhs.fHist);
    fPOT = rhs.fPOT;
    return *this;
  }

  //----------------------------------------------------------------------
  TH2D* ReweightableSpectrum::ToTH2(double pot) const
  {
    // Could have a file temporarily open
    DontAddDirectory guard;

    TH2D* ret = new TH2D(*fHist);

    ret->Scale(pot/fPOT);
    return ret;
  }

  //----------------------------------------------------------------------
  void ReweightableSpectrum::Fill(double x, double y, double w)
  {
    fHist->Fill(x, y, w);
  }

  //----------------------------------------------------------------------
  Spectrum ReweightableSpectrum::UnWeighted() const
  {
    return Spectrum(fHist->ProjectionX(UniqueName().c_str(), 0, -1), fPOT);
  }

  //----------------------------------------------------------------------
  Spectrum ReweightableSpectrum::WeightingVariable() const
  {
    return Spectrum(fHist->ProjectionY(UniqueName().c_str(), 0, -1), fPOT);
  }

  //----------------------------------------------------------------------
  Spectrum ReweightableSpectrum::WeightedBy(const TH1* ws) const
  {
    // This function is in the inner loop of oscillation fits, so some
    // optimization has been done.

    DontAddDirectory guard;

    assert(ws->GetNbinsX() == fHist->GetNbinsY());

    TH1D* hRet;
    TAxis* ax = fHist->GetXaxis();
    // ROOT: you suck
    if(ax->GetXbins()->GetArray())
      hRet = new TH1D("", (";"+std::string(ax->GetTitle())+";Events").c_str(),
                      ax->GetNbins(), ax->GetXbins()->GetArray());
    else
      hRet = new TH1D("", (";"+std::string(ax->GetTitle())+";Events").c_str(),
                      ax->GetNbins(), ax->GetXmin(), ax->GetXmax());

    const int X = fHist->GetNbinsX();
    const int Y = fHist->GetNbinsY();

    // Direct access to the bins is faster
    double* retArr = hRet->GetArray();
    double* histArr = fHist->GetArray();

    int bin = 0;
    for(int y = 0; y < Y+2; ++y){
      const double w = ws->GetBinContent(y);
      for(int x = 0; x < X+2; ++x){
        // Our loops go over the bins in the order they are internally in
        // fHist, and we do overflows, so we keep up exactly. If you get
        // paranoid, reenable this briefly.

        // assert(bin == fHist->GetBin(x, y));

        retArr[x] += histArr[bin]*w;
        ++bin;
      }
    }

    // TODO: can this all be more efficient?
    Spectrum ret(hRet, fPOT);
    delete hRet;
    return ret;
  }

  //----------------------------------------------------------------------
  void ReweightableSpectrum::Clear()
  {
    fHist->Reset();
  }

  //----------------------------------------------------------------------
  void ReweightableSpectrum::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();

    TObjString("ReweightableSpectrum").Write("type");

    fHist->Write("hist");
    TH1D hPot("", "", 1, 0, 1);
    hPot.Fill(.5, fPOT);
    hPot.Write("pot");
    tmp->cd();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<ReweightableSpectrum> ReweightableSpectrum::LoadFrom(TDirectory* dir)
  {
    TObjString* tag = (TObjString*)dir->Get("type");
    assert(tag);
    assert(tag->GetString() == "ReweightableSpectrum");

    TH2D* spect = (TH2D*)dir->Get("hist");
    assert(spect);
    TH1* hPot = (TH1*)dir->Get("pot");
    assert(hPot);
    return std::unique_ptr<ReweightableSpectrum>(new ReweightableSpectrum(spect, hPot->GetBinContent(1)));
  }
}
