#include "CAFAna/Spectrum.h"

#include "CAFAna/Ratio.h"
#include "CAFAna/Utilities.h"

#include "Utilities/func/MathUtil.h"

#include "TDirectory.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TObjString.h"
#include "TRandom3.h"

#include <cassert>
#include <iostream>

namespace ana
{
  //----------------------------------------------------------------------
  Spectrum::Spectrum(const std::string& label, const Binning& bins,
                     SpectrumLoaderBase& loader,
                     const Var& var,
                     const Cut& cut,
                     const Var& wei)
    : fHist(0), fPOT(0), fBinning(1, bins)
  {
    ConstructHistogram(label, bins);

    loader.AddSpectrum(*this, var, cut, wei);
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(SpectrumLoaderBase& loader,
                     const HistAxis& axis,
                     const Cut& cut,
                     const Var& wei)
    : Spectrum(axis.label, axis.bins, loader, axis.var, cut, wei)
  {
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(const std::string& label, const Binning& bins)
    : fHist(0), fPOT(0), fBinning(1, bins)
  {
    ConstructHistogram(label, bins);
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(const std::string& label, double pot, const Binning& bins)
    : fHist(0), fPOT(pot), fBinning(1, bins)
  {
    ConstructHistogram(label, bins);
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(TH1* h, double pot, std::string varName)
    : fHist(0), fPOT(pot)
  {
    if(!h){
      fHist = 0;
      return;
    }

    DontAddDirectory guard;

    const TString className = h->ClassName();

    if(className == "TH1D"){
      // Shortcut if types match
      fHist = new TH1D(*((TH1D*)h));
    }
    else{
      const TAxis* ax = h->GetXaxis();
      fHist = new TH1D("", (";"+std::string(ax->GetTitle())).c_str(),
                       ax->GetNbins(), ax->GetXbins()->GetArray());
      fHist->Add(h);
    }

    if(!varName.empty()) fHist->GetXaxis()->SetTitle(varName.c_str());
    fHist->GetYaxis()->SetTitle("Events");
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(const std::string& label, SpectrumLoaderBase& loader,
                     const Binning& binsx, const Var& varx,
                     const Binning& binsy, const Var& vary,
                     const Cut& cut,
                     const Var& wei)
    : fHist(0), fPOT(0)
  {
    fBinning.push_back(binsx);
    fBinning.push_back(binsy);

    Var multiDVar = Var2D(varx, binsx.NBins(), binsx.Min(), binsx.Max(),
                          vary, binsy.NBins(), binsy.Min(), binsy.Max());

    /// Multi-D binning
    const int nbins = binsx.NBins()*binsy.NBins();
    const Binning multiDBinning = Binning::Simple(nbins, 0.0, double(nbins));

    ConstructHistogram(label, multiDBinning);
    loader.AddSpectrum(*this, multiDVar, cut, wei);
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(const std::string& label, SpectrumLoaderBase& loader,
                     const Binning& binsx, const Var& varx,
                     const Binning& binsy, const Var& vary,
                     const Binning& binsz, const Var& varz,
                     const Cut& cut,
                     const Var& wei)
    : fHist(0), fPOT(0)
  {
    fBinning.push_back(binsx);
    fBinning.push_back(binsy);
    fBinning.push_back(binsz);

    Var multiDVar = Var3D(varx, binsx.NBins(), binsx.Min(), binsx.Max(),
                          vary, binsy.NBins(), binsy.Min(), binsy.Max(),
                          varz, binsz.NBins(), binsz.Min(), binsz.Max());

    /// Multi-D binning
    const int nbins = binsx.NBins()*binsy.NBins()*binsz.NBins();
    const Binning multiDBinning = Binning::Simple(nbins, 0.0, double(nbins));

    ConstructHistogram(label, multiDBinning);
    loader.AddSpectrum(*this, multiDVar, cut, wei);
  }

  //----------------------------------------------------------------------
  Spectrum::~Spectrum()
  {
    if(fHist && fHist->GetDirectory()){
      static bool once = true;
      if(once){
        once = false;
        std::cerr << "Spectrum's fHist is associated with a directory. How did that happen?" << std::endl;
      }
    }

    if(fHist) delete fHist;
  }

  //----------------------------------------------------------------------
  Spectrum::Spectrum(const Spectrum& rhs):
    fPOT    (rhs.fPOT),
    fBinning(rhs.fBinning)
  {
    DontAddDirectory guard;

    assert(rhs.fHist);
    fHist = new TH1D(*rhs.fHist);
  }

  //----------------------------------------------------------------------
  Spectrum& Spectrum::operator=(const Spectrum& rhs)
  {
    if(this == &rhs) return *this;

    DontAddDirectory guard;

    delete fHist;
    assert(rhs.fHist);
    fHist    = new TH1D(*rhs.fHist);
    fPOT     = rhs.fPOT;
    fBinning = rhs.fBinning;

    return *this;
  }

  //----------------------------------------------------------------------
  void Spectrum::ConstructHistogram(const std::string& label,
                                    const Binning& bins)
  {
    DontAddDirectory guard;

    assert(!fHist);

    if(bins.IsSimple()){
      fHist = new TH1D("", "", bins.NBins(), bins.Min(), bins.Max());
    }
    else{
      fHist = new TH1D("", "", bins.NBins(), &bins.Edges()[0]);
    }

    fHist->GetXaxis()->SetTitle(label.c_str());
    fHist->GetYaxis()->SetTitle("Events");

    // Ensure errors get accumulated properly
    fHist->Sumw2();
  }

  //----------------------------------------------------------------------
  TH1D* Spectrum::ToTH1(double pot) const
  {
    // Could have a file temporarily open
    DontAddDirectory guard;

    TH1D* ret = new TH1D(*fHist);
    if(fPOT){
      ret->Scale(pot/fPOT);
    }
    else{
      // How did it get events with no POT?
      assert(ret->Integral() == 0);
    }

    return ret;
  }

  //----------------------------------------------------------------------
  TH2* Spectrum::ToTH2(double pot) const
  {
    if(fBinning.size() != 2){
      std::cout<<"\nSorry, but dimension "<<fBinning.size()<<" is not two!!! Returning null.\n";
      return nullptr;
    }

    const Binning& binsx = fBinning[0];
    const Binning& binsy = fBinning[1];

    TH2* ret = ana::ToTH2(*this, pot,
                          binsx.NBins(), binsx.Min(), binsx.Max(),
                          binsy.NBins(), binsy.Min(), binsy.Max());

    ret->SetTitle(fHist->GetXaxis()->GetTitle());

    return ret;
  }

  //----------------------------------------------------------------------
  TH2* Spectrum::ToTH2NormX(double pot) const
  {
    TH2* xyhist = ToTH2(pot);
    if(!xyhist) return nullptr;

    const int nbinsx = fBinning[0].NBins();
    const int nbinsy = fBinning[1].NBins();

    // Normalize 2D histogram to X-axis spectrum
    for(int i=1; i<=nbinsx; ++i){
      double norm = 0.0;
      for(int j=1; j<=nbinsy; ++j){
        norm += xyhist->GetBinContent(i, j);
      }
      /// If no entries in the column, skip normalization
      if(norm < 0.0000001) continue;

      norm = 1.0 / norm;
      for(int j=1; j<=nbinsy; ++j){
        xyhist->SetBinContent(i,j, xyhist->GetBinContent(i, j) * norm);
      }
    }

    return xyhist;
  }

  //----------------------------------------------------------------------
  TH3* Spectrum::ToTH3(double pot) const
  {
    if(fBinning.size() != 3){
      std::cout<<"\nSorry, but dimension "<<fBinning.size()<<" is not 3!!! Returning null.\n";
      return nullptr;
    }

    const Binning& binsx = fBinning[0];
    const Binning& binsy = fBinning[1];
    const Binning& binsz = fBinning[2];

    TH3* ret = ana::ToTH3(*this, pot,
                          binsx.NBins(), binsx.Min(), binsx.Max(),
                          binsy.NBins(), binsy.Min(), binsy.Max(),
                          binsz.NBins(), binsz.Min(), binsz.Max());

    ret->SetTitle(fHist->GetXaxis()->GetTitle());

    return ret;
  }

  //----------------------------------------------------------------------
  double Spectrum::Integral(double pot, double* err) const
  {
    if(err){
      *err = 0;

      for(int i = 0; i < fHist->GetNbinsX()+2; ++i){
        *err += util::sqr(fHist->GetBinError(i));
      }
      *err = sqrt(*err) * pot/fPOT;
    }

    return fHist->Integral(0, -1) * pot/fPOT;
  }

  //----------------------------------------------------------------------
  void Spectrum::Fill(double x, double w)
  {
    fHist->Fill(x, w);
  }

  //----------------------------------------------------------------------
  Spectrum Spectrum::MockData(double pot) const
  {
    Spectrum ret = FakeData(pot);

    TRandom3 rnd(0); // zero seeds randomly

    for(int i = 0; i < ret.fHist->GetNbinsX()+2; ++i){
      ret.fHist->SetBinContent(i, rnd.Poisson(ret.fHist->GetBinContent(i)));
    }

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,34,12)
    // Drop old errors, which are based on the MC statistics, and create new
    // ones that are based on the prediction for the data
    ret.fHist->Sumw2(false);
#endif
    ret.fHist->Sumw2();

    return ret;
  }

  //----------------------------------------------------------------------
  Spectrum Spectrum::FakeData(double pot) const
  {
    Spectrum ret = *this;
    if(fPOT > 0.0)
        ret.fHist->Scale(pot/fPOT);
    ret.fPOT = pot;

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,34,12)
    // Drop old errors, which are based on the MC statistics, and create new
    // ones that are based on the prediction for the data
    ret.fHist->Sumw2(false);
#endif
    ret.fHist->Sumw2();

    return ret;
  }

  //----------------------------------------------------------------------
  void Spectrum::Clear()
  {
    fHist->Reset();
  }

  //----------------------------------------------------------------------
  Spectrum& Spectrum::operator+=(const Spectrum& rhs)
  {
    if(rhs.fPOT){
      fHist->Add(rhs.fHist, fPOT/rhs.fPOT);
    }
    else{
      // How can it have events but no POT?
      assert(rhs.fHist->Integral() == 0);
    }
    return *this;
  }

  //----------------------------------------------------------------------
  Spectrum Spectrum::operator+(const Spectrum& rhs) const
  {
    Spectrum ret = *this;
    ret += rhs;
    return ret;
  }

  //----------------------------------------------------------------------
  Spectrum& Spectrum::operator-=(const Spectrum& rhs)
  {
    if(rhs.fPOT){
      fHist->Add(rhs.fHist, -fPOT/rhs.fPOT);
    }
    else{
      // How can it have events but no POT?
      assert(rhs.fHist->Integral() == 0);
    }
    return *this;
  }

  //----------------------------------------------------------------------
  Spectrum Spectrum::operator-(const Spectrum& rhs) const
  {
    Spectrum ret = *this;
    ret -= rhs;
    return ret;
  }

  //----------------------------------------------------------------------
  Spectrum& Spectrum::operator*=(const Ratio& rhs)
  {
    fHist->Multiply(rhs.fHist);
    return *this;
  }

  //----------------------------------------------------------------------
  Spectrum Spectrum::operator*(const Ratio& rhs) const
  {
    Spectrum ret = *this;
    ret *= rhs;
    return ret;
  }

  //----------------------------------------------------------------------
  Spectrum& Spectrum::operator/=(const Ratio& rhs)
  {
    fHist->Divide(rhs.fHist);
    return *this;
  }

  //----------------------------------------------------------------------
  Spectrum Spectrum::operator/(const Ratio& rhs) const
  {
    Spectrum ret = *this;
    ret /= rhs;
    return ret;
  }

  //----------------------------------------------------------------------
  void Spectrum::SaveTo(TDirectory* dir) const
  {
    TDirectory* tmp = gDirectory;
    dir->cd();

    TObjString("Spectrum").Write("type");

    fHist->Write("hist");
    TH1D hPot("", "", 1, 0, 1);
    hPot.Fill(.5, fPOT);
    hPot.Write("pot");

    for(unsigned int i = 0; i < fBinning.size(); ++i){
      fBinning[i].SaveTo(dir->mkdir(TString::Format("bins%d", i)));
    }

    tmp->cd();
  }

  //----------------------------------------------------------------------
  std::unique_ptr<Spectrum> Spectrum::LoadFrom(TDirectory* dir)
  {
    TObjString* tag = (TObjString*)dir->Get("type");
    assert(tag);
    assert(tag->GetString() == "Spectrum");

    TH1D* spect = (TH1D*)dir->Get("hist");
    assert(spect);
    TH1* hPot = (TH1*)dir->Get("pot");
    assert(hPot);

    std::unique_ptr<Spectrum> ret(new Spectrum(spect, hPot->GetBinContent(1)));

    for(int i = 0; ; ++i){
      TDirectory* subdir = dir->GetDirectory(TString::Format("bins%d", i));
      if(!subdir) break;
      ret->fBinning.push_back(*Binning::LoadFrom(subdir));
    }

    return ret;
  }
}
