#include "CAFAna/SpectrumLoader.h"


#include "CAFAna/OscillatableSpectrum.h"
#include "CAFAna/Progress.h"
#include "CAFAna/Spectrum.h"
#include "CAFAna/SystematicShifter.h"
#include "CAFAna/Utilities.h"

#include "StandardRecord/StandardRecord.h"

#include <cassert>
#include <iostream>

#include <fenv.h>

#include "TFile.h"
#include "TH2.h"
#include "TTree.h"

namespace ana
{
  //----------------------------------------------------------------------
  SpectrumLoader::SpectrumLoader(const std::string& wildcard, DataSource src)
    : SpectrumLoaderBase(wildcard, src), fShifter(0)
  {
  }

  //----------------------------------------------------------------------
  SpectrumLoader::SpectrumLoader(const std::vector<std::string>& fnames,
                                 DataSource src)
    : SpectrumLoaderBase(fnames, src), fShifter(0)
  {
  }

  //----------------------------------------------------------------------
  SpectrumLoader::~SpectrumLoader()
  {
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::AddSpectrum(Spectrum& spect,
                                   const Var& var,
                                   const Cut& cut,
                                   const Var& wei)
  {
    SpectrumLoaderBase::AddSpectrum(spect, var, cut, wei);

    fHistDefs.back().shifter = fShifter;
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::AddOscillatableSpectrum(OscillatableSpectrum& spect,
                                               const Var& var,
                                               const Cut& cut,
                                               const Var& wei)
  {
    SpectrumLoaderBase::AddOscillatableSpectrum(spect, var, cut, wei);

    fHistDefs.back().shifter = fShifter;
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::Go()
  {
    // When embedded in Loaders, can be called even if no-one needs us
    if(fBranchNames.empty()){
      // But still need to notify Loaders itself
      for(auto c: fCallbacks) c();
      return;
    }

    feenableexcept(FE_INVALID);

    if(fGone){
      std::cerr << "Error: can only call Go() once on a SpectrumLoader" << std::endl;
      abort();
    }
    fGone = true;

    for(const HistDef& hd: fHistDefs){
      if(hd.shifter) RequireVariables(hd.shifter->Requires());
    }

    std::cout << "Active branches: ";
    for(const std::string& s: fBranchNames) std::cout << s << " ";
    std::cout << std::endl;

    const unsigned int Nfiles = fFileNames.size();

    double pot = 0;
    double livetime = 0;

    Progress prog(TString::Format("Filling %lu spectra from %d files matching '%s'", fHistDefs.size(), Nfiles, fWildcard.c_str()).Data());


    for(unsigned int fileIdx = 0; fileIdx < Nfiles; ++fileIdx){
      HandleFile(fFileNames[fileIdx], pot, livetime, Nfiles == 1 ? &prog : 0);

      if(Nfiles > 1) prog.SetProgress((fileIdx+1.)/Nfiles);
    } // end for fileIdx

    prog.Done();

    HandlePOT(pot, livetime);

    // Notify everyone this call to Go() has completed
    for(auto c: fCallbacks) c();

    fedisableexcept(FE_INVALID);
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::HandleFile(const std::string& fname,
                                  double& pot, double& livetime,
                                  Progress* prog)
  {
    TFile* f = TFile::Open(fname.c_str()); // This pattern allows xrootd
    assert(!f->IsZombie());

    TH1* hPOT = (TH1*)f->Get("TotalPOT");
    assert(hPOT);
    pot += hPOT->Integral();

    TH1* hLivetime = (TH1*)f->Get("TotalLivetime");
    assert(hLivetime);
    livetime += hLivetime->Integral();

    TTree* tr = (TTree*)f->Get("recTree");
    assert(tr);

    BranchList branches(tr, fBranchNames);

    const int Nentries = tr->GetEntries();
    for(int n = 0; n < Nentries; ++n){
      HandleRecord(branches.GetEntry(n));

      if(prog && n%10000 == 0) prog->SetProgress(double(n)/Nentries);
    } // end for n

    delete f;
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::HandleRecord(caf::StandardRecord* sr)
  {
    // TODO: what to really do about no true nu?
    const double trueE = (fAnyOsc && sr->mc.nnu == 1) ? sr->mc.nu[0].E : 0;

    SystematicShifter* lastShift = 0;

    double systWeight = 1;

    for(const HistDef& hd: fHistDefs){

      // When the systematic shifter changes, need to provide a clean slate for
      // the next one to work from. Unfortunately, copying the whole
      // StandardRecord is pretty expensive. So we need to rely on this
      // slightly dangerous "Unshift" mechanism.
      if(hd.shifter != lastShift){
        if(lastShift) lastShift->Unshift(sr);
        systWeight = 1;
        lastShift = hd.shifter;
        if(hd.shifter) hd.shifter->Shift(sr, systWeight);
      }

      // No cut, or cut passes
      if(hd.cut(sr)){
        const double wei = hd.wei(sr)*systWeight;
        if(wei){
          const double val = hd.var(sr);
          if(hd.spect) hd.spect->fHist->Fill(val, wei);
          // TODO: ignoring events with no true neutrino
          if(hd.oscSpect && trueE > 0){
            hd.oscSpect->fHist->Fill(val, trueE, wei);
          }
        }
      }
    } // end for hd
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::HandlePOT(double pot, double livetime)
  {
    if(fSource == kBeam){
      std::cout << "Total " << pot << " POT" << std::endl;
      assert(pot > 0);
    }
    else{
      std::cout << "Total livetime " << livetime << " seconds" << std::endl;
      assert(livetime > 0);
      // TODO: hardcoding these is gross, but it's the easiest way to get
      // reasonable cosmics normalization for now. Best long-term solution is
      // probably to get livetime saved in beam files and normalize everyone to
      // that.
      const double intensity = 5.0e13; // POT/spill
      const double window = 10e-6; // Beam timing cut in seconds
      pot = intensity*livetime/window;
      std::cout << "Scaled to an effective " << pot << " POT" << std::endl;
    }

    for(HistDef& hd: fHistDefs){
      // OK because we're a friend class
      if(hd.spect)
        hd.spect->fPOT += pot;
      else
        hd.oscSpect->fPOT += pot;
    }
  }

  //----------------------------------------------------------------------
  void SpectrumLoader::SetShifter(SystematicShifter* s)
  {
    fShifter = s;
  }
} // namespace
