#include "CAFAna/SpectrumLoaderBase.h"

#include "CAFAna/Progress.h"
#include "CAFAna/Utilities.h"

#include "StandardRecord/StandardRecord.h"

#include "TTree.h"

#include <algorithm>
#include <cassert>
#include <iostream>

#include "ifdh.h"

namespace ana
{
  //----------------------------------------------------------------------
  SpectrumLoaderBase::SpectrumLoaderBase(const std::string& wildcard,
                                         DataSource src)
    : fWildcard(wildcard), fFileNames(WildcardOrSAMQuery(wildcard)),
      fSource(src), fAnyOsc(false), fGone(false)
  {
  }

  //----------------------------------------------------------------------
  SpectrumLoaderBase::SpectrumLoaderBase(const std::vector<std::string>& fnames,
                                         DataSource src)
    : fWildcard("file list"), fSource(src), fAnyOsc(false), fGone(false)
  {
    fFileNames = fnames;
    assert(!fFileNames.empty());
    std::cout << "Loading from " << fFileNames.size() << " files" << std::endl;
  }

  //----------------------------------------------------------------------
  std::vector<std::string> SpectrumLoaderBase::
  WildcardOrSAMQuery(const std::string& str) const
  {
    // stat() blows up on strings with spaces
    if(str.find(' ') == std::string::npos){
      const std::vector<std::string> ret = Wildcard(str);
      if(!ret.empty()) return ret;
    }

    // Maybe this is a SAM dataset or query?
    TString query = str;
    if(!query.Contains(' ')){
      // This is a dataset name
      query = "defname: "+query;
    }

    ifdh i;
    i.set_debug("0"); // shut up
    std::vector<std::string> files;
    try{
      files = i.translateConstraints(query.Data());
    }
    catch(ifdh_util_ns::WebAPIException e){
      // I like my error message below better, since this could well be a
      // mistyped filename.
    }

    if(files.empty()){
      std::cerr << "\nCan't find any files matching '" << str
		<< "'. Aborting..." << std::endl;
      abort();
    }

    // IFDH likes to give back an empty string as the last response
    // https://cdcvs.fnal.gov/redmine/issues/6718
    if(!files.empty() && files.back().empty()){
      files.pop_back();
    }

    return LocateSAMFiles(files);
  }

  //----------------------------------------------------------------------
  std::vector<std::string> SpectrumLoaderBase::
  LocateSAMFiles(const std::vector<std::string>& fnames) const
  {
    // This is slow because it has to round-trip to the server for every file
    // https://cdcvs.fnal.gov/redmine/issues/6719

    std::vector<std::string> ret;

    ifdh i;
    i.set_debug("0"); // shut up

    bool needTickets = false;

    Progress prog(TString::Format("Looking up locations of %ld files in SAM",
				  fnames.size()).Data());
    for(unsigned int fIdx = 0; fIdx < fnames.size(); ++fIdx){
      const std::string f = fnames[fIdx];
      if(f.empty()) continue; // Last one is?

      const std::vector<std::string> locs = i.locateFile(f);

      int best = 0;
      
      std::string resolved;
      for(TString loc: locs){
	if(loc.BeginsWith("novadata:") && best < 2){
	  best = 2; // Prefer bluearc locations
	  loc.Remove(0, 9);
	  resolved = loc;
	}

	if(loc.BeginsWith("enstore:") && best < 1){
	  best = 1; // Otherwise, enstore via xrootd

	  loc.ReplaceAll("enstore:/pnfs/",
			 "root://fndca1.fnal.gov:1094//pnfs/fnal.gov/usr/");
	  // Strip the bit in brackets from the end
	  loc.Resize(loc.First('('));
	  resolved = loc;
	}
      } // end for loc

      if(best == 0 || resolved.empty()){
	std::cerr << "\nDidn't understand any of the locations of '" << f 
		  << "'. Aborting..." << std::endl;
	abort();
      }

      if(best == 1) needTickets = true;

      ret.push_back((resolved+'/'+f));

      prog.SetProgress((fIdx+1.)/fnames.size());
    } // end for fIdx
    prog.Done();

    if(needTickets){
      // Soon we'll need files via xrootd. Get every ticket known to man
      system("kx509");
      system("kxlist -p");
      system("voms-proxy-init --rfc --voms=fermilab:/fermilab/nova/Role=Analysis --noregen");
    }

    return ret;
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderBase::RequireVariable(std::string var)
  {
    fBranchNames.insert(var);
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderBase::RequireVariables(const std::set<std::string>& vars)
  {
    fBranchNames.insert(vars.begin(), vars.end());
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderBase::AddSpectrum(Spectrum& spect,
                                       const Var& var,
                                       const Cut& cut,
                                       const Var& wei)
  {
    if(fGone){
      std::cerr << "Error: can't add Spectra after the call to Go()" << std::endl;
      abort();
    }

    fHistDefs.push_back(HistDef(&spect, 0, var, cut, wei));

    RequireVariables(var.Requires());
    RequireVariables(cut.Requires());
    RequireVariables(wei.Requires());
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderBase::AddOscillatableSpectrum(OscillatableSpectrum& spect,
                                                   const Var& var,
                                                   const Cut& cut,
                                                   const Var& wei)
  {
    if(fGone){
      std::cerr << "Error: can't add Spectra after the call to Go()" << std::endl;
      abort();
    }

    fHistDefs.push_back(HistDef(0, &spect, var, cut, wei));
    RequireVariable("mc.nu");
    RequireVariable("mc.nnu");
    RequireVariable("mc.nu.E");
    fAnyOsc = true;

    RequireVariables(var.Requires());
    RequireVariables(cut.Requires());
    RequireVariables(wei.Requires());
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderBase::
  RegisterCompletionCallback(std::function<CallbackFunc_t> func)
  {
    fCallbacks.push_back(func);
  }

  //----------------------------------------------------------------------
  SpectrumLoaderBase::BranchList::
  BranchList(TTree* tr, const std::set<std::string>& names)
    : fTree(tr)
  {
    fSr = 0;
    tr->SetBranchAddress("rec", &fSr);

    // If we haven't been requested to do anything that actually signals to
    // load the entire StandardRecord object.
    if(names.empty()) return;

    tr->SetBranchStatus("*", 0);

    fBranches.reserve(names.size());
    for(std::string brName: names){
      tr->SetBranchStatus(brName.c_str(), 1);

      // If branch name ends with ".*" then it's right to set the branch status
      // with the star included, but GetBranch needs it removed. Because C++'s
      // string handling sucks this is pretty clumsy.
      if(brName.size() > 2 &&
         brName[brName.size()-2] == '.' &&
         brName[brName.size()-1] == '*')
        brName.resize(brName.size()-2);

      fBranches.push_back(tr->GetBranch(brName.c_str()));
      if(!fBranches.back()){
        std::cerr << "Tree has no branch \"" << brName
                  << "\". Aborting" << std::endl;
        abort();
      }
      tr->AddBranchToCache(fBranches.back());
    }
  }

  //----------------------------------------------------------------------
  caf::StandardRecord* SpectrumLoaderBase::BranchList::GetEntry(int n)
  {
    assert(n >= 0);
    const unsigned int NBr = fBranches.size();

    if(NBr == 0){
      // If we haven't been requested to do anything that actually signals to
      // load the entire StandardRecord object.
      fTree->GetEntry(n);
    }
    else{
      for(unsigned int i = 0; i < NBr; ++i){
        fBranches[i]->GetEntry(n);
      }
    }
    return fSr;
  }

} // namespace

