#pragma once

#include <functional>
#include <set>
#include <string>
#include <vector>

#include "CAFAna/Cut.h"
#include "CAFAna/Var.h"

namespace caf{class StandardRecord;}

class TBranch;
class TTree;

namespace ana
{
  class Spectrum;
  class OscillatableSpectrum;
  class SystematicShifter;

  /// Is this data-file representing beam spills or cosmic spills?
  enum DataSource{
    kBeam,
    kCosmic
  };

  /// Base class for the various types of spectrum loader
  class SpectrumLoaderBase
  {
  public:
    virtual ~SpectrumLoaderBase() {}

    /// For use by the \ref Spectrum constructor
    virtual void AddSpectrum(Spectrum& spect,
                             const Var& var,
                             const Cut& cut = kNoCut,
                             const Var& wei = kUnweighted);

    /// For use by the \ref OscillatableSpectrum constructor
    virtual void AddOscillatableSpectrum(OscillatableSpectrum& spect,
                                         const Var& var,
                                         const Cut& cut,
                                         const Var& wei);

    /// The type of the callback functions
    typedef void (CallbackFunc_t)();
    /// Register a function to be called when Go() is finished for this loader
    ///
    /// \param func A global function, lambda, or function object. The
    ///             signature is void function with no parameters.
    void RegisterCompletionCallback(std::function<CallbackFunc_t> func);

    /// Register a member function to be called when Go() is finished
    ///
    /// \param obj  Pointer to the object whose method to call
    /// \param meth The method to call, void, no arguments
    template<class T> void RegisterCompletionCallback(T* obj, void (T::*meth)())
    {
      // This has to be inline in the header so that the required versions get
      // instantiated when used.
      RegisterCompletionCallback(std::bind(meth, obj));
    }

    /// Load all the registered spectra
    virtual void Go() = 0;

  protected:
    SpectrumLoaderBase(const std::string& wildcard, DataSource src = kBeam);
    SpectrumLoaderBase(const std::vector<std::string>& fnames, DataSource src = kBeam);

    SpectrumLoaderBase(const SpectrumLoaderBase&);
    SpectrumLoaderBase& operator=(const SpectrumLoaderBase&);

    SpectrumLoaderBase() : fGone(false) {} ///< For use by \ref NullLoader

    /// Figure out if \a str is a wildcard or SAM query and return the results
    std::vector<std::string> WildcardOrSAMQuery(const std::string& str) const;

    /// Take filenames, return locations suitable for TFile::Open()
    std::vector<std::string> LocateSAMFiles(const std::vector<std::string>& fnames) const;

    /// Ensure this branch is activated
    void RequireVariable(std::string var);
    /// Ensure these branches are activated
    void RequireVariables(const std::set<std::string>& vars);

    std::string fWildcard;
    std::vector<std::string> fFileNames;
    std::set<std::string> fBranchNames;
    DataSource fSource;

    bool fAnyOsc; ///< Are any of \ref fHistDefs oscillatable spectra?

    bool fGone; ///< Has Go() been called? Can't add more histograms after that

    std::vector<std::function<CallbackFunc_t>> fCallbacks;

    /// Helper for SpectrumLoader, stores details of one thing to be loaded
    struct HistDef
    {
      HistDef(Spectrum* s,
              OscillatableSpectrum* os,
              const Var& _var,
              const Cut& _cut = kNoCut,
              const Var& _wei = kUnweighted)
        : spect(s), oscSpect(os), shifter(0), var(_var), cut(_cut), wei(_wei)
      {
      }

      Spectrum* spect;
      OscillatableSpectrum* oscSpect;
      SystematicShifter* shifter;
      const Var var;
      const Cut cut;
      const Var wei;
    };

    class BranchList
    {
    public:
      BranchList(TTree* tr, const std::set<std::string>& names);
      caf::StandardRecord* GetEntry(int n);
    protected:
      TTree* fTree;
      std::vector<TBranch*> fBranches;
      caf::StandardRecord* fSr;
    };

    std::vector<HistDef> fHistDefs;
  };

  /// \brief Dummy loader that doesn't load any files
  ///
  /// Useful when a loader is required for a component you want to ignore
  class NullLoader: public SpectrumLoaderBase
  {
  public:
    NullLoader() {}
    virtual void Go() override
    {
      // Polite to signal everyone in case they're relying on it
      for(auto c: fCallbacks) c();
    }
  };
  /// \brief Dummy loader that doesn't load any files
  ///
  /// Useful when a loader is required for a component you want to ignore
  static NullLoader kNullLoader;
}

