#include "CAFAna/SpectrumLoaderMockData.h"

#include "CAFAna/Spectrum.h"

#include "OscLib/func/IOscCalculator.h"

#include "StandardRecord/StandardRecord.h"

namespace ana
{
  /// Helper for \ref OscWeight
  class OscWeightFunc
  {
  public:
    OscWeightFunc(osc::IOscCalculator* calc) : fCalc(calc)
    {
    }

    double operator()(const caf::StandardRecord* sr) const
    {
      if(sr->mc.nnu == 0) return 0;

      const caf::SRNeutrino& nu = sr->mc.nu[0];

      if(!nu.iscc){
        // If we want to include NCs from fluxswap file this is all going to
        // have to get quite a bit more complicated.
        if(nu.pdg == nu.pdgorig) return 1; else return 0;
      }

      return fCalc->P(nu.pdgorig, nu.pdg, nu.E);
    }
  protected:
    osc::IOscCalculator* fCalc;
  };

  //----------------------------------------------------------------------
  Var* OscWeight(osc::IOscCalculator* calc)
  {
    return new Var({"mc.nu", "mc.nnu", "mc.nu.iscc",
                    "mc.nu.pdg", "mc.nu.pdgorig", "mc.nu.E"},
                   OscWeightFunc(calc));
  }

  //----------------------------------------------------------------------
  SpectrumLoaderMockData::SpectrumLoaderMockData(double pot,
                                                 const std::string& wildcard,
                                                 bool fluctuate)
    : SpectrumLoader(wildcard, kBeam),
      fDataPOT(pot), fFluctuate(fluctuate),
      fSwapLoader(0), fOscWeight(0)
  {
  }

  //----------------------------------------------------------------------
  SpectrumLoaderMockData::
  SpectrumLoaderMockData(double pot,
                         const std::vector<std::string>& fnames,
                         bool fluctuate)
    : SpectrumLoader(fnames, kBeam),
      fDataPOT(pot), fFluctuate(fluctuate),
      fSwapLoader(0), fOscWeight(0)
  {
  }

  //----------------------------------------------------------------------
  SpectrumLoaderMockData::
  SpectrumLoaderMockData(double pot,
                         const std::string& wildcardNonswap,
                         const std::string& wildcardSwap,
                         osc::IOscCalculator* calc,
                         bool fluctuate)
    : SpectrumLoader(wildcardNonswap, kBeam),
      fDataPOT(pot), fFluctuate(fluctuate)
  {
    fSwapLoader = new SpectrumLoader(wildcardSwap, kBeam);
    fOscWeight = OscWeight(calc);
  }

  //----------------------------------------------------------------------
  SpectrumLoaderMockData::
  SpectrumLoaderMockData(double pot,
                         const std::vector<std::string>& fnamesNonswap,
                         const std::vector<std::string>& fnamesSwap,
                         osc::IOscCalculator* calc,
                         bool fluctuate)
    : SpectrumLoader(fnamesNonswap, kBeam),
      fDataPOT(pot), fFluctuate(fluctuate)
  {
    fSwapLoader = new SpectrumLoader(fnamesSwap);
    fOscWeight = OscWeight(calc);
  }

  //----------------------------------------------------------------------
  SpectrumLoaderMockData::~SpectrumLoaderMockData()
  {
    delete fSwapLoader;
    for(Spectrum* s: fSwapSpectra) delete s;
    delete fOscWeight;
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderMockData::AddSpectrum(Spectrum& spect,
                                           const Var& var,
                                           const Cut& cut,
                                           const Var& wei)
  {
    const Var weiOsc = fOscWeight ? *fOscWeight * wei : wei;

    SpectrumLoader::AddSpectrum(spect, var, cut, weiOsc);

    if(fSwapLoader){
      assert(spect.POT() == 0);
      // Have to be on the heap so pointers to them are stable until Go() is
      // called.
      Spectrum* swapspect = new Spectrum(spect);
      fSwapLoader->AddSpectrum(*swapspect, var, cut, weiOsc);
      fSwapSpectra.push_back(swapspect);
    }
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderMockData::SetShifter(SystematicShifter* s)
  {
    SpectrumLoader::SetShifter(s);
    if(fSwapLoader) fSwapLoader->SetShifter(s);
  }

  //----------------------------------------------------------------------
  void SpectrumLoaderMockData::Go()
  {
    SpectrumLoader::Go();
    if(fSwapLoader) fSwapLoader->Go();

    if(fSwapLoader){
      // Get the swapped components and add them in. They should align
      // one-to-one with the non-swapped.
      assert(fSwapSpectra.size() == fHistDefs.size());
      for(unsigned int i = 0; i < fHistDefs.size(); ++i){
        *fHistDefs[i].spect += *fSwapSpectra[i];
      }
    }

    for(const HistDef& hd: fHistDefs){
      assert(hd.spect);
      if(fFluctuate)
        *hd.spect = hd.spect->MockData(fDataPOT);
      else
        *hd.spect = hd.spect->FakeData(fDataPOT);
    }
  }
} // namespace
