#include "CAFAna/Surface.h"

#include "CAFAna/Experiment.h"
#include "CAFAna/Fit.h"
#include "CAFAna/FitVars.h"
#include "CAFAna/Progress.h"
#include "CAFAna/ThreadPool.h"
#include "CAFAna/Utilities.h"

#include "OscLib/func/IOscCalculator.h"

#include "TCanvas.h"
#include "TGraph.h"
#include "TH2.h"
#include "TMarker.h"
#include "Minuit2/StackAllocator.h"
#include "TObjArray.h"
#include "TPad.h"
#include "TROOT.h"
#include "TStyle.h"

#include <iostream>
#include <functional>

namespace ana
{
  //----------------------------------------------------------------------
  Surface::Surface(const Experiment* expt,
                   osc::IOscCalculatorAdjustable* calc,
                   const FitVar* xvar, int nbinsx, double xmin, double xmax,
                   const FitVar* yvar, int nbinsy, double ymin, double ymax,
                   const std::vector<const FitVar*>& margVars,
                   const std::vector<Syst>& margSysts,
                   bool parallel)
    : fParallel(parallel)
  {
    fHist = ExpandedHistogram(";"+xvar->LatexName()+";"+yvar->LatexName(),
                              nbinsx, xmin, xmax,
                              nbinsy, ymin, ymax);

    for(unsigned int i = 0; i < margVars.size()+margSysts.size(); ++i){
      std::string title;
      if(i < margVars.size())
        title = margVars[i]->LatexName();
      else
        title = LatexName(margSysts[i-margVars.size()]);

      fMargHists.push_back(ExpandedHistogram(title+";"+xvar->LatexName()+";"+yvar->LatexName(),
                                             nbinsx, xmin, xmax,
                                             nbinsy, ymin, ymax));
    }

    std::string progTitle = "Filling surface for "+yvar->ShortName()+" vs "+xvar->ShortName();

    if(!margVars.empty() || !margSysts.empty()){
      progTitle += " (marginalizing ";

      for(const FitVar* v: margVars)
        progTitle += v->ShortName() + ", ";
      for(Syst s: margSysts)
        progTitle += ShortName(s) + ", ";

      // Always have one superfluous ", " at the end
      progTitle.resize(progTitle.size()-2);
      progTitle += ")";
    }

    FillSurface(progTitle, expt, calc, xvar, yvar, margVars, margSysts);

    // Location of the best minimum found from filled surface
    double minchi = 1e10;
    int minx = nbinsx/2;
    int miny = nbinsy/2;
    for(int x = 1; x < nbinsx+1; ++x){
      for(int y = 1; y < nbinsy+1; ++y){
        const double chi = fHist->GetBinContent(x, y);
        if(chi < minchi){
          minchi = chi;
          minx = x;
          miny = y;
        }
      }
    }

    std::vector<const FitVar*> allVars = {xvar, yvar};
    allVars.insert(allVars.end(), margVars.begin(), margVars.end());
    Fitter fit(expt, allVars, margSysts);
    // Seed from best grid point
    xvar->SetValue(calc, fHist->GetXaxis()->GetBinCenter(minx));
    yvar->SetValue(calc, fHist->GetYaxis()->GetBinCenter(miny));
    fMinChi = fit.Fit(calc);
    fMinX = xvar->GetValue(calc);
    fMinY = yvar->GetValue(calc);

    for(int x = 0; x < nbinsx+2; ++x){
      for(int y = 0; y < nbinsy+2; ++y){
        fHist->SetBinContent(x, y, fHist->GetBinContent(x, y)-fMinChi);
      }
    }

    fHist->SetMinimum(0);
  }

  //---------------------------------------------------------------------
  void Surface::FillSurface(const std::string& progTitle,
                            const Experiment* expt,
                            osc::IOscCalculatorAdjustable* calc,
                            const FitVar* xvar, const FitVar* yvar,
                            const std::vector<const FitVar*>& margVars,
                            const std::vector<Syst>& margSysts)
  {
    if(fParallel && !(margVars.empty() && margSysts.empty())){
      // Minuit calls this at some point, and it creates a static. Which
      // doesn't like happening on multiple threads at once. So do it upfront
      // while we're still single-threaded.
      ROOT::Minuit2::StackAllocatorHolder::Get();
    }

    // Nothing created during surface filling belongs in a
    // directory. Unfortunately the local guards in Spectrum etc are racey when
    // run in parallel. But this should cover the whole lot safely.
    DontAddDirectory guard;

    Progress* prog = 0;;
    // Difficult to keep a progress bar properly up to date when threaded
    if(!fParallel) prog = new Progress(progTitle);
    ThreadPool* pool = 0;
    if(fParallel){
      pool = new ThreadPool;
      pool->ShowProgress(progTitle);
    }

    const int Nx = fHist->GetNbinsX();
    const int Ny = fHist->GetNbinsY();
    for(int x = 1; x <= Nx; ++x){
      const double xv = fHist->GetXaxis()->GetBinCenter(x);
      // For parallel running need to set these at point of use otherwise we
      // race.
      if(!fParallel) xvar->SetValue(calc, xv);
      if(xvar->Penalty(xv) > 1e-10){
	std::cerr << "Warning! " << xvar->ShortName() << " = " << xv
		  << " has penalty of " << xvar->Penalty(xv)
		  << " that could have been applied in surface. "
		  << "This should never happen." << std::endl;
      }
      for(int y = 1; y <= Ny; ++y){
        const double yv = fHist->GetYaxis()->GetBinCenter(y);
        if(!fParallel) yvar->SetValue(calc, yv);
	if(yvar->Penalty(yv) > 1e-10){
	  std::cerr << "Warning! " << yvar->ShortName() << " = " << yv
		    << " has penalty of " << yvar->Penalty(yv)
		    << " that could have been applied in surface. "
		    << "This should never happen." << std::endl;
	}

        if(fParallel){
          pool->AddTask(this, &Surface::FillSurfacePoint,
                        expt, calc,
                        xvar, xv, yvar, yv,
                        margVars, margSysts);
        }
        else{
          FillSurfacePoint(expt, calc,
                           xvar, xv, yvar, yv,
                           margVars, margSysts);
          prog->SetProgress((x+y/double(Ny))/Nx);
        }
      } // end for y
    } // end for x

    if(fParallel){
      pool->Finish();
      delete pool;
    }
    else{
      prog->Done();
      delete prog;
    }
  }

  //----------------------------------------------------------------------
  void Surface::FillSurfacePoint(const Experiment* expt,
                                 osc::IOscCalculatorAdjustable* calc,
                                 const FitVar* xvar, double x,
                                 const FitVar* yvar, double y,
                                 const std::vector<const FitVar*>& margVars,
                                 const std::vector<Syst>& margSysts)
  {
    if(fParallel){
      // Need to take our own copy so that we don't get overwritten by someone
      // else's changes.
      calc = calc->Copy();
      xvar->SetValue(calc, x);
      yvar->SetValue(calc, y);
    }

    double chi;
    if(margVars.empty() && margSysts.empty()){
      chi = expt->ChiSq(calc);
    }
    else{
      Fitter fitter(expt, margVars, margSysts);
      SystShifts systSeed = SystShifts::Nominal();
      // TODO: this is going to lead to the marginalized values of the
      // oscillations persisting between steps, but the systematics not.
      // I don't know if that's OK, or if we want to do both, or neither.
      chi = fitter.Fit(calc, systSeed, Fitter::kQuiet);

      for(unsigned int i = 0; i < margVars.size(); ++i){
        fMargHists[i]->Fill(x, y, margVars[i]->GetValue(calc));
      }
      for(unsigned int j = 0; j < margSysts.size(); ++j){
        fMargHists[j+margVars.size()]->Fill(x, y, systSeed.GetShift(margSysts[j]));
      }
    }

    fHist->Fill(x, y, chi);

    if(fParallel) delete calc;
  }

  //---------------------------------------------------------------------
  void Surface::EnsureAxes() const
  {
    // If this pad has already been drawn in, already has axes
    if(gPad && !gPad->GetListOfPrimitives()->IsEmpty()) return;

    // Old, hackier solution
    /*
    std::cout << gPad->GetListOfPrimitives()->GetEntries() << std::endl;
    // Which pads have we already drawn axes in? Never draw axes in them
    // again. Unfortunately UniqueID() never seems to be set. If that's the
    // case, set it to a random value and hope...
    static std::set<UInt_t> already;
    if(already.count(gPad->GetUniqueID())) return;
    if(gPad->GetUniqueID() == 0) gPad->SetUniqueID(rand());
    already.insert(gPad->GetUniqueID());
    */

    const TAxis* ax = fHist->GetXaxis();
    const TAxis* ay = fHist->GetYaxis();
    const double Nx = ax->GetNbins();
    const double Ny = ay->GetNbins();

    // Axes with limits where the user originally requested them, which we
    // adjusted to be the centres of the first and last bins.
    TH2* axes = new TH2C(UniqueName().c_str(),
                         TString::Format(";%s;%s",
                                         ax->GetTitle(), ay->GetTitle()),
                         Nx-1, ax->GetBinCenter(1), ax->GetBinCenter(Nx),
                         Ny-1, ay->GetBinCenter(1), ay->GetBinCenter(Ny));
    axes->Draw();

    axes->SetTitle(fHist->GetTitle());
    axes->GetXaxis()->SetLabelSize(ax->GetLabelSize());
    axes->GetXaxis()->CenterTitle();
    axes->GetYaxis()->CenterTitle();
    gPad->Update();
  }

  //---------------------------------------------------------------------
  void Surface::Draw() const
  {
    EnsureAxes();

    fHist->Draw("colz same");

    // colz obliterated them
    gPad->RedrawAxis();

    gPad->Update();
  }

  //---------------------------------------------------------------------
  void Surface::DrawBestFit(Color_t color) const
  {
    EnsureAxes();

    TMarker* mark = new TMarker(fMinX, fMinY, kFullCircle);
    mark->SetMarkerSize(1.5);
    mark->SetMarkerColor(color);
    mark->Draw();
    gPad->Update();
  }

  //----------------------------------------------------------------------
  void Surface::DrawContour(TH2* fc, Style_t style, Color_t color,
                            double minchi)
  {
    EnsureAxes();

    if(minchi < 0) minchi = fMinChi;
    std::unique_ptr<TH2> surf((TH2*)fHist->Clone(UniqueName().c_str()));
    surf->Add(fc, -1);

    TVirtualPad* bak = gPad;

    const bool wasbatch = gROOT->IsBatch();
    gROOT->SetBatch(); // User doesn't want to see our temporary canvas
    TCanvas tmp;

    gStyle->SetOptStat(0);

    const double level = minchi-fMinChi;
    surf->SetContour(1, &level);
    surf->Draw("cont list");

    tmp.Update();
    tmp.Paint();

    gROOT->SetBatch(wasbatch);
    gPad = bak;

    // The graphs we need (contained inside TLists, contained inside
    // TObjArrays) are in the list of specials. But we need to be careful about
    // types, because other stuff can get in here too (TDatabasePDG for
    // example).
    TCollection* specs = gROOT->GetListOfSpecials();

    TIter nextSpec(specs);
    while(TObject* spec = nextSpec()){
      if(!spec->InheritsFrom(TObjArray::Class())) continue;
      TObjArray* conts = (TObjArray*)spec;

      if(conts->IsEmpty()) continue;

      if(!conts->At(0)->InheritsFrom(TList::Class())) continue;
      TList* cont = (TList*)conts->At(0);

      TIter nextObj(cont);
      // Contour could be split into multiple pieces
      while(TObject* obj = nextObj()){
        if(!obj->InheritsFrom(TGraph::Class())) continue;
        TGraph* g = (TGraph*)obj->Clone(UniqueName().c_str());

        g->SetLineWidth(2);
        g->SetLineStyle(style);
        g->SetLineColor(color);
        g->Draw("l");
      } // end for obj
    } // end for spec
    gPad->Update();
  }

  //----------------------------------------------------------------------
  TH2* Surface::ToTH2(double minchi) const
  {
    // Could have a file temporarily open
    DontAddDirectory guard;

    TH2* ret = new TH2F(*fHist);

    if(minchi >= 0){
      for(int x = 0; x < ret->GetNbinsX()+2; ++x){
        for(int y = 0; y < ret->GetNbinsY()+2; ++y){
          ret->SetBinContent(x, y, ret->GetBinContent(x, y)+fMinChi-minchi);
        }
      }
    }

    return ret;
  }

  //----------------------------------------------------------------------
  void Surface::SetTitle(const char* str)
  {
    fHist->SetTitle(str);
  }

  //----------------------------------------------------------------------
  /// Helper function for the gaussian approximation surfaces
  TH2* Flat(double level, const Surface& s)
  {
    TH2* h = s.ToTH2();

    for(int x = 0; x < h->GetNbinsX()+2; ++x)
      for(int y = 0; y < h->GetNbinsY()+2; ++y)
        h->SetBinContent(x, y, level);

    return h;
  }

  // See eg the statistics section of the PDG
  TH2* Gaussian68Percent2D(const Surface& s){return Flat(2.30, s);}
  TH2* Gaussian90Percent2D(const Surface& s){return Flat(4.61, s);}
  TH2* Gaussian95Percent2D(const Surface& s){return Flat(5.99, s);}
  TH2* Gaussian2Sigma2D   (const Surface& s){return Flat(6.18, s);}
  TH2* Gaussian99Percent2D(const Surface& s){return Flat(9.21, s);}
  TH2* Gaussian3Sigma2D   (const Surface& s){return Flat(11.83, s);}

  TH2* Gaussian68Percent1D(const Surface& s){return Flat(1.00, s);}
  TH2* Gaussian90Percent1D(const Surface& s){return Flat(2.71, s);}
  TH2* Gaussian95Percent1D(const Surface& s){return Flat(3.84, s);}
  TH2* Gaussian2Sigma1D   (const Surface& s){return Flat(4.00, s);}
  TH2* Gaussian99Percent1D(const Surface& s){return Flat(6.63, s);}
  TH2* Gaussian3Sigma1D   (const Surface& s){return Flat(9.00, s);}

} // namespace
