#include "CAFAna/SystematicShifter.h"

// Suppress warnings originating in GENIE that we can't do anything about
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Woverloaded-virtual"
#include "CAFReweight/CAFReweight.h"
#pragma GCC diagnostic pop

#include "StandardRecord/StandardRecord.h"

#include <cassert>

namespace ana
{
  //----------------------------------------------------------------------
  SystematicShifter::SystematicShifter(Syst syst, double shift)
    : fHaveGenie(false), fReweight(0)
  {
    SetShift(syst, shift);
  }

  //----------------------------------------------------------------------
  SystematicShifter::SystematicShifter(const std::map<Syst, double>& shifts)
    : fHaveGenie(false), fReweight(0)
  {
    for(auto it: shifts) SetShift(it.first, it.second);
  }

  //----------------------------------------------------------------------
  SystematicShifter::~SystematicShifter()
  {
    delete fReweight;
  }

  //----------------------------------------------------------------------
  void SystematicShifter::SetShift(Syst syst, double shift)
  {
    fShifts.SetShift(syst, shift);

    if(shift == 0){
      if(syst > kGENIESystInit && syst < kGENIESystEnd) fHaveGenie = false;
      return;
    }

    if(syst == kCaloESyst) fReqs.insert("slc.calE");
    if(syst == kNCScaleSyst){
      fReqs.insert("mc.nu");
      fReqs.insert("mc.nnu");
      fReqs.insert("mc.nu.iscc");
    }
    if(syst == kCCESyst){
      fReqs.insert("energy.numusimp.trknonqeE");
    }
    if(syst == kQEESyst){
      fReqs.insert("energy.numusimp.trkqeE");
    }
    if(syst == kUCESyst){
      fReqs.insert("energy.numusimp.ucccE");
    }

    assert(syst != kGENIESystInit && syst != kGENIESystEnd);
    if(syst > kGENIESystInit && syst < kGENIESystEnd){
      fHaveGenie = true;
      // Alter depending which method you plan to use (see Shift() function)
      fReqs.insert("mc.gen.*");
      //      fReqs.insert("mc.rwgt.*");
    }
  }

  //----------------------------------------------------------------------
  void SystematicShifter::Shift(caf::StandardRecord* sr,
                                double& weight) const
  {
    if(fShifts.GetShift(kNormSyst)) NormalizationShift(sr, weight);
    if(fShifts.GetShift(kCaloESyst)) CaloScaleShift(sr, weight);
    if(fShifts.GetShift(kNCScaleSyst)) NCBackgroundShift(sr, weight);
    if(fShifts.GetShift(kCCESyst)) CCEShift(sr, weight);
    if(fShifts.GetShift(kQEESyst)) QEEShift(sr, weight);
    if(fShifts.GetShift(kUCESyst)) UCEShift(sr, weight);

    if(fHaveGenie){
      // Alter to call one or the other depending on your CAFs.
      GenieShiftsFromGenBranch(sr, weight);
      //      GenieShiftsFromTable(sr, weight);
    }
  }

  //----------------------------------------------------------------------
  void SystematicShifter::Unshift(caf::StandardRecord* sr) const
  {
    if(fShifts.GetShift(kCaloESyst)) sr->slc.calE = fBakCalE;
    if(fShifts.GetShift(kCCESyst)) sr->energy.numusimp.trknonqeE = fBakCCE;
    if(fShifts.GetShift(kQEESyst)) sr->energy.numusimp.trkqeE = fBakQEE;
    if(fShifts.GetShift(kUCESyst)) sr->energy.numusimp.ucccE = fBakUCE;
  }

  //----------------------------------------------------------------------
  void SystematicShifter::NormalizationShift(caf::StandardRecord* /*sr*/,
                                             double& weight) const
  {
    weight *= 1+fShifts.GetShift(kNormSyst)*GetSigma(kNormSyst);
  }

  //----------------------------------------------------------------------
  void SystematicShifter::CaloScaleShift(caf::StandardRecord* sr,
                                         double& /*weight*/) const
  {
    fBakCalE = sr->slc.calE;
    sr->slc.calE *= 1+fShifts.GetShift(kCaloESyst)*GetSigma(kCaloESyst);
  }

  //----------------------------------------------------------------------
  void SystematicShifter::CCEShift(caf::StandardRecord* sr,
                                   double& /*weight*/) const
  {
    fBakCCE = sr->energy.numusimp.trknonqeE;
    sr->energy.numusimp.trknonqeE *= 1+fShifts.GetShift(kCCESyst)*GetSigma(kCCESyst);
  }

  //----------------------------------------------------------------------

  void SystematicShifter::QEEShift(caf::StandardRecord* sr,
                                   double& /*weight*/) const
  {
    fBakQEE = sr->energy.numusimp.trkqeE;
    sr->energy.numusimp.trkqeE *= 1+fShifts.GetShift(kQEESyst)*GetSigma(kQEESyst);
  }

  //----------------------------------------------------------------------

  void SystematicShifter::UCEShift(caf::StandardRecord* sr,
                                   double& /*weight*/) const
  {
    fBakUCE = sr->energy.numusimp.ucccE;
    sr->energy.numusimp.ucccE *= 1+fShifts.GetShift(kUCESyst)*GetSigma(kUCESyst);
  }

  //----------------------------------------------------------------------
  void SystematicShifter::NCBackgroundShift(caf::StandardRecord* sr,
                                            double& weight) const
  {
    if(sr->mc.nnu == 0) return;
    if(!sr->mc.nu[0].iscc) weight *= 1+fShifts.GetShift(kNCScaleSyst)*GetSigma(kNCScaleSyst);
  }

  //----------------------------------------------------------------------
  void SystematicShifter::GenieShiftsFromGenBranch(caf::StandardRecord* sr,
                                                   double& weight) const
  {
    // Avoid ~0.1% bad events. These seem to be (all?) nue/electron scattering,
    // which is not going to be reweightable anyway.
    if(sr->mc.gen.empty() || sr->mc.gen[0].fgQ2 < 0) return;

    if(!fReweight){
      fReweight = new cafrwgt::CAFReweight;
      bool any = false;
      for(int s = kGENIESystInit+1; s < kGENIESystEnd; ++s){
        const Syst syst = Syst(s);
        const double shift = fShifts.GetShift(syst)*GetSigma(syst);
        if(shift){
          any = true;
          fReweight->AddReweightValue(rwgt::EReweightLabel(s-kGENIESystInit),
                                      shift);
        }
      } // end for s
      assert(any); // We were told there were GENIE systs...
      fReweight->Configure();
    }

    weight *= fReweight->CalcWeight(sr->mc.gen);
  }

  //----------------------------------------------------------------------
  void SystematicShifter::GenieShiftsFromTable(caf::StandardRecord* sr,
                                               double& weight) const
  {
    if(sr->mc.rwgt.empty()) return;

    assert(sr->mc.rwgt.size() == kGENIESystEnd-kGENIESystInit);

    for(int s = kGENIESystInit+1; s < kGENIESystEnd; ++s){
      const Syst syst = Syst(s);
      const double shift = fShifts.GetShift(syst)*GetSigma(syst);
      const caf::SRGenieWeights& wgts = sr->mc.rwgt[s-kGENIESystInit];
      if(shift){
        // These tests for NaN don't actually work. We'd have to disable the
        // FE_INVALID exceptions we currently have turned on.
        if(isnan(wgts.minus2sigma) || isnan(wgts.minus1sigma) ||
           isnan(wgts.plus1sigma)  || isnan(wgts.plus2sigma)){
          std::cerr << "Warning: NaN GENIE weight found, ignoring" << std::endl;
          continue;
        }

        if(isinf(wgts.minus2sigma) || isinf(wgts.minus1sigma) || isinf(wgts.plus1sigma) || isinf(wgts.plus2sigma)){
          std::cerr << "Warning: Inf GENIE weight found, ignoring" << std::endl;
          continue;
        }

        /***/if(shift == -2) weight *= wgts.minus2sigma;
        else if(shift == -1) weight *= wgts.minus1sigma;
        else if(shift == +1) weight *= wgts.plus1sigma;
        else if(shift == +2) weight *= wgts.plus2sigma;
        else{
          // In principle we could implement some form of interpolation. Let's
          // leve that until (if) it's ever needed.
          std::cerr << "Unable to shift GENIE by " << shift << " sigma" << std::endl;
          abort();
        }
      }
    }
  }

  //----------------------------------------------------------------------
  double SystematicShifter::GetSigma(Syst syst) const
  {
    // TODO: should probably be configurable one day

    if(syst == kNormSyst) return 0.05;
    if(syst == kCaloESyst) return 0.02;
    if(syst == kNCScaleSyst) return 0.20;
    if(syst == kCCESyst) return 0.02;
    if(syst == kQEESyst) return 0.02;
    if(syst == kUCESyst) return 0.02;

    assert(syst != kGENIESystInit && syst != kGENIESystEnd);
    if(syst > kGENIESystInit && syst < kGENIESystEnd) return 1; // already in sigmas

    assert(0 && "Unknown syst");
  }
}
