#include "CAFAna/T2KToyExperiment.h"

#include "OscLib/func/OscCalculatorPMNS.h"
#include "Utilities/func/MathUtil.h"

#include "TFile.h"
#include "TH1.h"

#include <cassert>

namespace ana
{
  /// Baseline from JPARC to SK in km
  const int kT2KBaseline = 295;

  //----------------------------------------------------------------------
  T2KToyExperiment::T2KToyExperiment()
    : fPOT(0), fPOTAnti(0)
  {
    // The original version of this code was written by Ryan. Ask him about the
    // physics assumptions in it.

    // Scale everything from Neutrino 2013 results
    fBasePOT = 2.556e20;
    fBaseSig = 6.60;
    fBaseBkg  = 2.47;

    fTotSyst = 0.08; // 0.103 at Nu2012

    // using 3.8x reduction (15% from flux: 11/11/2012 IPMU talk, Nakaya;
    // from an old 2-deg off-axis simulation.  xsec reduction from those
    // usual xsec plots from Sam Z., taken at 0.8 GeV)
    // For BG: wrong-sign goes up by 3.8
    // For Signal: nu to nubar adjustment to fixed pot obtained by hacking
    // GetSignalScaleFactorT2K() to compare nubar to nu.  Result: 0.797151
    fBaseSigAnti = (fBaseSig / 3.8) * 0.797151;
    fBaseBkgAnti =  2.34 / 3.8 + 0.13 * 3.8;

    osc::OscCalculatorPMNS* osc = new osc::OscCalculatorPMNS;
    osc->SetL(kT2KBaseline);
    osc->SetRho(3.2);
    osc->SetDmsq21(7.6e-5);
    osc->SetDmsq32(2.4e-3);
    osc->SetTh12(.601);
    osc->SetTh13(asin(sqrt(.1))/2);
    osc->SetTh23(M_PI/4);
    osc->SetdCP(0);
    fOscBase = osc;

    LoadSpectra();

    // By default, fit to Neutrino results
    SetTruthParams(osc);
  }

  //----------------------------------------------------------------------
  void T2KToyExperiment::SetTruthParams(osc::IOscCalculatorAdjustable* osc)
  {
    // It seems very likely people are going to call us with NOvA
    // baseline. Ignore them.
    const double tmpL = osc->GetL();
    osc->SetL(kT2KBaseline);

    fSigFactorTrue = GetSignalScaleFactor(false, osc);
    fSigFactorAntiTrue = GetSignalScaleFactor(true, osc);

    osc->SetL(tmpL);
  }

  //----------------------------------------------------------------------
  T2KToyExperiment::~T2KToyExperiment()
  {
    delete fOscBase;
  }

  //----------------------------------------------------------------------
  void T2KToyExperiment::LoadSpectra()
  {
    // TODO: get from right location with cet::search_path or something
    TFile flux("flux_t2k.root");
    assert(!flux.IsZombie());
    fSpect = (TH1*)flux.Get("flux_t2k")->Clone();
    assert(fSpect);
    fSpect->SetDirectory(0);
    flux.Close();

    // cross section and osc weights
    for(int i = 1; i <= fSpect->GetNbinsX(); ++i){
      const double E = fSpect->GetXaxis()->GetBinCenter(i);
      double val = fSpect->GetBinContent(i);
      val *= E; // take QE xsec proportional to E

      // kill tiny bins at low E
      if(E < 0.4) val = 0;

      // kill bins above the E_reco cut at 1.25
      if(E > 1.25) val = 0;

      fSpect->SetBinContent(i, val);
    }
    // Normalize
    fSpect->Scale(1./fSpect->Integral());
  }

  //----------------------------------------------------------------------
  TH1* T2KToyExperiment::GetReweightedSignal(TH1* h, TH1* ret, bool anti,
                                             osc::IOscCalculator* calc) const
  {
    if(!ret || ret->GetNbinsX() != h->GetNbinsX())
      ret = (TH1*)h->Clone();
    // else, assume the histo has been provided

    // Average over the bin since the ratio can be quite spikey at any given E
    const int kNumSteps = 30;

    const int from = anti ? -14 : +14;
    const int to   = anti ? -12 : +12;
    for(int i = 1; i <= h->GetNbinsX(); ++i){
      if(h->GetBinContent(i) == 0){
        ret->SetBinContent(i, 0);
        continue;
      }

      double pTarget = 0;
      double pBase = 0;
      for(double E = h->GetXaxis()->GetBinLowEdge(i);
          E < h->GetXaxis()->GetBinUpEdge(i);
          E += h->GetXaxis()->GetBinWidth(i)/kNumSteps){
        pTarget += calc->P(from, to, E);
        pBase += fOscBase->P(from, to, E);
      }

      ret->SetBinContent(i, h->GetBinContent(i)*pTarget/pBase);
    }
    return ret;
  }

  //----------------------------------------------------------------------
  double T2KToyExperiment::GetSignalScaleFactor(bool anti,
                                                osc::IOscCalculator* calc) const
  {
    static TH1* hW = 0;
    hW = GetReweightedSignal(fSpect, hW, anti, calc);
    return hW->Integral()/ fSpect->Integral();
  }

  //----------------------------------------------------------------------
  double T2KToyExperiment::LogL(double data, double pred, double syst) const
  {
    // Scale as if it were a simple chi^2 to get some systematics in there
    const double syst_factor = pred ? pred/(pred + util::sqr(pred*syst)) : 1;

    return 2 * (pred-data+((data && pred) ? (data*log(data/pred)) : 0)) * syst_factor;
  }

  //----------------------------------------------------------------------
  double T2KToyExperiment::ChiSq(osc::IOscCalculatorAdjustable* calc,
                                 const SystShifts& /*shift*/) const
  {
    // It seems very likely people are going to call us with NOvA
    // baseline. Ignore them.
    const double tmpL = calc->GetL();
    calc->SetL(kT2KBaseline);

    double chi = 0;

    if(fPOT){
      // Scale for POTs
      const double scaledSig = (fPOT/fBasePOT) * fBaseSig;
      const double scaledBkg = (fPOT/fBasePOT) * fBaseBkg;

      // And then for two different sets of osc parameters
      const double data = scaledSig * fSigFactorTrue + scaledBkg;
      const double pred = scaledSig * GetSignalScaleFactor(false, calc) + scaledBkg;

      chi += LogL(data, pred, fTotSyst);
    }
    if(fPOTAnti){
      const double scaledSigAnti = (fPOTAnti/fBasePOT) * fBaseSigAnti;
      const double scaledBkgAnti = (fPOTAnti/fBasePOT) * fBaseBkgAnti;

      const double dataAnti = scaledSigAnti * fSigFactorAntiTrue + scaledBkgAnti;
      const double predAnti = scaledSigAnti * GetSignalScaleFactor(true,  calc) + scaledBkgAnti;

      chi += LogL(dataAnti, predAnti, fTotSyst);
    }

    calc->SetL(tmpL);

    return chi;
  }
} // end namespace
