#include "CAFAna/Utilities.h"

#include "CAFAna/Spectrum.h"

#include "TH2.h"
#include "TH3.h"
#include "TString.h"

#include <cassert>
#include <cmath>
#include <fstream>
#include <iostream>
#include <memory>
#include "sys/stat.h"
#include "wordexp.h"

namespace ana
{
  //----------------------------------------------------------------------
  std::string UniqueName()
  {
    static int N = 0;
    return TString::Format("cafanauniq%d", N++).Data();
  }

  //----------------------------------------------------------------------
  DontAddDirectory::DontAddDirectory()
  {
    fBackup = TH1::AddDirectoryStatus();
    TH1::AddDirectory(false);
  }

  //----------------------------------------------------------------------
  DontAddDirectory::~DontAddDirectory()
  {
    TH1::AddDirectory(fBackup);
  }

  //----------------------------------------------------------------------
  double LogLikelihood(const Spectrum& exp, const Spectrum& obs, double pot)
  {
    //    const double a = exp.ToTH1(18e20)->Integral(0, -1);
    //    const double b = obs.ToTH1(18e20)->Integral(0, -1);
    //    return util::sqr(a-b);

    if(pot == 0) pot = obs.POT();

    // With this value, negative expected events and one observed
    // event gives a chisq from this one bin of 182.
    const double minexp = 1e-40; // Don't let expectation go lower than this

    std::unique_ptr<TH1> eh(exp.ToTH1(pot));
    std::unique_ptr<TH1> oh(obs.ToTH1(pot));
    assert(eh->GetNbinsX() == oh->GetNbinsX());

    double chi = 0;

    for(int i = 0; i < eh->GetNbinsX()+2; ++i){
      double e = eh->GetBinContent(i);
      const double o = oh->GetBinContent(i);
      assert(o >= 0);
      if(e < minexp){
        if(!o) continue;
        e = minexp;
      }

      chi += 2*(e-o);
      if(o) chi += 2*o*log(o/e);
    }

    return chi;
  }

  //----------------------------------------------------------------------
  TH2F* ExpandedHistogram(const std::string& title,
                          int nbinsx, double xmin, double xmax,
                          int nbinsy, double ymin, double ymax)
  {
    // How wide the bins will be once we're done
    const double xwidth = (xmax-xmin)/(nbinsx-1);
    const double ywidth = (ymax-ymin)/(nbinsy-1);

    // Move the bin edges so that the limits occur at the centres
    xmin -= xwidth/2; ymin -= ywidth/2;
    xmax += xwidth/2; ymax += ywidth/2;

    return new TH2F(UniqueName().c_str(), title.c_str(),
                    nbinsx, xmin, xmax,
                    nbinsy, ymin, ymax);
  }

  //----------------------------------------------------------------------
  TH2* ToTH2(const Spectrum& s, double pot,
             int nx, double x0, double x1,
             int ny, double y0, double y1)
  {
    DontAddDirectory guard;

    TH1* h1 = s.ToTH1(pot);

    // Make sure it's compatible with having been made with this binning
    assert(h1->GetNbinsX() == nx*ny);

    TH2* ret = new TH2F(UniqueName().c_str(), "", nx, x0, x1, ny, y0, y1);

    for(int i = 0; i < h1->GetNbinsX(); ++i){
      const double val = h1->GetBinContent(i+1);
      const double err = h1->GetBinError(i+1);

      const int ix = i/ny;
      const int iy = i%ny;

      ret->SetBinContent(ix+1, iy+1, val);
      ret->SetBinError  (ix+1, iy+1, err);
    }

    delete h1;
    return ret;
  }

  //----------------------------------------------------------------------
  TH3* ToTH3(const Spectrum& s, double pot,
             int nx, double x0, double x1,
             int ny, double y0, double y1,
	     int nz, double z0, double z1)
  {
    DontAddDirectory guard;

    TH1* h1 = s.ToTH1(pot);

    // Make sure it's compatible with having been made with this binning
    assert(h1->GetNbinsX() == nx*ny*nz);

    TH3* ret = new TH3F(UniqueName().c_str(), "", 
			nx, x0, x1,
			ny, y0, y1,
			nz, z0, z1);

    for(int i = 0; i < h1->GetNbinsX(); ++i){
      const double val = h1->GetBinContent(i+1);
      const double err = h1->GetBinError(i+1);

      const int nynz = ny*nz;
      const int nmodnynz = i%nynz;
      const int ix = i/nynz;
      const int iy = nmodnynz/nz;
      const int iz = i%nz;

      ret->SetBinContent(ix+1, iy+1, iz+1, val);
      ret->SetBinError  (ix+1, iy+1, iz+1, err);
    }

    delete h1;
    return ret;
  }

  //----------------------------------------------------------------------
  std::vector<std::string> Wildcard(std::string wildcardString)
  {
    // Expand environment variables and wildcards like the shell would
    wordexp_t p;
    wordexp(wildcardString.c_str(), &p, WRDE_SHOWERR);

    std::vector<std::string> fileList;

    for(unsigned int i = 0; i < p.we_wordc; ++i){
      // Check the file exists before adding it
      struct stat sb;
      if(stat(p.we_wordv[i], &sb) == 0)
        fileList.push_back(p.we_wordv[i]);
    }

    wordfree(&p);

    return fileList;
  }

  //----------------------------------------------------------------------
  std::vector<std::string> LoadFileList(const std::string& listfile)
  {
    std::vector<std::string> ret;

    std::ifstream is(listfile);
    if(!is.good()){
      std::cerr << "Can't open file list '" << listfile << "'. Aborting." << std::endl;
      abort();
    }

    while(!is.eof()){
      std::string fname;
      is >> fname;
      if(!fname.empty()) ret.push_back(fname);
    }
    return ret;
  }
}
