#pragma once

#include <string>
#include <vector>

class TH1;
class TH2;
class TH3;
class TH2F;

namespace ana
{
  class Spectrum;

  /// Return a different string each time, for creating histograms
  std::string UniqueName();

  /// \brief Prevent histograms being added to the current directory
  ///
  /// Upon going out of scope, restores the previous setting
  class DontAddDirectory
  {
  public:
    DontAddDirectory();
    ~DontAddDirectory();
  protected:
    bool fBackup;
  };

  /** \brief The log-likelihood formula from the PDG.

      \param exp %A set of expected spectra
      \param obs The corresponding observed spectra
      \param pot POT to scale to. Leave at default to adopt POT from \a obs

      \returns The log-likelihood formula from the PDG
      \f[ \chi^2=2\sum_i^{\rm bins}\left(e_i-o_i+o_i\ln\left({o_i\over e_i}\right)\right) \f]

      Includes underflow and overflow bins and handles zero
      observed or expected events correctly.
  **/
  double LogLikelihood(const Spectrum& exp, const Spectrum& obs, double pot = 0);

  /// \brief Internal helper for \ref Surface and \ref FCSurface
  ///
  /// Creates a histogram having bins \em centred at the min and max
  /// coordinates
  TH2F* ExpandedHistogram(const std::string& title,
                          int nbinsx, double xmin, double xmax,
                          int nbinsy, double ymin, double ymax);

  /// \brief For use with \ref Var2D
  ///
  /// Re-expand a histogram flattened by \ref Var2D into a 2D histogram for
  /// plotting purposes. The binning scheme must match that used in the
  /// original Var.
  TH2* ToTH2(const Spectrum& s, double pot,
             int nx, double x0, double x1,
             int ny, double y0, double y1);

  /// Same as ToTH2, but with 3 dimensions
  TH3* ToTH3(const Spectrum& s, double pot,
             int nx, double x0, double x1,
             int ny, double y0, double y1,
             int nz, double z0, double z1);

  /// Find files matching a UNIX glob, plus expand environment variables
  std::vector<std::string> Wildcard(std::string wildcardString);

  /// Read list of input files from a text file, one per line
  std::vector<std::string> LoadFileList(const std::string& listfile);
}
