#include <iostream>
#include "CAFAna/Var.h"

namespace ana
{
  //----------------------------------------------------------------------
  /// Helper for \ref Var2D
  class Var2DFunc
  {
  public:
    Var2DFunc(const Var& a, int na, double a0, double a1,
              const Var& b, int nb, double b0, double b1)
      : fA(a), fNa(na), fAMin(a0), fAMax(a1),
        fB(b), fNb(nb), fBMin(b0), fBMax(b1)
    {
    }

    double operator()(const caf::StandardRecord* sr) const
    {
      /// Calculate current values of the variables in StandardRecord once
      const double a_value = fA(sr);
      const double b_value = fB(sr);

      /// Since there are no overflow/underflow bins, check the range
      if(a_value < fAMin || b_value < fBMin) { return -1.0;            }
      if(a_value > fAMax || b_value > fBMax) { return fNa * fNb + 1.0; }

      int ia = fNa*(a_value-fAMin)/(fAMax-fAMin);
      if(ia < 0) ia = 0;
      if(ia >= fNa) ia = fNa-1;

      int ib = fNb*(b_value-fBMin)/(fBMax-fBMin);
      if(ib < 0) ib = 0;
      if(ib >= fNb) ib = fNb-1;

      const int i = ia*fNb+ib;

      return i+.5;
    }
  protected:
    const Var& fA;
    int        fNa;
    double     fAMin, fAMax;
    const Var& fB;
    int        fNb;
    double     fBMin, fBMax;
  };

  class Var3DFunc
  {
  public:
    Var3DFunc(const Var& a, int na, double a0, double a1,
              const Var& b, int nb, double b0, double b1,
              const Var& c, int nc, double c0, double c1)
      : fA(a), fNa(na), fAMin(a0), fAMax(a1),
        fB(b), fNb(nb), fBMin(b0), fBMax(b1),
        fC(c), fNc(nc), fCMin(c0), fCMax(c1)
    {
    }

    double operator()(const caf::StandardRecord* sr) const
    {
      /// Calculate current values of the variables in StandardRecord once
      const double a_value = fA(sr);
      const double b_value = fB(sr);
      const double c_value = fC(sr);

      /// Since there are no overflow/underflow bins, check the range
      if(a_value < fAMin || b_value < fBMin || c_value < fCMin)
	{ return -1.0; }
      if(a_value > fAMax || b_value > fBMax || c_value > fCMax)
	{ return fNa * fNb * fNc + 1.0; }

      int ia = fNa*(a_value-fAMin)/(fAMax-fAMin);
      if(ia < 0) ia = 0;
      if(ia >= fNa) ia = fNa-1;

      int ib = fNb*(b_value-fBMin)/(fBMax-fBMin);
      if(ib < 0) ib = 0;
      if(ib >= fNb) ib = fNb-1;

      int ic = fNc*(c_value-fCMin)/(fCMax-fCMin);
      if(ic < 0) ic = 0;
      if(ic >= fNc) ic = fNc-1;

      const int i = ia*fNb*fNc + ib*fNc + ic;

      return i+.5;
    }
  protected:
    const Var& fA;
    int        fNa;
    double     fAMin, fAMax; 
    const Var& fB;
    int        fNb;
    double     fBMin, fBMax;
    const Var& fC;
    int        fNc;
    double     fCMin, fCMax;
  };


  //----------------------------------------------------------------------
  Var Var2D(const Var& a, int na, double a0, double a1,
            const Var& b, int nb, double b0, double b1)
  {
    std::set<std::string> areqs = a.Requires();
    std::set<std::string> breqs = b.Requires();
    areqs.insert(breqs.begin(), breqs.end());
    return Var(areqs, Var2DFunc(a, na, a0, a1, b, nb, b0, b1));
  }

  //----------------------------------------------------------------------
  Var Var3D(const Var& a, int na, double a0, double a1,
            const Var& b, int nb, double b0, double b1,
            const Var& c, int nc, double c0, double c1)
  {
    std::set<std::string> areqs = a.Requires();
    std::set<std::string> breqs = b.Requires();
    std::set<std::string> creqs = c.Requires();
    areqs.insert(breqs.begin(), breqs.end());
    areqs.insert(creqs.begin(), creqs.end());
    return Var(areqs, Var3DFunc(a, na, a0, a1, 
				b, nb, b0, b1,
				c, nc, c0, c1));
  }

  //----------------------------------------------------------------------
  Var Scaled(const Var& v, double s)
  {
    return Var(v.Requires(),
               [v, s](const caf::StandardRecord* sr){return s*v(sr);});
  }

  //----------------------------------------------------------------------
  Var Constant(double c)
  {
    return Var({}, [c](const caf::StandardRecord*){return c;});
  }

  //----------------------------------------------------------------------
  Var operator*(const Var& a, const Var& b)
  {
    std::set<std::string> areqs = a.Requires();
    std::set<std::string> breqs = b.Requires();
    areqs.insert(breqs.begin(), breqs.end());
    return Var(areqs,
               [a, b](const caf::StandardRecord* sr){return a(sr) * b(sr);});
  }
}
