#pragma once

// This file defines the basic Var object. For specific variables, and examples
// of how to implement your own, see Vars.h

#include <functional>
#include <set>
#include <string>

namespace caf{class StandardRecord;}

namespace ana
{
  /// \brief Representation of a variable to be retrieved from a \ref
  /// caf::StandardRecord object
  ///
  /// A Var consists of a function, taking a StandardRecord and returning the
  /// value of the variable (which may be some complicated function), plus a
  /// list of the branch names that must be active in the tree. This explicit
  /// naming of required branches leads to a significant speedup in CAF
  /// loading. However, if you neglect a necessary branch, that field will be
  /// uninitialized in the StandardRecord you are passed.
  class Var
  {
  public:
    /// The type of the function part of a var
    typedef double (VarFunc_t)(const caf::StandardRecord* sr);

    /// std::function can wrap a real function, function object, or lambda
    Var(const std::set<std::string>& reqs,
        const std::function<VarFunc_t>& fun) : fReqs(reqs), fFunc(fun) {}

    /// The list of branches required to be active
    const std::set<std::string>& Requires() const {return fReqs;}

    /// Allows a variable to be called with double value = myVar(sr) syntax
    double operator()(const caf::StandardRecord* sr) const
    {
      return fFunc(sr);
    }
  protected:
    std::set<std::string> fReqs;
    std::function<VarFunc_t> fFunc;
  };

  /// \brief For Vars where literally all you need is a single CAF variable
  ///
  /// eg Var myVar = SIMPLEVAR(my.var.str);
  /// NB lack of quotes quotes around my.var.str
#define SIMPLEVAR(CAFNAME) Var({#CAFNAME}, [](const caf::StandardRecord* sr){return sr->CAFNAME;})

  /// The simplest possible Var, always 1. Used as a default weight.
  const Var kUnweighted({}, [](const caf::StandardRecord*){return 1;});

  /// \brief Variable formed from two input variables
  ///
  /// The binning of each variable has to be given to allow conversion into a
  /// 1D binned representation.
  Var Var2D(const Var& a, int na, double a0, double a1,
            const Var& b, int nb, double b0, double b1);

  /// This is just like a Var2D, but usefull for 3D Spectra
  Var Var3D(const Var& a, int na, double a0, double a1,
            const Var& b, int nb, double b0, double b1,
            const Var& c, int nc, double c0, double c1);

  /// Use to rescale another variable.
  Var Scaled(const Var& v, double s);

  /// Use to weight events up and down by some factor
  Var Constant(double c);

  /// Most useful for combining weights
  Var operator*(const Var& a, const Var& b);
} // namespace
