////////////////////////////////////////////////////////////////////////
/// \brief   This module creates Common Analysis Files.
/// \author  rocco@physics.umn.edu
/// \date    July 2012
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <algorithm>
#include <cmath>
#include <fenv.h> 
#include <iomanip>
#include <iostream>
#include <iostream>
#include <map>
#include <string>
#include <time.h>
#include <vector>
#include <string>
#include <sstream>

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TSystem.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Core/FindOne.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "cetlib/exception.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
// Geo/Cal/Util
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/Geo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Metadata/MetadataManager.h"
#include "Utilities/func/MathUtil.h"
#include "RunHistory/service/RunHistoryService.h"

// Sim/BackTracker
#include "SimulationBase/GTruth.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/GENIEReweightTable.h"
#include "MCCheater/BackTracker.h"

// RecoBase
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Energy.h"
#include "RecoBase/Prong.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Shower.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"

// Other Reco
#include "CosRej/CosRejObj.h"
#include "CosRej/NueCosRej.h"
#include "RecoJMShower/EID.h"
#include "RecoJMShower/JMShower.h"
#include "RecVarPID/RVP.h"
#include "ReMId/ReMId.h"
#include "LEM/PIDDetails.h"
#include "LIDBuilder/EventLID.h"
#include "LIDBuilder/ShowerLID.h"
#include "ShowerLID/DedxDistribution.h"
#include "MichelEFilter/MichelECluster.h"
#include "MichelEFilter/SliceMEFCluster.h"
#include "MuonRemove/art/MRCCParent.h"
#include "NueSandbox/NueSandObj.h"
#include "NumuSandbox/NumuSandObj.h"
#include "NumuEnergy/NumuE.h"
#include "QEEventFinder/QePId.h"
#include "Preselection/Veto.h"

// SummData
#include "SummaryData/CosmicExposure.h"
#include "SummaryData/POTSum.h"
#include "SummaryData/SpillData.h"

// StandardRecord
#include "StandardRecord/StandardRecord.h"

// CAFMaker support
#include "CAFMaker/Blinding.h"



// note it is convention to NOT declare ANY "using namespace xxx" here
// especially "using namespace std"
namespace caf
{
  
  struct SliceInfo 
  {
    double nTotHit;
    double nCalHit;
    double nMipHit;
    double maxContig;
    
    unsigned int maxPlane;
    unsigned int minPlane;
    unsigned int maxCell;
    unsigned int minCell;
    unsigned int minCellwall;
    unsigned int minCellsFromEdge;
    
  };

  /// Module to create Common Analysis Files from ART files
  class CAFMaker : public art::EDProducer {
  public:
    explicit CAFMaker(fhicl::ParameterSet const& pset); // Required! explicit tag tells the compiler this is not a copy constructor      
    virtual ~CAFMaker();                        
    
    // Optional, read/write access to event
    void produce(art::Event& evt) noexcept; 
    
    void respondToOpenInputFile(const art::FileBlock& fb);
    
    // Optional if you want to be able to configure from event display, for example
    void reconfigure(const fhicl::ParameterSet& pset);
    
    // Optional use if you have histograms, ntuples, etc you want around for every event
    void beginJob();
    void endJob();
    virtual void beginRun(art::Run &r);
    virtual void endSubRun(art::SubRun& sr);
    
    
  private:
    ///< Sample configurable module parameters
    int number;
    std::string cafFilename;
    std::string dataTier;
    std::string generatorName;
    std::string reweightName;
    std::string exposureName;
    std::string clusterModName;
    std::string hitsModName;
    std::string kalmanTrackName;
    std::string discreteTrackName;
    std::string discreteTrack2dName;
    std::string cosmicTrackName;
    std::string fuzzyVertexName;
    std::string fuzzy3DAssnName;
    std::string fuzzy2DAssnName;
    std::string sliceMEFName;
    std::string elasticArmsName;
    std::string jmShowerName;
    std::string numuEnergyName;
    std::string vetoName;
    std::string remidName;
    std::string qepidName;
    std::string nueSelName;
    std::string lidName;
    std::string lemName;
    std::string rvpName;
    std::string cosrejName;
    std::string nuecosrejName;
    std::string kalmanMichelEName;
    std::string cosmicMichelEName;
    std::string discreteMichelEName;
    std::string nueSandboxName;
    std::string numuSandboxName;
    std::string numiBeamName;
    std::string mrccParentName;
    std::string fileExtension;

    bool generatorInfo;
    bool fullTruthInfo;
    bool enableBlindness;
    bool fIsRealData;
    double fTotalPOT;
    double fTotalEvents;
    double fTotalLivetime;
    int fMask;

    TFile*          fFile;
    TTree*          recTree;
    TH1D*           hPOT;
    TH1D*           hEvents;
    TH1D*           hLivetime;
    StandardRecord* rec;
    art::ServiceHandle<cheat::BackTracker>           bt;  
    art::ServiceHandle<calib::Calibrator>            cal;  
    art::ServiceHandle<geo::Geometry>                geom;  
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    
    novadaq::cnv::DetId fDetID;  ///< Detector ID in nova daq convention namespace typedef
    Det_t               fDet;    ///< Detector ID in caf namespace typedef
    
    //Creating separate friend tree and branches for spill info.
    //This tree will store one entry per event, even empty ones
    //so that POT counting can be done right.
    TTree*               spillTree;
    unsigned int         fRun;
    unsigned int         fSubrun;
    unsigned int         fEvt;
    bool                 fIsgoodspill;
    unsigned long int    fSpilltimesec;
    unsigned long int    fSpilltimensec;
    unsigned long int    fGpsspilltimesec;
    unsigned long int    fGpsspilltimensec;
    signed long long int fDeltaspilltimensec;
    double               fSpillpot;
    double               fHornI;
    double               fPosx;
    double               fPosy;
    double               fWidthx;
    double               fWidthy;
    
    
    
    void InitializeOutfile();
    
    void AddMetadataToFile(std::map<std::string, std::string> metadata);
    
    void AddCosmicTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                             std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRCosmic>* vec, const art::Event& evt);
    void AddGenInfoToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                         std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRGenerator>* vec, const art::Event& evt);

    void CalculateVisEFraction(std::vector<SRGenerator>* gen);

    void AddMCTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                         std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRNeutrino>* vec, const art::Event& evt);
    
    void AddParticleToVec(const sim::Particle &part, std::vector<cheat::TrackIDE> &allTracks, 
                          std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRTrueParticle>* vec);

    void AddTrueMichelEToVec(const sim::Particle &michel, const sim::Particle &motherMuon, 
                             std::vector<cheat::TrackIDE> &allTracks, std::vector<SRTrueMichelE>* vec);
    
    void FindAndAddMichels(std::vector<const sim::Particle*> particles,std::vector<cheat::TrackIDE> &allTracks, 
                                     std::vector<SRTrueMichelE>* michelVec);

    
    std::vector<int> getFaveIdsByEfficiency             (std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                         std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
    std::vector<int> getFaveIdsByEnergy                 (std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                         std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
    std::vector<int> getFaveIdsByPurity                 (std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                         std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
    std::vector<int> getFaveIdsByEfficiencyTimesPurity  (std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                         std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
    std::vector<int> getFaveIdsByEfficiencyThenPurity   (std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                         std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
    
    
    void drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable, 
                             const std::vector<int> &faveIds);
    void drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable, 
                             const std::vector<int> &faveIds, const std::vector<int> &otherFaveIds);
    
    SliceInfo getSliceInfo(const rb::Cluster& slice);

    void AddRBTrackToVec(const art::Ptr<rb::Track> &trk, std::vector<SRTrack>* trks);
    void AddTrkMEClusterToVec(const art::Ptr<me::MichelECluster> &michel, std::vector<SRTrkME>* vec, size_t trackId);
    void AddSlcMEClusterToVec(const art::Ptr<me::SliceMEFCluster> &michel, std::vector<SRSlcME>* vec);

    void FillTrackVars(const rb::Track& trk, SRTrack& srTrk);

    // Acts on any object that has a a superset of the same fields as the old SRProng
    template<class T> void FillProngVars(const rb::Prong &trk, T& srPrng);

    
    static bool EssentiallyEqual(double a, double b, double precision = 0.0001)
    { return a <= (b+precision) && a >= (b-precision); }
    
    static bool sortTrackLength(const SRTrack& a, const SRTrack& b) {return a.len>b.len; }

    static bool sortRBTrackLength(const art::Ptr<rb::Track>& a, const  art::Ptr<rb::Track>& b) {return a->TotalLength()>b->TotalLength(); }

    static bool sortRBProngEnergy(const art::Ptr<rb::Prong>& a, const art::Ptr<rb::Prong>& b) {return a->CalorimetricEnergy()>b->CalorimetricEnergy(); }
    
    static bool sortShowerEnergy(const SRJMShower& a, const SRJMShower& b) {return a.shwE>b.shwE; }

    static bool sortShowerLIDEnergy(const SRShowerLID& a, const SRShowerLID& b) {return a.shwE>b.shwE; }
    
    static bool sortElasticProng(const SRElastic& a, const SRElastic& b) {return a.png3D.size()>b.png3D.size(); }
    
    static bool sortMuIdLength(const SRMuId& a, const SRMuId& b) {return a.len>b.len; }
    
    static bool sortNuWithIdx(const cheat::NeutrinoWithIndex& a, const cheat::NeutrinoWithIndex& b) 
    {return a.truthColIndex<b.truthColIndex; }
    
    
    double SimpleOscProb(const simb::MCFlux& flux, const simb::MCNeutrino& nu) const;

    std::vector<caf::SRGenieWeights> GenieReweightTable(const rwgt::GENIEReweightTable& table) const;
  };
  
  //....................................................................... 
  
  //.......................................................................
  CAFMaker::CAFMaker(fhicl::ParameterSet const& pset):
  fIsRealData(false),
  fFile(0),
  bt(),
  cal(),
  geom()
  {
    //======================================================================
    // This is the constructor "nova" will use to create your module.
    //
    // Modules use configurations which have the same name as the
    // module. The variable "version" sets which version of this
    // configuration to use. Users are required to supply a configuration
    // called "default" but other versions ("test" eg.) are allowed.
    //
    // Which version to use is set at run time by the job description XML
    //======================================================================
    
    this->reconfigure(pset);    
  }
  
  //......................................................................
  CAFMaker::~CAFMaker()
  {
    //======================================================================
    // Clean up any memory allocated by your module
    //======================================================================
  }
  
  //......................................................................
  void CAFMaker::reconfigure(const fhicl::ParameterSet& pset)
  {
    // Keep everything lined up for clarity (stylistic)
    cafFilename         =   pset.get< std::string    >("CAFFilename");
    dataTier            =   pset.get< std::string    >("DataTier");
    generatorName       =   pset.get< std::string    >("GeneratorInput");
    reweightName        =   pset.get< std::string    >("ReweightInput");
    exposureName        =   pset.get< std::string    >("ExposureInput");
    clusterModName      =   pset.get< std::string    >("ClusterInput");
    hitsModName         =   pset.get< std::string    >("HitInput");
    kalmanTrackName     =   pset.get< std::string    >("KalmanInput");
    discreteTrackName   =   pset.get< std::string    >("DiscreteInput");
    discreteTrack2dName =   pset.get< std::string    >("Discrete2dInput");
    cosmicTrackName     =   pset.get< std::string    >("CosmicInput");
    fuzzyVertexName     =   pset.get< std::string    >("FuzzyInput");
    fuzzy3DAssnName     =   pset.get< std::string    >("Fuzzy3DAssnInput");
    fuzzy2DAssnName     =   pset.get< std::string    >("Fuzzy2DAssnInput");
    sliceMEFName        =   pset.get< std::string    >("SliceMEFInput");
    elasticArmsName     =   pset.get< std::string    >("ElasticInput");
    jmShowerName        =   pset.get< std::string    >("JmShowerInput");
    numuEnergyName      =   pset.get< std::string    >("NumuEnergyInput");
    vetoName            =   pset.get< std::string    >("VetoInput");
    remidName           =   pset.get< std::string    >("RemidInput");
    qepidName           =   pset.get< std::string    >("QepidInput");
    nueSelName          =   pset.get< std::string    >("NueSelInput");
    lidName             =   pset.get< std::string    >("LIDInput");
    lemName             =   pset.get< std::string    >("LemInput");
    rvpName             =   pset.get< std::string    >("RvpInput");
    cosrejName          =   pset.get< std::string    >("CosRejInput");
    nuecosrejName       =   pset.get< std::string    >("NueCosRejInput");
    kalmanMichelEName   =   pset.get< std::string    >("KalmanMichelEInput");
    cosmicMichelEName   =   pset.get< std::string    >("CosmicMichelEInput");
    discreteMichelEName =   pset.get< std::string    >("DiscreteMichelEInput");
    nueSandboxName      =   pset.get< std::string    >("NueSandboxInput");
    numuSandboxName     =   pset.get< std::string    >("NumuSandboxInput");
    numiBeamName        =   pset.get< std::string    >("NuMIBeamInput");
    generatorInfo       =   pset.get< bool           >("GeneratorInfo");
    fullTruthInfo       =   pset.get< bool           >("FullTruthInfo");
    enableBlindness     =   pset.get< bool           >("EnableBlindness");
    mrccParentName      =   pset.get<std::string     >("MRCCParentInput");
    fileExtension       =   pset.get<std::string     >("FileExtension");
  }
  
  //......................................................................
  void CAFMaker::respondToOpenInputFile(const art::FileBlock& fb)
  {
    if(!fFile)
    {
      // Filename wasn't set in the FCL, and this is the first file we've seen
      
      cafFilename = basename(fb.fileName().c_str());
      const size_t dotpos = cafFilename.find('.');
      assert(dotpos != std::string::npos); // Must have a dot, surely?
      cafFilename.resize(dotpos);
      cafFilename += fileExtension;
      
      InitializeOutfile();
    }
  }
  
  //......................................................................
  void CAFMaker::beginJob()
  {
    if(!cafFilename.empty()) InitializeOutfile();
  }

  //......................................................................
  void CAFMaker::beginRun(art::Run &r)
  {
    fDetID = geom->DetId();
    fDet   = (Det_t) fDetID;
    art::Handle<int> mask; // if good diblock mask is in file, make sure to set RH to use it
    r.getByLabel(hitsModName,mask);
    if(!mask.failedToGet()) {
      rh->SetGoodDiBlockMask(*mask);
    }
  }

  //......................................................................
  void CAFMaker::InitializeOutfile()
  {
    assert(!fFile);
    assert(!cafFilename.empty());
    
    mf::LogInfo("CAFMaker") << "Output filename is " << cafFilename;
    
    //      art::ServiceHandle<art::TFileService> tfs;
    //      recTree = tfs->make<TTree>("recTree", "records");
    //      recTree->Branch("rec", "caf::StandardRecord", &rec, 16000, 99);
    
    //        TVector3::Class()->IgnoreTObjectStreamer();
    
    fFile = new TFile(cafFilename.c_str(), "RECREATE");
    
    hPOT = new TH1D("TotalPOT", "TotalPOT;; POT", 1, 0, 1);
    hEvents = new TH1D("TotalEvents", "TotalEvents;; Events", 1, 0, 1);
    hLivetime = new TH1D("TotalLivetime", "TotalLiveTime;;Seconds", 1, 0, 1);
    
    recTree = new TTree("recTree", "records");
    
    // Create new StandardRecord object
    rec = new StandardRecord();

    recTree->Branch("rec", "caf::StandardRecord", &rec);
    
    delete rec;
    
    //create tree to store spill info for POT counting
    spillTree = new TTree("spillTree", "spills");
    spillTree->Branch("run",                &fRun,                "run/i");
    spillTree->Branch("subrun",             &fSubrun,             "subrun/i");
    spillTree->Branch("evt",                &fEvt,                "evt/i");
    spillTree->Branch("isgoodspill",        &fIsgoodspill,        "isgoodspill/O");
    spillTree->Branch("spilltimesec",       &fSpilltimesec,       "spillsimtesec/l");
    spillTree->Branch("spilltimensec",      &fSpilltimensec,      "spilltimensec/l");
    spillTree->Branch("gpsspilltimesec",    &fGpsspilltimesec,    "gpsspilltimesec/l");
    spillTree->Branch("gpsspilltimensec",   &fGpsspilltimensec,   "gpsspilltimensec/l");
    spillTree->Branch("deltaspilltimensec", &fDeltaspilltimensec, "deltaspilltimensec/L");
    spillTree->Branch("spillpot",           &fSpillpot,           "spillpot/D");
    spillTree->Branch("hornI",              &fHornI,              "hornI/D");
    spillTree->Branch("posx",               &fPosx,               "posx/D");
    spillTree->Branch("posy",               &fPosy,               "posy/D");
    spillTree->Branch("widthx",             &fWidthx,             "widthx/D");
    spillTree->Branch("widthy",             &fWidthy,             "widthy/D");
    
    //make the spillTree a friend of the recTree
    recTree->AddFriend("spillTree");
    
    
    fTotalPOT = 0;
    fTotalEvents = 0;
    fTotalLivetime = 0;
    
    // Global information about the processing details:
    
    std::map<TString, TString> envmap;
    std::string envstr;
    // environ comes from unistd.h
    for(char** penv = environ; *penv; ++penv)
    {
      const std::string pair = *penv;
      envstr += pair;
      envstr += "\n";
      const size_t split = pair.find("=");
      if(split == std::string::npos) continue; // Huh?
      const std::string key = pair.substr(0, split);
      const std::string value = pair.substr(split+1);
      envmap[key] = value;
    }
    
    // Get the command-line we were invoked with. What I'd really like is just
    // the fcl script and list of input filenames in a less hacky fashion. I'm
    // not sure that's actually possible in ART. TODO: ask the artists.
    FILE* cmdline = fopen("/proc/self/cmdline", "rb");
    char* arg = 0;
    size_t size = 0;
    std::string cmd;
    while(getdelim(&arg, &size, 0, cmdline) != -1)
    {
      cmd += arg;
      cmd += " ";
    }
    free(arg);
    fclose(cmdline);
    
    fFile->mkdir("env")->cd();
    
    TObjString(envmap["USER"]).Write("user");
    TObjString(envmap["HOSTNAME"]).Write("hostname");
    TObjString(envmap["PWD"]).Write("pwd");
    TObjString(envmap["SRT_PUBLIC_CONTEXT"]).Write("publiccontext");
    TObjString(envmap["SRT_PRIVATE_CONTEXT"]).Write("privatecontext");
    // default constructor is "now"
    TObjString(TTimeStamp().AsString()).Write("date");
    TObjString(cmd.c_str()).Write("cmd");
    TObjString(cafFilename.c_str()).Write("output");
    TObjString(envstr.c_str()).Write("env");
  }
  
  void CAFMaker::AddMetadataToFile(std::map<std::string, std::string> metadata)
  {
    assert(fFile && "CAFMaker: Trying to add metadata to an uninitialized file");
    
    fFile->mkdir("metadata")->cd();
    
    for(auto const& pair:metadata){
      TObjString(pair.second.c_str()).Write(pair.first.c_str());
    }
  }    
  
  
  
  //......................    //......................
  void CAFMaker::AddParticleToVec(const sim::Particle &part, std::vector<cheat::TrackIDE> &allTracks, 
                                  std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRTrueParticle>* vec)
  {
    int trackId = part.TrackId();
    //find visible energy
    double visE = 0;
    for (std::vector<cheat::TrackIDE>::iterator  it = allTracks.begin(); 
         it != allTracks.end(); it++)
    {
      if(trackId == it->trackID)
      {
        visE += it->energy;
        break;
      }
      
    }
    //find visible energy in slice
    double inSliceVisE = 0;
    for (std::vector<cheat::TrackIDE>::iterator  it = sliceTracks.begin(); 
         it != sliceTracks.end(); it++)
    {
      if(trackId == it->trackID)
      {
        inSliceVisE += it->energy;
        break;
      }
    }
    //make a new SRTrueParticle and get an accessor 
    vec->push_back(SRTrueParticle());
    std::vector<SRTrueParticle>::reverse_iterator srPart = vec->rbegin();
    //Set fields 
    srPart->pdg = part.PdgCode();
    srPart->visE = visE;
    srPart->visEinslc = inSliceVisE;
    srPart->p = part.Momentum();
    
  }
  
  //......................    //......................
  void CAFMaker::AddTrueMichelEToVec(const sim::Particle &michel, const sim::Particle &motherMuon, 
                                     std::vector<cheat::TrackIDE> &allTracks, std::vector<SRTrueMichelE>* vec)
  {

    vec->push_back(SRTrueMichelE());
    std::vector<SRTrueMichelE>::reverse_iterator srMichel = vec->rbegin();
        
    // find the visible energy from the FLS hits
    float visE = 0;
    for (std::vector<cheat::TrackIDE>::iterator  it = allTracks.begin(); 
         it != allTracks.end(); it++)
    {
      if(michel.TrackId() == it->trackID)
      {
        visE = it->energy;
        break;
      }
    }
    
    srMichel->E       = michel.E();
    srMichel->visE    = visE;
    srMichel->time    = michel.T();
    srMichel->mustop  = motherMuon.EndPosition();
    srMichel->p       = michel.Momentum();
    
  }
  
  //......................................................................  
  
  void CAFMaker::AddMCTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                                 std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRNeutrino>* vec, const art::Event& evt)
  {
    //get information about truNu
    
    vec->push_back(SRNeutrino());
    std::vector<SRNeutrino>::reverse_iterator srTruth =  vec->rbegin();
    
    const art::Ptr<simb::MCTruth> &tru = effPur->neutrinoInt;
    
    srTruth->pdg          = tru->GetNeutrino().Nu().PdgCode();
    srTruth->hitnuc       = tru->GetNeutrino().HitNuc();
    
    srTruth->mode         = tru->GetNeutrino().Mode();  // to avoid depending on simb
    srTruth->iscc         = (tru->GetNeutrino().CCNC() == simb::kCC);
    srTruth->inttype      = tru->GetNeutrino().InteractionType();
    
    srTruth->q2           = tru->GetNeutrino().QSqr();
    srTruth->x            = tru->GetNeutrino().X();
    srTruth->y            = tru->GetNeutrino().Y();
    srTruth->W2           = tru->GetNeutrino().W() * tru->GetNeutrino().W();
    
    srTruth->E            = tru->GetNeutrino().Nu().Momentum().E();
    srTruth->visE         = effPur->energyTotal;
    srTruth->visEinslc    = effPur->energySlice;
    srTruth->eff          = effPur->efficiency;
    srTruth->pur          = effPur->purity;
    srTruth->nhitslc      = effPur->nSliceHits;
    srTruth->nhittot      = effPur->nTotalHits;
    srTruth->time         = tru->GetNeutrino().Nu().Position().T();
    srTruth->vtx          = tru->GetNeutrino().Nu().Position().Vect();
    srTruth->p            = tru->GetNeutrino().Nu().Momentum();
    
    art::PtrVector<simb::MCTruth> pv;
    pv.push_back(tru);
    art::FindManyP<simb::MCFlux> fmFlux(pv, evt, generatorName);
    if(fmFlux.isValid())
    {
      std::vector< art::Ptr<simb::MCFlux> > fluxes = fmFlux.at(0);
      
      if(!fluxes.empty())
      {
        const simb::MCFlux& flux = *fluxes[0];
        
        srTruth->pdgorig  = flux.fntype;
        srTruth->woscdumb = SimpleOscProb(flux, tru->GetNeutrino());
        
        SRBeam& beam = srTruth->beam;
        
        beam.tv       = TVector3(flux.ftvx, flux.ftvy, flux.ftvz);
        beam.tp       = TVector3(flux.ftpx, flux.ftpy, flux.ftpz);
        beam.runjob   = flux.frun;
        beam.potnum   = flux.fevtno;
        beam.tptype   = flux.ftptype;
        beam.nimpwt   = flux.fnimpwt;
        beam.ndecay   = flux.fndecay;
        beam.v        = TVector3(flux.fvx, flux.fvy, flux.fvz);
        beam.pdp      = TVector3(flux.fpdpx, flux.fpdpy, flux.fpdpz);
        beam.ppdxdz   = flux.fppdxdz;
        beam.ppdydz   = flux.fppdydz;
        beam.pppz     = flux.fpppz;
        beam.ppenergy = flux.fppenergy;
        beam.ppmedium = flux.fppmedium;
        beam.ptype    = flux.fptype;
        beam.ppv      = TVector3(flux.fppvx, flux.fppvy, flux.fppvz);
      }
      else{
        throw cet::exception("Missing Association Error", "Failed to find MCFlux object assoicated with MCTruth.  This probably means something went wrong with overlays.");
      }
    }
    
    
    //get the neutrino's daughters (and other progeny)
    std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(tru);
    
    // We're going to loop over daughters to add other primaries to truNu.
    // First we're gonna get the lepton and add it as the zeroth entry, 
    // so let's get its information.
    int lepPdg = tru->GetNeutrino().Lepton().PdgCode();
    double lepEnergy = tru->GetNeutrino().Lepton().E();
    int lepCount = 0;  //count the found leptons, for debugging.
    int lepTrackId = 0; // We'll set this later
    // now we'll loop over the lepton and add it
    for(std::vector<const sim::Particle*>::iterator partIt = particles.begin(); 
        partIt != particles.end(); partIt++)
    {
      //Add particle if it's a primary and it has the correct PDG code and the correct energy.
      if(
         (*partIt)->Mother() == 0 && (*partIt)->PdgCode() == lepPdg
                   && EssentiallyEqual((*partIt)->E(), lepEnergy)
         )
      {
        lepTrackId = (*partIt)->TrackId();
        ++lepCount;
        AddParticleToVec(**partIt, allTracks, sliceTracks, &srTruth->prim);
      }
    }
    
    
    if(lepCount != 1)
    {
      switch (lepCount) {
        case 0:
          if (!srTruth->iscc) break;

          mf::LogError("CAFMaker")
          << "*****************************************" << "\n" 
          << "Uh, oh, CAFMaker did not find primary lepton. "<< "\n" 
          << "We're going to need to find a way to deal with this."  << "\n" 
          << "Info about the particle: "   << "\n" 
          << "  pdg: " << srTruth->pdg   << "\n" 
          << "  pdgorig: " << srTruth->pdgorig   << "\n" 
          << "  E: "<< srTruth->E        << "\n" 
          << "  visE: "<< srTruth->visE        << "\n" 
          << "  nhittot: "<< srTruth->nhittot        << "\n" 
          << "  X: "<< srTruth->vtx.X()  << "\n" 
          << "  Y: "<< srTruth->vtx.Y()  << "\n" 
          << "  Z: "<< srTruth->vtx.Z()  << "\n" 
          << "  iscc: "<< srTruth->iscc  << "\n" 
          << "  infid(nd): "<< (srTruth->vtx.X() < 200 && srTruth->vtx.X() > -200 && 
                                srTruth->vtx.Y() < 200 && srTruth->vtx.Y() > -200 && 
                                srTruth->vtx.Z() > 0 && srTruth->vtx.X() < 1600  ) << "\n" 
          
          << "*****************************************" << "\n" ;
          break;
        default:
          mf::LogError("CAFMaker")
          << "*****************************************" << "\n"
          << "Uh, oh, CAFMaker found too many primary leptons in event. "<< "\n"
          << "We're going to need to find a way to deal with this."  << "\n"
          << "*****************************************";
          break;
      }
      //      assert(lepCount == 1);
    } 
   
    for(std::vector<const sim::Particle*>::iterator partIt = particles.begin(); 
        partIt != particles.end(); partIt++)
    {
      //Add particle if it's a primary and it's not the lepton
      if((*partIt)->TrackId() != lepTrackId 
         && (*partIt)->Mother() == 0)
        AddParticleToVec(**partIt, allTracks, sliceTracks, &srTruth->prim);
      
    }
    
    FindAndAddMichels(particles, allTracks, &srTruth->michel);

  }
  
  void CAFMaker::AddGenInfoToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                                 std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRGenerator>* vec, const art::Event& evt) {
    
    vec->push_back(SRGenerator());
    std::vector<SRGenerator>::reverse_iterator gentruth =  vec->rbegin();
    
    const art::Ptr<simb::MCTruth> &tru = effPur->neutrinoInt;
    
    //Build the particle list first
    simb::MCNeutrino neutrino = tru->GetNeutrino();
    simb::MCParticle nu       = neutrino.Nu();
    
    for(int i = 0; i < tru->NParticles(); i++) {
      simb::MCParticle mcpart = tru->GetParticle(i);
      caf::SRGenParticle gpart;
      
      //int gmid = mcpart.PdgCode();
      int ndaughters = mcpart.NumberDaughters();
      int gmmo = mcpart.Mother();
      
      int fdtrkid = 0;
      int ldtrkid = 0;
      
      if(ndaughters !=0) {
        fdtrkid = mcpart.Daughter(0);
        if(ndaughters == 1) {
          ldtrkid = 1;
        }
        else if(ndaughters >1) {
          ldtrkid = mcpart.Daughter(ndaughters-1);
        }
      }      
      int gmfd = -1;
      int gmld = -1;
      //Genie uses the index in the particle array to reference the daughter particles.
      //MCTruth keeps the particles in the same order so use the track ID to find the proper index.
      for(int j = 0; j < tru->NParticles(); j++) {
        simb::MCParticle temp = tru->GetParticle(i);
        if(temp.TrackId() == fdtrkid) {
          gmfd = j;
        }
        if(temp.TrackId() == ldtrkid) {
          gmld = j;
        } 
      }
      gpart.fP4            = mcpart.Momentum();
      gpart.fVtx           = mcpart.Position();
      gpart.fPolz          = mcpart.Polarization();
      gpart.fNuclVtx       = mcpart.GetGvtx();
      
      gpart.fPdgCode       = mcpart.PdgCode();         
      gpart.fStatus        = mcpart.StatusCode();        
      gpart.fRescatterCode = mcpart.Rescatter();
      gpart.fFirstDaughter = gmfd;
      gpart.fLastDaughter  = gmld;   
      gpart.fMother        = gmmo;
      
      gentruth->fParticleList.push_back(gpart);
      if(i==0) {
        gentruth->fProbeIndx = 0; //&gentruth->fParticleList[0];
      }
      if( (mcpart.Mother()==tru->GetNeutrino().Nu().TrackId()) 
         && ((mcpart.PdgCode()  == tru->GetNeutrino().Nu().PdgCode()) 
             || ( mcpart.PdgCode()  == (tru->GetNeutrino().Nu().PdgCode())-1) ) ){
        gentruth->fFSLeptonIndx = i; //&gentruth->fParticleList[i];
      }
      
    }
    
    //Now the rest of the MCTruth stuff
    gentruth->fHitNucleon = tru->GetNeutrino().HitNuc();
    gentruth->fHitQuark   = tru->GetNeutrino().HitQuark();
    
    
    //GTruth stuff here
    art::PtrVector<simb::MCTruth> pv;
    pv.push_back(tru);
    art::FindManyP<simb::GTruth> fmGtruth(pv, evt, generatorName);
    if(fmGtruth.isValid()) {
      std::vector< art::Ptr<simb::GTruth> > gtruths = fmGtruth.at(0);
      
      if(!gtruths.empty()) {
        const simb::GTruth& gtruth = *gtruths[0];
	
	gentruth->fVisEfraction = 1.0;
	gentruth->visEinslc    = effPur->energySlice;
	gentruth->fProbePDG    = gtruth.fProbePDG;
        gentruth->fProbeP4     = gtruth.fProbeP4;
        gentruth->fVertex      = gtruth.fVertex;
        gentruth->fHitNucP4    = gtruth.fHitNucP4;
        gentruth->ftgtPDG      = gtruth.ftgtPDG;
        gentruth->ftgtA        = gtruth.ftgtA;
        gentruth->ftgtZ        = gtruth.ftgtZ;
        gentruth->fIsSeaQuark  = gtruth.fIsSeaQuark;
        gentruth->fFShadSystP4 = gtruth.fFShadSystP4;
        gentruth->fgY          = gtruth.fgY;
        gentruth->fgX          = gtruth.fgX;
        gentruth->fgT          = gtruth.fgT;
        gentruth->fgW          = gtruth.fgW;
        gentruth->fgq2         = gtruth.fgq2;
        gentruth->fgQ2         = gtruth.fgQ2;
        gentruth->fResNum      = gtruth.fResNum;
        gentruth->fIsCharm     = gtruth.fIsCharm;
        gentruth->fNumNeutron  = gtruth.fNumNeutron;
        gentruth->fNumProton   = gtruth.fNumProton;
        gentruth->fNumPi0      = gtruth.fNumPi0;
        gentruth->fNumPiMinus  = gtruth.fNumPiMinus;
        gentruth->fNumPiPlus   = gtruth.fNumPiPlus;
        gentruth->fweight      = gtruth.fweight;
        gentruth->fprobability = gtruth.fprobability;
        gentruth->fXsec        = gtruth.fXsec;
        gentruth->fDiffXsec    = gtruth.fDiffXsec;
        gentruth->fGint        = gtruth.fGint;
        gentruth->fGscatter    = gtruth.fGscatter;
	
      }
    }
  }

  void CAFMaker::CalculateVisEFraction(std::vector<SRGenerator>* gen) {
    double totalVisEinSlice = 0;
    for(unsigned int i = 0; i< gen->size(); i++) {
      totalVisEinSlice += (*gen)[i].visEinslc;
      //totalVisEinSlice += gen[i]->visEinslc;
    }
    for(unsigned int i = 0; i < gen->size(); i++) {
      (*gen)[i].fVisEfraction = ( (*gen)[i].visEinslc )/totalVisEinSlice;
      //gen[i]->fVisEfraction = gen[1]->visEinslc)/totalVisEinSlice;
    }

  }
  
  //.....................................................................
  void CAFMaker::AddCosmicTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                                     std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRCosmic>* vec, const art::Event& evt)
  {
    //get information about true cosmic
    
    vec->push_back(SRCosmic());
    std::vector<SRCosmic>::reverse_iterator srTruth =  vec->rbegin();
    
    const art::Ptr<simb::MCTruth> &tru = effPur->neutrinoInt;

    /// BUT GetParticle(int i) requires looping over list of particles in 
    /// single MCTruth object. So we have made an MCTruth object for each
    /// "cosmic" and GetParticle(0) corresponds to the only particle then
    srTruth->pdg          = tru->GetParticle(0).PdgCode();
    
    srTruth->E            = tru->GetParticle(0).Momentum().E();
    srTruth->visE         = effPur->energyTotal;
    srTruth->visEinslc    = effPur->energySlice;
    srTruth->eff          = effPur->efficiency;
    srTruth->pur          = effPur->purity;
    srTruth->nhitslc      = effPur->nSliceHits;
    srTruth->nhittot      = effPur->nTotalHits;
    
    srTruth->time         = tru->GetParticle(0).Position().T();
    srTruth->vtx          = tru->GetParticle(0).Position().Vect();
    srTruth->p            = tru->GetParticle(0).Momentum();
    
    
    srTruth->azimuth      = atan2(tru->GetParticle(0).Momentum().Pz(),
                                  tru->GetParticle(0).Momentum().Px());
    srTruth->zenith       = -(tru->GetParticle(0).Momentum().Py())/tru->GetParticle(0).Momentum().P();
    
    std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(tru);
    
    
    const simb::MCTrajectory& traj = particles[0]->Trajectory();
    
      
    
    bool entered = 0;    
    bool exited = 0;    

    for(size_t pt = 0; pt < traj.size(); ++pt)
    {
      TVector3 point = traj.Position(pt).Vect();
      
      // Check to see if trajectory point is in the detector 
      bool inDet = fabs(point.X()) <= geom->DetHalfWidth()   // Check X inside detector 
                && fabs(point.Y()) <= geom->DetHalfHeight()  /// Then check Y
                && point.Z() >= 0 && point.Z() <= geom->DetLength(); // then check Z

      if(inDet && !entered)
      {
        // then we just entered the detector 
        entered = true;
        
        if(pt > 0)
        {
          // Put xyz from last point into an array for geo function call
          TVector3 lastPoint = traj.Position(pt - 1).Vect();
          double pointArray[3] = {point.X(), point.Y(), point.Z()};
          
          // Put xyz from direction vector into an array.  
          TVector3 dir = (lastPoint - point).Unit();
          double dirArray[3] = {dir.X(), dir.Y(), dir.Z()};
          
          double enterArray[3] = {0,0,0};
          
          geo::ProjectToBoxEdge(pointArray, dirArray, 
                                           - geom->DetHalfWidth()  , geom->DetHalfWidth(), 
                                           - geom->DetHalfHeight() , geom->DetHalfHeight(), 
                                                                0  , geom->DetLength(), 
                                           enterArray);
          
          srTruth->enter.SetXYZ(enterArray[0], enterArray[1], enterArray[2]);
          
        }
        else
        {// if this is the case, we started inside the detector.
          srTruth->enter = point;
        }
      }
      
      if(!inDet && entered && !exited){
        // then we just exited the detector 
        exited= true;
        
        // Put xyz from last point into an array for geo function call
        TVector3 lastPoint = traj.Position(pt - 1).Vect();
        double lastPointArray[3] = {lastPoint.X(), lastPoint.Y(), lastPoint.Z()};
        
        // Put xyz from direction vector into an array.  
        TVector3 dir = (point - lastPoint).Unit();
        double dirArray[3] = {dir.X(), dir.Y(), dir.Z()};
        
        double exitArray[3] = {0,0,0};
        
        geo::ProjectToBoxEdge(lastPointArray, dirArray, 
                              - geom->DetHalfWidth()  , geom->DetHalfWidth(), 
                              - geom->DetHalfHeight() , geom->DetHalfHeight(), 
                              0  , geom->DetLength(), 
                              exitArray);
        
        srTruth->exit.SetXYZ(exitArray[0], exitArray[1], exitArray[2]);
        
      }
      
      
    }
    srTruth->stop = traj.Position(traj.size() - 1).Vect();
    
    // make sure the particle actually entered the detector  
    if(!entered){
      /// if the primary never entered the detector, set the coordinates to -5*10^9
      
      srTruth->enter.SetXYZ(-5e9, -5e9, -5e9);
      srTruth->exit.SetXYZ(-5e9, -5e9, -5e9);
      
    }

    /// If the particle entered and did not exit, set the exit position to stop
    if(entered && !exited){
      srTruth->exit = srTruth->stop;
    }
    // Find the Michel electrons
    FindAndAddMichels(particles, allTracks, &srTruth->michel);
    
  }
  //......................................................................
  
  void CAFMaker::FindAndAddMichels(std::vector<const sim::Particle*> particles,std::vector<cheat::TrackIDE> &allTracks, 
                                   std::vector<SRTrueMichelE>* michelVec)
  {
    for(std::vector<const sim::Particle*>::iterator partIt = particles.begin(); 
        partIt != particles.end(); partIt++)
    {
      
      // Check to see if we're dealing with an electron, continue if not.  
      int pdg    = (*partIt)->PdgCode();
      if (abs(pdg) != 11) continue;
      
      // Check to see if mother is a muon, continue if not 
      int motherId = (*partIt)->Mother();
      if( motherId == 0  )
        continue;
      const sim::Particle*  mother = bt->TrackIDToParticle( motherId );
      int pdgMother = 0;
      if(mother) pdgMother = mother->PdgCode();
      if(abs(pdgMother) != 13) continue;
      
      // Check to see if particle was produced in decay, continue if not.  
      if((*partIt)->Process().compare("Decay")) continue;
      
      // Add particle if it passed all criteria 
      AddTrueMichelEToVec(**partIt, *bt->TrackIDToParticle( motherId ), 
                          allTracks, michelVec);
      
      
    }
    
  }
  
  
  
  //......................................................................
  
  void CAFMaker::AddRBTrackToVec(const art::Ptr<rb::Track> &trk, std::vector<SRTrack>* trks)
  {
    trks->push_back(SRTrack());
    std::vector<SRTrack>::reverse_iterator srTrk = trks->rbegin();
    
    srTrk->nhit         = trk->NCell();
    srTrk->nhitx        = trk->NXCell();
    srTrk->nhity        = trk->NYCell();
    srTrk->nplane       = trk->ExtentPlane();
    srTrk->maxplanecont = trk->MostContiguousPlanes(geo::kXorY);
    srTrk->maxplanegap  = trk->MostMissingPlanes(geo::kXorY);
    srTrk->start        = trk->Start();
    srTrk->stop         = trk->Stop();
    srTrk->dir          = trk->Dir();
    srTrk->len          = trk->TotalLength();
    
    srTrk->lenE     = srTrk->len * 0.00157; // From N. Raddatz and K. Sachdev, see DocDB 7658-v2, slide 11 
    // trkLength * MIP dE/dx
    srTrk->calE     = trk->CalorimetricEnergy();
    
    if      (trk->NXCell() == 0)  srTrk->view = kY;
    else if (trk->NYCell() == 0)  srTrk->view = kX;
    else                          srTrk->view = kXorY;
    
    
  }
  
  //......................................................................  
  void CAFMaker::AddTrkMEClusterToVec(const art::Ptr<me::MichelECluster> &michel, std::vector<SRTrkME>* vec, size_t trackId)
  {

    vec->push_back(SRTrkME());
    std::vector<SRTrkME>::reverse_iterator srME = vec->rbegin();
    
    srME->nhit         =                     michel->NCell();         
    srME->calE         =                     michel->CalorimetricEnergy();
    srME->deltat       =                     michel->DeltaT();         
    srME->disttotrack  =                     michel->DistanceToTrackEnd();         
    srME->atend        =                     michel->AtTrackEnd();         
    srME->view         = (caf::_plane_proj)  michel->View();         
    srME->meanpos      =                     michel->MeanXYZ();         
    srME->trkidx       =      (short)        trackId;         
    
  }

  //......................................................................  
  void CAFMaker::AddSlcMEClusterToVec(const art::Ptr<me::SliceMEFCluster> &michel, std::vector<SRSlcME>* vec)
  {

    vec->push_back(SRSlcME());
    std::vector<SRSlcME>::reverse_iterator srME = vec->rbegin();
    
    srME->nhit         =                     michel->NCell();         
    srME->nmatches     =    (unsigned short) michel->NSlcMatches();         
    srME->calE         =                     michel->CalorimetricEnergy();
    srME->deltat       =                     michel->DeltaT();         
    srME->disttoslc    =                     michel->DistToSlc();         
    srME->meanpos      =                     michel->MeanXYZ();         
    
  }
  

  //......................................................................
  
  template<class T> void CAFMaker::FillProngVars(const rb::Prong &prng, T& srPrng)
  {    

    srPrng.nhit         = prng.NCell();
    srPrng.nhitx        = prng.NXCell();
    srPrng.nhity        = prng.NYCell();
    srPrng.nplane       = prng.ExtentPlane();
    srPrng.maxplanecont = prng.MostContiguousPlanes(geo::kXorY);
    srPrng.maxplanegap  = prng.MostMissingPlanes(geo::kXorY);
    srPrng.start        = prng.Start();
    srPrng.dir          = prng.Dir();
    srPrng.calE         = prng.CalorimetricEnergy();

    srPrng.len          = prng.TotalLength();

    if      (prng.NXCell() == 0)  srPrng.view = kY;
    else if (prng.NYCell() == 0)  srPrng.view = kX;
    else                           srPrng.view = kXorY;
    
  }  

  //......................................................................
  
  void CAFMaker::FillTrackVars(const rb::Track &trk, SRTrack& srTrk)
  {    

    FillProngVars(trk, srTrk);
    srTrk.stop         = trk.Stop();
    srTrk.lenE         = srTrk.len * 0.00157; // From N. Raddatz and K. Sachdev, see DocDB 7658-v2, slide 11 
                      // trkLength * MIP dE/dx
    
    
    
  }
  
  
  //......................................................................  
  
  std::vector<int> CAFMaker::getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                    std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu by slicer efficiency 
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use total slicer efficiency  
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use slicer efficiency 
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestNu = iNu;
          }
        }
      }
      if(bestNu >= 0 && bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }    
  std::vector<int> CAFMaker::getFaveIdsByPurity(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use slicer purity 
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].purity > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].purity;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use slicer purity
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].purity > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].purity;
            bestNu = iNu;
          }
        }
      }
      if(bestNu >= 0 && bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }
  
  //......................................................................  
  
  std::vector<int> CAFMaker::getFaveIdsByEnergy(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use total in slice energy 
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].energySlice > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].energySlice;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use total in slice energy 
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].energySlice > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].energySlice;
            bestNu = iNu;
          }
        }
      }
      if(bestNu >= 0 && bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }    
  
  //......................................................................  
  
  std::vector<int> CAFMaker::getFaveIdsByEfficiencyTimesPurity(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                               std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use the product of efficiency and purity  
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          float effPur = sliceTruthTable[iSlc][iNu].efficiency * sliceTruthTable[iSlc][iNu].purity;
          if(effPur > maxValue)
          {
            maxValue = effPur;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use total the product of efficiency and purity  
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          float effPur = sliceTruthTable[iSlc][iNu].efficiency * sliceTruthTable[iSlc][iNu].purity;
          if(effPur > maxValue)
          {
            maxValue = effPur;
            bestNu = iNu;
          }
        }
      }
      if(bestNu >= 0 && bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }    
  //......................................................................  
  
  std::vector<int> CAFMaker::getFaveIdsByEfficiencyThenPurity(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                              std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use slicer efficnecy  
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice by slicer purity 
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use total slicer purity  
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].purity > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].purity;
            bestNu = iNu;
          }
        }
      }
      if(bestNu >= 0 && bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }    
  
  //......................................................................  
  
  SliceInfo CAFMaker::getSliceInfo(const rb::Cluster& slice)
  {
    
    const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
    const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
    
    
    //Variables needed:
    //Holds contiguous planes info:
    int           planeNum     = -1;
    unsigned int  cellsFromEdge = -1;
    std::vector<int> planelist;
    
    //Hit counters:
    double nTotHit = 0.0;
    double nCalHit = 0.0;
    double nMipHit = 0.0;
    
    unsigned int maxPlane         = 0;
    unsigned int minPlane         = 99999999;
    unsigned int maxCell          = 0;
    unsigned int minCell          = 99999999;
    unsigned int minCellwall          = 99999999;
    unsigned int minCellsFromEdge = 99999999;
    
    //Loop over slice in question here:
    for(unsigned int hitIdx = 0; hitIdx < slice.NCell(); ++hitIdx)
    {
      //count up total number of cell hits
      nTotHit++;
      
      const art::Ptr<rb::CellHit>& chit = slice.Cell(hitIdx);
      //Get calbrated reco hit:
      const rb::RecoHit rhit(cal->MakeRecoHit(*chit,
					      chit->View() == geo::kX ? wx : wy));

      
      if(chit->Plane() < minPlane)
      {
        minPlane = chit->Plane();
      }
      if(chit->Plane() > maxPlane)
      {
        maxPlane = chit->Plane();
      }
      if(chit->Cell() < minCell)
      {
        minCell = chit->Cell();
      }
      if(chit->Cell() > maxCell)
      {
        maxCell = chit->Cell();
      }
      
      planeNum = chit->Plane();
      cellsFromEdge = std::min((unsigned int)chit->Cell(), geom->Plane(planeNum)->Ncells() - 1 - chit->Cell());
      
      if(cellsFromEdge < minCellsFromEdge)
      {
        minCellsFromEdge = cellsFromEdge;
      }
      
      //Fill a vector of plane numbers with hits; 
      if (std::find (planelist.begin(),planelist.end(),planeNum) == planelist.end()) {
        planelist.push_back(planeNum);
      }
      
      //Make sure things have been calibrated:
      if(!rhit.IsCalibrated())
      {
        LOG_DEBUG("CAFMaker") << "Not calibrated?! "
                              << chit->Plane() << " " << chit->Cell();
        continue;
      }
      
      //Count up both the total calibrated hits
      //and the MIP hits from 100-245 PECorr
      //Count up total hits:
      nCalHit++;
      //Count up MIP hits:
      if ( (rhit.PECorr()>100.0) && (rhit.PECorr()<245.0) ) nMipHit++;
      
    } //END OF SLICE LOOP HERE
    
    
    
    //calculate Contiguous Planes variable (maxContig) in the remaining code:
    int nContig=1;
    int maxContig=-1;
    
    
    
    //Calculate maximum contiguous planes
    if (planelist.size()>0) std::sort(planelist.begin(),planelist.end());
    for (unsigned int iplane=0; iplane<planelist.size(); iplane++)
    {
      
      if (iplane!=0)
      { 
        if (planelist[iplane] == (planelist[iplane-1]+1))
        {
          nContig++;
          if (nContig>maxContig) maxContig=nContig;
        } else {
          nContig=0;
        }
      }
    }
    
    if(maxContig < 0) maxContig = 0;
    
    SliceInfo slcInfo;
    slcInfo.nTotHit          = nTotHit;
    slcInfo.nCalHit          = nCalHit;
    slcInfo.nMipHit          = nMipHit;
    slcInfo.maxContig        = maxContig;
    slcInfo.maxPlane         = maxPlane;
    slcInfo.minPlane         = minPlane;
    slcInfo.maxCell          = maxCell;
    slcInfo.minCell          = minCell;
    slcInfo.minCellwall      = minCellwall;
    slcInfo.minCellsFromEdge = minCellsFromEdge;
    
    
    return slcInfo;
  }
  //......................................................................
  
  void CAFMaker::drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable, 
                                     const std::vector<int> &faveIds)
  {
    /// This function is here for debugging purposes.  
    /// It will draw a table of with slices as rows and neutrinos as columns.
    /// The favorites will be written with a star next to them.
    /// If two faveIds vectors are provided, the second will have an @ drawn next to them
    
    std::vector<int> otherFaveIds(faveIds.size(), -1);
    drawSliceTruthTable(sliceTruthTable, faveIds, otherFaveIds);
  }
  
  //......................................................................
  
  void CAFMaker::drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable, 
                                     const std::vector<int> &faveIds, const std::vector<int> &otherFaveIds)
  {
    /// This function is here for debugging purposes.  
    /// It will draw a table of with slices as rows and neutrinos as columns.
    /// The favorites will be written with a star next to them.
    /// If two faveIds vectors are provided, the second will have an @ drawn next to them
    
    std::cout.precision(2);    
    std::cout.width(2);
    std::cout.setf(std::ios::fixed,std::ios::floatfield); 
    int iSlc = 0;
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc, ++iSlc)
    {
      int iNu  = 0;
      for(auto nu = slc->begin(); nu != slc->end(); ++nu, ++iNu)
      {
        std::cout << nu->efficiency;
        if(faveIds[iSlc] == iNu)
          std::cout << "*"; 
        else
          std::cout << " "; 
        if(otherFaveIds[iSlc] == iNu)
          std::cout << "@"; 
        else
          std::cout << " "; 
        std::cout << "         " ;
        
      }
      
      std::cout << std::endl;
      
    }
    
  }
  
  //......................................................................
  
  void CAFMaker::produce(art::Event& evt) noexcept
  {
    feenableexcept(FE_INVALID);
    
    fTotalEvents += 1;
    fIsRealData = evt.isRealData();
 
    //get a handle on the slices
    art::Handle<std::vector<rb::Cluster> > hSlices;
    evt.getByLabel(clusterModName, hSlices);
    if(hSlices.failedToGet())
    {
      mf::LogError("CAFMaker") << "Slices not found in CAFMaker";   
    }
    
    // define a vector holding slices 
    art::PtrVector<rb::Cluster> subevts;
    for(unsigned int i = 0; i < hSlices->size(); ++i)
      subevts.push_back(art::Ptr<rb::Cluster>(hSlices, i));
    
    
    //Get all of the hits from calhits
    art::Handle< std::vector<rb::CellHit> > hitsHandle;
    evt.getByLabel(hitsModName, hitsHandle);
    art::PtrVector<rb::CellHit> allHits;
    for(size_t i = 0; i < hitsHandle->size(); ++i)
    {
      art::Ptr<rb::CellHit> hit(hitsHandle, i);
      allHits.push_back(hit);
    }

    // We're gonna get a vector of NeutrinoEffPurs for allHits
    std::vector<cheat::NeutrinoWithIndex> allNus = bt->allMCTruth();
    
    std::vector< std::vector<cheat::NeutrinoEffPur> > sliceTruthTable = bt->SlicesToMCTruthsTable(subevts);

  
  
    // Find favorite neutrino IDs using several metrics.  Total energy is the one we will use, for 
    // the favorite neutrino, but the others will be stored.
    std::vector<int> faveIdsEff         = getFaveIdsByEfficiency            (allNus, sliceTruthTable); 
    std::vector<int> faveIdsEnergy      = getFaveIdsByEnergy                (allNus, sliceTruthTable); 
    std::vector<int> faveIdsPur         = getFaveIdsByPurity                (allNus, sliceTruthTable); 
    std::vector<int> faveIdsEffPur      = getFaveIdsByEfficiencyTimesPurity (allNus, sliceTruthTable); 
    std::vector<int> faveIdsEffThenPur  = getFaveIdsByEfficiencyThenPurity  (allNus, sliceTruthTable); 

    //Get all truth track hits in the event
    std::vector<cheat::TrackIDE> allTracks;
    if(bt->HaveTruthInfo()) allTracks = bt->HitsToTrackIDE(allHits);
    
    

    
    //try to get pot spill info if this is real data
    art::Handle<sumdata::SpillData> spillPot;
    evt.getByLabel(numiBeamName, spillPot);
    //if we fail to identify a spill, we want it set to bad
    fIsgoodspill = false;
    fSpilltimesec = 0;
    fSpilltimensec = 0;
    fGpsspilltimesec = 0;
    fGpsspilltimensec = 0;
    fDeltaspilltimensec = 0;
    fSpillpot = 0.0;
    fHornI = 0.0;
    fPosx = 0.0;
    fPosy = 0.0;
    fWidthx = 0.0;
    fWidthy = 0.0;

    //need to flag each spill if the pot was good or bad
    //each slice will get the same tag
    //mc will flag each spill as good since mc does not make bad spills
    if (bt->HaveTruthInfo())
    {
      fIsgoodspill = true;
    }
    //if this is data try to get the spill info
    else {
      //if there was no spill info for this slice mark it bad. log message this should not happen
      if (spillPot.failedToGet())
      {
        mf::LogError("CAFMaker") << "Spill Data not found for real data event";   
        fIsgoodspill = false;
      }
      //if we have a SpillData object, get all the variables
      else {
        fIsgoodspill        = spillPot->goodbeam;
        fSpilltimesec       = spillPot->spilltimesec;
        fSpilltimensec      = spillPot->spilltimensec;
        fGpsspilltimesec    = spillPot->gpsspilltimensec;
        fGpsspilltimensec   = spillPot->gpsspilltimensec;
        fDeltaspilltimensec = spillPot->deltaspilltimensec;
        //NOTE: there is an implied factor of 1e12 when getting IFDBSpillInfo POT.  
        //Adding this back in to avoid confusion between data and mc
        fSpillpot           = (spillPot->spillpot)*(1e12);
        fHornI              = spillPot->hornI;
        fPosx               = spillPot->posx;
        fPosy               = spillPot->posy;
        fWidthx             = spillPot->widthx;
        fWidthy             = spillPot->widthy;
      }
    }
    
    //Add in run, subrun, evt information so that the spill tree can be 
    //mapped to the record tree.
    fRun    = evt.run();
    fSubrun = evt.subRun();
    fEvt    = evt.id().event();
    //Fill the spill tree once per event so that POT counting can be done properly.
    //This is tree will have an entry even for empty events with no physics slices.
    spillTree->Fill();
    
    
    
    
    // Get all association finder objects
    art::FindManyP<rb::Track>               fmDiscrete2d(hSlices, evt, discreteTrack2dName);
    art::FindManyP<rb::Track>               fmDiscrete(hSlices, evt, discreteTrackName);
    art::FindManyP<rb::Track>               fmKalman(hSlices, evt, kalmanTrackName);
    art::FindManyP<rb::Track>               fmCosmic(hSlices, evt, cosmicTrackName);
    art::FindManyP<rb::Vertex>              fmElastic(hSlices, evt, elasticArmsName);
    art::FindManyP<me::SliceMEFCluster>     fmSliceMEF(hSlices, evt, sliceMEFName);
    art::FindManyP<jmshower::JMShower>      fmJmShower(hSlices, evt, jmShowerName);
    art::FindManyP<jmshower::EID>           fmNueSel(hSlices, evt, nueSelName);
    art::FindOneP<slid::EventLID>           foLID(hSlices, evt, lidName); //one lid object per slice
    art::FindManyP<slid::ShowerLID>         fmShwLID(hSlices, evt, lidName);
    art::FindManyP<numue::NumuE>            fmNumuEnergy(hSlices, evt, numuEnergyName);
    art::FindManyP<presel::Veto>            fmVeto(hSlices,evt, vetoName);
    art::FindManyP<qeef::QePId>             fmQepid(hSlices, evt, qepidName);
    art::FindManyP<nuesand::NueSandObj>     fmNueSand(hSlices, evt, nueSandboxName);
    art::FindManyP<numusand::NumuSandObj>   fmNumuSand(hSlices, evt, numuSandboxName);
    art::FindManyP<lem::PIDDetails>         fmLem(hSlices, evt, lemName);
    art::FindManyP<rvp::RVP>                fmRvp(hSlices, evt, rvpName);
    art::FindManyP<cosrej::CosRejObj>       fmCosRej(hSlices,evt,cosrejName);
    art::FindManyP<cosrej::NueCosRej>       fmNueCosRej(hSlices,evt,nuecosrejName);
    art::FindManyP<murem::MRCCParent>       fmMrccPar(hSlices, evt, mrccParentName);

    int nNonNoise = 0;
    for(unsigned int subevtId = 0; subevtId < subevts.size(); ++subevtId )
    {
      if(subevts[subevtId]->IsNoise() || subevts[subevtId]->NCell() == 0)  
        continue; 
      ++nNonNoise;
    }    
    for(unsigned int subevtId = 0; subevtId < subevts.size(); ++subevtId )
    {
      if(subevts[subevtId]->IsNoise() || subevts[subevtId]->NCell() == 0)  continue; 
      // because we don't care about the noise slice and slices with no hits.
      
      rec = new StandardRecord();

      //#######################################################
      // Fill header.
      //#######################################################
      // Get metadata information for header
      unsigned int run = evt.run();
      unsigned int subrun = evt.subRun();
      unsigned int spillNum = evt.id().event();
      bool isMC = bt->HaveTruthInfo();

      art::Timestamp ts = evt.time();
      const time_t timeSec = ts.timeHigh();

      struct tm* timeStruct = localtime(&timeSec);

      unsigned int maskstatus = 0;
      unsigned int dibmask = (rh->GoodDiBlockMask()&rh->GetConfiguration());
      if(dibmask == ((1<<16)-1)) {
        dibmask = rh->GetConfiguration();
        maskstatus = 0;
      }
      if(dibmask == 0) {
        dibmask = rh->GetConfiguration();
        maskstatus = 2;
      }
      else maskstatus = 1;

      unsigned int dibfirst = 0;
      unsigned int diblast = 0;
      if (dibmask) {
	int iD;
	iD=0; while (!((dibmask>>iD)&1)) iD++; dibfirst=iD+1;
	iD=0; while (dibmask>>iD)        iD++; diblast=iD;
      }

      rec->hdr = SRHeader();

      rec->hdr.run             = run;
      rec->hdr.subrun          = subrun;
      rec->hdr.evt             = spillNum;
      rec->hdr.subevt          = subevtId;
      rec->hdr.ismc            = isMC;
      rec->hdr.det             = fDet;
      rec->hdr.blind           = 0;

      rec->hdr.subevtstarttime = subevts[subevtId]->MinTNS();
      rec->hdr.subevtendtime   = subevts[subevtId]->MaxTNS();
      rec->hdr.subevtmeantime  = subevts[subevtId]->MeanTNS();
      rec->hdr.unixtime        = timeSec;

      rec->hdr.dibfirst        = dibfirst;
      rec->hdr.diblast         = diblast;
      rec->hdr.dibmask         = dibmask;
      rec->hdr.maskstatus      = maskstatus;

      rec->hdr.year            = timeStruct->tm_year + 1900;  
      rec->hdr.month           = timeStruct->tm_mon + 1;  
      rec->hdr.day             = timeStruct->tm_mday;  
      rec->hdr.doy             = timeStruct->tm_yday + 1;  
      rec->hdr.hour            = timeStruct->tm_hour + 1;  
      rec->hdr.minute          = timeStruct->tm_min;  
      rec->hdr.second          = timeStruct->tm_sec;  
      
      //#######################################################
      // Fill Spill info.
      //#######################################################
      //Store info on NuMI beam quality and if the pot as good or bad.
      //This is meant to be used to refine a good spill, not for
      //pot counting since events have multiple slices and empty
      //spill are missing from this tree.  Use the friend tree for
      //pot counting.
      rec->spill.isgoodspill        = fIsgoodspill;
      rec->spill.spilltimesec       = fSpilltimesec;
      rec->spill.spilltimensec      = fSpilltimensec;
      rec->spill.gpsspilltimesec    = fGpsspilltimesec;
      rec->spill.gpsspilltimensec   = fGpsspilltimensec;
      rec->spill.deltaspilltimensec = fDeltaspilltimensec;
      rec->spill.spillpot           = fSpillpot;
      rec->spill.hornI              = fHornI;
      rec->spill.posx               = fPosx;
      rec->spill.posy               = fPosy;
      rec->spill.widthx             = fWidthx;
      rec->spill.widthy             = fWidthy;
      
      //#######################################################
      // Add slice info.
      //#######################################################
      SliceInfo slcInfo         = getSliceInfo(*subevts[subevtId]);
      rec->slc.nhit             = slcInfo.nTotHit;
      rec->slc.ncalhit          = slcInfo.nCalHit;
      rec->slc.nmiphit          = slcInfo.nMipHit;
      rec->slc.ncontplanes      = slcInfo.maxContig;
      rec->slc.firstplane       = slcInfo.minPlane;
      rec->slc.lastplane        = slcInfo.maxPlane;
      rec->slc.firstcell        = slcInfo.minCell;
      rec->slc.lastcell         = slcInfo.maxCell;
      rec->slc.ncellsfromedge   = slcInfo.minCellsFromEdge;
      rec->slc.nnonnoise        = nNonNoise;
      
      rec->slc.calE             = subevts[subevtId]->CalorimetricEnergy();
      rec->slc.starttime        = subevts[subevtId]->MinTNS();
      rec->slc.endtime          = subevts[subevtId]->MaxTNS();
      rec->slc.meantime         = subevts[subevtId]->MeanTNS();
      

      subevts[subevtId]->MinMaxMeanXYZ(rec->slc.boxmin, rec->slc.boxmax, rec->slc.meanpos);

      //#######################################################
      // Adding reconstructed objects.
      //#######################################################
      
      // Get all the tracks associated with this slice from discrete tracker.  
      //This depends on the findMany object created above.
      if(fmDiscrete.isValid())
      {      
        std::vector< art::Ptr<rb::Track> > discTracks = fmDiscrete.at(subevtId);  
        // Sort the tracks in descending length order 
        std::sort(discTracks.begin(), discTracks.end(), sortRBTrackLength);
        art::FindManyP<me::MichelECluster> fmDiscreteMichelE(discTracks, evt, discreteMichelEName);
        rec->trk.cosmic.reserve(discTracks.size());
        rec->trk.cosmic.resize(discTracks.size());
        for(size_t trackId = 0; trackId < discTracks.size(); ++trackId)
        {
          FillTrackVars((*discTracks[trackId]), rec->trk.cosmic[trackId]);
          if(fmDiscreteMichelE.isValid())
          {
            //Get MichelE from filter
            std::vector< art::Ptr<me::MichelECluster> > michels = fmDiscreteMichelE.at(trackId);
            
            for(auto itMichel = michels.begin(); itMichel != michels.end(); itMichel++)
            {
	      AddTrkMEClusterToVec(*itMichel, &rec->me.trkdiscrete, trackId);
            }
          }
        }
      }
      if(fmDiscrete2d.isValid())
      {      
        std::vector< art::Ptr<rb::Track> > discTracks = fmDiscrete2d.at(subevtId);
        // Sort the tracks in descending length order 
        std::sort(discTracks.begin(), discTracks.end(), sortRBTrackLength);
        rec->trk.cosmic.reserve(discTracks.size());
        rec->trk.cosmic.resize(discTracks.size());
        for(size_t trackId = 0; trackId < discTracks.size(); ++trackId)
        {
          FillTrackVars((*discTracks[trackId]), rec->trk.cosmic[trackId]);
        }
      }
      
      if(fmCosmic.isValid())
      {
        std::vector<art::Ptr<rb::Track>> cosTracks = fmCosmic.at(subevtId);
        // Sort the tracks in descending length order 
        std::sort(cosTracks.begin(), cosTracks.end(), sortRBTrackLength);

        art::FindManyP<me::MichelECluster> fmCosmicMichelE(cosTracks, evt, cosmicMichelEName);
        rec->trk.cosmic.reserve(cosTracks.size());
        rec->trk.cosmic.resize(cosTracks.size());
        for(size_t trackId = 0; trackId < cosTracks.size(); ++trackId)
        {
          FillTrackVars((*cosTracks[trackId]), rec->trk.cosmic[trackId]);
          if(fmCosmicMichelE.isValid())
          {
            //Get MichelE from filter
            std::vector< art::Ptr<me::MichelECluster> > michels = fmCosmicMichelE.at(trackId);
            
            for(auto itMichel = michels.begin(); itMichel != michels.end(); itMichel++)
            {
	      AddTrkMEClusterToVec(*itMichel, &rec->me.trkcosmic, trackId);
            }
            
          }
          
        }
      }
      
      // now do the same for kalmanTrack
      if(fmKalman.isValid())
      {
        std::vector< art::Ptr<rb::Track> > kalTracks = fmKalman.at(subevtId);
        // Sort the tracks in descending length order 
        std::sort(kalTracks.begin(), kalTracks.end(), sortRBTrackLength);
        art::FindManyP<remid::ReMId> fmRemid(kalTracks, evt, remidName);
        art::FindManyP<rb::Energy> fmMuTrkE(kalTracks, evt, numuEnergyName);
        art::FindManyP<me::MichelECluster> fmKalmanMichelE(kalTracks, evt, kalmanMichelEName);

        size_t bestPidIdx = 999;
        size_t bestPidIdx3d = 999;
        double bestPid = 0; 
        double bestPid3d = 0; 
	bool foundBestPid = false;
        
        rec->trk.kalman.reserve(kalTracks.size());
        rec->trk.kalman.resize(kalTracks.size());
        for(size_t trackId = 0; trackId < kalTracks.size(); ++trackId)
        {
          FillTrackVars((*kalTracks[trackId]), rec->trk.kalman[trackId]);
          
          if(fmRemid.isValid())
          {
            //Get MuID from ReMId
            std::vector< art::Ptr<remid::ReMId> > remids = fmRemid.at(trackId);
            
            if(remids.size() > 0)
            {
              rec->sel.remid.muid.push_back(SRMuId());
              std::vector<SRMuId>::reverse_iterator srMuId = rec->sel.remid.muid.rbegin();
              
              srMuId->pdg     = remids[0]->Pdg();
              srMuId->pid     = remids[0]->Value();
              srMuId->scatllh = remids[0]->ScatSeparation();
              srMuId->dedxllh = remids[0]->DedxSeparation();
              srMuId->len     = kalTracks[trackId]->TotalLength();
	      srMuId->measfrac = -5;
              if(remids[0]->NDedx() > 0) {
                srMuId->measfrac = ((float)remids[0]->NMeasurementPlanes()/(float)remids[0]->NDedx());
              }
              
              srMuId->cont    = remids[0]->Contained();
              
              if (srMuId->pid >= bestPid)
	      {
                bestPidIdx = trackId;
                bestPid    = srMuId->pid;
		foundBestPid = true;
              }
              if (srMuId->pid >= bestPid3d && kalTracks[trackId]->View() == geo::kXorY)
              {
                bestPidIdx3d = trackId;
                bestPid3d    = srMuId->pid;
              }
            }
	    else
	    {
	      throw cet::exception("Missing Association Error", "Failed to find a ReMId object with this track.");
	    }
          }
          
          
          if(fmMuTrkE.isValid())
          {
            //Get energy for track assuming it is a muon
            std::vector< art::Ptr<rb::Energy> > mutrkEs = fmMuTrkE.at(trackId);
            
            if(mutrkEs.size() > 0)
            {
              rec->energy.mutrkE.push_back(SREnergy());
              std::vector<SREnergy>::reverse_iterator srE = rec->energy.mutrkE.rbegin();
              srE->E     = mutrkEs[0]->E();
            }
	    else
	    {
	      throw cet::exception("Missing Association Error", "Failed to find a muon track energy object with this track.");
	    }
          }
          
          if(fmKalmanMichelE.isValid())
          {
            //Get MichelE from filter
            std::vector< art::Ptr<me::MichelECluster> > michels = fmKalmanMichelE.at(trackId);
            
            for(auto itMichel = michels.begin(); itMichel != michels.end(); itMichel++)
            {
	      AddTrkMEClusterToVec(*itMichel, &rec->me.trkkalman, trackId);
            }
              
          }  
        }

        //Give SRRemid the pid values of track with largest PID 
        if(rec->sel.remid.muid.size() > 0 && foundBestPid)
        {
          rec->sel.remid.pdg          = rec->sel.remid.muid[bestPidIdx].pdg;    
          rec->sel.remid.pid          = rec->sel.remid.muid[bestPidIdx].pid;    
          rec->sel.remid.scatllh      = rec->sel.remid.muid[bestPidIdx].scatllh;
          rec->sel.remid.dedxllh      = rec->sel.remid.muid[bestPidIdx].dedxllh;
          rec->sel.remid.len          = rec->sel.remid.muid[bestPidIdx].len;   
          rec->sel.remid.bestidx      = bestPidIdx;   
          rec->sel.remid.bestidx3d    = bestPidIdx3d;   
          
	  rec->sel.remid.measfrac = rec->sel.remid.muid[bestPidIdx].measfrac;
	  rec->sel.remid.cont    = rec->sel.remid.muid[bestPidIdx].cont;
        }
        else{
          rec->sel.remid.setDefault();
        }
        
        
      }else{
        rec->sel.remid.setDefault();
      }
      rec->trk.fillSizes();
      rec->sel.remid.fillSizes();
      
      
      //  Get all vertices associated with slice from FuzzyKVertex and ElasticArms
      if(fmElastic.isValid())
      {
        std::vector< art::Ptr<rb::Vertex> > elastics = fmElastic.at(subevtId);
        art::FindManyP<rb::Prong> fmFuzzyProng3D(elastics, evt, art::InputTag(fuzzyVertexName, fuzzy3DAssnName));
        art::FindManyP<rb::Prong> fmFuzzyProng2D(elastics, evt, art::InputTag(fuzzyVertexName, fuzzy2DAssnName));
        
        for(unsigned int vtxId = 0; vtxId < elastics.size(); ++vtxId)
        {
          rec->vtx.elastic.push_back(SRElastic());
          std::vector<SRElastic>::reverse_iterator srVtx = rec->vtx.elastic.rbegin();
          
          srVtx->time = elastics[vtxId]->GetT();
          srVtx->vtx = elastics[vtxId]->GetXYZ();
          if(fmFuzzyProng3D.isValid())
          {          
            
            std::vector< art::Ptr<rb::Prong> > prongs = fmFuzzyProng3D.at(vtxId);
            //short prongs in descending length order
            std::sort(prongs.begin(), prongs.end(),sortRBProngEnergy);
            for(unsigned int shwID = 0; shwID < prongs.size(); ++shwID)
            {
              if ((prongs[shwID]->NXCell()==0)||prongs[shwID]->NYCell()==0){
                srVtx->png2D.push_back(SRProng());
                FillProngVars((*prongs[shwID]), srVtx->png2D[srVtx->png2D.size() - 1]);
              }//end 2d prong case
              else{
                srVtx->png3D.push_back(SRProng());
                FillProngVars((*prongs[shwID]), srVtx->png3D[srVtx->png3D.size() - 1]);
              }//end 3d prong case

            }//end loop over prongs
            
          }//end if prongs are valid for this vertex
          //check if there is a valid instance label for 2D prongs
          if(fmFuzzyProng2D.isValid() && fuzzy2DAssnName != "")
          {          
            
            std::vector< art::Ptr<rb::Prong> > prongs = fmFuzzyProng2D.at(vtxId);
            //short prongs in descending length order
            std::sort(prongs.begin(), prongs.end(),sortRBProngEnergy);
            for(unsigned int shwID = 0; shwID < prongs.size(); ++shwID)
            {
              if ((prongs[shwID]->NXCell()==0)||prongs[shwID]->NYCell()==0){
                srVtx->png2D.push_back(SRProng());
                FillProngVars((*prongs[shwID]), srVtx->png2D[srVtx->png2D.size() - 1]);
              }//end 2d prong case
              else{
                srVtx->png3D.push_back(SRProng());
                FillProngVars((*prongs[shwID]), srVtx->png3D[srVtx->png3D.size() - 1]);
              }//end 3d prong case

            }//end loop over prongs
            
          }//end if prongs are valid for this vertex
          
        }
        //sort the vector
        std::sort(rec->vtx.elastic.begin(), rec->vtx.elastic.end(), sortElasticProng);
      }
      rec->vtx.fillSizes();

      if(fmSliceMEF.isValid())
      {
	std::vector< art::Ptr<me::SliceMEFCluster> > michels = fmSliceMEF.at(subevtId);
	for(auto itMichel = michels.begin(); itMichel != michels.end(); itMichel++)
        {
	  AddSlcMEClusterToVec(*itMichel, &rec->me.slc);
	}
      }
      rec->me.fillSizes();
      
      //#######################################################
      // Add Nue Presel info from slice and reconstructed objects
      //#######################################################

      //Nue presel vars:
      rec->sel.nuepre.calE                 =  rec->slc.calE;
      rec->sel.nuepre.ncell                =  slcInfo.nCalHit;
      rec->sel.nuepre.fracmip              =  (slcInfo.nCalHit>0) ? slcInfo.nMipHit/slcInfo.nCalHit : -1;
      rec->sel.nuepre.lengthdiscrete2d     =  (rec->trk.discrete2d.size()>0) ? rec->trk.discrete2d[0].len : 0;
      rec->sel.nuepre.lengthdiscretemerge  =  (rec->trk.discrete.size()>0) ? rec->trk.discrete[0].len : 0;
      rec->sel.nuepre.lengthkalman         =  (rec->trk.kalman.size()>0) ? rec->trk.kalman[0].len : 0;
      
      //Fiducial and containment vars:
      rec->sel.nuepre.cellstoedge     =  slcInfo.minCellsFromEdge;
      rec->sel.nuepre.planestofront   =  slcInfo.minPlane;
      rec->sel.nuepre.planestoback    =  geom->NPlanes() - 1 - slcInfo.maxPlane;
      rec->sel.nuepre.mean = subevts[subevtId]->MeanXYZ();
      if (rec->vtx.elastic.size()>0)
      {
        rec->sel.nuepre.vtx  = rec->vtx.elastic[0].vtx;
      } else { rec->sel.nuepre.vtx.SetXYZ(-99999,-99999,-99999); }
      
      
      //Booleans describing preselection:
      
      //Current fiducial cut:
      rec->sel.nuepre.passfid = 
      geom->isInsideFiducialVolume(rec->sel.nuepre.mean.X(),rec->sel.nuepre.mean.Y(),rec->sel.nuepre.mean.Z());
      
      //Current preselection cut:
      if (rec->sel.nuepre.ncell > 35 && rec->sel.nuepre.ncell <= 135 
          && rec->sel.nuepre.calE > 1.00 && rec->sel.nuepre.calE <= 3.00
          && rec->sel.nuepre.lengthdiscrete2d > 45 && rec->sel.nuepre.lengthdiscrete2d <= 385
          && rec->sel.nuepre.fracmip > 0.15 && rec->sel.nuepre.fracmip <= 0.65
          ) { rec->sel.nuepre.passpre = true; 
      } else { rec->sel.nuepre.passpre = false; }
      
      //####End Presel Fill####//
      
      //  Get all showers associated with slice from JMShower
      if(fmJmShower.isValid())
      {      
        std::vector< art::Ptr<jmshower::JMShower> > showers = fmJmShower.at(subevtId);
        for(unsigned int shwId = 0; shwId < showers.size(); ++shwId)
        {
          
          rec->sand.nue.jm.push_back(SRJMShower());
          std::vector<SRJMShower>::reverse_iterator srShw = rec->sand.nue.jm.rbegin();
          
          
          srShw->nhit  = showers[shwId]->NCell();
          srShw->len   = showers[shwId]->TotalLength();
          srShw->width = showers[shwId]->Radius();
          srShw->calE  = showers[shwId]->CalorimetricEnergy();
          srShw->vtxE  = showers[shwId]->VtxGeV();
          srShw->shwE  = showers[shwId]->Energy();
          srShw->start = showers[shwId]->Start();
          srShw->dir   = showers[shwId]->Dir();
          srShw->stop  = showers[shwId]->Stop();
          
          if      (showers[shwId]->NXCell() == 0)  srShw->view = kY;
          else if (showers[shwId]->NYCell() == 0)  srShw->view = kX;
          else                                     srShw->view = kXorY;
          
          
        }
        
        //sort the vector
        std::sort(rec->sand.nue.jm.begin(), rec->sand.nue.jm.end(), sortShowerEnergy);
        
      }
      
      rec->sand.nue.njm = rec->sand.nue.jm.size();
      
      
      // Get the JM EID objects and add them.
      if(fmNueSel.isValid())
      {
        std::vector< art::Ptr<jmshower::EID> > eids = fmNueSel.at(subevtId);
        if(eids.size() > 0)
        {
          rec->sand.nue.jmid.ann       = eids[0]->ANN();
          rec->sand.nue.jmid.annE      = eids[0]->ANNE();
          rec->sand.nue.jmid.annNoCos  = eids[0]->ANNNoCos();
          rec->sand.nue.jmid.annENoCos = eids[0]->ANNENoCos();
          rec->sand.nue.jmid.ismuon   = eids[0]->IsMuon();
          rec->sand.nue.jmid.eglll    = eids[0]->EGLLL();
          rec->sand.nue.jmid.egllt    = eids[0]->EGLLT();
          rec->sand.nue.jmid.emulll   = eids[0]->EMuLLL();
          rec->sand.nue.jmid.emullt   = eids[0]->EMuLLT();
          rec->sand.nue.jmid.epi0lll  = eids[0]->EPi0LLL();
          rec->sand.nue.jmid.epi0llt  = eids[0]->EPi0LLT();
          rec->sand.nue.jmid.eplll    = eids[0]->EPLLL();
          rec->sand.nue.jmid.epllt    = eids[0]->EPLLT();
          rec->sand.nue.jmid.enlll    = eids[0]->ENLLL();
          rec->sand.nue.jmid.enllt    = eids[0]->ENLLT();
          rec->sand.nue.jmid.epilll   = eids[0]->EPiLLL();
          rec->sand.nue.jmid.epillt   = eids[0]->EPiLLT();
          rec->sand.nue.jmid.vtxgev   = eids[0]->VtxGeV();
          rec->sand.nue.jmid.pi0mass  = eids[0]->Pi0Mass();
          rec->sand.nue.jmid.shwEFrac = eids[0]->ShE();
          rec->sand.nue.jmid.gap      = eids[0]->Gap();
          rec->sand.nue.jmid.elll     = eids[0]->ELLL();
          rec->sand.nue.jmid.ellt     = eids[0]->ELLT();
          rec->sand.nue.jmid.mulll    = eids[0]->MuLLL();
          rec->sand.nue.jmid.mullt    = eids[0]->MuLLT();

          rec->sand.nue.vtx      = eids[0]->Vtx();
          rec->sand.nue.time     = subevts[subevtId]->MeanTNS();

          rec->sand.nue.E     = eids[0]->NuEnergy();
          rec->sand.nue.depE  = eids[0]->DetEnergy();
          rec->sand.nue.shwE  = eids[0]->Sh1Energy();

	  rec->sand.nue.sh1totalL   = eids[0]->Sh1TotalLength();
	  rec->sand.nue.sh1start    = eids[0]->Sh1Start();
	  rec->sand.nue.sh1stop     = eids[0]->Sh1Stop();
	  rec->sand.nue.sh1dir 	    = eids[0]->Sh1Dir();

          rec->sand.nue.evtncell = eids[0]->EvtNCell();
          rec->sand.nue.evtsumcostheta = eids[0]->EvtSumCosTheta(); 
          rec->sand.nue.evtsump = eids[0]->EvtSumP();  
          rec->sand.nue.evtsumpt = eids[0]->EvtSumPt();
          rec->sand.nue.evtsump0 = eids[0]->EvtSumP0();
          rec->sand.nue.evtsumpt0 = eids[0]->EvtSumPt0();
          rec->sand.nue.evtetot = eids[0]->EvtEtot();  
          rec->sand.nue.evtminx = eids[0]->EvtMinX();  
          rec->sand.nue.evtminy = eids[0]->EvtMinY();  
          rec->sand.nue.evtminz = eids[0]->EvtMinZ();  
          rec->sand.nue.evtmaxx = eids[0]->EvtMaxX();
          rec->sand.nue.evtmaxy = eids[0]->EvtMaxY();
          rec->sand.nue.evtmaxz = eids[0]->EvtMaxZ();


          rec->sand.nue.evtgaptns = eids[0]->EvtGapTNS();  
          rec->sand.nue.evtncelltoedge = eids[0]->EvtNCellToEdge();  

          rec->sand.nue.sh1energy = eids[0]->Sh1Energy();     
          rec->sand.nue.sh1exclenergy = eids[0]->Sh1ExclEnergy(); 
          rec->sand.nue.sh1totalL = eids[0]->Sh1TotalLength(); 
          rec->sand.nue.sh1start = eids[0]->Sh1Start(); 
          rec->sand.nue.sh1stop = eids[0]->Sh1Stop();
	  rec->sand.nue.sh1dir = eids[0]->Sh1Dir(); 
          rec->sand.nue.sh1vtxdoca = eids[0]->Sh1VtxDoca(); 
          rec->sand.nue.sh1gap = eids[0]->Sh1Gap();
          rec->sand.nue.sh1nplane = eids[0]->Sh1NPlane(); 
          rec->sand.nue.sh1xnplane = eids[0]->Sh1XNPlane();
          rec->sand.nue.sh1ynplane = eids[0]->Sh1YNPlane();
          rec->sand.nue.sh1ncell = eids[0]->Sh1NCell();
          rec->sand.nue.sh1xncell = eids[0]->Sh1XNCell();
          rec->sand.nue.sh1yncell = eids[0]->Sh1YNCell();
          rec->sand.nue.sh1pid = eids[0]->Sh1PID();
          rec->sand.nue.sh1ncelltoedge = eids[0]->Sh1NCellToEdge();
          rec->sand.nue.sh1sh2dang = eids[0]->Sh1Sh2Dang();

          rec->sand.nue.sh2energy = eids[0]->Sh2Energy();
          rec->sand.nue.sh2exclenergy = eids[0]->Sh2ExclEnergy();
          rec->sand.nue.sh2totalL = eids[0]->Sh2TotalLength();
          rec->sand.nue.sh2start = eids[0]->Sh2Start(); 
          rec->sand.nue.sh2stop = eids[0]->Sh2Stop();
	  rec->sand.nue.sh2dir = eids[0]->Sh2Dir();
          rec->sand.nue.sh2vtxdoca = eids[0]->Sh2VtxDoca();
          rec->sand.nue.sh2gap = eids[0]->Sh2Gap();
          rec->sand.nue.sh2nplane = eids[0]->Sh2NPlane();
          rec->sand.nue.sh2xnplane = eids[0]->Sh2XNPlane();
          rec->sand.nue.sh2ynplane = eids[0]->Sh2YNPlane();
          rec->sand.nue.sh2ncell = eids[0]->Sh2NCell();
          rec->sand.nue.sh2xncell = eids[0]->Sh2XNCell();
          rec->sand.nue.sh2yncell = eids[0]->Sh2YNCell();
          rec->sand.nue.sh2pid = eids[0]->Sh2PID();
          rec->sand.nue.sh2ncelltoedge = eids[0]->Sh2NCellToEdge();

	  rec->sand.nue.elll        = eids[0]->ELLL();
	  rec->sand.nue.sh1d2edge   = eids[0]->Sh1DistToEdge();
	  rec->sand.nue.eedge2cell  = eids[0]->EEdge2Cell();
	  rec->sand.nue.eedge5cell  = eids[0]->EEdge5Cell();
	  rec->sand.nue.eedge10cell = eids[0]->EEdge10Cell();

          // nue containment boolean - nue group please feel free to update.  If variables move from sandbox this needs to be changed too
          if(std::min(eids[0]->Sh1Start().X(),eids[0]->Sh1Stop().X())>-750 && std::max(eids[0]->Sh1Start().X(),eids[0]->Sh1Stop().X())<750 &&
             std::min(eids[0]->Sh1Start().Y(),eids[0]->Sh1Stop().Y())>-750 && std::max(eids[0]->Sh1Start().Y(),eids[0]->Sh1Stop().Y())<750 &&
             std::min(eids[0]->Sh1Start().Z(),eids[0]->Sh1Stop().Z()) > 50 &&  std::max(eids[0]->Sh1Start().Z(),eids[0]->Sh1Stop().Z()) < 5900 &&
             (eids[0]->EvtSumPt()/eids[0]->EvtSumP()) < 0.6) rec->sel.contain.nuecontain = true;

        }
        else{
          rec->sand.nue.jmid    .setDefault();
        }
        
      }
      else{
      if(fmNueSel.isValid())
        rec->sand.nue.jmid    .setDefault();
      }

      //  Get slice level PID from LIDBuilder
      
      if(foLID.isValid())
      {
        cet::maybe_ref<art::Ptr<slid::EventLID> const> rlid(foLID.at(subevtId));
        art::Ptr<slid::EventLID> elid = rlid.ref();
        //if the pointer is valid
        if (elid){
          rec->sel.elecid.ann      = elid->Value();
          rec->sel.elecid.anne     = elid->ValueWithE();
          if(fmShwLID.isValid())
          {      
            std::vector< art::Ptr<slid::ShowerLID> > shwlids = fmShwLID.at(subevtId);
            art::FindOneP<rb::Shower> fos(shwlids, evt, lidName);
            for(unsigned int shwId = 0; shwId < shwlids.size(); ++shwId)
            {
              cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(shwId));
              art::Ptr<rb::Shower> shw = rshw.ref();
              if (shw){
                rec->shw.shwlid.push_back(SRShower());
                std::vector<SRShower>::reverse_iterator srShw = rec->shw.shwlid.rbegin();
            
                srShw->nhit         = shw->NCell();
                srShw->nhitx        = shw->NXCell();
                srShw->nhity        = shw->NYCell();
                srShw->nplane       = shw->ExtentPlane();
                srShw->nplanex      = shw->ExtentPlane(geo::kX);
                srShw->nplaney      = shw->ExtentPlane(geo::kY);
                srShw->gap          = shwlids[shwId]->Gap();
                srShw->maxplanecont = shw->MostContiguousPlanes(geo::kXorY);
                srShw->maxplanegap  = shw->MostMissingPlanes(geo::kXorY);
                srShw->len          = shw->TotalLength();
                srShw->width        = shwlids[shwId]->Radius();
                srShw->calE         = shw->CalorimetricEnergy();
                srShw->start        = shw->Start();
                srShw->dir          = shw->Dir();
                srShw->stop         = shw->Stop();
                if      (shw->NXCell() == 0)  srShw->view = kY;
                else if (shw->NYCell() == 0)  srShw->view = kX;
                else                                     srShw->view = kXorY;
              }

              rec->sel.elecid.shwlid.push_back(SRSLid());
              std::vector<SRSLid>::reverse_iterator slid = rec->sel.elecid.shwlid.rbegin();

              std::map<int, float> tPartLongLL = shwlids[shwId]->fPartLongLL;
              std::map<int, float> tPartTransLL = shwlids[shwId]->fPartTransLL;
  
              slid->ann      = shwlids[shwId]->Value();
              slid->anne     = shwlids[shwId]->ValueWithE();
              slid->ismuon   = shwlids[shwId]->IsMuon();
              slid->eglll    = shwlids[shwId]->EGLLL();
              slid->egllt    = shwlids[shwId]->EGLLT();
              slid->emulll   = shwlids[shwId]->EMuLLL();
              slid->emullt   = shwlids[shwId]->EMuLLT();
              slid->epi0lll  = shwlids[shwId]->EPi0LLL();
              slid->epi0llt  = shwlids[shwId]->EPi0LLT();
              slid->eplll    = shwlids[shwId]->EPLLL();
              slid->epllt    = shwlids[shwId]->EPLLT();
              slid->enlll    = shwlids[shwId]->ENLLL();
              slid->enllt    = shwlids[shwId]->ENLLT();
              slid->epilll   = shwlids[shwId]->EPiLLL();
              slid->epillt   = shwlids[shwId]->EPiLLT();
              slid->vtxgev   = shwlids[shwId]->VertexEnergy();
              slid->pi0mass  = shwlids[shwId]->Pi0mass();
              slid->shwEFrac = shwlids[shwId]->ShowerEFrac();
              slid->gap      = shwlids[shwId]->Gap();
              slid->costheta = shwlids[shwId]->CosTheta();

              std::map<int, float>::iterator itr = tPartLongLL.find(slid::DedxParticleType::kELECTRON);
              if (itr != tPartLongLL.end()) slid->elll = itr->second;
              else slid->elll = -5;
              itr = tPartTransLL.find(slid::DedxParticleType::kELECTRON);
              if (itr != tPartTransLL.end()) slid->ellt = itr->second;
              else slid->ellt = -5;
              itr = tPartLongLL.find(slid::DedxParticleType::kMUON);
              if (itr != tPartLongLL.end()) slid->mulll = itr->second;
              else slid->mulll = -5;
              itr = tPartTransLL.find(slid::DedxParticleType::kMUON);
              if (itr != tPartTransLL.end()) slid->mullt = itr->second;
              else slid->mullt = -5;

              slid->invglll  = shwlids[shwId]->InverseGLLL();

              //fill energy information for leading energy shower which also servers a slice pid
              if (shwId == 0){
                rec->energy.elecid.E    = shwlids[shwId]->NueEnergy();
                rec->energy.elecid.depE = shwlids[shwId]->ShowerDepEnergy();
                rec->energy.elecid.shwE = shwlids[shwId]->ShowerEnergy();
                rec->energy.elecid.hadE = shwlids[shwId]->HadronicEnergy();
              }

              //fill energy vector with the other shower energies
              rec->energy.shwlid.push_back(SRSLidEnergy());
              std::vector<SRSLidEnergy>::reverse_iterator senergy = rec->energy.shwlid.rbegin();
              senergy->E    = shwlids[shwId]->NueEnergy();
              senergy->depE = shwlids[shwId]->ShowerDepEnergy();
              senergy->shwE = shwlids[shwId]->ShowerEnergy();
              senergy->hadE = shwlids[shwId]->HadronicEnergy();
            } //end for loop on showerlids
          } //end if showerlids are valid
        } //end if event id is valid
        else{
          rec->sel.elecid    .setDefault();
          rec->energy.elecid .setDefault();
        }
      } //end if slice lids are valid
      else{
        rec->sel.elecid    .setDefault();
        rec->energy.elecid .setDefault();
      }
      
      rec->shw.fillSizes();
      rec->energy.fillSizes();
      rec->sel.elecid.fillSizes();

      // Get the LEM objects and add them.
      if(fmLem.isValid())
      {
        std::vector< art::Ptr<lem::PIDDetails> > lems = fmLem.at(subevtId);
        if(lems.size() > 0)
        {
          rec->sel.lem.pid               = lems[0]->Value();
          rec->sel.lem.pidfit            = lems[0]->vars.fPIDFit;
          rec->sel.lem.pidexp            = lems[0]->vars.fPIDExp; 
          rec->sel.lem.pidexpfit         = lems[0]->vars.fPIDExpFit;    
          rec->sel.lem.meany             = lems[0]->vars.fMeanY;  
          rec->sel.lem.meanyfit          = lems[0]->vars.fMeanYFit;  
          rec->sel.lem.meanyexp          = lems[0]->vars.fMeanYExp;    
          rec->sel.lem.meanysig          = lems[0]->vars.fMeanYSig;        
          rec->sel.lem.meanysigexp       = lems[0]->vars.fMeanYSigExp;        
          rec->sel.lem.meanqfrac         = lems[0]->vars.fMeanQFrac;        
          rec->sel.lem.meanqfracfit      = lems[0]->vars.fMeanQFracFit;        
          rec->sel.lem.meanqfracexp      = lems[0]->vars.fMeanQFracExp;        
          rec->sel.lem.meanqfracsig      = lems[0]->vars.fMeanQFracSig;        
          rec->sel.lem.meanqfracsigexp   = lems[0]->vars.fMeanQFracSigExp;            
          rec->sel.lem.meanqfracbkg      = lems[0]->vars.fMeanQFracBkg;    
          rec->sel.lem.meanqfracbkgexp   = lems[0]->vars.fMeanQFracBkgExp;        
          rec->sel.lem.energydiff        = lems[0]->vars.fEnergyDiff;  
          rec->sel.lem.energydiffexp     = lems[0]->vars.fEnergyDiffExp;        
          rec->sel.lem.enrichfrac        = lems[0]->vars.fEnrichFrac;        
          rec->sel.lem.enrichfracexp     = lems[0]->vars.fEnrichFracExp;            
          rec->sel.lem.fannid            = lems[0]->vars.fFannId;    
          rec->sel.lem.fannidenrich      = lems[0]->vars.fFannIdEnrich;        
          rec->sel.lem.chisig            = lems[0]->vars.fChiSig;  
          rec->sel.lem.chibkg            = lems[0]->vars.fChiBkg;  
          rec->sel.lem.chisigexp         = lems[0]->vars.fChiSigExp;        
          rec->sel.lem.chibkgexp         = lems[0]->vars.fChiBkgExp;
          rec->sel.lem.avginvE           = lems[0]->vars.fAvgInvE;
          rec->sel.lem.avgexpE           = lems[0]->vars.fAvgExpE;
          rec->sel.lem.avgsigE           = lems[0]->vars.fAvgSigE;
          rec->sel.lem.tmvabdtg          = lems[0]->vars.fTMVABDTG;
          rec->sel.lem.tmvabdt           = lems[0]->vars.fTMVABDT;
          rec->sel.lem.tmvabdtd          = lems[0]->vars.fTMVABDTD;
          rec->sel.lem.tmvabdtmf         = lems[0]->vars.fTMVABDTMF;
          rec->sel.lem.tmvamlp           = lems[0]->vars.fTMVAMLP;
          rec->sel.lem.dectree           = lems[0]->vars.fDecTreePID;
        }
        else{
          rec->sel.lem.setDefault();
        }
      }
      else{
        rec->sel.lem.setDefault();
      }
      
      
      if(fmRvp.isValid())
      {
        std::vector< art::Ptr<rvp::RVP> > rvps = fmRvp.at(subevtId);
        if(rvps.size() > 0)
        {
          rec->sel.rvp.pid         = rvps[0]->Value();
          rec->sel.rvp.ncell       = rvps[0]->NCells();
          rec->sel.rvp.recoE       = rvps[0]->RecoE();
          rec->sel.rvp.longtr      = rvps[0]->LongTr();
          rec->sel.rvp.longtrfrac  = rvps[0]->LongTrFrac();
          rec->sel.rvp.mipfrac     = rvps[0]->MipFrac();
          rec->sel.rvp.miphits     = rvps[0]->MipHits();
          rec->sel.rvp.epl20frac   = rvps[0]->EPl20Frac();
          rec->sel.rvp.efrac2plwin = rvps[0]->EFrac2PlWin();
          rec->sel.rvp.efrac6plwin = rvps[0]->EFrac6PlWin();
          rec->sel.rvp.efrac2sigrd = rvps[0]->EFrac2SigRd();
          rec->sel.rvp.prongs3D    = rvps[0]->Prongs3D();
          rec->sel.rvp.prongEbal3D = rvps[0]->ProngEBal3D();
          rec->sel.rvp.prongs2D    = rvps[0]->Prongs2D();
          rec->sel.rvp.prongEbal2D = rvps[0]->ProngEBal2D();
          rec->sel.rvp.eiso3sig    = rvps[0]->EIso3Sig();
          rec->sel.rvp.rvp12       = rvps[0]->RVP12();
        }
        else{
          rec->sel.rvp.setDefault();
        }
      }
      else{
        rec->sel.rvp.setDefault();
      }

      if(fmCosRej.isValid())
      {
        std::vector< art::Ptr<cosrej::CosRejObj> > cosRejs = fmCosRej.at(subevtId);
        if(cosRejs.size() > 0)
        {
          rec->sel.cosrej.anglekal      = cosRejs[0]->AngleKal();
          rec->sel.cosrej.anglecos      = cosRejs[0]->AngleCos();
          rec->sel.cosrej.numucontpid   = cosRejs[0]->ConCosPID();
          rec->sel.cosrej.numuuncontpid = cosRejs[0]->UnconCosPID();
          rec->sel.cosrej.nkal3d        = cosRejs[0]->NTracks3D();
          rec->sel.cosrej.mindist       = cosRejs[0]->MinDist();
          rec->sel.cosrej.mincell       = cosRejs[0]->MinCell();
          rec->sel.cosrej.eratio        = cosRejs[0]->ERatio();
          rec->sel.cosrej.scatt         = cosRejs[0]->Scatt();
          rec->sel.cosrej.hadE          = cosRejs[0]->StartAct();
          rec->sel.cosrej.costhetatrue  = cosRejs[0]->CosThetaTrue();
          rec->sel.cosrej.kalthetatrue  = cosRejs[0]->KalThetaTrue();
          rec->sel.cosrej.kaldist       = cosRejs[0]->KalDist();
          rec->sel.cosrej.fscattmax     = cosRejs[0]->FScattMax();
          rec->sel.cosrej.fscattsum     = cosRejs[0]->FScattSum();
          rec->sel.cosrej.fscattext     = cosRejs[0]->FScattExt();
          rec->sel.cosrej.fscattsig     = cosRejs[0]->FScattSig();
          rec->sel.cosrej.cvetocont     = cosRejs[0]->CVCont();
          rec->sel.cosrej.cvetoangleweak   = cosRejs[0]->CVAngleW();
          rec->sel.cosrej.cvetoanglestrong = cosRejs[0]->CVAngleS();
          rec->sel.cosrej.cvetofwddist  = cosRejs[0]->CVFwdDist();
          rec->sel.cosrej.cvetobakdist  = cosRejs[0]->CVBakDist();
          rec->sel.cosrej.nbbc          = cosRejs[0]->BBC();
          rec->sel.cosrej.rpid          = cosRejs[0]->RPID();
          rec->sel.cosrej.kfitspeed    = cosRejs[0]->KFitSpeed();
          rec->sel.cosrej.cfitspeed    = cosRejs[0]->CFitSpeed();
          rec->sel.cosrej.kdirscore    = cosRejs[0]->KDirScore();
          rec->sel.cosrej.cdirscore    = cosRejs[0]->CDirScore();
          rec->sel.cosrej.kscorediff   = cosRejs[0]->KScoreDiff();
          rec->sel.cosrej.cscorediff   = cosRejs[0]->CScoreDiff();
          //          rec->sel.cosrej.kalslope     = cosRejs[0]->KalSlope();
          //          rec->sel.cosrej.kalchisq     = cosRejs[0]->KalChisq();
          //          rec->sel.cosrej.kalchidiff   = cosRejs[0]->KalChiDiff();
          //          rec->sel.cosrej.cosslope     = cosRejs[0]->CosSlope();
          //          rec->sel.cosrej.coschisq     = cosRejs[0]->CosChisq();
          //          rec->sel.cosrej.coschidiff   = cosRejs[0]->CosChiDiff();

          rec->sel.contain.cosfwddist    = cosRejs[0]->CosFwdDist();
          rec->sel.contain.cosbakdist    = cosRejs[0]->CosBakDist();
          rec->sel.contain.cosfwdcell    = cosRejs[0]->CosFwdCell();
          rec->sel.contain.cosbakcell    = cosRejs[0]->CosBakCell();
          rec->sel.contain.cosfwdair     = cosRejs[0]->CosFwdAir();
          rec->sel.contain.cosfwdsteel   = cosRejs[0]->CosFwdSteel();
          rec->sel.contain.cosbakair     = cosRejs[0]->CosBakAir();
          rec->sel.contain.cosbaksteel   = cosRejs[0]->CosBakSteel();
          rec->sel.contain.kalfwdair     = cosRejs[0]->KalFwdAir();
          rec->sel.contain.kalfwdsteel   = cosRejs[0]->KalFwdSteel();
          rec->sel.contain.kalbakair     = cosRejs[0]->KalBakAir();
          rec->sel.contain.kalbaksteel   = cosRejs[0]->KalBakSteel();
          rec->sel.contain.kalfwddist    = cosRejs[0]->KalFwdDist();
          rec->sel.contain.kalbakdist    = cosRejs[0]->KalBakDist();
          rec->sel.contain.kalfwdcell    = cosRejs[0]->KalFwdCell();
          rec->sel.contain.kalbakcell    = cosRejs[0]->KalBakCell();
          rec->sel.contain.cosyposattrans= cosRejs[0]->CosYPosAtTrans();
          rec->sel.contain.kalyposattrans= cosRejs[0]->KalYPosAtTrans();

          if(cosRejs[0]->CosFwdCell() > 0 && cosRejs[0]->CosBakCell() > 0 &&
             cosRejs[0]->KalFwdCell() >10 && cosRejs[0]->KalBakCell() > 10 &&
             slcInfo.minCellsFromEdge > 1 && int(slcInfo.minPlane) > (((rec->hdr.dibfirst-1)*64)+1) &&
             int(slcInfo.maxPlane) < (rec->hdr.diblast*64)-2) rec->sel.contain.numucontain = true;
        }
        else{
          rec->sel.cosrej.setDefault();
        }
      }
      else{
        rec->sel.cosrej.setDefault();
      }

      if(fmNueCosRej.isValid())
      {
        std::vector< art::Ptr<cosrej::NueCosRej> > nuecosrejs = fmNueCosRej.at(subevtId);
        if(nuecosrejs.size() > 0)
        {

          rec->sel.nuecosrej.hitplane  	  = nuecosrejs[0]->HitsPerPlane();
          rec->sel.nuecosrej.pngptp   	  = nuecosrejs[0]->ProngTransMom();
          rec->sel.nuecosrej.partptp	  = nuecosrejs[0]->ParticleShowerTransMom();
          rec->sel.nuecosrej.photptp	  = nuecosrejs[0]->PhotonShowerTransMom();
          rec->sel.nuecosrej.starttop	  = nuecosrejs[0]->StartDistToTop();
          rec->sel.nuecosrej.startbottom  = nuecosrejs[0]->StartDistToBottom();
          rec->sel.nuecosrej.startfront	  = nuecosrejs[0]->StartDistToFront();
          rec->sel.nuecosrej.startback	  = nuecosrejs[0]->StartDistToBack();
          rec->sel.nuecosrej.startwest	  = nuecosrejs[0]->StartDistToWest();
          rec->sel.nuecosrej.starteast	  = nuecosrejs[0]->StartDistToEast();
          rec->sel.nuecosrej.stoptop	  = nuecosrejs[0]->StopDistToTop();
          rec->sel.nuecosrej.stopbottom	  = nuecosrejs[0]->StopDistToBottom();
          rec->sel.nuecosrej.stopfront	  = nuecosrejs[0]->StopDistToFront();
          rec->sel.nuecosrej.stopback	  = nuecosrejs[0]->StopDistToBack();
          rec->sel.nuecosrej.stopwest	  = nuecosrejs[0]->StopDistToWest();
          rec->sel.nuecosrej.stopeast	  = nuecosrejs[0]->StopDistToEast();
          rec->sel.nuecosrej.dang	  = nuecosrejs[0]->AngleToNextShower();
          rec->sel.nuecosrej.vtxdoca	  = nuecosrejs[0]->ShowerDistToVtx();
          rec->sel.nuecosrej.showermaxx	  = nuecosrejs[0]->ShowerMaxX();
          rec->sel.nuecosrej.showermaxy	  = nuecosrejs[0]->ShowerMaxY();
          rec->sel.nuecosrej.showermaxz	  = nuecosrejs[0]->ShowerMaxZ();
        }
        else{
          rec->sel.nuecosrej.setDefault();
        }
      }
      else{
        rec->sel.nuecosrej.setDefault();
      }

      if(fmQepid.isValid())
      {
        std::vector< art::Ptr<qeef::QePId> > qepids = fmQepid.at(subevtId);
        if(qepids.size() > 0)
        {
          rec->sel.qepid.pdg = qepids[0]->Pdg();
          rec->sel.qepid.mode = qepids[0]->Mode();
          rec->sel.qepid.ntrk = qepids[0]->Ntrk();
          rec->sel.qepid.pid = qepids[0]->Value();
          rec->sel.qepid.offE = qepids[0]->OffTrkE();
          rec->sel.qepid.ediff = qepids[0]->Ediff();
          rec->sel.qepid.ediffz = qepids[0]->EdiffZ();
          rec->sel.qepid.dedx = qepids[0]->Dedx();
        }
        else{
          rec->sel.qepid.setDefault();
        }
      }
      else{
        rec->sel.qepid.setDefault();
      }

      // Get the Veto objects and add them.
      if(fmVeto.isValid())
      {
        std::vector< art::Ptr<presel::Veto> > vetoCuts = fmVeto.at(subevtId);
        if( !vetoCuts.empty() )
        {
          rec->sel.veto.ncell      = vetoCuts[0]->NCell();
          rec->sel.veto.fwddist    = vetoCuts[0]->FwdDist();
          rec->sel.veto.bakdist    = vetoCuts[0]->BakDist();
          rec->sel.veto.ratio      = vetoCuts[0]->Ratio();
          rec->sel.veto.diry       = vetoCuts[0]->DirY();
          rec->sel.veto.angle      = vetoCuts[0]->Angle();
          rec->sel.veto.anglevar   = vetoCuts[0]->AngleVar();
          rec->sel.veto.anglevarold= vetoCuts[0]->AngleVarOld();
          rec->sel.veto.passthru   = vetoCuts[0]->PassThru();
          rec->sel.veto.passthruold= vetoCuts[0]->PassThruOld();
          rec->sel.veto.passangle  = vetoCuts[0]->PassAngle();
          rec->sel.veto.passanglefirstanal  = vetoCuts[0]->PassAngleFirstAnal();
          rec->sel.veto.keep       = vetoCuts[0]->Keep();
        }
        else{
          rec->sel.veto.setDefault();
        }
      }
      else{
        rec->sel.veto.setDefault();
      }

      if( fmMrccPar.isValid() ){
	std::vector< art::Ptr<murem::MRCCParent> > mrccParent = fmMrccPar.at(subevtId);
	if( !mrccParent.empty() ){
	  rec->parent.mrccpar.eff             = mrccParent[0]->ParentEff();
	  rec->parent.mrccpar.pur             = mrccParent[0]->ParentPur();
	  rec->parent.mrccpar.slcidx          = mrccParent[0]->ParentIndex();

	  }// end if mrccParent is not empty
      }// end if fmMrccPar is valid
      else
	rec->parent.mrccpar.setDefault();

      // Get the NumuE objects and add them.
      if(fmNumuEnergy.isValid())
      {
        std::vector< art::Ptr<numue::NumuE> > numuEnergies = fmNumuEnergy.at(subevtId);
        if( !numuEnergies.empty() )
        {
          rec->energy.numusimp.E             =  numuEnergies[0]->E();
          rec->energy.numusimp.calccE        =  numuEnergies[0]->CalCCE();
          rec->energy.numusimp.trkqeE        =  numuEnergies[0]->TrkQEE();
          rec->energy.numusimp.trknonqeE     =  numuEnergies[0]->TrkNonQEE();
          rec->energy.numusimp.angleE        =  numuEnergies[0]->AngleQEE();
          rec->energy.numusimp.angleerror    =  numuEnergies[0]->AngleQEError();
          rec->energy.numusimp.hadcalE       =  numuEnergies[0]->HadCalE();
          rec->energy.numusimp.hadtrkE       =  numuEnergies[0]->HadTrkE();
          rec->energy.numusimp.ucccE         =  numuEnergies[0]->UCCCE();
          rec->energy.numusimp.fsE           =  numuEnergies[0]->FScatterE();
          rec->energy.numusimp.ndtrklenact   =  numuEnergies[0]->NDTrkLenAct();
          rec->energy.numusimp.ndtrklencat   =  numuEnergies[0]->NDTrkLenCat();
          rec->energy.numusimp.ndtrkcalactE  =  numuEnergies[0]->NDTrkCalAct();
          rec->energy.numusimp.ndtrkcaltranE =  numuEnergies[0]->NDTrkCalTran();
          rec->energy.numusimp.ndtrkcalcatE  =  numuEnergies[0]->NDTrkCalCat();
          rec->energy.numusimp.ndhadcalactE  =  numuEnergies[0]->NDHadCalAct();
          rec->energy.numusimp.ndhadcaltranE =  numuEnergies[0]->NDHadCalTran();
          rec->energy.numusimp.ndhadcalcatE  =  numuEnergies[0]->NDHadCalCat();
          rec->energy.numusimp.ndhadtrkactE  =  numuEnergies[0]->NDHadTrkAct();
          rec->energy.numusimp.ndhadtrktranE =  numuEnergies[0]->NDHadTrkTran();
          rec->energy.numusimp.ndhadtrkcatE  =  numuEnergies[0]->NDHadTrkCat();
          rec->energy.numusimp.ndtrktranx    =  numuEnergies[0]->NDTrkTranX();
          rec->energy.numusimp.ndtrktrany    =  numuEnergies[0]->NDTrkTranY();
        }
        else{
          rec->energy.numusimp.setDefault();
        }
        
      }
      else{
        rec->energy.numusimp.setDefault();
      }
      
      if(fmNueSand.isValid())
      {
        std::vector<art::Ptr<nuesand::NueSandObj> > nueSands = fmNueSand.at(subevtId);
        if(!nueSands.empty())
        {
          rec->sand.nue.infid      = nueSands[0]->InFiducialVolume();
          rec->sand.nue.slicemeanx = nueSands[0]->SliceMeanPosition().X();
          rec->sand.nue.slicemeany = nueSands[0]->SliceMeanPosition().Y();
          rec->sand.nue.slicemeanz = nueSands[0]->SliceMeanPosition().Z();
          
          rec->sand.nue.npi0 = nueSands[0]->fNPi0;
          rec->sand.nue.npip = nueSands[0]->fNPiPlus;
          rec->sand.nue.npim = nueSands[0]->fNPiMinus;
          
          rec->sand.nue.pi0E = nueSands[0]->fPi0Energy;
          rec->sand.nue.pipE = nueSands[0]->fPiPlusEnergy;
          rec->sand.nue.pimE = nueSands[0]->fPiMinusEnergy;
          
          rec->sand.nue.phot0E  = nueSands[0]->fPhot0E;
          rec->sand.nue.phot1E  = nueSands[0]->fPhot1E;
          rec->sand.nue.opencos = nueSands[0]->fOpenCos;
          rec->sand.nue.conv0   = nueSands[0]->fConvL0;
          rec->sand.nue.conv1   = nueSands[0]->fConvL1;

	  rec->sand.nue.mincellcosmic     = nueSands[0]->fMincellCosmic;
	  rec->sand.nue.mincellwallcosmic = nueSands[0]->fMincellWallCosmic;
	  rec->sand.nue.anglecosmic       = nueSands[0]->fAngleCosmic;
	  rec->sand.nue.nplanesfilled     = nueSands[0]->fFilledPlanes;
	  rec->sand.nue.nxplanes          = nueSands[0]->fExtentXPlanes;
	  rec->sand.nue.nyplanes          = nueSands[0]->fExtentYPlanes;
	  rec->sand.nue.ncellrow          = nueSands[0]->fMaxRowOfCells;
	  rec->sand.nue.ncellsum          = nueSands[0]->fMaxCellsInPlane;
	  rec->sand.nue.mincellavg        = nueSands[0]->fMincellAvg;
	  rec->sand.nue.mincellkalman     = nueSands[0]->fMincellKalman;
	  rec->sand.nue.nmiphits          = nueSands[0]->fNMIPHits;
	  rec->sand.nue.mipmin            = nueSands[0]->fMIPMin;
	  rec->sand.nue.mipmax            = nueSands[0]->fMIPMax;

          rec->sand.nue.fracangchanges      = nueSands[0]->fFracAngChanges;
          rec->sand.nue.fracangchangesloose = nueSands[0]->fFracAngChangesLoose;
          rec->sand.nue.fracmodalhits       = nueSands[0]->fFracModalHits;
          rec->sand.nue.fracnoncontig       = nueSands[0]->fFracNonContig;
          rec->sand.nue.fracnoncontigloose  = nueSands[0]->fFracNonContigLoose;
          rec->sand.nue.maxgap              = nueSands[0]->fMaxGap;
        }
      }
      
      if(fmNumuSand.isValid())
      {
        std::vector< art::Ptr<numusand::NumuSandObj> > numuSands = fmNumuSand.at(subevtId);
        if(numuSands.size() > 0)
        {
          rec->sand.numu.pimudecay = numuSands[0]->PiMuDecay();
          rec->sand.numu.vtxdst    = numuSands[0]->VtxDst();
          rec->sand.numu.mudst     = numuSands[0]->MuDst();

          rec->sel.contain.missE   = numuSands[0]->MissingE();

          rec->sel.contain.xmin2   = numuSands[0]->XMin2();
          rec->sel.contain.ymin2   = numuSands[0]->YMin2();
          rec->sel.contain.zmin2   = numuSands[0]->ZMin2();
          rec->sel.contain.xmax2   = numuSands[0]->XMax2();
          rec->sel.contain.ymax2   = numuSands[0]->YMax2();
          rec->sel.contain.zmax2   = numuSands[0]->ZMax2();

          rec->sel.contain.xmint   = numuSands[0]->XMint();
          rec->sel.contain.ymint   = numuSands[0]->YMint();
          rec->sel.contain.zmint   = numuSands[0]->ZMint();
          rec->sel.contain.xmaxt   = numuSands[0]->XMaxt();
          rec->sel.contain.ymaxt   = numuSands[0]->YMaxt();
          rec->sel.contain.zmaxt   = numuSands[0]->ZMaxt();
        }
        else{
          rec->sand.numu.setDefault();
        }
        
      }
      else{
        rec->sand.numu.setDefault();
      }

      //#######################################################
      //Fill truth information
      //#######################################################
      
      // Set mc branch values to default 
      rec->mc.setDefault();
      if(bt->HaveTruthInfo())
      {
        
        
        //Get all truth track hits associated with slice 
        std::vector<cheat::TrackIDE> sliceTracks = bt->HitsToTrackIDE(subevts[subevtId]->AllCells());
        
        int           currentNuIdx        = 0;
        unsigned int  nNuGlobal           = 0; /// count of neutrino interactions in this slice 

        for(unsigned int iEffPur = 0; iEffPur < sliceTruthTable[subevtId].size(); ++iEffPur)
        {
          // continue if truth left no hits in slice
          if(! (sliceTruthTable[subevtId][iEffPur].nSliceHits > 0) ) continue;

          if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEnergy[subevtId])
          {
            cheat::NeutrinoEffPur* faveNu = &sliceTruthTable[subevtId][iEffPur];
            
            //If NOT Neutrinos
            if(! faveNu->neutrinoInt->NeutrinoSet())
            {  
              // Add faveNu to allcosmics
              if(fullTruthInfo) AddCosmicTruthToVec(faveNu, allTracks, sliceTracks, &rec->mc.allcosmics, evt);
              // Add faveNu to cosmic
              AddCosmicTruthToVec(faveNu, allTracks, sliceTracks, &rec->mc.cosmic, evt);            
            }
            // Otherwise we are a neutrino
            else
            {

              // Add faveNu to allnus
              if(fullTruthInfo) AddMCTruthToVec(faveNu, allTracks, sliceTracks, &rec->mc.allnus, evt);
              // Add faveNu to nu        
              AddMCTruthToVec(faveNu, allTracks, sliceTracks, &rec->mc.nu, evt);            
              rec->mc.faveidxenergy = currentNuIdx;
              
              if(generatorInfo) {
                //Add faveNu to allgen generator info branch
                if(fullTruthInfo) AddGenInfoToVec(faveNu, allTracks, sliceTracks, &rec->mc.allgen, evt);
                //Add faveNu to generator info branch
                AddGenInfoToVec(faveNu, allTracks, sliceTracks, &rec->mc.gen, evt);
              }

              art::PtrVector<simb::MCTruth> pv;
              pv.push_back(faveNu->neutrinoInt);
              art::FindManyP<rwgt::GENIEReweightTable> fmRW(pv, evt, reweightName);
              if(fmRW.isValid()) {
                const std::vector<art::Ptr<rwgt::GENIEReweightTable>> rws = fmRW.at(0);

                if(!rws.empty()) rec->mc.rwgt = GenieReweightTable(*rws[0]);
              }

              
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEff[subevtId])
                rec->mc.faveidxeff        = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsPur[subevtId])
                rec->mc.faveidxpur        = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffPur[subevtId])
                rec->mc.faveidxeffpur     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffThenPur[subevtId])
                rec->mc.faveidxeffthenpur = currentNuIdx;
              
              currentNuIdx++;
              
            }
          }
        }

        for(unsigned int iEffPur = 0; iEffPur < sliceTruthTable[subevtId].size(); ++iEffPur) 
        {

          if(sliceTruthTable[subevtId][iEffPur].neutrinoInt->NeutrinoSet())
            nNuGlobal++;
          // continue if truth left no hits in slice
          if(! (sliceTruthTable[subevtId][iEffPur].nSliceHits > 0) ) continue;
          
          if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex != faveIdsEnergy[subevtId])
          {
            
            // Cosmics
            if(! sliceTruthTable[subevtId][iEffPur].neutrinoInt->NeutrinoSet())
            {
              if(fullTruthInfo) AddCosmicTruthToVec(&sliceTruthTable[subevtId][iEffPur], allTracks, sliceTracks, &rec->mc.allcosmics,  evt);
            }
            // Neutrino
            else
            {
              
              if(fullTruthInfo) AddMCTruthToVec(&sliceTruthTable[subevtId][iEffPur], allTracks, sliceTracks, &rec->mc.allnus, evt);
              
              
              
              if(generatorInfo && fullTruthInfo) {
                AddGenInfoToVec(&sliceTruthTable[subevtId][iEffPur], allTracks, sliceTracks, &rec->mc.allgen, evt);
              }
              
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEff[subevtId])
                rec->mc.faveidxeff     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsPur[subevtId])
                rec->mc.faveidxpur     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffPur[subevtId])
                rec->mc.faveidxeffpur     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffThenPur[subevtId])
                rec->mc.faveidxeffthenpur = currentNuIdx;
              
              currentNuIdx++;
              
            }
          }
        }// end loop to fill all* vectors

        // fill size fields for mc branch
        rec->mc.nnu           =  rec->mc.nu         .size();
        rec->mc.ncosmic       =  rec->mc.cosmic     .size();
        rec->mc.ngen          =  rec->mc.gen        .size();
        if(fullTruthInfo){
          rec->mc.nallnus     =  rec->mc.allnus     .size();
          rec->mc.nallcosmics =  rec->mc.allcosmics .size();
          rec->mc.nallgen     =  rec->mc.allgen     .size();
        }
        rec->mc.global.nnu    =  nNuGlobal;
      }//end MC block

      if (enableBlindness) BlindThisRecord(rec);

      recTree->Fill();
      delete rec;
    }// end loop over slices
    
    return;
  }
  
  
  void CAFMaker::endSubRun(art::SubRun& sr)
  {
    art::Handle<sumdata::POTSum> pot;
    //if this is mc get the spill info from the generator
    if (bt->HaveTruthInfo())
    {
      sr.getByLabel(generatorName, pot);
    }
    //if we have real data try to get the spill info from the NuMI beam
    else {
      sr.getByLabel(numiBeamName, pot);
    }
    
    if(!pot.failedToGet())
    {
      
      //NOTE: The pot that comes out of IFDBSpillInfo has an implied factor of 1E12 that is being left off
      //For now the data POT is being multiplied by this factor so that the number is consistent with what we get in MC
      //Is this what we want to do in the future?
      float subrunpot = pot->totgoodpot;
      if (!bt->HaveTruthInfo()) subrunpot *= 1e12;
      fTotalPOT += subrunpot;
      
      mf::LogInfo("CAFMaker") << "POT in subrun " << sr.subRun()
      << " = " << pot->totgoodpot;
    }
    
    art::Handle<sumdata::CosmicExposure> expo;
    sr.getByLabel(exposureName, expo);
    
    if(!expo.failedToGet())
    {
      fTotalLivetime += expo->totlivetime;
      mf::LogInfo("CAFMaker") << "Exposure in subrun " << sr.subRun()
      << " = " << expo->totlivetime;
    }
  }
  
  
  
  //......................................................................
  void CAFMaker::endJob()
  {
    fFile->cd();
    hPOT->Fill(.5, fTotalPOT);
    hEvents->Fill(.5, fTotalEvents);
    hLivetime->Fill(.5, fTotalLivetime);
    
    hPOT->Write();
    hEvents->Write();
    hLivetime->Write();
    fFile->Write();
    
    std::map<std::string, std::string> metadata = meta::MetadataManager::getInstance().GetMetadata();
    metadata["data_tier"] = dataTier;
    if(fIsRealData)   metadata["file_type"] = "importedDetector";
    else                   metadata["file_type"] = "importedSimulated";

    
    if(metadata.count("online.stream"))
    {
      // Then this is a real data file and the stream has been filled in the metadata 
      if( ! metadata["online.stream"].compare("0"))
      {
        // Then this is a numi file
        if(!enableBlindness)
        {
          // Then we are not blinding it, need to mark it as such in the metadata
          std::string tag("restrictedbox");
          if(std::string("none").compare(metadata["nova.special"]))
          { // Then the metadata is not the same as none, we need to append to it
            // Use a stringstream to jam the strings together and append the tag 
            std::stringstream ss;
            ss.str();
            ss << metadata["nova.special"] << "-" << tag;
            metadata["nova.special"] = ss.str();
          }
          else
            metadata["nova.special"] = tag;
        }
      }
    } 
    
    AddMetadataToFile(metadata);
  }

  //......................................................................
  double CAFMaker::SimpleOscProb(const simb::MCFlux& flux,
                                 const simb::MCNeutrino& nu) const
  { 
    const double L      = 810.;    // km
    const double ldm    = 2.35e-3; // large delta m^2, m23
    const double ss2t13 = 0.1;     // sin^2(2theta_13)
    const double ss2t23 = 1.;      // maximal sin^2(2theta_23)
    const double ssth23 = 0.5;     // sin^2(theta_23) corresponding to above
    
    // Signal
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux.fntype) == 14) 
    {
      // Nue appearance
      return ss2t13*ssth23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux.fntype) == 14)
    {
      // CC mu disappearance
      return 1-ss2t23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    
    // Background
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux.fntype) == 12)
    {
      // Beam nue
      return 1-ss2t13*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    if(nu.CCNC() == 1 && nu.Nu().PdgCode() == flux.fntype)
    {
      // NC
      return 1;
    }
    if(nu.CCNC() == 1 && nu.Nu().PdgCode() != flux.fntype)
    {
      // Swapped NC. Doesn't exist. If people want to play clever reweighting
      // games to reclaim these, they can figure that out, but this allows the
      // naive implementation to work.
      return 0;
    }
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux.fntype) == 12)
    {
      // CC mu appearance
      return ss2t13*ssth23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    // Don't know what this is
    return 0; 
  }

  //......................................................................
  std::vector<caf::SRGenieWeights> CAFMaker::
  GenieReweightTable(const rwgt::GENIEReweightTable& table) const
  {
    const unsigned int N = table.NShifts();

    std::vector<caf::SRGenieWeights> ret(N);

    for(unsigned int i = 0; i < N; ++i){
      ret[i].minus2sigma = table.Minus2Sigma(i);
      ret[i].minus1sigma = table.Minus1Sigma(i);
      ret[i].plus1sigma = table.Plus1Sigma(i);
      ret[i].plus2sigma = table.Plus2Sigma(i);
    }

    return ret;
  }

  DEFINE_ART_MODULE(CAFMaker);

} // end namespace caf
////////////////////////////////////////////////////////////////////////
