////////////////////////////////////////////////////////////////////////                  
/// \file    Mpi0.cxx
/// \brief   TODO                                                                         
/// \version $Id: Mpi0.cxx,v 1.3 2012-10-22 04:27:49 gsdavies Exp $                                                                       
/// \author  xbbu@fnal.gov
////////////////////////////////////////////////////////////////////////
#include "CalibrationUtils/NNfunction.h"

// ROOT includes
#include "TH1F.h"
#include "TMath.h"
#include "TLorentzVector.h"
#include "TVector3.h"
#include "TNtuple.h"
#include "TTree.h"
#include "TH2F.h"
#include "TProfile.h"
#include "TProfile2D.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "Utilities/func/MathUtil.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>

//include TMVA
#include "TMVA/Tools.h"
#include "TMVA/Reader.h"
#include "TMVA/MethodCuts.h"

class TLorentzVector;
class TH1F;
class TTree;

namespace Ed {
  class Mpi0 : public art::EDAnalyzer {
  public:
    explicit Mpi0(fhicl::ParameterSet const& pset);
    virtual ~Mpi0();

    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    double DeltaR(TLorentzVector v1, TLorentzVector v2);

  private:

    //folders to get objects from
    std::string fSliceLabel;     ///label for slices
    std::string fTrackLabel;     ///label for tracks
    std::string fProngLabel;     ///label for prong 
    std::string fInstLabel3D;    ///instance label for prongs 3D
    std::string fVertexLabel;    ///label for vertex

    //save event info to ntuples
    TTree *_otree;
    /*** Vertex ***/
    Float_t recoVx;//reco vertex X
    Float_t recoVy;//reco vertex Y
    Float_t recoVz;//reco vertex Z
    //for pi0->2gamma
    Int_t HasPi0;
    Float_t Epi0;
    Float_t gam1E;
    Float_t gam1Px;
    Float_t gam1Py;
    Float_t gam1Pz;
    Float_t gam2E;
    Float_t gam2Px;
    Float_t gam2Py;
    Float_t gam2Pz;
    /*** slicer information ***/
    Int_t Ncells;//# of associate cells
    Int_t Nplanes;//# of extended planes
    Float_t recoE;//reco energy
    //longitudial shower
    Float_t E40planes;
    Float_t Efrac_40planes;
    Float_t Efrac_20planes;
    Float_t Efrac_2slides;
    Float_t Efrac_4slides;
    Float_t Efrac_6slides;
    //transverse shower
    Float_t EDxDy;
    Float_t E2Dx2Dy2;
    Float_t Efrac_2sigma;
    Float_t Efrac_4sigma;
    Float_t Efrac_6sigma;
    /*** longest track ***/
    Float_t trackLength;//length of longest track
    Int_t trackNcells;//# of cells associated to longest track
    Float_t trackEnergy;//energy associated to longest track
    Float_t trackNcells_ratio;//# of cells associated to longest track over Ncells
    Float_t trackEnergy_ratio;//energy associated to longest track over recoE
    /*** mip information ***/
    Float_t Nmip;//# of mip cells
    Float_t Fmip;//# of mip cells over Ncells
    /*** prong information ***/
    Int_t Nprongs;
    Float_t Ebalance;
    Float_t Dphi;
    Float_t Dtheta;
    Float_t gam1_Ereco;
    Float_t gam2_Ereco;
    Float_t gam1_ncells;
    Float_t gam2_ncells;
    Float_t Mreco;
    Float_t Mnncorr;
    Float_t prongEnergy[20];
    Float_t prongPhi[20];
    Float_t prongTheta[20];
    Int_t prongNCells[20];//# of cells
    Float_t pi0BDT;
  };
}//end namespace


using namespace TMVA;

namespace Ed{

  TMVA::Reader *_pi0bdt=new TMVA::Reader();
  Float_t inputvars_pi0[15]={0.};
  TString inputstr_pi0[15]={"_ncells","_energy","_ltrack","_nmip","_efrac_slide2",
                            "_efrac_slide6","_efrac_2sigma","_dphi","_dtheta",
                            "_ebalance","_mreco","_Eg1","_Eg2","_Ng1","_Ng2"};


  //......................................................................
  Mpi0::Mpi0(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    std::cout<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  Mpi0::~Mpi0()
  {
  }
  
  //......................................................................
  void Mpi0::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSliceLabel   = pset.get< std::string >("SliceLabel");
    fTrackLabel   = pset.get< std::string >("TrackLabel");
    fProngLabel   = pset.get< std::string >("ProngLabel");
    fInstLabel3D  = pset.get< std::string >("InstLabel3D");
    fVertexLabel  = pset.get< std::string >("VertexLabel");
  }

  //......................................................................
  void Mpi0::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    
 
    for( int i=0; i<15; ++i )
      _pi0bdt->AddVariable(inputstr_pi0[i],&inputvars_pi0[i]);
    
    _pi0bdt->BookMVA("BDT method","pi0BDT_20130114.weights.xml");

    _otree = tfs->make<TTree>("Mpi0","Reconstruct pi0 mass");
    _otree->Branch("recoVx",&recoVx,"recoVx/F");
    _otree->Branch("recoVy",&recoVy,"recoVy/F");
    _otree->Branch("recoVz",&recoVz,"recoVz/F");
    
    _otree->Branch("HasPi0",&HasPi0,"HasPi0/I");
    _otree->Branch("Epi0",&Epi0,"Epi0/F");
    _otree->Branch("gam1E",&gam1E,"gam1E/F");
    _otree->Branch("gam1Px",&gam1Px,"gam1Px/F");
    _otree->Branch("gam1Py",&gam1Py,"gam1Py/F");
    _otree->Branch("gam1Pz",&gam1Pz,"gam1Pz/F");
    _otree->Branch("gam2E",&gam2E,"gam2E/F");
    _otree->Branch("gam2Px",&gam2Px,"gam2Px/F");
    _otree->Branch("gam2Py",&gam2Py,"gam2Py/F");
    _otree->Branch("gam2Pz",&gam2Pz,"gam2Pz/F");

    _otree->Branch("Ncells",&Ncells,"Ncells/I");
    _otree->Branch("Nplanes",&Nplanes,"Nplanes/I");
    _otree->Branch("recoE",&recoE,"recoE/F");
    _otree->Branch("E40planes",&E40planes,"E40planes/F");
    _otree->Branch("Efrac_20planes",&Efrac_20planes,"Efrac_20planes/F");
    _otree->Branch("Efrac_40planes",&Efrac_40planes,"Efrac_40planes/F");
    _otree->Branch("Efrac_2slides",&Efrac_2slides,"Efrac_2slides/F");
    _otree->Branch("Efrac_4slides",&Efrac_4slides,"Efrac_4slides/F");
    _otree->Branch("Efrac_6slides",&Efrac_6slides,"Efrac_6slides/F");
    _otree->Branch("Efrac_2sigma",&Efrac_2sigma,"Efrac_2sigma/F");
    _otree->Branch("Efrac_4sigma",&Efrac_4sigma,"Efrac_4sigma/F");
    _otree->Branch("Efrac_6sigma",&Efrac_6sigma,"Efrac_6sigma/F");
    _otree->Branch("EDxDy",&EDxDy,"EDxDy/F");
    _otree->Branch("E2Dx2Dy2",&E2Dx2Dy2,"E2Dx2Dy2/F");

    _otree->Branch("Nmip",&Nmip,"Nmip/F");
    _otree->Branch("Fmip",&Fmip,"Fmip/F");

    _otree->Branch("Nprongs",&Nprongs,"Nprongs/I");
    _otree->Branch("Ebalance",&Ebalance,"Ebalance/F");
    _otree->Branch("Dphi",&Dphi,"Dphi/F");
    _otree->Branch("Dtheta",&Dtheta,"Dtheta/F");
    _otree->Branch("Mreco",&Mreco,"Mreco/F");
    _otree->Branch("Mnncorr",&Mnncorr,"Mnncorr/F");
    
    _otree->Branch("gam1_Ereco",&gam1_Ereco,"gam1_Ereco/F");
    _otree->Branch("gam2_Ereco",&gam2_Ereco,"gam2_Ereco/F");
    _otree->Branch("gam1_ncells",&gam1_ncells,"gam1_ncells/F");
    _otree->Branch("gam2_ncells",&gam2_ncells,"gam2_ncells/F");

    _otree->Branch("prongEnergy",&prongEnergy,"prongEnergy[Nprongs]/F");
    _otree->Branch("prongPhi",&prongPhi,"prongPhi[Nprongs]/F");
    _otree->Branch("prongTheta",&prongTheta,"prongTheta[Nprongs]/F");
    _otree->Branch("prongNCells",&prongNCells,"prongNCells[Nprongs]/I");
    
    _otree->Branch("pi0BDT",&pi0BDT,"pi0BDT/F");
  }

  
  void Mpi0::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<cheat::BackTracker> bt;

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel, slices);
    int bestSlice = -1;
    unsigned int mostCells = 0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;
      if(slice.NCell() > mostCells){
        mostCells = slice.NCell();
        bestSlice = sliceIdx;
      }
    }
    if(bestSlice < 0) return;
    const rb::Cluster& slice = (*slices)[bestSlice];

    //get tracks, verticies and showers associated with the slices
    art::FindManyP<rb::Track> fmTrack(slices, evt, fTrackLabel);
    art::FindManyP<rb::Vertex> fmVertex(slices, evt, fVertexLabel);
    if (!(fmVertex.isValid())) return;
    std::vector<art::Ptr<rb::Vertex>> vert = fmVertex.at(bestSlice);
    if (vert.size() != 1) return;

    //select the true pi0
    int pi0Idx=-1;
    double epi0=-99.;
    const sim::ParticleNavigator& pn = bt->ParticleNavigator();
    for(int ip = 0; ip < pn.NumberOfPrimaries(); ++ip){
      const sim::Particle* genpar = pn.Primary(ip);
      if(genpar->PdgCode() != 111) continue;
      if(genpar->E() > epi0){
          pi0Idx = ip;
          epi0 = genpar->E();
      }
    }//loop true particles
    bool hasPi0=false;
    if( pi0Idx>=0 ){
      const sim::Particle* selectedpi0 = pn.Primary(pi0Idx);
      if(selectedpi0->NumberDaughters() == 2){
          const sim::Particle* gam1 = pn[selectedpi0->Daughter(0)];
          const sim::Particle* gam2 = pn[selectedpi0->Daughter(1)];
          if( gam1->PdgCode()==22 && gam2->PdgCode()==22 ){
            hasPi0=true;
            HasPi0=1;
            Epi0=selectedpi0->E();
            gam1E=gam1->E();
            gam1Px=gam1->Px();
            gam1Py=gam1->Py();
            gam1Pz=gam1->Pz();
            gam2E=gam2->E();
            gam2Px=gam2->Px();
            gam2Py=gam2->Py();
            gam2Pz=gam2->Pz();
       }//
     }//has 2 photons
    }//has pi0

    if( !hasPi0 ){
      HasPi0=0;
      Epi0=-1.;
      gam1E=-1.;
      gam1Px=-1.;
      gam1Py=-1.;
      gam1Pz=-1.;
      gam2E=-1.;
      gam2Px=-1.;
      gam2Py=-1.;
      gam2Pz=-1.;
    }

    //select the longest track
    double longestTrack = 0;
    int longestTrack_ncells=0;
    double longestTrack_energy=0.;
    std::vector<art::Ptr<rb::Track>> tracks = fmTrack.at(bestSlice);

    for(unsigned int n = 0; n < tracks.size(); ++n){
        const double L = tracks[n]->TotalLength();
        if(L > longestTrack){
           longestTrack = L;
           longestTrack_ncells=tracks[n]->NCell();
           longestTrack_energy=tracks[n]->TotalGeV();
       }
      }

    //mip cells and reco energy for slicer
    const double nCells = slice.NCell();
    const int lengthPlanes = slice.ExtentPlane();
    double recoEnergy=0.;
    double nmip=0.;
    double Ntotal=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

       if( !rhit.IsCalibrated() ) continue;
       recoEnergy += rhit.GeV();

       if( rhit.PECorr()>100. && rhit.PECorr()<245. ) nmip++;
       Ntotal++;
    }
    //select 3D prongs
    art::FindManyP<rb::Track> fmProng(vert, evt, fProngLabel);
    std::vector<art::Ptr<rb::Track>> prongs = fmProng.at(0);
    std::vector<art::Ptr<rb::Track>> SelectedProng;
    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
       if( prongs[ip]->IsNoise() ) continue;
       if(prongs[ip]->View() != geo::kXorY) continue;
       SelectedProng.push_back(prongs[ip]);
    }
    int nprongs=SelectedProng.size();

    //preselection cuts
    if( recoEnergy<0.1||recoEnergy>1.5 ) return;
    if( nCells<10||nCells>100 ) return;
    if( lengthPlanes>140 ) return;
    if( longestTrack>400. ) return;
    if( nprongs<2 ) return;

     //calculate per plane cell energy
    std::vector<double> PlaneSumEnergy;
    double sum_edx=0.;//energy*dx
    double sum_e2dx2=0.;//energy*energy*dx*dx
    double sum_edy=0.;//energy*dy
    double sum_e2dy2=0.;//energy*energy*dy*dy
    double totalE_road20=0.;
    double totalE_road40=0.;
    double totalE_road60=0.;

    for( int iplane=0; iplane<lengthPlanes; ++iplane ){
      double cellEnergy=0.;//sum of cells energy in each plane
      int planecells=0;//# of cells in each plane
      double cellX=0.;//energetic cell X in each plane
      double cellY=0.;//energetic cell Y in each plane
      double Ecell=0.;
      double EXcell=0.;
      double EYcell=0.;

       for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

         const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

        if( !rhit.IsCalibrated() ) continue;

        int Dplane=chit->Plane()-slice.MinPlane();
        if( Dplane<0 ) std::cout<<"plane # of this cell is beyond the minimal plane."<<std::endl;
        if( Dplane==iplane ){
          cellEnergy += rhit.GeV();
          planecells++;

          Ecell += rhit.GeV();
          EXcell += rhit.GeV()*rhit.X();
          EYcell += rhit.GeV()*rhit.Y();
         }
       }
     if( Ecell>0. ){
        cellX = EXcell/Ecell;
        cellY = EYcell/Ecell;
     }

     PlaneSumEnergy.push_back(cellEnergy);
     
     for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
	 
         const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

         if( !rhit.IsCalibrated() ) continue;
         int Dplane=chit->Plane()-slice.MinPlane();
         if( Dplane==iplane ){
           double dx=fabs(rhit.X()-cellX);
           double dy=fabs(rhit.Y()-cellY);
           if( chit->View()==geo::kX ){
            sum_edx += rhit.GeV()*dx;
            sum_e2dx2 += rhit.GeV()*rhit.GeV()*dx*dx;
            if( dx<2.*2. ) totalE_road20+=rhit.GeV();
             if( dx<2.*4. ) totalE_road40+=rhit.GeV();
             if( dx<2.*6. ) totalE_road60+=rhit.GeV();
           }//in X view
           else{
             sum_edy += rhit.GeV()*dy;
             sum_e2dy2 += rhit.GeV()*rhit.GeV()*dy*dy;
             if( dy<2.*2. ) totalE_road20+=rhit.GeV();
             if( dy<2.*4. ) totalE_road40+=rhit.GeV();
             if( dy<2.*6. ) totalE_road60+=rhit.GeV();
           }//in Y view
         }
     }//loop all cells

   }

    double efrac_plane2=-1.;
    double efrac_plane4=-1.;
    double efrac_plane6=-1.;
    double Eplane20=0.;
    double Eplane40=0.;
   int nplanes=PlaneSumEnergy.size();
   for( int i=0; i<nplanes; ++i ){
      if( i<20 ) Eplane20 += PlaneSumEnergy[i];
      if( i<40 ) Eplane40 += PlaneSumEnergy[i];

      //maximal energy fraction in 2 sliding window
      if( i<nplanes-2 ){
        double e_frac2=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1])/recoEnergy;
        if( efrac_plane2<e_frac2 ) efrac_plane2=e_frac2;        }
      else if( nplanes<=2 ) efrac_plane2=1.;
      //maximal energy fraction in 4 sliding window
      if( i<nplanes-4 ){
        double e_frac4=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3])/recoEnergy;
        if( efrac_plane4<e_frac4 ) efrac_plane4=e_frac4;        }
      else if( nplanes<=4 ) efrac_plane4=1.;
      //maximal energy fraction in 6 sliding window
      if( i<nplanes-6 ){
        double e_frac6=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3]+PlaneSumEnergy[i+4]+PlaneSumEnergy[i+5])/recoEnergy;
        if( efrac_plane6<e_frac6 ) efrac_plane6=e_frac6; }
      else if( nplanes<=6 ) efrac_plane6=1.;
    }

    if( nprongs>20 ) nprongs=20;
    Nprongs=nprongs;
    std::vector<double> ProngEnergy;
    std::vector<double> ProngPhi;
    std::vector<double> ProngPx;
    std::vector<double> ProngPy;
    std::vector<double> ProngPz;
    std::vector<double> ProngNcells;

    if( SelectedProng.size()>0 ){
    for(int ip=0; ip<nprongs; ++ip ){
      //reco direction
     TVector3 dir1 = SelectedProng[ip]->Dir();
     prongPhi[ip]=dir1.Phi();
     prongTheta[ip]=dir1.Theta();
     prongNCells[ip]=SelectedProng[ip]->NCell();

     ProngNcells.push_back(SelectedProng[ip]->NCell());

     //reco momentum
     double Eprong=0.;
     double EprongX=0.;
     double EprongY=0.;
     for( unsigned int icell=0; icell<SelectedProng[ip]->NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = SelectedProng[ip]->Cell(icell);
       const TVector3 mean = SelectedProng[ip]->MeanXYZ();
       const double wx = (SelectedProng[ip]->NXCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->XCell(0).get()) : 0;
       const double wy = (SelectedProng[ip]->NYCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->YCell(0).get()) : 0;

       const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

       if( !rhit.IsCalibrated() ) continue;
       Eprong += rhit.GeV();
       if( chit->View() == geo::kX ) EprongX += rhit.GeV();
       if( chit->View() == geo::kY ) EprongY += rhit.GeV();
     }
     ProngEnergy.push_back(Eprong);
     ProngPhi.push_back(dir1.Phi());
     ProngPx.push_back(Eprong*sin(dir1.Theta())*sin(dir1.Phi()));
     ProngPy.push_back(Eprong*sin(dir1.Theta())*cos(dir1.Phi()));
     ProngPz.push_back(Eprong*cos(dir1.Theta()));

     prongEnergy[ip]=Eprong;
    }
   }//has at least one prong

   double ratio=1.;
   double dphi=0.;
      int Index1=-1;
      int Index2=-1;
      double MaxEnergy1=-999.;
      double MaxEnergy2=-999.;
      for( int i=0;i<nprongs; ++i ){
         if( MaxEnergy1<ProngEnergy[i] ){
           MaxEnergy1=ProngEnergy[i];
           Index1=i;
        }
      }//leading
      for( int i=0;i<nprongs; ++i ){
         if( i==Index1 ) continue;
         if( MaxEnergy2<ProngEnergy[i] ){
           MaxEnergy2=ProngEnergy[i];
           Index2=i;
        }
      }//sub-leading
      double Eden=ProngEnergy[Index1]+ProngEnergy[Index2];
      double Enum=ProngEnergy[Index1]-ProngEnergy[Index2];
      if( Eden>0.0 ) ratio=Enum/Eden;
      dphi=fabs(ProngPhi[Index1]-ProngPhi[Index2]);
      if( dphi>2.*TMath::Pi() ) dphi=dphi-2.*TMath::Pi();
      if( dphi>TMath::Pi() ) dphi=2.*TMath::Pi()-dphi;

   if( Index1<0 ) return;
   if( Index2<0 ) return;

   TLorentzVector reco1(ProngPx[Index1],ProngPy[Index1],ProngPz[Index1],ProngEnergy[Index1]);
   TLorentzVector reco2(ProngPx[Index2],ProngPy[Index2],ProngPz[Index2],ProngEnergy[Index2]);

   double mgg_reco=(reco1+reco2).M();
   double a2=reco1.Angle(reco2.Vect());

    //fill root tree
    Ebalance=ratio;
    Dphi=dphi;
    Dtheta=a2;
    Mreco=mgg_reco;
    gam1_Ereco=ProngEnergy[Index1];
    gam1_ncells=ProngNcells[Index1];
    gam2_Ereco=ProngEnergy[Index2];
    gam2_ncells=ProngNcells[Index2];
     E40planes=Eplane40;
     Efrac_20planes=Eplane20/recoEnergy;
     Efrac_40planes=Eplane40/recoEnergy;
     Efrac_2slides=efrac_plane2;
     Efrac_4slides=efrac_plane4;
     Efrac_6slides=efrac_plane6;
     Efrac_2sigma=totalE_road20/recoEnergy;
     Efrac_4sigma=totalE_road40/recoEnergy;
     Efrac_6sigma=totalE_road60/recoEnergy;

     EDxDy=sum_edx+sum_edy;
     E2Dx2Dy2=sum_e2dx2+sum_e2dy2;

    recoVx=vert[0]->GetX();
    recoVy=vert[0]->GetY();
    recoVz=vert[0]->GetZ();     

    trackLength = longestTrack;
    trackEnergy = longestTrack_energy;
    trackNcells = longestTrack_ncells;
    trackEnergy_ratio = longestTrack_energy/recoEnergy;
    trackNcells_ratio = longestTrack_ncells/nCells;
    recoE = recoEnergy;
    Ncells = nCells;
    Nplanes = lengthPlanes;
    Nmip=nmip;
    Fmip=nmip/Ntotal;

    inputvars_pi0[0]=(Float_t)nCells;
    inputvars_pi0[1]=(Float_t)recoEnergy;
    inputvars_pi0[2]=(Float_t)longestTrack;
    inputvars_pi0[3]=(Float_t)nmip;
    inputvars_pi0[4]=(Float_t)efrac_plane2;
    inputvars_pi0[5]=(Float_t)efrac_plane6;
    inputvars_pi0[6]=(Float_t)totalE_road20/recoEnergy;
    inputvars_pi0[7]=(Float_t)dphi;
    inputvars_pi0[8]=(Float_t)a2;
    inputvars_pi0[9]=(Float_t)ratio;
    inputvars_pi0[10]=(Float_t)mgg_reco;
    inputvars_pi0[11]=(Float_t)ProngEnergy[Index1];
    inputvars_pi0[12]=(Float_t)ProngEnergy[Index2];
    inputvars_pi0[13]=(Float_t)ProngNcells[Index1];
    inputvars_pi0[14]=(Float_t)ProngNcells[Index2];

    pi0BDT=_pi0bdt->EvaluateMVA("BDT method");

    NNfunction NNcorr;
    double NNinput[4]={ProngEnergy[Index1],ProngEnergy[Index2],recoEnergy,a2};
    Mnncorr=NNcorr.Value(0,NNinput);

    _otree->Fill();

    return;
  }//end analyze

}//end namespace

namespace Ed{

  DEFINE_ART_MODULE(Mpi0);

}

