// \file    PCHitsList_module.cc
// \brief   Module to fill vector of PCHits, calibration hit info from cosmic files
// \version $Id: PCHitsList_module.cc,v 1.0 2013/01/13 18:20:04 rtoner Exp $
// \author  rtoner
// \date    $Date: 2013/01/13 22:06:10 $

//C++ and C:
#include <cmath>
#include <iostream>
#include <vector>
#include <utility>

//Framework:
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "cetlib/exception.h"
#include "messagefacility/MessageLogger/MessageLogger.h" 
#include "Utilities/AssociationUtil.h"

//NovaSoft:
#include "CalibrationDataProducts/PCHit.h"
#include "ChannelInfo/BadChanList.h"
#include "DAQChannelMap/DAQChannelMap.h"
#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Utilities/func/MathUtil.h"
#include "CalibrationUtils/CalibUtil.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/HitMap.h"
#include "RecoBase/Track.h"


namespace caldp { class PCHit; }

namespace calib { 
  
  class PCHitsList : public art::EDProducer {
    
  public:
    explicit PCHitsList(fhicl::ParameterSet const & pset);
    virtual ~PCHitsList();
    
    void produce(art::Event & e) override;
    
    void reconfigure(fhicl::ParameterSet const & p) override;
    
  private:
    void ProcessTrack(const rb::Track* track,
                      art::ServiceHandle<cheat::BackTracker> bt,
                      art::ServiceHandle<geo::Geometry> geom,
                      daqchannelmap::DAQChannelMap* cmap,
                      std::vector<caldp::PCHit>* pcHitsXY,
                      std::vector<caldp::PCHit>* pcHitsZ,
                      std::vector<caldp::PCHit>* pcHitsAvg) const;

    void ProcessTrackForBelowThresholdHits(const rb::Track* track,
                                           art::ServiceHandle<cheat::BackTracker> bt,
                                           art::ServiceHandle<geo::Geometry> geom,
                                           daqchannelmap::DAQChannelMap* cmap,
                                           std::vector<caldp::PCHit>* pcHitsXYThresh) const;

    void ProcessTrackForFLSHits(const rb::Track* track2,
                                art::ServiceHandle<cheat::BackTracker> bt,
                                art::ServiceHandle<geo::Geometry> geom,
                                daqchannelmap::DAQChannelMap* cmap,
                                std::vector<caldp::PCHit>* pcHitsXY) const;

    caldp::PCHit CellHitToPCHit(const rb::CellHit* chit,
                                const rb::Track* trk,
                                double path, double mev,
                                double truePE, double poissonLambda,
                                art::ServiceHandle<geo::Geometry> geom,
                                daqchannelmap::DAQChannelMap* cmap) const;

    std::string fCosmicLabel;
    std::string fQualXYName;
    std::string fQualZName;
    std::string fQualAvgName;
    std::string fQualXYThreshName;
    std::string fQualXYFLSName;

    double fPathCut;         ///< Only keep hits with pathlength less than this
    int    fExtentPlanesCut; ///< Only keep tracks covering more planes
    double fDirZCut;         ///< Only keep tracks shallower than this
    double fCompletenessCut; ///< Only keep tracks with more than this fraction of slice hits included in each view
  };
}

namespace calib { 

    //---------------------------------------------------------------------//    

    PCHitsList::PCHitsList(fhicl::ParameterSet const & pset)
    {
      reconfigure(pset);
      
      produces< std::vector<caldp::PCHit> >(fQualXYName);
      produces< std::vector<caldp::PCHit> >(fQualZName);
      produces< std::vector<caldp::PCHit> >(fQualAvgName);
      produces< std::vector<caldp::PCHit> >(fQualXYThreshName);
      produces< std::vector<caldp::PCHit> >(fQualXYFLSName);

      produces< art::Assns<caldp::PCHit, rb::Track> >(fQualXYName);
      produces< art::Assns<caldp::PCHit, rb::Track> >(fQualZName);
      produces< art::Assns<caldp::PCHit, rb::Track> >(fQualAvgName);
      produces< art::Assns<caldp::PCHit, rb::Track> >(fQualXYThreshName);
      produces< art::Assns<caldp::PCHit, rb::Track> >(fQualXYFLSName);
    }

    //---------------------------------------------------------------------//
    
    PCHitsList::~PCHitsList()
    {
    }

    //---------------------------------------------------------------------//    

    void PCHitsList::reconfigure(fhicl::ParameterSet const & pset)
    {
      fCosmicLabel  =   pset.get< std::string >("CosmicLabel"); //Label of tracker module
      fQualXYName       =   pset.get< std::string >("QualXYName"); //Instance label, "XY" quality hits
      fQualZName        =   pset.get< std::string >("QualZName"); //Instance label, "Z" quality hits
      fQualAvgName      =   pset.get< std::string >("QualAvgName"); //Instance label, "Avg" quality hits

      fQualXYThreshName =   pset.get< std::string >("QualXYThreshName"); //Instance label, "XY" quality hits falling below threshold
      fQualXYFLSName    =   pset.get< std::string >("QualXYFLSName"); //Instance label, "XY" quality hits from FLS

      fPathCut         = pset.get< double >("PathCut");
      fExtentPlanesCut = pset.get< int    >("ExtentPlanesCut");
      fDirZCut         = pset.get< double >("DirZCut");
      fCompletenessCut = pset.get< double >("CompletenessCut");
    }
    
    //---------------------------------------------------------------------//    
    
    void PCHitsList::produce(art::Event & e)
    {
      //Ptrs to vectors to be filled (one instance for each degree of hit quality)
      std::unique_ptr< std::vector<caldp::PCHit>  > pcHitsXY(new std::vector<caldp::PCHit>);
      std::unique_ptr< std::vector<caldp::PCHit>  > pcHitsZ(new std::vector<caldp::PCHit>);
      std::unique_ptr< std::vector<caldp::PCHit>  > pcHitsAvg(new std::vector<caldp::PCHit>);
      std::unique_ptr< std::vector<caldp::PCHit>  > pcHitsXYThresh(new std::vector<caldp::PCHit>);
      std::unique_ptr< std::vector<caldp::PCHit>  > pcHitsXYFLS(new std::vector<caldp::PCHit>);
      
      std::unique_ptr<art::Assns<caldp::PCHit, rb::Track> > hitAssnsXY(new art::Assns<caldp::PCHit, rb::Track>);
      std::unique_ptr<art::Assns<caldp::PCHit, rb::Track> > hitAssnsZ(new art::Assns<caldp::PCHit, rb::Track>);
      std::unique_ptr<art::Assns<caldp::PCHit, rb::Track> > hitAssnsAvg(new art::Assns<caldp::PCHit, rb::Track>);
      std::unique_ptr<art::Assns<caldp::PCHit, rb::Track> > hitAssnsXYThresh(new art::Assns<caldp::PCHit, rb::Track>);
      std::unique_ptr<art::Assns<caldp::PCHit, rb::Track> > hitAssnsXYFLS(new art::Assns<caldp::PCHit, rb::Track>);

      //Get cosmic tracks to use for calibration:
      art::Handle<std::vector<rb::Track> > trackcol;
      e.getByLabel(fCosmicLabel, trackcol);
      
      std::vector< art::Ptr<rb::Track> > tracks;
      art::fill_ptr_vector(tracks, trackcol);

      art::FindManyP<rb::Cluster> fmpslc(trackcol, e, fCosmicLabel);

      art::ServiceHandle<cheat::BackTracker> bt;
      art::ServiceHandle<geo::Geometry> geom;

      daqchannelmap::DAQChannelMap* cmap = daqchannelmap::DAQChannelMap::getInstance(geom->DetId());

      for (size_t itrk = 0; itrk < tracks.size(); ++itrk){

	// Find the slice associated with this track
	const std::vector<art::Ptr<rb::Cluster> > slices = fmpslc.at(itrk);
        assert(slices.size() == 1); // There should be exactly one
        const rb::Cluster& slice = *slices[0];

        // Remember how many hits of each type we had so far, so we know what
        // index to start associating from when we're done with this track.
        const unsigned int xyBeginIdx = pcHitsXY->size();
        const unsigned int zBeginIdx = pcHitsZ->size();
        const unsigned int avgBeginIdx = pcHitsAvg->size();
        const unsigned int xyThreshBeginIdx = pcHitsXYThresh->size();
        const unsigned int xyFLSBeginIdx = pcHitsXYFLS->size();

	//const rb::Track *track = &tracks->at(itrk);
	const art::Ptr<rb::Track> track = tracks.at(itrk);

	LOG_DEBUG("PCHitsList") << "TRACK " << itrk << ": " << track->NCell() << " cells to process.";
	
	//-------Track cuts:----------------//
	
	LOG_DEBUG("PCHitsList") << "NCellsX=" << track->NXCell() << " NCellsY=" << track->NYCell();

	//Track must have at least two X cells and 2 Y cells:
	if(track->NXCell() < 2 || track->NYCell() < 2) { 
	  LOG_DEBUG("PCHitsList") << "!---->Failed NCell Cut!"; 
	  continue; 
	}
	
	//Check that direction vector is well-defined:
	const TVector3 dr = track->Dir();
	const double dx = fabs(dr.X());
	const double dy = fabs(dr.Y());
	const double dz = fabs(dr.Z());

	//No undefined directions:
	if(isnan(dx) || isnan(dy) || isnan(dz)) continue;

	//At least one direction must be non-zero:
	if(dx == 0 && dy == 0 && dz == 0) {
	  LOG_DEBUG("PCHitsList") << "!---->Failed Direction Cut!"; 
	  continue;
	}

	//Must be a unit vector (otherwise, abort event, something went very wrong):
	if ( fabs( util::pythag(dx, dy, dz)-1 ) > 1e-4){
	  throw cet::exception("SkipEvent") << "ERROR: Direction was not a unit vector.  Skipping Event.\n";
	}

	// Additional quality cuts. These remove tracks that often have
	// badly-reconstructed W positions.

	// Cut very steep tracks two ways. These can be misreconstructed, or
	// they can simply make it difficult for the W estimation to do a good
	// job.
	if(int(track->ExtentPlane()) <= fExtentPlanesCut) continue;
	if(fabs(track->Dir().Z()) < fDirZCut) continue;

	// Cut out actual reconstruction failures, where the track doesn't use
	// a significant amount of the slice hits.  If either view is not
	// complete enough discard the whole thing
	if(track->NXCell() < fCompletenessCut*slice.NXCell()) continue;
	if(track->NYCell() < fCompletenessCut*slice.NYCell()) continue;
	
	//-----End track cuts-------------//

        ProcessTrack(track.get(), bt, geom, cmap,
                     pcHitsXY.get(), pcHitsZ.get(), pcHitsAvg.get());

        ProcessTrackForBelowThresholdHits(track.get(), bt, geom, cmap,
                                          pcHitsXYThresh.get());

        ProcessTrackForFLSHits(track.get(), bt, geom, cmap, pcHitsXYFLS.get());


        // Associate all the PCHits we created for this track. We do this in
        // bulk because making lots of individual calls as we go is slow.
        util::CreateAssn(*this, e, *pcHitsXY, track, *hitAssnsXY, xyBeginIdx, UINT_MAX, fQualXYName);
        util::CreateAssn(*this, e, *pcHitsZ, track, *hitAssnsZ, zBeginIdx, UINT_MAX, fQualZName);
        util::CreateAssn(*this, e, *pcHitsAvg, track, *hitAssnsAvg, avgBeginIdx, UINT_MAX, fQualAvgName);
        util::CreateAssn(*this, e, *pcHitsXYThresh, track, *hitAssnsXYThresh, xyThreshBeginIdx, UINT_MAX, fQualXYThreshName);
        util::CreateAssn(*this, e, *pcHitsXYFLS, track, *hitAssnsXYFLS, xyFLSBeginIdx, UINT_MAX, fQualXYFLSName);

      }//End track loop
      
      LOG_DEBUG("PCHitsList") << "SUMMARY: Size of XY vector  = " << pcHitsXY->size() 
                              << ", Size of Z vector   = " << pcHitsZ->size() 
                              << ", Size of Avg vector = " << pcHitsAvg->size();
      
      e.put(std::move(pcHitsXY),fQualXYName);
      e.put(std::move(pcHitsZ),fQualZName);
      e.put(std::move(pcHitsAvg),fQualAvgName);
      e.put(std::move(pcHitsXYThresh),fQualXYThreshName);
      e.put(std::move(pcHitsXYFLS),fQualXYFLSName);
      e.put(std::move(hitAssnsXY),fQualXYName);
      e.put(std::move(hitAssnsZ),fQualZName);
      e.put(std::move(hitAssnsAvg),fQualAvgName);
      e.put(std::move(hitAssnsXYThresh), fQualXYThreshName);
      e.put(std::move(hitAssnsXYFLS), fQualXYFLSName);
     
      LOG_DEBUG("PCHitsList") << "Event finished successfully.";

    }//end produce


    //---------------------------------------------------------------------//

    void PCHitsList::
    ProcessTrack(const rb::Track* track,
                 art::ServiceHandle<cheat::BackTracker> bt,
                 art::ServiceHandle<geo::Geometry> geom,
                 daqchannelmap::DAQChannelMap* cmap,
                 std::vector<caldp::PCHit>* pcHitsXY,
                 std::vector<caldp::PCHit>* pcHitsZ,
                 std::vector<caldp::PCHit>* pcHitsAvg) const
    {
      std::vector<rb::CellHit> chits;
      for(unsigned int ihit = 0; ihit < track->NCell(); ++ihit){
        chits.push_back(*track->Cell(ihit).get());
      }

      //Calculate path quality:
      std::vector<EPathQuality> quals;
      const std::vector<double> paths = BestPathEstimates(track->PlaneCellPairs(), track->Dir(), quals);

      //Check there is a path and quality measure for each cell; if not, something wrong, abort event:
      if (chits.size() != quals.size() ||
          chits.size() != paths.size()){
        throw cet::exception("SkipEvent") << "ERROR: Number of pathlengths or qualities != number of cells.  Skipping Event.\n";
      }

      //Loop through cells in track:
      for(unsigned int ihit = 0; ihit < chits.size(); ++ihit){
	  
        //Get hit
        const rb::CellHit& chit = chits[ihit];

        //Get path:
        const double path = paths[ihit]; //RBT: Way of ensuring cell and path line up??
        //Check path not too long:
        if(path > fPathCut) continue;

        //Get the correct MeV and true PE in this hit
        double mev = 0;
        double truePE = 0;
        double poissonLambda = 0;
        if (bt->HaveTruthInfo()){
          const std::vector<cheat::TrackIDE> ides = bt->HitToTrackIDE(track->Cell(ihit));
          for(unsigned int n = 0; n < ides.size(); ++n)
            mev += ides[n].energy*1000;

          const std::vector<sim::PhotonSignal> pes = bt->HitToPhotonSignal(track->Cell(ihit));
          for(unsigned int n = 0; n < pes.size(); ++n){
            truePE += pes[n].NPhoton();
            poissonLambda += pes[n].PoissonLambda();
          }
        }

        caldp::PCHit pchit = CellHitToPCHit(&chit, track,
                                            path, mev, truePE, poissonLambda,
                                            geom, cmap);

        switch(quals[ihit]){

        case kXYAdjacents:
          LOG_DEBUG("PCHitsList") << "Quality XY!  - hit=" << ihit << " FILLS TO -> plane=" << chit.Plane() 
                                  << " cell=" << chit.Cell() << " pe=" << chit.PE() << " path=" << path << " W=" << track->W(&chit) << " mev=" << mev;
          pcHitsXY->push_back(pchit);
          break;

        case kZAdjacents:
          LOG_DEBUG("PCHitsList") << "Quality Z!   - hit=" << ihit << " FILLS TO -> plane=" << chit.Plane() 
                                  << " cell=" << chit.Cell() << " pe=" << chit.PE() << " path=" << path << " W=" << track->W(&chit) << " mev=" << mev;
          pcHitsZ->push_back(pchit);
          break;

        case kAverage:
          LOG_DEBUG("PCHitsList") << "Quality Avg! - hit=" << ihit << " FILLS TO -> plane=" << chit.Plane() 
                                  << " cell=" << chit.Cell() << " pe=" << chit.PE() << " path=" << path << " W=" << track->W(&chit) << " mev=" << mev;
          pcHitsAvg->push_back(pchit);
          break;

        case kBadTrack:
          LOG_DEBUG("PCHitsList") << "!!!!!!BAD hit/track!!!!!! - hit=" << ihit;
          // Something wrong with this hit/track, so do nothing
          break;

        default:
          throw cet::exception("SkipEvent") << "ERROR: Unknown Path Quality type.  Skipping Event.\n";
          //Unknown path type; something went very wrong; abort this event.
        }
	
      }//End cell loop

    }//end ProcessTrack

    //---------------------------------------------------------------------//

    void PCHitsList::
    ProcessTrackForBelowThresholdHits(const rb::Track* track,
                                      art::ServiceHandle<cheat::BackTracker> bt,
                                      art::ServiceHandle<geo::Geometry> geom,
                                      daqchannelmap::DAQChannelMap* cmap,
                                      std::vector<caldp::PCHit>* pcHitsXYThresh) const
    {
      std::map<geo::OfflineChan, double> trueE;
      std::map<geo::OfflineChan, double> truePE;
      std::map<geo::OfflineChan, double> poissonLambda;
      if(bt->HaveTruthInfo()){
        const std::vector<const sim::Particle*> parts = bt->HitsToParticle(track->AllCells());
        if(!parts.empty()){
          const int id = parts[0]->TrackId(); // Biggest contributor
          std::set<geo::OfflineChan> flsChans;
          const std::vector<sim::FLSHit> flss = bt->ParticleToFLSHit(id);
          for(const sim::FLSHit& fls: flss){
            const geo::OfflineChan chan(fls.GetPlaneID(), fls.GetCellID());
            trueE[chan] += 1000*fls.GetEdep();
            flsChans.insert(chan);
          }

          for(geo::OfflineChan chan: flsChans){
            // We won't have an actual CellHit for this cell, so have to use
            // CellTo not HitTo. Could try to do something clever with matching
            // the times up. But this must be a small effect.
            const std::vector<sim::PhotonSignal> pes = bt->CellToPhotonSignal(chan.Plane(), chan.Cell());
            for(sim::PhotonSignal pe: pes){
              truePE[chan] += pe.NPhoton();
              poissonLambda[chan] += pe.PoissonLambda();
            }
          }
        }
      }

      // Find cells that would have made it into the sample if they'd been
      // hit. Downstream analyzers can interpret these as a threshold effect.
      std::vector<EPathQuality> threshQuals;
      std::vector<double> threshPaths;
      std::vector<geo::OfflineChan> threshChans;

      FindBelowThresholdCalibCandidates(track->PlaneCellPairs(), track->Dir(), threshQuals, threshPaths, threshChans);

      for(unsigned int threshIdx = 0; threshIdx < threshQuals.size(); ++threshIdx){
        const double path = threshPaths[threshIdx];
        // Check path not too long. Apply same cut as normal candidates so as
        // not to bias ourselves.
        if(path > 10) continue;

        const int plane = threshChans[threshIdx].Plane();
        const int cell = threshChans[threshIdx].Cell();
        const geo::OfflineChan chan(plane, cell);
        const geo::View_t view = geom->Plane(plane)->View();
        rb::CellHit dummy;
        dummy.SetPlane(plane);
        dummy.SetCell(cell);
        dummy.SetView(view);

        caldp::PCHit pchit = CellHitToPCHit(&dummy, track,
                                            path, trueE[chan],
                                            truePE[chan], poissonLambda[chan],
                                            geom, cmap);

        if(threshQuals[threshIdx] == kXYAdjacents)
          pcHitsXYThresh->push_back(pchit);
      } // end for threshIdx

    } // end ProcessTrackForBelowThresholdHits

    //---------------------------------------------------------------------//

    void PCHitsList::
    ProcessTrackForFLSHits(const rb::Track* track,
                           art::ServiceHandle<cheat::BackTracker> bt,
                           art::ServiceHandle<geo::Geometry> geom,
                           daqchannelmap::DAQChannelMap* cmap,
                           std::vector<caldp::PCHit>* pcHitsXY) const
    {
      if(!bt->HaveTruthInfo()) return;

      rb::HitMap hmap(track->AllCells());

      // BackTracker won't be able to figure out the true energy, since
      // there's no CellHit to work back from. Accumulate it here.
      std::map<geo::OfflineChan, double> extrasTruthMap;
      std::map<geo::OfflineChan, double> extrasPEMap;
      std::map<geo::OfflineChan, double> extrasPoissonLambda;
      std::vector<geo::OfflineChan> extras;

      const std::vector<const sim::Particle*> parts = bt->HitsToParticle(track->AllCells());
      if(!parts.empty()){
        const int id = parts[0]->TrackId(); // Biggest contributor
        const std::vector<sim::FLSHit> flss = bt->ParticleToFLSHit(id);
        const unsigned int nFLS = flss.size();
        std::set<geo::OfflineChan> extraset;
        std::set<geo::OfflineChan> flsCells;
        for(unsigned int flsIdx = 0; flsIdx < nFLS; ++flsIdx){
          const sim::FLSHit& fls = flss[flsIdx];
          const geo::OfflineChan chan(fls.GetPlaneID(), fls.GetCellID());
          extraset.insert(chan);
          extrasTruthMap[chan] += 1000*fls.GetEdep();
          flsCells.insert(chan);
        }

        for(geo::OfflineChan chan: flsCells){
          // We might not have an actual CellHit for this cell, so have to use
          // CellTo not HitTo. Could try to do something clever with matching
          // the times up. But this must be a small effect.
          const std::vector<sim::PhotonSignal> pes = bt->CellToPhotonSignal(chan.Plane(), chan.Cell());
          for(sim::PhotonSignal pe: pes){
            extrasPEMap[chan] += pe.NPhoton();
            extrasPoissonLambda[chan] += pe.PoissonLambda();
          }
        }

        // extraset is a set to get a collection of distinct cells, but
        // later on we need it to be a vector.
        extras.insert(extras.end(), extraset.begin(), extraset.end());
      }


      const double meant = track->MeanTNS();

      rb::Track trackNoHits(*track);
      trackNoHits.Clear();

      //Calculate path quality
      std::vector<EPathQuality> quals;
      const std::vector<double> paths = BestPathEstimates(trackNoHits.PlaneCellPairs(), trackNoHits.Dir(), quals, extras);

      //Loop through cells in track:
      for(unsigned int ihit = 0; ihit < extras.size(); ++ihit){
        const geo::OfflineChan chan = extras[ihit];

        //Get path:
        const double path = paths[ihit];
        //Check path not too long:
        if(path > 10) continue;

        if(quals[ihit] == kXYAdjacents){
          //Get the correct MeV in this hit
          const double mev = extrasTruthMap[chan];
          const double truePE = extrasPEMap[chan];
          const double poissonLambda = extrasPoissonLambda[chan];

          rb::CellHit chit;
          chit.SetPlane(chan.Plane());
          chit.SetCell(chan.Cell());
          chit.SetView(geom->Plane(chan.Plane())->View());
          if(hmap.CellExists(chan.Plane(), chan.Cell())){
            chit.SetPE(hmap.Cell(chan.Plane(), chan.Cell())->PE());
          }
          else{
            chit.SetPE(0);
          }
          chit.SetTNS(meant, true);

          caldp::PCHit pchit = CellHitToPCHit(&chit, track,
                                              path, mev, truePE, poissonLambda,
                                              geom, cmap);

          pcHitsXY->push_back(pchit);
        }
      }//End cell loop

    }//end ProcessTrack

    //---------------------------------------------------------------------//

    caldp::PCHit PCHitsList::CellHitToPCHit(const rb::CellHit* chit,
                                            const rb::Track* trk,
                                            double path, double mev, double truePE, double poissonLambda,
                                            art::ServiceHandle<geo::Geometry> geom,
                                            daqchannelmap::DAQChannelMap* cmap) const
    {
      //get more information on hit location
      uint32_t lchan   = cmap->encodeLChan(geom->DetId(), chit->Plane(), chit->Cell());
      uint32_t dchan   = cmap->encodeDChan(lchan);
      uint32_t pixel   = cmap->getPixel(dchan);
      uint32_t apd     = cmap->getFEB(dchan);
      uint32_t dcm     = cmap->getDCM(dchan);
      uint32_t diblock = cmap->getDiBlock(dchan);

      double w = trk->W(chit);

      //get distance to readout electronics
      double xyzd[4];
      GetXYZD(chit->PlaneCellPair(), w, xyzd);
      //make sure the reconstructed hit coordinate is inside the physical
      //detector if not then change the distance to the electronics to -1.0
      if ((xyzd[0] < -geom->DetHalfWidth()) || (xyzd[0] > geom->DetHalfWidth()) || 
          (xyzd[1] < -geom->DetHalfHeight()) || (xyzd[1] > geom->DetHalfHeight()) ||
          (xyzd[3] < 0.0) ) xyzd[3] = -1.0;

      double flightlen = (trk->Start()-TVector3(xyzd)).Mag();

      //Define and fill a PCHit for this cell hit
      caldp::PCHit pchit;
      pchit.SetPlane    ( chit->Plane()    ); //Set plane number
      pchit.SetCell     ( chit->Cell()     ); //Set cell number
      pchit.SetPixel    ( pixel            ); //Set pixel number
      pchit.SetAPD      ( apd              ); //Set apd number
      pchit.SetDCM      ( dcm              ); //Set dcm number
      pchit.SetDiblock  ( diblock          ); //set diblock number
      pchit.SetView     ( chit->View()     ); //set hit view
      pchit.SetPE       ( chit->PE()       ); //Set uncalibrated PE
      pchit.SetPath     ( path             ); //Set path length
      pchit.SetW        ( w                ); //Set W (unmeasured coordinate) value
      pchit.SetTrueMeV  ( mev              ); //Set true energy of hit 
      pchit.SetTruePE   ( truePE           ); //Set true PE of hit 
      pchit.SetPoissonLambda(poissonLambda );
      pchit.SetTNS      ( chit->TNS()      ); //Set time of hit
      pchit.SetFlightLen( flightlen        ); //Set path from start of track to hit
      pchit.SetReadDist ( xyzd[3]          ); //Set distance from hit to readout electronics
      pchit.SetGoodTime ( chit->GoodTiming()); //Set if cell had a proper time fit

      return pchit;
    } // end CellHitToPCHit
  
} //end namespace

//---------------------------------------------------------------------//    

DEFINE_ART_MODULE(calib::PCHitsList)
