
#include "Calibrator/Calibrator.h"

// ROOT includes
#include "TVector3.h"
#include "TH1I.h"
#include "TH1F.h"
#include "TMath.h"
#include "TNtuple.h"
#include "TTree.h"
#include "TLorentzVector.h"
#include "TH2F.h"
#include "TProfile.h"

// NOvA includes
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/ParticleNavigator.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/RecoHit.h"

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>

class TVector3;
class TH1F;
class TH2F;
class TTree;
class TProfile;

namespace Ed {
  class SingleParticle : public art::EDAnalyzer {
  public:
    explicit SingleParticle(fhicl::ParameterSet const& pset);
    virtual ~SingleParticle();

    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);

  private:
    int inputPDG;//PDG for the used single particle

    //histograms
    TH1F *Egen;//input generator energy
    TH1F *Etrue;//sum energy for true hits energy deposit
    TH1F *Ereco;//sum energy for reco hits
    TH1F *Eratio_gen_true;//energy ratio for Egen/Etrue
    TH1F *Eratio_true_reco;//energy ratio for Etrue/Ereco
    TH1F *Eratio_gen_reco;//energy ratio for Egen/Ereco
    TH2F *Egen_vs_Ereco;
    TH2F *Etrue_vs_Ereco;
    TH2F *Egen_vs_Etrue;
    TH1F *Nslice;//# of selected slices

    TProfile *Eratio_gen_true_vx;//Egen/Etrue vs. vertex X
    TProfile *Eratio_gen_true_vy;//Egen/Etrue vs. vertex Y
    TProfile *Eratio_gen_true_vz;//Egen/Etrue vs. vertex Z
    TProfile *Eratio_gen_true_Etrue;//Egen/Etrue vs. Etrue

    TProfile *Eratio_true_reco_vx;//Etrue/Ereco vs. vertex X
    TProfile *Eratio_true_reco_vy;//Etrue/Ereco vs. vertex Y
    TProfile *Eratio_true_reco_vz;//Etrue/Ereco vs. vertex Z
    TProfile *Eratio_true_reco_Ereco;//Etrue/Ereco vs. Ereco

    TProfile *Eratio_gen_reco_vx;//Egen/Ereco vs. vertex X
    TProfile *Eratio_gen_reco_vy;//Egen/Ereco vs. vertex Y
    TProfile *Eratio_gen_reco_vz;//Egen/Ereco vs. vertex Z
    TProfile *Eratio_gen_reco_Ereco;//Egen/Ereco vs. Ereco

    TProfile *Eratio_gen_true_dx_Xview;//Egen/Etrue vs. dx in X view
    TProfile *Eratio_gen_true_dx_Yview;//Egen/Etrue vs. dx in Y view
    TProfile *Eratio_gen_reco_dx_Xview;//Egen/Ereco vs. dx in X view
    TProfile *Eratio_gen_reco_dx_Yview;//Egen/Ereco vs. dx in Y view
    TProfile *Eratio_true_reco_dx_Xview;//Etrue/Ereco vs. dx in X view
    TProfile *Eratio_true_reco_dx_Yview;//Etrue/Ereco vs. dx in Y view

    TProfile *Eratio_gen_true_dy_Xview;//Egen/Etrue vs. dy in X view
    TProfile *Eratio_gen_true_dy_Yview;//Egen/Etrue vs. dy in Y view
    TProfile *Eratio_gen_reco_dy_Xview;//Egen/Ereco vs. dy in X view
    TProfile *Eratio_gen_reco_dy_Yview;//Egen/Ereco vs. dy in Y view
    TProfile *Eratio_true_reco_dy_Xview;//Etrue/Ereco vs. dy in X view
    TProfile *Eratio_true_reco_dy_Yview;//Etrue/Ereco vs. dy in Y view

    //save even info to ntuples
    TTree *_otree;
    Double_t Vx;//true vertex X
    Double_t Vy;//true vertex Y
    Double_t Vz;//true vertex Z
    Double_t recoVx;//reco vertex X
    Double_t recoVy;//reco vertex Y
    Double_t recoVz;//reco vertex Z
    Double_t truePx;//true Px
    Double_t truePy;//true Py
    Double_t truePz;//true Pz
    Double_t trueE;//true E
    Int_t nslice;//# of selected slices

    Double_t trueEntry;//input energy of true hits
    Double_t trueExit;//output energy of true hits
    Double_t trueDeposit;//deposit energy of true hits

    Int_t goodcells;//# of good cells
    Double_t Xcell[5000];//cell X
    Double_t Ycell[5000];//cell Y
    Double_t Zcell[5000];//cell Z
    Double_t ecell[5000];//cell energy
    Int_t inXVIEW[5000];//cell energy
  };
}//end namespace


namespace Ed{

  //......................................................................
  SingleParticle::SingleParticle(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    std::cout<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  SingleParticle::~SingleParticle()
  {
  }
  
  //......................................................................
  void SingleParticle::reconfigure(const fhicl::ParameterSet& pset)
  {
    inputPDG = pset.get<int>("pdg"); 
  }

  //......................................................................
  void SingleParticle::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    

    //histo
    Ereco=tfs->make<TH1F>("Ereco","Ereco;Ereco;# of events",100,0.,5.);
    Egen=tfs->make<TH1F>("Egen","Egen;Egen;# of events",100,0.,5.);
    Etrue=tfs->make<TH1F>("Etrue","Etrue;Etrue;# of events",100,0.,5.);
    Eratio_gen_true=tfs->make<TH1F>("Eratio_gen_true","Egen/Etrue;Egen/Etrue;# of events",100,0,5);
    Eratio_gen_reco=tfs->make<TH1F>("Eratio_gen_reco","Egen/Ereco;Egen/Ereco;# of events",100,0,5);
    Eratio_true_reco=tfs->make<TH1F>("Eratio_true_reco","Etrue/Ereco;Etrue/Ereco;# of events",100,0,5);
    Egen_vs_Ereco=tfs->make<TH2F>("Egen_vs_Ereco","Egen_vs_Ereco;Ereco;Egen",100,0,5,100,0,5);
    Egen_vs_Etrue=tfs->make<TH2F>("Egen_vs_Etrue","Egen_vs_Etrue;Etrue;Egen",100,0,5,100,0,5);
    Etrue_vs_Ereco=tfs->make<TH2F>("Etrue_vs_Ereco","Etrue_vs_Ereco;Ereco;Etrue",100,0,5,100,0,5);
    Nslice=tfs->make<TH1F>("Nslice","Nslice;# of slices;# of events",10,0.,10.);
    
    Eratio_gen_true_vx=tfs->make<TProfile>("Eratio_gen_true_vx","Eratio_gen_true_vx;Vx (cm);Egen/Etrue",50,-700,700,0,10);
    Eratio_gen_true_vy=tfs->make<TProfile>("Eratio_gen_true_vy","Eratio_gen_true_vy;Vy (cm);Egen/Etrue",50,-700,700,0,10);
    Eratio_gen_true_vz=tfs->make<TProfile>("Eratio_gen_true_vz","Eratio_gen_true_vz;Vz (cm);Egen/Etrue",50,0,6000,0,10);
    Eratio_gen_true_Etrue=tfs->make<TProfile>("Eratio_gen_true_Ereco","Eratio_gen_true_Etrue;Etrue (GeV);Egen/Etrue",100,0,5,0,10);
    
    Eratio_gen_reco_vx=tfs->make<TProfile>("Eratio_gen_reco_vx","Eratio_gen_reco_vx;Vx (cm);Egen/Ereco",50,-700,700,0,10);
    Eratio_gen_reco_vy=tfs->make<TProfile>("Eratio_gen_reco_vy","Eratio_gen_reco_vy;Vy (cm);Egen/Ereco",50,-700,700,0,10);
    Eratio_gen_reco_vz=tfs->make<TProfile>("Eratio_gen_reco_vz","Eratio_gen_reco_vz;Vz (cm);Egen/Ereco",50,0,6000,0,10);
    Eratio_gen_reco_Ereco=tfs->make<TProfile>("Eratio_gen_reco_Ereco","Eratio_gen_reco_Ereco;Ereco (GeV);Egen/Ereco",100,0,5,0,10);

    Eratio_true_reco_vx=tfs->make<TProfile>("Eratio_true_reco_vx","Eratio_true_reco_vx;Vx (cm);Etrue/Ereco",50,-700,700,0,10);
    Eratio_true_reco_vy=tfs->make<TProfile>("Eratio_true_reco_vy","Eratio_true_reco_vy;Vy (cm);Etrue/Ereco",50,-700,700,0,10);
    Eratio_true_reco_vz=tfs->make<TProfile>("Eratio_true_reco_vz","Eratio_true_reco_vz;Vz (cm);Etrue/Ereco",50,0,6000,0,10);
    Eratio_true_reco_Ereco=tfs->make<TProfile>("Eratio_true_reco_Ereco","Eratio_true_reco_Ereco;Ereco (GeV);Etrue/Ereco",100,0,5,0,10);

    Eratio_gen_true_dx_Xview=tfs->make<TProfile>("Eratio_gen_true_dx_Xview","Egen/Etrue vs #Delta X in X view;#Delta X (cm);Egen/Etrue",50,0,1500,0,10);
    Eratio_gen_true_dx_Yview=tfs->make<TProfile>("Eratio_gen_true_dx_Yview","Egen/Etrue vs #Delta X in Y view;#Delta X (cm);Egen/Etrue",50,0,1500,0,10);
    Eratio_gen_true_dy_Xview=tfs->make<TProfile>("Eratio_gen_true_dy_Xview","Egen/Etrue vs #Delta Y in X view;#Delta Y (cm);Egen/Etrue",50,0,1500,0,10);
    Eratio_gen_true_dy_Yview=tfs->make<TProfile>("Eratio_gen_true_dy_Yview","Egen/Etrue vs #Delta Y in Y view;#Delta Y (cm);Egen/Etrue",50,0,1500,0,10);

    Eratio_gen_reco_dx_Xview=tfs->make<TProfile>("Eratio_gen_reco_dx_Xview","Egen/Ereco vs #Delta X in X view;#Delta X (cm);Egen/Ereco",50,0,1500,0,10);
    Eratio_gen_reco_dx_Yview=tfs->make<TProfile>("Eratio_gen_reco_dx_Yview","Egen/Ereco vs #Delta X in Y view;#Delta X (cm);Egen/Ereco",50,0,1500,0,10);
    Eratio_gen_reco_dy_Xview=tfs->make<TProfile>("Eratio_gen_reco_dy_Xview","Egen/Ereco vs #Delta Y in X view;#Delta Y (cm);Egen/Ereco",50,0,1500,0,10);
    Eratio_gen_reco_dy_Yview=tfs->make<TProfile>("Eratio_gen_reco_dy_Yview","Egen/Ereco vs #Delta Y in Y view;#Delta Y (cm);Egen/Ereco",50,0,1500,0,10);

    Eratio_true_reco_dx_Xview=tfs->make<TProfile>("Eratio_true_reco_dx_Xview","Etrue/Ereco vs #Delta X in X view;#Delta X (cm);Etrue/Ereco",50,0,1500,0,10);
    Eratio_true_reco_dx_Yview=tfs->make<TProfile>("Eratio_true_reco_dx_Yview","Etrue/Ereco vs #Delta X in Y view;#Delta X (cm);Etrue/Ereco",50,0,1500,0,10);
    Eratio_true_reco_dy_Xview=tfs->make<TProfile>("Eratio_true_reco_dy_Xview","Etrue/Ereco vs #Delta Y in X view;#Delta Y (cm);Etrue/Ereco",50,0,1500,0,10);
    Eratio_true_reco_dy_Yview=tfs->make<TProfile>("Eratio_true_reco_dy_Yview","Etrue/Ereco vs #Delta Y in Y view;#Delta Y (cm);Etrue/Ereco",50,0,1500,0,10);

    //personal tuples
    _otree = tfs->make<TTree>("calibration","energy calibration");
    _otree->Branch("Vx",&Vx,"Vx/D");
    _otree->Branch("Vy",&Vy,"Vy/D");
    _otree->Branch("Vz",&Vz,"Vz/D");
    _otree->Branch("recoVx",&recoVx,"recoVx/D");
    _otree->Branch("recoVy",&recoVy,"recoVy/D");
    _otree->Branch("recoVz",&recoVz,"recoVz/D");
    _otree->Branch("truePx",&truePx,"truePx/D");
    _otree->Branch("truePy",&truePy,"truePy/D");
    _otree->Branch("truePz",&truePz,"truePz/D");
    _otree->Branch("trueE",&trueE,"trueE/D");
   
    _otree->Branch("trueEntry",&trueEntry,"trueEntry/D");    
    _otree->Branch("trueExit",&trueExit,"trueExit/D");    
    _otree->Branch("trueDeposit",&trueDeposit,"trueDeposit/D");    
    _otree->Branch("goodcells",&goodcells,"goodcells/I");
    _otree->Branch("Xcell",&Xcell,"Xcell[goodcells]/D");
    _otree->Branch("Ycell",&Ycell,"Ycell[goodcells]/D");
    _otree->Branch("Zcell",&Zcell,"Zcell[goodcells]/D");
    _otree->Branch("ecell",&ecell,"ecell[goodcells]/D");
    _otree->Branch("inXVIEW",&inXVIEW,"inXVIEW[goodcells]/I");
 
  }

  
  void SingleParticle::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<cheat::BackTracker> bt;

    //++Ntotal; 
    //get MC true info
    const sim::ParticleNavigator pnav = bt->ParticleNavigator();

    //select single photon
    TLorentzVector gamma1;
    double vx=0.;
    double vy=0.;
    double vz=0.;
    sim::Particle part = *(pnav.Particle(0));
    if( part.PdgCode() != inputPDG ){
      mf::LogWarning("SingleParticle")<<"The first particle PDG is not "<<inputPDG<<" !";
      return;
    }
    gamma1.SetPxPyPzE(part.Px(),part.Py(),part.Pz(),part.E());
    vx=part.Vx();
    vy=part.Vy();
    vz=part.Vz();

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel("slicer", slices);

    int bestSlice = -1;
    unsigned int mostCells = 0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;
      if(slice.NCell() > mostCells){        
        mostCells = slice.NCell();
        bestSlice = sliceIdx;
      }
    }
    if(bestSlice < 0) return;
    const rb::Cluster& slice = (*slices)[bestSlice];

    //true vertex
    Vx=vx;
    Vy=vy;
    Vz=vz;
    //true momentum
    truePx=gamma1.Px();
    truePy=gamma1.Py();
    truePz=gamma1.Pz();
    trueE=gamma1.E();

    Nslice->Fill(slices->size());
    double recoEnergy=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

       if( !rhit.IsCalibrated() ) continue;
       recoEnergy += rhit.GeV();
    }
  
    //select the true hits
    art::Handle< std::vector<sim::FLSHitList> > flslistHandle;
    evt.getByLabel("geantgen", flslistHandle);
    double eEntry=0.;
    double eExit=0.;
    double eDeposit=0.;
    if( flslistHandle.failedToGet() ) return;
    //for( size_t l = 0; l < flslistHandle->size(); ++l){
      const std::vector<sim::FLSHit> &hits = flslistHandle->at(0).fHits;
      for( size_t f = 0; f < hits.size(); ++f){
        //if( hits[f].GetPDG() != 22 ) continue;
        eEntry += hits[f].GetEntryEnergy();
        eExit += hits[f].GetExitEnergy();
        eDeposit += hits[f].GetEdep();
      }
     trueEntry = eEntry;
     trueExit = eExit;
     trueDeposit = eDeposit;

    double ratio1=gamma1.E()/recoEnergy;//Egen/Ereco
    double ratio2=gamma1.E()/trueDeposit;//Egen/Etrue
    double ratio3=trueDeposit/recoEnergy;//Etrue/Ereco

    Egen->Fill(gamma1.E());
    Ereco->Fill(recoEnergy);
    Etrue->Fill(trueDeposit);
    Eratio_gen_true->Fill(ratio2);
    Eratio_true_reco->Fill(ratio3);
    Eratio_gen_reco->Fill(ratio1);
    
    Egen_vs_Ereco->Fill(recoEnergy,gamma1.E());
    Egen_vs_Etrue->Fill(trueDeposit,gamma1.E());
    Etrue_vs_Ereco->Fill(recoEnergy,trueDeposit);

    Eratio_gen_reco_vx->Fill(vx,ratio1);
    Eratio_gen_reco_vy->Fill(vy,ratio1);
    Eratio_gen_reco_vz->Fill(vz,ratio1);
    Eratio_gen_reco_Ereco->Fill(recoEnergy,ratio1);

    Eratio_gen_true_vx->Fill(vx,ratio2);
    Eratio_gen_true_vy->Fill(vy,ratio2);
    Eratio_gen_true_vz->Fill(vz,ratio2);
    Eratio_gen_true_Etrue->Fill(trueDeposit,ratio2);

    Eratio_true_reco_vx->Fill(vx,ratio3);
    Eratio_true_reco_vy->Fill(vy,ratio3);
    Eratio_true_reco_vz->Fill(vz,ratio3);
    Eratio_true_reco_Ereco->Fill(recoEnergy,ratio3);

    int ncells=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

       if( !rhit.IsCalibrated() ) continue;
       if( icell<5000 ){
         ncells++;
         Xcell[icell]=rhit.X();
         Ycell[icell]=rhit.Y();
         Zcell[icell]=rhit.Z();
         ecell[icell]=rhit.GeV();
         if( chit->View()==geo::kX ) inXVIEW[icell]=1;
         else inXVIEW[icell]=0;
       }
      //distance from readout to each cell position
      double dX = 1549.4/2.-rhit.X();
      double dY = 1549.2/2.-rhit.Y();
      
      if( chit->View() == geo::kX ){ //in X view
         Eratio_gen_reco_dx_Xview->Fill(dX,ratio1);
         Eratio_gen_reco_dy_Xview->Fill(dY,ratio1);
         Eratio_gen_true_dx_Xview->Fill(dX,ratio2);
         Eratio_gen_true_dy_Xview->Fill(dY,ratio2);
         Eratio_true_reco_dx_Xview->Fill(dX,ratio3);
         Eratio_true_reco_dy_Xview->Fill(dY,ratio3);
       }
      else if( chit->View() == geo::kY ){//in Y view
         Eratio_gen_reco_dx_Yview->Fill(dX,ratio1);
         Eratio_gen_reco_dy_Yview->Fill(dY,ratio1);
         Eratio_gen_true_dx_Yview->Fill(dX,ratio2);
         Eratio_gen_true_dy_Yview->Fill(dY,ratio2);
         Eratio_true_reco_dx_Yview->Fill(dX,ratio3);
         Eratio_true_reco_dy_Yview->Fill(dY,ratio3);
       }
   
     }
    goodcells=ncells;

    _otree->Fill();

    return;
  }//end analyze
  
}//end namespace

namespace Ed{

  DEFINE_ART_MODULE(SingleParticle);

}

