////////////////////////////////////////////////////////////////////////
// \file    StopperSelection.cxx
// \brief   Selects stopping cosmics
// \author  bckhouse@caltech.edu, bays@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

// NOvASoft includes
#include "MCCheater/BackTracker.h"
#include "MichelEFilter/MichelECluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Track.h"
#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Utilities/AssociationUtil.h"

namespace calib
{
  /// Selects stopping cosmics
  class StopperSelection: public art::EDFilter
  {
  public:
    explicit StopperSelection(const fhicl::ParameterSet& pset);
    ~StopperSelection();

    virtual bool filter(art::Event& evt);

    void reconfigure(const fhicl::ParameterSet& pset);

    void beginJob();
    void endJob();

  protected:
    bool IsContainedStopper(const rb::Track* trk, const me::MichelECluster& me) const;
    bool IsTrulyContainedStopper(const rb::Track* trk) const;

    std::string fMichelLabel;

    bool fApplyFilter;

    int fTable[2][2];
  };

  //......................................................................
  StopperSelection::StopperSelection(const fhicl::ParameterSet& pset)
  {
    produces< std::vector<rb::Track> >();
    produces< art::Assns<rb::Track, rb::Cluster> >();

    reconfigure(pset);

    for(int i = 0; i < 2; ++i) for(int j = 0; j < 2; ++j) fTable[i][j] = 0;
  }

  //......................................................................
  StopperSelection::~StopperSelection()
  {
  }

  //......................................................................
  void StopperSelection::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMichelLabel = pset.get<std::string>("MichelLabel");
    fApplyFilter = pset.get<bool>("ApplyFilter");
  }

  //......................................................................
  void StopperSelection::beginJob()
  {
  }

  //......................................................................
  bool StopperSelection::filter(art::Event& evt)
  {
    std::unique_ptr<std::vector<rb::Track> > outTracks(new std::vector<rb::Track>);
    std::unique_ptr<art::Assns<rb::Track, rb::Cluster> > assns(new art::Assns<rb::Track, rb::Cluster>);

    art::Handle<std::vector<me::MichelECluster> > meinfo;
    evt.getByLabel(fMichelLabel, meinfo);

    art::FindManyP<rb::Cluster> fmpslc(meinfo, evt, fMichelLabel);

    for(size_t meIdx = 0; meIdx < meinfo->size(); ++meIdx){
      const rb::Track* track = (*meinfo)[meIdx].ParentTrack().get();

      const bool stopper = IsContainedStopper(track, (*meinfo)[meIdx]);      

      if(stopper){
	outTracks->push_back(*track);
 
	// Find the slice the original track was associated to
	const std::vector<art::Ptr<rb::Cluster> > slices = fmpslc.at(meIdx);
	assert(slices.size() == 1);
	util::CreateAssn(*this, evt, *outTracks, slices[0], *assns);
      }

      if(!evt.isRealData()){
	const bool trueStopper = IsTrulyContainedStopper(track);
	++fTable[stopper][trueStopper];
      }
    } // end for meIdx

    const bool ret = !outTracks->empty();
    evt.put(std::move(outTracks));
    evt.put(std::move(assns));
    if(fApplyFilter) return ret; else return true;
  }

  //......................................................................
  void StopperSelection::endJob()
  {
    if(fTable[0][0]+fTable[1][0] == 0 || fTable[0][1]+fTable[1][1] == 0) return;
    std::cout << "Stopper efficiency table. x-axis selection, y-axis truth:" << std::endl;
    std::cout << fTable[0][0] << "\t" << fTable[1][0] << "\t" << .1*(1000*fTable[0][0]/(fTable[0][0]+fTable[1][0])) << "%" << std::endl;
    std::cout << fTable[0][1] << "\t" << fTable[1][1] << "\t" << .1*(1000*fTable[1][1]/(fTable[0][1]+fTable[1][1])) << "%" << std::endl;
    std::cout << .1*(1000*fTable[0][0]/(fTable[0][0]+fTable[0][1])) << "%\t" << .1*(1000*fTable[1][1]/(fTable[1][0]+fTable[1][1])) << "%" << std::endl;
    std::cout << "Purity = " << .1*(1000*fTable[1][1]/(fTable[1][0]+fTable[1][1])) << "%" << std::endl;
    std::cout << "Efficieny = " << .1*(1000*fTable[1][1]/(fTable[0][1]+fTable[1][1])) << "%" << std::endl;
  }

  //......................................................................
  bool StopperSelection::
  IsContainedStopper(const rb::Track* trk, const me::MichelECluster& me) const
  {
    // Start and end don't mean much, pick whichever is lowest in y.
    const TVector3 v1 = trk->Start();
    const TVector3 v2 = trk->Stop();
    const TVector3 stop = (v1.Y() < v2.Y()) ? v1 : v2;

    const double x = stop.X();
    const double y = stop.Y();
    const double z = stop.Z();

    art::ServiceHandle<geo::Geometry> geom;
    switch(geom->DetId()){
    case novadaq::cnv::kNDOS:
      if(fabs(x) > 110) return false; // Fiducial X
      // Fall through
    case novadaq::cnv::kNEARDET:
      if(fabs(x) > 180) return false; // Fiducial X
      if(fabs(y) > 180) return false; // Fiducial Y
      if(z < 20 || z > 1400) return false; // Fiducial Z
      if(me.DeltaT() > 12000) return false; // time difference
      if(me.TotalGeV() > 0.035) return false; // total Michel-e charge
      if(v1.X()==v2.X()||v1.Y()==v2.Y()) return false; // dx or dy = 0 means bad fit
      break;
    case novadaq::cnv::kFARDET:
      if(fabs(x) > 730 || fabs(y) > 730) return false;
      if(z < 50 || z > 6300) return false;
      if(me.DeltaT() > 12000) return false;
      if(me.TotalGeV() > 0.035) return false;
      if(v1.X()==v2.X()||v1.Y()==v2.Y()) return false;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    return true;
  }

  //......................................................................
  bool StopperSelection::
  IsTrulyContainedStopper(const rb::Track* trk) const
  {
    art::ServiceHandle<cheat::BackTracker> bt;

    std::vector<const sim::Particle*> parts = bt->HitsToParticle(trk->AllCells());
    if(parts.empty()) return false;

    const sim::Particle* part = parts.front();

    const TVector3 stop = part->EndPosition().Vect();

    art::ServiceHandle<geo::Geometry> geom;

    if(fabs(stop.X()) > geom->DetHalfWidth()) return false;
    if(fabs(stop.Y()) > geom->DetHalfHeight()) return false;
    if(stop.Z() < 0) return false;
    if(stop.Z() > geom->DetLength()) return false;
    if(abs(part->PdgCode())!=13) return false;

    return true;
  }

  DEFINE_ART_MODULE(StopperSelection);

} // end namespace
////////////////////////////////////////////////////////////////////////

