////////////////////////////////////////////////////////////////////////
/// \file    Calibrator_service.cc
/// \brief   Service that handles calibrations for a particular
///          validity context (eg: detector+timestamp).
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////
#include "Calibrator/Calibrator.h"
#include "RecoBaseHit/CellHit.h"
#include "RecoBaseHit/RecoHit.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Calibrator/func/AbsCache.h"
#include "Calibrator/func/ADCShapeFit.h"
#include "Calibrator/func/AttenCurve.h"
#include "Calibrator/func/DriftCache.h"
#include "Calibrator/func/TimingCache.h"

#include "CalibrationUtils/CalibUtil.h"

#include "DAQDataFormats/NanoSliceVersionConvention.h"

#include "GeometryObjects/CellGeo.h"

namespace calib 
{
  // Number of points we expect to receive, to find the best peak match with
  const unsigned int kNumFineTimingADCPoints = 4; // Including the DCS origin

  //------------------------------------------------------------
  Calibrator::Calibrator(fhicl::ParameterSet const& pset, 
			 art::ActivityRegistry& reg)
    : fCurrTimeStamp(0)
    , fShapeTable(0)
    , fAttenCache(new AttenCache)
    , fDriftCache(new DriftCache)
    , fAbsCache(new AbsCache)
    , fTimingCache(new TimingCache)
    , fStatsNumCalibrated(0)
    , fStatsNumUncalibrated(0)
    , fStatsNumAveraged(0)
  {
    this->reconfigure(pset);
    reg.sPreProcessEvent.watch(this, &Calibrator::preEvent);
    reg.sPostBeginSubRun.watch(this, &Calibrator::postBeginSubRun);
    reg.sPostBeginRun.watch(this, &Calibrator::postBeginRun);
  }
  
  //------------------------------------------------------------
  Calibrator::~Calibrator()  
  {
    if(fShapeTable){
      // Expecting about 95%
      mf::LogInfo("Calibrator") << "ADCShapeFitTable hit rate was "
                                << int(100*fShapeTable->TableHitFraction())
                                << "%";
      delete fShapeTable;
    }

    // NumCalibrated includes NumAveraged
    const unsigned int numRet = fStatsNumCalibrated+fStatsNumUncalibrated;
    if(numRet > 0){
      // Hopefully around 1%
      mf::LogInfo("Calibrator")
        << "Provided " << numRet << " RecoHits, of which "
        << ((1000*fStatsNumUncalibrated)/numRet)*.1 << "% were uncalibrated and "
        << ((1000*fStatsNumAveraged)/numRet)*.1 << "% were average constants.";
    }

    delete fAttenCache;
    delete fDriftCache;
    delete fAbsCache;
    delete fTimingCache;
  }
  
  //------------------------------------------------------------
  void Calibrator::reconfigure(const fhicl::ParameterSet& pset)
  {
    // Fix up the filename by looking in the search path
    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(pset.get< std::string >("ShapeTableFilenameFD"), fShapeTableFilenameFD);
    assert(!fShapeTableFilenameFD.empty());
    sp.find_file(pset.get< std::string >("ShapeTableFilenameNDMC"), fShapeTableFilenameNDMC);
    assert(!fShapeTableFilenameNDMC.empty());

    sp.find_file(pset.get< std::string >("ShapeTableFilenameNDData"), fShapeTableFilenameNDData);
    assert(!fShapeTableFilenameNDData.empty());

    fAttenCache->SetConstsCSVSource(pset.get<std::string>("AttenConstsCSVPath"),
				    pset.get<std::string>("AttenPointsCSVPath"));

    fAttenCache->SetTag(pset.get<std::string>("Tag"));

    // Chunked loading is only effective for non-CSV sources
    if(pset.get<std::string>("AttenPointsCSVPath").empty())
      fAttenCache->SetMaxDiblock(pset.get<int>("ChunkedLoadingMaxDiblock"));

    fAbsCache->SetCSVSource(pset.get<std::string>("AbsConstsCSVPath"));
    fAbsCache->SetTag(pset.get<std::string>("Tag"));
    
    fTimingCache->SetCSVSource(pset.get<std::string>("TimingConstsCSVPath"));
    fTimingCache->SetTag(pset.get<std::string>("TimingTag"));

    fUseMCcalib = pset.get<bool>("UseMCcalib");
    fMakeSinglePointCellHits = pset.get<bool>("MakeSinglePointCellHits");

    fUseTimingOffsets = pset.get<bool>("UseTimingOffsets");
  }
  
  //------------------------------------------------------------
  void Calibrator::preEvent(art::Event const& evt)
  {
    // use the art::Timestamp to create the fCurrTrimeStamp value
    unsigned long long int tsval = evt.time().value();
    
    // taking it apart
    // the masking isn't strictly necessary *if* "long" is truly 32bits
    // but this can vary w/ compiler/platform
    const unsigned long int mask32 = 0xFFFFFFFFUL;
    time_t lup = ( tsval >> 32 ) & mask32;
    int llo = tsval & mask32;
    TTimeStamp tts(lup, llo);
    fCurrTimeStamp = tts;
  }

  //------------------------------------------------------------
  void Calibrator::postBeginSubRun(const art::SubRun& subrun)
  {
    fDriftCache->SetRun(fGeom->DetId(), subrun.run(), subrun.subRun());
  }

  //------------------------------------------------------------
  void Calibrator::postBeginRun(art::Run const& run)
  {
    //std::cout<<"Beginning run "<<fGeom->DetId()<<" "<<run.beginTime().timeHigh()<<" "<<run.endTime().timeHigh()<<std::endl;
    //TODO: run.endTime() is returning 0 for all runs.  These constants are only set on a run by run basis at most, so just look for startTime + 1 second for now.
    fTimingCache->SetVldTime(fGeom->DetId(), run.beginTime().timeHigh(), run.beginTime().timeHigh());
  }


  //------------------------------------------------------------
  //   RawDigit-to-CellHit methods
  //------------------------------------------------------------
  
  rb::CellHit Calibrator::MakeCellHit(const rawdata::RawDigit *rawdigit)
  {
    // Copy of the hit that's modified to look like single-point data
    rawdata::RawDigit digSingle = *rawdigit;
    if(fMakeSinglePointCellHits && rawdigit->Version() != 0){
      digSingle.SetVersion(0);
      digSingle.fADC.clear();
      digSingle.SetADC(0, rawdigit->ADC());

      // Rest of function uses modified hit in place of argument provided
      rawdigit = &digSingle;
    }

    rb::CellHit chit(*rawdigit);
    
    // new info
    chit.SetCell  (fCMap->GetCell(rawdigit));
    chit.SetPlane (fCMap->GetPlane(rawdigit));
    chit.SetView  (fGeom->Plane(chit.Plane())->View());
    bool goodTime;
    const double tns = GetTNS(rawdigit, goodTime);

    const daqchannelmap::lchan lchan = fCMap->Map()->encodeLChan(fGeom->DetId(),
							  chit.Plane(),
							  chit.Cell());
    const daqchannelmap::dchan dchan = fCMap->Map()->encodeDChan(lchan);

    //check if timing offsets should be added in, 
    if (fUseTimingOffsets){
      const double offset = fTimingCache->GetOffset(fCMap->Map()->getDCM(dchan), fCMap->Map()->getDiBlock(dchan), fCurrTimeStamp.GetSec(), chit.IsRealData());
      chit.SetTNS   (tns+offset, goodTime);
    }
    else{
      chit.SetTNS   (tns, goodTime);
    }

    chit.SetPE    (GetPE(rawdigit));

    return chit;
  }

  //-----
  float Calibrator::GetPE(rb::CellHit const& ch)
  {
    return this->GetPE(&ch);
  }

  //-----
  float Calibrator::GetPE(const rawdata::RawDigit* dig)
  {
    // Look up gain from DB.  For now, must hard code the
    // number.  (This should not be put in any XML file.
    // That'll just confuse the issue.)
    const int maxadc = (1<<12)-1; // 4095
    // 1800 is the value of ADCMaxPE in ReadoutSim
    double gain = maxadc/1800.0;
    // That was the NDOS value. In FD, and ND, it's these values.
    // FEB 4.1
    if(fGeom->DetId() == novadaq::cnv::kFARDET ) gain = maxadc/2170.0;
    // FEB 5.2
    if(fGeom->DetId() == novadaq::cnv::kNEARDET) gain = maxadc/2048.0;

    // PE comes from the value determined as a side-effect of the timing fit,
    // the best-fit height of the curve when t0 is at its best value.

    // Doesn't matter whether or not timing fit suceeds, good fallbacks are in
    // place.
    bool junk;
    double peakadc;
    GetTNS(dig, junk, &peakadc);

    return peakadc/gain;
  }
  
  //-----
  float Calibrator::GetTNS(const rawdata::RawDigit* dig, bool& goodTime,
			   double* maxadc)
  {
    // Allows the rest of the function to assume maxadc is writeable
    double junk;
    if(!maxadc) maxadc = &junk;

    const int det = fGeom->DetId();

    // This is how the old MC fast-timing was encoded. Three
    // baseline-subtracted points, in non-chronological order. Treat specially
    // below.
    const bool isOldMC = (dig->IsMC() &&
			  dig->Version() == 0 &&
			  dig->NADC() == 3);

    static daqdataformats::NanoSliceVersionConvention conv;

    const unsigned int nSamples = conv.getNSamples(dig->Version());
    assert(isOldMC || dig->NADC() == nSamples);

    const double tns0 = dig->TDC()*15.625;

    // Version zero means this is a single DCS point
    if(dig->Version() == 0 && !isOldMC){
      goodTime = false;
      *maxadc = dig->ADC(); // Make sure this gets set reasonably
      // Simply convert the time stored in TDC to nanoseconds
      return tns0;
    }

    // Otherwise we can do the timing fit
    const unsigned int nPretrig = conv.getNPretriggeredSamples(dig->Version());

    // Need to have enough points to be able to see back to the baseline
    assert(isOldMC || nPretrig >= 3);
    // Also need enough points to see up to the peak
    assert(isOldMC || nSamples >= 4);

    // Create lookup table if required
    if(!fShapeTable){
      const bool isND = (det == novadaq::cnv::kNEARDET);
      std::string fname = fShapeTableFilenameFD;
      if(isND){
        if(dig->IsMC()){
          fname = fShapeTableFilenameNDMC;
        }
        else{
          fname = fShapeTableFilenameNDData;
        }
      }
      fShapeTable = new ADCShapeFitTable(fname, isND, dig->IsMC());
    }

    double tns;

    if(isOldMC){
      double junk;
      // ShapeTable wants to see the points in chronological order, but old MC
      // had ADC0 as the peak for backwards compatibility.
      tns = fShapeTable->TNS(tns0,
			     dig->ADC(1),
			     dig->ADC(2),
			     dig->ADC(0),
			     *maxadc, goodTime, junk);
    }
    else{
      double junk;
      // Figure out which sample is the baseline sample (ie the one combined
      // with the peak in the DCS calculation).
      const int baseIdx = nPretrig-3;
      // ShapeTable expects points in order, baseline subtracted
      tns = fShapeTable->TNS(tns0,
			     dig->ADC(baseIdx+1)-dig->ADC(baseIdx),
			     dig->ADC(baseIdx+2)-dig->ADC(baseIdx),
			     dig->ADC(baseIdx+3)-dig->ADC(baseIdx),
			     *maxadc, goodTime, junk);
    }

    return tns;
  }


  //------------------------------------------------------------
  //   CellHit-to-RecoHit methods
  //------------------------------------------------------------

  // For these two, calculate the misssing piece of information, then forward
  // to the protected version.
  rb::RecoHit Calibrator::MakeRecoHit(rb::CellHit const& cellhit, double w)
  {
    Double_t xyzd[4];
    GetXYZD(cellhit.PlaneCellPair(),w,xyzd);
    return MakeRecoHit(cellhit, w, xyzd);
  }

  //----
  rb::RecoHit Calibrator::MakeRecoHit(rb::CellHit const& cellhit, double *xyz)
  {
    double w = xyz[0];
    if(cellhit.View() == geo::kX) w = xyz[1];

    return MakeRecoHit(cellhit, w, xyz);
  }


  //----
  rb::RecoHit Calibrator::MakeRecoHit(rb::CellHit const& chit,
					double w, double *xyz)
  {
    rb::RecoHit rhit;

    rhit.SetX(xyz[0]);
    rhit.SetY(xyz[1]);
    rhit.SetZ(xyz[2]);

    rhit.SetT(GetT(chit, xyz[3]));

    // Set hit up uncalibrated in case we don't fill it below
    rhit.SetPECorr(-1);
    rhit.SetMIP   (-1);
    rhit.SetGeV   (-1);

    if(isnan(w)){
      mf::LogError("Calibrator")
        << "W position provided is NaN. This hit will be uncalibrated.";
    }

    const double peCorr = isnan(w) ? -1 : GetPECorr(chit, w);

    if(peCorr > 0){
      rhit.SetPECorr(peCorr);
      const double gevPerPECorr = GetPECorrToGeVScale(chit);
      if(gevPerPECorr > 0){
        rhit.SetGeV(peCorr     * gevPerPECorr);
        rhit.SetMIP(rhit.GeV() * GetGeVToMIPScale(chit));
        ++fStatsNumCalibrated;
      }
      else{
        // Uncalibrated
        rhit.SetPECorr(-1);
        ++fStatsNumUncalibrated;
      }
    }
    else{
      ++fStatsNumUncalibrated;
    }

    return rhit;
  }

  //-----
  double Calibrator::GetT(rb::CellHit const& cellhit, double dist_to_readout) {
    // This needs to be calibrated and put in a DB table
    Double_t speedOfFiberTransport = 15.3; // cm/ns, "first principles" calib.
    // Differs from c/n due to zigzag
    // paths down fiber.  But, this
    // value may be way off (10%? 30%?).
    return (cellhit.TNS()-dist_to_readout/speedOfFiberTransport);
  }

  //-----
  double Calibrator::GetPECorr(rb::CellHit const& cellhit, double w)
  {
    const double attenFactor = GetAttenScale(cellhit, w);
    if(attenFactor < 0) return -1;

    //    const double driftFactor = GetDriftScale(cellhit);
    // Disable for now
    const double driftFactor = 1;
    if(driftFactor < 0) return -1;

    return cellhit.PE()*attenFactor*driftFactor;
  }

  //----
  double Calibrator::GetTimeRes(rb::CellHit const& cellhit)
  {
    const int det = fGeom->DetId();
    double pe = cellhit.PE();
    //This is a fit for the timing resolution as a function of PE.  
    //For FD single point data this is derived by plotting the time difference between 
    //pairs of hits within a DCM on a cosmic track.  More information is in docdb-11571
    //TODO: the resolution for FD multipoint readout is still derived from MC truth, will be updated when there is data.

    double stdev = 144.0; //simple assumption of 500ns/sqrt(12)
    if (det == novadaq::cnv::kFARDET) {
      if (cellhit.GoodTiming()){
        stdev = 6884.58/(49.1465+pow(pe,1.378)) + 2.74427; //2014-05-06 fit, multipoint readout from mc
      }
      else {
        stdev = 848357.0/(5734.75+pow(pe,2.20060)) + 142.221; //2014-06-13 fit, from data
      }
    }
    else if (det == novadaq::cnv::kNEARDET) {
      if (cellhit.GoodTiming()){
        stdev = 2229.53/(77.0043+pow(pe,1.22743)) + 4.89355; //2014-09-15 fit from data with 140 ns rise time, 4.5 us fall time
      }
      //TODO:  ND single point timing resolution needs to be updated.  Currently simulation and data are all in
      //multipoint.  For now take the FD singlpe point resolution and dived by 4 since the sampling is 4 times as fast
      else {
        stdev = 6136470.0/(185518.0+pow(pe,2.92360)) + 31.6071; //2014-09-15 singe point fit of data
      }
    }

    return stdev;

  }

  //-----
  double Calibrator::GetAttenScale(rb::CellHit const& cellhit, double w)
  {

    const calib::AttenCurve* curve = 0;
    double                   maxW  = 0.0;

    if(!GetAttenCurve(cellhit.Plane(), cellhit.Cell(), cellhit.IsRealData(), curve, maxW)) return -1;

    if(curve->chan.Plane() >= 1000) ++fStatsNumAveraged;

    return GetAttenScale(curve, maxW, w);
  }

  //-----
  double Calibrator::GetAttenScale(const calib::AttenCurve* curve, const double maxW, double w)
  {

    if(w < -maxW) w = -maxW;
    if(w > +maxW) w = +maxW;

    const double mean = curve->MeanPEPerCmAt(w);

    if(mean < 0) return -1;

    // The goal of this factor is to make PECorrs approximately the same size
    // as PEs. But the exact values are terrible historical accident. If we
    // decide to clean this up in future calibration versions we can do so and
    // then condition this on the tag version.
    double pecorrscale = 0;
    switch(fGeom->DetId()){
    case novadaq::cnv::kNDOS:    pecorrscale = 75.00; break;
    case novadaq::cnv::kNEARDET: pecorrscale = 37.51; break;
    case novadaq::cnv::kFARDET:  pecorrscale = 39.91; break;
    default:
      assert(0 && "Unknown detector");
    }

    const double factor = pecorrscale/mean;

    return factor;
  }

  //-----
  bool Calibrator::GetAttenCurve(const int plane, const int cell, const bool is_realdata
                                ,const calib::AttenCurve*& curve, double& maxW) const
  {
    const geo::OfflineChan offline_channel(plane, cell);

    if(is_realdata)
      //
      //! \todo Expert use only! Remove once FD database is populated
      //        with calibration constants
      //        This will just apply the MC constants to FD data
      //
      if(fUseMCcalib)
	curve = fAttenCache->ForMC(&(*fGeom), offline_channel);
      else
	curve = fAttenCache->ForData(&(*fGeom), offline_channel);
    else
      curve = fAttenCache->ForMC(&(*fGeom), offline_channel);
    if(!curve) return false;
    if(!curve->IsCalibrated()) return false;

    // Clamp given w-coordinate into the detector to avoid bad behaviour of the
    // attenuation function.
    maxW = fGeom->Plane(plane)->Cell(cell)->HalfL();

    return true;
  }

  //-----
  double Calibrator::GetDriftScale(rb::CellHit const& cellhit)
  {
    // MC doesn't drift
    if(!cellhit.IsRealData()) return 1;

    return fDriftCache->GetFactor(cellhit.Plane(), cellhit.Cell());
  }

  //-----
  double Calibrator::GetGeVToMIPScale(rb::CellHit const& /* cellhit */ )
  {
    // MIPs for one GeV times one cell in z in NOvA scintillator
    return 94.62;
  }

  //-----
  double Calibrator::GetPECorrToGeVScale(rb::CellHit const& chit)
  {
    const daqchannelmap::lchan lchan = fCMap->Map()->encodeLChan(fGeom->DetId(),
							  chit.Plane(),
							  chit.Cell());
    const daqchannelmap::dchan dchan = fCMap->Map()->encodeDChan(lchan);

    const bool data = fUseMCcalib ? false : chit.IsRealData();

    return fAbsCache->GetPECorrToGeVScale(fGeom->DetId(),
					  fCMap->Map()->getDiBlock(dchan),
					  data);
  }

  //-----
  double Calibrator::GetGeVToCalorimetricScale() const
  {
    // From Kanika's study: docdb 11528
    //
    // Could vary by detector or calibration version etc. We have all that
    // information available here. If it gets complicated enough may need to
    // come from database. For now it's a single number, and there's a
    // reasonable chance it might never need to change.
    return 1.78;
  }

  //-----
  double Calibrator::GetShowerEnergy(TVector3 const& start,
				     TVector3 const& stop,
				     double   const& totalGeV)
  {//now only for electron
    double ecor =  5.66050e-01;//for dead material correction

    double avgx = (start.X()+stop.X())/2.;
    double avgy = (start.Y()+stop.Y())/2.;
    double poscor = 1.0+7.00279e-05*avgx+6.91714e-05*avgy;//for position correction

    return totalGeV/(ecor*poscor);
  }

  DEFINE_ART_SERVICE(Calibrator);

} // end namespace calib
///////////////////////////////////////////////////////////////////////
