////////////////////////////////////////////////////////////////////////
/// \brief   Helper for AttenCurve
/// \author  Chris Backhouse - bckhouse@caltech.edu
///
////////////////////////////////////////////////////////////////////////
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Calibrator/func/AttenCache.h"
#include "Calibrator/func/AttenCurve.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/GeometryBase.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

namespace calib
{
  //......................................................................
  AttenCache::AttenCache(int maxdiblock)
    : fMaxDiblock(maxdiblock), fInitialized(false)
  {
  }

  //......................................................................
  AttenCache::~AttenCache()
  {
    for(std::map<Key_t, AttenCurve*>::iterator it = fMap.begin();
	it != fMap.end(); ++it){
      delete it->second;
    }
  }

  //......................................................................
  void AttenCache::Add(AttenCurve* res)
  {
    fMap[Key_t(res->det, res->chan)] = res;
  }

  //......................................................................
  const AttenCurve* AttenCache::ForData(const geo::GeometryBase* geom,
					geo::OfflineChan chan)
  {
    // TODO - this won't behave right if we ask for more than one detector
    // TODO - or data and MC in the same job.
    if(!Initialized()) Fill(geom, true);

    const int db = chan.Plane()/64+1;
    if(!fDiblocks.count(db)){
      // No calibration for this diblock, use the averages instead.
      // Stored at channel 1,000,000+1000*view+cell
      const geo::View_t view = geom->Plane(chan.Plane())->View();
      return Get(geom->DetId(), geo::OfflineChan(1000+int(view), chan.Cell()));
    }

    return Get(geom->DetId(), chan);
  }

  //......................................................................
  const AttenCurve* AttenCache::ForMC(const geo::GeometryBase* geom,
				      geo::OfflineChan chan)
  {
    // TODO - this won't behave right if we ask for more than one detector
    // TODO - or data and MC in the same job.
    if(!Initialized()) Fill(geom, false);

    const novadaq::cnv::DetId det = geom->DetId();
    const int plane = chan.Plane();
    const int cell = chan.Cell();

    const geo::View_t view = geom->Plane(plane)->View();

    // If it's a muon catcher plane, check to see if there is a special entry
    // for it
    if(geom->DetId() == novadaq::cnv::kNEARDET &&
       plane >= int(geom->FirstPlaneInMuonCatcher())){
      const AttenCurve* ret = Get(det, geo::OfflineChan(10+int(view), cell));
      // If not, fall back to the non-muon-catcher constants
      if(ret) return ret;
    }

    // This is the way we encode the MC parameters. Plane represents the view.
    // TODO: this view-merging logic is duplicated in
    // AttenCache::LoadFromDatabase and is fragile.
    return Get(det, geo::OfflineChan(int(view), cell));
  }

  //......................................................................
  const AttenCurve*
  AttenCache::Get(int det, geo::OfflineChan chan) const
  {
    Map_t::const_iterator it = fMap.find(Key_t(det, chan));
    if(it != fMap.end()) return it->second;
    return 0;
  }

  //......................................................................
  void AttenCache::Fill(const geo::GeometryBase* geom, bool isData)
  {
    if(fInitialized) return;
    fInitialized = true;

    LoadFromDatabase(geom, isData);
  }

  //......................................................................
  void AttenCache::LoadFromDatabase(const geo::GeometryBase* geom, bool isData)
  {
    // For the FD this will make a cache ~80MB. That's OK, but denser LOWESS
    // points would increase this rapidly.

    mf::LogInfo("AttenCache")
      << "Initializing attenuation cache for " << novadaq::cnv::DetInfo::GetName(geom->DetId()) << " from database...\n";


    nova::dbi::Table tblConsts("Calibration/tables/CalibAttenConstsVld.xml");
    tblConsts.SetDetector(geom->DetId());
    tblConsts.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);

    if(!fTag.empty()) tblConsts.SetTag(fTag);
    tblConsts.SetVerbosity(100);

    tblConsts.SetMinTSVld(1);
    tblConsts.SetMaxTSVld(100);

    bool ok = true;

    if(!fConstsCSVSource.empty()){
      mf::LogInfo("AttenCache") << "Loading attenuation cache from "
				<< fConstsCSVSource << std::endl;
      ok = ok && tblConsts.LoadFromCSV(fConstsCSVSource);
    }
    else{
      ok = ok && tblConsts.Load();
    }

    if(!ok){
      // No safe way to handle this
      mf::LogError("AttenCache") << "Unable to load attenuation tables. Aborting"
                                 << std::endl;
      // I've definitely seen mf lose races with shutdown
      std::cerr << "Unable to load attenuation tables. Aborting"
                << std::endl;
      abort();
    }

    const int coeffCol   = tblConsts.GetColIndex("coeff_exp");
    const int attenCol   = tblConsts.GetColIndex("atten_length");
    const int bkgCol     = tblConsts.GetColIndex("background");
    const int loEdgeCol  = tblConsts.GetColIndex("edge_low");
    const int hiEdgeCol  = tblConsts.GetColIndex("edge_high");
    const int loCoeffCol = tblConsts.GetColIndex("coeff_low");
    const int hiCoeffCol = tblConsts.GetColIndex("coeff_high");
    const int chisqCol   = tblConsts.GetColIndex("chisq");

    const int Nconsts = tblConsts.NRow();
    for(int nconsts = 0; nconsts < Nconsts; ++nconsts){
      nova::dbi::Row* row = tblConsts.GetRow(nconsts);

      AttenCurve* res = AttenCurve::Uninitialized(geom->DetId(),
						  geo::OfflineChan(-1, -1));

      res->chan = geo::OfflineChan::FromDBValidityChan(row->Channel());

      ok = ok && row->Col(coeffCol  ).Get(res->coeff_exp);
      ok = ok && row->Col(attenCol  ).Get(res->atten_length);
      ok = ok && row->Col(bkgCol    ).Get(res->background);

      ok = ok && row->Col(loEdgeCol ).Get(res->edge_low);
      ok = ok && row->Col(hiEdgeCol ).Get(res->edge_high);
      ok = ok && row->Col(loCoeffCol).Get(res->coeff_low);
      ok = ok && row->Col(hiCoeffCol).Get(res->coeff_high);
      ok = ok && row->Col(chisqCol  ).Get(res->chisq);
      assert(ok);

      const geo::PlaneGeo* geoplane = 0;

      bool inMuonCatcher = false;

      const int plane = res->chan.Plane();
      if(plane == 1000 || plane == 1001){
        // These two plane numbers indicate the "average" constants for use in
        // uncalibrated diblocks.
        geoplane = geom->Plane(*geom->GetPlanesByView(geo::View_t(plane-1000)).begin());
      }
      else{
        if(isData){
          // Normal data just has regular channel numbers
          geoplane = geom->Plane(plane);
          if(geom->DetId() == novadaq::cnv::kNEARDET &&
             plane >= int(geom->FirstPlaneInMuonCatcher())){
            inMuonCatcher = true;
          }
        }
        else{
          // MC has all planes in a view merged, and stores them in planes 0
          // and 1.

          // TODO: this view-merging logic is duplicated in AttenCurve::ForMC
          // and is fragile.
          if(plane < 2){
            geoplane = geom->Plane(*geom->GetPlanesByView(geo::View_t(plane)).begin());
          }
          else if(plane == 10 || plane == 11){
            assert(geom->DetId() == novadaq::cnv::kNEARDET);
            inMuonCatcher = true;
            const int mcp = geom->FirstPlaneInMuonCatcher();
            geoplane = geom->Plane(mcp);
            if(geoplane->View() != geo::View_t(plane-10))
              geoplane = geom->Plane(geom->NextPlaneOtherView(mcp, +1));
          }
          else{
            assert(0 && "Bad encoded view");
          }
        }
      }

      assert(geoplane);
      const geo::CellGeo* geocell = geoplane->Cell(res->chan.Cell());
      assert(geocell);

      res->cell_length = 2*geocell->HalfL();

      if(inMuonCatcher){
        // Find a plane from the same view that's not in the muon catcher, to
        // find out how much the muon catcher cell must be displaced.
        int i = 0;
        const geo::PlaneGeo* regularPlane = geom->Plane(i);
        while(regularPlane->View() != geoplane->View())
          regularPlane = geom->Plane(++i);
        const geo::CellGeo* regularCell = regularPlane->Cell(res->chan.Cell());

        // Assume that the cells are aligned on the readout side
        res->center_offset = regularCell->HalfL()-geocell->HalfL();
      }

      res->initialized = true;

      Add(res);

      // Figure out which diblock, mark it calibrated
      const int db = plane/64+1;
      fDiblocks.insert(db);
    } // end for nconsts

    // Now fetch the LOWESS points and add them to the curves we've just
    // created.

    int totPoints = 0;

    // Only do chunked loading for FD data
    if(!isData || geom->DetId() != novadaq::cnv::kFARDET)
      fMaxDiblock = -1;

    for(int diblock = 1; diblock <= (fMaxDiblock >= 0 ? fMaxDiblock : 1); ++diblock){
      nova::dbi::Table tblPoints("Calibration/tables/CalibAttenPointsVld.xml");
      tblPoints.SetDetector(geom->DetId());
      tblPoints.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);

      if(!fTag.empty()) tblPoints.SetTag(fTag);
      tblPoints.SetVerbosity(100);

      tblPoints.SetMinTSVld(1);
      tblPoints.SetMaxTSVld(100);

      if(fMaxDiblock > 0){
	// Difference in vldChan encoding block-to-block
	const int stride = 1000*64*1000;
	tblPoints.SetChannelRange((diblock-1)*stride, diblock*stride-1);
	if(diblock == fMaxDiblock){
	  // If this is the last iteration, make sure to also pick up the
	  // special averaged constants out at plane 1000+
	  tblPoints.SetMaxChannel(1000*1000*1000+stride);
	}
      }

      if(!fPointsCSVSource.empty()){
	mf::LogInfo("AttenCache") << "Loading attenuation cache from "
				  << fPointsCSVSource << std::endl;
	ok = ok && tblPoints.LoadFromCSV(fPointsCSVSource);
	mf::LogInfo("AttenCache") << "Got " << tblPoints.NRow() << " rows from CSV" << std::endl;
      }
      else{
	ok = ok && tblPoints.Load();
      }

      if(!ok){
	// No safe way to handle this
	mf::LogError("AttenCache") << "Unable to load attenuation tables. Aborting"
				   << std::endl;
	// I've definitely seen mf lose races with shutdown
	std::cerr << "Unable to load attenuation tables. Aborting"
		  << std::endl;
	abort();
      }

      const int Npoints = tblPoints.NRow();
      if(Npoints == 0) continue;
      totPoints += Npoints;

      const int wColPts      = tblPoints.GetColIndex("w");
      const int factorColPts = tblPoints.GetColIndex("factor");

      for(int npoints = 0; npoints < Npoints; ++npoints){
	nova::dbi::Row* row = tblPoints.GetRow(npoints);

	const geo::OfflineChan chan = geo::OfflineChan::FromDBValidityChan(row->Channel()/1000);

	float w, factor;
	ok = ok && row->Col(wColPts     ).Get(w);
	ok = ok && row->Col(factorColPts).Get(factor);
	assert(ok);

	auto it = fMap.find(Key_t(geom->DetId(), chan));
	assert(it != fMap.end());
	AttenCurve* curve = it->second;

	const unsigned int interpIdx = row->Channel()%1000;
	if(curve->interp_pts.size() <= interpIdx)
	  curve->interp_pts.resize(interpIdx+1);
	curve->interp_pts[interpIdx] = {w, factor};
      } // end for npoints
    } // end for diblock

    // Check that the number of points we got from the DB is the same as the
    // total in all the AttenCurves. If these numbers don't match, it means
    // there were duplicates or gaps.
    int pointsCount = 0;
    for(auto it: fMap) pointsCount += it.second->interp_pts.size();
    assert(pointsCount == totPoints);
  }

  //......................................................................
  bool AttenCache::Key_t::operator<(const Key_t& rhs) const
  {
    if(fDet < rhs.fDet) return true;
    if(fDet > rhs.fDet) return false;
    return fChan < rhs.fChan;
  }

} // end namespace calib
////////////////////////////////////////////////////////////////////////
