////////////////////////////////////////////////////////////////////////
/// \brief   Helper for AttenCurve
/// \author bckhouse@caltech.edu
/// \date
////////////////////////////////////////////////////////////////////////
#ifndef CALIB_ATTENCACHE_H
#define CALIB_ATTENCACHE_H

#include <map>
#include <string>

#include "NovaDAQConventions/DAQConventions.h"
#include "GeometryObjects/OfflineChan.h"

namespace geo{class GeometryBase;}

namespace calib
{
  class AttenCurve;

  class AttenCache
  {
  public:
    /// \param maxdiblock Largest diblock in DB when loading points table
    ///                   in chunks. Pass -1 not to do chunking.
    AttenCache(int maxdiblock = -1);
    virtual ~AttenCache();

    /// Return an AttenCurve object matching your parameters, or NULL
    /// if there isn't one in the database. Do not delete the resulting object,
    /// they are cached internally.
    const AttenCurve* ForData(const geo::GeometryBase* geom,
			      geo::OfflineChan chan);
    const AttenCurve* ForMC(const geo::GeometryBase* geom,
			    geo::OfflineChan chan);

    void SetConstsCSVSource(const std::string& s, const std::string& p)
    {
      fConstsCSVSource = s;
      fPointsCSVSource = p;
    }

    void SetTag(const std::string& tag) {fTag = tag;}

    bool Initialized() const {return fInitialized;}

    void SetMaxDiblock(int db) {fMaxDiblock = db;}

  protected:
    // Loading curves from the database as required is too slow (latency).
    // Instead, just read the whole detector in and cache it.
    void Fill(const geo::GeometryBase* geom, bool isData);

    void Add(AttenCurve* res);
    /// NULL if entry isn't in the cache. Do not delete the returned pointer.
    const AttenCurve* Get(int det, geo::OfflineChan chan) const;

    void LoadFromDatabase(const geo::GeometryBase* geom, bool isData);

    int fMaxDiblock; ///< -1 if not chunking loading by diblock

    std::string fConstsCSVSource, fPointsCSVSource;
    std::string fTag;

    // Whether we have (attempted to) read from db or csv yet
    bool fInitialized;

    class Key_t
    {
    public:
      Key_t(int d, geo::OfflineChan chan) : fDet(d), fChan(chan) {}
      bool operator<(const Key_t& rhs) const;
    protected:
      int fDet;
      geo::OfflineChan fChan;
    };

    typedef std::map<Key_t, AttenCurve*> Map_t;
    Map_t fMap;

    std::set<int> fDiblocks; ///< Diblocks with calibrations
  };

} // end namespace calib

#endif
////////////////////////////////////////////////////////////////////////
