////////////////////////////////////////////////////////////////////////
/// \brief   Hold drift constants from current run
/// \author  bckhouse@caltech.edu
/// \date
/////////////////////////////////////////////////////////////////////////
#include "Calibrator/func/DriftCache.h"

#include "messagefacility/MessageLogger/MessageLogger.h"

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include <cassert>

namespace calib
{
  //......................................................................
  DriftCache::DriftCache()
    : fDirty(true),
      fCurDet(novadaq::cnv::kUNKNOWN_DET),
      fCurRun(-1),
      fCurSubrun(-1)
  {
  }

  //......................................................................
  DriftCache::~DriftCache()
  {
  }

  //......................................................................
  double DriftCache::GetFactor(int plane, int cell)
  {
    // Ensure we were told what run it was at some point
    assert(fCurRun >= 0);

    if(fDirty){
      fDirty = false;
      LoadFromDatabase();
    }

    std::map<geo::OfflineChan, double>::iterator it = fMap.find(geo::OfflineChan(plane, cell));

    if(it == fMap.end()){
      mf::LogWarning("DriftCache") << "No drift constants for plane "
                                   << plane << " cell " << cell;
      return 1;
    }

    return it->second;
  }

  //......................................................................
  void DriftCache::SetRun(int det, int run, int subrun)
  {
    if(det != fCurDet || run != fCurRun || subrun != fCurSubrun){
      fDirty = true;
      fMap.clear();

      fCurDet = det;
      fCurRun = run;
      fCurSubrun = subrun;
    }
  }

  //......................................................................
  void DriftCache::LoadFromDatabase()
  {
    std::string detStr;
    switch(fCurDet) {
    case novadaq::cnv::kNEARDET:
      detStr = "nd";
      break;
    case novadaq::cnv::kFARDET:
      detStr = "fd";
      break;
    case novadaq::cnv::kNDOS:
      detStr = "ndos";
      break;
    default:
      assert(0 && "Unknown detector ID");
    }

    mf::LogInfo("DriftCache")
      << "Initializing drift cache for " << detStr << " from database...\n";


    nova::dbi::Table tbl("CellToCell/tables/DriftConstants.xml");
    tbl.SetDetector(fCurDet);
    tbl.SetVerbosity(100);

    tbl.SetValidityRange("detector", detStr);
    tbl.SetValidityRange("release", "S12.02.14"); // TODO

    std::stringstream sql;
    sql << fCurRun << ">=minrun and " << fCurRun << "<maxrun";
    tbl.SetValiditySQL(sql.str());

    bool ok = tbl.GetConnection();
    ok = ok && tbl.LoadFromDB();

    if(!ok || tbl.NRow() == 0){
      // No safe way to handle this
      mf::LogError("DriftCache") << "Unable to load drift tables. Aborting"
                                 << std::endl;
      // I've definitely seen mf race with shutdown
      std::cerr << "Unable to load drift tables. Aborting"
                << std::endl;
      abort();
    }

    const int planeCol = tbl.GetColIndex("plane");
    const int cellCol  = tbl.GetColIndex("cell");
    const int driftCol = tbl.GetColIndex("driftconst");

    const int maxRow = tbl.NRow();
    for(int rowIdx = 0; rowIdx < maxRow; ++rowIdx){
      nova::dbi::Row* row = tbl.GetRow(rowIdx);

      int plane, cell;
      bool ok = row->Col(planeCol).Get(plane);
      ok = (ok && row->Col(cellCol).Get(cell));
      double drift;
      ok = (ok && row->Col(driftCol).Get(drift));
      assert(ok);

      fMap[geo::OfflineChan(plane, cell)] = drift;
    } // end for rowIdx

    tbl.CloseConnection();
  }

} // end namespace calib
////////////////////////////////////////////////////////////////////////
