////////////////////////////////////////////////////////////////////////
/// \brief   Hold drift constants from current run
/// \author  bckhouse@caltech.edu
/// \date
/////////////////////////////////////////////////////////////////////////
#include "Calibrator/func/TimingCache.h"

#include "messagefacility/MessageLogger/MessageLogger.h"

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include <cassert>

namespace calib
{
  //......................................................................
  TimingCache::TimingCache()
    : fInitialized(false),
      fCurDet(novadaq::cnv::kUNKNOWN_DET),
      fMinTime(std::numeric_limits<uint32_t>::max()-1),
      fMaxTime(std::numeric_limits<uint32_t>::min())
  {
  }

  //......................................................................
  TimingCache::~TimingCache()
  {
  }

  //......................................................................
  double TimingCache::GetOffset(int dcm, int diblock, std::uint32_t time, bool isData)
  {

    if(!fInitialized)
      LoadFromDatabase(isData);

    // Ensure we were told what run it was at some point
    //TODO: leave this out until run start and end times make sense
    /*if (!(time >= fMinTime && time <= fMaxTime)){
         //mf::LogWarning("TimingCache") << "Time outside of valid range, no timing offset for dcm "<<dcm;
         //TODO: Eventually for data we would want to abort in this case, leave this out till the db is populated
        return 0;
    }*/

    int key = (diblock*1000)+dcm;
    std::map<int, DelayRec>::iterator it;
    it = fMap.find(key);
    if(it == fMap.end()){
      mf::LogDebug("TimingCache") << "No timing offset for dcm "<<dcm;
      return 0;
    }


    const DelayRec& rec = it->second;
    return rec.dcmoffset;
  }

  //......................................................................
  double TimingCache::GetSigma(int dcm, int diblock, std::uint32_t time, bool isData)
  {

    if(!fInitialized)
      LoadFromDatabase(isData);

    // Ensure we were told what run it was at some point
    if (!(time >= fMinTime && time <= fMaxTime)){
         //mf::LogWarning("TimingCache") << "Time outside of valid range, no timing offset sigma for dcm "<<dcm;
        return 0;
    }


    std::map<int, DelayRec>::iterator it;
    int key = (diblock*1000)+dcm;
    it = fMap.find(key);
    if(it == fMap.end()){
      mf::LogDebug("TimingCache") << "No timing offset sigma for dcm "<<dcm;
      return 0;
    }

    const DelayRec& rec = it->second;
    return rec.sigma;
  }

  //......................................................................
  void TimingCache::SetVldTime(novadaq::cnv::DetId det, std::uint32_t minTime, std::uint32_t maxTime)
  {
    if(det != fCurDet || minTime != fMinTime || maxTime != fMaxTime){
      fInitialized = false;
      fMap.clear();

      fCurDet = det;
      fMinTime= minTime;
      fMaxTime = maxTime;
    }
  }

  //......................................................................
  void TimingCache::LoadFromDatabase(bool isData)
  {
    // Even if it goes wrong there's no point trying again
    fInitialized = true;

    mf::LogInfo("TimingCache")
      << "Initializing timing delay cache for " << fCurDet << " from database...\n";


    nova::dbi::Table tblConsts("Calibration/tables/CalibTimingConsts.xml");
    tblConsts.SetDetector(fCurDet);
    tblConsts.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);
    if(!fTag.empty()) tblConsts.SetTag(fTag);
    tblConsts.SetVerbosity(100);

    tblConsts.SetMinTSVld(fMinTime);
    tblConsts.SetMaxTSVld(fMaxTime);

    bool ok = true;

    if(!fCSVSource.empty()){
      mf::LogInfo("TimingCache") << "Loading from "
			      << fCSVSource << std::endl;
      ok = ok && tblConsts.LoadFromCSV(fCSVSource);
    }
    else{
      ok = ok && tblConsts.Load();
    }


    if(!ok){
      // No safe way to handle this
      mf::LogError("TimingCache") << "Unable to load timing delay tables. No timing delays will be set."
                                 << std::endl;

      //TODO: For data abort if table cannot be loaded, commented out since the db in not in production, this is a placeholder
      //if (isData) abort();
    }
    else{
      const int dcmoffsetCol    = tblConsts.GetColIndex("dcmoffset");
      const int sigmaCol = tblConsts.GetColIndex("sigma");
      const int statsCol = tblConsts.GetColIndex("stats");

      const int Nconsts = tblConsts.NRow();
      for(int nconsts = 0; nconsts < Nconsts; ++nconsts){
        nova::dbi::Row* row = tblConsts.GetRow(nconsts);

        const int dcmkey = row->Channel();

        DelayRec rec;

        ok = ok && row->Col(dcmoffsetCol).Get(rec.dcmoffset);
        ok = ok && row->Col(sigmaCol).Get(rec.sigma);
        ok = ok && row->Col(statsCol).Get(rec.stats);

        //For data, abort if the value is bad,
        if (isData) assert(ok);

        if (ok) fMap[dcmkey] = rec;
      } // end for nconsts
    }

  }

} // end namespace calib
////////////////////////////////////////////////////////////////////////
