////////////////////////////////////////////////////////////////////////
// \file    CellToCell.cxx
// \brief   TODO
// \version $Id: CellToCell.cxx,v 1.19 2012-09-25 16:52:48 gsdavies Exp $
// \author  timdkut@iastate.edu
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <vector>

// ROOT includes
#include "TH1.h"
#include "TH2.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"

// NOvASoft includes
#include "Calibrator/Calibrator.h"
#include "CellToCell/CellSummary.h"
#include "ChannelInfo/BadChanList.h"
#include "CMap/CMap.h"
#include "Commissioning/CanaNt.h"
#include "CalibrationUtils/CalibUtil.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/PlaneGeo.h"
#include "RawData/RawDigit.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"


namespace c2c { class CellSummary; }

/// Cell-to-cell calibration
namespace c2c { 
  class CellToCell : public art::EDAnalyzer {
    
  public:
    explicit CellToCell(fhicl::ParameterSet const &pset); // Required! explicit tag tells the compiler this is not a copy constructor      
    virtual ~CellToCell();                        
    
    void analyze(art::Event const& evt); 
    void beginJob();
    
  private:
    std::string pTracks;
    std::string pCellHits;
    std::string pRawDigits;
    
    TTree*           fSummaryTree;
    CellSummary*     fsummary;
    TH2*             fBadChannels;

  private:

    time_t fFirstEventTimeStamp;
   
  };
} // end namespace c2c

////////////////////////////////////////////////////////////////////////
namespace c2c
{
  int EVENT = 0;

  //...................................................................
  CellToCell::CellToCell(fhicl::ParameterSet const &pset): 
    EDAnalyzer(pset),
    pTracks(pset.get < std::string >("Tracks")),
    pCellHits(pset.get < std::string >("CellHits")),
    pRawDigits(pset.get < std::string >("RawDigits")),
    fFirstEventTimeStamp(0)
  {
  }
  
  //......................................................................
  
  CellToCell::~CellToCell() { } 
  
  //......................................................................
  void CellToCell::beginJob()
  {
    //
    // Book histograms, ntuples, initialize counts etc., etc., ...
    //
    art::ServiceHandle<art::TFileService> tfs;
    
    fsummary = new CellSummary;
    fSummaryTree = tfs->make<TTree>("Cosmics","Cosmic Information Tree");
    fSummaryTree->Branch("fCellSummary", "c2c::CellSummary", &fsummary, 1600, 99);
    
    fBadChannels = tfs->make<TH2D>("Masked Channels in Detector",";Plane;Cell",210,0,209,100,0,99);

    return;
  }
  
  //......................................................................
  void CellToCell::analyze(art::Event const& evt)
  {
    
    art::ServiceHandle<geo::Geometry> geom;
    
    art::ServiceHandle<cmap::CMap> cmap;

    //    art::ServiceHandle<cmap::CMap> cmap;

    art::ServiceHandle<calib::Calibrator> cal;
    
    art::Handle< std::vector<rb::Track> > trks;
    evt.getByLabel(pTracks, trks);
    
    art::Handle< std::vector<rb::CellHit> >chits;
    evt.getByLabel(pCellHits, chits);
    
    art::Handle< std::vector<rawdata::RawDigit> > rawhitcol;
    evt.getByLabel(pRawDigits, rawhitcol);

    //Make it an art::PtrVector
    art::PtrVector<rawdata::RawDigit> rawhitlist;
    for(unsigned int i = 0; i < rawhitcol->size(); ++i){
      art::Ptr<rawdata::RawDigit> rawhit(rawhitcol,i);
      rawhitlist.push_back(rawhit);
    }
    if (EVENT == 0){
      for(unsigned int iPlane = 0; iPlane < geom->NPlanes(); iPlane++){
	const geo::PlaneGeo* p = geom->Plane(iPlane);
	for(unsigned int iCell = 0; iCell < p->Ncells(); iCell++){
	  art::ServiceHandle<chaninfo::BadChanList> badc;
	  //std::cout << badc->IsBad(iPlane, iCell) << std::endl;
	  if(badc->IsBad(iPlane,iCell)){
	    fBadChannels->Fill(iPlane,iCell);
	  }
	}
      }
    }
    EVENT += 1;

    int geomplanes    = geom->NPlanes();
    //    double geomwidth  = geom->DetHalfWidth()*2;
    //    double geomheight = geom->DetHalfHeight()*2;
    //    double geomlength = geom->DetLength();
    
    int TrackSliceNumber  = 0;            //labels the track in the order the slicer creates the track (this helps with counting tracks with cuts)
    int TrackNumber       = trks->size(); //gives the number of tracks in an event
   
    for(unsigned int a = 0; a < trks->size(); ++a){ //loop over tracks
      
      art::Ptr<rb::Track> track(trks,a);
      
      //track information : Start and End points, length and direction of track
      double trksx       = track->Start().X();
      double trksy       = track->Start().Y();
      double trksz       = track->Start().Z();
      double dx          = track->Dir().X();
      double dy          = track->Dir().Y();
      double dz          = track->Dir().Z();
      double tracklength = track->TotalLength();
      double trkex       = track->Start().X()+tracklength*dx;
      double trkey       = track->Start().Y()+tracklength*dy;
      double trkez       = track->Start().Z()+tracklength*dz;      
      TrackSliceNumber += 1;

      std::vector<calib::EPathQuality> quals;
      const std::vector<double> pathlengths = calib::BestPathEstimates(track->PlaneCellPairs(), track->Dir(), quals);

      //Other necessary information about the data

      int EventTimeHigh  = evt.time().timeHigh(); //time and date in seconds (eg. date and clock time)
      int EventTimeLow   = evt.time().timeLow();  //time less than a second represented in nanoseconds (eg. fine timing for event by event basis)

      int RunNumber      = evt.id().run();
      int SubrunNumber   = evt.id().subRun();
      int EventNumber    = evt.id().event();

      
      ///////////////////////////////////////////////
      ////                                       ////
      ////                 Cuts                  ////
      ////                                       ////
      ///////////////////////////////////////////////
      
      //Need to go throught the entire track to see if the track is quality
      //    Then the cut (ie. boolean output) is saved to the tree for future use
      //
      
      bool NPLANE          = false;
      bool NPLANEX         = false;
      bool NPLANEY         = false;
      bool NCELL           = false;
      bool NCELLX          = false;
      bool NCELLY          = false;
      bool CONTINUOUSfour  = false;
      bool CONTINUOUSfive  = false;
      bool NPLANETOTAL     = false;
      
      int planetotal [201] = {0};
      
      int numbercell         = 0;//number of total cells (counted by this script)
      int npt                = 0;//number of total planes (counted by this script)
      int nplanex            = 0;//number of total x planes (counted by this script)
      int nplaney            = 0;//number of total y planes (counted by this script)
      
      int ncell              = track->NCell();
      int nxcell             = track->NXCell();
      int nycell             = track->NYCell();
      int nplane             = 0;
      int nxplane            = 0;
      int nyplane            = 0;
      
      int planecounterx[210] = {0};
      int planecountery[210] = {0};
      int cellcounterx[210]  = {0};
      int cellcountery[210]  = {0};
      
      //this allocates for each plane and cell a 0 or 1 for a no hit or hit
      if (nxcell > 0){
	for(signed int d = 0; d < nxcell; ++d){
	  art::Ptr<rb::CellHit> cellhits=track->Cell(geo::kX, d);
	  
	  int planecount = cellhits->Plane();
	  planecounterx[planecount] = 1;
	  cellcounterx[planecount] += 1;
	}	
      }
      if (nycell > 0){
	for(signed int d = 0; d < nycell; ++d){
	  art::Ptr<rb::CellHit> cellhits=track->Cell(geo::kY, d);
	  
	  int planecount = cellhits->Plane();
	  planecountery[planecount] = 1;
	  cellcountery[planecount] += 1;
	}
      }
      //loops through the entire planes of the detector
      for(int jjj = 0; jjj < geomplanes; ++jjj){
	
	nplanex += planecounterx[jjj];   //Counts the number of planes for x-view
	nplaney += planecountery[jjj];   //Counts the number of planes for y-view
	
	if (cellcounterx[jjj] > 3)      //////////////////////////////////////////////////////////////////
	  NCELLX = true;                //                   "Trackyness" of a track :                  //
	if (cellcountery[jjj] > 3)      //  Do the amount of cells make what should look like a track?  //
	  NCELLY = true;                //////////////////////////////////////////////////////////////////
	
	numbercell += cellcounterx[jjj];
	numbercell += cellcountery[jjj];
	
      }

      //Quality Cut for how messy a track may be sliced (trackyness)
      if ((NCELLX == false)&&(NCELLY == false))
	NCELL = true;

      //Quality Cut for both views having at more than 3 planes in a track (stricter than the total 8 plane cut)
      if (nplanex > 3)
	NPLANEX = true;
      if (nplaney > 3)
	NPLANEY = true;
      if ((NPLANEX == true)&&(NPLANEY == true))
	NPLANE = true;

      //Quality Cut for a track to have more than 8 planes in total of both view (total 8 plane cut)
      npt = nplanex + nplaney;
      if (npt > 8)
	NPLANETOTAL = true;

      //Quality Cut that requires that the track have a hit in sequential/continuous planes
      for(int jjj = 0; jjj < geomplanes; ++jjj){  
	if ((planetotal[jjj] == 1)&&(planetotal[jjj-1] == 1)&&(planetotal[jjj-2] == 1)&&(planetotal[jjj-3] == 1)){// 4 plane cut
	  CONTINUOUSfour = true;
	}
	if ((planetotal[jjj] == 1)&&(planetotal[jjj-1] == 1)&&(planetotal[jjj-2] == 1)&&(planetotal[jjj-3] == 1)&&(planetotal[jjj-4] == 1)){// 5 plane cut
	  CONTINUOUSfive = true;
	}
      }
      
      //END CUTS
      
      
      ///////////////////////////////////////////////
      ////                                       ////
      ////               Cell Hits               ////
      ////                                       ////
      ///////////////////////////////////////////////
      
      
      nplane   = npt;
      nxplane  = nplanex;
      nyplane  = nplaney;

      int daq_channel = 0;
      int feb         = 0;
      int dcm         = 0;
      int diblock     = 0;
      int rawplane    = 0;
      int rawcell     = 0;

      for(signed int d = 0; d < ncell; ++d){
	
	art::Ptr<rb::CellHit> cellhits=track->Cell(d);
	
	const double www = track->W(cellhits.get());

	rb::RecoHit recohits(cal->MakeRecoHit(*cellhits,www));

        // No point writing uncalibrated hits to TTrree
        if(!recohits.IsCalibrated()) continue;
	
	short plane        = cellhits->Plane();             //Plane Number
	short cell         = cellhits->Cell();              //Cell Number
	double photoel     = cellhits->PE();                //Number of Photoelectrons of Hit
	double photoelcorr = recohits.PECorr();             //Number of Photoelectrons of Hit Corrected with Attenuation
	double CellTime    = cellhits->TNS();               //Time of hit with no channel offset (in nanoseconds)
	double x           = recohits.X();                  //Reconstructed x-position
	double y           = recohits.Y();                  //Reconstructed y-position
	double z           = recohits.Z();                  //Reconstructed z-position
	double RecoTime    = recohits.T();                  //Reconstructed time
	
	int DIB = 0;  //diblock
	int DCM = 0;  //d
	int FEB = 0;  //front end board

	int Qual = 0;

	double AverageCellPathLength = -99;
	double BestEstCellPathLength = -99;
	
	if (cellhits->View()==geo::kX){
	  if ((dx > 0)||(dy > 0)||(dz > 0)){
	    AverageCellPathLength = geo::AverageCellPathLength(geo::kX,dx,dy,dz);
	  }
	}
	if (cellhits->View()==geo::kY){
	  if ((dx > 0)||(dy > 0)||(dz > 0)){
	    AverageCellPathLength = geo::AverageCellPathLength(geo::kY,dx,dy,dz);
	  }
	}
	
	BestEstCellPathLength = pathlengths[d];
	
	if (quals[d]==calib::kXYAdjacents){
	  Qual = 1;
	}
	if (quals[d]==calib::kZAdjacents){
	  Qual = 2;
	}
	if (quals[d]==calib::kAverage){
	  Qual = 3;
	}
	if (quals[d]==calib::kBadTrack){
	  Qual = 4;
	}
	for (unsigned int delta = 0; delta < rawhitcol->size(); delta++){
	  daq_channel                      = rawhitlist[delta]->DaqChannel();
	  daqchannelmap::lchan log_channel = cmap->Map()->encodeLChan(daq_channel);
          feb                              = cmap->Map()->getFEB(daq_channel);
          dcm                              = cmap->Map()->getDCM(daq_channel);
          diblock                          = cmap->Map()->getDiBlock(daq_channel);
          rawplane                         = cmap->Map()->getPlane(log_channel);
          rawcell                          = cmap->Map()->getCell(log_channel);

	  if ((rawplane == plane)&&(rawcell == cell)){
	    DIB = diblock;
	    DCM = dcm;
	    FEB = feb;
	  }
	}
	
	//-------------------------------------------------Tree Fill---------------------------------------------------------
	
	fsummary->EventNumber             = EventNumber;
	fsummary->RunNumber               = RunNumber;
	fsummary->SubrunNumber            = SubrunNumber;
	fsummary->SubrunTime              = EventTimeHigh;
	fsummary->EventTime               = EventTimeLow;
	fsummary->TrackNumber             = TrackNumber;
	fsummary->TrackSliceNumber        = TrackSliceNumber;
	
	fsummary->NCell                   = ncell;
	fsummary->NXCell                  = nxcell;
	fsummary->NYCell                  = nycell;
	fsummary->NPlane                  = nplane;
	fsummary->NXPlane                 = nxplane;
	fsummary->NYPlane                 = nyplane;
	
	fsummary->TrackLength             = tracklength;
	fsummary->xTrackStart             = trksx;
	fsummary->yTrackStart             = trksy;
	fsummary->zTrackStart             = trksz;	  
	fsummary->xTrackEnd               = trkex;
	fsummary->yTrackEnd               = trkey;
	fsummary->zTrackEnd               = trkez;
	fsummary->dX                      = dx;
	fsummary->dY                      = dy;
	fsummary->dZ                      = dz;
	
	if (cellhits->View()==geo::kX){
	  fsummary->view                  = 0;
	}
	if (cellhits->View()==geo::kY){
	  fsummary->view                  = 1;
	}
	fsummary->X                       = x;
	fsummary->Y                       = y;
	fsummary->Z                       = z;
	fsummary->AverageCellPathLength   = AverageCellPathLength;
	fsummary->BECPL                   = BestEstCellPathLength;
	fsummary->TrackQuality            = Qual;
	fsummary->cell                    = cell;
	fsummary->plane                   = plane;	
	fsummary->CellTime                = CellTime;
	fsummary->RecoTime                = RecoTime;
	fsummary->PE                      = photoel;          //energy deposition of every cell
	fsummary->PECorr                  = photoelcorr;      //energy deposition of every cell after attenuation correction
	
	fsummary->DiBlock                 = DIB;
	fsummary->DCM                     = DCM;
	fsummary->FEB                     = FEB;
	fsummary->DAQChannel              = daq_channel;
	
	fsummary->CutPlaneNumber          = NPLANE;           //plane number > 3 planes in each view
	fsummary->CutPlaneNumberTotal     = NPLANETOTAL;      //plane number > 8 planes
	fsummary->CutContinuousTrackFour  = CONTINUOUSfour;
	fsummary->CutContinuousTrackFive  = CONTINUOUSfive;
	fsummary->CutTrackiness           = NCELL;
	
	fSummaryTree->Fill();
	
	    //-------------------------------------------------Tree Fill---------------------------------------------------------
      }
      //END CELL HITS     
    } 
    //END X-VIEW
    
    return;
  }
} // end namespace c2c
/////////////////////////////////////////////////////////////////////////  

namespace c2c
{
  DEFINE_ART_MODULE(CellToCell);
}

