/// 
/// \brief   Interface to the run-by-run list of bad channels
/// \author  betancourt@physics.umn.edu, jpaley@anl.gov
/// \date
///
#ifndef CHANINFO_BADCHANLIST_H
#define CHANINFO_BADCHANLIST_H

#include <map>
#include <string>

// NOvA includes
#include "RawData/RawDigit.h"
#include "GeometryObjects/OfflineChan.h"
#include "NovaDAQConventions/DAQConventions.h"

//Framework includes
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Provenance/Timestamp.h"
#include "fhiclcpp/ParameterSet.h"

namespace nova{namespace dbi{class Table;}}

namespace chaninfo {

  ///
  /// Interface to the run-by-run list of bad channels
  ///
  class BadChanList {
  public:
    // Get a BadChanList instance here
    BadChanList(fhicl::ParameterSet const& pset, art::ActivityRegistry &reg);
    ~BadChanList();

    void postBeginRun(art::Run const& run);
    //    void postBeginSubRun(art::SubRun const& subrun);
    void preBeginEvent ( art::Event const & evt);

    int  Apply(art::PtrVector<rawdata::RawDigit>& rd);
    int  Apply(std::vector< art::Ptr<rawdata::RawDigit> >& rd);
    int  Apply(std::vector< const rawdata::RawDigit*>&);

    bool IsBad(int plane, int cell);
    bool IsLowEfficiency(int plane, int cell);
    void reconfigure(fhicl::ParameterSet p);

    void LoadNewData();

  private:
    void SetRandomBadChannels();

    void LoadCosmicEff(int newrun, int newsubrun);

    // private data members for BadChanList class
    int                                  fApplyBadChan;           ///< Option to activate the bad channels   
    int                                  fMCRunNumber;            ///< Run number for the MC
    int                                  fStateIdx;  ///< index of "state" column in chan_rate_state dB table

    bool                                 fApplyCoincidenceCut;    ///< Option to cut on the coincidence rate variable

    bool                                 fIsMC;

    bool                                 fNeedNewData;

    float                                fCoincidenceThreshold;
    std::map<geo::OfflineChan, bool >    fLowEfficiency;

    std::map<geo::OfflineChan, bool >    fRandomBCMap;   ///< random bad channel map

    std::string                          fBadChanCSVFile;

    bool fNewCosmicEffTable; ///< Use new-style table?

    float fRandomBadChannelFraction; ///< If >0, ignore DB and set this fraction bad at random

    std::unique_ptr<nova::dbi::Table> fBCTable;

    art::Timestamp fEvtTime;

    bool fIsPerfect;

    std::string fValidityTag;

    time_t fMCVldTime;
  };
}
DECLARE_ART_SERVICE(chaninfo::BadChanList, LEGACY)
#endif
////////////////////////////////////////////////////////////////////////
