///////////////////////////////////////////////////////////////////////////
/// \brief   Interface to the run-by-run list of bad channels
/// \author  betancourt@physics.umn.edu
//////////////////////////////////////////////////////////////////////////
#include "ChannelInfo/BadChanList.h"

#include <sys/stat.h>
#include <sstream>

// NOvA includes
#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "SummaryData/DataRunNumInMC.h"
#include "SummaryData/RunData.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "DAQChannelMap/DAQChannelMap.h"

//#include "NovaDatabase/Table.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/service/RunHistoryService.h"

// Framework includes
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "cetlib/search_path.h"
#include "cetlib/exception.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "TRandom3.h"
#include "TString.h"

namespace chaninfo {

  //------------------------------------------------------------
  BadChanList::BadChanList(fhicl::ParameterSet const& pset,
                           art::ActivityRegistry &reg) :
    fNewCosmicEffTable(false)
  {
    fRandomBCMap.clear();
    fIsMC = false;
    fNeedNewData = true;
    fValidityTag = "";
    fMCRunNumber = -1;
    fMCVldTime = 0;
    this->reconfigure(pset);
    reg.sPostBeginRun.watch(this, &BadChanList::postBeginRun);
    //    reg.sPostBeginSubRun.watch(this, &BadChanList::postBeginSubRun);
    reg.sPreProcessEvent.watch(this, &BadChanList::preBeginEvent);    
  }

  //----------------------------------------------------------

  BadChanList::~BadChanList()
  {
    fRandomBCMap.clear();
    fBCTable.reset();

  }

  //----------------------------------------------------------
  void BadChanList::postBeginRun(art::Run const & run)
  {
    if (run.run() >= 1000000 || run.run() < 10000)
      fIsMC = true;
    else
      fIsMC = false;

    fNeedNewData = true;

  }

  //----------------------------------------------------------
  void BadChanList::preBeginEvent(art::Event const & event)
  {
    fEvtTime = event.time();
  }

  //-----------------------------------------------------------
  void BadChanList::SetRandomBadChannels()
  {
    if (!fRandomBCMap.empty()) return;

    std::cerr << " RANDOM BAD CHANNELS at rate "
              << fRandomBadChannelFraction << std::endl;

    fLowEfficiency.clear();

    // Intentionally don't seed for now, to always get reproducible results
    TRandom3 rnd;

    art::ServiceHandle<geo::Geometry> geom;
    for(unsigned int plane: geom->GetPlanesByView()){
      const unsigned int nCells = geom->Plane(plane)->Ncells();

      for(unsigned int cell = 0; cell < nCells; ++cell){
	geo::OfflineChan chan(plane, cell);
	fRandomBCMap[chan] = (rnd.Uniform(0, 1) > fRandomBadChannelFraction);
      }
    }
  }

  //-----------------------------------------------------------
  void BadChanList::LoadCosmicEff(int newrun, int newsubrun)
  {
    fLowEfficiency.clear();

    nova::dbi::Table tblEff(fNewCosmicEffTable ?
			    "/ChannelInfo/CosmicEffNew.xml" :
			    "/ChannelInfo/CosmicEff.xml");
//    tblEff.SetDetector("NDOS");
//    tblEff.SetDetector(fDetector);

    //    tblEff.SetValidityRange("release", fTableVersion);


    /*
    if(fNewCosmicEffTable){
      tblEff.SetRunVld(newrun);
      tblEff.SetSubrunVld(newsubrun);
    }
    else {
    */
    tblEff.SetValidityRange("detector", "ndos");
    tblEff.SetValiditySQL(TString::Format("(beginrun<%d or (beginrun=%d and beginsubrun<=%d)) and "
					  "(endrun>%d or (endrun=%d and endsubrun>=%d))",
					  newrun, newrun, newsubrun, newrun, newrun, newsubrun).Data());
    //    }

    // Only ask for the bad channels. We won't get any "good" ones back, so
    // they won't go in the map, which is fine.
    tblEff.SetValidityRange("good", false);

    tblEff.SetVerbosity(100);
    const bool ok = tblEff.Load();
    if(!ok){
      // No safe way to handle this
      std::cerr << "Couldn't load efficiencies channels from DB. Aborting" << std::endl;
      abort();
    }

    int planeIdx = tblEff.GetColIndex("plane");
    int cellIdx = tblEff.GetColIndex("cell");
    int goodIdx = tblEff.GetColIndex("good");

    for(int i = 0; i < tblEff.NRow(); ++i){
      int plane, cell;
      bool good;
      bool ok = tblEff.GetRow(i)->Col(planeIdx).Get(plane);
      ok = ok && tblEff.GetRow(i)->Col(cellIdx).Get(cell);
      ok = ok && tblEff.GetRow(i)->Col(goodIdx).Get(good);
      if(!ok) throw;
      geo::OfflineChan chan(plane, cell);
      // DB stores "goodness", we store "badness"
      fLowEfficiency[chan] = !good;
    }
  }

  //-----------------------------------------------------------
  void BadChanList::reconfigure(fhicl::ParameterSet pset)
  {
    fCoincidenceThreshold    = pset.get<float>("CoincidenceThreshold");
    fApplyBadChan            = pset.get<int>("ApplyBadChan");
    fApplyCoincidenceCut     = pset.get<bool>("ApplyCoincidenceCut");
    fMCRunNumber             = pset.get<int>("MCRunNumber");
    fNewCosmicEffTable       = pset.get<bool>("NewCosmicEffTable");
    fRandomBadChannelFraction= pset.get<float>("RandomBadChannelFraction");
    fBadChanCSVFile = pset.get<std::string>("BadChanCSVFile");
    fValidityTag = pset.get<std::string>("ValidityTag");
  }

  //----------------------------------------------------------
  void BadChanList::LoadNewData()
  {
    art::ServiceHandle<geo::Geometry> geom;

    // reset bad channel table; set vld time to 0, so that EnsureRun 
    // will correctly set validity times
    fBCTable.reset(new nova::dbi::Table("/ChannelInfo/ChanRateState.xml"));
    fBCTable->SetVerbosity(100);
    fBCTable->SetDetector(geom->DetId());
    if (fValidityTag != "")
      fBCTable->SetTag(fValidityTag);
    fStateIdx = fBCTable->GetColIndex("state");

    // for now, just deal with FarDet
    if (geom->DetId() == novadaq::cnv::kFARDET ||
	geom->DetId() == novadaq::cnv::kNEARDET) {
      art::ServiceHandle<nova::dbi::RunHistoryService> rh;
      fBCTable->SetDataTypeMask(nova::dbi::kMCOnly|nova::dbi::kDataOnly);
      fBCTable->SetMinTSVld(rh->TStart()-120);
      fBCTable->SetMaxTSVld(rh->TStop()+120);	  
      
      // add a couple of minutes of buffer to start and stop times, 
      // just in case
      if (!fBCTable->Load()) 
	throw cet::exception("BadChannelTableLoadFail") << "Bad Channels table failed to load data!";
      if (fBCTable->NRow() == 0) 
	std::cerr << "******** WARNING: BadChanList_service found no bad " 
		  << "channels!" << std::endl; 
      fBCTable->FillChanRowMap();
    }
    fNeedNewData = false;
  }

  //----------------------------------------------------------
  bool BadChanList::IsBad(int plane, int cell)
  {
    if(!fApplyBadChan) return false;

    geo::OfflineChan chan(plane,cell);

    if(fRandomBadChannelFraction > 0){
      SetRandomBadChannels();
      return fRandomBCMap[chan];
    }

    if (fNeedNewData) LoadNewData();

    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    art::ServiceHandle<geo::Geometry> geom;

    // for now, just deal with FarDet.
    if (geom->DetId() == novadaq::cnv::kFARDET ||
	geom->DetId() == novadaq::cnv::kNEARDET) {

      // check with RH whether this channel was instrumented, part of the partition, or masked
      if (!rh->HasAPD(chan)) return true;

      time_t tvld = time_t(fEvtTime.timeHigh());
      int tchan = chan.ToDBValidityChan();

      // get validity row for this timestamp
      nova::dbi::Row* row = fBCTable->GetVldRow(tchan,tvld);

      // if there is no entry that corresponds to this timestamp, assume
      // all is ok
      if (!row) return false;

      // there is an entry for this timestamp (expected behavior), so
      // make sure that the state is "ok" (==0)
      int state;
      bool ok = row->Col(fStateIdx).Get(state);
      assert(ok);

      if (state == 0) return false;

      else return true;

    }

    // If we can't find anything wrong with it then it must not be bad
    return false;
  }

  //----------------------------------------------------------
  bool BadChanList::IsLowEfficiency(int plane, int cell)
  {
    if(!fApplyBadChan) return false;

    const geo::OfflineChan chan(plane, cell);
    return fLowEfficiency[chan];
  }

  //--------------------------------------------------------
  
  int BadChanList::Apply(std::vector< art::Ptr<rawdata::RawDigit> >& rd)
  {
    art::ServiceHandle<cmap::CMap> cmap;
    int ch;
    std::vector< art::Ptr<rawdata::RawDigit> > good;
    for(unsigned int i=0; i<rd.size(); ++i) {
      int plane = cmap->GetPlane(rd[i].get());
      int cell =  cmap->GetCell(rd[i].get());
      if(!IsBad(plane,cell)) good.push_back(rd[i]);
    }
    rd.swap(good);
    ch=rd.size();
    return ch;
  }
  
  //--------------------------------------------------------
  
  int BadChanList::Apply(art::PtrVector<rawdata::RawDigit>& rd)
  {
    art::ServiceHandle<cmap::CMap> cmap;
    int ch;
    art::PtrVector<rawdata::RawDigit> good;
    for(unsigned int i=0; i<rd.size(); ++i) {
      int plane = cmap->GetPlane(rd[i].get());
      int cell =  cmap->GetCell(rd[i].get());
      if(!IsBad(plane,cell)) good.push_back(rd[i]);
    }
    rd.swap(good);
    ch=rd.size();
    return ch;
  }

  //--------------------------------------------------------

  int BadChanList::Apply( std::vector<const rawdata::RawDigit*>& rd)
  {
    art::ServiceHandle<cmap::CMap> cmap;
    int ch;
    std::vector<const rawdata::RawDigit*> good;
    for(unsigned int i=0; i<rd.size(); ++i) {
      int plane = cmap->GetPlane(rd[i]);
      int cell =  cmap->GetCell(rd[i]);
      if(!IsBad(plane,cell)) good.push_back(rd[i]);
    }
    rd.swap(good);
    ch=rd.size();
    return ch;
  }



  DEFINE_ART_SERVICE(BadChanList);

}
////////////////////////////////////////////////////////////////////////
