#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <map>
#include <cstdlib>
#include <ctime>
#include <getopt.h>
#include <boost/filesystem.hpp>

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/RunHistory.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "DAQChannelMap/HardwareDisplay.h"

int run=0;
bool write2CSV=false;
std::string csvFileName="";
std::string validityTag="";
std::string det="";

//------------------------------------------------------------

void PrintUsage()
{
  std::cout << "Usage: showBadChannels [options] [detector] [run]" << std::endl;
  std::cout << "options:\n";
  std::cout << "\t -t (--tag) [tag]: set validity tag" << std::endl;
  std::cout << "\t -c (--tocsv) [filename]: write results to csv file" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int argc, char* argv[])
{
  if (argc == 1) {
    PrintUsage();
    exit(0);
  }

  struct option long_options[] = {
    {"tocsv", 0, 0, 'c'},
    {"tag", 0, 0, 't'},
    {0,0,0,0}
  };
  
  while (1) {
    int optindx;
    int c = getopt_long(argc,argv,"c:t:",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
    case 'c':
      write2CSV=true;
      csvFileName=optarg;
      break;
    case 't':
      validityTag=optarg;
      break;
    default:
      break;
    }
  }
  
  if ((argc-optind) < 2) {
    std::cerr << "Missing arguments." << std::endl;
    PrintUsage();
    exit(0);
  }
  
  if ((argc-optind) > 2) {
    std::cerr << "Unrecognized arguments." << std::endl;
    PrintUsage();
    exit(0);
  }
  
  det = argv[optind++];
  run = atoi(argv[optind++]);

  return true;
}

//------------------------------------------------------------

int main(int argc, char *argv[])
{
  ParseCLArgs(argc, argv);

  int detId = novadaq::cnv::DetInfo::GetId(det);
  if (detId != novadaq::cnv::kFARDET) {
    std::cout << "No bad channels available for " << det << " at this time"
	      << std::endl;
    exit(0);
  }

  nova::dbi::Table* t = new nova::dbi::Table("ChannelInfo/ChanRateState.xml");
  t->SetDetector(det);
  t->SetDataTypeMask(nova::dbi::kMCOnly|nova::dbi::kDataOnly);
  if (validityTag != "") 
    t->SetTag(validityTag);

  if (run > 1000000) run -= 1000000;
  nova::dbi::RunHistory rh(run,det);

  if (!rh.LoadDAQRunHistory()) {
    std::cerr << "RunHistory::LoadDAQRunHistory failed, aborting." << std::endl;
    exit(1);
  }

  rh.LoadHardwareInfo();

  t->SetMinTSVld(rh.TStart()+1);
  t->SetMaxTSVld(rh.TStop()-1);	  
  
  if (!t->Load()) {
    std::cerr << "Failed to load bad channels data!" << std::endl;
    exit(1);
  }
  if (t->NRow() == 0) 
    std::cerr << "No bad channels found!" << std::endl; 

  t->FillChanRowMap();

  daqchannelmap::DAQChannelMap* daqcmap = daqchannelmap::DAQChannelMap::getInstance(detId);

  int ipix=1;
  int state;
  int nbad=0;
  int stateIdx=t->GetColIndex("state");

  for (int idb=0; idb<rh.NDiBlocks(); ++idb) {
    nova::dbi::RunHistory::DiBlock dblk = rh.GetDiBlock(idb,false);
    for (unsigned int idcm=0; idcm<dblk.dcm.size(); ++idcm) {
      for (unsigned int ifeb=0; ifeb<dblk.dcm[idcm].feb.size(); ++ifeb) {
	if (dblk.dcm[idcm].feb[ifeb].hasAPD) {
	  daqchannelmap::dchan daqChan = daqcmap->encodeDChan(detId,dblk.num,
							      idcm+1,ifeb,
							      ipix);
	  daqchannelmap::lchan logChan = daqcmap->encodeLChan(daqChan);
	  
	  int plane = daqcmap->getPlane(logChan);
	  int cell = daqcmap->getCell(logChan);
	  geo::OfflineChan chan(plane, cell);
	  int tchan = chan.ToDBValidityChan();

	  std::vector<nova::dbi::Row*> rowVec = t->GetVldRows(tchan);
	  if (rowVec.empty()) continue;
	  assert(rowVec[0]->Col(stateIdx).Get(state));
	  if (rowVec.size() == 1 && state == 0) continue; // channel is ok
	  std::cout << detId << "-" << dblk.num << "-" << idcm << "-"
		    << ifeb << ": " << rowVec.size() << " bad channel entries"
		    << std::endl;
	  ++nbad;
	}
      }
    }
  }
  std::cout << nbad << " bad channels found for run " << run << std::endl;
      
  return 0;
}
