/////////////////////////////////////////////////////////
/// \brief  Cosmic Rejection PIDs
/// \author Kirk Bays
////////////////////////////////////////////////////////
#include <string>

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "CosRej/CosRejFxs.h"
#include "CosRej/CosRejObj.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "MCCheater/BackTracker.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "ReMId/ReMId.h"
#include "TimingFit/TimingFitAlg.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/EnvExpand.h"

#include "TFile.h"
#include "TMVA/Reader.h"

namespace cosrej {

  class CosRej : public art::EDProducer {
  public:
    explicit CosRej(fhicl::ParameterSet const & pset);
    virtual ~CosRej();

    void produce (art::Event & evt);
    void beginRun(art::Run& run);

  protected:

    cosrej::CosRejFxs* fCosRejFxs;
    std::string        fSlicerLabel;
    std::string        fTrackLabel;
    std::string        fPIDLabel;
    std::string        fTMVApath;
    bool fUseLongestTrack;
    TMVA::Reader*      fReader;
    TMVA::Reader*      fReaderuc;
    float              TMVAvars[23];
  };
}


namespace cosrej
{

  //----------------------------------------------------------------------
  CosRej::CosRej(fhicl::ParameterSet const & pset):
    fCosRejFxs        (new CosRejFxs()),
    fSlicerLabel      (pset.get< std::string >("SlicerLabel")),
    fTrackLabel       (pset.get< std::string >("TrackLabel")),
    fPIDLabel         (pset.get< std::string >("PIDLabel")),
    fTMVApath         (pset.get< std::string >("tmvapath")),
    fUseLongestTrack  (pset.get< bool        >("UseLongestTrack")),
    fReader           (nullptr)
  {
    produces< std::vector<cosrej::CosRejObj> >();
    produces< art::Assns<cosrej::CosRejObj, rb::Cluster> >();

  }


  //----------------------------------------------------------------------
  CosRej::~CosRej()
  {
      if(fCosRejFxs) delete fCosRejFxs;
      if(fReader)    delete fReader;
  }

  //-------------------------------------------------------------------
  void CosRej::beginRun(art::Run& run)
  {

    art::ServiceHandle<geo::Geometry>  geom;
    if(geom->DetId() == novadaq::cnv::kFARDET) {
      fReader = new TMVA::Reader;
      fReader->AddVariable("anglekal",&TMVAvars[0]);
      fReader->AddVariable("anglecos",&TMVAvars[1]);
      fReader->AddVariable("cosdiry",&TMVAvars[2]);
      fReader->AddVariable("kaldiry",&TMVAvars[3]);
      fReader->AddVariable("maxy",&TMVAvars[4]);
      fReader->AddVariable("cosratio",&TMVAvars[5]);
      fReader->AddVariable("cosfwdcell",&TMVAvars[6]);
      fReader->AddVariable("cosbakcell",&TMVAvars[7]);
      fReader->AddVariable("kalfwdcell",&TMVAvars[8]);
      fReader->AddVariable("kalbakcell",&TMVAvars[9]);
      fReader->AddVariable("scatt/tracklen",&TMVAvars[10]);
      fReader->AddVariable("hade",&TMVAvars[11]);
      fReader->AddVariable("visE",&TMVAvars[12]);
      fReader->AddSpectator("slicetime",&TMVAvars[13]);
      fReader->AddSpectator("remid",&TMVAvars[14]);
      fReader->AddSpectator("cbdtg",&TMVAvars[15]);
      fReader->AddSpectator("osc",&TMVAvars[16]);
      fReader->AddSpectator("misse",&TMVAvars[17]);
      fReader->AddSpectator("nhit",&TMVAvars[18]);
      fReader->AddSpectator("trueE",&TMVAvars[19]);
      fReader->AddSpectator("ntracks3d",&TMVAvars[20]);
      fReader->AddSpectator("type",&TMVAvars[21]);
      fReader->AddSpectator("vtx",&TMVAvars[22]);
      std::string pidpath = util::EnvExpansion(fTMVApath);
      fReader->BookMVA("CC_BDTA",pidpath+"CC_BDTA.weights.xml");

      fReaderuc = new TMVA::Reader;
      fReaderuc->AddVariable("anglekal",&TMVAvars[0]);
      fReaderuc->AddVariable("anglecos",&TMVAvars[1]);
      fReaderuc->AddVariable("cosdiry",&TMVAvars[2]);
      fReaderuc->AddVariable("kaldiry",&TMVAvars[3]);
      fReaderuc->AddVariable("maxy",&TMVAvars[4]);
      fReaderuc->AddVariable("cosratio",&TMVAvars[5]);
      fReaderuc->AddVariable("cosfwdcell",&TMVAvars[6]);
      fReaderuc->AddVariable("cosbakcell",&TMVAvars[7]);
      fReaderuc->AddVariable("kalfwdcell",&TMVAvars[8]);
      fReaderuc->AddVariable("kalbakcell",&TMVAvars[9]);
      fReaderuc->AddVariable("scatt/tracklen",&TMVAvars[10]);
      fReaderuc->AddVariable("hade",&TMVAvars[11]);
      fReaderuc->AddVariable("visE",&TMVAvars[12]);
      fReaderuc->AddSpectator("slicetime",&TMVAvars[13]);
      fReaderuc->AddSpectator("remid",&TMVAvars[14]);
      fReaderuc->AddSpectator("cbdtg",&TMVAvars[15]);
      fReaderuc->AddSpectator("osc",&TMVAvars[16]);
      fReaderuc->AddSpectator("misse",&TMVAvars[17]);
      fReaderuc->AddSpectator("nhit",&TMVAvars[18]);
      fReaderuc->AddSpectator("trueE",&TMVAvars[19]);
      fReaderuc->AddSpectator("ntracks3d",&TMVAvars[20]);
      fReaderuc->AddSpectator("type",&TMVAvars[21]);
      fReaderuc->AddSpectator("vtx",&TMVAvars[22]);
      fReaderuc->BookMVA("UC_BDTA",pidpath+"UC_BDTA.weights.xml");
    }

  } //end beginRun

  //-------------------------------------------------------------------
  void CosRej::produce(art::Event& evt)
  {

    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;
    art::ServiceHandle<cheat::BackTracker> bt;
    art::Handle<std::vector<rb::Cluster> > slicevec;
    evt.getByLabel(fSlicerLabel,slicevec);

    if(slicevec->empty()) {
      mf::LogWarning ("No Slices")<<"No Slices in the input file";
      return;
    }

    art::FindManyP<rb::Track> trackAssnList(slicevec, evt, fTrackLabel);
    art::FindManyP<rb::Track> costrackAssnList(slicevec, evt, "cosmictrack");
    std::unique_ptr< std::vector<cosrej::CosRejObj> > outputObjects(new std::vector<cosrej::CosRejObj>);
    std::unique_ptr< art::Assns<cosrej::CosRejObj, rb::Cluster> > assoc(new art::Assns<cosrej::CosRejObj, rb::Cluster>);

    tf::TimingFitAlg tfalg;

    for(size_t i = 0; i < slicevec->size(); ++i){

      art::Ptr<rb::Cluster> slice(slicevec,i);

      if(slice->IsNoise()) {
        continue;
      }

      cosrej::CosRejObj cosrejobj; // initialization in constructor

      int bestIdx = -1;  // for index in list of the highest ReMId PID track
      int longIdx = -1;  // for index in list of the longest track
      int useIdx = -1;   // which index we use
      double bestval = -999., longest = -999.;
      TVector3 minp, maxp; // minimum and maximum extent vectors for slice
      int nhadtracks = 0;
      double fwddist = 0, bakdist = 0,  scatt = 0;
      int fwdcell = 0, bakcell = 0;
      bool badtrack = false;    // check for messed up tracks; a few kalman tracks have nans for directions
      float  walldist;

      const std::vector< art::Ptr<rb::Track> > cosTracks = costrackAssnList.at(i);

      minp = slice->MinXYZ();
      maxp = slice->MaxXYZ();

      if(!costrackAssnList.isValid()) {
        mf::LogError("CosRej") << "CosRej: No Cosmic Tracks! Do not go any further!";
        outputObjects->push_back(cosrejobj);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        return;
      }

      if(cosTracks.size()==1) {
        const art::Ptr<rb::Track> track = cosTracks[0];
        double angleOfBestTrack = fCosRejFxs->getAngle(track->Dir());
        if(angleOfBestTrack>-6&&angleOfBestTrack<2) cosrejobj.SetAngleCos(angleOfBestTrack);
        fCosRejFxs->getCosmicVars(track,fwddist,bakdist,fwdcell,bakcell,walldist);

        cosrejobj.SetCosFwdDist(fwddist);
        cosrejobj.SetCosBakDist(bakdist);
        cosrejobj.SetCosFwdCell(fwdcell);
        cosrejobj.SetCosBakCell(bakcell);

        if(geom->DetId() == novadaq::cnv::kNEARDET) {
          TVector3 dirfor = (track->TrajectoryPoint(track->NTrajectoryPoints()-1) - track->TrajectoryPoint(track->NTrajectoryPoints()-2)).Unit();

          cosrejobj.SetCosFwdSteel(livegeom->ProjectedDistanceInMC(track->Stop(), dirfor));
          cosrejobj.SetCosBakSteel(livegeom->ProjectedDistanceInMC(track->Start(), -(track->Dir())));
          cosrejobj.SetCosYPosAtTrans(livegeom->YPositionAtTransition(track->Start(), track->Dir()));
        }

        TMVAvars[2] = track->Dir().Y();
        TMVAvars[5] = (1.0*track->NCell())/(1.0*slice->NCell());
        //        if(track->AllCells().size()>2)fCosRejFxs->getFits(track,slope,chisq,chisqdiff);  Timing Fits are currently broken
        //        cosrejobj.SetCosSlope(slope);
        //        cosrejobj.SetCosChisq(chisq);
        //        cosrejobj.SetCosChiDiff(chisqdiff);
        tf::TimingFitResult tfres = tfalg.HoughFit(track.get());
        cosrejobj.SetCFitSpeed(tfres.bestInvSpeed);
        cosrejobj.SetCDirScore(tfres.fwdScore-tfres.revScore);
        cosrejobj.SetCScoreDiff(tfres.bestScore-std::max(tfres.fwdScore,tfres.revScore));  // Chris's timing fit (Hough)
        cosrejobj.SetCosHitRatio(1.0*track->NCell()/(1.0*slice->NCell()));
        bool cont, weakangle, strongangle;
        float cvfd, cvbd;
        fCosRejFxs->getCVVars(track, slice->NCell(), strongangle, weakangle, cont, cvfd, cvbd);
        cosrejobj.SetCVCont(cont);
        cosrejobj.SetCVWeakAngle(weakangle);
        cosrejobj.SetCVStrongAngle(strongangle);
        cosrejobj.SetCVFwdDist(cvfd);
        cosrejobj.SetCVBakDist(cvbd);

        if(bt->HaveTruthInfo()) {
          art::PtrVector<rb::CellHit> trkhits = track->AllCells();
          const std::vector<const sim::Particle*> parts = bt->HitsToParticle(trkhits);

          if(!parts.empty()){
            const TVector3 trueDir = parts[0]->Momentum().Vect().Unit();
            cosrejobj.SetCosThetaTrue(trueDir.Dot(track->Dir()));
          }
        } // end cosmic truth
      } // end cosmics

      if(!trackAssnList.isValid()) {
        mf::LogWarning("CosRej") << "CosRej: No Kalman Tracks!";
        outputObjects->push_back(cosrejobj);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        continue;
      }

      const std::vector< art::Ptr<rb::Track> > sliceTracks = trackAssnList.at(i);
      art::FindOneP<remid::ReMId> remidVec(sliceTracks, evt, fPIDLabel);

      if(fUseLongestTrack) {
        for(size_t j = 0; j < sliceTracks.size(); ++j) {
          const art::Ptr<rb::Track> track = sliceTracks[j];
          if(!track->Is3D()) continue;
          if(cosrejobj.NTracks3D()==-5) cosrejobj.SetNTracks3D(1);
          else cosrejobj.SetNTracks3D(cosrejobj.NTracks3D()+1);
        }
        longIdx = 0;
      }

      else if(!remidVec.isValid()) {
        mf::LogWarning("CosRej") << "CosRej: No ReMId information!";
        outputObjects->push_back(cosrejobj);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        continue;
      }

      else { for(size_t j = 0; j < sliceTracks.size(); ++j){
        const art::Ptr<rb::Track> track = sliceTracks[j];
        art::Ptr<remid::ReMId> trackRemid = remidVec.at(j);

        if(!track->Is3D() || trackRemid->Value() < 0) {
          continue;
        }

        if(cosrejobj.NTracks3D()==-5) cosrejobj.SetNTracks3D(1);
        else cosrejobj.SetNTracks3D(cosrejobj.NTracks3D()+1);

        if(trackRemid->Value() < 0.5) nhadtracks++;

        if(trackRemid->Value() > bestval) {
          bestval = trackRemid->Value();
          bestIdx = j;
        }

        if(track->TotalLength() > longest) {
          longest = sliceTracks[j]->TotalLength();
          longIdx = j;
        }

        art::PtrVector<rb::CellHit> trkhits = track->AllCells();

        cosrejobj.SetRPID(bestval);

        if(bt->HaveTruthInfo()) {
          const std::vector<const sim::Particle*> parts = bt->HitsToParticle(trkhits);
          const std::vector< cheat::TrackIDE > trids = bt->HitsToTrackIDE(trkhits);

          if(trackRemid->Value() == bestval) {  // get cos(theta_true) for best kal track

            if(!parts.empty()){
              const TVector3 trueDir = parts[0]->Momentum().Vect().Unit();
              cosrejobj.SetKalThetaTrue(trueDir.Dot(track->Dir()));
              cosrejobj.SetPDGBest(parts[0]->PdgCode());
            }
          }
        }
        }} //end loop over tracks and else

      if(cosrejobj.NTracks3D()>0 && (bestIdx>-1||longIdx>-1)) {

        if(bestIdx>-1 && bestval >= 0 && bestval <=1 ) useIdx = bestIdx;
        else if (longIdx>-1) useIdx = longIdx;
        else {
          mf::LogError("CosRej") << "Couldn't find a good track; this should never happen!";
          return;
        }

        const art::Ptr<rb::Track> track = sliceTracks[useIdx];
        double tdx = track->Dir().X();
        double tdy = track->Dir().Y();
        double tdz = track->Dir().Z();

        if(!(tdx>=-1&&tdx<=1&&tdy>=-1&&tdy<=1&&tdz>=-1&&tdz<=1)) badtrack = true;

        if(badtrack) continue;

        TVector3 dirfor = (track->TrajectoryPoint(track->NTrajectoryPoints()-1) - track->TrajectoryPoint(track->NTrajectoryPoints()-2)).Unit();

        fwddist = livegeom->ProjectedLiveDistanceToEdge(track->Stop(), dirfor);
        bakdist = livegeom->ProjectedLiveDistanceToEdge(track->Start(), -(track->Dir()));
        fwdcell = livegeom->ProjectedLiveCellsToEdge(track->Stop(), dirfor);
        bakcell = livegeom->ProjectedLiveCellsToEdge(track->Start(), -(track->Dir()));

        if(geom->DetId() == novadaq::cnv::kNEARDET) {
          cosrejobj.SetKalFwdSteel(livegeom->ProjectedDistanceInMC(track->Stop(), dirfor));
          cosrejobj.SetKalBakSteel(livegeom->ProjectedDistanceInMC(track->Start(), -(track->Dir())));
          cosrejobj.SetKalYPosAtTrans(livegeom->YPositionAtTransition(track->Start(), track->Dir()));
        }

        if(std::min(fwddist,bakdist) < cosrejobj.MinDist()) cosrejobj.SetMinDist(std::min(fwddist,bakdist));
        if(std::min(fwdcell,bakcell) < cosrejobj.MinCell()) cosrejobj.SetMinCell(std::min(fwdcell,bakcell));

        //        bestkaly = track->Dir().Y();
        //        nhitkal = track->NCell();
        cosrejobj.SetERatio(track->TotalGeV()/slice->TotalGeV());
        scatt = fCosRejFxs->getScatt(track);
        if(track->TotalLength()>0) cosrejobj.SetScatt(scatt/track->TotalLength());

        double ferpars[4];
        bool fdef = 0;
        bool fFlag = fCosRejFxs->FScatterEstim(track,fdef,ferpars[0],ferpars[1],ferpars[2],ferpars[3]); // Fernandas scattering energy estimate
        if(fFlag==1) {
          cosrejobj.SetFScattExt(ferpars[0]);
          cosrejobj.SetFScattSig(ferpars[1]);
          cosrejobj.SetFScattMax(ferpars[2]);
          cosrejobj.SetFScattSum(ferpars[3]);
        }
        //        if(track->AllCells().size()>2)fCosRejFxs->getFits(track,slope,chisq,chisqdiff);
        //        cosrejobj.SetKalSlope(slope);
        //        cosrejobj.SetKalChisq(chisq);
        //        cosrejobj.SetKalChiDiff(chisqdiff);
        tf::TimingFitResult tfres = tfalg.HoughFit(track.get());
        cosrejobj.SetKFitSpeed(tfres.bestInvSpeed);
        cosrejobj.SetKDirScore(tfres.fwdScore-tfres.revScore);
        cosrejobj.SetKScoreDiff(tfres.bestScore-std::max(tfres.fwdScore,tfres.revScore));
        cosrejobj.SetStartAct(fCosRejFxs->getActivity(track->AllCells(),slice->AllCells()));
        fCosRejFxs->getCosmicVars(track,fwddist,bakdist,fwdcell,bakcell,walldist);
        cosrejobj.SetKalFwdDist(fwddist);
        cosrejobj.SetKalBakDist(bakdist);
        cosrejobj.SetKalFwdCell(fwdcell);
        cosrejobj.SetKalBakCell(bakcell);
        cosrejobj.SetKalDist(walldist);
        cosrejobj.SetNHadTracks(nhadtracks);

      }

      if(!badtrack && useIdx>=0) {
        double angleOfBestTrack = fCosRejFxs->getAngle(sliceTracks[useIdx]->Dir());
        if(angleOfBestTrack>-6&&angleOfBestTrack<2) cosrejobj.SetAngleKal(angleOfBestTrack);

        if(geom->DetId() == novadaq::cnv::kFARDET) {
          const art::Ptr<rb::Track> track = sliceTracks[useIdx];
          TMVAvars[0] = cosrejobj.AngleKal();
          TMVAvars[1] = cosrejobj.AngleCos();
          TMVAvars[3] = track->Dir().Y();
          TMVAvars[4] = maxp.Y();
          TMVAvars[6] = cosrejobj.CosFwdCell();
          TMVAvars[7] = cosrejobj.CosBakCell();
          TMVAvars[8] = cosrejobj.KalFwdCell();
          TMVAvars[9] = cosrejobj.KalBakCell();
          TMVAvars[10] = cosrejobj.Scatt() / track->TotalLength();
          TMVAvars[11] = cosrejobj.StartAct();
          TMVAvars[12] = slice->CalorimetricEnergy();
          TMVAvars[13] = 0;  // after this is just spectator variables, don't need to be set even
          TMVAvars[14] = 0;
          TMVAvars[15] = 0;
          TMVAvars[16] = 0;
          TMVAvars[17] = 0;
          TMVAvars[18] = 0;
          TMVAvars[19] = 0;
          TMVAvars[20] = 0;
          TMVAvars[21] = 0;
          TMVAvars[22] = 0;

          cosrejobj.SetConCosPID(fReader->EvaluateMVA("CC_BDTA"));
          cosrejobj.SetUnconCosPID(fReaderuc->EvaluateMVA("UC_BDTA"));
        }

      } // end if good track

      outputObjects->push_back(cosrejobj);  // write out object with all information
      util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));

    } //end loop over slices

    evt.put(std::move(outputObjects));
    evt.put(std::move(assoc));

  } //end produce

} //end namespace


DEFINE_ART_MODULE(cosrej::CosRej)
