////////////////////////////////////////////////////////////////////////
// \file    NueCosRej.h
// \brief   Object collecting Cosmic Rejection variables for Nue analysis
// \version
// \author  Kanika Sachdev ksachdev@physics.umn.edu
////////////////////////////////////////////////////////////////////////
#ifndef NueCosRej_h
#define NueCosRej_h

namespace cosrej
{
  class NueCosRej {
  public:

    NueCosRej();
    ~NueCosRej();

    /// \brief Number of hits in slice per number of planes in leading shower. 
    /// This is a handle on rejecting FEB flashes. We don't use number
    /// of planes in slice because there might be too many if a far off 
    /// noise hit is sliced with physics
    double HitsPerPlane     () const { return fHitsPerPlane;}
    
    /// \brief Transverse component of the energy-weighted average 
    /// of prong directions
    double ProngTransMom         () const { return fProngTransMom;}

    /// \brief Transverse component of the event momentum. Particle masses
    /// in the momentum calculation are inferred from log-likelihoods
    /// \ref ShowerLID
    double ParticleShowerTransMom() const { return fParticleShowerTransMom;}

    /// \brief Transverse component of the event momentum. Particle masses
    /// in the momentum calculation are assumed to be 0
    double PhotonShowerTransMom  () const { return fPhotonShowerTransMom;}

    /// \brief Perpendicular distance of start point of the leading shower to the 
    /// top edge of the detector
    double StartDistToTop        () const { return fStartDistToTop;}

    /// \brief Perpendicular distance of start point of the leading shower to the 
    /// bottom edge of the detector
    double StartDistToBottom     () const { return fStartDistToBottom;}

    /// \brief Perpendicular distance of start point of the leading shower to the 
    /// front edge of the detector
    double StartDistToFront      () const { return fStartDistToFront;}

    /// \brief Perpendicular distance of start point of the leading shower to the 
    /// back edge of the detector
    double StartDistToBack       () const { return fStartDistToBack;}

    /// \brief Perpendicular distance of start point of the leading shower to the 
    /// west edge of the detector (west is positive x)    
    double StartDistToWest       () const { return fStartDistToWest;}

    /// \brief Perpendicular distance of start point of the leading shower to the 
    /// east edge of the detector (east is negative x)
    double StartDistToEast       () const { return fStartDistToEast;}

    /// \brief Perpendicular distance of stop point of the leading shower to the 
    /// top edge of the detector
    double StopDistToTop         () const { return fStopDistToTop;}

    /// \brief Perpendicular distance of stop point of the leading shower to the
    /// bottom edge of the detector
    double StopDistToBottom      () const { return fStopDistToBottom;}

    /// \brief Perpendicular distance of stop point of the leading shower to the 
    /// front edge of the detector
    double StopDistToFront       () const { return fStopDistToFront;}

    /// \brief Perpendicular distance of stop point of the leading shower to the 
    /// back edge of the detector
    double StopDistToBack        () const { return fStopDistToBack;}
    
    /// \brief Perpendicular distance of stop point of the leading shower to the 
    /// west edge of the detector (west is positive x)
    double StopDistToWest        () const { return fStopDistToWest;}

    /// \brief Perpendicular distance of stop point of the leading shower to the 
    /// east edge of the detector (east is negative x)
    double StopDistToEast        () const { return fStopDistToEast;}

    /// \brief Angle between the leading and the next leading shower in degrees
    double AngleToNextShower     () const { return fAngleToNextShower;}

    /// \brief Perpendicular distance of closes approach between the leading shower 
    /// and the event vertex
    double ShowerDistToVtx       () const { return fShowerDistToVtx;}

    /// \brief The maximum X between the start and stop points of the leading 
    /// shower in the slice
    double ShowerMaxX            () const { return fShowerMaxX;}

    /// \brief The maximum Y between the start and stop points of the leading 
    /// shower in the slice
    double ShowerMaxY            () const { return fShowerMaxY;}

    /// \brief The maximum Z between the start and stop points of the leading 
    /// shower in the slice
    double ShowerMaxZ            () const { return fShowerMaxZ;}
    
    /// \name Setters
    //@{
    void SetHitsPerPlane          (double input){  fHitsPerPlane           = input;}
    void SetProngTransMom         (double input){  fProngTransMom          = input;}
    void SetParticleShowerTransMom(double input){  fParticleShowerTransMom = input;}
    void SetPhotonShowerTransMom  (double input){  fPhotonShowerTransMom   = input;}
    void SetStartDistToTop        (double input){  fStartDistToTop         = input;}
    void SetStartDistToBottom     (double input){  fStartDistToBottom      = input;}
    void SetStartDistToFront      (double input){  fStartDistToFront       = input;}
    void SetStartDistToBack       (double input){  fStartDistToBack        = input;}
    void SetStartDistToWest       (double input){  fStartDistToWest        = input;}
    void SetStartDistToEast       (double input){  fStartDistToEast        = input;}
    void SetStopDistToTop         (double input){  fStopDistToTop          = input;}
    void SetStopDistToBottom      (double input){  fStopDistToBottom       = input;}
    void SetStopDistToFront       (double input){  fStopDistToFront        = input;}
    void SetStopDistToBack        (double input){  fStopDistToBack         = input;}
    void SetStopDistToWest        (double input){  fStopDistToWest         = input;}
    void SetStopDistToEast        (double input){  fStopDistToEast         = input;}
    void SetAngleToNextShower     (double input){  fAngleToNextShower      = input;}
    void SetShowerDistToVtx       (double input){  fShowerDistToVtx        = input;}
    void SetShowerMaxX            (double input){  fShowerMaxX             = input;}
    void SetShowerMaxY            (double input){  fShowerMaxY             = input;}
    void SetShowerMaxZ            (double input){  fShowerMaxZ             = input;}
    //@}

  private:
    double fHitsPerPlane;
    double fProngTransMom;
    double fParticleShowerTransMom;
    double fPhotonShowerTransMom;
    double fStartDistToTop;
    double fStartDistToBottom;
    double fStartDistToFront;
    double fStartDistToBack;
    double fStartDistToWest;
    double fStartDistToEast;
    double fStopDistToTop;
    double fStopDistToBottom;
    double fStopDistToFront;
    double fStopDistToBack;
    double fStopDistToWest;
    double fStopDistToEast;
    double fShowerDistToVtx;
    double fAngleToNextShower;
    double fShowerMaxX;
    double fShowerMaxY;
    double fShowerMaxZ;
  };

}
#endif
