////////////////////////////////////////////////////////////////////////
// Class:       UpMuAna
// Module Type: analyzer
// File:        UpMuAna_module.cc
//
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"

#include "RawData/RawTrigger.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/Track.h"
#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/HitList.h"
#include "TrackFit/LocatedChan.h"
#include "RecoBase/CellHit.h"
#include <TF1.h>
#include <TH1F.h>
#include <TCanvas.h>
#include <TApplication.h>
#include <TGraphErrors.h>
#include <TNtuple.h>
#include <TPaveText.h>
#include <TMath.h>
#include <algorithm>
#include <vector>
#include <numeric>
#include <boost/math/special_functions/gamma.hpp>
#include <boost/math/special_functions.hpp>


namespace novaddt {
  class UpMuAna;
}

class novaddt::UpMuAna : public art::EDAnalyzer {
public:
  explicit UpMuAna(fhicl::ParameterSet const & p);
  virtual ~UpMuAna();

  void analyze(art::Event const & e) override;

  void beginJob() override;

private:
  bool isMC;
  bool UseTimeOffsets;
  std::string _hitslabel;
  std::string _instance;
  unsigned int _nSlices           = 0;
  unsigned int _nTracks           = 0;
  const float c0 = -759.5;    // x or y coordinates of cell=0   [cm]
  const float p0 = 4.57;      // z coordinates of plane=0       [cm]
  const float cw = 3.97;      // Cell width                     [cm]
  const float pw = 6.654;     // Plane width                    [cm]
  uint64_t event_time = 0;

  std::string  _trackModuleLabel;   ///< label of module making the Tracks
  std::string  _trackInstanceLabel; ///< instance label making the Tracks
  std::string  _trackToSliceInstanceLabel;
  std::string  _sliceModuleLabel;   ///< label of module making the HitList
  std::string  _sliceInstanceLabel; ///< instance label making the HitList
  std::string  _FLSHitListLabel;
  std::string  _RawDigitLabel;

  void GetXYZD(DAQHit const& daqhit, double w, double *xyzd);

  float GetZ(int plane);
  float GetT(int TDC, float dist_to_readout);
  float GetRes(int ADC);

  float GetX(int cell);
  void GetXYZ(DAQHit const& daqhit, TVector3 start, TVector3 end, float *xyzd);
  void LinFit(const std::vector<double>& x, const std::vector<double>& y, float *fitpar);
  void LinFit(const std::vector<double>& x, const std::vector<double>& y, const std::vector<double>& ye, float *fitpar);

  void LinFitLLR(std::vector<double>& x_hit, std::vector<double>& y_hit, std::vector<double>& y_error, 
		 double& slope, double& chi2, double& P_up, double& P_dn);


  float RootFitLLR(std::vector<double>& x_hit, std::vector<double>& y_hit, std::vector<double>& y_error, float slope);

  double GetPigtail(DAQHit const& daqhit) const;

  void populateFLSmap(std::map<std::pair<unsigned int,
       unsigned int>, const sim::FLSHit*> &pcpflsmap,
       std::map<std::pair<unsigned int, unsigned int>,
           unsigned short> &pcpnummap,
       art::Event const &e);

  std::map<std::string, TNtuple*> ntup_;
  TH1F *dTdist;
  TCanvas *can;

  art::ServiceHandle<art::TFileService> tfs_;
};


novaddt::UpMuAna::UpMuAna(fhicl::ParameterSet const & p) 
  : EDAnalyzer(p)
  , isMC (p.get<bool> ("isMC"))
  , UseTimeOffsets (p.get<bool> ("UseTimeOffsets"))
  , _trackModuleLabel  (p.get<std::string >("TrackModuleLabel"      ))
  , _trackInstanceLabel(p.get<std::string >("TrackInstanceLabel", ""))
  , _trackToSliceInstanceLabel(p.get<std::string >("TrackToSliceInstanceLabel", ""))
  , _sliceModuleLabel  (p.get<std::string >("SliceModuleLabel"      ))
  , _sliceInstanceLabel(p.get<std::string >("SliceInstanceLabel", ""))
{
  if (isMC) {
    _FLSHitListLabel = p.get<std::string>("FLSHitListLabel");
    _RawDigitLabel = p.get<std::string>("RawDigitLabel");
  }
}

novaddt::UpMuAna::~UpMuAna()
{
}

void novaddt::UpMuAna::beginJob()
{
   ntup_["hit_ntuple"] = tfs_->make<TNtuple>
     ("hit_ntuple", "Per hit variables",
      "Run:Event:TrackID:iHit:View:eT:mT:mT_uncor:compMT:truthT:TDC:ADC:Plane:Cell:DCM:X:Y:Z:D:TrackLen");
   
   ntup_["track_ntuple"] = tfs_->make<TNtuple>
     ("track_ntuple", "Slope NTuple",
      "StartX:StartY:StartZ:EndX:EndY:EndZ:dTX:dTY:TrackLen:TrackID:R2x:R2y:nXhits:nYhits:SlopeX:SlopeY:SlopeXY:Chi2X:Chi2Y:Chi2XY:ProbUpX:ProbDnX:ProbUpY:ProbDnY:ProbUpXY:ProbDnXY:ProbUpRoot:ProbDnRoot");

     can = tfs_->make<TCanvas> ("can", "can", 800, 600);

}

void novaddt::UpMuAna::analyze(art::Event const & e)
{

   std::map<std::pair<unsigned int, unsigned int>, 
            const sim::FLSHit*> pcpflsmap;
   std::map<std::pair<unsigned int, unsigned int>,
           unsigned short> pcpnummap;
   if (isMC) populateFLSmap(pcpflsmap, pcpnummap, e);
   
   art::Handle<std::vector<novaddt::HitList> > slices;
   e.getByLabel(_sliceModuleLabel, _sliceInstanceLabel, slices);

   art::Handle<std::vector<novaddt::Track> > tracks;
   e.getByLabel("track", "", tracks);
   
   std::map<unsigned short, std::set<unsigned short>> planesAndCells;
   
   // get the hit list for each track
   art::FindOneP<novaddt::HitList> fohl(tracks, e, _trackModuleLabel);
   // get the tracks for each hit list
   const std::string the_slices = _trackModuleLabel+":"+_trackToSliceInstanceLabel;
   
   art::FindManyP<novaddt::Track3D> find_tracks(slices, e, the_slices);
   assert(find_tracks.isValid());


   

   // Loop through the slices
    for(size_t i_slice = 0; i_slice < slices->size(); ++i_slice){

      novaddt::HitList hit_list = slices->at(i_slice);
      
      // find the tracks for this hit list
      std::vector< art::Ptr<novaddt::Track3D> > tracks = find_tracks.at(i_slice);    
      if(tracks.size() == 0) continue;
      // get the hit list for each track
      art::FindOneP<novaddt::HitList> fohl(tracks, e, _trackModuleLabel);
      assert(fohl.isValid());
      assert(fohl.size() == tracks.size());

      for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
	art::Ptr<novaddt::Track3D> track = tracks[i_track];
	// get the hit list for this track
	art::Ptr<novaddt::HitList> this_hit_list = fohl.at(i_track);
	
	int StartX = track->Start().X();
	int StartY = track->Start().Y();
	int StartZ = track->Start().Z();

	int EndX = track->End().X();
	int EndY = track->End().Y();
	int EndZ = track->End().Z();
	if(StartY > EndY){
	  std::swap(StartX, EndX);
	  std::swap(StartY, EndY);
	  std::swap(StartZ, EndZ);
	}

	TVector3 length;
	length.SetXYZ(cw*(track->End().X() - track->Start().X()),
		      cw*(track->End().Y() - track->Start().Y()),
		      pw*(track->End().Z() - track->Start().Z())
		      );

	float tr_length = std::abs(length.Mag());            
	float exp_trav_time = tr_length/29.97;   // c=29.97 cm/ns; 1/64E6=15.625 ns

	if (tr_length < 200.) continue; // Let's select tracks 20 m long.
	if (!track->Is3D()) continue;

	std::cout.precision(4);
	if(e.id().event() == 0){
	  std::cout << "\t3D-track[" << i_track
		    << "]\t view: " << track->View()
		    << "\thits: "   << this_hit_list->size() 
		    << "\tX: "      << StartX << " - " << EndX
		    << "\tY: "      << StartY << " - " << EndY
		    << "\tZ: "      << StartZ << " - " << EndZ
		    << "\tLen: " << tr_length/100 
		    << "\tExp time:  " << exp_trav_time << "\t3D: " << track->Is3D()
		    << std::endl;
	}


	uint64_t T0 = 0;
	uint64_t T0X = 0;
	uint64_t T0Y = 0;
	uint64_t T1X = 0;
	uint64_t T1Y = 0;

	std::vector<double> x_hit;
	std::vector<double> y_hit;
	std::vector<double> zy_hit;
	std::vector<double> zx_hit;

	// Get T0 that corresponds to the minimum cell number in Y-view
	for(size_t i_hit = 0; i_hit < this_hit_list->size(); ++i_hit){
	  uint16_t iC = this_hit_list->at(i_hit).Cell().val;
	  uint16_t iP = this_hit_list->at(i_hit).Plane().val;
	  uint8_t view = this_hit_list->at(i_hit).View().val;
	  uint64_t iTDC = this_hit_list->at(i_hit).TDC().val;	

	  if(view == daqchannelmap::X_VIEW){
	    if(iC == StartX)
	      T0X = iTDC;	  	
	    if(iC == EndX)
	      T1X = iTDC;	
	    x_hit.push_back(iC);
	    zx_hit.push_back(iP);
	  }

	  if(view == daqchannelmap::Y_VIEW){
	    if(iC == StartY)
	      T0Y = iTDC;
	    if(iC == EndY)
	      T1Y = iTDC;
	    y_hit.push_back(iC);
	    zy_hit.push_back(iP);
	  }
	  
	  T0 = T0Y;    
	}


	// Linear fit of the hits in XZ/YZ views. Drop tracks that has bad linear fit
	float fitpar[3];
	LinFit(x_hit, zx_hit, fitpar);
	float r2x = fitpar[2];
	LinFit(y_hit, zy_hit, fitpar);
	float r2y = fitpar[2];

	

	int dTX = static_cast<int>(T0X - T1X);
	int dTY = static_cast<int>(T0Y - T1Y);
	// Vectors of expected and measured times for TGraph 
	std::vector<Double_t> eT_vecx;
	std::vector<Double_t> mT_vecx;
	std::vector<Double_t> eT_vecy;
	std::vector<Double_t> mT_vecy;
	std::vector<Double_t> eT_vecxy;
	std::vector<Double_t> mT_vecxy;

	// errors
	std::vector<Double_t> mT_vecx_e;
	std::vector<Double_t> mT_vecy_e;
	std::vector<Double_t> mT_vecxy_e;
	std::vector<Double_t> eT_vecx_e;
	std::vector<Double_t> eT_vecy_e;
	std::vector<Double_t> eT_vecxy_e;

	for(size_t i_hit = 0; i_hit < this_hit_list->size(); ++i_hit){
	  uint16_t iC = this_hit_list->at(i_hit).Cell().val;
	  uint16_t iP = this_hit_list->at(i_hit).Plane().val;
	  int meas_time = static_cast<int>(this_hit_list->at(i_hit).TDC().val - T0);


	  int hit_adc = this_hit_list->at(i_hit).ADC().val;
	  if (hit_adc < 50) continue;

	  // Calculate time of the hit
	  DAQHit const& dqhit = this_hit_list->at(i_hit);
	  float pos[4];
	  TVector3 start = track->Start();
	  TVector3 end = track->End();
	  GetXYZ(dqhit, start, end, pos);
	  float pigtail = GetPigtail(dqhit);
	  float tns = GetT(meas_time, pos[3] + pigtail);

          float compMT = 0;
          double truthT = 0;

	  float tns2 = tns;
	  // Subtract DCM offsects
 	  float DCM_off = 0.;
	  if(UseTimeOffsets)
	    DCM_off = (iC/64)*40.;

 	  tns = tns - DCM_off;


	  if( this_hit_list->at(i_hit).View().val == daqchannelmap::X_VIEW){
	    if(iP > TMath::Max(StartZ, EndZ)) continue;
	    if(iP < TMath::Min(StartZ, EndZ)) continue;

	    if(iC > TMath::Max(StartX, EndX)) continue;
	    if(iC < TMath::Min(StartX, EndX)) continue;
	  }
	  else{
	    if(iP > TMath::Max(StartZ, EndZ)) continue;
	    if(iP < TMath::Min(StartZ, EndZ)) continue;

	    if(iC > TMath::Max(StartY, EndY)) continue;
	    if(iC < TMath::Min(StartY, EndY)) continue;
	  }

	  float exp_time = -999;
	  std::string h_view = "y";	   

	  if( this_hit_list->at(i_hit).View().val == daqchannelmap::X_VIEW){
	    h_view = "x";
	    exp_time = exp_trav_time*(iC-StartX)/(EndX-StartX);
	  }
	  else	    	    
	    exp_time = exp_trav_time*(iC-StartY)/(EndY-StartY);

	  const bool debug = false;
	  if(debug){
	    std::cout<< "hit["   << i_hit
		     << "]: \tp: " << this_hit_list->at(i_hit).Plane().val 
		     << ",\tc: "    << this_hit_list->at(i_hit).Cell().val 
		     << ",\tDCM: "    << this_hit_list->at(i_hit).getDCM().val 
		     << ",\t view: "   << h_view
		     << ",\tTDC: "     << this_hit_list->at(i_hit).TDC().val
		     << ",\tT0: "     << T0
		     << ",\texp_t: "     << exp_time
		     << ",\tmt: "    << tns
		     << ",\tmt2: "    << tns2
		     << ",\tXP: " << pos[0]
		     << ",\tYP: " << pos[1]
		     << std::endl;
	  }

	  assert(pos[3]>0);

	  if(this_hit_list->at(i_hit).View().val == daqchannelmap::Y_VIEW){
	    eT_vecy.push_back(exp_time);
	    mT_vecy.push_back(tns);
	    eT_vecy_e.push_back(0);
	    mT_vecy_e.push_back(GetRes(hit_adc));
	  }

	  if(this_hit_list->at(i_hit).View().val == daqchannelmap::X_VIEW){
	    eT_vecx.push_back(exp_time);
	    mT_vecx.push_back(tns);
	    eT_vecx_e.push_back(0);
	    mT_vecx_e.push_back(GetRes(hit_adc));
	  }	  

	  eT_vecxy.push_back(exp_time);
	  mT_vecxy.push_back(tns);
	  eT_vecxy_e.push_back(0);
	  mT_vecxy_e.push_back(GetRes(hit_adc));

	    
	  float ntuple_entries[20] = {
	      (float)e.id().run(), (float)e.id().event(), (float)_nTracks, 
	      (float)i_hit, (float)(this_hit_list->at(i_hit).View().val == daqchannelmap::Y_VIEW), exp_time,
	      tns, tns2, compMT, 
	      (float) truthT, (float)meas_time, (float)this_hit_list->at(i_hit).ADC().val, 
	      (float)this_hit_list->at(i_hit).Plane().val, (float)this_hit_list->at(i_hit).Cell().val, (float)this_hit_list->at(i_hit).getDCM().val, 
	      pos[0],  pos[1],  pos[2],  
	      pos[3], tr_length 
	  };
	  ntup_["hit_ntuple"]->Fill(ntuple_entries);
	  
	} // end of loop on hits


	int nxhit = eT_vecx.size();
	int nyhit = eT_vecy.size();

	// Remove events with less than 20 hits in X+Y views
	if((nxhit+nyhit)>20){	

	// Calculate the likelihood of upward/downward going hypothesis
	float P_up_root   = 0;
	float P_down_root = 0;
	P_up_root   = RootFitLLR(eT_vecxy, mT_vecxy, mT_vecxy_e, 1);
	P_down_root = RootFitLLR(eT_vecxy, mT_vecxy, mT_vecxy_e, -1);
	
	if(P_down_root<1e-30) P_down_root = 1e-30;
	if(P_up_root<1e-30)   P_up_root   = 1e-30;

	double slope_x,  chi_x,  P_up_x,  P_dn_x;
	double slope_y,  chi_y,  P_up_y,  P_dn_y;
	double slope_xy, chi_xy, P_up_xy, P_dn_xy;

	LinFitLLR(eT_vecx,  mT_vecx,  mT_vecx_e,  slope_x,  chi_x,  P_up_x,  P_dn_x);
	LinFitLLR(eT_vecy,  mT_vecy,  mT_vecy_e,  slope_y,  chi_y,  P_up_y,  P_dn_y);
	LinFitLLR(eT_vecxy, mT_vecxy, mT_vecxy_e, slope_xy, chi_xy, P_up_xy, P_dn_xy);


	float tr_ntuple[28] = {(float)StartX,    (float)StartY, (float)StartZ, (float)EndX, (float)EndY, (float)EndZ, (float)dTX, (float)dTY, (float)tr_length, (float)_nTracks,
			       (float)r2x,       (float)r2y, 
			       (float)nxhit,     (float)nyhit,
			       (float)slope_x,   (float)slope_y, (float)slope_xy,
			       (float)chi_x,     (float)chi_y,   (float)chi_xy,
			       (float)P_up_x,    (float)P_dn_x,
			       (float)P_up_y,    (float)P_dn_y,
			       (float)P_up_xy,   (float)P_dn_xy,
			       (float)P_up_root, (float)P_down_root};

	ntup_["track_ntuple"]->Fill(tr_ntuple);

	}

	_nTracks++;

      } // end of loop on tracks

    } // end of loop on slices

}

void novaddt::UpMuAna::GetXYZ(DAQHit const& daqhit, TVector3 start, TVector3 end, float *xyzd) {
  UInt_t plane = daqhit.Plane().val;
  UInt_t cell  = daqhit.Cell().val;

  // The parameters are taken from the fit of plane vs Z
  float x = c0 + cw*cell;
  float y = x;
  float z = p0 + pw*plane;
  float d = -999.9;

  float x0 = c0 + cw*start.X(); float x1 = c0 + cw*end.X();  
  float y0 = c0 + cw*start.Y(); float y1 = c0 + cw*end.Y();    
  float z0 = p0 + pw*start.Z(); float z1 = p0 + pw*end.Z();

  if(daqhit.View().val == daqchannelmap::X_VIEW){
    y = (y1 - y0)/(z1 - z0);
    y = y*(z - z0);
    y = y + y0;
    d = 800 - y;
  }
  if(daqhit.View().val == daqchannelmap::Y_VIEW){
    x = (x1 - x0)/(z1 - z0);
    x = x*(z - z0);
    x = x + x0;
    d = 800 - x;
  }

  if(d<0) d=0;
  xyzd[0] = x;
  xyzd[1] = y;
  xyzd[2] = z;
  xyzd[3] = d;
}

float novaddt::UpMuAna::GetT(int TDC, float dist_to_readout) {
  // This needs to be calibrated and put in a DB table
  float speedOfFiberTransport = 15.3; // cm/ns, "first principles" calib.
  float TDC_to_ns = 15.625; // 1/64E6=15.625 ns
  
  // Differs from c/n due to zigzag
  // paths down fiber.  But, this
  // value may be way off (10%? 30%?).
  return (TDC*TDC_to_ns - dist_to_readout/speedOfFiberTransport);
}


float novaddt::UpMuAna::GetRes(int ADC) {

  float res = 3000.0/ADC;
  if(ADC<50)  res = 60.;
  if(ADC>400) res = 7.5;

  return res*2.0;
}


double novaddt::UpMuAna::GetPigtail(DAQHit const& daqhit) const{
    UInt_t cellid  = daqhit.Cell().val;
    //NOTE: Do we think 32 cells per module can ever change
    //This number can be found in DAQChannelMap but did not want to add that
    //dependancy to geometry unless needed.
    const int kCellsPerModule = 32;
    cellid = cellid % kCellsPerModule;
    // In vertical planes, high cell numbers have longer fibres.
    // In horizontal planes it's the opposite, so flip it round.
    if(daqhit.View().val == daqchannelmap::X_VIEW)
      cellid = kCellsPerModule-cellid-1;

    // This should really never happen, but just to be safe...
    if(cellid < 0 || cellid >= kCellsPerModule) return 100;
    // Email from Tom Chase 2011-04-29
    // NB: this isn't just a ~3.8cm change per cell. Every 8 cells something
    // different happens.
    const double kPigtails[kCellsPerModule] = {
      34.5738, 38.4379, 42.3020, 46.1660, 50.0301, 53.8942, 57.7582, 61.6223,
      64.7504, 68.6144, 72.4785, 76.3426, 80.2067, 84.0707, 87.9348, 91.0790,
      95.3301, 99.1941, 103.058, 106.922, 110.786, 114.650, 118.514, 122.379,
      125.507, 129.371, 133.235, 137.099, 140.963, 144.827, 148.691, 150.751
    };
    return kPigtails[cellid];
  }


void novaddt::UpMuAna::LinFit(const std::vector<double>& x_val, const std::vector<double>& y_val, float *fitpar){
  // http://en.wikipedia.org/wiki/Simple_linear_regression
  const auto n    = x_val.size();
  const auto x  = (std::accumulate(x_val.begin(), x_val.end(), 0.0))/n;                       // <x>
  const auto y  = (std::accumulate(y_val.begin(), y_val.end(), 0.0))/n;                       // <y>
  const auto xx = (std::inner_product(x_val.begin(), x_val.end(), x_val.begin(), 0.0))/n;     // <xx>
  const auto yy = (std::inner_product(y_val.begin(), y_val.end(), y_val.begin(), 0.0))/n;     // <yy>
  const auto xy = (std::inner_product(x_val.begin(), x_val.end(), y_val.begin(), 0.0))/n;     // <xy>

  const auto b    = (xy - x*y)/(xx - x*x);                                                    // slope
  const auto a    = y - b*x;                                                                  // intercept
  const auto r    = (xy - x*y)/sqrt((xx - x*x)*(yy - y*y));                                   // Rxy - coeffcient of determination
  fitpar[0] = a;
  fitpar[1] = b;
  fitpar[2] = r*r;

}


void novaddt::UpMuAna::LinFit(const std::vector<double>& x_val, const std::vector<double>& y_val, const std::vector<double>& y_err, float *fitpar){
  // http://en.wikipedia.org/wiki/Simple_linear_regression
  int n    = x_val.size();
  double x = 0; 
  double y = 0; 
  double xx = 0;
  double yy = 0;
  double xy = 0;
  double ee = 0;

  for ( int i=0; i<n; i++ ){

    x = x + x_val[i]/y_err[i]/y_err[i];
    y = y + y_val[i]/y_err[i]/y_err[i];

    xx = xx + x_val[i]*x_val[i]/y_err[i]/y_err[i];
    yy = yy + y_val[i]*y_val[i]/y_err[i]/y_err[i];
    xy = xy + x_val[i]*y_val[i]/y_err[i]/y_err[i];
    ee = ee + 1./y_err[i]/y_err[i];
  }

  const auto b    = (xy*ee - x*y)/(xx*ee - x*x);                                              // slope
  const auto a    = (xy - b*xx)/x;                                                            // intercept
  const auto r    = (xy - x*y)/sqrt((xx - x*x)*(yy - y*y));                                   // Rxy - coeffcient of determination
  fitpar[0] = a;
  fitpar[1] = b;
  fitpar[2] = r*r;


}



void novaddt::UpMuAna::LinFitLLR(std::vector<double>& x_hit, std::vector<double>& y_hit, std::vector<double>& y_error, 
				 double& slope, double& chi2, double& P_up, double& P_dn) {

  int n = x_hit.size();

  float fitpar[3];
  LinFit(x_hit, y_hit, y_error, fitpar);	
  float a = fitpar[0];
  float b = fitpar[1];


  int totAllowOutlier = (int)(0.1*n);
  int iOutlier = 0;
  std::vector<double> x_filt, y_filt, ye_filt, x_filt_e;
  for (int i = 0; i < n; i++) {
    float y_fit = a + b*x_hit[i];
    if ( fabs(y_hit[i] - y_fit) < 5*y_error[i] || iOutlier>totAllowOutlier) 
      { // > 5 * sigma. Accept 10% of 5sigma outliers. These numbers should be optimized.
	x_filt.push_back(x_hit[i]);
	y_filt.push_back(y_hit[i]);
	ye_filt.push_back(y_error[i]);
	x_filt_e.push_back(0);
      }    
    else
      iOutlier++;

  }

  LinFit(x_filt, y_filt, ye_filt, fitpar);	
  a = fitpar[0];
  b = fitpar[1];
  n = x_filt.size();
  if (n < 5){
    slope = 0;
    chi2  = 999;
    P_dn   = 1e-30;
    P_dn   = 1e-30;
    return;
  }


  slope = fitpar[1];
  
  // chi2
  chi2=0.0;
  for (int i=0; i<n; i++) {
    double y_expected = a + b*x_filt.at(i);
    chi2+=pow((y_filt.at(i)-y_expected) / ye_filt.at(i), 2.0);
  }

  chi2/=(double)(n-2);      // NDF = N points - 2 parameters in the fit


  // Calculate up/down intercepts
  double one_over_ee = 0.0; 
  double x_over_ee   = 0.0;
  double y_over_ee   = 0.0;

  for (int i=0; i<n; i++) {
    double e = ye_filt.at(i);
    one_over_ee+=1.0/e/e;
    x_over_ee+=x_filt.at(i)/e/e;
    y_over_ee+=y_filt.at(i)/e/e;
  }

  // up/down intercepts defined below
  double up_inter   = (y_over_ee-x_over_ee)/one_over_ee;
  double down_inter = (y_over_ee+x_over_ee)/one_over_ee;


  // Calculate up/down chi2  			
  double up_chi2=0.0, down_chi2=0.0;
  for (int i=0;i<n;i++) {

    double e = ye_filt.at(i);
    double up_expected   = up_inter + x_filt.at(i);
    double down_expected = down_inter - x_filt.at(i);
    up_chi2   += pow((y_filt.at(i)-up_expected) / e, 2.0);
    down_chi2 += pow((y_filt.at(i)-down_expected) / e, 2.0);
  }
  
  // if statement is here for safety to prevent gamma_q failing

    double prob_up   = boost::math::gamma_q((double)(n-2)/2.0,up_chi2/2.0);
    double prob_down = boost::math::gamma_q((double)(n-2)/2.0,down_chi2/2.0);

    // Use Root TMath if you prefer. Root and boost libraries give identical results.
    // double prob_up = TMath::Prob(up_chi2, n-2);
    // double prob_down = TMath::Prob(down_chi2, n-2);

    if (prob_up<1e-30) prob_up = 1e-30;
    if (prob_down<1e-30) prob_down = 1e-30;

    P_up = prob_up;
    P_dn = prob_down;


    if(false){ 
    // Debugging part. Will produce the time distribution fits using Root fitter and analytical fuction
    TGraphErrors *eTmT_grx = new TGraphErrors(x_filt.size(), &x_filt[0], &y_filt[0], &x_filt_e[0], &ye_filt[0]);
    TF1 *SlopeFit = new TF1("SlopeFit", "pol1", -10, 150);
    TF1 *UpFit = new TF1("UpFit", "pol1", -10, 150);
    TF1 *UpFit1 = new TF1("UpFit1", "pol1", -10, 150);
    TF1 *DnFit2 = new TF1("DnFit2", "pol1", -10, 150);
    TF1 *DnFit = new TF1("DnFit", "pol1", -10, 150);
    UpFit->SetLineColor(4);
    DnFit->SetLineColor(4);
    DnFit->SetLineStyle(3);
    UpFit->SetParameter(0, up_inter);
    UpFit->SetParameter(1, 1);

    DnFit2->SetParameter(0, fitpar[0]);
    DnFit2->SetParameter(1, fitpar[1]);
    DnFit2->SetLineColor(kGreen);
    DnFit2->SetLineStyle(3);


    DnFit->SetParameter(0, down_inter);
    DnFit->SetParameter(1, -1);
    


    SlopeFit->SetParameter(0, a);
    SlopeFit->SetParameter(1, b);
    SlopeFit->SetLineColor(3);
    SlopeFit->SetLineStyle(3);
    can->Draw();
    can->cd();

    eTmT_grx->GetXaxis()->SetRangeUser(-10, 100);
    eTmT_grx->Draw("AP");
    SlopeFit->Draw("same");
    //    UpFit->Draw("same");
    DnFit2->Draw("same");
    DnFit->Draw("same");

    UpFit1->SetLineColor(kRed);
    UpFit1->SetLineWidth(1);
    //    UpFit1->SetLineStyle(4);

    eTmT_grx->Fit("UpFit1", "BR", "", -10, 150);

    TPaveText *pt = new TPaveText(.75, .9, 0.9, 1.0, "brNDC");    
    pt->SetBorderSize(1);
    pt->AddText(Form("P_up = %.1e",   P_up            ));
    pt->AddText(Form("P_dn = %.1e",   P_dn            ));
    pt->AddText(Form("LLR =  %.1f",   log(P_up/P_dn)  ));  
    pt->AddText(Form("Chi2 = %.2f",   chi2  ));   
    pt->AddText(Form("NDF =  %d",   n-2  ));   
 
    pt->Draw();
    can->Update();
    can->SaveAs(Form("plots/Chi2_%f.pdf", chi2));
  }

}


float novaddt::UpMuAna::RootFitLLR(std::vector<double>& x_hit, std::vector<double>& y_hit, std::vector<double>& y_error, float slope){

  std::vector<double> x_error(x_hit.size(), 0.0);
  TGraphErrors *gr = new TGraphErrors(x_hit.size(), &x_hit[0], &y_hit[0], &x_error[0], &y_error[0]);

  TF1 *SlopeFit = new TF1("SlopeFit", "pol1", -10, 150);
  SlopeFit->FixParameter(1, slope);
  gr->Fit("SlopeFit", "BRQ");

  int tot_out = (int)(0.05*gr->GetN());
  int iout = 0;

  for(int i=0; i<gr->GetN(); i++){
    float x = gr->GetX()[i];
    float y = gr->GetY()[i];
    float ye = gr->GetEY()[i];
    float p0 = SlopeFit->GetParameter(0);
    float p1 = SlopeFit->GetParameter(1);
    float eTime = p0 + p1*x;
    float sigma = TMath::Abs((eTime - y)/ye);
    
    const float sigma_cut = 5;
    if(sigma>sigma_cut && iout<tot_out){
      gr->RemovePoint(i); 
      iout++;
      i--;
    }    
  }

  gr->Fit("SlopeFit", "BRQ");
  float fit_prob = SlopeFit->GetProb();

  SlopeFit->Delete();
  gr->Delete();

  return fit_prob;
}


// Make a mapping from plane:cell pair to a FLSHit for later comparison.
// This method only works well for events with a single track.
void novaddt::UpMuAna::populateFLSmap
(std::map<std::pair<unsigned int, unsigned int>,
                    const sim::FLSHit*> &pcpflsmap,
 std::map<std::pair<unsigned int, unsigned int>,
                     unsigned short> &pcpnummap,
                    art::Event const &e)
{
   // Get the event time to remove offset in later comparison.
   art::Handle< std::vector<rawdata::RawTrigger> > RawTriggers;
   art::Handle<std::vector<sim::FLSHitList> > flshitlist;
   e.getByLabel(_RawDigitLabel, RawTriggers);
   event_time = RawTriggers->at(0).fTriggerTimingMarker_TimeStart;
   e.getByLabel(_FLSHitListLabel, flshitlist);

   if (flshitlist->empty()) std::cout << "FLSHitList is empty!" << std::endl;

   for (unsigned int i=0; i<flshitlist->size(); ++i) {
      art::Ptr<sim::FLSHitList> hlist(flshitlist, i);
      const std::vector<sim::FLSHit>& hits = hlist->fHits;

      for (unsigned int j=0; j<hits.size(); ++j) {
        if (hits[j].GetTrackID() < 0) continue; // this is incoming particle

	unsigned int planeid = hits[j].GetPlaneID();
        unsigned int cellid = hits[j].GetCellID();

        std::pair<unsigned int, unsigned int> pcpair(planeid, cellid);
        pcpflsmap[pcpair]= &hits[j];
        pcpnummap[pcpair]++;
      }
   }
}

DEFINE_ART_MODULE(novaddt::UpMuAna)
