///
/// \file    GeometryDrawer.cxx
/// \brief   Class to aid in the rendering of Geometry objects
/// \author  messier@indiana.edu
/// \version $Id: GeometryDrawer.cxx,v 1.12 2012-12-05 02:24:46 bckhouse Exp $
///
#include "TBox.h"
#include "TLine.h"
#include "TMarker3DBox.h"
#include "TPolyLine.h"
#include "TPolyLine3D.h"
#include "TText.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "ChannelInfo/BadChanList.h"
#include "EventDisplay/GeometryDrawer.h"
#include "EventDisplay/GeoTransform.h"
#include "EventDisplay/GeometryDrawingOptions.h"
#include "EventDisplayBase/View2D.h"
#include "EventDisplayBase/View3D.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/CellGeo.h"
#include "CMap/CMap.h"

using namespace evd;

//......................................................................

GeometryDrawer::GeometryDrawer() { }

//......................................................................

GeometryDrawer::~GeometryDrawer() { }

//......................................................................

void GeometryDrawer::DetOutline2D(const art::Event& /*evt*/,
				  evdb::View2D*        xzview,
				  evdb::View2D*        yzview)
{
  art::ServiceHandle<evd::GeometryDrawingOptions> drawopt;
  if ( (drawopt->fOutline & 
	evd::GeometryDrawingOptions::kOUTLINE_DETECTOR)==0) {
    return;
  }
  
  art::ServiceHandle<geo::Geometry> geo;
  
  double xlo = -geo->DetHalfWidth();
  double xhi =  geo->DetHalfWidth();
  double ylo = -geo->DetHalfHeight();
  double yhi =  geo->DetHalfHeight();
  double zlo =  0.0;
  double zhi =  geo->DetLength();
  
  GeoTransform::XYZ(&xlo, &ylo, &zlo);
  GeoTransform::XYZ(&xhi, &yhi, &zhi);
  
  int c = 16;
  int s = 1;
  int w = 2;
  if (xzview) {
    TPolyLine& l = xzview->AddPolyLine(5, c, w, s);
    l.SetPoint(0, zlo, xlo);
    l.SetPoint(1, zhi, xlo);
    l.SetPoint(2, zhi, xhi);
    l.SetPoint(3, zlo, xhi);
    l.SetPoint(4, zlo, xlo);
    l.SetBit(kCannotPick);
  }
  if (yzview) {
    TPolyLine& l = yzview->AddPolyLine(5, c, w, s);
    l.SetPoint(0, zlo, ylo);
    l.SetPoint(1, zhi, ylo);
    l.SetPoint(2, zhi, yhi);
    l.SetPoint(3, zlo, yhi);
    l.SetPoint(4, zlo, ylo);
    l.SetBit(kCannotPick);
  }
}

//......................................................................

void GeometryDrawer::DetOutline3D(const art::Event& /*evt*/,
				  evdb::View3D*        view)
{
  art::ServiceHandle<geo::Geometry> geo;
  
  double xlo = -geo->DetHalfWidth();
  double xhi =  geo->DetHalfWidth();
  double ylo = -geo->DetHalfHeight();
  double yhi =  geo->DetHalfHeight();
  double zlo =  0.0;
  double zhi =  geo->DetLength();
  
  int c = 19;
  int s = 1;
  int w = 2;
  TPolyLine3D& top = view->AddPolyLine3D(5, c, w, s);
  top.SetPoint(0, xlo, yhi, zlo);
  top.SetPoint(1, xhi, yhi, zlo);
  top.SetPoint(2, xhi, yhi, zhi);
  top.SetPoint(3, xlo, yhi, zhi);
  top.SetPoint(4, xlo, yhi, zlo);
  
  TPolyLine3D& bottom = view->AddPolyLine3D(5, 10, w, s);
  bottom.SetPoint(0, xlo, ylo, zlo);
  bottom.SetPoint(1, xhi, ylo, zlo);
  bottom.SetPoint(2, xhi, ylo, zhi);
  bottom.SetPoint(3, xlo, ylo, zhi);
  bottom.SetPoint(4, xlo, ylo, zlo);
  
  TPolyLine3D& front = view->AddPolyLine3D(3, 10, w, s);
  front.SetPoint(0, xlo, yhi, zlo);
  front.SetPoint(1, xlo, ylo, zlo);
  front.SetPoint(2, xhi, ylo, zlo);
  front.SetPoint(3, xhi, yhi, zlo);
  front.SetPoint(4, xlo, yhi, zlo);
  
  TPolyLine3D& wside = view->AddPolyLine3D(5, c, w, s);
  wside.SetPoint(0, xhi, yhi, zlo);
  wside.SetPoint(1, xhi, ylo, zlo);
  wside.SetPoint(2, xhi, ylo, zhi);
  wside.SetPoint(3, xhi, yhi, zhi);
  wside.SetPoint(4, xhi, yhi, zlo);
  
  TPolyLine3D& eside = view->AddPolyLine3D(5, 10, w, s);
  eside.SetPoint(0, xlo, yhi, zlo);
  eside.SetPoint(1, xlo, ylo, zlo);
  eside.SetPoint(2, xlo, ylo, zhi);
  eside.SetPoint(3, xlo, yhi, zhi);
  eside.SetPoint(4, xlo, yhi, zlo);
  
  c = 15;
  s = 1;
  w = 1;
  double z = zlo;
  // Grid running along x and y at constant z
  for (;;) {
    TPolyLine3D& gridt = view->AddPolyLine3D(2, c, s, w);
    gridt.SetPoint(0, xlo, ylo, z);
    gridt.SetPoint(1, xhi, ylo, z);
    
    TPolyLine3D& grids = view->AddPolyLine3D(2, c, s, w);
    grids.SetPoint(0, xhi, ylo, z);
    grids.SetPoint(1, xhi, yhi, z);
    
    z += 100.0;
    if (z>zhi) break;
  }
  
  // Grid running along z at constant x
  double x = 0.0;
  for (;;) {
    TPolyLine3D& gridt = view->AddPolyLine3D(2, c, s, w);
    gridt.SetPoint(0, x, ylo, zlo);
    gridt.SetPoint(1, x, ylo, zhi);
    x += 100.0;
    if (x>xhi) break;
  }
  x = -100.0;
  for (;;) {
    TPolyLine3D& gridt = view->AddPolyLine3D(2, c, s, w);
    gridt.SetPoint(0, x, ylo, zlo);
    gridt.SetPoint(1, x, ylo, zhi);
    x -= 100.0;
    if (x<xlo) break;
  }
  
  // Grid running along z at constant y
  double y = 0.0;
  for (;;) {
    TPolyLine3D& grids = view->AddPolyLine3D(2, c, s, w);
    grids.SetPoint(0, xhi, y, zlo);
    grids.SetPoint(1, xhi, y, zhi);
    y += 100.0;
    if (y>yhi) break;
  }
  y = -100.0;
  for (;;) {
    TPolyLine3D& grids = view->AddPolyLine3D(2, c, s, w);
    grids.SetPoint(0, xhi, y, zlo);
    grids.SetPoint(1, xhi, y, zhi);
    y -= 100.0;
    if (y<ylo) break;
  }
  
  /*
  // Indicate north and an arrow
  double x0 =  1.5*xhi;          // Center location of the key
  double z0 = (4.0*zlo+zhi)/5.0; // Center location of the key  
  double sz =  0.02*zhi;         // Scale size of the key in z direction
  double sx =  0.75*sz;          // Scale size of the key in z direction
  
  TPolyLine3D& arrow1 = view->AddPolyLine3D(2, c, s, w);
  TPolyLine3D& arrow2 = view->AddPolyLine3D(3, c, s, w);
  arrow1.SetPoint(0, x0, ylo, z0-0.5*sz);
  arrow1.SetPoint(1, x0, ylo, z0+0.5*sz);
  
  arrow2.SetPoint(0, x0-0.5*sx, ylo, z0+0.5*sz-0.5*sx);
  arrow2.SetPoint(1, x0,        ylo, z0+0.5*sz);
  arrow2.SetPoint(2, x0+0.5*sx, ylo, z0+0.5*sz-0.5*sx);
  
  TPolyLine3D& north = view->AddPolyLine3D(4, c, s, w);  
  north.SetPoint(0, x0+0.5*sx, ylo, z0+0.5*sz);
  north.SetPoint(1, x0+0.5*sx, ylo, z0+1.5*sz);
  north.SetPoint(2, x0-0.5*sx, ylo, z0+0.5*sz);
  north.SetPoint(3, x0-0.5*sx, ylo, z0+1.5*sz);
  
  TPolyLine3D& south = view->AddPolyLine3D(6, c, s, w);  
  south.SetPoint(0, x0-0.5*sx, ylo, z0-0.75*sz);
  south.SetPoint(1, x0,        ylo, z0-0.50*sz);
  south.SetPoint(2, x0+0.5*sx, ylo, z0-0.75*sz);
  south.SetPoint(3, x0-0.5*sx, ylo, z0-1.25*sz);
  south.SetPoint(4, x0,        ylo, z0-1.50*sz);
  south.SetPoint(5, x0+0.5*sx, ylo, z0-1.25*sz);
  
  TPolyLine3D& west = view->AddPolyLine3D(5, c, s, w);  
  x0 += sz;
  west.SetPoint(0, x0+0.5*sx, ylo, z0+0.5*sz); 
  west.SetPoint(1, x0+0.5*sx, ylo, z0-0.5*sz);
  west.SetPoint(2, x0,        ylo, z0);
  west.SetPoint(3, x0-0.5*sx, ylo, z0-0.5*sz);
  west.SetPoint(4, x0-0.5*sx, ylo, z0+0.5*sz);
  
  TPolyLine3D& east = view->AddPolyLine3D(7, c, s, w);  
  x0 -= 2.0*sz;
  east.SetPoint(0, x0-0.5*sx, ylo, z0+0.5*sz);
  east.SetPoint(1, x0+0.5*sx, ylo, z0+0.5*sz);	
  east.SetPoint(2, x0+0.5*sx, ylo, z0);
  east.SetPoint(3, x0,        ylo, z0);
  east.SetPoint(4, x0+0.5*sx, ylo, z0);
  east.SetPoint(5, x0+0.5*sx, ylo, z0-0.5*sz);
  east.SetPoint(6, x0-0.5*sx, ylo, z0-0.5*sz);
  */
}


//......................................................................

void GeometryDrawer::FiducialVolumeUser2D(const art::Event& /*evt*/,
					  evdb::View2D* xzview,
					  evdb::View2D* yzview)
{
  art::ServiceHandle<evd::GeometryDrawingOptions> drawopt;
  if ( (drawopt->fOutline & 
	evd::GeometryDrawingOptions::kOUTLINE_FIDUCIAL_USER)==0 ||
       (drawopt->fFiducialBounds).size() !=6 ) {
    return;
  }

  art::ServiceHandle<geo::Geometry> geom;
  std::vector< float > bounds = drawopt->fFiducialBounds;
  TVector3 lo( bounds[0], bounds[1], bounds[2]);
  TVector3 hi( bounds[3], bounds[4], bounds[5]);

  if(xzview){
    TPolyLine& l = xzview->AddPolyLine(5, kBlue-9, 2, 0);
    l.SetPoint(0, lo.Z(), lo.X());
    l.SetPoint(1, hi.Z(), lo.X());
    l.SetPoint(2, hi.Z(), hi.X());
    l.SetPoint(3, lo.Z(), hi.X());
    l.SetPoint(4, lo.Z(), lo.X());
    l.SetBit(kCannotPick);
  }
  if(yzview){
    TPolyLine& l = yzview->AddPolyLine(5, kBlue-9, 2, 0);
    l.SetPoint(0, lo.Z(), lo.Y());
    l.SetPoint(1, hi.Z(), lo.Y());
    l.SetPoint(2, hi.Z(), hi.Y());
    l.SetPoint(3, lo.Z(), hi.Y());
    l.SetPoint(4, lo.Z(), lo.Y());
    l.SetBit(kCannotPick);
  }
}

//......................................................................


void GeometryDrawer::DrawCells2D(const art::Event& evt,
				 evdb::View2D*     xzview,
				 evdb::View2D*     yzview,
				 double xlo, double xhi,
				 double ylo, double yhi,
				 double zlo, double zhi)
{
  art::ServiceHandle<evd::GeometryDrawingOptions> drawopt;
  bool draw_cells =
    (drawopt->fOutline&evd::GeometryDrawingOptions::kOUTLINE_CELLS)!=0;
  bool label_cells =
    (drawopt->fLabel&evd::GeometryDrawingOptions::kLABEL_PLANECELL)!=0;
  if (!draw_cells && !label_cells) return;

  art::ServiceHandle<geo::Geometry>         geo;
  art::ServiceHandle<chaninfo::BadChanList> bcl;
  art::ServiceHandle<cmap::CMap>            cmap;  

  unsigned register int ip, ic;
  const geo::PlaneGeo* plane;
  const geo::CellGeo*  cell;
  for (ip=0; ip<geo->NPlanes(); ++ip) {
    plane = geo->Plane(ip);
    
    geo::View_t view = plane->View();
    if (view==geo::kX && xzview==0) continue;
    if (view==geo::kY && yzview==0) continue;

    evdb::View2D* v2d = 0;
    if (view==geo::kX && xzview) { v2d = xzview; }
    if (view==geo::kY && yzview) { v2d = yzview; }
    if (v2d==0) continue;
    
    for (ic=0; ic<plane->Ncells(); ++ic ) {
      cell = plane->Cell(ic);

      //
      // Cell location and size
      //
      double pos[3];
      cell->GetCenter(pos);
      GeoTransform::XYZ(pos);

      //
      // Skip channels that are out of view range
      //
      if (pos[2]<zlo || pos[2]>zhi)    continue;
      if (view==geo::kX && pos[0]<xlo) continue;
      if (view==geo::kX && pos[0]>xhi) continue;
      if (view==geo::kY && pos[1]<ylo) continue;
      if (view==geo::kY && pos[1]>yhi) continue;

      //
      // Corner locations
      //
      double halfw = cell->HalfW();
      double halfd = cell->HalfD();
      double x0    = pos[2]-halfd;
      double x1    = pos[2]+halfd;
      double y0    = pos[view]-halfw;
      double y1    = pos[view]+halfw;
      
      if (draw_cells) {
	//
	// Change style for bad channels
	//
	static const int kBox=1;
	static const int kXed=2;
	int style = kBox;
	int color = drawopt->fEnabledColor;
	if (drawopt->fDimDisabled && bcl->IsBad(ip, ic) ) {
	  style = kXed;
	  color = drawopt->fDisabledColor;
	}
	if (style==kBox) {
	  TBox& b = v2d->AddBox(x0, y0, x1, y1);
	  b.SetFillStyle(0);
	  b.SetLineColor(color);
	  b.SetLineStyle(1);
	  b.SetLineWidth(1);
	  b.SetBit(kCannotPick);
	}
	if (style==kXed) {
	  TLine& l1 = v2d->AddLine(x0,y0,x1,y1);
	  TLine& l2 = v2d->AddLine(x0,y1,x1,y0);
	  
	  l1.SetLineColor(color);
	  l1.SetLineStyle(1);
	  l1.SetLineWidth(1);
	  l1.SetBit(kCannotPick);
	  
	  l2.SetLineColor(color);
	  l2.SetLineStyle(1);
	  l2.SetLineWidth(1);
	  l2.SetBit(kCannotPick);
	}
      } // if (draw_cells) ...
      
      if (label_cells && (ip%5)==0) {
	this->ShowCellAndPlaneNumbers(v2d,
				      ip,
				      ic,
				      0.5*(x1+x0),
				      0.5*(x1-x0),
				      0.5*(y1+y0),
				      0.5*(y1-y0));
      }
    } // loop on cells
  } // loop on planes
}


//......................................................................


void GeometryDrawer::DrawHighlightCell(const art::Event& /*evt*/,
                                       evdb::View2D*     xzview,
                                       evdb::View2D*     yzview)
{
  art::ServiceHandle<evd::GeometryDrawingOptions> drawopt;
  const int planeId =  drawopt->fHighlightPlane;
  const int cellId =  drawopt->fHighlightCell;

  art::ServiceHandle<geo::Geometry> geom;

  const geo::PlaneGeo* plane;
  const geo::CellGeo* cell;
  try{
    plane = geom->Plane(planeId);
    if(!plane) return;

    cell = plane->Cell(cellId);
    if(!cell) return;
  }
  catch(...){
    return;
  }

  const geo::View_t view = plane->View();
  evdb::View2D* v2d = view == geo::kX ? xzview : yzview;
  if(!v2d) return;

  // Cell location and size
  double pos[3];
  cell->GetCenter(pos);
  GeoTransform::XYZ(pos);

  const double dx = cell->HalfD();
  const double dy = cell->HalfW();
  const double x0 = pos[2]-dx;
  const double x1 = pos[2]+dx;
  const double y0 = pos[view]-dy;
  const double y1 = pos[view]+dy;

  TBox& b = v2d->AddBox(x0, y0, x1, y1);
  b.SetFillStyle(0);
  b.SetLineColor(kRed);
  b.SetLineStyle(1);
  b.SetLineWidth(3);
  b.SetBit(kCannotPick);
}

//......................................................................

void GeometryDrawer::DrawDCMBoxes(const art::Event&    evt,
				  evdb::View2D*        xzview,
				  evdb::View2D*        yzview)
{
  art::ServiceHandle<evd::GeometryDrawingOptions> drawopt;
  if ( (drawopt->fOutline & 
	evd::GeometryDrawingOptions::kOUTLINE_DCMS) == 0) {
    return;
  }
  
  art::ServiceHandle<cmap::CMap>    cmap;  
  art::ServiceHandle<geo::Geometry> geo;
  
  //
  // These maps will hold the geographic extent of each DCM
  //
  static std::map<unsigned int, double> xviewDCMxlo;
  static std::map<unsigned int, double> xviewDCMxhi;
  static std::map<unsigned int, double> xviewDCMzlo;
  static std::map<unsigned int, double> xviewDCMzhi;

  static std::map<unsigned int, double> yviewDCMylo;
  static std::map<unsigned int, double> yviewDCMyhi;
  static std::map<unsigned int, double> yviewDCMzlo;
  static std::map<unsigned int, double> yviewDCMzhi;
  
  static bool ifirst = true;
  
  //
  // First time through, discover the extent of each DCM
  //
  if (ifirst==true) {
    unsigned register int ip, ic;
    for (ip=0; ip<geo->NPlanes(); ++ip) {
      for (ic=0; ic<geo->Plane(ip)->Ncells(); ++ic) {
	geo::View_t v;
	double pos[3], dpos[3];
	geo->CellInfo(ip, ic, &v, pos, dpos);
	
	const daqchannelmap::lchan lc  = cmap->Map()->encodeLChan(geo->DetId(),ip,ic);
	const daqchannelmap::dchan dc  = cmap->Map()->encodeDChan(lc);
	const uint32_t             dib = cmap->Map()->getDiBlock(dc);
	const uint32_t             dcm = cmap->Map()->getDCM(dc);
	
	// Create a unique identifier for this DCM
	unsigned int dcmid = dcm+dib*1000;
	
	double xlo, ylo, zlo;
	double xhi, yhi, zhi;
	if (v==geo::kX) {
	  xlo = pos[0]-dpos[0];
	  zlo = pos[2]-dpos[2];
	  
	  xhi = pos[0]+dpos[0];
	  zhi = pos[2]+dpos[2];
	  
	  //
	  // If we've never seen this DCM before, initialize its values
	  //
	  if (xviewDCMxlo.find(dcmid)==xviewDCMxlo.end()) {
	    xviewDCMxlo[dcmid] =  9E9;
	    xviewDCMzlo[dcmid] =  9E9;
	    xviewDCMxhi[dcmid] = -9E9;
	    xviewDCMzhi[dcmid] = -9E9;
	  }
	  //
	  // Record the minimum and maximum spatial extent of this DCM
	  //
	  if (xlo<xviewDCMxlo[dcmid]) xviewDCMxlo[dcmid] = xlo;
	  if (zlo<xviewDCMzlo[dcmid]) xviewDCMzlo[dcmid] = zlo;
	  if (xhi>xviewDCMxhi[dcmid]) xviewDCMxhi[dcmid] = xhi;
	  if (zhi>xviewDCMzhi[dcmid]) xviewDCMzhi[dcmid] = zhi;
	}
	if (v==geo::kY) {
	  ylo = pos[1]-dpos[1];
	  zlo = pos[2]-dpos[2];
	  
	  yhi = pos[1]+dpos[1];
	  zhi = pos[2]+dpos[2];
	  
	  //
	  // If we've never seen this DCM before, initialize its values
	  //
	  if (yviewDCMylo.find(dcmid)==yviewDCMylo.end()) {
	    yviewDCMylo[dcmid] =  9E9;
	    yviewDCMzlo[dcmid] =  9E9;
	    yviewDCMyhi[dcmid] = -9E9;
	    yviewDCMzhi[dcmid] = -9E9;
	  }
	  //
	  // Record the minimum and maximum spatial extent of this DCM
	  //
	  if (ylo<yviewDCMylo[dcmid]) yviewDCMylo[dcmid] = ylo;
	  if (zlo<yviewDCMzlo[dcmid]) yviewDCMzlo[dcmid] = zlo;
	  if (yhi>yviewDCMyhi[dcmid]) yviewDCMyhi[dcmid] = yhi;
	  if (zhi>yviewDCMzhi[dcmid]) yviewDCMzhi[dcmid] = zhi;
	}
      }
    }
    ifirst = false;
  }
  
  unsigned int dcmid;
  double ll[3]; // Box position, lower left
  double ur[3]; // Box position, upper right
  if (xzview) {
    std::map<unsigned int,double>::iterator ix(xviewDCMxlo.begin());
    std::map<unsigned int,double>::iterator ixe(xviewDCMxlo.end());
    for (; ix!=ixe; ++ix) {
      dcmid = ix->first;
      
      ll[0] = xviewDCMxlo[dcmid];
      ur[0] = xviewDCMxhi[dcmid];
      
      ll[2] = xviewDCMzlo[dcmid];
      ur[2] = xviewDCMzhi[dcmid];

      GeoTransform::XYZ(ll);
      GeoTransform::XYZ(ur);
      
      TBox& b = xzview->AddBox(ll[2],ll[0],ur[2],ur[0]);

      b.SetFillStyle(0);
      b.SetLineColor(kGreen-8);
      b.SetBit(kCannotPick);
    }
  }

  if (yzview) {
    std::map<unsigned int,double>::iterator iy(yviewDCMylo.begin());
    std::map<unsigned int,double>::iterator iye(yviewDCMylo.end());
    for (; iy!=iye; ++iy) {
      dcmid = iy->first;
      
      ll[1] = yviewDCMylo[dcmid];
      ur[1] = yviewDCMyhi[dcmid];
      
      ll[2] = yviewDCMzlo[dcmid];
      ur[2] = yviewDCMzhi[dcmid];
      
      GeoTransform::XYZ(ll);
      GeoTransform::XYZ(ur);
      
      TBox& b = yzview->AddBox(ll[2],ll[1],ur[2],ur[1]);

      b.SetFillStyle(0);
      b.SetLineColor(kGreen-8);
      b.SetBit(kCannotPick);
    }
  }
}

//......................................................................
///
/// Draw text indicating the plane and cell numbers on the display
///
/// \param view - A 2D drawing view
/// \param ip   - Plane number
/// \param ic   - Cell number
/// \param z    - z center of the cell (cm)
/// \param dz   - z width of the cell (cm)
/// \param t    - transverse location (x or y) of the cell (cm)
/// \param dt   - transverse width of the cell (cm)
///
void GeometryDrawer::ShowCellAndPlaneNumbers(evdb::View2D* view,
					     int ip,
					     int ic,
					     double z,
					     double dz,
					     double t,
					     double /*dt*/)
{
  int    f = 42;    // Font style
  int    c = 33;    // Text color
  double s = 0.035; // Text size
  
  if (ic%4==0) {
    char buff[256];
    sprintf(buff,"%d-",ic);
    TText& t1 = view->AddText(z-1.1*dz,t, buff);
    t1.SetTextAlign(32);
    t1.SetTextSize(s);
    t1.SetTextFont(f);
    t1.SetTextColor(c);
    t1.SetBit(kCannotPick);
  }
  if ((ic+5)%4==0) {
    char buff[256];
    sprintf(buff,"  -");
    TText& t1 = view->AddText(z-1.1*dz,t, buff);
    t1.SetTextAlign(32);
    t1.SetTextSize(s);
    t1.SetTextFont(f);
    t1.SetTextColor(c);
    t1.SetBit(kCannotPick);
  }
  if ((ic+5)%4==0) {
    char buff[256];
    sprintf(buff,"..%d..",ip);
    TText& t2 = view->AddText(z+1.6*dz,t, buff);
    t2.SetTextAngle(-90);
    t2.SetTextAlign(22);
    t2.SetTextSize(s);
    t2.SetTextFont(f);
    t2.SetTextColor(c);
  }
}
//......................................................................

void GeometryDrawer::DrawBadBoxes(const art::Event& /*evt*/,
					  evdb::View2D* xzview,
					  evdb::View2D* yzview)
{
  art::ServiceHandle<evd::GeometryDrawingOptions> drawopt;
  if ( (drawopt->fOutline &
	evd::GeometryDrawingOptions::kOUTLINE_BADBOX)==0) {
    return;
  }

  art::ServiceHandle<geo::LiveGeometry> livegeom;

  if(xzview){
    for(int i=0; i < livegeom->NumBadBoxesX(); i++) {
      TPolyLine& l = xzview->AddPolyLine(5, kRed, 2, 0);
      l.SetPoint(0, livegeom->GetBadBoxCorner(false,3,i), livegeom->GetBadBoxCorner(false,1,i));
      l.SetPoint(1, livegeom->GetBadBoxCorner(false,4,i), livegeom->GetBadBoxCorner(false,1,i));
      l.SetPoint(2, livegeom->GetBadBoxCorner(false,4,i), livegeom->GetBadBoxCorner(false,2,i));
      l.SetPoint(3, livegeom->GetBadBoxCorner(false,3,i), livegeom->GetBadBoxCorner(false,2,i));
      l.SetPoint(4, livegeom->GetBadBoxCorner(false,3,i), livegeom->GetBadBoxCorner(false,1,i));
      l.SetBit(kCannotPick);
    }
  }
  if(yzview){
    for(int i=0; i < livegeom->NumBadBoxesY(); i++) {
      TPolyLine& l = yzview->AddPolyLine(5, kRed, 2, 0);
      l.SetPoint(0, livegeom->GetBadBoxCorner(true,3,i), livegeom->GetBadBoxCorner(true,1,i));
      l.SetPoint(1, livegeom->GetBadBoxCorner(true,4,i), livegeom->GetBadBoxCorner(true,1,i));
      l.SetPoint(2, livegeom->GetBadBoxCorner(true,4,i), livegeom->GetBadBoxCorner(true,2,i));
      l.SetPoint(3, livegeom->GetBadBoxCorner(true,3,i), livegeom->GetBadBoxCorner(true,2,i));
      l.SetPoint(4, livegeom->GetBadBoxCorner(true,3,i), livegeom->GetBadBoxCorner(true,1,i));
      l.SetBit(kCannotPick);
    }
  }
}
////////////////////////////////////////////////////////////////////////
