///
/// \file    RawDataDrawer.cxx
/// \brief   Class to aid in the rendering of RawData objects
/// \author  messier@indiana.edu
/// \version $Id: RawDataDrawer.cxx,v 1.30 2012-08-28 22:00:57 denis Exp $
///
#include <cmath>

#include "TBox.h"
#include "TLine.h"
#include "TText.h"
#include "TH1F.h"
#include "TMarker3DBox.h"
#include "TPolyLine3D.h"
#include "TPolyMarker3D.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"

#include "DAQChannelMap/HardwareDisplay.h"
#include "ChannelInfo/BadChanList.h"
#include "CMap/CMap.h"
#include "EventDisplay/ColorScaleTable.h"
#include "EventDisplay/HitTower.h"
#include "EventDisplay/RawDataDrawer.h"
#include "EventDisplay/RawDrawingOptions.h"
#include "EventDisplay/GeometryDrawingOptions.h"
#include "EventDisplay/GeoTransform.h"
#include "EventDisplayBase/View2D.h"
#include "EventDisplayBase/View3D.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RawData/RawDigit.h"
#include "RecoBase/CellHit.h"
#include "RawData/RawTrigger.h"
#include "Calibrator/Calibrator.h"
#include "EventDisplay/SliceNavigator.h"

namespace {
  // Utility function to make uniform error messages.
  void writeErrMsg(const char* fcn,
		   cet::exception const& e)
  {
    std::cout << "RawDataDrawer::" << fcn
	      << " failed with message:\n"
	      << e
	      << std::endl;
  }
}

namespace evd
{
  //
  // Conversion constant from NOvA clock ticks to micro-seconds
  //
  static double kUSEC_PER_TDC = 1.e6/64000000.0;

  //.....................................................................

  RawDataDrawer::RawDataDrawer() { }

  //......................................................................

  RawDataDrawer::~RawDataDrawer() { }

  //......................................................................

  void RawDataDrawer::FillTQHisto(const art::Event& evt,
				  TH1F*             thisto,
				  TH1F*             qhisto) 
  {
    // Check if we're supposed to draw raw hits at all
    art::ServiceHandle<evd::RawDrawingOptions> drawopt;
    if (drawopt->fWhichHits!=evd::RawDrawingOptions::kRAW_HITS) {
      return;
    }
    
    unsigned int imod;
    for (imod=0; imod<drawopt->fRawDigitsModules.size(); ++imod) {
      const std::string& which    = drawopt->fRawDigitsModules[imod];
      const std::string& instance = drawopt->fRawDigitsInstances[imod];
      
      std::vector<const rawdata::RawDigit*> rawhits;
      this->GetRawDigits(evt, which, instance, rawhits);
      
      for (unsigned int i=0; i<rawhits.size(); ++i) {
	double t = 0;
	double q = 0;
	t = rawhits[i]->TDC() * kUSEC_PER_TDC;
	q = rawhits[i]->ADC();
	if (thisto!=0) thisto->Fill(t);
	if (qhisto!=0) {
	  //
	  // Only histogram ADC counts that are in time with the
	  // selected time range
	  //
	  if (t>drawopt->fTimeRange[0] && t<drawopt->fTimeRange[1]) {
	    qhisto->Fill(q);
	  }
	}
      }
    }// end loop over imod
  }

  //......................................................................
  ///
  /// Draw raw data according to hardware address
  ///
  void RawDataDrawer::RawDigitHardware(const art::Event& evt,
				       evdb::View2D*     view)
  {
    art::ServiceHandle<evd::RawDrawingOptions> drawopt;
    if (drawopt->fWhichHits!=evd::RawDrawingOptions::kRAW_HITS) return;

    art::ServiceHandle<cmap::CMap>                  cmap;
    art::ServiceHandle<geo::Geometry>               geo;
    art::ServiceHandle<evd::GeometryDrawingOptions> geoopt;
    art::ServiceHandle<evd::SliceNavigator>         nav;

    daqchannelmap::HardwareDisplay hd;
    hd.SetupDet(geo->DetId());
    
    std::vector<const rb::Cluster*> slicesToNavigate = nav->SlicesToNavigate(evt);
    std::vector<rawdata::RawDigit> unfiltHits;
    int n = slicesToNavigate.size();
    for(int i=0 ; i < n; ++i ){
      // If DimOthers option is chosen, hits from slices
      // other than the current slice should be dimmed.

      if( nav->DimOthers() && !nav->FullEventMode() ){
	if( i != nav->CurrentSliceIndex() ) 
	  continue;
      }
      for( int c = 0, nc = slicesToNavigate[i]->NCell(); c < nc; ++c){
	rb::CellHit chit = *(slicesToNavigate[i]->Cell(c));
	unfiltHits.push_back( (rawdata::RawDigit)chit );
      }
    }

    unsigned int ix; // hardware x coordinate
    unsigned int iy; // hardware y coordinate
    
    for (unsigned int imod=0; imod<drawopt->fRawDigitsModules.size(); ++imod) {
      const char* which = drawopt->fRawDigitsModules[imod].c_str();
      const std::string& instance = drawopt->fRawDigitsInstances[imod];
      
      std::vector< const rawdata::RawDigit*> rawhits;
      this->GetRawDigits(evt, which, instance, rawhits);
      
      for (unsigned int i=0; i<rawhits.size(); ++i) {
	unsigned int ch = rawhits[i]->DaqChannel();
	unsigned int diblock = (ch&0x0FC00000)>>22;
	unsigned int dcm     = (ch&0x003F0000)>>16;
	unsigned int feb     = (ch&0x0000FF00)>>8;
	unsigned int pix     = (ch&0x000000FF)>>0;
	
	hd.DAQToXY(diblock, dcm, feb, pix, &ix, &iy);
	double t = 0;
	double q = 0;
	
	t = rawhits[i]->TDC() * kUSEC_PER_TDC;
	q = rawhits[i]->ADC();
	if (q<=0.0) q=1.;
	
	//
	// Draw markers representing hits
	//
	int co = this->HitColor(rawhits[i], unfiltHits);
	if (drawopt->fRawDrawingOpt & evd::RawDrawingOptions::kSUPPRESS_GHOSTED_HITS) {
	  if (co==evd::kGhostedColor) continue;
	}
	double sf = this->ScaleFactor(t, q);

	TBox& b1 = view->AddBox(ix-0.5*sf,iy-0.5*sf,ix+0.5*sf,iy+0.5*sf);
	b1.SetFillStyle(1001);
	b1.SetFillColor(co);    
	b1.SetBit(kCannotPick);

	if (geoopt->fOutline&evd::GeometryDrawingOptions::kOUTLINE_CELLS) {
	  TBox& b2 = view->AddBox(ix-0.5*sf,iy-0.5*sf,ix+0.5*sf,iy+0.5*sf);
	  b2.SetFillStyle(0);
	  if (co==evd::kGhostedColor) b2.SetLineColor(17); // Greyed out hit
	  else                        b2.SetLineColor(14);
	  b2.SetBit(kCannotPick);
	}
      }// end loop over hits
    }// end loop over imod
    
    // Now draw hardware boundaries and labels
    
    // Vertical lines marking the FEB boundaries. Also mark diblock
    // boundaries periodically
    
    for (unsigned int i=0; i<=hd.fNdiblocks*hd.fNfebCols; ++i) {
      int lw = 1;
      int co = 33;
      if (i%hd.fNfebCols==0) { lw = 1; co = 38; }
      TLine& l = view->AddLine(i*hd.fNpixCols-0.5, -0.5,
			       i*hd.fNpixCols-0.5, hd.fYmax-0.5);
      l.SetLineStyle(1);
      l.SetLineColor(co);
      l.SetLineWidth(lw);
      l.SetBit(kCannotPick);
    }
    for (unsigned int i=0; i<(hd.fNdcmTop+hd.fNdcmSide)*hd.fNfebRows; ++i) {
      int lw = 1;
      int co = 33;
      if (i%hd.fNfebRows==0) { lw = 1; co = 38; }
      TLine& l = view->AddLine(-0.5,        i*hd.fNpixRows-0.5,
			       hd.fXmax-0.5,i*hd.fNpixRows-0.5);
      l.SetLineStyle(1);
      l.SetLineColor(co);
      l.SetLineWidth(lw);
      l.SetBit(kCannotPick);
    }

    for (unsigned int i=1; i<=hd.fNdiblocks; ++i) {
      char buff[128];
      sprintf(buff,"DB%.2d",i);
      iy = -hd.fYmax/200;
      hd.DAQToXY(i,1,0,0,&ix,&iy);
      TText& t = view->AddText(ix,-1,buff);
      t.SetTextSize(0.04);
      t.SetTextAlign(13);
    }

    for (unsigned int i=1; i<=hd.fNdcmTop+hd.fNdcmSide; ++i) {
      char buff[128];
      sprintf(buff,"DCM%.2d",i);
      hd.DAQToXY(1,i,32,0,&ix,&iy);
      ix = -hd.fXmax/200;
      TText& t = view->AddText(-1,iy,buff);
      t.SetTextSize(0.04);
      t.SetTextAlign(31);
    }
  }

  //......................................................................

  void RawDataDrawer::RawDigit2D(const art::Event& evt,
				 evdb::View2D*     xzview,
				 evdb::View2D*     yzview) 
  {
    // Check if we're supposed to draw raw hits at all
    art::ServiceHandle<evd::RawDrawingOptions> drawopt;
    if (drawopt->fWhichHits!=evd::RawDrawingOptions::kRAW_HITS) return;

    art::ServiceHandle<geo::Geometry>               geo;
    art::ServiceHandle<cmap::CMap>                  cmap;
    art::ServiceHandle<evd::GeometryDrawingOptions> geoopt;
    art::ServiceHandle<evd::SliceNavigator> nav;

    std::vector<const rb::Cluster*> slicesToNavigate = nav->SlicesToNavigate(evt);
    std::vector<rawdata::RawDigit> unfiltHits;
    for(int i=0, n = slicesToNavigate.size(); i < n; ++i ){
      // If DimOthers option is chosen, hits from slices
      // other than the current slice should be dimmed.
      if( nav->DimOthers() && !nav->FullEventMode() ){
	if( i != nav->CurrentSliceIndex() ) 
	  continue;
      }
      for( int c = 0, nc = slicesToNavigate[i]->NCell(); c < nc; ++c){
	rb::CellHit chit = *(slicesToNavigate[i]->Cell(c));
	unfiltHits.push_back( (rawdata::RawDigit)chit );
      }
    }

    fXmin =  9.E9;
    fYmin =  9.E9;
    fZmin =  9.E9;
    fXmax = -9.E9;
    fYmax = -9.E9;
    fZmax = -9.E9;

    for (unsigned int imod=0; imod<drawopt->fRawDigitsModules.size(); ++imod) {
      const char* which = drawopt->fRawDigitsModules[imod].c_str();
      const std::string& instance = drawopt->fRawDigitsInstances[imod];

      std::vector< const rawdata::RawDigit*> rawhits;
      this->GetRawDigits(evt, which, instance, rawhits);
      
      for (unsigned int i=0; i<rawhits.size(); ++i) {
	unsigned int plane = cmap->GetPlane(rawhits[i]);
	unsigned int cell  = cmap->GetCell(rawhits[i]);
	
	if (plane>=geo->NPlanes()) {
	  std::cerr << "*- Illegal plane number " << plane
		    << " DaqChannel = " 
		    << std::hex << rawhits[i]->DaqChannel() 
		    << std::endl;
	  std::cerr << std::dec;
	  continue;
	}
	const geo::PlaneGeo *p = geo->Plane(plane);
	
	if (cell>=p->Ncells()) {
	  std::cerr << "*- Illegal cell number " << cell 
		    << " in plane " << plane
		    << "*- DaqChannel = " 
		    << std::hex << rawhits[i]->DaqChannel()
		    << std::endl;
	  std::cerr << std::dec;
	  continue;
	}
	const geo::CellGeo *c = p->Cell(cell);
	
	double xyz[3];
	double tpos = 0.0, zpos = 0.0;
	double dt, dz;
	dz = c->HalfD();
	dt = c->HalfW();
	evdb::View2D* v = 0;
	if (p->View()==geo::kX) {
	  c->GetCenter(xyz);
	  GeoTransform::XYZ(xyz);
	  
	  tpos = xyz[0];
	  zpos = xyz[2];
      
	  if (tpos<fXmin) fXmin = tpos;
	  if (tpos>fXmax) fXmax = tpos;
	  if (zpos<fZmin) fZmin = zpos;
	  if (zpos>fZmax) fZmax = zpos;
	}
	if (p->View()==geo::kY) {
	  c->GetCenter(xyz);
	  GeoTransform::XYZ(xyz);

	  tpos = xyz[1];
	  zpos = xyz[2];
	  
	  if (tpos<fYmin) fYmin = tpos;
	  if (tpos>fYmax) fYmax = tpos;
	  if (zpos<fZmin) fZmin = zpos;
	  if (zpos>fZmax) fZmax = zpos;
	}
	// Decide if this hit should be drawn on this view
	if (p->View()==geo::kX && xzview==0) continue;
	if (p->View()==geo::kY && yzview==0) continue;
	if (xzview!=0) v = xzview;
	if (yzview!=0) v = yzview;
    
	double t = 0;
	double q = 0;
	t = rawhits[i]->TDC()*kUSEC_PER_TDC;
	q = rawhits[i]->ADC();
	
	// Hack for now...
	if (q<=0.0) q=100.;
	
	//
	// Draw boxes representing hit cells
	//
	int co = this->HitColor( rawhits[i], unfiltHits);
	if (drawopt->fRawDrawingOpt & evd::RawDrawingOptions::kSUPPRESS_GHOSTED_HITS) {
	  if (co==evd::kGhostedColor) continue;
	}

	double sf = this->ScaleFactor(t, q);
	TBox& b1 = v->AddBox(zpos-sf*dz,tpos-sf*dt,zpos+sf*dz,tpos+sf*dt);
	b1.SetFillStyle(1001);
	b1.SetFillColor(co);    
	b1.SetBit(kCannotPick);

	if (geoopt->fOutline&evd::GeometryDrawingOptions::kOUTLINE_CELLS) {
	  TBox& b2 = v->AddBox(zpos-dz,tpos-dt,zpos+dz,tpos+dt);
	  b2.SetFillStyle(0);
	  if (co==evd::kGhostedColor) b2.SetLineColor(17); // Greyed out hit
	  else                        b2.SetLineColor(14);
	  b2.SetBit(kCannotPick);
	}
      }//end loop over raw hits
    }// end loop over imod
  }

  //......................................................................

  void RawDataDrawer::RawDigit3D(const art::Event& evt,
				 evdb::View3D*     view)
  {
    // Check if we're supposed to draw raw hits at all
    art::ServiceHandle<evd::RawDrawingOptions> drawopt;
    if (drawopt->fWhichHits!=evd::RawDrawingOptions::kRAW_HITS) return;
    
    art::ServiceHandle<cmap::CMap>    cmap;
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<evd::SliceNavigator> nav;

    std::vector<const rb::Cluster*> slicesToNavigate = nav->SlicesToNavigate(evt);
    std::vector<rawdata::RawDigit> unfiltHits;

    for(int i=0, n = slicesToNavigate.size(); i < n; ++i ){
      // If DimOthers option is chosen, hits from slices
      // other than the current slice should be dimmed.
      if( nav->DimOthers() && !nav->FullEventMode()){
	if( i != nav->CurrentSliceIndex() ) 
	  continue;
      }
      for( int c = 0, nc = slicesToNavigate[i]->NCell(); c < nc; ++c){
	rb::CellHit chit = *(slicesToNavigate[i]->Cell(c));
	unfiltHits.push_back( (rawdata::RawDigit)chit );
      }
    }// end loop over slices to navigate

    HitTower tower;
    tower.fQscale = 100.0/2024;
    for (unsigned int imod=0; imod<drawopt->fRawDigitsModules.size(); ++imod) {
      const char* which = drawopt->fRawDigitsModules[imod].c_str();
      const std::string& instance = drawopt->fRawDigitsInstances[imod];
      
      std::vector< const rawdata::RawDigit*> rawhits;
      this->GetRawDigits(evt, which, instance, rawhits);

      double xmin =  99e99;
      double xmax = -99e99;      
      double ymin =  99e99;
      double ymax = -99e99;
      double zmin =  99e99;
      double zmax = -99e99;
      for (unsigned int i=0; i<rawhits.size(); ++i) {

	int co = this->HitColor( rawhits[i], unfiltHits);
	if (co == evd::kGhostedColor) continue;
	
	int iplane = cmap->GetPlane(rawhits[i]);
	int icell  = cmap->GetCell(rawhits[i]);
	
	double      xyz[3];
	double      dpos[3];
	geo::View_t v;
	geom->CellInfo(iplane, icell, &v, xyz, dpos);
	if (v==geo::kX && xyz[0]<xmin) xmin = xyz[0];
	if (v==geo::kX && xyz[0]>xmax) xmax = xyz[0];
	if (v==geo::kY && xyz[1]<ymin) ymin = xyz[1];
	if (v==geo::kY && xyz[1]>ymax) ymax = xyz[1];
	if (xyz[2]<zmin) zmin = xyz[2];
	if (xyz[2]>zmax) zmax = xyz[2];
      }
      xmin -= 75.0;
      ymin -= 75.0;
      zmin -= 75.0;
      xmax += 75.0;
      ymax += 75.0;
      zmax += 75.0;
      if (xmax> geom->DetHalfWidth())  xmax =  geom->DetHalfWidth();
      if (ymax> geom->DetHalfHeight()) ymax =  geom->DetHalfHeight();
      if (xmin<-geom->DetHalfWidth())  xmin = -geom->DetHalfWidth();
      if (ymin<-geom->DetHalfHeight()) ymin = -geom->DetHalfHeight();
      if (zmin<0)                      zmin = 0;
      if (zmax>geom->DetLength())      zmax = geom->DetLength();
      
      //
      // Draw intersections using hits in adjacent planes
      //
      int nmarker = 0;
      for (unsigned int i=0; i<rawhits.size(); ++i) {
	double ti = 0;
	double qi = 0;
	ti = rawhits[i]->TDC()*kUSEC_PER_TDC;
	qi = rawhits[i]->ADC();
	
	int iplanei = cmap->GetPlane(rawhits[i]);
	int icelli  = cmap->GetCell(rawhits[i]);

	double      xyzi[3];
	double      dposi[3];
	geo::View_t vi;
	geom->CellInfo(iplanei, icelli, &vi, xyzi, dposi);
	
	for (unsigned int j=i+1; j<rawhits.size(); ++j) {
	  double tj = 0;
	  double qj = 0;
	  tj = rawhits[j]->TDC()*kUSEC_PER_TDC;
	  qj = rawhits[j]->ADC();
	  
	  int iplanej = cmap->GetPlane(rawhits[j]);
	  int icellj  = cmap->GetCell(rawhits[j]);

	  double      xyzj[3];
	  double      dposj[3];
	  geo::View_t vj;
	  geom->CellInfo(iplanej, icellj, &vj, xyzj, dposj);
	  
	  //
	  // Cells must be in time with each other and in adjacent
	  // orthogonal views
	  //
	  if (fabs(ti-tj)>1.5)      continue;
	  if ((iplanej-iplanei)!=1) continue;
	  if (vi==vj)               continue;
	  
	  double qave = 2.0*(qi*qj)/(qi+qj);
	  if (qave<200.0) continue;

	  double x, y, z;
	  if (vi==geo::kX) { x = xyzi[0]; }
	  else             { x = xyzj[0]; }
	  if (vi==geo::kY) { y = xyzi[1]; }
	  else             { y = xyzj[1]; }
	  z  = 0.5*(xyzi[2]+xyzj[2]);
	  
	  int    co = this->HitColor( rawhits[i], unfiltHits);
	  int    st = kFullDotMedium;
	  double sz = 10.0*sqrt(qave/200.0);
	  TPolyMarker3D& pm = view->AddPolyMarker3D(1, co, st, sz);
	  pm.SetPoint(0,x,y,z);
	  ++nmarker;
	  // TMarker3DBox& mb = view->AddMarker3DBox(x,y,z,dx,dy,dz);
	  // mb.SetLineColor(co);
	  // mb.SetFillStyle(1);
	  // mb.SetFillColor(co);
	} // Loop on j rawhits
      } // Loop on i rawhits
      
      for (unsigned int i=0; i<rawhits.size(); ++i) {
	double t = 0;
	double q = 0;
	t = rawhits[i]->TDC()*kUSEC_PER_TDC;
	q = rawhits[i]->ADC();
	
	// Hack for now...
	if (q<=0.0) q = 100.;
	
	// Get the cell geometry for the hit
	int         iplane = cmap->GetPlane(rawhits[i]);
	int         icell  = cmap->GetCell(rawhits[i]);
	double      xyz[3];
	double      dpos[3];
	geo::View_t v;
	geom->CellInfo(iplane, icell, &v, xyz, dpos);
	
	int co = this->HitColor( rawhits[i], unfiltHits);
	if (co == evd::kGhostedColor) continue;

	if (1) {
	  //
	  // Render digits as towers
	  //
	  if (v==geo::kX) {
	    /*
	      double h = q*tower.fQscale;
	      TMarker3DBox& m = 
	      view->AddMarker3DBox(xyz[0],
	      geom.DetHalfHeight()+0.5*h,
	      xyz[2],
	      dpos[0],
	      0.5*h,
	      0.0);
	      m.SetLineColor(kRed);
	      m.SetFillColor(kRed);
	    */
	    tower.AddHit(v, iplane, icell, xyz[0], xyz[2], q, t);
	  }
	  else if (v==geo::kY) {
	    /*
	      double h = q*tower.fQscale;
	      TMarker3DBox& m = 
	      view->AddMarker3DBox(geom.DetHalfWidth()+0.5*h,
	      xyz[1],
	      xyz[2],
	      0.5*h,
	      dpos[1],
	      0.0);
	      m.SetLineColor(kRed);
	      m.SetFillColor(kRed);
	    */
	    tower.AddHit(v, iplane, icell, xyz[1], xyz[2], q, t);
	  }
	  else abort();
	}

	if (1) {
	  //
	  // Render Digits as boxes 
	  //
	  TPolyLine3D& p = view->AddPolyLine3D(5,co,1,1);
	  double sf = 1.0;
	  if (v==geo::kX) {
	    double x1 = xyz[0] - sf*dpos[0];
	    double x2 = xyz[0] + sf*dpos[0];
	    double z1 = xyz[2] - sf*dpos[2];
	    double z2 = xyz[2] + sf*dpos[2];
	    p.SetPoint(0, x1, ymin, z1);
	    p.SetPoint(1, x2, ymin, z1);
	    p.SetPoint(2, x2, ymin, z2);
	    p.SetPoint(3, x1, ymin, z2);
	    p.SetPoint(4, x1, ymin, z1);
	  }
	  else if (v==geo::kY) {
	    double y1 = xyz[1] - sf*dpos[1];
	    double y2 = xyz[1] + sf*dpos[1];
	    double z1 = xyz[2] - sf*dpos[2];
	    double z2 = xyz[2] + sf*dpos[2];
	    p.SetPoint(0, xmax, y1, z1);
	    p.SetPoint(1, xmax, y2, z1);
	    p.SetPoint(2, xmax, y2, z2);
	    p.SetPoint(3, xmax, y1, z2);
	    p.SetPoint(4, xmax, y1, z1);
	  }
	  else abort();
	}
      }//end loop over raw digits
      
      // Render the towers for that style choice
      if (drawopt->fHit3DStyle&evd::RawDrawingOptions::kHIT3D_TOWERS) {
	tower.Draw(view);
      }

      // Draw the zoom boxes
      TPolyLine3D& xz = view->AddPolyLine3D(5,18,1,1);
      xz.SetPoint(0,xmin+10,ymin,zmin);
      xz.SetPoint(1,xmin+10,ymin,zmax);
      xz.SetPoint(2,xmax-10,ymin,zmax);
      xz.SetPoint(3,xmax-10,ymin,zmin);
      xz.SetPoint(4,xmin+10,ymin,zmin);

      TPolyLine3D& yz = view->AddPolyLine3D(5,18,1,1);
      yz.SetPoint(0,xmax,ymin+10,zmin);
      yz.SetPoint(1,xmax,ymin+10,zmax);
      yz.SetPoint(2,xmax,ymax-10,zmax);
      yz.SetPoint(3,xmax,ymax-10,zmin);
      yz.SetPoint(4,xmax,ymin+10,zmin);
      
    }// end loop over imod
    
    //
    // A stick figure for scale
    //
    this->StickFigure(view,
		      -geom->DetHalfWidth(),
		      -geom->DetHalfHeight(),
		      0.0);
  }

  //......................................................................    

  void RawDataDrawer::GetLimits(double* xmin, double* xmax,
				double* ymin, double* ymax,
				double* zmin, double* zmax) 
  {
    *xmin = fXmin;
    *xmax = fXmax;
    *ymin = fYmin;
    *ymax = fYmax;
    *zmin = fZmin;
    *zmax = fZmax;
  }

  //......................................................................    

  int RawDataDrawer::GetRawDigits(const art::Event&                  evt,
				  std::string const&                 which,
                                  std::string const&                 instance,
				  std::vector< const rawdata::RawDigit*>& rawhits)
  {
    rawhits.clear();

    /// Temporary vector of rawdigits
    std::vector<const rawdata::RawDigit*> temp(rawhits);

    try{
        art::Handle< std::vector<rawdata::RawDigit> > rdcol;
        evt.getByLabel(which, instance, rdcol);
        /// Extract rawdigits from Handle
        for(unsigned int j = 0; j < rdcol->size(); ++j)
                temp.push_back( &(*rdcol)[j] );
    }// end try
    catch(cet::exception& e) {
            std::cout<<__PRETTY_FUNCTION__<<": Did not find RawDigits with\""<<which<<"\" label trying all labels\n";

            /// Try to find RawDigits in other folders
            std::vector<art::Handle< std::vector<rawdata::RawDigit> > > rdcol_vec;
            evt.getManyByType(rdcol_vec);

            /// Extract rawdigits from vector of Handle
            for(unsigned int i = 0; i < rdcol_vec.size()    ; ++i)
            for(unsigned int j = 0; j < rdcol_vec[i]->size(); ++j)
                temp.push_back( &(*rdcol_vec[i])[j] );
            
            /// If still couldn't find, throw exception
            if(rdcol_vec.size() == 0)
                writeErrMsg("GetRawDigits", e);
    }// end of catch
                                               
      
    /// Fill in the temporary vector of rawdigits to output reference rawhits
    temp.swap(rawhits);

    // check for bad channels
    art::ServiceHandle<evd::RawDrawingOptions> drawopt;
    if (drawopt->fRawDrawingOpt & evd::RawDrawingOptions::kMASK_BAD_CHANNELS) {
      art::ServiceHandle<chaninfo::BadChanList> bcl;
      bcl->Apply(rawhits);
    }

    return rawhits.size();
  }

  //......................................................................    

  int RawDataDrawer::GetRawTrigger(const art::Event&                    e,
				   art::PtrVector<rawdata::RawTrigger>& r)
  {
    r.clear();
    art::PtrVector<rawdata::RawTrigger> temp;

    // using getManyByType so that we don't have to specify a module 
    // label
    std::vector< art::Handle< std::vector<rawdata::RawTrigger> > > rtcol;
  
    try{
      e.getManyByType(rtcol);
      for(size_t rtc = 0; rtc < rtcol.size(); ++rtc){
	for(unsigned int i=0; i < rtcol[rtc]->size(); ++i){
	  art::Ptr<rawdata::RawTrigger> rt(rtcol[rtc], i);
	  temp.push_back(rt);
	}
      }
      temp.swap(r);
    }// end try
    catch(cet::exception& e) {
      writeErrMsg("GetRawTrigger", e);
    }

    return r.size();
  }

  //......................................................................    

  ///
  /// Determine the color to use for the hit. If the hit is flagged
  /// to be ghosted out by either its time or charge, ghost the hit
  ///
  int RawDataDrawer::HitColor( const rawdata::RawDigit* hit,
			       const std::vector<rawdata::RawDigit>&  unfiltered)
			       
  {
    art::ServiceHandle<evd::RawDrawingOptions>  drawopt;
    art::ServiceHandle<evd::SliceNavigator>     sliceopt;

    int co = 0;
    double t = hit->TDC() * kUSEC_PER_TDC;
    double q = hit->ADC();

    int t_color = ColorScaleTable::RawT().GetColor(t);
    int q_color = ColorScaleTable::RawQ().GetColor(q);
    if(t_color==kGhostedColor || q_color==kGhostedColor || 
       DimFiltered(hit, unfiltered ) ) {
      co = kGhostedColor;
    }
    else {
      switch (drawopt->fColor) {
      case evd::RawDrawingOptions::kCOLOR_BY_CHARGE:
	co = q_color; 
	break;
      case evd::RawDrawingOptions::kCOLOR_BY_TIME:
	co = t_color; 
	break;
      default:
	co = 1;
      }
    }
    return co;
  }

  //......................................................................


  bool RawDataDrawer::DimFiltered(const rawdata::RawDigit* hit,
				  const std::vector<rawdata::RawDigit>&  unfiltered)
  {
    art::ServiceHandle<SliceNavigator> nav;
    if(!nav->DimHits() && !nav->DimOthers() ) return false;

    int nrd = unfiltered.size();
    bool notfound = true;
    
    for( int ird = 0; ird < nrd; ++ird){
      if( unfiltered[ird].ADC() == hit->ADC() &&
	  unfiltered[ird].TDC() == hit->TDC() ){
	notfound = false;
	break;
      }
    }
    return notfound;
  }

  //......................................................................                 

  ///
  /// Determine the scale factor to use when rendering hit cells
  ///
  double RawDataDrawer::ScaleFactor(double /*t*/, double q)
  {
    art::ServiceHandle<evd::RawDrawingOptions> drawopt;
    
    double sf = drawopt->fScaleFactor;
    double q0 = 4096.0;
    if (drawopt->fRawDrawingOpt & evd::RawDrawingOptions::kSCALE_HITS_BY_CHARGE) {
      sf *= 3.0*sqrt(q/q0);
    }
    return sf;
  }
  
  ///
  /// Draw a stick figure to scale
  ///
  void RawDataDrawer::StickFigure(evdb::View3D* view,
				  double x,
				  double y,
				  double z)
  {
    double h = 178.0; // Average height of a man
    double xlo= 9E9;
    double xhi=-9E9;
    double ylo= 9E9;
    double yhi=-9E9;
    double leg[8][2] = {
      {406,-618}, // 0
      {392,-576}, // 1
      {481,-509}, // 2
      {511,-403}, // 3
      {444,-385}, // 4
      {451,-495}, // 5
      {451,-618}, // 6
      {429,-617}  // 7
    };
    double arms[8][2] = {
      {537,-414}, // 0
      {535,-394}, // 1
      {552,-309}, // 2
      {550,-228}, // 3
      {462,-239}, // 4
      {401,-290}, // 5
      {453,-345}, // 6
      {447,-363}  // 7
    };
    double body[9][2] = {
      {479,-394}, // 0
      {504,-221}, // 1
      {493,-220}, // 2
      {478,-212}, // 3
      {473,-181}, // 4
      {491,-165}, // 5
      {521,-174}, // 6
      {522,-211}, // 7
      {504,-221}  // 8
    };
    int i;
    for (i=0; i<8; ++i) {
      if (leg[i][0]<xlo) xlo = leg[i][0];
      if (leg[i][0]>xhi) xhi = leg[i][0];
      if (leg[i][1]<ylo) ylo = leg[i][1];
      if (leg[i][1]>yhi) yhi = leg[i][1];
    }
    for (i=0; i<8; ++i) {
      if (arms[i][0]<xlo) xlo = arms[i][0];
      if (arms[i][0]>xhi) xhi = arms[i][0];
      if (arms[i][1]<ylo) ylo = arms[i][1];
      if (arms[i][1]>yhi) yhi = arms[i][1];
    }
    for (i=0; i<8; ++i) {
      if (body[i][0]<xlo) xlo = body[i][0];
      if (body[i][0]>xhi) xhi = body[i][0];
      if (body[i][1]<ylo) ylo = body[i][1];
      if (body[i][1]>yhi) yhi = body[i][1];
    }
    
    double sf = h/fabs(yhi-ylo);
    
    double x0 = x-sf*fabs(xhi-xlo)-2.0;
    double y0 = y+2.0;
    for (i=0; i<8; ++i) {
      leg[i][0] = x0+sf*(leg[i][0]-xlo);
      leg[i][1] = y0+sf*(leg[i][1]-ylo);
    }
    for (i=0; i<8; ++i) {
      arms[i][0] = x0+sf*(arms[i][0]-xlo);
      arms[i][1] = y0+sf*(arms[i][1]-ylo);
    }
    for (i=0; i<9; ++i) {
      body[i][0] = x0+sf*(body[i][0]-xlo);
      body[i][1] = y0+sf*(body[i][1]-ylo);
    }
    
    TPolyLine3D& l = view->AddPolyLine3D(8,kCyan,1,2);
    TPolyLine3D& a = view->AddPolyLine3D(8,kCyan,1,2);
    TPolyLine3D& b = view->AddPolyLine3D(9,kCyan,1,2);
    for (i=0; i<8; ++i) {
      l.SetPoint(i,leg[i][0],leg[i][1],z);
    }
    for (i=0; i<8; ++i) {
      a.SetPoint(i,arms[i][0],arms[i][1],z);
    }
    for (i=0; i<9; ++i) {
      b.SetPoint(i,body[i][0],body[i][1],z);
    }
  }

} // end namespace evd
////////////////////////////////////////////////////////////////////////
