////////////////////////////////////////////////////////////////////////
// \file    CAFFilter.cxx
// \brief   Filter events based on simple expressions and CAF variables
// \author  Eric Flumerfelt - eflumerf@utk.edu
////////////////////////////////////////////////////////////////////////


// Framework includes

#include "TFormula.h"

// C/C++ includes
#include <cmath>
#include <iostream>
#include <unordered_map>
#include <vector>
#include <string>
#include <sstream>
#include <iomanip>
#include <algorithm>
#include <time.h>
#include <fenv.h>

// Framework includes
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Core/FindOne.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

// NOvASoft includes
// Geo/Cal/Util
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/Geo.h"
#include "Calibrator/Calibrator.h"
#include "Utilities/func/MathUtil.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RawData/RawDigit.h"
#include "Simulation/FLSHitList.h"

// Sim/BackTracker
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/GTruth.h"
#include "MCCheater/BackTracker.h"
// RecoBase
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Shower.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/Energy.h"

// Other Reco
#include "RecoJMShower/JMShower.h"
#include "RecoJMShower/EID.h"
#include "LEM/PIDDetails.h"
#include "ReMId/ReMId.h"
#include "QEEventFinder/QePId.h"
#include "LIDBuilder/EventLID.h"
#include "LIDBuilder/ShowerLID.h"
#include "RecVarPID/RVP.h"
#include "NumuEnergy/NumuE.h"
#include "NueSandbox/NueSandObj.h"
#include "NumuSandbox/NumuSandObj.h"
#include "MichelEFilter/MichelECluster.h"
#include "CosRej/CosRejObj.h"
#include "MuonRemove/art/MRCCParent.h"

// SummData
#include "SummaryData/CosmicExposure.h"
#include "SummaryData/POTSum.h"
#include "SummaryData/SpillData.h"

namespace filter
{
    struct SliceInfo
    {
        double nTotHit;
        double nCalHit;
        double nMipHit;
        double maxContig;

        unsigned int maxPlane;
        unsigned int minPlane;
        unsigned int maxCell;
        unsigned int minCell;
        unsigned int minCellwall;
        unsigned int minCellsFromEdge;

    };

    typedef std::unordered_map<std::string, double> valMap;

    const valMap CAFFilterMap({
        { "rec.energy.elecid.E", NAN }, ///< Energy [GeV]
        { "rec.energy.elecid.depE", NAN }, ///< Total energy deposited [GeV]
        { "rec.energy.elecid.shwE", NAN }, ///< Energy of primary shower [GeV]
        { "rec.energy.numusimp.E", NAN }, ///< Energy [GeV]
        { "rec.energy.numusimp.angleE", NAN }, ///< Quasielastic angle formula neutrino energy [GeV]
        { "rec.energy.numusimp.angleerror", NAN }, ///< 1-sigma error from using quasielastic angle formula for neutrino energy
        { "rec.energy.numusimp.calccE", NAN }, ///< Calorometric charged current neutrino energy [GeV]
        { "rec.energy.numusimp.fsE", NAN }, ///< Fernanda's scattering energy
        { "rec.energy.numusimp.hadcalE", NAN }, ///< Hadronic calorometric energy NOT on the muon track[GeV]
        { "rec.energy.numusimp.hadtrkE", NAN }, ///< Hadronic calorometric energy on the muon track[GeV]
        { "rec.energy.numusimp.ndhadcalactE", NAN }, ///< Near detector -- hadronic calorometric energy in active region [GeV]
        { "rec.energy.numusimp.ndhadcalcatE", NAN }, ///< Near detector -- hadronic calorometric energy in muon catcher [GeV]
        { "rec.energy.numusimp.ndhadcaltranE", NAN }, ///< Near detector -- hadronic calorometric energy in transition plane [GeV]
        { "rec.energy.numusimp.ndtrkcalactE", NAN }, ///< Near detector -- muon calorometric energy in active region [GeV]
        { "rec.energy.numusimp.ndtrkcalcatE", NAN }, ///< Near detector -- muon calorometric energy in muon catcher [GeV]
        { "rec.energy.numusimp.ndtrkcaltranE", NAN }, ///< Near detector -- muon calorometric energy in transition plane [GeV]
        { "rec.energy.numusimp.ndtrklenact", NAN }, ///< Near detector -- muon track length in active region [cm]
        { "rec.energy.numusimp.ndtrklencat", NAN }, ///< Near detector -- muon track length in muon catcher [cm]
        { "rec.energy.numusimp.trknonqeE", NAN }, ///< Track length non-quasielastic neutrino energy [GeV]
        { "rec.energy.numusimp.trkqeE", NAN }, ///< Track length quasielastic neutrino energy [GeV]
        { "rec.energy.numusimp.ucccE", NAN }, ///< Uncontained Energy of numu events (TMVA BDT)
        { "rec.hdr.day", NAN }, ///< day of spill within month
        { "rec.hdr.det", NAN }, ///< Detector, NAN }, ND = 1, NAN }, FD = 2, NAN }, NDOS = 3
        { "rec.hdr.doy", NAN }, ///< day of spill within year
        { "rec.hdr.evt", NAN }, ///< ART event number, NAN }, indexes trigger windows.
        { "rec.hdr.hour", NAN }, ///< hour of spill
        { "rec.hdr.ismc", NAN }, ///< data or MC? True if MC
        { "rec.hdr.minute", NAN }, ///< minute of spill
        { "rec.hdr.month", NAN }, ///< month of spill
        { "rec.hdr.run", NAN }, ///< run number
        { "rec.hdr.second", NAN }, ///< second of spill
        { "rec.hdr.subevt", NAN }, ///< slice number within spill
        { "rec.hdr.subevtendtime", NAN }, ///< Slice end time [ns]
        { "rec.hdr.subevtmeantime", NAN }, ///< Slice mean time [ns]
        { "rec.hdr.subevtstarttime", NAN }, ///< time of beginning of slice within spill [ns]
        { "rec.hdr.subrun", NAN }, ///< subrun number
        { "rec.hdr.unixtime", NAN }, ///< unix time of spill
        { "rec.hdr.year", NAN }, ///< year of spill
        { "rec.parent.mrccpar.eff", NAN },
        { "rec.parent.mrccpar.pur", NAN },
        { "rec.parent.mrccpar.slcidx", NAN },
        { "rec.sand.nue.anglecosmic", NAN }, ///Angle of longest cosmictrack
        { "rec.sand.nue.conv0", NAN }, ///< Conversion length of photon 0
        { "rec.sand.nue.conv1", NAN }, ///< Conversion length of photon 1
        { "rec.sand.nue.eedge10cell", NAN }, ///Energy deposit within 10 cells to detector edges
        { "rec.sand.nue.eedge2cell", NAN }, ///Energy deposit within 2 cells to detector edges
        { "rec.sand.nue.eedge5cell", NAN }, ///Energy deposit within 5 cells to detector edges
        { "rec.sand.nue.eidannE", NAN }, ///EID ann trained with E
        { "rec.sand.nue.elll", NAN }, ///Electron likelihood
        { "rec.sand.nue.fracangchanges", NAN },
        { "rec.sand.nue.fracangchangesloose", NAN },
        { "rec.sand.nue.fracmodalhits", NAN },
        { "rec.sand.nue.fracnoncontig", NAN },
        { "rec.sand.nue.fracnoncontigloose", NAN },
        { "rec.sand.nue.infid", NAN }, ///< Geometry::isInFiducialVolume() on slice MeanXYZ
        { "rec.sand.nue.maxgap", NAN },
        { "rec.sand.nue.mincellavg", NAN }, ///Mincell using regression through event
        { "rec.sand.nue.mincellcosmic", NAN }, ///Mincell using all cosmictracker
        { "rec.sand.nue.mincellkalman", NAN }, ///Mincell using regression through event (MIP frac events)
        { "rec.sand.nue.mincellwallcosmic", NAN }, ///Mincellwall using all cosmictracker
        { "rec.sand.nue.mipmax", NAN }, ///Maximum coordinates of box with MIP range hits
        { "rec.sand.nue.mipmin", NAN }, ///Minimum coordinates of box with MIP range hits
        { "rec.sand.nue.ncellrow", NAN }, ///Max number contiguous cells, NAN }, single plane
        { "rec.sand.nue.ncellsum", NAN }, ///Max number cells in total, NAN }, single plane
        { "rec.sand.nue.nmiphits", NAN }, ///Number of MIP Hits (Adjustable boundary)
        { "rec.sand.nue.npi0", NAN }, ///< Number of primary pi0
        { "rec.sand.nue.npim", NAN }, ///< Number of primary pi-
        { "rec.sand.nue.npip", NAN }, ///< Number of primary pi+
        { "rec.sand.nue.nplanesfilled", NAN }, ///Number of filled planes
        { "rec.sand.nue.nxplanes", NAN }, ///Extent in planes for slice X-view
        { "rec.sand.nue.nyplanes", NAN }, ///Extent in planes for slice Y-view
        { "rec.sand.nue.opencos", NAN }, ///< Opening angle of largest pi0's photons
        { "rec.sand.nue.phot0E", NAN }, ///< Energy of largest pi0's largest photon
        { "rec.sand.nue.phot1E", NAN }, ///< Energy of largest pi0's smallest photon
        { "rec.sand.nue.pi0E", NAN }, ///< Energy of largest primary pi0
        { "rec.sand.nue.pimE", NAN }, ///< Energy of largest primary pi-
        { "rec.sand.nue.pipE", NAN }, ///< Energy of largest primary pi+
        { "rec.sand.nue.sh1d2edge", NAN }, ///Shower distance to edge
        { "rec.sand.nue.sh1dir", NAN }, ///Leading shower dir
        { "rec.sand.nue.sh1start", NAN }, ///Leading shower start pt
        { "rec.sand.nue.sh1stop", NAN }, ///Leading shower stop pt
        { "rec.sand.nue.sh1totalL", NAN }, ///Leading shower length
        { "rec.sand.nue.slicemeanx", NAN }, ///< slice's Cluster::MeanXYZ().X()
        { "rec.sand.nue.slicemeany", NAN }, ///< slice's Cluster::MeanXYZ().Y()
        { "rec.sand.nue.slicemeanz", NAN }, ///< slice's Cluster::MeanXYZ().Z()
        { "rec.sand.numu.mudst", NAN }, ///< projected distance from muon end to detector edge
        { "rec.sand.numu.pimudecay", NAN }, ///< did slice contain pi->mu decay?
        { "rec.sand.numu.vtxdst", NAN }, ///< distance from vertex to detector edge
        { "rec.sel.cosrej.anglecos", NAN }, ///< cos of angle of Cosmic Track
        { "rec.sel.cosrej.anglekal", NAN }, ///< cos of angle of best ReMId Kalman track
        { "rec.sel.cosrej.cdirscore", NAN }, ///< chisq score difference between +c and -c assumptions in Hough timing fit (TimingFit module) for cosmic track
        { "rec.sel.cosrej.cfitspeed", NAN }, ///< fit inverse speed (ns/cm) of Hough timing fit (TimingFit module) for cosmic track
        { "rec.sel.cosrej.contcospid", NAN }, ///< cosmic rejection PID for contained events", NAN },
        { "rec.sel.cosrej.cosbakcell", NAN }, ///< cells along distance projected backwards (cosmictrack)
        { "rec.sel.cosrej.cosbakdist", NAN }, ///< distance projected backwards (track end to det wall) (cosmictrack)
        { "rec.sel.cosrej.coschidiff", NAN }, ///< chisq difference of track timing fits (cosmictrack)
        { "rec.sel.cosrej.coschisq", NAN }, ///< chisq value of slope fit (cosmictrack)
        { "rec.sel.cosrej.coscontain", NAN }, ///< cosmic containment cut
        { "rec.sel.cosrej.cosfwdcell", NAN }, ///< cells along distance projected forwards (cosmictrack)
        { "rec.sel.cosrej.cosfwddist", NAN }, ///< distance projected forwards (track end to det wall) (cosmictrack)
        { "rec.sel.cosrej.cosslope", NAN }, ///< slope of track timing fit (cosmictrack)
        { "rec.sel.cosrej.costhetatrue", NAN }, ///< cosine of angle between cosmic trk dir and true dir of most contributing particle
        { "rec.sel.cosrej.cscorediff", NAN }, ///< chisq score difference between best +c/-c assumption and free fit in Hough timing fit (TimingFit module) for Cosmic track
        { "rec.sel.cosrej.endact", NAN }, ///< sum of activity in 50 cm of best kal trk end pos (GeV)
        { "rec.sel.cosrej.eratio", NAN }, ///< ratio of best kalman track GeV / slice GeV
        { "rec.sel.cosrej.fscattext", NAN }, ///< scattering variable (Fernanda)
        { "rec.sel.cosrej.fscattmax", NAN }, ///< maximum scattering angle (Fernanda)
        { "rec.sel.cosrej.fscattsig", NAN }, ///< sigma of scattering variable distribution (Fernanda)
        { "rec.sel.cosrej.fscattsum", NAN }, ///< sum of scattering angles (Fernanda)
        { "rec.sel.cosrej.kalbakcell", NAN }, ///< cells along distance projected backwards (kalmantrack)
        { "rec.sel.cosrej.kalbakdist", NAN }, ///< distance projected backwards (track end to det wall) (kalmantrack)
        { "rec.sel.cosrej.kalchidiff", NAN }, ///< chisq difference of track timing fits (kalmantrack)
        { "rec.sel.cosrej.kalchisq", NAN }, ///< chisq value of slope fit (kalmantrack)
        { "rec.sel.cosrej.kaldist", NAN }, ///< distance from best kal track end to det wall (negative means outside det)
        { "rec.sel.cosrej.kalfwdcell", NAN }, ///< cells along distance projected forwards (kalmantrack)
        { "rec.sel.cosrej.kalfwddist", NAN }, ///< distance projected forwards (track end to det wall) (kalmantrack)
        { "rec.sel.cosrej.kalslope", NAN }, ///< slope of track timing fit (kalmantrack)
        { "rec.sel.cosrej.kalthetatrue", NAN }, ///< cosine of angle between kalman trk dir and true dir of most contributing particle
        { "rec.sel.cosrej.kdirscore", NAN }, ///< chisq score difference between +c and -c assumptions in Hough timing fit (TimingFit module) for best ReMId Kalman track
        { "rec.sel.cosrej.kfitspeed", NAN }, ///< fit inverse speed (ns/cm) of Hough timing fit (TimingFit module) for best ReMId Kalman track
        { "rec.sel.cosrej.kscorediff", NAN }, ///< chisq score difference between best +c/-c assumption and free fit in Hough timing fit (TimingFit module) for best ReMId Kalman track
        { "rec.sel.cosrej.nhadtracks", NAN }, ///< number of 3D kalman tracks w/ ReMId < 0.4
        { "rec.sel.cosrej.ntracks3d", NAN }, ///< number of 3D tracks as determined by Kalman tracker
        { "rec.sel.cosrej.pdgbest", NAN }, ///< pdg code of most contributing true particle to best kalman trk
        { "rec.sel.cosrej.scatt", NAN }, ///< sum of all all angular deviation (looping over kaltrk traj pts) / trk len
        { "rec.sel.cosrej.startact", NAN }, ///< sum of activity in 50 cm of best kal trk start pos (GeV)
        { "rec.sel.cosrej.uncontcospid", NAN }, ///< cosmic rejection PID for uncontained events", NAN },
        { "rec.sel.elecid.ann", NAN }, ///< ann output
        { "rec.sel.elecid.eglll", NAN }, ///<Electron - gamma ll for longitudinal shower
        { "rec.sel.elecid.egllt", NAN }, ///<Electron - gamma ll for ltransverse shower
        { "rec.sel.elecid.elll", NAN }, ///<electron ll for longitudinal shower
        { "rec.sel.elecid.ellt", NAN }, ///<electron ll for transverse shower
        { "rec.sel.elecid.emulll", NAN }, ///<Electron - muon ll for longitudinal shower
        { "rec.sel.elecid.emullt", NAN }, ///<Electron - muon ll for ltransverse shower
        { "rec.sel.elecid.enlll", NAN }, ///<Electron - neutron ll for longitudinal shower
        { "rec.sel.elecid.enllt", NAN }, ///<Electron - neutron ll for ltransverse shower
        { "rec.sel.elecid.epi0lll", NAN }, ///<Electron - Pi0 ll for longitudinal shower
        { "rec.sel.elecid.epi0llt", NAN }, ///<Electron - Pi0 ll for ltransverse shower
        { "rec.sel.elecid.epilll", NAN }, ///<Electron - pion ll for longitudinal shower
        { "rec.sel.elecid.epillt", NAN }, ///<Electron - pion ll for ltransverse shower
        { "rec.sel.elecid.eplll", NAN }, ///<Electron - proton ll for longitudinal shower
        { "rec.sel.elecid.epllt", NAN }, ///<Electron - proton ll for ltransverse shower
        { "rec.sel.elecid.gap", NAN }, ///<gap from vertex to start of shower
        { "rec.sel.elecid.ismuon", NAN }, ///< Boolean to distinguish muons from electrons.
        { "rec.sel.elecid.mulll", NAN }, ///<muon ll for longitudinal shower
        { "rec.sel.elecid.mullt", NAN }, ///<muon ll for transverse shower
        { "rec.sel.elecid.pi0mass", NAN }, ///<best pi0 mass hypothesis coming from combinations of JMShowers
        { "rec.sel.elecid.shE", NAN }, ///<fraction of energy of leading shower out of total energy of slice
        { "rec.sel.elecid.vtxgev", NAN }, ///<Energy of slice in vertex region
        { "rec.sel.lem.avgexpE", NAN }, ///< Fraction of signal matches, NAN }, "exp" weighting
        { "rec.sel.lem.avginvE", NAN }, ///< Fraction of signal matches, NAN }, 1/E weighting
        { "rec.sel.lem.avgsigE", NAN }, ///< Fraction of signal matches, NAN }, sigmoid weighting
        { "rec.sel.lem.chibkg", NAN }, ///< chisq of a fit assuming best match is background
        { "rec.sel.lem.chibkgexp", NAN }, ///< chisq of a fit assuming best match is bkg, NAN }, "exp"
        { "rec.sel.lem.chisig", NAN }, ///< chi-square of a fit assuming best match is signal
        { "rec.sel.lem.chisigexp", NAN }, ///< chisq of a fit assuming best match is signal, NAN }, "exp"
        { "rec.sel.lem.dectree", NAN }, ///< Decision tree PID
        { "rec.sel.lem.energydiff", NAN }, ///< Potential diff between sig and bkg matches
        { "rec.sel.lem.energydiffexp", NAN }, ///< Pot. diff between sig and bkg matches, NAN }, "exp"
        { "rec.sel.lem.enrichfrac", NAN }, ///< Fraction of matches that are "enriched"
        { "rec.sel.lem.enrichfracexp", NAN }, ///< Fraction of matches that are "enriched", NAN }, "exp"
        { "rec.sel.lem.fannid", NAN }, ///< ANN output, NAN }, "fann" package
        { "rec.sel.lem.fannidenrich", NAN }, ///< ANN output, NAN }, including \a enrichfracexp
        { "rec.sel.lem.meanqfrac", NAN }, ///< Mean fraction of charge matched
        { "rec.sel.lem.meanqfracbkg", NAN }, ///< Mean fraction matched in background matches
        { "rec.sel.lem.meanqfracbkgexp", NAN }, ///< Fraction matched in background matches, NAN }, "exp"
        { "rec.sel.lem.meanqfracexp", NAN }, ///< Fraction of charge matched, NAN }, "exp"
        { "rec.sel.lem.meanqfracfit", NAN }, ///< Fraction of charge matched, NAN }, "fit"
        { "rec.sel.lem.meanqfracsig", NAN }, ///< Fraction matched in signal matches
        { "rec.sel.lem.meanqfracsigexp", NAN }, ///< Fraction matched in signal matches "exp"
        { "rec.sel.lem.meany", NAN }, ///< Mean hadronic y of matches
        { "rec.sel.lem.meanyexp", NAN }, ///< Hadronic y of matches, NAN }, "exp"
        { "rec.sel.lem.meanyfit", NAN }, ///< Hadronic y of matches, NAN }, "fit"
        { "rec.sel.lem.meanysig", NAN }, ///< Mean hadronic y of matches that are signal
        { "rec.sel.lem.meanysigexp", NAN }, ///< Hadronic y of matches that are signal, NAN }, "exp"
        { "rec.sel.lem.pid", NAN }, ///< The default PID value - normally use this one
        { "rec.sel.lem.pidexp", NAN }, ///< Fraction of matches that are signal, NAN }, "exp"
        { "rec.sel.lem.pidexpfit", NAN }, ///< Fraction of matches that are signal, NAN }, "expfit"
        { "rec.sel.lem.pidfit", NAN }, ///< Fraction of matches that are signal, NAN }, "fit"
        { "rec.sel.lem.tmvabdt", NAN }, ///< TMVA boosted decision tree, NAN }, adaptive boost
        { "rec.sel.lem.tmvabdtd", NAN }, ///< TMVA boosted decision tree, NAN }, decorrelation
        { "rec.sel.lem.tmvabdtg", NAN }, ///< TMVA boosted decision tree, NAN }, gradient boost
        { "rec.sel.lem.tmvabdtmf", NAN }, ///< TMVA boosted decision tree, NAN }, Fisher discriminant
        { "rec.sel.lem.tmvamlp", NAN }, ///< TMVA ANN ("multi layer perceptron")
        { "rec.sel.nuepre.calE", NAN }, ///< Energy (sum of calibrated hit GeV in slice)
        { "rec.sel.nuepre.cellstoedge", NAN }, ///< Number of cells to side of detector
        { "rec.sel.nuepre.fracmip", NAN }, ///< Fraction of MIP hits
        { "rec.sel.nuepre.lengthdiscrete2d", NAN }, ///< Longest track (discrete)
        { "rec.sel.nuepre.lengthdiscretemerge", NAN }, ///< Longest track (discretetrack merge)
        { "rec.sel.nuepre.lengthkalman", NAN }, ///< Longest track (kalman)
        { "rec.sel.nuepre.mean", NAN }, ///< Mean of slice
        { "rec.sel.nuepre.ncell", NAN }, ///< Number of cell hits in slice
        { "rec.sel.nuepre.passfid", NAN }, ///< Bool saying if event passed Nue Fiducial (current guess)
        { "rec.sel.nuepre.passpre", NAN }, ///< Bool saying if event passed Nue Presel (current guess)
        { "rec.sel.nuepre.planestoback", NAN }, ///< Number of planes to back end of detector
        { "rec.sel.nuepre.planestofront", NAN }, ///< Number of planes to front of detector
        { "rec.sel.nuepre.vtx", NAN }, ///< Elastic vertex
        { "rec.sel.qepid.dedx", NAN }, ///< Ratio of average dE/dx in non-muon track to muon identified track
        { "rec.sel.qepid.ediff", NAN }, ///< Difference in energy between QE angle energy formula and weighted fit, NAN }, normalized to fit energy
        { "rec.sel.qepid.ediffz", NAN }, ///< Difference in energy between QE angle energy formula and weighted fit, NAN }, normalized to error in QE angle energy formula
        { "rec.sel.qepid.mode", NAN }, ///< Neutrino mode of slice
        { "rec.sel.qepid.ntrk", NAN }, ///< Number of tracks used by qepid
        { "rec.sel.qepid.offE", NAN }, ///< Ratio of energy not assoicated with tracks to energy on tracks
        { "rec.sel.qepid.pdg", NAN }, ///< PDG code of slice
        { "rec.sel.qepid.pid", NAN }, ///< PID value of output kNN
        { "rec.sel.remid.bestidx", NAN }, ///< Index of best track and muid object in kalman and muid vectors
        { "rec.sel.remid.bestidx3d", NAN }, ///< Index of best 3d track and muid object in kalman and muid vectors
        { "rec.sel.remid.cont", NAN }, ///< True if track is contained as defined by ReMId
        { "rec.sel.remid.dedxllh", NAN }, ///< Log-likelihood value from dE/dx
        { "rec.sel.remid.len", NAN }, ///< Track length of identified track
        { "rec.sel.remid.measfrac", NAN }, ///< Fraction of planes used to measure dE/dx
        { "rec.sel.remid.nmuid", NAN }, ///< Number of SRMuid in vector
        { "rec.sel.remid.pdg", NAN }, ///< PDG code of identified track, NAN }, pion or muon
        { "rec.sel.remid.pid", NAN }, ///< PID value output by kNN
        { "rec.sel.remid.scatllh", NAN }, ///< Log-likelihood value from scattering angle
        { "rec.sel.rvp.efrac2plwin", NAN }, ///<largest fraction of energy in a 2 plane window out of total slice
        { "rec.sel.rvp.efrac2sigrd", NAN }, ///<fraction of slice energy within 2sigma of mean energy weighted slice position
        { "rec.sel.rvp.efrac6plwin", NAN }, ///<largest fraction of energy in a 6 plane window out of total slice
        { "rec.sel.rvp.eiso3sig", NAN }, ///<fraction of slice energy more then 3sigma from mean energy weighted slice position
        { "rec.sel.rvp.epl20frac", NAN }, ///<fraction of energy in first 20 planes out of total slice energy
        { "rec.sel.rvp.longtr", NAN }, ///<longest discrete track merge track
        { "rec.sel.rvp.longtrfrac", NAN }, ///<fraction of cells in longest track out of total cells in slice
        { "rec.sel.rvp.mipfrac", NAN }, ///<fraction of mip hits out of total number of hits
        { "rec.sel.rvp.miphits", NAN }, ///<number of mip hits in slice
        { "rec.sel.rvp.ncell", NAN }, ///<number of cells in slice
        { "rec.sel.rvp.pid", NAN }, ///< The output value from the BDT
        { "rec.sel.rvp.prongEbal2D", NAN }, ///<the energy balance between highest energy 2D prong and 3D prong in the opposite view
        { "rec.sel.rvp.prongEbal3D", NAN }, ///<the energy balance between the two hightest energy 3D prongs in slice
        { "rec.sel.rvp.prongs2D", NAN }, ///<number of 2D FuzzyK prongs associated with slice
        { "rec.sel.rvp.prongs3D", NAN }, ///<number of 3D FuzzyK prongs associated with slice
        { "rec.sel.rvp.recoE", NAN }, ///<reco energy of slice
        { "rec.sel.rvp.rvp12", NAN }, ///<PID value for 12 variable RVP for comparison purposes
        { "rec.shw.njm", NAN },
        { "rec.slc.boxmax", NAN }, ///< Maximum coordinates box containing all the hits [cm]
        { "rec.slc.boxmin", NAN }, ///< Minumum coordinates box containing all the hits [cm]
        { "rec.slc.calE", NAN }, ///< sum of uncalibrated ADC
        { "rec.slc.endtime", NAN }, ///< end time [ns]
        { "rec.slc.firstcell", NAN }, ///< first cell
        { "rec.slc.firstplane", NAN }, ///< first plane
        { "rec.slc.lastcell", NAN }, ///< last cell
        { "rec.slc.lastplane", NAN }, ///< last plane
        { "rec.slc.meantime", NAN }, ///< mean time [ns]
        { "rec.slc.ncalhit", NAN }, ///< number of hits with calibration
        { "rec.slc.ncellsfromedge", NAN }, ///< minumum number of cells to edge of detector
        { "rec.slc.ncontplanes", NAN }, ///< number of continuous planes
        { "rec.slc.nhit", NAN }, ///< number of hits
        { "rec.slc.nmiphit", NAN }, ///< number of minimum ionizing hits
        { "rec.slc.nnonnoise", NAN }, ///< number of non noise slices in event from which this slice came
        { "rec.slc.starttime", NAN }, ///< start time [ns]
        { "rec.spill.deltaspilltimensec", NAN }, ///< Delta time [ns]
        { "rec.spill.gpsspilltimensec", NAN }, ///< Spill time from GPS [ns]
        { "rec.spill.gpsspilltimesec", NAN }, ///< Spill time from GPS [s]
        { "rec.spill.hornI", NAN }, ///< Horn current
        { "rec.spill.isgoodspill", NAN }, ///< Was the pot for a spill good? (only applicable to data, NAN }, default true)
        { "rec.spill.posx", NAN }, ///< x position on target
        { "rec.spill.posy", NAN }, ///< y position on target
        { "rec.spill.spillpot", NAN }, ///< POT in spill including factor of 1e12 so that a user does not have to apply this correction
        { "rec.spill.spilltimensec", NAN }, ///< Spill time in naonoseconds [ns]
        { "rec.spill.spilltimesec", NAN }, ///< Spill time in seconds [s]
        { "rec.spill.widthx", NAN }, ///< Spill width in x dimension
        { "rec.spill.widthy", NAN }, ///< Spill width in y dimension
        { "rec.trk.ncosmic", NAN }, ///< Number of tracks produced by CosmicTrack (size of vector)
        { "rec.trk.ndiscrete", NAN }, ///< Number of 3D tracks produced by DiscreteTrack (size of vector)
        { "rec.trk.ndiscrete2d", NAN }, ///< Number of 2D tracks produced by DiscreteTrack (size of vector)
        { "rec.trk.nkalman", NAN }, ///< Number of tracks produced by KalmanTrack (size of vector)
        { "rec.vtx.elecid.time", NAN }, ///< Time [ns]
        { "rec.vtx.elecid.vtx", NAN }, ///< Vertex position in detector coordinates. [cm]
        { "rec.vtx.nelastic", NAN }, ///< Number of vertices in elastic vector (elastic.size())
        { "rec.sel.lid.ann", NAN },
        { "rec.sel.lid.anne", NAN },
        { "rec.shw.shwlid.nhit", NAN },         
        { "rec.shw.shwlid.nhitx", NAN },        
        { "rec.shw.shwlid.nhity", NAN },        
        { "rec.shw.shwlid.nplane", NAN },       
        { "rec.shw.shwlid.nplanex", NAN },      
        { "rec.shw.shwlid.nplaney", NAN },      
        { "rec.shw.shwlid.gap", NAN },          
        { "rec.shw.shwlid.maxplanecont", NAN }, 
        { "rec.shw.shwlid.maxplanegap", NAN },  
        { "rec.shw.shwlid.len", NAN },          
        { "rec.shw.shwlid.width", NAN },        
        { "rec.shw.shwlid.calE", NAN },         
        { "rec.shw.shwlid.start", NAN },        
        { "rec.shw.shwlid.dir", NAN },          
        { "rec.shw.shwlid.stop", NAN },         
        { "rec.shw.shwlid.view", NAN }, 
        { "rec.sel.lid.shwlid.ann", NAN },      
        { "rec.sel.lid.shwlid.anne", NAN },     
        { "rec.sel.lid.shwlid.ismuon", NAN },   
        { "rec.sel.lid.shwlid.eglll", NAN },    
        { "rec.sel.lid.shwlid.egllt", NAN },    
        { "rec.sel.lid.shwlid.emulll", NAN },   
        { "rec.sel.lid.shwlid.emullt", NAN },   
        { "rec.sel.lid.shwlid.epi0lll", NAN },  
        { "rec.sel.lid.shwlid.epi0llt", NAN },  
        { "rec.sel.lid.shwlid.eplll", NAN },    
        { "rec.sel.lid.shwlid.epllt", NAN }, 
        { "rec.sel.lid.shwlid.enlll", NAN },  
        { "rec.sel.lid.shwlid.enllt", NAN },  
        { "rec.sel.lid.shwlid.epilll", NAN }, 
        { "rec.sel.lid.shwlid.epillt", NAN }, 
        { "rec.sel.lid.shwlid.vtxgev", NAN },   
        { "rec.sel.lid.shwlid.pi0mass", NAN },  
        { "rec.sel.lid.shwlid.shwEFrac", NAN }, 
        { "rec.sel.lid.shwlid.gap", NAN },      
        { "rec.sel.lid.shwlid.costheta", NAN }, 
        { "rec.sel.lid.shwlid.invglll", NAN },  
        { "rec.energy.lid.E", NAN },    
        { "rec.energy.lid.depE", NAN }, 
        { "rec.energy.lid.shwE", NAN }, 
        { "rec.energy.lid.hadE", NAN }

    });

    /// Filter events based on simple expressions
    class CAFFilter : public art::EDFilter
    {
    public:
        explicit CAFFilter(const fhicl::ParameterSet& pset);
        ~CAFFilter();

        bool filter(art::Event& evt);

        void reconfigure(const fhicl::ParameterSet& pset);

        void beginJob();

    protected:
                bool generatorInfo;
        std::string generatorName;
        std::string exposureName;
        std::string clusterModName;
        std::string hitsModName;
        std::string kalmanTrackName;
        std::string discreteTrackName;
        std::string discreteTrack2dName;
        std::string cosmicTrackName;
        std::string fuzzyVertexName;
        std::string elasticArmsName;
        std::string jmShowerName;
        std::string numuEnergyName;
        std::string remidName;
        std::string qepidName;
        std::string nueSelName;
                std::string lidName;
        std::string lemName;
        std::string rvpName;
        std::string cosrejName;
        std::string kalmanMichelEName;
        std::string cosmicMichelEName;
        std::string discreteMichelEName;
        std::string nueSandboxName;
        std::string numuSandboxName;
        std::string numiBeamName;
        std::string mrccParentName;

        bool isActiveVar(TString var);
        bool isActiveCategory(TString category);
                double convertTVector3ToDouble(TVector3 input);
        valMap FillVals();
        SliceInfo getSliceInfo(const rb::Cluster& slice);  

        TFormula* fFormula;
        std::vector<std::string> exprVars;
    art::ServiceHandle<cheat::BackTracker> bt;  
    art::ServiceHandle<calib::Calibrator>  cal;  
    art::ServiceHandle<geo::Geometry>      geom;    
    
    novadaq::cnv::DetId fDetID;  ///< Detector ID in nova daq convention namespace typedef
    
    unsigned int         fRun;
    unsigned int         fSubrun;
    unsigned int         fEvt;
    bool                 fIsgoodspill;
    unsigned long int    fSpilltimesec;
    unsigned long int    fSpilltimensec;
    unsigned long int    fGpsspilltimesec;
    unsigned long int    fGpsspilltimensec;
    signed long long int fDeltaspilltimensec;
    double               fSpillpot;
    double               fHornI;
    double               fPosx;
    double               fPosy;
    double               fWidthx;
    double               fWidthy;


static bool sortRBTrackLength(const art::Ptr<rb::Track>& a, const  art::Ptr<rb::Track>& b) {return a->TotalLength()>b->TotalLength(); }
    };

    //......................................................................
    CAFFilter::CAFFilter(const fhicl::ParameterSet& pset)
        : fFormula(0),
                  bt(),
                  cal(),
                  geom()
    {
        reconfigure(pset);
    }

    //......................................................................
    CAFFilter::~CAFFilter()
    {
        delete fFormula;
    }

    //......................................................................
    void CAFFilter::reconfigure(const fhicl::ParameterSet& pset)
    {
        // Keep everything lined up for clarity (stylistic)
        generatorName = pset.get< std::string    >("GeneratorInput");
        exposureName = pset.get< std::string    >("ExposureInput");
        clusterModName = pset.get< std::string    >("ClusterInput");
        hitsModName = pset.get< std::string    >("HitInput");
        kalmanTrackName = pset.get< std::string    >("KalmanInput");
        discreteTrackName = pset.get< std::string    >("DiscreteInput");
        discreteTrack2dName = pset.get< std::string    >("Discrete2dInput");
        cosmicTrackName = pset.get< std::string    >("CosmicInput");
        fuzzyVertexName = pset.get< std::string    >("FuzzyInput");
        elasticArmsName = pset.get< std::string    >("ElasticInput");
        jmShowerName = pset.get< std::string    >("JmShowerInput");
        numuEnergyName = pset.get< std::string    >("NumuEnergyInput");
        remidName = pset.get< std::string    >("RemidInput");
        qepidName = pset.get< std::string    >("QepidInput");
        nueSelName = pset.get< std::string    >("NueSelInput");
                lidName =   pset.get< std::string    >("LIDInput");
        lemName = pset.get< std::string    >("LemInput");
        rvpName = pset.get< std::string    >("RvpInput");
        cosrejName = pset.get< std::string    >("CosRejInput");
        kalmanMichelEName = pset.get< std::string    >("KalmanMichelEInput");
        cosmicMichelEName = pset.get< std::string    >("CosmicMichelEInput");
        discreteMichelEName = pset.get< std::string    >("DiscreteMichelEInput");
        nueSandboxName = pset.get< std::string    >("NueSandboxInput");
        numuSandboxName = pset.get< std::string    >("NumuSandboxInput");
        numiBeamName = pset.get< std::string    >("NuMIBeamInput");
        generatorInfo = pset.get< bool           >("GeneratorInfo");
        mrccParentName = pset.get<std::string     >("MRCCParentInput");

        TString expr = pset.get<std::string>("Expression");

        // Empty string evaluates true
        if (expr == "") 
{
expr = "1";
                mf::LogWarning("CAFFilter") << "Null Expression";
}

        std::cout << "Expression is: " << expr << std::endl;
                exprVars.clear();
mf::LogWarning("CAFFilter") << "CAFFilterMap size is: " << CAFFilterMap.size();
        for( auto it = CAFFilterMap.begin(); it != CAFFilterMap.end(); ++it)
        {
                mf::LogWarning("CAFFilter") << "Comparing: " << it->first << " to " << expr;
            if (expr.Index(it->first,it->first.length(),0,TString::kIgnoreCase) != kNPOS)
            {
                mf::LogWarning("CAFFilter") << "Match Found!" ;
                exprVars.push_back(it->first);
                expr.ReplaceAll(it->first, TString::Format("[%lu]", exprVars.size() - 1));
                mf::LogWarning("CAFFilter") << "Expression is now: " << expr;
            }
        }

        delete fFormula;
        fFormula = new TFormula("form", expr);
    }

    //......................................................................
    void CAFFilter::beginJob()
    {
    }

    //......................................................................
    bool CAFFilter::isActiveCategory(TString category)
    {        
        for (unsigned int i = 0; i < exprVars.size(); i++)
        {
            if (exprVars[i].find(category) != std::string::npos)
            {
                return true;
            }
        }
        
        return false;
    }


        //.....................................................................
        double CAFFilter::convertTVector3ToDouble(TVector3 input)
        {
             double output = input.X();
             double y = input.Y() * 100000;
             double z = input.Z() * 100000 * 100000;
             return output + y + z;
        }


    //.....................................................................
    SliceInfo CAFFilter::getSliceInfo(const rb::Cluster& slice)
    {

        const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
        const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;


        //Variables needed:
        //Holds contiguous planes info:
        int           planeNum = -1;
        unsigned int  cellsFromEdge = -1;
        std::vector<int> planelist;

        //Hit counters:
        double nTotHit = 0.0;
        double nCalHit = 0.0;
        double nMipHit = 0.0;

        unsigned int maxPlane = 0;
        unsigned int minPlane = 99999999;
        unsigned int maxCell = 0;
        unsigned int minCell = 99999999;
        unsigned int minCellwall = 99999999;
        unsigned int minCellsFromEdge = 99999999;

        //Loop over slice in question here:
        for (unsigned int hitIdx = 0; hitIdx < slice.NCell(); ++hitIdx)
        {
            //count up total number of cell hits
            nTotHit++;

            const art::Ptr<rb::CellHit>& chit = slice.Cell(hitIdx);
            //Get calbrated reco hit:

              const rb::RecoHit rhit(cal->MakeRecoHit(*chit,
                                  chit->View() == geo::kX ? wx : wy));

            //Make sure things have been calibrated:
            if (!rhit.IsCalibrated())
            {
                mf::LogWarning("CAFMaker") << "Not calibrated?!";
                continue;
            }

            if (chit->Plane() < minPlane)
            {
                minPlane = chit->Plane();
            }
            if (chit->Plane() > maxPlane)
            {
                maxPlane = chit->Plane();
            }
            if (chit->Cell() < minCell)
            {
                minCell = chit->Cell();
            }
            if (chit->Cell() > maxCell)
            {
                maxCell = chit->Cell();
            }

            planeNum = chit->Plane();
            cellsFromEdge = std::min((unsigned int)chit->Cell(), geom->Plane(planeNum)->Ncells() - 1 - chit->Cell());

            if (cellsFromEdge < minCellsFromEdge)
            {
                minCellsFromEdge = cellsFromEdge;
            }

            //Fill a vector of plane numbers with hits;
            if (std::find(planelist.begin(), planelist.end(), planeNum) == planelist.end()) {
                planelist.push_back(planeNum);
            }


            //Count up both the total calibrated hits
            //and the MIP hits from 100-245 PECorr
            //Count up total hits:
            nCalHit++;
            //Count up MIP hits:
            if ((rhit.PECorr() > 100.0) && (rhit.PECorr() < 245.0)) nMipHit++;

        } //END OF SLICE LOOP HERE



        //calculate Contiguous Planes variable (maxContig) in the remaining code:
        int nContig = 1;
        int maxContig = -1;



        //Calculate maximum contiguous planes
        if (planelist.size() > 0) std::sort(planelist.begin(), planelist.end());
        for (unsigned int iplane = 0; iplane<planelist.size(); iplane++)
        {

            if (iplane != 0)
            {
                if (planelist[iplane] == (planelist[iplane - 1] + 1))
                {
                    nContig++;
                    if (nContig>maxContig) maxContig = nContig;
                }
                else {
                    nContig = 0;
                }
            }
        }

        if (maxContig < 0) maxContig = 0;

        SliceInfo slcInfo;
        slcInfo.nTotHit = nTotHit;
        slcInfo.nCalHit = nCalHit;
        slcInfo.nMipHit = nMipHit;
        slcInfo.maxContig = maxContig;
        slcInfo.maxPlane = maxPlane;
        slcInfo.minPlane = minPlane;
        slcInfo.maxCell = maxCell;
        slcInfo.minCell = minCell;
        slcInfo.minCell = minCellwall;
        slcInfo.minCellsFromEdge = minCellsFromEdge;


        return slcInfo;
    }

    //......................................................................
    bool CAFFilter::filter(art::Event& evt)
    {
        feenableexcept(FE_INVALID);

        //get a handle on the slices
        art::Handle<std::vector<rb::Cluster> > hSlices;
        evt.getByLabel(clusterModName, hSlices);
        if (hSlices.failedToGet())
        {
            mf::LogError("CAFFilter") << "Slices not found in CAFFilter";
        }

        // define a vector holding slices 
        art::PtrVector<rb::Cluster> subevts;
        for (unsigned int i = 0; i < hSlices->size(); ++i)
            subevts.push_back(art::Ptr<rb::Cluster>(hSlices, i));


        //Get all of the hits from calhits
        art::Handle< std::vector<rb::CellHit> > hitsHandle;
        evt.getByLabel(hitsModName, hitsHandle);
        art::PtrVector<rb::CellHit> allHits;
        for (size_t i = 0; i < hitsHandle->size(); ++i)
        {
            art::Ptr<rb::CellHit> hit(hitsHandle, i);
            allHits.push_back(hit);
        }

        // We're gonna get a vector of NeutrinoEffPurs for allHits
        std::vector<cheat::NeutrinoWithIndex> allNus = bt->allMCTruth();

        std::vector< std::vector<cheat::NeutrinoEffPur> > sliceTruthTable = bt->SlicesToMCTruthsTable(subevts);



        // Find favorite neutrino IDs using several metrics.  Total energy is the one we will use, for 
        // the favorite neutrino, but the others will be stored.
        //std::vector<int> faveIdsEff = getFaveIdsByEfficiency(allNus, sliceTruthTable);
        //std::vector<int> faveIdsEnergy = getFaveIdsByEnergy(allNus, sliceTruthTable);
        //std::vector<int> faveIdsPur = getFaveIdsByPurity(allNus, sliceTruthTable);
        //std::vector<int> faveIdsEffPur = getFaveIdsByEfficiencyTimesPurity(allNus, sliceTruthTable);
        //std::vector<int> faveIdsEffThenPur = getFaveIdsByEfficiencyThenPurity(allNus, sliceTruthTable);

        //Get all truth track hits in the event
        std::vector<cheat::TrackIDE> allTracks;
        if (bt->HaveTruthInfo()) allTracks = bt->HitsToTrackIDE(allHits);




        //try to get pot spill info if this is real data
        art::Handle<sumdata::SpillData> spillPot;
        evt.getByLabel(numiBeamName, spillPot);
        //if we fail to identify a spill, we want it set to bad
        fIsgoodspill = false;
        fSpilltimesec = 0;
        fSpilltimensec = 0;
        fGpsspilltimesec = 0;
        fGpsspilltimensec = 0;
        fDeltaspilltimensec = 0;
        fSpillpot = 0.0;
        fHornI = 0.0;
        fPosx = 0.0;
        fPosy = 0.0;
        fWidthx = 0.0;
        fWidthy = 0.0;

        //need to flag each spill if the pot was good or bad
        //each slice will get the same tag
        //mc will flag each spill as good since mc does not make bad spills
        if (bt->HaveTruthInfo())
        {
            fIsgoodspill = true;
        }
        //if this is data try to get the spill info
        else {
            //if there was no spill info for this slice mark it bad. log message this should not happen
            if (spillPot.failedToGet())
            {
                mf::LogError("CAFMaker") << "Spill Data not found for real data event";
                fIsgoodspill = false;
            }
            //if we have a SpillData object, get all the variables
            else {
                fIsgoodspill = spillPot->goodbeam;
                fSpilltimesec = spillPot->spilltimesec;
                fSpilltimensec = spillPot->spilltimensec;
                fGpsspilltimesec = spillPot->gpsspilltimensec;
                fGpsspilltimensec = spillPot->gpsspilltimensec;
                fDeltaspilltimensec = spillPot->deltaspilltimensec;
                //NOTE: there is an implied factor of 1e12 when getting IFDBSpillInfo POT.  
                //Adding this back in to avoid confusion between data and mc
                fSpillpot = (spillPot->spillpot)*(1e12);
                fHornI = spillPot->hornI;
                fPosx = spillPot->posx;
                fPosy = spillPot->posy;
                fWidthx = spillPot->widthx;
                fWidthy = spillPot->widthy;
            }
        }

        //Add in run, subrun, evt information so that the spill tree can be 
        //mapped to the record tree.
        fRun = evt.run();
        fSubrun = evt.subRun();
        fEvt = evt.id().event();



        // Get all association finder objects
        art::FindManyP<rb::Track>               fmDiscrete2d(hSlices, evt, discreteTrack2dName);
        art::FindManyP<rb::Track>               fmDiscrete(hSlices, evt, discreteTrackName);
        art::FindManyP<rb::Track>               fmKalman(hSlices, evt, kalmanTrackName);
        art::FindManyP<rb::Track>               fmCosmic(hSlices, evt, cosmicTrackName);
        art::FindManyP<rb::Vertex>              fmElastic(hSlices, evt, elasticArmsName);
        art::FindManyP<jmshower::JMShower>      fmJmShower(hSlices, evt, jmShowerName);
        art::FindManyP<jmshower::EID>           fmNueSel(hSlices, evt, nueSelName);
        art::FindManyP<numue::NumuE>            fmNumuEnergy(hSlices, evt, numuEnergyName);
        art::FindManyP<qeef::QePId>             fmQepid(hSlices, evt, qepidName);
                art::FindOneP<slid::EventLID>           foLID(hSlices, evt, lidName); //one lid object per slice
            art::FindManyP<slid::ShowerLID>         fmShwLID(hSlices, evt, lidName);
        art::FindManyP<nuesand::NueSandObj>     fmNueSand(hSlices, evt, nueSandboxName);
        art::FindManyP<numusand::NumuSandObj>   fmNumuSand(hSlices, evt, numuSandboxName);
        art::FindManyP<lem::PIDDetails>         fmLem(hSlices, evt, lemName);
        art::FindManyP<rvp::RVP>                fmRvp(hSlices, evt, rvpName);
        art::FindManyP<cosrej::CosRejObj>       fmCosRej(hSlices, evt, cosrejName);
        art::FindManyP<murem::MRCCParent>       fmMrccPar(hSlices, evt, mrccParentName);

        int nNonNoise = 0;
        for (unsigned int subevtId = 0; subevtId < subevts.size(); ++subevtId)
        {
            if (subevts[subevtId]->IsNoise() || subevts[subevtId]->NCell() == 0)
                continue;
            ++nNonNoise;
        }
        for (unsigned int subevtId = 0; subevtId < subevts.size(); ++subevtId)
        {
            if (subevts[subevtId]->IsNoise() || subevts[subevtId]->NCell() == 0)  continue;

            valMap output(CAFFilterMap);

            if (isActiveCategory("rec.hdr"))
            {
                //#######################################################
                // Fill header.
                //#######################################################
                // Get metadata information for header
                unsigned int run = evt.run();
                unsigned int subrun = evt.subRun();
                unsigned int spillNum = evt.id().event();
                bool isMC = bt->HaveTruthInfo();

                art::Timestamp ts = evt.time();
                const time_t timeSec = ts.timeHigh();

                struct tm* timeStruct = localtime(&timeSec);

                output["rec.hdr.run"] = run;
                output["rec.hdr.subrun"] = subrun;
                output["rec.hdr.evt"] = spillNum;
                output["rec.hdr.subevt"] = subevtId;
                output["rec.hdr.ismc"] = isMC;
                output["rec.hdr.det"] = fDetID;

                output["rec.hdr.subevtstarttime"] = subevts[subevtId]->MinTNS();
                output["rec.hdr.subevtendtime"] = subevts[subevtId]->MaxTNS();
                output["rec.hdr.subevtmeantime"] = subevts[subevtId]->MeanTNS();
                output["rec.hdr.unixtime"] = timeSec;

                output["rec.hdr.year"] = timeStruct->tm_year + 1900;
                output["rec.hdr.month"] = timeStruct->tm_mon + 1;
                output["rec.hdr.day"] = timeStruct->tm_mday;
                output["rec.hdr.doy"] = timeStruct->tm_yday + 1;
                output["rec.hdr.hour"] = timeStruct->tm_hour + 1;
                output["rec.hdr.minute"] = timeStruct->tm_min;
                output["rec.hdr.second"] = timeStruct->tm_sec;
            }
            if (isActiveCategory("rec.spill"))
            {
                //#######################################################
                // Fill Spill info.
                //#######################################################
                //Store info on NuMI beam quality and if the pot as good or bad.
                //This is meant to be used to refine a good spill, not for
                //pot counting since events have multiple slices and empty
                //spill are missing from this tree.  Use the friend tree for
                //pot counting.
                output["rec.spill.isgoodspill"] = fIsgoodspill;
                output["rec.spill.spilltimesec"] = fSpilltimesec;
                output["rec.spill.spilltimensec"] = fSpilltimensec;
                output["rec.spill.gpsspilltimesec"] = fGpsspilltimesec;
                output["rec.spill.gpsspilltimensec"] = fGpsspilltimensec;
                output["rec.spill.deltaspilltimensec"] = fDeltaspilltimensec;
                output["rec.spill.spillpot"] = fSpillpot;
                output["rec.spill.hornI"] = fHornI;
                output["rec.spill.posx"] = fPosx;
                output["rec.spill.posy"] = fPosy;
                output["rec.spill.widthx"] = fWidthx;
                output["rec.spill.widthy"] = fWidthy;
            }
            if (isActiveCategory("rec.slc"))
            {
                //#######################################################
                // Add slice info.
                //#######################################################
                SliceInfo slcInfo = getSliceInfo(*subevts[subevtId]);
                output["rec.slc.nhit"] = slcInfo.nTotHit;
                output["rec.slc.ncalhit"] = slcInfo.nCalHit;
                output["rec.slc.nmiphit"] = slcInfo.nMipHit;
                output["rec.slc.ncontplanes"] = slcInfo.maxContig;
                output["rec.slc.firstplane"] = slcInfo.minPlane;
                output["rec.slc.lastplane"] = slcInfo.maxPlane;
                output["rec.slc.firstcell"] = slcInfo.minCell;
                output["rec.slc.firstcell"] = slcInfo.minCellwall;
                output["rec.slc.lastcell"] = slcInfo.maxCell;
                output["rec.slc.ncellsfromedge"] = slcInfo.minCellsFromEdge;
                output["rec.slc.nnonnoise"] = nNonNoise;

                output["rec.slc.calE"] = subevts[subevtId]->TotalGeV();
                output["rec.slc.starttime"] = subevts[subevtId]->MinTNS();
                output["rec.slc.endtime"] = subevts[subevtId]->MaxTNS();
                output["rec.slc.meantime"] = subevts[subevtId]->MeanTNS();

                output["rec.slc.boxmin"] = convertTVector3ToDouble(subevts[subevtId]->MinXYZ());
                output["rec.slc.boxmax"] = convertTVector3ToDouble(subevts[subevtId]->MaxXYZ());
            }
            if (isActiveCategory("rec.energy"))
            {
                if (isActiveCategory("rec.energy.elecid"))
                {
                    // Get the JM EID objects and add them.
                    if (fmNueSel.isValid())
                    {
                        std::vector< art::Ptr<jmshower::EID> > eids = fmNueSel.at(subevtId);
                        if (eids.size() > 0)
                        {
                            output["rec.energy.elecid.E"] = eids[0]->NuEnergy();
                            output["rec.energy.elecid.depE"] = eids[0]->DetEnergy();
                            output["rec.energy.elecid.shwE"] = eids[0]->Sh1Energy();
                        }
                    }
                }

                if(isActiveCategory("rec.energy.lid"))
                {
                    std::cout << "foLID.isValid()" << std::endl;
                    //  Get slice level PID from LIDBuilder
                    if(foLID.isValid())
                    {
                        cet::maybe_ref<art::Ptr<slid::EventLID> const> rlid(foLID.at(subevtId));
                        art::Ptr<slid::EventLID> elid = rlid.ref();
                        //if the pointer is valid
                        if (elid)
                        {
                          std::cout << "fmShwLID.isValid()" << std::endl;
                          if(fmShwLID.isValid())
                          {      
                                std::vector< art::Ptr<slid::ShowerLID> > shwlids = fmShwLID.at(subevtId);
  
                                output["rec.energy.lid.E"]    = shwlids[0]->NueEnergy();
                                output["rec.energy.lid.depE"] = shwlids[0]->ShowerDepEnergy();
                                output["rec.energy.lid.shwE"] = shwlids[0]->ShowerEnergy();
                                output["rec.energy.lid.hadE"] = shwlids[0]->HadronicEnergy();
                          }
                       }
                    }
                }

                if (isActiveCategory("rec.energy.numusimp"))
                {
                    if (fmNumuEnergy.isValid())
                    {
                        std::vector< art::Ptr<numue::NumuE> > numuEnergies = fmNumuEnergy.at(subevtId);
                        if (!numuEnergies.empty())
                        {
                            output["rec.energy.numusimp.E"] = numuEnergies[0]->E();
                            output["rec.energy.numusimp.calccE"] = numuEnergies[0]->CalCCE();
                            output["rec.energy.numusimp.trkqeE"] = numuEnergies[0]->TrkQEE();
                            output["rec.energy.numusimp.trknonqeE"] = numuEnergies[0]->TrkNonQEE();
                            output["rec.energy.numusimp.angleE"] = numuEnergies[0]->AngleQEE();
                            output["rec.energy.numusimp.angleerror"] = numuEnergies[0]->AngleQEError();
                            output["rec.energy.numusimp.hadcalE"] = numuEnergies[0]->HadCalE();
                            output["rec.energy.numusimp.hadtrkE"] = numuEnergies[0]->HadTrkE();
                            output["rec.energy.numusimp.ucccE"] = numuEnergies[0]->UCCCE();
                            output["rec.energy.numusimp.fsE"] = numuEnergies[0]->FScatterE();
                            output["rec.energy.numusimp.ndtrklenact"] = numuEnergies[0]->NDTrkLenAct();
                            output["rec.energy.numusimp.ndtrklencat"] = numuEnergies[0]->NDTrkLenCat();
                            output["rec.energy.numusimp.ndtrkcalactE"] = numuEnergies[0]->NDTrkCalAct();
                            output["rec.energy.numusimp.ndtrkcaltranE"] = numuEnergies[0]->NDTrkCalTran();
                            output["rec.energy.numusimp.ndtrkcalcatE"] = numuEnergies[0]->NDTrkCalCat();
                            output["rec.energy.numusimp.ndhadcalactE"] = numuEnergies[0]->NDHadCalAct();
                            output["rec.energy.numusimp.ndhadcaltranE"] = numuEnergies[0]->NDHadCalTran();
                            output["rec.energy.numusimp.ndhadcalcatE"] = numuEnergies[0]->NDHadCalCat();
                        }
                    }
                }
            }
            if (isActiveCategory("rec.parent.mrccpar"))
            {
                if (fmMrccPar.isValid()){
                    std::vector< art::Ptr<murem::MRCCParent> > mrccParent = fmMrccPar.at(subevtId);
                    if (!mrccParent.empty()){
                        output["rec.parent.mrccpar.eff"] = mrccParent[0]->ParentEff();
                        output["rec.parent.mrccpar.pur"] = mrccParent[0]->ParentPur();
                        output["rec.parent.mrccpar.slcidx"] = mrccParent[0]->ParentIndex();
                    }
                }
            }
            if (isActiveCategory("rec.sand"))
            {
                if (isActiveCategory("rec.sand.nue"))
                {
                    if (fmNueSand.isValid())
                    {
                        std::vector<art::Ptr<nuesand::NueSandObj> > nueSands = fmNueSand.at(subevtId);
                        if (!nueSands.empty())
                        {
                            output["rec.sand.nue.infid"] = nueSands[0]->InFiducialVolume();
                            output["rec.sand.nue.slicemeanx"] = nueSands[0]->SliceMeanPosition().X();
                            output["rec.sand.nue.slicemeany"] = nueSands[0]->SliceMeanPosition().Y();
                            output["rec.sand.nue.slicemeanz"] = nueSands[0]->SliceMeanPosition().Z();

                            output["rec.sand.nue.npi0"] = nueSands[0]->fNPi0;
                            output["rec.sand.nue.npip"] = nueSands[0]->fNPiPlus;
                            output["rec.sand.nue.npim"] = nueSands[0]->fNPiMinus;

                            output["rec.sand.nue.pi0E"] = nueSands[0]->fPi0Energy;
                            output["rec.sand.nue.pipE"] = nueSands[0]->fPiPlusEnergy;
                            output["rec.sand.nue.pimE"] = nueSands[0]->fPiMinusEnergy;

                            output["rec.sand.nue.phot0E"] = nueSands[0]->fPhot0E;
                            output["rec.sand.nue.phot1E"] = nueSands[0]->fPhot1E;
                            output["rec.sand.nue.opencos"] = nueSands[0]->fOpenCos;
                            output["rec.sand.nue.conv0"] = nueSands[0]->fConvL0;
                            output["rec.sand.nue.conv1"] = nueSands[0]->fConvL1;

                            output["rec.sand.nue.mincellcosmic"] = nueSands[0]->fMincellCosmic;
                            output["rec.sand.nue.mincellwallcosmic"] = nueSands[0]->fMincellWallCosmic;
                            output["rec.sand.nue.anglecosmic"] = nueSands[0]->fAngleCosmic;
                            output["rec.sand.nue.nplanesfilled"] = nueSands[0]->fFilledPlanes;
                            output["rec.sand.nue.nxplanes"] = nueSands[0]->fExtentXPlanes;
                            output["rec.sand.nue.nyplanes"] = nueSands[0]->fExtentYPlanes;
                            output["rec.sand.nue.ncellrow"] = nueSands[0]->fMaxRowOfCells;
                            output["rec.sand.nue.ncellsum"] = nueSands[0]->fMaxCellsInPlane;
                            output["rec.sand.nue.mincellavg"] = nueSands[0]->fMincellAvg;
                            output["rec.sand.nue.mincellkalman"] = nueSands[0]->fMincellKalman;
                            output["rec.sand.nue.nmiphits"] = nueSands[0]->fNMIPHits;
                            output["rec.sand.nue.mipmin"] = convertTVector3ToDouble(nueSands[0]->fMIPMin);
                            output["rec.sand.nue.mipmax"] = convertTVector3ToDouble(nueSands[0]->fMIPMax);

                            output["rec.sand.nue.fracangchanges"] = nueSands[0]->fFracAngChanges;
                            output["rec.sand.nue.fracangchangesloose"] = nueSands[0]->fFracAngChangesLoose;
                            output["rec.sand.nue.fracmodalhits"] = nueSands[0]->fFracModalHits;
                            output["rec.sand.nue.fracnoncontig"] = nueSands[0]->fFracNonContig;
                            output["rec.sand.nue.fracnoncontigloose"] = nueSands[0]->fFracNonContigLoose;
                            output["rec.sand.nue.maxgap"] = nueSands[0]->fMaxGap;
                        }
                    }

                    if (fmNueSel.isValid())
                    {
                        std::vector< art::Ptr<jmshower::EID> > eids = fmNueSel.at(subevtId);
                        if (eids.size() > 0)
                        {
                            output["rec.sand.nue.eidannE"] = eids[0]->ANNE();
                            output["rec.sand.nue.sh1totalL"] = eids[0]->Sh1TotalLength();
                            output["rec.sand.nue.sh1start"] = convertTVector3ToDouble(eids[0]->Sh1Start());
                            output["rec.sand.nue.sh1stop"] = convertTVector3ToDouble(eids[0]->Sh1Stop());
                            output["rec.sand.nue.sh1dir"] = convertTVector3ToDouble(eids[0]->Sh1Dir());
                            output["rec.sand.nue.elll"] = eids[0]->ELLL();
                            output["rec.sand.nue.sh1d2edge"] = eids[0]->Sh1DistToEdge();
                            output["rec.sand.nue.eedge2cell"] = eids[0]->EEdge2Cell();
                            output["rec.sand.nue.eedge5cell"] = eids[0]->EEdge5Cell();
                            output["rec.sand.nue.eedge10cell"] = eids[0]->EEdge10Cell();
                        }
                    }
                }
                if (isActiveCategory("rec.sand.numu"))
                {
                    if (fmNumuSand.isValid())
                    {
                        std::vector< art::Ptr<numusand::NumuSandObj> > numuSands = fmNumuSand.at(subevtId);
                        if (numuSands.size() > 0)
                        {
                            output["rec.sand.numu.pimudecay"] = numuSands[0]->PiMuDecay();
                            output["rec.sand.numu.vtxdst"] = numuSands[0]->VtxDst();
                            output["rec.sand.numu.mudst"] = numuSands[0]->MuDst();
                        }
                    }
                }
            }
            if (isActiveCategory("rec.sel.cosrej"))
            {
                if (fmCosRej.isValid())
                {
                    std::vector< art::Ptr<cosrej::CosRejObj> > cosRejs = fmCosRej.at(subevtId);
                    if (cosRejs.size() > 0)
                    {
                        output["rec.sel.cosrej.anglekal"] = cosRejs[0]->AngleKal();
                        output["rec.sel.cosrej.anglecos"] = cosRejs[0]->AngleCos();
                        output["rec.sel.cosrej.contcospid"] = cosRejs[0]->ConCosPID();
                        output["rec.sel.cosrej.uncontcospid"] = cosRejs[0]->UnconCosPID();
                        output["rec.sel.cosrej.coscontain"] = cosRejs[0]->CosContain();
                        output["rec.sel.cosrej.ntracks3d"] = cosRejs[0]->NTracks3D();
                        output["rec.sel.cosrej.nhadtracks"] = cosRejs[0]->NHadTracks();
                        output["rec.sel.cosrej.kfitspeed"] = cosRejs[0]->KFitSpeed();
                        output["rec.sel.cosrej.cfitspeed"] = cosRejs[0]->CFitSpeed();
                        output["rec.sel.cosrej.kdirscore"] = cosRejs[0]->KDirScore();
                        output["rec.sel.cosrej.cdirscore"] = cosRejs[0]->CDirScore();
                        output["rec.sel.cosrej.kscorediff"] = cosRejs[0]->KScoreDiff();
                        output["rec.sel.cosrej.cscorediff"] = cosRejs[0]->CScoreDiff();
                        output["rec.sel.cosrej.kalslope"] = cosRejs[0]->KalSlope();
                        output["rec.sel.cosrej.kalchisq"] = cosRejs[0]->KalChisq();
                        output["rec.sel.cosrej.kalchidiff"] = cosRejs[0]->KalChiDiff();
                        output["rec.sel.cosrej.cosslope"] = cosRejs[0]->CosSlope();
                        output["rec.sel.cosrej.coschisq"] = cosRejs[0]->CosChisq();
                        output["rec.sel.cosrej.coschidiff"] = cosRejs[0]->CosChiDiff();
                        output["rec.sel.cosrej.cosfwddist"] = cosRejs[0]->CosFwdDist();
                        output["rec.sel.cosrej.cosbakdist"] = cosRejs[0]->CosBakDist();
                        output["rec.sel.cosrej.cosfwdcell"] = cosRejs[0]->CosFwdCell();
                        output["rec.sel.cosrej.cosbakcell"] = cosRejs[0]->CosBakCell();
                        output["rec.sel.cosrej.kalfwddist"] = cosRejs[0]->KalFwdDist();
                        output["rec.sel.cosrej.kalbakdist"] = cosRejs[0]->KalBakDist();
                        output["rec.sel.cosrej.kalfwdcell"] = cosRejs[0]->KalFwdCell();
                        output["rec.sel.cosrej.kalbakcell"] = cosRejs[0]->KalBakCell();
                        output["rec.sel.cosrej.eratio"] = cosRejs[0]->ERatio();
                        output["rec.sel.cosrej.scatt"] = cosRejs[0]->Scatt();
                        output["rec.sel.cosrej.startact"] = cosRejs[0]->StartAct();
                        output["rec.sel.cosrej.endact"] = cosRejs[0]->EndAct();
                        output["rec.sel.cosrej.pdgbest"] = cosRejs[0]->PDGBest();
                        output["rec.sel.cosrej.costhetatrue"] = cosRejs[0]->CosThetaTrue();
                        output["rec.sel.cosrej.kalthetatrue"] = cosRejs[0]->KalThetaTrue();
                        output["rec.sel.cosrej.kaldist"] = cosRejs[0]->KalDist();
                        output["rec.sel.cosrej.fscattmax"] = cosRejs[0]->FScattMax();
                        output["rec.sel.cosrej.fscattsum"] = cosRejs[0]->FScattSum();
                        output["rec.sel.cosrej.fscattext"] = cosRejs[0]->FScattExt();
                        output["rec.sel.cosrej.fscattsig"] = cosRejs[0]->FScattSig();
                    }
                }
            }
            if (isActiveCategory("rec.sel.elecid"))
            {
                if (fmNueSel.isValid())
                {
                    std::vector< art::Ptr<jmshower::EID> > eids = fmNueSel.at(subevtId);
                    if (eids.size() > 0)
                    {
                        output["rec.sel.elecid.ann"] = eids[0]->ANN();
                        output["rec.sel.elecid.ismuon"] = eids[0]->IsMuon();
                        output["rec.sel.elecid.eglll"] = eids[0]->EGLLL();
                        output["rec.sel.elecid.egllt"] = eids[0]->EGLLT();
                        output["rec.sel.elecid.emulll"] = eids[0]->EMuLLL();
                        output["rec.sel.elecid.emullt"] = eids[0]->EMuLLT();
                        output["rec.sel.elecid.epi0lll"] = eids[0]->EPi0LLL();
                        output["rec.sel.elecid.epi0llt"] = eids[0]->EPi0LLT();
                        output["rec.sel.elecid.eplll"] = eids[0]->EPLLL();
                        output["rec.sel.elecid.epllt"] = eids[0]->EPLLT();
                        output["rec.sel.elecid.enlll"] = eids[0]->ENLLL();
                        output["rec.sel.elecid.enllt"] = eids[0]->ENLLT();
                        output["rec.sel.elecid.epilll"] = eids[0]->EPiLLL();
                        output["rec.sel.elecid.epillt"] = eids[0]->EPiLLT();
                        output["rec.sel.elecid.vtxgev"] = eids[0]->VtxGeV();
                        output["rec.sel.elecid.pi0mass"] = eids[0]->Pi0Mass();
                        output["rec.sel.elecid.shE"] = eids[0]->ShE();
                        output["rec.sel.elecid.gap"] = eids[0]->Gap();
                        output["rec.sel.elecid.elll"] = eids[0]->ELLL();
                        output["rec.sel.elecid.ellt"] = eids[0]->ELLT();
                        output["rec.sel.elecid.mulll"] = eids[0]->MuLLL();
                        output["rec.sel.elecid.mullt"] = eids[0]->MuLLT();

                    }
                }
            }

            if (isActiveCategory("rec.sel.lid"))
            {

                //  Get slice level PID from LIDBuilder
      
                      if(foLID.isValid())
                      {
                        cet::maybe_ref<art::Ptr<slid::EventLID> const> rlid(foLID.at(subevtId));
                        art::Ptr<slid::EventLID> elid = rlid.ref();
                        //if the pointer is valid
                        if (elid){
                              output["rec.sel.lid.ann"]    = elid->Value();
                              output["rec.sel.lid.anne"]     = elid->ValueWithE();
                              
                        if(fmShwLID.isValid())
                              {      
                                    std::vector< art::Ptr<slid::ShowerLID> > shwlids = fmShwLID.at(subevtId);
  
                                      output["rec.sel.lid.shwlid.ann"]      = shwlids[0]->Value();
                                      output["rec.sel.lid.shwlid.anne"]     = shwlids[0]->ValueWithE();
                                      output["rec.sel.lid.shwlid.ismuon"]   = shwlids[0]->IsMuon();
                                      output["rec.sel.lid.shwlid.eglll"]    = shwlids[0]->EGLLL();
                                      output["rec.sel.lid.shwlid.egllt"]    = shwlids[0]->EGLLT();
                                      output["rec.sel.lid.shwlid.emulll"]   = shwlids[0]->EMuLLL();
                                      output["rec.sel.lid.shwlid.emullt"]   = shwlids[0]->EMuLLT();
                                      output["rec.sel.lid.shwlid.epi0lll"]  = shwlids[0]->EPi0LLL();
                                      output["rec.sel.lid.shwlid.epi0llt"]  = shwlids[0]->EPi0LLT();
                                      output["rec.sel.lid.shwlid.eplll"]    = shwlids[0]->EPLLL();
                                      output["rec.sel.lid.shwlid.epllt"] = shwlids[0]->EPLLT();
                                      output["rec.sel.lid.shwlid.enlll"]  = shwlids[0]->ENLLL();
                                      output["rec.sel.lid.shwlid.enllt"]  = shwlids[0]->ENLLT();
                                      output["rec.sel.lid.shwlid.epilll"] = shwlids[0]->EPiLLL();
                                      output["rec.sel.lid.shwlid.epillt"] = shwlids[0]->EPiLLT();
                                      output["rec.sel.lid.shwlid.vtxgev"]   = shwlids[0]->VertexEnergy();
                                      output["rec.sel.lid.shwlid.pi0mass"]  = shwlids[0]->Pi0mass();
                                      output["rec.sel.lid.shwlid.shwEFrac"] = shwlids[0]->ShowerEFrac();
                                      output["rec.sel.lid.shwlid.gap"]      = shwlids[0]->Gap();
                                      output["rec.sel.lid.shwlid.costheta"] = shwlids[0]->CosTheta();
                            output["rec.sel.lid.shwlid.invglll"]  = shwlids[0]->InverseGLLL();
                        }
                    }
                }
            }

            if (isActiveCategory("rec.sel.lem"))
            {
                // Get the LEM objects and add them.
                if (fmLem.isValid())
                {
                    std::vector< art::Ptr<lem::PIDDetails> > lems = fmLem.at(subevtId);
                    if (lems.size() > 0)
                    {
                        output["rec.sel.lem.pid"] = lems[0]->Value();
                        output["rec.sel.lem.pidfit"] = lems[0]->vars.fPIDFit;
                        output["rec.sel.lem.pidexp"] = lems[0]->vars.fPIDExp;
                        output["rec.sel.lem.pidexpfit"] = lems[0]->vars.fPIDExpFit;
                        output["rec.sel.lem.meany"] = lems[0]->vars.fMeanY;
                        output["rec.sel.lem.meanyfit"] = lems[0]->vars.fMeanYFit;
                        output["rec.sel.lem.meanyexp"] = lems[0]->vars.fMeanYExp;
                        output["rec.sel.lem.meanysig"] = lems[0]->vars.fMeanYSig;
                        output["rec.sel.lem.meanysigexp"] = lems[0]->vars.fMeanYSigExp;
                        output["rec.sel.lem.meanqfrac"] = lems[0]->vars.fMeanQFrac;
                        output["rec.sel.lem.meanqfracfit"] = lems[0]->vars.fMeanQFracFit;
                        output["rec.sel.lem.meanqfracexp"] = lems[0]->vars.fMeanQFracExp;
                        output["rec.sel.lem.meanqfracsig"] = lems[0]->vars.fMeanQFracSig;
                        output["rec.sel.lem.meanqfracsigexp"] = lems[0]->vars.fMeanQFracSigExp;
                        output["rec.sel.lem.meanqfracbkg"] = lems[0]->vars.fMeanQFracBkg;
                        output["rec.sel.lem.meanqfracbkgexp"] = lems[0]->vars.fMeanQFracBkgExp;
                        output["rec.sel.lem.energydiff"] = lems[0]->vars.fEnergyDiff;
                        output["rec.sel.lem.energydiffexp"] = lems[0]->vars.fEnergyDiffExp;
                        output["rec.sel.lem.enrichfrac"] = lems[0]->vars.fEnrichFrac;
                        output["rec.sel.lem.enrichfracexp"] = lems[0]->vars.fEnrichFracExp;
                        output["rec.sel.lem.fannid"] = lems[0]->vars.fFannId;
                        output["rec.sel.lem.fannidenrich"] = lems[0]->vars.fFannIdEnrich;
                        output["rec.sel.lem.chisig"] = lems[0]->vars.fChiSig;
                        output["rec.sel.lem.chibkg"] = lems[0]->vars.fChiBkg;
                        output["rec.sel.lem.chisigexp"] = lems[0]->vars.fChiSigExp;
                        output["rec.sel.lem.chibkgexp"] = lems[0]->vars.fChiBkgExp;
                        output["rec.sel.lem.avginvE"] = lems[0]->vars.fAvgInvE;
                        output["rec.sel.lem.avgexpE"] = lems[0]->vars.fAvgExpE;
                        output["rec.sel.lem.avgsigE"] = lems[0]->vars.fAvgSigE;
                        output["rec.sel.lem.tmvabdtg"] = lems[0]->vars.fTMVABDTG;
                        output["rec.sel.lem.tmvabdt"] = lems[0]->vars.fTMVABDT;
                        output["rec.sel.lem.tmvabdtd"] = lems[0]->vars.fTMVABDTD;
                        output["rec.sel.lem.tmvabdtmf"] = lems[0]->vars.fTMVABDTMF;
                        output["rec.sel.lem.tmvamlp"] = lems[0]->vars.fTMVAMLP;
                        output["rec.sel.lem.dectree"] = lems[0]->vars.fDecTreePID;
                    }
                }
            }
            if (isActiveCategory("rec.sel.nuepre"))
            {
                //#######################################################
                // Add Nue Presel info from slice and reconstructed objects
                //#######################################################

                //Nue presel vars:
                SliceInfo slcInfo = getSliceInfo(*subevts[subevtId]);
                output["rec.sel.nuepre.calE"] = subevts[subevtId]->TotalGeV();
                output["rec.sel.nuepre.ncell"] = slcInfo.nCalHit;
                output["rec.sel.nuepre.fracmip"] = (slcInfo.nCalHit > 0) ? slcInfo.nMipHit / slcInfo.nCalHit : -1;

                if (fmDiscrete.isValid() && fmDiscrete.at(subevtId).size() > 0)
                {
                    output["rec->sel.nuepre.lengthdiscretemerge"] = fmDiscrete.at(subevtId).at(0)->TotalLength();
                }

                if (fmDiscrete2d.isValid() && fmDiscrete2d.at(subevtId).size() > 0)
                {
                    output["rec->sel.nuepre.lengthdiscrete2d"] = fmDiscrete2d.at(subevtId).at(0)->TotalLength();
                }

                if (fmKalman.isValid() && fmKalman.at(subevtId).size() > 0)
                {
                    output["rec->sel.nuepre.lengthkalman"] = fmKalman.at(subevtId).at(0)->TotalLength();
                }

                //Fiducial and containment vars:
                output["rec.sel.nuepre.cellstoedge"] = slcInfo.minCellsFromEdge;
                output["rec.sel.nuepre.planestofront"] = slcInfo.minPlane;
                output["rec.sel.nuepre.planestoback"] = geom->NPlanes() - 1 - slcInfo.maxPlane;
                output["rec.sel.nuepre.mean"] = convertTVector3ToDouble(subevts[subevtId]->MeanXYZ());
                if (fmElastic.isValid())
                {
                    if (fmElastic.at(subevtId).size() > 0)
                    {
                        output["rec.sel.nuepre.vtx"] = convertTVector3ToDouble(fmElastic.at(subevtId).at(0)->GetXYZ());
                    }
                    else
                    {
                        output["rec.sel.nuepre.vtx"] = convertTVector3ToDouble(TVector3(-99999, -99999, -99999));
                    }
                }

                //Booleans describing preselection:

                //Current fiducial cut:
                output["rec.sel.nuepre.passfid"] =
                    geom->isInsideFiducialVolume(subevts[subevtId]->MeanXYZ().X(), subevts[subevtId]->MeanXYZ().Y(), subevts[subevtId]->MeanXYZ().Z());

                //Current preselection cut:
                if (slcInfo.nCalHit > 35 && slcInfo.nCalHit <= 135
                    && subevts[subevtId]->TotalGeV() > 0.50 && subevts[subevtId]->TotalGeV() <= 1.50
                    && output["rec.sel.nuepre.lengthdiscrete2d"] > 45 && output["rec.sel.nuepre.lengthdiscrete2d"] <= 385
                    && output["rec.sel.nuepre.fracmip"] > 0.15 && output["rec.sel.nuepre.fracmip"] <= 0.65
                    ) {
                    output["rec.sel.nuepre.passpre"] = true;
                }
                else {
                    output["rec.sel.nuepre.passpre"] = false;
                }

                //####End Presel Fill####//
            }
            if (isActiveCategory("rec.sel.qepid"))
            {
                if (fmQepid.isValid())
                {
                    std::vector< art::Ptr<qeef::QePId> > qepids = fmQepid.at(subevtId);
                    if (qepids.size() > 0)
                    {
                        output["rec.sel.qepid.pdg"] = qepids[0]->Pdg();
                        output["rec.sel.qepid.mode"] = qepids[0]->Mode();
                        output["rec.sel.qepid.ntrk"] = qepids[0]->Ntrk();
                        output["rec.sel.qepid.pid"] = qepids[0]->Value();
                        output["rec.sel.qepid.offE"] = qepids[0]->OffTrkE();
                        output["rec.sel.qepid.ediff"] = qepids[0]->Ediff();
                        output["rec.sel.qepid.ediffz"] = qepids[0]->EdiffZ();
                        output["rec.sel.qepid.dedx"] = qepids[0]->Dedx();
                    }
                }
            }
            if (isActiveCategory("rec.sel.remid"))
            {
                // now do the same for kalmanTrack
                if (fmKalman.isValid())
                {
                    std::vector< art::Ptr<rb::Track> > kalTracks = fmKalman.at(subevtId);
                    // Sort the tracks in descending length order 
                    std::sort(kalTracks.begin(), kalTracks.end(), sortRBTrackLength);
                    art::FindManyP<remid::ReMId> fmRemid(kalTracks, evt, remidName);
                    art::FindManyP<rb::Energy> fmMuTrkE(kalTracks, evt, numuEnergyName);
                    art::FindManyP<me::MichelECluster> fmKalmanMichelE(kalTracks, evt, kalmanMichelEName);

                    size_t bestPidIdx = 0;
                    size_t bestPidIdx3d = 0;
                    double bestPid = 0;
                    double bestPid3d = 0;
                    int nMuid = 0;
                    for (size_t trackId = 0; trackId < kalTracks.size(); ++trackId)
                    {
                        if (fmRemid.isValid())
                        {
                            //Get MuID from ReMId
                            std::vector< art::Ptr<remid::ReMId> > remids = fmRemid.at(trackId);

                            if (remids.size() > 0)
                            {
                                nMuid++;
                                float measfrac = 0;
                                if (remids[0]->NDedx() > 0) {
                                    measfrac = ((float)remids[0]->NMeasurementPlanes() / (float)remids[0]->NDedx());
                                }
                                
                                if (remids[0]->Value() > bestPid)
                                {
                                    bestPidIdx = trackId;
                                    bestPid = remids[0]->Value();

                                    output["rec.sel.remid.pdg"] = remids[0]->Pdg();
                                    output["rec.sel.remid.pid"] = remids[0]->Value();
                                    output["rec.sel.remid.scatllh"] = remids[0]->ScatSeparation();
                                    output["rec.sel.remid.dedxllh"] = remids[0]->DedxSeparation();
                                    output["rec.sel.remid.len"] = kalTracks[trackId]->TotalLength();
                                    output["rec.sel.remid.measfrac"] = measfrac;
                                    output["rec.sel.remid.cont"] = remids[0]->Contained();
                                }
                                if (remids[0]->Value() > bestPid3d && kalTracks[trackId]->View() == geo::kXorY)
                                {
                                    bestPidIdx3d = trackId;
                                    bestPid3d = remids[0]->Value();
                                }
                            }
                        }
                    }

                    //Give SRRemid the pid values of track with largest PID 
                    if (nMuid > 0)
                    {
                        output["rec.sel.remid.bestidx"] = bestPidIdx;
                        output["rec.sel.remid.bestidx3d"] = bestPidIdx3d;
                        output["rec.sel.remid.nmuid"] = nMuid;
                    }
                }
            }
            if (isActiveCategory("rec.sel.rvp"))
            {
                if (fmRvp.isValid())
                {
                    std::vector< art::Ptr<rvp::RVP> > rvps = fmRvp.at(subevtId);
                    if (rvps.size() > 0)
                    {
                        output["rec.sel.rvp.pid"] = rvps[0]->Value();
                        output["rec.sel.rvp.ncell"] = rvps[0]->NCells();
                        output["rec.sel.rvp.recoE"] = rvps[0]->RecoE();
                        output["rec.sel.rvp.longtr"] = rvps[0]->LongTr();
                        output["rec.sel.rvp.longtrfrac"] = rvps[0]->LongTrFrac();
                        output["rec.sel.rvp.mipfrac"] = rvps[0]->MipFrac();
                        output["rec.sel.rvp.miphits"] = rvps[0]->MipHits();
                        output["rec.sel.rvp.epl20frac"] = rvps[0]->EPl20Frac();
                        output["rec.sel.rvp.efrac2plwin"] = rvps[0]->EFrac2PlWin();
                        output["rec.sel.rvp.efrac6plwin"] = rvps[0]->EFrac6PlWin();
                        output["rec.sel.rvp.efrac2sigrd"] = rvps[0]->EFrac2SigRd();
                        output["rec.sel.rvp.prongs3D"] = rvps[0]->Prongs3D();
                        output["rec.sel.rvp.prongEbal3D"] = rvps[0]->ProngEBal3D();
                        output["rec.sel.rvp.prongs2D"] = rvps[0]->Prongs2D();
                        output["rec.sel.rvp.prongEbal2D"] = rvps[0]->ProngEBal2D();
                        output["rec.sel.rvp.eiso3sig"] = rvps[0]->EIso3Sig();
                        output["rec.sel.rvp.rvp12"] = rvps[0]->RVP12();
                    }
                }
            }
            if (isActiveCategory("rec.shw"))
            {
                if (fmJmShower.isValid()) output["rec.shw.njm"] = fmJmShower.at(subevtId).size();

                if(foLID.isValid())
                {
                        cet::maybe_ref<art::Ptr<slid::EventLID> const> rlid(foLID.at(subevtId));
                        art::Ptr<slid::EventLID> elid = rlid.ref();
                        //if the pointer is valid
                        if (elid){
                              if(fmShwLID.isValid())
                              {      
                                    std::vector< art::Ptr<slid::ShowerLID> > shwlids = fmShwLID.at(subevtId);
                                art::FindOneP<rb::Shower> fos(shwlids, evt, lidName);
                                cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(0));
                                    art::Ptr<rb::Shower> shw = rshw.ref();
                                if (shw){            
                                        output["rec.shw.shwlid.nhit"]         = shw->NCell();
                                        output["rec.shw.shwlid.nhitx"]        = shw->NXCell();
                                        output["rec.shw.shwlid.nhity"]        = shw->NYCell();
                                        output["rec.shw.shwlid.nplane"]       = shw->ExtentPlane();
                                        output["rec.shw.shwlid.nplanex"]      = shw->ExtentPlane(geo::kX);
                                        output["rec.shw.shwlid.nplaney"]      = shw->ExtentPlane(geo::kY);
                                        output["rec.shw.shwlid.gap"]          = shwlids[0]->Gap();
                                        output["rec.shw.shwlid.maxplanecont"] = shw->MostContiguousPlanes(geo::kXorY);
                                        output["rec.shw.shwlid.maxplanegap"]  = shw->MostMissingPlanes(geo::kXorY);
                                        output["rec.shw.shwlid.len"]          = shw->TotalLength();
                                        output["rec.shw.shwlid.width"]        = shwlids[0]->Radius();
                                        output["rec.shw.shwlid.calE"]         = shw->CalorimetricEnergy();
                                        if      (shw->NXCell() == 0)  output["rec.shw.shwlid.view"] = geo::kY;
                                        else if (shw->NYCell() == 0)  output["rec.shw.shwlid.view"] = geo::kX;
                                        else                                     output["rec.shw.shwlid.view"] = geo::kXorY;
                                  }
                        }
                    }
                }
            }
            if (isActiveCategory("rec.trk"))
            {
                if (fmCosmic.isValid()) output["rec.trk.ncosmic"] = fmCosmic.at(subevtId).size();
                if (fmDiscrete.isValid()) output["rec.trk.ndiscrete"] = fmDiscrete.at(subevtId).size();
                if (fmDiscrete2d.isValid()) output["rec.trk.ndiscrete2d"] = fmDiscrete2d.at(subevtId).size();
                if (fmKalman.isValid()) output["rec.trk.nkalman"] = fmKalman.at(subevtId).size();
            }
            if (isActiveCategory("rec.vtx"))
            {
                if (isActiveCategory("rec.vtx.elecid"))
                {
                    if (fmNueSel.isValid())
                    {
                        std::vector< art::Ptr<jmshower::EID> > eids = fmNueSel.at(subevtId);
                        if (eids.size() > 0)
                        {
                            output["rec.vtx.elecid.vtx"] = convertTVector3ToDouble(eids[0]->Vtx());
                            output["rec.vtx.elecid.time"] = subevts[subevtId]->MeanTNS();
                        }
                    }
                }
                if (isActiveCategory("rec.vtx.nelastic"))
                {
                    if (fmElastic.isValid())
                    {
                        output["rec.vtx.nelastic"] = fmElastic.at(subevtId).size();
                    }
                }
            }

            std::vector<double> vals(exprVars.size(), 0.);
            bool goodVals = true;

            std::stringstream evtString;
                        evtString << "Run " << fRun << ", Subrun " << fSubrun << ", Event " << fEvt << ", Slice " << subevtId << ": ";

            for (unsigned int i = 0; i < exprVars.size(); i++)
            {
                if (std::isinf(vals[i]))
                {
                    goodVals = false;
                    break;
                }
                vals[i] = output.at(exprVars[i]);
                if(i > 0)
                                {
                                  evtString << ", ";
                                }
                evtString << exprVars[i] << ": " << vals[i];
            }

            if (goodVals && fFormula->EvalPar(0, vals.data()))
            {
                std::cerr << evtString.str() << std::endl;;
                return true;
            }
        }

                std::cerr << "Run " << fRun << ", Subrun " << fSubrun << ", Event " << fEvt << ": No matching slices" << std::endl;

        return false;
    }

    DEFINE_ART_MODULE(CAFFilter);

} // namespace
////////////////////////////////////////////////////////////////////////


