///////////////////////////////////////////////////////////////////////////
/// \brief   Validate anything that creates HoughResults
/// \author  mibaird@indiana.edu
/// \date
//////////////////////////////////////////////////////////////////////////

#include "TTree.h"
#include "TH1F.h"

#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h" 
#include "art/Framework/Core/ModuleMacros.h" 
#include "art/Framework/Services/Optional/TFileService.h" 	
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h" 
#include "messagefacility/MessageLogger/MessageLogger.h" 

#include "GeometryObjects/Geo.h"
#include "HoughTransform/HValidateNt.h"
#include "RecoBase/HoughResult.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "Utilities/func/MathUtil.h"





// HoughValidate header
namespace hough {
  ///
  /// \brief Validate Hough transform algorithms
  ///
  class HoughValidate : public art::EDAnalyzer {

  public:
    explicit HoughValidate(fhicl::ParameterSet const &p);
    virtual ~HoughValidate();
    
    void      beginJob();
    void      analyze(const art::Event &e);
    void      reconfigure(const fhicl::ParameterSet& p);

  private:

    TTree*        fNtTree;
    HValidateNt*  fNt;

    std::string   fMCLabel;      ///< label for process that made MCTruth
    std::string   fHoughLabel;   ///< label for process that made HoughResult
    int           fIsMC;         ///< flag for is MC or not

    // Histograms used for production validation
    TH1F* fRho1x;
    TH1F* fRho2x;
    TH1F* fRho3x;
    TH1F* fRho1y;
    TH1F* fRho2y;
    TH1F* fRho3y;

    TH1F* fTheta1x;
    TH1F* fTheta2x;
    TH1F* fTheta3x;
    TH1F* fTheta1y;
    TH1F* fTheta2y;
    TH1F* fTheta3y;

    TH1F* fNumHLx;
    TH1F* fNumHLy;

    TH1F* fThreshx;
    TH1F* fThreshy;

    TH1F* fPeak1x;
    TH1F* fPeakThresh1x;
    TH1F* fPeak2x;
    TH1F* fPeakThresh2x;
    TH1F* fPeak3x;
    TH1F* fPeakThresh3x;
    TH1F* fPeak1y;
    TH1F* fPeakThresh1y;
    TH1F* fPeak2y;
    TH1F* fPeakThresh2y;
    TH1F* fPeak3y;
    TH1F* fPeakThresh3y;


    TH1F* fPerpD1x;
    TH1F* fPerpD2x;
    TH1F* fPerpD3x;
    TH1F* fPerpD1y;
    TH1F* fPerpD2y;
    TH1F* fPerpD3y;
  };
}



namespace hough
{
  HoughValidate::HoughValidate(fhicl::ParameterSet const &p)
    : EDAnalyzer(p)
  {
    this->reconfigure(p);
  }

  //......................................................................
  


  void HoughValidate::beginJob() 
  {      

      art::ServiceHandle<art::TFileService> tfs;

      // book Ntuple
      fNt = new HValidateNt;
      fNtTree = tfs->make<TTree>("hough", "Hough Analysis Tree");
      fNtTree->Branch("fHValidateNt", "hough::HValidateNt", &fNt, 1600, 99);
      
      // book histos
      fRho1x = tfs->make<TH1F>("fRho1x",
			       "Rho 1st line x-view;rho [cm];count",
			       200,-400.0,400.0);
      fRho2x = tfs->make<TH1F>("fRho2x",
			       "Rho 2nd line x-view;rho [cm];count",
			       200,-400.0,400.0);
      fRho3x = tfs->make<TH1F>("fRho3x",
			       "Rho 3rd line x-view;rho [cm];count",
			       200,-400.0,400.0);
      fRho1y = tfs->make<TH1F>("fRho1y",
			       "Rho 1st line y-view;rho [cm];count",
			       200,-400.0,400.0);
      fRho2y = tfs->make<TH1F>("fRho2y",
			       "Rho 2nd line y-view;rho [cm];count",
			       200,-400.0,400.0);
      fRho3y = tfs->make<TH1F>("fRho3y",
			       "Rho 3rd line y-view;rho [cm];count",
			       200,-400.0,400.0);

      fTheta1x = tfs->make<TH1F>("fTheta1x",
				 "Theta 1st line x-view;theta [rad];count",
				 40,0.0,4.0);
      fTheta2x = tfs->make<TH1F>("fTheta2x",
				 "Theta 2nd line x-view;theta [rad];count",
				 40,0.0,4.0);
      fTheta3x = tfs->make<TH1F>("fTheta3x",
				 "Theta 3rd line x-view;theta [rad];count",
				 40,0.0,4.0);
      fTheta1y = tfs->make<TH1F>("fTheta1y",
				 "Theta 1st line y-view;theta [rad];count",
				 40,0.0,4.0);
      fTheta2y = tfs->make<TH1F>("fTheta2y",
				 "Theta 2nd line y-view;theta [rad];count",
				 40,0.0,4.0);
      fTheta3y = tfs->make<TH1F>("fTheta3y",
				 "Theta 3rd line y-view;theta [rad];count",
				 40,0.0,4.0);

      fNumHLx = tfs->make<TH1F>("fNumHLx",
				"Num HL x-view;# of lines;count",
				11,-0.5,10.5);
      fNumHLy = tfs->make<TH1F>("fNumHLy",
				"Num HL y-view;# of lines;count",
				11,-0.5,10.5);

      fThreshx = tfs->make<TH1F>("fThreshx",
				 "Threshold value for Hough space map (x-view);threshold (arbitrary units);count",
				 100,0.0,5.0);
      fThreshy = tfs->make<TH1F>("fThreshy",
				 "Threshold value for Hough space map (y-view);threshold (arbitrary units);count",
				 100,0.0,5.0);

      fPeak1x = tfs->make<TH1F>("fPeak1x",
				"Peak value in Hough space map for 1st line (x-view);peak (arbitrary units);count",
				200,0.0,2000.0);
      fPeakThresh1x = tfs->make<TH1F>("fPeakThresh1x",
				      "Peak over Threshold 1st line (x-view);peak/threshold;count",
				      200,0.0,10000.0);
      fPeak2x = tfs->make<TH1F>("fPeak2x",
				"Peak value in Hough space map for 2nd line (x-view);peak (arbitrary units);count",
				200,0.0,2000.0);
      fPeakThresh2x = tfs->make<TH1F>("fPeakThresh2x",
				      "Peak over Threshold 2nd line (x-view);peak/threshold;count",
				      200,0.0,10000.0);
      fPeak3x = tfs->make<TH1F>("fPeak3x",
				"Peak value in Hough space map for 3rd line (x-view);peak (arbitrary units);count",
				200,0.0,2000.0);
      fPeakThresh3x = tfs->make<TH1F>("fPeakThresh3x",
				      "Peak over Threshold 3rd line (x-view);peak/threshold;count",
				      200,0.0,10000.0);
      fPeak1y = tfs->make<TH1F>("fPeak1y",
				"Peak value in Hough space map for 1st line (y-view);peak (arbitrary units);count",
				200,0.0,2000.0);
      fPeakThresh1y = tfs->make<TH1F>("fPeakThresh1y",
				      "Peak over Threshold 1st line (y-view);peak/threshold;count",
				      200,0.0,10000.0);
      fPeak2y = tfs->make<TH1F>("fPeak2y",
				"Peak value in Hough space map for 2nd line (y-view);peak (arbitrary units);count",
				200,0.0,2000.0);
      fPeakThresh2y = tfs->make<TH1F>("fPeakThresh2y",
				      "Peak over Threshold 2nd line (y-view);peak/threshold;count",
				      200,0.0,10000.0);
      fPeak3y = tfs->make<TH1F>("fPeak3y",
				"Peak value in Hough space map for 3rd line (y-view);peak (arbitrary units);count",
				200,0.0,2000.0);
      fPeakThresh3y = tfs->make<TH1F>("fPeakThresh3y",
				      "Peak over Threshold 3rd line (y-view);peak/threshold;count",
				      200,0.0,10000.0);

      fPerpD1x = tfs->make<TH1F>("fPerpD1x",
				 "perpendicular dist from true vertex to 1st line x-view;dist [cm];count",
				 100,0.0,100.0);
      fPerpD2x = tfs->make<TH1F>("fPerpD2x",
				 "perpendicular dist from true vertex to 2nd line x-view;dist [cm];count",
				 100,0.0,100.0);
      fPerpD3x = tfs->make<TH1F>("fPerpD3x",
				 "perpendicular dist from true vertex to 3rd line x-view;dist [cm];count",
				 100,0.0,100.0);
      fPerpD1y = tfs->make<TH1F>("fPerpD1y",
				 "perpendicular dist from true vertex to 1st line y-view;dist [cm];count",
				 100,0.0,100.0);
      fPerpD2y = tfs->make<TH1F>("fPerpD2y",
				 "perpendicular dist from true vertex to 2nd line y-view;dist [cm];count",
				 100,0.0,100.0);
      fPerpD3y = tfs->make<TH1F>("fPerpD3y",
				 "perpendicular dist from true vertex to 3rd line y-view;dist [cm];count",
				 100,0.0,100.0);
  }

  //......................................................................
  


  void HoughValidate::reconfigure(const fhicl::ParameterSet& p) 
  {

    fMCLabel    = p.get<std::string> ("MCLabel");
    fHoughLabel = p.get<std::string> ("HoughLabel");
    fIsMC       = p.get<int>         ("IsMC"); 

  }
  
  //......................................................................
  


  HoughValidate::~HoughValidate() { }
  
  //......................................................................
  


  void HoughValidate::analyze(const art::Event& evt) 
  {

    // get the Hough Results
    art::Handle< std::vector< rb::HoughResult > > hr;
    evt.getByLabel(fHoughLabel, hr);

    // get the MCTruth info
    art::Handle< std::vector< simb::MCTruth > > mclist;
    if(fIsMC == 1) evt.getByLabel(fMCLabel, mclist);

    /*
    // NOTE: I was using this for the MC dot product validation, but I'm
    // temporarily skipping that since sim::ParticleList is no longer being used
    
    // get the particle list for trajectory lengths
    art::Handle< std::vector<sim::ParticleList> > particlelist;
    evt.getByLabel("geantgen", particlelist);
    */
   


    // load event ID info into Ntuple
    fNt->run    = evt.run();
    fNt->subrun = evt.subRun();
    fNt->evt    = evt.id().event();



    // a set of temp vectors to fill the Ntuple
    std::vector<int> numHLXtemp;
    std::vector<int> numHLYtemp;
    
    std::vector<double> nuXtemp;
    std::vector<double> nuYtemp;
    std::vector<double> nuZtemp;
    std::vector<double> nuEtemp;
    
    std::vector<int> nupdgtemp;
    std::vector<int> nuCCNCtemp;

    std::vector<double> perpdist1xtemp;
    std::vector<double> perpdist1ytemp;
    std::vector<double> perpdist2xtemp;
    std::vector<double> perpdist2ytemp;
    std::vector<double> perpdist3xtemp;
    std::vector<double> perpdist3ytemp;

    std::vector<double> mcdp1xtemp;
    std::vector<double> mcdp1ytemp;
    std::vector<double> mcdp2xtemp;
    std::vector<double> mcdp2ytemp;
    std::vector<double> mcdp3xtemp;
    std::vector<double> mcdp3ytemp;

    std::vector<double> rho1xtemp;
    std::vector<double> rho2xtemp;
    std::vector<double> rho3xtemp;
    std::vector<double> rho1ytemp;
    std::vector<double> rho2ytemp;
    std::vector<double> rho3ytemp;
    std::vector<double> theta1xtemp;
    std::vector<double> theta2xtemp;
    std::vector<double> theta3xtemp;
    std::vector<double> theta1ytemp;
    std::vector<double> theta2ytemp;
    std::vector<double> theta3ytemp;

    std::vector<double> threshxtemp;
    std::vector<double> threshytemp;
    std::vector<double> peak1xtemp;
    std::vector<double> peak2xtemp;
    std::vector<double> peak3xtemp;
    std::vector<double> peak1ytemp;
    std::vector<double> peak2ytemp;
    std::vector<double> peak3ytemp;
    
    // This validation information is gathered for both data and MC
    for(unsigned int k = 0; k < hr->size(); ++k) {
      if((*hr)[k].fView == geo::kX) {

	if((*hr)[k].fPeak.size() >= 1) {
	  numHLXtemp.push_back((*hr)[k].fPeak.size());
	  fNumHLx->Fill((*hr)[k].fPeak.size());
	  fThreshx->Fill((*hr)[k].fPeak[0].fThr);
	  threshxtemp.push_back((*hr)[k].fPeak[0].fThr);

	  fRho1x->Fill((*hr)[k].fPeak[0].fRho);
	  fTheta1x->Fill((*hr)[k].fPeak[0].fTheta);
	  fPeak1x->Fill((*hr)[k].fPeak[0].fH);
	  fPeakThresh1x->Fill(((*hr)[k].fPeak[0].fH)/((*hr)[k].fPeak[0].fThr));

	  rho1xtemp.push_back((*hr)[k].fPeak[0].fRho);
	  theta1xtemp.push_back((*hr)[k].fPeak[0].fTheta);
	  peak1xtemp.push_back((*hr)[k].fPeak[0].fH);
	}

	if((*hr)[k].fPeak.size() >= 2) {
	  fRho2x->Fill((*hr)[k].fPeak[1].fRho);
	  fTheta2x->Fill((*hr)[k].fPeak[1].fTheta);
	  fPeak2x->Fill((*hr)[k].fPeak[1].fH);
	  fPeakThresh2x->Fill(((*hr)[k].fPeak[1].fH)/((*hr)[k].fPeak[1].fThr));

	  rho2xtemp.push_back((*hr)[k].fPeak[1].fRho);
	  theta2xtemp.push_back((*hr)[k].fPeak[1].fTheta);
	  peak2xtemp.push_back((*hr)[k].fPeak[1].fH);
	}
	if((*hr)[k].fPeak.size() >= 3) {
	  fRho3x->Fill((*hr)[k].fPeak[2].fRho);
	  fTheta3x->Fill((*hr)[k].fPeak[2].fTheta);
	  fPeak3x->Fill((*hr)[k].fPeak[2].fH);
	  fPeakThresh3x->Fill(((*hr)[k].fPeak[2].fH)/((*hr)[k].fPeak[2].fThr));

	  rho3xtemp.push_back((*hr)[k].fPeak[2].fRho);
	  theta3xtemp.push_back((*hr)[k].fPeak[2].fTheta);
	  peak3xtemp.push_back((*hr)[k].fPeak[2].fH);
	}

      }

      if((*hr)[k].fView == geo::kY) {

	if((*hr)[k].fPeak.size() >= 1) {
	  numHLYtemp.push_back((*hr)[k].fPeak.size());
	  fNumHLy->Fill((*hr)[k].fPeak.size());
	  fThreshy->Fill((*hr)[k].fPeak[0].fThr);
	  threshytemp.push_back((*hr)[k].fPeak[0].fThr);

	  fRho1y->Fill((*hr)[k].fPeak[0].fRho);
	  fTheta1y->Fill((*hr)[k].fPeak[0].fTheta);
	  fPeak1y->Fill((*hr)[k].fPeak[0].fH);
	  fPeakThresh1y->Fill(((*hr)[k].fPeak[0].fH)/((*hr)[k].fPeak[0].fThr));

	  rho1ytemp.push_back((*hr)[k].fPeak[0].fRho);
	  theta1ytemp.push_back((*hr)[k].fPeak[0].fTheta);
	  peak1ytemp.push_back((*hr)[k].fPeak[0].fH);
	}
	if((*hr)[k].fPeak.size() >= 2) {
	  fRho2y->Fill((*hr)[k].fPeak[1].fRho);
	  fTheta2y->Fill((*hr)[k].fPeak[1].fTheta);
	  fPeak2y->Fill((*hr)[k].fPeak[1].fH);
	  fPeakThresh2y->Fill(((*hr)[k].fPeak[1].fH)/((*hr)[k].fPeak[1].fThr));

	  rho2ytemp.push_back((*hr)[k].fPeak[1].fRho);
	  theta2ytemp.push_back((*hr)[k].fPeak[1].fTheta);
	  peak2ytemp.push_back((*hr)[k].fPeak[1].fH);
	}
	if((*hr)[k].fPeak.size() >= 3) {
	  fRho3y->Fill((*hr)[k].fPeak[2].fRho);
	  fTheta3y->Fill((*hr)[k].fPeak[2].fTheta);
	  fPeak3y->Fill((*hr)[k].fPeak[2].fH);
	  fPeakThresh3y->Fill(((*hr)[k].fPeak[2].fH)/((*hr)[k].fPeak[2].fThr));

	  rho3ytemp.push_back((*hr)[k].fPeak[2].fRho);
	  theta3ytemp.push_back((*hr)[k].fPeak[2].fTheta);
	  peak3ytemp.push_back((*hr)[k].fPeak[2].fH);
	}

      }

    }





    // additional validation info gathered for MC files
    if(fIsMC == 1) {

      // validate perpendicular distance to top three lines
      if(hr->size() > 0) {
	
	for(unsigned int i = 0; i < mclist->size(); i++) {

	  simb::MCNeutrino nu = (*mclist)[i].GetNeutrino();
	  simb::MCParticle Nu = nu.Nu();

	  // BEWARE OF SEGFAULTS!!!  If this MCTruth did NOT come from a neutrino,
	  // then we must skip the next bit to avoid segfaluts.
	  if(!(Nu.PdgCode() == 12 || Nu.PdgCode() == 14 || Nu.PdgCode() == 16)) continue;
      
	  double xt = Nu.Vx();
	  double yt = Nu.Vy();
	  double zt = Nu.Vz();

	  nuXtemp.push_back(xt);
	  nuYtemp.push_back(yt);
	  nuZtemp.push_back(zt);
	  nuEtemp.push_back(Nu.E());
	  nupdgtemp.push_back(Nu.PdgCode());
	  nuCCNCtemp.push_back(nu.CCNC());



	  for(unsigned int k = 0; k < hr->size(); ++k) {
	    if((*hr)[k].fPeak.size() != 0) {
	
	      // variables used in validation
	      double m1 = 0.0, m2 = 0.0, m3 = 0.0;
	      double b1 = 0.0, b2 = 0.0, b3 = 0.0;
	      double distx = 0.0, disty = 0.0;



	      // x-view validation
	      if((*hr)[k].fView == geo::kX) {

		// perpendicular distance to main line validation
		(*hr)[k].SlopeIcept(0, &m1, &b1);
		distx = sqrt(geo::DsqrToLine(zt, xt, 0.0, b1, 100.0, 100.0*m1 + b1));
		perpdist1xtemp.push_back(distx);
		fPerpD1x->Fill(distx);

		// perpendicular distance to second line validation
		if((*hr)[k].fPeak.size() >= 2) {
		  (*hr)[k].SlopeIcept(1, &m2, &b2);
		  distx = sqrt(geo::DsqrToLine(zt, xt, 0.0, b2, 100.0, 100.0*m2 + b2));
		  perpdist2xtemp.push_back(distx);
		  fPerpD2x->Fill(distx);
		} // end if peak size >= 2

		// perpendicular distance to third line validation
		if((*hr)[k].fPeak.size() >= 3) {
		  (*hr)[k].SlopeIcept(1, &m3, &b3);
		  distx = sqrt(geo::DsqrToLine(zt, xt, 0.0, b3, 100.0, 100.0*m3 + b3));
		  perpdist3xtemp.push_back(distx);
		  fPerpD3x->Fill(distx);
		} // end if peak size >= 3
	      } // end if fView == geo::kX



	      // y-view validation
	      if((*hr)[k].fView == geo::kY) {

		// perpendicular distance to main line validation
		(*hr)[k].SlopeIcept(0, &m1, &b1);
		disty = sqrt(geo::DsqrToLine(zt, yt, 0.0, b1, 100.0, 100.0*m1 + b1));
		perpdist1ytemp.push_back(disty);
		fPerpD1y->Fill(disty);

		// perpendicular distance to second line validation
		if((*hr)[k].fPeak.size() >= 2) {
		  (*hr)[k].SlopeIcept(1, &m2, &b2);
		  disty = sqrt(geo::DsqrToLine(zt, yt, 0.0, b2, 100.0, 100.0*m2 + b2));
		  perpdist2ytemp.push_back(disty);
		  fPerpD2y->Fill(disty);
		} // end if peak size >= 2

		// perpendicular distance to third line validation
		if((*hr)[k].fPeak.size() >= 3) {
		  (*hr)[k].SlopeIcept(1, &m3, &b3);
		  disty = sqrt(geo::DsqrToLine(zt, yt, 0.0, b3, 100.0, 100.0*m3 + b3));
		  perpdist3ytemp.push_back(disty);
		  fPerpD3y->Fill(disty);
		} // end if peak size >= 3
	      } // end if fView == geo::kY

	    } // end if peak size != 0
	  } // end for k (loop over size of hough results)
	} // end for i (loop over size of MClist)
      } // end if hr->size > 0
    } // end if(fIsMC)
    




    // NOTE: I was using this for the MC dot product validation, but I'm
    // temporarily skipping that since sim::ParticleList is no longer being used
    //
    // If I decide to put this back in, it belongs in the if(hr->size() > 0) loop found above
    /*
    if(hr->size() > 0) {

      // MC dot product validation
      //
      // NOTE:  This could be a little more efficient if I had integrated this section into the other loops
      // (instead of looping over mclist two seperate times.)  BUT this already runs very quickly and it
      // was cleaner to code and debug this way...

      for(unsigned int k = 0; k < hr->size(); ++k) {
	if((*hr)[k].fPeak.size() != 0) {

	// variables used in MC-trajectory dot product validation
	double dp = -999.0;
	unsigned int track1 = 999, track2 = 999;

	double xy = 0.0, z = 0.0;
	double m = 0.0, b = 0.0;

	// x-view
	if((*hr)[k].fView == geo::kX) {
	
	  //std::cout << "\n\nevent ID:   " << evt.id() << "\n";

	  // loop over the number of MC reactions
	  for(unsigned int i = 0; i < particlelist->size(); ++i) {

	    art::Ptr<sim::ParticleList> plist(particlelist, i);
	    
	    // loop over all particles and find the one closest to first hough line   
	    (*hr)[k].SlopeIcept(0, &m, &b);
	    for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {
          
	      int pdg = (*plist).fParticles[c].PdgCode();

	      // don't let a hough line match a particle that wouldn't leave a signal in the detector
	      // (nu-e, nu-mu, nu-tau, neutron, pi0)



	      // ADD ANTI-MATTER VERSIONS OF THESE PARTICLES AS WELL!!!!!!!



	      if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		double mcslope = 0.0;
		double xytemp  = 0.0;
		double dptemp  = 0.0;
		double d = 0.0;

		xytemp = (*plist).fParticles[c].Px();
		z      = (*plist).fParticles[c].Pz();
		// make (xy,z) a unit vector
		xy = xytemp/sqrt(xytemp*xytemp + z*z);
		z  = z/sqrt(xytemp*xytemp + z*z);

		if(z != 0.0) mcslope = xy/z;
		else         mcslope = 1.0E9;

		dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		if(dptemp < 0.0) dptemp = -dptemp;
	      
		d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vx(), 0.0, b, 100.0, 100.0*m + b));

		// dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		// AND the track-length of the particle must be > 20.0cm
		if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track1 = c;}
	      
	      }  // end if(pdg codes)

	    }  // end for c
	    
	    mcdp1xtemp.push_back(dp);
	    
	    //std::cout << "height/threshold = " << (*hr)[k].fPeak[0].fH/(*hr)[k].fPeak[0].fThr 
		      //<< "   peak = " << (*hr)[k].fPeak[0].fH
		      //<< "   threshold = " << (*hr)[k].fPeak[0].fThr
		      //<< "   dp = " << dp << "\n";
	    
	    //if(dp < 0.0) std::cout << "\n\nFirst line X:   (no match)\n";
	    //else std::cout << "\n\nFirst line X:    particle# = " << track1 << "   PDG = " << (*plist).fParticles[track1].PdgCode() << "   dp = " << dp 
	    //		   << "   track length = " << (*plist).fParticles[track1].Trajectory().TotalLength() << "\n";



	    // loop over all particles and find the one closest to second hough line
	    dp = -999.0;
	    if((*hr)[k].fPeak.size() > 1) {
	      
	      (*hr)[k].SlopeIcept(1, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {


		// skip particle if it is already associated with the first hough line
		if(c != track1) {

		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Px();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vx(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track2 = c;}

		  }  // end if(pdg codes)

		} // end if(c != track1)
	      }  // end for c

	      mcdp2xtemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nSecond line X:   (no match)\n";
	      //else std::cout << "\n\nSecond line X:    particle# = " << track2 << "   PDG = " << (*plist).fParticles[track2].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track2].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 1)



	    // loop over all particles and find the one closest to third hough line
	    dp = -999.0;
	    if((*hr)[k].fPeak.size() > 2) {
	      
	      (*hr)[k].SlopeIcept(2, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {

		// skip particle if it is already associated with the first or second hough line
		if(c != track1 && c != track2) {
		  
		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Px();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vx(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) dp = dptemp;

		  }  // end if(pdg codes)

		} // end if(c != track1 && c != track2)
	      }  // end for c

	      mcdp3xtemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nThird line X:   (no match)\n";
	      //else std::cout << "\n\nThird line X:    particle# = " << track3 << "   PDG = " << (*plist).fParticles[track3].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track3].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 2)
	    
	  }  // end for i
	}  // end if(x-view) 


	
	// y-view
	if((*hr)[k].fView == geo::kY) {
	
	  // loop over the number of MC reactions
	  for(unsigned int i = 0; i < particlelist->size(); ++i) {

	    art::Ptr<sim::ParticleList> plist(particlelist, i);
	    
	    // loop over all particles and find the one closest to first hough line   
	    (*hr)[k].SlopeIcept(0, &m, &b);
	    for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {
          
	      int pdg = (*plist).fParticles[c].PdgCode();

	      // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
	      if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		double mcslope = 0.0;
		double xytemp  = 0.0;
		double dptemp  = 0.0;
		double d = 0.0;

		xytemp = (*plist).fParticles[c].Py();
		z      = (*plist).fParticles[c].Pz();
		// make (xy,z) a unit vector
		xy = xytemp/sqrt(xytemp*xytemp + z*z);
		z  = z/sqrt(xytemp*xytemp + z*z);

		if(z != 0.0) mcslope = xy/z;
		else         mcslope = 1.0E9;

		dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		if(dptemp < 0.0) dptemp = -dptemp;
	      
		d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vy(), 0.0, b, 100.0, 100.0*m + b));

		// dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		// AND the track-length of the particle must be > 20.0cm
		if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track1 = c;}
	      
	      }  // end if(pdg codes)

	    }  // end for c

	    mcdp1ytemp.push_back(dp);

	    //if(dp < 0.0) std::cout << "\n\nFirst line Y:   (no match)\n";
	    //else std::cout << "\n\nFirst line Y:    particle# = " << track1 << "   PDG = " << (*plist).fParticles[track1].PdgCode() << "   dp = " << dp 
	    //		   << "   track length = " << (*plist).fParticles[track1].Trajectory().TotalLength() << "\n";


	    // loop over all particles and find the one closest to second hough line
	    dp = -999.0;
	    if((*hr)[k].fPeak.size() > 1) {
	      
	      (*hr)[k].SlopeIcept(1, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {


		// skip particle if it is already associated with the first hough line
		if(c != track1) {

		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Py();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vy(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track2 = c;}

		  }  // end if(pdg codes)

		} // end if(c != track1)
	      }  // end for c

	      mcdp2ytemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nSecond line Y:   (no match)\n";
	      //else std::cout << "\n\nSecond line Y:    particle# = " << track2 << "   PDG = " << (*plist).fParticles[track2].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track2].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 1)



	    // loop over all particles and find the one closest to third hough line
	    dp = -999.0;
	    if((*hr)[k].fPeak.size() > 2) {
	      
	      (*hr)[k].SlopeIcept(2, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {

		// skip particle if it is already associated with the first or second hough line
		if(c != track1 && c != track2) {
		  
		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Py();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vy(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) dp = dptemp;

		  }  // end if(pdg codes)

		} // end if(c != track1 && c != track2)
	      }  // end for c

	      mcdp3ytemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nThird line Y:   (no match)\n";
	      //else std::cout << "\n\nThird line Y:    particle# = " << track3 << "   PDG = " << (*plist).fParticles[track3].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track3].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 2)
    
	  }  // end for i
	}  // end if(y-view)

	}  // end if((*hr)[k].size() != 0)
      }  // end for k
    }  // end dot product validation
    */




    fNt->numHLX = numHLXtemp;
    fNt->numHLY = numHLYtemp;
    
    fNt->nuX = nuXtemp;
    fNt->nuY = nuYtemp;
    fNt->nuZ = nuZtemp;
    fNt->nuE = nuEtemp;
    
    fNt->nupdg  = nupdgtemp;
    fNt->nuCCNC = nuCCNCtemp;

    fNt->perpdist1x = perpdist1xtemp;
    fNt->perpdist1y = perpdist1ytemp;
    fNt->perpdist2x = perpdist2xtemp;
    fNt->perpdist2y = perpdist2ytemp;
    fNt->perpdist3x = perpdist3xtemp;
    fNt->perpdist3y = perpdist3ytemp;

    fNt->mcdp1x = mcdp1xtemp;
    fNt->mcdp1y = mcdp1ytemp;
    fNt->mcdp2x = mcdp2xtemp;
    fNt->mcdp2y = mcdp2ytemp;
    fNt->mcdp3x = mcdp3xtemp;
    fNt->mcdp3y = mcdp3ytemp;

    fNt->rho1x = rho1xtemp;
    fNt->rho2x = rho2xtemp;
    fNt->rho3x = rho3xtemp;
    fNt->rho1y = rho1ytemp;
    fNt->rho2y = rho2ytemp;
    fNt->rho3y = rho3ytemp;
    fNt->theta1x = theta1xtemp;
    fNt->theta2x = theta2xtemp;
    fNt->theta3x = theta3xtemp;
    fNt->theta1y = theta1ytemp;
    fNt->theta2y = theta2ytemp;
    fNt->theta3y = theta3ytemp;

    fNt->threshx = threshxtemp;
    fNt->threshy = threshytemp;
    fNt->peak1x  = peak1xtemp;
    fNt->peak2x  = peak2xtemp;
    fNt->peak3x  = peak3xtemp;
    fNt->peak1y  = peak1ytemp;
    fNt->peak2y  = peak2ytemp;
    fNt->peak3y  = peak3ytemp;

    fNtTree->Fill();

    return;

  }

} // end namespace hough  
////////////////////////////////////////////////////////////////////////  



#include "art/Framework/Core/ModuleMacros.h"
namespace hough
{
  DEFINE_ART_MODULE(HoughValidate);
}
