////////////////////////////////////////////////////////////////////////
/// \brief   This module reads IFDB DB and then stores NuMI spill info
/// \authors  lgoodenough@anl.gov and sphanbudd@winona.edu
/// \date
////////////////////////////////////////////////////////////////////////

//C++ includes 
#include <string>
#include <iostream>
#include <fstream>
#include <math.h>
#include <time.h> 

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "SummaryData/POTSum.h"
#include "SummaryData/SpillData.h"
#include "NovaTimingUtilities/TimingUtilities.h"
#include "RawData/RawTrigger.h"

//IFDB includes
#include "IFBeam_service.h"
#include "Munits.h"

//ROOT includes
#include <TF1.h>
#include <TGraph.h>
#include <TMinuit.h>



/// Information about the IFDB
namespace ifdb {
	class IFDBSpillInfo : public art::EDProducer {
	public:
		explicit IFDBSpillInfo(fhicl::ParameterSet const& pset); // Required! explicit tag tells the compiler this is not a copy constructor      
		virtual ~IFDBSpillInfo();                        
    
		void produce(art::Event& evt);
    
		void reconfigure(const fhicl::ParameterSet& pset);
    
		void beginSubRun(art::SubRun &sr);
		void endSubRun(art::SubRun &sr);
    
	  
		double extrapolate_position(double, double, double, double, double);
				
		void BpmProjection(std::vector<double> &, std::vector<double> &,
		       std::vector<double> &, std::vector<double> &, std::vector<std::vector<double>>);
		
		double BpmAtTarget(double &xpmean, double &ypmean, double &xpstd, double &ypstd,
						   std::vector<double>TargBpmX, std::vector<double>TargBpmY,
						   std::vector<double>BpmIntX, std::vector<double>BpmIntY);

		// Project Profile centers and widths to target.
		// Return int giving status 0 if no device data
		// Units are Munits.
		int ProfileProjection(double &, double &, double &, double &, std::vector<double>, std::vector<double>);
		
		double GetStats(TGraph *, double &, double &);

		double GetGaussFit(double &, double &, std::vector<double>);
	  
    
  private:
	double fMinPOTCut;
    double fMinHornICut;
    double fMaxHornICut;
    double fMinPosXCut;
    double fMaxPosXCut;
    double fMinPosYCut;
    double fMaxPosYCut;
    double fMinWidthXCut;
    double fMaxWidthXCut;
    double fMinWidthYCut;
    double fMaxWidthYCut;
	double fMaxDeltaTCut;
	std::string pRawDataLabel;

    art::ServiceHandle<ifbeam_ns::IFBeam> ifbeam_handle;
    std::unique_ptr<BeamFolder> bfp;
    sumdata::POTSum *totpots;
    std::string pPotLabel;
    bool useBPMS = false;
    bool useBeamP = false;
    bool isMC;
	  

    static constexpr double
	PI      =		3.14159265,
    z_hp121 =		-72.2309*Munits::foot,	// HP121
    z_vp121 =		-71.3142*Munits::foot,// VP121
    z_pm121 =		-70.5267*Munits::foot,	// PM121
    z_hptgt =		-33.1564*Munits::foot,	// HPTGT
    z_vptgt =		-32.2397*Munits::foot,	// VPTGT
    z_pmtgt =		-31.5980*Munits::foot,	// PMTGT
    z_target_le =	0.0*Munits::foot, // LE location
    actrn1  =		1.2467192*Munits::foot, // ACTRN1	
	wirespacing =	0.5*Munits::mm; // wire spacing on SWIC

				
  };
}


////////////////////////////////////////////////////////////////////////
namespace ifdb
{
  //..............................................................
  IFDBSpillInfo::IFDBSpillInfo(fhicl::ParameterSet const& pset) :
    fMinPOTCut     (pset.get< double      >("MinPOTCut")     ),
    fMinHornICut   (pset.get< double      >("MinHornICut")   ),
    fMaxHornICut   (pset.get< double      >("MaxHornICut")   ),
    fMinPosXCut    (pset.get< double      >("MinPosXCut")    ),
    fMaxPosXCut    (pset.get< double      >("MaxPosXCut")    ),
    fMinPosYCut    (pset.get< double      >("MinPosYCut")    ),
    fMaxPosYCut    (pset.get< double      >("MaxPosYCut")    ),
    fMinWidthXCut  (pset.get< double      >("MinWidthXCut")  ),
    fMaxWidthXCut  (pset.get< double      >("MaxWidthXCut")  ),
    fMinWidthYCut  (pset.get< double      >("MinWidthYCut")  ),
    fMaxWidthYCut  (pset.get< double      >("MaxWidthYCut")  ),
	fMaxDeltaTCut  (pset.get< double      >("MaxDeltaTCut")  ),
	pRawDataLabel  (pset.get< std::string >("RawDataLabel")  ),
    bfp( ifbeam_handle->getBeamFolder(pset.get< std::string >("Bundle"), pset.get< std::string >("URL"), pset.get< double >("TimeWindow"))),
    pPotLabel      (pset.get< std::string >("PotLabel")      ),
    useBPMS        (pset.get< bool        >("UseBPMS")       ),
    useBeamP       (pset.get< bool         >("UseBeamP")      ),
    isMC(false)
  {
    // tell the module what it is making as this is a EDProducer
    produces< sumdata::SpillData >();
    produces< sumdata::POTSum, art::InSubRun >();
    // Be noisy to demonstrate what's happening
    mf::LogInfo("IFDBSpillInfo") << " In IFDBSpillInfo Constructor"<<std::endl;
  }
  
  //......................................................................
  IFDBSpillInfo::~IFDBSpillInfo()
  {
  }
  
  //......................................................................
  void IFDBSpillInfo::beginSubRun(art::SubRun &sr)
  {
    totpots = new sumdata::POTSum(); 
  }
  
  //......................................................................
  void IFDBSpillInfo::endSubRun(art::SubRun &sr)
  {
    if(isMC){
      art::Handle< sumdata::POTSum >mcpots;
      
      //if mc, get the old POTSum object that was filled in GENIEGen, and copy it to my current one
      sr.getByLabel(pPotLabel,mcpots);
      totpots->totpot = mcpots->totpot;
      totpots->totgoodpot = mcpots->totgoodpot;
      totpots->totspills = mcpots->totspills;
      totpots->goodspills = mcpots->goodspills;
     }
	  
    std::unique_ptr< sumdata::POTSum > p(totpots);
    sr.put(std::move(p));
  }
  
  //......................................................................
  void IFDBSpillInfo::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMinPOTCut      = pset.get< double >("MinPOTCut"     );
    fMinHornICut    = pset.get< double >("MinHornICut"   );
    fMaxHornICut    = pset.get< double >("MaxHornICut"   );
    fMinPosXCut     = pset.get< double >("MinPosXCut"    );
    fMaxPosXCut     = pset.get< double >("MaxPosXCut"    );
    fMinPosYCut     = pset.get< double >("MinPosYCut"    );
    fMaxPosYCut     = pset.get< double >("MaxPosYCut"    );
    fMinWidthXCut   = pset.get< double >("MinWidthXCut"  );
    fMaxWidthXCut   = pset.get< double >("MaxWidthXCut"  );
    fMinWidthYCut   = pset.get< double >("MinWidthYCut"  );
    fMaxWidthYCut   = pset.get< double >("MaxWidthYCut"  );
	fMaxDeltaTCut   = pset.get< double >("MaxDeltaTCut"  );
  }

	// Linear extrapolation to location z3
	double IFDBSpillInfo::extrapolate_position(double t1, double z1, double t2, double z2, double z3)
	{
		return t1+(t2-t1)*(z3-z1)/(z2-z1);
	}
	
	
	// Calculate mean and standard deviation for points in TGraph to use as starting parameters values in Fit
	double IFDBSpillInfo::GetStats(TGraph *prof, double &mean, double &stdev)
	{
		
		double qx = 0, qtot = 0;
		
		double X = 0.0;		
		double Qx = 0.0;		
		
		double minprof = 1E6;
		double maxprof = -1E6;
		
		// Find min and max values of profile.  Profile is upside down Gaussian with offset.  All profile values
		// have already been inverted to make Gaussian right-side up - offset is now negative.
		// To get good starting param values, shift all profile values up by the most negative of all
		// profile values, ie minprof, as found above.
		for (int i = 0; i < prof->GetN(); i++) {
					
			Qx = prof->GetY()[i];
		
			if (Qx <= minprof) minprof = Qx;
			if (Qx >= maxprof) maxprof = Qx;
			
		}
				
		for (int i = 0; i < prof->GetN(); i++) {
			// for each point in profile, get wire position and charge values
			X = prof->GetX()[i];
			Qx = prof->GetY()[i];
						
			prof->SetPoint(i,X,Qx-minprof);
			Qx = prof->GetY()[i];
						
			// Total charge on all wires
			qtot += Qx;
			
			// Total charge-weighted position value
			qx += Qx*X;
		}
		
		// If total charge is 0, mean and stdev are 0
		if (qtot == 0.0) {
			mean = 0.0;
			stdev = 0.0;
		}
		
		// Otherwise, mean position is charge-weighted mean
		else {
		
			mean = qx/qtot;
			
			// Charge-weighted variance
			double var = 0.0;
			
			for (int i = 0; i < prof->GetN(); i++) {
							
				X = prof->GetX()[i];
				Qx = prof->GetY()[i];
				
				var +=  Qx*(X-mean)*(X-mean);
			}
			
			var = var/qtot;
			
			if (var>0.0){
				stdev = sqrt(var);
							}
			else {
				stdev=0.0;

			}
		}
	
		return qtot;	
	}// GetStats
	
	
	void IFDBSpillInfo::BpmProjection(std::vector<double> &xp, std::vector<double> &yp, 
									std::vector<double> &xi, std::vector<double> &yi,
									std::vector<std::vector<double>> BPMS)
	{
		xp.clear();
		yp.clear();
		xi.clear();
		yi.clear();
  
		// In NOvA era (for now) we will assume the following
		// In MINOS era, target location was 3937 after time period 1155564632
		double z_targ = 0;
	
		// Return shortest array length of all BPM positions and intensities used
		int size = BPMS.size();
		int n = BPMS[0].size();
		for (int ind=1; ind<size; ++ind) {
			int n2 = BPMS[ind].size();
			if (n2<n) n = n2;
		}

		int start = 0;
		if (n>1) ++start; // Skip first BPMS value (average used in auto-tune) but only if there are multiple batches
		
		// Check to see if elements of BPMS[4] and BPMS[5] (BPM intensities) are 0.
		// If they are, set position values to 0.  Otherwise, use corresponding values of
		// vertical and horizontal position monitors to extrapolate vertical and horizontal
		// position at target.  Simply transfer intensities over.
		for (int ind=start; ind<n; ++ind) {

			if (BPMS[4][ind] == 0.0 || BPMS[5][ind]== 0.0) {
				yp.push_back(0.0);
				xp.push_back(0.0);
				yi.push_back(0.0);
				xi.push_back(0.0);
				continue;
			}

			//Returns transverse position (x or y) at z_targ assuming a linear extrapolation from z_vp121 to z_vptgt
			yp.push_back
				(extrapolate_position(BPMS[0][ind]*Munits::mm,z_vp121,BPMS[2][ind]*Munits::mm,z_vptgt,z_targ));
			xp.push_back
				(extrapolate_position(BPMS[1][ind]*Munits::mm,z_hp121,BPMS[3][ind]*Munits::mm,z_hptgt,z_targ));
			//Returns the horizontal or vertical intensity of beam	
			yi.push_back(BPMS[4][ind]);
			xi.push_back(BPMS[5][ind]);
		}
		
		return;

}// BpmProjection
	

	int IFDBSpillInfo::ProfileProjection(double &x, double &y, double &xstdev, double &ystdev,
									std::vector<double> PM121, std::vector<double> PMTGT)
	{	
		//double z_targ=0;

		//double x121 = 0., y121 = 0., xstdev121 = 0., ystdev121 = 0.;
		double xtgt = 0., ytgt = 0., xstdevtgt = 0., ystdevtgt = 0.;
		
		std::vector<double> hchannels121, vchannels121, hchannelstgt, vchannelstgt;
				
		// Need to add line here to return 0 if no data from devices 
		
		// Get relevant channels from the device		
		for (int ind=0; ind <= 47; ++ind) {
			
			hchannels121.push_back(PM121[103+ind]);
			vchannels121.push_back(PM121[103+48+ind]);
			hchannelstgt.push_back(PMTGT[103+ind]);
			vchannelstgt.push_back(PMTGT[103+48+ind]);
		}
			
		// currently no M121DS device, so we will not use it to extrapolate
		//GetGaussFit(x121,xstdev121,hchannels121);
		//GetGaussFit(y121,ystdev121,vchannels121);
		
		GetGaussFit(xtgt,xstdevtgt,hchannelstgt);
		GetGaussFit(ytgt,ystdevtgt,vchannelstgt);

		// currently no M121DS device, so no extrapolation
		/*if (fabs(x121) < 11*Munits::mm &&
			fabs(xtgt) < 11*Munits::mm &&
			fabs(y121) < 11*Munits::mm &&
			fabs(ytgt) < 11*Munits::mm) {
			x = extrapolate_position(x121,z_pm121,xtgt,z_pmtgt,z_targ);
			y = extrapolate_position(y121,z_pm121,ytgt,z_pmtgt,z_targ);
		}*/
		
		// If values at MTGTDS are reasonable, set them to values at target 
		if (fabs(xtgt) < 11*Munits::mm && fabs(ytgt) < 11*Munits::mm) {
			x = xtgt;
			y = ytgt;
		}
				
		// Standard deviation values are not extrapolated to target
		xstdev = xstdevtgt;
		ystdev = ystdevtgt;
		
		return 1;
	}// ProfileProjection
	
	
	double IFDBSpillInfo::GetGaussFit(double &mean, double &sigma, std::vector<double> profile)
	{		
		TGraph Prof;	
		Prof.Set(0);
		//Prof.Clear("");

		int npoints = 0;
		
		// For each of 48 channels of profile monitor
		for (int ch = 1; ch <= 48; ++ch) {
			
			// Get position of wire. Origin is halfway betwwen channels 24 and 25
			double X = (ch-24.5)*wirespacing;

			// Get the voltage value.
			double Qx = profile[ch-1];
				
			// ensure that they are greater than or equal to zero
			// NOTE: FOR NOW DON'T DO THIS - INSTEAD FIT TO AN OFFSET
			// Qx = Qx > 0 ? Qx : 0;
			
			// Set values of points in TGraph, x=X/Munits::mm and y=-1*Qx/Munits::millivolt
			// Invert voltage to make it positive in Gaussian peak - may lead to better fitting
			Prof.SetPoint(npoints,X/Munits::mm,-1*Qx/Munits::millivolt);
						
			npoints++;

		}// for channels 1 to 48
		
		
		// Now that all of channel information has been loaded...
		
		// Remove dead and hot channels from profiles being fitted.
		// I DON'T DO THIS BUT I WILL REMOVE DEAD CHANNELS (IE THOSE WITH ZERO READING)
		//SuppressDeadHot(&xProf);
		
		for (int i = 0; i < Prof.GetN(); i++) {
			if (Prof.GetY()[i] == 0.0){
				Prof.RemovePoint(i);
				i--;
			}
		}
		
		// For some reason this is faster than if I use a TMath::Gaus:
		TF1 gausF("gausF","[3]+([0]*exp((-1*(x-[1])*(x-[1]))/(2*[2]*[2]))/(sqrt(2*3.142)*[2]))");
		//TF1 gausF("gausF","[3]+[0]*TMath::Gaus(x,[1],[2],1)");
		
		gausF.SetParName(0,"Area");
		gausF.SetParName(1,"Mean");
		gausF.SetParName(2,"Sigma");
		gausF.SetParName(3,"Offset");
		
		//NOT SURE WE NEED THIS
		//gausF.SetParLimits(0,0.00,100000000);
		//gausF.SetParLimits(2,0,1000);
		
		// Gaussian fit parameters
		double area = -9.;
		double offset = -9.;
		mean  = -9., sigma = -9.;
		
		// Get statistical mean and standard deviation of profile distribution
		double StatMean = 0.0, stdev = 0.0;
		area = GetStats(&Prof,StatMean,stdev);
		area*= Munits::millivolt;
		
			
		// Check we have enough points and enough charge for a reasonable fit 
		if (Prof.GetN() > 5 && fabs(area) > 100*Munits::millivolt &&
			fabs(StatMean) < 20*wirespacing/Munits::mm && fabs(stdev) < 12*wirespacing/Munits::mm) {
			
			gausF.SetParameter(0,area/Munits::millivolt);
			gausF.SetParameter(1,StatMean);
			gausF.SetParameter(2,stdev/2);
			// Starting offset value is a guess based on looking at data
			offset = -3.0;
			gausF.SetParameter(3,offset);
						
			if (Prof.Fit("gausF","q","",-24*wirespacing/Munits::mm,24*wirespacing/Munits::mm) == 0) {
							
				bool goodfit = true;
				
				// Are fit results physical?
				if (gausF.GetParameter(0) < 10){
					goodfit = false;
				}
				
				if (fabs(gausF.GetParameter(1)) > 24*wirespacing/Munits::mm){
					goodfit = false;
				}
				
				if (gausF.GetParameter(2) > 24*wirespacing/Munits::mm || gausF.GetParameter(2) < 0.0){
					goodfit = false;
				}
									
				if (goodfit) {
					// Remember to correct fit result for area by dividing by "bin size" and correcting units
					area = gausF.GetParameter(0)/wirespacing*Munits::mm*Munits::millivolt;
					mean = gausF.GetParameter(1)*Munits::mm;
					sigma = gausF.GetParameter(2)*Munits::mm;
					offset = gausF.GetParameter(3)*Munits::millivolt;					
				}// if goodfit
				
			}
		}// if we can make a reasonable fit
		
		else {
			//std::cout << "POINTS DID NOT PASS BASIC CUTS" << std::endl;
			//std::cout << std::endl;			
		}
		
		return area;
	}// GetGaussFit
	

	
	//......................................................................
	void IFDBSpillInfo::produce(art::Event& evt)
	{
		sumdata::SpillData *spilldata = new sumdata::SpillData();
		
		//count the total number of spills
//		totpots->totspills++;
		
    
		if(!evt.isRealData()){
			// remove empty spill info write from here, move to GENIEGen and fill
			isMC=true;
			return;
		}
                
		//this will be the CORRECT UTC time we want ultimately
		struct timespec unixTime;

		// the following is from Cana_module.cc
		// Get trigger information to extract trigger time
		art::Handle< std::vector<rawdata::RawTrigger> > trigv;
		evt.getByLabel(pRawDataLabel, trigv);
                assert(!trigv->empty());
		const rawdata::RawTrigger& trig = (*trigv)[0];

                if(trig.fTriggerMask_TriggerType != 0){
                  // This isn't a NuMI trigger. Don't try to get beam info
                  return;
                }

		novadaq::timeutils::convertNovaTimeToUnixTime(trig.fTriggerTimingMarker_TimeStart, unixTime);
				
		unsigned long int uevt_sec = unixTime.tv_sec; //utval >> 32 & 0xFFFFFFFFF;
		unsigned long int uevt_nsec = unixTime.tv_nsec; //utval & 0xFFFFFFFFF;
			  	
		//DAQ event time 
		double DAQtime = uevt_sec + (0.000000001)*uevt_nsec;
			  	  
		//torroid values
		double trtgtd=0.,tr101d=0.;
	
		//for horn current
		double nslina=0.,nslinb=0.,nslinc=0.,nslind=0.;
		//for horn polarity
		double horndir=0.;
		

	try{
		
		bfp->GetNamedData(DAQtime,"E:TRTGTD,E:TR101D", &trtgtd, &tr101d);
				
		// This checks that event time is contained within time interval pulled from database
		std::vector<double> times = bfp->GetTimeList();
		//double time_low = times[0];
		//double time_high = times[(times.size()-1)];
		
		double time_closest = 0.;	
		double diff = 1e6;//a large number
		
		// Find DB time closest to event time and time difference between the two 	
		for (size_t i = 0; i < times.size(); i++) {		
			
			double time_in_query = times[i];
			
			if(fabs(DAQtime-time_in_query) < diff){
				diff = fabs(DAQtime-time_in_query);
				time_closest = time_in_query;
			}
			
		}//for times.size
		
		spilldata->deltaspilltimensec = (DAQtime-time_closest)*1e9;
		
		unsigned long int time_closest_int = (int) time_closest;
		double time_closest_ns = (time_closest - time_closest_int)*1e9;
		
		spilldata->spilltimesec = time_closest_int;
		spilldata->spilltimensec = time_closest_ns;
		
		// If spill meets time requirement, calculate all quantities and fill appropriate fields
		if (fabs(spilldata->deltaspilltimensec)<fMaxDeltaTCut){
			
			double temppot = trtgtd;
			
			if(temppot<0.02) temppot = tr101d;
			
			// This guarantees that when toroid values go negative, total POT is not reduced
			if(temppot<0.0) temppot = 0.0;
			
			spilldata->spillpot = temppot;
			
			// Total POT
			totpots->totpot += spilldata->spillpot;
			totpots->totspills++;
									
		//These were 0 in old code, so set them to 0 here.
		spilldata->gpsspilltimesec = 0.;
		spilldata->gpsspilltimensec = 0.;

		// Horn polarity: note, this is a fairly new variable, so it is not available in older files
		bfp->GetNamedData(DAQtime,"E:HRNDIR",&horndir);

		// Horn current
		bfp->GetNamedData(DAQtime,"E:NSLINA,E:NSLINB,E:NSLINC,E:NSLIND",&nslina, &nslinb, &nslinc, &nslind);
		
		double ihorn = 0.;
		
		// Updated numbers from Jim Hylen, email Oct 4, 2013 
		ihorn += (nslina - (+0.01)) / 0.9951;
		ihorn += (nslinb - (-0.14)) / 0.9957;
		ihorn += (nslinc - (-0.05)) / 0.9965;
		ihorn += (nslind - (-0.07)) / 0.9945;
	
		spilldata->hornI = ihorn;
		

		// Beam position
		if(useBPMS==true){
			std::vector<double> VP121, HP121, VPTGT, HPTGT, VITGT, HITGT;
			std::vector<std::vector<double>> BPMS;
		
			// BPM Positions and Intensities - they each have 7 elements
			// First is an average value of a few batches (often 2,3,4) used for auto-tuning, so we should disregard it
			VP121 = bfp->GetNamedVector(DAQtime,"E:VP121[]");
			HP121 = bfp->GetNamedVector(DAQtime,"E:HP121[]");
			VPTGT = bfp->GetNamedVector(DAQtime,"E:VPTGT[]");
			HPTGT = bfp->GetNamedVector(DAQtime,"E:HPTGT[]");
			VITGT = bfp->GetNamedVector(DAQtime,"E:VITGT[]");
			HITGT = bfp->GetNamedVector(DAQtime,"E:HITGT[]");

			BPMS.push_back(VP121);
			BPMS.push_back(HP121);
			BPMS.push_back(VPTGT);
			BPMS.push_back(HPTGT);
			BPMS.push_back(VITGT);
			BPMS.push_back(HITGT);
		
			std::vector<double> xp, yp, xi, yi;
		
			// Use Beam Position Monitors to estimate x and y position of beam (xp,yp) and intensities (xi,yi) at target
			// Do this for all six readings of positions monitors
			BpmProjection(xp,yp,xi,yi,BPMS);
						
			double xpmean = -9., ypmean = -9., xpstdev = -9., ypstdev=-9.;
			// Use 6 position and intensity values to get mean and stdev values of positions
			BpmAtTarget(xpmean,ypmean,xpstdev,ypstdev,xp,yp,xi,yi);
			spilldata->posx = xpmean*1000.;//make it in mm
			spilldata->posy = ypmean*1000.;//make it in mm
			
		}//if useBPMS
    
		
		if(useBeamP==true){
		  		
			//BPM Positions	
			std::vector<double> M121DS, MTGTDS;
			M121DS = bfp->GetNamedVector(DAQtime,"E:M121DS[]");
			MTGTDS = bfp->GetNamedVector(DAQtime,"E:MTGTDS[]");
		
			double x = -9., y = -9., xstdev= -9., ystdev= -9.;
			// Use Beam Profile Monitors to estimate x and y widths and positions of beam at target
			ProfileProjection(x,y,xstdev,ystdev,M121DS,MTGTDS);
		
			spilldata->widthx = xstdev*1000.;//make it in mm
			spilldata->widthy = ystdev*1000.;//make it in mm
		
		}//if useBeamP
		
		spilldata->goodbeam = 0;
		
    // Apply good beam cuts
		if(   (spilldata->spillpot>fMinPOTCut)												// POT
		   && (spilldata->hornI>fMinHornICut && spilldata->hornI<fMaxHornICut)				// Horn I
		   && (spilldata->posx>fMinPosXCut && spilldata->posx<fMaxPosXCut)					// x position
		   && (spilldata->posy>fMinPosYCut && spilldata->posy<fMaxPosYCut)					// y position
		   && (spilldata->widthx>fMinWidthXCut && spilldata->widthx<fMaxWidthXCut)			// x width
		   && (spilldata->widthy>fMinWidthYCut&&spilldata->widthy<fMaxWidthYCut)			// y width
		   && (fabs(spilldata->deltaspilltimensec)<fMaxDeltaTCut)				){						// delta t
			
			spilldata->goodbeam=1;

		}// Good beam cuts

    
		// Total good POT	
		if(spilldata->goodbeam==1){
			totpots->totgoodpot += spilldata->spillpot;
			totpots->goodspills++;
		}

/*			
		if(!(spilldata->spillpot>fMinPOTCut)){
			std::cout << "Failed POT cut" << std::endl;
		}
		if(!(spilldata->hornI>fMinHornICut && spilldata->hornI<fMaxHornICut)){
			std::cout << "Failed horn I cut" << std::endl;
		}
		
		if(!(spilldata->posx>fMinPosXCut && spilldata->posx<fMaxPosXCut)){
			std::cout << "Failed pos x cut" << std::endl;
		}
		
		if(!(spilldata->posy>fMinPosYCut && spilldata->posy<fMaxPosYCut)){
			std::cout << "Failed pos y cut" << std::endl;
		}
		
		if(!(spilldata->widthx>fMinWidthXCut && spilldata->widthx<fMaxWidthXCut)){
			std::cout << "Failed width x cut" << std::endl;
		}
		
		if(!(spilldata->widthy>fMinWidthYCut && spilldata->widthy<fMaxWidthYCut)){
			std::cout << "Failed width y cut" << std::endl;
		}
		
		if(!(fabs(spilldata->deltaspilltimensec)<1e9)){
			std::cout << "Failed time cut"<< std::endl;
		}
		      
		std::cout << "Spill pot: " << spilldata->spillpot << " Cuts: >" << fMinPOTCut << std::endl;
		std::cout << "Horn I: " << spilldata->hornI << " Cuts: (" << fMinHornICut << "," << fMaxHornICut << ")" << std::endl;
		std::cout << "Beam Pos X: " << spilldata->posx << " Cuts: (" << fMinPosXCut << "," << fMaxPosXCut << ")" << std::endl;
		std::cout << "Beam Pos Y: " << spilldata->posy << " Cuts: (" << fMinPosYCut << "," << fMaxPosYCut << ")" << std::endl;
		std::cout << "Beam Width X: " << spilldata->widthx << " Cuts: (" << fMinWidthXCut << "," << fMaxWidthXCut << ")" << std::endl;
		std::cout << "Beam Width Y: " << spilldata->widthy << " Cuts: (" << fMinWidthYCut << "," << fMaxWidthYCut << ")" << std::endl;
		std::cout << "Delta T: " << spilldata->deltaspilltimensec << " Cuts: <" << 1e9 << std::endl;
		std::cout << "Good beam is " << spilldata->goodbeam << std::endl;
		std::cout << "##########################" << std::endl;
*/
			
		}// if deltaspilltimensec<fMaxDeltaTCut
		
		else std::cout << "spill did not meet time requirement " << std::endl;

		
		std::unique_ptr<sumdata::SpillData> sd(spilldata);
		evt.put(std::move(sd));
		
    
    }catch (WebAPIException we){
		// Eventually put in better error catching. What if each device isn't filled?
		std::cout << "Exception:" << we.what() << std::endl;
    }//end catch

  }//end event loop
  
    
double IFDBSpillInfo::BpmAtTarget(double &xpmean, double &ypmean, double &xpstdev, double &ypstdev,
								  std::vector<double>TargBpmX, std::vector<double>TargBpmY,
								  std::vector<double>BpmIntX, std::vector<double>BpmIntY)
	{	
		//xpmean = ypmean = xpstdev = ypstdev = 0.0;
		double IXtot = 0.0, IYtot = 0.0;
		double ix = 0.0, iy = 0.0, ix2 = 0.0, iy2 = 0.0;
	
		// for intensity-weighted position values...
		// Remember, TargBpmX, TargBpmY, BpmIntX and BpmIntY have one fewer element than
		// the devices because first device value was an average used for beam tuning
		for (size_t ind = 0; ind < BpmIntX.size(); ++ind) {
				
			// if average intensity is 0, skip to next value
			if (BpmIntX[ind] == 0.0 || BpmIntY[ind] == 0.0) continue;
	
			double X = TargBpmX[ind];//xp[ind]
			double Y = TargBpmY[ind];//yp[ind]
			double IX = BpmIntX[ind];//xi[ind]
			double IY = BpmIntX[ind];//yi[ind]
		
			// if I>0, set I=I, else set I=0	
			IX = IX > 0 ? IX : 0;
			IY = IY > 0 ? IY : 0;
		
			// sum intensities
			IXtot += IX;
			IYtot += IY;

			// intensity-weighted position value
			ix += IX*X;
			iy += IY*Y;
		
			ix2 += IX*X*X;
			iy2 += IY*Y*Y;
		
			}// end for-loop over elements of xp, yp, xi and yi

			if (IXtot <= 0.0) return 0.0;
			if (IYtot <= 0.0) return 0.0;

			// intensity-weighted average positions
			xpmean = ix/IXtot;
			ypmean = iy/IYtot;
		
			// standard deviation
			double xvar = ix2/IXtot-xpmean*xpmean;
			if (xvar > 0.0) xpstdev = sqrt(xvar);
			double yvar = iy2/IYtot-ypmean*ypmean;
			if (yvar > 0.0) ypstdev = sqrt(yvar);
		
			return xpmean;
	
	}// BpmAtTarget
	
	
} // end namespace ifdb

////////////////////////////////////////////////////////////////////////
namespace ifdb
{
  DEFINE_ART_MODULE(IFDBSpillInfo);
}

  
  
