////////////////////////////////////////////////////////////////////////
// \file    ApplyFilter_module.cc
// \brief   Filter according to FilterList status of largest slice
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"

namespace lem
{
  /// Filter according to FilterList status of largest slice
  class ApplyFilter: public art::EDFilter
  {
  public:
    explicit ApplyFilter(const fhicl::ParameterSet& pset);
    virtual bool filter(art::Event& evt);
    virtual void reconfigure(const fhicl::ParameterSet& pset);

  protected:
    std::string fSlicerLabel;
  };

  //......................................................................
  ApplyFilter::ApplyFilter(const fhicl::ParameterSet&)
  {
  }

  //......................................................................
  void ApplyFilter::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlicerLabel = pset.get<std::string>("SlicerLabel");
  }

  //......................................................................
  bool ApplyFilter::filter(art::Event& evt)
  {
    art::Handle<std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // Find the biggest slice
    int bestIdx = -1;
    double bestADC = 0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      if((*slices)[sliceIdx].IsNoise()) continue;
      const double adc = (*slices)[sliceIdx].TotalADC();
      if(adc > bestADC){
        bestADC = adc;
        bestIdx = sliceIdx;
      }
    }

    if(bestIdx == -1) return false;

    return !rb::IsFiltered(evt, slices, bestIdx);
  }

  DEFINE_ART_MODULE(ApplyFilter);

} // end namespace lem
////////////////////////////////////////////////////////////////////////

