////////////////////////////////////////////////////////////////////////
/// \file    BackgroundEnricher_module.cc
/// \brief   Filter to remove non-pizero interactions
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////


#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCParticle.h"

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"

namespace lem
{
  /// Filter to remove non-pizero interactions
  class BackgroundEnricher: public art::EDFilter
  {
  public:
    explicit BackgroundEnricher(const fhicl::ParameterSet& pset);
    virtual ~BackgroundEnricher();

    virtual bool filter(art::Event& evt);
    virtual void reconfigure(const fhicl::ParameterSet& pset);

  protected:
    std::string fMCTruthLabel;
  };

  //----------------------------------------------------------------------
  BackgroundEnricher::BackgroundEnricher(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);
  }

  //----------------------------------------------------------------------
  BackgroundEnricher::~BackgroundEnricher()
  {
  }

  //----------------------------------------------------------------------
  void BackgroundEnricher::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMCTruthLabel = pset.get<std::string>("MCTruthLabel");
  }

  //----------------------------------------------------------------------
  bool BackgroundEnricher::filter(art::Event& evt)
  {
    art::Handle<std::vector<simb::MCTruth> > truths;
    evt.getByLabel(fMCTruthLabel, truths);
    assert(truths->size() == 1);

    const simb::MCTruth& truth = (*truths)[0];

    if(truth.GetNeutrino().CCNC() != simb::kNC) return false;

    // Find the biggest primary pi0
    double maxE = 0;
    const int N = truth.NParticles();
    for(int n = 0; n < N; ++n){
      const simb::MCParticle& part = truth.GetParticle(n);
      const int pdg = part.PdgCode();
      const double E = part.E();
      if(pdg == 111 && E > maxE) maxE = E;
    } // end for primIdx

    return (maxE > 0);

    // There seems to be no limit to how low energy a pi0 can be a source of
    // background
    //    return (maxE > 0.5);
  }

  DEFINE_ART_MODULE(BackgroundEnricher);

} // end namespace
