////////////////////////////////////////////////////////////////////////
// \file    FindMatchesTranspose.cxx
// \brief   TODO
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Persistency/Common/Ptr.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "LEM/LEMInput.h"
#include "LEM/Util.h"

#include "LEM/func/DistanceMap.h"
#include "LEM/func/EventSummary.h"
#include "LEM/func/FindMatches.h"
#include "LEM/func/Heads.h"
#include "LEM/func/Library.h"
#include "LEM/func/MatchList.h"
#include "LEM/func/OnDemand.h"

#include "GeometryObjects/OfflineChan.h"
#include "Utilities/func/EnvExpand.h"
#include "Utilities/func/MathUtil.h"

#include <iostream>

namespace lem
{
  class FindMatchesTranspose: public art::EDProducer
  {
  public:
    explicit FindMatchesTranspose(const fhicl::ParameterSet& pset);
    ~FindMatchesTranspose();

    virtual void beginJob();

    virtual void reconfigure(const fhicl::ParameterSet& pset);

    virtual void produce(art::Event& evt);

    virtual void endRun(art::Run& run);

  protected:
    std::string fLibPath;
    int fDownsampleFactor;
    bool fOnDemand;

    std::string fInputLabel;

    std::vector<MatchableEvent> fTrials;
    std::vector<lem::MatchableEvent> fTrialsDownsample;
    std::vector<Potential> fVs;
    std::vector<Potential> fVsDownsample;
  };

  //.......................................................................
  FindMatchesTranspose::FindMatchesTranspose(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);

    produces<std::vector<geo::OfflineChan>>();

    produces<lem::LibrarySummary, art::InRun>();

    produces<std::vector<lem::MatchList>, art::InRun>();
  }

  //......................................................................
  FindMatchesTranspose::~FindMatchesTranspose()
  {
  }

  //......................................................................
  void FindMatchesTranspose::reconfigure(const fhicl::ParameterSet& pset)
  {
    std::string libpath = pset.get<std::string>("LibraryPath");
    fLibPath            = util::EnvExpansion(libpath);

    fDownsampleFactor = pset.get<int>("DownsampleFactor");

    fInputLabel = pset.get<std::string>("InputLabel");

    // This needs to be before anyone asks the DistanceMap anything, should be
    // safe enough here.
    DistanceMap::SetPlaneScale(pset.get<double>("DistPlaneScale"));
    DistanceMap::SetCellScale(pset.get<double>("DistCellScale"));
    DistanceMap::SetDecayPower(pset.get<double>("DistDecayPower"));
    DistanceMap::SetExpMode(pset.get<bool>("DistExpMode"));

    EventSummary::fgDecayRate = pset.get<double>("HitDecayRate");
    EventSummary::fgChargePower = pset.get<double>("HitChargePower");

    SetOnDemandMemoryLimit(pset.get<int>("LibraryMemoryLimit"));
    fOnDemand = (pset.get<int>("LibraryMemoryLimit") > 0);
  }

  //......................................................................
  void FindMatchesTranspose::beginJob()
  {
  }

  //......................................................................
  void FindMatchesTranspose::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<geo::OfflineChan> > chancol(new std::vector<geo::OfflineChan>);

    art::Handle<std::vector<LEMInput>> inputs;
    evt.getByLabel(fInputLabel, inputs);

    const int inputMax = inputs->size();
    for(int inputIdx = 0; inputIdx < inputMax; ++inputIdx){
      const LEMInput& input = (*inputs)[inputIdx];

      EventSummary sum(input.hits, -1, std::vector<int>());

      sum.run = evt.run();
      sum.subrun = evt.subRun();
      sum.event = evt.event();

      sum.origPdg = 0;

      sum.pdg = 0;
      sum.ccnc = 0;
      sum.y = 0;
      sum.trueEVis = 0;


      //      chancol->push_back(geo::OfflineChan(vtxPlane, vtxCell));
      // Should really be nextPlane, but that's lost inside of Util.cxx. This
      // is almost always right, and only for display purposes anyway.
      //      chancol->push_back(geo::OfflineChan(vtxPlane+1, vtxCellOther));

      const MatchableEvent evt(sum);
      fTrials.push_back(evt);
      fTrialsDownsample.push_back(evt.Downsampled(fDownsampleFactor));

      Potential V;
      FillPotential(sum, V, false, false);
      fVs.push_back(V);

      Potential V_down;
      V_down = V.Downsampled(fDownsampleFactor);
      fVsDownsample.push_back(V_down);

    } // end for sliceIdx

    evt.put(std::move(chancol));
  }

  // TODO: don't copy and paste with LoadLibrary
  //......................................................................
  double SimpleSurvivalProb(const EventSummary& evt)
  {
    if(evt.IsSig()){
      // Scale up to keep as much signal as possible in sample, weighted
      // correctly internally. The missing prefactor is (ss2th13*ssth23)
      return util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
    }
    if(evt.ccnc == 0 && abs(evt.pdg) == 14 && abs(evt.origPdg) == 14){
      // CC mu
      return 1-util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
    }
    if(evt.ccnc == 0 && abs(evt.pdg) == 12 && abs(evt.origPdg) == 12){
      // Beam nue
      //      return 1-.1*util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
      // Want library matching to ignore nues
      return 0;
    }
    if(evt.ccnc == 1 && evt.pdg == evt.origPdg){
      // NC
      return 1;
    }
    if(evt.ccnc == 1 && evt.pdg != evt.origPdg){
      // Swapped NC. We play clever reweighting games so this is OK
      return 1;
    }
    if(evt.ccnc == 0 && abs(evt.pdg) == 14 && abs(evt.origPdg) == 12){
      // CC mu appearance, take ss2th13=0.1, maximal th23
      return .1*.5*util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
    }
    // Don't know what this is
    return 0;
  }

  //......................................................................
  void FindMatchesTranspose::endRun(art::Run& run)
  {
    // One thing that can happen is that there's an exception thrown before
    // anything's even happened. But endRun gets called first. Don't want to
    // stall confusingly in that case.
    if(fTrials.empty()) return;

    const unsigned int Ntrials = fTrials.size();
    std::cout << "Have " << Ntrials << " trial events" << std::endl;

    const HeadsTranspose* heads = fOnDemand ? HeadsTranspose::FromMMapOnDemand(fLibPath) : HeadsTranspose::FromMMap(fLibPath);
    std::cout << "Loaded " << heads << std::endl;
    const Library* lib = fOnDemand ? Library::FromMMapOnDemand(fLibPath) : Library::FromMMap(fLibPath, false);
    std::cout << "Loaded " << lib->Summary() << std::endl;

    std::cout << "Matching heads..." << std::endl;

    std::vector<int> bestHeads(Ntrials, -1);
    std::vector<float> bestHeadEs(Ntrials, FLT_MAX);

    const int nHeads = heads->NHeads();
    for(int i = 0; i < nHeads; ++i){
      if(i%100 == 0) std::cout << i/100 << "/" << nHeads/100 << std::endl;

      const int headIdx = heads->HeadIdx(i);

      for(unsigned int trialIdx = 0; trialIdx < Ntrials; ++trialIdx){
        const MatchableEvent& trial = fTrials[trialIdx];
        const MatchableEvent& trialDown = fTrialsDownsample[trialIdx];

        for(bool flipEven: {false, true}){
          for(bool flipOdd: {false, true}){

            const MatchableEvent headevt_down = lib->DownsampledEvent(headIdx).Flipped(flipEven, flipOdd);

            // Compare the downsampled event to the downsampled potential. The
            // resulting energy is a lower-bound on what the real match would
            // have done.
            float E = CalcEnergy(fVsDownsample[trialIdx],
                                 trialDown, headevt_down);

            if(E < bestHeadEs[trialIdx]){
              const MatchableEvent headevt = lib->Event(headIdx).Flipped(flipEven, flipOdd);

              // Calculate the energy fully
              E = CalcEnergy(fVs[trialIdx], trial, headevt);

              // If it's still good, use it
              if(E < bestHeadEs[trialIdx]){
                bestHeads[trialIdx] = i;
                bestHeadEs[trialIdx] = E;
              }
            }
          } // end for flipOdd
        } // end for flipEven
      } // end for trialIdx
    } // end for i

    for(unsigned int trialIdx = 0; trialIdx < Ntrials; ++trialIdx){
      std::cout << "Best head " << bestHeads[trialIdx] << " with " << bestHeadEs[trialIdx]/fTrials[trialIdx].selfEnergy << std::endl;
    }

    std::cout << "Main PID..." << std::endl;

    const unsigned int kNumMatches = 1001;

    // Candidates for storing, remains sorted
    std::vector<std::multiset<MatchSummary>> Elists(Ntrials);

    int* pHeadSeqs = heads->pHeadSeqs();

    const unsigned int libSize = lib->NEvents();
    for(unsigned int libIdx = 0; libIdx < libSize; ++libIdx){
      if(libIdx%10000 == 0) std::cout << libIdx/10000 << "/" << libSize/10000 << std::endl;

      const int numHeadsForLibEvt = *pHeadSeqs;
      ++pHeadSeqs;
      std::vector<bool> isHeadForLibEvt(nHeads, false);
      for(int i = 0; i < numHeadsForLibEvt; ++i){
        isHeadForLibEvt[*pHeadSeqs] = true;
        ++pHeadSeqs;
      }

      for(unsigned int trialIdx = 0; trialIdx < Ntrials; ++trialIdx){
        // Don't try to match this library event and this trial if none of this
        // library events heads are the one the trial event wants.
        if(!isHeadForLibEvt[bestHeads[trialIdx]]) continue;

        const MatchableEvent& trial = fTrials[trialIdx];
        const MatchableEvent& trialDown = fTrialsDownsample[trialIdx];
        std::multiset<MatchSummary>& Elist = Elists[trialIdx];

        float target = (Elist.size() == kNumMatches) ? Elist.rbegin()->potential : FLT_MAX;

        for(bool flipEven: {false, true}){
          for(bool flipOdd: {false, true}){
            const lem::MatchableEvent libevt_down = lib->DownsampledEvent(libIdx).Flipped(flipEven, flipOdd);

            // Compare the downsampled event to the downsampled potential. The
            // resulting energy is a lower-bound on what the real match would
            // have done.
            float E = CalcEnergy(fVsDownsample[trialIdx],
                                 trialDown, libevt_down);

            if(E < target){
              const lem::MatchableEvent libevt = lib->Event(libIdx).Flipped(flipEven, flipOdd);
              // Calculate the energy fully
              E = CalcEnergy(fVs[trialIdx], trial, libevt);

              // If it's still good, use it
              if(E < target){
                Match m(E, &libevt, flipEven, flipOdd);
                // This is extra work for matches we're not even sure we're
                // going to keep. But it's better than having to do a whole
                // extra pass through the library at the end.
                m.qfrac = FracChargeMatched(trial, libevt,
                                            flipEven, flipOdd);
                MatchSummary ms(m);
                Elist.insert(ms);

                if(Elist.size() > kNumMatches){
                  // Always drop the last one, to keep list correct size
                  auto itLargest = Elist.end();
                  --itLargest;
                  Elist.erase(itLargest);
                  target = Elist.rbegin()->potential;
                }
              }
            }
          } // end for flipOdd
        } // end for flipEven

      } // end for trialIdx
    } // end for n

    std::cout << "Writing results out..." << std::endl;

    std::unique_ptr<std::vector<lem::MatchList>> ms(new std::vector<lem::MatchList>);
    ms->resize(Ntrials);

    for(unsigned int trialIdx = 0; trialIdx < Ntrials; ++trialIdx){
      std::cout << trialIdx << "/" << Ntrials << std::endl;

      (*ms)[trialIdx].trial = MatchSummary(Match(0, &fTrials[trialIdx]));
      for(MatchSummary m: Elists[trialIdx]){
        m.potential /= fTrials[trialIdx].selfEnergy;
        (*ms)[trialIdx].matches.push_back(m);
      }
    }

    run.put(std::move(ms));

    std::unique_ptr<lem::LibrarySummary> ls(new lem::LibrarySummary);
    *ls = lib->Summary();
    run.put(std::move(ls));

    delete lib;
    delete heads;
  }

  DEFINE_ART_MODULE(FindMatchesTranspose);

} // end namespace lem
////////////////////////////////////////////////////////////////////////

