////////////////////////////////////////////////////////////////////////
// \file    GenFromLib_source.cc
// \brief   Reconstitute ART events from a LEM library
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "art/Framework/Core/InputSourceMacros.h"
#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/IO/Sources/SourceHelper.h"
#include "art/Framework/Core/ProductRegistryHelper.h"
#include "art/Framework/IO/Sources/put_product_in_principal.h"
#include "art/Framework/IO/Sources/Source.h"
#include "art/Framework/Principal/EventPrincipal.h"
#include "art/Framework/Principal/RunPrincipal.h"
#include "art/Framework/Principal/SubRunPrincipal.h"
#include "art/Utilities/Exception.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"


#include "TFile.h"
#include "TTree.h"

#include "Calibrator/Calibrator.h"
#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "RawData/RawDigit.h"
#include "RawData/RawTrigger.h"
#include "SummaryData/RunData.h"
#include "RecoBase/CellHit.h"
#include "LEM/func/EventSummary.h"

namespace lem
{
  /// Reconstitute ART events from a %LEM library
  class GenFromLib
  {
  public:
    GenFromLib(const fhicl::ParameterSet& pset,
               art::ProductRegistryHelper& help,
               const art::SourceHelper& pm);

    ~GenFromLib();

    void readFile(const std::string& fname,
                  art::FileBlock*& fb);

    bool readNext(const art::RunPrincipal* inR,
                  const art::SubRunPrincipal* inSR,
                  art::RunPrincipal*& outR,
                  art::SubRunPrincipal*& outSR,
                  art::EventPrincipal*& outE);

    void closeCurrentFile();

    void reconfigure(const fhicl::ParameterSet& pset);

  protected:
    art::SourceHelper fSourceHelper; ///< Class to help with {Run, SubRun, Event}Principal construction.

    bool fOnce; ///< Flag that this is the first call to \ref readNext

    int fTreeIdx;
    int fEvtNum;

    TFile* fFile;
    TTree* fTree;
  };

  //.......................................................................
  GenFromLib::GenFromLib(const fhicl::ParameterSet& pset,
                           art::ProductRegistryHelper& help,
                           const art::SourceHelper& pm)
    : fSourceHelper(pm), fOnce(true), fEvtNum(1), fTree(0)
  {
    reconfigure(pset);

    help.reconstitutes<std::vector<rawdata::RawTrigger>, art::InEvent>("daq");
    help.reconstitutes<std::vector<rawdata::RawDigit>, art::InEvent>("daq");
    help.reconstitutes<sumdata::RunData, art::InRun>("daq");

    //    help.reconstitutes<lem::EventSummary, art::InEvent>("daq");
  }

  //......................................................................
  GenFromLib::~GenFromLib()
  {
  }

  //......................................................................
  void GenFromLib::reconfigure(const fhicl::ParameterSet& pset)
  {
    // We want to store exactly the values of the hits, only rescale them later
    // when actually doing matching.
    //
    // NB: static nature of these means it's impossible to regen an event and
    // then match it using different values here.
    EventSummary::fgDecayRate = 0;
    EventSummary::fgChargePower = 1;
  }

  //......................................................................
  void GenFromLib::readFile(const std::string& fname, art::FileBlock*& fb)
  {
    fFile = new TFile(fname.c_str());
    if(fFile->IsZombie()) throw art::Exception(art::errors::FileOpenError);
    fTree = (TTree*)fFile->Get("makelib/tree");
    if(!fTree) throw art::Exception(art::errors::FileReadError);
    fTreeIdx = 0;

    EventSummary::InitFromTree(fTree);

    fb = new art::FileBlock(art::FileFormatVersion(1, "GenFromLib 12-08-10"),
                            fname);
  }

  //......................................................................
  bool GenFromLib::readNext(const art::RunPrincipal* /*inR*/,
                             const art::SubRunPrincipal* /*inSR*/,
                             art::RunPrincipal*& outR,
                             art::SubRunPrincipal*& outSR,
                             art::EventPrincipal*& outE)
  {
    const int tstamp = 0;

    if(fOnce){
      fOnce = false;

      outR = fSourceHelper.makeRunPrincipal(1, tstamp);
      outSR = fSourceHelper.makeSubRunPrincipal(1, 1, tstamp);

      // Need to write the detector into the file, because we're the first
      art::ServiceHandle<geo::Geometry> geom;

      std::unique_ptr<sumdata::RunData> runcol(new sumdata::RunData(geom->DetId(),
                                                                    geom->FileBaseName()));
      art::put_product_in_principal(std::move(runcol), *outR, "daq");
    }

    // Ran off the end of this tree. Get another one, or quit
    if(fTreeIdx >= fTree->GetEntries()) return false;

    outE = fSourceHelper.makeEventPrincipal(1, 1, fEvtNum, tstamp);
    ++fEvtNum;

    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<cmap::CMap> cmap;

    std::unique_ptr<std::vector<rawdata::RawDigit> > digcol(new std::vector<rawdata::RawDigit>);
    std::unique_ptr<std::vector<rawdata::RawTrigger> > trigcol(new std::vector<rawdata::RawTrigger>);

    // Doesn't matter if we say it's enriched or not. Not serialized.
    const EventSummary sum = EventSummary::FromTree(fTree, fTreeIdx, false);
    ++fTreeIdx;

    // Stick the original event summary in the record as well
    // TODO: at least in the TBrowser this looks corrupted
    //    std::unique_ptr<lem::EventSummary> sumcol(new lem::EventSummary(sum));
    //    art::put_product_in_principal(std::move(sumcol), *outE, "daq");

    // Note: AddVtxToLib depends on these being the offsets. If they're changed
    // here they need to change there as well.
    const int plane0 = geom->NPlanes()/2-128;
    const int cell0 = geom->Plane(0)->Ncells()/2-128;
    assert(plane0 >= 0);
    assert(cell0 >= 0);

    for(int hitIdx = 0; hitIdx < sum.nhits; ++hitIdx){
      const LiteHit& hit = sum.hits[hitIdx];

      const daqchannelmap::lchan chan = cmap->Map()->encodeLChan(geom->DetId(), hit.Plane()+plane0, hit.Cell()+cell0);

      // Put in 100 as a reference value
      const short adc = 100;
      const int tdc = 10000;

      rawdata::RawDigit dig(chan, 0, {adc}, tdc);
      dig.SetMC();

      // Push the hit through the procedure CalHit will use, and find out
      // what PECorr it'll end up with. Use that to scale to the PECorr value
      // we actually want it to obtain. It's not so vital this is right, I'm
      // not sure anyone downstream cares about the overall scaling.
      const rb::CellHit chit(cal->MakeCellHit(&dig));
      const double control = cal->GetPECorr(chit, 0);
      assert(control > 0);
      const short newadc = 100*hit.pecorr/control;
      dig.SetADC(0, newadc);
      digcol->push_back(dig);
    } // end for hitIdx

    // Event display for one insists on a RawTrigger, give it junk
    trigcol->push_back(rawdata::RawTrigger());

    art::put_product_in_principal(std::move(digcol), *outE, "daq");
    art::put_product_in_principal(std::move(trigcol), *outE, "daq");

    return true;
  }

  //......................................................................
  void GenFromLib::closeCurrentFile()
  {
    fFile->Close();
    fTree = 0;
    fTreeIdx = 0;
  }

  typedef art::Source<GenFromLib> GenFromLibSource;

  DEFINE_ART_INPUT_SOURCE(GenFromLibSource);

} // end namespace lem
////////////////////////////////////////////////////////////////////////

