////////////////////////////////////////////////////////////////////////
// \file    LEMMixer_module.cc
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/PIDDetails.h"

// Framework includes
#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Modules/MixFilter.h"

namespace lem
{
  class LEMMixer
  {
  public:
    explicit LEMMixer(const fhicl::ParameterSet& pset,
		      art::MixHelper& helper);

    size_t nSecondaries() {return 1;}

    /// Fetch the matching lempart file from SAM
    virtual void respondToOpenInputFile(const art::FileBlock& fb);

  protected:
    /// Tell ART what file to mix with
    std::string getSecondaryFilename();

    /// Rewrite filename to find the corresponding lempart file
    std::string LEMPartFor(std::string fname) const;

    /// Do the actual mixing
    bool MixPIDDetails(const std::vector<const std::vector<PIDDetails>*>& pids,
                       std::vector<PIDDetails>& out,
                       const art::PtrRemapper& rm);

    bool fFromSAM; ///< Are we fetching lempart files from SAM?
    std::string fLempartFilename; ///< If so, this is the file we got
  };

  //......................................................................
  LEMMixer::LEMMixer(const fhicl::ParameterSet& pset,
		     art::MixHelper& helper)
    : fFromSAM(pset.get<bool>("FetchLempartFromSAM"))
  {
    helper.declareMixOp(art::InputTag(pset.get<std::string>("PIDLabel"), ""),
                        &LEMMixer::MixPIDDetails, *this);

    // If we're getting files from SAM, register this so that ART knows what
    // file we fetched. Otherwise it comes from the fclx
    if(fFromSAM) helper.registerSecondaryFileNameProvider(std::bind(&LEMMixer::getSecondaryFilename, this));
  }

  //......................................................................
  std::string LEMMixer::getSecondaryFilename()
  {
    assert(!fLempartFilename.empty());
    mf::LogInfo("LEMMixer") << "Merging with " << fLempartFilename << " from SAM...";
    return fLempartFilename;
  }

  //......................................................................
  bool LEMMixer::MixPIDDetails(const std::vector<const std::vector<PIDDetails>*>& pids,
			       std::vector<PIDDetails>& out,
			       const art::PtrRemapper& rm)
  {
    assert(pids.size() == 1);

    out = *pids[0];

    return true;
  }

  //......................................................................
  std::string LEMMixer::LEMPartFor(std::string fname) const
  {
    // Remove any path information from the input file to get the basename
    size_t pos = fname.rfind("/");
    if(pos != std::string::npos) fname = fname.substr(pos+1);

    mf::LogInfo("LEMMixer") << "Input file basename is " << fname << std::endl;

    // Convert the filename to the matching lempart file
    pos = fname.find("pidpart");
    assert(pos != std::string::npos);
    fname.replace(pos, 7, "lempart");

    return fname;
  }

  //......................................................................
  void LEMMixer::respondToOpenInputFile(const art::FileBlock& fb)
  {
    if(!fFromSAM) return;

    fLempartFilename = LEMPartFor(fb.fileName());    

    mf::LogInfo("LEMMixer") << "Fetching " << fLempartFilename << " from SAM...";

    // lempart file is small, safe to fetch into current directory
    const int status = system(("ifdh_fetch -e nova "+fLempartFilename).c_str());
    assert(status == 0);
  }

  DEFINE_ART_MODULE(art::MixFilter<LEMMixer>);

} //namespace lem
////////////////////////////////////////////////////////////////////////

