////////////////////////////////////////////////////////////////////////
// \file   LEMSummarizer_module.cc
// \brief  Transform slices into the minimal description %LEM needs to PID them
// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/EventSummary.h"
#include "LEM/LEMInput.h"
#include "LEM/Util.h"

// NOvASoft includes
#include "Calibrator/Calibrator.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Vertex.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"

namespace lem
{
  /// Transform slices into the minimal description %LEM needs to PID them
  class LEMSummarizer: public art::EDProducer
  {
  public:
    explicit LEMSummarizer(const fhicl::ParameterSet& pset);
    ~LEMSummarizer();

    virtual void reconfigure(const fhicl::ParameterSet& pset);
    virtual void produce(art::Event& evt);

  protected:
    // Scale ND input event by this much to match the FD-based library. This
    // wouldn't be necessary if the library were in GeV rather than pecorr.
    double fNDEnergyScaleFactor;

    std::string fSlicerLabel;
    std::string fVertexLabel;
  };

  //......................................................................
  LEMSummarizer::LEMSummarizer(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);

    produces<std::vector<LEMInput>>();
    produces<art::Assns<LEMInput, rb::Cluster>>();
  }

  //......................................................................
  LEMSummarizer::~LEMSummarizer()
  {
  }

  //......................................................................
  void LEMSummarizer::reconfigure(const fhicl::ParameterSet& pset)
  {
    fNDEnergyScaleFactor = pset.get<double>("NDEnergyScaleFactor");

    fSlicerLabel = pset.get<std::string>("SlicerLabel");
    fVertexLabel = pset.get<std::string>("VertexLabel");
  }

  //......................................................................
  void LEMSummarizer::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<LEMInput>> sumcol(new std::vector<LEMInput>);
    std::unique_ptr<art::Assns<LEMInput, rb::Cluster> > assns(new art::Assns<LEMInput, rb::Cluster>);

    art::Handle<std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    art::ServiceHandle<calib::Calibrator> cal;

    art::ServiceHandle<geo::Geometry> geom;
    const bool isND = (geom->DetId() == novadaq::cnv::kNEARDET);

    const int sliceMax = slices->size();
    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      if(rb::IsFiltered(evt, slices, sliceIdx)) continue;

      const art::Ptr<rb::Cluster> slice(slices, sliceIdx);

      if(slice->IsNoise()) continue;

      // Find the associated vertex and transform it into plane/cell
      art::FindManyP<rb::Vertex> fmt(slices, evt, fVertexLabel);
      std::vector<art::Ptr<rb::Vertex> > vtxs = fmt.at(sliceIdx);
      int vtxPlane, vtxCell, vtxCellOther;
      if(vtxs.empty()){
        DefaultVertex(*slice, vtxPlane, vtxCell, vtxCellOther);
      }
      else{
        assert(vtxs.size() == 1);
        VertexToPlaneAndCell(vtxs[0]->GetXYZ(), *slice,
                             vtxPlane, vtxCell, vtxCellOther);
      }

      LEMInput sum;

      // TODO: merge hits that fall on already hit cells. This apparently
      // happens. Does it cause the self-energy to be wrong?
      for(unsigned int cellIdx = 0; cellIdx < slice->NCell(); ++cellIdx){
        const art::Ptr<rb::CellHit>& chit = slice->Cell(cellIdx);

        const int planeDelta = chit->Plane()-vtxPlane;
        const int cellDelta = ((abs(planeDelta)%2) ?
                               chit->Cell()-vtxCellOther :
                               chit->Cell()-vtxCell);

        // Align to standard position and clamp into allowed range
        int plane = std::min(255, std::max(0, planeDelta + lem::kVertexPlane));
        int cell  = std::min(255, std::max(0, cellDelta  + lem::kVertexCell ));

        const rb::RecoHit rhit = slice->RecoHit(chit);
        if(rhit.IsCalibrated()){
          double pecorr = rhit.PECorr();
          if(isND) pecorr *= fNDEnergyScaleFactor;
          sum.hits.push_back(LiteHit(plane, cell, pecorr, 0));
        }
      } // end for cellIdx

      sum.totalGeV = 0;
      // Mean position in each view
      const double wx = (slice->NXCell() > 0) ? slice->W(slice->XCell(0).get()) : 0;
      const double wy = (slice->NYCell() > 0) ? slice->W(slice->YCell(0).get()) : 0;

      for(unsigned int hitIdx = 0; hitIdx < slice->NCell(); ++hitIdx){
        // This is slow because we have to keep recalculating the mean
        // positions. The manual version below is much better.
        // const rb::RecoHit rhit = slice->RecoHit(hitIdx);

        const art::Ptr<rb::CellHit>& chit = slice->Cell(hitIdx);

        const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

        if(!rhit.IsCalibrated()){
          LOG_DEBUG("LEMSummarizer") << "Not calibrated?! "
                                     << chit->Plane() << " " << chit->Cell()
                                     << std::endl;
          continue;
        }
        sum.totalGeV += rhit.GeV();
      }

      sumcol->push_back(sum);
      util::CreateAssn(*this, evt, *sumcol, slice, *assns);
    } // end for sliceIdx

    evt.put(std::move(sumcol));
    evt.put(std::move(assns));
  }

  DEFINE_ART_MODULE(LEMSummarizer);

} //namespace lem
////////////////////////////////////////////////////////////////////////

