////////////////////////////////////////////////////////////////////////
// \file    MakeLibrary_module.cc
// \brief   Summarize ART events into LEM libary format
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/EventSummary.h"
#include "LEM/Util.h"

// ROOT includes
#include "TH1.h"
#include "TTree.h"

// NOvASoft includes
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"

#include "MCCheater/BackTracker.h"

#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"

#include "Simulation/Particle.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"

#include "SummaryData/POTSum.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"

namespace lem
{
  /// Summarize ART events into %LEM libary format
  class MakeLibrary: public art::EDProducer
  {
  public:
    explicit MakeLibrary(const fhicl::ParameterSet& pset);
    ~MakeLibrary();

    virtual void produce(art::Event& evt);

    virtual void reconfigure(const fhicl::ParameterSet& pset);

    virtual void beginJob();
    virtual void endSubRun(art::SubRun& sr);
    virtual void endJob();
  protected:
    TVector3 SliceMeanPosEWeighted(const rb::Cluster& slice) const;

    TTree* fTree;
    double fTotalPOT;

    std::string fSlicerLabel;
    std::string fMCTruthLabel;
    std::string fParticleLabel;
  };

  //......................................................................
  MakeLibrary::MakeLibrary(const fhicl::ParameterSet& pset)
    : fTree(0), fTotalPOT(0)
  {
    reconfigure(pset);
  }

  //......................................................................
  MakeLibrary::~MakeLibrary()
  {
  }

  //......................................................................
  void MakeLibrary::reconfigure(const fhicl::ParameterSet& pset)
  {
    fParticleLabel = pset.get<std::string>("ParticleLabel");
    fSlicerLabel   = pset.get<std::string>("SlicerLabel");
    fMCTruthLabel  = pset.get<std::string>("MCTruthLabel");

    // We want to store exactly the values of the hits, only rescale them later
    // when actually doing matching.
    EventSummary::fgDecayRate = 0;
    EventSummary::fgChargePower = 1;
  }

  //......................................................................
  void MakeLibrary::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    fTree = tfs->make<TTree>("tree", "tree");

    EventSummary::InitToTree(fTree);
  }

  //......................................................................
  void MakeLibrary::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    TH1* hPOT = tfs->make<TH1F>("hTotalPOT", ";; POT", 1, 0, 1);
    hPOT->Fill(.5, fTotalPOT);
  }

  //----------------------------------------------------------------------
  void MakeLibrary::endSubRun(art::SubRun& sr)
  {
    art::Handle<sumdata::POTSum> pot;
    sr.getByLabel(fMCTruthLabel, pot);

    fTotalPOT += pot->totgoodpot;

    std::cout << "POT " << pot->totgoodpot << std::endl;
  }

  //......................................................................
  void MakeLibrary::produce(art::Event& evt)
  {
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<cheat::BackTracker> bt;

    art::Handle<std::vector<simb::MCFlux> > fluxs;
    evt.getByLabel(fMCTruthLabel, fluxs);

    assert(fluxs->size() == 1);
    const simb::MCFlux& flux = (*fluxs)[0];

    art::Handle<std::vector<simb::MCTruth> > truths;
    evt.getByLabel(fMCTruthLabel, truths);
    assert(truths->size() == 1);
    const simb::MCTruth& truth = (*truths)[0];

    const simb::MCNeutrino& nu = truth.GetNeutrino();
    const TVector3 vtx = nu.Nu().Position(0).Vect();

    art::Handle<std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // Find the biggest slice
    int bestIdx = -1;
    double bestADC = 0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      if((*slices)[sliceIdx].IsNoise()) continue;
      const double adc = (*slices)[sliceIdx].TotalADC();
      if(adc > bestADC){
        bestADC = adc;
        bestIdx = sliceIdx;
      }
    }

    if(bestIdx == -1) return;

    const rb::Cluster& slice = (*slices)[bestIdx];

    const int firstPlane = slice.MinPlane();
    const int meanCellX = (slice.MinCell(geo::kX)+slice.MaxCell(geo::kX))/2;
    const int meanCellY = (slice.MinCell(geo::kY)+slice.MaxCell(geo::kY))/2;

    // Collect all the track IDs that are ever primary for a hit
    std::set<int> ids;
    for(unsigned int cellIdx = 0; cellIdx < slice.NCell(); ++cellIdx){
      const art::Ptr<rb::CellHit>& chit = slice.Cell(cellIdx);
      const std::vector<cheat::TrackIDE> ides = bt->HitToTrackIDE(chit);
      if(!ides.empty()) ids.insert(ides[0].trackID);
    }
    if(ids.size() > 255) mf::LogWarning("MakeLibrary") << "Track ID overflow";
    // Map them onto small integers
    std::map<int, unsigned char> idMap;
    for(int id : ids){
      const unsigned char next = std::min(idMap.size(), 255ul);
      idMap[id] = next;
    }

    // TODO: merge hits that fall on already hit cells. This apparently
    // happens. Does it cause the self-energy to be wrong?
    std::vector<LiteHit> hs;
    for(unsigned int cellIdx = 0; cellIdx < slice.NCell(); ++cellIdx){
      const art::Ptr<rb::CellHit>& chit = slice.Cell(cellIdx);

      const int meanCell = (chit->View() == geo::kX ? meanCellX : meanCellY);

      int plane = chit->Plane()-firstPlane+lem::kVertexPlane;
      // Can't underflow, but could overflow
      plane = std::min(255, plane);

      // Get it central in the window
      int cell = chit->Cell()-meanCell+lem::kVertexCell;
      // Clamp into allowed range
      cell = std::max(0, cell);
      cell = std::min(255, cell);

      const std::vector<cheat::TrackIDE> ides = bt->HitToTrackIDE(chit);
      const unsigned char idIdx = ides.empty() ? 255 : idMap[ides[0].trackID];

      hs.push_back(LiteHit(plane, cell,
                           slice.RecoHit(chit).PECorr(), idIdx));
    } // end for cellIdx

    std::vector<int> pdgs;
    for(int id : ids){
      if(id != 255){
        const sim::Particle* part = bt->TrackIDToParticle(id);
        if(part) pdgs.push_back(part->PdgCode());
      }
    }

    EventSummary sum(hs, -1, pdgs);

    sum.run = evt.run();
    sum.subrun = evt.subRun();
    sum.event = evt.event();

    sum.origPdg = flux.fntype;

    sum.pdg = nu.Nu().PdgCode();
    sum.ccnc = nu.CCNC();
    sum.mode = nu.Mode();
    sum.y = nu.Y();
    sum.trueEVis = nu.Nu().E()*(nu.CCNC() ? nu.Y() : 1);


    // Convert the true vertex to an integer cell, and store in the summary

    VertexToPlaneAndCell(vtx, slice, sum.trueVtxPlane, sum.trueVtxCell, sum.trueVtxCellOther);
    const int meanCell      = (geom->Plane(sum.trueVtxPlane)->View() == geo::kX) ? meanCellX : meanCellY;
    const int meanCellOther = (geom->Plane(sum.trueVtxPlane)->View() == geo::kX) ? meanCellY : meanCellX;

    sum.trueVtxPlane = sum.trueVtxPlane-firstPlane+lem::kVertexPlane;
    // Can't underflow, but could overflow
    sum.trueVtxPlane = std::min(255, sum.trueVtxPlane);

    // Get it central in the window
    sum.trueVtxCell = sum.trueVtxCell-meanCell+lem::kVertexCell;
    // Clamp into allowed range
    sum.trueVtxCell = std::max(0, sum.trueVtxCell);
    sum.trueVtxCell = std::min(255, sum.trueVtxCell);

    // Get it central in the window
    sum.trueVtxCellOther = sum.trueVtxCellOther-meanCellOther+lem::kVertexCell;
    // Clamp into allowed range
    sum.trueVtxCellOther = std::max(0, sum.trueVtxCellOther);
    sum.trueVtxCellOther = std::min(255, sum.trueVtxCellOther);

    // Conversion length on photon pi0 children
    art::Handle<std::vector<sim::Particle> > parts;
    evt.getByLabel(fParticleLabel, parts);
    if(!parts.failedToGet()){
      int pi0Id = -10;
      std::map<double, double> photLs; // -energy, length, just for sorting
      for(unsigned int n = 0; n < parts->size(); ++n){
        const sim::Particle& p = (*parts)[n];
        if(p.Process() == "Primary" && p.PdgCode() == 111) pi0Id = p.TrackId();
        if(p.PdgCode() == 22 && p.Mother() == pi0Id){
          photLs[-p.E()] = (p.EndPosition().Vect()-vtx).Mag();
        }
      }
      if(!photLs.empty()){
        sum.photE0 = -photLs.begin()->first;
        sum.photL0 = photLs.begin()->second;

        photLs.erase(photLs.begin());
        if(!photLs.empty()){
          sum.photE1 = -photLs.begin()->first;
          sum.photL1 = photLs.begin()->second;
        }
      }
    }

    sum.ToTree(fTree);
  }

  //......................................................................
  TVector3 MakeLibrary::SliceMeanPosEWeighted(const rb::Cluster& slice) const
  {
    art::ServiceHandle<geo::Geometry> geom;

    double Ex = 0;
    double Ey = 0;

    TVector3 mean;

    for(unsigned int i = 0; i < slice.NCell(); ++i){
      const art::Ptr<rb::CellHit>& h = slice.Cell(i);
      double xyz[3];
      geom->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
      const double x = xyz[0];
      const double y = xyz[1];
      const double z = xyz[2];
      const geo::View_t view = geom->Plane(h->Plane())->View();

      const double E = slice.RecoHit(h).PECorr();

      if(view == geo::kX){
        mean.SetX(mean.X()+x*E);
        Ex += E;
      }
      if(view == geo::kY){
        mean.SetY(mean.Y()+y*E);
        Ey += E;
      }
      mean.SetZ(mean.Z()+z*E);
    }

    if(Ex) mean.SetX(mean.X()/Ex);
    if(Ey) mean.SetY(mean.Y()/Ey);
    if(Ex || Ey) mean.SetZ(mean.Z()/(Ex+Ey));

    return mean;
  }

  DEFINE_ART_MODULE(MakeLibrary);

} //namespace lem
////////////////////////////////////////////////////////////////////////

