////////////////////////////////////////////////////////////////////////
// \file    MergeMatches_module.cc
// \brief   Merge LEM match lists from split running
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/FindMatches.h" // Just for the list size
#include "LEM/func/LibrarySummary.h"
#include "LEM/func/MatchList.h"

#include "RecoBase/Cluster.h"
#include "GeometryObjects/OfflineChan.h"
#include "Utilities/AssociationUtil.h"

#include "TH2.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"

#include <iostream>

namespace lem
{
  /// Merge %LEM match lists from split running
  class MergeMatches: public art::EDProducer
  {
  public:
    explicit MergeMatches(const fhicl::ParameterSet& pset);
    ~MergeMatches();

    virtual void beginRun(art::Run& run);
    virtual void produce(art::Event& evt);

    virtual void reconfigure(const fhicl::ParameterSet& pset);

  protected:
    std::string fSplitLabel;
  };

  //......................................................................
  MergeMatches::MergeMatches(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);

    produces<std::vector<lem::MatchList>>();
    // MatchLists are associated with the slice they are against
    produces<art::Assns<lem::MatchList, rb::Cluster>>();

    // Really should be file level, not run level, but don't know how to do
    // that
    produces<lem::LibrarySummary, art::InRun>();

    // Copies from one of the split parts
    produces<std::vector<geo::OfflineChan> >();

    produces<std::vector<TH2F>>(); // copies from best match part
  }

  //......................................................................
  MergeMatches::~MergeMatches()
  {
  }

  //......................................................................
  void MergeMatches::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSplitLabel = pset.get<std::string>("SplitLabel");
  }

  //......................................................................
  void MergeMatches::beginRun(art::Run& run)
  {
    std::vector<art::Handle<lem::LibrarySummary> > libsums;
    run.getManyByType(libsums);

    std::unique_ptr<lem::LibrarySummary> libout(new lem::LibrarySummary);

    for(unsigned int n = 0; n < libsums.size(); ++n){
      *libout += *libsums[n];
    }

    run.put(std::move(libout));
  }

  //......................................................................
  void MergeMatches::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<lem::MatchList>> outcol(new std::vector<lem::MatchList>);
    std::unique_ptr<std::vector<TH2F>> th2outcol(new std::vector<TH2F>);

    // Associate matches with the slice they came from
    std::unique_ptr<art::Assns<lem::MatchList, rb::Cluster>> assns(new art::Assns<lem::MatchList, rb::Cluster>);

    std::unique_ptr<std::vector<geo::OfflineChan>> chanoutcol(new std::vector<geo::OfflineChan>);


    std::vector<art::Handle<std::vector<lem::MatchList>>> matchcol;
    std::vector<art::Handle<std::vector<TH2F>>> th2col;
    std::vector<art::Handle<std::vector<geo::OfflineChan>>> chancol;

    evt.getManyByType(matchcol);
    // TODO: this is bad if anyone else puts TH2s or chans in the event
    evt.getManyByType(th2col);
    evt.getManyByType(chancol);
    assert(!matchcol.empty());
    assert(th2col.size() == matchcol.size());

    // Just copy them from the first split job
    chanoutcol->insert(chanoutcol->end(),
                       chancol[0]->begin(), chancol[0]->end());

    const unsigned int partMax = matchcol.size();
    const unsigned int sliceMax = matchcol[0]->size();

    std::cout << "Merging from " << partMax << " jobs over "
              << sliceMax << " slices" << std::endl;

    // Note: these are indexes into the list of slices that got PID'd: not all
    // slices.
    for(unsigned int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){

      std::vector<MatchSummary> matches;
      std::vector<MatchSummary> enrichedMatches;

      for(unsigned int partIdx = 0; partIdx < partMax; ++partIdx){
        assert(matchcol[partIdx]->size() == sliceMax);

        const MatchList& ms = (*matchcol[partIdx])[sliceIdx];

        matches.insert(matches.end(), ms.matches.begin(), ms.matches.end());
      }


      std::sort(matches.begin(), matches.end());
      matches.resize(kMaxNumMatches);

      int bestMatchPart = -1;
      const double bestPotential = matches[0].potential;
      for(unsigned int partIdx = 0; partIdx < partMax; ++partIdx){
        if((*matchcol[partIdx])[sliceIdx].matches.empty()) continue;
        if((*matchcol[partIdx])[sliceIdx].matches[0].potential == bestPotential){
          bestMatchPart = partIdx;
        }
      } // end for partIdx

      if(bestMatchPart >= 0){
        th2outcol->insert(th2outcol->end(),
                          th2col[bestMatchPart]->begin(),
                          th2col[bestMatchPart]->end());
      }
      else{
        mf::LogWarning("MergeMatches") << "Couldn't figure out which part had the best match to copy its histograms";
      }

      art::FindManyP<rb::Cluster> fmt(matchcol[0], evt, fSplitLabel);
      // Get all the slices associated with this match list
      std::vector<art::Ptr<rb::Cluster> > slice = fmt.at(sliceIdx);
      assert(slice.size() == 1);

      // Duplicate the same association in our output
      util::CreateAssn(*this, evt, *outcol, slice[0], *assns);
    }

    evt.put(std::move(outcol));
    evt.put(std::move(assns));
    evt.put(std::move(th2outcol));
    evt.put(std::move(chanoutcol));
  }

  DEFINE_ART_MODULE(MergeMatches);

} //namespace lem
////////////////////////////////////////////////////////////////////////

