////////////////////////////////////////////////////////////////////////
// \file    UnTranspose_module.cc
// \brief   TODO
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "LEM/func/LibrarySummary.h"
#include "LEM/func/MatchList.h"

#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "Utilities/AssociationUtil.h"

namespace lem
{
  class UnTranspose: public art::EDProducer
  {
  public:
    explicit UnTranspose(const fhicl::ParameterSet& pset);
    ~UnTranspose();

    virtual void beginRun(art::Run& run);

    void produce(art::Event& evt);

    void reconfigure(const fhicl::ParameterSet& pset);

    void beginJob();

  protected:
    std::string fSlicerLabel;
    std::string fMatchLabel;

    std::vector<lem::MatchList> fMatchLists;
    unsigned int fMatchIdx;
  };

  //.......................................................................
  UnTranspose::UnTranspose(const fhicl::ParameterSet& pset)
    : fMatchIdx(0)
  {
    reconfigure(pset);

    produces<lem::LibrarySummary, art::InRun>();

    produces<std::vector<lem::MatchList>>();
    // MatchLists are associated with the slice they are against
    produces<art::Assns<lem::MatchList, rb::Cluster>>();
  }

  //......................................................................
  UnTranspose::~UnTranspose()
  {
  }

  //......................................................................
  void UnTranspose::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlicerLabel = pset.get<std::string>("SlicerLabel");
    fMatchLabel = pset.get<std::string>("MatchLabel");
  }

  //......................................................................
  void UnTranspose::beginJob()
  {
  }

  //......................................................................
  void UnTranspose::beginRun(art::Run& run)
  {
    art::Handle<std::vector<lem::MatchList>> ms;
    run.getByLabel(fMatchLabel, ms);
    fMatchLists = *ms;
    fMatchIdx = 0;

    // Copy library summary
    art::Handle<lem::LibrarySummary> ls;
    run.getByLabel(fMatchLabel, ls);
    std::unique_ptr<lem::LibrarySummary> lsout(new lem::LibrarySummary(*ls));
    run.put(std::move(lsout));
  }

  //......................................................................
  void UnTranspose::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<lem::MatchList> > matchcol(new std::vector<lem::MatchList>);

    // Associate matches with the slice they came from
    std::unique_ptr<art::Assns<lem::MatchList, rb::Cluster> > assns(new art::Assns<lem::MatchList, rb::Cluster>);

    art::Handle<std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    const int sliceMax = slices->size();
    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      if(rb::IsFiltered(evt, slices, sliceIdx)) continue;

      const rb::Cluster& slice = (*slices)[sliceIdx];

      if(slice.IsNoise()) continue;

      assert(fMatchIdx < fMatchLists.size());
      const lem::MatchList& ml = fMatchLists[fMatchIdx++];
      // Make sure we're still in sync
      assert(ml.trial.run == int(evt.run()));
      assert(ml.trial.subrun == int(evt.subRun()));
      assert(ml.trial.event == int(evt.event()));

      matchcol->push_back(ml);

      util::CreateAssn(*this, evt, *matchcol,
                       art::Ptr<rb::Cluster>(slices, sliceIdx), *assns);
    } // end for sliceIdx

    evt.put(std::move(matchcol));
    evt.put(std::move(assns));
  }

  DEFINE_ART_MODULE(UnTranspose);

} // end namespace lem
////////////////////////////////////////////////////////////////////////

