////////////////////////////////////////////////////////////////////////
/// \file    DecisionTree.cxx
/// \brief   Decision Tree PID
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/DecisionTree.h"

#include "Utilities/func/MathUtil.h"

#include "TFile.h"
#include "TRandom3.h"
#include "TTree.h"
#include "TVectorD.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <map>
#include <iostream>

#include "pthread.h"

namespace lem
{
  namespace dec
  {
    // For serializing to/from a TTree. Anonymous namespace, no-one else needs
    // to see these
    namespace
    {
      int gThis;
      int gLeft, gRight;
      int gCutDim;
      double gCutVal;
      double gSig, gBkg;
      // Maps memory addresses to rows in the TTree, allowing serialization of pointers
      std::map<const TreeNode*, int> gPtrs;
    }

    //......................................................................
    Forest::~Forest()
    {
      for(Tree* t: fTrees) delete t;
    }

    //......................................................................
    double Forest::Classify(const Evt& evt) const
    {
      // Average all the constituent trees
      double mean = 0;
      for(Tree* t: fTrees) mean += t->Classify(evt);
      mean /= fTrees.size();
      return mean;
    }

    //......................................................................
    Forest* Forest::FromFile(const std::string& fname)
    {
      TFile f(fname.c_str());
      assert(!f.IsZombie());
      TTree* tr = (TTree*)f.Get("tree");
      assert(tr);

      tr->SetBranchAddress("this", &gThis);
      tr->SetBranchAddress("left", &gLeft);
      tr->SetBranchAddress("right", &gRight);
      tr->SetBranchAddress("cutdim", &gCutDim);
      tr->SetBranchAddress("cutval", &gCutVal);
      tr->SetBranchAddress("sig", &gSig);
      tr->SetBranchAddress("bkg", &gBkg);

      // Map from TTree rows to pointers, use to connect up nodes
      std::map<int, TreeNode*> ptrs;

      // For every row in the tree
      const int N = tr->GetEntries();
      for(int n = 0; n < N; ++n){
        tr->GetEntry(n);

        if(gLeft == -1){
          assert(gRight == -1);

          // If it has no children it's a simple Leaf
          ptrs[gThis] = new Leaf(gSig, gBkg);
        }
        else{
          assert(gRight != -1);

          // If it does have children, it's a Cut. We should have seen its
          // children already, so look them up in the map.

          TreeNode* left = ptrs[gLeft];
          TreeNode* right = ptrs[gRight];
          assert(left && right);

          // Only retain unparented nodes in the map
          ptrs.erase(ptrs.find(gLeft));
          ptrs.erase(ptrs.find(gRight));

          ptrs[gThis] = new Cut(gCutDim, gCutVal, left, right);
        }
      }

      // Anything that never found a parent is a tree root
      std::cout << "Loaded " << ptrs.size() << " trees" << std::endl;
      Forest* ret = new Forest;

      int matIdx = 0;
      // Fallback in case file is old
      TMatrixD unit(kNumPIDVars+1, kNumPIDVars+1);
      unit.UnitMatrix();

      for(auto it: ptrs){
        TMatrixD* mat = (TMatrixD*)f.Get(TString::Format("mat_%d", matIdx++));
        ret->fTrees.push_back(new Tree(it.second, mat ? *mat : unit));
      }

      return ret;
    }

    //......................................................................
    void Forest::ToFile(const std::string& fname)
    {
      TFile fout(fname.c_str(), "RECREATE");
      TTree tr("tree", "tree");

      // The ToTree functions of the individual nodes require us to have done this setup
      tr.Branch("this", &gThis);
      tr.Branch("left", &gLeft);
      tr.Branch("right", &gRight);
      tr.Branch("cutdim", &gCutDim);
      tr.Branch("cutval", &gCutVal);
      tr.Branch("sig", &gSig);
      tr.Branch("bkg", &gBkg);

      // And they maintain the state of gPtrs
      gPtrs.clear();

      for(unsigned int i = 0; i < fTrees.size(); ++i){
        fTrees[i]->ToTree(&tr);
        fTrees[i]->GetMatrix().Write(TString::Format("mat_%u", i));
      }

      tr.Write("tree");
    }

    //......................................................................
    TMatrixD Forest::RandomOrthoMatrix()
    {
      // Start with the identity matrix
      TMatrixD ret(kNumPIDVars+1, kNumPIDVars+1);
      ret.UnitMatrix();

      TRandom3 r(0); // auto-seed

      // Random rotation by all the Euler angles
      for(int i = 0; i < kNumPIDVars-1; ++i){
        for(int j = i+1; j < kNumPIDVars; ++j){
          TMatrixD mul(kNumPIDVars+1, kNumPIDVars+1);
          mul.UnitMatrix();

          double ang = r.Uniform(2*M_PI);
          mul(i, i) = mul(j, j) = cos(ang);
          mul(i, j) = +sin(ang);
          mul(j, i) = -sin(ang);

          ret = mul*ret;
        }
      }

      return ret;
    }

    //......................................................................
    Evt TransformEvent(const Evt& evt, const TMatrixD& mat)
    {
      TVectorD vec(kNumPIDVars+1);
      for(int i = 0; i < kNumPIDVars; ++i) vec[i] = evt.vars[i];
      vec[kNumPIDVars] = 1;
      vec = mat*vec;

      Evt ret = evt;
      for(int i = 0; i < kNumPIDVars; ++i) ret.vars[i] = vec[i];
      return ret;
    }

    // Boilerplate required to call thread function with the correct arguments
    struct ThreadArgs
    {
      const std::vector<Evt>* evts;
      TMatrixD scale, trans;
    };

    void* thread_func(void* a)
    {
      ThreadArgs* args = (ThreadArgs*)a;
      return Forest::TrainSingle(*args->evts, args->scale, args->trans);
    }

    //......................................................................
    Forest Forest::Train(std::vector<Evt>& trainEvts, unsigned int nTrees,
                         bool parallel)
    {
      Forest ret;

      const unsigned int N = trainEvts.size();

      // Because we only take half the sample each time below
      for(unsigned int n = 0; n < N; ++n) trainEvts[n].weight *= 2;

      // Accumulate the mean and RMS for each variable
      double mean[kNumPIDVars] = {0,};
      double rms[kNumPIDVars] = {0,};
      double W = 0;
      for(unsigned int n = 0; n < N; ++n){
        const Evt& e = trainEvts[n];
        for(int i = 0; i < kNumPIDVars; ++i){
          if(fabs(e.vars[i]) < 10){ // There are some crazy values in there
            mean[i] += e.weight*e.vars[i];
            rms[i] += e.weight*util::sqr(e.vars[i]);
          }
        }
        W += e.weight;
      }

      // Translation matrix. To learn how this works, read
      // http://en.wikipedia.org/wiki/Translation_matrix
      TMatrixD transMat(kNumPIDVars+1, kNumPIDVars+1);
      transMat.UnitMatrix();

      // Matrix to normalize coordinates into sigmas
      TMatrixD scaleMat(kNumPIDVars+1, kNumPIDVars+1);
      scaleMat(kNumPIDVars, kNumPIDVars) = 1;

      for(int i = 0; i < kNumPIDVars; ++i){
        mean[i] /= W;
        rms[i] = rms[i]/W-util::sqr(mean[i]);

        transMat(i, kNumPIDVars) = -mean[i];
        scaleMat(i, i) = 1/sqrt(rms[i]);
      }


      if(parallel){
        // Somewhat oversubscribe the number of cores, to keep them all busy
        const int nThreads = sysconf(_SC_NPROCESSORS_ONLN)*1.5;

        std::vector<pthread_t*> ths;
        int live = 0; // counter how many threads are currently running

        ThreadArgs args = {&trainEvts, scaleMat, transMat};

        while(true){
          // Start new threads if needed
          while(live < nThreads && ret.fTrees.size()+live < nTrees){
            ths.push_back(new pthread_t);
            pthread_create(ths.back(), 0, &thread_func, &args);
            ++live;
          }

          // Break when we're done
          if(live == 0 && ret.fTrees.size() >= nTrees) break;

          // Have to busy-wait because sleep() suspends all the threads

          // Check all the threads to see if they've finished
          for(unsigned int i = 0; i < ths.size(); ++i){
            void* tr;
            if(ths[i] && pthread_tryjoin_np(*ths[i], &tr) == 0){
              // If so, clean them up and store their results
              delete ths[i];
              ths[i] = 0;
              ret.fTrees.push_back((Tree*)tr);
              --live;
              std::cout << "." << std::flush;
            }
          } // end for i
        } // end while
        std::cout << std::endl;
      }
      else{
        for(unsigned int n = 0; n < nTrees; ++n){
          ret.fTrees.push_back(TrainSingle(trainEvts, scaleMat, transMat));
          std::cout << "." << std::flush;
        }
        std::cout << std::endl;
      }

      // Try to put things back as we found them
      for(unsigned int n = 0; n < N; ++n) trainEvts[n].weight /= 2;

      return ret;
    }

    //......................................................................
    Tree* Forest::TrainSingle(const std::vector<Evt>& trainEvts,
                              const TMatrixD& scaleMat,
                              const TMatrixD& transMat)
    {
      // To avoid linear artifacts in the output, we rotate into a random
      // basis. Note that this is different for each tree in the forest.
      const TMatrixD rotMat = RandomOrthoMatrix();

      // Before rotating, center and normalize the variables so they're all
      // treated on an equal footing.
      const TMatrixD mat = rotMat * scaleMat * transMat;

      const unsigned int N = trainEvts.size();

      std::vector<Evt> transformedEvts;
      for(unsigned int i = 0; i < N; ++i){
        transformedEvts.push_back(TransformEvent(trainEvts[i], mat));
      }

      // Shuffle the training events and take the first half. This ensures
      // each tree gets a different sample, so on average they don't
      // overtrain.
      std::random_shuffle(transformedEvts.begin(), transformedEvts.end());

      // The training algorithm requires multiple lists of the events, sorted
      // according to each PID variable.
      std::vector<std::list<Evt*>> sorted(kNumPIDVars);
      for(int v = 0; v < kNumPIDVars; ++v){
        for(unsigned int i = 0; i < N/2; ++i){
          sorted[v].push_back(&transformedEvts[i]);
        }

        sorted[v].sort([v](const Evt* a, const Evt* b)
                       {
                         return a->vars[v] < b->vars[v];
                       });
      }

      // Train the tree and add it to the forest
      TreeNode* tree = TrainSingleTransformed(sorted, 0);
      std::vector<Evt> test(transformedEvts.begin()+N/2, transformedEvts.end());
      tree = tree->Prune(test);
      return new Tree(tree, mat);
    }

    /// Don't build a tree deeper than this
    const unsigned int kDepthLimit = 20;
    /// Once there are less than this many events in a subtree, stop subdividing
    const unsigned int kMinBucketSize = 20;

    //......................................................................
    TreeNode* Forest::
    TrainSingleTransformed(std::vector<std::list<Evt*> >& sorted,
                           unsigned int depth)
    {
      const unsigned int N = sorted[0].size();

      double totSig = 0;
      double totBkg = 0;

      for(Evt* evt : sorted[0]){
        const double w = evt->weight;
        if(evt->isSig) totSig += w; else totBkg += w;
      }

      // Variety of reasons not to try to cut anymore
      if(depth == kDepthLimit || N < kMinBucketSize || totSig == 0 || totBkg == 0)
        return new Leaf(totSig, totBkg);

      double bestFOMSq = 0;
      // Convention that the invalid iterator is taken from the first list
      std::list<Evt*>::iterator bestCutIt = sorted[0].end();
      int bestCutDim = 0;

      // Try a cut in every dimension
      for(int dim = 0; dim < kNumPIDVars; ++dim){
        // Counts below the cut
        double s1 = 0;
        double b1 = 0;
        for(std::list<Evt*>::iterator it = sorted[dim].begin(); it != sorted[dim].end(); ++it){
          // Keep the counts below and above up to date
          const double w = (*it)->weight;
          if((*it)->isSig) s1 += w; else b1 += w;

          const double s2 = totSig-s1;
          const double b2 = totBkg-b1;

          // Figure of merit if we were to cut here
          const double fomSq = util::sqr(s1)/(s1+b1)+util::sqr(s2)/(s2+b2);

          if(fomSq > bestFOMSq){
            bestFOMSq = fomSq;
            if(it != sorted[dim].begin()){
              // Definition of the cut is the last iterator in the "left" sample
              bestCutIt = it;
              --bestCutIt;
            }
            bestCutDim = dim;
          }
        } // end for n
      } // end for dim

      // Didn't find a cut, or it's no better
      if(bestCutIt == sorted[0].end() ||
         bestFOMSq <= util::sqr(totSig)/(totSig+totBkg))
        return new Leaf(totSig, totBkg);

      // First iterator in the "right" sample
      std::list<Evt*>::iterator next = bestCutIt;
      ++next;

      // Cut is halfway between the two points we want to seperate
      const double cutPos = ((*bestCutIt)->vars[bestCutDim]+(*next)->vars[bestCutDim])/2;

      // Children require the same sorted lists that we got
      std::vector<std::list<Evt*> > sortedLeft(kNumPIDVars);
      std::vector<std::list<Evt*> > sortedRight(kNumPIDVars);

      for(int v = 0; v < kNumPIDVars; ++v){
        for(Evt* evt : sorted[v]){
          if(evt->vars[bestCutDim] < cutPos)
            sortedLeft[v].push_back(evt);
          else
            sortedRight[v].push_back(evt);
        }

        // Make sure to clear lists out once we don't need them, save memory
        // while recursing.
        sorted[v].clear();
      }
      sorted.clear();

      TreeNode* left = TrainSingleTransformed(sortedLeft, depth+1);
      TreeNode* right = TrainSingleTransformed(sortedRight, depth+1);
      return new Cut(bestCutDim, cutPos, left, right);
    }

    //......................................................................
    double Tree::Classify(const Evt& e) const
    {
      return fHead->Classify(TransformEvent(e, fMatrix));
    }

    //......................................................................
    Cut::Cut(int d, double v, TreeNode* l, TreeNode* r)
      : fCutDim(d), fCutVal(v), fLeft(l), fRight(r)
    {
      assert(fLeft->NSig()+fRight->NSig() > 0);
      assert(fLeft->NBkg()+fRight->NBkg() > 0);
    }

    //......................................................................
    double Cut::FOM() const
    {
      // FOM is the sum in quadrature of our children
      return sqrt(util::sqr(fLeft->FOM())+util::sqr(fRight->FOM()));
    }

    //......................................................................
    double Cut::Classify(const Evt& e) const
    {
      if(e.vars[fCutDim] < fCutVal){
        return fLeft->Classify(e);
      }
      else{
        return fRight->Classify(e);
      }
    }

    //......................................................................
    double Cut::NSig() const
    {
      return fLeft->NSig()+fRight->NSig();
    }

    //......................................................................
    double Cut::NBkg() const
    {
      return fLeft->NBkg()+fRight->NBkg();
    }

    //......................................................................
    TreeNode* Cut::Prune(std::vector<Evt>& evts)
    {
      std::vector<Evt> leftEvts, rightEvts;
      const unsigned int N = evts.size();
      double sigL = 0, bkgL = 0, sigR = 0, bkgR = 0;
      for(unsigned int n = 0; n < N; ++n){
        const Evt& e = evts[n];
        if(e.vars[fCutDim] < fCutVal){
          leftEvts.push_back(e);
          if(e.isSig) sigL += e.weight; else bkgL += e.weight;
        }
        else{
          rightEvts.push_back(e);
          if(e.isSig) sigR += e.weight; else bkgR += e.weight;
        }
      }
      evts.clear();

      fLeft = fLeft->Prune(leftEvts);
      fRight = fRight->Prune(rightEvts);

      // If originally trained with higher s/b on the right but test sample has
      // it higher on the left, or vice versa, then overtrained, merge the
      // trees.
      const bool trainSign = fLeft->NSig()*fRight->NBkg() < fRight->NSig()*fLeft->NBkg();
      const bool testSign = sigL*bkgR < sigR*bkgL;

      if(fLeft->IsLeaf() && fRight->IsLeaf() && testSign != trainSign){
        TreeNode* ret = new Leaf(fLeft->NSig()+fRight->NSig(),
                                 fLeft->NBkg()+fRight->NBkg());
        delete this;
        return ret;
      }

      return this;
    }

    //......................................................................
    void Cut::ToTree(TTree* tr) const
    {
      // Recurse first so there'll be entries in gPtrs to look left and right up in
      fLeft->ToTree(tr);
      fRight->ToTree(tr);

      gPtrs[this] = tr->GetEntries();

      gThis = gPtrs[this];
      gLeft = gPtrs[fLeft];
      gRight = gPtrs[fRight];
      gCutDim = fCutDim;
      gCutVal = fCutVal;
      gSig = -1;
      gBkg = -1;

      tr->Fill();
    }


    //......................................................................
    Leaf::Leaf(double s, double b) : fSig(s), fBkg(b)
    {
    }

    //......................................................................
    double Leaf::FOM() const
    {
      if(fSig == 0) return 0;
      return fSig/sqrt(fSig+fBkg);
    }

    //......................................................................
    double Leaf::Classify(const Evt&) const
    {
      if(fSig == 0) return 0;
      return fSig/(fSig+fBkg);
    }

    //......................................................................
    void Leaf::ToTree(TTree* tr) const
    {
      gPtrs[this] = tr->GetEntries();

      gThis = gPtrs[this];
      gLeft = -1;
      gRight = -1;
      gCutDim = -1;
      gCutVal = -1;
      gSig = fSig;
      gBkg = fBkg;

      tr->Fill();
    }

  } // namespace dec
} // namespace lem

