////////////////////////////////////////////////////////////////////////
/// \file    DistanceMap.cxx
/// \brief   Calculate and cache electrostatic potential between cells
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/DistanceMap.h"

#include "TH2.h"

#include <cstdlib>

#include "Utilities/func/MathUtil.h"


namespace lem
{
  // Best aspect ratio empirically. Matches fcl default
  double DistanceMap::fgPlaneScale = 3.5;
  double DistanceMap::fgCellScale = 10.5;
  // 1/sqrt(dr). Best power empirically. EM would be -1
  double DistanceMap::fgDecayPower = -0.25;
  bool DistanceMap::fgExpMode = false;

  //......................................................................
  const DistanceMap& DistanceMap::Instance()
  {
    static DistanceMap instance;
    return instance;
  }

  //......................................................................
  DistanceMap::DistanceMap()
  {
    for(int p = 0; p < kNumPlanes; ++p){
      for(int c = 0; c < kNumCells; ++c){
        // For long enough distances don't bother to do the integral over cells
        if(p > 5 || c > 5){
          fInvDistMap[p][c] = DistFunc(p, c);
          continue;
        }

        double E = 0;

        const int kNumSteps = 10;
        for(int dp1 = 0; dp1 < kNumSteps; ++dp1){
          for(int dp2 = 0; dp2 < kNumSteps; ++dp2){
            const double dp = (dp1-dp2)/double(kNumSteps);
            for(int dc1 = 0; dc1 < kNumSteps; ++dc1){
              for(int dc2 = 0; dc2 < kNumSteps; ++dc2){
                // Avoid division by zero
                if(p == 0 && c == 0 && dp1 == dp2 && dc1 == dc2) continue;

                const double dc = (dc1-dc2)/double(kNumSteps);

                E += DistFunc(p+dp, c+dc);
              }
            }
          }
        } // end for dp1

        // Don't count the terms that we continue'd out of
        int count = util::sqr(util::sqr(kNumSteps));
        if(p == 0 && c == 0) count -= util::sqr(kNumSteps);

        fInvDistMap[p][c] = E/count;
      } // end for c
    } // end for p
  }

  //......................................................................
  TH2* DistanceMap::MakeHist() const
  {
    TH2* hDistMap = new TH2F("", ";#Deltaplanes;#Deltacells", 5, 0, 5, 5, 0, 5);

    for(int p = 0; p < kNumPlanes; ++p){
      for(int c = 0; c < kNumCells; ++c){
        hDistMap->Fill(p, c, 1/fInvDistMap[p][c]);
      } // end for c
    } // end for p

    return hDistMap;
  }

  //......................................................................
  double DistanceMap::DistFunc(double dp, double dc)
  {
    const double dr = util::pythag(fgPlaneScale*dp, fgCellScale*dc);

    if(fgExpMode){
      // Can control the coefficient with planeScale and cellScale
      return exp(-dr);
    }

    return pow(dr, fgDecayPower);
  }

  //......................................................................
  double DistanceMap::InvDist(int aplane, int bplane, int acell, int bcell) const
  {
    const int dp = abs(aplane-bplane);
    const int dc = abs(acell-bcell);

    // Very far away
    if(dp < 0) return 0;
    if(dp >= kNumPlanes) return 0;
    if(dc < 0) return 0;
    if(dc >= kNumCells) return 0;

    // Comparison between different views shouldn't happen
    assert(dp%2 == 0);

    return fInvDistMap[dp][dc];
  }

  //......................................................................
  double DistanceMap::InvDist(const lem::LiteHit& a, const lem::LiteHit& b) const
  {
    return InvDist(a.Plane(), b.Plane(), a.Cell(), b.Cell());
  }

} // namespace
