////////////////////////////////////////////////////////////////////////
/// \file    EventSummary.cxx
/// \brief   Simple representation of event for LEM use
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/EventSummary.h"

#include "TCanvas.h"
#include "TH2.h"
#include "TPad.h"
#include "TTree.h"

#include <algorithm>
#include <cassert>
#include <map>

#include "pthread.h"

#include "Utilities/func/MathUtil.h"

namespace lem
{
  double EventSummary::fgDecayRate = 0;
  double EventSummary::fgChargePower = .5;


  //......................................................................
  EventSummary::EventSummary(const std::vector<LiteHit>& hs, int i, std::vector<int> pdgs)
    : id(i), photL0(0), photL1(0), photE0(0), photE1(0), enrich(false)
  {
    nhits = hs.size();
    hits = new LiteHit[nhits];
    for(unsigned int i = 0; i < nhits; ++i) hits[i] = hs[i];

    nParts = pdgs.size();
    partPdgs = new int[nParts];
    for(unsigned int i = 0; i < nParts; ++i) partPdgs[i] = pdgs[i];

    // TODO: the ordering of this with charge decay below makes a difference
    if(fgChargePower != 1){
      for(unsigned int i = 0; i < nhits; ++i){
        hits[i].pecorr = pow(hits[i].pecorr, fgChargePower);
      }
    }

    if(fgDecayRate != 0){
      for(unsigned int i = 0; i < nhits; ++i){
        LiteHit& h = hits[i];

        const int planeDelta = h.Plane()-kVertexPlane;
        const int cellDelta = h.Cell()-kVertexCell;

        // TODO: probably this can be optimized...
        const double dist = util::pythag(planeDelta, cellDelta);

        const double decay = exp(-fgDecayRate*dist);

        h.pecorr *= decay;
      } // end for i
    } // end if nonzero

    Finalize();
  }

  //......................................................................
  EventSummary::EventSummary(const EventSummary& evt)
  {
    *this = evt;
  }

  //......................................................................
  EventSummary::~EventSummary()
  {
    delete[] hits;
    delete[] partPdgs;
  }

  //......................................................................
  EventSummary& EventSummary::operator=(const EventSummary& evt)
  {
    // TODO: it would be much nicer for most of this to come from a default
    // implementation.

    nhits = evt.nhits;
    hits = new LiteHit[nhits];
    for(int i = 0; i < nhits; ++i) hits[i] = evt.hits[i];

    totalPE = evt.totalPE;
    trueEVis = evt.trueEVis;
    y = evt.y;
    id = evt.id;

    run = evt.run;
    subrun = evt.subrun;
    event = evt.event;

    ccnc = evt.ccnc;
    pdg = evt.pdg;
    origPdg = evt.origPdg;
    mode = evt.mode;

    trueVtxPlane = evt.trueVtxPlane;
    trueVtxCell = evt.trueVtxCell;
    trueVtxCellOther = evt.trueVtxCellOther;

    nParts = evt.nParts;
    partPdgs = new int[nParts];
    for(int i = 0; i < nParts; ++i) partPdgs[i] = evt.partPdgs[i];

    photL0 = evt.photL0;
    photL1 = evt.photL1;
    photE0 = evt.photE0;
    photE1 = evt.photE1;

    enrich = evt.enrich;

    return *this;
  }

  //......................................................................
  void EventSummary::FillHists(TH2* h1, TH2* h2, bool flipEven, bool flipOdd, bool idxs) const
  {
    TH2* h[2] = {h1, h2};

    for(unsigned int i = 0; i < nhits; ++i){
      const int plane = hits[i].Plane();
      int cell = hits[i].Cell();
      const int v = plane%2;
      if(v == 0 && flipEven) cell = 255-cell;
      if(v == 1 && flipOdd) cell = 255-cell;
      if(idxs){
        h[plane%2]->Fill(plane, cell, hits[i].partIdx);
      }
      else{
        h[plane%2]->Fill(plane, cell, hits[i].pecorr);
      }
    }
    h[0]->SetMaximum(std::max(h[0]->GetMaximum(), h[1]->GetMaximum()));
    h[1]->SetMaximum(std::max(h[0]->GetMaximum(), h[1]->GetMaximum()));
  }

  //......................................................................
  std::string EventSummary::Description() const
  {
    std::map<int, const char*> partNames;
    partNames[12] = "#nu_{e}";
    partNames[14] = "#nu_{#mu}";
    partNames[16] = "#nu_{#tau}";
    partNames[-12] = "#bar{#nu}_{e}";
    partNames[-14] = "#bar{#nu}_{#mu}";
    partNames[-16] = "#bar{#nu}_{#tau}";

    return TString::Format("%s %s#rightarrow%s, y=%.2lf, %dPE",
                           ccnc ? "NC" : "CC",
                           partNames[origPdg],
                           partNames[pdg],
                           y,
                           int(totalPE)).Data();
  }

  //......................................................................
  void EventSummary::SaveImage(TString fname, TString title) const
  {
    static pthread_mutex_t lock;
    pthread_mutex_lock(&lock);

    TCanvas c("a", "b", 1200, 600);
    c.Divide(2, 1);

    TH2F h[2] = {TH2F("", ";plane;cell", 128, 32, 160, 128, 64, 192),
                 TH2F("", ";plane;cell", 128, 32, 160, 128, 64, 192)};

    h[0].SetTitle(Description().c_str());

    h[1].SetTitle(title);

    FillHists(&h[0], &h[1]);

    for(int v = 0; v < 2; ++v){
      c.cd(v+1);
      h[v].Draw("colz");
    }

    c.cd(0);
    gPad->Print(fname);

    pthread_mutex_unlock(&lock);
  }

  // Anonymous namespace, no-one else needs to see these
  namespace{
    // globals for tree
    int gRun, gSubrun, gEvent;
    int gPdg, gOrigPdg, gCCNC, gMode;
    double gTrueEVis, gY;
    int gNHits;
    unsigned short gCellIdx[256*256];
    float gPECorr[256*256];
    unsigned char gPartIdx[256*256];
    int gNParts;
    int gPartPdgs[256];
    double gPhotL0, gPhotL1;
    double gPhotE0, gPhotE1;

    int gTrueVtxPlane, gTrueVtxCell, gTrueVtxCellOther;
  }

  //......................................................................
  void EventSummary::InitToTree(TTree* tr)
  {
    // Bookkeeping
    tr->Branch("run", &gRun);
    tr->Branch("subrun", &gSubrun);
    tr->Branch("event", &gEvent);

    // Truth information
    tr->Branch("pdg", &gPdg);
    tr->Branch("origpdg", &gOrigPdg);
    tr->Branch("ccnc", &gCCNC);
    tr->Branch("mode", &gMode);
    tr->Branch("trueEVis", &gTrueEVis);
    tr->Branch("y", &gY);

    // Length of the arrays
    tr->Branch("nhits", &gNHits);

    // Hit positions and charges
    tr->Branch("cellIdx", &gCellIdx, "cellIdx[nhits]/s");
    tr->Branch("pecorr", &gPECorr, "pecorr[nhits]/F");
    tr->Branch("partIdx", &gPartIdx, "partIdx[nhits]/b");

    // True Vertex
    tr->Branch("trueVtxPlane", &gTrueVtxPlane);
    tr->Branch("trueVtxCell", &gTrueVtxCell);
    tr->Branch("trueVtxCellOther", &gTrueVtxCellOther);

    // Particles
    tr->Branch("nParts", &gNParts);
    tr->Branch("partPdgs", &gPartPdgs, "partPdgs[nParts]/I");

    // Photons
    tr->Branch("photL0", &gPhotL0);
    tr->Branch("photL1", &gPhotL1);
    tr->Branch("photE0", &gPhotE0);
    tr->Branch("photE1", &gPhotE1);
  }

  //......................................................................
  void EventSummary::ToTree(TTree* tr) const
  {
    gRun = run;
    gSubrun = subrun;
    gEvent = event;

    gPdg = pdg;
    gOrigPdg = origPdg;
    gCCNC = ccnc;
    gMode = mode;

    gTrueEVis = trueEVis;
    gY = y;

    assert(nhits < 256*256);
    gNHits = nhits;

    for(unsigned int n = 0; n < nhits; ++n){
      gCellIdx[n] = hits[n].cellIdx;
      gPECorr[n] = hits[n].pecorr;
      gPartIdx[n] = hits[n].partIdx;
    }

    gNParts = nParts;
    assert(gNParts <= 255);
    for(int n = 0; n < gNParts; ++n) gPartPdgs[n] = partPdgs[n];

    gTrueVtxPlane = trueVtxPlane;
    gTrueVtxCell = trueVtxCell;
    gTrueVtxCellOther = trueVtxCellOther;

    gPhotL0 = photL0;
    gPhotL1 = photL1;
    gPhotE0 = photE0;
    gPhotE1 = photE1;

    tr->Fill();
  }

  //......................................................................
  void EventSummary::InitFromTree(TTree* tr)
  {
    // Bookkeeping
    tr->SetBranchAddress("run", &gRun);
    tr->SetBranchAddress("subrun", &gSubrun);
    tr->SetBranchAddress("event", &gEvent);

    // Truth
    tr->SetBranchAddress("pdg", &gPdg);
    tr->SetBranchAddress("origpdg", &gOrigPdg);
    tr->SetBranchAddress("ccnc", &gCCNC);
    tr->SetBranchAddress("mode", &gMode);
    tr->SetBranchAddress("trueEVis", &gTrueEVis);
    tr->SetBranchAddress("y", &gY);

    // Hit positions and charges
    tr->SetBranchAddress("nhits", &gNHits);
    tr->SetBranchAddress("cellIdx", &gCellIdx);
    tr->SetBranchAddress("pecorr", &gPECorr);
    tr->SetBranchAddress("partIdx", &gPartIdx);

    // Particles
    tr->SetBranchAddress("nParts", &gNParts);
    tr->SetBranchAddress("partPdgs", &gPartPdgs);

    // True vertex
    tr->SetBranchAddress("trueVtxPlane", &gTrueVtxPlane);
    tr->SetBranchAddress("trueVtxCell", &gTrueVtxCell);
    tr->SetBranchAddress("trueVtxCellOther", &gTrueVtxCellOther);

    // Photons
    tr->SetBranchAddress("photL0", &gPhotL0);
    tr->SetBranchAddress("photL1", &gPhotL1);
    tr->SetBranchAddress("photE0", &gPhotE0);
    tr->SetBranchAddress("photE1", &gPhotE1);
  }

  //......................................................................
  EventSummary EventSummary::FromTree(TTree* tr, int evtIdx, bool enrich)
  {
    tr->GetEntry(evtIdx);

    std::vector<LiteHit> hs;
    hs.reserve(gNHits);
    for(int n = 0; n < gNHits; ++n){
      lem::LiteHit h(0, 0, gPECorr[n], gPartIdx[n]);
      h.cellIdx = gCellIdx[n];

      hs.push_back(h);
    } // end for n

    std::vector<int> pdgs(gNParts);
    for(int n = 0; n < gNParts; ++n) pdgs[n] = gPartPdgs[n];

    EventSummary ret(hs, -1, pdgs);

    ret.ccnc = gCCNC;
    ret.pdg = gPdg;
    ret.origPdg = gOrigPdg;
    ret.mode = gMode;
    ret.trueEVis = gTrueEVis;
    ret.y = gY;

    ret.trueVtxPlane = gTrueVtxPlane;
    ret.trueVtxCell = gTrueVtxCell;
    ret.trueVtxCellOther = gTrueVtxCellOther;

    ret.photL0 = gPhotL0;
    ret.photL1 = gPhotL1;
    ret.photE0 = gPhotE0;
    ret.photE1 = gPhotE1;

    ret.enrich = enrich;

    return ret;
  }

  //......................................................................
  void EventSummary::Finalize()
  {
    totalPE = 0;

    for(unsigned int i = 0; i < nhits; ++i){
      LiteHit& h = hits[i];
      totalPE += h.pecorr;
    } // end for i

    // Put in order of cellIdx, might improve locality
    std::sort(hits, hits+nhits);
  }

  //......................................................................
  int EventSummary::Length() const
  {
    if(nhits < 2) return 0;
    // Hits are sorted
    return hits[nhits-1].Plane() - hits[0].Plane();
  }
}
