////////////////////////////////////////////////////////////////////////
/// \file    Heads.cxx
/// \brief   Mapping from a subset of the library to their best matches
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/Heads.h"

#include "LEM/func/OnDemand.h"
#include "LEM/func/Util.h"

#include <cassert>
#include <sys/mman.h>

namespace lem
{
  // Private to this file
  namespace{
    char* kHeadsAddr = (char*)0x00000300CAFE0000;
  }

  //......................................................................
  Heads* Heads::FromMMap(const std::string& libPath, bool touchAll)
  {
    Heads* heads = new Heads;

    heads->fMappingSize = MMapFileAtAddress(libPath+"/heads_map.bin", kHeadsAddr,
                                            touchAll, "/tmp/lem_preload_heads.lock");

    heads->fNHeads = ((int*)kHeadsAddr)[0];
    heads->fHeadSeqsLen = ((int*)kHeadsAddr)[1];

    heads->fHeadIdxs = ((int*)kHeadsAddr)+2;

    heads->fHeadSeqs = new int*[heads->fNHeads];
    heads->fHeadSeqs[0] = ((int*)kHeadsAddr)+2+heads->fNHeads;
    for(long i = 0; i < heads->fNHeads; ++i){
      heads->fHeadSeqs[i] = heads->fHeadSeqs[0]+heads->fHeadSeqsLen*i;

      // If "i" is an int, the calculation overflows. Check nothing similar is
      // reintroduced.
      assert((char*)heads->fHeadSeqs[i] > kHeadsAddr);
    }

    return heads;
  }

  //......................................................................
  Heads* Heads::FromMMapOnDemand(const std::string& libPath)
  {
    CreateOnDemandMapping((libPath+"/heads_map.bin").c_str(), kHeadsAddr);

    Heads* heads = new Heads;
    heads->fMappingSize = 0;

    heads->fNHeads = ((int*)kHeadsAddr)[0];
    heads->fHeadSeqsLen = ((int*)kHeadsAddr)[1];

    heads->fHeadIdxs = ((int*)kHeadsAddr)+2;

    heads->fHeadSeqs = new int*[heads->fNHeads];
    heads->fHeadSeqs[0] = ((int*)kHeadsAddr)+2+heads->fNHeads;
    for(int i = 0; i < heads->fNHeads; ++i){
      heads->fHeadSeqs[i] = heads->fHeadSeqs[0]+heads->fHeadSeqsLen*i;
    }

    return heads;
  }

  //......................................................................
  Heads::~Heads()
  {
    //    munmap(kHeadsAddr, fMappingSize);
  }


  //......................................................................
  int Heads::ChildIdxFor(int head, int child) const
  {
    // Strip out flip information
    return fHeadSeqs[head][child]>>2;
  }

  //......................................................................
  bool Heads::FlipEvenFor(int head, int child) const
  {
    // Extract information in this bit
    return fHeadSeqs[head][child] & 1<<1;
  }

  //......................................................................
  bool Heads::FlipOddFor(int head, int child) const
  {
    // Extract information in this bit
    return fHeadSeqs[head][child] & 1<<0;
  }

  //......................................................................
  HeadsTranspose* HeadsTranspose::FromMMap(const std::string& libPath)
  {
    HeadsTranspose* heads = new HeadsTranspose;

    heads->fMappingSize = MMapFileAtAddress(libPath+"/heads_map_transpose.bin", kHeadsAddr, false);

    heads->fNHeads = ((int*)kHeadsAddr)[0];
    heads->fLibSize = ((int*)kHeadsAddr)[1];

    heads->fHeadIdxs = ((int*)kHeadsAddr)+2;

    heads->fPSeqs = (int*)((char*)heads->fHeadIdxs+sizeof(int)*heads->fNHeads);

    return heads;
  }

  //......................................................................
  HeadsTranspose* HeadsTranspose::FromMMapOnDemand(const std::string& libPath)
  {
    CreateOnDemandMapping((libPath+"/heads_map_transpose.bin").c_str(), kHeadsAddr);

    HeadsTranspose* heads = new HeadsTranspose;
    heads->fMappingSize = 0;

    heads->fNHeads = ((int*)kHeadsAddr)[0];
    heads->fLibSize = ((int*)kHeadsAddr)[1];

    heads->fHeadIdxs = ((int*)kHeadsAddr)+2;

    heads->fPSeqs = (int*)((char*)heads->fHeadIdxs+sizeof(int)*heads->fNHeads);

    return heads;
  }

  //......................................................................
  HeadsTranspose::~HeadsTranspose()
  {
    munmap(kHeadsAddr, fMappingSize);
  }

  //......................................................................
  std::ostream& operator<<(std::ostream& os, const lem::Heads& hs)
  {
    os << hs.fNHeads << " heads, with " << hs.fHeadSeqsLen << " events each.";
    return os;
  }

  //......................................................................
  std::ostream& operator<<(std::ostream& os, const lem::HeadsTranspose& hs)
  {
    os << hs.fNHeads << " heads, for a librar of " << hs.fLibSize << " events.";
    return os;
  }

} // namespace
