////////////////////////////////////////////////////////////////////////
/// \file    Library.cxx
/// \brief   Collection of events for matching
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/Library.h"

#include "LEM/func/OnDemand.h"
#include "LEM/func/Util.h"

#include "Utilities/func/MathUtil.h"

#include "TChain.h"

#include <iostream>

#include <sys/mman.h>

namespace lem
{
  // Private to this file
  namespace{
    char* kLibAddr = (char*)0x00000100CAFE0000;
  }

  //......................................................................
  double Library::SimpleSurvivalProb(const EventSummary& evt)
  {
    if(evt.IsSig()){
      // Scale up to keep as much signal as possible in sample, weighted
      // correctly internally. The missing prefactor is (ss2th13*ssth23)
      return util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
    }
    if(evt.ccnc == 0 && abs(evt.pdg) == 14 && abs(evt.origPdg) == 14){
      // CC mu
      return 1-util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
    }
    if(evt.ccnc == 0 && abs(evt.pdg) == 12 && abs(evt.origPdg) == 12){
      // Beam nue
      //      return 1-.1*util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
      // Want library matching to ignore nues
      return 0;
    }
    if(evt.ccnc == 1 && evt.pdg == evt.origPdg){
      // NC
      return 1;
    }
    if(evt.ccnc == 1 && evt.pdg != evt.origPdg){
      // Swapped NC. We play clever reweighting games so this is OK
      return 1;
    }
    if(evt.ccnc == 0 && abs(evt.pdg) == 14 && abs(evt.origPdg) == 12){
      // CC mu appearance, take ss2th13=0.1, maximal th23
      return .1*.5*util::sqr(sin(1.267*2.35e-3*810/evt.trueEVis));
    }
    // Don't know what this is
    return 0;
  }

  //......................................................................
  Library* Library::FromMMap(const std::string& libPath,
                             bool touchAll)
  {
    Library* lib = new Library;

    lib->fMappingSize = MMapFileAtAddress(libPath+"/map.bin", kLibAddr,
                                          touchAll, "/tmp/lem_preload_lib.lock");

    lib->fSummary = *((LibrarySummary*)kLibAddr);

    lib->fLibSize = lib->fSummary.N;

    lib->fLib = (MatchableEvent*)((char*)kLibAddr+sizeof(LibrarySummary));

    lib->fLibDownsample = (MatchableEvent*)((char*)lib->fLib+lib->fLibSize*sizeof(MatchableEvent));

    return lib;
  }

  //......................................................................
  Library* Library::FromMMapOnDemand(const std::string& libPath)
  {
    CreateOnDemandMapping((libPath+"/map.bin").c_str(), kLibAddr);

    Library* lib = new Library;

    lib->fSummary = *((LibrarySummary*)kLibAddr);

    lib->fLibSize = lib->fSummary.N;

    lib->fLib = (MatchableEvent*)((char*)kLibAddr+sizeof(LibrarySummary));

    lib->fLibDownsample = (MatchableEvent*)((char*)lib->fLib+lib->fLibSize*sizeof(MatchableEvent));

    return lib;
  }

  //......................................................................
  Library* Library::FromTrees(const std::string& libPath,
                              int splitLibIndex,
                              int splitLibFactor,
                              int downsampleFactor)
  {
    // TODO: this order won't match what heads is expecting

    TChain ch("makelib/tree");
    ch.Add((libPath+"/lem_hist_sig.root").c_str());
    ch.Add((libPath+"/lem_hist_bkg.root").c_str());
    const int Nunenrich = ch.GetEntries();
    ch.Add((libPath+"/lem_hist_enrich.root").c_str());

    EventSummary::InitFromTree(&ch);

    const int startIdx = (splitLibIndex*ch.GetEntries())/splitLibFactor;
    const int endIdx = ((splitLibIndex+1)*ch.GetEntries())/splitLibFactor;

    const int N = endIdx-startIdx;

    assert(N > 0);

    std::cout << "Loading libraries..." << std::endl;
    int nSig = 0, nBkg = 0;
    double totSig = 0, totBkg = 0, totEnrich = 0;
    int nTrueNC = 0, nSwapNC = 0;

    Library* lib = new Library;
    lib->fMappingSize = 0;

    lib->fLibSize = N;
    lib->fLib = new MatchableEvent[N];
    lib->fLibDownsample = new MatchableEvent[N];

    for(int n = startIdx; n < endIdx; ++n){
      if((n-startIdx)%10000 == 0)
        std::cout << (n-startIdx)/10000 << "/" << N/10000 << std::endl;

      const bool isEnrich = n >= Nunenrich;
      EventSummary sum = EventSummary::FromTree(&ch, n, isEnrich);
      sum.id = n;

      // This has to happen here, and not in MakePID where the other
      // reweighting happens, because it feeds into the totals for the
      // different event classes that we write out.
      const double P = isEnrich ? 1 : SimpleSurvivalProb(sum);
      // Screws up indexing, don't do it. FindMatches can cope
      //      if(P == 0) continue; // beam nue

      lem::MatchableEvent evt(sum);
      evt.weight = P;

      if(!isEnrich && sum.ccnc == 1){
        if(sum.pdg == sum.origPdg) ++nTrueNC; else ++nSwapNC;
      }

      lib->fLib[n-startIdx] = evt;

      lib->fLibDownsample[n-startIdx] = evt.Downsampled(downsampleFactor);

      if(isEnrich){
        ++totEnrich;
      }
      else{
        if(evt.IsSig()){
          ++nSig;
          totSig += P;
        }
        else{
          ++nBkg;
          totBkg += P;
        }
      }
    } // end for n

    lib->fSummary.nTrueNC = nTrueNC;
    lib->fSummary.nSwapNC = nSwapNC;
    lib->fSummary.totBkg = totBkg;
    lib->fSummary.totSig = totSig;
    lib->fSummary.totEnrich = totEnrich;

    std::cout << "Loaded " << lib->fSummary << std::endl;

    return lib;
  }

  //......................................................................
  Library::~Library()
  {
    if(fMappingSize != 0){
      //      munmap(kLibAddr, fMappingSize);
    }
    else{
      // Was loaded from trees
      delete[] fLib;
      delete[] fLibDownsample;
    }
  }
} // namespace
