////////////////////////////////////////////////////////////////////////
/// \file    OnDemand.h
/// \brief   Utilities for mapping in library files on-demand
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "LEM/func/OnDemand.h"

#include <cassert>
#include <cstring>
#include <deque>
#include <vector>

#include <sys/mman.h>
#include <sys/stat.h>

#include <signal.h>
#include <cstdio>
#include <cstdlib>


// There's no way to pass any context to sigaction, so we have to keep our
// state globally here. That's OK. It lets us check for overlapping mappings at
// least...

// Private to this file
namespace
{
  unsigned int gMemoryLimit = 0; // Bytes
  const int kPageSize = 10*1024*1024; // Our internal pages (10MB)

  /// Definition of a memory mapping
  struct Mapping
  {
    Mapping(const std::string& f, const char* b, size_t l)
      : fname(f), base(b), length(l) {}

    std::string fname; ///< Name of file to map in
    const char* base; ///< Where the start of the file should go
    size_t length; ///< How long the file is (bytes)
  };

  /// All the mappings we've been asked to make
  std::vector<Mapping> gMappings;

  /// The addresses of all pages we currently have mapped in, oldest first
  std::deque<void*> gMappedList;

  /// Handler for segfault, maps in the required part of a file, if possible
  void segfault_sigaction(int signal, siginfo_t* si, void* /*arg*/)
  {
    assert(si->si_signo == SIGSEGV);

    // Find what mapping this address is in
    int mappingIdx = -1;
    for(unsigned int n = 0; n < gMappings.size(); ++n){
      if(si->si_addr >= gMappings[n].base &&
         si->si_addr <  gMappings[n].base+gMappings[n].length){
        assert(mappingIdx == -1); // No overlapping mappings
        mappingIdx = n;
      }
    }

    // Out of range of our mappings
    if(mappingIdx < 0){
      // Reset the handler to nothing and resume. The program will die at the
      // correct point without our fingerprints all over it.
      struct sigaction sa;
      memset(&sa, 0, sizeof(sa));
      sigemptyset(&sa.sa_mask);

      sa.sa_sigaction = 0;

      sigaction(SIGSEGV, &sa, NULL);
      return;
    }


    // If there's a memory limit set
    if(gMemoryLimit > 0){
      // Keep discarding memory until we're below it
      while(gMappedList.size()*kPageSize >= gMemoryLimit){
        // Discard least recently used
        munmap(gMappedList.front(), kPageSize);
        gMappedList.pop_front();
      }
    }


    Mapping map = gMappings[mappingIdx];

    FILE* f = fopen(map.fname.c_str(), "r");
    assert(f);

    // Where are we in the file?
    long offset = long((char*)si->si_addr-(char*)map.base);
    // Round down to the start of the page this is in
    offset = (offset/kPageSize)*kPageSize;

    void* ret = mmap((char*)map.base+offset,
                     kPageSize,
                     PROT_READ,
                     MAP_SHARED | MAP_FIXED,
                     fileno(f),
                     offset);

    fclose(f);
    // Make sure we got our mapping
    assert(ret != (void*)-1);

    gMappedList.push_back((char*)map.base+offset);
  }
}

namespace lem
{
  void SetOnDemandMemoryLimit(long limit_mb)
  {
    // Convert to bytes
    gMemoryLimit = limit_mb<<20;
  }

  void CreateOnDemandMapping(const std::string& fname, const char* base)
  {
    // Must be aligned on OS pages
    assert(long(base)%(4*1024) == 0);

    // Open the file to check it exists and find out its length
    FILE* f = fopen(fname.c_str(), "r");
    assert(f);

    struct stat st;
    fstat(fileno(f), &st);
    const long fsize = st.st_size;
    fclose(f);


    gMappings.push_back(Mapping(fname, base, fsize));

    // Set up signal handler to fault this file (and others) in on demand. It
    // should be OK to do this multiple times.
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    sigemptyset(&sa.sa_mask);

    sa.sa_sigaction = segfault_sigaction;
    sa.sa_flags = SA_SIGINFO;

    sigaction(SIGSEGV, &sa, NULL);
  }
}
