////////////////////////////////////////////////////////////////////////
///
/// \brief  TODO
/// \author brebel@fnal.gov
///
////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>

// NOvASoft includes
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"

namespace cheat {
  class ClusterCheater : public art::EDProducer {
  public:
    explicit ClusterCheater(fhicl::ParameterSet const& pset);
    virtual ~ClusterCheater();

    void produce(art::Event& evt);

    void reconfigure(fhicl::ParameterSet const& pset);

 private:

    std::string fHitModuleLabel;    ///< label for module creating rb::CellHit objects	   
    std::string fG4ModuleLabel;     ///< label for module running G4 and making particles, etc

  };
}

namespace cheat{

  //--------------------------------------------------------------------
  ClusterCheater::ClusterCheater(fhicl::ParameterSet const& pset)
  {
    this->reconfigure(pset);

    produces< std::vector<rb::Cluster> >();
    produces< art::Assns<rb::Cluster, rb::CellHit> >();
  }

  //--------------------------------------------------------------------
  ClusterCheater::~ClusterCheater()
  {
  }

  //--------------------------------------------------------------------
  void ClusterCheater::reconfigure(fhicl::ParameterSet const& pset)
  {
    fHitModuleLabel    = pset.get< std::string >("HitModuleLabel",    "calhit"  );
    fG4ModuleLabel     = pset.get< std::string >("G4ModuleLabel",     "geantgen");

    return;
  }

  //--------------------------------------------------------------------
  void ClusterCheater::produce(art::Event& evt)
  {
    // get a back tracker
    art::ServiceHandle<cheat::BackTracker> bt;

    sim::ParticleNavigator pnav = bt->ParticleNavigator();

    // grab the hits that have been reconstructed
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fHitModuleLabel, hitcol);

    // make a vector of them
    std::vector< art::Ptr<rb::CellHit> > hits;
    art::fill_ptr_vector(hits, hitcol);
    
    // loop over the hits and figure out which particle contributed to each one
    std::vector< art::Ptr<rb::CellHit> >::iterator itr = hits.begin();

    // make a map of vectors of art::Ptrs keyed by trackID values
    std::map< int, std::vector< art::Ptr<rb::CellHit> > > trackHitMap;
    std::map< int, std::vector< art::Ptr<rb::CellHit> > >::iterator hitMapItr = trackHitMap.begin();

    // loop over all hits and fill in the map
    while( itr != hits.end() ){

      std::vector<cheat::TrackIDE> trackides = bt->HitToTrackIDE((*itr));

      // loop over all trackides for this hit
      for(size_t e = 0; e < trackides.size(); ++e){

	// don't worry about eve particles that contribute less than 10% of the
	// energy in the current hit
	if( trackides[e].energyFrac < 0.1) continue;

	trackHitMap[trackides[e].trackID].push_back((*itr));
	
      } // end loop over eve IDs for this hit

      itr++;
    }// end loop over hits

    // loop over the map and make clusters
    std::unique_ptr< std::vector<rb::Cluster> > clustercol(new std::vector<rb::Cluster>);
    std::unique_ptr< art::Assns<rb::Cluster, rb::CellHit> > assn(new art::Assns<rb::Cluster, rb::CellHit>);

    for(hitMapItr = trackHitMap.begin(); hitMapItr != trackHitMap.end(); hitMapItr++){

      // separate out the hits for each particle into the different views
      std::vector< art::Ptr<rb::CellHit> > trackHits( (*hitMapItr).second );

      art::PtrVector<rb::CellHit> ptrvsX;
      art::PtrVector<rb::CellHit> ptrvsY;
	
      for(size_t h = 0; h < trackHits.size(); ++h){

	if(trackHits[h]->View() == geo::kX) ptrvsX.push_back(trackHits[h]);
	if(trackHits[h]->View() == geo::kY) ptrvsY.push_back(trackHits[h]);
	  
      } // end loop over hits for this particle	
      
      // check to see if the particle id for this cluster lives in the
      // ParticleNavigator, if not, don't make a cluster and spew a warning
      if( pnav.find((*hitMapItr).first) == pnav.end() ){
	mf::LogWarning("ClusterCheater") << " no particle in ParticleNavigator has track id "
					 << (*hitMapItr).first
					 << " this could be a problem.  Certainly won't make a"
					 << " Cluster for it.";
	continue;
      }

      if(ptrvsX.size() > 0){
	clustercol->push_back(rb::Cluster(geo::kX,
					  ptrvsX,
					  ((*hitMapItr).first * 1000) + geo::kX*10));
	// associate the hits to this cluster
	util::CreateAssn(*this, evt, *(clustercol.get()), ptrvsX, *(assn.get()));
      }

      LOG_DEBUG("ClusterCheater") << "adding cluster: \n" 
				  << clustercol->back()
				  << "\nto collection.";

      if(ptrvsY.size() > 0){
	clustercol->push_back(rb::Cluster(geo::kY,
					  ptrvsY,
					  ((*hitMapItr).first * 1000) + geo::kY*10));
	// associate the hits to this cluster
	util::CreateAssn(*this, evt, *(clustercol.get()), ptrvsY, *(assn.get()));
      }
                
      LOG_DEBUG("ClusterCheater") << "adding cluster: \n" 
				  << clustercol->back()
				  << "\nto collection.";

      


    } // end loop over the map

    evt.put(std::move(clustercol));
    evt.put(std::move(assn));

    return;

  } // end produce

} // end namespace cheat

DEFINE_ART_MODULE(cheat::ClusterCheater);
