////////////////////////////////////////////////////////////////////////
// Class:       TestTrackIds
// Module Type: analyzer
// File:        TestTrackIds_module.cc
////////////////////////////////////////////////////////////////////////
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "MCCheater/BackTracker.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/ParticleNavigator.h"

namespace cheat
{
  class TestTrackIds : public art::EDAnalyzer
  {
  public:
    explicit TestTrackIds(const fhicl::ParameterSet& p);
    virtual ~TestTrackIds();

    virtual void analyze(const art::Event& e);
  };

//-------------------------------------------------------------------
  TestTrackIds::TestTrackIds(const fhicl::ParameterSet& pset)
    : EDAnalyzer(pset)
  {
  }

//-------------------------------------------------------------------
  TestTrackIds::~TestTrackIds()
  {
  }

//-------------------------------------------------------------------
void TestTrackIds::analyze(const art::Event& evt)
  {
    // check that this is MC, stop if it isn't
    assert(!evt.isRealData());
  
    art::Handle<std::vector<simb::MCTruth> > truths;
    evt.getByLabel("generator"/*fGeneratorModuleLabel*/, truths);
    if(truths->size() > 1){
      mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() << " has " 
  				    << truths->size() << " truth records";
    }
  
    art::ServiceHandle<cheat::BackTracker> bt;
  
    // Invariant tested here is that every photon signal has a matching FLSHit stored
  
    art::Handle<std::vector<sim::PhotonSignal> > phots;
    evt.getByLabel("photrans"/*fPhotonModuleLabel*/, phots);
  
    for(unsigned int photIdx = 0; photIdx < phots->size(); ++photIdx){
      const sim::PhotonSignal& phot = (*phots)[photIdx];
      const std::vector<sim::FLSHit> flss = bt->CellToFLSHit(phot.Plane(),
                                                             phot.Cell());
  
      bool matched = false;
      for(unsigned int flsIdx = 0; flsIdx < flss.size(); ++flsIdx){
        if(flss[flsIdx].GetTrackID() == phot.TrackId()){
          matched = true;
          break;
        }
      }
  
      if(!matched){
        mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() 
  				      << ": PhotonSignal in plane " << phot.Plane() 
  				      << " cell " << phot.Cell() 
  				      << " trackId " << phot.TrackId() 
  				      << " does not have any matching FLSHits";
        // This never goes wrong in practice, so leave the abort in so we'll notice if it ever does
        abort();
      }
    } // end for photIdx
  
    // now get the sim::Particles from the event
    sim::ParticleNavigator pnav = bt->ParticleNavigator();
  
    // Sanity check of pnav
    const std::set<int> ids = bt->GetTrackIDList();
  
    // Invariant here is that every FLSHit's trackId corresponds to a particle stored in the event
 
    art::Handle< std::vector<sim::FLSHitList> > flslistHandle;
    evt.getByLabel("geantgen"/*fGeantModuleLabel*/, flslistHandle);

    for(unsigned int listIdx = 0; listIdx < flslistHandle->size(); ++listIdx){
      const sim::FLSHitList flslist = (*flslistHandle)[listIdx];
      for(unsigned int flsIdx = 0; flsIdx < flslist.fHits.size(); ++flsIdx){
        const sim::FLSHit fls = flslist.fHits[flsIdx];
        const sim::ParticleNavigator::iterator part = pnav.find(fls.GetTrackID());
        if(part == pnav.end()){
          mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() 
  					<< ": FLS hit in plane " << fls.GetPlaneID()
  					<< " cell " << fls.GetCellID() << " trackId "
  					<< fls.GetTrackID() << " PDG code " << fls.GetPDG()
  					<< " has no matching sim::Particle";
          assert(ids.count(fls.GetTrackID()) == 0); // Make sure it really isn't there
	  abort();
        }
      }
    }
  
    // Invariant here is that every non-primary particle has its mother stored

    for(sim::ParticleNavigator::iterator it = pnav.begin(); it != pnav.end(); ++it){
      const int id = it->first;
      if(pnav.IsPrimary(id)) continue;
      const int mom = it->second->Mother();
      const int pdg = it->second->PdgCode();
      const std::string proc = it->second->Process();
      if(pnav.find(mom) == pnav.end()){
        mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() 
  				      << " Particle " << id 
  				      << " PDG code " << pdg 
  				      << " formed by process " << proc 
  				      << " can't find mother, index " << mom;
        assert(ids.count(mom) == 0); // Make sure it really isn't there
	abort();
      }
    }

    // Invaraint is that all daughters of a particle should exist

    std::set<int> trackIDs = bt->GetTrackIDList();
    for(std::set<int>::iterator it = trackIDs.begin(); it != trackIDs.end(); ++it){
      const sim::Particle* mother = bt->TrackIDToParticle(*it);
      if(!mother) {
	mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() 
					<< " can't find particle corresponding to track id:  " << *it; 
	abort();
      }
      
      for(int i = 0; i < mother->NumberDaughters(); ++i){
	const int id = mother->Daughter(i);
	//assert(bt->TrackIDToParticle(id));
	if (!bt->TrackIDToParticle(id)){
	  mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() 
					  << " can't find the daughter of particle:  " << *it
					  << " and the daughter has track id: "<<id;
	  abort();
	}
      }//Loop over the daughters of this particle


 }

  }// end analyze

  DEFINE_ART_MODULE(TestTrackIds);

}// end namespace cheat
