////////////////////////////////////////////////////////////////////////
/// \brief  TrackCheater module
///
/// \author brebel@fnal.gov
////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>

// NOvASoft includes
#include "MCCheater/BackTracker.h"
#include "RecoBase/Track.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"

namespace cheat {
  class TrackCheater : public art::EDProducer {
  public:
    explicit TrackCheater(fhicl::ParameterSet const& pset);
    virtual ~TrackCheater();

    void produce(art::Event& evt);

    void reconfigure(fhicl::ParameterSet const& pset);

 private:

    std::string fCheatedClusterLabel; ///< label for module creating recob::Cluster objects	   
    std::string fG4ModuleLabel;       ///< label for module running G4 and making particles, etc

  };
}

namespace cheat{

  //--------------------------------------------------------------------
  TrackCheater::TrackCheater(fhicl::ParameterSet const& pset)
  {
    this->reconfigure(pset);

    produces< std::vector<rb::Track> >();
    produces< art::Assns<rb::Track, rb::CellHit> >();
  }

  //--------------------------------------------------------------------
  TrackCheater::~TrackCheater()
  {
  }

  //--------------------------------------------------------------------
  void TrackCheater::reconfigure(fhicl::ParameterSet const& pset)
  {
    fCheatedClusterLabel = pset.get< std::string >("CheatedClusterLabel", "cluster" );
    fG4ModuleLabel       = pset.get< std::string >("G4ModuleLabel",       "geantgen");

    return;
  }

  //--------------------------------------------------------------------
  void TrackCheater::produce(art::Event& evt)
  {
    art::ServiceHandle<cheat::BackTracker> bt;
    sim::ParticleNavigator pnav = bt->ParticleNavigator();

    // grab the clusters that have been reconstructed
    art::Handle< std::vector<rb::Cluster> > clustercol;
    evt.getByLabel(fCheatedClusterLabel, clustercol);

    // make a vector of them - we aren't writing anything out to a file
    // so no need for a art::PtrVector here
    std::vector< art::Ptr<rb::Cluster> > clusters;
    art::fill_ptr_vector(clusters, clustercol);

    // loop over the clusters and figure out which particle contributed to each one
    std::vector< art::Ptr<rb::Cluster> >::iterator itr = clusters.begin();

    // make a map of vectors of art::Ptrs keyed by trackID values
    std::map< int, std::vector< art::Ptr<rb::Cluster> > > tidClusterMap;
    std::map< int, std::vector< art::Ptr<rb::Cluster> > >::iterator clusterMapItr = tidClusterMap.begin();

    // loop over all clusters and fill in the map
    while( itr != clusters.end() ){

      // in the ClusterCheater module we set the cluster ID to be 
      // the  particle track ID*1000 + view*10.  The
      // floor function on the cluster ID / 1000 will give us
      // the track ID
      int tID = floor((*itr)->ID()/1000.);

      tidClusterMap[tID].push_back((*itr));

      itr++;
    }// end loop over clusters

    
    // loop over the map and make prongs
    std::unique_ptr< std::vector<rb::Track> >  trackcol(new std::vector<rb::Track>);
    std::unique_ptr< art::Assns<rb::Track, rb::CellHit> >  assn(new art::Assns<rb::Track, rb::CellHit>);

    for(clusterMapItr = tidClusterMap.begin(); clusterMapItr != tidClusterMap.end(); clusterMapItr++){

      // is this prong electro-magnetic in nature or 
      // hadronic/muonic?  EM --> shower, everything else is a track
      if( abs(pnav[(*clusterMapItr).first]->PdgCode()) != 11  &&
	  abs(pnav[(*clusterMapItr).first]->PdgCode()) != 22  &&
	  abs(pnav[(*clusterMapItr).first]->PdgCode()) != 111 ){

	LOG_DEBUG("TrackCheater") << "prong of " << (*clusterMapItr).first 
				  << " is a track with pdg code "
				  << pnav[(*clusterMapItr).first]->PdgCode();

	// separate out the hits for each particle into the different views
	std::vector< art::Ptr<rb::Cluster> > tidClusters( (*clusterMapItr).second );

	// get the hits from the different clusters
	art::PtrVector<rb::CellHit> ptrvs;
	for(size_t c = 0; c < tidClusters.size(); ++c){
	  for(size_t h = 0; h < tidClusters[c]->NCell(); ++h)
	    ptrvs.push_back(tidClusters[c]->Cell(h));
	}

	// get the initial position and direction of the particle
	TVector3 pos(pnav[(*clusterMapItr).first]->Vx(),
		     pnav[(*clusterMapItr).first]->Vy(),
		     pnav[(*clusterMapItr).first]->Vz());

	TVector3 dir(pnav[(*clusterMapItr).first]->Px()/pnav[(*clusterMapItr).first]->P(),
		     pnav[(*clusterMapItr).first]->Py()/pnav[(*clusterMapItr).first]->P(),
		     pnav[(*clusterMapItr).first]->Pz()/pnav[(*clusterMapItr).first]->P());
	
	// add a prong to the collection.  Make the prong
	// ID be the same as the track ID for the eve particle
	trackcol->push_back(rb::Track(ptrvs, pos, dir, (*clusterMapItr).first));

	// need the trajectory points for this track - get them using
	// the MCParticle
	unsigned int numTraj = pnav[(*clusterMapItr).first]->NumberTrajectoryPoints();
	for(unsigned int t = 1; t < numTraj; ++t){
	  TVector3 traj(pnav[(*clusterMapItr).first]->Vx(t),
			pnav[(*clusterMapItr).first]->Vy(t),
			pnav[(*clusterMapItr).first]->Vz(t));
	  trackcol->back().AppendTrajectoryPoint(traj);
	}


	// get the hits associated with each cluster and associate those with the track
	art::FindManyP<rb::CellHit> fmh(tidClusters, evt, fCheatedClusterLabel);
        for(size_t c = 0; c < tidClusters.size(); ++c){
          std::vector< art::Ptr<rb::CellHit> > hits = fmh.at(c);
          util::CreateAssn(*this, evt, *trackcol, hits, *assn);
        }
	
	LOG_DEBUG("TrackCheater") << "adding track: \n" 
				  << trackcol->back()
				  << "\nto collection.";

      }//end if this is a track

    } // end loop over the map

    evt.put(std::move(trackcol));
    evt.put(std::move(assn));

    return;

  } // end produce

} // end namespace

namespace cheat{

  DEFINE_ART_MODULE(TrackCheater);

}

