////////////////////////////////////////////////////////////////////////
/// \brief   Simple module to analyze MC cosmics distributions
/// \author  messier@indiana.edu
////////////////////////////////////////////////////////////////////////
#include <vector>
#include <cmath>

// ROOT includes
#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Persistency/Common/Ptr.h" 
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 

// NOvA includes
#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "RawData/DAQHeader.h"
#include "RawData/RawDigit.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/func/MathUtil.h"


typedef struct{
  double dcosX;
  double dcosY;
  double dcosZ;
  double vtxX;
  double vtxY;
  double vtxZ;
  double endX;
  double endY;
  double endZ;
  double avCellsX;
  double avCellsY;
  double energy;
  double length;
  int run;
  int event;
  int pdg;
} CRINFO;

namespace mcchk {
  /// A module to make some plots of generated cosmic-rays
  class CosmicAna : public art::EDAnalyzer {

  public:
    explicit CosmicAna(fhicl::ParameterSet const& pset);
    virtual ~CosmicAna();

    void analyze(art::Event const& evt);
    void beginJob();
  
  private:

    std::string fFLSHitListLabel;   ///< label of module making fls hits
    std::string fRawDataLabel;      ///< label of module making raw digits

    TH1F* fNhitHisto;    ///< Histogram of number of hits in spill
    TH1F* fDminHisto0;   ///< Closest approach for particles leaving no hits
    TH1F* fDminHisto1;   ///< Closest approach for particles leaving hits
    TH2F* fPhotonAngles;   ///< Photon rate vs angle
    TH2F* fPhotonAnglesLo; ///< Photon rate vs angle, low momenta
    TH2F* fPhotonAnglesMi; ///< Photon rate vs angle, middle momenta
    TH2F* fPhotonAnglesHi; ///< Photon rate vs angle, high momenta
    TH1F* fPhotonCosQ;     ///< Photon rate vs cos(Q)
    TH1F* fPhotonEnergy;   ///< Photon energy (GeV)

    TH2F* fElectronAngles;   ///< Electron rate vs angle
    TH2F* fElectronAnglesLo; ///< Electron rate vs angle, low momenta
    TH2F* fElectronAnglesMi; ///< Electron rate vs angle, middle momenta
    TH2F* fElectronAnglesHi; ///< Electron rate vs angle, high momenta
    TH1F* fElectronCosQ;     ///< Electron rate vs cos(Q)
    TH1F* fElectronEnergy;   ///< Electron energy (GeV)

    TH2F* fMuonAngles;   ///< Muon rate vs angle
    TH2F* fMuonAnglesLo; ///< Muon rate vs angle, low momenta
    TH2F* fMuonAnglesMi; ///< Muon rate vs angle, middle momenta
    TH2F* fMuonAnglesHi; ///< Muon rate vs angle, high momenta
    TH1F* fMuonCosQ;     ///< Muon rate vs cos(Q)
    TH1F* fMuonEnergy;   ///< Muon energy (GeV)

    TH2F* fHitsYZ; ///< hit locations in YZ view
    TH2F* fHitsXZ; ///< hit locations in XZ view

    CRINFO  fCRInfo;   ///< cosmic ray muon info

    TTree   *fTree;   ///< tree holding CRINFO
  };
};


////////////////////////////////////////////////////////////////////////
namespace mcchk {
  
  //......................................................................
  
  CosmicAna::CosmicAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
    , fFLSHitListLabel  (pset.get< std::string >("FLSHitListLabel")    )
    , fRawDataLabel     (pset.get< std::string >("RawDataLabel", "daq"))
  {

  }

  //......................................................................
  
  CosmicAna::~CosmicAna() { }
  
  //......................................................................
  
  void CosmicAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    fNhitHisto = tfs->make<TH1F>("fNhitHisto",";hits/20 usec;spills",
				 200,0.0,20000.0);
    
    fDminHisto0 = tfs->make<TH1F>("fDminHisto0",";d (cm);",100,0.0,500.E2);
    fDminHisto1 = tfs->make<TH1F>("fDminHisto1",";d (cm);",100,0.0,500.E2);
    
    fPhotonAngles = tfs->make<TH2F>("fPhotonAngles",";x;y",
			     36,-180.0,180.0,50,-1.0,1.0);
    fPhotonAnglesLo = tfs->make<TH2F>("fPhotonAnglesLo",";x;y",
			       36,-180.0,180.0,50,-1.0,1.0);
    fPhotonAnglesMi = tfs->make<TH2F>("fPhotonAnglesMi",";x;y",
			       36,-180.0,180.0,50,-1.0,1.0);
    fPhotonAnglesHi = tfs->make<TH2F>("fPhotonAnglesHi",";x;y",
			       36,-180.0,180.0,50,-1.0,1.0);
    fPhotonCosQ   = tfs->make<TH1F>("fPhotonCosQ",";cos#theta;tracks",50,-1.0,1.0);
    fPhotonEnergy = tfs->make<TH1F>("fPhotonEnergy",";E (GeV)",200,0.0,5.0);
    
    fElectronAngles = tfs->make<TH2F>("fElectronAngles",";x;y",
			       36,-180.0,180.0,50,-1.0,1.0);
    fElectronAnglesLo = tfs->make<TH2F>("fElectronAnglesLo",";x;y",
				 36,-180.0,180.0,50,-1.0,1.0);
    fElectronAnglesMi = tfs->make<TH2F>("fElectronAnglesMi",";x;y",
				 36,-180.0,180.0,50,-1.0,1.0);
    fElectronAnglesHi = tfs->make<TH2F>("fElectronAnglesHi",";x;y",
				 36,-180.0,180.0,50,-1.0,1.0);
    fElectronCosQ   = tfs->make<TH1F>("fElectronCosQ",";cos#theta;tracks",50,-1.0,1.0);
    fElectronEnergy = tfs->make<TH1F>("fElectronEnergy",";E (GeV)",200,0.0,10.0);
    
    fMuonAngles = tfs->make<TH2F>("fMuonAngles",";x;y",
			   36,-180.0,180.0,50,-1.0,1.0);
    fMuonAnglesLo = tfs->make<TH2F>("fMuonAnglesLo",";x;y",
			     36,-180.0,180.0,50,-1.0,1.0);
    fMuonAnglesMi = tfs->make<TH2F>("fMuonAnglesMi",";x;y",
			     36,-180.0,180.0,50,-1.0,1.0);
    fMuonAnglesHi = tfs->make<TH2F>("fMuonAnglesHi",";x;y",
			     36,-180.0,180.0,50,-1.0,1.0);
    fMuonCosQ   = tfs->make<TH1F>("fMuonCosQ",";cos#theta;tracks",50,-1.0,1.0);
    fMuonEnergy = tfs->make<TH1F>("fMuonEnergy",";E (GeV)",200,0.0,100.0);
    
    fHitsYZ = tfs->make<TH2F>("hitsYZ", ";z (cm); y (cm)", 1500, 0., 1500., 400, -200., 200.); 
    fHitsXZ = tfs->make<TH2F>("hitsXZ", ";z (cm); X (cm)", 1500, 0., 1500., 400, -200., 200.); 
    
    fTree = tfs->make<TTree>("cosmics", "cosmics");
    fTree->Branch("cr", &fCRInfo, "dcosX/D:dcosY/D:dcosZ/D:vtxX/D:vtxY/D:vtxZ/D:endX/D:endY/D:endZ/D:avCellsX/D:avCellsY/D:energy/D:length/D:run/I:event/I:pdg/I");
    
  }
  
  //......................................................................

  void CosmicAna::analyze(art::Event const& evt)
  { 

    //  get rawdata::RawDigits from the event
    art::Handle< std::vector<rawdata::RawDigit> > digits;
    evt.getByLabel(fRawDataLabel, digits);
    
    art::Handle< std::vector<sim::FLSHitList> > hitlist;
    evt.getByLabel(fFLSHitListLabel,hitlist);
    
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<cmap::CMap> cmap;

    art::Ptr<sim::FLSHitList> hlist(hitlist,0);
    const std::vector<sim::FLSHit>& hits = hlist->fHits;

    fNhitHisto->Fill((float)hits.size());

    ///loop over the cellhits and see where they are
    for(unsigned int i = 0; i < digits->size(); ++i){
      double xyz[3] = {0.};

      art::Ptr<rawdata::RawDigit> dig(digits,i);
      // Need to ask channel map for plane and cell number.
      geom->Plane(cmap->GetPlane(dig.get()))->Cell(cmap->GetCell(dig.get()))->
	GetCenter(xyz);
      
    }
    
    ///loop over the hits and fill their entry positions in the hit histograms
    double wpos[3] = {0.};
    for(unsigned int i = 0; i < hits.size(); ++i) {
      ///convert hit positions from local to world coordinates
      double lpos[3] = {0.5*(hits[i].GetEntryX()+hits[i].GetExitX()),
			0.5*(hits[i].GetEntryY()+hits[i].GetExitY()),
			0.5*(hits[i].GetEntryZ()+hits[i].GetExitZ())};
      geom->Plane(hits[i].GetPlaneID())->Cell(hits[i].GetCellID())->LocalToWorld(lpos,wpos);
      
      if(geom->Plane(hits[i].GetPlaneID())->View() == geo::kY)
	fHitsYZ->Fill(wpos[2], wpos[1]);
      //fHitsYZ->Fill(hits[i].fPlaneId, hits[i].fCellId);
      if(geom->Plane(hits[i].GetPlaneID())->View() == geo::kX)
	fHitsXZ->Fill(wpos[2], wpos[0]);
      //fHitsXZ->Fill(hits[i].fPlaneId, hits[i].fCellId);
    }
    
    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator &pnav = bt->ParticleNavigator();
    
    double electronE = -1.0;
    double photonE   = -1.0;
    for (sim::ParticleNavigator::const_iterator i = pnav.begin(); i != pnav.end(); ++i) {
      const sim::Particle *particle = (*i).second;
      
      const TLorentzVector& v4 = particle->Position();
      const TLorentzVector& p4 = particle->Momentum();
      double x0[3] = {v4.X(),  v4.Y(),  v4.Z()};
      double dx[3] = {p4.Px(), p4.Py(), p4.Pz()};
      double x1[3] = {0.0,     0.0,     0.5*geom->DetLength()};
      double x2[3];
      double d = geo::ClosestApproach(x1, x0, dx, x2);
      if (bt->ParticleToFLSHit(particle->TrackId()).size()==0) {fDminHisto0->Fill(d); }
      else{
	fDminHisto1->Fill(d);
	if (d > 100.E2) {
	  mf::LogVerbatim("CosmicAna") << "dmin.gt.100 " << evt.run() << " " << evt.id().event() << " " 
				       << particle->PdgCode() << " "
				       << v4.X()  << " " << v4.Y()  << " " << v4.Z()  << " "
				       << p4.Px() << " " << p4.Py() << " " << p4.Pz() << " "
				       << d;
	}
      }
      
      TH1F* hCosQ     = 0;
      TH2F* hAngles   = 0;
      TH2F* hAnglesLo = 0;
      TH2F* hAnglesMi = 0;
      TH2F* hAnglesHi = 0;
      TH1F* hEnergy   = 0;
      if (abs(particle->PdgCode())==13) {
	hCosQ     = fMuonCosQ;
	hAngles   = fMuonAngles;
	hAnglesLo = fMuonAnglesLo;
	hAnglesMi = fMuonAnglesMi;
	hAnglesHi = fMuonAnglesHi;
	hEnergy   = fMuonEnergy;
      }
      else if (abs(particle->PdgCode())==22) {
	hCosQ     = fPhotonCosQ;
	hAngles   = fPhotonAngles;
	hAnglesLo = fPhotonAnglesLo;
	hAnglesMi = fPhotonAnglesMi;
	hAnglesHi = fPhotonAnglesHi;
	hEnergy   = fPhotonEnergy;
      }
      else if (abs(particle->PdgCode())==11) {
	hCosQ     = fElectronCosQ;
	hAngles   = fElectronAngles;
	hAnglesLo = fElectronAnglesLo;
	hAnglesMi = fElectronAnglesMi;
	hAnglesHi = fElectronAnglesHi;
	hEnergy   = fElectronEnergy;
      }
      if (hCosQ!=0) {
	if (bt->ParticleToFLSHit(particle->TrackId()).size()>0) {
	  double cosq = -p4.Py()/p4.P();
	  double phi  = atan2(p4.Pz(),p4.Px());
	  phi *= 180/M_PI;
	  hCosQ->Fill(cosq);
	  hAngles->Fill(phi,cosq);
	  if      (p4.E()<1.0)  hAnglesLo->Fill(phi,cosq);
	  else if (p4.E()<10.0) hAnglesMi->Fill(phi,cosq);
	  else                  hAnglesHi->Fill(phi,cosq);
	  hEnergy->Fill(p4.E());
	  
	  fCRInfo.run = evt.run();
	  fCRInfo.event = evt.id().event();
	  fCRInfo.pdg = particle->PdgCode();
	  fCRInfo.dcosX = p4.Px()/p4.P();
	  fCRInfo.dcosY = p4.Py()/p4.P();
	  fCRInfo.dcosZ = p4.Pz()/p4.P();
	  fCRInfo.energy = p4.E();
	  
	  std::vector< std::vector<double> > hitsP(geom->GetPlanesByView().size());
	  unsigned int numCells = geom->Plane(0)->Ncells();
	  if(geom->Plane(1)->Ncells() > numCells) numCells = geom->Plane(1)->Ncells();
	  for(unsigned int p = 0; p < hitsP.size(); ++p)
	    hitsP[p].resize(numCells,0.);
	  
	  std::vector<geo::View_t> viewP(geom->GetPlanesByView().size());
	  
	  double entryPos[3] = {0.};
	  double exitPos[3] = {0.};
	  const std::vector<sim::FLSHit>& flshits = bt->ParticleToFLSHit(particle->TrackId());

	  /*
	  for(unsigned int h = 0; h < flshits.size(); ++h){
	    if(h == 0){
	      double lpos[3] = {0.5*(flshits[h].GetEntryX()+flshits[h].GetExitX()),
				0.5*(flshits[h].GetEntryY()+flshits[h].GetExitY()),
				0.5*(flshits[h].GetEntryZ()+flshits[h].GetExitZ())};
	      geom->Plane(flshits[h].GetPlaneID())->Cell(flshits[h].GetCellID())->LocalToWorld(lpos,entryPos);
	    }
	    double lpos[3] = {0.5*(flshits[h].GetEntryX()+flshits[h].GetExitX()),
			      0.5*(flshits[h].GetEntryY()+flshits[h].GetExitY()),
			      0.5*(flshits[h].GetEntryZ()+flshits[h].GetExitZ())};
	    geom->Plane(flshits[h].GetPlaneID())->Cell(flshits[h].GetCellID())->LocalToWorld(lpos,exitPos);
	    
	    hitsP[flshits[h].GetPlaneID()][flshits[h].GetCellID()] += 1.;
	    viewP[flshits[h].GetPlaneID()] = geom->Plane(flshits[h].GetPlaneID())->View();
	  }
	  */
	  //Look at mine:
	  double startT = flshits[0].GetEntryT();
	  double endT = flshits[0].GetExitT();
	  int Plane1, Cell1, Plane2, Cell2;
	  Plane1 = flshits[0].GetPlaneID();
	  Plane2 = Plane1;
	  Cell1  = flshits[0].GetCellID();
	  Cell2  = Cell1;

	  double localStartP[3] = {flshits[0].GetEntryX(),
				   flshits[0].GetEntryY(),
				   flshits[0].GetEntryZ()};
	  double localEndP[3]   = {flshits[0].GetExitX(),
				   flshits[0].GetExitY(),
				   flshits[0].GetExitZ()};

	  hitsP[Plane1][Cell1] += 1.;
	  viewP[Plane1] = geom->Plane(Plane1)->View();

	  for(unsigned int h = 1; h < flshits.size(); ++h){
	    if (flshits[h].GetEntryT()<startT) {
	      Plane1 = flshits[h].GetPlaneID();
	      Cell1  = flshits[h].GetCellID();
	      localStartP[0] = flshits[h].GetEntryX();
	      localStartP[1] = flshits[h].GetEntryY();
	      localStartP[2] = flshits[h].GetEntryZ();
	    }

	    else if (flshits[h].GetExitT()>endT) {
	      Plane2 = flshits[h].GetPlaneID();
	      Cell2  = flshits[h].GetCellID();
	      localEndP[0] = flshits[h].GetExitX();
	      localEndP[1] = flshits[h].GetExitY();
	      localEndP[2] = flshits[h].GetExitZ();
	    }
	    
	    hitsP[flshits[h].GetPlaneID()][flshits[h].GetCellID()] += 1.;
	    viewP[flshits[h].GetPlaneID()] = geom->Plane(flshits[h].GetPlaneID())->View();
	  }

	  geom->Plane(Plane1)->Cell(Cell1)->LocalToWorld(localStartP,entryPos);
	  geom->Plane(Plane2)->Cell(Cell2)->LocalToWorld(localEndP,exitPos);
	  //Hmm, done.
	  
	  double planesX = 0.;
	  double cellsX = 0.;
	  double planesY = 0.;
	  double cellsY = 0.;
	  for(unsigned int p = 0; p < hitsP.size(); ++p){
	    double cells = 0.;
	    for(unsigned int c = 0; c < hitsP[p].size(); ++c){
	      if(hitsP[p][c] > 0.) cells += 1.;
	    }
	    
	    if(viewP[p] == geo::kX && cells > 0.){
	      planesX += 1.;
	      cellsX += cells;
	    }
	    if(viewP[p] == geo::kY && cells > 0.){
	      planesY += 1.;
	      cellsY += cells;
	    }
	    
	  }//end loop over hit cells
	  
	  
	  if(planesX > 0.) fCRInfo.avCellsX = cellsX/planesX;
	  else             fCRInfo.avCellsX = 0.;
	  
	  if(planesY > 0.) fCRInfo.avCellsY = cellsY/planesY;
	  else             fCRInfo.avCellsY = 0.;
	  
	  fCRInfo.vtxX = entryPos[0];
	  fCRInfo.vtxY = entryPos[1];
	  fCRInfo.vtxZ = entryPos[2];
	  fCRInfo.endX = exitPos[0];
	  fCRInfo.endY = exitPos[1];
	  fCRInfo.endZ = exitPos[2];
	  
	  ///cosmic rays always come from above
	  if(exitPos[1] > entryPos[1]){
	    fCRInfo.endX = entryPos[0];
	    fCRInfo.endY = entryPos[1];
	    fCRInfo.endZ = entryPos[2];
	    fCRInfo.vtxX = exitPos[0];
	    fCRInfo.vtxY = exitPos[1];
	    fCRInfo.vtxZ = exitPos[2];
	  }
	  
	  //check this calculation
	  fCRInfo.length = util::pythag(exitPos[0]-entryPos[0],
					exitPos[1]-entryPos[1],
					exitPos[2]-entryPos[2]);
	  
	  
	  fTree->Fill();
	  
	  // Check for interesting particles
	  if (p4.E()>0.500 && p4.E()<5.0) {
	    if (abs(particle->PdgCode())==11) electronE = p4.E();
	    if (abs(particle->PdgCode())==22) photonE   = p4.E();
	  }
	}
      } // loop on particles
    }
    if (electronE>0.0 || photonE>0.0) {
      mf::LogVerbatim("CosmicAna") << "Run/Event " << evt.getRun().run() << " "
				   << evt.id().event() << " " 
				   << electronE << " " << photonE;
    }

  }
  
}

namespace mcchk
{
  DEFINE_ART_MODULE(CosmicAna);
}

