////////////////////////////////////////////////////////////////////////
/// \brief
/// \author messier@indiana.edu
////////////////////////////////////////////////////////////////////////
#include <vector>

#include "TH1.h"
#include "TH2.h"
#include "TDatabasePDG.h"
#include "TSystem.h"

#include "RawData/DAQHeader.h"
#include "RawData/RawDigit.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "CMap/CMap.h"

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h" 
#include "fhiclcpp/ParameterSet.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Persistency/Common/Ptr.h" 
#include "art/Persistency/Common/PtrVector.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Services/Optional/TFileService.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h" 
#include "messagefacility/MessageLogger/MessageLogger.h" 


namespace mcchk {
  /// A module to check the results from the Monte Carlo generator

  class DetAna : public art::EDAnalyzer {

  public:
    explicit DetAna(fhicl::ParameterSet const& pset);
    virtual ~DetAna();

    void analyze(art::Event const& evt);
    void beginRun(const art::Run& run);

  private:

    TH1F*       fNumDigits;     ///< number of digits per event 
    TH1F* 	fADCs;          ///< adc values for each digit  
    TH2F* 	fHitCellsXView; ///< cells hit in the x view	  
    TH2F* 	fHitCellsYView; ///< cells hit in the y view    

    int         fCellsX;        ///< number of cells in x view  
    int 	fCellsY;        ///< number of cells in y view  
    int 	fPlanes;        ///< number of planes           

    std::string fRawDataLabel;  ///< module label for the raw data

  };
};


////////////////////////////////////////////////////////////////////////
namespace mcchk {
  
  static int msg1cnt = 0;
  
  //......................................................................
  
  DetAna::DetAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
    , fNumDigits(0) // Flag that the others are uninitialized
    , fRawDataLabel(pset.get< std::string >("RawDataLabel", "daq"))
  {

  }

  //......................................................................
  
  DetAna::~DetAna() { }
  
  //......................................................................
  
  void DetAna::beginRun(const art::Run& /*run*/)
  {
    if(fNumDigits) return; // Don't make histograms twice

    art::ServiceHandle<art::TFileService> tfs;

    art::ServiceHandle<geo::Geometry> geo;
    const std::set<unsigned int> planesX = geo->GetPlanesByView(geo::kX);
    const std::set<unsigned int> planesY = geo->GetPlanesByView(geo::kY);
    fCellsX = geo->Plane(*planesX.begin())->Ncells();
    fCellsY = geo->Plane(*planesY.begin())->Ncells();
    fPlanes = planesX.size() + planesY.size();
    
    
    fNumDigits     = tfs->make<TH1F>("fNumDigits", ";Digits;", 
				     1000, 0., 1000.);
    fADCs          = tfs->make<TH1F>("fADCs", ";ADC;", 
				     1000, 0., 1000.);
    fHitCellsXView = tfs->make<TH2F>("fHitCellsXView", ";Plane; Cell", 
				     fPlanes, 0., 1.*fPlanes, 
				     fCellsX, 0., 1.*fCellsX);
    fHitCellsYView = tfs->make<TH2F>("fHitCellsYView", ";Plane; Cell", 
				     fPlanes, 0., 1.*fPlanes, 
				     fCellsY, 0., 1.*fCellsY);
    
  }
  
  //......................................................................
  
  void DetAna::analyze(art::Event const& evt)
  { 
    //  get rawdata::RawDigits from the event
    art::Handle< std::vector<rawdata::RawDigit> > diglist;
    evt.getByLabel(fRawDataLabel, diglist);

    if (diglist->empty()) {
      if(msg1cnt < 5){
	std::cerr << "no raw digit list for this event" << std::endl;
	++msg1cnt;
      }
      fNumDigits->Fill(0.);
      return;
    }
    
    fNumDigits->Fill(diglist->size());

    art::ServiceHandle<geo::Geometry> geo;
    art::ServiceHandle<cmap::CMap> cmap;
    geo::View_t view;
    
    for(unsigned int i = 0; i < diglist->size(); ++i){      
      art::Ptr<rawdata::RawDigit> rd(diglist,i);

      for(unsigned int adc = 0; adc < rd->NADC(); ++adc)
	fADCs->Fill(rd->ADC(adc)); 
      
      // Need to ask channel map for plane and cell number.
      view = geo->Plane(cmap->GetPlane(rd.get()))->View();
      if(view == geo::kX)
	fHitCellsXView->Fill(1.*cmap->GetPlane(rd.get())+0.5, 1.*cmap->GetCell(rd.get())+0.5);
      else if(view == geo::kY)
	fHitCellsYView->Fill(1.*cmap->GetPlane(rd.get())+0.5, 1.*cmap->GetCell(rd.get())+0.5);
    }///end loop over digits
  }
} // end mcchk 

///////////////////////////////////////////////////////////////////////
namespace mcchk
{
  DEFINE_ART_MODULE(DetAna);
}


