////////////////////////////////////////////////////////////////////////
/// \brief
/// \author messier@indiana.edu
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <vector>

#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TParticle.h"
#include "TDatabasePDG.h"
#include "TSystem.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Persistency/Common/Ptr.h" 
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 

// NOvA includes
#include "MCCheater/BackTracker.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/ParticleNavigator.h"


namespace mcchk {
  /// A module to check the results from the Monte Carlo generator
  
  class DetSimAna : public art::EDAnalyzer {
    
  public:
    explicit DetSimAna(fhicl::ParameterSet const& pset);
    virtual ~DetSimAna();

    void analyze(art::Event const& evt);
    void beginJob();

  private:
    TH1F* fNevents;             ///< Total number of events

    TH1F* fPDF_momentum;       ///< Momentum of the first particle in MCTruth

    TH1F* fNParticles_detsim;    ///< Number of particles after DetSim
    TH1F* fNParticles_truth;    ///< Number of particles after MCTruth

    TH1F* fPDG_detsim;          ///< PDG of all particles after DetSim
    TH1F* fPDG_truth;          ///< PDG of all particles after MCTruth

    TH1F* fFLS_hits;           ///< FLS hits
    TH1F* fFLS_EDep;           ///< FLS Energy deposition
    TH1F* fFLS_EDep_perhit;           ///< FLS Energy deposition

    TH1F* fFLS_X_entry[3];      ///< X entry 0=all, 1=primary, 2=rest
    TH1F* fFLS_Y_entry[3];      ///< 
    TH1F* fFLS_Z_entry[3];      ///< 

    TH1F* fFLS_dX[3];      ///< dX entry 0=all, 1=primary, 2=rest
    TH1F* fFLS_dY[3];      ///< 
    TH1F* fFLS_dZ[3];      ///< 
    TH1F* fFLS_dR[3];      ///< Passing distance in the cell (for primary particle only)

    TH2F* fFLS_XY_entry[3];      ///< X entry
    TH2F* fFLS_XZ_entry[3];      ///< X entry
    TH2F* fFLS_YZ_entry[3];      ///< X entry

//Event number vs...

    TH1F* fEvent_FLS_hits;           ///< Event number vs FLSHits
    TH1F* fEvent_NParticles_detsim;  ///< Event number vs Number of particles
    
    TH2F* fEvent_FLS_hits_forPID;    ///< Event number and PID vs FLSHits
    
    
//Individual FLS hit
    TH1F* fFLSHit_Eventnumber;
    TH1F* fFLSHit_cellid;
    TH1F* fFLSHit_planeid;
    TH1F* fFLSHit_PDG;

    TH1F* fFLSHit_EDep;
    TH1F* fFLSHit_Entry_XYZ[3];
    TH1F* fFLSHit_Exit_XYZ[3];


    std::string fFLSHitListLabel;
    std::string fMCTruthModuleLabel;
  };
};


////////////////////////////////////////////////////////////////////////
namespace mcchk {

  static int cntEvent = 0; //event number
  static int cntFLSHit = 0; //FLS hit number

  //......................................................................
  
  DetSimAna::DetSimAna(fhicl::ParameterSet const& pset) :
    EDAnalyzer(pset),
    fFLSHitListLabel   ( pset.get< std::string >("FLSHitListLabel")    ),
    fMCTruthModuleLabel( pset.get< std::string >("MCTruthModuleLabel" ))
  {
  }

  //......................................................................
  
  DetSimAna::~DetSimAna() { }
  
  //......................................................................
  
  void DetSimAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    
    fPDG_detsim = tfs->make<TH1F>("fPDG_detsim", "After DetSim;PDG number;", 1001, -500.5, 500.5);
    fPDG_truth  = tfs->make<TH1F>("fPDG_truth" , "In MCTruth;PDG number;"  , 1001, -500.5, 500.5);

    fNParticles_detsim  = tfs->make<TH1F>("fNParticles_detsim", "After DetSim;Number of particles in event;" , 101, -2.50, 502.5);
    fNParticles_truth   = tfs->make<TH1F>("fNParticles_truth" , "After MCTruth;Number of particles in event;", 51, -0.25, 50.25);

    fNevents  = tfs->make<TH1F>("fNevents", ";Total number of events", 1, -0.5, 0.5);

    fFLS_hits = tfs->make<TH1F>("fFLS_hits", ";FLS hits", 201, -10.0, 4010.0);
    fFLS_EDep = tfs->make<TH1F>("fFLS_EDep", ";Total FLS_Edep (GeV)", 1000, 0.0, 10.0);
    fFLS_EDep_perhit = tfs->make<TH1F>("fFLS_EDep_perhit", ";FLS_Edep per hit (GeV)", 1000, 0.0, .01);

    fEvent_FLS_hits = tfs->make<TH1F>("fEvent_FLS_hits", ";Event number;FLS hits", 5001, -0.5, 5000.5);
  
    fEvent_NParticles_detsim = tfs->make<TH1F>("fEvent_NParticles_detsim", ";Event number;Number of particles after detsim", 5001, -0.5, 5000.5);


    int   MaximumFLShits=100000;
    float MM = float(MaximumFLShits) - 0.5;

    fFLSHit_Eventnumber= tfs->make<TH1F>("fFLSHit_Eventnumber",";FLS hit number; Event number", MaximumFLShits, -0.5, MM);
    fFLSHit_cellid     = tfs->make<TH1F>("fFLSHit_cellid",";FLS hit number; Cell ID", MaximumFLShits, -0.5, MM);
    fFLSHit_planeid    = tfs->make<TH1F>("fFLSHit_planeid",";FLS hit number; Plane ID", MaximumFLShits, -0.5, MM);
    fFLSHit_EDep       = tfs->make<TH1F>("fFLSHit_EDep",";FLS hit number; EDep", MaximumFLShits, -0.5, MM);
    fFLSHit_PDG        = tfs->make<TH1F>("fFLSHit_PDG",";FLS hit number; PDG", MaximumFLShits, -0.5, MM);

    char str[20];
    char HistoName[200];
    char HistoTitle[200];

    for(int i=0; i<3; i++){
 	
      if(i==0)sprintf(str,"X");
      if(i==1)sprintf(str,"Y");
      if(i==2)sprintf(str,"Z");
	
      sprintf(HistoName , "fFLSHit_Entry_XYZ[%d]", i);
      sprintf(HistoTitle, " fFLSHit_Entry_%s (cm)", str);
      fFLSHit_Entry_XYZ[i] = tfs->make<TH1F>(HistoName,HistoTitle, MaximumFLShits, -0.5, MM);
 
      sprintf(HistoName , "fFLSHit_Exit_XYZ[%d]", i);
      sprintf(HistoTitle, " fFLSHit_Exit_%s (cm)", str);
      fFLSHit_Exit_XYZ[i]  = tfs->make<TH1F>(HistoName,HistoTitle, MaximumFLShits, -0.5, MM);
    }

    for(int i=0; i<3; i++){
      if(i==0)sprintf(str,"for all particles");
      if(i==1)sprintf(str,"for the primary particle");
      if(i==2)sprintf(str,"for the rest particles");

      sprintf(HistoName , "fFLS_X_entry[%d]", i);
      sprintf(HistoTitle, "FLS_X_entry %s (cm)", str);
      fFLS_X_entry[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 1000 , -10.0 , 10.0);

      sprintf(HistoName , "fFLS_Y_entry[%d]",i);
      sprintf(HistoTitle, "FLS_Y_entry %s (cm)", str);
      fFLS_Y_entry[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 1000 , -10.0 , 10.0);

      sprintf(HistoName , "fFLS_Z_entry[%d]",i);
      sprintf(HistoTitle, "FLS_Z_entry %s (cm)", str);
      fFLS_Z_entry[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 10000, -210.0, 210.);

      sprintf(HistoName , "fFLS_dX[%d]", i);
      sprintf(HistoTitle, "FLS_dX %s (cm)", str);
      fFLS_dX[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 20000, -10.0, 10.0);

      sprintf(HistoName , "fFLS_dY[%d]",i);
      sprintf(HistoTitle, "FLS_dY %s (cm)", str);
      fFLS_dY[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 20000, -10.0, 10.0);

      sprintf(HistoName , "fFLS_dZ[%d]",i);
      sprintf(HistoTitle, "FLS_dZ %s (cm)", str);
      fFLS_dZ[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 20000, -10.0, 10.0);

      sprintf(HistoName , "fFLS_dR[%d]",i);
      sprintf(HistoTitle, "FLS_dR %s (cm)", str);
      fFLS_dR[i]  = tfs->make<TH1F>(HistoName,HistoTitle, 10000,   0.0, 10.0);

      sprintf(HistoName , "fFLS_XY_entry[%d]", i);
      sprintf(HistoTitle, "%s;FLS_X_entry (cm);FLS_Y_entry (cm);", str);
      fFLS_XY_entry[i] = tfs->make<TH2F>(HistoName,HistoTitle, 1000, -3.5, 3.5, 1000, -3.5  , 3.5  );

      sprintf(HistoName , "fFLS_XZ_entry[%d]", i);
      sprintf(HistoTitle, "%s;FLS_X_entry (cm);FLS_Z_entry (cm);", str);
      fFLS_XZ_entry[i] = tfs->make<TH2F>(HistoName,HistoTitle, 1000, -3.5, 3.5, 5000, -210.0, 210.0);

      sprintf(HistoName , "fFLS_YZ_entry[%d]", i);
      sprintf(HistoTitle, "%s;FLS_Y_entry (cm);FLS_Z_entry (cm);", str);
      fFLS_YZ_entry[i] = tfs->make<TH2F>(HistoName,HistoTitle, 1000, -3.5, 3.5, 5000, -210.0, 210.0);
  
    }
    
  }
  
  //......................................................................
  
  void DetSimAna::analyze(art::Event const& evt)
  { 
    // int Present_diglist = 1;

    //    std::cout << "From DetSimAna run    : " << evt.run() 
    //      //	      << " subrun : " << evt.Subrun() 
    //	      << " event  : " << evt.id().event() << std::endl;

    /// Increment the number of events
    fNevents->Fill(0.);

    bool is_mctruth_list_empty  = false;
    bool is_flshit_list_empty   = false;

    /// Pull the MC generator information out of the event
    art::Handle<std::vector<simb::MCTruth> > mclist;
    try{
      evt.getByLabel(fMCTruthModuleLabel,mclist);
      if (mclist->size() == 0) is_mctruth_list_empty = true; 
    }
    catch(...){
      is_mctruth_list_empty = true; 
    }
	
    /// Pull the FLSHit list
    art::Handle<std::vector<sim::FLSHitList> > hitlist;
    try{
      evt.getByLabel(fFLSHitListLabel,hitlist);
      if (hitlist->size() == 0) is_flshit_list_empty = true; 
    }
    catch(...){
      is_flshit_list_empty = true; 
    }


    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();
	    
    /*
    // Pull the Raw Digits information out
    art::Handle<std::vector<rawdata::RawDigit> > diglist;
    try{
      evt.getByLabel(fRawDataLabel, diglist);
      if (diglist->size() == 0) is_digit_list_empty = true;
    }
    catch(...){
      is_digit_list_empty = true;
    }
    */

    //std::cout<<is_particle_list_empty<<" "<<is_mctruth_list_empty<<" "<<is_flshit_list_empty<<" "<<is_digit_list_empty<<"\n";
    
    /// If there are no particle list or fls hits, do't bother filling FLS hit list info
    if(!is_flshit_list_empty){
      /// Looping over interactions
      for(unsigned int iint = 0; iint < hitlist->size(); ++iint ){

	const std::vector<sim::FLSHit>&   hits = hitlist->at(iint).fHits;
	float EDep_Total=0.;
  
	/// Fill number of FLS Hits
	fFLS_hits->Fill(hits.size());
	fEvent_FLS_hits->Fill(cntEvent, hits.size());

	/// Loop over FLSHits
	for(unsigned int i = 0; i < hits.size(); ++i){
	  EDep_Total += hits[i].GetEdep();
	  fFLS_EDep_perhit->Fill(hits[i].GetEdep());

	  float xx = hits[i].GetEntryX() - hits[i].GetExitX();
	  float yy = hits[i].GetEntryY() - hits[i].GetExitY();
	  float zz = hits[i].GetEntryZ() - hits[i].GetExitZ();
	  float rr = sqrt(xx*xx+yy*yy+zz*zz);

	  int FillingArray = 2;

	  /// Fill it for all particles and whatever FillingArray is
	  for(int j=0; j<FillingArray+1; j+=FillingArray){
	    fFLS_X_entry[j]->Fill(hits[i].GetEntryX());
	    fFLS_Y_entry[j]->Fill(hits[i].GetEntryY());
	    fFLS_Z_entry[j]->Fill(hits[i].GetEntryZ());
	
	    fFLS_dX[j]->Fill(xx);
	    fFLS_dY[j]->Fill(yy);
	    fFLS_dZ[j]->Fill(zz);
	    fFLS_dR[j]->Fill(rr);;

	    fFLS_XY_entry[j]->Fill(hits[i].GetEntryX(), hits[i].GetEntryY());
	    fFLS_XZ_entry[j]->Fill(hits[i].GetEntryX(), hits[i].GetEntryZ());
	    fFLS_YZ_entry[j]->Fill(hits[i].GetEntryY(), hits[i].GetEntryZ());
	  }//end of loop over j
	
	  fFLSHit_Eventnumber->Fill(cntFLSHit,cntEvent);
	  fFLSHit_cellid->Fill(cntFLSHit,hits[i].GetCellID());
	  fFLSHit_planeid->Fill(cntFLSHit,hits[i].GetPlaneID());
	  fFLSHit_EDep->Fill(cntFLSHit,hits[i].GetEdep());
	  fFLSHit_PDG->Fill(cntFLSHit,hits[i].GetPDG());
	  fFLSHit_Entry_XYZ[0]->Fill(cntFLSHit,hits[i].GetEntryX());
	  fFLSHit_Entry_XYZ[1]->Fill(cntFLSHit,hits[i].GetEntryY());
	  fFLSHit_Entry_XYZ[2]->Fill(cntFLSHit,hits[i].GetEntryZ());
	  fFLSHit_Exit_XYZ[0]->Fill(cntFLSHit,hits[i].GetExitX());
	  fFLSHit_Exit_XYZ[1]->Fill(cntFLSHit,hits[i].GetExitY());
	  fFLSHit_Exit_XYZ[2]->Fill(cntFLSHit,hits[i].GetExitZ());
	
	  ++cntFLSHit;
	}// end of loop over FLSHits
      		
	fFLS_EDep->Fill(EDep_Total);// Fill total energy deposition
    
      }// end of loop over interactions
    }// end of if there are no particle list or flshits

    if(!is_mctruth_list_empty){
      /// Filling the MCTruth information
      for(unsigned int iint = 0; iint < mclist->size(); ++iint){

	fNParticles_truth->Fill((float)mclist->at(iint).NParticles());

      }//end loop over interactions
    }// end if there is no mctruth

    
    /// Looping over particles
    for (sim::ParticleNavigator::const_iterator itr = pnav.begin(); itr != pnav.end(); ++itr) {
      fPDG_detsim->Fill((*itr).second->PdgCode());
    }// end of loop over particles in the interaction
    
    fNParticles_detsim->Fill((float)pnav.size());
    fEvent_NParticles_detsim->Fill(cntEvent,(float)pnav.size());

    
    /*
      if(Present_diglist){
      // std::cout<<"Now Raw digit"<<std::endl;
      // Looping over RawDigit
      for(unsigned int i = 0; i < diglist.size(); ++i){
      const rawdata::RawDigit* rd = diglist[i];
      int trackid_rawdigit = rd->TrackID();
      std::cout<<"TrackID = "<<trackid_rawdigit<<std::endl;
      }
      }// end of Present_diglist
    */

    ++cntEvent;
  }

} // end mcchk

////////////////////////////////////////////////////////////////////////
namespace mcchk
{
  DEFINE_ART_MODULE(DetSimAna);
}
