////////////////////////////////////////////////////////////////////////
/// \brief  version 1
///
/// \authors Dan Pershey - Raphael Schroeter
//           rschroet@physics.harvard.edu
/// \date    12/01/25
//
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <vector>

#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TParticle.h"
#include "TDatabasePDG.h"
#include "TSystem.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Principal/SubRun.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Persistency/Common/Ptr.h" 
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 

// NOvA includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RawData/DAQHeader.h"
#include "RawData/RawDigit.h"
#include "Simulation/Particle.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCFlux.h"
#include "Simulation/FLSHitList.h"
#include "Utilities/func/MathUtil.h"


namespace mcchk {
  /// Module to check the daughter leptons coming from neutrino interactions
  class LeptonAna : public art::EDAnalyzer {
    
  public:
    explicit LeptonAna(fhicl::ParameterSet const& pset);
    virtual ~LeptonAna();

    void reconfigure(const fhicl::ParameterSet& pset);
    void analyze(art::Event const& evt);
    void beginRun(const art::Run& run);

  private:

    // Truth information
    std::string      fMCTruthModuleLabel;  ///< Label mc truth information module 
    std::string      fParticleListLabel;   ///< Label particle information module
    std::string      fFLSHitListLabel;     ///< Label FLSHit information module

    ///  Vertex Information for Leptons
    TH1F* fCosThetaX;                      ///< Cosine of angle between lepton momentum and xhat
    TH1F* fCosThetaY;                      ///< Cosine of angle between lepton momentum and yhat
    TH1F* fCosThetaZ;                      ///< Cosine of angle between lepton momentum and zhat
    TH1F* fmuSpectrum;                     ///< Energy spectrum of daughter particle mu- and mu+
    TH1F* feSpectrum;                      ///< Energy spectrum of daughter particle e- and e+
    TH1F* fnumuSpectrum;                   ///< Energy spectrum of daughter particle nu_mu and nu_mu_bar
    TH1F* fnueSpectrum;                    ///< Energy spectrum of daughter nu_e and nu_e_bar
    TH1F* fmuTrackLength;                  ///< Lepton track lenghth distribution
    TH1F* feTrackLength;                   ///< Lepton track lenghth distribution

    ///  Global Information for CC Leptons
    TH1F* fmuEDiff;                        ///< Energy Deposited - Truth Energy of for muons
    TH1F* feEDiff;                         ///< Energy Deposited - Truth Energy of for electrons
    TH1F* ftotalEDiff;                     ///< Energy Deposited - Truth Energy Total
    TH1F* fmudEdx;                         ///< Average dE/dx for muon along its track or shower
    TH1F* fedEdx;                          ///< Average dE/dx for electron along its track or shower
    TH2F* fXZEDep;                         ///< X vs Z view of Energy Deposited by charged lepton
    TH2F* fYZEDep;                         ///< X vs Z view of Energy Deposited by charged lepton

    ///  End Point Information for Contained CC Leptons
    TH2F* fXZEndPoint;                     ///< X vs Z view of where contained lepton tracks end
    TH2F* fYZEndPoint;                     ///< Y vs Z view of where contained lepton tracks end
    TH2F* fXEEndPoint;                     ///< X vs Energy view of where contained lepton tracks end
    TH2F* fYEEndPoint;                     ///< Y vs Energy view of where contained lepton tracks end
    TH2F* fZEEndPoint;                     ///< Z vs Energy view of where contained lepton tracks end

    /// Escape Point Information for Not Contained CC Leptons
    TH2F* fXZEscape;                       ///< X vs Z view of where non-contained leptons leave the detector
    TH2F* fYZEscape;                       ///< Y vs Z view of where non-contained leptons leave the detector
    TH2F* fXEEscape;                       ///< X vs Lepton Energy view of where non-contained leptons leave the detector
    TH2F* fYEEscape;                       ///< Y vs Lepton Energy view of where non-contained leptons leave the detector
    TH2F* fZEEscape;                       ///< Z vs Lepton Energy view of where non-contained leptons leave the detector

  };
};


////////////////////////////////////////////////////////////////////////
namespace mcchk {
    
  LeptonAna::LeptonAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
    fCosThetaX(0) // Flag that the others are uninitialized too
  { 
    this->reconfigure(pset);
  }
    
////////////////////////////////////////////////////////////////////////
  
  LeptonAna::~LeptonAna()  { }
  
////////////////////////////////////////////////////////////////////////
  
  void LeptonAna::reconfigure(const fhicl::ParameterSet& pset)
  {
    fParticleListLabel         = pset.get< std::string >("ParticleListLabel");
    fFLSHitListLabel           = pset.get< std::string >("FLSHitListLabel");
    fMCTruthModuleLabel        = pset.get< std::string >("MCTruthModuleLabel");
  }

////////////////////////////////////////////////////////////////////////
  
  void LeptonAna::beginRun(const art::Run& /*run*/)
  {
    if(fCosThetaX) return; // Don't make histograms twice

    art::ServiceHandle<art::TFileService> tfs;
    art::ServiceHandle<geo::Geometry> geo;

    double x = 2.1*geo->DetHalfWidth();
    double y = 2.1*geo->DetHalfHeight();
    double z = 1.1*geo->DetLength();
    int xdiv = TMath::Nint(2*x/5.);
    int ydiv = TMath::Nint(2*y/5.);
    int zdiv = TMath::Nint(2*z/5.);
    
    //  Histograms for Daughter Lepton Information at the Vertex
    fCosThetaX = tfs->make<TH1F>("fCosThetaX","Cosine of Angle Between Daughter Lepton Momentum and X Axis; Cos #theta_{x}; ",200,-1,1);
    fCosThetaY = tfs->make<TH1F>("fCosThetaY","Cosine of Angle Between Daughter Lepton Momentum and Y Axis; Cos #theta_{y}; ",200,-1,1);
    fCosThetaZ = tfs->make<TH1F>("fCosThetaZ","Cosine of Angle Between Daughter Lepton Momentum and Z Axis; Cos #theta_{z}; ",200,-1,1);
    fmuSpectrum = tfs->make<TH1F>("fmuSpectrum","Energy Spectrum of Daughter Muons; E_{#mu (GeV)}; ",200,0,10);
    feSpectrum = tfs->make<TH1F>("feSpectrum","Energy Spectrum of Daughter Electrons; E_{e} (GeV); ",200,0,10);
    fnumuSpectrum = tfs->make<TH1F>("fnumuSpectrum","Energy Spectrum of Daughter Muons; E_{#nu} (GeV); ",200,0,10);
    fnueSpectrum = tfs->make<TH1F>("fnueSpectrum","Energy Spectrum of Daughter Electrons; E_{#nu} (GeV); ",200,0,10);
    fmuTrackLength = tfs->make<TH1F>("fmuTrackLength","Track length ; (cm); ",1000,0,2000);
    feTrackLength = tfs->make<TH1F>("feTrackLength","Track length ; (cm); ",500,0,1000);

    //  histograms for Global Daughter Lepton Information
    fmuEDiff = tfs->make<TH1F>("fmuEDiff","Deposited Energy - Truth Energy for Muon; E_{Truth}-E_{FLS} (GeV); ",200,0,1);
    feEDiff = tfs->make<TH1F>("feEDiff","Deposited Energy - Truth Energy for Electron; E_{Truth}-E_{FLS} (GeV); ",200,0,1);
    ftotalEDiff = tfs->make<TH1F>("ftotalEDiff","Deposited Energy - Truth Energy Total; E_{Truth}-E_{FLS} (GeV); ",200,0,1);
    fmudEdx = tfs->make<TH1F>("fmudEdx","Average Energy Loss for Muon; dE/dx (MeV/cm); ",400,0,10);
    fedEdx = tfs->make<TH1F>("fedEdx","Average Energy Loss for Electron; dE/dx (MeV/cm); ",400,0,10);
    fXZEDep = tfs->make<TH2F>("fXZEDep","Energy Deposited by Lepton; Z (cm); X (cm)", zdiv, -0.02*z, z, xdiv, -x, x );
    fYZEDep = tfs->make<TH2F>("fYZEDep","Energy Deposited by Lepton; Z (cm); Y (cm)", zdiv, -0.02*z, z, ydiv, -y, y);
					
    //  Histograms for Endpoint of Contained Daugther Charged Leptons
    fXZEndPoint = tfs->make<TH2F>("fXZEndPoint","X vs Z View of End Point of Contained Charged Lepton Tracks; Z (cm); X (cm)",zdiv, -0.02*z, z, xdiv, -x, x);
    fYZEndPoint = tfs->make<TH2F>("fYZEndPoint","Y vs Z View of End Point of Contained Charged Lepton Tracks; Z (cm); Y (cm)",zdiv, -0.02*z, z, ydiv, -y, y);
    fXEEndPoint = tfs->make<TH2F>("fXEEndPoint","Truth Lepton Energy vs X View of End Point of Contained Charged Lepton Tracks; X (cm); E (GeV)",xdiv, -x, x,200,0,10);
    fYEEndPoint = tfs->make<TH2F>("fYEEndPoint","Truth Lepton Energy vs Y View of End Point of Contained Charged Lepton Tracks; Y (cm); E (GeV)",ydiv, -y, y,200,0,10);
    fZEEndPoint = tfs->make<TH2F>("fZEEndPoint","Truth Lepton Energy vs Z View of End Point of Contained Charged Lepton Tracks; Z (cm); E (GeV)",zdiv, -0.02*z, z,200,0,10);

    //  Histograms for Escape Point of Non-Contained Daugther Charged Leptons
    fXZEscape = tfs->make<TH2F>("fXZEscape","X vs Z View of Where Not-Contained Charged Lepton Tracks Leave the Detector; Z (cm); X (cm)",zdiv, -0.02*z, z, xdiv, -x, x);
    fYZEscape = tfs->make<TH2F>("fYZEscape","Y vs Z View of Where Not-Contained Charged Lepton Tracks Leave the Detector; Z (cm); Y (cm)",zdiv, -0.02*z, z, ydiv, -y, y);
    fXEEscape = tfs->make<TH2F>("fXEEscape","Truth Lepton Energy vs X View of Where Not-Contained Charged Lepton Tracks Leave the Detector; X (cm); E (GeV)",xdiv, -x, x,200,0,10);
    fYEEscape = tfs->make<TH2F>("fYEEscape","Truth Lepton Energy vs Y View of Where Not-Contained Charged Lepton Tracks Leave the Detector; Y (cm); E (GeV)",ydiv, -y, y,200,0,10);
    fZEEscape = tfs->make<TH2F>("fZEEscape","Truth Lepton Energy vs Z View of Where Not-Contained Charged Lepton Tracks Leave the Detector; Z (cm); E (GeV)",zdiv, -0.02*z, z,200,0,10);
  }
  
////////////////////////////////////////////////////////////////////////
    
  void LeptonAna::analyze(art::Event const& evt)
  {
    
    art::ServiceHandle<geo::Geometry> geo;

    double det_width  = 2*geo->DetHalfWidth();
    double det_height = 2*geo->DetHalfHeight();
    double det_length = geo->DetLength();


    // Pull the MC generator information out of the event
    art::Handle< std::vector<simb::MCTruth> > mclist;
    evt.getByLabel(fMCTruthModuleLabel, mclist);
    if (mclist->empty()){
      std::cerr << "no mc list for this event" << std::endl;
      return;
    }
    
    //  Pull the particle list information out of the event
    art::Handle< std::vector<sim::Particle> > partlist;
    evt.getByLabel(fParticleListLabel, partlist);
    if (partlist->empty()){
      std::cerr << "no particle list for this event" << std::endl;
      return;
    }

    //  Pull the FLS list information out of the event
    art::Handle< std::vector<sim::FLSHitList>  > hitlist;
    evt.getByLabel(fFLSHitListLabel, hitlist);   
    if (hitlist->empty()){
      std::cerr << "no flshit list for this event" << std::endl;
      return;
    }     

    //  Get out the information on particles, hits, and incoming neutrino
    art::Ptr<simb::MCTruth> mc(mclist,0);
    const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
    /// Incoming neutrino
    const simb::MCParticle& nu = mc_neutrino.Nu();
    /// Outgoing lepton
    const simb::MCParticle& lep = mc_neutrino.Lepton();

    const std::vector<sim::Particle>& parts = *partlist;

    art::Ptr<sim::FLSHitList> hlist(hitlist, 0);
    const std::vector<sim::FLSHit>& hits = hlist->fHits;

    //  Process type
    bool is_muCC = 0;
    bool is_eCC  = 0;
    bool is_muNC = 0;
    bool is_eNC  = 0;
    if (mc_neutrino.CCNC() == simb::kCC && abs(nu.PdgCode()) == 14) is_muCC = 1;
    if (mc_neutrino.CCNC() == simb::kCC && abs(nu.PdgCode()) == 12) is_eCC = 1;
    if (mc_neutrino.CCNC() == simb::kNC && abs(nu.PdgCode()) == 14) is_muNC = 1;
    if (mc_neutrino.CCNC() == simb::kNC && abs(nu.PdgCode()) == 12) is_eNC = 1;

    //  scan the particle stack to find the Daughter Lepton index and track id, the first one is good enough
    int lepton_index = 0;
    int lepton_track_id = 0;
    for (unsigned int j = 0; j < parts.size(); ++j){
     if (parts[j].Mother() >= 0){
	if (is_muCC && abs(parts[j].PdgCode()) == 13){
	  lepton_index = j;
	  lepton_track_id = parts[j].TrackId();
	  break;
	}
	else if (is_eCC && abs(parts[j].PdgCode()) == 11){
	  lepton_index = j;
	  lepton_track_id = parts[j].TrackId();
	  break;
	}	
	else if (is_muNC && abs(parts[j].PdgCode()) == 14){
	  lepton_index = j;
	  lepton_track_id = parts[j].TrackId();
	  break;
	}	
	else if (is_eNC && abs(parts[j].PdgCode()) == 12){
	  lepton_index = j;
	  lepton_track_id = parts[j].TrackId();
	  break;
	}
      }
    }

    /// Get total energy of particles
    float total_energy = 0;
    for (unsigned int j = 0; j < parts.size(); ++j){
      if (parts[j].Mother() >= 0)
	total_energy += parts[j].E();
    }
    
    // std::cout << "NC CC " << mc_neutrino.CCNC() << std::endl;
    // std::cout << "mu CC " << is_muCC << std::endl;
    // std::cout << "e CC " << is_eCC << std::endl;
    // std::cout << "mu NC " << is_muNC << std::endl;
    // std::cout << "e NC " << is_eNC << std::endl;
    // std::cout << "lepton track id " << lepton_track_id << std::endl;
    // std::cout << "new lepton track id " << lep.TrackId() << std::endl;
    // std::cout << "lepton PX " << parts[lepton_index].Px() << std::endl;
    // std::cout << "new lepton PX " << lep.Px() << std::endl;
    // std::cout << "lepton E " << parts[lepton_index].E() << std::endl;
    // std::cout << "new lepton E " << lep.E() << std::endl;
    // std::cout << "lepton end x " << parts[lepton_index].EndX() << std::endl;
    // std::cout << "new lepton end x " << lep.EndX() << std::endl;
    // std::cout << "lepton end y " << parts[lepton_index].EndY() << std::endl;
    // std::cout << "new lepton end y " << lep.EndY() << std::endl;
    // std::cout << "lepton end z " << parts[lepton_index].EndZ() << std::endl;
    // std::cout << "new lepton end z " << lep.EndZ() << std::endl;
    
    // Fill histograms with Vertex information
    fCosThetaX->Fill(lep.Px()/lep.P());
    fCosThetaY->Fill(lep.Py()/lep.P());
    fCosThetaZ->Fill(lep.Pz()/lep.P());
    if (is_muCC == 1) fmuSpectrum->Fill(lep.E());
    if (is_eCC == 1) feSpectrum->Fill(lep.E());
    if (is_muNC == 1) fnumuSpectrum->Fill(lep.E());
    if (is_eNC == 1) fnueSpectrum->Fill(lep.E());

    // for (unsigned int j = 1; j < hits.size(); ++j){
    //   //std::cout << "fls hit track id " << hits[j].GetTrackID() << std::endl;
    //   if (hits[j].GetTrackID() == lepton_track_id){
    // 	std::cout << "E dep fls for i + " << j << " " << hits[j].GetEdep() << std::endl;
    //   }
    // }
					
    float deposited_energy = 0;
    float deposited_energy_total = 0;
    double xyz[3];
    double xyzworld[3];
    for (unsigned int j = 1; j < hits.size(); ++j){
      if (hits[j].GetTrackID() == lepton_track_id){
    	xyz[0] = 0.5*(hits[j].GetEntryX() + hits[j].GetExitX());
    	xyz[1] = 0.5*(hits[j].GetEntryY() + hits[j].GetExitY());
    	xyz[2] = 0.5*(hits[j].GetEntryZ() + hits[j].GetExitZ());
    	geo->Plane(hits[j].GetPlaneID())->Cell(hits[j].GetCellID())->LocalToWorld(xyz,xyzworld);
    	deposited_energy += hits[j].GetEdep();
    	fXZEDep->Fill(xyzworld[2],xyzworld[0],hits[j].GetEdep());
    	fYZEDep->Fill(xyzworld[2],xyzworld[1],hits[j].GetEdep());
      }
      deposited_energy_total += hits[j].GetEdep();
    }

    //  Find the last point in the track of the charged daugther lepton in the detector
    float final_lepton_track_x;
    float final_lepton_track_y;
    float final_lepton_track_z;
    float track_length;
    //unused//bool is_Contained = 0;

    if (abs(parts[lepton_index].EndX()) < det_width && abs(parts[lepton_index].EndY()) < det_height && parts[lepton_index].EndZ() < det_length 
    	&& parts[lepton_index].EndZ() > 0)
      {
    	final_lepton_track_x = parts[lepton_index].EndX();
    	final_lepton_track_y = parts[lepton_index].EndY();
    	final_lepton_track_z = parts[lepton_index].EndZ();      
    	//unused//is_Contained = 1;
    	fXZEndPoint->Fill(final_lepton_track_z, final_lepton_track_x);
    	fYZEndPoint->Fill(final_lepton_track_z, final_lepton_track_y);
    	fXEEndPoint->Fill(final_lepton_track_x, parts[lepton_index].E());
    	fYEEndPoint->Fill(final_lepton_track_y, parts[lepton_index].E());
    	fZEEndPoint->Fill(final_lepton_track_z, parts[lepton_index].E());
      }
    else{
      for (unsigned int j = 0; j < parts[lepton_index].NumberTrajectoryPoints(); ++j){
	
    	if (abs(parts[lepton_index].Vx(j)) < det_width)
    	  final_lepton_track_x = parts[lepton_index].Vx(j);
    	else break;
	
    	if (abs(parts[lepton_index].Vy(j)) < det_height) 
    	  final_lepton_track_y = parts[lepton_index].Vy(j);
    	else break;
	
    	if (parts[lepton_index].Vz(j) < det_length && parts[lepton_index].Vz(j) > 0) 
    	  final_lepton_track_z = parts[lepton_index].Vz(j);
    	else break;
      }
    	fXZEscape->Fill(final_lepton_track_z, final_lepton_track_x);
    	fYZEscape->Fill(final_lepton_track_z, final_lepton_track_y);
    	fXEEscape->Fill(final_lepton_track_x, parts[lepton_index].E());
    	fYEEscape->Fill(final_lepton_track_y, parts[lepton_index].E());
    	fZEEscape->Fill(final_lepton_track_z, parts[lepton_index].E());
    }

    //  Track length of the charged daughter lepton
    
    track_length = util::pythag(parts[lepton_index].Vx()-final_lepton_track_x,
                                parts[lepton_index].Vy()-final_lepton_track_y,
                                parts[lepton_index].Vz()-final_lepton_track_z);

    if (is_muCC)
      fmuTrackLength->Fill(track_length);
    else if (is_eCC)
      feTrackLength->Fill(track_length);

    //  dE/dx measurement    
    if( track_length > 0. ){
      float dEdx = deposited_energy/track_length;
      if (is_muCC == 1) {
	fmudEdx->Fill(dEdx);
	fmuEDiff->Fill(deposited_energy/parts[lepton_index].E());
      }
      if (is_eCC == 1) {
	fedEdx->Fill(dEdx);
	feEDiff->Fill(deposited_energy/parts[lepton_index].E());
      }
      ftotalEDiff->Fill(deposited_energy_total/total_energy);
    }
 
  }  // end LeptonAna
}  // end mcchk

////////////////////////////////////////////////////////////////////////
namespace mcchk
{
  DEFINE_ART_MODULE(LeptonAna);
}

