////////////////////////////////////////////////////////////////////////
/// \brief
/// \author messier@indiana.edu
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <map>
#include <vector>

#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TF1.h"
#include "TDatabasePDG.h"
#include "TSystem.h"

// NOvA includes
#include "GeometryObjects/CoordinateTransformation.h"
#include "Geometry/Geometry.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCFlux.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "fhiclcpp/ParameterSet.h"


namespace mcchk {
  /// check the results from the Monte Carlo generator
  class NeutrinoAna : public art::EDAnalyzer {
    
  public:
    explicit NeutrinoAna(fhicl::ParameterSet const& pset);
    virtual ~NeutrinoAna();

    void analyze(art::Event const& evt);
    void beginJob();
    void endJob();


  private:

    /// Calculating true L, E, L/E from neutrino information
    /// and those measured from the final state particles
    bool analyzeNeutrinoInteraction(const art::Ptr<simb::MCTruth>& , const art::Ptr<simb::MCFlux>& );

    /// Calculate Neutrino energy and direction based on Lepton and Proton information
    bool calculateNeutrino(const simb::MCParticle& input_lepton, const simb::MCParticle& input_proton, double* neutrino_direction, double& Enu);

    /// Sets the most energetic particle. Returns the number of such particles
    int getMostEnergeticParticle(const art::Ptr<simb::MCTruth>& mc, const std::vector <int>& pids, simb::MCParticle& most_energetic_particle, bool print=false);

    /// Calculate L from the start and end point. Also returns the direction (U)
    double calculateL(double* start_point, double* end_point, double* U);

    /// Calculate dot product between two vectors
    double dotProduct(double*, double*);

  private:

      //I Giuseppe Ferone and my supervisor Denis Perevalov are making an histogram with the neutrino starting positions; plese do not modify

      std::vector<double> fNeutrinoStartingPositionX;
      std::vector<double> fNeutrinoStartingPositionY;
      std::vector<double> fNeutrinoStartingPositionZ;
      //********************************************************

    //geo::CoordinateTransformation _CoordinateTransformation; ///< CoordiateTransformation class

    /// Neutrino information
    std::string fMCTruthModuleLabel;                ///< label for module producing mc truth information
    std::vector <TH1F*> fEnuForDifferentProcesses; ///< Neutrino energy spectra as generated for different processes and neutrinos

    int   fEnu_Nbins;  ///< number of bins for Enu histograms
    float fEnu_Emin;   ///< in GeV. E_min for Enu histograms
    float fEnu_Emax;   ///< in GeV. E_max for Enu histograms


    TH1F* fVertexX;    ///< vertex location of generated events in x
    TH1F* fVertexY;    ///< vertex location of generated events in y
    TH1F* fVertexZ;    ///< vertex location of generated events in z

    TH2F* fVertexXY;   ///< vertex location in xy
    TH2F* fVertexXZ;   ///< vertex location in xz
    TH2F* fVertexYZ;   ///< vertex location in yz

    TH3F* fVertex;     ///< 3D vertex location

    TH1F* fDCosX;      ///< direction cosine in x
    TH1F* fDCosY;      ///< direction cosine in y
    TH1F* fDCosZ;      ///< direction cosine in z

    TH1F* fMuMomentum; ///< momentum of outgoing muons
    TH1F* fMuDCosX;    ///< direction cosine of outgoing mu in x
    TH1F* fMuDCosY;    ///< direction cosine of outgoing mu in y
    TH1F* fMuDCosZ;    ///< direction cosine of outgoing mu in z

    TH1F* fEMomentum;  ///< momentum of outgoing electrons
    TH1F* fEDCosX;     ///< direction cosine of outgoing e in x
    TH1F* fEDCosY;     ///< direction cosine of outgoing e in y
    TH1F* fEDCosZ;     ///< direction cosine of outgoing e in z

    TH1F* fMode;       ///< interaction mode

    TH1F* fECons;      ///< histogram to determine if energy is conserved in the event

    TH1F* fProtonMultiplicityAllNucl;  ///< Proton multiplicity 
    TH1F* fProtonMultiplicityTi48;
    TH1F* fProtonMultiplicityFe56;
    TH1F* fProtonMultiplicityCl35;
    TH1F* fProtonMultiplicityC12;
    TH1F* fProtonMultiplicityH;
    TH1F* fProtonMultiplicityO16;
    TH1F* fProtonMultiplicityN14;
    TH1F* fProtonMultiplicityS32;


    TH1F* fnuQEEnergy; ///< neutrino energy in QE interactions
    TH1F* fnuResEnergy; ///< neutrino energy in Res interactions
    TH1F* fnuDISEnergy; ///< neutrino energy in DIS interactions
    TH1F* fnuCohEnergy; ///< neutrino energy in Coh interactions

    TH1F* fnuKlongEnergy; ///< neutrino energy from K Long decays
    TH1F* fnueKlongEnergy; ///< nu_e energy from K Long decays
    TH1F* fnumuKlongEnergy; ///< nu_mu energy from K Long decays
    TH1F* fnuKchargedEnergy; ///< neutrino energy from charged kaon decays
    TH1F* fnueKchargedEnergy; ///< nu_e energy from charged kaon decays
    TH1F* fnumuKchargedEnergy; ///< nu_mu energy from charged kaon decays
    TH1F* fnuPiEnergy; ///< neutrino energy from pion decays
    TH1F* fnuePiEnergy; ///< nu_e energy from pion decays
    TH1F* fnumuPiEnergy; ///< nu_mu energy from pion decays
    TH1F* fnuMuEnergy; ///< neutrino energy from pion decays
    TH1F* fnueMuEnergy; ///< nu_e energy from pion decays
    TH1F* fnumuMuEnergy; ///< nu_mu energy from pion decays

    TH1F* fW; ///< W (shower mass)
    TH1F* fX; ///< Xbj
    TH1F* fY; ///< Ybj
    TH1F* fQSqr; ///< Q^2

    /// Parent Hadron Information
    std::vector <TH1F*> fVertexHadronDecay;    ///< vertex location of generated events in beam coordinates

    double fEnuTrue;    ///< True neutrino energy. Dimensional (whatever units)

    /// Parent Hadron Information
    TH1F* fNeutrinoTrueL;   ///< True L for neutrino

    TH2F* fNeutrinoTrueETrueL;   ///<  True L and E of the neutrino

    TH1F* fNeutrinoEnuDiffLeptonBest; ///< Difference between Enu_true and Enue from Lepton information only
    TH1F* fNeutrinoEnuDiffProtonBest; ///< Difference between Enu_true and Enue from Proton information only
    TH1F* fNeutrinoEnuDiffBothBest;   ///< Difference between Enu_true and Enue from Both Lepton and Proton (average)

    TH1F* fNeutrinoDirectionDiffBest;      ///< Angle between true and best reconstructed neutrino direction
    TH1F* fNeutrinoDirectionDiffHPndGenie; ///< Angle between the neutrino directions that calculated from the Hadron production  and what Genie calculated

    std::vector <TH1F*> fNeutrinoU_Genie;        ///< Direction cosines (detector coodrinates) of neutrinos that Genie produced
    std::vector <TH1F*> fNeutrinoU_HP;           ///< Direction cosines (detector coodrinates) of neutrinos calculated from Hadron production
    std::vector <TH1F*> fNeutrinoU_DiffHPGenie;  ///< Direction cosines (detector coodrinates) of the difference between of direction calculated from Hadron production and the one produced by Genie

    std::vector <int> fBadEventNumber;    ///< Vector of Bad events
    int               fCurrentEvent;      ///< current event number

    
    //service handle
    art::ServiceHandle <art::TFileService> tfs;
    art::ServiceHandle<geo::Geometry> geo;

  };
};

////////////////////////////////////////////////////////////////////////
namespace mcchk {

  static int msg1cnt = 0;
  
  //......................................................................
  
  NeutrinoAna::NeutrinoAna(fhicl::ParameterSet const& pset):
    EDAnalyzer(pset)
    ,fMCTruthModuleLabel(pset.get< std::string>("MCTruthModuleLabel"))
    ,fEnu_Nbins(pset.get<int>   ("Enu_Nbins"))
    ,fEnu_Emin (pset.get<double>("Enu_Emin"))
    ,fEnu_Emax (pset.get<double>("Enu_Emax"))
  { 
  }
    
  //......................................................................
  
  NeutrinoAna::~NeutrinoAna(){

      std::cout<<"~NeutrinoAna\n";

      std::cout<<"Bad events:\n";
      for(unsigned int i = 0; i < fBadEventNumber.size(); i++)
          std::cout<<fBadEventNumber[i]<<"\n";
  }
  

  //......................................................................
  
  void NeutrinoAna::beginJob(){

    // double x = 2.5*geo->DetHalfWidth();
    // double y = 2.5*geo->DetHalfHeight();
    // double z = 2.*geo->DetLength();  

    double x = 400;
    double y = 400;
    double z = 2000;

    int xdiv = TMath::Nint(2*x/5.);
    int ydiv = TMath::Nint(2*y/5.);
    int zdiv = TMath::Nint(2*z/5.);
    
    fVertexX = tfs->make<TH1F>("fVertexX", ";x (cm)", xdiv,  -x, x);
    fVertexY = tfs->make<TH1F>("fVertexY", ";y (cm)", ydiv,  -y, y);
    fVertexZ = tfs->make<TH1F>("fVertexZ", ";z (cm)", zdiv, -0.1*z, z);
    
    fVertexXY = tfs->make<TH2F>("fVertexXY", ";x (cm);y (cm)",
    				xdiv,-x, x, ydiv, -y, y);
    fVertexXZ = tfs->make<TH2F>("fVertexXZ", ";z (cm);x (cm)",
				zdiv, -0.2*z, z, xdiv, -x, x);
    fVertexYZ = tfs->make<TH2F>("fVertexYZ", ";z (cm);y (cm)",
    				zdiv, -0.2*z, z, ydiv, -y, y);
    
    //fVertex   = tfs->make<TH3F>("fVertex", "", zdiv, -0.2*z, z, xdiv, -x, x, ydiv, -y, y);
    
    
    /// Creating Enu histograms
    {
      //int   n_bins = fEnu_Nbins;       ///< Number of bins
      float e_min  = fEnu_Emin * CLHEP::GeV;  ///< Enu minimal
      float e_max  = fEnu_Emax * CLHEP::GeV;  ///< Enu maximum
      
      /// Base neutrino interactions
      std::vector <std::string> initial_name;
      initial_name.push_back("fEnu_nue_CC");
      initial_name.push_back("fEnu_nuebar_CC");
      initial_name.push_back("fEnu_numu_CC");
      initial_name.push_back("fEnu_numubar_CC");
      initial_name.push_back("fEnu_nu_NC");
      initial_name.push_back("fEnu_nubar_NC");
      
      /// Additional information about the final states
      std::vector <std::string> addition_to_name;
      addition_to_name.push_back("");
      addition_to_name.push_back("_QE");
      addition_to_name.push_back("_ProtonPi0");
      addition_to_name.push_back("_NeutronPi0");
      addition_to_name.push_back("_ProtonPiPlus");
      addition_to_name.push_back("_NeutronPiPlus");
      addition_to_name.push_back("_ProtonPiMinus");
      addition_to_name.push_back("_NeutronPiMinus");
      
      /// Loop over additional and base names (in that order!) and create histograms
      for(unsigned int i_additional_name = 0; i_additional_name < addition_to_name.size(); i_additional_name++){
        for(unsigned int i_initial_name    = 0; i_initial_name    < initial_name.size()      ; i_initial_name++){
	  std::string final_name = initial_name[i_initial_name]
	    + addition_to_name[i_additional_name];
	  //fEnuForDifferentProcesses.push_back(tfs->make<TH1F>(final_name.c_str(), "; E_{#nu} (GeV)", n_bins, e_min/CLHEP::GeV, e_max/CLHEP::GeV));
	  fEnuForDifferentProcesses.push_back(tfs->make<TH1F>(final_name.c_str(), "; E_{#nu} (GeV)", 80, e_min/CLHEP::GeV, e_max/CLHEP::GeV));
        }
      }// end of loops over names
      
    }// end of creating Enu histograms
    
    fDCosX = tfs->make<TH1F>("fDCosX", ";dx/ds", 200, -1., 1.);
    fDCosY = tfs->make<TH1F>("fDCosY", ";dy/ds", 200, -1., 1.);
    fDCosZ = tfs->make<TH1F>("fDCosZ", ";dz/ds", 200, -1., 1.);
    
    fMuMomentum = tfs->make<TH1F>("fMuMomentum", ";p_{#mu} (GeV/c)", 500, 0., 50.);
    fMuDCosX = tfs->make<TH1F>("fMuDCosX", ";dx/ds;", 200, -1., 1.);
    fMuDCosY = tfs->make<TH1F>("fMuDCosY", ";dy/ds;", 200, -1., 1.);
    fMuDCosZ = tfs->make<TH1F>("fMuDCosZ", ";dz/ds;", 200, -1., 1.);
    
    fEMomentum  = tfs->make<TH1F>("fEMomentum", ";p_{e} (GeV/c)", 500, 0., 50.);
    fEDCosX = tfs->make<TH1F>("fEDCosX", ";dx/ds;", 200, -1., 1.);
    fEDCosY = tfs->make<TH1F>("fEDCosY", ";dy/ds;", 200, -1., 1.);
    fEDCosZ = tfs->make<TH1F>("fEDCosZ", ";dz/ds;", 200, -1., 1.);
    
    fMode = tfs->make<TH1F>("fMode", ";Interaction Mode;", 4, 0., 4.);
    fMode->GetXaxis()->SetBinLabel(1, "QE");
    fMode->GetXaxis()->SetBinLabel(2, "Res");
    fMode->GetXaxis()->SetBinLabel(3, "DIS");
    fMode->GetXaxis()->SetBinLabel(4, "Coh");
    fMode->GetXaxis()->CenterLabels();
    
    fECons  = tfs->make<TH1F>("fECons", ";#Delta E(#nu,lepton);", 500, -5., 5.);

    fProtonMultiplicityTi48 = tfs->make<TH1F>("fProtonMultiplicityTi48", ";Proton Multiplicity Ti_{48};", 15, 0., 15.);
    fProtonMultiplicityFe56 = tfs->make<TH1F>("fProtonMultiplicityFe56", ";Proton Multiplicity Fe_{56};", 15, 0., 15.);
    fProtonMultiplicityCl35 = tfs->make<TH1F>("fProtonMultiplicityCl35", ";Proton Multiplicity Cl_{35};", 15, 0., 15.);
    fProtonMultiplicityC12 = tfs->make<TH1F>("fProtonMultiplicityC12", ";Proton Multiplicity C_{12};", 15, 0., 15.);
    fProtonMultiplicityH = tfs->make<TH1F>("fProtonMultiplicityH", ";Proton Multiplicity H;", 15, 0., 15.);
    fProtonMultiplicityO16 = tfs->make<TH1F>("fProtonMultiplicityO16", ";Proton Multiplicity O_{16};", 15, 0., 15.);
    fProtonMultiplicityN14 = tfs->make<TH1F>("fProtonMultiplicityN14", ";Proton Multiplicity N_{14};", 15, 0., 15.);
    fProtonMultiplicityS32 = tfs->make<TH1F>("fProtonMultiplicityS32", ";Proton Multiplicity S_{32};", 15, 0., 15.);
    fProtonMultiplicityAllNucl = tfs->make<TH1F>("fProtonMultiplicityAllNucl", ";Proton Multiplicity All Nuclei;", 15, 0., 15.);

    fVertexHadronDecay.push_back(tfs->make<TH1F>("fVertexHadronDecayX", ";X (m);", 200, -1, 1.));
    fVertexHadronDecay.push_back(tfs->make<TH1F>("fVertexHadronDecayY", ";Y (m);", 200, -1, 1.));
    fVertexHadronDecay.push_back(tfs->make<TH1F>("fVertexHadronDecayZ", ";Z (m);", 200, -100, 1000.));

    fNeutrinoTrueL = tfs->make<TH1F>("fNeutrinoTrueL", ";L (m);", 201, -0.05, 2000.05);

    fnuQEEnergy = tfs->make<TH1F>("fnuQEEnergy",";#nu energy in QE;",150,0,30);
    fnuResEnergy= tfs->make<TH1F>("fnuResEnergy",";#nu energy in Res;",150,0,30);
    fnuDISEnergy= tfs->make<TH1F>("fnuDISEnergy",";#nu energy in DIS;",150,0,30);
    fnuCohEnergy= tfs->make<TH1F>("fnuCohEnergy",";#nu energy in Coh;",150,0,30);
    
    fnuKlongEnergy = tfs->make<TH1F>("fnuKlongEnergy",";#nu energy from K_{L};",150,0,30);
    fnueKlongEnergy = tfs->make<TH1F>("fnueKlongEnergy",";#nu_{e} energy from K_{L};",150,0,30);
    fnumuKlongEnergy = tfs->make<TH1F>("fnumuKlongEnergy",";#nu_{#mu} energy from K_{L};",150,0,30);
    fnuKchargedEnergy = tfs->make<TH1F>("fnuKchargedEnergy",";#nu energy from K^{+/-};",150,0,30);
    fnueKchargedEnergy = tfs->make<TH1F>("fnueKchargedEnergy",";#nu_{e} energy from K^{+/-};",150,0,30);
    fnumuKchargedEnergy = tfs->make<TH1F>("fnumuKchargedEnergy",";#nu_{#mu} energy from K^{+/-};",150,0,30);
    fnuPiEnergy = tfs->make<TH1F>("fnuPiEnergy",";#nu energy from pions;",150,0,30);
    fnuePiEnergy = tfs->make<TH1F>("fnuePiEnergy",";#nu_{e} energy from pions;",150,0,30);
    fnumuPiEnergy = tfs->make<TH1F>("fnumuPiEnergy",";#nu_{#mu} energy from pions;",150,0,30);
    fnuMuEnergy = tfs->make<TH1F>("fnuMuEnergy",";#nu energy from muons;",150,0,30);
    fnueMuEnergy = tfs->make<TH1F>("fnueMuEnergy",";#nu_{e} energy from muons;",150,0,30);
    fnumuMuEnergy = tfs->make<TH1F>("fnumuMuEnergy",";#nu_{#mu} energy from muons;",150,0,30);
    
    fW = tfs->make<TH1F>("fW", ";W (shower mass);", 100, 0., 2.);
    fX = tfs->make<TH1F>("fX", ";X_{Bj};", 100, 0., 1.);
    fY = tfs->make<TH1F>("fY", ";Y_{Bj};", 100, 0., 1.);
    fQSqr = tfs->make<TH1F>("fQSqr", ";Q^{2};", 100, 0., 2.);
    
    fNeutrinoEnuDiffLeptonBest = tfs->make<TH1F>("fNeutrinoEnuDiffLeptonBest", ";#Delta E (GeV);", 51, -0.5, 0.5);
    fNeutrinoEnuDiffProtonBest = tfs->make<TH1F>("fNeutrinoEnuDiffProtonBest", ";#Delta E (GeV);", 51, -0.5, 0.5);
    fNeutrinoEnuDiffBothBest   = tfs->make<TH1F>("fNeutrinoEnuDiffBothBest"  , ";#Delta E (GeV);", 51, -0.5, 0.5);
    
    fNeutrinoDirectionDiffBest = tfs->make<TH1F>("fNeutrinoDirectionDiffBest", ";angle (degrees);", 1800, 0, 180.0);
    fNeutrinoDirectionDiffHPndGenie = tfs->make<TH1F>("fNeutrinoDirectionDiffHPndGenie", ";angle (degrees);", 1800, 0, 180.0);
    
    
    fNeutrinoU_Genie.push_back(tfs->make<TH1F>("fNeutrinoUX_Genie", ";Ux", 1800, -1, 1));
    fNeutrinoU_Genie.push_back(tfs->make<TH1F>("fNeutrinoUY_Genie", ";Uy", 1800, -1, 1));
    fNeutrinoU_Genie.push_back(tfs->make<TH1F>("fNeutrinoUZ_Genie", ";Uz", 1800, -1, 1));
    
    
    fNeutrinoU_HP.push_back(tfs->make<TH1F>("fNeutrinoUX_HP"    , ";Ux", 1800, -1, 1));
    fNeutrinoU_HP.push_back(tfs->make<TH1F>("fNeutrinoUY_HP"    , ";Uy", 1800, -1, 1));
    fNeutrinoU_HP.push_back(tfs->make<TH1F>("fNeutrinoUZ_HP"    , ";Uz", 1800, -1, 1));
    
    
    fNeutrinoTrueETrueL = tfs->make<TH2F>("fNeutrinoTrueETrueL"  , ";E (GeV); L (m)", 1000, 0.0, 10.0 ,10000, 100, 1100);
    
  }
  
  //......................................................................
  
  void NeutrinoAna::analyze(art::Event const& evt)  {

    /// Current event number
    fCurrentEvent = evt.id().event();
    
    // Pull the MC generator information out of the event
    art::Handle< std::vector<simb::MCTruth> > mclist;
    evt.getByLabel(fMCTruthModuleLabel, mclist);
    if (mclist->empty()) {
      std::cerr << "Error retrieving MCTruth list" << std::endl;
      return;
    }
    
    art::Handle< std::vector<simb::MCFlux> > fllist;
    evt.getByLabel(fMCTruthModuleLabel, fllist);
    if (fllist->empty()) {
      if(msg1cnt < 5){
	std::cerr << "ooops file made before MCFlux existed.  let it slide" 
		  << std::endl;
	++msg1cnt;
      }
    }


    std::cout<<"List sizes:"<<mclist->size()<<" "<<fllist->size()<<"\n";
    
    // uncomment me when ready to use
    // double neutrino_angle;
    
    /// Loop over neutrino interactions
    for(unsigned int i_intx = 0; i_intx < mclist->size(); ++i_intx){
      
      /// Pointers to current MCTruth and MCFlux
      art::Ptr<simb::MCTruth> mc    (mclist,i_intx);
      art::Ptr<simb::MCFlux>  mcflux(fllist,i_intx);
      
      /// Calculating true L, E, L/E from neutrino information
      /// and those measured from the final state particles
      analyzeNeutrinoInteraction(mc, mcflux);
      
      /// Link to the MCNeutrino class.
      /// The class contains information not only about
      /// the incoming neutrino, but about the products of the decay
      const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
      
      // Decide which histograms to put the spectrum in
      std::vector <int> hist_id;
      
      if (mc_neutrino.CCNC()==simb::kCC) {
	if      (mc_neutrino.Nu().PdgCode() ==  12) hist_id.push_back(0);
	else if (mc_neutrino.Nu().PdgCode() == -12) hist_id.push_back(1);
	else if (mc_neutrino.Nu().PdgCode() ==  14) hist_id.push_back(2);
	else if (mc_neutrino.Nu().PdgCode() == -14) hist_id.push_back(3);
	else return;
      }
      else {
        /// neutrino or antineutrino
	if (mc_neutrino.Nu().PdgCode() > 0) hist_id.push_back(4);
	else                                hist_id.push_back(5);
      }/// end of checking whether it's CC interaction
      
      /// If id is still undefined, something is wrong
      if (hist_id.size() == 0) abort();
      
      /// Figure out the neutrino interaction type
      {
        /*
	  1 = QE
	  2 = Pi0_NoChargedLepton
	  3 = Pi0_ChargedLepton
	  4 = Pi+_NoChargedLepton
	  5 = Pi+_ChargedLepton
	  6 = Pi-_NoChargedLepton
	  7 = Pi-_ChargedLepton
        */
	
        /// Interaction types are defined in SimulationBase/MCNeutrino.h
        /// my_interaction_type is the one identified by fEnuForDifferentProcesses histograms in the constructor
        int my_interaction_type = -1;
        switch(mc_neutrino.InteractionType()){
	case simb::kCCQE:
	case simb::kNCQE:
	  my_interaction_type = 1;
	  break;
	case simb::kResNCNuProtonPi0:
	case simb::kResNCNuBarProtonPi0:
	case simb::kResCCNuBarProtonPi0:
	  my_interaction_type = 2;
	  break;
	case simb::kResNCNuNeutronPi0:
	case simb::kResNCNuBarNeutronPi0:
	case simb::kResCCNuNeutronPi0:
	  my_interaction_type = 3;
	  break;
	case simb::kResCCNuProtonPiPlus:
	case simb::kResNCNuProtonPiPlus:
	case simb::kResNCNuBarProtonPiPlus:
	  my_interaction_type = 4;
	  break;
	case simb::kResCCNuNeutronPiPlus:
	  my_interaction_type = 5;
	  break;
	case simb::kResCCNuBarProtonPiMinus:
	  my_interaction_type = 6;
	  break;
	case simb::kResNCNuNeutronPiMinus:
	case simb::kResCCNuBarNeutronPiMinus:
	case simb::kResNCNuBarNeutronPiMinus:
	  my_interaction_type = 7;
	  break;
	default:
	  break;
        }// end of switch
	
        /// Add a histogram id for the histoghram with the interaction type to fill
        if(my_interaction_type > 0){
	  int number_of_base_types = 6; // number of different base neutrino interactions (numuCC, numubarCC, etc defined in the constructor)
	  hist_id.push_back(my_interaction_type*number_of_base_types + hist_id[0]);
        }// end of filling histogram id
      }// end of figuring out the neutrino interaction type
      
      //std::cout<<"Filling histos:";
      /// Fill the neutrino specta histograms
      for(unsigned int i_id = 0; i_id < hist_id.size(); i_id++){
	int histo_id = hist_id[i_id];
	//std::cout<<histo_id<<" ";
	fEnuForDifferentProcesses[histo_id]->Fill( mc_neutrino.Nu().E() );
	//std::cout<<fEnuForDifferentProcesses[histo_id]->GetName()<<" ";
	
      }
      //std::cout<<"\n";
      
      
      fMode->Fill(mc_neutrino.Mode());
      
      const simb::MCParticle& nu = mc_neutrino.Nu();
      
      fVertexX->Fill(nu.Vx());
      fVertexY->Fill(nu.Vy());
      fVertexZ->Fill(nu.Vz());
      
      fVertexXY->Fill(nu.Vx(), nu.Vy());
      fVertexXZ->Fill(nu.Vz(), nu.Vx());
      fVertexYZ->Fill(nu.Vz(), nu.Vy());
      
      //fVertex->Fill(nu.Vz(), nu.Vx(), nu.Vy());
      
      if(fabs(nu.P()) > 0.){
	fDCosX->Fill(nu.Px()/nu.P());
	fDCosY->Fill(nu.Py()/nu.P());
	fDCosZ->Fill(nu.Pz()/nu.P());
      }/// end of checking whether P is nonzero
      
      //angle between neutrino direction and the normal to the detector
      
      //neutrino direction's components
      //unused//double neutrino_compX,neutrino_compY,neutrino_compZ;

      //unused//neutrino_compX=nu.Px()/nu.P();
      //unused//neutrino_compY=nu.Py()/nu.P();
      //unused//neutrino_compZ=nu.Pz()/nu.P();

      //normal to the detector components
      //unused//double normal_X,normal_Y,normal_Z;

      //unused//normal_X=0;
      //unused//normal_Y=0;
      //unused//normal_Z=1;

      // uncomment me when ready to actually use me
      // calculating neutrino angle
      // double dot_product=neutrino_compX*normal_X+neutrino_compY*normal_Y+neutrino_compZ*normal_Z;
      // neutrino_angle=acos(dot_product);
    
      if(mc_neutrino.CCNC() == simb::kCC){
	
	/// Look for the outgoing lepton in the particle stack
	/// just interested in the first one
	for(int i_particles = 0; i_particles < mc->NParticles(); ++i_particles){
	  const simb::MCParticle& part = mc->GetParticle(i_particles);
	  if(abs(part.PdgCode()) == 11){
	    fEMomentum->Fill(part.P());
	    fEDCosX->Fill(part.Px()/part.P());
	    fEDCosY->Fill(part.Py()/part.P());
	    fEDCosZ->Fill(part.Pz()/part.P());
	    fECons->Fill(nu.E() - part.E());
	    break;
	  }
	  else if(abs(part.PdgCode()) == 13){
	    fMuMomentum->Fill(part.P());
	    fMuDCosX->Fill(part.Px()/part.P());
	    fMuDCosY->Fill(part.Py()/part.P());
	    fMuDCosZ->Fill(part.Pz()/part.P());
	    fECons->Fill(nu.E() - part.E());
	    break;
	  }
	}//end loop over particles
      }//end if CC
      
      //Proton multiplicity -- Look for protons in the particle stack
      int nprotons = 0; ///< number of protons
      
      ///Loop over particles produced in the event
      for(int i_particles = 0; i_particles < mc->NParticles(); ++i_particles){
	const simb::MCParticle& part = mc->GetParticle(i_particles);
	if(part.PdgCode() == 2212)
	  nprotons++;
      }
      
      if(mc_neutrino.Target() == 1000220480) fProtonMultiplicityTi48->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000260560) fProtonMultiplicityFe56->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000170350) fProtonMultiplicityCl35->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000060120) fProtonMultiplicityC12->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000010010) fProtonMultiplicityH->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000080160) fProtonMultiplicityO16->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000070140) fProtonMultiplicityN14->Fill(nprotons);
      else if(mc_neutrino.Target() == 1000160320) fProtonMultiplicityS32->Fill(nprotons);
      fProtonMultiplicityAllNucl->Fill(nprotons);
      
      if(mc_neutrino.Mode()==0) fnuQEEnergy->Fill(nu.E());
      else if(mc_neutrino.Mode()==1) fnuResEnergy->Fill(nu.E());
      else if(mc_neutrino.Mode()==2) fnuDISEnergy->Fill(nu.E());
      else if(mc_neutrino.Mode()==3) fnuCohEnergy->Fill(nu.E());
      
      if(mcflux->fndecay>0 && mcflux->fndecay<5){
      	fnuKlongEnergy->Fill(nu.E());
      	if(abs(nu.PdgCode()) == 12) fnueKlongEnergy->Fill(nu.E());
      	else if(abs(nu.PdgCode()) == 14) fnumuKlongEnergy->Fill(nu.E());
      }
      else if(mcflux->fndecay>4 && mcflux->fndecay<11){ 
      	fnuKchargedEnergy->Fill(nu.E());
      	if(abs(nu.PdgCode()) == 12) fnueKchargedEnergy->Fill(nu.E());
      	else if(abs(nu.PdgCode()) == 14) fnumuKchargedEnergy->Fill(nu.E());
      }
      else if(mcflux->fndecay>12 && mcflux->fndecay<15){ 
      	fnuPiEnergy->Fill(nu.E());
      	if(abs(nu.PdgCode()) == 12) fnuePiEnergy->Fill(nu.E());
      	else if(abs(nu.PdgCode()) == 14) fnumuPiEnergy->Fill(nu.E());
      }
      else if(mcflux->fndecay>10 && mcflux->fndecay<13){ 
      	fnuMuEnergy->Fill(nu.E());
      	if(abs(nu.PdgCode()) == 12) fnueMuEnergy->Fill(nu.E());
      	else if(abs(nu.PdgCode()) == 14) fnumuMuEnergy->Fill(nu.E());
      }

      fW->Fill(mc_neutrino.W());
      fX->Fill(mc_neutrino.X());
      fY->Fill(mc_neutrino.Y());
      fQSqr->Fill(mc_neutrino.QSqr());
    }//end loop over interactions
    
  }
  ////////////////////////////////////////////////////////////////////////
  
  
  //----------------------------------------------------------------------------
  bool NeutrinoAna::analyzeNeutrinoInteraction(const art::Ptr<simb::MCTruth>& mc,
                                               const art::Ptr<simb::MCFlux>&  mcflux) {
    

    /// Link to the MCNeutrino class.
    /// The class contains information not only about
    /// the incoming neutrino, but about the products of the decay
    const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
    
    /// Incoming neutrino
    const simb::MCParticle& incoming_neutrino = mc_neutrino.Nu();
    /// Outgoing lepton
    const simb::MCParticle& outgoing_lepton   = mc_neutrino.Lepton();

    
    /// Neutrino interaction vertex in detector coordinates
    double neutrino_true_end_detector_coords[3];
    neutrino_true_end_detector_coords[0] = outgoing_lepton.Vx() * CLHEP::cm;
    neutrino_true_end_detector_coords[1] = outgoing_lepton.Vy() * CLHEP::cm;
    neutrino_true_end_detector_coords[2] = outgoing_lepton.Vz() * CLHEP::cm;
    
    /// Neutrino starting positions added by Giuseppe Ferone and Denis Perevalov
    fNeutrinoStartingPositionX.push_back(mcflux->fvx);
    fNeutrinoStartingPositionY.push_back(mcflux->fvy);
    fNeutrinoStartingPositionZ.push_back(mcflux->fvz);
    
    
    
    /// If outside of the detector, don't bother
    if(!geo->isInsideFiducialVolume(neutrino_true_end_detector_coords)) return true;
    
    /// True neutrino energy. Dimensionless.
    fEnuTrue = incoming_neutrino.E() * CLHEP::GeV;
    
    /// True neutrino direction
    double UnuTrue[3];
    UnuTrue[0] = incoming_neutrino.Px() / incoming_neutrino.P();
    UnuTrue[1] = incoming_neutrino.Py() / incoming_neutrino.P();
    UnuTrue[2] = incoming_neutrino.Pz() / incoming_neutrino.P();
    
    
    /// Neutrino origin vertex in beam coordinates. Dimensionless.
    double neutrino_true_start_beam_coords[3];
    
    /// Position of hadron/muon decay (from mcflux)
    neutrino_true_start_beam_coords[0] = mcflux->fvx * CLHEP::cm;
    neutrino_true_start_beam_coords[1] = mcflux->fvy * CLHEP::cm;
    neutrino_true_start_beam_coords[2] = mcflux->fvz * CLHEP::cm;
    
    /// Neutrino Rays
    //neutrino_true_start_beam_coords[0] = mcflux->fgenx * CLHEP::cm;
    //neutrino_true_start_beam_coords[1] = mcflux->fgeny * CLHEP::cm;
    //neutrino_true_start_beam_coords[2] = mcflux->fgenz * CLHEP::cm;
    
    /// Fill Neutrino production vertex
    fVertexHadronDecay[0]->Fill(neutrino_true_start_beam_coords[0] / CLHEP::m);
    fVertexHadronDecay[1]->Fill(neutrino_true_start_beam_coords[1] / CLHEP::m);
    fVertexHadronDecay[2]->Fill(neutrino_true_start_beam_coords[2] / CLHEP::m);
    
    /// Neutrino origin vertex in detector coordinates
    double neutrino_true_start_detector_coords[3];
    
    geo::CoordinateTransformation* coordinate_transformation = geo->getCoordinateTransformation();
    
    /// Try to calculate detector coordinates. If can't calculate, return.
    if(! coordinate_transformation->transformBeamToDetectorCoordinates(neutrino_true_start_beam_coords,
								       neutrino_true_start_detector_coords ) ) return false;
    
    /// Neutrino direction in the detector coordinates
    double Unu_beam[3];
    
    /// Calculate L and Unu_beam
    double L = calculateL(neutrino_true_start_detector_coords, neutrino_true_end_detector_coords, Unu_beam);
    
    /// Fill L in meters
    fNeutrinoTrueL->Fill(L/CLHEP::m);
    
    /// Fill E vs L
    fNeutrinoTrueETrueL->Fill(fEnuTrue/CLHEP::GeV, L/CLHEP::m);
    
    /// If not CC, return now
    if(mc_neutrino.CCNC() != simb::kCC) return false;
    
    
    /// Find most energetic proton
    simb::MCParticle most_energetic_proton;
    std::vector <int> proton_pid;
    proton_pid.push_back(2212);
    int nprotons = getMostEnergeticParticle(mc, proton_pid, most_energetic_proton, false);
    
    /// Find most energetic neutron
    simb::MCParticle most_energetic_neutron;
    std::vector <int> neutron_pid;
    neutron_pid.push_back(2112);
    int nneutrons = getMostEnergeticParticle(mc, neutron_pid, most_energetic_neutron, false);
    
    /*
   /// Find most energetic lepton
   simb::MCParticle most_energetic_lepton;
   std::vector <int> lepton_pid;
   lepton_pid.push_back(11);
   lepton_pid.push_back(13);
   lepton_pid.push_back(15);
   int nleptons = getMostEnergeticParticle(mc, lepton_pid, most_energetic_lepton);
    */
    
    /// Find most energetic pion
    simb::MCParticle most_energetic_pion;
    std::vector <int> pion_pid;
    pion_pid.push_back(111);
    pion_pid.push_back(211);
    pion_pid.push_back(-211);
    int npions = getMostEnergeticParticle(mc, pion_pid, most_energetic_pion);
    
    /// Only one proton
    if(nprotons != 1 ) return false;
    
    /// No neutrons
    if(nneutrons > 0) return false;
    
    /// No pions
    if(npions > 0) return false;
    
    double Enu;
    double Unu[3];
    
    calculateNeutrino(outgoing_lepton, most_energetic_proton, Unu, Enu);
    //calculateNeutrino(most_energetic_lepton, most_energetic_proton, Unu, Enu);
    
    double angle_between_true_and_best = acos(dotProduct(Unu, UnuTrue)) / CLHEP::degree;
    double angle_between_true_and_HP   = acos(dotProduct(Unu_beam, UnuTrue)) / CLHEP::degree;
    
    
    
    fNeutrinoDirectionDiffBest->Fill(angle_between_true_and_best);
    fNeutrinoDirectionDiffHPndGenie->Fill(angle_between_true_and_HP);
    
    
    if(angle_between_true_and_best>20)
      fBadEventNumber.push_back(fCurrentEvent);
    
    
    double Ux[3]={1,0,0};
    double Uy[3]={0,1,0};
    double Uz[3]={0,0,1};
    
    fNeutrinoU_Genie[0]->Fill(dotProduct(UnuTrue, Ux));
    fNeutrinoU_HP[0]->Fill(dotProduct(Unu_beam, Ux));
    
    fNeutrinoU_Genie[1]->Fill(dotProduct(UnuTrue, Uy));
    fNeutrinoU_HP[1]->Fill(dotProduct(Unu_beam, Uy));
    
    fNeutrinoU_Genie[2]->Fill(dotProduct(UnuTrue, Uz));
    fNeutrinoU_HP[2]->Fill(dotProduct(Unu_beam, Uz));
    
    //std::cout<<"Number of protons "<< nprotons<<"\n";
    
    return true;
  }
  
  //------------------------------------------------------------------------
  /// Calculate L from the start and end point
  double NeutrinoAna::calculateL(double* start_point, double* end_point, double* U){
    
    double L = sqrt(
		    (start_point[0] - end_point[0]) * (start_point[0] - end_point[0])
		    + (start_point[1] - end_point[1]) * (start_point[1] - end_point[1])
		    + (start_point[2] - end_point[2]) * (start_point[2] - end_point[2])
		    );
    
    /// Fill direction
    for(int i = 0; i < 3; i++)
      U[i] = (end_point[i] - start_point[i]) / L;
    
    return L;
  }
  
  
  //--------------------------------------------------------------------------
  bool NeutrinoAna::calculateNeutrino(const simb::MCParticle& input_lepton,
                                      const simb::MCParticle& input_proton,
                                      double* neutrino_direction,
                                      double& Enu) 
  {
    
    double ml = input_lepton.Mass() * CLHEP::GeV; ///< Lepton mass
    double mp = input_proton.Mass() * CLHEP::GeV; ///< Proton mass
    
    double El = input_lepton.E() * CLHEP::GeV;    ///< Lepton energy
    double Ep = input_proton.E() * CLHEP::GeV;    ///< Proton energy
    
    double Pl = input_lepton.P() * CLHEP::GeV;    ///< Lepton momentum
    double Pp = input_proton.P() * CLHEP::GeV;    ///< Proton momentum
    
    
    /// Lepton and proton momentum must be non-zero
    if(Pp < 0.01 / CLHEP::GeV) return false;
    if(Pl < 0.01 / CLHEP::GeV) return false;
    
    /// Can't be NAN
    //if(Pl != Pl) return false;
    //if(Pp != Pp) return false;
    
    /// Lepton direction
    double Ul[3] = { input_lepton.Px()/input_lepton.P(),
		     input_lepton.Py()/input_lepton.P(),
		     input_lepton.Pz()/input_lepton.P() };
    
    /// Proton direction
    double Up[3] = { input_proton.Px()/input_proton.P(),
		     input_proton.Py()/input_proton.P(),
		     input_proton.Pz()/input_proton.P() };
    
    /// angle between lepton and proton
    double alpha = acos(dotProduct(Ul, Up)) * CLHEP::radian;
    
    /*
    // Calculating Enu from proton and from lepton information
    {
    double A = (ml*ml/2.0 + mp*mp + Ep*mp) / (mp*El - ml*ml/2.0);
    
    double B = mp - Ep;
    
    double C = mp - El;
    
    //double D = A * Pl;
    
    double E = B - A*C;
    
    double F = Pp*cos(alpha/CLHEP::radian) - A*Pl;
    
    double G = Pp*sin(alpha/CLHEP::radian);
    
    double H = F*F + G*G;
    
    double I = E*E - G*G;
    
    double J = 2.0*F*E;
    
    double costheta1 = 0.5*(J + sqrt(J*J-4*H*I)) / H;
    double costheta2 = 0.5*(J - sqrt(J*J-4*H*I)) / H;
    
    double phi1 = alpha - acos(costheta1);
    double phi2 = alpha - acos(costheta2);
    
    
    double Enu_lepton1 = (mp*El - ml*ml/2.0) / (mp - El + Pl*costheta1);
    double Enu_lepton2 = (mp*El - ml*ml/2.0) / (mp - El + Pl*costheta2);
    
    double Enu_proton1 = (ml*ml/2.0 - mp*mp + Ep*mp)/(mp-Ep+Pp*cos(phi1));
    double Enu_proton2 = (ml*ml/2.0 - mp*mp + Ep*mp)/(mp-Ep+Pp*cos(phi2));
    }
    */
    
    // Calculate best costh
    double costheta_best = -1;
    double dcosth = 0.001;
    double chi2_best = 1000000000000.0;
    for(double costh = -1.0; costh < 1.0; costh += dcosth){
      
      double phi = alpha - acos(costh);
      
      double Enu_lepton = (mp*El - ml*ml/2.0) / (mp - El + Pl*costh);
      double Enu_proton = (ml*ml/2.0 - mp*mp + Ep*mp)/(mp-Ep+Pp*cos(phi));
      
      if(Enu_lepton < 0.0 || Enu_proton < 0.0)continue;
      
      double chi2 = (Enu_lepton - Enu_proton) * (Enu_lepton - Enu_proton);
      
      if(chi2 < chi2_best){
	costheta_best = costh;
	chi2_best = chi2;
      }
      
    }/// end of loop over costh bins
    
    double phi = alpha - acos(costheta_best);
    double Enu_lepton = (mp*El - ml*ml/2.0) / (mp - El + Pl*costheta_best);
    double Enu_proton = (ml*ml/2.0 - mp*mp + Ep*mp)/(mp-Ep+Pp*cos(phi));
    
    fNeutrinoEnuDiffLeptonBest->Fill( (Enu_lepton - fEnuTrue) / CLHEP::GeV );
    fNeutrinoEnuDiffProtonBest->Fill( (Enu_proton - fEnuTrue) / CLHEP::GeV );
    fNeutrinoEnuDiffBothBest->Fill( (0.5*Enu_lepton + 0.5*Enu_proton - fEnuTrue) / CLHEP::GeV );
    
    
    /// Fill Enu reconstructed
    Enu = 0.5*Enu_lepton + 0.5*Enu_proton - fEnuTrue;
    
    
    /// Now get the neutrino direction
    double C1 = (costheta_best - cos(alpha)*cos(phi)     ) /sin(alpha)/sin(alpha);
    double C2 = (cos(phi)      - cos(alpha)*costheta_best) /sin(alpha)/sin(alpha);
    
    for(int i = 0; i <  3; i++){
      neutrino_direction[i] = C1*Ul[i] + C2*Up[i];
    }
    
    return true;
  }
  
  
  //-------------------------------------------------------------------------
  /// Calculate dot product between two vectors
  double NeutrinoAna::dotProduct(double* v1, double* v2) 
  {
    double product = 0.0;
    
    for(int i = 0; i < 3; i++)
      product += v1[i] * v2[i];
    
    return product;
  }
  
  //-------------------------------------------------------------------------
  int NeutrinoAna::getMostEnergeticParticle(const art::Ptr<simb::MCTruth>& mc,
                                            const std::vector <int>& pids,
                                            simb::MCParticle& most_energetic_particle,
                                            bool print) 
  {
    
    /// Find most energetic proton
    int nparticles = 0; ///< number of protons
    
    
    for(int iparticle = 0; iparticle < mc->NParticles(); iparticle++){
      simb::MCParticle* particle = (simb::MCParticle*)&mc->GetParticle(iparticle);
      if(print)std::cout<<" pdg="<<particle->PdgCode()<<" P="<<particle->P();
    }
    
    
    if(print)std::cout<<"\n";
    
    
    for(int iparticle = 0; iparticle < mc->NParticles(); iparticle++){
      simb::MCParticle* particle = (simb::MCParticle*)&mc->GetParticle(iparticle);
      if(particle->StatusCode() != 1) continue;
      if(particle->P() < 0.05) continue;
      if(print)std::cout<<" pdggood="<<particle->PdgCode()<<" P="<<particle->P()<<" SC="<<particle->StatusCode();
    }
    if(print)std::cout<<"\n";
    
    
    /// Loop over particles
    for(int iparticle = 0; iparticle < mc->NParticles(); iparticle++){
      
      simb::MCParticle* particle = (simb::MCParticle*)&mc->GetParticle(iparticle);
      
      
      /// If the particle doesn't leave the nucleus, forget it
      if(particle->StatusCode() != 1 ) continue;
      
      bool required_pid = false;
      for(unsigned int ipid = 0; ipid < pids.size(); ipid++)
	if(particle->PdgCode() == pids[ipid]) required_pid=true;
      
      if(!required_pid) continue;
      
      //std::cout<<" pdgVeryGood="<<particle->PdgCode()<<" E="<<particle->E()<<" Mass="<<particle->Mass() * GeV;
      
      /// If it's the first proton, it's going to be the most energetic for now
      if(nparticles == 0){
	most_energetic_particle = *particle;
	nparticles = 1;
	continue;
      }// end if it's the first proton
      
      if(particle->E() > most_energetic_particle.E())
	most_energetic_particle = *particle;
      
      nparticles++;
    } // end of loop over particles in MC truth
      //std::cout<<"\n";
    //std::cout<<"Out Mass="<<most_energetic_particle.Mass() * GeV<<"\n";
    return nparticles;
  }
  
  void NeutrinoAna::endJob()
  {  
    if(fNeutrinoStartingPositionX.size() == 0) return;
    
    
    TH1F* neutrino_starting_position_x;
    TH1F* neutrino_starting_position_y;
    TH1F* neutrino_starting_position_z;
    TH2F* neutrino_starting_position_xy;
    TH2F* neutrino_starting_position_xz;
    TH2F* neutrino_starting_position_yz;
    
    double min_x,max_x;
    min_x=fNeutrinoStartingPositionX[0];
    max_x=fNeutrinoStartingPositionX[0];
    
    for(unsigned int j=0;j< fNeutrinoStartingPositionX.size();j++)
      
      {
	if(fNeutrinoStartingPositionX[j]>max_x)       max_x=fNeutrinoStartingPositionX[j];
	else if(fNeutrinoStartingPositionX[j]<min_x)  min_x=fNeutrinoStartingPositionX[j];
      }
    
    neutrino_starting_position_x= tfs->make<TH1F>("neutrino_starting_position_x","Starting neutrino position x coordinate",200,min_x-0.01,max_x +0.01);
    
    double min_y,max_y;
    min_y=fNeutrinoStartingPositionY[0];
    max_y=fNeutrinoStartingPositionY[0];
    
    for(unsigned int j=0;j< fNeutrinoStartingPositionY.size();j++)
      
      {
	if(fNeutrinoStartingPositionY[j]>max_y)       max_y=fNeutrinoStartingPositionY[j];
	else if(fNeutrinoStartingPositionY[j]<min_y)  min_y=fNeutrinoStartingPositionY[j];
	
      }
    
    neutrino_starting_position_y= tfs->make<TH1F>("neutrino_starting_position_y","Starting neutrino position y coordinate",200,min_y-0.01, max_y +0.01);
    
    
    double min_z,max_z;
    
    min_z=fNeutrinoStartingPositionZ[0];
    max_z=fNeutrinoStartingPositionZ[0];
    
    for(unsigned int j=0;j< fNeutrinoStartingPositionZ.size();j++)
      
      {
	if(fNeutrinoStartingPositionZ[j]>max_z)       max_z=fNeutrinoStartingPositionZ[j];
	else if(fNeutrinoStartingPositionZ[j]<min_z)  min_z=fNeutrinoStartingPositionZ[j];
	
      }
    
    neutrino_starting_position_z= tfs->make<TH1F>("neutrino_starting_position_z","Starting neutrino position z coordinate",200,min_z-0.01,max_z +0.01);
    
    neutrino_starting_position_x->GetXaxis()->SetTitle("x (cm)");
    neutrino_starting_position_y->GetXaxis()->SetTitle("y (cm)");
    neutrino_starting_position_z->GetXaxis()->SetTitle("z (cm)");
    
    for(unsigned int i = 0; i< fNeutrinoStartingPositionX.size(); i++)       neutrino_starting_position_x->Fill(fNeutrinoStartingPositionX[i]);
    for(unsigned int i = 0; i< fNeutrinoStartingPositionY.size(); i++)       neutrino_starting_position_y->Fill(fNeutrinoStartingPositionY[i]);
    for(unsigned int i = 0; i< fNeutrinoStartingPositionZ.size(); i++)       neutrino_starting_position_z->Fill(fNeutrinoStartingPositionZ[i]);
    
    
    neutrino_starting_position_xy=tfs->make<TH2F>("neutrino_starting_position_xy","Starting neutrino position xy view",200,min_x-0.01,max_x +0.01,200,min_y-0.01,max_y +0.01);
    neutrino_starting_position_xz=tfs->make<TH2F>("neutrino_starting_position_xz","Starting neutrino position xz view",200,min_x-0.01,max_x +0.01,200,min_z-0.01,max_z +0.01);
    neutrino_starting_position_yz=tfs->make<TH2F>("neutrino_starting_position_yz","Starting neutrino position yz view",200,min_y-0.01,max_y +0.01,200,min_z-0.01,max_z +0.01);
    
    if(fNeutrinoStartingPositionX.size()==fNeutrinoStartingPositionZ.size()){
      for(unsigned int i = 0; i< fNeutrinoStartingPositionX.size(); i++){
        neutrino_starting_position_xy->Fill(fNeutrinoStartingPositionX[i],fNeutrinoStartingPositionY[i]);
      }
    }
    if(fNeutrinoStartingPositionX.size()==fNeutrinoStartingPositionZ.size()){
      for(unsigned int i = 0; i< fNeutrinoStartingPositionX.size(); i++){
        neutrino_starting_position_xz->Fill(fNeutrinoStartingPositionX[i],fNeutrinoStartingPositionZ[i]);
      }
    }
    if(fNeutrinoStartingPositionX.size()==fNeutrinoStartingPositionZ.size()){
      for(unsigned int i = 0; i< fNeutrinoStartingPositionY.size(); i++){
        neutrino_starting_position_yz->Fill(fNeutrinoStartingPositionY[i],fNeutrinoStartingPositionZ[i]); 
      }
    }
    neutrino_starting_position_xy->GetXaxis()->SetTitle("x (cm)");
    neutrino_starting_position_xy->GetYaxis()->SetTitle("y (cm)");
    neutrino_starting_position_xz->GetXaxis()->SetTitle("x (cm)");
    neutrino_starting_position_xy->GetYaxis()->SetTitle("z (cm)");
    neutrino_starting_position_yz->GetXaxis()->SetTitle("y (cm)");
    neutrino_starting_position_xy->GetYaxis()->SetTitle("z (cm)");
    
    
    
    
  }
  
}// end of namespace

//////////////////////////////////////////////////////////////
namespace mcchk
{
  DEFINE_ART_MODULE(NeutrinoAna);
}



