////////////////////////////////////////////////////////////////////////
/// \brief
/// \author 
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <vector>

#include "TH1.h"

#include "SummaryData/POTSum.h"

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Principal/SubRun.h" 
#include "fhiclcpp/ParameterSet.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Services/Optional/TFileService.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h" 
#include "messagefacility/MessageLogger/MessageLogger.h" 


namespace mcchk {
  /// A module to check the results from the Monte Carlo generator
  
  class POTAna : public art::EDAnalyzer {
    
  public:
    explicit POTAna(fhicl::ParameterSet const& pset);
    virtual ~POTAna();

    void analyze(art::Event const& evt);
    void beginSubRun(art::SubRun const& sr);
    void beginJob();
    void endJob();
    void endSubRun(art::SubRun const& sr);
    void reconfigure(const fhicl::ParameterSet& p);

  private:
    void LoadMCInfo(art::Event const& evt);

  private:
  
    int                 fNevents;           ///< Total number of events
    std::vector<double> fNeventsSubrun;     ///< Total number of events per subrun
    std::vector<double> fNeventsPOTSubrun;  ///< Events per POT per subrun
    std::vector<double> fPOTSubrun;         ///< POT per subrun
    std::vector<double> fPOTSpillSubrun;    ///< POT per spill per subrun

    sumdata::POTSum* fPOTSum;
    
    std::string fMCTruthModuleLabel; ///< label for module producing mc truth information
    std::string fPotLabel;           ///< Module that produced the POTSum object

  };
};

////////////////////////////////////////////////////////////////////////
namespace mcchk {

  static int cntEvent = 0; //event number

  //......................................................................
  
  
  POTAna::POTAna(fhicl::ParameterSet const& p)
    : EDAnalyzer(p),
      fNevents(0),
      fPotLabel(p.get< std::string>("PotLabel"))
  {
  }
  

  //......................................................................
  
  POTAna::~POTAna() { }
  
  //......................................................................
  void POTAna::beginJob()
  {
    fPOTSum = new sumdata::POTSum();
  }
  
  //......................................................................
  void POTAna::beginSubRun(art::SubRun const& sr) {

    art::Handle< sumdata::POTSum > p;
    sr.getByLabel(fPotLabel,p);

    fPOTSum->totpot     = p->totpot;
    fPOTSum->totgoodpot = p->totgoodpot;
    fPOTSum->totspills  = p->totspills;
    fPOTSum->goodspills = p->goodspills;

    mf::LogInfo ("POTAna") << "fPOTSum This sub run has "<< fPOTSum->totspills << " total spills with an exposure of " << fPOTSum->totpot << " POT " << std::endl
			   << "with cuts on beam quality, there are " << fPOTSum->goodspills << " good spills with an exposure of " 
			   << fPOTSum->totgoodpot << std::endl;


  }

  //......................................................................
  void POTAna::analyze(art::Event const& evt)
  { 
    fNevents++;

    ++cntEvent;
    
  }
  

  //......................................................................
  
  void POTAna::LoadMCInfo(art::Event const& evt)
  {
  }
    
  
  //......................................................................
  void POTAna::reconfigure(const fhicl::ParameterSet& p)
  {
    fMCTruthModuleLabel = p.get< std::string>("MCTruthModuleLabel"); 
    fPotLabel           = p.get< std::string>("PotLabel");
  }
  
  //......................................................................
  void POTAna::endSubRun(art::SubRun const& sr)
  {

    mf::LogInfo ("POTAna") << "Event Number " << cntEvent << std::endl;
    mf::LogInfo ("POTAna") << "Subrun Number " << sr.subRun() << std::endl;
    mf::LogInfo ("POTAna") << "Event per POT " << cntEvent/fPOTSum->totpot << std::endl;

    
    fNeventsPOTSubrun.push_back(cntEvent/fPOTSum->totpot);
    fPOTSubrun.push_back(fPOTSum->totpot);
    fPOTSpillSubrun.push_back(fPOTSum->totpot/fPOTSum->totspills);
    fNeventsSubrun.push_back( cntEvent);

    cntEvent = 0;
  }

  //......................................................................
  void POTAna::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    TH1F* nevents;
    TH1F* nevents_subrun;
    TH1F* nevents_pot_subrun;
    TH1F* pot_subrun;
    TH1F* pot_spill_subrun;
    
    nevents_pot_subrun  = tfs->make<TH1F>("fNeventsPOTSubrun", "Total number of events", fNeventsPOTSubrun.size()+1, -0.5, fNeventsPOTSubrun.size()+0.5);
    nevents_pot_subrun->GetXaxis()->SetTitle("Run number");
    nevents_pot_subrun->GetYaxis()->SetTitle("Events per POT");

    pot_subrun  = tfs->make<TH1F>("pot_subrun", "POT", fPOTSubrun.size()+1, -0.5, fPOTSubrun.size()+0.5);
    pot_subrun->GetXaxis()->SetTitle("Run number");
    pot_subrun->GetYaxis()->SetTitle("POT");
    pot_subrun->GetYaxis()->SetTitleOffset(1.2);

    pot_spill_subrun  = tfs->make<TH1F>("pot_spill_subrun", "POT per spill", fPOTSpillSubrun.size()+1, -0.5, fPOTSpillSubrun.size()+0.5);
    pot_spill_subrun->GetXaxis()->SetTitle("Run number");
    pot_spill_subrun->GetYaxis()->SetTitle("POT per spill");
    pot_spill_subrun->GetYaxis()->SetTitleOffset(1.2);

    nevents_subrun  = tfs->make<TH1F>("nevents_subrun", "Total number of events", fNeventsSubrun.size()+1, -0.5, fNeventsSubrun.size()+0.5);
    nevents_subrun->GetXaxis()->SetTitle("Run number");
    nevents_subrun->GetYaxis()->SetTitle("Events");

    nevents  = tfs->make<TH1F>("fNevents", "Total number of events", 1, -0.5,0.5);    
    nevents->Fill(fNevents);

    for(unsigned int i = 0; i<fNeventsPOTSubrun.size(); i++)       nevents_pot_subrun->Fill(i, fNeventsPOTSubrun[i]);
    for(unsigned int i = 0; i<fPOTSubrun.size(); i++)              pot_subrun->Fill( i,fPOTSubrun[i]);
    for(unsigned int i = 0; i<fPOTSpillSubrun.size(); i++)         pot_spill_subrun->Fill(i,fPOTSpillSubrun[i]);
    for(unsigned int i = 0; i<fNeventsSubrun.size(); i++)          nevents_subrun->Fill(i, fNeventsSubrun[i]);
    
  } 
} // end mcchk
////////////////////////////////////////////////////////////////////////
namespace mcchk
{
  DEFINE_ART_MODULE(POTAna);
}
