#include "Metadata/MetadataManager.h"

#include <algorithm>
#include <iterator>
#include <cctype>
#include <cstdlib>
#include <sstream>
#include <memory>

#include "TString.h"

using std::pair;
using std::vector;
using std::string;

namespace meta {
  
  //.................................................................
  MetadataManager::MetadataManager() 
  {
    fKeys.clear();
  }
  
  //..................................................................
  std::map< std::string, std::string >& MetadataManager::GetMetadata() 
  {
    //update metadata with information from parsing the pset
    for(auto& it: fPSetMetaMap){
      fMetaMap[it.first] = it.second;
    }
    return fMetaMap;
  }

  //..................................................................
  std::string MetadataManager::ToLower(std::string str)
  {
    std::transform(str.begin(), 
		   str.end(), 
		   str.begin(), 
		   (int(*)(int))std::tolower );
    return str;
  }
  
  //..................................................................
  std::string MetadataManager::CleanWhiteSpace(std::string str)
  {
    std::string out;
    std::copy_if(str.begin(), 
		 str.end(), 
		 std::back_inserter(out), 
		 [](char c){ return !(std::isspace(c));});
    return out;
  }
  
  //......................................................................
  void MetadataManager::ParsePSet(string key, 
				  fhicl::ParameterSet const & inPSet )
  {
    fhicl::ParameterSet pset;
    if(key == "") pset = inPSet;
    else pset = inPSet.get<fhicl::ParameterSet>(key);
    
    std::vector<std::string> pset_keys = pset.get_pset_keys();
    std::vector<std::string> all_keys  = pset.get_keys();
    std::vector<std::string> nonpset_keys;
    
    std::sort(pset_keys.begin(), pset_keys.end());
    std::sort(all_keys.begin(), all_keys.end());
    
    // get keys that are in all, but not in pset 
    std::set_difference(all_keys.begin(), all_keys.end(),
			pset_keys.begin(), pset_keys.end(),
			std::inserter( nonpset_keys, nonpset_keys.begin())
			);
  
  //all non-pset keys are ready to be saved in the vector of keys
    for(auto& it: nonpset_keys){
      std::string newkey;
      if(key == "") newkey = it;
      else newkey = key + (std::string)"." + it;
      fKeys.push_back(newkey);
    }
    //all pset keys should be made into fully qualified keys and 
    //put into a recursive call
  for(auto& it: pset_keys){
    std::string newkey;
    if(key == "") newkey = it;
    else newkey = key + (std::string)"." + it;
    ParsePSet( newkey, inPSet );
  }
  //when it reaches here, it's made it through the pset_key loop and is done
  return;
}

  //..................................................................
  void MetadataManager::DoParsing(fhicl::ParameterSet const & inPSet)
  {
    //first, clear fKeys to reset the parsing
    fKeys.clear();
    
    ParsePSet("", inPSet);
    //loop through keys, find values & fill metadata map 
    
    for(auto& it: fKeys){
      //get value from parameter set
      std::string key   = ToLower(it);
      key   = CleanWhiteSpace(key);
      
      std::string value;
      try{
	// If it's an int, format it to a string ourselves (because
	// the default formatting goes sometimes uses floating point
	// notation).
	value = TString::Format("%d", inPSet.get<int>(it));
      }
      catch(...){
	// Otherwise it's safe to treat as a string
	value = inPSet.get<std::string>(it, "");
	value = CleanWhiteSpace(value);
      }
      
      //is this an unfilled base_release parameter?
      if(key.find("base_release") != std::string::npos && value == "") value = std::getenv("SRT_BASE_RELEASE");
      
      fPSetMetaMap[ key ] = value;
    }
  }
  
  //..............................................................
  void MetadataManager::AddMetadata(std::string key, std::string value)
  {
    key = ToLower(key);
    key = CleanWhiteSpace(key);
    
    value = CleanWhiteSpace(value);
    
    fMetaMap[ key ] = value;
  }
  
} // end namespace meta
