#include "Monopole/SlowMonopoleTrigger.h"

#include <DDTBaseDataProducts/CompareDAQHit.h>

#include <boost/format.hpp>

#include <algorithm>
#include <iostream>



mono::SlowMonopoleTrigger::SlowMonopoleTrigger(novaddt::HitList const& h) : 
  hits_(h)
{
  mf_trigger_decisions_.clear();
  zukai_trigger_decisions_.clear();
  
  // Zukai's FHICL Parameters:
  // FD13DB_smmtrigger:
  // hits_label:     smmhighadcfilter
  // instance:       HEHits
  _hf_min  =  0.2;
  _minHits =   50;
  _deltP   =    2;
  _deltC   =    6;
  _sigmaL  = 0.01;
  _Pmax    =  831;
  _Cmax    =  383;
  _DistMin =  100;
  _vproMin =  2.3;

  // Zukai's Hard-Coded Parameters:
  _deltaT  =  300;
  _sigmaT  =   80;
  _vproMax =  200;

  // my translated parameters:
  min_hits_               = _minHits;
  view_match_time_cut_    = _deltaT;
  view_match_plane_cut_   = _deltP;

  total_planes_           = static_cast<unsigned>(_Pmax);
  total_cells_            = static_cast<unsigned>(_Cmax);
  surface_plane_cut_      = static_cast<unsigned>(_deltP);
  surface_cell_cut_       = static_cast<unsigned>(_deltC);

  track_time_deviation_   = static_cast<double>(_sigmaT);
  
  aspect_ratio_           = 2.82072;

  hits_on_track_fraction_ = _hf_min;

  trigger_decision_time_buffer_ = _deltaT;

  mf_algorithm();
  zukai_algorithm();
}



std::vector<novaddt::TriggerDecision> 
mono::SlowMonopoleTrigger::trigger_decisions() const
{
  print(mf_trigger_decisions_);
  print(zukai_trigger_decisions_, "ZW:");

  assert(is_subset_of(zukai_trigger_decisions_.at(0), 
		      mf_trigger_decisions_.at(0)));
  // assert(mf_trigger_decisions_.at(0) == zukai_trigger_decisions_.at(0));

  return mf_trigger_decisions_;
}



novaddt::HitList
mono::SlowMonopoleTrigger::essential_hits() const
{
  std::cout << "MF: SlowMonopoleTrigger: n essential hits = " 
	    << essential_hits_.size()
	    << std::endl;
  return essential_hits_;
}



bool mono::SlowMonopoleTrigger::mf_algorithm()
{
  bool result = false;
  novaddt::DAQHit trigger_entry_hit, trigger_exit_hit;

  std::map<std::string, novaddt::HitList> hitmap;
  split_by_view(hits_, hitmap);

  for (auto & h : hitmap)
    if (h.second.size() < min_hits_ + 2)
      return false;

  std::cout << "MF: Total Hits: " << hits_.size()
	    << "; xz hits: " << hitmap["x"].size()
	    << "; yz hits: " << hitmap["y"].size()
	    << std::endl;

  hitmap["x_surface"];
  hitmap["x_contained"];
  
  find_view_matched_hits(hitmap);
  
  std::cout << "MF: Surface X hits: " << hitmap["x_surface"].size()
	    << ";  Contained X Hits: " << hitmap["x_contained"].size()
	    << std::endl;
  
  // note that this loop will end as soon as a result is found
  for (auto entry_hit = hitmap["x_surface"].cbegin();
       entry_hit     != hitmap["x_surface"].cend() && !result; 
       ++entry_hit)
  {
    // we add 1, so that we avoid combinations between the same hit
    novaddt::HitList::const_reverse_iterator until(entry_hit + 1);
    
    auto trigger_exit_hit_iterator = std::find_if
      (hitmap["x_surface"].crbegin(), until, 
       [&](novaddt::DAQHit const& exit_hit)
       {
	 auto contained_first =
	 find(hitmap["x_contained"], entry_hit->TDC().val);
	 auto contained_last =
	 find(hitmap["x_contained"], exit_hit.TDC().val);
	 
	 unsigned hits_on_track = std::count_if
	 (contained_first, contained_last, [&](novaddt::DAQHit const& hit)
          { return hit_is_on_track(hit, *entry_hit, exit_hit); });

	 unsigned min_hits_on_track = 
	   hits_on_track_fraction_ * distance(*entry_hit, exit_hit);

	 if (hits_on_track > min_hits_on_track)
	   return true;

	 return false;
       });
    
    if (trigger_exit_hit_iterator != until)
    {
      result = true;
      trigger_entry_hit = *entry_hit;
      trigger_exit_hit = *trigger_exit_hit_iterator;
    }
  }

  if (result)
  {
    auto t0 = trigger_entry_hit.TDC().val;
    auto dt = trigger_exit_hit.TDC().val - t0;
    
    mf_trigger_decisions_.
      emplace_back(t0 - 2 * trigger_decision_time_buffer_,
		   dt + 4 * trigger_decision_time_buffer_,
		   daqdataformats::trigID::TRIG_ID_DATA_SLOWMONO, 1);


    std::cout << "MF: Adding the hits to the essential hits." 
	      << std::endl;
    essential_hits_.push_back(trigger_entry_hit);
    essential_hits_.push_back(trigger_exit_hit);
  }
  
  return result;
}



void mono::SlowMonopoleTrigger::split_by_view
(novaddt::HitList hits,
 std::map<std::string, novaddt::HitList> & hitmap) const
{
  auto bound = std::stable_partition
    (hits.begin(), hits.end(),
     [](novaddt::DAQHit const& h)
     { return h.View().val == daqchannelmap::X_VIEW; });

  hitmap["x"] = novaddt::HitList(std::distance(hits.begin(), bound));
  std::move(hits.begin(), bound, hitmap["x"].begin());

  hitmap["y"] = novaddt::HitList(std::distance(bound, hits.end()));
  std::move(bound, hits.end(), hitmap["y"].begin());
}



void mono::SlowMonopoleTrigger::find_view_matched_hits
(std::map<std::string, novaddt::HitList> & hitmap) const
{
  for (auto const& x_hit : hitmap["x"])
  {
    auto y_first = find(hitmap["y"], x_hit.TDC().val - view_match_time_cut_);
    auto y_last  = find(hitmap["y"], x_hit.TDC().val + view_match_time_cut_);

    /*
      I think that the matching should happen before we check whether 
      the hit is on the surface, but in the interest to retain comparability
      to Zukai's code, let's just do it the way he does it for now.
     */
    // auto y_match = std::find_if
    //   (y_first, y_last, [&](novaddt::DAQHit const& y_hit)
    //    { 
    // 	 return 
    // 	 abs(int(x_hit.Plane().val) - int(y_hit.Plane().val)) 
    // 	 < view_match_plane_cut_; 
    //    });

    // if (y_match != y_last)
    // {
    //   if (hit_is_on_surface(x_hit) || hit_is_on_surface(*y_match))
    //   {
    // 	hitmap["x_surface"].push_back(x_hit);
    //   } else {
    // 	hitmap["x_contained"].push_back(x_hit);
    //   }
    // }

    // Zukai's Method:
    bool met = false;
    for (auto y_hit = y_first; y_hit != y_last; ++y_hit)
    {
      if (abs(int(x_hit.Plane().val) - int(y_hit->Plane().val))
	  < view_match_plane_cut_)
      {
	if (hit_is_on_surface(x_hit) || hit_is_on_surface(*y_hit))
	{
	  hitmap["x_surface"].push_back(x_hit);
	  break;
	} else if (!met) {
	  hitmap["x_contained"].push_back(x_hit);
	}
	met = true;
      }
    }
  }
}



bool mono::SlowMonopoleTrigger::hit_is_on_surface
(novaddt::DAQHit const& hit) const
{
  if (hit.Cell().val < surface_cell_cut_)
    return true;

  if (hit.Cell().val > total_cells_ - surface_cell_cut_)
    return true;

  if (hit.Plane().val < surface_plane_cut_)
    return true;

  if (hit.Plane().val > total_planes_ - surface_plane_cut_)
    return true;

  return false;
}



bool mono::SlowMonopoleTrigger::hit_is_on_track
(novaddt::DAQHit const& hit,
 novaddt::DAQHit const& track_start, novaddt::DAQHit const& track_end) const
{
  if (hit.TDC()       < track_start.TDC() ||
      track_end.TDC() < hit.TDC())
  {
    std::cerr << "MF: Hits do not match time ordering: start <= hit <= end."
	      << "MF:\tTrack Start TDC = " << track_start.TDC().val
	      << "MF:\tHit TDC         = " << hit.TDC().val
	      << "MF:\tTrack End TDC   = " << track_end.TDC().val
	      << std::endl;

    assert(false);
  }

  // all this information can be calculated before we start
  // looking at individual hits
  // perhaps a Track class/struct could help here
  double track_distance = distance(track_start, track_end);
  double track_time = track_end.TDC().val - track_start.TDC().val;
  double track_vpro = track_time / track_distance;

  double distance_to_start = distance(track_start, hit);
  double distance_to_end   = distance(track_end,   hit);

  double time_to_start = hit.TDC().val       - track_start.TDC().val;
  double time_to_end   = track_end.TDC().val - hit.TDC().val;

  if (distance_to_start > distance_to_end)
    return hit_is_in_time(time_to_start, distance_to_start * track_vpro);
  else
    return hit_is_in_time(time_to_end,   distance_to_end   * track_vpro);

  // is this last return statement even necessary?
  return false;
}



bool mono::SlowMonopoleTrigger::hit_is_in_time
(double hit_time, double time_to_hit) const
{
  if (hit_time > time_to_hit - track_time_deviation_ &&
      hit_time < time_to_hit + track_time_deviation_)
    return true;

  return false;
}



double mono::SlowMonopoleTrigger::distance
(novaddt::DAQHit const& hit1,
 novaddt::DAQHit const& hit2) const
{
  double dp = double(hit1.Plane().val) - double(hit2.Plane().val);
  double dc = double(hit1.Cell().val)  - double(hit2.Cell().val);

  // working with the square instead of the distance itself might help
  return sqrt(dp * dp * aspect_ratio_ + dc * dc);
}



novaddt::HitList::const_iterator mono::SlowMonopoleTrigger::find
(novaddt::HitList const& hits, novaddt::TDC tdc) const
{
  return std::lower_bound(hits.begin(), hits.end(), tdc,
			  novaddt::CompareDAQHit<novaddt::TDC>());
}



void mono::SlowMonopoleTrigger::print
(novaddt::DAQHit const& h, std::string prefix) const
{
  std::string view;
  if (h.View().val == daqchannelmap::DetView_TYPE::X_VIEW)
    view = "xz";
  else if (h.View().val == daqchannelmap::DetView_TYPE::Y_VIEW)
    view = "yz";
  else
    view = "bad";

  boost::format form("%3s %4u %4u %5u %20u");
  std::cout << prefix << "\t" << 
    form % view % h.Plane().val % h.Cell().val % 
    h.ADC().val % (h.TDC().val - 9223640785000000)
	    << std::endl;
}



void mono::SlowMonopoleTrigger::print
(std::vector<novaddt::TriggerDecision> const& decisions, 
 std::string prefix) const
{
  boost::format form("%20u %6u %6u");

  for (auto const& d : decisions)
  {
    auto t0 = d.start() - 9223640785000000;
    std::cout << prefix << "\t" << 
      form % t0 % (t0 + d.duration()) % d.duration()
	      << std::endl;
  }
}



/******************/
/* DDT Extensions */
/******************/

bool novaddt::operator== 
(novaddt::TriggerDecision const& lhs, novaddt::TriggerDecision const& rhs)
{
  if (lhs.start() != rhs.start())
  {
    std::cout << "MF: The trigger decision start times do not agree!"
	      << "\nMF:  lhs = " << lhs.start()
	      << "\nMF:  rhs = " << rhs.start()
	      << "\nMF:  delta = " << lhs.start() - rhs.start()
	      << std::endl;

    return false;
  }

  if (lhs.duration() != rhs.duration())
  {
    std::cout << "MF: The trigger decision duration times do not agree!"
	      << "\nMF:  lhs = " << lhs.duration()
	      << "\nMF:  rhs = " << rhs.duration()
	      << "\nMF:  delta = " << lhs.duration() - rhs.duration()
	      << std::endl;
    
    return false;
  }
  
  return true;
}



bool novaddt::is_subset_of
(novaddt::TriggerDecision const& subset, novaddt::TriggerDecision const& set)
{
  if (subset.start() < set.start())
  {
    std::cout << "MF: The trigger decision is not a subset!"
	      << "\nMF:  subset start = " << subset.start()
	      << "\nMF:     set start = " << set.start()
	      << std::endl;

    return false;
  }

  auto subset_end = subset.start() + subset.duration();
  auto set_end = set.start() + set.duration();

  if (subset_end > set_end)
  {
    std::cout << "MF: The trigger decision is not a subset!"
	      << "\nMF:  subset end = " << subset_end
	      << "\nMF:     set end = " << set_end
	      << std::endl;

    return false;
  }

  return true;
}



/***************************/
/* Original Implementation */
/***************************/

bool mono::SlowMonopoleTrigger::zukai_algorithm()
{
  bool result = false;

  //  std::cout<<"Total input hits: "<< hits->size()<<std::endl;

  novaddt::HitList hl[2];
  //Seperate the hits by view, preserve the time order:
  for (unsigned i = 0; i!= hits_.size(); ++i ) {
    unsigned v = 0;
    if ( ((hits_)[i]).View().val==daqchannelmap::Y_VIEW ) v = 1;
    hl[v].emplace_back((hits_)[i]);
  }

  std::cout << "ZW: Total Hits: " << hits_.size()
	    << "; xz hits: " << hl[0].size()
	    << "; yz hits: " << hl[1].size()
	    << std::endl;


  if (hl[0].size()<_minHits+2 || hl[1].size()<_minHits+2) {
    // ofs<<"0"<<std::endl;
    return false;
  }

  //Only record X hits
  novaddt::HitList c_PS;     //They will contain all Valid non surface hits on each view
  novaddt::HitList s_PS;     //They will contain all Valid surface and mate hits on each view

  //View check
  //Goal: doing view check for each X hit, and preserve the time order
  for (unsigned i = 0; i != hl[0].size(); ++i) 
  {
    novaddt::DAQHit const& hx = hl[0][i];
    novaddt::TDC::value_type Tx = hx.TDC().val;
    unsigned p1= hx.Plane().val;
    //Do index locator calculation to locate the indices range:
    unsigned start_id = IndexLocator(hl[1], Tx - _deltaT);
    unsigned stop_id  = IndexLocator(hl[1], Tx + _deltaT); 
    bool met = false;
    //Look for your mate, once found, push back and break:
    for (unsigned j = start_id; j < stop_id; ++j) 
    {
      novaddt::DAQHit const& hy = hl[1][j];
      unsigned p2 = hy.Plane().val;
      novaddt::TDC::value_type Ty = hy.TDC().val;
 
      if (ViewCheck(p1,Tx,p2,Ty)) 
      {
	if (SurfAssign(hx) || SurfAssign(hy)) 
	{
	  s_PS.emplace_back(hx);
	  break;
	} else if (!met) {
	  c_PS.emplace_back(hx);
	}
	met = true;
      }
    }
  }

  std::cout<<"ZW: Surface X hits: "<<s_PS.size()<<";  Contained X Hits: "<<c_PS.size()<<std::endl;
  
  unsigned x_con = c_PS.size();
  if (x_con < _minHits+2 ) {
    // ofs<<"0"<<std::endl;
    return false;
  }

  if (s_PS.size()<2){
    //  ofs<<"0"<<std::endl;
    return false;
  }
  //Making possible combinations of entry & exit hits:
  //Hints: you usually get more y hits than x hits, so look for candidates in X view 
  for (unsigned i=0; i<s_PS.size()-1 && !result; ++i) {

    //    std::cout<<"i: "<<i<<std::endl;
    novaddt::DAQHit const& entry = s_PS[i]; 
    novaddt::TDC::value_type  Ti = entry.TDC().val;
    unsigned pi = entry.Plane().val;
    unsigned ci = entry.Cell().val;
    
    unsigned ini = IndexLocator(c_PS, Ti)+1;
    if (ini+_minHits+1 > x_con ) break;  //Don't go to the next hit, you won't get enough hits
    if (Ti > (c_PS[x_con-_minHits-1]).TDC().val ) break; 

    //Notice: the entry hit and the exit hit may not be in the same view:
    //Notice: exit hit should be looping from back to prevent double triggering!
    //According to the chronological order of hits, once j is triggered, 
    //you don't have to go down further, later triggers will be included in the time window!
    for (unsigned j=s_PS.size()-1; j>i && !result; --j) {

      if (ini+_minHits+1 > j ) break;

      novaddt::DAQHit const&  exit = s_PS[j]; 
      novaddt::TDC::value_type Tf  = exit.TDC().val;
      unsigned pf = exit.Plane().val;
      unsigned cf = exit.Cell().val;
      
      if (Tf < c_PS[ini+1+_minHits].TDC().val ) {
	//	std::cout<<"Break due to lack of hits in beetween time window!"<<std::endl;
	break;
      }
      if (Tf < Ti+ unsigned(_vproMin*_DistMin)) {
	//	std::cout<<"Break due to end and start time too close!"<<std::endl;
	break;
      }
      if (Tf > c_PS[x_con-1].TDC().val+_sigmaT ) {
	//	std::cout<<"continue due to Tf too large!"<<std::endl;
	continue;
      }
      //Set the plane range:
      unsigned p_max = std::max(pi,pf);
      unsigned p_min = std::min(pi,pf);
      //Set the cell range:
      unsigned c_max = std::max(ci,cf);
      unsigned c_min = std::min(ci,cf);

      //If you are not traveling through enough planes, you also get cut:
      if (p_max < p_min+2*_deltP) continue;


      float Distance = Dist_(ci,cf,pi,pf);
      float vpro = (Tf-Ti)/Distance; // Always positive since hits are sorted
      //     std::cout<<"vpro:"<<vpro<<std::endl;

      if (Distance < _DistMin) continue;

      if (Tf > Ti + unsigned(_vproMax*Distance)+_sigmaT ) {
	//	std::cout<<"vpro too large!"<<std::endl;
	continue;
      }

      if (Tf < Ti + unsigned(_vproMin*Distance) ){
	//	std::cout<<"vpro too small!"<<std::endl;
	continue;
      }
      unsigned fin =  IndexLocator(c_PS, Tf);
      
      if (fin<ini+Distance*_hf_min){
	//	std::cout<<"fin: "<<fin<<"  ini: "<<ini<<"   Distance: "<<Distance<<std::endl;
	continue;
      }
      //Hmm, here comes the meat of this piece of code:
      //Loop through the non-boundary hits in the view, in the given range:
      unsigned h_counts = 0;
 
      float disCut = Distance*_sigmaL;
      unsigned h_cut = Distance*_hf_min;	

      for (unsigned k=ini; k!=fin;++k) {
	novaddt::DAQHit const& midP = c_PS[k];
	unsigned pk = midP.Plane().val;
	unsigned ck = midP.Cell().val; 

	if (pk>p_max) continue;
	if (pk<p_min) continue;
	if (ck>c_max) continue;
	if (ck<c_min) continue;

	novaddt::TDC::value_type Tk  = midP.TDC().val;

	//	std::cout<<"mid T: "<<midP.TDC().val<<std::endl;	
	float d1 = Dist_(ci,ck,pi,pk);
	float d2 = Dist_(cf,ck,pf,pk);

	if (d1+d2 > Distance+disCut) continue;
	float t1 = Tk-Ti;   // From start to mid
	float t2 = Tf-Tk;   // From mid to end

	if (d1>d2) {
	  if (t1 < d1*vpro + _sigmaT && t1 + _sigmaT > d1*vpro) 
	    ++h_counts;
	}

	else {
	  if (t2 < d2*vpro + _sigmaT && t2 + _sigmaT > d2*vpro )
	    ++h_counts;
	}	
		  
      }
     
      
      if (h_counts>h_cut && h_counts>_minHits-2) {
	/*
	std::cout<<"counts on line: "<<h_counts
		 <<"  Distance: "<<Distance<<"  vpro: "<<vpro
		 <<"  hits number density: "<<h_counts/Distance<<std::endl;
	
	std::cout<<"c0: "<<ci<<", p0: "<<pi<<std::endl;
	std::cout<<"c1: "<<cf<<", p1: "<<pf<<std::endl;
	std::cout<<std::endl;
	*/
	result = true;
	zukai_trigger_decisions_.
	  emplace_back(Ti-2*_deltaT, Tf - Ti+4*_deltaT, 
		       daqdataformats::trigID::TRIG_ID_DATA_SLOWMONO, 1);

	break;
      }
    }
  }	
  /*
  if (result) ofs<<"1"<<std::endl;
  else ofs<<"0"<<std::endl;
  ofs.close();
  */

  return result;
}



bool mono::SlowMonopoleTrigger::SurfAssign(novaddt::DAQHit const & hit) 
{
  //Check if this is close to surface
  if      ( hit.Cell().val  > _Cmax - _deltC   ) return true;
  else if ( hit.Cell().val  < _deltC           ) return true;
  else if ( hit.Plane().val < _deltP           ) return true;
  else if ( hit.Plane().val > _Pmax - _deltP   ) return true;

  return false;
}



//Note: Ts < Tm (Looking forwards). Time check is already done in the main loop.
bool mono::SlowMonopoleTrigger::ViewCheck(unsigned p1, novaddt::TDC::value_type T1,
				       unsigned p2, novaddt::TDC::value_type T2)
{
  if (T1>T2+_deltaT) return false;
  if (T2>T1+_deltaT) return false;
  if (p1 > p2 ) {
    if (p1-p2 < _deltP) return true;
  }
  else if (p2-p1 < _deltP) return true;

  return false; 
}



//It should return the index before T
unsigned mono::SlowMonopoleTrigger::IndexLocator
(novaddt::HitList & hl,
 novaddt::TDC::value_type T)
{
  unsigned i = 0;
  unsigned f = hl.size() - 1;

  if (T + _deltaT < hl[i].TDC().val) 
    return i; 

  if (T > hl[f].TDC().val + _deltaT) 
    return f;

  while (f > i + 1)
  {
    unsigned m = (f + i) / 2;
    if (T >= hl[m].TDC().val) 
      i = m;
    else 
      f = m; 
  }

  return i;
}



float mono::SlowMonopoleTrigger::Dist_(float c1, float c2,
				    float p1, float p2)
{ 
  return sqrt((p1-p2)*(p1-p2)*2.82072+(c1-c2)*(c1-c2));
}
