///////////////////////////////////////////////////////////////////////
// \file    MRCCSystematicsAna_module.cc
//
// \brief   MRCC Systematics and Analysis module. Fills TTrees with a lot 
//          of information one may need about original and mrcc events. It 
//          contains Pid values and variables, fiducial cuts, neutrino
//          energies, some truth information etc. It uses NuReweight to
//          evaluate event weights when some GENIE parameters are changed
//          by +/- 1 sigma, for systematics studies. For the Mrcc tree, it 
//          also stores information related to the original event, like 
//          ReMId PID value original fiducial cut pass/fail info etc.
//
// \author  Kanika Sachdev (ksachdev@physics.umn.edu)
////////////////////////////////////////////////////////////////////////


// C/C++ & root includes
#include <iostream>
#include <vector>
#include "TTree.h"
#include "TH1D.h"
#include <map>


// ART includes      
#include "art/Framework/Core/ModuleMacros.h" 
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Principal/Run.h"                                                                       
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"

//NOvA includes
#include "MuonRemove/art/MRCCParent.h"
#include "NuReweight/art/NuReweight.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/GTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "ReMId/ReMId.h"
#include "NumuEnergy/NumuE.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Vertex.h"
#include "MCCheater/BackTracker.h"
#include "LEM/PIDDetails.h"
#include "RecoJMShower/EID.h"
#include "RecVarPID/RVP.h"
#include "NumuSandbox/NumuSandFxs.h"
#include "SummaryData/POTSum.h"

namespace murem 
{                                                                                                               
  
  class MRCCSystematics : public art::EDAnalyzer 
  {
  public:
    explicit MRCCSystematics(fhicl::ParameterSet const& pset);
    virtual ~MRCCSystematics();
  
    void analyze(art::Event const& evt);
    void beginRun( const art::Run& run);
    void beginJob();
    void reconfigure(const fhicl::ParameterSet& pset);


  private:
    
    numusand::NumuSandFxs*     fNumuSandFxs;
    
    void SetDefaults();


    void CalculateWeights( const simb::MCTruth& truth, 
			   const simb::GTruth&  gTruth,
			   int sliceIndex);

    void StoreWeights();

    void FindWeights( int origSliceIndex );

    void ClearMaps();

    virtual void endSubRun( const art::SubRun& sr);
    
    void endJob();

    TTree* fSys;
    TTree* fSysMrcc;
    TH1D*  hPOT;

    double fTotalPOT;

    float fgMaRes[2];
    float fgMvRes[2];
    float fgDeltaPi[2];
    float fgNoDeltaPi[2];
    float fgMaCOHpi[2];
    float fgROCOHpi[2];
    float fgMaNCEL[2];
    float fgEtaNCEL[2];
    float fgMaCCQE[2];
    float fgDelta2Pi[2];
    float fgNoDelta2Pi[2];
    float fgAhtBY[2];
    float fgBhtBY[2];
    float fgCV1uBY[2];
    float fgCV2uBY[2];
    float fgAGKY_xF1pi[2];
    float fgAGKY_pT1pi[2];
    float fgMFP_pi[2];
    float fgMFP_N[2];
    float fgFrCEx_pi[2];
    float fgFrElas_pi[2];
    float fgFrInel_pi[2];
    float fgFrAbs_pi[2];
    float fgFrPiProd_pi[2];
    float fgFrCEx_N[2];
    float fgFrElas_N[2];
    float fgFrInel_N[2];
    float fgFrAbs_N[2];
    float fgFrPiProd_N[2];
    float fgCCQEPauliSupViaKF[2];
    float fgCCQEMomDistroFGtoSF[2];
    float fgBR1gamma[2];
    float fgBR1eta[2];
    float fgTheta_Delta2Npi[2];

    // The following maps are to map each leading neutrino to its GENIE weight
    // for that given parameter change. The key is the original slice index.
    // These maps are built so that we don't have to recalculate weights for
    // mrcc slices: since the true neutrinos are still going to be the same,
    // the weights will be the same too. The vector will contain only two 
    // components- for +/- sigma changes in the GENIE parameter.

    std::map< int, float > mgMaRes[2];
    std::map< int, float > mgMvRes[2];
    std::map< int, float > mgDeltaPi[2];
    std::map< int, float > mgNoDeltaPi[2];
    std::map< int, float > mgMaCOHpi[2];
    std::map< int, float > mgROCOHpi[2];
    std::map< int, float > mgMaNCEL[2];
    std::map< int, float > mgEtaNCEL[2];
    std::map< int, float > mgMaCCQE[2];
    std::map< int, float > mgDelta2Pi[2];
    std::map< int, float > mgNoDelta2Pi[2];
    std::map< int, float > mgAhtBY[2];
    std::map< int, float > mgBhtBY[2];
    std::map< int, float > mgCV1uBY[2];
    std::map< int, float > mgCV2uBY[2];
    std::map< int, float > mgAGKY_xF1pi[2];
    std::map< int, float > mgAGKY_pT1pi[2];
    std::map< int, float > mgMFP_pi[2];
    std::map< int, float > mgMFP_N[2];
    std::map< int, float > mgFrCEx_pi[2];
    std::map< int, float > mgFrElas_pi[2];
    std::map< int, float > mgFrInel_pi[2];
    std::map< int, float > mgFrAbs_pi[2];
    std::map< int, float > mgFrPiProd_pi[2];
    std::map< int, float > mgFrCEx_N[2];
    std::map< int, float > mgFrElas_N[2];
    std::map< int, float > mgFrInel_N[2];
    std::map< int, float > mgFrAbs_N[2];
    std::map< int, float > mgFrPiProd_N[2];
    std::map< int, float > mgCCQEPauliSupViaKF[2];
    std::map< int, float > mgCCQEMomDistroFGtoSF[2];
    std::map< int, float > mgBR1gamma[2];
    std::map< int, float > mgBR1eta[2];
    std::map< int, float > mgTheta_Delta2Npi[2];
 
    float fNuTrueE;
    int   fEvt;
    int   fPdg;
    int   fIsCc;
    float fQSq;
    float fW;
    float fX;
    float fHadY;

    float fNuSusanE;
    float fNuJmE;
    float fNuTotGeV;
    float fJmid;
    float fRemid;
    float fLemid;

    float fNuSusanEMrcc;
    float fNuJmEMrcc;
    float fNuTotGeVMrcc;
    float fJmidMrcc;
    float fRemidMrcc;
    float fLemidMrcc;
    float fMuon3D;

    TVector3 fVert;
    float fTrkStart;
    float fTrkEnd; 
    TVector3 fVertMrcc;

    float fRvp;
    float fRvpNCells;
    float fRvpRecoE;
    float fRvpLongTr;
    float fRvpLongTrFrac;
    float fRvpMipFrac;
    float fRvpMipHits;
    float fRvpEPl20Frac;
    float fRvpEFrac2PlWin;
    float fRvpEFrac6PlWin;
    float fRvpEFrac2SigRd;
    float fRvpShowers3D;
    float fRvpProngEBal3D;
    float fRvpPdg;

    float fRvpMrcc;
    float fRvpNCellsMrcc;
    float fRvpRecoEMrcc;
    float fRvpLongTrMrcc;
    float fRvpLongTrFracMrcc;
    float fRvpMipFracMrcc;
    float fRvpMipHitsMrcc;
    float fRvpEPl20FracMrcc;
    float fRvpEFrac2PlWinMrcc;
    float fRvpEFrac6PlWinMrcc;
    float fRvpEFrac2SigRdMrcc;
    float fRvpShowers3DMrcc;
    float fRvpProngEBal3DMrcc;
    float fRvpPdgMrcc;

    TVector3  fShowerStopMrcc;
    TVector3  fShowerStartMrcc;
    float  fShowerEMrcc;        
    float  fegLMrcc;            
    float  fegTMrcc;            
    float  femuLMrcc;            
    float  femuTMrcc;            
    float  fepi0LMrcc;          
    float  fepi0TMrcc;          
    float  fepiLMrcc;           
    float  fepiTMrcc;           
    float  fenLMrcc;
    float  fenTMrcc;
    float  fepLMrcc;
    float  fepTMrcc;
    float  feLMrcc;
    float  feTMrcc;
    float  fgapMrcc;
    float  fERatioMrcc;
    float  fpi0massMrcc;

    TVector3 fShowerStop;
    TVector3 fShowerStart;
    float  fShowerE;        
    float  fegL;            
    float  fegT;            
    float  femuL;            
    float  femuT;            
    float  fepi0L;          
    float  fepi0T;          
    float  fepiL;           
    float  fepiT;           
    float  fenL;
    float  fenT;
    float  fepL;
    float  fepT;
    float  feL;
    float  feT;
    float  fgap;
    float  fERatio;
    float  fpi0mass;

    float fOrigJmid;
    float fOrigLemid;
    float fOrigRemid;
    float fOrigMuon3D;
    float fOrigTrkStart;
    float fOrigTrkEnd; 
    TVector3 fOrigVert;

    float fSlicePurity;
    float fSliceEfficiency;

    bool  fDoSystematicsCalc;
  
    std::vector<float> fDaughterPdg;
    std::vector<float> fDaughterPt;
    std::vector<float> fDaughterE;

    std::vector<float> fDaughterPdgMrcc;
    std::vector<float> fDaughterPtMrcc;
    std::vector<float> fDaughterEMrcc;
  
    // Cross section weight calculator                                                                         
                          
    rwgt::NuReweight* fGMaRes[2];
    rwgt::NuReweight* fGMvRes[2];
    rwgt::NuReweight* fGDeltaPi[2];
    rwgt::NuReweight* fGNoDeltaPi[2];
    rwgt::NuReweight* fGMaCOHpi[2];
    rwgt::NuReweight* fGROCOHpi[2];
    rwgt::NuReweight* fGMaNCEL[2];
    rwgt::NuReweight* fGEtaNCEL[2];
    rwgt::NuReweight* fGMaCCQE[2];
    rwgt::NuReweight* fGDelta2Pi[2];
    rwgt::NuReweight* fGNoDelta2Pi[2];
    rwgt::NuReweight* fGAhtBY[2];
    rwgt::NuReweight* fGBhtBY[2];
    rwgt::NuReweight* fGCV1uBY[2];
    rwgt::NuReweight* fGCV2uBY[2];
    rwgt::NuReweight* fGAGKY_xF1pi[2];
    rwgt::NuReweight* fGAGKY_pT1pi[2];
    rwgt::NuReweight* fGMFP_pi[2];
    rwgt::NuReweight* fGMFP_N [2];
    rwgt::NuReweight* fGFrCEx_pi[2];
    rwgt::NuReweight* fGFrElas_pi[2];
    rwgt::NuReweight* fGFrInel_pi[2];
    rwgt::NuReweight* fGFrAbs_pi[2];
    rwgt::NuReweight* fGFrPiProd_pi[2];
    rwgt::NuReweight* fGFrCEx_N[2];
    rwgt::NuReweight* fGFrElas_N[2];
    rwgt::NuReweight* fGFrInel_N[2];
    rwgt::NuReweight* fGFrAbs_N[2];
    rwgt::NuReweight* fGFrPiProd_N[2];
    rwgt::NuReweight* fGCCQEPauliSupViaKF[2];
    rwgt::NuReweight* fGCCQEMomDistroFGtoSF[2];
    rwgt::NuReweight* fGBR1gamma[2];
    rwgt::NuReweight* fGBR1eta[2];
    rwgt::NuReweight* fGTheta_Delta2Npi[2];

    std::string fMCTruthLabel;
    std::string fNumiBeamLabel;

    std::string fVertexLabel;
    std::string fClusterLabel;
    std::string fTrackLabel;
    std::string fEnergyLabel;
    std::string fHitLabel;
    std::string fNumuIdLabel;
    std::string fJmIdLabel;
    std::string fLemIdLabel;
    std::string fRvpLabel;

    std::string fMRParentLabel;
    std::string fMRVertexLabel;
    std::string fMRClusterLabel;
    std::string fMRTrackLabel;
    std::string fMREnergyLabel;
    std::string fMRHitLabel;
    std::string fMRNumuIdLabel;
    std::string fMRJmIdLabel;
    std::string fMRLemIdLabel;
    std::string fMRRvpLabel;
  };

}

murem::MRCCSystematics::MRCCSystematics(fhicl::ParameterSet const& pset)
  : EDAnalyzer(pset)
{
  this->reconfigure(pset);
  mf::LogInfo("MRCCSystematics") <<"MRCCSystematics::MRCCSystematics()\n";
}

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

murem::MRCCSystematics::~MRCCSystematics()
{
}

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::reconfigure(const fhicl::ParameterSet& pset)
{

  fMCTruthLabel       = pset.get< std::string      >("MCTruthLabel");  
  fNumiBeamLabel      = pset.get< std::string      >("NumiBeamLabel");  

  fMRParentLabel      = pset.get< std::string      >("MRParentLabel");   
  fVertexLabel        = pset.get< std::string      >("VertexLabel");   
  fClusterLabel       = pset.get< std::string      >("ClusterLabel");   
  fTrackLabel         = pset.get< std::string      >("TrackLabel");   
  fEnergyLabel        = pset.get< std::string      >("EnergyLabel");   
  fHitLabel           = pset.get< std::string      >("HitLabel");   
  fNumuIdLabel        = pset.get< std::string      >("NumuIdLabel");   
  fJmIdLabel          = pset.get< std::string      >("JmIdLabel");   
  fLemIdLabel         = pset.get< std::string      >("LemIdLabel");   
  fRvpLabel           = pset.get< std::string      >("RvpLabel");   

  fMRVertexLabel      = pset.get< std::string      >("MRVertexLabel");   
  fMRClusterLabel     = pset.get< std::string      >("MRClusterLabel");   
  fMRTrackLabel       = pset.get< std::string      >("MRTrackLabel");   
  fMREnergyLabel      = pset.get< std::string      >("MREnergyLabel");   
  fMRHitLabel         = pset.get< std::string      >("MRHitLabel");   
  fMRNumuIdLabel      = pset.get< std::string      >("MRNumuIdLabel");   
  fMRJmIdLabel        = pset.get< std::string      >("MRJmIdLabel");   
  fMRLemIdLabel       = pset.get< std::string      >("MRLemIdLabel");   
  fMRRvpLabel         = pset.get< std::string      >("MRRvpLabel");   

  fDoSystematicsCalc  = pset.get< bool             >("DoSystematicsCalc");   
}

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////


void murem::MRCCSystematics::beginRun(const art::Run& run)
{
  fNumuSandFxs = new numusand::NumuSandFxs();
}

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::beginJob()
{
  
  art::ServiceHandle<art::TFileService> tfs;

  hPOT = tfs->make<TH1D>("TotalPOT", "TotalPOT;; POT", 1, 0, 1);

  fSys = tfs->make<TTree>("fSys","fSys");

  fSys->Branch("Evt",                     &fEvt);
  fSys->Branch("Pdg",                     &fPdg);
  fSys->Branch("IsCc",                    &fIsCc);
  fSys->Branch("QSq",                     &fQSq);
  fSys->Branch("W",                       &fW);
  fSys->Branch("X",                       &fX);
  fSys->Branch("HadY",                    &fHadY);
  fSys->Branch("NuTrueE",                 &fNuTrueE);
  fSys->Branch("DaughterPdg",             &fDaughterPdg);
  fSys->Branch("DaughterPt",              &fDaughterPt);
  fSys->Branch("DaughterE",               &fDaughterE);

  fSys->Branch("SlicePur",                &fSlicePurity);
  fSys->Branch("SliceEff",                &fSliceEfficiency);
  
  fSys->Branch("Vert",                    &fVert);
  fSys->Branch("TrkStart",                &fTrkStart);
  fSys->Branch("TrkEnd",                  &fTrkEnd);

  fSys->Branch("NuSusanE",                &fNuSusanE);
  fSys->Branch("NuTotGeV",                &fNuTotGeV);
  fSys->Branch("NuJmE",                   &fNuJmE);
  fSys->Branch("Remid",                   &fRemid);
  fSys->Branch("Muon3D",                  &fMuon3D);
  fSys->Branch("Jmid",                    &fJmid);
  fSys->Branch("Lemid",                   &fLemid);

  fSys->Branch("ShowerE",                 &fShowerE);        
  fSys->Branch("ShowerStop",              &fShowerStop);                                                                  
  fSys->Branch("ShowerStart",             &fShowerStart);  
  fSys->Branch("egL",                     &fegL);            
  fSys->Branch("egT",                     &fegT);            
  fSys->Branch("emL",                     &femuL);            
  fSys->Branch("emT",                     &femuT);                
  fSys->Branch("epi0L",                   &fepi0L);          
  fSys->Branch("epi0T",                   &fepi0T);          
  fSys->Branch("epiL",                    &fepiL);           
  fSys->Branch("epiT",                    &fepiT);           
  fSys->Branch("enL",                     &fenL);
  fSys->Branch("enT",                     &fenT);
  fSys->Branch("eL",                      &feL);
  fSys->Branch("eT",                      &feT);
  fSys->Branch("gap",                     &fgap);
  fSys->Branch("ERatio",                  &fERatio);
  fSys->Branch("pi0mass",                 &fpi0mass);

  fSys->Branch( "Rvp",                    &fRvp);
  fSys->Branch( "RvpNCells",              &fRvpNCells);
  fSys->Branch( "RvpRecoE",               &fRvpRecoE);
  fSys->Branch( "RvpLongTr",              &fRvpLongTr);
  fSys->Branch( "RvpLongTrFrac",          &fRvpLongTrFrac);
  fSys->Branch( "RvpMipFrac",             &fRvpMipFrac);
  fSys->Branch( "RvpMipHits",             &fRvpMipHits);
  fSys->Branch( "RvpEPl20Frac",           &fRvpEPl20Frac);
  fSys->Branch( "RvpEFrac2PlWin",         &fRvpEFrac2PlWin);
  fSys->Branch( "RvpEFrac6PlWin",         &fRvpEFrac6PlWin);
  fSys->Branch( "RvpEFrac2SigRd",         &fRvpEFrac2SigRd);
  fSys->Branch( "RvpShowers3D",            &fRvpShowers3D);
  fSys->Branch( "RvpProngEBal3D",         &fRvpProngEBal3D);
  fSys->Branch( "RvpPdg",                 &fRvpPdg);

  fSysMrcc = tfs->make<TTree>("fSysMrcc","fSysMrcc");

  fSysMrcc->Branch("EvtMrcc",             &fEvt);
  fSysMrcc->Branch("PdgMrcc",             &fPdg);
  fSysMrcc->Branch("IsCcMrcc",            &fIsCc);
  fSysMrcc->Branch("QSqMrcc",             &fQSq);
  fSysMrcc->Branch("WMrcc",               &fW);
  fSysMrcc->Branch("XMrcc",               &fX);
  fSysMrcc->Branch("HadYMrcc",            &fHadY);
  fSysMrcc->Branch("NuTrueEMrcc",         &fNuTrueE);
  fSysMrcc->Branch("DaughterPdgMrcc",     &fDaughterPdgMrcc);
  fSysMrcc->Branch("DaughterPtMrcc",      &fDaughterPtMrcc);
  fSysMrcc->Branch("DaughterEMrcc",       &fDaughterEMrcc);

  fSysMrcc->Branch("VertMrcc",            &fVertMrcc);

  fSysMrcc->Branch("OrigRemid",           &fOrigRemid);
  fSysMrcc->Branch("OrigMuon3D",          &fOrigMuon3D);
  fSysMrcc->Branch("OrigLemid",           &fOrigLemid);
  fSysMrcc->Branch("OrigJmid",            &fOrigJmid);
  fSysMrcc->Branch("OrigVert",            &fOrigVert);
  fSysMrcc->Branch("OrigTrkStart",        &fOrigTrkStart);
  fSysMrcc->Branch("OrigTrkEnd",          &fOrigTrkEnd);

  fSysMrcc->Branch("SlicePur",            &fSlicePurity);
  fSysMrcc->Branch("SliceEff",            &fSliceEfficiency);

  fSysMrcc->Branch("NuSusanEMrcc",        &fNuSusanEMrcc);
  fSysMrcc->Branch("NuTotGeVMrcc",        &fNuTotGeVMrcc);
  fSysMrcc->Branch("NuJmEMrcc",           &fNuJmEMrcc);
  fSysMrcc->Branch("RemidMrcc",           &fRemidMrcc);
  fSysMrcc->Branch("JmidMrcc",            &fJmidMrcc);
  fSysMrcc->Branch("LemidMrcc",           &fLemidMrcc);

  fSysMrcc->Branch("ShowerEMrcc",         &fShowerEMrcc);        
  fSysMrcc->Branch("ShowerStopMrcc",      &fShowerStopMrcc);                                                             
  fSysMrcc->Branch("ShowerStartMrcc",     &fShowerStartMrcc);
  fSysMrcc->Branch("egLMrcc",             &fegLMrcc);            
  fSysMrcc->Branch("egTMrcc",             &fegTMrcc);            
  fSysMrcc->Branch("emLMrcc",             &femuLMrcc);            
  fSysMrcc->Branch("emTMrcc",             &femuTMrcc);            
  fSysMrcc->Branch("epLMrcc",             &fepLMrcc);            
  fSysMrcc->Branch("epTMrcc",             &fepTMrcc);                
  fSysMrcc->Branch("epi0LMrcc",           &fepi0LMrcc);          
  fSysMrcc->Branch("epi0TMrcc",           &fepi0TMrcc);          
  fSysMrcc->Branch("epiLMrcc",            &fepiLMrcc);           
  fSysMrcc->Branch("epiTMrcc",            &fepiTMrcc);           
  fSysMrcc->Branch("enLMrcc",             &fenLMrcc);
  fSysMrcc->Branch("enTMrcc",             &fenTMrcc);
  fSysMrcc->Branch("eLMrcc",              &feLMrcc);
  fSysMrcc->Branch("eTMrcc",              &feTMrcc);
  fSysMrcc->Branch("gapMrcc",             &fgapMrcc);
  fSysMrcc->Branch("ERatioMrcc",          &fERatioMrcc);
  fSysMrcc->Branch("pi0massMrcc",         &fpi0massMrcc);
 
  fSysMrcc->Branch( "RvpMrcc",            &fRvpMrcc);
  fSysMrcc->Branch( "RvpNCellsMrcc",      &fRvpNCellsMrcc);
  fSysMrcc->Branch( "RvpRecoEMrcc",       &fRvpRecoEMrcc);
  fSysMrcc->Branch( "RvpLongTrMrcc",      &fRvpLongTrMrcc);
  fSysMrcc->Branch( "RvpLongTrFracMrcc",  &fRvpLongTrFracMrcc);
  fSysMrcc->Branch( "RvpMipFracMrcc",     &fRvpMipFracMrcc);
  fSysMrcc->Branch( "RvpMipHitsMrcc",     &fRvpMipHitsMrcc);
  fSysMrcc->Branch( "RvpEPl20FracMrcc",   &fRvpEPl20FracMrcc);
  fSysMrcc->Branch( "RvpEFrac2PlWinMrcc", &fRvpEFrac2PlWinMrcc);
  fSysMrcc->Branch( "RvpEFrac6PlWinMrcc", &fRvpEFrac6PlWinMrcc);
  fSysMrcc->Branch( "RvpEFrac2SigRdMrcc", &fRvpEFrac2SigRdMrcc);
  fSysMrcc->Branch( "RvpShowers3DMrcc",    &fRvpShowers3DMrcc);
  fSysMrcc->Branch( "RvpProngEBal3DMrcc", &fRvpProngEBal3DMrcc);
  fSysMrcc->Branch( "RvpPdgMrcc",         &fRvpPdgMrcc);

  
  if(fDoSystematicsCalc){
    fSys->Branch("gMaRes[2]",       fgMaRes);
    fSys->Branch("gMvRes[2]",       fgMvRes); 
    fSys->Branch("gDeltaPi[2]",     fgDeltaPi);  
    fSys->Branch("gNoDeltaPi[2]",   fgNoDeltaPi);
    fSys->Branch("gMaCOHpi[2]",     fgMaCOHpi); 
    fSys->Branch("gROCOHpi[2]",     fgROCOHpi);  
    fSys->Branch("gMaNCEL[2]" ,     fgMaNCEL);
    fSys->Branch("gEtaNCEL[2]" ,    fgEtaNCEL);
    fSys->Branch("gMaCCQE[2]" ,     fgMaCCQE);
    fSys->Branch("gDelta2Pi[2]" ,   fgDelta2Pi);
    fSys->Branch("gNoDelta2Pi[2]" , fgNoDelta2Pi);
    fSys->Branch("gAhtBY[2]" ,      fgAhtBY);
    fSys->Branch("gBhtBY[2]" ,      fgBhtBY);
    fSys->Branch("gCV1uBY[2]" ,     fgCV1uBY);
    fSys->Branch("gCV2uBY[2]" ,     fgCV2uBY);
    fSys->Branch("gAGKY_xF1pi[2]" , fgAGKY_xF1pi);
    fSys->Branch("gAGKY_pT1pi[2]" , fgAGKY_pT1pi);
    fSys->Branch("gMFP_pi[2]" ,     fgMFP_pi);
    fSys->Branch("gMFP_N[2]" ,      fgMFP_N );
    fSys->Branch("gFrCEx_pi[2]" ,   fgFrCEx_pi);
    fSys->Branch("gFrElas_pi[2]" ,  fgFrElas_pi);
    fSys->Branch("gFrInel_pi[2]" ,  fgFrInel_pi);
    fSys->Branch("gFrAbs_pi[2]" ,   fgFrAbs_pi);
    fSys->Branch("gFrPiProd_pi[2]" ,fgFrPiProd_pi);
    fSys->Branch("gFrCEx_N[2]" ,    fgFrCEx_N);
    fSys->Branch("gFrElas_N[2]" ,   fgFrElas_N);
    fSys->Branch("gFrInel_N[2]" ,   fgFrInel_N);
    fSys->Branch("gFrAbs_N[2]" ,    fgFrAbs_N);
    fSys->Branch("gFrPiProd_N[2]" , fgFrPiProd_N);
    fSys->Branch("gBR1gamma[2]" ,   fgBR1gamma);
    fSys->Branch("gBR1eta[2]" ,     fgBR1eta);
    fSys->Branch("gTheta_Delta2Npi[2]" ,   fgTheta_Delta2Npi);
    fSys->Branch("gCCQEPauliSupViaKF[2]" , fgCCQEPauliSupViaKF);
    fSys->Branch("gCCQEMomDistroFGtoSF[2]",fgCCQEMomDistroFGtoSF);

    fSysMrcc->Branch("gMaRes[2]",       fgMaRes);
    fSysMrcc->Branch("gMvRes[2]",       fgMvRes); 
    fSysMrcc->Branch("gDeltaPi[2]",     fgDeltaPi);  
    fSysMrcc->Branch("gNoDeltaPi[2]",   fgNoDeltaPi);
    fSysMrcc->Branch("gMaCOHpi[2]",     fgMaCOHpi); 
    fSysMrcc->Branch("gROCOHpi[2]",     fgROCOHpi);  
    fSysMrcc->Branch("gMaNCEL[2]" ,     fgMaNCEL);
    fSysMrcc->Branch("gEtaNCEL[2]" ,    fgEtaNCEL);
    fSysMrcc->Branch("gMaCCQE[2]" ,     fgMaCCQE);
    fSysMrcc->Branch("gDelta2Pi[2]" ,   fgDelta2Pi);
    fSysMrcc->Branch("gNoDelta2Pi[2]" , fgNoDelta2Pi);
    fSysMrcc->Branch("gAhtBY[2]" ,      fgAhtBY);
    fSysMrcc->Branch("gBhtBY[2]" ,      fgBhtBY);
    fSysMrcc->Branch("gCV1uBY[2]" ,     fgCV1uBY);
    fSysMrcc->Branch("gCV2uBY[2]" ,     fgCV2uBY);
    fSysMrcc->Branch("gAGKY_xF1pi[2]" , fgAGKY_xF1pi);
    fSysMrcc->Branch("gAGKY_pT1pi[2]" , fgAGKY_pT1pi);
    fSysMrcc->Branch("gMFP_pi[2]" ,     fgMFP_pi);
    fSysMrcc->Branch("gMFP_N[2]" ,      fgMFP_N );
    fSysMrcc->Branch("gFrCEx_pi[2]" ,   fgFrCEx_pi);
    fSysMrcc->Branch("gFrElas_pi[2]" ,  fgFrElas_pi);
    fSysMrcc->Branch("gFrInel_pi[2]" ,  fgFrInel_pi);
    fSysMrcc->Branch("gFrAbs_pi[2]" ,   fgFrAbs_pi);
    fSysMrcc->Branch("gFrPiProd_pi[2]" ,fgFrPiProd_pi);
    fSysMrcc->Branch("gFrCEx_N[2]" ,    fgFrCEx_N);
    fSysMrcc->Branch("gFrElas_N[2]" ,   fgFrElas_N);
    fSysMrcc->Branch("gFrInel_N[2]" ,   fgFrInel_N);
    fSysMrcc->Branch("gFrAbs_N[2]" ,    fgFrAbs_N);
    fSysMrcc->Branch("gFrPiProd_N[2]" , fgFrPiProd_N);
    fSysMrcc->Branch("gBR1gamma[2]" ,   fgBR1gamma);
    fSysMrcc->Branch("gBR1eta[2]" ,     fgBR1eta);
    fSysMrcc->Branch("gTheta_Delta2Npi[2]" ,   fgTheta_Delta2Npi);
    fSysMrcc->Branch("gCCQEPauliSupViaKF[2]" , fgCCQEPauliSupViaKF);
    fSysMrcc->Branch("gCCQEMomDistroFGtoSF[2]",fgCCQEMomDistroFGtoSF);
  

    // The following creates two instances of the NuReweight object.
    // The objects alter the cross section parameters by 1 and -1 sigma.

    for(int i = 0; i < 2; i++) {
      double sigma = 0;
      if ( i == 0 )
	sigma = -1;
      else
	sigma = 1;
      

      fGMaRes[i]       = new rwgt::NuReweight();
      fGMvRes[i]       = new rwgt::NuReweight();
      fGDeltaPi[i]     = new rwgt::NuReweight();
      fGNoDeltaPi[i]   = new rwgt::NuReweight();
      fGMaCOHpi[i]     = new rwgt::NuReweight();
      fGROCOHpi[i]     = new rwgt::NuReweight();
      fGMaNCEL[i]      = new rwgt::NuReweight();
      fGEtaNCEL[i]     = new rwgt::NuReweight();
      fGMaCCQE[i]      = new rwgt::NuReweight();
      fGDelta2Pi[i]    = new rwgt::NuReweight();
      fGNoDelta2Pi[i]  = new rwgt::NuReweight();
      fGAhtBY[i]       = new rwgt::NuReweight();
      fGBhtBY[i]       = new rwgt::NuReweight();
      fGCV1uBY[i]      = new rwgt::NuReweight();
      fGCV2uBY[i]      = new rwgt::NuReweight();
      fGAGKY_xF1pi[i]  = new rwgt::NuReweight();
      fGAGKY_pT1pi[i]  = new rwgt::NuReweight();
      fGMFP_pi[i]      = new rwgt::NuReweight();
      fGMFP_N [i]      = new rwgt::NuReweight();
      fGFrCEx_pi[i]    = new rwgt::NuReweight();
      fGFrElas_pi[i]   = new rwgt::NuReweight();
      fGFrInel_pi[i]   = new rwgt::NuReweight();
      fGFrAbs_pi[i]    = new rwgt::NuReweight();
      fGFrPiProd_pi[i] = new rwgt::NuReweight();
      fGFrCEx_N[i]     = new rwgt::NuReweight();
      fGFrElas_N[i]    = new rwgt::NuReweight();
      fGFrInel_N[i]    = new rwgt::NuReweight();
      fGFrAbs_N[i]     = new rwgt::NuReweight();
      fGFrPiProd_N[i]  = new rwgt::NuReweight();
      fGBR1gamma[i]    = new rwgt::NuReweight();
      fGBR1eta[i]      = new rwgt::NuReweight();
      fGTheta_Delta2Npi[i]    = new rwgt::NuReweight();
      fGCCQEPauliSupViaKF[i]  = new rwgt::NuReweight();
      fGCCQEMomDistroFGtoSF[i]= new rwgt::NuReweight();
      
      // The following two paramteres alter the Delta production form factor
      // This results in different pi+/- production cross-section in Neutral Currents.
      
      fGMaRes[i]->AddReweightValue(rwgt::fReweightMaNCRES, sigma);
      fGMvRes[i]->AddReweightValue(rwgt::fReweightMvNCRES, sigma);
      fGMaRes[i]->AddReweightValue(rwgt::fReweightMaCCRES, sigma);
      fGMvRes[i]->AddReweightValue(rwgt::fReweightMvCCRES, sigma);

      // In the resonance/DIS transition region these parameter directly
      // reweight events with a single pion in 
      // the final state (charged as well as neutral) by changing the delta form factor 
      // The following 4 parameter should all be changed by the same amount.
      
      fGDeltaPi[i]->AddReweightValue(rwgt::fReweightRvpCC1pi, sigma);
      fGDeltaPi[i]->AddReweightValue(rwgt::fReweightRvpNC1pi, sigma);
      fGDeltaPi[i]->AddReweightValue(rwgt::fReweightRvbarnCC1pi, sigma);
      fGDeltaPi[i]->AddReweightValue(rwgt::fReweightRvbarnNC1pi, sigma);
      
      // The following 4 parameter should all be changed by the same amount. 
      // They produce a non-resonance background to the nucleon-pion final state.
      
      fGNoDeltaPi[i]->AddReweightValue(rwgt::fReweightRvnCC1pi, sigma);
      fGNoDeltaPi[i]->AddReweightValue(rwgt::fReweightRvnNC1pi, sigma);
      fGNoDeltaPi[i]->AddReweightValue(rwgt::fReweightRvbarpCC1pi, sigma);
      fGNoDeltaPi[i]->AddReweightValue(rwgt::fReweightRvbarpNC1pi, sigma);
      

      // The following parameters alter the form factors for coherent 
      // interaction. This changes the neutral pi production in NC and
      // charged pi production in CC.
      
      fGMaCOHpi[i]->AddReweightValue(rwgt::fReweightMaCOHpi, sigma);
      fGROCOHpi[i]->AddReweightValue(rwgt::fReweightR0COHpi, sigma);
      fGMaNCEL[i]->AddReweightValue(rwgt::fReweightMaNCEL, sigma);
      fGEtaNCEL[i]->AddReweightValue(rwgt::fReweightEtaNCEL, sigma);
      
      
      fGMaCCQE[i]->AddReweightValue(rwgt::fReweightMaCCQE, sigma);

      fGDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvpCC2pi, sigma);
      fGDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvpNC2pi, sigma);
      fGDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvbarnCC2pi ,  sigma);
      fGDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvbarnNC2pi, sigma);

      fGNoDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvnCC2pi, sigma);
      fGNoDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvnNC2pi, sigma);
      fGNoDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvbarpCC2pi, sigma);
      fGNoDelta2Pi[i]->AddReweightValue(rwgt::fReweightRvbarpNC2pi, sigma);
      
      fGAhtBY[i]->AddReweightValue(rwgt::fReweightAhtBY, sigma);
      fGBhtBY[i]->AddReweightValue(rwgt::fReweightBhtBY, sigma);
      fGCV1uBY[i]->AddReweightValue(rwgt::fReweightCV1uBY, sigma);
      fGCV2uBY[i]->AddReweightValue(rwgt::fReweightCV2uBY, sigma);

      fGAGKY_xF1pi[i]->AddReweightValue(rwgt::fReweightAGKY_xF1pi, sigma);
      fGAGKY_pT1pi[i]->AddReweightValue(rwgt::fReweightAGKY_pT1pi, sigma);

      fGMFP_pi[i]->AddReweightValue(rwgt::fReweightMFP_pi, sigma);
      fGMFP_N[i]->AddReweightValue(rwgt::fReweightMFP_N, sigma);
      fGFrCEx_pi[i]->AddReweightValue(rwgt::fReweightFrCEx_pi, sigma);
      fGFrElas_pi[i]->AddReweightValue(rwgt::fReweightFrElas_pi, sigma);
      fGFrInel_pi[i]->AddReweightValue(rwgt::fReweightFrInel_pi, sigma);
      fGFrAbs_pi[i]->AddReweightValue(rwgt::fReweightFrAbs_pi, sigma);
      fGFrPiProd_pi[i]->AddReweightValue(rwgt::fReweightFrPiProd_pi, sigma);
      fGFrCEx_N[i]->AddReweightValue(rwgt::fReweightFrCEx_N, sigma);
      fGFrElas_N[i]->AddReweightValue(rwgt::fReweightFrElas_N, sigma);
      fGFrInel_N[i]->AddReweightValue(rwgt::fReweightFrInel_N, sigma);
      fGFrAbs_N[i]->AddReweightValue(rwgt::fReweightFrAbs_N, sigma);
      fGFrPiProd_N[i]->AddReweightValue(rwgt::fReweightFrPiProd_N, sigma);
      
      fGCCQEPauliSupViaKF[i]->AddReweightValue(rwgt::fReweightCCQEPauliSupViaKF, sigma);
      fGCCQEMomDistroFGtoSF[i]->AddReweightValue(rwgt::fReweightCCQEMomDistroFGtoSF, sigma);
      
      fGBR1gamma[i]->AddReweightValue(rwgt::fReweightBR1gamma, sigma);
      fGBR1eta[i]->AddReweightValue(rwgt::fReweightBR1eta, sigma);
      fGTheta_Delta2Npi[i]->AddReweightValue(rwgt::fReweightTheta_Delta2Npi, sigma);

      
      fGMaRes[i]->Configure();
      fGMvRes[i]->Configure();
      fGDeltaPi[i]->Configure();
      fGNoDeltaPi[i]->Configure();
      fGMaCOHpi[i]->Configure();
      fGROCOHpi[i]->Configure();
      fGMaNCEL[i]->Configure();
      fGEtaNCEL[i]->Configure();
      fGMaCCQE[i]->Configure();
      fGDelta2Pi[i]->Configure();
      fGNoDelta2Pi[i]->Configure();
      fGAhtBY[i]->Configure();
      fGBhtBY[i]->Configure();
      fGCV1uBY[i]->Configure();
      fGCV2uBY[i]->Configure();

      fGAGKY_xF1pi[i]->Configure();
      fGAGKY_pT1pi[i]->Configure();
      fGMFP_pi[i]->Configure();
      fGMFP_N [i]->Configure();
      fGFrCEx_pi[i]->Configure();
      fGFrElas_pi[i]->Configure();
      fGFrInel_pi[i]->Configure();
      fGFrAbs_pi[i]->Configure();
      fGFrPiProd_pi[i]->Configure();
      fGFrCEx_N[i]->Configure();
      fGFrElas_N[i]->Configure();
      fGFrInel_N[i]->Configure();
      fGFrAbs_N[i]->Configure();
      fGFrPiProd_N[i]->Configure();
      fGCCQEPauliSupViaKF[i]->Configure();
      fGCCQEMomDistroFGtoSF[i]->Configure();
      fGBR1gamma[i]->Configure();
      fGBR1eta[i]->Configure();
      fGTheta_Delta2Npi[i]->Configure();
    }
  }// if fDoSystematics is true

}// End of beginJob

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::analyze(art::Event const& evt)
{

  // Intialize variables
  fEvt        =  evt.id().event();
  ClearMaps();
  art::ServiceHandle<cheat::BackTracker> bt;
  if( !bt->HaveTruthInfo() )
    throw cet::exception("NO TRUTH INFO")
      <<" The input ile does not have truth information.\n"
      << "This module only runs on simulation.\n";
      
  art::Handle< std::vector<simb::MCTruth> > mctruthlist;
  evt.getByLabel(fMCTruthLabel, mctruthlist);
  
  if (mctruthlist->empty()) {
    mf::LogWarning("MRCCSystematics") << "Error retrieving mctruthlist list" ;
    return;
  }

  art::Handle< std::vector<rb::Cluster> > slicelist;
  evt.getByLabel(fClusterLabel, slicelist);
  
  if (slicelist->empty()) {
    mf::LogWarning("MRCCSystematics") << "Error retrieving slice list" ;
    return;
  }


  art::Handle< std::vector<rb::Cluster> > mrslicelist;
  try{
    evt.getByLabel(fMRClusterLabel, mrslicelist);
  }
  catch(...){

  }
  // Grab all the cell hits in the event and lump them into a PtrVector called allhits. 
  // Need this for a few BackTracker functions.

  art::Handle< std::vector<rb::CellHit> > hitlist;
  evt.getByLabel(fHitLabel, hitlist);

  art::PtrVector<rb::CellHit> allhits;
  for(size_t h = 0; h < hitlist->size(); ++h){
    art::Ptr<rb::CellHit> hit(hitlist, h);
    allhits.push_back(hit);
  }


  art::Handle< std::vector<rb::CellHit> > mrHitlist;
  evt.getByLabel(fMRHitLabel, mrHitlist);

  // Grab the tracks and energy object associated with these slices
  art::FindManyP<rb::Track>   sliceToTracks(slicelist, evt, fTrackLabel);
  art::FindOneP<numue::NumuE> sliceToEnergy(slicelist, evt, fEnergyLabel);

  art::FindOneP<murem::MRCCParent>    mrSliceToParent(mrslicelist, evt, fMRParentLabel);
  art::FindManyP<rb::Track>           mrSliceToTracks(mrslicelist, evt, fMRTrackLabel);
  art::FindOneP<numue::NumuE>         mrSliceToEnergy(mrslicelist, evt, fMREnergyLabel);


  art::FindOneP<simb::GTruth> mcTruthToGTruth(mctruthlist, evt, fMCTruthLabel);

  art::Handle< std::vector<simb::GTruth> > gtlist;
  evt.getByLabel(fMCTruthLabel, gtlist);
  
  if (gtlist->empty()) {
    mf::LogWarning("MRCCSystematics") << "Error retrieving GTruth list" ;
    return;
  }

  std::map<int, float > sliRemid;
  std::map<int, int   > sliMuon3D;
  std::map<int, float > sliJmid;
  std::map<int, float > sliLemid;
  std::map<int, TVector3 > sliVert;
  std::map<int, float > sliTrkEnd;
  std::map<int, float > sliTrkStart;
  std::map<int, float > sliPdg;
  std::map<int, float > sliIsCC;
  std::map<int, float > sliQSg;
  std::map<int, float > sliW;
  std::map<int, float > sliX;
  std::map<int, float > sliHadY;
  std::map<int, float > sliNuE;
  std::map<int, int   > sliMC;

  for(int sli = 0; sli < (int)slicelist->size(); sli++){
    art::Ptr<rb::Cluster> thisSlice(slicelist,sli);
    if( thisSlice->IsNoise() )
      continue;
    
    SetDefaults();
   
    const art::Ptr<numue::NumuE>             sliceEnergy = sliceToEnergy.at(sli);
    const std::vector< art::Ptr<rb::Track> > sliceTracks = sliceToTracks.at(sli);    
    
    bool sortByPurity = true;
    std::vector<cheat::NeutrinoEffPur> sliceNus = bt->SliceToMCTruth( thisSlice, allhits, sortByPurity);
    
    if( !sliceNus.empty() ){

      // For the GENIE weights calculations, choose the slice that has the 
      fSlicePurity     = sliceNus[0].purity;
      fSliceEfficiency = sliceNus[0].efficiency;
      
      int truthIndex   =  sliceNus[0].truthColIndex;
      art::Ptr<simb::MCTruth> truth       = sliceNus[0].neutrinoInt;
      simb::MCNeutrino const& sliNeutrino = truth->GetNeutrino();
      
      if(sliNeutrino.CCNC() == simb::kCC )
	fIsCc = true;
      else
	fIsCc = false;
      
      if(fDoSystematicsCalc){
	const simb::GTruth& gTruth = gtlist->at(truthIndex);  
	sliMC[sli] = truthIndex;
	CalculateWeights(*truth, gTruth, sli);
      }// if fDoSystematicsCalc is true
      
      fNuTrueE    = sliNeutrino.Nu().E();
      fQSq        = sliNeutrino.QSqr();
      fW          = sliNeutrino.W();
      fX          = sliNeutrino.X();
      fHadY       = sliNeutrino.Y();
      fPdg        = sliNeutrino.Nu().PdgCode();
      
      int nDaughters = truth->NParticles();
      
      for( int d = 0; d< nDaughters; d++){
	const simb::MCParticle & daughter = truth->GetParticle(d);
	
	if(daughter.Process() == "primary" || daughter.Process() == "Primary"){
	  fDaughterPdg.push_back( daughter.PdgCode() );
	  fDaughterE.push_back( daughter.E() );
	  fDaughterPt.push_back( daughter.Pt() );	
	}
      }// end loop over daughters
      
    }// End if sliceNus not empty
    
    art::Ptr<rb::Track> maxRemidTrack;
    if(sliceTracks.size() > 0){
      art::FindOneP<remid::ReMId> trackRemid(sliceTracks, evt, fNumuIdLabel); 
      float maxId = -1;
      if (trackRemid.isValid()){
	for(int t = 0; t < (int)sliceTracks.size(); t++){
	  art::Ptr<remid::ReMId> remid = trackRemid.at(t);
	  if(remid){
	    if( remid->Value() > maxId ){
	      maxRemidTrack = sliceTracks[t];
	      maxId         = remid->Value();
	      fRemid        = maxId;
              fMuon3D       = maxRemidTrack->Is3D();
	    }
	  }
	}// End loop over slice tracks

	fNumuSandFxs->getContainment( maxRemidTrack, fTrkStart, fTrkEnd);
      }// trackRemid is valid
    }// sliceTracks > 0

    art::FindOneP<rb::Vertex> sliceVertex(slicelist, evt, fVertexLabel);
    if(sliceVertex.isValid()){
      art::Ptr<rb::Vertex> vert = sliceVertex.at(sli);
      if(vert)
	fVert = vert->GetXYZ();
    }
    
    art::FindOneP<jmshower::EID> sliceJmid(slicelist, evt, fJmIdLabel); 
    if(sliceJmid.isValid()){
      
      art::Ptr<jmshower::EID> jmid = sliceJmid.at(sli);
      if(jmid){
	std::cout<<"JMID FOUND AND VALUE AND ENERGY ARE "<<jmid->ANN()<<" "<<jmid->NuEnergy()<<std::endl;
	fJmid       = jmid->ANN();
	fNuJmE      = jmid->NuEnergy();
	fShowerE    = jmid->Sh1Energy();
	fShowerStart= jmid->Sh1Start();                                                                               
	fShowerStop = jmid->Sh1Stop();               
	fegL        = jmid->EGLLL();
	fegT        = jmid->EGLLT();
	femuL       = jmid->EMuLLL();
	femuT       = jmid->EMuLLT();
	fepi0L      = jmid->EPi0LLL();
	fepi0T      = jmid->EPi0LLT();
	fepiL       = jmid->EPiLLL();
	fepiT       = jmid->EPiLLT();
	fepL        = jmid->EPLLL();
	fepT        = jmid->EPLLT();
	fenL        = jmid->ENLLL();
	fenT        = jmid->ENLLT();
	feL         = jmid->ELLL();
	feT         = jmid->ELLT();
	fgap        = jmid->Gap();
	fERatio     = jmid->ShE();
	fpi0mass    = jmid->Pi0Mass();	
      }
    }


    art::FindOneP<lem::PIDDetails> sliceLemid(slicelist, evt, fLemIdLabel);
    if(sliceLemid.isValid()){
      art::Ptr<lem::PIDDetails> lemid = sliceLemid.at(sli);
      if(lemid)
	fLemid          = lemid->Value();
    }
      
    
    art::FindOneP<rvp::RVP>    sliceRvp(slicelist, evt, fRvpLabel);
    if(sliceRvp.isValid()){

      art::Ptr<rvp::RVP> rvpObj = sliceRvp.at(sli);
      if(rvpObj){

	fRvp            = rvpObj->Value();
        fRvpNCells      = rvpObj->NCells();
	fRvpRecoE       = rvpObj->RecoE();
	fRvpLongTr      = rvpObj->LongTr();
	fRvpLongTrFrac  = rvpObj->LongTrFrac();
	fRvpMipFrac     = rvpObj->MipFrac();
	fRvpMipHits     = rvpObj->MipHits();
	fRvpEPl20Frac   = rvpObj->EPl20Frac();
	fRvpEFrac2PlWin = rvpObj->EFrac2PlWin();
	fRvpEFrac6PlWin = rvpObj->EFrac6PlWin();
	fRvpEFrac2SigRd = rvpObj->EFrac2SigRd();
	fRvpShowers3D    = rvpObj->Prongs3D();
	fRvpProngEBal3D = rvpObj->ProngEBal3D();
	fRvpPdg         = rvpObj->Pdg();
      }
    }

    fNuTotGeV     = thisSlice->TotalGeV();    
    if(sliceEnergy)
      fNuSusanE   = sliceEnergy->TrkNonQEE();

   
    fSys->Fill();

    sliRemid[sli] = fRemid;
    sliMuon3D[sli]= fMuon3D;
    sliVert[sli]  = fVert;
    sliJmid[sli]  = fJmid;
    sliLemid[sli] = fLemid;
    sliNuE[sli]   = fNuTrueE;
    sliX[sli]     = fX;
    sliHadY[sli]  = fHadY;
    sliW[sli]     = fW;
    sliQSg[sli]   = fQSq;
    sliPdg[sli]   = fPdg;
    sliIsCC[sli]  = fIsCc;
    sliTrkEnd[sli]   = fTrkEnd;
    sliTrkStart[sli] = fTrkStart;

  }// End loop over slices 

  //  StoreWeights();

  // Fill MRCC variables
  
  int nMrSlices = mrslicelist->size();

  for(int mrSli = 0; mrSli < nMrSlices; mrSli++ ){
    
    art::Ptr<rb::Cluster> thisMrSlice(mrslicelist, mrSli);
    if(thisMrSlice->IsNoise())
      continue;

    SetDefaults();
    
    int origSliceIndex;

    if( mrSliceToParent.isValid() ){
      const  art::Ptr<murem::MRCCParent>  mrSliceParent = mrSliceToParent.at( mrSli );       
      
      if( mrSliceParent ){
	origSliceIndex   = mrSliceParent->ParentIndex();
	fSliceEfficiency = mrSliceParent->ParentEff();
	fSlicePurity     = mrSliceParent->ParentPur();
      }
      
    }// is mrslice to parent valid?

    if(origSliceIndex > -1){
      
      fOrigRemid = sliRemid[origSliceIndex];
      fOrigMuon3D= sliMuon3D[origSliceIndex];
      fOrigLemid = sliLemid[origSliceIndex];
      fOrigJmid  = sliJmid[origSliceIndex];
      fNuTrueE   = sliNuE[origSliceIndex];   
      fX         = sliX[origSliceIndex];     
      fHadY      = sliHadY[origSliceIndex]; 
      fW         = sliW[origSliceIndex]; 
      fQSq       = sliQSg[origSliceIndex];
      fPdg       = sliPdg[origSliceIndex];
      fIsCc      = sliIsCC[origSliceIndex];
      fOrigVert  = sliVert[origSliceIndex];
      fOrigTrkEnd   = sliTrkEnd[origSliceIndex];
      fOrigTrkStart = sliTrkStart[origSliceIndex];

      if(fDoSystematicsCalc)
	FindWeights( origSliceIndex );

      const simb::MCTruth& mrccTruth = mctruthlist->at(sliMC[origSliceIndex]);
      int nDaughters = mrccTruth.NParticles();
      
      for( int d = 0; d< nDaughters; d++){
	const simb::MCParticle & daughter = mrccTruth.GetParticle(d);

	if(daughter.Process() == "primary" || daughter.Process() == "Primary"){
	  fDaughterPdgMrcc.push_back( daughter.PdgCode() );
	  fDaughterEMrcc.push_back( daughter.E() );
	  fDaughterPtMrcc.push_back( daughter.Pt() );	
	}
      }// end loop over daughters
      
      
      float mrMaxId = -1;

      if( mrSliceToTracks.isValid() ){
	const std::vector< art::Ptr<rb::Track> > mrSliceTracks = mrSliceToTracks.at( mrSli );    
	
	if (mrSliceTracks.size() != 0 ){
	  art::FindOneP<remid::ReMId> mrTrackRemid(mrSliceTracks, evt, fMRNumuIdLabel);  
	  int nTracks = mrSliceTracks.size(); 
	  
	  for(int mrt = 0; mrt < nTracks; mrt++ ){
	    if (mrTrackRemid.isValid()){
	      art::Ptr<remid::ReMId> mrRemid = mrTrackRemid.at(mrt);
	      if( mrRemid ){
		if( mrRemid->Value() > mrMaxId )
		  mrMaxId = mrRemid->Value();
	      }
	    }
	  }// End loop over mr tracks in thisMrSlice
	}// if mrSliceTracks
      }

      art::FindOneP<rb::Vertex> mrSliceVertex(mrslicelist, evt, fMRVertexLabel);
      if(mrSliceVertex.isValid()){
	art::Ptr<rb::Vertex> vert = mrSliceVertex.at(mrSli);
	if(vert)
	  fVertMrcc = vert->GetXYZ();	
      }
      
      art::FindOneP<jmshower::EID> mrSliceJmid(mrslicelist, evt, fMRJmIdLabel);  
      if(mrSliceJmid.isValid() ){
	art::Ptr<jmshower::EID> mrJmid = mrSliceJmid.at(mrSli);
	if (mrJmid){
	  std::cout<<" MRJMID FOUND AND VALUE IS "<<mrJmid->ANN()<<" "<<mrJmid->NuEnergy()<<std::endl;
	  fJmidMrcc       = mrJmid->ANN();
	  fNuJmEMrcc      = mrJmid->NuEnergy();
	  fShowerEMrcc    = mrJmid->Sh1Energy();
	  fShowerStopMrcc = mrJmid->Sh1Start();                                                                       
	  fShowerStartMrcc= mrJmid->Sh1Stop();   
	  fegLMrcc        = mrJmid->EGLLL();
	  fegTMrcc        = mrJmid->EGLLT();
	  femuLMrcc       = mrJmid->EMuLLL();
	  femuTMrcc       = mrJmid->EMuLLT();
	  fepi0LMrcc      = mrJmid->EPi0LLL();
	  fepi0TMrcc      = mrJmid->EPi0LLT();
	  fepiLMrcc       = mrJmid->EPiLLL();
	  fepiTMrcc       = mrJmid->EPiLLT();
	  fepLMrcc        = mrJmid->EPLLL();
	  fepTMrcc        = mrJmid->EPLLT();
	  fenLMrcc        = mrJmid->ENLLL();
	  fenTMrcc        = mrJmid->ENLLT();
	  feLMrcc         = mrJmid->ELLL();
	  feTMrcc         = mrJmid->ELLT();
	  fgapMrcc        = mrJmid->Gap();
	  fERatioMrcc     = mrJmid->ShE();
	  fpi0massMrcc    = mrJmid->Pi0Mass();
	}
      }

      art::FindOneP<lem::PIDDetails> mrSliceLemid(mrslicelist, evt, fMRLemIdLabel);  
      if(mrSliceLemid.isValid()){
	art::Ptr<lem::PIDDetails> mrLemid = mrSliceLemid.at(mrSli);
	if(mrLemid)
	  fLemidMrcc      = mrLemid->Value();
      }
    
      art::FindOneP<rvp::RVP>    mrSliceRvp(mrslicelist, evt, fMRRvpLabel);
      if(mrSliceRvp.isValid()){
	art::Ptr<rvp::RVP> mrRvpObj = mrSliceRvp.at(mrSli);
	if(mrRvpObj){
	  fRvpMrcc            = mrRvpObj->Value();
	  fRvpNCellsMrcc      = mrRvpObj->NCells();
	  fRvpRecoEMrcc       = mrRvpObj->RecoE();
	  fRvpLongTrMrcc      = mrRvpObj->LongTr();
	  fRvpLongTrFracMrcc  = mrRvpObj->LongTrFrac();
	  fRvpMipFracMrcc     = mrRvpObj->MipFrac();
	  fRvpMipHitsMrcc     = mrRvpObj->MipHits();
	  fRvpEPl20FracMrcc   = mrRvpObj->EPl20Frac();
	  fRvpEFrac2PlWinMrcc = mrRvpObj->EFrac2PlWin();
	  fRvpEFrac6PlWinMrcc = mrRvpObj->EFrac6PlWin();
	  fRvpEFrac2SigRdMrcc = mrRvpObj->EFrac2SigRd();
	  fRvpShowers3DMrcc    = mrRvpObj->Prongs3D();
	  fRvpProngEBal3DMrcc = mrRvpObj->ProngEBal3D();
	  fRvpPdgMrcc         = mrRvpObj->Pdg();
	}
      }

      fNuTotGeVMrcc   = thisMrSlice->TotalGeV();
      fRemidMrcc      = mrMaxId;            
      
      
      if( mrSliceToEnergy.isValid() ){
	const art::Ptr<numue::NumuE> mrSliceEnergy =  mrSliceToEnergy.at( mrSli );
	if (mrSliceEnergy) 
	  fNuSusanEMrcc = mrSliceEnergy->TrkNonQEE();
      }
      
      fSysMrcc->Fill();
    }// end if origSliceIndex > 0   
  }// end loop over mrcc slices
  
  return;

} // End of analyser

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::SetDefaults()
{
 
  fDaughterPdgMrcc.clear();
  fDaughterPtMrcc.clear();
  fDaughterEMrcc.clear();
  fDaughterPdg.clear();
  fDaughterE.clear();
  fDaughterPt.clear();

  
  fPdg                = -5;
  fIsCc               = -5;
  fQSq                = -5;
  fW                  = -5;
  fX                  = -5;
  fHadY               = -5;
  fNuTrueE            = -5;
 
  fRemid              = -5;
  fMuon3D             = -5;
  fLemid              = -5;
  fNuSusanE           = -5;
  fNuJmE              = -5;
  fNuTotGeV           = -5;
 
  fRemidMrcc          = -5;
  fJmidMrcc           = -5;
  fLemidMrcc          = -5;
  fNuSusanEMrcc       = -5;
  fNuJmEMrcc          = -5;
  fNuTotGeVMrcc       = -5;

  fVert.SetXYZ( -5, -5, -5);
  
  fTrkStart           = -5;
  fTrkEnd             = -5;

  fVertMrcc.SetXYZ( -5, -5, -5);

  fOrigTrkStart       = -5;
  fOrigTrkEnd         = -5;

  fOrigRemid          = -5;
  fOrigMuon3D         = -5;
  fOrigLemid          = -5;
  fOrigJmid           = -5;
  fSlicePurity        = -5;
  fSliceEfficiency    = -5;
  fOrigVert.SetXYZ( -5, -5, -5);

  fJmid               = -5;
  fShowerE            = -5;
  fShowerStart.SetXYZ( -5, -5, -5);
  fShowerStop.SetXYZ( -5, -5, -5);
  fegL                = -5;
  fegT                = -5;
  femuL               = -5;
  femuL               = -5;
  fepL                = -5;
  fepT                = -5;
  fepi0L              = -5;
  fepi0T              = -5;
  fepiL               = -5;
  fepiT               = -5;
  fenL                = -5;
  fenT                = -5;
  feL                 = -5;
  feT                 = -5;
  fgap                = -5;
  fERatio             = -5;
  fpi0mass            = -5;

  fJmidMrcc           = -5;
  fShowerEMrcc        = -5;
  fShowerStartMrcc.SetXYZ( -5, -5, -5);
  fShowerStopMrcc.SetXYZ(  -5, -5, -5);
  fegLMrcc            = -5;
  fegTMrcc            = -5;
  femuLMrcc           = -5;
  femuLMrcc           = -5;
  fepLMrcc            = -5;
  fepTMrcc            = -5;
  fepi0LMrcc          = -5;
  fepi0TMrcc          = -5;
  fepiLMrcc           = -5;
  fepiTMrcc           = -5;
  fenLMrcc            = -5;
  fenTMrcc            = -5;
  feLMrcc             = -5;
  feTMrcc             = -5;
  fgapMrcc            = -5;
  fERatioMrcc         = -5;
  fpi0massMrcc        = -5;

  fRvp                = -5;
  fRvpNCells          = -5;
  fRvpRecoE           = -5;
  fRvpLongTr          = -5;
  fRvpLongTrFrac      = -5;
  fRvpMipFrac         = -5;
  fRvpMipHits         = -5;
  fRvpEPl20Frac       = -5;
  fRvpEFrac2PlWin     = -5;
  fRvpEFrac6PlWin     = -5;
  fRvpEFrac2SigRd     = -5;
  fRvpShowers3D        = -5;
  fRvpProngEBal3D     = -5;
  fRvpPdg             = -5;

  fRvpMrcc            = -5;
  fRvpNCellsMrcc      = -5;
  fRvpRecoEMrcc       = -5;
  fRvpLongTrMrcc      = -5;
  fRvpLongTrFracMrcc  = -5;
  fRvpMipFracMrcc     = -5;
  fRvpMipHitsMrcc     = -5;
  fRvpEPl20FracMrcc   = -5;
  fRvpEFrac2PlWinMrcc = -5;
  fRvpEFrac6PlWinMrcc = -5;
  fRvpEFrac2SigRdMrcc = -5;
  fRvpShowers3DMrcc    = -5;
  fRvpProngEBal3DMrcc = -5;
  fRvpPdgMrcc         = -5;


  if(fDoSystematicsCalc){    
    for(int i = 0; i < 2; i++){
      fgMaRes[i]      = 1;
      fgMvRes[i]      = 1;
      fgDeltaPi[i]      = 1;
      fgNoDeltaPi[i]    = 1;
      fgMaCOHpi[i]      = 1;
      fgROCOHpi[i]      = 1;
      fgMaNCEL[i]       = 1;
      fgEtaNCEL[i]      = 1;
      fgMaCCQE[i]       = 1;
      fgDelta2Pi[i]     = 1;
      fgNoDelta2Pi[i]   = 1;
      fgAhtBY[i]        = 1;
      fgBhtBY[i]        = 1;
      fgCV1uBY[i]       = 1;
      fgCV2uBY[i]       = 1;

      fgAGKY_xF1pi[i]   = 1;
      fgAGKY_pT1pi[i]   = 1;
      fgMFP_pi[i]       = 1;
      fgMFP_N[i]        = 1;
      fgFrCEx_pi[i]     = 1;
      fgFrElas_pi[i]    = 1;
      fgFrInel_pi[i]    = 1;
      fgFrAbs_pi[i]     = 1;
      fgFrPiProd_pi[i]  = 1;
      fgFrCEx_N[i]      = 1;
      fgFrElas_N[i]     = 1;
      fgFrInel_N[i]     = 1;
      fgFrAbs_N[i]      = 1;
      fgFrPiProd_N[i]   = 1;
      fgBR1gamma[i]     = 1;
      fgBR1eta[i]       = 1;
      fgCCQEPauliSupViaKF[i]  = 1;
      fgCCQEMomDistroFGtoSF[i]= 1;
      fgTheta_Delta2Npi[i]    = 1;
    }
  }// end if(fDoSystematicsCalc)

}// End of SetDefaults


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////


void murem::MRCCSystematics::CalculateWeights( const simb::MCTruth& truth, 
					       const simb::GTruth&  gTruth,
					       int   sliceIndex)
{

  std::cout<<"Inside calculate weights \n";

  for(int sig = 0; sig < 2; sig++) {       
    
    std::cout<<"Inside sig "<<sig<<"\n";
    std::map<int, float> m;

    fgMaRes[sig]   =  fGMaRes[sig]->CalcWeight(truth, gTruth);
    mgMaRes[ sig ][ sliceIndex ] = fgMaRes[ sig  ];

    std::cout<<"sliceindex, sig, weight, weightMap "<<sliceIndex<<" "<<sig<<" "<<fgMaRes[sig]<<" "<<mgMaRes[sig][sliceIndex]<<"\n";

    fgMvRes[sig] =  fGMvRes[sig]->CalcWeight(truth, gTruth);
    mgMvRes[ sig ][ sliceIndex ] = fgMvRes[ sig  ];
    
 
    fgDeltaPi[sig] =  fGDeltaPi[sig]->CalcWeight(truth, gTruth);
    mgDeltaPi[ sig ][ sliceIndex ] = fgDeltaPi[ sig  ];
    
 
    fgNoDeltaPi[sig] =  fGNoDeltaPi[sig]->CalcWeight(truth, gTruth);
    mgNoDeltaPi[ sig ][ sliceIndex ] = fgNoDeltaPi[ sig  ];
    
 
    fgMaCOHpi[sig] =  fGMaCOHpi[sig]->CalcWeight(truth, gTruth);
    mgMaCOHpi[ sig ][ sliceIndex ] = fgMaCOHpi[ sig  ];
    
 
    fgROCOHpi[sig] =  fGROCOHpi[sig]->CalcWeight(truth, gTruth);
    mgROCOHpi[ sig ][ sliceIndex ] = fgROCOHpi[ sig  ];
    
 
    fgMaNCEL[sig] =  fGMaNCEL[sig]->CalcWeight(truth, gTruth) ;
    mgMaNCEL[ sig ][ sliceIndex ] = fgMaNCEL[ sig  ];
    

    fgEtaNCEL[sig] =  fGEtaNCEL[sig]->CalcWeight(truth, gTruth) ;
    mgEtaNCEL[ sig ][ sliceIndex ] = fgEtaNCEL[ sig  ];
    

    fgMaCCQE[sig] =  fGMaCCQE[sig]->CalcWeight(truth, gTruth) ;
    mgMaCCQE[ sig ][ sliceIndex ] = fgMaCCQE[ sig  ];
    

    fgDelta2Pi[sig] =  fGDelta2Pi[sig]->CalcWeight(truth, gTruth) ;
    mgDelta2Pi[ sig ][ sliceIndex ] = fgDelta2Pi[ sig  ];
    

    fgNoDelta2Pi[sig] =  fGNoDelta2Pi[sig]->CalcWeight(truth, gTruth) ;
    mgNoDelta2Pi[ sig ][ sliceIndex ] = fgNoDelta2Pi[ sig  ];
    

    fgAhtBY[sig] =  fGAhtBY[sig]->CalcWeight(truth, gTruth) ;
    mgAhtBY[ sig ][ sliceIndex ] = fgAhtBY[ sig  ];
    

    fgBhtBY[sig] =  fGBhtBY[sig]->CalcWeight(truth, gTruth) ;
    mgBhtBY[ sig ][ sliceIndex ] = fgBhtBY[ sig  ];
    

    fgCV1uBY[sig] =  fGCV1uBY[sig]->CalcWeight(truth, gTruth) ;
    mgCV1uBY[ sig ][ sliceIndex ] = fgCV1uBY[ sig  ];
    

    fgCV2uBY[sig] =  fGCV2uBY[sig]->CalcWeight(truth, gTruth) ;
    mgCV2uBY[ sig ][ sliceIndex ] = fgCV2uBY[ sig  ];
    

    fgAGKY_xF1pi[sig] =  fGAGKY_xF1pi[sig]->CalcWeight(truth, gTruth) ;
    mgAGKY_xF1pi[ sig ][ sliceIndex ] = fgAGKY_xF1pi[ sig  ];
    

    fgAGKY_pT1pi[sig] =  fGAGKY_pT1pi[sig]->CalcWeight(truth, gTruth) ;
    mgAGKY_pT1pi[ sig ][ sliceIndex ] = fgAGKY_pT1pi[ sig  ];
    

    fgMFP_pi[sig] =  fGMFP_pi[sig]->CalcWeight(truth, gTruth) ;
    mgMFP_pi[ sig ][ sliceIndex ] = fgMFP_pi[ sig  ];
    

    fgMFP_N [sig] =  fGMFP_N[sig]->CalcWeight(truth, gTruth) ;
    mgMFP_N[ sig ][ sliceIndex ] = fgMFP_N[ sig  ];
    

    fgFrCEx_pi[sig] =  fGFrCEx_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrCEx_pi[ sig ][ sliceIndex ] = fgFrCEx_pi[ sig  ];
    

    fgFrElas_pi[sig] =  fGFrElas_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrElas_pi[ sig ][ sliceIndex ] = fgFrElas_pi[ sig  ];
    

    fgFrInel_pi[sig] =  fGFrInel_pi[sig]->CalcWeight(truth, gTruth) ;
    mgFrInel_pi[ sig ][ sliceIndex ] = fgFrInel_pi[ sig  ];
    

    fgFrAbs_pi[sig] =  fGFrAbs_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrAbs_pi[ sig ][ sliceIndex ] = fgFrAbs_pi[ sig  ];
    

    fgFrPiProd_pi[sig] =  fGFrPiProd_pi[sig]->CalcWeight(truth, gTruth) ;
    mgFrPiProd_pi[ sig ][ sliceIndex ] = fgFrPiProd_pi[ sig  ];
    

    fgFrCEx_N[sig] =  fGFrElas_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrCEx_N[ sig ][ sliceIndex ] = fgFrCEx_N[ sig  ];
    

    fgFrElas_N[sig] =  fGFrElas_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrElas_N[ sig ][ sliceIndex ] = fgFrElas_N[ sig  ];
    

    fgFrInel_N[sig] =  fGFrInel_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrInel_N[ sig ][ sliceIndex ] = fgFrInel_N[ sig  ];
    

    fgFrAbs_N[sig] =  fGFrAbs_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrAbs_N[ sig ][ sliceIndex ] = fgFrAbs_N[ sig  ];
    

    fgFrPiProd_N[sig] =  fGFrPiProd_N[sig]->CalcWeight(truth, gTruth) ;
    mgFrPiProd_N[ sig ][ sliceIndex ] = fgFrPiProd_N[ sig  ];
    

    fgBR1gamma[sig] =  fGBR1gamma[sig]->CalcWeight(truth, gTruth) ;
    mgBR1gamma[ sig ][ sliceIndex ] = fgBR1gamma[ sig  ];
    

    fgBR1eta[sig] =  fGBR1eta[sig]->CalcWeight(truth, gTruth) ;
    mgBR1eta[ sig ][ sliceIndex ] = fgBR1eta[ sig  ];
    
	  
    fgCCQEPauliSupViaKF[sig] =  fGCCQEPauliSupViaKF[sig]->CalcWeight(truth, gTruth) ;
    mgCCQEPauliSupViaKF[ sig ][ sliceIndex ] = fgCCQEPauliSupViaKF[ sig  ];
    

    fgCCQEMomDistroFGtoSF[sig] =  fGCCQEMomDistroFGtoSF[sig]->CalcWeight(truth, gTruth) ;
    mgCCQEMomDistroFGtoSF[ sig ][ sliceIndex ] = fgCCQEMomDistroFGtoSF[ sig  ];
    

    fgTheta_Delta2Npi[sig] =  fGTheta_Delta2Npi[sig]->CalcWeight(truth, gTruth) ;
    mgTheta_Delta2Npi[ sig ][ sliceIndex ] = fgTheta_Delta2Npi[ sig  ];
    

	
  }
  
  return;
}// End of CalculateWeights
 


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////


void murem::MRCCSystematics::FindWeights( int origSliceIndex )
{
  std::cout<<"Inside findweight. origSliceindex  "<<origSliceIndex<<"\n";

  for(int sig = 0; sig < 2; sig++){

    fgMaRes[sig]           =      mgMaRes[sig][ origSliceIndex ];
    fgMvRes[sig]           =      mgMvRes[sig][ origSliceIndex ];
    fgDeltaPi[sig]         =      mgDeltaPi[sig][ origSliceIndex ];
    fgNoDeltaPi[sig]       =      mgNoDeltaPi[sig][ origSliceIndex ];
    fgMaCOHpi[sig]         =      mgMaCOHpi[sig][ origSliceIndex ];
    fgROCOHpi[sig]         =      mgROCOHpi[sig][ origSliceIndex ];
    fgMaNCEL[sig]          =      mgMaNCEL[sig][ origSliceIndex ];
    fgEtaNCEL[sig]         =      mgEtaNCEL[sig][ origSliceIndex ];
    fgMaCCQE[sig]          =      mgMaCCQE[sig][ origSliceIndex ];
    fgDeltaPi[sig]         =      mgDeltaPi[sig][ origSliceIndex ];
    fgNoDeltaPi[sig]       =      mgNoDeltaPi[sig][ origSliceIndex ];
    fgAhtBY[sig]           =      mgAhtBY[sig][ origSliceIndex ];
    fgBhtBY[sig]           =      mgBhtBY[sig][ origSliceIndex ];
    fgCV1uBY[sig]          =      mgCV1uBY[sig][ origSliceIndex ];
    fgCV1uBY[sig]          =      mgCV1uBY[sig][ origSliceIndex ];
    fgAGKY_xF1pi[sig]      =      mgAGKY_xF1pi[sig][ origSliceIndex ];
    fgAGKY_pT1pi[sig]      =      mgAGKY_pT1pi[sig][ origSliceIndex ];
    fgMFP_pi[sig]          =      mgMFP_pi[sig][ origSliceIndex ];
    fgMFP_N[sig]           =      mgMFP_N[sig][ origSliceIndex ];
    fgFrCEx_pi[sig]        =      mgFrCEx_pi[sig][ origSliceIndex ];
    fgFrElas_pi[sig]       =      mgFrElas_pi[sig][ origSliceIndex ];
    fgFrInel_pi[sig]       =      mgFrInel_pi[sig][ origSliceIndex ];
    fgFrAbs_pi[sig]        =      mgFrAbs_pi[sig][ origSliceIndex ];
    fgFrPiProd_pi[sig]     =      mgFrPiProd_pi[sig][ origSliceIndex ];
    fgFrCEx_N[sig]         =      mgFrCEx_N[sig][ origSliceIndex ];
    fgFrElas_N[sig]        =      mgFrElas_N[sig][ origSliceIndex ];
    fgFrInel_N[sig]        =      mgFrInel_N[sig][ origSliceIndex ];
    fgFrAbs_N[sig]         =      mgFrAbs_N[sig][ origSliceIndex ];
    fgFrPiProd_N[sig]      =      mgFrPiProd_N[sig][ origSliceIndex ];
    fgBR1gamma[sig]        =      mgBR1gamma[sig][ origSliceIndex ];
    fgBR1eta[sig]          =      mgBR1eta[sig][ origSliceIndex ];
    fgTheta_Delta2Npi[sig] =      mgTheta_Delta2Npi[sig][ origSliceIndex ];
    fgCCQEPauliSupViaKF[sig]   =      mgCCQEPauliSupViaKF[sig][ origSliceIndex ];
    fgCCQEMomDistroFGtoSF[sig] =      mgCCQEMomDistroFGtoSF[sig][ origSliceIndex ];


  }

  return;
}// End of FindWeights


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::ClearMaps()
{

  for( int sig = 0; sig < 2; sig++){

    mgMaRes[sig].clear();
    mgMvRes[sig].clear();
    mgDeltaPi[sig].clear();
    mgNoDeltaPi[sig].clear();
    mgMaCOHpi[sig].clear();
    mgROCOHpi[sig].clear();
    mgMaNCEL[sig].clear();
    mgEtaNCEL[sig].clear();
    mgMaCCQE[sig].clear();
    mgDelta2Pi[sig].clear();
    mgNoDelta2Pi[sig].clear();
    mgAhtBY[sig].clear();
    mgBhtBY[sig].clear();
    mgCV1uBY[sig].clear();
    mgCV2uBY[sig].clear();
    mgAGKY_xF1pi[sig].clear();
    mgAGKY_pT1pi[sig].clear();
    mgMFP_pi[sig].clear();
    mgMFP_N[sig].clear();
    mgFrCEx_pi[sig].clear();
    mgFrElas_pi[sig].clear();
    mgFrInel_pi[sig].clear();
    mgFrAbs_pi[sig].clear();
    mgFrPiProd_pi[sig].clear();
    mgFrCEx_N[sig].clear();
    mgFrElas_N[sig].clear();
    mgFrInel_N[sig].clear();
    mgFrAbs_N[sig].clear();
    mgFrPiProd_N[sig].clear();
    mgCCQEPauliSupViaKF[sig].clear();
    mgCCQEMomDistroFGtoSF[sig].clear();
    mgBR1gamma[sig].clear();
    mgBR1eta[sig].clear();
    mgTheta_Delta2Npi[sig].clear();

  }   
  return;

}// End of ClearMaps



/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::endSubRun( const art::SubRun& sr)
{

  art::Handle<sumdata::POTSum> pot;
  art::ServiceHandle<cheat::BackTracker> bt;

  //if this is mc get the spill info from the generator

  if (bt->HaveTruthInfo())
      sr.getByLabel( fMCTruthLabel, pot);
  
  //if we have real data try to get the spill info from the NuMI beam
  
  else
    sr.getByLabel( fNumiBeamLabel, pot);
  

  if(!pot.failedToGet())

    {
      float subrunpot = pot->totgoodpot;
      if (!bt->HaveTruthInfo()) subrunpot *= 1e12;
      fTotalPOT += subrunpot;
    }

}// End of endSubRun

 
/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

void murem::MRCCSystematics::endJob()
{
  hPOT->Fill(.5, fTotalPOT);
}// End of endJob


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////


DEFINE_ART_MODULE(murem::MRCCSystematics);



/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////
