///
/// \brief
/// \author Xuebing Bu - xbbu@fnal.gov
///
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"

// ROOT includes
#include "TMath.h"
#include "TTree.h"

#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"

// NOvA includes
#include "MCCheater/BackTracker.h"

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <map>

class TTree;

namespace rb  { class Cluster; class Track; }
namespace sim { class Particle; }
namespace simb{ class MCFlux; class MCTruth; class MCNeutrino; }

namespace ncs {

  class DumpMC : public art::EDAnalyzer {
  public:
    explicit DumpMC(fhicl::ParameterSet const& pset);
    virtual ~DumpMC();

    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
  private:

    TTree *_otree;
    //for pi0->2gamma
    int   HasPi0;
    float Epi0;
    float gam1E;
    float gam1Px;
    float gam1Py;
    float gam1Pz;
    float gam2E;
    float gam2Px;
    float gam2Py;
    float gam2Pz;

  };
}

namespace ncs{

  //......................................................................
  DumpMC::DumpMC(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    mf::LogInfo("NCAna") <<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  DumpMC::~DumpMC()
  {
  }
  
  //......................................................................
  void DumpMC::reconfigure(const fhicl::ParameterSet& pset)
  {
    //inputPDG = pset.get<int>("pdg"); 
  }

  //......................................................................
  void DumpMC::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    

    _otree=tfs->make<TTree>("DumpMC","dump MC decay chain");

    _otree->Branch("HasPi0",&HasPi0,"HasPi0/I");
    _otree->Branch("Epi0",&Epi0,"Epi0/F");
    _otree->Branch("gam1E",&gam1E,"gam1E/F");
    _otree->Branch("gam1Px",&gam1Px,"gam1Px/F");
    _otree->Branch("gam1Py",&gam1Py,"gam1Py/F");
    _otree->Branch("gam1Pz",&gam1Pz,"gam1Pz/F");
    _otree->Branch("gam2E",&gam2E,"gam2E/F");
    _otree->Branch("gam2Px",&gam2Px,"gam2Px/F");
    _otree->Branch("gam2Py",&gam2Py,"gam2Py/F");
    _otree->Branch("gam2Pz",&gam2Pz,"gam2Pz/F"); 
  }

  void DumpMC::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<geo::Geometry> geom;

    art::ServiceHandle<cheat::BackTracker> bt;
/*
    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel("slicer", slices);

    int bestSlice = -1;
    unsigned int mostCells = 0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;
      if(slice.NCell() > mostCells){        
        mostCells = slice.NCell();
        bestSlice = sliceIdx;
      }
    }
    if(bestSlice < 0) return;

    art::Handle<std::vector<rb::CellHit> > chits;
    evt.getByLabel("calhit", chits);
    art::PtrVector<rb::CellHit> allhits;
    for(size_t h = 0; h < chits->size(); ++h)
      allhits.push_back(art::Ptr<rb::CellHit>(chits, h));
    const art::Ptr<rb::Cluster> allslice(slices,bestSlice);
    std::vector<cheat::NeutrinoEffPur> StoNINT = bt->SliceToNeutrinoInteractions(allslice, allhits);
    if( StoNINT.empty() ) return;
      art::Ptr<simb::MCTruth>& truth = StoNINT[0].neutrinoInt;
      art::PtrVector<simb::MCTruth> pv;
      pv.push_back(truth);
      art::FindManyP<simb::MCFlux> fmf(pv, evt, "generator");
      std::vector< art::Ptr<simb::MCFlux> > fluxs = fmf.at(0);
      simb::MCFlux flux;
      if(fluxs.size() == 1){
        flux = *fluxs[0];
      }
      else{
        std::cout<<"There is no assoicated neutrino intereaction !"<<std::endl;
      }
    const simb::MCNeutrino& nu = truth->GetNeutrino();
    
    ///for NC
    if( nu.CCNC() != 1 ) return;
    
    std::cout<<"Neutrino Info: ";
    std::cout<<nu.Nu().E()<<" "<<nu.CCNC()<<" "<<nu.Nu().PdgCode()<<" "<<flux.fntype<<std::endl;
*/

    //print out the MC particles
    sim::ParticleNavigator const& pn = bt->ParticleNavigator();
    
    //select the true pi0
    int pi0Idx=-1;
    double epi0=-99.;
    for(int ip = 0; ip < pn.NumberOfPrimaries(); ++ip){
      const sim::Particle* genpar = pn.Primary(ip);
      if(genpar->PdgCode() != 111) continue;
      if(genpar->E() > epi0){
	pi0Idx = ip;
	epi0 = genpar->E();
      }
    }//loop true particles
    
    bool hasPi0=false;
    
    if( pi0Idx>=0 ){
      const sim::Particle* selectedpi0 = pn.Primary(pi0Idx);
      if(selectedpi0->NumberDaughters() == 2){
	const sim::Particle* gam1 = pn[selectedpi0->Daughter(0)];
	const sim::Particle* gam2 = pn[selectedpi0->Daughter(1)];
	if( gam1->PdgCode()==22 && gam2->PdgCode()==22 ){
	  hasPi0=true;
	  HasPi0=1;
	  Epi0=selectedpi0->E();
	  gam1E=gam1->E();
	  gam1Px=gam1->Px();
	  gam1Py=gam1->Py();
	  gam1Pz=gam1->Pz();
	  gam2E=gam2->E();
	  gam2Px=gam2->Px();
	  gam2Py=gam2->Py();
	  gam2Pz=gam2->Pz();
	}//
      }//has 2 photons
    }//has pi0
    
    if( !hasPi0 ) return;
    
    if( !hasPi0 ){
      HasPi0=0;
      Epi0=-1.;
      gam1E=-1.;
      gam1Px=-1.;
      gam1Py=-1.;
      gam1Pz=-1.;
      gam2E=-1.;
      gam2Px=-1.;
      gam2Py=-1.;
      gam2Pz=-1.;
    }
    
    int numberOfParticles = pn.size();
    mf::LogDebug("NCAna")<<"# of particles = "<<numberOfParticles;
    
    //build a map for indexing and track ID
    std::map<std::string,int> IndexMap;
    IndexMap.clear();
    
    for( int i=0; i<numberOfParticles; ++i ){
      sim::Particle* particle = pn.Particle(i);
      std::stringstream ss;//create a stringstream
      ss << particle->TrackId();//add number to the stream
      IndexMap[ss.str()] = i;
    }
    
    for( int i=0; i<numberOfParticles; ++i ){
      sim::Particle* particle = pn.Particle(i);
      int pdgid=particle->PdgCode();
      //int status=particle->StatusCode();
      double energy=particle->E(); 
      int ndau=particle->NumberDaughters();
      //int trackid=particle->TrackId();
      
      mf::LogDebug("NCAna")<<i<<" "<<pdgid<<" "<<energy<<" ";
      
      //find mother
     int MotherIndex=particle->Mother();
     if( MotherIndex>0 ){
       mf::LogDebug("NCAna")<<"m: "; 
       std::stringstream ss;
       ss << pn[MotherIndex]->TrackId();
       std::map<std::string, int>::const_iterator mitr=IndexMap.find(ss.str());
       if( mitr != IndexMap.end() )
	 mf::LogDebug("NCAna")<<mitr->second<<" "; 
     } 
     
     //find daughter
     if( ndau>0 ){
       mf::LogDebug("NCAna")<<"d: ";
       for( int j=0; j<ndau; ++j ){
         //std::cout<<"dau: "<<j<<": ";
         int dauID=particle->Daughter(j);
         //std::cout<<pn[dauID]->TrackId()<<" "<<pn[dauID]->PdgCode()<<" ";
         std::stringstream ss;
         ss << pn[dauID]->TrackId();
         std::map<std::string, int>::const_iterator mitr=IndexMap.find(ss.str());
         if( mitr != IndexMap.end() )
	   mf::LogDebug("NCAna")<<mitr->second<<" ";
       }
     }//has daughters
     mf::LogDebug("NCAna")<<"\n";
    }
    
    _otree->Fill();
    
    return;
  }//end analyze
  
}//end namespace

namespace ncs{

  DEFINE_ART_MODULE(DumpMC);

}
