///
/// \brief 
/// \author Xuebing Bu - xbbu@fnal.gov
///
#include "RecoJMShower/JMShower.h"
#include "RecoJMShower/EID.h"
#include "RecVarPID/RVP.h"

#include "Calibrator/Calibrator.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"

// ROOT includes
#include "TVector3.h"
#include "TH1F.h"
#include "TMath.h"
#include "TTree.h"
#include "TProfile2D.h"

#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "SummaryData/POTSum.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"

#include "Utilities/AssociationUtil.h"
#include "MCCheater/BackTracker.h"
#include "Utilities/func/MathUtil.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>
#include <vector>

//include TMVA
#include "TMVA/Tools.h"
#include "TMVA/Reader.h"
#include "TMVA/MethodCuts.h"

class TVector3;
class TH1F;
class TTree;

namespace rb{class Cluster; class Track;}
namespace sim{class Particle;}
namespace simb{class MCFlux; class MCTruth; class MCNeutrino;}

namespace ncs {
  class NCAna : public art::EDAnalyzer {
  public:
    explicit NCAna(fhicl::ParameterSet const& pset);
    virtual ~NCAna();
    
    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    void endSubRun(const art::SubRun& sr);
    void endJob(); 
    double SimpleOscProb(const simb::MCFlux& flux,
                         const simb::MCNeutrino& nu) const;
    double fTotalPOT;

  private:

    const rb::RecoHit MakeRecoHit(art::Ptr<rb::CellHit> const& chit,
				  double const& w);

    //folders to get objects from
    std::string fSliceLabel;     ///label for slices
    //std::string fDTrackLabel;     ///label for discrete tracks
    std::string fKTrackLabel;     ///label for kalman tracks
    std::string fProngLabel;     ///label for prong 
    std::string fInstLabel3D;    ///instance label for prongs 3D
    std::string fInstLabel2D;    ///instance label for prongs 2D
    std::string fMCFluxLabel;    ///label for generator information
    std::string fVertexLabel;    ///label for vertex
    std::string fCellHitLabel;   ///label for cell hits
    std::string fShowerLabel;    ///label for jmshower
    std::string frvpLabel;       ///label for rvp
    int isMC; //mode for MC

    //save event info to ntuples
    TTree *_otree;
    /*** Vertex ***/
    float Vx;//true vertex X
    float Vy;//true vertex Y
    float Vz;//true vertex Z
    float recoVx;//reco vertex X
    float recoVy;//reco vertex Y
    float recoVz;//reco vertex Z

    /*** true info ***/
    float truePt;//true Pt
    float trueE;//true E
    float trueTheta;//true theta
    int ccnc;
    int PDG;
    int origPDG;
    int mode;
    bool isQE;
    int intType;
    int hitNucl;
    float HadX;
    float HadY;
    float HadW;
    float qsqr;
    float OscWt; 

    //for pi0->2gamma from most energetic pi0
    int HasPi0;
    int HasP2GG;//has pi0->2gamma
    int PMPDG;//pdg code for pi0 mother
    int Ndaus;//# of pi0 daughters
    int NdausPDG[5];//pdg code for pi0 daughters
    float Epi0;
    float gam1E;
    float gam1Px;
    float gam1Py;
    float gam1Pz;
    float gam2E;
    float gam2Px;
    float gam2Py;
    float gam2Pz;

    //all pi0, first 20 if possible ?
    int Npi0;
    float pi0Px[20];
    float pi0Py[20];
    float pi0Pz[20];
    float pi0E[20];
    float pi0Vx[20];
    float pi0Vy[20];
    float pi0Vz[20];

    /*** slicer information ***/
    int Ncells;//# of associate cells
    int Nplanes;//# of extended planes
    float recoE;//reco energy
    float PlaneEnergy[140];//set the maximal plane # to 140 
    int MinPlane;//min plane index
    int PlaneNcells[140];//# of cells in each plane
    float PlaneDedx[140];//sum Dedx in each plane
    float PlaneZcell[140];

    //longitudial shower
    float E40planes;
    float Efrac_40planes;
    float Efrac_20planes;
    float Efrac_2slides;
    float Efrac_4slides;
    float Efrac_6slides;
    //transverse shower
    float EDxDy;
    float E2Dx2Dy2;
    float Efrac_2sigma;
    float Efrac_4sigma;
    float Efrac_6sigma;
    float Eiso[10];

    /*** longest discrete track ***/
/*
    float DtrackLength;//length of longest track
    int DtrackNcells;//# of cells associated to longest track
    float DtrackEnergy;//energy associated to longest track
    float DtrackNcells_ratio;//# of cells associated to longest track over Ncells
    float DtrackEnergy_ratio;//energy associated to longest track over recoE
    int NDtracks;//# of tracks
*/
    /*** longest kalman track ***/
    float KtrackLength;//length of longest track
    int KtrackNcells;//# of cells associated to longest track
    float KtrackEnergy;//energy associated to longest track
    float KtrackNcells_ratio;//# of cells associated to longest track over Ncells
    float KtrackEnergy_ratio;//energy associated to longest track over recoE
    float KtrackNplanes;
    float KtrackTheta;
    float KtrackPhi;
    float KtrackStartX;
    float KtrackStartY;
    float KtrackStartZ;
    float KtrackStopX;
    float KtrackStopY;
    float KtrackStopZ;
    int NKtracks;//# of tracks

    /*** mip information ***/
    float Nmip;//# of mip cells
    float Fmip;//# of mip cells over Ncells


    /*** cell information ***/
    int goodcells;//# of good cells
    float Xcell[300];//cell X
    float Ycell[300];//cell Y
    float Zcell[300];//cell Z
    float ecell[300];//cell energy
    float cellPECorr[300];//cell PECorr
    //int MinCell[300];
    int inXVIEW[300];//cell view
    int cellPlane[300];
    //int cellCell[300];

    /*** prong information ***/
    int Nprongs;
    float Ebalance;
    float Dphi;

    float prongEnergy[20];
    float prongPhi[20];
    float prongTheta[20];
    float prongStartX[20];
    float prongStartY[20];
    float prongStartZ[20];
    float prongLength[20];
    float prongStopX[20];
    float prongStopY[20];
    float prongStopZ[20];
    
    int prong1Ncell;
    float prong1X[300];
    float prong1Y[300];
    float prong1Z[300];
    float prong1E[300];

    int prong2Ncell;
    float prong2X[300];
    float prong2Y[300];
    float prong2Z[300];
    float prong2E[300];

    /*
    float prongPx[20];
    float prongPy[20];
    float prongPz[20];
    */
    float prongEnergyXView[20];//energy for XView cells
    float prongEnergyYView[20];//energy for YView cells
    int prongNCells[20];//# of cells
    int prongNCellsXView[20];//# of cells for XView
    int prongNCellsYView[20];//# of cells for YView
    //2D prongs
    int Nprongs2D;
    int prong2DInXView[20];//prong in X View or not
    float prongEnergy2D[20];//energy
    int prongNCells2D[20];//# of cells
    float Ebalance2D;

    /*** BDT ***/
    float BDT12;
    float rvp;

    /*** shower width ***/
    float Wshower;

    /*** JM shower ***/
/*
    float TSeg;
    float LSeg;
    float TSepi0;
    float LSepi0;
    float TSepi;
    float LSepi;
    float TSep;
    float LSep;
    float TSen;
    float LSen;
*/
    float EID;
  };
}
 
namespace ncs{
  
  //BDT reader
  TMVA::Reader *_bdt_ncana=new TMVA::Reader();
  //input variable values
  float inputvars_ncana[12]={0.};
  //input variable names
  TString inputs_ncana[12]={"_ncells","_energy","_ltrack","_ltrack_ncells_ratio",
			     "_fmip","_nmip","_efrac_20planes","_efrac_slide2",
                           "_efrac_slide6","_efrac_2sigma","_nprongs","_ebalance"};
  
  //......................................................................
  NCAna::NCAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    mf::LogInfo("NCAna")<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  NCAna::~NCAna()
  {
  }
  
  //......................................................................
  void NCAna::reconfigure(const fhicl::ParameterSet& pset)
  {
    //inputPDG = pset.get<int>("pdg"); 
    fSliceLabel   = pset.get< std::string >("SliceLabel");
    //fDTrackLabel   = pset.get< std::string >("DTrackLabel");
    fKTrackLabel   = pset.get< std::string >("KTrackLabel");
    fProngLabel   = pset.get< std::string >("ProngLabel");
    fInstLabel3D  = pset.get< std::string >("InstLabel3D");
    fInstLabel2D  = pset.get< std::string >("InstLabel2D");
    fMCFluxLabel  = pset.get< std::string >("MCFluxLabel");
    fCellHitLabel = pset.get< std::string >("CellHitLabel");
    fVertexLabel  = pset.get< std::string >("VertexLabel");
    fShowerLabel  = pset.get< std::string >("ShowerLabel");
    frvpLabel     = pset.get< std::string>("rvpLabel");
    isMC          = pset.get<int>("MCevts");
  }

  //......................................................................
  void NCAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    

    fTotalPOT=0.;

    for( int i=0; i<12; ++i ){
      _bdt_ncana->AddVariable(inputs_ncana[i],&inputvars_ncana[i]);
    }

    _bdt_ncana->BookMVA("BDT method","BDT_12vars_20121102.xml");

    //histo
    //Ereco=tfs->make<TH1F>("Ereco","Ereco;Ereco;# of events",100,0.,5.);
    //personal tuples
    _otree = tfs->make<TTree>("NCAna","NCAna particle");
    _otree->Branch("Vx",&Vx,"Vx/F");
    _otree->Branch("Vy",&Vy,"Vy/F");
    _otree->Branch("Vz",&Vz,"Vz/F");
    _otree->Branch("recoVx",&recoVx,"recoVx/F");
    _otree->Branch("recoVy",&recoVy,"recoVy/F");
    _otree->Branch("recoVz",&recoVz,"recoVz/F");

    _otree->Branch("truePt",&truePt,"truePt/F");
    _otree->Branch("trueE",&trueE,"trueE/F");
    _otree->Branch("trueTheta",&trueTheta,"trueTheta/F");
    _otree->Branch("ccnc",&ccnc,"ccnc/I");
    _otree->Branch("PDG",&PDG,"PDG/I"); 
    _otree->Branch("origPDG",&origPDG,"origPDG/I"); 
    _otree->Branch("mode",&mode,"mode/I");
    _otree->Branch("isQE",&isQE,"isQE/B");
    _otree->Branch("intType",&intType,"intType/I");
    _otree->Branch("hitNucl",&hitNucl,"hitNucl/I");
    _otree->Branch("HadX",&HadX,"HadX/F");
    _otree->Branch("HadY",&HadY,"HadY/F");
    _otree->Branch("qsqr",&qsqr,"qsqr/F");
    _otree->Branch("OscWt",&OscWt,"OscWt/F");

    _otree->Branch("HasPi0",&HasPi0,"HasPi0/I");
    _otree->Branch("Epi0",&Epi0,"Epi0/F");
    _otree->Branch("gam1E",&gam1E,"gam1E/F");
    _otree->Branch("gam1Px",&gam1Px,"gam1Px/F");
    _otree->Branch("gam1Py",&gam1Py,"gam1Py/F");
    _otree->Branch("gam1Pz",&gam1Pz,"gam1Pz/F");
    _otree->Branch("HasP2GG",&HasP2GG,"HasP2GG/I");
    _otree->Branch("PMPDG",&PMPDG,"PMPDG/I");
    _otree->Branch("Ndaus",&Ndaus,"Ndaus/I");
    _otree->Branch("NdausPDG",&NdausPDG,"NdausPDG[Ndaus]/I");

    _otree->Branch("gam2E",&gam2E,"gam2E/F");
    _otree->Branch("gam2Px",&gam2Px,"gam2Px/F");
    _otree->Branch("gam2Py",&gam2Py,"gam2Py/F");
    _otree->Branch("gam2Pz",&gam2Pz,"gam2Pz/F");

    _otree->Branch("Npi0",&Npi0,"Npi0/I");
    _otree->Branch("pi0Px",pi0Px,"pi0Px[Npi0]/F");
    _otree->Branch("pi0Py",pi0Py,"pi0Py[Npi0]/F");
    _otree->Branch("pi0Pz",pi0Pz,"pi0Pz[Npi0]/F");
    _otree->Branch("pi0E",pi0E,"pi0E[Npi0]/F");
    _otree->Branch("pi0Vx",pi0Vx,"pi0Vx[Npi0]/F");
    _otree->Branch("pi0Vy",pi0Vy,"pi0Vy[Npi0]/F");
    _otree->Branch("pi0Vz",pi0Vz,"pi0Vz[Npi0]/F");

    _otree->Branch("goodcells",&goodcells,"goodcells/I");
    _otree->Branch("Xcell",&Xcell,"Xcell[goodcells]/F");
    _otree->Branch("Ycell",&Ycell,"Ycell[goodcells]/F");
    _otree->Branch("Zcell",&Zcell,"Zcell[goodcells]/F");
    _otree->Branch("ecell",&ecell,"ecell[goodcells]/F");
    _otree->Branch("cellPECorr",&cellPECorr,"cellPECorr[goodcells]/F");
    _otree->Branch("inXVIEW",&inXVIEW,"inXVIEW[goodcells]/I");
    //_otree->Branch("MinCell",&MinCell,"MinCell[goodcells]/I");
    //_otree->Branch("cellCell",&cellCell,"cellCell[goodcells]/I");
    _otree->Branch("cellPlane",&cellPlane,"cellPlane[goodcells]/I");

/*
    _otree->Branch("NDtracks",&NDtracks,"NDtracks/I");
    _otree->Branch("DtrackLength",&DtrackLength,"DtrackLength/F");
    _otree->Branch("DtrackNcells",&DtrackNcells,"DtrackNcells/I");
    _otree->Branch("DtrackEnergy",&DtrackEnergy,"DtrackEnergy/F");
    _otree->Branch("DtrackNcells_ratio",&DtrackNcells_ratio,"DtrackNcells_ratio/F");
    _otree->Branch("DtrackEnergy_ratio",&DtrackEnergy_ratio,"DtrackEnergy_ratio/F");
*/
    _otree->Branch("NKtracks",&NKtracks,"NKtracks/I");
    _otree->Branch("KtrackLength",&KtrackLength,"KtrackLength/F");
    _otree->Branch("KtrackNcells",&KtrackNcells,"KtrackNcells/I");
    _otree->Branch("KtrackEnergy",&KtrackEnergy,"KtrackEnergy/F");
    _otree->Branch("KtrackNcells_ratio",&KtrackNcells_ratio,"KtrackNcells_ratio/F");
    _otree->Branch("KtrackEnergy_ratio",&KtrackEnergy_ratio,"KtrackEnergy_ratio/F");
    _otree->Branch("KtrackNplanes",&KtrackNplanes,"KtrackNplanes/F");
    _otree->Branch("KtrackPhi",&KtrackPhi,"KtrackPhi/F");
    _otree->Branch("KtrackTheta",&KtrackTheta,"KtrackTheta/F");
    _otree->Branch("KtrackStartX",&KtrackStartX,"KtrackStartX/F");
    _otree->Branch("KtrackStartY",&KtrackStartY,"KtrackStartY/F");
    _otree->Branch("KtrackStartZ",&KtrackStartZ,"KtrackStartZ/F");
    _otree->Branch("KtrackStopX",&KtrackStopX,"KtrackStopX/F");
    _otree->Branch("KtrackStopY",&KtrackStopY,"KtrackStopY/F");
    _otree->Branch("KtrackStopZ",&KtrackStopZ,"KtrackStopZ/F");

    _otree->Branch("Ncells",&Ncells,"Ncells/I");
    _otree->Branch("Nplanes",&Nplanes,"Nplanes/I");
    _otree->Branch("recoE",&recoE,"recoE/F");
    _otree->Branch("PlaneEnergy",&PlaneEnergy,"PlaneEnergy[Nplanes]/F");
    _otree->Branch("PlaneNcells",&PlaneNcells,"PlaneNcells[Nplanes]/I");
    _otree->Branch("PlaneDedx",&PlaneDedx,"PlaneDedx[Nplanes]/F");
    _otree->Branch("PlaneZcell",&PlaneZcell,"PlaneZcell[Nplanes]/F");
    _otree->Branch("MinPlane",&MinPlane,"MinPlane/I");
    _otree->Branch("E40planes",&E40planes,"E40planes/F");
    _otree->Branch("Efrac_20planes",&Efrac_20planes,"Efrac_20planes/F");
    _otree->Branch("Efrac_40planes",&Efrac_40planes,"Efrac_40planes/F");
    _otree->Branch("Efrac_2slides",&Efrac_2slides,"Efrac_2slides/F");
    _otree->Branch("Efrac_4slides",&Efrac_4slides,"Efrac_4slides/F");
    _otree->Branch("Efrac_6slides",&Efrac_6slides,"Efrac_6slides/F");
    _otree->Branch("Efrac_2sigma",&Efrac_2sigma,"Efrac_2sigma/F");
    _otree->Branch("Efrac_4sigma",&Efrac_4sigma,"Efrac_4sigma/F");
    _otree->Branch("Efrac_6sigma",&Efrac_6sigma,"Efrac_6sigma/F");
    _otree->Branch("EDxDy",&EDxDy,"EDxDy/F");
    _otree->Branch("E2Dx2Dy2",&E2Dx2Dy2,"E2Dx2Dy2/F");

    for( int i=1; i<11; ++i ){
      TString Bname=Form("Eiso_%i",i);
      _otree->Branch(Bname,&Eiso[i-1],Bname+"/F");
    }
    
    _otree->Branch("Nmip",&Nmip,"Nmip/F");
    _otree->Branch("Fmip",&Fmip,"Fmip/F");

    _otree->Branch("Nprongs",&Nprongs,"Nprongs/I");
    _otree->Branch("Ebalance",&Ebalance,"Ebalance/F");
    _otree->Branch("Dphi",&Dphi,"Dphi/F");
    
    _otree->Branch("prongLength",&prongLength,"prongLength[Nprongs]/F");
    _otree->Branch("prongStopX",&prongStopX,"prongStopX[Nprongs]/F");
    _otree->Branch("prongStopY",&prongStopY,"prongStopY[Nprongs]/F");
    _otree->Branch("prongStopZ",&prongStopZ,"prongStopZ[Nprongs]/F");
    _otree->Branch("prong1Ncell",&prong1Ncell,"prong1Ncell/I");
    _otree->Branch("prong1X",&prong1X,"prong1X[prong1Ncell]/F");
    _otree->Branch("prong1Y",&prong1Y,"prong1Y[prong1Ncell]/F");
    _otree->Branch("prong1Z",&prong1Z,"prong1Z[prong1Ncell]/F");
    _otree->Branch("prong1E",&prong1E,"prong1E[prong1Ncell]/F");

    _otree->Branch("prong2Ncell",&prong2Ncell,"prong2Ncell/I");
    _otree->Branch("prong2X",&prong2X,"prong2X[prong2Ncell]/F");
    _otree->Branch("prong2Y",&prong2Y,"prong2Y[prong2Ncell]/F");
    _otree->Branch("prong2Z",&prong2Z,"prong2Z[prong2Ncell]/F");
    _otree->Branch("prong2E",&prong2E,"prong2E[prong2Ncell]/F");

    _otree->Branch("prongEnergy",&prongEnergy,"prongEnergy[Nprongs]/F");
    //_otree->Branch("prongPx",&prongPx,"prongPx[Nprongs]/F");
    //_otree->Branch("prongPy",&prongPy,"prongPy[Nprongs]/F");
    //_otree->Branch("prongPz",&prongPz,"prongPz[Nprongs]/F");
    _otree->Branch("prongPhi",&prongPhi,"prongPhi[Nprongs]/F");
    _otree->Branch("prongTheta",&prongTheta,"prongTheta[Nprongs]/F");
    _otree->Branch("prongStartX",&prongStartX,"prongStartX[Nprongs]/F");
    _otree->Branch("prongStartY",&prongStartY,"prongStartY[Nprongs]/F");
    _otree->Branch("prongStartZ",&prongStartZ,"prongStartZ[Nprongs]/F");
    _otree->Branch("prongEnergyXView",&prongEnergyXView,"prongEnergyXView[Nprongs]/F");
    _otree->Branch("prongEnergyYView",&prongEnergyYView,"prongEnergyYView[Nprongs]/F");
    _otree->Branch("prongNCells",&prongNCells,"prongNCells[Nprongs]/I");
    _otree->Branch("prongNCellsXView",&prongNCellsXView,"prongNCellsXView[Nprongs]/I");
    _otree->Branch("prongNCellsYView",&prongNCellsYView,"prongNCellsYView[Nprongs]/I");
    _otree->Branch("Nprongs2D",&Nprongs2D,"Nprongs2D/I");
    _otree->Branch("prong2DInXView",&prong2DInXView,"prong2DInXView[Nprongs2D]/I");
    _otree->Branch("prongEnergy2D",&prongEnergy2D,"prongEnergy2D[Nprongs2D]/F");
    _otree->Branch("prongNCells2D",&prongNCells2D,"prongNCells2D[Nprongs2D]/I");
    _otree->Branch("Ebalance2D",&Ebalance2D,"Ebalance2D/F");
    
    _otree->Branch("BDT12",&BDT12,"BDT12/F");
    _otree->Branch("rvp",&rvp,"rvp/F");

    _otree->Branch("Wshower",&Wshower,"Wshower/F");
/*
    _otree->Branch("LSeg",&LSeg,"LSeg/F");
    _otree->Branch("TSeg",&TSeg,"TSeg/F");
    _otree->Branch("LSepi0",&LSepi0,"LSepi0/F");
    _otree->Branch("TSepi0",&TSepi0,"TSepi0/F");
    _otree->Branch("LSepi",&LSepi,"LSepi/F");
    _otree->Branch("TSepi",&TSepi,"TSepi/F");
    _otree->Branch("LSep",&LSep,"LSep/F");
    _otree->Branch("TSep",&TSep,"TSep/F");
    _otree->Branch("LSen",&LSen,"LSen/F");
    _otree->Branch("TSen",&TSen,"TSen/F");
*/
    _otree->Branch("EID",&EID,"EID/F");
  }

  //......................................................................
  void NCAna::endSubRun(const art::SubRun& sr)
  {
    if( isMC ){
      art::Handle<sumdata::POTSum> pot;
      sr.getByLabel(fMCFluxLabel, pot);
      
      fTotalPOT += pot->totgoodpot;
      
      std::cout << "POT in subrun " << sr.subRun()
		<< " = " << pot->totgoodpot << std::endl;
    }
  }

  //......................................................................
  const rb::RecoHit NCAna::MakeRecoHit(art::Ptr<rb::CellHit> const& chit,
				       double const& w)
  {
    art::ServiceHandle<calib::Calibrator> cal;
    const rb::RecoHit rhit(cal->MakeRecoHit(*chit, w));
    
    return rhit;
  }
  
  //......................................................................
  void NCAna::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel, slices);

    //get tracks, verticies and showers associated with the slices
    //art::FindManyP<rb::Track> fmDTrack(slices, evt, fDTrackLabel);
    art::FindManyP<rb::Track> fmKTrack(slices, evt, fKTrackLabel);
    art::FindManyP<rb::Vertex> fmVertex(slices, evt, fVertexLabel);
    art::FindManyP<jmshower::JMShower> fmShower(slices, evt, fShowerLabel);

    //std::cout<<evt.run()<<" "<<evt.subRun()<<" "<<evt.id().event()<<std::endl;

    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){

      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;

    //reco veretx
    //art::Handle<std::vector<rb::Vertex> > vert;
    //evt.getByLabel(fVertexLabel, vert);
    //now getting the list of verticies associated with the best slice
    //NOTE: right now it is not possible for a given slice to have more then 1
    //vertex so the size of vert should never be larger then 1,
    //so it should be safe to always use the first index
    if (!(fmVertex.isValid())) continue;
    std::vector<art::Ptr<rb::Vertex>> vert = fmVertex.at(sliceIdx);

    if (vert.size() != 1) continue;
    //std::cout<<"# of vertex: "<<vert->size()<<std::endl;
    //art::Ptr<rb::Vertex> v(vert, 0);
    //std::cout<<"Vertex Position: "<<v->GetX()<<" "<<v->GetY()<<" "<<v->GetZ()<<std::endl;

/*
    //select the longest discrete track
    std::vector<art::Ptr<rb::Track>> tracks = fmDTrack.at(sliceIdx);
    double longestTrack = 0;
    int longestTrack_ncells=0;
    double longestTrack_energy=0.;

    if( fmDTrack.isValid() ){
    for(unsigned int n = 0; n < tracks.size(); ++n){
      const double L = tracks[n]->TotalLength();
      if(L > longestTrack){
        longestTrack = L;
        longestTrack_ncells=tracks[n]->NCell();
        longestTrack_energy=tracks[n]->TotalGeV();
      }
     }
    }
*/
    //select the longest kalman track
    std::vector<art::Ptr<rb::Track>> Ktracks = fmKTrack.at(sliceIdx);
    double longestKTrack = 0;
    int longestKTrack_ncells=0;
    double longestKTrack_energy=0.;
    int longestKTrack_nplanes=0;
    float longestKTrack_phi=0.;
    float longestKTrack_theta=0.; 
    float longestKTrack_startx=0.;
    float longestKTrack_starty=0.;
    float longestKTrack_startz=0.;
    float longestKTrack_stopx=0.;
    float longestKTrack_stopy=0.;
    float longestKTrack_stopz=0.;

    //std::vector<art::Ptr<rb::Track>> tracks3d;
    //tracks3d.clear();

   if( fmKTrack.isValid() ){
    for(unsigned int n = 0; n < Ktracks.size(); ++n){
      //if( !(Ktracks[n]->Is3D()) ) continue;
      //tracks3d.push_back(Ktracks[n]);

      const double L = Ktracks[n]->TotalLength();
      if(L > longestKTrack){
        longestKTrack = L;
        longestKTrack_ncells=Ktracks[n]->NCell();
        longestKTrack_energy=Ktracks[n]->TotalGeV();
        longestKTrack_nplanes=Ktracks[n]->ExtentPlane();
        TVector3 dir = Ktracks[n]->Dir();
        longestKTrack_phi=dir.Phi();
        longestKTrack_theta=dir.Theta();
        TVector3 start = Ktracks[n]->Start();
        TVector3 stop = Ktracks[n]->Stop();
        longestKTrack_startx = start.X();
        longestKTrack_starty = start.Y();
        longestKTrack_startz = start.Z();
        longestKTrack_stopx = stop.X();
        longestKTrack_stopy = stop.Y();
        longestKTrack_stopz = stop.Z();
      }
     }
    }
    //mip cells and reco energy for slicer
    const double nCells = slice.NCell();
    const int lengthPlanes = slice.ExtentPlane();
    double recoEnergy=0.;
    double nmip=0.;
    double Ntotal=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);

       if( !rhit.IsCalibrated() ) continue;
       recoEnergy += rhit.GeV();

       if( rhit.PECorr()>100. && rhit.PECorr()<245. ) nmip++;
       Ntotal++;
    }

    //preselection cuts
    if( recoEnergy<0.1||recoEnergy>2. ) continue;
    if( nCells<10 || nCells>300 ) continue;
    if( longestKTrack>500. ) continue;
/*
    NDtracks = tracks.size();
    DtrackLength = longestTrack;
    DtrackEnergy = longestTrack_energy;
    DtrackNcells = longestTrack_ncells;
    DtrackEnergy_ratio = longestTrack_energy/recoEnergy;
    DtrackNcells_ratio = longestTrack_ncells/nCells;
*/
    NKtracks = Ktracks.size();
    KtrackLength = longestKTrack;
    KtrackEnergy = longestKTrack_energy;
    KtrackNcells = longestKTrack_ncells;
    KtrackEnergy_ratio = longestKTrack_energy/recoEnergy;
    KtrackNcells_ratio = longestKTrack_ncells/nCells;
    KtrackNplanes = longestKTrack_nplanes;
    KtrackPhi = longestKTrack_phi;
    KtrackTheta = longestKTrack_theta;
    KtrackStartX = longestKTrack_startx;
    KtrackStartY = longestKTrack_starty;
    KtrackStartZ = longestKTrack_startz;
    KtrackStopX = longestKTrack_stopx;
    KtrackStopY = longestKTrack_stopy;
    KtrackStopZ = longestKTrack_stopz;

    recoE = recoEnergy;
    Ncells = nCells;
    Nplanes = lengthPlanes;
    MinPlane = slice.MinPlane();
    Nmip=nmip;
    Fmip=nmip/Ntotal;

    recoVx=vert[0]->GetX();
    recoVy=vert[0]->GetY();
    recoVz=vert[0]->GetZ();

    //RVP
    art::FindManyP<rvp::RVP> fmRvp(slices, evt, frvpLabel);
    std::vector<art::Ptr<rvp::RVP>> rvps = fmRvp.at(sliceIdx);
    if( rvps.size()>0 ){
     rvp = rvps[0]->Value();
    }
    else rvp=-1.;

    //get MC truth info
    double vx=0.;
    double vy=0.;
    double vz=0.;

    if( isMC ){

      art::ServiceHandle<cheat::BackTracker> bt;
      art::Handle<std::vector<rb::CellHit> > chits;
      evt.getByLabel(fCellHitLabel, chits);
      art::PtrVector<rb::CellHit> allhits;
      for(size_t h = 0; h < chits->size(); ++h)
	allhits.push_back(art::Ptr<rb::CellHit>(chits, h));
      const art::Ptr<rb::Cluster> allslice(slices,sliceIdx);
      std::vector<cheat::NeutrinoEffPur> iacts = bt->SliceToNeutrinoInteractions(allslice, allhits);
      
      if( iacts.empty() ) return;
      
      // Pick the best-matching one
      art::Ptr<simb::MCTruth>& truth = iacts[0].neutrinoInt;
      // Assn interface is needlessly confusing. Have to pack stuff up like this
      art::PtrVector<simb::MCTruth> pv;
      pv.push_back(truth);
      // Get the flux associated with this truth
      //art::PtrVector<simb::MCFlux> fluxs = util::FindManyP<simb::MCFlux>(pv, evt, "geantgen", 0);
      art::FindManyP<simb::MCFlux> fmf(pv, evt, fMCFluxLabel);
      std::vector< art::Ptr<simb::MCFlux> > fluxs = fmf.at(0);
      simb::MCFlux flux;
      if(fluxs.size() == 1){
        flux = *fluxs[0];
      }
      else{
        std::cout<<"There is no assoicated neutrino intereaction !"<<std::endl;
      }
      const simb::MCNeutrino& nu = truth->GetNeutrino();
      
      trueE   = nu.Nu().E();
      truePt  = nu.Pt();
      trueTheta = nu.Theta();
      ccnc    = nu.CCNC();
      PDG     = nu.Nu().PdgCode();
      origPDG = flux.fntype;
      mode    = nu.Mode();
      isQE    = nu.InteractionType() == simb::kQE ||
	nu.InteractionType() == simb::kCCQE;
      intType = nu.InteractionType();
      hitNucl = nu.HitNuc();
      HadX    = nu.X();
      HadY    = nu.Y();
      HadW    = nu.W();
      qsqr    = nu.QSqr();
      OscWt   = SimpleOscProb(flux, nu);
      
      vx=nu.Nu().Vx();
      vy=nu.Nu().Vy();
      vz=nu.Nu().Vz();
      
      //std::cout<<"Truth info: "<<nu.CCNC()<<" "<<nu.Nu().PdgCode()<<" "<<flux.fntype<<std::endl;
      //}
      
      //true vertex
      Vx=vx;
      Vy=vy;
      Vz=vz;
      
      //select the most energetic true pi0
      int pi0Idx=-1;
      double epi0=-99.;
      const sim::ParticleNavigator& pn = bt->ParticleNavigator();
      std::vector<const sim::Particle*> AllPi0;
      AllPi0.clear();
      for(int ip = 0; ip < pn.NumberOfPrimaries(); ++ip){
	const sim::Particle* genpar = pn.Primary(ip);
	if(genpar->PdgCode() != 111) continue;

        AllPi0.push_back(pn.Primary(ip));

	if(genpar->E() > epi0){
          pi0Idx = ip;
          epi0 = genpar->E();
	}
      }//loop true particles

      if( AllPi0.size()>0 ){
         int npi0=20;
         if( AllPi0.size()<20 ) npi0=AllPi0.size(); 
         Npi0=npi0;
         for( int i=0; i<npi0; ++i ){
            pi0Px[i]=AllPi0[i]->Px();
            pi0Py[i]=AllPi0[i]->Py();
            pi0Pz[i]=AllPi0[i]->Pz();
            pi0E[i]=AllPi0[i]->E();
            pi0Vx[i]=AllPi0[i]->Vx();
            pi0Vy[i]=AllPi0[i]->Vy();
            pi0Vz[i]=AllPi0[i]->Vz();
         }
      }
      else Npi0=0;

      bool hasP2GG=false;
      if( pi0Idx>=0 ){
	const sim::Particle* selectedpi0 = pn.Primary(pi0Idx);
	HasPi0=1;
	
	int Nds=selectedpi0->NumberDaughters();
	if( Nds>5 ) Nds=5;
	Ndaus=Nds;
	if( Nds>0 ){
	  for( int i=0; i<selectedpi0->NumberDaughters(); ++i ){
	    NdausPDG[i]=pn[selectedpi0->Daughter(i)]->PdgCode();
	  }
	}//has daughters
	
	int MotherIndex=selectedpi0->Mother();
	if( MotherIndex>0 )  PMPDG=pn[MotherIndex]->PdgCode();
	else  PMPDG=-1;
	
	if(selectedpi0->NumberDaughters() == 2){
          const sim::Particle* gam1 = pn[selectedpi0->Daughter(0)];
          const sim::Particle* gam2 = pn[selectedpi0->Daughter(1)];
          if( gam1->PdgCode()==22 && gam2->PdgCode()==22 ){
            hasP2GG=true;
            HasP2GG=1;
            Epi0=selectedpi0->E();
            gam1E=gam1->E();
            gam1Px=gam1->Px();
            gam1Py=gam1->Py();
            gam1Pz=gam1->Pz();
            gam2E=gam2->E();
            gam2Px=gam2->Px();
            gam2Py=gam2->Py();
            gam2Pz=gam2->Pz();
       }//
	}//has 2 photons
      }//has pi0
      else{
	HasPi0=0;
	Ndaus=0;
	PMPDG=-1;
      }
      
      if( !hasP2GG ){
	HasP2GG=0;
	Epi0=-1.;
	gam1E=-1.;
	gam1Px=-1.;
	gam1Py=-1.;
	gam1Pz=-1.;
	gam2E=-1.;
	gam2Px=-1.;
	gam2Py=-1.;
	gam2Pz=-1.;
      }
    }//only for MC events
    
      //reco prongs
    //FuzzyK by default makes tracks.  Get these and convert to prongs
    //pull out both 2D and 3D info in case needed
    //art::Handle< std::vector<rb::Track> > shwcol3D;
    //art::Handle< std::vector<rb::Track> > shwcol2D;
    //evt.getByLabel("shower", shwcol);
    //evt.getByLabel(fProngLabel, fInstLabel3D, shwcol3D);
    //evt.getByLabel(fProngLabel, fInstLabel2D, shwcol2D);

    //get the prongs associated with the vertex. Always use
    //vertex 0 since there  should never be more then one.  If 
    //there ever is more then one vertex per slice this needs to become a loop
    //NOTE: 3D and 2D prongs will both be grabbed with FindManyP
    //need to then pull them apart into separate lists by checking view
    //art::FindManyP<rb::Track> fmProng(vert, evt, fProngLabel);
    //std::vector<art::Ptr<rb::Track>> prongs = fmProng.at(0);
    //std::vector<art::Ptr<rb::Track>> SelectedProng;

    art::FindManyP<rb::Prong> fmProng(vert, evt, fProngLabel);
    std::vector<art::Ptr<rb::Prong>> prongs = fmProng.at(0);
    std::vector<art::Ptr<rb::Prong>> SelectedProng;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
       //const rb::Prong* iprong = &(*shwcol3D)[ip];
       if( prongs[ip]->IsNoise() ) continue;
       if(prongs[ip]->View() != geo::kXorY) continue;
       //SelectedProng.push_back(&(*shwcol3D)[ip]);
       SelectedProng.push_back(prongs[ip]);
    }
    int nprongs=SelectedProng.size();
    if( nprongs>20 ) nprongs=20;
    Nprongs=nprongs;
    std::vector<double> ProngEnergy;
    std::vector<double> ProngEnergyXView;
    std::vector<double> ProngEnergyYView;
    std::vector<double> ProngPhi;
    
    double EWidth=0.;

    if( SelectedProng.size()>0 ){
    for(int ip=0; ip<nprongs; ++ip ){ 
      //reco direction
     TVector3 dir1 = SelectedProng[ip]->Dir();
     //recoEta[ip]=dir1.Eta();
     prongPhi[ip]=dir1.Phi();
     prongTheta[ip]=dir1.Theta();

     TVector3 start1 = SelectedProng[ip]->Start();
     prongStartX[ip] = start1.X();
     prongStartY[ip] = start1.Y();
     prongStartZ[ip] = start1.Z();

     prongLength[ip] = SelectedProng[ip]->TotalLength();

     TVector3 Pstop = start1 + (SelectedProng[ip]->TotalLength())*dir1;
     prongStopX[ip] = Pstop.X();
     prongStopY[ip] = Pstop.Y();
     prongStopZ[ip] = Pstop.Z();

     prongNCells[ip]=SelectedProng[ip]->NCell();
     prongNCellsXView[ip]=SelectedProng[ip]->NCell(geo::kX);
     prongNCellsYView[ip]=SelectedProng[ip]->NCell(geo::kY);

     //reco momentum
     double Eprong=0.;     
     double EprongX=0.;
     double EprongY=0.;
     for( unsigned int icell=0; icell<SelectedProng[ip]->NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = SelectedProng[ip]->Cell(icell);
       const TVector3 mean = SelectedProng[ip]->MeanXYZ();
       const double wx = (SelectedProng[ip]->NXCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->XCell(0).get()) : 0;
       const double wy = (SelectedProng[ip]->NYCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->YCell(0).get()) : 0;
       const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
       Eprong += rhit.GeV();
       if( chit->View() == geo::kX ) EprongX += rhit.GeV();
       if( chit->View() == geo::kY ) EprongY += rhit.GeV();
     }
     ProngEnergy.push_back(Eprong);
     ProngEnergyXView.push_back(EprongX);
     ProngEnergyYView.push_back(EprongY);
     ProngPhi.push_back(dir1.Phi());
     prongEnergy[ip]=Eprong;
     prongEnergyXView[ip]=EprongX;
     prongEnergyYView[ip]=EprongY;
     //prongPx[ip]=Eprong*sin(dir1.Theta())*sin(dir1.Phi());
     //prongPy[ip]=Eprong*sin(dir1.Theta())*cos(dir1.Phi());
     //prongPz[ip]=Eprong*cos(dir1.Theta());
    }
   }//has at least one prong

   double cosTheta=0.;
   if( nprongs==1 ) cosTheta=cos(SelectedProng[0]->Dir().Theta());

   double ratio=1.;
   double dphi=0.;
   int Index1=-1;
   int Index2=-1;
   double MaxEnergy1=-999.;
   double MaxEnergy2=-999.;
   if( nprongs>0 ){
      for( int i=0;i<nprongs; ++i ){
         if( MaxEnergy1<ProngEnergy[i] ){
           MaxEnergy1=ProngEnergy[i];
           Index1=i;
        }
      }//leading
    }
    if( nprongs>1 ){
      for( int i=0;i<nprongs; ++i ){
         if( i==Index1 ) continue;
         if( MaxEnergy2<ProngEnergy[i] ){
           MaxEnergy2=ProngEnergy[i];
           Index2=i;
        }
      }//sub-leading
      double Eden=ProngEnergy[Index1]+ProngEnergy[Index2];
      double Enum=ProngEnergy[Index1]-ProngEnergy[Index2];
      if( Eden>0.0 ) ratio=Enum/Eden;
      dphi=fabs(ProngPhi[Index1]-ProngPhi[Index2]);
      if( dphi>2.*TMath::Pi() ) dphi=dphi-2.*TMath::Pi();
      if( dphi>TMath::Pi() ) dphi=2.*TMath::Pi()-dphi;

      cosTheta = cos(SelectedProng[Index1]->Dir().Theta());
    }
    Ebalance=ratio;
    Dphi=dphi;

    if( Index1>=0 ){
       prong1Ncell=SelectedProng[Index1]->NCell();
       for( unsigned int icell=0; icell<SelectedProng[Index1]->NCell(); ++ icell){
        const art::Ptr<rb::CellHit>& chit = SelectedProng[Index1]->Cell(icell);
        const TVector3 mean = SelectedProng[Index1]->MeanXYZ();
        const double wx = (SelectedProng[Index1]->NXCell() > 0) ? SelectedProng[Index1]->W(SelectedProng[Index1]->XCell(0).get()) : 0;
        const double wy = (SelectedProng[Index1]->NYCell() > 0) ? SelectedProng[Index1]->W(SelectedProng[Index1]->YCell(0).get()) : 0;
        const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
        prong1X[icell] = rhit.X();
        prong1Y[icell] = rhit.Y();
        prong1Z[icell] = rhit.Z();
        prong1E[icell] = rhit.GeV();
     }
    }
    else prong1Ncell=0;

    if( Index2>=0 ){
        prong2Ncell=SelectedProng[Index2]->NCell();
        for( unsigned int icell=0; icell<SelectedProng[Index2]->NCell(); ++ icell){
        const art::Ptr<rb::CellHit>& chit = SelectedProng[Index2]->Cell(icell);
        const TVector3 mean = SelectedProng[Index2]->MeanXYZ();        const double wx = (SelectedProng[Index2]->NXCell() > 0) ? SelectedProng[Index2]->W(SelectedProng[Index2]->XCell(0).get()) : 0;
        const double wy = (SelectedProng[Index2]->NYCell() > 0) ? SelectedProng[Index2]->W(SelectedProng[Index2]->YCell(0).get()) : 0;
        const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
        prong2X[icell] = rhit.X();
        prong2Y[icell] = rhit.Y();
        prong2Z[icell] = rhit.Z();
        prong2E[icell] = rhit.GeV();
      }
    }
    else prong2Ncell=0;

    //calculate shower width
    if( nprongs>0 ){
    double ecell_sum=0.;
    double ewidth=0.;
    TVector3 PleadStart = SelectedProng[Index1]->Start();   
    TVector3 PleadStop = SelectedProng[Index1]->TotalLength()*(SelectedProng[Index1]->Dir());
    TVector3 Plead=PleadStop-PleadStart;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
         const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);

        if( !rhit.IsCalibrated() ) continue;

        TVector3 VcellStop(rhit.X(),rhit.Y(),rhit.Z());
        TVector3 Vcell=VcellStop-PleadStart;
        double dist=Vcell.Mag()*sin(Vcell.Angle(Plead));

        ecell_sum += rhit.GeV(); 
        ewidth += rhit.GeV()*dist*dist;
      }    

    if( ecell_sum>0. && SelectedProng[Index1]->TotalLength()>0. ) EWidth = ewidth/SelectedProng[Index1]->TotalLength()/ecell_sum; 
    }//has 3D prongs


    //calculate per plane cell energy
    std::vector<double> PlaneSumEnergy;
    double sum_edx=0.;//energy*dx
    double sum_e2dx2=0.;//energy*energy*dx*dx
    double sum_edy=0.;//energy*dy
    double sum_e2dy2=0.;//energy*energy*dy*dy
    double totalE_road20=0.;
    double totalE_road40=0.;
    double totalE_road60=0.;

    double Esig[10]={0.};

    for( int iplane=0; iplane<lengthPlanes; ++iplane ){
      double cellEnergy=0.;//sum of cells energy in each plane
      int planecells=0;//# of cells in each plane
      double cellX=0.;//energetic cell X in each plane
      double cellY=0.;//energetic cell Y in each plane
      //double cellMax=-99.;//energy of energetic cell
      double Ecell=0.;
      double EXcell=0.;
      double EYcell=0.;
      double DEDX=0.;

      double EZcell=0.;
      double cellZ=0.;

      std::vector<double> Xxview;
      std::vector<double> Exview;
      std::vector<double> Yyview;
      std::vector<double> Eyview;
      Xxview.clear();
      Exview.clear();
      Yyview.clear();
      Eyview.clear();


      //if( slice.NCell()>0 ){//already required there exists at least 20 cells
       for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
         const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);

        if( !rhit.IsCalibrated() ) continue;

        int Dplane=chit->Plane()-slice.MinPlane();
        if( Dplane<0 ) std::cout<<"plane # of this cell is beyond the minimal plane."<<std::endl;

        //int mincell=-1;
        //if( chit->View()==geo::kX ) mincell=slice.MinCell(geo::kX);
        //else if( chit->View()==geo::kY ) mincell=slice.MinCell(geo::kY);
        //int Dcell=chit->Cell()-mincell;

        if( Dplane==iplane ){
          cellEnergy += rhit.GeV();
          planecells++;
          //if( cellMax<rhit.GeV() ){
             //cellMax=rhit.GeV();
             //cellX=rhit.X();
            // cellY=rhit.Y();
          //}
          Ecell += rhit.GeV();
          EXcell += rhit.GeV()*rhit.X();
          EYcell += rhit.GeV()*rhit.Y();         
          EZcell += rhit.GeV()*rhit.Z();
         //std::cout<<"plane "<<iplane<<": "<<chit->Cell()<<"-"<<mincell<<" = "<<Dcell<<std::endl;

          if( chit->View() == geo::kX ){
            Xxview.push_back(rhit.X());
            Exview.push_back(rhit.GeV());
          }//in X view
          else {
            Yyview.push_back(rhit.Y());
            Eyview.push_back(rhit.GeV());
          }//in Y view

          int Iplane=chit->Plane();
          int Icell=chit->Cell();
          double dcell=2.*geom->Plane(Iplane)->Cell(Icell)->HalfD();
          double dedx=rhit.GeV()*cosTheta/dcell;
          DEDX += dedx;

         }//Dplane==iplane
       }//loop cells
      //}//has cells
     if( Ecell>0. ){
        cellX = EXcell/Ecell;
        cellY = EYcell/Ecell;
        cellZ = EZcell/Ecell;
     }
     
     PlaneSumEnergy.push_back(cellEnergy);
  
     int Nxviews=Xxview.size();
      int Nyviews=Yyview.size();
      if( Nxviews>0 ){
        for( int ixv=0; ixv<Nxviews; ++ixv ){
          if( fabs(Xxview[ixv]-cellX) < 2.*2. ) totalE_road20+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*4. ) totalE_road40+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*6. ) totalE_road60+=Exview[ixv];

            for( int iw=1; iw<11; ++iw ){
              if( fabs(Xxview[ixv]-cellX) < 2.*iw ) Esig[iw-1]+=Exview[ixv];
           }
        }
      }//in X view
      if( Nyviews>0 ){
        for( int iyv=0; iyv<Nyviews; ++iyv ){
          if( fabs(Yyview[iyv]-cellY) < 2.*2. ) totalE_road20+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*4. ) totalE_road40+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*6. ) totalE_road60+=Eyview[iyv];
            for( int iw=1; iw<11; ++iw ){
             if( fabs(Yyview[iyv]-cellY) < 2.*iw ) Esig[iw-1]+=Eyview[iyv];
          }
        }
      }//in Y view

/*   
     for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
         const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
         if( !rhit.IsCalibrated() ) continue;
         int Dplane=chit->Plane()-slice.MinPlane();
         if( Dplane==iplane ){
           double dx=fabs(rhit.X()-cellX);     
           double dy=fabs(rhit.Y()-cellY);
           if( chit->View()==geo::kX ){
            sum_edx += rhit.GeV()*dx;
            sum_e2dx2 += rhit.GeV()*rhit.GeV()*dx*dx;
            if( dx<2.*2. ) totalE_road20+=rhit.GeV();
             if( dx<2.*4. ) totalE_road40+=rhit.GeV();
             if( dx<2.*6. ) totalE_road60+=rhit.GeV();
           }//in X view
           else{
             sum_edy += rhit.GeV()*dy;
             sum_e2dy2 += rhit.GeV()*rhit.GeV()*dy*dy;
             if( dy<2.*2. ) totalE_road20+=rhit.GeV();
             if( dy<2.*4. ) totalE_road40+=rhit.GeV();
             if( dy<2.*6. ) totalE_road60+=rhit.GeV();
           }//in Y view
         }
     }//loop all cells
*/
     PlaneEnergy[iplane] = cellEnergy; 
     PlaneNcells[iplane] = planecells;
     PlaneDedx[iplane] = DEDX;
     PlaneZcell[iplane] = cellZ;
   }

    double efrac_plane2=-1.;
    double efrac_plane4=-1.;
    double efrac_plane6=-1.;
    double Eplane20=0.;
    double Eplane40=0.;
   int nplanes=PlaneSumEnergy.size();
   for( int i=0; i<nplanes; ++i ){
      if( i<20 ) Eplane20 += PlaneSumEnergy[i];
      if( i<40 ) Eplane40 += PlaneSumEnergy[i];

      //maximal energy fraction in 2 sliding window
      if( i<nplanes-2 ){
        double e_frac2=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1])/recoEnergy;
        if( efrac_plane2<e_frac2 ) efrac_plane2=e_frac2;        }
      else if( nplanes<=2 ) efrac_plane2=1.;
      //maximal energy fraction in 4 sliding window
      if( i<nplanes-4 ){
        double e_frac4=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3])/recoEnergy;
        if( efrac_plane4<e_frac4 ) efrac_plane4=e_frac4;        }
      else if( nplanes<=4 ) efrac_plane4=1.;
      //maximal energy fraction in 6 sliding window
      if( i<nplanes-6 ){
        double e_frac6=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3]+PlaneSumEnergy[i+4]+PlaneSumEnergy[i+5])/recoEnergy;
        if( efrac_plane6<e_frac6 ) efrac_plane6=e_frac6; }
      else if( nplanes<=6 ) efrac_plane6=1.;
    }

     E40planes=Eplane40;
     Efrac_20planes=Eplane20/recoEnergy;
     Efrac_40planes=Eplane40/recoEnergy;
     Efrac_2slides=efrac_plane2;
     Efrac_4slides=efrac_plane4;
     Efrac_6slides=efrac_plane6;
     Efrac_2sigma=totalE_road20/recoEnergy;
     Efrac_4sigma=totalE_road40/recoEnergy;
     Efrac_6sigma=totalE_road60/recoEnergy;

     EDxDy=sum_edx+sum_edy;
     E2Dx2Dy2=sum_e2dx2+sum_e2dy2;

     double iso[10]={0.};
     for( int iw=0; iw<10; ++iw ){
       if( Esig[iw]>0. ) iso[iw]=(recoEnergy-Esig[iw])/Esig[iw];
     }
     
     for( int iw=0; iw<10; ++iw ) Eiso[iw] = iso[iw];

    //cell information
    int ncells=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
       //if( icell<5000 ){//already require there is at most 300 cells
         ncells++;
         Xcell[icell]=rhit.X();
         Ycell[icell]=rhit.Y();
         Zcell[icell]=rhit.Z();
         ecell[icell]=rhit.GeV();
         cellPECorr[icell]=rhit.PECorr();
         cellPlane[icell]=chit->Plane();
         //cellCell[icell]=chit->Cell();

         if( chit->View()==geo::kX ){
           inXVIEW[icell]=1;
           //MinCell[icell]=slice.MinCell(geo::kX);
         }//in X view
         else {
           inXVIEW[icell]=0;
           //MinCell[icell]=slice.MinCell(geo::kY);
         }//in Y view

       //}
      //distance from readout to each cell position
      //double dX = 1549.4/2.-rhit.X();
      //double dY = 1549.2/2.-rhit.Y();
     }
    goodcells=ncells;
    
    inputvars_ncana[0]=(float)nCells;
    inputvars_ncana[1]=(float)recoEnergy;
    inputvars_ncana[2]=(float)longestKTrack;
    inputvars_ncana[3]=(float)longestKTrack_ncells/nCells;
    inputvars_ncana[4]=(float)nmip/Ntotal;
    inputvars_ncana[5]=(float)nmip; 
    inputvars_ncana[6] = (float)Eplane20/recoEnergy;
    inputvars_ncana[7] = (float)efrac_plane2;
    inputvars_ncana[8] = (float)efrac_plane6;
    inputvars_ncana[9] = (float)totalE_road20/recoEnergy;
    inputvars_ncana[10] = (float)nprongs;
    inputvars_ncana[11] = (float)ratio;

    BDT12=_bdt_ncana->EvaluateMVA("BDT method");

    Wshower = EWidth;

    //reco prong View(): 0 for XZ, 1 for YZ, 2 for 3D
    //study 2D prongs
    //std::vector<art::Ptr<rb::Track>> SelectedProng2D;
    std::vector<art::Ptr<rb::Prong>> SelectedProng2D;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
      //const rb::Prong* iprong = &(*shwcol2D)[ip];
      if( prongs[ip]->IsNoise() ) continue; 
      if( prongs[ip]->View() == geo::kXorY) continue;
      //SelectedProng2D.push_back(&(*shwcol2D)[ip]);
      SelectedProng2D.push_back(prongs[ip]);
      //std::cout<<ip<<" "<<iprong->NCell()<<std::endl;
    }     
    std::vector<int> Prong2DInXView;
    std::vector<double> ProngEnergy2D;
    int nprongs2D=SelectedProng2D.size();
    if( nprongs2D>20 ) nprongs2D=20;
    Nprongs2D = nprongs2D;
    for(int ip=0; ip<nprongs2D; ++ip ){
      prongNCells2D[ip]=SelectedProng2D[ip]->NCell();
      int prongview=SelectedProng2D[ip]->View();
      prong2DInXView[ip]=prongview; 
      Prong2DInXView.push_back(prongview);
      //reco momentum
      double Eprong=0.;
      for( unsigned int icell=0; icell<SelectedProng2D[ip]->NCell(); ++ icell){
        const art::Ptr<rb::CellHit>& chit = SelectedProng2D[ip]->Cell(icell);
        const TVector3 mean = SelectedProng2D[ip]->MeanXYZ();
        const double wx = (SelectedProng2D[ip]->NXCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->XCell(0).get()) : 0;
        const double wy = (SelectedProng2D[ip]->NYCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->YCell(0).get()) : 0;
        const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
        Eprong += rhit.GeV();
      }
      
      prongEnergy2D[ip]=Eprong;
      ProngEnergy2D.push_back(Eprong);
    }

      int index3D=-1;
      double max3D=-99.;
      for( int i=0; i<nprongs; ++i ){
        if( max3D<ProngEnergy[i] ){
          max3D=ProngEnergy[i];
          index3D=i;
        }
      }//loop 3D prongs
      int index2D=-1;
      double max2D=-99.;
      for( int i=0;i<nprongs2D; ++i ){
        if( max2D<ProngEnergy2D[i] ){
          max2D=ProngEnergy2D[i];
          index2D=i;
        }
      }//loop 2D prongs
     double ebalance2D=1.;
     if( nprongs>0 && nprongs2D>0 ){
        double Eden=0.;
        double Enum=0.;
        if( Prong2DInXView[index2D]==0 ){//XZ view
          Eden=ProngEnergy2D[index2D]+ProngEnergyXView[index3D];
          Enum=ProngEnergy2D[index2D]-ProngEnergyXView[index3D];
        }
        else {//YZ view
          Eden=ProngEnergy2D[index2D]+ProngEnergyYView[index3D];
          Enum=ProngEnergy2D[index2D]-ProngEnergyYView[index3D];
        }
        if( Eden>0. ) ebalance2D=fabs(Enum)/Eden;
     }
     Ebalance2D=ebalance2D;


    //reco jmshower
    //art::Handle< std::vector<jmshower::JMShower> > showerhandle;
    //evt.getByLabel(fShowerLabel, showerhandle);
    //grab list of showers associated with best slice
    std::vector<art::Ptr<jmshower::JMShower> >  showercol = fmShower.at(sliceIdx);
    /*for(unsigned int i = 0; i < showerhandle->size(); i++){
      art::Ptr<jmshower::JMShower> p(showerhandle, i);
      showercol.push_back(p);
    }*/

    //int bestShowerIdx=-1;//index for the best shower (highest ANN output)
    double maxANN=-1.;
    //std::cout<<"# of showers: "<<showercol.size()<<std::endl;
    for( unsigned int ish=0; ish<showercol.size(); ++ish ){
      if( maxANN<showercol[ish]->ANN(0) ){
          maxANN=showercol[ish]->ANN(0);
          //bestShowerIdx=ish;
      }
    }
   EID=maxANN;

/*
   double Leg=-1.;
   double Teg=-1.;
   double Lepi0=-1.;
   double Tepi0=-1.;
   double Lep=-1.;
   double Tep=-1.;
   double Len=-1.;
   double Ten=-1.;
   double Lepi=-1.;
   double Tepi=-1.;
   if( showercol.size()>0 ){
      //std::cout<<"Best shower index is "<<bestShowerIdx<<std::endl;
      //std::cout<<"ANN output is "<<showercol[bestShowerIdx]->ANN(0)<<std::endl;
      double ellmax = -9999.;      
      double elllmax = -9999.;
      double elltmax = -9999.;

      for(int itype = int(jmshower::kNull)+1; itype != int(jmshower::kLast); ++itype){
        if(itype == int(jmshower::kElectron) || itype == int(jmshower::kElectronQE) 
          || itype == int(jmshower::kElectronRES) || itype == int(jmshower::kElectronDIS)
          || itype == int(jmshower::kElectronCOH)){
          if(ellmax<showercol[bestShowerIdx]->DedxLongLL(itype)+showercol[bestShowerIdx]->DedxTransLL(itype)){
            ellmax = showercol[bestShowerIdx]->DedxLongLL(itype)+showercol[bestShowerIdx]->DedxTransLL(itype);
            elllmax = showercol[bestShowerIdx]->DedxLongLL(itype);
            elltmax = showercol[bestShowerIdx]->DedxTransLL(itype);
          }
        }
      }
     
      Leg = elllmax - showercol[bestShowerIdx]->DedxLongLL(int(jmshower::kGamma));
      Teg = elltmax - showercol[bestShowerIdx]->DedxTransLL(int(jmshower::kGamma));
      Lepi0 = elllmax - showercol[bestShowerIdx]->DedxLongLL(int(jmshower::kPi0));
      Tepi0 = elltmax - showercol[bestShowerIdx]->DedxTransLL(int(jmshower::kPi0));
      Lep = elllmax - showercol[bestShowerIdx]->DedxLongLL(int(jmshower::kProton));
      Tep = elltmax - showercol[bestShowerIdx]->DedxTransLL(int(jmshower::kProton));
      Len = elllmax - showercol[bestShowerIdx]->DedxLongLL(int(jmshower::kNeutron));
      Ten = elltmax - showercol[bestShowerIdx]->DedxTransLL(int(jmshower::kNeutron));
      Lepi = elllmax - showercol[bestShowerIdx]->DedxLongLL(int(jmshower::kPion));
      Tepi = elltmax - showercol[bestShowerIdx]->DedxTransLL(int(jmshower::kPion));
   }
    LSeg=Leg;
    TSeg=Teg;
    LSep=Lep;
    TSep=Tep;
    LSen=Len;
    TSen=Ten;
    LSepi=Lepi;
    TSepi=Tepi;
    LSepi0=Lepi0;
    TSepi0=Tepi0;
*/
    _otree->Fill();

   }//end loop of slice

    return;
  }//end analyze

 void NCAna::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    TH1* hPOT = tfs->make<TH1F>("hTotalPOT", ";; POT", 1, 0, 1);
    hPOT->Fill(.5, fTotalPOT);
  }
  
  double NCAna::SimpleOscProb(const simb::MCFlux& flux,
                                 const simb::MCNeutrino& nu) const
  {
    const double L        = 810.;    // km
    const double ldm      = 2.35e-3; // large delta m^2, m23
    const double sin22t13 = 0.1;     // sin^2(2theta_13)
    const double sin22t23 = 1.;      // maximal sin^2(2theta_23)
    const double ssth23 = 0.5;     // sin^2(theta_23) corresponding to above

    // Signal
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux.fntype) == 14){
      // Nue appearance
      return sin22t13*ssth23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux.fntype) == 14){
      // CC mu disappearance
      return 1-sin22t23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }

    // Background
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux.fntype) == 12){
      // Beam nue
      return 1-sin22t13*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    if(nu.CCNC() == 1 && nu.Nu().PdgCode() == flux.fntype){
      // NC
      return 1;
    }
    if(nu.CCNC() == 1 && nu.Nu().PdgCode() != flux.fntype){
      // Swapped NC. We play clever reweighting games so this is OK
      return 1;
    }
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux.fntype) == 12){
      // CC mu appearance
      return sin22t13*ssth23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }
    // Don't know what this is
    return 0;
  }
}//end namespace

namespace ncs{

  DEFINE_ART_MODULE(NCAna);

}

