///
/// \brief study RVP input and output distributions
///

#include "Calibrator/Calibrator.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"

// ROOT includes
#include "TVector3.h"
#include "TH1F.h"
#include "TMath.h"
#include "TTree.h"
#include "TProfile2D.h"

#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "SummaryData/POTSum.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"

#include "Utilities/AssociationUtil.h"
#include "MCCheater/BackTracker.h"
#include "Utilities/func/MathUtil.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>
#include <vector>

//include TMVA
#include "TMVA/Tools.h"
#include "TMVA/Reader.h"
#include "TMVA/MethodCuts.h"

class TVector3;
class TH1F;
class TTree;

namespace rb{class Cluster; class Track;}
namespace sim{class Particle;}
namespace simb{class MCFlux; class MCTruth; class MCNeutrino;}

namespace ncs {
  class RVPVal : public art::EDAnalyzer {
  public:
    explicit RVPVal(fhicl::ParameterSet const& pset);
    virtual ~RVPVal();
    
    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    void endSubRun(const art::SubRun& sr);
    void endJob(); 
    double fTotalPOT;

  private:
    const rb::RecoHit MakeRecoHit(art::Ptr<rb::CellHit> const& chit,
				  double const& w);

    //folders to get objects from
    std::string fSliceLabel;     ///label for slices
    std::string fDTrackLabel;     ///label for discrete tracks
    std::string fKTrackLabel;     ///label for kalman tracks
    std::string fProngLabel;     ///label for prong 
    std::string fInstLabel3D;    ///instance label for prongs 3D
    std::string fInstLabel2D;    ///instance label for prongs 2D
    std::string fMCFluxLabel;    ///label for generator information
    std::string fVertexLabel;    ///label for vertex
    std::string fCellHitLabel;   ///label for cell hits
    std::string frvpLabel;       ///label for rvp
    int isMC; //mode for MC


    //save event info to ntuples
    TTree *_otree;
    /*** Vertex ***/
    float recoVx;//reco vertex X
    float recoVy;//reco vertex Y
    float recoVz;//reco vertex Z

    /*** slicer information ***/
    int Ncells;//# of associate cells
    int Nplanes;//# of extended planes
    float recoE;//reco energy
    float PlaneEnergy[140];//set the maximal plane # to 140 
    int MinPlane;//min plane index
    int PlaneNcells[140];//# of cells in each plane
    float PlaneDedx[140];//sum Dedx in each plane
    float PlaneZcell[140];

    //longitudial shower
    float E40planes;
    float Efrac_40planes;
    float Efrac_20planes;
    float Efrac_2slides;
    float Efrac_4slides;
    float Efrac_6slides;
    //transverse shower
    float EDxDy;
    float E2Dx2Dy2;
    float Efrac_2sigma;
    float Efrac_4sigma;
    float Efrac_6sigma;
    float Eiso[10];


    /*** longest discrete track ***/
/*    float DtrackLength;//length of longest track
    int DtrackNcells;//# of cells associated to longest track
    float DtrackEnergy;//energy associated to longest track
    float DtrackNcells_ratio;//# of cells associated to longest track over Ncells
    float DtrackEnergy_ratio;//energy associated to longest track over recoE
    int NDtracks;//# of tracks
*/
    /*** longest kalman track ***/
    float KtrackLength;//length of longest track
    int KtrackNcells;//# of cells associated to longest track
    float KtrackEnergy;//energy associated to longest track
    float KtrackNcells_ratio;//# of cells associated to longest track over Ncells
    float KtrackEnergy_ratio;//energy associated to longest track over recoE
    int NKtracks;//# of tracks
    int KtrackNplanes;//# of extend planes
    float KtrackPhi;//phi
    float KtrackTheta;//theta
    int Ktracks_ncells[20];
    float Ktracks_energy[20];
    float Ktracks_length[20];
    float Ktracks_phi[20];
    float Ktracks_theta[20];
    int Ktracks_nplanes[20];

    /*** mip information ***/
    float Nmip;//# of mip cells
    float Fmip;//# of mip cells over Ncells

    /*** cell information ***/
    int goodcells;//# of good cells
    float Xcell[300];//cell X
    float Ycell[300];//cell Y
    float Zcell[300];//cell Z
    float ecell[300];//cell energy
    float cellPECorr[300];//cell PECorr
    //int MinCell[300];
    int inXVIEW[300];//cell view
    int cellPlane[300];
    //int cellCell[300];

    /*** prong information ***/
    int Nprongs;
    float Ebalance;
    float Dphi;

    float prongEnergy[20];
    float prongPhi[20];
    float prongTheta[20];
    /*
    float prongPx[20];
    float prongPy[20];
    float prongPz[20];
*/
    float prongEnergyXView[20];//energy for XView cells
    float prongEnergyYView[20];//energy for YView cells
    int prongNCells[20];//# of cells
    int prongNCellsXView[20];//# of cells for XView
    int prongNCellsYView[20];//# of cells for YView
    //2D prongs
    int Nprongs2D;
    int prong2DInXView[20];//prong in X View or not
    float prongEnergy2D[20];//energy
    int prongNCells2D[20];//# of cells
    float Ebalance2D;

    /*** BDT ***/
    float BDT12;
    float BDT15;
    float rvp;

   };
}
 
namespace ncs{
  
  //BDT reader
  TMVA::Reader *_bdt12_ncana=new TMVA::Reader();
  TMVA::Reader *_bdt15_ncana=new TMVA::Reader();
  //input variable values
  float input12vars_ncana[12]={0.};
  float input15vars_ncana[15]={0.};
  //input variable names
  TString inputs12_ncana[12]={ "_ncells","_energy","_ltrack",
                               "_ltrack_ncells_ratio", "_fmip","_nmip",
                               "_efrac_20planes", "_efrac_slide2", 
                       "_efrac_slide6","_efrac_2sigma","_nprongs","_ebalance"};
  
  TString inputs15_ncana[15]={ "_ncells","_energy","_ltrack",
                               "_ltrack_ncells_ratio", "_fmip","_nmip",
                               "_efrac_20planes", "_efrac_slide2", 
                       "_efrac_slide6","_efrac_2sigma","_nprongs","_ebalance",
                       "_nprongs2D", "_ebalance2D", "_eiso_3"};
  
  //......................................................................
  RVPVal::RVPVal(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    mf::LogInfo("RVPVal")<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  RVPVal::~RVPVal()
  {
  }
  
  //......................................................................
  void RVPVal::reconfigure(const fhicl::ParameterSet& pset)
  {
    //inputPDG = pset.get<int>("pdg"); 
    fSliceLabel   = pset.get< std::string >("SliceLabel");
    fDTrackLabel   = pset.get< std::string >("DTrackLabel");
    fKTrackLabel   = pset.get< std::string >("KTrackLabel");
    fProngLabel   = pset.get< std::string >("ProngLabel");
    fInstLabel3D  = pset.get< std::string >("InstLabel3D");
    fInstLabel2D  = pset.get< std::string >("InstLabel2D");
    fMCFluxLabel  = pset.get< std::string >("MCFluxLabel");
    fCellHitLabel = pset.get< std::string >("CellHitLabel");
    fVertexLabel  = pset.get< std::string >("VertexLabel");
    frvpLabel     = pset.get< std::string>("rvpLabel");
    isMC          = pset.get<int>("MCevts");
  }

  //......................................................................
  void RVPVal::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    

    fTotalPOT=0.;

    for( int i=0; i<12; ++i ){
      _bdt12_ncana->AddVariable(inputs12_ncana[i],&input12vars_ncana[i]);
    }
    _bdt12_ncana->BookMVA("BDT method","BDT_12vars_20121102.xml");

    for( int i=0; i<15; ++i ){
      _bdt15_ncana->AddVariable(inputs15_ncana[i],&input15vars_ncana[i]);
    }
    _bdt15_ncana->BookMVA("BDT method","BDT_15vars_20130408.xml");

    //personal tuples
    _otree = tfs->make<TTree>("RVPVal","RVP Val");
    _otree->Branch("recoVx",&recoVx,"recoVx/F");
    _otree->Branch("recoVy",&recoVy,"recoVy/F");
    _otree->Branch("recoVz",&recoVz,"recoVz/F");

    _otree->Branch("goodcells",&goodcells,"goodcells/I");
    _otree->Branch("Xcell",&Xcell,"Xcell[goodcells]/F");
    _otree->Branch("Ycell",&Ycell,"Ycell[goodcells]/F");
    _otree->Branch("Zcell",&Zcell,"Zcell[goodcells]/F");
    _otree->Branch("ecell",&ecell,"ecell[goodcells]/F");
    _otree->Branch("cellPECorr",&cellPECorr,"cellPECorr[goodcells]/F");
    _otree->Branch("inXVIEW",&inXVIEW,"inXVIEW[goodcells]/I");
    //_otree->Branch("MinCell",&MinCell,"MinCell[goodcells]/I");
    //_otree->Branch("cellCell",&cellCell,"cellCell[goodcells]/I");
    _otree->Branch("cellPlane",&cellPlane,"cellPlane[goodcells]/I");

/*
    _otree->Branch("NDtracks",&NDtracks,"NDtracks/I");
    _otree->Branch("DtrackLength",&DtrackLength,"DtrackLength/F");
    _otree->Branch("DtrackNcells",&DtrackNcells,"DtrackNcells/I");
    _otree->Branch("DtrackEnergy",&DtrackEnergy,"DtrackEnergy/F");
    _otree->Branch("DtrackNcells_ratio",&DtrackNcells_ratio,"DtrackNcells_ratio/F");
    _otree->Branch("DtrackEnergy_ratio",&DtrackEnergy_ratio,"DtrackEnergy_ratio/F");
*/
    _otree->Branch("NKtracks",&NKtracks,"NKtracks/I");
    _otree->Branch("KtrackLength",&KtrackLength,"KtrackLength/F");
    _otree->Branch("KtrackNcells",&KtrackNcells,"KtrackNcells/I");
    _otree->Branch("KtrackEnergy",&KtrackEnergy,"KtrackEnergy/F");
    _otree->Branch("KtrackNcells_ratio",&KtrackNcells_ratio,"KtrackNcells_ratio/F");
    _otree->Branch("KtrackEnergy_ratio",&KtrackEnergy_ratio,"KtrackEnergy_ratio/F");
    _otree->Branch("KtrackNplanes",&KtrackNplanes,"KtrackNplanes/I");
    _otree->Branch("KtrackPhi",&KtrackPhi,"KtrackPhi/F");
    _otree->Branch("KtrackTheta",&KtrackTheta,"KtrackTheta/F");

    _otree->Branch("Ktracks_length",&Ktracks_length,"Ktracks_length[NKtracks]/F");
    _otree->Branch("Ktracks_ncells",&Ktracks_ncells,"Ktracks_ncells[NKtracks]/I");
    _otree->Branch("Ktracks_energy",&Ktracks_energy,"Ktracks_energy[NKtracks]/F");
    _otree->Branch("Ktracks_nplanes",&Ktracks_nplanes,"Ktracks_nplanes[NKtracks]/I");
    _otree->Branch("Ktracks_phi",&Ktracks_phi,"Ktracks_phi[NKtracks]/F");
    _otree->Branch("Ktracks_theta",&Ktracks_theta,"Ktracks_theta[NKtracks]/F");

    _otree->Branch("Ncells",&Ncells,"Ncells/I");
    _otree->Branch("Nplanes",&Nplanes,"Nplanes/I");
    _otree->Branch("recoE",&recoE,"recoE/F");
    _otree->Branch("PlaneEnergy",&PlaneEnergy,"PlaneEnergy[Nplanes]/F");
    _otree->Branch("PlaneNcells",&PlaneNcells,"PlaneNcells[Nplanes]/I");
    _otree->Branch("PlaneDedx",&PlaneDedx,"PlaneDedx[Nplanes]/F");
    _otree->Branch("PlaneZcell",&PlaneZcell,"PlaneZcell[Nplanes]/F");
    _otree->Branch("MinPlane",&MinPlane,"MinPlane/I");
    _otree->Branch("E40planes",&E40planes,"E40planes/F");
    _otree->Branch("Efrac_20planes",&Efrac_20planes,"Efrac_20planes/F");
    _otree->Branch("Efrac_40planes",&Efrac_40planes,"Efrac_40planes/F");
    _otree->Branch("Efrac_2slides",&Efrac_2slides,"Efrac_2slides/F");
    _otree->Branch("Efrac_4slides",&Efrac_4slides,"Efrac_4slides/F");
    _otree->Branch("Efrac_6slides",&Efrac_6slides,"Efrac_6slides/F");
    _otree->Branch("Efrac_2sigma",&Efrac_2sigma,"Efrac_2sigma/F");
    _otree->Branch("Efrac_4sigma",&Efrac_4sigma,"Efrac_4sigma/F");
    _otree->Branch("Efrac_6sigma",&Efrac_6sigma,"Efrac_6sigma/F");
    _otree->Branch("EDxDy",&EDxDy,"EDxDy/F");
    _otree->Branch("E2Dx2Dy2",&E2Dx2Dy2,"E2Dx2Dy2/F");

    for( int i=1; i<11; ++i ){
      TString Bname=Form("Eiso_%i",i);
      _otree->Branch(Bname,&Eiso[i-1],Bname+"/F");
    }
    
    _otree->Branch("Nmip",&Nmip,"Nmip/F");
    _otree->Branch("Fmip",&Fmip,"Fmip/F");

    _otree->Branch("Nprongs",&Nprongs,"Nprongs/I");
    _otree->Branch("Ebalance",&Ebalance,"Ebalance/F");
    _otree->Branch("Dphi",&Dphi,"Dphi/F");
    
    _otree->Branch("prongEnergy",&prongEnergy,"prongEnergy[Nprongs]/F");
    //_otree->Branch("prongPx",&prongPx,"prongPx[Nprongs]/F");
    //_otree->Branch("prongPy",&prongPy,"prongPy[Nprongs]/F");
    //_otree->Branch("prongPz",&prongPz,"prongPz[Nprongs]/F");
    _otree->Branch("prongPhi",&prongPhi,"prongPhi[Nprongs]/F");
    _otree->Branch("prongTheta",&prongTheta,"prongTheta[Nprongs]/F");
    _otree->Branch("prongEnergyXView",&prongEnergyXView,"prongEnergyXView[Nprongs]/F");
    _otree->Branch("prongEnergyYView",&prongEnergyYView,"prongEnergyYView[Nprongs]/F");
    _otree->Branch("prongNCells",&prongNCells,"prongNCells[Nprongs]/I");
    _otree->Branch("prongNCellsXView",&prongNCellsXView,"prongNCellsXView[Nprongs]/I");
    _otree->Branch("prongNCellsYView",&prongNCellsYView,"prongNCellsYView[Nprongs]/I");
    _otree->Branch("Nprongs2D",&Nprongs2D,"Nprongs2D/I");
    _otree->Branch("prong2DInXView",&prong2DInXView,"prong2DInXView[Nprongs2D]/I");
    _otree->Branch("prongEnergy2D",&prongEnergy2D,"prongEnergy2D[Nprongs2D]/F");
    _otree->Branch("prongNCells2D",&prongNCells2D,"prongNCells2D[Nprongs2D]/I");
    _otree->Branch("Ebalance2D",&Ebalance2D,"Ebalance2D/F");
    
    _otree->Branch("BDT12",&BDT12,"BDT12/F");
    _otree->Branch("BDT15",&BDT15,"BDT15/F");
    _otree->Branch("rvp",&rvp,"rvp/F");

  }

  void RVPVal::endSubRun(const art::SubRun& sr)
  {
/*
    if( isMC ){
      art::Handle<sumdata::POTSum> pot;
      sr.getByLabel(fMCFluxLabel, pot);
      
      fTotalPOT += pot->totgoodpot;
      
      std::cout << "POT in subrun " << sr.subRun()
		<< " = " << pot->totgoodpot << std::endl;
    }*/
  }

  //......................................................................
  const rb::RecoHit RVPVal::MakeRecoHit(art::Ptr<rb::CellHit> const& chit,
					double const& w)
  {
    art::ServiceHandle<calib::Calibrator> cal;
    const rb::RecoHit rhit(cal->MakeRecoHit(*chit, w));
    
    return rhit;
  }
  
  void RVPVal::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel, slices);

    //get tracks, verticies and showers associated with the slices
    //art::FindManyP<rb::Track> fmDTrack(slices, evt, fDTrackLabel);
    art::FindManyP<rb::Track> fmKTrack(slices, evt, fKTrackLabel);
    art::FindManyP<rb::Vertex> fmVertex(slices, evt, fVertexLabel);

     if (!(fmVertex.isValid())) return;
     if( !(fmKTrack.isValid())) return;

    //std::cout<<"# of slices are "<<slices->size()<<std::endl;
    //loop over slices
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){

      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;

     

    std::vector<art::Ptr<rb::Vertex>> vert = fmVertex.at(sliceIdx);
    if (vert.size() != 1) return;

/*
    //select the longest discrete track
    double longestTrack = 0;
    int longestTrack_ncells=0;
    double longestTrack_energy=0.;

    std::vector<art::Ptr<rb::Track>> tracks = fmDTrack.at(sliceIdx);

    if( tracks.size()>0 ){
    for(unsigned int n = 0; n < tracks.size(); ++n){
      const double L = tracks[n]->TotalLength();
      if(L > longestTrack){
	longestTrack = L;
	longestTrack_ncells=tracks[n]->NCell();
	longestTrack_energy=tracks[n]->TotalGeV();
      } 
    }
   }//has 3D discrete track
*/
    //select the longest kalman track
    std::vector<art::Ptr<rb::Track>> Ktracks = fmKTrack.at(sliceIdx);
    double longestKTrack = 0;
    int longestKTrack_ncells=0;
    double longestKTrack_energy=0.;
    int longestKTrack_nplanes=0;
    float longestKTrack_phi=0.;
    float longestKTrack_theta=0.;

    for(unsigned int n = 0; n < Ktracks.size(); ++n){
      const double L = Ktracks[n]->TotalLength();
      if(L > longestKTrack){
        longestKTrack = L;
        longestKTrack_ncells=Ktracks[n]->NCell();
        longestKTrack_energy=Ktracks[n]->TotalGeV();
        longestKTrack_nplanes=Ktracks[n]->ExtentPlane();
        TVector3 dir = Ktracks[n]->Dir();
        longestKTrack_phi=dir.Phi();
        longestKTrack_theta=dir.Theta();
      }
    }

    //mip cells and reco energy for slicer
    const double nCells = slice.NCell();
    const int lengthPlanes = slice.ExtentPlane();
    double recoEnergy=0.;
    double nmip=0.;
    double Ntotal=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       
       //std::cout<<chit->Plane()<<std::endl;
       //veto on # planes
       if( chit->Plane() > 126 ) continue;
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);

       if( !rhit.IsCalibrated() ) continue;
       recoEnergy += rhit.GeV();
      
       //std::cout<<icell<<" "<<rhit.GeV()<<std::endl;

       if( rhit.PECorr()>100. && rhit.PECorr()<245. ) nmip++;
       Ntotal++;
    }

    //std::cout<<"reco info: "<<recoEnergy<<" "<<nCells<<" "<<lengthPlanes<<" "<<longestTrack<<" "<<longestKTrack<<std::endl;

    //preselection cuts
    if( recoEnergy<0.1 ) return;
    if( Ntotal>300 ) return;
  
/*  
    NDtracks = tracks.size();
    DtrackLength = longestTrack;
    DtrackEnergy = longestTrack_energy;
    DtrackNcells = longestTrack_ncells;
    DtrackEnergy_ratio = longestTrack_energy/recoEnergy;
    DtrackNcells_ratio = longestTrack_ncells/nCells;
*/
    //NKtracks = Ktracks.size();
    int ntrks=Ktracks.size();
    if( ntrks>20 ) ntrks=20;
    NKtracks = ntrks;
    for(int n = 0; n < ntrks; ++n){
        Ktracks_ncells[n] = Ktracks[n]->NCell();
        Ktracks_energy[n] = Ktracks[n]->TotalGeV();
        Ktracks_nplanes[n] = Ktracks[n]->ExtentPlane();
        Ktracks_length[n] = Ktracks[n]->TotalLength();
        TVector3 dir1 = Ktracks[n]->Dir();
        Ktracks_phi[n] = dir1.Phi();
        Ktracks_theta[n] = dir1.Theta();
     }

    KtrackLength = longestKTrack;
    KtrackEnergy = longestKTrack_energy;
    KtrackNcells = longestKTrack_ncells;
    KtrackEnergy_ratio = longestKTrack_energy/recoEnergy;
    KtrackNcells_ratio = longestKTrack_ncells/nCells;
    KtrackNplanes = longestKTrack_nplanes;
    KtrackPhi = longestKTrack_phi;
    KtrackTheta = longestKTrack_theta;

    recoE = recoEnergy;
    Ncells = nCells;
    Nplanes = lengthPlanes;
    MinPlane = slice.MinPlane();
    Nmip=nmip;
    Fmip=nmip/Ntotal;

    recoVx=vert[0]->GetX();
    recoVy=vert[0]->GetY();
    recoVz=vert[0]->GetZ();

    //RVP
/*
    art::FindManyP<rvp::RVP> fmRvp(slices, evt, frvpLabel);
    std::vector<art::Ptr<rvp::RVP>> rvps = fmRvp.at(sliceIdx);
    if( rvps.size()>0 ){
     rvp = rvps[0]->Value();
    }
    else rvp=-1.;
*/ 
    //reco prongs
    art::FindManyP<rb::Prong> fmProng(vert, evt, fProngLabel);
    std::vector<art::Ptr<rb::Prong>> prongs = fmProng.at(0);
    std::vector<art::Ptr<rb::Prong>> SelectedProng;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
       if( prongs[ip]->IsNoise() ) continue;
       if(prongs[ip]->View() != geo::kXorY) continue;
       SelectedProng.push_back(prongs[ip]);
    }
    int nprongs=SelectedProng.size();
    if( nprongs>20 ) nprongs=20;
    Nprongs=nprongs;
    std::vector<double> ProngEnergy;
    std::vector<double> ProngEnergyXView;
    std::vector<double> ProngEnergyYView;
    std::vector<double> ProngPhi;
    
    if( SelectedProng.size()>0 ){
    for(int ip=0; ip<nprongs; ++ip ){ 
      //reco direction
     TVector3 dir1 = SelectedProng[ip]->Dir();
     //recoEta[ip]=dir1.Eta();
     prongPhi[ip]=dir1.Phi();
     prongTheta[ip]=dir1.Theta();

     prongNCells[ip]=SelectedProng[ip]->NCell();
     prongNCellsXView[ip]=SelectedProng[ip]->NCell(geo::kX);
     prongNCellsYView[ip]=SelectedProng[ip]->NCell(geo::kY);

     //reco momentum
     double Eprong=0.;     
     double EprongX=0.;
     double EprongY=0.;
     for( unsigned int icell=0; icell<SelectedProng[ip]->NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = SelectedProng[ip]->Cell(icell);
       if( chit->Plane()>126 ) continue;
       const TVector3 mean = SelectedProng[ip]->MeanXYZ();
       const double wx = (SelectedProng[ip]->NXCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->XCell(0).get()) : 0;
       const double wy = (SelectedProng[ip]->NYCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->YCell(0).get()) : 0;
       const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
       Eprong += rhit.GeV();
       if( chit->View() == geo::kX ) EprongX += rhit.GeV();
       if( chit->View() == geo::kY ) EprongY += rhit.GeV();
     }
     ProngEnergy.push_back(Eprong);
     ProngEnergyXView.push_back(EprongX);
     ProngEnergyYView.push_back(EprongY);
     ProngPhi.push_back(dir1.Phi());
     prongEnergy[ip]=Eprong;
     prongEnergyXView[ip]=EprongX;
     prongEnergyYView[ip]=EprongY;
     //prongPx[ip]=Eprong*sin(dir1.Theta())*sin(dir1.Phi());
     //prongPy[ip]=Eprong*sin(dir1.Theta())*cos(dir1.Phi());
     //prongPz[ip]=Eprong*cos(dir1.Theta());
    }
   }//has at least one prong

   double cosTheta=0.;
   if( nprongs==1 ) cosTheta=cos(SelectedProng[0]->Dir().Theta());

   double ratio=1.;
   double dphi=0.;
   if( nprongs>1 ){
      int Index1=-1;
      int Index2=-1;
      double MaxEnergy1=-999.;
      double MaxEnergy2=-999.;
      for( int i=0;i<nprongs; ++i ){
         if( MaxEnergy1<ProngEnergy[i] ){
           MaxEnergy1=ProngEnergy[i];
           Index1=i;
        }
      }//leading
      for( int i=0;i<nprongs; ++i ){
         if( i==Index1 ) continue;
         if( MaxEnergy2<ProngEnergy[i] ){
           MaxEnergy2=ProngEnergy[i];
           Index2=i;
        }
      }//sub-leading
      double Eden=ProngEnergy[Index1]+ProngEnergy[Index2];
      double Enum=ProngEnergy[Index1]-ProngEnergy[Index2];
      if( Eden>0.0 ) ratio=Enum/Eden;
      dphi=fabs(ProngPhi[Index1]-ProngPhi[Index2]);
      if( dphi>2.*TMath::Pi() ) dphi=dphi-2.*TMath::Pi();
      if( dphi>TMath::Pi() ) dphi=2.*TMath::Pi()-dphi;

      cosTheta = cos(SelectedProng[Index1]->Dir().Theta());
    }
    Ebalance=ratio;
    Dphi=dphi;

    //calculate per plane cell energy
    std::vector<double> PlaneSumEnergy;
    double sum_edx=0.;//energy*dx
    double sum_e2dx2=0.;//energy*energy*dx*dx
    double sum_edy=0.;//energy*dy
    double sum_e2dy2=0.;//energy*energy*dy*dy
    double totalE_road20=0.;
    double totalE_road40=0.;
    double totalE_road60=0.;

    double Esig[10]={0.};

    for( int iplane=0; iplane<lengthPlanes; ++iplane ){
      double cellEnergy=0.;//sum of cells energy in each plane
      int planecells=0;//# of cells in each plane
      double cellX=0.;//energetic cell X in each plane
      double cellY=0.;//energetic cell Y in each plane
      //double cellMax=-99.;//energy of energetic cell
      double Ecell=0.;
      double EXcell=0.;
      double EYcell=0.;
      double DEDX=0.;

      double EZcell=0.;
      double cellZ=0.;

      std::vector<double> Xxview;
      std::vector<double> Exview;
      std::vector<double> Yyview;
      std::vector<double> Eyview;
      Xxview.clear();
      Exview.clear();
      Yyview.clear();
      Eyview.clear();


      //if( slice.NCell()>0 ){//already required there exists at least 20 cells
       for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         if( chit->Plane()>126 ) continue;
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
         const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;

        int Dplane=chit->Plane()-slice.MinPlane();
        if( Dplane<0 ) std::cout<<"plane # of this cell is beyond the minimal plane."<<std::endl;

        //int mincell=-1;
        //if( chit->View()==geo::kX ) mincell=slice.MinCell(geo::kX);
        //else if( chit->View()==geo::kY ) mincell=slice.MinCell(geo::kY);
        //int Dcell=chit->Cell()-mincell;

        if( Dplane==iplane ){
          cellEnergy += rhit.GeV();
          planecells++;
          //if( cellMax<rhit.GeV() ){
             //cellMax=rhit.GeV();
             //cellX=rhit.X();
            // cellY=rhit.Y();
          //}
          Ecell += rhit.GeV();
          EXcell += rhit.GeV()*rhit.X();
          EYcell += rhit.GeV()*rhit.Y();         
          EZcell += rhit.GeV()*rhit.Z();
         //std::cout<<"plane "<<iplane<<": "<<chit->Cell()<<"-"<<mincell<<" = "<<Dcell<<std::endl;

          if( chit->View() == geo::kX ){
            Xxview.push_back(rhit.X());
            Exview.push_back(rhit.GeV());
          }//in X view
          else {
            Yyview.push_back(rhit.Y());
            Eyview.push_back(rhit.GeV());
          }//in Y view

          int Iplane=chit->Plane();
          int Icell=chit->Cell();
          double dcell=2.*geom->Plane(Iplane)->Cell(Icell)->HalfD();
          double dedx=rhit.GeV()*cosTheta/dcell;
          DEDX += dedx;

         }//Dplane==iplane
       }//loop cells
      //}//has cells
     if( Ecell>0. ){
        cellX = EXcell/Ecell;
        cellY = EYcell/Ecell;
        cellZ = EZcell/Ecell;
     }
     
     PlaneSumEnergy.push_back(cellEnergy);
  
     int Nxviews=Xxview.size();
      int Nyviews=Yyview.size();
      if( Nxviews>0 ){
        for( int ixv=0; ixv<Nxviews; ++ixv ){
          if( fabs(Xxview[ixv]-cellX) < 2.*2. ) totalE_road20+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*4. ) totalE_road40+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*6. ) totalE_road60+=Exview[ixv];

            for( int iw=1; iw<11; ++iw ){
              if( fabs(Xxview[ixv]-cellX) < 2.*iw ) Esig[iw-1]+=Exview[ixv];
           }
        }
      }//in X view
      if( Nyviews>0 ){
        for( int iyv=0; iyv<Nyviews; ++iyv ){
          if( fabs(Yyview[iyv]-cellY) < 2.*2. ) totalE_road20+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*4. ) totalE_road40+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*6. ) totalE_road60+=Eyview[iyv];
            for( int iw=1; iw<11; ++iw ){
             if( fabs(Yyview[iyv]-cellY) < 2.*iw ) Esig[iw-1]+=Eyview[iyv];
          }
        }
      }//in Y view

     PlaneEnergy[iplane] = cellEnergy; 
     PlaneNcells[iplane] = planecells;
     PlaneDedx[iplane] = DEDX;
     PlaneZcell[iplane] = cellZ;
   }

    double efrac_plane2=-1.;
    double efrac_plane4=-1.;
    double efrac_plane6=-1.;
    double Eplane20=0.;
    double Eplane40=0.;
   int nplanes=PlaneSumEnergy.size();
   for( int i=0; i<nplanes; ++i ){
      if( i<20 ) Eplane20 += PlaneSumEnergy[i];
      if( i<40 ) Eplane40 += PlaneSumEnergy[i];

      //maximal energy fraction in 2 sliding window
      if( i<nplanes-2 ){
        double e_frac2=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1])/recoEnergy;
        if( efrac_plane2<e_frac2 ) efrac_plane2=e_frac2;        }
      else if( nplanes<=2 ) efrac_plane2=1.;
      //maximal energy fraction in 4 sliding window
      if( i<nplanes-4 ){
        double e_frac4=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3])/recoEnergy;
        if( efrac_plane4<e_frac4 ) efrac_plane4=e_frac4;        }
      else if( nplanes<=4 ) efrac_plane4=1.;
      //maximal energy fraction in 6 sliding window
      if( i<nplanes-6 ){
        double e_frac6=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3]+PlaneSumEnergy[i+4]+PlaneSumEnergy[i+5])/recoEnergy;
        if( efrac_plane6<e_frac6 ) efrac_plane6=e_frac6; }
      else if( nplanes<=6 ) efrac_plane6=1.;
    }

     E40planes=Eplane40;
     Efrac_20planes=Eplane20/recoEnergy;
     Efrac_40planes=Eplane40/recoEnergy;
     Efrac_2slides=efrac_plane2;
     Efrac_4slides=efrac_plane4;
     Efrac_6slides=efrac_plane6;
     Efrac_2sigma=totalE_road20/recoEnergy;
     Efrac_4sigma=totalE_road40/recoEnergy;
     Efrac_6sigma=totalE_road60/recoEnergy;

     EDxDy=sum_edx+sum_edy;
     E2Dx2Dy2=sum_e2dx2+sum_e2dy2;

     double iso[10]={0.};
     for( int iw=0; iw<10; ++iw ){
       if( Esig[iw]>0. ) iso[iw]=(recoEnergy-Esig[iw])/Esig[iw];
     }
     
     for( int iw=0; iw<10; ++iw ) Eiso[iw] = iso[iw];

    //cell information
    int ncells=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       if( chit->Plane()>126 ) continue;
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
         ncells++;
         Xcell[icell]=rhit.X();
         Ycell[icell]=rhit.Y();
         Zcell[icell]=rhit.Z();
         ecell[icell]=rhit.GeV();
         cellPECorr[icell]=rhit.PECorr();
         cellPlane[icell]=chit->Plane();
         //cellCell[icell]=chit->Cell();

         if( chit->View()==geo::kX ){
           inXVIEW[icell]=1;
           //MinCell[icell]=slice.MinCell(geo::kX);
         }//in X view
         else {
           inXVIEW[icell]=0;
           //MinCell[icell]=slice.MinCell(geo::kY);
         }//in Y view
     }
    goodcells=ncells;
    
    input12vars_ncana[0]=(float)nCells;
    input12vars_ncana[1]=(float)recoEnergy;
    input12vars_ncana[2]=(float)longestKTrack; //not retraining, just use kalman track
    input12vars_ncana[3]=(float)longestKTrack_ncells/nCells;
    input12vars_ncana[4]=(float)nmip/Ntotal;
    input12vars_ncana[5]=(float)nmip; 
    input12vars_ncana[6] = (float)Eplane20/recoEnergy;
    input12vars_ncana[7] = (float)efrac_plane2;
    input12vars_ncana[8] = (float)efrac_plane6;
    input12vars_ncana[9] = (float)totalE_road20/recoEnergy;
    input12vars_ncana[10] = (float)nprongs;
    input12vars_ncana[11] = (float)ratio;

    BDT12=_bdt12_ncana->EvaluateMVA("BDT method");

    for( int i=0; i<12; ++i ) input15vars_ncana[i] = input12vars_ncana[i];

    //2D prongs
    std::vector<art::Ptr<rb::Prong>> SelectedProng2D;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
      if( prongs[ip]->IsNoise() ) continue; 
      if( prongs[ip]->View() == geo::kXorY) continue;
      SelectedProng2D.push_back(prongs[ip]);
    }     
    std::vector<int> Prong2DInXView;
    std::vector<double> ProngEnergy2D;
    int nprongs2D=SelectedProng2D.size();
    if( nprongs2D>20 ) nprongs2D=20;
    Nprongs2D = nprongs2D;
    for(int ip=0; ip<nprongs2D; ++ip ){
      prongNCells2D[ip]=SelectedProng2D[ip]->NCell();
      int prongview=SelectedProng2D[ip]->View();
      prong2DInXView[ip]=prongview; 
      Prong2DInXView.push_back(prongview);
      //reco momentum
      double Eprong=0.;
      for( unsigned int icell=0; icell<SelectedProng2D[ip]->NCell(); ++ icell){
        const art::Ptr<rb::CellHit>& chit = SelectedProng2D[ip]->Cell(icell);
        if( chit->Plane()>126 ) continue;
        const TVector3 mean = SelectedProng2D[ip]->MeanXYZ();
        const double wx = (SelectedProng2D[ip]->NXCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->XCell(0).get()) : 0;
        const double wy = (SelectedProng2D[ip]->NYCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->YCell(0).get()) : 0;
        const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
        Eprong += rhit.GeV();
      }
      
      prongEnergy2D[ip]=Eprong;
      ProngEnergy2D.push_back(Eprong);
    }

      int index3D=-1;
      double max3D=-99.;
      for( int i=0; i<nprongs; ++i ){
        if( max3D<ProngEnergy[i] ){
          max3D=ProngEnergy[i];
          index3D=i;
        }
      }//loop 3D prongs
      int index2D=-1;
      double max2D=-99.;
      for( int i=0;i<nprongs2D; ++i ){
        if( max2D<ProngEnergy2D[i] ){
          max2D=ProngEnergy2D[i];
          index2D=i;
        }
      }//loop 2D prongs
     double ebalance2D=1.;
     if( nprongs>0 && nprongs2D>0 ){
        double Eden=0.;
        double Enum=0.;
        if( Prong2DInXView[index2D]==0 ){//XZ view
          Eden=ProngEnergy2D[index2D]+ProngEnergyXView[index3D];
          Enum=ProngEnergy2D[index2D]-ProngEnergyXView[index3D];
        }
        else {//YZ view
          Eden=ProngEnergy2D[index2D]+ProngEnergyYView[index3D];
          Enum=ProngEnergy2D[index2D]-ProngEnergyYView[index3D];
        }
        if( Eden>0. ) ebalance2D=fabs(Enum)/Eden;
     }
     Ebalance2D=ebalance2D;

     input15vars_ncana[12] = nprongs2D;
     input15vars_ncana[13] = ebalance2D;
     input15vars_ncana[14] = iso[2];

     BDT15=_bdt15_ncana->EvaluateMVA("BDT method");

     _otree->Fill();
    }//end of loop slice
    return;
  }//end analyze

 void RVPVal::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    TH1* hPOT = tfs->make<TH1F>("hTotalPOT", ";; POT", 1, 0, 1);
    hPOT->Fill(.5, fTotalPOT);
  }
  
}//end namespace

namespace ncs{

  DEFINE_ART_MODULE(RVPVal);

}

