#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"

// ROOT includes
#include "TVector3.h"
#include "TH1F.h"
#include "TMath.h"
#include "TTree.h"
#include "TLorentzVector.h"
#include "TH2F.h"
#include "TProfile.h"
#include "TProfile2D.h"

// NOvA includes
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/func/MathUtil.h"

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>

#ifndef SINGLE_H
#define SINGLE_H

class TVector3;
class TH1F;
class TH2F;
class TTree;
class TProfile;
class TProfile2D;

namespace ncs {
  class Single : public art::EDAnalyzer {
  public:
    explicit Single(fhicl::ParameterSet const& pset);
    virtual ~Single();

    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);

  private:
    
    const rb::RecoHit MakeRecoHit(art::Ptr<rb::CellHit> const& chit,
				  double const& w);
    
    int inputPDG;//PDG for the used single particle

    //folders to get objects from
    std::string fSliceLabel;     ///label for slices
    std::string fTrackLabel;     ///label for tracks
    std::string fProngLabel;     ///label for prong 
    std::string fInstLabel3D;    ///instance label for prongs 3D
    std::string fInstLabel2D;    ///instance label for prongs 2D
    std::string fVertexLabel;    ///label of vertex
    std::string fMCLabel;        ///label of MC

    //save even info to ntuples
    TTree *_otree;
    float Vx;//true vertex X
    float Vy;//true vertex Y
    float Vz;//true vertex Z
    float recoVx;//reco vertex X
    float recoVy;//reco vertex Y
    float recoVz;//reco vertex Z
    float truePx;//true Px
    float truePy;//true Py
    float truePz;//true Pz
    float trueE;//true E
    int nslice;//# of selected slices

    float gam1E;
    float gam1Px;
    float gam1Py;
    float gam1Pz;
    float gam2E;
    float gam2Px;
    float gam2Py;
    float gam2Pz;

    float trueEntry;//input energy of true hits
    float trueExit;//output energy of true hits
    float trueDeposit;//deposit energy of true hits

    //slice information
    int Ncells;
    int Nplanes;
    float recoE;//reco energy
    float PlaneEnergy[140];//only consider the planes<=140 events 
    int PlaneNcells[140];//# of cells in each plane
    float PlaneDedx[140];//Log(dE/dX) for each plane

    int MinPlane;
    //longest track
    float trackLength;
    int trackNcells;
    float trackEnergy;
    int Ntracks;

    //cell information
    int goodcells;//# of good cells
    float Xcell[300];//cell X
    float Ycell[300];//cell Y
    float Zcell[300];//cell Z
    float ecell[300];//cell energy
    float cellPECorr[300];//cell PECorr
    int MinCell[300];
    int inXVIEW[300];//cell energy
    int cellPlane[300];
    int cellCell[300];

    //3D prongs
    int Nprongs;
    float prongEnergy[20];//total energy
    float prongEnergyXView[20];//energy for XView cells
    float prongEnergyYView[20];//energy for YView cells
    int prongNCells[30];//# of cells
    int prongNCellsXView[30];//# of cells for XView
    int prongNCellsYView[30];//# of cells for YView
    float prongPhi[20];
    float prongTheta[20];
    float prongPx[20];
    float prongPy[20];
    float prongPz[20];

   //2D prongs
   int Nprongs2D;
   int prong2DInXView[20];//prong in X View or not
   float prongEnergy2D[20];//energy
   int prongNCells2D[20];//# of cells

  };
}
#endif

namespace ncs{

  //......................................................................
  Single::Single(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    mf::LogDebug("Single")<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  Single::~Single()
  {
  }
  
  //......................................................................
  void Single::reconfigure(const fhicl::ParameterSet& pset)
  {
    inputPDG = pset.get<int>("pdg"); 
    fSliceLabel   = pset.get< std::string >("SliceLabel");
    fTrackLabel   = pset.get< std::string >("TrackLabel");
    fProngLabel   = pset.get< std::string >("ProngLabel");
    fInstLabel3D  = pset.get< std::string >("InstLabel3D");
    fInstLabel2D  = pset.get< std::string >("InstLabel2D");
    fVertexLabel  = pset.get< std::string >("VertexLabel");
    fMCLabel   = pset.get< std::string >("MCLabel");
  }
  
  //......................................................................
  void Single::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    
    
    //personal tuples
    _otree = tfs->make<TTree>("NCAna","single particle");
    _otree->Branch("Vx",&Vx,"Vx/F");
    _otree->Branch("Vy",&Vy,"Vy/F");
    _otree->Branch("Vz",&Vz,"Vz/F");
    _otree->Branch("recoVx",&recoVx,"recoVx/F");
    _otree->Branch("recoVy",&recoVy,"recoVy/F");
    _otree->Branch("recoVz",&recoVz,"recoVz/F");
    _otree->Branch("truePx",&truePx,"truePx/F");
    _otree->Branch("truePy",&truePy,"truePy/F");
    _otree->Branch("truePz",&truePz,"truePz/F");
    _otree->Branch("trueE",&trueE,"trueE/F");
    
    _otree->Branch("trueEntry",&trueEntry,"trueEntry/F");    
    _otree->Branch("trueExit",&trueExit,"trueExit/F");    
    _otree->Branch("trueDeposit",&trueDeposit,"trueDeposit/F");    
    _otree->Branch("goodcells",&goodcells,"goodcells/I");
    _otree->Branch("Xcell",&Xcell,"Xcell[goodcells]/F");
    _otree->Branch("Ycell",&Ycell,"Ycell[goodcells]/F");
    _otree->Branch("Zcell",&Zcell,"Zcell[goodcells]/F");
    _otree->Branch("ecell",&ecell,"ecell[goodcells]/F");
    _otree->Branch("cellPECorr",&cellPECorr,"cellPECorr[goodcells]/F");
    _otree->Branch("inXVIEW",&inXVIEW,"inXVIEW[goodcells]/I");
    _otree->Branch("MinCell",&MinCell,"MinCell[goodcells]/I");
    _otree->Branch("cellCell",&cellCell,"cellCell[goodcells]/I");
    _otree->Branch("cellPlane",&cellPlane,"cellPlane[goodcells]/I");
 
    _otree->Branch("trackLength",&trackLength,"trackLength/F");
    _otree->Branch("trackNcells",&trackNcells,"trackNcells/I");
    _otree->Branch("trackEnergy",&trackEnergy,"trackEnergy/F");
    
    _otree->Branch("Ncells",&Ncells,"Ncells/I");
    _otree->Branch("Nplanes",&Nplanes,"Nplanes/I");
    _otree->Branch("recoE",&recoE,"recoE/F");
    _otree->Branch("PlaneEnergy",&PlaneEnergy,"PlaneEnergy[Nplanes]/F");
    _otree->Branch("PlaneNcells",&PlaneNcells,"PlaneNcells[Nplanes]/I");
    _otree->Branch("PlaneDedx",&PlaneDedx,"PlaneDedx[Nplanes]/F");
    _otree->Branch("MinPlane",&MinPlane,"MinPlane/I");
    _otree->Branch("Ntracks",&Ntracks,"Ntracks/I");

    _otree->Branch("Nprongs",&Nprongs,"Nprongs/I");
    _otree->Branch("prongEnergy",&prongEnergy,"prongEnergy[Nprongs]/F");
    _otree->Branch("prongEnergyXView",&prongEnergyXView,"prongEnergyXView[Nprongs]/F");
    _otree->Branch("prongEnergyYView",&prongEnergyYView,"prongEnergyYView[Nprongs]/F");
    _otree->Branch("prongNCells",&prongNCells,"prongNCells[Nprongs]/I");
    _otree->Branch("prongNCellsXView",&prongNCellsXView,"prongNCellsXView[Nprongs]/I");
    _otree->Branch("prongNCellsYView",&prongNCellsYView,"prongNCellsYView[Nprongs]/I");
    _otree->Branch("prongPx",&prongPx,"prongPx[Nprongs]/F");
    _otree->Branch("prongPy",&prongPy,"prongPy[Nprongs]/F");
    _otree->Branch("prongPz",&prongPz,"prongPz[Nprongs]/F");
    _otree->Branch("prongPhi",&prongPhi,"prongPhi[Nprongs]/F");
    _otree->Branch("prongTheta",&prongTheta,"prongTheta[Nprongs]/F");

    _otree->Branch("Nprongs2D",&Nprongs2D,"Nprongs2D/I");
    _otree->Branch("prong2DInXView",&prong2DInXView,"prong2DInXView[Nprongs2D]/I");
    _otree->Branch("prongEnergy2D",&prongEnergy2D,"prongEnergy2D[Nprongs2D]/F");
    _otree->Branch("prongNCells2D",&prongNCells2D,"prongNCells2D[Nprongs2D]/I");
 
  }

  //......................................................................
  const rb::RecoHit Single::MakeRecoHit(art::Ptr<rb::CellHit> const& chit, 
					double const& w)
  {
    art::ServiceHandle<calib::Calibrator> cal;
    const rb::RecoHit rhit(cal->MakeRecoHit(*chit, w));
    
    return rhit;
  }
  
  void Single::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<cheat::BackTracker> bt;

    //get MC true info
    sim::ParticleNavigator const& pn = bt->ParticleNavigator();
    if( pn.size()<1 ) return;
    sim::Particle* particle = pn.Particle(0);
    if( particle->PdgCode() != inputPDG ){      
      mf::LogDebug("Single")<<"The first particle PDG is not "<<inputPDG<<" !"<<std::endl;
      return;
    }
    //select single particle
    TLorentzVector truepart;
    double vx=0.;
    double vy=0.;
    double vz=0.;
    truepart.SetPxPyPzE(particle->Px(),particle->Py(),particle->Pz(),particle->E());
    vx=particle->Vx();
    vy=particle->Vy();
    vz=particle->Vz();
    
   //pick pi0->2gam
   if( particle->PdgCode()==111 ){
      if( particle->NumberDaughters()==2 ){
        const sim::Particle* gam1 = pn[particle->Daughter(0)];
        const sim::Particle* gam2 = pn[particle->Daughter(1)];
        gam1E=gam1->E();
        gam1Px=gam1->Px();
        gam1Py=gam1->Py();
        gam1Pz=gam1->Pz();
        gam2E=gam2->E();
        gam2Px=gam2->Px();
        gam2Py=gam2->Py();
        gam2Pz=gam2->Pz();
     }
    }

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel, slices);
    
    int bestSlice = -1;
    unsigned int mostCells = 0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;
      if(slice.NCell() > mostCells){        
        mostCells = slice.NCell();
        bestSlice = sliceIdx;
      }
    }
    if(bestSlice < 0) return;
    
    const rb::Cluster& slice = (*slices)[bestSlice];
    //std::cout<<"slice info: "<<slice.ExtentPlane()<<", "<<slice.NCell()<<std::endl;

    //get tracks, verticies and showers associated with the slices
    art::FindManyP<rb::Track> fmTrack(slices, evt, fTrackLabel);
    art::FindManyP<rb::Vertex> fmVertex(slices, evt, fVertexLabel);
    
    if (!(fmVertex.isValid())) return;
    std::vector<art::Ptr<rb::Vertex>> vert = fmVertex.at(bestSlice);
    if (vert.size() != 1) return;
    
    //const rb::Cluster& slice = (*slices)[bestSlice];
    
    double longestTrack = 0;
    int longestTrack_ncells=0;
    double longestTrack_energy=0.;
   
    std::vector<art::Ptr<rb::Track>> tracks = fmTrack.at(bestSlice);

    for(unsigned int n = 0; n < tracks.size(); ++n){
      const double L = tracks[n]->TotalLength();
      if(L > longestTrack){
	longestTrack = L;
	longestTrack_ncells=tracks[n]->NCell();
	longestTrack_energy=tracks[n]->TotalGeV();
      } 
    }
    
    const double nCells = slice.NCell();
    const int lengthPlanes = slice.ExtentPlane();
    
    double recoEnergy=0.;
    double Nmip=0.;
    double Ntotal=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
      const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
      const TVector3 mean = slice.MeanXYZ();
      const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
      const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
      
      const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
      if( !rhit.IsCalibrated() ) continue;
      recoEnergy += rhit.GeV();
      
      if( rhit.PECorr()>100. && rhit.PECorr()<245. ) Nmip++;
      Ntotal++;

    }//loop slice assoicated cells
    
    //preselection cuts
    if( recoEnergy<0.1 ) return;
    //if( nCells<20 || nCells>300 ) return;
    //if( lengthPlanes>140 ) return;
    //if( longestTrack>400. ) return;  
    
    recoVx=vert[0]->GetX();
    recoVy=vert[0]->GetY();
    recoVz=vert[0]->GetZ();

    Ntracks = tracks.size();
    trackLength = longestTrack;
    trackEnergy = longestTrack_energy;
    trackNcells = longestTrack_ncells;
    recoE = recoEnergy;
    Ncells = nCells;
    Nplanes = lengthPlanes;
    MinPlane = slice.MinPlane();
    
    //reco prongs
    //art::FindManyP<rb::Track> fmProng(vert, evt, fProngLabel);
    //std::vector<art::Ptr<rb::Track>> prongs = fmProng.at(0);
    //std::vector<art::Ptr<rb::Track>> SelectedProng;
    art::FindManyP<rb::Prong> fmProng(vert, evt, fProngLabel);
    std::vector<art::Ptr<rb::Prong>> prongs = fmProng.at(0);
    std::vector<art::Ptr<rb::Prong>> SelectedProng;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
      if( prongs[ip]->IsNoise() ) continue;
      if(prongs[ip]->View() != geo::kXorY) continue;
      
      SelectedProng.push_back(prongs[ip]);
    }
    //mf::LogDebug("Single")<<"# of prongs: "<<shwcol->size()<<std::endl;
    int nprongs=SelectedProng.size();
    if( nprongs>20 ) nprongs=20;
    Nprongs=nprongs;
    //double Ebalance=1.;
    //if( SelectedProng.size()>0 ){
    if( SelectedProng.size()<1 ) return;
    
    // Ebalance is not used, only set below, so comment out
    //double Ebalance=1.;

    for(int ip=0; ip<nprongs; ++ip ){ 
      //reco direction
      TVector3 dir1 = SelectedProng[ip]->Dir();
      //recoEta[ip]=dir1.Eta();
      prongPhi[ip]=dir1.Phi();
      prongTheta[ip]=dir1.Theta();
      
      prongNCells[ip]=SelectedProng[ip]->NCell();
      prongNCellsXView[ip]=SelectedProng[ip]->NCell(geo::kX);
      prongNCellsYView[ip]=SelectedProng[ip]->NCell(geo::kY);
      
      //reco momentum
      double Eprong=0.;     
      double EprongX=0.;     
      double EprongY=0.;     
      for( unsigned int icell=0; icell<SelectedProng[ip]->NCell(); ++ icell){
	const art::Ptr<rb::CellHit>& chit = SelectedProng[ip]->Cell(icell);
	const TVector3 mean = SelectedProng[ip]->MeanXYZ();
	const double wx = (SelectedProng[ip]->NXCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->XCell(0).get()) : 0;
	const double wy = (SelectedProng[ip]->NYCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->YCell(0).get()) : 0;
	const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
	if( !rhit.IsCalibrated() ) continue;
	Eprong += rhit.GeV();
	if( chit->View() == geo::kX ) EprongX += rhit.GeV();
	if( chit->View() == geo::kY ) EprongY += rhit.GeV();
      }
      
      prongEnergy[ip]=Eprong;
      prongPx[ip]=Eprong*sin(dir1.Theta())*sin(dir1.Phi());
      prongPy[ip]=Eprong*sin(dir1.Theta())*cos(dir1.Phi());
      prongPz[ip]=Eprong*cos(dir1.Theta());
      prongEnergyXView[ip]=EprongX;
      prongEnergyYView[ip]=EprongY;
      
    }//loop all prongs
    
/*
    int Index1=-1;
    //int Index2=-1;
    if( nprongs>1 ){
      
      double MaxEnergy1=-999.;
      double MaxEnergy2=-999.;
      for( int i=0;i<nprongs; ++i ){
	if( MaxEnergy1<prongEnergy[i] ){
	  MaxEnergy1=prongEnergy[i];
	  Index1=i;
	}
      }//leading
      for( int i=0;i<nprongs; ++i ){
	if( i==Index1 ) continue;
	if( MaxEnergy2<prongEnergy[i] ){
	  MaxEnergy2=prongEnergy[i];
	  //Index2=i;
	}
      }//sub-leading
      //double Eden=prongEnergy[Index1]+prongEnergy[Index2];
      //double Enum=prongEnergy[Index1]-prongEnergy[Index2];
      //if( Eden>0.0 ) Ebalance=Enum/Eden;
      
    }//has at least two prongs 
*/
    //double cosTheta=cos((SelectedProng[0]->Dir()).Theta());
    //if( nprongs>1 ) cosTheta=cos((SelectedProng[Index1]->Dir()).Theta());
    
    //calculate per plane cell energy
    std::vector<double> PlaneSumEnergy;
    double totalE_road20=0.;
    double totalE_road40=0.;
    double totalE_road60=0.;

    for( int iplane=0; iplane<lengthPlanes; ++iplane ){
      double cellEnergy=0.;//sum of cells energy in each plane
      int planecells=0;//# of cells in each plane
      double cellX=0.;
      double cellY=0.;
      double Ecell=0.;
      double EXcell=0.;
      double EYcell=0.;

      std::vector<double> Xxview;
      std::vector<double> Exview;
      std::vector<double> Yyview;
      std::vector<double> Eyview;
      Xxview.clear();
      Exview.clear();
      Yyview.clear();
      Eyview.clear();

      double DEDX=0.;

      for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
	const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
	const TVector3 mean = slice.MeanXYZ();
	const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
	const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
	const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
	
        int Dplane=chit->Plane()-slice.MinPlane();
        if( Dplane<0 ) mf::LogDebug("Single")<<"plane # of this cell is beyond the minimal plane."<<std::endl;
	
        //int mincell=-1;
        //if( chit->View()==geo::kX ) mincell=slice.MinCell(geo::kX);
        //else if( chit->View()==geo::kY ) mincell=slice.MinCell(geo::kY);
        //int Dcell=chit->Cell()-mincell;
        
        if( Dplane==iplane ){
          cellEnergy += rhit.GeV();
          planecells++;
	  
          Ecell += rhit.GeV();
          EXcell += rhit.GeV()*rhit.X();
          EYcell += rhit.GeV()*rhit.Y();
	  
          if( chit->View() == geo::kX ){
	    Xxview.push_back(rhit.X());
	    Exview.push_back(rhit.GeV());
          }//in X view
          else {
	    Yyview.push_back(rhit.Y());
	    Eyview.push_back(rhit.GeV());
          }//in Y view
	}//Dplane==iplane
      }//loop cells
      //}//has cells

      if( Ecell>0. ){
	cellX = EXcell/Ecell;
	cellY = EYcell/Ecell;
      }
      PlaneSumEnergy.push_back(cellEnergy);
      
      int Nxviews=Xxview.size();
      int Nyviews=Yyview.size();
      if( Nxviews>0 ){
	for( int ixv=0; ixv<Nxviews; ++ixv ){
	  if( fabs(Xxview[ixv]-cellX) < 2.*2. ) totalE_road20+=Exview[ixv];
	  if( fabs(Xxview[ixv]-cellX) < 2.*4. ) totalE_road40+=Exview[ixv];
	  if( fabs(Xxview[ixv]-cellX) < 2.*6. ) totalE_road60+=Exview[ixv];
	}
      }//in X view
      if( Nyviews>0 ){
	for( int iyv=0; iyv<Nyviews; ++iyv ){
	  if( fabs(Yyview[iyv]-cellY) < 2.*2. ) totalE_road20+=Eyview[iyv];
	  if( fabs(Yyview[iyv]-cellY) < 2.*4. ) totalE_road40+=Eyview[iyv];
	  if( fabs(Yyview[iyv]-cellY) < 2.*6. ) totalE_road60+=Eyview[iyv];
	}
      }//in Y view
      
      
      PlaneEnergy[iplane] = cellEnergy; 
      PlaneNcells[iplane] = planecells;
      PlaneDedx[iplane] = DEDX;
    }//loop planes
    
    double efrac_plane2=-1.;
    double efrac_plane4=-1.;
    double efrac_plane6=-1.;
    double Eplane20=0.;
    double Eplane40=0.;
    int nplanes=PlaneSumEnergy.size();
    for( int i=0; i<nplanes; ++i ){
      if( i<20 ) Eplane20 += PlaneSumEnergy[i];
      if( i<40 ) Eplane40 += PlaneSumEnergy[i];
      
      //maximal energy fraction in 2 sliding window
      if( i<nplanes-2 ){
        double e_frac2=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1])/recoEnergy;
        if( efrac_plane2<e_frac2 ) efrac_plane2=e_frac2;        }
      else if( nplanes<=2 ) efrac_plane2=1.;
      //maximal energy fraction in 4 sliding window
      if( i<nplanes-4 ){
        double e_frac4=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3])/recoEnergy;
        if( efrac_plane4<e_frac4 ) efrac_plane4=e_frac4;        }
      else if( nplanes<=4 ) efrac_plane4=1.;
      //maximal energy fraction in 6 sliding window
      if( i<nplanes-6 ){
        double e_frac6=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3]+PlaneSumEnergy[i+4]+PlaneSumEnergy[i+5])/recoEnergy;
        if( efrac_plane6<e_frac6 ) efrac_plane6=e_frac6; }
      else if( nplanes<=6 ) efrac_plane6=1.;
    }
    //select the true hits
    art::Handle< std::vector<sim::FLSHitList> > flslistHandle;
    evt.getByLabel(fMCLabel, flslistHandle);
    double eEntry=0.;
    double eExit=0.;
    double eDeposit=0.;
    if( flslistHandle.failedToGet() ) return;
    //for( size_t l = 0; l < flslistHandle->size(); ++l){
    const std::vector<sim::FLSHit> &hits = flslistHandle->at(0).fHits;
    for( size_t f = 0; f < hits.size(); ++f){
      //if( hits[f].GetPDG() != 22 ) continue;
      eEntry += hits[f].GetEntryEnergy();
      eExit += hits[f].GetExitEnergy();
      eDeposit += hits[f].GetEdep();
    }
    trueEntry = eEntry;
    trueExit = eExit;
    trueDeposit = eDeposit;
    
    //true vertex
    Vx=vx;
    Vy=vy;
    Vz=vz;
    //true momentum
    truePx=truepart.Px();
    truePy=truepart.Py();
    truePz=truepart.Pz();
    trueE=truepart.E();
    
    int ncells=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
      const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
      const TVector3 mean = slice.MeanXYZ();
      const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
      const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
      
      const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
      if( !rhit.IsCalibrated() ) continue;
      //if( icell<5000 ){//already require there is at most 300 cells
      ncells++;
      Xcell[icell]=rhit.X();
      Ycell[icell]=rhit.Y();
      Zcell[icell]=rhit.Z();
      ecell[icell]=rhit.GeV();
      cellPECorr[icell]=rhit.PECorr();
      cellPlane[icell]=chit->Plane();
      cellCell[icell]=chit->Cell();
      
      if( chit->View()==geo::kX ){
	inXVIEW[icell]=1;
	MinCell[icell]=slice.MinCell(geo::kX);
      }//in X view
      else {
	inXVIEW[icell]=0;
	MinCell[icell]=slice.MinCell(geo::kY);
      }//in Y view
      
       //}
      //distance from readout to each cell position
      //double dX = 1549.4/2.-rhit.X();
      //double dY = 1549.2/2.-rhit.Y();
      
    }
    goodcells=ncells;
    
    //reco prong View(): 0 for XZ, 1 for YZ, 2 for 3D
    //study 2D prongs
    //std::vector<art::Ptr<rb::Track>> SelectedProng2D;
    std::vector<art::Ptr<rb::Prong>> SelectedProng2D;
    
    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
      if( prongs[ip]->IsNoise() ) continue;
      if( prongs[ip]->View() == geo::kXorY) continue;
      SelectedProng2D.push_back(prongs[ip]);
    }
    //mf::LogDebug("Single")<<"# of prongs: "<<SelectedProng.size()<<" "<<SelectedProng2D.size()<<std::endl;
    int nprongs2D=SelectedProng2D.size();
    if( nprongs2D>20 ) nprongs2D=20;
    Nprongs2D = nprongs2D;
    for(int ip=0; ip<nprongs2D; ++ip ){
      prongNCells2D[ip]=SelectedProng2D[ip]->NCell();
      int prongview=SelectedProng2D[ip]->View();
      prong2DInXView[ip]=prongview; 
      
      //reco momentum
      double Eprong=0.; 
      for( unsigned int icell=0; icell<SelectedProng2D[ip]->NCell(); ++ icell){
        const art::Ptr<rb::CellHit>& chit = SelectedProng2D[ip]->Cell(icell);
        const TVector3 mean = SelectedProng2D[ip]->MeanXYZ();
        const double wx = (SelectedProng2D[ip]->NXCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->XCell(0).get()) : 0;
        const double wy = (SelectedProng2D[ip]->NYCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->YCell(0).get()) : 0;
        const rb::RecoHit rhit = this->MakeRecoHit(chit, chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
        Eprong += rhit.GeV();
      }
      
      prongEnergy2D[ip]=Eprong;
    }
    
    _otree->Fill();
    
    return;
  }//end analyze
  
}//end namespace


namespace ncs{

  DEFINE_ART_MODULE(Single);

}

