///
/// \brief 
/// \author Xuebing Bu - xbbu@fnal.gov
///
#include "Calibrator/Calibrator.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"

// ROOT includes
#include "TVector3.h"
#include "TH1F.h"
#include "TMath.h"
#include "TTree.h"
#include "TProfile2D.h"

#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/GTruth.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "SummaryData/POTSum.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"

#include "Utilities/AssociationUtil.h"
#include "MCCheater/BackTracker.h"
#include "Utilities/func/MathUtil.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>
#include <vector>
#include <algorithm>

//include TMVA
#include "TMVA/Tools.h"
#include "TMVA/Reader.h"
#include "TMVA/MethodCuts.h"

class TVector3;
class TH1F;
class TTree;

namespace rb{class Cluster; class Track;}
namespace sim{class Particle;}
namespace simb{class MCFlux; class MCTruth; class MCNeutrino;}

namespace ncs {
  class Xbeam : public art::EDAnalyzer {
  public:
    explicit Xbeam(fhicl::ParameterSet const& pset);
    virtual ~Xbeam();
    
    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    void endSubRun(const art::SubRun& sr);
    void endJob(); 
    double SimpleOscProb(const simb::MCFlux& flux,
                         const simb::MCNeutrino& nu) const;
    double fTotalPOT;
    bool isTrueEnergyGreater(const sim::Particle* p1, const sim::Particle* p2);

  private:
    //folders to get objects from
    std::string fSliceLabel;     ///label for slices
    std::string fKTrackLabel;     ///label for kalman tracks
    std::string fProngLabel;     ///label for prong 
    std::string fInstLabel3D;    ///instance label for prongs 3D
    std::string fInstLabel2D;    ///instance label for prongs 2D
    std::string fMCFluxLabel;    ///label for generator information
    std::string fVertexLabel;    ///label for vertex
    std::string fCellHitLabel;   ///label for cell hits
    int isMC; //mode for MC

    //save event info to ntuples
    TTree *_otree;
    /*** reco Vertex ***/
    float recoVx;//reco vertex X
    float recoVy;//reco vertex Y
    float recoVz;//reco vertex Z

    /*** true info ***/
/*
    int hasTrueInfo;
    int Nnu;
    float nu_Vx[100];//true vertex X
    float nu_Vy[100];//true vertex Y
    float nu_Vz[100];//true vertex Z
    float nu_truePt[100];//true Pt
    float nu_trueE[100];//true E
    float nu_trueTheta[100];//true theta
    float nu_truePhi[100];//true phi
    int nu_ccnc[100];
    int nu_PDG[100];
    int nu_origPDG[100];
    int nu_mode[100];
    bool nu_isQE[100];
    int nu_intType[100];
    int nu_hitNucl[100];
    float nu_HadX[100];
    float nu_HadY[100];
    float nu_HadW[100];
    float nu_qsqr[100];
    float nu_OscWt[100]; 
    int meson_PDG[100];
    float meson_Px[100];
    float meson_Py[100];
    float meson_Pz[100];
    int lep_PDG[100];
    float lep_Px[100];
    float lep_Py[100];
    float lep_Pz[100];
    float lep_E[100];
*/
    int hasTrueInfo;
    int nu_origPDG;
    int mesonPDG;
    float mesonPx;
    float mesonPy;
    float mesonPz;

    /*** true info for slice matched nu ***/
    int SlNeuMatched;
    int hasFlux;
    float sl_truePt;//true Pt
    float sl_truePx;//true Px
    float sl_truePy;//true Py
    float sl_truePz;//true Pz
    float sl_trueE;//true E
    float sl_trueTheta;//true theta
    int sl_ccnc;
    int sl_PDG;
    int sl_origPDG;
    int sl_mode;
    bool sl_isQE;
    int sl_intType;
    int sl_hitNucl;
    float sl_HadX;
    float sl_HadY;
    float sl_HadW;
    float sl_qsqr;
    float sl_Vx;
    float sl_Vy;
    float sl_Vz;
    int sl_lepPDG; 
    float sl_lepE;
    float sl_lepPx;
    float sl_lepPy;
    float sl_lepPz;

    /*** pi0->2gamma ***/
    int Npi0;
    float pi0E[50];
    float pi0Vx[50];
    float pi0Vy[50];
    float pi0Vz[50];
    float gam1E[50];
    float gam1Px[50];
    float gam1Py[50];
    float gam1Pz[50];
    float gam2E[50];
    float gam2Px[50];
    float gam2Py[50];
    float gam2Pz[50];

    /*** event count ***/
    int hasSlice;
    int hasVertex;
    int hasOneVertex;
    int Nvertex;
    int passPre;

    /*** slicer information ***/
    int Ncells;//# of associate cells
    int Nplanes;//# of extended planes
    float recoE;//reco energy
    float PlaneEnergy[140];//set the maximal plane # to 140 
    int MinPlane;//min plane index
    int PlaneNcells[140];//# of cells in each plane
    float PlaneDedx[140];//sum Dedx in each plane
    float PlaneZcell[140];

    //longitudial shower
    float E40planes;
    float Efrac_40planes;
    float Efrac_20planes;
    float Efrac_2slides;
    float Efrac_4slides;
    float Efrac_6slides;
    //transverse shower
    float Efrac_2sigma;
    float Efrac_4sigma;
    float Efrac_6sigma;
    float Eiso[10];

    /*** longest kalman track ***/
    float KtrackLength;//length of longest track
    int KtrackNcells;//# of cells associated to longest track
    float KtrackEnergy;//energy associated to longest track
    float KtrackNcells_ratio;//# of cells associated to longest track over Ncells
    float KtrackEnergy_ratio;//energy associated to longest track over recoE
    int NKtracks;//# of tracks
    int KtrackNplanes;
    float KtrackTheta;
    float KtrackPhi;
    float KtrackStartX;
    float KtrackStartY;
    float KtrackStartZ;
    float KtrackStopX;
    float KtrackStopY;
    float KtrackStopZ;

    int NKtracks3D;//# of 3D tracks
    float trackEnergy[20];
    float trackLength[20];
    float trackPhi[20];
    float trackTheta[20];
    float trackEnergyXView[20];//energy for XView cells
    float trackEnergyYView[20];//energy for YView cells
    int trackNCells[20];//# of cells
    int trackNCellsXView[20];//# of cells for XView
    int trackNCellsYView[20];//# of cells for YView

    /*** mip information ***/
    float Nmip;//# of mip cells
    float Fmip;//# of mip cells over Ncells

    /*** cell information ***/
    int goodcells;//# of good cells
    float Xcell[300];//cell X
    float Ycell[300];//cell Y
    float Zcell[300];//cell Z
    float ecell[300];//cell energy
    float cellPECorr[300];//cell PECorr
    //int MinCell[300];
    int inXVIEW[300];//cell view
    int cellPlane[300];
    //int cellCell[300];

    /*** prong information ***/
    int Nprongs;
    float Ebalance;
    float Dphi;

    float prongEnergy[20];
    float prongPhi[20];
    float prongTheta[20];
/*
    float prongPx[20];
    float prongPy[20];
    float prongPz[20];
*/
    float prongEnergyXView[20];//energy for XView cells
    float prongEnergyYView[20];//energy for YView cells
    int prongNCells[20];//# of cells
    int prongNCellsXView[20];//# of cells for XView
    int prongNCellsYView[20];//# of cells for YView
    //2D prongs
    int Nprongs2D;
    int prong2DInXView[20];//prong in X View or not
    float prongEnergy2D[20];//energy
    int prongNCells2D[20];//# of cells
    float Ebalance2D;

  };
}
 
namespace ncs{
  
  //......................................................................
  Xbeam::Xbeam(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    mf::LogInfo("Xbeam")<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  Xbeam::~Xbeam()
  {
  }
  
  //......................................................................
  void Xbeam::reconfigure(const fhicl::ParameterSet& pset)
  {
    //inputPDG = pset.get<int>("pdg"); 
    fSliceLabel   = pset.get< std::string >("SliceLabel");
    fKTrackLabel   = pset.get< std::string >("KTrackLabel");
    fProngLabel   = pset.get< std::string >("ProngLabel");
    fInstLabel3D  = pset.get< std::string >("InstLabel3D");
    fInstLabel2D  = pset.get< std::string >("InstLabel2D");
    fMCFluxLabel  = pset.get< std::string >("MCFluxLabel");
    fCellHitLabel = pset.get< std::string >("CellHitLabel");
    fVertexLabel  = pset.get< std::string >("VertexLabel");
    isMC          = pset.get<int>("MCevts");
  }

  //......................................................................
  void Xbeam::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    

    fTotalPOT=0.;


    //histo
    //Ereco=tfs->make<TH1F>("Ereco","Ereco;Ereco;# of events",100,0.,5.);
    //personal tuples
    _otree = tfs->make<TTree>("Xbeam","Xbeam particle");
    _otree->Branch("recoVx",&recoVx,"recoVx/F");
    _otree->Branch("recoVy",&recoVy,"recoVy/F");
    _otree->Branch("recoVz",&recoVz,"recoVz/F");
/*
    _otree->Branch("hasTrueInfo",&hasTrueInfo,"hasTrueInfo/I");
    _otree->Branch("Nnu",&Nnu,"Nnu/I");
    _otree->Branch("nu_Vx",&nu_Vx,"nu_Vx[Nnu]/F");
    _otree->Branch("nu_Vy",&nu_Vy,"nu_Vy[Nnu]/F");
    _otree->Branch("nu_Vz",&nu_Vz,"nu_Vz[Nnu]/F");
    _otree->Branch("nu_truePt",&nu_truePt,"nu_truePt[Nnu]/F");
    _otree->Branch("nu_trueE",&nu_trueE,"nu_trueE[Nnu]/F");
    _otree->Branch("nu_trueTheta",&nu_trueTheta,"nu_trueTheta[Nnu]/F");
    _otree->Branch("nu_truePhi",&nu_truePhi,"nu_truePhi[Nnu]/F");
    _otree->Branch("nu_ccnc",&nu_ccnc,"nu_ccnc[Nnu]/I");
    _otree->Branch("nu_PDG",&nu_PDG,"nu_PDG[Nnu]/I"); 
    _otree->Branch("nu_origPDG",&nu_origPDG,"nu_origPDG[Nnu]/I"); 
    _otree->Branch("nu_mode",&nu_mode,"nu_mode[Nnu]/I");
    _otree->Branch("nu_isQE",&nu_isQE,"nu_isQE[Nnu]/B");
    _otree->Branch("nu_intType",&nu_intType,"nu_intType[Nnu]/I");
    _otree->Branch("nu_hitNucl",&nu_hitNucl,"nu_hitNucl[Nnu]/I");
    _otree->Branch("nu_HadX",&nu_HadX,"nu_HadX[Nnu]/F");
    _otree->Branch("nu_HadY",&nu_HadY,"nu_HadY[Nnu]/F");
    _otree->Branch("nu_HadW",&nu_HadW,"nu_HadW[Nnu]/F");
    _otree->Branch("nu_qsqr",&nu_qsqr,"nu_qsqr[Nnu]/F");
    _otree->Branch("nu_OscWt",&nu_OscWt,"nu_OscWt[Nnu]/F");
    _otree->Branch("meson_PDG",&meson_PDG,"meson_PDG[Nnu]/I");
    _otree->Branch("meson_Px",&meson_Px,"meson_Px[Nnu]/F");
    _otree->Branch("meson_Py",&meson_Py,"meson_Py[Nnu]/F");
    _otree->Branch("meson_Pz",&meson_Pz,"meson_Pz[Nnu]/F");
    _otree->Branch("lep_PDG",&lep_PDG,"lep_PDG[Nnu]/I");
    _otree->Branch("lep_Px",&lep_Px,"lep_Px[Nnu]/F");
    _otree->Branch("lep_Py",&lep_Py,"lep_Py[Nnu]/F");
    _otree->Branch("lep_Pz",&lep_Pz,"lep_Pz[Nnu]/F");
    _otree->Branch("lep_E",&lep_E,"lep_E[Nnu]/F");
*/

    _otree->Branch("hasTrueInfo",&hasTrueInfo,"hasTrueInfo/I");
    _otree->Branch("nu_origPDG",&nu_origPDG,"nu_origPDG/I"); 
    _otree->Branch("mesonPDG",&mesonPDG,"mesonPDG/I");
    _otree->Branch("mesonPx",&mesonPx,"mesonPx/F");
    _otree->Branch("mesonPy",&mesonPy,"mesonPy/F");
    _otree->Branch("mesonPz",&mesonPz,"mesonPz/F");

    _otree->Branch("sl_truePt",&sl_truePt,"sl_truePt/F");
    _otree->Branch("sl_truePx",&sl_truePx,"sl_truePx/F");
    _otree->Branch("sl_truePy",&sl_truePy,"sl_truePy/F");
    _otree->Branch("sl_truePz",&sl_truePz,"sl_truePz/F");
    _otree->Branch("sl_trueE",&sl_trueE,"sl_trueE/F");
    _otree->Branch("sl_trueTheta",&sl_trueTheta,"sl_trueTheta/F");
    _otree->Branch("sl_ccnc",&sl_ccnc,"sl_ccnc/I");
    _otree->Branch("sl_PDG",&sl_PDG,"sl_PDG/I");
    _otree->Branch("sl_origPDG",&sl_origPDG,"sl_origPDG/I");
    _otree->Branch("sl_mode",&sl_mode,"sl_mode/I");
    _otree->Branch("sl_isQE",&sl_isQE,"sl_isQE/B");
    _otree->Branch("sl_intType",&sl_intType,"sl_intType/I");
    _otree->Branch("sl_hitNucl",&sl_hitNucl,"sl_hitNucl/I");
    _otree->Branch("sl_HadX",&sl_HadX,"sl_HadX/F");
    _otree->Branch("sl_HadY",&sl_HadY,"sl_HadY/F");
    _otree->Branch("sl_HadW",&sl_HadW,"sl_HadW/F");
    _otree->Branch("sl_qsqr",&sl_qsqr,"sl_qsqr/F");
    _otree->Branch("SlNeuMatched",&SlNeuMatched,"SlNeuMatched/I");
    _otree->Branch("hasFlux",&hasFlux,"hasFlux/I");
    _otree->Branch("sl_Vx",&sl_Vx,"sl_Vx/F");
    _otree->Branch("sl_Vy",&sl_Vy,"sl_Vy/F");
    _otree->Branch("sl_Vz",&sl_Vz,"sl_Vz/F");
    _otree->Branch("sl_lepPDG",&sl_lepPDG,"sl_lepPDG/F");
    _otree->Branch("sl_lepPx",&sl_lepPx,"sl_lepPx/F");
    _otree->Branch("sl_lepPy",&sl_lepPy,"sl_lepPy/F");
    _otree->Branch("sl_lepPz",&sl_lepPz,"sl_lepPz/F");
    _otree->Branch("sl_lepE",&sl_lepE,"sl_lepE/F");

    _otree->Branch("Npi0",&Npi0,"Npi0/I");
    _otree->Branch("pi0E",&pi0E,"pi0E[Npi0]/F");
    _otree->Branch("pi0Vx",&pi0Vx,"pi0Vx[Npi0]/F");
    _otree->Branch("pi0Vy",&pi0Vy,"pi0Vy[Npi0]/F");
    _otree->Branch("pi0Vz",&pi0Vz,"pi0Vz[Npi0]/F");
    _otree->Branch("gam1E",&gam1E,"gam1E[Npi0]/F");
    _otree->Branch("gam1Px",&gam1Px,"gam1Px[Npi0]/F");
    _otree->Branch("gam1Py",&gam1Py,"gam1Py[Npi0]/F");
    _otree->Branch("gam1Pz",&gam1Pz,"gam1Pz[Npi0]/F");
    _otree->Branch("gam2E",&gam2E,"gam2E[Npi0]/F");
    _otree->Branch("gam2Px",&gam2Px,"gam2Px[Npi0]/F");
    _otree->Branch("gam2Py",&gam2Py,"gam2Py[Npi0]/F");
    _otree->Branch("gam2Pz",&gam2Pz,"gam2Pz[Npi0]/F");

    _otree->Branch("hasSlice",&hasSlice,"hasSlice/I");
    _otree->Branch("hasVertex",&hasVertex,"hasVertex/I");
    _otree->Branch("hasOneVertex",&hasOneVertex,"hasOneVertex/I");
    _otree->Branch("Nvertex",&Nvertex,"Nvertex/I");
    _otree->Branch("passPre",&passPre,"passPre/I");

    _otree->Branch("goodcells",&goodcells,"goodcells/I");
    _otree->Branch("Xcell",&Xcell,"Xcell[goodcells]/F");
    _otree->Branch("Ycell",&Ycell,"Ycell[goodcells]/F");
    _otree->Branch("Zcell",&Zcell,"Zcell[goodcells]/F");
    _otree->Branch("ecell",&ecell,"ecell[goodcells]/F");
    _otree->Branch("cellPECorr",&cellPECorr,"cellPECorr[goodcells]/F");
    _otree->Branch("inXVIEW",&inXVIEW,"inXVIEW[goodcells]/I");
    //_otree->Branch("MinCell",&MinCell,"MinCell[goodcells]/I");
    //_otree->Branch("cellCell",&cellCell,"cellCell[goodcells]/I");
    _otree->Branch("cellPlane",&cellPlane,"cellPlane[goodcells]/I");

    _otree->Branch("NKtracks",&NKtracks,"NKtracks/I");
    _otree->Branch("KtrackLength",&KtrackLength,"KtrackLength/F");
    _otree->Branch("KtrackNcells",&KtrackNcells,"KtrackNcells/I");
    _otree->Branch("KtrackEnergy",&KtrackEnergy,"KtrackEnergy/F");
    _otree->Branch("KtrackNcells_ratio",&KtrackNcells_ratio,"KtrackNcells_ratio/F");
    _otree->Branch("KtrackEnergy_ratio",&KtrackEnergy_ratio,"KtrackEnergy_ratio/F");
    _otree->Branch("KtrackNplanes",&KtrackNplanes,"KtrackNplanes/I");
    _otree->Branch("KtrackPhi",&KtrackPhi,"KtrackPhi/F");
    _otree->Branch("KtrackTheta",&KtrackTheta,"KtrackTheta/F");
    _otree->Branch("KtrackStartX",&KtrackStartX,"KtrackStartX/F");
    _otree->Branch("KtrackStartY",&KtrackStartY,"KtrackStartY/F");
    _otree->Branch("KtrackStartZ",&KtrackStartZ,"KtrackStartZ/F");
    _otree->Branch("KtrackStopX",&KtrackStopX,"KtrackStopX/F");
    _otree->Branch("KtrackStopY",&KtrackStopY,"KtrackStopY/F");
    _otree->Branch("KtrackStopZ",&KtrackStopZ,"KtrackStopZ/F");

    _otree->Branch("NKtracks3D",&NKtracks3D,"NKtracks3D/I");
    _otree->Branch("trackEnergy",&trackEnergy,"trackEnergy[NKtracks3D]/F");
    _otree->Branch("trackLength",&trackLength,"trackLength[NKtracks3D]/F");
    _otree->Branch("trackPhi",&trackPhi,"trackPhi[NKtracks3D]/F");
    _otree->Branch("trackTheta",&trackTheta,"trackTheta[NKtracks3D]/F");
    _otree->Branch("trackEnergyXView",&trackEnergyXView,"trackEnergyXView[NKtracks3D]/F");
    _otree->Branch("trackEnergyYView",&trackEnergyYView,"trackEnergyYView[NKtracks3D]/F");
    _otree->Branch("trackNCells",&trackNCells,"trackNCells[NKtracks3D]/I");
    _otree->Branch("trackNCellsXView",&trackNCellsXView,"trackNCellsXView[NKtracks3D]/I");
    _otree->Branch("trackNCellsYView",&trackNCellsYView,"trackNCellsYView[NKtracks3D]/I");

    _otree->Branch("Ncells",&Ncells,"Ncells/I");
    _otree->Branch("Nplanes",&Nplanes,"Nplanes/I");
    _otree->Branch("recoE",&recoE,"recoE/F");
    _otree->Branch("PlaneEnergy",&PlaneEnergy,"PlaneEnergy[Nplanes]/F");
    _otree->Branch("PlaneNcells",&PlaneNcells,"PlaneNcells[Nplanes]/I");
    _otree->Branch("PlaneDedx",&PlaneDedx,"PlaneDedx[Nplanes]/F");
    _otree->Branch("PlaneZcell",&PlaneZcell,"PlaneZcell[Nplanes]/F");
    _otree->Branch("MinPlane",&MinPlane,"MinPlane/I");
    _otree->Branch("E40planes",&E40planes,"E40planes/F");
    _otree->Branch("Efrac_20planes",&Efrac_20planes,"Efrac_20planes/F");
    _otree->Branch("Efrac_40planes",&Efrac_40planes,"Efrac_40planes/F");
    _otree->Branch("Efrac_2slides",&Efrac_2slides,"Efrac_2slides/F");
    _otree->Branch("Efrac_4slides",&Efrac_4slides,"Efrac_4slides/F");
    _otree->Branch("Efrac_6slides",&Efrac_6slides,"Efrac_6slides/F");
    _otree->Branch("Efrac_2sigma",&Efrac_2sigma,"Efrac_2sigma/F");
    _otree->Branch("Efrac_4sigma",&Efrac_4sigma,"Efrac_4sigma/F");
    _otree->Branch("Efrac_6sigma",&Efrac_6sigma,"Efrac_6sigma/F");

    for( int i=1; i<11; ++i ){
      TString Bname=Form("Eiso_%i",i);
      _otree->Branch(Bname,&Eiso[i-1],Bname+"/F");
    }
    
    _otree->Branch("Nmip",&Nmip,"Nmip/F");
    _otree->Branch("Fmip",&Fmip,"Fmip/F");

    _otree->Branch("Nprongs",&Nprongs,"Nprongs/I");
    _otree->Branch("Ebalance",&Ebalance,"Ebalance/F");
    _otree->Branch("Dphi",&Dphi,"Dphi/F");
    
    _otree->Branch("prongEnergy",&prongEnergy,"prongEnergy[Nprongs]/F");
    //_otree->Branch("prongPx",&prongPx,"prongPx[Nprongs]/F");
    //_otree->Branch("prongPy",&prongPy,"prongPy[Nprongs]/F");
    //_otree->Branch("prongPz",&prongPz,"prongPz[Nprongs]/F");
    _otree->Branch("prongPhi",&prongPhi,"prongPhi[Nprongs]/F");
    _otree->Branch("prongTheta",&prongTheta,"prongTheta[Nprongs]/F");
    _otree->Branch("prongEnergyXView",&prongEnergyXView,"prongEnergyXView[Nprongs]/F");
    _otree->Branch("prongEnergyYView",&prongEnergyYView,"prongEnergyYView[Nprongs]/F");
    _otree->Branch("prongNCells",&prongNCells,"prongNCells[Nprongs]/I");
    _otree->Branch("prongNCellsXView",&prongNCellsXView,"prongNCellsXView[Nprongs]/I");
    _otree->Branch("prongNCellsYView",&prongNCellsYView,"prongNCellsYView[Nprongs]/I");
    _otree->Branch("Nprongs2D",&Nprongs2D,"Nprongs2D/I");
    _otree->Branch("prong2DInXView",&prong2DInXView,"prong2DInXView[Nprongs2D]/I");
    _otree->Branch("prongEnergy2D",&prongEnergy2D,"prongEnergy2D[Nprongs2D]/F");
    _otree->Branch("prongNCells2D",&prongNCells2D,"prongNCells2D[Nprongs2D]/I");
    _otree->Branch("Ebalance2D",&Ebalance2D,"Ebalance2D/F");
    
  }

  void Xbeam::endSubRun(const art::SubRun& sr)
  {
    if( isMC ){
      art::Handle<sumdata::POTSum> pot;
      sr.getByLabel(fMCFluxLabel, pot);
      
      fTotalPOT += pot->totgoodpot;
      
      std::cout << "POT in subrun " << sr.subRun()
		<< " = " << pot->totgoodpot << std::endl;
    }
  }
  
  void Xbeam::analyze(const art::Event& evt) 
  {
    //initialization for tree leaves
    recoVx=-999.;
    recoVy=-999.;
    recoVz=-999.;
    hasTrueInfo=-1;
    nu_origPDG=-1;
    mesonPDG=-1;
    mesonPx=0.;
    mesonPy=0.;
    mesonPz=0.;
    
    SlNeuMatched=-1;
    hasFlux=-1;
    sl_truePt=0.;
    sl_truePx=0.;
    sl_truePy=0.;
    sl_truePz=0.;
    sl_trueE=0.;
    sl_trueTheta=0.;
    sl_ccnc=-1;
    sl_PDG=-1;
    sl_origPDG=-1;
    sl_intType=-1;
    sl_mode=-1; 
    sl_isQE=0;
    sl_hitNucl=-1;
    sl_HadX=-1.;
    sl_HadY=-1.;
    sl_HadW=-1.;
    sl_qsqr=-1.;
    sl_Vx=-999.;
    sl_Vy=-999.;
    sl_Vz=-999.;
    sl_lepPDG=-1;
    sl_lepE=0.;
    sl_lepPx=0.;
    sl_lepPy=0.;
    sl_lepPz=0.;

    Npi0=0;
    for( int i=0; i<50; ++i ){
       pi0E[i]=0.;
       pi0Vx[i]=-999.;
       pi0Vy[i]=-999.;
       pi0Vz[i]=-999.;
       gam1E[i]=0.;
       gam1Px[i]=0.;
       gam1Py[i]=0.;
       gam1Pz[i]=0.;
       gam2E[i]=0.;
       gam2Px[i]=0.;
       gam2Py[i]=0.;
       gam2Pz[i]=0.;
    }

    hasSlice=-1;
    hasVertex=-1;
    hasOneVertex=-1;
    Nvertex=-1;
    passPre=-1;

    Ncells=0;
    Nplanes=0;
    recoE=0.;
    MinPlane=0;
    for(int i=0; i<140.; ++i ){
       PlaneEnergy[i]=0.;
       PlaneNcells[i]=0;
       PlaneDedx[i]=0.;
       PlaneZcell[i]=0.;
    }
     
    E40planes=-1.;
    Efrac_20planes=-1.;
    Efrac_20planes=-1.;
    Efrac_2slides=-1.;
    Efrac_4slides=-1.;
    Efrac_6slides=-1.;
   
    Efrac_2sigma=-1.;
    Efrac_4sigma=-1.;
    Efrac_6sigma=-1.;

    for( int i=0; i<10; ++i ) Eiso[i]=-1.;
  
    KtrackLength=0.;
    KtrackNcells=0;
    KtrackEnergy=0.;
    KtrackNcells_ratio=-1.;
    KtrackEnergy_ratio=-1.;
    NKtracks=0;
    NKtracks3D=0;
    KtrackNplanes=-1;
    KtrackTheta=-1.;
    KtrackPhi=-1.;
    KtrackStartX=-999.;
    KtrackStartY=-999.;
    KtrackStartZ=-999.;
    KtrackStopX=-999.;
    KtrackStopY=-999.;
    KtrackStopZ=-999.;

    for( int i=0; i<20; ++i ){
      trackEnergy[i]=-1.;
      trackLength[i]=-1.;
      trackPhi[i]=-1.;
      trackTheta[i]=-1.;
      trackEnergyXView[i]=-1.;
      trackEnergyYView[i]=-1.;
      trackNCells[i]=-1;
      trackNCellsXView[i]=-1;
      trackNCellsYView[i]=-1;
    }


    Nmip=-1.;
    Fmip=-1.;

    goodcells=0.;
    for( int i=0; i<300; ++i ){
       Xcell[i]=-1.;
       Ycell[i]=-1.;
       Zcell[i]=-1.;
       ecell[i]=-1.;
       cellPECorr[i]=-1.;
       inXVIEW[i]=-1;
       cellPlane[i]=-1;
    }

    Nprongs=-1;
    Ebalance=-1.;
    Dphi=-1.;
    Nprongs2D=-1;
    Ebalance2D=-1.;
    for( int i=0; i<20; ++i ){
      prongEnergy[i]=-1.;
      prongPhi[i]=-1.;
      prongTheta[i]=-1.;
      prongEnergyXView[i]=-1.;
      prongEnergyYView[i]=-1.;
      prongNCells[i]=-1;
      prongNCellsXView[i]=-1;
      prongNCellsYView[i]=-1;
      prong2DInXView[i]=-1;
      prongEnergy2D[i]=-1.;
      prongNCells2D[i]=-1;
    }

    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;

/*
  if( isMC && isNue ){
    art::Handle<std::vector<simb::MCTruth> > mclist;
    evt.getByLabel(fMCFluxLabel,mclist);
    if( mclist->size()<1 ){
     std::cout<<"No true info ???"<<std::endl;
    }
    art::Handle< std::vector<simb::MCFlux> > fllist;
    evt.getByLabel(fMCFluxLabel, fllist);
    if( fllist->size()<1 ){
      std::cout<<"No flux info ???"<<std::endl;
    }

    if( mclist->size()>0 && fllist->size()>0 ){
    std::cout<<"# of interactions: "<<mclist->size()<<std::endl;
    for(unsigned int i_intx = 0; i_intx < mclist->size(); ++i_intx){
      art::Ptr<simb::MCTruth> mctruth(mclist,i_intx);
      const simb::MCNeutrino& nu = mctruth->GetNeutrino();
      //const simb::MCParticle& nu  = gen.Nu();
  
      nu_trueE[i_intx]   = nu.Nu().E();
      nu_PDG[i_intx]     = nu.Nu().PdgCode();
      nu_Vx[i_intx]=nu.Nu().Vx();
      nu_Vy[i_intx]=nu.Nu().Vy();
      nu_Vz[i_intx]=nu.Nu().Vz();

      nu_truePt[i_intx]  = nu.Pt();
      nu_trueTheta[i_intx] = nu.Theta();
      nu_truePhi[i_intx] = nu.Phi();
      nu_ccnc[i_intx]    = nu.CCNC();
      nu_mode[i_intx]    = nu.Mode();
      nu_isQE[i_intx]    = nu.InteractionType() == simb::kQE ||
        nu.InteractionType() == simb::kCCQE;
      nu_intType[i_intx] = nu.InteractionType();
      nu_hitNucl[i_intx] = nu.HitNuc();
      nu_HadX[i_intx]    = nu.X();
      nu_HadY[i_intx]    = nu.Y();
      nu_HadW[i_intx]    = nu.W();
      nu_qsqr[i_intx]    = nu.QSqr();

      lep_PDG[i_intx] = nu.Lepton().PdgCode();
      lep_Px[i_intx] = nu.Lepton().Px();
      lep_Py[i_intx] = nu.Lepton().Py();
      lep_Pz[i_intx] = nu.Lepton().Pz();
      lep_E[i_intx] = nu.Lepton().E();

      art::Ptr<simb::MCFlux> flux(fllist,i_intx);
      nu_origPDG[i_intx] = flux->fntype;
      meson_PDG[i_intx] = flux->ftptype;
      meson_Px[i_intx] = flux->ftpx;
      meson_Py[i_intx] = flux->ftpy;
      meson_Pz[i_intx] = flux->ftpz;

      std::cout<<i_intx<<" "<<nu.CCNC()<<" "<<flux->fntype<<" "<<nu.Nu().PdgCode()<<std::endl;
      std::cout<<"meson is "<<flux->ftptype<<std::endl;
     }//end loop of interactions
      std::cout<<"============================="<<std::endl;
    }//has true info
   }//for MC only
*/

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel, slices);

    //get tracks, verticies and showers associated with the slices
    art::FindManyP<rb::Track> fmKTrack(slices, evt, fKTrackLabel);
    art::FindManyP<rb::Vertex> fmVertex(slices, evt, fVertexLabel);

    int bestSlice=-1;
    unsigned int mostCells=0;
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;
      if( slice.NCell() > mostCells ){
        mostCells = slice.NCell();
        bestSlice = sliceIdx;
      }
    }//select slice
    if( bestSlice<0 ){
       hasSlice=0;
     }
    else{
     hasSlice=1;
     const rb::Cluster& slice = (*slices)[bestSlice];
   
     //select nu interaction matched to slice using bt
     if( isMC ){
       art::ServiceHandle<cheat::BackTracker> bt;
       art::Handle<std::vector<rb::CellHit> > chits;
       evt.getByLabel(fCellHitLabel, chits);
       art::PtrVector<rb::CellHit> allhits;
       for(size_t h = 0; h < chits->size(); ++h)
          allhits.push_back(art::Ptr<rb::CellHit>(chits, h));      
       const art::Ptr<rb::Cluster> allslice(slices,bestSlice);
       std::vector<cheat::NeutrinoEffPur> iacts = 
                                  bt->SliceToNeutrinoInteractions(allslice, allhits);
       if( iacts.empty() ){
          SlNeuMatched=0;
          std::cout<<"There is no matched interaction to selected slice."<<std::endl;
        }
        else {
          SlNeuMatched=1;
          art::Ptr<simb::MCTruth>& truth = iacts[0].neutrinoInt;
          art::PtrVector<simb::MCTruth> pv;
          pv.push_back(truth);
          // Get the flux associated with this truth
          art::FindManyP<simb::MCFlux> fmf(pv, evt, fMCFluxLabel);
          std::vector< art::Ptr<simb::MCFlux> > fluxs = fmf.at(0);
          simb::MCFlux flux;
          if(fluxs.size() == 1){
            flux = *fluxs[0];
            hasFlux=1;
          }
          else{
            hasFlux=0;
            std::cout<<"There is no assoicated neutrino intereaction !"<<std::endl;
          }
          const simb::MCNeutrino& nu = truth->GetNeutrino();
          sl_trueE   = nu.Nu().E();
          sl_truePx  = nu.Nu().Px();
          sl_truePy  = nu.Nu().Py();
          sl_truePz  = nu.Nu().Pz();
          sl_truePt  = nu.Pt();
          sl_trueTheta = nu.Theta();
          sl_ccnc    = nu.CCNC();
          sl_PDG     = nu.Nu().PdgCode();
          sl_origPDG = flux.fntype;
          sl_mode    = nu.Mode();
          sl_isQE    = nu.InteractionType() == simb::kQE ||
                       nu.InteractionType() == simb::kCCQE;      
          sl_intType = nu.InteractionType();
          sl_hitNucl = nu.HitNuc();
          sl_HadX    = nu.X();
          sl_HadY    = nu.Y();
          sl_HadW    = nu.W();
          sl_qsqr    = nu.QSqr();

          sl_Vx=nu.Nu().Vx();
          sl_Vy=nu.Nu().Vy();
          sl_Vz=nu.Nu().Vz();

          sl_lepPDG = nu.Lepton().PdgCode();
          sl_lepE = nu.Lepton().E();
          sl_lepPx = nu.Lepton().Px();
          sl_lepPy = nu.Lepton().Py();
          sl_lepPz = nu.Lepton().Pz();

          //select pi0 
          const sim::ParticleNavigator& pn = bt->ParticleNavigator();
          std::vector<const sim::Particle*> AllPi0;
          AllPi0.clear();
          for(int ip = 0; ip < pn.NumberOfPrimaries(); ++ip){
            const sim::Particle* genpar = pn.Primary(ip);
            if(genpar->PdgCode() != 111) continue;
            if( genpar->Vz()<0. || genpar->Vz()>1300. ) continue;//ND only, excluding the MC
            if( genpar->Vx()>200. || genpar->Vx()<-200. ) continue;
            if( genpar->Vy()>200. || genpar->Vy()<-200. ) continue;
            if( genpar->NumberDaughters() != 2 ) continue;
            const sim::Particle* gam1 = pn[genpar->Daughter(0)];
            const sim::Particle* gam2 = pn[genpar->Daughter(1)];
            if( gam1->PdgCode() != 22 ) continue;
            if( gam2->PdgCode() != 22 ) continue;
            AllPi0.push_back(pn.Primary(ip));
          } 

          int npi0=AllPi0.size();
          if( AllPi0.size()> 50 ) npi0=50;
          Npi0=npi0;
          if( npi0>0 ){
            //std::sort( AllPi0.begin(), AllPi0.end(), isTrueEnergyGreater );//order the pi0 based on the energy
            for( int ipi=0; ipi<npi0; ++ipi ){
               pi0E[ipi] = AllPi0[ipi]->E();
               pi0Vx[ipi] = AllPi0[ipi]->Vx();
               pi0Vy[ipi] = AllPi0[ipi]->Vy();
               pi0Vz[ipi] = AllPi0[ipi]->Vz();
               const sim::Particle* gam1 = pn[AllPi0[ipi]->Daughter(0)];
               const sim::Particle* gam2 = pn[AllPi0[ipi]->Daughter(1)];
               gam1E[ipi]=gam1->E();
               gam1Px[ipi]=gam1->Px();
               gam1Py[ipi]=gam1->Py();
               gam1Pz[ipi]=gam1->Pz();
               gam2E[ipi]=gam2->E();
               gam2Px[ipi]=gam2->Px();
               gam2Py[ipi]=gam2->Py();
               gam2Pz[ipi]=gam2->Pz();
            }//first 10 energetic pi0            
          }//has pi0

/*
          std::cout<<nu.Nu().PdgCode()<<" "<<flux.fntype<<" "<<nu.CCNC()<<std::endl;
          std::cout<<nu.Nu().E()<<" "<<nu.Theta()<<std::endl;
          std::cout<<nu.Lepton().PdgCode()<<" "<<nu.Lepton().E()<<std::endl;
*/
          art::Handle<std::vector<simb::MCTruth> > mclist;
          evt.getByLabel(fMCFluxLabel,mclist);
          art::Handle< std::vector<simb::MCFlux> > fllist;
          evt.getByLabel(fMCFluxLabel, fllist);
          if( mclist->size()>0 && fllist->size()>0 ){
            hasTrueInfo=1;
            int nmatched=0;
            for( unsigned int i_intx=0; i_intx<mclist->size(); ++i_intx){
               art::Ptr<simb::MCTruth> mctruth(mclist,i_intx);
               const simb::MCNeutrino& mcnu = mctruth->GetNeutrino();
               TLorentzVector Tslnu(nu.Nu().Px(),nu.Nu().Py(),
                                    nu.Nu().Pz(),nu.Nu().E());
               TLorentzVector Tmcnu(mcnu.Nu().Px(),mcnu.Nu().Py(),
                                    mcnu.Nu().Pz(),mcnu.Nu().E());
               if(mcnu.Nu().PdgCode()==nu.Nu().PdgCode() && 
                  mcnu.CCNC()==nu.CCNC() && mcnu.Mode()==nu.Mode() &&
                  mcnu.Lepton().PdgCode()==nu.Lepton().PdgCode() &&
                        Tslnu==Tmcnu ){
                  art::Ptr<simb::MCFlux> flux(fllist,i_intx);
                  nu_origPDG = flux->fntype;
                  mesonPDG = flux->ftptype;
                  mesonPx = flux->ftpx;
                  mesonPy = flux->ftpy;
                  mesonPz = flux->ftpz;
          /*       
               std::cout<<nu.Nu().PdgCode()<<" "<<mcnu.Nu().PdgCode()<<std::endl;
               std::cout<<nu.Lepton().PdgCode()<<" "<<mcnu.Lepton().PdgCode()<<std::endl;
               std::cout<<nu.Nu().E()<<" "<<mcnu.Nu().E()<<std::endl;
               std::cout<<flux->fntype<<" "<<flux->ftptype<<std::endl;
          */
               ++nmatched;
               }
            }//loop all nu interactions
            if( nmatched==0 ) std::cout<<"no matched interaction."<<std::endl;
            else if( nmatched>1 ) std::cout<<"more than one matched interaction."<<std::endl;
          }//has true info
          else hasTrueInfo=0;
       }//slice matched to nu interaction
     }//for MC only

     if (!(fmVertex.isValid())){
       hasVertex=0;
     }
     else{
       hasVertex=1;
       std::vector<art::Ptr<rb::Vertex>> vert = fmVertex.at(bestSlice);
       Nvertex=vert.size();
       if (vert.size() != 1){
         hasOneVertex=0;
       }
       else {
        hasOneVertex=1;

       recoVx=vert[0]->GetX();
       recoVy=vert[0]->GetY();
       recoVz=vert[0]->GetZ();

    //select the longest kalman track
    std::vector<art::Ptr<rb::Track>> Ktracks = fmKTrack.at(bestSlice);
    double longestKTrack = 0;
    int longestKTrack_ncells=0;
    double longestKTrack_energy=0.;
    int longestKTrack_nplanes=0;
    float longestKTrack_phi=0.;
    float longestKTrack_theta=0.;
    float longestKTrack_startx=0.;
    float longestKTrack_starty=0.;
    float longestKTrack_startz=0.;
    float longestKTrack_stopx=0.;
    float longestKTrack_stopy=0.;
    float longestKTrack_stopz=0.;

    for(unsigned int n = 0; n < Ktracks.size(); ++n){
      if( !(Ktracks[n]->Is3D()) ) continue;
      const double L = Ktracks[n]->TotalLength();
      if(L > longestKTrack){
        longestKTrack = L;
        longestKTrack_ncells=Ktracks[n]->NCell();
        longestKTrack_energy=Ktracks[n]->TotalGeV();
        longestKTrack_nplanes=Ktracks[n]->ExtentPlane();
        TVector3 dir = Ktracks[n]->Dir();
        longestKTrack_phi=dir.Phi();
        longestKTrack_theta=dir.Theta();
        TVector3 start = Ktracks[n]->Start();
        TVector3 stop = Ktracks[n]->Stop();
        longestKTrack_startx = start.X();
        longestKTrack_starty = start.Y();
        longestKTrack_startz = start.Z();
        longestKTrack_stopx = stop.X();
        longestKTrack_stopy = stop.Y();
        longestKTrack_stopz = stop.Z();
      }
    }

      //mip cells and reco energy for slicer
    const double nCells = slice.NCell();
    const int lengthPlanes = slice.ExtentPlane();
    double recoEnergy=0.;
    double nmip=0.;
    double Ntotal=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit = cal->MakeRecoHit(*chit,
						 chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
       recoEnergy += rhit.GeV();

       if( rhit.PECorr()>100. && rhit.PECorr()<245. ) nmip++;
       Ntotal++;
    }

    //preselection cuts
    if( recoEnergy>0.1 && nCells>10 && nCells<300 
        && lengthPlanes<140 && longestKTrack<400. )
    {
      passPre=1;
    
    NKtracks = Ktracks.size();
    KtrackLength = longestKTrack;
    KtrackEnergy = longestKTrack_energy;
    KtrackNcells = longestKTrack_ncells;
    KtrackEnergy_ratio = longestKTrack_energy/recoEnergy;
    KtrackNcells_ratio = longestKTrack_ncells/nCells;
    KtrackNplanes = longestKTrack_nplanes;
    KtrackPhi = longestKTrack_phi;
    KtrackTheta = longestKTrack_theta;
    KtrackStartX = longestKTrack_startx;
    KtrackStartY = longestKTrack_starty;
    KtrackStartZ = longestKTrack_startz;
    KtrackStopX = longestKTrack_stopx;
    KtrackStopY = longestKTrack_stopy;
    KtrackStopZ = longestKTrack_stopz;

    recoE = recoEnergy;
    Ncells = nCells;
    Nplanes = lengthPlanes;
    MinPlane = slice.MinPlane();
    Nmip=nmip;
    Fmip=nmip/Ntotal;


   //select 3D tracks
   std::vector<art::Ptr<rb::Track>> SelectedTrack;
   for(unsigned int n = 0; n < Ktracks.size(); ++n){
      if( !(Ktracks[n]->Is3D()) ) continue;
      SelectedTrack.push_back(Ktracks[n]);
   }
   //std::cout<<"# of 3D track is "<<SelectedTrack.size()<<std::endl;
   int ntracks=SelectedTrack.size();
   if( ntracks>20 ) ntracks=20;
   NKtracks3D=ntracks;

   if( ntracks>0 ){
    for( int it=0; it<ntracks; ++it ){
       TVector3 dir = SelectedTrack[it]->Dir();
       trackPhi[it]=dir.Phi();
       trackTheta[it]=dir.Theta();
       trackNCells[it]=SelectedTrack[it]->NCell();
       trackNCellsXView[it]=SelectedTrack[it]->NCell(geo::kX);
       trackNCellsYView[it]=SelectedTrack[it]->NCell(geo::kY);
       trackLength[it]=SelectedTrack[it]->TotalLength();
    }
   }

    art::FindManyP<rb::Prong> fmProng(vert, evt, fProngLabel);
    std::vector<art::Ptr<rb::Prong>> prongs = fmProng.at(0);
    std::vector<art::Ptr<rb::Prong>> SelectedProng;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
       //const rb::Prong* iprong = &(*shwcol3D)[ip];
       if( prongs[ip]->IsNoise() ) continue;
       if(prongs[ip]->View() != geo::kXorY) continue;
       //SelectedProng.push_back(&(*shwcol3D)[ip]);
       SelectedProng.push_back(prongs[ip]);
    }
    //std::cout<<SelectedProng.size()<<std::endl;
    int nprongs=SelectedProng.size();
    if( nprongs>20 ) nprongs=20;
    Nprongs=nprongs;
    std::vector<double> ProngEnergy;
    std::vector<double> ProngEnergyXView;
    std::vector<double> ProngEnergyYView;
    std::vector<double> ProngPhi;
    
    if( SelectedProng.size()>0 ){
    for(int ip=0; ip<nprongs; ++ip ){ 
      //reco direction
     TVector3 dir1 = SelectedProng[ip]->Dir();
     //recoEta[ip]=dir1.Eta();
     prongPhi[ip]=dir1.Phi();
     prongTheta[ip]=dir1.Theta();

     prongNCells[ip]=SelectedProng[ip]->NCell();
     prongNCellsXView[ip]=SelectedProng[ip]->NCell(geo::kX);
     prongNCellsYView[ip]=SelectedProng[ip]->NCell(geo::kY);

     //reco momentum
     double Eprong=0.;     
     double EprongX=0.;
     double EprongY=0.;
     for( unsigned int icell=0; icell<SelectedProng[ip]->NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = SelectedProng[ip]->Cell(icell);
       const TVector3 mean = SelectedProng[ip]->MeanXYZ();
       const double wx = (SelectedProng[ip]->NXCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->XCell(0).get()) : 0;
       const double wy = (SelectedProng[ip]->NYCell() > 0) ? SelectedProng[ip]->W(SelectedProng[ip]->YCell(0).get()) : 0;
       const rb::RecoHit rhit = cal->MakeRecoHit(*chit,
						 chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
       Eprong += rhit.GeV();
       if( chit->View() == geo::kX ) EprongX += rhit.GeV();
       if( chit->View() == geo::kY ) EprongY += rhit.GeV();
     }
     ProngEnergy.push_back(Eprong);
     ProngEnergyXView.push_back(EprongX);
     ProngEnergyYView.push_back(EprongY);
     ProngPhi.push_back(dir1.Phi());
     prongEnergy[ip]=Eprong;
     prongEnergyXView[ip]=EprongX;
     prongEnergyYView[ip]=EprongY;
/*
     prongPx[ip]=Eprong*sin(dir1.Theta())*sin(dir1.Phi());
     prongPy[ip]=Eprong*sin(dir1.Theta())*cos(dir1.Phi());
     prongPz[ip]=Eprong*cos(dir1.Theta());
*/
    }
   }//has at least one prong

   double cosTheta=0.;
   if( nprongs==1 ) cosTheta=cos(SelectedProng[0]->Dir().Theta());

   double ratio=1.;
   double dphi=0.;
   if( nprongs>1 ){
      int Index1=-1;
      int Index2=-1;
      double MaxEnergy1=-999.;
      double MaxEnergy2=-999.;
      for( int i=0;i<nprongs; ++i ){
         if( MaxEnergy1<ProngEnergy[i] ){
           MaxEnergy1=ProngEnergy[i];
           Index1=i;
        }
      }//leading
      for( int i=0;i<nprongs; ++i ){
         if( i==Index1 ) continue;
         if( MaxEnergy2<ProngEnergy[i] ){
           MaxEnergy2=ProngEnergy[i];
           Index2=i;
        }
      }//sub-leading
      double Eden=ProngEnergy[Index1]+ProngEnergy[Index2];
      double Enum=ProngEnergy[Index1]-ProngEnergy[Index2];
      if( Eden>0.0 ) ratio=Enum/Eden;
      dphi=fabs(ProngPhi[Index1]-ProngPhi[Index2]);
      if( dphi>2.*TMath::Pi() ) dphi=dphi-2.*TMath::Pi();
      if( dphi>TMath::Pi() ) dphi=2.*TMath::Pi()-dphi;

      cosTheta = cos(SelectedProng[Index1]->Dir().Theta());
    }
    Ebalance=ratio;
    Dphi=dphi;

    //calculate per plane cell energy
    std::vector<double> PlaneSumEnergy;
    double totalE_road20=0.;
    double totalE_road40=0.;
    double totalE_road60=0.;

    double Esig[10]={0.};

    for( int iplane=0; iplane<lengthPlanes; ++iplane ){
      double cellEnergy=0.;//sum of cells energy in each plane
      int planecells=0;//# of cells in each plane
      double cellX=0.;//energetic cell X in each plane
      double cellY=0.;//energetic cell Y in each plane
      //double cellMax=-99.;//energy of energetic cell
      double Ecell=0.;
      double EXcell=0.;
      double EYcell=0.;
      double DEDX=0.;

      double EZcell=0.;
      double cellZ=0.;

      std::vector<double> Xxview;
      std::vector<double> Exview;
      std::vector<double> Yyview;
      std::vector<double> Eyview;
      Xxview.clear();
      Exview.clear();
      Yyview.clear();
      Eyview.clear();


      //if( slice.NCell()>0 ){//already required there exists at least 20 cells
       for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
         const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
         const TVector3 mean = slice.MeanXYZ();
         const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
         const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
         const rb::RecoHit rhit = cal->MakeRecoHit(*chit,
						   chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;

        int Dplane=chit->Plane()-slice.MinPlane();
        if( Dplane<0 ) std::cout<<"plane # of this cell is beyond the minimal plane."<<std::endl;

        //int mincell=-1;
        //if( chit->View()==geo::kX ) mincell=slice.MinCell(geo::kX);
        //else if( chit->View()==geo::kY ) mincell=slice.MinCell(geo::kY);
        //int Dcell=chit->Cell()-mincell;

        if( Dplane==iplane ){
          cellEnergy += rhit.GeV();
          planecells++;
          //if( cellMax<rhit.GeV() ){
             //cellMax=rhit.GeV();
             //cellX=rhit.X();
            // cellY=rhit.Y();
          //}
          Ecell += rhit.GeV();
          EXcell += rhit.GeV()*rhit.X();
          EYcell += rhit.GeV()*rhit.Y();         
          EZcell += rhit.GeV()*rhit.Z();
         //std::cout<<"plane "<<iplane<<": "<<chit->Cell()<<"-"<<mincell<<" = "<<Dcell<<std::endl;

          if( chit->View() == geo::kX ){
            Xxview.push_back(rhit.X());
            Exview.push_back(rhit.GeV());
          }//in X view
          else {
            Yyview.push_back(rhit.Y());
            Eyview.push_back(rhit.GeV());
          }//in Y view

          int Iplane=chit->Plane();
          int Icell=chit->Cell();
          double dcell=2.*geom->Plane(Iplane)->Cell(Icell)->HalfD();
          double dedx=rhit.GeV()*cosTheta/dcell;
          DEDX += dedx;

         }//Dplane==iplane
       }//loop cells
      //}//has cells
     if( Ecell>0. ){
        cellX = EXcell/Ecell;
        cellY = EYcell/Ecell;
        cellZ = EZcell/Ecell;
     }
     
     PlaneSumEnergy.push_back(cellEnergy);
  
     int Nxviews=Xxview.size();
      int Nyviews=Yyview.size();
      if( Nxviews>0 ){
        for( int ixv=0; ixv<Nxviews; ++ixv ){
          if( fabs(Xxview[ixv]-cellX) < 2.*2. ) totalE_road20+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*4. ) totalE_road40+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*6. ) totalE_road60+=Exview[ixv];

            for( int iw=1; iw<11; ++iw ){
              if( fabs(Xxview[ixv]-cellX) < 2.*iw ) Esig[iw-1]+=Exview[ixv];
           }
        }
      }//in X view
      if( Nyviews>0 ){
        for( int iyv=0; iyv<Nyviews; ++iyv ){
          if( fabs(Yyview[iyv]-cellY) < 2.*2. ) totalE_road20+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*4. ) totalE_road40+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*6. ) totalE_road60+=Eyview[iyv];
            for( int iw=1; iw<11; ++iw ){
             if( fabs(Yyview[iyv]-cellY) < 2.*iw ) Esig[iw-1]+=Eyview[iyv];
          }
        }
      }//in Y view

     PlaneEnergy[iplane] = cellEnergy; 
     PlaneNcells[iplane] = planecells;
     PlaneDedx[iplane] = DEDX;
     PlaneZcell[iplane] = cellZ;
   }

    double efrac_plane2=-1.;
    double efrac_plane4=-1.;
    double efrac_plane6=-1.;
    double Eplane20=0.;
    double Eplane40=0.;
   int nplanes=PlaneSumEnergy.size();
   for( int i=0; i<nplanes; ++i ){
      if( i<20 ) Eplane20 += PlaneSumEnergy[i];
      if( i<40 ) Eplane40 += PlaneSumEnergy[i];

      //maximal energy fraction in 2 sliding window
      if( i<nplanes-2 ){
        double e_frac2=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1])/recoEnergy;
        if( efrac_plane2<e_frac2 ) efrac_plane2=e_frac2;        }
      else if( nplanes<=2 ) efrac_plane2=1.;
      //maximal energy fraction in 4 sliding window
      if( i<nplanes-4 ){
        double e_frac4=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3])/recoEnergy;
        if( efrac_plane4<e_frac4 ) efrac_plane4=e_frac4;        }
      else if( nplanes<=4 ) efrac_plane4=1.;
      //maximal energy fraction in 6 sliding window
      if( i<nplanes-6 ){
        double e_frac6=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3]+PlaneSumEnergy[i+4]+PlaneSumEnergy[i+5])/recoEnergy;
        if( efrac_plane6<e_frac6 ) efrac_plane6=e_frac6; }
      else if( nplanes<=6 ) efrac_plane6=1.;
    }

     E40planes=Eplane40;
     Efrac_20planes=Eplane20/recoEnergy;
     Efrac_40planes=Eplane40/recoEnergy;
     Efrac_2slides=efrac_plane2;
     Efrac_4slides=efrac_plane4;
     Efrac_6slides=efrac_plane6;
     Efrac_2sigma=totalE_road20/recoEnergy;
     Efrac_4sigma=totalE_road40/recoEnergy;
     Efrac_6sigma=totalE_road60/recoEnergy;

     double iso[10]={0.};
     for( int iw=0; iw<10; ++iw ){
       if( Esig[iw]>0. ) iso[iw]=(recoEnergy-Esig[iw])/Esig[iw];
     }
     
     for( int iw=0; iw<10; ++iw ) Eiso[iw] = iso[iw];

    //cell information
    int ncells=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
       const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
       const TVector3 mean = slice.MeanXYZ();
       const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
       const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

       const rb::RecoHit rhit = cal->MakeRecoHit(*chit,
						 chit->View() == geo::kX ? wx : wy);
       if( !rhit.IsCalibrated() ) continue;
       //if( icell<5000 ){//already require there is at most 300 cells
         ncells++;
         Xcell[icell]=rhit.X();
         Ycell[icell]=rhit.Y();
         Zcell[icell]=rhit.Z();
         ecell[icell]=rhit.GeV();
         cellPECorr[icell]=rhit.PECorr();
         cellPlane[icell]=chit->Plane();
         //cellCell[icell]=chit->Cell();

         if( chit->View()==geo::kX ){
           inXVIEW[icell]=1;
           //MinCell[icell]=slice.MinCell(geo::kX);
         }//in X view
         else {
           inXVIEW[icell]=0;
           //MinCell[icell]=slice.MinCell(geo::kY);
         }//in Y view

       //}
      //distance from readout to each cell position
      //double dX = 1549.4/2.-rhit.X();
      //double dY = 1549.2/2.-rhit.Y();
     }
    goodcells=ncells;
    
    //reco prong View(): 0 for XZ, 1 for YZ, 2 for 3D
    //study 2D prongs
    //std::vector<art::Ptr<rb::Track>> SelectedProng2D;
    std::vector<art::Ptr<rb::Prong>> SelectedProng2D;

    for( unsigned int ip=0; ip<prongs.size(); ++ip ){
      //const rb::Prong* iprong = &(*shwcol2D)[ip];
      if( prongs[ip]->IsNoise() ) continue; 
      if( prongs[ip]->View() == geo::kXorY) continue;
      //SelectedProng2D.push_back(&(*shwcol2D)[ip]);
      SelectedProng2D.push_back(prongs[ip]);
      //std::cout<<ip<<" "<<iprong->NCell()<<std::endl;
    }     
    std::vector<int> Prong2DInXView;
    std::vector<double> ProngEnergy2D;
    int nprongs2D=SelectedProng2D.size();
    if( nprongs2D>20 ) nprongs2D=20;
    Nprongs2D = nprongs2D;
    for(int ip=0; ip<nprongs2D; ++ip ){
      prongNCells2D[ip]=SelectedProng2D[ip]->NCell();
      int prongview=SelectedProng2D[ip]->View();
      prong2DInXView[ip]=prongview; 
      Prong2DInXView.push_back(prongview);
      //reco momentum
      double Eprong=0.;
      for( unsigned int icell=0; icell<SelectedProng2D[ip]->NCell(); ++ icell){
        const art::Ptr<rb::CellHit>& chit = SelectedProng2D[ip]->Cell(icell);
        const TVector3 mean = SelectedProng2D[ip]->MeanXYZ();
        const double wx = (SelectedProng2D[ip]->NXCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->XCell(0).get()) : 0;
        const double wy = (SelectedProng2D[ip]->NYCell() > 0) ? SelectedProng2D[ip]->W(SelectedProng2D[ip]->YCell(0).get()) : 0;
        const rb::RecoHit rhit = cal->MakeRecoHit(*chit,
                                                  chit->View() == geo::kX ? wx : wy);
        if( !rhit.IsCalibrated() ) continue;
        Eprong += rhit.GeV();
      }
      
      prongEnergy2D[ip]=Eprong;
      ProngEnergy2D.push_back(Eprong);
    }

      int index3D=-1;
      double max3D=-99.;
      for( int i=0; i<nprongs; ++i ){
        if( max3D<ProngEnergy[i] ){
          max3D=ProngEnergy[i];
          index3D=i;
        }
      }//loop 3D prongs
      int index2D=-1;
      double max2D=-99.;
      for( int i=0;i<nprongs2D; ++i ){
        if( max2D<ProngEnergy2D[i] ){
          max2D=ProngEnergy2D[i];
          index2D=i;
        }
      }//loop 2D prongs
     double ebalance2D=1.;
     if( nprongs>0 && nprongs2D>0 ){
        double Eden=0.;
        double Enum=0.;
        if( Prong2DInXView[index2D]==0 ){//XZ view
          Eden=ProngEnergy2D[index2D]+ProngEnergyXView[index3D];
          Enum=ProngEnergy2D[index2D]-ProngEnergyXView[index3D];
        }
        else {//YZ view
          Eden=ProngEnergy2D[index2D]+ProngEnergyYView[index3D];
          Enum=ProngEnergy2D[index2D]-ProngEnergyYView[index3D];
        }
        if( Eden>0. ) ebalance2D=fabs(Enum)/Eden;
     }
     Ebalance2D=ebalance2D;

     }//pass preselection
     else{
       passPre=0;
       goodcells=0;
       Nplanes=0;
       Nprongs=0;
       Nprongs2D=0;
     }
     }//has only one vertex
    }//has vertex
   }//has slice

    _otree->Fill();
    return;
  }//end analyze

 void Xbeam::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    TH1* hPOT = tfs->make<TH1F>("hTotalPOT", ";; POT", 1, 0, 1);
    hPOT->Fill(.5, fTotalPOT);
  }
 
 bool Xbeam::isTrueEnergyGreater(const sim::Particle* p1, const sim::Particle* p2)
  {
  return (p1->E() > p2->E());
  }

  
}//end namespace

namespace ncs{

  DEFINE_ART_MODULE(Xbeam);

}

