////////////////////////////////////////////////////////////////////           
/// \brief   Boosted decision tree training class for numu cc selector
/// \author  xue@physics.umn.edu                                               
//////////////////////////////////////////////////////////////////// 
// C/C++ includes
#include <cmath>
#include <iostream>
#include <map>
#include <string>
#include <vector>

// ROOT includes
#include "TTree.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"

// NOvASoft includes
#include "NuMuBDTSelector/VarCalc.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"


namespace numubdtsel {
  class NuMuBDTSelectorTrain : public art::EDAnalyzer {
  public:
    explicit NuMuBDTSelectorTrain(fhicl::ParameterSet const& pset); 
    ~NuMuBDTSelectorTrain();                        
    
    void beginJob();      
    void analyze(art::Event const& evt); 
    void endJob();
    void reconfigure(const fhicl::ParameterSet& pset);
    
  private:
    
    TTree* varCC;
    TTree* varNC;
    
    VarCalc * vc;
    
    float fLongestTrackLength;
    float fNormVertexEnergy;
    float fLongestTrackAvgEnergy;
    float fTrackLengthRatio;
    float fTotalGeV;
    
  };
}

////////////////////////////////////////////////////////////////////////
namespace numubdtsel
{
  //.......................................................................
  NuMuBDTSelectorTrain::NuMuBDTSelectorTrain(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    reconfigure(pset);
  }

  //......................................................................
  NuMuBDTSelectorTrain::~NuMuBDTSelectorTrain()
  {
  }

  //......................................................................
  
  void NuMuBDTSelectorTrain::beginJob()
  { 

    art::ServiceHandle<art::TFileService> tfs;
    
    varCC = tfs->make<TTree>("fVarCC", "variables for CC events");
    varCC->Branch("longestTrackLength", &fLongestTrackLength, "longestTrackLength/F");
    varCC->Branch("normVertexEnergy", &fNormVertexEnergy, "normVertexEnergy/F");
    varCC->Branch("longestTrackAvgEnergy", &fLongestTrackAvgEnergy, "longestTrackAvgEnergy/F"); 
    varCC->Branch("trackLengthRatio", &fTrackLengthRatio, "trackLengthRatio/F");
    varCC->Branch("totalGeV", &fTotalGeV, "totalGeV/F");

    varNC = tfs->make<TTree>("fVarNC", "variables for NC events");
    varNC->Branch("longestTrackLength", &fLongestTrackLength, "longestTrackLength/F");
    varNC->Branch("normVertexEnergy", &fNormVertexEnergy, "normVertexEnergy/F");
    varNC->Branch("longestTrackAvgEnergy", &fLongestTrackAvgEnergy, "longestTrackAvgEnergy/F"); 
    varNC->Branch("trackLengthRatio", &fTrackLengthRatio, "trackLengthRatio/F");
    varNC->Branch("totalGeV", &fTotalGeV, "totalGeV/F");
    
  }

  //.................................................................
  
  void NuMuBDTSelectorTrain::analyze(art::Event const& evt)
  {
    
    art::Handle< std::vector<simb::MCTruth> > mclist;
    evt.getByLabel("generator", mclist);

    if(mclist->size() == 0)
      return;
    
    art::Ptr<simb::MCTruth> mc    (mclist,0);
    const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
    int currType = mc_neutrino.CCNC();

    if (!vc->PassPreselectionCut(evt))
      return;
    
    vc->SetInputVariables(evt, 
			  fLongestTrackLength, 
			  fNormVertexEnergy, 
			  fLongestTrackAvgEnergy, 
			  fTrackLengthRatio, 
			  fTotalGeV);

    if (currType == 0)
      varCC->Fill();
    else
      varNC->Fill();

    return;
  }
  
  //.................................................................

  void NuMuBDTSelectorTrain::endJob()
  {
  }
  
  //.................................................................
  
  void NuMuBDTSelectorTrain::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

} // end namespace numubdtsel

////////////////////////////////////////////////////////////////////////
namespace numubdtsel
{
  DEFINE_ART_MODULE(NuMuBDTSelectorTrain);
}
