////////////////////////////////////////////////////////////////////           
/// \brief   Boosted decision tree-based numu CC event selector
/// \author  xue@physics.umn.edu                                                
///////////////////////////////////////////////////////////////////// 
//C/C++ includes
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <map>
#include <string>
#include <vector>

// ROOT includes
#include "TFile.h"
#include "TTree.h"
#include "TString.h"
#include "TSystem.h"
#include "TROOT.h"
#include "TStopwatch.h"

#include "TMVA/Tools.h"
#include "TMVA/Reader.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"

// NOvASoft includes
#include "NuMuBDTSelector/VarCalc.h"
#include "NuMuBDTSelector/EID.h"


namespace numubdtsel {
  class NuMuBDTSelector : public art::EDProducer {
    
    public:
      
      explicit NuMuBDTSelector(fhicl::ParameterSet const& pset);
      ~NuMuBDTSelector();                        
      
      void produce(art::Event &evt);
      void reconfigure(const fhicl::ParameterSet& p);

    private:

      VarCalc * vc;

      TMVA::Reader * reader;

      float fLongestTrackLength;
      float fNormVertexEnergy;
      float fLongestTrackAvgEnergy;
      float fTrackLengthRatio;
      float fTotalGeV;
  };
}

////////////////////////////////////////////////////////////////////////
namespace numubdtsel
{
  //.......................................................................
  NuMuBDTSelector::NuMuBDTSelector(fhicl::ParameterSet const& pset)
  {
    
    reconfigure(pset);
    
    produces< std::vector<numubdtsel::EID>   >();
   
    reader = new TMVA::Reader( "!Color:!Silent" );

    reader->AddVariable( "longestTrackLength",    &fLongestTrackLength );
    reader->AddVariable( "normVertexEnergy",      &fNormVertexEnergy );
    reader->AddVariable( "longestTrackAvgEnergy", &fLongestTrackAvgEnergy );
    reader->AddVariable( "trackLengthRatio",      &fTrackLengthRatio );
    reader->AddVariable( "totalGeV",              &fTotalGeV );

    reader->BookMVA( "BDTG", "./weights/TMVAClassification_BDTG.weights.xml" );

  }

  //......................................................................
  NuMuBDTSelector::~NuMuBDTSelector()
  {
  }

  //......................................................................
  
  void NuMuBDTSelector::produce(art::Event & evt)
  {

    std::unique_ptr< std::vector<numubdtsel::EID> > eventids(new std::vector<numubdtsel::EID>);
    
    if (!vc->PassPreselectionCut(evt))
      return;

    vc->SetInputVariables(evt, fLongestTrackLength, fNormVertexEnergy, fLongestTrackAvgEnergy, fTrackLengthRatio, fTotalGeV);

    double mvaValue = reader->EvaluateMVA( "BDTG" );

    eventids->push_back(mvaValue);
    evt.put(std::move(eventids));

    return;
  }

  //....................................................................

  void NuMuBDTSelector::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

} // end namespace numubdtsel

////////////////////////////////////////////////////////////////////////
namespace numubdtsel
{
  DEFINE_ART_MODULE(NuMuBDTSelector);
}

