////////////////////////////////////////////////////////////////////
// \file    EID.cxx
// \brief   Brief please!!
// \version $Id: VarCalc.cxx,v 1.3 2012-09-21 02:32:52 greenc Exp $
// \author  xue@physics.umn.edu
// \author  $author: $
////////////////////////////////////////////////////////////////////
#include "NuMuBDTSelector/VarCalc.h"

// C/C++ includes
#include <iostream>
#include <map>
#include <vector>
#include <cstdlib>

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"

// NOvASoft includes
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"


namespace numubdtsel
{
  //.......................................................................
  VarCalc::VarCalc()
  {

  }

  //......................................................................
  VarCalc::~VarCalc()
  {

  }

  //......................................................................

  // returns true if event passes preselection cuts, false otherwise

  bool VarCalc::PassPreselectionCut(art::Event const& evt) {

    art::Handle< std::vector<simb::MCTruth> > mclist;
    evt.getByLabel("generator", mclist);

    if(mclist->size() == 0)
      return false;

    art::Ptr<simb::MCTruth> mc    (mclist,0);
    const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
    int neutrinoPDG = mc_neutrino.Nu().PdgCode();

    // ignore non numu events for now
    if (neutrinoPDG != 14 && neutrinoPDG != -14)
      return false;

    // must have at least one non-noise slice - necessary?
    art::Handle< std::vector<rb::Cluster> > clusterhandle;
    evt.getByLabel("slicer",clusterhandle);

    bool nonNoiseSliceExists = false;
    for (unsigned int i=0; i<clusterhandle->size(); i++) {
      art::Ptr<rb::Cluster> cluster(clusterhandle,i);
      if ( !cluster->IsNoise() ) {
        nonNoiseSliceExists = true;
        break;
      }
    }
    if (!nonNoiseSliceExists)
      return false;

    // at least one 3d reconstructed track
    art::Handle< std::vector<rb::Track> > trkcol;
    evt.getByLabel("kalmantrackmerge",trkcol);

    bool RecoTrackExists = false;
    for(unsigned int i = 0; i<trkcol->size(); ++i) {
      art::Ptr<rb::Track> track(trkcol,i);
      if ( track->NCell()!=track->NXCell() && track->NCell()!=track->NYCell() ) {
        RecoTrackExists = true;
        break;
      }
    }
    if (!RecoTrackExists)
      return false;

    // at least one reconstructed vertex
    art::Handle< std::vector<rb::Vertex> > vertexhandle;
    evt.getByLabel("elasticarmshs", vertexhandle);
    if (vertexhandle->size() == 0)
      return false;

    //fiducial cut

    art::Ptr<rb::Vertex> vertex(vertexhandle,0);

    art::ServiceHandle<geo::Geometry> geom;

    double radius = 100;

    if (vertex->GetZ() < radius || vertex->GetZ() > geom->DetLength()-radius)
      return false;
    if (vertex->GetX() < -geom->DetHalfWidth()+radius || vertex->GetX() > geom->DetHalfWidth()-radius)
      return false;
    if (vertex->GetY() < -geom->DetHalfHeight()+radius || vertex->GetY() > geom->DetHalfHeight()-radius)
      return false;

    return true;
  }


  //......................................................................


  void VarCalc::SetInputVariables(art::Event const& evt, float &longestTrackLength, float &normVertexEnergy, float &longestTrackAvgEnergy, float &trackLengthRatio, float &totalGeV)
  {

    // get (non-noise) slices
    art::Handle< std::vector<rb::Cluster> > clusterhandle;
    evt.getByLabel("slicer",clusterhandle);

    art::PtrVector<rb::Cluster> clusterlist;
    for (unsigned int i=0; i<clusterhandle->size(); i++) {
      art::Ptr<rb::Cluster> cluster(clusterhandle,i);
      if ( !cluster->IsNoise() ) {
        clusterlist.push_back(cluster);
      }
    }

    // get tracks
    art::Handle< std::vector<rb::Track> > trkcol;
    evt.getByLabel("kalmantrackmerge",trkcol);

    // 3d tracks only
    art::PtrVector<rb::Track> trklist;
    for(unsigned int i = 0; i<trkcol->size(); ++i) {
      art::Ptr<rb::Track> track(trkcol,i);
      if ( track->NCell()!=track->NXCell() && track->NCell()!=track->NYCell() ) {
        trklist.push_back(track);
      }
    }

    // find longest track

    longestTrackLength = 0;
    unsigned int longestTrackIndex = 0;

    for (unsigned int i=0; i<trklist.size(); i++) {
      if (longestTrackLength < trklist[i]->TotalLength()) {
        longestTrackLength = trklist[i]->TotalLength();
        longestTrackIndex = i;
      }
    }


    // find second longest track

    double secondLongestTrackLength = 0;

    for (unsigned int i=0; i<trklist.size(); i++) {
      if (secondLongestTrackLength < trklist[i]->TotalLength() && i != longestTrackIndex) {
        secondLongestTrackLength = trklist[i]->TotalLength();
      }
    }



    // use slice corresponding to neutrino interaction

    int neutrinoSliceIndex = 0;

    // use slice with most cellhits as slice containing neutrino interaction, for now
    if (clusterlist.size() > 1) {
      for (unsigned int i=1; i<clusterlist.size(); i++) {
        if (clusterlist[i]->NCell() > clusterlist[neutrinoSliceIndex]->NCell())
          neutrinoSliceIndex = i;
      }
    }

    art::Ptr<rb::Cluster> neutrinoSlice = clusterlist[neutrinoSliceIndex];

    // get vertex
    art::Handle< std::vector<rb::Vertex> > vertexhandle;
    evt.getByLabel("elasticarmshs", vertexhandle);

    art::Ptr<rb::Vertex> vertex(vertexhandle,0);


    // loop over cellhits to get vertex energy
    float vertexPE = 0;

    art::ServiceHandle<geo::Geometry> geom;

    double radius = 100;

    // x view
    const art::PtrVector<rb::CellHit>& xhits = neutrinoSlice->XCells();
    for (unsigned int i=0; i<xhits.size(); i++) {
      double xyz[3];
      geom->Plane(xhits[i]->Plane())->Cell(xhits[i]->Cell())->GetCenter(xyz);
      double dx = xyz[0] - vertex->GetX();
      double dz = xyz[2] - vertex->GetZ();
      if ( dx*dx+dz*dz <= radius*radius ) {
        rb::RecoHit rhit = neutrinoSlice->RecoHit(xhits[i]);
        vertexPE += rhit.GeV();
      }
    }

    // y view
    const art::PtrVector<rb::CellHit>& yhits = neutrinoSlice->YCells();
    for (unsigned int i=0; i<yhits.size(); i++) {
      double xyz[3];
      geom->Plane(yhits[i]->Plane())->Cell(yhits[i]->Cell())->GetCenter(xyz);
      double dy = xyz[1] - vertex->GetY();
      double dz = xyz[2] - vertex->GetZ();
      if ( dy*dy+dz*dz <= radius*radius ) {
        rb::RecoHit rhit = neutrinoSlice->RecoHit(yhits[i]);
        vertexPE += rhit.GeV();
      }
    }

    //set variables

    double area = 3.14 * radius * radius;
    normVertexEnergy = vertexPE / area / longestTrackLength;

    double longestTrackEnergy = trklist[longestTrackIndex]->TotalGeV();
    longestTrackAvgEnergy = longestTrackEnergy / longestTrackLength;

    trackLengthRatio = secondLongestTrackLength / longestTrackLength;

    totalGeV = neutrinoSlice->TotalGeV();

  }

} // end namespace numubdtsel
////////////////////////////////////////////////////////////////////////

