///////////////////////////////////////////////////////////////////////
// Class:       NumuEAna
// Module Type: analyzer
// File:        NumuEAna_module.cc
////////////////////////////////////////////////////////////////////////

#include "TH1.h"
#include "TH2.h"

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "MCCheater/BackTracker.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Energy.h"
#include "Simulation/Simulation.h"
#include "Simulation/Particle.h"
#include "Geometry/Geometry.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"
#include "ReMId/ReMId.h"
#include "NumuEnergy/NumuE.h"
#include "NovaDAQConventions/DAQConventions.h"

class TH1D;
class TH2D;

namespace numue {
  
  class NumuEAna : public art::EDAnalyzer {
  public:
    explicit NumuEAna(fhicl::ParameterSet const &p);
    virtual ~NumuEAna();
    
    virtual void analyze(art::Event const &e);
    
    virtual void reconfigure(fhicl::ParameterSet const & p);
    virtual void beginJob();
    virtual void beginRun(art::Run const &r);
    
  private:
    
    std::string fHitModuleLabel;      ///< lable for module making the hits
    std::string fTrackModuleLabel;    ///< label for module making the tracks
    std::string fSlicerModuleLabel;   ///< label for module making the slices
    std::string fEnergyModuleLabel;   ///< label for module writing energy objects
    std::string fPIDModuleLabel;      ///< label for module writing pid objects
    
    TH1D*       fEnergy;              ///< histogram of best energy
    TH1D*       fCalCCEnergy;         ///< histogram of calorimetric neutrino energy for general CC population
    TH1D*       fTrkQEEnergy;         ///< histogram of muon track length neutrino energy for QE super pure population
    TH1D*       fTrkNonQEEnergy;      ///< histogram of muon track length neutrino energy for NonQE super pure population
    TH1D*       fAngleQEEnergy;       ///< histogram of neutrino energy for QE population using angle formula
    
    TH1D*       fAngleQEError;        ///< histogram of error on the neutrino energy for QE population using angle formula
    
    TH1D*       fEnergyDiff;          ///< histogram of the diff. between best E and true neutrino E (for pure pop.)
    TH1D*       fCalCCEnergyDiff;     ///< histogram of the diff. between calorimetric neutrino E and true neutrino E (for pure pop.)
    TH1D*       fTrkQEEnergyDiff;     ///< histogram of the diff. between muon track length neutrino E and true neutrino E (for pure QE pop.)
    TH1D*       fTrkNonQEEnergyDiff;  ///< histogram of the diff. between muon track length neutrino E and true neutrino E (for pure NonQE pop.)
    TH1D*       fAngleQEEnergyDiff;   ///< histogram of the diff. between neutrino E using angle formula and true neutrino E (for pure QE pop.)
    
    TH2D*       fEnergy2D;            ///< histogram of best E and true neutrino E (for pure pop.)
    TH2D*       fCalCCEnergy2D;       ///< histogram of calorimetric neutrino E and true neutrino E (for pure pop.)
    TH2D*       fTrkQEEnergy2D;       ///< histogram of muon track length neutrino E and true neutrino E (for pure QE pop.)
    TH2D*       fTrkNonQEEnergy2D;    ///< histogram of muon track length neutrino E and true neutrino E (for pure NonQE pop.)
    TH2D*       fAngleQEEnergy2D;     ///< histogram of neutrino E using angle formula and true neutrino E (for pure QE pop.)
    
    TH1D*       fHadronCalEnergy;     ///< histogram of hadronic calorimetric energy NOT on the muon track
    TH1D*       fHadronTrkEnergy;     ///< histogram of hadronic calorimetric energy on the muon track
    
    TH1D*       fNDTrkLenActive;      ///< histogram of ND muon track length in the active region
    TH1D*       fNDTrkLenCatcher;     ///< histogram of ND muon track length in the muon catcher region
    TH1D*       fNDTrkCalActive;      ///< histogram of ND muon calorimetric energy in the active region
    TH1D*       fNDTrkCalTransition;  ///< histogram of ND muon calorimetric energy in the transition region
    TH1D*       fNDTrkCalCatcher;     ///< histogram of ND muon calorimetric energy in the muon catcher region
    TH1D*       fNDHadCalActive;      ///< histogram of ND hadron calorimetric energy in the active region
    TH1D*       fNDHadCalTransition;  ///< histogram of ND hadron calorimetric energy in the transition region
    TH1D*       fNDHadCalCatcher;     ///< histogram of ND hadron calorimetric energy in the muon catcher region
    
    double      fMaxTrkLen;           ///< length of detector from corner to corner    
    
    bool        fND;                  ///< is detector ND?
    bool        fFD;                  ///< is detector FD?
    
  };
  
  //-------------------------------------------------------------------
  NumuEAna::NumuEAna(fhicl::ParameterSet const &p):
    EDAnalyzer(p),
    fMaxTrkLen(-1),
    fND(false),
    fFD(false)
  {
    this->reconfigure(p);
  }
  
  //-------------------------------------------------------------------
  NumuEAna::~NumuEAna() 
  {
  }
  
  //-------------------------------------------------------------------
  void NumuEAna::analyze(art::Event const &e) 
  { 
    if(!(fND || fFD)){
      mf::LogWarning("NumuEAna") << "attempting to run on detector that is not "
                                 << "nd or fd, bail";
      return;
    }
    
    
    // get all hits in the event 
    art::Handle< std::vector<rb::CellHit> > hithdl;
    e.getByLabel(fHitModuleLabel, hithdl);
    
    
    art::PtrVector<rb::CellHit> allhits;
    for(size_t h = 0; h < hithdl->size(); ++h){
      art::Ptr<rb::CellHit> hit(hithdl, h);
      allhits.push_back(hit);
    }
    
    
    // get the tracks out of the event
    art::Handle<std::vector<rb::Track> > trkcol;
    e.getByLabel(fTrackModuleLabel,trkcol);
    
    //Load the slicer list from the event
    art::Handle<std::vector<rb::Cluster> > slicecol;
    e.getByLabel(fSlicerModuleLabel,slicecol);
    if(slicecol->empty()){
      mf::LogWarning ("No Slices")<<"No Slices in the input file";
      return;
    }
    
    art::PtrVector<rb::Cluster>  slicelist;
    for(unsigned int i = 0; i<slicecol->size();++i){
      art::Ptr<rb::Cluster> clust(slicecol,i);
      slicelist.push_back(clust);
    }
    
    art::ServiceHandle<cheat::BackTracker> bt;
    if(!bt->HaveTruthInfo()){
      mf::LogWarning("NumuEAna") << "attempting to run MC truth check on "
                                 << "things without truth, bail";
      return;
    }
    
    //Associating slices to tracks and energies
    art::FindManyP<rb::Track>   sliceToTracks(slicecol, e, fTrackModuleLabel);
    art::FindOneP<NumuE> sliceToEnergy(slicecol, e, fEnergyModuleLabel);
    
    //Loop over slices
    for(size_t s = 0; s < slicelist.size(); ++s){
      
      const art::Ptr<NumuE> sliceEnergy = sliceToEnergy.at(s);
      if (!sliceEnergy) continue;
      
      //Pulling out all info from the energy object and plotting it - these plots are without any cuts
      float BestEnergy    = sliceEnergy->E();
      float CalCCEnergy   = sliceEnergy->CalCCE();
      float QEEnergy      = sliceEnergy->TrkQEE();
      float NonQEEnergy   = sliceEnergy->TrkNonQEE();
      float AngleQEEnergy = sliceEnergy->AngleQEE();
      float AngleQEError  = sliceEnergy->AngleQEError();
      float HadCalEnergy  = sliceEnergy->HadCalE();
      float HadTrkEnergy  = sliceEnergy->HadTrkE();
      float ndTrkLenAct   = sliceEnergy->NDTrkLenAct();
      float ndTrkLenCat   = sliceEnergy->NDTrkLenCat();
      float ndTrkCalAct   = sliceEnergy->NDTrkCalAct();
      float ndTrkCalTran  = sliceEnergy->NDTrkCalTran();
      float ndTrkCalCat   = sliceEnergy->NDTrkCalCat();
      float ndHadCalAct   = sliceEnergy->NDHadCalAct();
      float ndHadCalTran  = sliceEnergy->NDHadCalTran();
      float ndHadCalCat   = sliceEnergy->NDHadCalCat();
      
      fEnergy->Fill(BestEnergy);
      fCalCCEnergy->Fill(CalCCEnergy);
      fTrkQEEnergy->Fill(QEEnergy);
      fTrkNonQEEnergy->Fill(NonQEEnergy);
      fAngleQEEnergy->Fill(AngleQEEnergy);
      fAngleQEError->Fill(AngleQEError);
      fHadronCalEnergy->Fill(HadCalEnergy);
      fHadronTrkEnergy->Fill(HadTrkEnergy);
      fNDTrkLenActive->Fill(ndTrkLenAct);
      fNDTrkLenCatcher->Fill(ndTrkLenCat);
      fNDTrkCalActive->Fill(ndTrkCalAct);
      fNDTrkCalTransition->Fill(ndTrkCalTran);
      fNDTrkCalCatcher->Fill(ndTrkCalCat);
      fNDHadCalActive->Fill(ndHadCalAct);
      fNDHadCalTransition->Fill(ndHadCalTran);
      fNDHadCalCatcher->Fill(ndHadCalCat);
      
      
      //Now start trying to make super pure samples with truth to fill resolution plots for neutrino energy  
      art::PtrVector< rb::CellHit > sliceHits;
      sliceHits = slicelist[s]->AllCells();
      
      std::vector<cheat::NeutrinoEffPur> funcReturn = bt->SliceToNeutrinoInteractions(sliceHits,allhits);
      if (funcReturn.empty()) continue;
      
      double funcPurity = funcReturn[0].purity;
      double funcEfficiency = funcReturn[0].efficiency;
      
      //Requirement that slice is mostly one neutrino interaction (only relevant for nd)
      if ((funcPurity < 0.8) || (funcEfficiency < 0.8)) continue;
      
      //Only using numu CC events      
      const simb::MCNeutrino& test_neutrino = funcReturn[0].neutrinoInt->GetNeutrino();
      if (!(test_neutrino.CCNC()==simb::kCC)||!(std::abs(test_neutrino.Nu().PdgCode()) == 14)) continue;
      
      //Checking if QE or not
      bool QEInt = false;
      if ((test_neutrino.InteractionType()==simb::kCCQE)) QEInt = true;
      
      //Determining information about the muon - does it have enough hits to make this a decent interaction?
      std::vector<const sim::Particle*> particleList = bt->MCTruthToParticles(funcReturn[0].neutrinoInt);
      std::set<int> neutrinoTrackIDs;
      for (unsigned int i = 0; i < particleList.size(); ++i){
        neutrinoTrackIDs.insert(particleList[i]->TrackId());
      }
      
      std::set<int>::iterator itr = neutrinoTrackIDs.begin();
      
      int goodMuon = -1;
      while( itr != neutrinoTrackIDs.end() ){
        
        int id_int = *itr;
        const sim::Particle* particle = bt->TrackIDToParticle(id_int); 
        int PDG = particle->PdgCode();
        
        if ((PDG != 13)&&(PDG != -13)){  
          ++itr;
          continue;
        }//End of loop over non-muons
        
        //Trying to find main muon for the event
        const sim::Particle* mother = bt->TrackIDToMotherParticle(id_int);
        if (mother->TrackId() != id_int){
          ++itr;
          continue;
        }//End of loop over muons who don't come from the neutrino
        
        bool passCuts = bt->PassMuonCuts(id_int,sliceHits);
        
        if (!passCuts){
          ++itr;
          continue;
        }//End of loop over muons who don't pass cuts
        
        goodMuon = id_int;
        break;      
      }//End of while loop over neutrino particles
      
      if (goodMuon == -1) continue;
      
      //Now we know we have an interaction of the correct type, passes muon cuts. We now loop over tracks.
      
      int    bestTrack  = -1;
      double trackLen   = 0.0;
      
      // Get all of the tracks in the slice
      const std::vector< art::Ptr<rb::Track> > sliceTracks = sliceToTracks.at(s);    
      if(sliceTracks.empty()) continue; 
      
      if (fND){
        
        for(size_t c = 0; c < sliceTracks.size(); ++c){
          
          geo::View_t view = sliceTracks[c]->View();       
          //Only use 3D tracks
          if (view != geo::kXorY){  
            continue;
          } //End of loop over 2d tracks
          
          if ((sliceTracks[c]->TotalLength() > trackLen)&&(sliceTracks[c]->TotalLength() < fMaxTrkLen)) {
            trackLen  = sliceTracks[c]->TotalLength();
            bestTrack = c;
          }
          
        }//End of for loop over tracks      
      }//End of best track finding for nd
      
      
      if (fFD){

        double bestPID    = 0.0;
        
        art::FindOneP<remid::ReMId> trackRemid(sliceTracks, e, fPIDModuleLabel);  
        
        for(size_t c = 0; c < sliceTracks.size(); ++c){  
          
          geo::View_t view = sliceTracks[c]->View();        
          //Only use 3D tracks
          if (view != geo::kXorY){  
            continue;
          } //End of loop over 2d tracks  
          
          if (trackRemid.isValid()){
            art::Ptr<remid::ReMId> remid = trackRemid.at(c);
            
            if ((sliceTracks[c]->TotalLength() > 0) && (sliceTracks[c]->TotalLength() < fMaxTrkLen) && (remid->Value() > bestPID)){
              trackLen = sliceTracks[c]->TotalLength();
              bestPID  = remid->Value();
              bestTrack = c;
            }//End of tracks with higher PID values
          }//End of valid remid object for track
        }//End of for loop over tracks
        
      }//End of best track finding for fd
      
      if(bestTrack == -1) continue;
      
      //Checking if reco muon track matches truth muon well enough
      bool trackMatchesMuon = false;
      std::set<int> muonTrackID;
      muonTrackID.insert(goodMuon);
      art::PtrVector< rb::CellHit > bestTrackHits;
      bestTrackHits = sliceTracks[bestTrack]->AllCells();  
      double trackMuonEffHits = bt->HitCollectionEfficiency(muonTrackID, bestTrackHits, allhits, geo::kXorY, 0, false);  
      double trackMuonPurHits = bt->HitCollectionPurity(muonTrackID, bestTrackHits, 0, 0, false); 
      
      if ((trackMuonEffHits > 0.8)&&(trackMuonPurHits > 0.8)) trackMatchesMuon = true;  
      
      if(!trackMatchesMuon) continue;
      
      //Determining if track is contained enough
      bool contained = false;  
            
      if (sliceTracks[bestTrack]->NCell()>0){
        
        double start_x_reco = sliceTracks[bestTrack]->Start().X();
        double start_y_reco = sliceTracks[bestTrack]->Start().Y();
        double start_z_reco = sliceTracks[bestTrack]->Start().Z();
        
        double end_x_reco   = sliceTracks[bestTrack]->Stop().X();
        double end_y_reco   = sliceTracks[bestTrack]->Stop().Y();
        double end_z_reco   = sliceTracks[bestTrack]->Stop().Z();
        
        if (fFD){
          if ((TMath::Abs(start_x_reco)<650.213)&&(TMath::Abs(start_y_reco)<650.213)&&(start_z_reco<5845.29)&&(start_z_reco>119.9156)){
            if ((TMath::Abs(end_x_reco)<750.213)&&(TMath::Abs(end_y_reco)<750.213)&&(end_z_reco<5945.29)&&(end_z_reco>19.9156)){
              contained = true;
            }
          } 
        }//End of loop for fd contained defnitions
        
        if (fND){
          if ((TMath::Abs(start_x_reco)<128.455)&&(TMath::Abs(start_y_reco)<128.455)&&(start_z_reco<1223.8)&&(start_z_reco>69.9156)){
            if ((TMath::Abs(end_x_reco)<178.455)&&(TMath::Abs(end_y_reco)<178.455)&&(end_z_reco<1493.445)&&(end_z_reco>19.9156)){
              contained = true;
            }
          }
        }//End of loop for nd contained definitions        
      }//End of loop over best tracks to find start/stop
      
      if (!contained) continue;
      
      //Filling remaining histograms with relevant super pure populations      
      double NuE = test_neutrino.Nu().E();
      bool energyRange = true;
      if ((NuE < 1.0) || (NuE > 3.0)) energyRange = false;
      
      if (energyRange){
        fEnergyDiff     ->Fill((BestEnergy-NuE)/NuE);
        fCalCCEnergyDiff->Fill((CalCCEnergy-NuE)/NuE);
      }
      fEnergy2D     ->Fill(BestEnergy,NuE);
      fCalCCEnergy2D->Fill(CalCCEnergy,NuE);
      
      if (QEInt){
        if (energyRange){
          fTrkQEEnergyDiff  ->Fill((QEEnergy-NuE)/NuE);
          fAngleQEEnergyDiff->Fill((AngleQEEnergy-NuE)/NuE);  
        }  
        fTrkQEEnergy2D  ->Fill(QEEnergy,NuE);
        fAngleQEEnergy2D->Fill(AngleQEEnergy,NuE);
      }//End of pure QE slices
      else{
        if (energyRange){
          fTrkNonQEEnergyDiff->Fill((NonQEEnergy-NuE)/NuE);
        }
        fTrkNonQEEnergy2D->Fill(NonQEEnergy,NuE);
      }//End of pure NonQE slices
      
    }//End of loop over slices
    return;
  }//End of analyze
  
  //-------------------------------------------------------------------
  void NumuEAna::reconfigure(fhicl::ParameterSet const & p) 
  {
    fHitModuleLabel     = p.get< std::string >("HitModuleLabel"    );
    fTrackModuleLabel   = p.get< std::string >("TrackModuleLabel"  );
    fSlicerModuleLabel  = p.get< std::string >("SlicerModuleLabel" );
    fEnergyModuleLabel  = p.get< std::string >("EnergyModuleLabel" );
    fPIDModuleLabel     = p.get< std::string >("PIDModuleLabel"    );
  }
  
  //-------------------------------------------------------------------
  void NumuEAna::beginJob() 
  {
    art::ServiceHandle<art::TFileService> tfs;
    
    fEnergy             = tfs->make<TH1D>("Energy",             ";Reco Neutrino Energy (GeV);Slices",200, -0.1, 10);
    fCalCCEnergy        = tfs->make<TH1D>("CalCCEnergy",        ";Reco Neutrino Energy (GeV);Slices",200, -0.1, 10);
    fTrkQEEnergy        = tfs->make<TH1D>("TrkQEEnergy",        ";Reco Neutrino Energy (GeV);Slices",200, -0.1, 10);
    fTrkNonQEEnergy     = tfs->make<TH1D>("TrkNonQEEnergy",     ";Reco Neutrino Energy (GeV);Slices",200, -0.1, 10);
    fAngleQEEnergy      = tfs->make<TH1D>("AngleQEEnergy",      ";Reco Neutrino Energy (GeV);Slices",200, -0.1, 10);
    
    fAngleQEError       = tfs->make<TH1D>("AngleQEError",      ";Error on Neutrino Energy;Slices",200, -0.1, 5.0);
    
    fEnergyDiff         = tfs->make<TH1D>("EnergyDiff",         ";(Reco - Neutrino Energy)/Neutrino Energy;Slices",500, -3.0, 3.0);
    fCalCCEnergyDiff    = tfs->make<TH1D>("CalCCEnergyDiff",    ";(Reco - Neutrino Energy)/Neutrino Energy;Slices",500, -3.0, 3.0);
    fTrkQEEnergyDiff    = tfs->make<TH1D>("TrkQEEnergyDiff",    ";(Reco - Neutrino Energy)/Neutrino Energy;Slices",500, -3.0, 3.0);
    fTrkNonQEEnergyDiff = tfs->make<TH1D>("TrkNonQEEnergyDiff", ";(Reco - Neutrino Energy)/Neutrino Energy;Slices",500, -3.0, 3.0);
    fAngleQEEnergyDiff  = tfs->make<TH1D>("AngleQEEnergyDiff",  ";(Reco - Neutrino Energy)/Neutrino Energy;Slices",500, -3.0, 3.0);
    
    fEnergy2D           = tfs->make<TH2D>("Energy2D",           ";Reco Energy (GeV);Neutrino Energy (GeV)",200, 0., 20, 200, 0., 20);
    fCalCCEnergy2D      = tfs->make<TH2D>("CalCCEnergy2D",      ";Reco Energy (GeV);Neutrino Energy (GeV)",200, 0., 20, 200, 0., 20);
    fTrkQEEnergy2D      = tfs->make<TH2D>("TrkQEEnergy2D",      ";Reco Energy (GeV);Neutrino Energy (GeV)",200, 0., 20, 200, 0., 20);
    fTrkNonQEEnergy2D   = tfs->make<TH2D>("TrkNonQEEnergy2D",   ";Reco Energy (GeV);Neutrino Energy (GeV)",200, 0., 20, 200, 0., 20);
    fAngleQEEnergy2D    = tfs->make<TH2D>("AngleQEEnergy2D",    ";Reco Energy (GeV);Neutrino Energy (GeV)",200, 0., 20, 200, 0., 20);
    
    fHadronCalEnergy    = tfs->make<TH1D>("HadronCalEnergy",    ";Reco Hadron Energy (GeV);Slices",200, -0.1, 10);
    fHadronTrkEnergy    = tfs->make<TH1D>("HadronTrkEnergy",    ";Track Hadron Energy (GeV);Slices",200, -0.1, 5);
    
    fNDTrkLenActive     = tfs->make<TH1D>("NDTrkLenActive",     ";Track Length (cm) in Active Region;Slices",200, 0., 1400);  
    fNDTrkLenCatcher    = tfs->make<TH1D>("NDTrkLenCatcher",    ";Track Length (cm) in Catcher Region;Slices",200, 0., 800);
    fNDTrkCalActive     = tfs->make<TH1D>("NDTrkCalActive",     ";Track Energy (GeV) in Active Region;Slices",200, 0., 10);
    fNDTrkCalTransition = tfs->make<TH1D>("NDTrkCalTransition", ";Track Energy (GeV) in Tran. Plane;Slices",200, 0., 0.5);
    fNDTrkCalCatcher    = tfs->make<TH1D>("NDTrkCalCatcher",    ";Track Energy (GeV) in Catcher Region;Slices",200, 0., 2);
    fNDHadCalActive     = tfs->make<TH1D>("NDHadCalActive",     ";Hadron Energy (GeV) in Active Region;Slices",200, 0., 10);
    fNDHadCalTransition = tfs->make<TH1D>("NDHadCalTransition", ";Hadron Energy (GeV) in Tran. Plane;Slices",200, 0., 0.5);
    fNDHadCalCatcher    = tfs->make<TH1D>("NDHadCalCatcher",  ";Hadron Energy (GeV) in Catcher Region;Slices",200, 0., 2);
    
  }//End of begin Run
  
  //-------------------------------------------------------------------
  void NumuEAna::beginRun(art::Run const &r) 
  {
    art::ServiceHandle<geo::Geometry>  geom;    
    double detHalfWidth   =  geom->DetHalfWidth();
    double detHalfHeight  =  geom->DetHalfHeight();
    double detLength      =  geom->DetLength();
    
    fMaxTrkLen = util::pythag(detHalfWidth*2, detHalfHeight*2, detLength);
    
    fND = false;
    fFD = false;
    
    novadaq::cnv::DetId detID = geom->DetId();
    if (detID == novadaq::cnv::kNEARDET) fND = true;
    if (detID == novadaq::cnv::kFARDET)  fFD = true;
    
  }//End of beginRun
  
  //----------------------------------------------------------
  DEFINE_ART_MODULE(NumuEAna);
  
}//End of namespace
