#include "NumuSandbox/NumuSandFxs.h"

#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/Geo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/RecoHit.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/Cluster.h"

#include "TMath.h"
#include "TVector3.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"

#define PI 3.14159265

namespace numusand
{

  //-------------------------------------------------------
  NumuSandFxs::NumuSandFxs() //constructor
  {
    fKGeo = new trk::KalmanGeoHelper(true);
  }

  //-------------------------------------------------------
  NumuSandFxs::~NumuSandFxs() //destructor
  {
  }

  //-------------------------------------------------------
  // getAngle: function to get the angle of the lepton wrt the incoming neutrino.
  // currently the incoming neutrino direction estimate is hardcoded.

  float NumuSandFxs::getAngle(TVector3 trackdir)
  {
    art::ServiceHandle<geo::Geometry> geom;

    TVector3 beamdir = geom->NuMIBeamDirection();
    float    ret     = trackdir*beamdir;

    return ret;
  }

  //-------------------------------------------------------
  // getERemain: function to get the 'remaining energy' (slice E - best track E)
  // with a consistent calibration.  All hits will be calibrated with the w of clus 1

  float NumuSandFxs::getERemain(const rb::Cluster& clus1, const rb::Cluster& clus2)
  {
    art::ServiceHandle<calib::Calibrator> cal;

    const double wx = (clus1.NXCell() > 0) ? clus1.W(clus1.XCell(0).get()) : 0;
    const double wy = (clus1.NYCell() > 0) ? clus1.W(clus1.YCell(0).get()) : 0;

    double E1 = 0;
    double E2 = 0;

    for(unsigned int hit = 0; hit < clus1.NCell(); ++hit) {
      const art::Ptr<rb::CellHit>& chit = clus1.Cell(hit);
      rb::CellHit chit2 = *chit;
      chit2.SetMC();
      const rb::RecoHit rhit(cal->MakeRecoHit(chit2, 
					      chit->View() == geo::kX ? wx : wy));
      if(rhit.IsCalibrated()) {
        E1+=rhit.GeV();
      }
    }

    for(unsigned int hit = 0; hit < clus2.NCell(); ++hit) {
      const art::Ptr<rb::CellHit>& chit = clus2.Cell(hit);
      rb::CellHit chit2 = *chit;
      chit2.SetMC();
      const rb::RecoHit rhit(cal->MakeRecoHit(chit2, 
					      chit->View() == geo::kX ? wx : wy));
      if(rhit.IsCalibrated()) {
        E2+=rhit.GeV();
      }
    }

    return fabs(E1-E2);

  }

  //-------------------------------------------------------
  void NumuSandFxs::getContainment(art::Ptr<rb::Track> track, float &vtxcontdist, float &mucontdist)
  {
    mucontdist = 0.0;
    vtxcontdist = 0.0;
    double xmin,xmax,ymin,ymax,zmin,zmax;

    // define the boundaries of containment
    art::ServiceHandle<geo::Geometry>  geom;    

    double detHalfWidth  = geom->DetHalfWidth();
    double detHalfHeight = geom->DetHalfHeight();
    double detLength     = geom->DetLength();

    // fd detector edges determined from event display for 28 block detector
    // get rid of offset between geom->DetHalfWidth()(Height()) and actual detector edges
    double voffset = 15.0;
    xmin = -detHalfWidth+voffset;
    xmax = detHalfWidth-voffset;
    ymin = -detHalfHeight+voffset;
    ymax = detHalfHeight-voffset;
    zmin = 0;
    zmax = detLength;

    TVector3 mins(xmin,ymin,zmin);
    TVector3 maxs(xmax,ymax,zmax);

    float mindist = 1e10;

    TVector3 start = track->Start();
    TVector3 stop = track->Stop();
    if(start.X() > xmin && start.X() < xmax && 
       start.Y() > ymin && start.Y() < ymax && 
       start.Z() > zmin && start.Z() < zmax){
      TVector3 mindiffs = mins - start;
      TVector3 maxdiffs = maxs - start;
      if(TMath::Abs(mindiffs.X()) < mindist){ mindist = TMath::Abs(mindiffs.X()); }
      if(TMath::Abs(mindiffs.Y()) < mindist){ mindist = TMath::Abs(mindiffs.Y()); }
      if(TMath::Abs(mindiffs.Z()) < mindist){ mindist = TMath::Abs(mindiffs.Z()); }
      if(TMath::Abs(maxdiffs.X()) < mindist){ mindist = TMath::Abs(maxdiffs.X()); }
      if(TMath::Abs(maxdiffs.Y()) < mindist){ mindist = TMath::Abs(maxdiffs.Y()); }
      if(TMath::Abs(maxdiffs.Z()) < mindist){ mindist = TMath::Abs(maxdiffs.Z()); }
      vtxcontdist = mindist;
    }

    // check the distance from the end of the track to the detector edge. If stop point is outside the detector, return 0
    if(stop.X() > maxs.X() || stop.X()< mins.X() ||
       stop.Y() > maxs.Y() || stop.Y()< mins.Y() ||
       stop.Z() > maxs.Z() || stop.Z()< mins.Z()){vtxcontdist =  0.0; }

    // get the end point and direction of the track end
    if(track->NTrajectoryPoints() < 2){ return;}
    double end[3] = {stop.X(), stop.Y(), stop.Z()};
    TVector3 traj = track->TrajectoryPoint(track->NTrajectoryPoints()-2);
    TVector3 enddir = (stop-traj).Unit();
    double dir[3] = {enddir.X(), enddir.Y(), enddir.Z()};
    // ensure that 2d tracks have 0 elements in stop point and direction of view with no info
    if(track->View() == geo::kX){
      end[1] = 0.0;
      dir[1] = 0.0;
    }
    else if(track->View() == geo::kY){
      end[0] = 0.0;
      dir[0] = 0.0;
    }

    // get the detector location that will be intersected
    double edge[3];
    // if the end in outside of the detector return zero for muon containment
    if(end[0] > detHalfWidth || end[0] < -detHalfWidth || 
       end[1] > detHalfHeight || end[1] < -detHalfHeight ||
       end[2] > detLength || end[2] < 0){ return; }
    geo::ProjectToBoxEdge(end,dir,-detHalfWidth,detHalfWidth,-detHalfHeight,detHalfHeight,0.0,detLength,edge);

    // get the length of active material between the track stop and the edge it intersects
    std::vector<geo::OfflineChan> missCells;
    std::vector<double> dists;
    if(track->View() == geo::kX){ fKGeo->CountMissedCellsOnLine(geo::kX,end[0],end[2],edge[0],edge[2],missCells,dists); }
    else if(track->View() == geo::kY){ fKGeo->CountMissedCellsOnLine(geo::kY,end[1],end[2],edge[1],edge[2],missCells,dists); }
    else{
      fKGeo->CountMissedCellsOnLine(geo::kX,end[0],end[2],edge[0],edge[2],missCells,dists); 
      fKGeo->CountMissedCellsOnLine(geo::kY,end[1],end[2],edge[1],edge[2],missCells,dists); 
    }
    
    float dist = 0.0;
    // count the active distance
    for(unsigned int i = 0; i < dists.size(); ++i){
      dist+=((float)dists[i]);
    }

    mucontdist = dist;
  }


  //-------------------------------------------------------
  // getCosmicVars: fill variables for distance and # cells from track end projected to detector wall.
  // assume detector dimensions are -762 to +762 cm in x and y, 0 to 5960 cm in z (based on looking at MC event displays)

  void NumuSandFxs::getCosmicVars(const art::Ptr<rb::Track> track, double &fwddist, double &bakdist, int &fwdcell, int &bakcell)
  {

    //    double fwddist, bakdist, fwdcell, bakcell;

    art::ServiceHandle<geo::Geometry>  geom;

    std::vector< geo::OfflineChan> xhits, yhits, xhits2, yhits2;
    TVector3 start = track->Start();
    TVector3 end = track->Stop();
    TVector3 dirfor = track->Dir().Unit();
    TVector3 dirbak = -dirfor;  // future: use last two trajectory points for end direction.
    TVector3 temp;
    double point[3], dir[3], out[3];

    end.GetXYZ(point);  // starting at track end point projected forwards along track dir = fwd
    dirfor.GetXYZ(dir);

    if(fabs(point[0])>762||fabs(point[1])>762||point[2]<0||point[2]>5960) { // if track end is outside of detector, consider it uncontained
      fwddist = -111;
      fwdcell = -111;
    }
    else {
      geo::ProjectToBoxEdge(point,dir,-762,762,-762,762,0,5960,out);
      temp.SetXYZ(out[0],out[1],out[2]);
      fwddist = (end-out).Mag();
      geom->CountCellsOnLineFast(point[0],point[1],point[2],out[0],out[1],out[2],xhits,yhits);
      fwdcell = (xhits.size()+yhits.size());
    }

    start.GetXYZ(point);  // starting at track start point projected backwards along track dir = bak
    dirbak.GetXYZ(dir);

    if(fabs(point[0])>762||fabs(point[1])>762||point[2]<0||point[2]>5960) {  // if track start is outside detector, consider it uncontained
      bakdist = -111;
      bakcell = -111;
    }
    else {
      geo::ProjectToBoxEdge(point,dir,-762,762,-762,762,0,5960,out);
      temp.SetXYZ(out[0],out[1],out[2]);
      bakdist = (start-out).Mag();
      geom->CountCellsOnLineFast(point[0],point[1],point[2],out[0],out[1],out[2],xhits2,yhits2); // use x/yhits2 here to avoid reinitializing; the output of this function appends to vector
      bakcell = (xhits2.size()+yhits2.size());
    }

    //    dist = std::min(bakdist,fwddist);
    //    cell = std::min(bakcell,fwdcell);

  } // end getCosmicVars

  //------------------------------------------------------------------------------

  void NumuSandFxs::getFits(const art::Ptr<rb::Track> track, double &slope, double &chisq, double &chisqdiff)
  {
    double xyz[3], getx, gety, getz, intercept;
    std::vector<double> zs, ts, qs; // vectors for fit: distance along track, time, weight
    TVector3 tdir = track->Dir();
    TVector3 tstart = track->Start();
    TVector3 tend = track->Stop();

    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<calib::Calibrator> cal;

    const double wx = (track->NXCell() > 0) ? track->W(track->XCell(0).get()) : 0;
    const double wy = (track->NYCell() > 0) ? track->W(track->YCell(0).get()) : 0;

    for(unsigned int hitIdx = 0; hitIdx < track->NCell(); ++hitIdx){
      art::Ptr<rb::CellHit> chit = track->Cell(hitIdx);
      rb::CellHit chit2 = *chit;
      chit2.SetMC();
      geom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);
      const rb::RecoHit rhit(cal->MakeRecoHit(chit2, 
					      chit->View() == geo::kX ? wx : wy));
      if(!chit->GoodTiming()) {
        continue;
      }

      if(chit->View()==geo::kX) { // each view is different

        double px, pz = 1;  // make a direction vector perpindicular to track in 2d
        px = -tdir.Z()/tdir.X();
        double norm = sqrt(px*px+pz*pz);
        px = px/norm;
        pz = pz/norm;

        geo::LineIntersection(tstart.X(), tstart.Z(), tend.X(), tend.Z(), xyz[0], xyz[2],
                              xyz[0]+50*px, xyz[2]+50*pz, getx, getz);  // find intersection with track
        zs.push_back((getz-tstart.Z())/tdir.Z());  // how far along the track is in 3d
      }

      if(chit->View()==geo::kY) {

        double py, pz = 1;  // make a direction vector perindicular to track in 2d
        py = -tdir.Z()/tdir.Y();
        double norm = sqrt(py*py+pz*pz);
        py = py/norm;
        pz = pz/norm;

        geo::LineIntersection(tstart.Y(), tstart.Z(), tend.Y(), tend.Z(), xyz[1], xyz[2],
                              xyz[1]+50*py, xyz[2]+50*pz, gety, getz);  // find intersection with track
        zs.push_back((getz-tstart.Z())/tdir.Z());
      }

      ts.push_back(rhit.T());
      double par1 = -0.377*rhit.PECorr();
      double par2 = 0.000772*rhit.PECorr()*rhit.PECorr();
      double par3 = -7.2e-7*rhit.PECorr()*rhit.PECorr()*rhit.PECorr();
      double par4 = 2.486e-10*rhit.PECorr()*rhit.PECorr()*rhit.PECorr()*rhit.PECorr();
      double error = 87.3+par1+par2+par3+par4;  //timing resolution (ns), fit from timing res plot from Evan
      if(error<1) error = 1; //keep 0 < 1/error^2 < 1
      qs.push_back(1/(error*error));  // weight by timing resolution;
    }

    if(zs.size()>1) {
      slope = fitslope(zs, ts, qs, chisq, intercept, 1);
      chisqdiff = fitchisq(zs, ts, 0, 30.0) - fitchisq(zs, ts, 1, 30.0);
    }
    else {
      slope = -5;
      chisqdiff = -5;
      chisq = -5;
    }

  }

  //---------------------------------------------------------------------------

  float NumuSandFxs::fitslope(const std::vector< double > &l, const std::vector< double > &t, const std::vector< double > &q, double &chisq, double &inter, bool linflag) {

    assert(t.size()==l.size());
    assert(t.size()>=2);

    double ztStart[2] = {0.};
    double ztEnd[2]   = {0.};
    if(linflag==false) {
      chisq = geo::LinFit(t, l, q,
                          &ztStart[0], &ztStart[1],
                          &ztEnd[0], &ztEnd[1]);
    }
    else if(linflag==true) {
      chisq = geo::LinFitMinDperp(t, l, q,
                                  &ztStart[0], &ztStart[1],
                                  &ztEnd[0], &ztEnd[1]);
    }
    const double ztSlope = (ztEnd[1]-ztStart[1])/(ztEnd[0]-ztStart[0]);

    inter = ztStart[1]-(ztSlope*ztStart[0]);

    return ztSlope;
  }

  //----------------------------------------------------------------------------------------

  float NumuSandFxs::fitchisq(const std::vector< double > &t, const std::vector< double > &l, int flag, double c) {

    assert(t.size()==l.size());
    assert(t.size()>=2);

    double chi2 = 0.0;
    double tave = 0;
    double lave = 0;

    for (unsigned int i=1; i<t.size(); ++i) {
      tave += t[i];
      lave += l[i];
    }

    double norm = t.size();

    tave = tave / norm;
    lave = lave / norm;

    double tstart = tave - 500;
    double tend = tave + 500;
    double lstart, lend;

    assert(flag == 0 || flag == 1);
    if(flag==0) {
      lstart = lave - (c * 500);
      lend = lave + (c * 500);
    }
    else {
      lstart = lave + (c * 500);
      lend = lave - (c * 500);
    }

    for (unsigned int i=0; i<t.size(); ++i) {
      chi2 += geo::DsqrToLine(t[i],l[i],tstart,lstart,tend,lend);
    }
    return chi2;
  }

  //----------------------------------------------------------------------------------

  float NumuSandFxs::getScatt(const art::Ptr<rb::Track> track) {
    double tscattmax=0, tscatttot=0, tscatt;
    TVector3 traj1, traj2, dir, olddir;

    for(unsigned int p=0;p<track->NTrajectoryPoints()-1;p++) {
      traj1 = track->TrajectoryPoint(p);
      traj2 = track->TrajectoryPoint(p+1);
      dir = (traj2-traj1).Unit();
      if((dir-olddir).Mag()<1e-7) continue;
      if(p>0&&dir.Mag()<=1.0&&olddir.Mag()<=1.0) {
        tscatt = std::acos(dir.Dot(olddir)) * 180 / PI;
        if(tscatt > tscattmax) tscattmax = tscatt;
        tscatttot += tscatt;
      }
      olddir = dir;
    }
    return tscatttot;  // calculated both maximum and total scatter, total seems more useful, ignore max for now
  }

  //-----------------------------------------------------------------------------------

  void NumuSandFxs::getActivity(const art::Ptr<rb::Track> track, double &startact, double &endact) {
    double xyz[3];
    TVector3 tstart = track->Start();
    TVector3 tend = track->Stop();
    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<calib::Calibrator> cal;

    startact = 0;
    endact = 0;

    const double wx = (track->NXCell() > 0) ? track->W(track->XCell(0).get()) : 0;
    const double wy = (track->NYCell() > 0) ? track->W(track->YCell(0).get()) : 0;

    for(unsigned int hitIdx = 0; hitIdx < track->NCell(); ++hitIdx){
      art::Ptr<rb::CellHit> chit = track->Cell(hitIdx);
      geom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);
      rb::CellHit chit2 = *chit;
      chit2.SetMC();
      const rb::RecoHit rhit(cal->MakeRecoHit(chit2, 
					      chit->View() == geo::kX ? wx : wy));
      if(chit->View()==geo::kX) { // ignore y position of cell center
        if(sqrt((tstart.X()-xyz[0])*(tstart.X()-xyz[0]) + (tstart.Z()-xyz[2])*(tstart.Z()-xyz[2])) < 50) startact+=rhit.GeV();
        if(sqrt((tend.X()-xyz[0])*(tend.X()-xyz[0]) + (tend.Z()-xyz[2])*(tend.Z()-xyz[2])) < 50) endact+=rhit.GeV();
      }

      if(chit->View()==geo::kY) { // ignore x position of cell center
        if(sqrt((tstart.Y()-xyz[1])*(tstart.Y()-xyz[1]) + (tstart.Z()-xyz[2])*(tstart.Z()-xyz[2])) < 50) startact+=rhit.GeV();
        if(sqrt((tend.Y()-xyz[1])*(tend.Y()-xyz[1]) + (tend.Z()-xyz[2])*(tend.Z()-xyz[2])) < 50) endact+=rhit.GeV();
      }
    }
  }

  //--------------------------------------------------------------------------- start get missing energy

  bool NumuSandFxs::getMissingE(const art::Ptr<rb::Cluster> slice, float &missE1, float &missE2)
  {

    art::ServiceHandle<cheat::BackTracker> bt;

    if(!bt->HaveTruthInfo()) return false; // No Truth Info!

    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;

    art::PtrVector<rb::CellHit> allhits = slice->AllCells();

    const std::vector<const sim::Particle*> parts = bt->HitsToParticle(allhits);
    //    const std::vector< cheat::TrackIDE > trids = bt->HitsToTrackIDE(allhits);

    float endZ, sum1=0;
    endZ = livegeom->InstrumentedDetBack();

    if(parts.size()==0) return false;

    for(unsigned int i = 0; i < parts.size(); i++) { // loop over particles
      if(!(abs(parts[i]->PdgCode())==11 || abs(parts[i]->PdgCode())==13 || parts[i]->PdgCode()==2212 ||  abs(parts[i]->PdgCode())==211 ||  parts[i]->PdgCode()==22)) continue; // not interesting particle
      if(parts[i]->E() < 0.001) continue;  // ignore things < 1 MeV
      if(fabs(parts[i]->Vx())>762. || fabs(parts[i]->Vy())>762. || parts[i]->Vz()<livegeom->InstrumentedDetFront() || parts[i]->Vz()>endZ) { // particle doesn't start in detector
        sum1+=parts[i]->E();
        continue;
      }
      if(fabs(parts[i]->EndX())<762. && fabs(parts[i]->EndY())<762. && parts[i]->EndZ()>livegeom->InstrumentedDetFront() && parts[i]->EndZ()<endZ) continue; // only care about exiting
      sum1+=parts[i]->E();
      //      for(unsigned int j = 0; j < parts[i].Trajectory().  // TODO loop through trajectory points to see when leaves det and get E missing only
    }

    missE1 = sum1;
    return true;

  }
  //---------------------------------------------------------------- end missing energy, start new containment

  bool NumuSandFxs::getNewContain(const art::Ptr<rb::Cluster> slice, int &bbc, float &dfe2, float &dfe3, float &dfeY2, float &dfeY3, float &dfeth1, float &dfeth2, float &dfeYth1, float &dfeYth2) {

    float fMinXFid = -700;
    float fMinYFid = -700;
    float fMaxXFid = 700;
    float fMaxYFid = 660;

    int stot[6] = {0};

    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;
    art::ServiceHandle<cheat::BackTracker> bt;

    double dl;
    if(geom->DetId() == novadaq::cnv::kFARDET) dl = livegeom->InstrumentedDetBack();
    else dl = geom->DetLength(); // ND
    float fMaxZFid = dl - 50;

    float fMinZFid;
    if(geom->DetId() == novadaq::cnv::kFARDET) fMinZFid = livegeom->InstrumentedDetFront() + 50;
    else fMinZFid = 50; // ND

    float y1=9999, y2=9999, y3=9999, yt;
    float o1=9999, o2=9999, o3=9999, ot;
    float yt1=9999, yt2=9999;
    float ot1=9999, ot2=9999;

    //loop over all hits
    for(unsigned int j = 0; j < slice->NCell(); ++j) {
      const rb::CellHit* h = slice->Cell(j).get();

      double xyz[3];
      geom->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
      const double x = xyz[0];
      const double y = xyz[1];
      const double z = xyz[2];

      if(h->View()==geo::kX) {
        ot = std::min(762-fabs(x), std::min(z-3,5959-z));
        yt = 9999;
      }

      else {
        ot = std::min(y+762.0, std::min(z-3,5959-z));
        yt = 762.0-y;
      }

      if(ot < o1) {
        o3 = o2;
        o2 = o1;
        o1 = ot;
      }
      else if(ot < o2) {
        o3 = o2;
        o2 = ot;
      }
      else if(ot < o3) o3 = ot;

      if(yt < y1) {
        y3 = y2;
        y2 = y1;
        y1 = yt;
      }
      else if(yt < y2) {
        y3 = y2;
        y2 = yt;
      }
      else if(yt < y3) y3 = yt;

      if(h->PE() > 100) {

        if(ot < ot1) {
          ot2 = ot1;
          ot1 = ot;
        }
        else if(ot < ot2) ot2 = ot;

        if(yt < yt1) {
          yt2 = yt1;
          yt1 = yt;
        }
        else if(yt < yt2) yt2 = yt;

      }

      //Original BBC
      // Total up number of cells > Max (X/Y/Z) and < Min (X/Y/Z)
      //TOP
      if(y > fMaxYFid) {
        stot[0]++;
        continue;
      }
      //BOTTOM
      if(y < fMinYFid) {
        stot[1]++;
        continue;
      }
      //LEFT
      if(x > fMaxXFid) {
        stot[2]++;
        continue;
      }
      //RIGHT
      if(x < fMinXFid) {
        stot[3]++;
        continue;
      }
      //FRONT
      if(z < fMinZFid) {
        stot[4]++;
        continue;
      }
      //BACK
      if(z > fMaxZFid) {
        stot[5]++;
        continue;
      }

    } // end loop over hits

    bbc = *std::max_element(stot,stot+6);
    dfe2 = o2;
    dfe3 = o3;
    dfeY2 = y2;
    dfeY3 = y3;
    dfeth1 = ot1;
    dfeth2 = ot2;
    dfeYth1 = yt1;
    dfeYth2 = yt2;

    return true;

  } // end getNewContain

} //end namespace
