///
/// \file  FEBRateAnalysis.cxx
/// \brief Summarize FEB rates produced by the online monitoring
///

#include "DAQChannelMap/DAQChannelMap.h"
#include "DAQChannelMap/HardwareDisplay.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "OnlineMonitoring/ana/FEBRateAnalysis.h"
#include "OnlineMonitoring/ana/RateSummaryNt.h"
#include <iostream>
#include "TFile.h"
#include "TH1.h"
#include "TH2F.h"
#include "TTree.h"

using namespace om;


FEBRateAnalysis::FEBRateAnalysis(std::string det) :
  fHardwareMapping(new daqchannelmap::HardwareDisplay)
{
  if(det == "fd"   || det == "FD" || det == "FarDet"  || det == "far")
    fHardwareMapping->SetupDet(novadaq::cnv::kFARDET);
  if(det == "nd"   || det == "ND" || det == "NearDet" || det == "near")
    fHardwareMapping->SetupDet(novadaq::cnv::kNEARDET);
  if(det == "ndos" || det == "NDOS")
    fHardwareMapping->SetupDet(novadaq::cnv::kNDOS);
}

//......................................................................

void FEBRateAnalysis::AnaFile(const TFile& f, int run, int subrun)
{
  std::cout << "FEBRateAnalysis " 
	    << run << ":"
	    << subrun << " " 
	    << f.GetName() << std::endl;

  //  Open output file for this subrun
  char filename[100];
  if(fHardwareMapping->fDetId == novadaq::cnv::kFARDET)
    sprintf(filename, "FarDet-Summary-onmon_ana-subrun-%.2u-%.2u.root", run, subrun);
  if(fHardwareMapping->fDetId == novadaq::cnv::kNEARDET)
    sprintf(filename, "NearDet-Summary-onmon_ana-subrun-%.2u-%.2u.root", run, subrun);
  if(fHardwareMapping->fDetId == novadaq::cnv::kNDOS)
    sprintf(filename, "NDOS-Summary-onmon_ana-subrun-%.2u-%.2u.root", run, subrun);
  TFile* out_file = new TFile(filename, "RECREATE");


  // Summary tree
  fNt = new RateSummaryNt;

  fRatesTree = new TTree("OnMonSummary","OnMon Summary Tree");
  fRatesTree->Branch("fRateSummaryNt", "om::RateSummaryNt", &fNt, 1600, 99);

  // Copy the histos of interest into these temp ones
  TH2F* fcPixelsRateDCM[PixelsDCM_size][PixelsDCM_size];
  TH2F* fcHighADCPixelsRateDCM[PixelsDCM_size][PixelsDCM_size];
  TH2F* fcLowADCPixelsRateDCM[PixelsDCM_size][PixelsDCM_size];
  TH2F* fcMipADCPixelsRateDCM[PixelsDCM_size][PixelsDCM_size];
  TH2F* fcMinADCPixelsDCM[PixelsDCM_size][PixelsDCM_size];


  for(unsigned int i = 0; i < PixelsDCM_size; ++i) {
    for(unsigned int j = 0; j < PixelsDCM_size; ++j) {
      char histoname[30];
      sprintf(histoname, "PixelsRateDCM_%.2u_%.2u", i, j);
      fcPixelsRateDCM[i][j]    =  (TH2F*)f.FindObjectAny(histoname);
      if(fcPixelsRateDCM[i][j]!=0){
	char histonameH[30], histonameL[30], histonameM[30], histonameMIN[30];
	sprintf(histonameH, "HighADCPixelsRateDCM_%.2u_%.2u", i, j);
	sprintf(histonameL, "LowADCPixelsRateDCM_%.2u_%.2u", i, j);
	sprintf(histonameM, "MipADCPixelsRateDCM_%.2u_%.2u", i, j);
	sprintf(histonameMIN, "MinADCPixelsDCM_%.2u_%.2u", i, j);
	fcHighADCPixelsRateDCM[i][j]   =  (TH2F*)f.FindObjectAny(histonameH);
	fcLowADCPixelsRateDCM[i][j]    =  (TH2F*)f.FindObjectAny(histonameL);
	fcMipADCPixelsRateDCM[i][j]    =  (TH2F*)f.FindObjectAny(histonameM);
	fcMinADCPixelsDCM[i][j]        =  (TH2F*)f.FindObjectAny(histonameMIN);
	this->StoreContent(fcPixelsRateDCM[i][j], fcHighADCPixelsRateDCM[i][j], fcLowADCPixelsRateDCM[i][j], 
			   fcMipADCPixelsRateDCM[i][j], fcMinADCPixelsDCM[i][j], run, subrun, (int)i, (int)j);
      }
    } // end for j
  } // end for i (for every pixelDCM histo)

  fRatesTree->Write();
  out_file->Close();
}
//......................................................................
void FEBRateAnalysis::StoreContent(TH2F*& histPix, TH2F*& histPixHigh, TH2F*& histPixLow, 
				   TH2F*& histPixMip, TH2F*& histPixMin, int run, int subrun, int db, int dcm)
{
  // Stores histo content in summary tree

  bool existsL=0, existsH=0, existsM=0; 
  if(histPixHigh!=0) existsH=1;
  if(histPixLow!=0)  existsL=1;
  if(histPixMip!=0)  existsM=1;

  int NbinsX = histPix->GetNbinsX();
  int NbinsY = histPix->GetNbinsY();

  for(int i=1;i<NbinsX+1;i++){
    for(int j=1;j<NbinsY+1;j++){
      double x = histPix->GetXaxis()->GetBinCenter(i);
      double y = histPix->GetYaxis()->GetBinCenter(j);
      unsigned int NOTthisDB, NOTthisDCM, thisFEB, thisPIX, thisX, thisY;
      unsigned int thisBLK, thisPINB, thisVIEW, thisMOD, thisCELL;
      double contentH=0, contentL=0, contentM=0;
      fHardwareMapping->XYToDAQ(&NOTthisDB,&NOTthisDCM,&thisFEB,
				&thisPIX,(unsigned int)x,(unsigned int)y);
      fHardwareMapping->DAQToXY((unsigned int)db, (unsigned int)dcm,
				thisFEB, thisPIX, &thisX, &thisY);
      fHardwareMapping->DAQToPlaneInBlockModCell(&thisBLK,&thisPINB,&thisVIEW,&thisMOD,&thisCELL,
						 (unsigned int)db, (unsigned int)dcm, thisFEB, thisPIX);


      double content    = histPix->GetBinContent(i,j);
      double contentMin = histPixMin->GetBinContent(i,j);
      if(existsH) contentH = histPixHigh->GetBinContent(i,j);
      if(existsL) contentL = histPixLow->GetBinContent(i,j);
      if(existsM) contentM = histPixMip->GetBinContent(i,j);

      fNt->run      = run;
      fNt->subrun   = subrun;
      fNt->db       = db;
      fNt->dcm      = dcm;
      fNt->feb      = (int)thisFEB;
      fNt->pix      = (int)thisPIX;
      fNt->hwx      = (int)thisX;
      fNt->hwy      = (int)thisY;
      fNt->blk      = (int)thisBLK;
      fNt->pinb     = (int)thisPINB;
      fNt->view     = (int)thisVIEW;
      fNt->mod      = (int)thisMOD;
      fNt->cell     = (int)thisCELL;
      fNt->minADC   = contentMin;
      fNt->rate     = content;
      fNt->rateHigh = contentH;
      fNt->rateLow  = contentL;
      fNt->rateMip  = contentM;
      fRatesTree->Fill();

    }//j
  }//i  (for every pixel)
}

////////////////////////////////////////////////////////////////////////
