#include "OnlineMonitoring/producer/FEBRatesByDiblock.h"
#include <iostream>
#include <cmath>
#include "TH1F.h"
#include "TH2F.h"
#include "DAQChannelMap/HardwareDisplay.h"
#include "OnlineMonitoring/producer/HistoSet.h"
#include "OnlineMonitoring/producer/RawEventSummary.h"
#include "OnlineMonitoring/producer/MicroSliceSummary.h"
#include "OnlineMonitoring/producer/NanoSliceSummary.h"
#include "OnlineMonitoring/util/HistoData.h"
#include "OnlineMonitoring/util/Settings.h"

using namespace om;

FEBRatesByDiblock::FEBRatesByDiblock() : 
  SummarySubscriber(SummarySubscriber::kNanoSlice),
  fHwMap(new daqchannelmap::HardwareDisplay)
{
  fHwMap->SetupDet(Settings::Instance().fDet);
  fDCMpDB = fHwMap->fNdcmTop + fHwMap->fNdcmSide;

  HistoSet& h = HistoSet::Instance();

  for(unsigned int i = 0; i < numDB; ++i) {
    fFEBRatesDB[i] = 0;
    fFEBHitsDB[i]  = 0;
  }

  fDCMRates  = h.GetTH2F("DCMRatesVsHour");
  fDCMHits   = h.GetTH2F("DCMHitsVsHour");
  fEmptyDCMs = h.GetTH2F("EmptyDCMsVsHour");
  fLiveTime  = h.GetTH1F("LiveTimeVsHour");

  fDCMs.clear();
  fDCMs.resize(fDCMpDB*(fHwMap->fNdiblocks)+1);

  fLastTime = time(0);
}

//......................................................................

FEBRatesByDiblock::~FEBRatesByDiblock() {
  if(fHwMap) {delete fHwMap; fHwMap = 0;}
}

//......................................................................

void FEBRatesByDiblock::GetNanoSliceSummary(const RawEventSummary&   r,
					    const DataBlockSummary&  dbs,
					    const MicroSliceSummary& mss,
					    const NanoSliceSummary&  nss)
{
  HistoSet& h = HistoSet::Instance();

  static unsigned int lastRun=0, lastEvent=0;

  // Don't do the standard histogram filling if this is the VERY start of
  // OnMon running.
  static bool start = true;

  if((lastRun!=r.fRun || lastEvent!=r.fEvent) && !start) {

    // Update empty DCMs plot and reset the list of DCMs
    for(unsigned int i = 1; i <= fDCMpDB*(fHwMap->fNdiblocks); ++i)
      if(fDCMs[i] == 0) fEmptyDCMs->Fill(r.fHour,i);

    fDCMs.clear();
    fDCMs.resize(fDCMpDB*(fHwMap->fNdiblocks)+1);
    


    // Get the current UNIX time in seconds
    time_t t = time(0);
    
    // only update the rate plots every 120 seconds
    if((t - fLastTime) >= 120) {
      fLastTime = t;
      this->UpdateRatePlots();
    }

    double range = (double)r.fRange;

    // Adjust the range due to the fact that we always get any micro blocks
    // that touch the edge of the trigger range.
    range = 100.0*floor(range/100.0)+100.0;

    fLiveTime->Fill(r.fHour,range);

  }

  lastRun   = r.fRun;
  lastEvent = r.fEvent;
  start     = false;
  
  if(fFEBHitsDB[mss.fDiblock] == 0) {
    char histoname[15];
    sprintf(histoname,"FEBHitsVsHourDB_%.2u",mss.fDiblock);
    fFEBHitsDB[mss.fDiblock] = h.GetTH2F(histoname);   
  }

  if(fFEBRatesDB[mss.fDiblock] == 0) {
    char histoname[15];
    sprintf(histoname,"FEBRatesVsHourDB_%.2u",mss.fDiblock);
    fFEBRatesDB[mss.fDiblock] = h.GetTH2F(histoname);   
  }
  
  // Fill the FEB hit plot
  fFEBHitsDB[mss.fDiblock]->Fill(r.fHour,(mss.fDCM-1)*64+nss.fFEB);

  // Fill the DCM hit plot
  fDCMHits->Fill(r.fHour,(mss.fDiblock-1)*fDCMpDB+mss.fDCM);  

  // Record if this DCM has reported yet or not
  fDCMs[(mss.fDiblock-1)*fDCMpDB+mss.fDCM] = 1;

}

//......................................................................

void FEBRatesByDiblock::UpdateRatePlots() {
  
  static const double secPerRange = 5E-7;     // fRange is in units of 500nsec

  //
  // Compute FEB rates
  //
  for(unsigned int db = 0; db < numDB; ++db) {

    if(fFEBHitsDB[db] == 0) continue;

    unsigned int nx = fFEBHitsDB[db]->GetNbinsX();
    unsigned int ny = fFEBHitsDB[db]->GetNbinsY();

    for(unsigned int i = 1; i <= nx; ++i) {

      unsigned int tbin = fLiveTime->GetBin(i);
      double time = (fLiveTime->GetBinContent(tbin))*secPerRange;
      
      if(time == 0.0) continue;

      for(unsigned int j = 1; j <= ny; ++j) {

	unsigned int hbin = fFEBHitsDB[db]->GetBin(i,j);
	double content = fFEBHitsDB[db]->GetBinContent(hbin);

	fFEBRatesDB[db]->SetBinContent(hbin,content/time);

      } // end loop over j
    } // end loop over i

  } // end loop over db



  //
  // Compute DCM rates
  //
  unsigned int nx = fDCMHits->GetNbinsX();
  unsigned int ny = fDCMHits->GetNbinsY();
  
  for(unsigned int i = 1; i <= nx; ++i) {
    
    unsigned int tbin = fLiveTime->GetBin(i);
    double time = (fLiveTime->GetBinContent(tbin))*secPerRange;
    
    if(time == 0.0) continue;
    
    for(unsigned int j = 1; j <= ny; ++j) {
      
      unsigned int hbin = fDCMHits->GetBin(i,j);
      double content = fDCMHits->GetBinContent(hbin);
      
      fDCMRates->SetBinContent(hbin,content/time);
      
    } // end loop over j
  } // end loop over i
  
}


////////////////////////////////////////////////////////////////////////
