#include "OnlineMonitoring/producer/MicroErrors.h"
#include <iostream>
#include "TH2F.h"
#include "TH1F.h"
#include "DAQChannelMap/HardwareDisplay.h"
#include "OnlineMonitoring/producer/HistoSet.h"
#include "OnlineMonitoring/producer/RawEventSummary.h"
#include "OnlineMonitoring/producer/MicroSliceSummary.h"
#include "OnlineMonitoring/util/HistoData.h"
#include "OnlineMonitoring/util/Settings.h"

using namespace om;
class RawEventSummary;
class DataBlockSummary;

MicroErrors::MicroErrors() : 
  SummarySubscriber(SummarySubscriber::kMicroSlice),
  fHwMap(new daqchannelmap::HardwareDisplay)
{
  fHwMap->SetupDet(Settings::Instance().fDet);

  // initalize the arrays of histograms to zero
  for(unsigned int i = 0; i < DCM_size_micro; ++i) {
    for(unsigned int j = 0; j < DCM_size_micro; ++j) {
       fBCperDCM[i][j] = 0;
    } // end for j
  } // end for i

  // create the initial Error Plots
  HistoSet& h = HistoSet::Instance();

  fDPEbyDCM = h.GetTH2F("DPEbyDCM");     
  fMCbyDCM  = h.GetTH2F("MCbyDCM");
  fBCbyDCM  = h.GetTH2F("BCbyDCM");
  fAEVsHour = h.GetTH2F("AEVsHour");
  fAAVsHour = h.GetTH2F("AAVsHour");


}

//......................................................................

MicroErrors::~MicroErrors() {
  if(fHwMap) {delete fHwMap; fHwMap = 0;}
}

//......................................................................

void MicroErrors::GetMicroSliceSummary(const RawEventSummary&   r,
				       const DataBlockSummary&  dbs,
				       const MicroSliceSummary& mss)
{
  HistoSet& h = HistoSet::Instance();

  //
  // Get the detector coordinates for this micro slice entry
  //
  unsigned int x1, x2, y1, y2;
  fHwMap->DiblockXY(mss.fDiblock, &x1, &y1);
  fHwMap->DCMXY(mss.fDCM, &x2, &y2);
  unsigned int X = x1+x2, Y = y1+y2;

  float AlertFill = 0.0;





  //
  // fill the Data Present histos
  //
  // NOTE:  Unlike DataPresent reported at the nano level, this is NOT reported as an "error."
  //        It is reported as a "status" so if it is false, then the histo should be filled.
  if(!mss.fDataPresent) {
    fDPEbyDCM->Fill(X,Y);
    fAEVsHour->Fill(r.fHour, 3);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 3);
  }



  //
  // fill the Simulations Flag histos
  // NOTE:   This is an "alert" histo and should only be filled with a status
  //         indicating good (0) or bad (1).
  //         We fill the histo with the OR of the current and previous values
  //         so that once an alert gets set, the histo will continue to display
  //         the set status until the histo is reset.
  //

  // record error in detailed all alerts histo
  fAAVsHour->Fill(r.fHour, 0.0, mss.fMCflag);
  if(r.fHour >= 23.0) fAAVsHour->Fill(r.fHour - 24.0, 0.0, mss.fMCflag);

  // fill DCM level histo of MC flag reported by DCM
  // 
  // NOTE:  We add 0.1 to distinguish bins that are empty from bins that have been filled
  //        with zero.
  if(mss.fMCflag || fMCbyDCM->GetBinContent(fMCbyDCM->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
  else AlertFill = 0.1;
  fMCbyDCM->SetBinContent(fMCbyDCM->FindBin(X,Y), AlertFill);





  //
  // fill the Byte Count histos
  //

  // fill the whole detector histo
  fBCbyDCM->Fill(X,Y,mss.fByteCount);

  // fill the DCM specific histos
  if(fBCperDCM[mss.fDiblock][mss.fDCM] == 0) {
    // create and name the histogram by Diblock and DCM
    char histoname[15];
    sprintf(histoname, "BCperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
    fBCperDCM[mss.fDiblock][mss.fDCM] = h.GetTH1F(histoname);
  }

  fBCperDCM[mss.fDiblock][mss.fDCM]->Fill(mss.fByteCount);

}

////////////////////////////////////////////////////////////////////////
