#include "OnlineMonitoring/producer/NanoErrors.h"
#include <iostream>
#include "TH2F.h"
#include "DAQChannelMap/HardwareDisplay.h"
#include "OnlineMonitoring/producer/HistoSet.h"
#include "OnlineMonitoring/producer/RawEventSummary.h"
#include "OnlineMonitoring/producer/MicroSliceSummary.h"
#include "OnlineMonitoring/producer/NanoSliceSummary.h"
#include "OnlineMonitoring/util/HistoData.h"
#include "OnlineMonitoring/util/Settings.h"

using namespace om;
class RawEventSummary;
class DataBlockSummary;
class MicroSliceSummary;

NanoErrors::NanoErrors() : 
  SummarySubscriber(SummarySubscriber::kNanoSlice),
  fHwMap(new daqchannelmap::HardwareDisplay)
{
  fHwMap->SetupDet(Settings::Instance().fDet);

  // initalize the arrays of histograms to zero
  for(unsigned int i = 0; i < DCM_size_nano; ++i) {
    for(unsigned int j = 0; j < DCM_size_nano; ++j) {
      fDPEperDCM[i][j]  = 0;
      fLSEperDCM[i][j]  = 0;
      fDbMperDCM[i][j]  = 0;
      fFEBFperDCM[i][j] = 0;
      fFEBEperDCM[i][j] = 0;
      fEDAQperDCM[i][j] = 0;
      fADCEperDCM[i][j] = 0;
      fOFEperDCM[i][j]  = 0;
      fPEperDCM[i][j]   = 0;
      fCEperDCM[i][j]   = 0;
    } // end for j
  } // end for i

  // create the initial Error Plots
  HistoSet& h = HistoSet::Instance();

  fDPEbyPixelDCM  = h.GetTH2F("DPEbyPixelDCM");
  fLSEbyPixelDCM  = h.GetTH2F("LSEbyPixelDCM");
  fDbMbyPixelDCM  = h.GetTH2F("DbMbyPixelDCM");
  fFEBFbyPixelDCM = h.GetTH2F("FEBFbyPixelDCM");
  fFEBEbyPixelDCM = h.GetTH2F("FEBEbyPixelDCM");
  fEDAQbyPixelDCM = h.GetTH2F("EDAQbyPixelDCM");
  fADCEbyPixelDCM = h.GetTH2F("ADCEbyPixelDCM");
  fOFEbyPixelDCM  = h.GetTH2F("OFEbyPixelDCM");
  fPEbyPixelDCM   = h.GetTH2F("PEbyPixelDCM");
  fCEbyPixelDCM   = h.GetTH2F("CEbyPixelDCM");
  fAEVsHour       = h.GetTH2F("AEVsHour");
  fAAVsHour       = h.GetTH2F("AAVsHour");


}

//......................................................................

NanoErrors::~NanoErrors() {
  if(fHwMap) {delete fHwMap; fHwMap = 0;}
}

//......................................................................

void NanoErrors::GetNanoSliceSummary(const RawEventSummary&   r,
				     const DataBlockSummary&  dbs,
				     const MicroSliceSummary& mss,
				     const NanoSliceSummary&  nss)
{
  HistoSet& h = HistoSet::Instance();

  //
  // Get the detector coordinates for this nano slice entry
  //
  unsigned int X = nss.fHdwX, Y = nss.fHdwY;
  unsigned int x1, x2, y1, y2;
  fHwMap->FEBXY(nss.fFEB, &x1, &y1);
  fHwMap->PixXY(nss.fPix, &x2, &y2);

  uint32_t FEBstatus = nss.fFEBStatus;
  float    AlertFill = 0.0;

  // fake an error message (must also add (1<<4) since Enable DAQ is negative logic)
  /*
  FEBstatus = (1<<4);
  if(r.fHour > 14.85 && r.fHour < 14.95 &&
     nss.fFEB == 33  && mss.fDiblock == 2 && mss.fDCM == 2) FEBstatus = (1<<6)+(1<<4);
  */

 
  //
  // fill Data Present Error (DPE) histos
  // 
  if(nss.fDataPresent != 1) {

    // record error in detailed all errors histo
    fAEVsHour->Fill(r.fHour, 9);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 9);

    // fill DCM level histo of DPE reported by pixel
    fDPEbyPixelDCM->Fill(X,Y);
    
    // fill pixel level histo of DPE by specific DCM
    if(fDPEperDCM[mss.fDiblock][mss.fDCM] == 0) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "DPEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fDPEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
    }

    fDPEperDCM[mss.fDiblock][mss.fDCM]->Fill(x1+x2, y1+y2);
    
  }   // end if fDataPresent != 1
  



  
  //
  // fill Link Status Error (LSE) histos
  //
  if(nss.fLinkStatus != 1) {

    // record error in detailed all errors histo
    fAEVsHour->Fill(r.fHour, 8);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 8);

    // fill DCM level histo of LSE reported by pixel
    fLSEbyPixelDCM->Fill(X,Y);

    // fill pixel level histo of LSE by specific DCM
    if(fLSEperDCM[mss.fDiblock][mss.fDCM] == 0) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "LSEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fLSEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
    }

    fLSEperDCM[mss.fDiblock][mss.fDCM]->Fill(x1+x2, y1+y2);

  }   // end if fLinkStatus != 1





  //
  // fill Debug Mode (DbM) histos
  // NOTE:   This is an "alert" histo and should only be filled with a status
  //         indicating good (0) or bad (1).
  //         We fill the histo with the OR of the current and previous values
  //         so that once an alert gets set, the histo will continue to display
  //         the set status until the histo is reset.
  //

  // record error in detailed all alerts histo
  fAAVsHour->Fill(r.fHour, 4, nss.fDebugMode);
  if(r.fHour >= 23.0) fAAVsHour->Fill(r.fHour - 24.0, 4, nss.fDebugMode);


  // fill DCM level histo of DbM reported by pixel
  // 
  // NOTE:  We add 0.1 to distinguish bins that are empty from bins that have been filled
  //        with zero.
  if(nss.fDebugMode || fDbMbyPixelDCM->GetBinContent(fDbMbyPixelDCM->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
  else AlertFill = 0.1;
  fDbMbyPixelDCM->SetBinContent(fDbMbyPixelDCM->FindBin(X,Y), AlertFill);

  // fill pixel level histo of DbM by specific DCM

  // if the histogram already exists, then fill it...
  if(fDbMperDCM[mss.fDiblock][mss.fDCM] != 0) {
    if(nss.fDebugMode ||
       fDbMperDCM[mss.fDiblock][mss.fDCM]
       ->GetBinContent(fDbMperDCM[mss.fDiblock][mss.fDCM]->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
    else AlertFill = 0.1;
    fDbMperDCM[mss.fDiblock][mss.fDCM]
      ->SetBinContent(fDbMperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), AlertFill);
  }
  // ...otherwise, only book and fill the histogram if the alert status is high
  else {
    if(nss.fDebugMode) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "DbMperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fDbMperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
      fDbMperDCM[mss.fDiblock][mss.fDCM]
	->SetBinContent(fDbMperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), 1.1);      
    }
  }





  //
  // fill the FEB buffer Full histos
  // NOTE:   This is an "alert" histo and should only be filled with a status
  //         indicating good (0) or bad (1).
  //         We fill the histo with the OR of the current and previous values
  //         so that once an alert gets set, the histo will continue to display
  //         the set status until the histo is reset.
  //
  int FEBfull;
  if((FEBstatus & (1<<6))) FEBfull = 1;
  else                     FEBfull = 0;

  // record error in detailed all alerts histo
  fAAVsHour->Fill(r.fHour, 3, FEBfull);
  if(r.fHour >= 23.0) fAAVsHour->Fill(r.fHour - 24.0, 3, FEBfull);

  // fill DCM level histo of FEBfull reported by pixel
  // 
  // NOTE:  We add 0.1 to distinguish bins that are empty from bins that have been filled
  //        with zero.
  if(FEBfull || fFEBFbyPixelDCM->GetBinContent(fFEBFbyPixelDCM->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
  else AlertFill = 0.1;

  fFEBFbyPixelDCM->SetBinContent(fFEBFbyPixelDCM->FindBin(X,Y), AlertFill);
  
  // fill pixel level histo of FEBfull by specific DCM

  // if the histogram already exists, then fill it...
  if(fFEBFperDCM[mss.fDiblock][mss.fDCM] != 0) {
    if(FEBfull ||
       fFEBFperDCM[mss.fDiblock][mss.fDCM]
       ->GetBinContent(fFEBFperDCM[mss.fDiblock][mss.fDCM]->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
    else AlertFill = 0.1;
    fFEBFperDCM[mss.fDiblock][mss.fDCM]
      ->SetBinContent(fFEBFperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), AlertFill);
  }
  // ...otherwise, only book and fill the histogram if the alert status is high
  else {
    if(FEBfull) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "FEBFperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fFEBFperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
      fFEBFperDCM[mss.fDiblock][mss.fDCM]
	->SetBinContent(fFEBFperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), 1.1);      
    }
  }




  
  //
  // fill the FEB buffer Empty histos
  // NOTE:   This is an "alert" histo and should only be filled with a status
  //         indicating good (0) or bad (1).
  //         We fill the histo with the OR of the current and previous values
  //         so that once an alert gets set, the histo will continue to display
  //         the set status until the histo is reset.
  //
  int FEBempty;
  if((FEBstatus & (1<<5))) FEBempty = 1;
  else                     FEBempty = 0;

  // record error in detailed all alerts histo
  fAAVsHour->Fill(r.fHour, 2, FEBempty);
  if(r.fHour >= 23.0) fAAVsHour->Fill(r.fHour - 24.0, 2, FEBempty);

  // fill DCM level histo of FEBempty reported by pixel
  // 
  // NOTE:  We add 0.1 to distinguish bins that are empty from bins that have been filled
  //        with zero.
  if(FEBempty || fFEBEbyPixelDCM->GetBinContent(fFEBEbyPixelDCM->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
  else AlertFill = 0.1;
  fFEBEbyPixelDCM->SetBinContent(fFEBEbyPixelDCM->FindBin(X,Y), AlertFill);

  // fill pixel level histo of FEBempty by specific DCM

  // if the histogram already exists, then fill it...
  if(fFEBEperDCM[mss.fDiblock][mss.fDCM] != 0) {
    if(FEBempty ||
       fFEBEperDCM[mss.fDiblock][mss.fDCM]
       ->GetBinContent(fFEBEperDCM[mss.fDiblock][mss.fDCM]->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
    else AlertFill = 0.1;
    fFEBEperDCM[mss.fDiblock][mss.fDCM]
      ->SetBinContent(fFEBEperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), AlertFill);
  }
  // ...otherwise, only book and fill the histogram if the alert status is high
  else {
    if(FEBempty) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "FEBEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fFEBEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
      fFEBEperDCM[mss.fDiblock][mss.fDCM]
	->SetBinContent(fFEBEperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), 1.1);      
    }
  }





  //
  // fill the Enable DAQ histos
  // NOTE:   This is an "alert" histo and should only be filled with a status
  //         indicating good (0) or bad (1).
  //         We fill the histo with the OR of the current and previous values
  //         so that once an alert gets set, the histo will continue to display
  //         the set status until the histo is reset.
  //

  // NOTE:   The status flag for Enable DAQ is "negative logic" with respect to the alert-status
  //         histogram color scheme (0 = bad, 1 = good.)
  int EDAQ;
  if((FEBstatus & (1<<4))) EDAQ = 0;
  else                     EDAQ = 1;

  // record error in detailed all alerts histo
  fAAVsHour->Fill(r.fHour, 1, EDAQ);
  if(r.fHour >= 23.0) fAAVsHour->Fill(r.fHour - 24.0, 1, EDAQ);

  // fill DCM level histo of EDAQ reported by pixel
  // 
  // NOTE:  We add 0.1 to distinguish bins that are empty from bins that have been filled
  //        with zero.
  if(EDAQ || fEDAQbyPixelDCM->GetBinContent(fEDAQbyPixelDCM->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
  else AlertFill = 0.1;
  fEDAQbyPixelDCM->SetBinContent(fEDAQbyPixelDCM->FindBin(X,Y), AlertFill);

  // fill pixel level histo of EDAQ by specific DCM

  // if the histogram already exists, then fill it...
  if(fEDAQperDCM[mss.fDiblock][mss.fDCM] != 0) {
    if(EDAQ ||
       fEDAQperDCM[mss.fDiblock][mss.fDCM]
       ->GetBinContent(fEDAQperDCM[mss.fDiblock][mss.fDCM]->FindBin(X,Y)) > 1.0) AlertFill = 1.1;
    else AlertFill = 0.1;
    fEDAQperDCM[mss.fDiblock][mss.fDCM]
      ->SetBinContent(fEDAQperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), AlertFill);
  }
  // ...otherwise, only book and fill the histogram if the alert status is high
  else {
    if(EDAQ) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "EDAQperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fEDAQperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
      fEDAQperDCM[mss.fDiblock][mss.fDCM]
	->SetBinContent(fEDAQperDCM[mss.fDiblock][mss.fDCM]->FindBin(x1+x2,y1+y2), 1.1);      
    }
  }





  //
  // fill the ADC Error histos
  //
  if((FEBstatus & (1<<3))) {

    // record error in detailed all errors histo
    fAEVsHour->Fill(r.fHour, 7);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 7);

    // fill DCM level histo of ADCE reported by pixel
    fADCEbyPixelDCM->Fill(X,Y);

    // fill pixel level histo of ADCE by specific DCM
    if(fADCEperDCM[mss.fDiblock][mss.fDCM] == 0) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "ADCEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fADCEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
    }

    fADCEperDCM[mss.fDiblock][mss.fDCM]->Fill(x1+x2, y1+y2);

  }   // end if ADCE == true





  //
  // fill the Overflow Error histos
  //
  if((FEBstatus & (1<<2))) {

    // record error in detailed all errors histo
    fAEVsHour->Fill(r.fHour, 6);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 6);

    // fill DCM level histo of OFE reported by pixel
    fOFEbyPixelDCM->Fill(X,Y);

    // fill pixel level histo of OFE by specific DCM
    if(fOFEperDCM[mss.fDiblock][mss.fDCM] == 0) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "OFEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fOFEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
    }

    fOFEperDCM[mss.fDiblock][mss.fDCM]->Fill(x1+x2, y1+y2);

  }   // end if OFE == true





  //
  // fill the Packet Error histos
  //
  if((FEBstatus & (1<<1))) {

    // record error in detailed all errors histo
    fAEVsHour->Fill(r.fHour, 5);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 5);

    // fill DCM level histo of PE reported by pixel
    fPEbyPixelDCM->Fill(X,Y);

    // fill pixel level histo of PE by specific DCM
    if(fPEperDCM[mss.fDiblock][mss.fDCM] == 0) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "PEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fPEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
    }

    fPEperDCM[mss.fDiblock][mss.fDCM]->Fill(x1+x2, y1+y2);

  }   // end if PE == true





  //
  // fill the Communication Error histos
  //
  if((FEBstatus & (1<<0))) {

    // record error in detailed all errors histo
    fAEVsHour->Fill(r.fHour, 4);
    if(r.fHour >= 23.0) fAEVsHour->Fill(r.fHour - 24.0, 4);

    // fill DCM level histo of CE reported by pixel
    fCEbyPixelDCM->Fill(X,Y);

    // fill pixel level histo of OFE by specific DCM
    if(fCEperDCM[mss.fDiblock][mss.fDCM] == 0) {
      // create and name the histogram by Diblock and DCM
      char histoname[15];
      sprintf(histoname, "CEperDCM_%.2u_%.2u", mss.fDiblock, mss.fDCM);
      fCEperDCM[mss.fDiblock][mss.fDCM] = h.GetTH2F(histoname);
    }

    fCEperDCM[mss.fDiblock][mss.fDCM]->Fill(x1+x2, y1+y2);

  }   // end if CE == true

}

////////////////////////////////////////////////////////////////////////
