#include "OnlineMonitoring/viewer/PlotClickHandler.h"
#include <cstdio>
#include <cmath>
#include <iostream>
#include "TObject.h"
#include "TPad.h"
#include "TH1.h"
#include "TH2.h"
#include "OnlineMonitoring/viewer/GUIModel.h"
#include "OnlineMonitoring/viewer/PlotOptions.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "DAQChannelMap/HardwareDisplay.h"

using namespace om;

PlotClickHandler::PlotClickHandler() :
  fHwMap(new daqchannelmap::HardwareDisplay)
{
  fHwMap->SetupDet(GUIModel::Instance().Data().fDetector);
}

//......................................................................

PlotClickHandler::~PlotClickHandler() {
  if(fHwMap) {delete fHwMap; fHwMap = 0;}
}

//......................................................................

void PlotClickHandler::SingleClick(const PlotOptions& plotopt) 
{
  int decoded = this->DecodeClick();
  if (decoded==0) return;
  
  if      (plotopt.fHwLblDet) this->HwDetHistogramInfo();
  else if (plotopt.fHwLblDCM) this->HwDCMHistogramInfo();
  else if (plotopt.fHwLblPCX) this->HwPCHistogramInfoX();
  else if (plotopt.fHwLblPCY) this->HwPCHistogramInfoY();
  else if (fHistoName.find("FEBRatesVsHourDB") < fHistoName.length() ||
	   fHistoName.find("FEBHitsVsHourDB")  < fHistoName.length())
    this->FEButcInfo();
  else if (fHistoName.find("DCMRatesVsHour")  < fHistoName.length() ||
	   fHistoName.find("DCMHitsVsHour")   < fHistoName.length() ||
	   fHistoName.find("EmptyDCMsVsHour") < fHistoName.length())
    this->DCMutcInfo();
  else                        this->DefaultHistogramInfo();
}

//......................................................................

void PlotClickHandler::DoubleClick(const PlotOptions& plotopt) 
{
  int decoded = this->DecodeClick();
  if (decoded==0) return;
  
  //
  // Double clicks supported for detector view histograms. Other would
  // be listed here...
  //
  if (plotopt.fHwLblDet) this->HwDetDoubleClick();
}

//......................................................................

int PlotClickHandler::DecodeClick() 
{
  //
  // Get current histo name
  //
  fHistoName = GUIModel::Instance().Data().fCurrentHistogram.Current();

  //
  // Find out where the click occured
  //
  TObject* h = gPad->GetSelected();
  if (h==0) return 0;

  fH1 = 0;
  if (h->InheritsFrom(TH1::Class())) fH1 = (TH1*)h;
  if (fH1==0) return 0;
  
  fH2 = 0;
  if (h->InheritsFrom(TH2::Class())) fH2 = (TH2*)h;
  if (fH2==0) return 0;
  

  int px = gPad->GetEventX();
  double xx = gPad->AbsPixeltoX(px);
  
  fX = gPad->PadtoX(xx);
  if (fX<fH1->GetXaxis()->GetXmin()) return 0;
  if (fX>fH1->GetXaxis()->GetXmax()) return 0;

  //
  // Handle TH2s here. Do them first since ROOT defines TH2 as a
  // subset of TH1's
  //
  fY    = 0;
  fYbin = 0;
  if (fH2) {
    int    py = gPad->GetEventY();
    double yy = gPad->AbsPixeltoY(py);
    
    fY = gPad->PadtoY(yy);
    
    double ymin = fH2->GetYaxis()->GetXmin();
    if (fY<ymin) return 0;

    double ymax = fH2->GetYaxis()->GetXmax();
    if (fY>ymax) return 0;

    fXbin = fH2->GetXaxis()->FindBin(fX);
    fYbin = fH2->GetYaxis()->FindBin(fY);
    return 1;
  }
  if (fH1) {
    fXbin = fH1->GetXaxis()->FindBin(fX);
    return 1;
  }

  return 0;
}

//......................................................................

void PlotClickHandler::DefaultHistogramInfo() 
{
  if (fH2) {
    char buff[256];
    sprintf(buff, "Histogram %s: bin %d,%d (x=%.3f,y=%.3f) content=%.3f",
	    fHistoName.c_str(), fXbin, fYbin, fX, fY,
	    fH2->GetBinContent(fXbin,fYbin));
    GUIModel::Instance().SetHistogramInfoText(buff);
    return;
  }
  if (fH1) {
    char buff[256];
    sprintf(buff, "Histogram %s: Bin %d (x=%.3f) content=%.3f",
	    fHistoName.c_str(), fXbin, fX, fH1->GetBinContent(fXbin));
    GUIModel::Instance().SetHistogramInfoText(buff);
    return;
  }
}

//......................................................................

void PlotClickHandler::HwDetHistogramInfo() 
{
  char buff[256];

  int ix = (int)floor(fX);
  int iy = (int)floor(fY);
  
  unsigned int diblock, dcm, feb, pix;
  unsigned int block, plane, view, module, cell;
  
  fHwMap->XYToDAQ(&diblock, &dcm, &feb, &pix, ix, iy);

  // For the NearDet we display global plane, for the FarDet & NDOS we
  // display block and plane-in-block.

  if(GUIModel::Instance().Data().fDetector == kNEARDET) {

    // For NearDet
    fHwMap->DAQToPlaneModCell(&block, &plane, &view, &module, &cell,
			      diblock, dcm, feb, pix);

    // Check that we are in an instrumented region
    if (module<=2) {
      const char* v = "?";
      if (view==daqchannelmap::X_VIEW) v = "V";
      if (view==daqchannelmap::Y_VIEW) v = "H";
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %d %s / Module %d] "
	      "[DB %.2d / DCM %.2d / FEB %2d]",
	      fHistoName.c_str(),
	      fH2->GetBinContent(fXbin,fYbin),
	      plane,
	      v,
	      module,
	      diblock,
	      dcm,
	      feb);

    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane n/a / Module n/a] "
	      "[DB %.2d / DCM %.2d / FEB %2d]",
	      fHistoName.c_str(),
	      fH2->GetBinContent(fXbin,fYbin),
	      diblock,
	      dcm,
	      feb);

    }
  }
  else {

    // For FarDet & NDOS
    fHwMap->DAQToPlaneInBlockModCell(&block, &plane, &view, &module, &cell,
				     diblock, dcm, feb, pix);

    // Check that we are in an instrumented region
    if (module<=11) {
      const char* v = "?";
      if (view==daqchannelmap::X_VIEW) v = "V";
      if (view==daqchannelmap::Y_VIEW) v = "H";
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Block %.2d / Plane in block %d %s / Module %d] "
	      "[DB %.2d / DCM %.2d / FEB %2d]",
	      fHistoName.c_str(),
	      fH2->GetBinContent(fXbin,fYbin),
	      block,
	      plane,
	      v,
	      module,
	      diblock,
	      dcm,
	      feb);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Block %.2d / Plane n/a / Module n/a] "
	      "[DB %.2d / DCM %.2d / FEB %2d]",
	      fHistoName.c_str(),
	      fH2->GetBinContent(fXbin,fYbin),
	      block,
	      diblock,
	      dcm,
	      feb);
    }

  }

  GUIModel::Instance().SetHistogramInfoText(buff);
}

//......................................................................

void PlotClickHandler::HwDCMHistogramInfo() 
{
  char buff[256];

  int ix = (int)floor(fX);
  int iy = (int)floor(fY);
  
  unsigned int diblock, dcm, feb, pix;
  unsigned int block, plane, view, module, cell;

  fHwMap->XYToDAQ(&diblock, &dcm, &feb, &pix, ix, iy);
  
  //
  // Figure out which diblock/dcm we are in. Histograms must conform
  // to naming convention test_[db]_[dcm]
  //
  const char* nm = strchr(fHistoName.c_str(), '_');
  if (nm!=NULL) {
    ++nm;
    unsigned int n = sscanf(nm,"%d_%d",&diblock,&dcm);
    if (n!=2) {
      std::cerr << "Bad histogram name decode " << fHistoName << std::endl;
      diblock = 0;
      dcm     = 0;
    }
  }

  // For the NearDet we display global plane, for the FarDet & NDOS we
  // display block and plane-in-block.

  if(GUIModel::Instance().Data().fDetector == kNEARDET) {

    // For NearDet
    fHwMap->DAQToPlaneModCell(&block, &plane, &view, &module, &cell,
			      diblock, dcm, feb, pix);

    // Check that we are in an instrumented region
    if (cell<=3*32) {
      const char* v = "?";
      if (view==daqchannelmap::X_VIEW) v = "V";
      if (view==daqchannelmap::Y_VIEW) v = "H";
      sprintf(buff,
	      "%s: Contents = %e\n"
	      "[Plane %d %s / Module %d / Cell %d] "
	      " [DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(),
	      fH2->GetBinContent(fXbin,fYbin),
	      plane,
	      v,
	      module,
	      cell,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e"
	      "[Plane n/a / Module n/a / Cell n/a] "
	      "[DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      diblock,
	      dcm,
	      feb,
	      pix);
    }

  }
  else {

    // For FarDet & NDOS
    fHwMap->DAQToPlaneInBlockModCell(&block, &plane, &view, &module, &cell,
				     diblock, dcm, feb, pix);

    // Check that we are in an instrumented region
    if (cell<=12*32) {
      const char* v = "?";
      if (view==daqchannelmap::X_VIEW) v = "V";
      if (view==daqchannelmap::Y_VIEW) v = "H";
      sprintf(buff,
	      "%s: Contents = %e\n"
	      "[Block %.2d / Plane in block %d %s / Module %d / Cell %d] "
	      " [DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(),
	      fH2->GetBinContent(fXbin,fYbin),
	      block,
	      plane,
	      v,
	      module,
	      cell,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e"
	      "[Block %.2d / Plane n/a / Module n/a / Cell n/a] "
	      "[DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      block,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }

  }

  GUIModel::Instance().SetHistogramInfoText(buff);
}

//......................................................................

void PlotClickHandler::HwPCHistogramInfoX() 
{
  unsigned int diblock, block, dcm, feb, pix;
  unsigned int plane, pinb, view, module, cell;

  plane = fH2->GetNbinsX() - fXbin;
  cell  = fH2->GetNbinsY() - fYbin;

  fHwMap->PlaneCellToDAQ(plane, &pinb, &view, &module, cell, 
			 &diblock, &block, &dcm, &feb, &pix);
  
  char buff[256];

  // For the NearDet we display global plane, for the FarDet & NDOS we
  // display block and plane-in-block.
  
  if(GUIModel::Instance().Data().fDetector == kNEARDET) {

    // For NearDet

    // Check to see if we clicked on a bin cooresponding to a plane in the
    // other view (in which case db, dcm, etc. will be wrong.)  
    if(view == daqchannelmap::X_VIEW) {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %d V / Module %d / Cell %d] "
	      "[DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      plane,
	      module,
	      cell,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %dH]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      plane);
    }

  }
  else {

    // For FarDet & NDOS

    // Check to see if we clicked on a bin cooresponding to a plane in the
    // other view (in which case db, dcm, etc. will be wrong.)  
    if(view == daqchannelmap::X_VIEW) {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Block %.2d / Plane in block %d V / Module %d / Cell %d] "
	      "[DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      block,
	      pinb,
	      module,
	      cell,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %dH]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      plane);
    }

  }

  GUIModel::Instance().SetHistogramInfoText(buff);
}

//......................................................................

void PlotClickHandler::HwPCHistogramInfoY() 
{
  unsigned int diblock, block, dcm, feb, pix;
  unsigned int plane, pinb, view, module, cell;

  plane = fH2->GetNbinsX() - fXbin;
  cell  = fYbin - 1;

  fHwMap->PlaneCellToDAQ(plane, &pinb, &view, &module, cell, 
			 &diblock, &block, &dcm, &feb, &pix);

  char buff[256];

  // For the NearDet we display global plane, for the FarDet & NDOS we
  // display block and plane-in-block.

  if(GUIModel::Instance().Data().fDetector == kNEARDET) {

    // For NearDet

    // Check to see if we clicked on a bin cooresponding to a plane in the
    // other view (in which case db, dcm, etc. will be wrong.)
    if(view == daqchannelmap::Y_VIEW) {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %d H / Module %d / Cell %d] "
	      "[DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      plane,
	      module,
	      cell,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %dV]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      plane);
    }

  }
  else {
    
    // For FarDet & NDOS

    // Check to see if we clicked on a bin cooresponding to a plane in the
    // other view (in which case db, dcm, etc. will be wrong.)
    if(view == daqchannelmap::Y_VIEW) {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Block %.2d / Plane in block %d H / Module %d / Cell %d] "
	      "[DB %.2d / DCM %.2d / FEB %2d / PIX %2d]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      block,
	      pinb,
	      module,
	      cell,
	      diblock,
	      dcm,
	      feb,
	      pix);
    }
    else {
      sprintf(buff, 
	      "%s: Contents = %e\n"
	      "[Plane %dV]",
	      fHistoName.c_str(), 
	      fH2->GetBinContent(fXbin,fYbin),
	      plane);
    }

  }

  GUIModel::Instance().SetHistogramInfoText(buff);
}

//......................................................................

void PlotClickHandler::FEButcInfo() 
{
  unsigned int diblock, dcm, feb, pix;
  unsigned int block, plane, view, module, cell;

  //
  // Figure out which diblock we are in. Histograms must conform
  // to naming convention test_[db]
  //
  const char* nm = strchr(fHistoName.c_str(), '_');
  if (nm!=NULL) {
    ++nm;
    unsigned int n = sscanf(nm,"%d",&diblock);
    if (n!=1) {
      std::cerr << "Bad histogram name decode " << fHistoName << std::endl;
      diblock = 0;
    }
  }

  char buff[256];

  dcm = (fYbin-1)/64+1;
  feb = (fYbin-1)%64;
  pix = 0;   // pix is not used here so just set it to a throw away value

  // For the NearDet we display global plane, for the FarDet & NDOS we
  // display block and plane-in-block.

  if(GUIModel::Instance().Data().fDetector == kNEARDET) {

    // For NearDet

    fHwMap->DAQToPlaneModCell(&block, &plane, &view, &module, &cell, 
			      diblock, dcm, feb, pix);
    
    const char* v = "?";
    if (view==daqchannelmap::X_VIEW) v = "V";
    if (view==daqchannelmap::Y_VIEW) v = "H";
    sprintf(buff, 
	    "%s: Contents = %e\n"
	    "[Plane %d %s / Module %d] "
	    "[DB %.2d / DCM %.2d / FEB %2d ]",
	    fHistoName.c_str(),
	    fH2->GetBinContent(fXbin,fYbin),
	    plane,
	    v,
	    module,
	    diblock,
	    dcm,
	    feb);

  }
  else {

    // For FarDet & NDOS

    fHwMap->DAQToPlaneInBlockModCell(&block, &plane, &view, &module, &cell, 
				     diblock, dcm, feb, pix);
    
    const char* v = "?";
    if (view==daqchannelmap::X_VIEW) v = "V";
    if (view==daqchannelmap::Y_VIEW) v = "H";
    sprintf(buff, 
	    "%s: Contents = %e\n"
	    "[Block %.2d / Plane in block %d %s / Module %d] "
	    "[DB %.2d / DCM %.2d / FEB %2d ]",
	    fHistoName.c_str(),
	    fH2->GetBinContent(fXbin,fYbin),
	    block,
	    plane,
	    v,
	    module,
	    diblock,
	    dcm,
	    feb);

  }

  GUIModel::Instance().SetHistogramInfoText(buff);
}

//......................................................................

void PlotClickHandler::DCMutcInfo() 
{
  char buff[256];

  unsigned int diblock, dcm;

  diblock = (fYbin-1)/(fHwMap->fNdcmTop + fHwMap->fNdcmSide) + 1;
  dcm     = (fYbin-1)%(fHwMap->fNdcmTop + fHwMap->fNdcmSide) + 1;

  sprintf(buff,
	  "%s: Contents = %e\n"
	  "[DB %.2d / DCM %.2d ]",
	  fHistoName.c_str(),
	  fH2->GetBinContent(fXbin,fYbin),
	  diblock,
	  dcm);
  
  GUIModel::Instance().SetHistogramInfoText(buff);
}

//......................................................................

void PlotClickHandler::HwDetDoubleClick() 
{
  int ix = (int)floor(fX);
  int iy = (int)floor(fY);
  
  unsigned int diblock, dcm;
  fHwMap->XYDiblock(&diblock, ix);
  fHwMap->XYDCM    (&dcm,     iy);

  //
  // Histograms which drill down to pixel views
  //
  const std::string DCMHitRateMap       ("DCMHitRateMap");
  const std::string DCMHitRateMapHighADC("DCMHitRateMapHighADC");
  const std::string DCMHitRateMapLowADC ("DCMHitRateMapLowADC");
  const std::string DCMHitRateMapMipADC ("DCMHitRateMapMipADC");
  const std::string FEBHitRateMap       ("FEBHitRateMap");
  const std::string FEBHitRateMapHighADC("FEBHitRateMapHighADC");
  const std::string FEBHitRateMapLowADC ("FEBHitRateMapLowADC");
  const std::string FEBHitRateMapMipADC ("FEBHitRateMapMipADC");
  const std::string PixelHitRateMap     ("PixelHitRateMap");

  const std::string DCMHitMap       ("DCMHitMap");
  const std::string DCMHitMapHighADC("DCMHitMapHighADC");
  const std::string DCMHitMapLowADC ("DCMHitMapLowADC");
  const std::string DCMHitMapMipADC ("DCMHitMapMipADC");
  const std::string FEBHitMap       ("FEBHitMap");
  const std::string FEBHitMapHighADC("FEBHitMapHighADC");
  const std::string FEBHitMapLowADC ("FEBHitMapLowADC");
  const std::string FEBHitMapMipADC ("FEBHitMapMipADC");
  const std::string PixelHitMap     ("PixelHitMap");

  const std::string TotADCFEBHitMap ("TotADCFEBHitMap");

  const std::string DPEbyPixelDCM ("DPEbyPixelDCM");
  const std::string LSEbyPixelDCM ("LSEbyPixelDCM");
  const std::string DbMbyPixelDCM ("DbMbyPixelDCM");
  const std::string FEBFbyPixelDCM("FEBFbyPixelDCM");
  const std::string FEBEbyPixelDCM("FEBEbyPixelDCM");
  const std::string EDAQbyPixelDCM("EDAQbyPixelDCM");
  const std::string ADCEbyPixelDCM("ADCEbyPixelDCM");
  const std::string OFEbyPixelDCM ("OFEbyPixelDCM");
  const std::string PEbyPixelDCM  ("PEbyPixelDCM");
  const std::string CEbyPixelDCM  ("CEbyPixelDCM");

  const std::string BCbyDCM       ("BCbyDCM");



  ///\todo: remove all of the hard-coded drill-down histos listed above
  // and rename them according to the general scheme outlined below

  //
  // Determine if histo is a drilldown histo from the name and set the
  // current histogram to the appropriate DCM histo.
  //

  bool drilldown = fHistoName.find("DET_dd") < fHistoName.length();

  if(drilldown) {
    unsigned int pos = fHistoName.find("DET_dd");
    std::string name = fHistoName.substr(0,pos);
    char temp[256];
    sprintf(temp, "DCM_%.2d_%.2d",diblock,dcm);
    name += temp;
    GUIModel::Instance().SetCurrentHistogram(name.c_str());
  }


  if (DCMHitMap   == fHistoName || 
      FEBHitMap   == fHistoName ||
      PixelHitMap == fHistoName) {
    char hist[256];
    sprintf(hist, "PixelsDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitMapHighADC == fHistoName ||
      FEBHitMapHighADC == fHistoName) {
    char hist[256];
    sprintf(hist, "HighADCPixelsDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitMapLowADC == fHistoName ||
      FEBHitMapLowADC == fHistoName) {
    char hist[256];
    sprintf(hist, "LowADCPixelsDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitMapMipADC == fHistoName ||
      FEBHitMapMipADC == fHistoName) {
    char hist[256];
    sprintf(hist, "MipADCPixelsDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitRateMap   == fHistoName || 
      FEBHitRateMap   == fHistoName ||
      PixelHitRateMap == fHistoName) {
    char hist[256];
    sprintf(hist, "PixelsRateDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitRateMapHighADC == fHistoName ||
      FEBHitRateMapHighADC == fHistoName) {
    char hist[256];
    sprintf(hist, "HighADCPixelsRateDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitRateMapLowADC == fHistoName ||
      FEBHitRateMapLowADC == fHistoName) {
    char hist[256];
    sprintf(hist, "LowADCPixelsRateDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DCMHitRateMapMipADC == fHistoName ||
      FEBHitRateMapMipADC == fHistoName) {
    char hist[256];
    sprintf(hist, "MipADCPixelsRateDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }


  if (TotADCFEBHitMap == fHistoName) {
    char hist[256];
    sprintf(hist, "TotADCPixelsDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }


  if (DPEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "DPEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (LSEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "LSEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (DbMbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "DbMperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (FEBFbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "FEBFperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (FEBEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "FEBEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (EDAQbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "EDAQperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (ADCEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "ADCEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (OFEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "OFEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (PEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "PEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (CEbyPixelDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "CEperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
  if (BCbyDCM == fHistoName) {
    char hist[256];
    sprintf(hist, "BCperDCM_%.2d_%.2d",diblock,dcm);
    GUIModel::Instance().SetCurrentHistogram(hist);
    return;
  }
}

////////////////////////////////////////////////////////////////////////
