////////////////////////////////////////////////////////////////////////
// \file    ReMId.cxx
// \brief   A pid object for muon identification
// \version $Id: ReMId.cxx,v 1.1.1.1 2012-09-26 18:38:44 raddatz Exp $
// \author  Nicholas Raddatz - raddatz@physics.umn.edu
////////////////////////////////////////////////////////////////////////
#include "ReMId.h"
#include "TVector3.h"

namespace remid
{

  //......................................
  ReMId::ReMId()
    : rb::PID()
  {
    SetMuonDedxLL(0);
    SetPionDedxLL(0);
    SetMuonScatLL(0);
    SetMuonScatLL(0);
  }

  //......................................
  ReMId::ReMId(int pdg, double val)
    : rb::PID(pdg,val)
  {
    SetMuonDedxLL(0);
    SetPionDedxLL(0);
    SetMuonScatLL(0);
    SetMuonScatLL(0);
  }

  //......................................
  void    ReMId::AddScatter(TVector3 pos,double cosScat)
  {
    fscatLocation.push_back(pos);
    fcosScatAngle.push_back(cosScat);
  }

  //.....................................
  void    ReMId::AddDedx(double dedx, double distFromStart, bool used, bool vert)
  {
    fdedx.push_back(dedx);
    ffracDistFromStart.push_back(distFromStart);
    fdedxused.push_back(used);
    fdedxvert.push_back(vert);
  }

  //.....................................
  void    ReMId::SetMuonDedxLL(double ll)
  {
    fMuonDedxLL = ll;
  }

  //.....................................
  void    ReMId::SetPionDedxLL(double ll)
  {
    fPionDedxLL = ll;
  }

  //.....................................
  double  ReMId::AvgDedx() const
  {
    if(fdedx.size() == 0){ return 0;}
    else{
      double totDedx = 0;
      for(unsigned int i = 0; i<fdedx.size();++i){
        totDedx+=fdedx[i];
      } 
      return totDedx/((double)fdedx.size());
    }
  }

  //.....................................
  double  ReMId::AvgDedxUsed() const
  {
    int ndedxused = 0;
    double totDedx = 0;
    for(unsigned int i = 0; i < fdedx.size(); ++i){
      if(fdedxused[i]){
	++ndedxused;
	totDedx+=fdedx[i];
      }
    }
    if(ndedxused == 0){ return 0;}
    else{return totDedx/((double)ndedxused); }
  }

  //.....................................
  unsigned int  ReMId::NScatters() const
  {
    return fscatLocation.size();
  }

  //.....................................
  unsigned int ReMId::NDedx() const
  {
    return fdedx.size();
  }

  //.....................................
  TVector3  ReMId::ScatLocation(unsigned int i) const
  {
    return fscatLocation[i];
  }

  //.....................................
  double  ReMId::CosScat(unsigned int i) const
  {
    return fcosScatAngle[i];
  }

  //.....................................
  double  ReMId::DedxLocation(unsigned int i) const
  {
    return ffracDistFromStart[i];
  }

  //.....................................
  double  ReMId::Dedx(unsigned int i) const
  {
    return fdedx[i];
  }

  //.....................................
  bool  ReMId::DedxUsed(unsigned int i) const
  {
    return fdedxused[i];
  }

  //.....................................
  bool  ReMId::DedxVertex(unsigned int i) const
  {
    return fdedxvert[i];
  }

  //.....................................
  double  ReMId::MuonDedxLL() const
  {
    return fMuonDedxLL;
  }

  //.....................................
  double  ReMId::PionDedxLL() const
  {
    return fPionDedxLL;
  }

  //.....................................
  void    ReMId::SetMuonScatLL(double ll)
  {
    fMuonScatLL = ll;
  }

  //.....................................
  void    ReMId::SetPionScatLL(double ll)
  {
    fPionScatLL = ll;
  }

  //.....................................
  void    ReMId::SetDedxUsed(unsigned int idx, bool used)
  {
    fdedxused[idx] = used;
  }

  //.....................................
  void    ReMId::SetDedxVertex(unsigned int idx, bool vert)
  {
    fdedxvert[idx] = vert;
  }

  //.....................................
  void    ReMId::SetContained(bool cont)
  {
    fCont = cont;
  }

  //.....................................
  double  ReMId::MuonScatLL() const
  {
    return fMuonScatLL;
  }

  //.....................................
  double  ReMId::PionScatLL() const
  {
    return fPionScatLL;
  }

  //.....................................
  /// Return the dE/dx separation variable used as an input to the kNN that determines the pid value
  double  ReMId::DedxSeparation() const
  {
    return fMuonDedxLL-fPionDedxLL;
  }

  //.....................................
  /// Return the scattering separation variable used as an input to the kNN that determines the pid value
  double  ReMId::ScatSeparation() const
  {
    return fMuonScatLL-fPionScatLL;
  }

  //.....................................
  int  ReMId::NMeasurementPlanes() const
  {
    int npl = 0;
    for(unsigned int i = 0; i < fdedxused.size(); ++i){
      if(fdedxused[i]){ ++npl; }
    }
    return npl;
  }

  //.....................................
  int  ReMId::NVertexPlanes() const
  {
    int npl = 0;
    for(unsigned int i = 0; i< fdedxvert.size(); ++i){
      if(fdedxvert[i]){ ++npl; }
    }
    return npl;
  }

  //.....................................
  bool    ReMId::Contained() const
  {
    return fCont;
  }

}
