///////////////////////////////////////////////////////////////////////////
/// \brief  Simple service to provide a RunHistory configured to the right run
/// \author Christopher Backhouse - bckhouse@caltech.edu
//////////////////////////////////////////////////////////////////////////

// NOvA includes
#include "RunHistory/service/RunHistoryService.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include "SummaryData/RunData.h"

// Framework includes
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// ROOT includes
#include "TTimeStamp.h"

namespace nova
{
namespace dbi
{

  //------------------------------------------------------------
  RunHistoryService::RunHistoryService(const fhicl::ParameterSet& pset,
                                       art::ActivityRegistry &reg)
  {
    reconfigure(pset);
    reg.sPreBeginRun.watch(this, &RunHistoryService::preBeginRun);
  }

  //----------------------------------------------------------

  RunHistoryService::~RunHistoryService()
  {
  }

  //-----------------------------------------------------------
  void RunHistoryService::reconfigure(const fhicl::ParameterSet& pset)
  {
    fCalHitLabel            = pset.get< std::string >("CalHitLabel");
    fFakeRunNumber          = pset.get< int >("FakeRunNumber", 0);
    fTimeQueries            = pset.get< bool >("TimeQueries", false);
    fUseMasking             = pset.get< bool >("UseMasking", false);
    fAnalysisDiBlockMasking = pset.get< bool >("AnalysisDiBlockMasking", false);
    fCustomMasking          = pset.get< bool >("CustomMasking", false);
    fCustomMask             = pset.get< std::vector<int> >("CustomMask");
  }

  //----------------------------------------------------------
  void RunHistoryService::preBeginRun(const art::Run& run)
  {
    std::vector<art::Handle<sumdata::RunData>> rdcol;
    run.getManyByType(rdcol);
    if(!rdcol.empty()) RunHistory::SetDetector(rdcol[0]->DetId());

    int runNumber = (fFakeRunNumber == 0) ? run.run() : fFakeRunNumber;

    RunHistory::SetRunNumber(runNumber);

    this->SetTimeQueries(fTimeQueries);
    this->SetTimeParsing(fTimeQueries);
    this->SetIfAnyMasking(fUseMasking);
    this->SetAnalysisDiBlockMasking(fAnalysisDiBlockMasking);
    this->SetIfCustomMasking(fCustomMasking);
    std::vector<bool> cmask;  // let people put 1s and 0s into fcl instead of trues and falses
    for(unsigned int i = 0; i < 14; i++) {
      if(fCustomMask[i] == 1) cmask.push_back(true);
      else cmask.push_back(false);
    }
    this->SetCustomMask(cmask);

    art::Handle<int> mask;
    run.getByLabel(fCalHitLabel,mask);
    if(mask.failedToGet()) {
      if(RunHistory::GoodDiBlockMask()==0) std::cout << "RunHistory: did not find mask in file, using current fcl settings of: no mask" << std::endl;
      else std::cout << "RunHistory: did not find mask in file, using current fcl settings of: " << RunHistory::DiBlockMaskToString(RunHistory::GoodDiBlockMask()) << std::endl;
    }
    else {
      RunHistory::SetGoodDiBlockMask(*mask);
      if(*mask==0) std::cout << "RunHistory: found mask in file: no mask used " << std::endl;
      else std::cout << "RunHistory: found mask in file: " << RunHistory::DiBlockMaskToString(*mask) << std::endl;
    }

    // determine whether we are dealing with MC, data, or a mixture of the two
    // for now just test on the date of the file.  MC has a timestamp before 2010
    art::Timestamp ts = run.beginTime();

    // make it into a TTimeStamp
    TTimeStamp tts(ts.timeHigh(), ts.timeLow());

    // Get the year of the timestamp from YYYYMMDD format
    uint32_t year = tts.GetDate()/10000;

    this->SetDataType(nova::dbi::kNDataType);
    if(year < 2010) this->SetDataType(nova::dbi::kMCOnly);
    else            this->SetDataType(nova::dbi::kDataOnly);

    if (runNumber > 1000000 || runNumber < 10000) 
      this->SetDataType(nova::dbi::kMCOnly);

    return;
  }

  DEFINE_ART_SERVICE(RunHistoryService)

}
}
////////////////////////////////////////////////////////////////////////
