//////////////////////////////////////////////////////////////////////////
/// \brief  Cosmic track analysis module CosmicTrackAna.
///         Makes some validation plots.
/// \author Gavin S. Davies - gsdavies@iastate.edu
/// \date 
//////////////////////////////////////////////////////////////////////////
#include <string>
#include <vector>

// ROOT includes
#include "TH1D.h"
#include "TH2D.h"
#include "TTree.h"
#include "TVector3.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "GeometryObjects/Geo.h"
#include "MCCheater/BackTracker.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "SummaryData/CosmicExposure.h"
#include "Utilities/func/MathUtil.h"
#include "GeometryObjects/CellGeo.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/FindOne.h" 

typedef struct{
  int run;
  int subrun;
  int event;
  int nevents;
  int ntracks;
} CosmicTrackTree;

typedef struct{
  float efficiency;
  float purity;
} CosmicPurityEfficiency;

typedef struct{
  float signalFraction;
  float hitFraction;
  float signal;
  float hits;
} CosmicSignalHitFraction;


typedef std::map<double, std::pair<std::pair<TVector3, TVector3>, TVector3> > zBoundMap;

namespace trk {
  // A module to analyse cosmic tracks produced by CosmicTrack module
  class CosmicTrackAna : public art::EDAnalyzer {
  public:
    explicit CosmicTrackAna(fhicl::ParameterSet const& pset);
    virtual ~CosmicTrackAna();

    void analyze(art::Event const& evt);
    void reconfigure(const fhicl::ParameterSet& pset);

    virtual void beginSubRun(const art::SubRun& sr);

    void beginJob();
    void endJob();
        
  private:

    void FillEventMCTruthHistograms(size_t             trkMax);   
    void FillTrackMCTruthHistograms(rb::Track   const& track,
				    rb::Cluster const& slice);   
    void FillTrackHistograms(rb::Track   const& track,
			     rb::Cluster const& slice);   

    zBoundMap::const_iterator Bounds(zBoundMap const& bounds,
				     double    const& hitZ);
    
    bool CheckZInBounds(std::pair<TVector3, TVector3> const& bounds,
			double                        const& hitZ);
      
    std::string   fMCTruthListLabel;  ///< Where to find MCTruth to reconstruct
    std::string   fParticleListLabel; ///< Where to find ParticleList to reconstruct
    std::string   fFLSHitListLabel;   ///< Where to find FLSHits to reconstruct
    std::string   fCellHitsListLabel; ///< Where to find CellHits to reconstruct
    std::string   fTrackLabel;        ///< Where to find Tracks to reconstruct
    std::string   fExposureLabel;     ///< Where to find CosmicExposure info
    
    double        fMinSig;            ///< Minimum signal of cell hit to use in a track - using PE values
    double        fTrkLengthCut;      ///< Minimum length of track to accept (cm)
    
    int           fTicks;             ///< # of clock ticks to read out one event
    int           fTickWindow;        ///< Size of window for ticks of TDC to slice event in time
    double        fLiveTime;          ///< Livetime in sec according to file
    unsigned int  fNevents;	      ///< total number of events counter
    int           fNTracks;           ///< total number of reconstructed tracks counter
    int           fNTracksMC;         ///< total number of MC particles
    int           fRun;		      ///< Run number
    int           fSubRun;            ///< Subrun number

    double        fCellHalfDepthZ;    ///< extent of a cell in the z direction
    
    //......................................................................
       
    TH2D *fDCosXRecoVsTrue;           ///< Direction cosine in x
    TH2D *fDCosYRecoVsTrue;           ///< Direction cosine in y
    TH2D *fDCosZRecoVsTrue;           ///< Direction cosine in z
    TH2D *fDCosXDiffVsZLen;           ///< Fraction difference in direction cosine in x vs z length
    TH2D *fDCosYDiffVsZLen;           ///< Fraction difference in direction cosine in y vs z length
    TH2D *fDCosZDiffVsZLen;           ///< Fraction difference in direction cosine in z vs z length
    TH2D *fFracHitsUsedVsNum;         ///< Fraction hits used vs the number of hits that could be used
    
    TH1D *fDCosXDiff;                 ///< Fraction difference in direction cosine in x
    TH1D *fDCosYDiff;                 ///< Fraction difference in direction cosine in y
    TH1D *fDCosZDiff;                 ///< Fraction difference in direction cosine in z
    TH1D *fRecoEff;
    TH2D *fTrackFindEff;
    TH1D *fPrimaryTrueEnergy;         ///< True energy of particle zero (GeV)

    TH1D *fDCosX;                     ///< direction cosine in x
    TH1D *fDCosY;                     ///< direction cosine in y
    TH1D *fDCosZ;                     ///< direction cosine in z
    TH1D *fEndZ;                      ///< track end position in z
    TH1D *fEndX;                      ///< track end position in x
    TH1D *fEndY;                      ///< track end position in y
    TH2D *fEndZX;                     ///< track end position in zx
    TH2D *fEndXY;                     ///< track end position in xy
    TH2D *fEndZY;                     ///< track end position in zy
    TH1D *fStartZ;                    ///< track start position in z
    TH1D *fStartX;                    ///< track start position in x
    TH1D *fStartY;                    ///< track start position in y
    TH2D *fStartZX;                   ///< track start position in zx
    TH2D *fStartXY;                   ///< track start position in xy
    TH2D *fStartZY;                   ///< track start position in zy
    TH1D *fDiffZ;                     ///< difference in track start and end position in z
    TH1D *fDiffX;                     ///< difference in track start and end position in x
    TH1D *fDiffY;                     ///< difference in track start and end position in y
    TH1D *fDirZ;                      ///< direction in z
    TH1D *fDirX;                      ///< direction in x
    TH1D *fDirY;                      ///< direction in y
    TH2D *fDirXY;		      ///< direction in x versus y
  
    TH1D *fFractionHitsUsed;          ///< Fraction of hits in event used in track
    TH1D *fFractionSigUsed;           ///< Fraction of hits in event used in track
    TH2D *fTracksFoundVsLeft;         ///< number of tracks found in event vs particles leaving tracks
    TH1D *fTrkCells;
    TH1D *fTracksFound;               ///< number of tracks found
    TH1D *fResiduals[2];              ///< residuals for the tracks
    TH1D *fCellHitSig;                ///< signal size of calibrated cell hits
    TH1D *fTrkSig;                    ///< track signal (Q)
    TH1D *fVtxDistToEdge;             ///< distance in cm to closest edge of detector at track vertex
    TH1D *fEndDistToEdge;             ///< distance in cm to closest edge of detector at track end
    TH2D *fResidVsCos[2];             ///< residual for each hit vs cos(theta)
    TH2D *fResidVsAz[2];              ///< residual for each hit vs azimuth

    TH1D *fHitXPosition;              ///< position of reco hits in x
    TH1D *fHitYPosition;              ///< position of reco hits in y
    TH1D *fHitZPosition;              ///< position of reco hits in z
    TH1D *fTrackLength;               ///< length of track

    TH2D *fHitXYPosition;             ///< reco hit positions in xy
    TH2D *fHitXZPosition;             ///< reco hit positions in xz
    TH2D *fHitYZPosition;             ///< reco hit positions in yz
    TH1D *fCellHit[2];		      ///<
    TH1D *fPlaneHit[2];      	      ///<
 
    TH2D *fCellPlane[2]; 	      ///<
    
    TH1D *fCosCosmic;                 ///< cosine theta of cosmic track with respect to zenith
    TH1D *fAzimuth;                   ///< azimuthal angle of the cosmic ray
    TH1D *fCosmicRate;                ///< histogram holding single value for cosmic rate at end of job
    TH1F *fTotNEvents;                ///< total number of events
    TH1D *fTNS[2];                    ///< distribution of x/y cellhit times in nano seconds

    TH1D *fDeltaZEnd;                 ///< difference in z endpoint of reconstructed track vs simulated
    TH1D *fDeltaEnd;                  ///< difference in endpoint of reconstructed track vs simulated

    TH2D *fDeltaZEndVsDCosZ;          ///< difference in z endpoint of recoed track & simulated vs dcosZ
    TH2D *fDeltaZBegVsDCosZ;          ///< difference in z start point of recoed track & simulated vs dcosZ
    TH2D *fDeltaZEndVsDistToEdge;     ///< difference in z endpoint of recoed track & simulated vs dcosZ

    TH1D *fTrackEfficiency;           ///< efficiency in selecting hits for the track 
    TH1D *fTrackPurity;               ///< purity in selecting hits for the track
 
    CosmicTrackTree                      fCosmicTrackTree;    ///< event information
    std::vector<CosmicPurityEfficiency>  fPurityEfficiency;   ///< purity and efficiency information
    std::vector<CosmicSignalHitFraction> fSignalHitFraction;  ///< fraction of signal and hits used
    std::vector<TVector3>                fTrackStartPos;      ///< starting position of tracks
    std::vector<TVector3>                fTrackEndPos;        ///< ending position of tracks
    CosmicPurityEfficiency               fPurEff;             ///< purity and efficiency information
    CosmicSignalHitFraction              fSigHitFrac;         ///< fraction of signal and hits used
    TVector3                             fTrkStPos;           ///< starting position of tracks
    TVector3                             fTrkEndPos;          ///< ending position of tracks
    TTree*                               fTree;               ///< initializer of TTree for this module    

    art::ServiceHandle<geo::Geometry> fGeom;
    art::ServiceHandle<cheat::BackTracker> fBT;

  };
};

static const double ns            = 1.0E-9;

namespace trk{

  bool sortTVector3ByZ(TVector3 const& a, TVector3 const& b)
  {
    return a.Z() < b.Z();
  }
  
  //......................................................................
  CosmicTrackAna::CosmicTrackAna(fhicl::ParameterSet const& pset):
    EDAnalyzer(pset),
    fMCTruthListLabel (pset.get< std::string >("MCTruthListLabel", "generator") ),
    fCellHitsListLabel(pset.get< std::string >("CellHitsListLabel")             ),
    fTrackLabel       (pset.get< std::string >("TrackLabel")                    ),
    fExposureLabel    (pset.get< std::string >("ExposureLabel")                 ),
    fMinSig           (pset.get< double      >("MinSig")                        ),
    fTrkLengthCut     (pset.get< double      >("TrkLengthCut")                  ),
    fLiveTime(0), 
    fNevents(0), 
    fNTracks(0), 
    fNTracksMC(0), 
    fRun(-999), 
    fSubRun(-999),
    fCellHalfDepthZ(0.)
  {
    reconfigure(pset);    
  }
  
  //......................................................................  
  CosmicTrackAna::~CosmicTrackAna()
  {
  }
  
  //......................................................................  
  void CosmicTrackAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    double maxlength = 6600.0;
    double xmax      = 1200.0;
    double ymax      = 1200.0;
    double zmax      = 6500.0;
    int    xbins     = 480;
    int    ybins     = 480;
    int    zbins     = 1300;

    
    fDCosXRecoVsTrue   = tfs->make<TH2D>("dcosXRecoVsTrue", ";dcos X_{true};dcos X_{reco}", 250, -1., 1., 250, -1., 1.);
    fDCosYRecoVsTrue   = tfs->make<TH2D>("dcosYRecoVsTrue", ";dcos Y_{true};dcos Y_{reco}", 250, -1., 1., 250, -1., 1.);
    fDCosZRecoVsTrue   = tfs->make<TH2D>("dcosZRecoVsTrue", ";dcos Z_{true};dcos Z_{reco}", 250, -1., 1., 250, -1., 1.);

    fDCosXDiffVsZLen   = tfs->make<TH2D>("dcosXDiffVsZLen", ";Z length (cm);dcos_{X} Fractional Difference", 1320, 0., maxlength, 200, -1., 5.);
    fDCosYDiffVsZLen   = tfs->make<TH2D>("dcosYDiffVsZLen", ";Z length (cm);dcos_{Y} Fractional Difference", 1320, 0., maxlength, 200, -1., 3.);
    fDCosZDiffVsZLen   = tfs->make<TH2D>("dcosZDiffVsZLen", ";Z length (cm);dcos_{Z} Fractional Difference", 1320, 0., maxlength, 200, -1., 3.);
    
    fDCosXDiff         = tfs->make<TH1D>("dcosXDiff", ";dcos_{X} Fractional Difference;", 200, -1., 3.);
    fDCosYDiff         = tfs->make<TH1D>("dcosYDiff", ";dcos_{Y} Fractional Difference;", 200, -1., 3.);
    fDCosZDiff         = tfs->make<TH1D>("dcosZDiff", ";dcos_{Z} Fractional Difference;", 200, -1., 3.);

    fPrimaryTrueEnergy = tfs->make<TH1D>("primaryTrueEnergy", ";True Energy (GeV);", 200, 0., 50);
    
    fDCosX             = tfs->make<TH1D>("dcosX", ";dcos_{X};", 101, -1., 1.);
    fDCosY             = tfs->make<TH1D>("dcosY", ";dcos_{Y};", 101, -1., 1.);
    fDCosZ             = tfs->make<TH1D>("dcosZ", ";dcos_{Z};", 101, -1., 1.);
    
    fEndX 	       = tfs->make<TH1D>("endX",";End X (cm)",xbins, -xmax, xmax);
    fEndY 	       = tfs->make<TH1D>("endY",";End Y (cm)",ybins, -ymax, ymax);
    fEndZ 	       = tfs->make<TH1D>("endZ",";End Z (cm)",zbins, 0., zmax);
    
    fEndXY             = tfs->make<TH2D>("endXY",";End X (cm); End Y (cm)",xbins, -xmax, xmax, ybins, -ymax, ymax);
    fEndZY             = tfs->make<TH2D>("endZY",";End Z (cm); End Y (cm)",zbins, 0., zmax, ybins, -ymax, ymax);
    fEndZX             = tfs->make<TH2D>("endZX",";End Z (cm); End X (cm)",zbins, 0., zmax, xbins, -xmax, xmax);
    
    fDirX 	       = tfs->make<TH1D>("dirX",";Direction X (cm)",100, -1.1, 1.1);
    fDirY 	       = tfs->make<TH1D>("dirY",";Direction Y (cm)",100, -1.1, 1.1);
    fDirZ 	       = tfs->make<TH1D>("dirZ",";Direction Z (cm)",100, -1.1, 1.1);
    
    fStartX            = tfs->make<TH1D>("startX",";Start X (cm)",xbins, -xmax, xmax);
    fStartY            = tfs->make<TH1D>("startY",";Start Y (cm)",ybins, -ymax, ymax);
    fStartZ            = tfs->make<TH1D>("startZ",";Start Z (cm)",zbins, 0., zmax);
    
    fStartXY           = tfs->make<TH2D>("startXY",";Start X (cm); Start Y (cm)",xbins, -xmax, xmax, ybins, -ymax, ymax);
    fStartZY           = tfs->make<TH2D>("startZY",";Start Z (cm); Start Y (cm)",zbins, 0., zmax, ybins, -ymax, ymax);
    fStartZX           = tfs->make<TH2D>("startZX",";Start Z (cm); Start X (cm)",zbins, 0., zmax, xbins, -xmax, xmax);
    
    fDiffX             = tfs->make<TH1D>("diffX",";End-Start X (cm)",1000, 0., 1000.);
    fDiffY             = tfs->make<TH1D>("diffY",";End-Start Y (cm)",1000, 0., 1000.);
    fDiffZ             = tfs->make<TH1D>("diffZ",";End-Start Z (cm)",3000, 0., 1500.);
    fDirXY             = tfs->make<TH2D>("dirXY",";Prong Direction X;Prong Direction Y",101,-1.,1.,101,-1.,1.);
    
    fFractionHitsUsed  = tfs->make<TH1D>("fractionHitsUsed", ";Used/Total (Hits);Spills", 100, 0., 1.);
    fFractionSigUsed   = tfs->make<TH1D>("fractionSigUsed", ";Used/Total (Signal);Spills", 100, 0., 1.);
    fFracHitsUsedVsNum = tfs->make<TH2D>("fractionHitsUsedVsNum", ";Total Hits; Fraction Used", 250, 0., 500., 100, 0., 1.);    
    fTracksFoundVsLeft = tfs->make<TH2D>("tracksFoundVsLeft", ";True tracks;Reco tracks", 25, 0, 125, 25, 0, 125);
    fTracksFound       = tfs->make<TH1D>("tracksFound", ";Reco tracks;", 25, 0, 125);
    fTrackFindEff      = tfs->make<TH2D>("trackfindeff",";True tracks;Reco Efficiency",20,0,10,100,0,1);
    fRecoEff           = tfs->make<TH1D>("recoeff",";Reco Efficiency",100,0,1);
    fTrackPurity       = tfs->make<TH1D>("trackPurity",     ";Tracking Purity; Tracks",     100, 0., 2.);
    fTrackEfficiency   = tfs->make<TH1D>("trackEfficiency", ";Tracking Efficiency; Tracks", 100, 0., 2.);

    fCellHitSig        = tfs->make<TH1D>("cellHitSig", ";PE;Cell Hits", 100, 0., 100.);
    fTrkSig            = tfs->make<TH1D>("trksig", ";Track Signal (Q);counts",100,0.,100000.);
    fTrkCells          = tfs->make<TH1D>("fTrkCells",";Number of Track Cell Hits",500,0,500);
    fEndDistToEdge     = tfs->make<TH1D>("endDistToEdge", ";Distance (cm);", 50,  0., 500.);
    fVtxDistToEdge     = tfs->make<TH1D>("vtxDistToEdge", ";Distance (cm);", 50,  0., 500.);
    
    fTrackLength       = tfs->make<TH1D>("trackLength",   ";Track Length (cm);"  , 550, 0., maxlength);
    
    fCosCosmic         = tfs->make<TH1D>("coscosmic",";cos#theta",100,0.,1.);
    fAzimuth           = tfs->make<TH1D>("azimuth",  ";#phi",     36,0.,360.);
    fCosmicRate        = tfs->make<TH1D>("cosmicrate",";Cosmic Rate (Hz)",1,0.,10000);
    
    fTotNEvents        = tfs->make<TH1F>("fNEvents",";;Number of events",2,0,2);
    
    fHitXPosition      = tfs->make<TH1D>("hitXPosition", ";Hit X (cm);Reco Hits", xbins, -xmax, xmax);	
    fHitYPosition      = tfs->make<TH1D>("hitYPosition", ";Hit Y (cm);Reco Hits", ybins, -ymax, ymax);
    fHitZPosition      = tfs->make<TH1D>("hitZPosition", ";Hit Z (cm);Reco Hits", zbins, 0., zmax);
    
    fHitXYPosition     = tfs->make<TH2D>("hitXYPosition", ";Hit X (cm);Hit Y (cm)", 20, -200., 200., 50, -250., 250.);	
    fHitXZPosition     = tfs->make<TH2D>("hitXZPosition", ";Hit Z (cm);Hit X (cm)", 150, 0.,  1500., 40, -200., 200.);
    fHitYZPosition     = tfs->make<TH2D>("hitYZPosition", ";Hit Z (cm);Hit Y (cm)", 150, 0.,  1500., 50, -250., 250.);

    fDeltaZEnd         = tfs->make<TH1D>("deltaZEnd",     ";#Delta Z_{end} (cm);Tracks", 200, -100., 100.);
    fDeltaEnd          = tfs->make<TH1D>("deltaEnd",      ";#Delta_{end} (cm);Tracks",   200, -100., 100.);
    
    fDeltaZEndVsDCosZ  = tfs->make<TH2D>("deltaZEndVsDCosZ", ";#Delta Z_{end} (cm);dZ/dS;", 200, -100., 100., 20, -1., 1.);
    fDeltaZBegVsDCosZ  = tfs->make<TH2D>("deltaZBegVsDCosZ", ";#Delta Z_{start} (cm);dZ/dS;", 200, -100., 100., 20, -1., 1.);
    fDeltaZEndVsDistToEdge  = tfs->make<TH2D>("deltaZEndVsDistToEdge", ";#Delta Z_{end} (cm);Distance to Edge (cm);", 200, -100., 100., 200, -100., 100.);

    double  ncell = 384;
    double nplane = 960;
    for(int view = geo::kX; view <= geo::kY; ++view){	
      const TString viewStr = (view == geo::kX) ? "X" : "Y";
      fCellHit[view]          = tfs->make<TH1D>("fCell"+viewStr,";cell #",ncell,-0.5,ncell-0.5);
      fPlaneHit[view]         = tfs->make<TH1D>("fPlane"+viewStr,";plane #",nplane,0,nplane-0.5);
      
      fCellPlane[view]        = tfs->make<TH2D>("View"+viewStr,";Plane no. ;Cell no.",nplane,0.,nplane-0.5,ncell,0.,ncell-0.5);
      
      fResiduals[view]        = tfs->make<TH1D>("residuals"+viewStr, ";Residual (cm);Hits", 400, 0., 1000.);
      fResidVsCos[view]       = tfs->make<TH2D>("residVsCos"+viewStr, ";Residual (cm);cos(#theta)", 200, 0., 1000., 100, -1., 1.);
      fResidVsAz[view]        = tfs->make<TH2D>("residVsAz"+viewStr, ";Residual (cm);#phi", 200, 0., 1000., 180, 0., 360.);
      fTNS[view]              = tfs->make<TH1D>("HitTNS"+viewStr, ";Time (ns);hits", 70000, -100000.0, 600000.);     
    }

    
    fTree = tfs->make<TTree>("Cosmics","CosmicTrackAna Tree");
    fTree->Branch("fCosmicTrackTree",  &fCosmicTrackTree,       "run/I:subrun/I:event/I:nevents/I:ntracks/I");
    fTree->Branch("efficiencyPurity",  &fPurEff,                "efficiency/F:purity/F");
    fTree->Branch("signalHitFraction", &fSignalHitFraction,     "signalFraction/F:hitFraction/F:signal/F:hits/F");
    fTree->Branch("trackStartPos",     &fTrkStPos);
    fTree->Branch("trackEndPos",       &fTrkEndPos);
  }

  //......................................................................
  void CosmicTrackAna::beginSubRun(const art::SubRun& sr)
  {
    art::Handle<sumdata::CosmicExposure> ce;
    sr.getByLabel(fExposureLabel, ce);
    if(ce.failedToGet()){
      mf::LogError("CosmicTrackAna") << "No Cosmic Exposure Info";
      return;
    }
    fLiveTime += ce->totlivetime;

  }
  
  //......................................................................
  ///
  /// The analysis method is intended to perform checks of the
  /// performance of the reconstruction for the purposes of
  /// optimization, testing, and validation.
  ///
  /// \param evt - Read-only access to the event data
  ///
  // checks to make include - 
  // 1. are the direction cosines of the track correct
  // 2. fraction of hits used in tracks
  // 3. number of particles leaving tracks vs number tracks found
  // 4. fraction of hits in a track that belong to a given particle
  // 5. cosmic flux rate and angular distribtion stability
  // 6. cosmic track efficiency
  void CosmicTrackAna::analyze(art::Event const& evt)
  {  

    if(fCellHalfDepthZ < 1.){
      fCellHalfDepthZ = fGeom->Plane(0)->Cell(0)->HalfD();    
      
      LOG_DEBUG("CosmicTrackAna") << "cell half depth: " << fGeom->Plane(0)->Cell(0)->HalfD() << "\n"
				  << "cell half width: " << fGeom->Plane(0)->Cell(0)->HalfW() << "\n"
				  << "cell half leng:  " << fGeom->Plane(0)->Cell(0)->HalfL();
    }

    ++fNevents; //Updating counter of number of events in sub-run
    if(fRun < 0)    fRun = evt.run();
    if(fSubRun < 0) fSubRun = evt.subRun();
    
    fCosmicTrackTree.run      = evt.run();	
    fCosmicTrackTree.subrun   = evt.subRun();	
    fCosmicTrackTree.event    = evt.id().event(); 
    fCosmicTrackTree.nevents = fNevents;
    fTotNEvents->Fill(1);    
    
    art::ServiceHandle<calib::Calibrator> cal;
    
    // now get the tracks that were made by CosmicTrack
    // and test that there are tracks in the event
    art::Handle< std::vector<rb::Track> > trks;
    evt.getByLabel(fTrackLabel,trks);
    if (trks->size() < 1) {
      mf::LogWarning("CTAWarn") << "There are " << trks->size() 
 	                        << " Tracks in this event, skip";
      return;
    }
    LOG_DEBUG("CosmicTrackAna") << "There are " << trks->size() << " Tracks in this event";

    const unsigned int trkMax = trks->size();

    this->FillEventMCTruthHistograms(trkMax);

    // get a FindMany to grab the slice associated with each track
    art::FindOne<rb::Cluster> foc(trks, evt, fTrackLabel);
        
    unsigned int numtracks   = 0;

    fSignalHitFraction.clear();
    fPurityEfficiency .clear();
    fTrackStartPos    .clear();
    fTrackEndPos      .clear();

    // Loop over tracks
    for(unsigned int trkIdx = 0; trkIdx < trkMax; ++trkIdx){

      // find the slice associated to the track
      cet::maybe_ref<rb::Cluster const> sliceRef(foc.at(trkIdx));
      if( !sliceRef.isValid() ) continue;
      rb::Cluster const& slice(sliceRef.ref());
                  
      rb::Track const& track = trks->at(trkIdx);
      if(!track.Is3D()) continue;

//       std::cout << "Track starts at " 
// 		<< track.Start().X() << " " 
// 		<< track.Start().Y() << " " 
// 		<< track.Start().Z() << std::endl;


      //only fill the histograms if the track crosses more than fTrkLengthCut cm in z 
      if(std::abs(track.Stop().Z() - track.Start().Z()) > fTrkLengthCut){

	this->FillTrackMCTruthHistograms(track, slice);
	
	//count number of tracks that cross fTrkLengthCut
	++numtracks;
	
	this->FillTrackHistograms(track, slice);

      }// end if passes track length in z cut

    }//end loop over tracks
    
    fTracksFound->Fill(numtracks);
    
    fNTracks +=numtracks;
    
    fCosmicTrackTree.ntracks = numtracks;

    if(fPurityEfficiency.size() != fSignalHitFraction.size() ||
       fPurityEfficiency.size() != fTrackStartPos.size()     ||
       fPurityEfficiency.size() != fTrackEndPos.size()){
      LOG_WARNING("CosmicTrackAna") << "vectors for filling tree are not of the same size: "
				    << fPurityEfficiency.size() << " " 
				    << fSignalHitFraction.size() << " " 
				    << fTrackStartPos.size() << " " 
				    << fTrackEndPos.size();
    }
    else{
      // fill the tree in a flat way, one line for every track
      for(size_t i = 0; i < fPurityEfficiency.size(); ++i){
	fPurEff     = fPurityEfficiency[i];
	fSigHitFrac = fSignalHitFraction[i];
	fTrkStPos   = fTrackStartPos[i];
	fTrkEndPos  = fTrackEndPos[i];

	fTree->Fill();
      }
    }

    return;
  }


  //......................................................................  
  void CosmicTrackAna::FillEventMCTruthHistograms(size_t trkMax)
  {

    if(fBT->HaveTruthInfo()){ 

      unsigned int numtracksMC = 0;

      std::vector<sim::Particle> detParticles;
     
      const sim::ParticleNavigator& pnav = fBT->ParticleNavigator();
    
      // count how many tracks were left by primary particles in the event
      for(int p = 0; p < pnav.NumberOfPrimaries(); ++p){
        if( fBT->ParticleToFLSHit(pnav.Primary(p)->TrackId()).size() > 0){
	  if(std::abs(pnav.Primary(p)->PdgCode()) == 13){
	    ++fNTracksMC;
	    ++numtracksMC;
	    detParticles.push_back(*(pnav.Primary(p)));
	  }
        }
      }
      
      // no point filling histograms for MC if particles don't leave hits
      if (detParticles.size() < 1 ) {
	mf::LogWarning("CTAWarn") << "Failed to find any Mother particles leaving"
				  << " hits in the detector, bail";
        return;
      }
      else {
	LOG_DEBUG("CosmicTrackAna") << "There are " << detParticles.size() 
			            << " Mother particles leaving hits in the detector";

	for(auto detPart : detParticles){
	  fPrimaryTrueEnergy->Fill(detPart.E());
	}// end loop over particles (muons)

	double ratio = 0.;
	fTracksFoundVsLeft->Fill(numtracksMC*1., trkMax*1.); // fill only for MC

	if(trkMax > numtracksMC) 
	  LOG_VERBATIM("CosmicTrackAna") << "reconstructed more tracks than generated: "
					 << trkMax << " vs " << numtracksMC;

	ratio = (trkMax*1.)/(numtracksMC*1.);
	fTrackFindEff->Fill(numtracksMC*1.,ratio);
	fRecoEff->Fill(ratio);
	
      }// end if particles were found

    }// end if bt has MCTruth information

    return;
  }

  //......................................................................  
  void CosmicTrackAna::FillTrackMCTruthHistograms(rb::Track   const& track,
						  rb::Cluster const& slice)
  {

    if(fBT->HaveTruthInfo()){ 

      TVector3 dir(track.Dir());
      double dcosReco[3] = {dir.X()/dir.Mag(),
			    dir.Y()/dir.Mag(),
			    dir.Z()/dir.Mag()};
      
      // get the truth information for this track
      std::vector<cheat::TrackIDE> ides = fBT->HitsToTrackIDE(track.AllCells());
      if(ides.size() > 0){
	
// 	if(ides.size() > 1){
// 	  LOG_VERBATIM("CosmicTrackAna") << "found more than one trackID corresponding to the hits in the track:";
// 	  for(auto tid : ides) LOG_VERBATIM("CosmicTrackAna") << "\t" << tid.trackID << " " << tid.energyFrac 
// 							      << " " << fBT->TrackIDToParticle(ides.front().trackID)->PdgCode();
// 	}
	
	// get the particle for the first IDE as it represents the track ID with the most 
	// energy in the collection of hits
	auto part = fBT->TrackIDToParticle(ides.front().trackID);

	double dcosTrue[3] = {part->Px()/part->P(),
			      part->Py()/part->P(),
			      part->Pz()/part->P()};
	
	fDCosXRecoVsTrue->Fill(dcosTrue[0], dcosReco[0]);
	fDCosYRecoVsTrue->Fill(dcosTrue[1], dcosReco[1]);
	fDCosZRecoVsTrue->Fill(dcosTrue[2], dcosReco[2]);
	
	fDCosXDiff->Fill((dcosReco[0] - dcosTrue[0])/dcosTrue[0]);
	fDCosYDiff->Fill((dcosReco[1] - dcosTrue[1])/dcosTrue[1]);
	fDCosZDiff->Fill((dcosReco[2] - dcosTrue[2])/dcosTrue[2]);
	
	fDCosXDiffVsZLen->Fill(track.ExtentZ(), (dcosReco[0] - dcosTrue[0])/dcosTrue[0]);
	fDCosYDiffVsZLen->Fill(track.ExtentZ(), (dcosReco[1] - dcosTrue[1])/dcosTrue[1]);
	fDCosZDiffVsZLen->Fill(track.ExtentZ(), (dcosReco[2] - dcosTrue[2])/dcosTrue[2]);
	
	// check the reconstructed end point vs the true end point
	TVector3 simEnd(part->EndX(), part->EndY(), part->EndZ());
	TVector3 simBeg(part->Vx(),   part->Vy(),   part->Vz());
	TVector3 diff = track.Stop() - simEnd;

// 	if(track.Dir().Z() > 0)
// 	LOG_VERBATIM("CosmicTrackAna") << "simulation z end point: " << simEnd.Z()
// 				       << "\ntrack z end point: " << track.Stop().Z();

	fDeltaZEnd->Fill(diff.Z());
	fDeltaEnd->Fill(diff.Mag());
	fDeltaZEndVsDCosZ->Fill(diff.Z(), track.Dir().Z());
	double xyz[3] = { track.Stop().X(), track.Stop().Y(), track.Stop().Z() };
	
	double distEdge = geo::DistToEdge(xyz, 
					  fGeom->DetHalfWidth(),
					  fGeom->DetHalfHeight(),
					  fGeom->DetLength()); 
	fDeltaZEndVsDistToEdge->Fill(diff.Z(), distEdge);

	diff = track.Start() - simBeg;
	fDeltaZBegVsDCosZ->Fill(diff.Z(), track.Dir().Z());

	// check the efficiency and purity of the track hits for
	// this MC particle
	std::set<int> trackID;
	trackID.insert(ides.front().trackID);

	CosmicPurityEfficiency pe;
	pe.purity     = fBT->HitCollectionPurity(trackID, track.AllCells());
	pe.efficiency = fBT->HitCollectionEfficiency(trackID, track.AllCells(), slice.AllCells(), geo::kXorY);

	fTrackPurity    ->Fill(pe.purity);
	fTrackEfficiency->Fill(pe.efficiency);
	
	fPurityEfficiency.push_back(pe);

      } // end if at least one particle matches for the track
            
    } // end if(fBT->HaveTruthInfo())
    
    return;
  }

  //......................................................................  
  void CosmicTrackAna::FillTrackHistograms(rb::Track   const& track,
					   rb::Cluster const& slice)
  {
    float cosCosmic = 0.;
    float trkcells  = 0.;
    float trksig    = 0.;
    float azimuth   = 0.;
    float totcell   = 0.;
    float totsig    = 0.;

    // loop over the cell hits from the cluster
    for(size_t i = 0; i < slice.NCell(); ++i){
      art::Ptr<rb::CellHit> chit = slice.Cell(i);
      if(chit->PE() > fMinSig){
	fCellHitSig->Fill(chit->PE()); 
	totcell += 1.;
	totsig  += chit->PE();
      }
    }

    // Find the track residuals.  get vectors for the end points of the track
    TVector3 start(track.Start());
    TVector3 end(track.Stop());
    TVector3 dir(track.Dir());
    double length = track.TotalLength();    

    // Angular Distribution with respect to zenith
    cosCosmic = -dir.Y()/dir.Mag();
    azimuth = std::atan2(dir.x()/dir.Mag(), -dir.z()/dir.Mag()) * 180./TMath::Pi(); 
    if(azimuth < 0.) azimuth += 360.;
    fAzimuth->Fill(azimuth);
    fCosCosmic->Fill(cosCosmic);

    fTrackStartPos.push_back(start);
    fTrackEndPos  .push_back(end);
      
    fStartX->Fill(start.X()); 
    fStartY->Fill(start.Y());
    fStartZ->Fill(start.Z());
    fEndX->Fill(end.X());
    fEndY->Fill(end.Y());
    fEndZ->Fill(end.Z());
    fStartXY->Fill(start.X(), start.Y());
    fStartZX->Fill(start.Z(), start.X());
    fStartZY->Fill(start.Z(), start.Y());	
    fEndXY->Fill(end.X(), end.Y());
    fEndZX->Fill(end.Z(), end.X());
    fEndZY->Fill(end.Z(), end.Y());
    fTrackLength->Fill(length);
      
    // make a map of the z position of each trajectory point of the
    // track to its corresponding trajectory point.  the map will
    // be used to find the residual of the reconstruction to the 
    // cell hits.
    TVector3 dirCos;
    TVector3 diff01;
    TVector3 diff02;
    TVector3 diff12;
    TVector3 cross;
    TVector3 point;
    TVector3 lowerBound;
    TVector3 upperBound;
    double resid = 0.;
    double xyz[3] = {0.};

    //std::map<double, TVector3> zToTrajectory;
    zBoundMap zToPlaneBounds;
    std::pair<TVector3, TVector3> planeBounds;
    std::vector<TVector3> const& traj = track.Trajectory();
    //std::cout << "Trajectory points for track: " << std::endl;
    for(size_t tp = 0; tp < traj.size(); ++tp){
      
      // find the intersection of the trajectory with the boundary
      // of the plane it represents
      if(tp < traj.size() - 1) dirCos = traj[tp+1] - traj[tp];
      else                     dirCos = traj[tp]   - traj[tp-1];

      planeBounds.first  = TVector3(traj[tp].X() - fCellHalfDepthZ*dirCos.X()/dirCos.Z(),
				    traj[tp].Y() - fCellHalfDepthZ*dirCos.Y()/dirCos.Z(),
				    traj[tp].Z() - fCellHalfDepthZ);
      
      planeBounds.second = TVector3(traj[tp].X() + fCellHalfDepthZ*dirCos.X()/dirCos.Z(),
				    traj[tp].Y() + fCellHalfDepthZ*dirCos.Y()/dirCos.Z(),
				    traj[tp].Z() + fCellHalfDepthZ);

      zToPlaneBounds[traj[tp].Z()] = std::pair< std::pair<TVector3, TVector3>, TVector3>(planeBounds, dirCos);

      //dirCos.Print();
      //planeBounds.first.Print();
      //traj[tp].Print();
      //planeBounds.second.Print();
    }
//     std::cout << " Direction Cosines " << std::endl;
//     zToPlaneBounds[traj[0].Z()].second.Print();

    // Loop through both views and grab every CellHit in this track
    for(size_t c = 0; c < track.NCell(); ++c){	
      const art::Ptr<rb::CellHit> chit = track.Cell(c);
      const geo::View_t view = chit->View();
      
      fCellHit[view]->Fill(chit->Cell());
      fPlaneHit[view]->Fill(chit->Plane());
      fCellPlane[view]->Fill(chit->Plane(), chit->Cell());
	
      const rb::RecoHit rhit = track.RecoHit(chit);
      if(!rhit.IsCalibrated()){
	//mf::LogError("CosmicTrackAna") << "Not calibrated?!";
	continue;
      }
      
      fGeom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);

      // find the trajectory points of the track that bound this
      // hit in z and determine where the hit is expected to be
      // based on the direction of the track between the trajectory 
      // points
      auto lbitr = this->Bounds(zToPlaneBounds, xyz[2]);

      // have to check if indeed our z for this cell is 
      // between the bounds or if it is upstream/downstream of the
      // extrema for the trajectory    
      lowerBound = lbitr->second.first.first;
      upperBound = lbitr->second.first.second;
      dirCos     = lbitr->second.second;

      point = TVector3(rhit.X(), rhit.Y(), rhit.Z());
	
      if(xyz[2] < lowerBound.Z())
	lowerBound  = TVector3(lowerBound.X() + (lowerBound.Z() - xyz[2])*dirCos.X()/dirCos.Z(),
			       lowerBound.Y() + (lowerBound.Z() - xyz[2])*dirCos.Y()/dirCos.Z(),
			       xyz[2]);

      if(xyz[2] > upperBound.Z())
	upperBound  = TVector3(upperBound.X() + (xyz[2] - upperBound.Z())*dirCos.X()/dirCos.Z(),
			       upperBound.Y() + (xyz[2] - upperBound.Z())*dirCos.Y()/dirCos.Z(),
			       xyz[2]);

      diff01 = point - lowerBound;
      diff02 = point - upperBound;
      diff12 = upperBound - lowerBound;
      
      cross  = diff01.Cross(diff02);
      resid  = cross.Mag()/diff12.Mag();

//       if(resid > 100.) 
//       std::cout << "view " << view 
// 		<< " lower bound: (" << lowerBound.X() << "," << lowerBound.Y() << "," << lowerBound.Z() << ") "
// 		<< " cell center: (" << rhit.X() << "," << rhit.Y() << "," << rhit.Z() << ") " 
// 		<< " upper bound: (" << upperBound.X() << "," << upperBound.Y() << "," << upperBound.Z() << ") "
// 		<< " residual: " << resid << std::endl;      
//       planeBounds.first.Print();
//       point.Print();
//       planeBounds.second.Print();
//       cross.Print();
//       diff12.Print();
      
      fResidVsCos[view]->Fill(resid, cosCosmic);
      fResidVsAz[view]->Fill(resid, azimuth);
      fResiduals[view]->Fill(resid);
      fHitXPosition->Fill(rhit.X());
      fHitYPosition->Fill(rhit.Y());
      fHitZPosition->Fill(rhit.Z());
      fHitXYPosition->Fill(rhit.X(), rhit.Y());
      fHitXZPosition->Fill(rhit.Z(), rhit.X());
      fHitYZPosition->Fill(rhit.Z(), rhit.Y());
	
      fTNS[view]->Fill(chit->TNS());
	
    } // end for cells
      
    trksig   = track.TotalPE();  
    trkcells = track.NCell();
      
    double dcosReco[3] = {dir.X()/dir.Mag(),
			  dir.Y()/dir.Mag(),
			  dir.Z()/dir.Mag()};
      
    fDCosX->Fill(dcosReco[0]);
    fDCosY->Fill(dcosReco[1]);
    fDCosZ->Fill(dcosReco[2]);
      
    fStartX->Fill(start.X());
    fStartY->Fill(start.Y());
    fStartZ->Fill(start.Z());
      
    fDirX->Fill(dir.X());
    fDirY->Fill(dir.Y());
    fDirZ->Fill(dir.Z());
    fDirXY->Fill(dir.X(),dir.Y());
      
    fEndX->Fill(end.X());
    fEndY->Fill(end.Y());
    fEndZ->Fill(end.Z());
      
    fDiffX->Fill(track.ExtentX());
    fDiffY->Fill(track.ExtentY());
    fDiffZ->Fill(track.ExtentZ());
      
//     if(cosCosmic < 0.1){
//       LOG_VERBATIM("CosmicTrackAna") << "track params " << track.TotalLength() << " " << cosCosmic << " " << azimuth
// 				     << "\ndirection " << dir.X() << " " << dir.Y() << " " << dir.Z()
// 				     << "\nminPlane: " << track.MinPlane() << " maxPlane: " << track.MaxPlane();
//       for(size_t p = 0; p < track.NTrajectoryPoints(); ++p){
// 	TVector3 point = track.TrajectoryPoint(p);
// 	LOG_VERBATIM("CosmicTrackAna") << point.X() << " " << point.Y() << " " << point.Z();
//       }
//     }
      
    double startpt[3] = {start.X(), start.Y(), start.Z()};
    double endpt[3]   = {  end.X(),   end.Y(),   end.Z()};
      
    fVtxDistToEdge->Fill(geo::DistToEdge(startpt, 
					 fGeom->DetHalfWidth(),
					 fGeom->DetHalfHeight(), 
					 fGeom->DetLength()));
    fEndDistToEdge->Fill(geo::DistToEdge(endpt, 
					 fGeom->DetHalfWidth(),
					 fGeom->DetHalfHeight(), 
					 fGeom->DetLength()));
    fTrkSig->Fill(trksig);
    fTrkCells->Fill(trkcells);

    CosmicSignalHitFraction shf;
    shf.signalFraction = trksig/totsig;
    shf.hitFraction    = trkcells/totcell;
    shf.signal         = trksig;
    shf.hits           = trkcells;
    fSignalHitFraction.push_back(shf);
    
    fFracHitsUsedVsNum->Fill(totcell, shf.hitFraction);
    fFractionHitsUsed->Fill(shf.hitFraction);
    fFractionSigUsed->Fill(shf.signalFraction);  
      
    return;
  }

  //......................................................................  
  void CosmicTrackAna::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMinSig       = pset.get< double >("MinSig"      );
    fTrkLengthCut = pset.get< double >("TrkLengthCut");
  }
  
  //......................................................................  
  void CosmicTrackAna::endJob()
  { 
    if(fNevents != 0 && fLiveTime != 0){
      float cosmicrate = fNTracks/fLiveTime;

      LOG_VERBATIM("CosmicTrackAna") << "There are: " << fNTracks 
				     << " cosmic tracks selected in this job.\n"
				     << "Livetime is " << fLiveTime 
				     << " according to CosmicExposure.\n"
				     << "Note, livetime is for FULL SUBRUN, "
				     << "so rate is wrong if haven't run full subrun.\n"
				     << "The cosmic rate is therefore " 
				     << cosmicrate << " Hz \n";
      
      fCosmicRate->Fill(cosmicrate);
    }
  } 

  //----------------------------------------------------------------------------
  zBoundMap::const_iterator CosmicTrackAna::Bounds(zBoundMap const& bounds,
						   double    const& hitZ)
  {
    zBoundMap::const_iterator lbitr = bounds.lower_bound(hitZ);

    // check that the returned bounds actually bound the point, if not
    // look before and after the returned bounds in the map
    while( !this->CheckZInBounds(lbitr->second.first, hitZ) ){
      if(lbitr != bounds.begin()) --lbitr;
      else break;
    }
    
    return lbitr;
  }

  //----------------------------------------------------------------------------
  bool CosmicTrackAna::CheckZInBounds(std::pair<TVector3, TVector3> const& bounds,
				      double                        const& hitZ)
  {
    if(hitZ >= bounds.first.Z() && 
       hitZ <= bounds.second.Z()) return true;

    return false;
  }

  
} // end namespace trk

namespace trk{
  DEFINE_ART_MODULE(CosmicTrackAna);
}

