////////////////////////////////////////////////////////////////////////
/// \brief   A module for propagating particles through Geant4 and
///          generating energy depositions
/// \author  messier@indiana.edu brebel@fnal.gov rhatcher@fnal.gov
////////////////////////////////////////////////////////////////////////


// NOvA includes 
#include "g4nova/G4Alg.h"
// #include "g4nova/debug_tools/FLSHitListDebug.h" //Debug stuff
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "Simulation/Simulation.h"
#include "SimulationBase/MCTruth.h"
#include "Utilities/AssociationUtil.h"

#include <cassert>
#include <cstdlib>
#include <string>
#include <sstream>
#include <sys/stat.h>
#include <vector>
#include <map>
#include <unistd.h>

// ROOT include
#include <TDatabasePDG.h>
#include <TSystem.h>
#include <TStopwatch.h>

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/search_path.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"

#include "art/Framework/Core/ModuleMacros.h"

namespace art { 
  class Event; 
  class ParameterSet;
}

namespace g4n {

  /// A module to pass interaction information to Geant4 and 
  /// create hits and particle lists
  class G4Gen : public art::EDProducer {
  public:
    explicit G4Gen(fhicl::ParameterSet const &pset);
    virtual ~G4Gen();                        

    void produce (art::Event& evt);
    void beginRun(art::Run& run);

  private:

    g4n::G4Alg*         fG4Alg;     ///< G4Helper object
    fhicl::ParameterSet fG4AlgPSet; ///< parameter set to configure the G4Alg object
    TStopwatch          fStopwatch; ///< keep track of how long it takes to run the job

    //bool                fDoFLSHitListDebug; ///< Are we going to debug the FLSHitList?
    
  };
}

namespace g4n {

  //___________________________________________________________________________
  // Constructor
  G4Gen::G4Gen(fhicl::ParameterSet const& pset)
    : fG4Alg(0)
    , fG4AlgPSet(pset.get< fhicl::ParameterSet >("G4AlgPSet") )
	      //, fDoFLSHitListDebug(false)
  {

    // get the random number seed, use a random default if not specified    
    // in the configuration file. 
    unsigned int seed = pset.get< unsigned int >("Seed", sim::GetRandomNumberSeed());
    // setup the random number service for Geant4, the "G4Engine" label is a 
    // special tag setting up a global engine for use by Geant4/CLHEP
    createEngine(seed, "G4Engine");

    // Start counting the time
    fStopwatch.Start();

    produces< std::vector<sim::FLSHitList >            >();
    produces< std::vector<sim::Particle>     	       >();
    produces< art::Assns<sim::Particle, simb::MCTruth> >();

    /// Setup the FLSHitList debugger
    //fDoFLSHitListDebug = doG4Debug<FLSHitListDebug>();
    // if(fDoFLSHitListDebug){
//         mf::LogVerbatim("G4Gen") << "Will be Debugging FLSHitLists";
//     }
//     else{
//         mf::LogVerbatim("G4Gen") << "Will NOT be Debugging FLSHitLists";
//     }


  }// end of constructor

  //___________________________________________________________________________
  // Destructor
  G4Gen::~G4Gen()  
  {  
    fStopwatch.Stop();
    mf::LogVerbatim("G4Gen") << "real time to run through Geant4: " 
                             << fStopwatch.RealTime();
    if(fG4Alg) delete fG4Alg;
  }// end of destructor

  //___________________________________________________________________________
  void G4Gen::beginRun(art::Run& run) 
  {
    fG4Alg = new G4Alg(fG4AlgPSet);
    
    return;
  }// end of begin job

  //___________________________________________________________________________
  void G4Gen::produce(art::Event& evt)  
  {
    // need to instantiate the G4Helper in order to make the 
    // G4RunManager and pass it the physics list - G4 really dislikes
    // it if you try to make a G4UserRunAction (ie UserActionManager)
    // before passing a physics list to the run manager
  
    // Define the FLSHitList and Particle vectors.
    std::unique_ptr<std::vector<sim::FLSHitList>   >            flshlcol(new std::vector<sim::FLSHitList>  );
    std::unique_ptr<std::vector<sim::Particle>     >            pcol    (new std::vector<sim::Particle>    );
    std::unique_ptr< art::Assns<sim::Particle, simb::MCTruth> > tpassn  (new art::Assns<sim::Particle, simb::MCTruth>);

    //look to see if there is any MCTruth information for this
    //event
    std::vector< art::Handle< std::vector<simb::MCTruth> > > mclists;
    evt.getManyByType(mclists);

    std::vector< std::vector< std::pair<size_t, size_t> > > pListLimits;

    fG4Alg->RunGeant(mclists, *flshlcol, *pcol, pListLimits);

    /// If we need to debug FLSHitLists, do it now
//     if(fDoFLSHitListDebug){
//         art::ServiceHandle<FLSHitListDebug> fls_hitlist_debug;
//         fls_hitlist_debug->DEBUG_FLSHITLISTS(flshlcol);
//     }

    // put the particles into the product to be stored in the event
    for(size_t mcl = 0; mcl < mclists.size(); ++mcl){
      art::Handle< std::vector<simb::MCTruth> > mclistHandle = mclists[mcl];

      for(size_t i = 0; i < mclistHandle->size(); ++i){
	art::Ptr<simb::MCTruth> mct(mclistHandle, i);
	
	for( size_t p = pListLimits[mcl].at(i).first; p < pListLimits[mcl].at(i).second; ++p) 
	  util::CreateAssn(*this, evt, *pcol, mct, *tpassn, p);
      }
    }// end loop over handles

    // Put the data products into the event
    evt.put(std::move(flshlcol));
    evt.put(std::move(pcol));
    evt.put(std::move(tpassn));

    return;
  }// end of produce
 
} // end namespace

namespace g4n { DEFINE_ART_MODULE(G4Gen); }
